# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
Helpers for testing the project.
"""
import os
import random
import shutil
import string
import tempfile
import unittest


def _choices(population, k=1):
    """
    Select k individuals from population, with replacement.

    Replaces random.choices from Python 3.
    """
    return (random.choice(population) for _ in range(k))


def _maybe_segments_to_path(segments_or_path):
    """
    Convert segments or path to string path.
    """
    if isinstance(segments_or_path, list):
        return os.sep.join(segments_or_path)
    return segments_or_path


class BrinkTestCase(unittest.TestCase):
    """
    TestCase dedicated to brink project.
    """

    @staticmethod
    def string(length=8):
        """
        Generate a random string.

        FIXME:5648:
        Add a non-ASCII character, and make Unicode work everywhere.
        """
        return 'string123_' + ''.join(_choices(string.ascii_letters, k=length))

    def createFileInTemp(
            self, content=b'', prefix='', suffix=''):
        """
        Returns path of a new temporary file in a default temporary folder.
        Ensure cleanup after the test.
        """
        prefix = 'brink-' + self._testMethodName + '-' + prefix
        fd, abspath = tempfile.mkstemp(prefix=prefix, suffix=suffix)
        self.addCleanup(os.remove, abspath)

        if content:
            with os.fdopen(fd, 'wb') as f:
                f.write(content)

        return abspath

    def createFolderInTemp(self, prefix):
        """
        Returns path to a new temporary folder with given prefix.
        Ensure its deletion after the test.
        """
        prefix = 'brink-' + self._testMethodName + '-' + prefix
        abspath = tempfile.mkdtemp(prefix=prefix)
        self.addCleanup(shutil.rmtree, abspath)

        return abspath

    def createFile(self, path=None, directory=None, name=None, content=b''):
        """
        Create a file with a specific path or directory and name.
        Ensure its deletion after the test.
        """

        if path is not None and (directory or name):
            raise ValueError(
                "Can only specify either path or directory and name.")
        if path is None:
            path = os.path.join(directory, name)

        with open(path, 'wb') as f:
            f.write(content)

        self.addCleanup(os.remove, path)

        return path

    @staticmethod
    def exists(segments_or_path):
        """
        Checks whether a file or directory exists.

        Can be specified as segments or string path.
        """
        path = _maybe_segments_to_path(segments_or_path)
        return os.path.exists(path)

    @staticmethod
    def getFileContent(segments_or_path, utf8=True):
        """
        Return the content of file (specified as segments or string path).

        By default, the content is returned as Unicode.
        """
        with open(_maybe_segments_to_path(segments_or_path), 'rb') as f:
            result = f.read()

            if utf8:
                return result.decode('utf-8')
            return result
