# coding=utf-8
# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
# We don't want to depend on six/future so we have our own unicode handling.
# pylint: disable=unicode-builtin
"""
Unit tests for brink filesystem.
"""
import os

from unittest.mock import Mock

from brink.filesystem import BrinkFilesystem
from brink.testing import BrinkTestCase


class TestBrinkFilesystem(BrinkTestCase):
    """
    Unit tests for `BrinkFilesystem`.
    """

    def setUp(self):
        super(TestBrinkFilesystem, self).setUp()

        self.sut = BrinkFilesystem()
        self.sut._isValidSystemPath = Mock(return_value=True)

    def test_which_extra_paths_not_defined(self):
        """
        Returns `None` if the command could not be found in system path or
        in the `extra_paths` argument.
        """
        command = self.string()
        extra_paths = [self.string()]
        self.sut.listFolder = Mock(return_value=[])

        result = self.sut.which(command, extra_paths=extra_paths)

        self.assertIsNone(result)

    def test_which_extra_paths_file_found(self):
        """
        Returns the full path to the specified command if found and the
        executable file exists.
        """
        folder = self.string()
        command = self.string()
        full_path_to_command = os.path.join(folder, command)
        extra_paths = [self.string(), folder]

        def _folderListing(path):
            result = []
            if folder == path:
                result.append(command)
            return result

        self.sut.listFolder = _folderListing

        result = self.sut.which(command, extra_paths=extra_paths)

        self.assertEqual(full_path_to_command, result)

    def test_which_extra_paths_file_not_exists(self):
        """
        Returns `None` if the specified command is defined in `extra_paths`
        but the actual, executable, file does not exist.
        """
        command = self.string()
        path_exe_file = '%s.exe' % command
        self.sut.listFolder = Mock(return_value=[])
        extra_paths = [self.string(), path_exe_file]

        result = self.sut.which(command, extra_paths=extra_paths)

        self.assertIsNone(result)

    def test_parseUnixPaths(self):
        """
        It uses : to separate path entries.
        """
        path1 = self.string()
        path2 = self.string()
        paths = '{}:{}'.format(path1, path2)

        result = self.sut._parseUnixPaths(paths)

        self.assertEqual(2, len(result))
        self.assertIn(path1, result)
        self.assertIn(path2, result)

    def test_parseWindowsPaths(self):
        """
        It uses ; to separate path entries.
        """
        path1 = self.string()
        path2 = self.string()
        paths = '{};{}'.format(path1, path2)

        result = self.sut._parseWindowsPaths(paths)

        self.assertEqual(2, len(result))
        self.assertIn(path1, result)
        self.assertIn(path2, result)

    def test_findCommand_ok_posix(self):
        """
        On Unix systems it does not validate Windows type executable files
        as valid commands.
        """
        if os.name == 'nt':
            raise self.skipTest("Unix specific test.")

        command = self.string()
        path = self.string()
        full_path = '{}/{}'.format(path, command)
        exe_command = '%s.exe' % command
        bat_command = '%s.bat' % command
        cmd_command = '%s.cmd' % command
        files = [
            exe_command,
            bat_command,
            cmd_command,
            ]
        self.sut.listFolder = Mock(return_value=files)

        result = self.sut._findCommand(command, path)

        self.assertIsNone(result)

        files.append(command)

        result = self.sut._findCommand(command, path)

        self.assertEqual(full_path, result)

    def test_findCommand_ok_windows(self):
        """
        On Windows systems it validates files with .exe extension before
        checking files without extension.
        """
        if os.name != 'nt':
            raise self.skipTest("Windows specific test.")

        command = self.string()
        path = self.string()
        exe_command = '%s.exe' % command
        full_path = r'{}\{}'.format(path, exe_command)
        files = [
            command,
            exe_command,
            ]
        self.sut.listFolder = Mock(return_value=files)

        result = self.sut._findCommand(command, path)

        self.assertEqual(full_path, result)

    def test_findCommand_ok_windows_with_extension(self):
        """
        On Windows systems it validates files with .exe extension, even
        when the command is specified with explicit .exe extension.
        """
        if os.name != 'nt':
            raise self.skipTest("Windows specific test.")

        command = self.string()
        path = self.string()
        exe_command = '%s.exe' % command
        full_path = r'{}\{}'.format(path, exe_command)
        files = [
            command,
            exe_command,
            ]
        self.sut.listFolder = Mock(return_value=files)

        result = self.sut._findCommand(exe_command, path)

        self.assertEqual(full_path, result)

    def test_findCommand_checks_folders_only(self):
        """
        Returns `None` if the path argument is not a folder.
        """
        command = self.string()
        path = self.string()
        self.sut._isValidSystemPath = Mock(return_value=False)

        result = self.sut._findCommand(command, path)

        self.assertIsNone(result)

    def test_findCommand_no_result(self):
        """
        Returns `None` if the command could not be found in the specified
        path.
        """
        command = self.string()
        path = self.string()
        files = [self.string(), self.string()]
        self.sut.listFolder = Mock(return_value=files)

        result = self.sut._findCommand(command, path)

        self.assertIsNone(result)

    def test_getSearchPaths_default(self):
        """
        It returns a list of paths from operating system environment
        variable.
        """
        if os.name == 'posix':
            paths = self.sut._parseUnixPaths(os.environ['PATH'])
        else:
            paths = self.sut._parseWindowsPaths(os.environ['PATH'])

        result = self.sut._getSearchPaths()

        self.assertEqual(paths, result)

    def test_getSearchPaths_with_extra(self):
        """
        If a list of extra paths if provided, it return a list with paths
        defined in the operating system, but with the list of extra paths
        prepended to the result.
        """
        path1 = self.string
        path2 = self.string
        if os.name == 'posix':
            path3 = self.sut._parseUnixPaths(os.environ['PATH'])[0]
        else:
            path3 = self.sut._parseWindowsPaths(os.environ['PATH'])[0]

        result = self.sut._getSearchPaths(extra_paths=[path1, path2])

        self.assertEqual(path1, result[0])
        self.assertEqual(path2, result[1])
        self.assertEqual(path3, result[2])

    def test_copyFolder_no_overwrite(self):
        """
        It can copy the folder without overwriting existing files.
        """
        source_directory_path = self.createFolderInTemp(prefix='src-')
        destination_directory_path = self.createFolderInTemp(prefix='dst-')

        existing_name = self.string()
        self.createFile(
            directory=source_directory_path,
            name=existing_name,
            content=b'source-exist')
        existing_destination_path = self.createFile(
            directory=destination_directory_path,
            name=existing_name,
            content=b'destination-exist')
        self.createFile(
            directory=source_directory_path,
            name='other-file',
            content=b'other')

        source_directory_segments = source_directory_path.split(os.sep)
        destination_directory_segments = destination_directory_path.split(
            os.sep)

        self.sut.copyFolder(
            source=['/'] + source_directory_segments,
            destination=['/'] + destination_directory_segments,
            overwrite=False,
            )

        self.assertTrue(
            self.exists(destination_directory_segments + ['other-file']))
        self.assertTrue(
            self.exists(existing_destination_path))
        self.assertEqual(
            b'destination-exist',
            self.getFileContent(existing_destination_path, utf8=False))

    def test_join_rejoin_unicode(self):
        """
        It can join and re-join the result from unicode path.
        """
        first_join = self.sut.join([self.string(), self.string()])
        second_join = self.sut.join([first_join, self.string()])

        self.assertIsInstance(second_join, str)
