# Copyright (c) 2013 Adi Roiban.
# See LICENSE for details.
"""
Tests for brink utilities.
"""
import hashlib
import os

from brink import __version__
from brink.git_command import parse_diff_list
from brink.testing import BrinkTestCase

from brink.utils import BrinkPaver


MINIMAL_SETUP = {
    'folders': {
        'source': 'brink',
        'dist': 'dist',
        'publish': 'publish',
        }
    }


class TestBrinkPaver(BrinkTestCase):
    """
    Tests for BrinkPaver.
    """

    def test_initialization(self):
        """
        It is initialized with a setup dictionary.
        """
        result = BrinkPaver(setup=MINIMAL_SETUP)

        self.assertEqual(MINIMAL_SETUP, result.setup)

    def test_convertToDOSNewlines(self):
        """
        Convert unix newlines to dos newlines.
        """
        sut = BrinkPaver(setup=MINIMAL_SETUP)
        content = (
            b'one line\t \n'
            b' \tspaces\n'
            )
        path = self.createFileInTemp(content=content)

        sut._convertToDOSNewlines(path)

        result = self.getFileContent(path)

        self.assertEqual((
            'one line\t \r\n'
            ' \tspaces\r\n'
            ),
            result)

    def test_createMD5Sum(self):
        """
        Return the MD5 of file at path, which is specified as segments
        """
        sut = BrinkPaver(setup=MINIMAL_SETUP)
        content = (
            b'one line\t \n'
            b' \tspaces\n'
            )
        path = self.createFileInTemp(content=content)

        result = sut.createMD5Sum([path])

        expected = hashlib.md5(content).hexdigest()
        self.assertEqual(expected, result)

    def test_rsync_unix(self):
        """
        On Unix it used the default SSH.
        """
        if os.name != 'posix':
            raise self.skipTest("Unix specific test.")

        sut = BrinkPaver(setup=MINIMAL_SETUP)
        command = []
        sut.execute = (
            lambda **kwargs: not command.append(kwargs) and (0, ''))

        sut.rsync(
            username='some-user',
            hostname='some-host',
            source=['source', 'folder'],
            destination='path/on/server',
            )

        self.assertEqual([{
            'command': [
                'rsync',
                '-cza',
                '-e',
                'ssh',
                'source/folder',
                'some-user@some-host:path/on/server',
                ],
            }], command)

    def test_rsync_windows(self):
        """
        On Windows it used the dedicated SSH with a dedicated config file.
        """
        if os.name != 'nt':
            raise self.skipTest("Windows specific test.")

        sut = BrinkPaver(setup=MINIMAL_SETUP)
        command = []
        sut.execute = (
            lambda **kwargs: not command.append(kwargs) and (0, ''))

        sut.rsync(
            username='some-user',
            hostname='some-host',
            source=['source', 'folder'],
            destination='path/on/server',
            )

        self.assertEqual([{
            'command': [
                'rsync',
                '-rcz',
                '-no-p',
                '--chmod=D755,F644',
                '--chown=some-user:www-data',
                '-e',
                'ssh-rsync -F %s\\.ssh\\config' % os.getenv('USERPROFILE'),
                'source/folder',
                'some-user@some-host:path/on/server',
                ],
            }], command)

    def test_getPythonLibPath_default(self):
        """
        It will return a path which exists.
        """
        sut = BrinkPaver(setup=MINIMAL_SETUP)

        result = sut.getPythonLibPath()

        # We test `chevah-brink.egg-link` which should be the setup by pip
        # in development mode.
        # To make sure we don't get a generic Python path.
        # chevah_brink-0.79.0.dist-info
        site_packages = list(os.listdir(result))
        self.assertIn(
            'chevah_brink-%s.dist-info' % (__version__,), site_packages)

    def test_default_values(self):
        """
        It will make the runtime information and path available.

        To check brink.sh we also test the environment variables
        """
        sut = BrinkPaver(setup=MINIMAL_SETUP)

        self.assertEqual('python3.11', sut.python_version)

        if os.getenv('CHEVAH_BUILD', '') != '':
            # We run with custom builddir
            build_dir = 'build-brink-ț'
        else:
            build_dir = 'build-brink'

        # On Windows, brink path is always with backslashes.
        expected_path = os.path.join(
            os.getcwd(), build_dir).replace('\\', '/')

        self.assertEqual(expected_path, sut.path.build)

        self.assertEqual(build_dir, os.environ['PYTHONPATH'])
        self.assertEqual('python3.11', os.environ['CHEVAH_PYTHON'])


class TestBrinkGitParsing(BrinkTestCase):
    """
    Tests for parsing Git output.
    """
    def test_parse_many_cases(self):
        """
        Parse Git output showing multiple file actions.
        In particular, include a rename action (R100), and use its new filename.
        We use the `replace()` call to improve readability.
        """
        git_output = (
            "R100 chevah/pool.py chevah/slim_server/pool.py\n"
            "A chevah/slim_server/raw_config_parser.py\n"
            "M pavement.py\n"
            "A release-notes/5677-1.bugfix\n"
            "A release-notes/5677.bugfix"
        ).replace(' ', '\t')

        action_and_file_list = [
            ('r100',    'chevah/slim_server/pool.py'),
            ('a',       'chevah/slim_server/raw_config_parser.py'),
            ('m',       'pavement.py'),
            ('a',       'release-notes/5677-1.bugfix'),
            ('a',       'release-notes/5677.bugfix'),
        ]
        self.assertListEqual(action_and_file_list, parse_diff_list(git_output))
