# Copyright (c) 2015 Adi Roiban.
# See LICENSE for details.
"""
Authentication implementing anonymous account as discussed in RFC 1635.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

from zope.interface import implementer

from chevah.server.commons.exception import (
    CredentialsForbidden,
    InvalidAuthentication,
    )
from chevah.server.commons.interface import (
    IApplicationAccountConfiguration,
    IAuthentication,
    )
from chevah.server.commons.runnable import Runnable


@implementer(IAuthentication)
class AnonymousAuthentication(Runnable):
    """
    Simple authentication which does authentication without a password.

    Authentication is only done based on a fixed username.
    """
    PROPERTIES_REQUIRING_RESTART = ()

    _ANONYMOUS_ACCOUNT_NAME = 'anonymous'

    def _onStart(self, avatar=None):
        """
        See: Runnable.

        No action is taken for start.
        """

    def _onStop(self, avatar=None):
        """
        See: Runnable.

        Nothing is required for this auth method.
        """

    def authenticate(self, credentials):
        """
        See: `IAuthentication`
        """
        if credentials.username != self._ANONYMOUS_ACCOUNT_NAME:
            # Non-anonymous account. Allow other methods to check the account.
            return None

        anonymous_account_uuid = self.configuration.anonymous_account_uuid
        accounts = self.root.configuration.accounts
        account_configuration = accounts.getSection(
            self.configuration.anonymous_account_uuid)

        if not account_configuration:
            raise CredentialsForbidden(
                'Invalid account uuid "%s" associated with the anonymous '
                'account for "%s".' % (
                    anonymous_account_uuid, self.name,))

        if not IApplicationAccountConfiguration.providedBy(
                account_configuration):
            raise CredentialsForbidden(
                'Account uuid "%s" associated with the anonymous '
                'account for "%s" is not an application account.' % (
                    anonymous_account_uuid, self.name,))

        return account_configuration, ()

    def authenticateAdministrator(self, credentials):
        """
        See: `IAuthentication`
        """
        raise InvalidAuthentication(
            'Anonymous administrators are not supported.')
