# Copyright (c) 2015 Adi Roiban.
# See LICENSE for details.
"""
Authentication done for application accounts.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

from zope.interface import implementer

from chevah.server.authentication.credential_checker import (
    CredentialsValidator,
    )
from chevah.server.commons.exception import (
    CredentialsForbidden,
    InvalidAuthentication,
    )
from chevah.server.commons.interface import IAuthentication
from chevah.server.commons.runnable import Runnable


@implementer(IAuthentication)
class ApplicationAuthentication(Runnable):
    """
    Do authentication for application accounts with passwords stored in the
    configuration.
    """
    PROPERTIES_REQUIRING_RESTART = ()

    def _onStart(self, avatar=None):
        """
        See: Runnable.
        """
        # Does nothing as the configuration file should be always available.
        pass

    def _onStop(self, avatar=None):
        """
        See: Runnable.
        """
        # Does nothing as the configuration file should be always available.
        pass

    def authenticate(self, credentials):
        """
        See: `IAuthentication`
        """
        accounts = self.root.configuration.accounts
        account_configuration = accounts.getApplicationAccountConfiguration(
            credentials.username)

        if account_configuration is None:
            # User does not exist in the application.
            # Allow other methods to check it.
            return None

        allowed_groups = self._configuration.allowed_groups

        identity_groups = self.root.configuration.groups

        groups = []

        has_allowed_group = False

        if not allowed_groups:
            # We consider that group is allowed when no groups are filtered.
            has_allowed_group = True

        for group_uuid in account_configuration.group:

            group = identity_groups.getSection(group_uuid)

            if not group:
                raise InvalidAuthentication(
                    'Group does not exist: %s' % (group_uuid,))

            if allowed_groups and group_uuid in allowed_groups:
                has_allowed_group = True

            groups.append(group)

        if not has_allowed_group:
            raise CredentialsForbidden(
                'No group allowed for "%s".' % (self._configuration.name,))

        validator = CredentialsValidator(
            me=account_configuration,
            groups=groups,
            )
        result = validator.validate(credentials)

        if not result:
            raise CredentialsForbidden('Credentials denied')

        return account_configuration, groups

    def authenticateAdministrator(self, credentials):
        """
        See: `IAuthentication`
        """
        administrators = self.root.configuration.administrators
        administrator = administrators.getSectionByName(credentials.username)

        if not administrator:
            # Admin don't exists.
            # Allow other methods to check it.
            return None

        roles = []
        identity_roles = self.root.configuration.roles
        for role_uuid in administrator.roles:
            role = identity_roles.getSection(role_uuid)
            if not role:
                raise InvalidAuthentication(
                    'Role does not exist: %s' % (role_uuid,))
            roles.append(role)

        validator = CredentialsValidator(
            me=administrator,
            groups=roles,
            )
        result = validator.validate(credentials)

        if not result:
            raise CredentialsForbidden('Administrator credentials denied.')

        return administrator, roles
