# Copyright (c) 2014 Adi Roiban.
# See LICENSE for details.
"""
Code shared by authentication methods.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

import base64
import six

import pyotp
from passlib.context import CryptContext
from passlib.hash import (
    pbkdf2_sha256,
    pbkdf2_sha512,
    sha256_crypt,
    sha512_crypt,
    )
from twisted.internet import defer

from chevah.server.commons.constant import PASSWORD_CRYPT_ROUNDS, TYPE_NAME
from chevah.server.commons.exception import (
    CredentialsForbidden,
    InvalidAuthentication,
    ServerException,
    )
from chevah.server.commons.match import get_matched_groups
from chevah.server.commons.interface import IPasswordCredentials


hash_context = CryptContext(
    schemes=[
        'pbkdf2_sha1',
        'pbkdf2_sha256',
        'pbkdf2_sha512',
        'sha256_crypt',
        'sha512_crypt',
        ],
    default='sha512_crypt',
    )


class PasswordHandler(object):
    """
    Handler for passwords.
    """

    def identify(self, hash):
        """
        Return True if this is the hash for a password.
        """
        if not hash:
            return False

        # The context returns None for plain text.
        return bool(hash_context.identify(hash))

    def verify(self, plain, hash):
        """
        Return True if plain password matches the `hash`.

        Raise ValueError if hash is not a hashed password.
        """
        return hash_context.verify(plain, hash)

    def encrypt(
        self, plain,
        rounds=PASSWORD_CRYPT_ROUNDS, format=TYPE_NAME.CRYPT_SHA512,
            ):
        """
        Return the hashed version of the `plain`.
        """
        if not plain:
            raise ServerException('No value provided for password.')

        if len(plain) > 128:
            raise ServerException('Password is too long.')

        def get_hash():
            """
            Wrap all know methods to shared the Py2 Unicode conversion code.
            """
            if format == TYPE_NAME.CRYPT_SHA512:
                return sha512_crypt.hash(plain, rounds=rounds)

            if format == TYPE_NAME.CRYPT_SHA256:
                return sha256_crypt.hash(plain, rounds=rounds)

            if format == TYPE_NAME.PBKDF2_SHA512:
                return pbkdf2_sha512.hash(plain, rounds=rounds)

            if format == TYPE_NAME.PBKDF2_SHA256:
                return pbkdf2_sha256.hash(plain, rounds=rounds)

            raise ServerException('Unknown password format: "%s".' % (format,))

        result = get_hash()
        if six.PY2:
            result = result.decode('ascii')

        return result


def check_password_credentials_validity(method_name, credentials):
    """
    High level check for credentials.

    Returns None if the password credentials are valid.

    Return a failed deferred on errors.
    """
    if not IPasswordCredentials.providedBy(credentials):
        return defer.fail(CredentialsForbidden(
            '%s method only accepts username/password credentials.' % (
                method_name,)))

    if not credentials.password or not credentials.password.strip():
        return defer.fail(CredentialsForbidden(
            '%s method does not allow empty passwords.' % (method_name,)))

    return defer.succeed(None)


def apply_group_mapping(group_mapping, external, entry):
    """
    Set the group UUID for the `external` configuration based on the
    match between the `entry` attribute key and list of value and the
    provided `group_mapping`.

    `entry` keys should be lowercase to help implement the case insensitive
    matching.

    This is code shared between LDAP and RADIUS.

    See the guides/ldap.rst

    An attribute can have multiple values.

    It will stop at the first match.

    It assumes that `group_mapping` was already validated for format.

    It return the updated raw account configuration.

    The account configuration is not validated here.
    """
    external = external.copy()
    if not group_mapping:
        # No special group mapping
        return external

    # We have a fallback group, so start with it.
    fallback_group = tuple(group_mapping[0])
    external['group'] = fallback_group

    # We skip the fallback groups.
    for rule in group_mapping[1:]:
        # We can have multiple values for the same attribute.
        attribute_name = rule[0].lower()
        matching_rule = rule[1]
        groups_uuid = rule[2:]
        values = entry.get(attribute_name, [])
        for value in values:
            # Check that expression is valid for each valid.
            value = value.decode('utf-8')
            if not get_matched_groups(
                    value=value, rule=matching_rule, usage='group_mapping'):
                continue
            external['group'] = tuple(groups_uuid)
            return external

    return external


def get_TOTP_secret(url):
    """
    Return the secret used for TOTP from the `url`.

    Raise `InvalidAuthentication` if the provided `url` is empty.
    """
    # Here to prevent circular imports.
    from chevah.server.commons.configuration import parse_url

    if not url:
        raise InvalidAuthentication('Empty MFA configuration.')

    try:
        url = parse_url(url)
    except ServerException as error:
        raise InvalidAuthentication(error.message)

    if not url['protocol'] == 'otpauth':
        raise InvalidAuthentication(
            'Invalid "multi_factor_authentication" URL.')

    if not url['location'] == 'totp':
        raise InvalidAuthentication(
            'Only TOTP is supported for '
            '"multi_factor_authentication" URL.')

    query = url['query']
    if not query:
        raise InvalidAuthentication(
            'Missing parameters from '
            '"multi_factor_authentication" URL.')

    secret = query.get('secret', [''])[0]
    if not secret:
        raise InvalidAuthentication(
            'Missing "secret" parameter from '
            '"multi_factor_authentication" URL.')

    try:
        base64.b32decode(secret)
    except (TypeError, ValueError):
        raise InvalidAuthentication(
            'Invalid "secret" parameter from '
            '"multi_factor_authentication" URL.')

    return secret


def split_TOTP_password(raw_password):
    """
    Return a tuple of (password, totp_code) for the `raw_password` which
    is expected to contain the TOTP code appended.
    """
    provided_totp = raw_password[-6:]
    provided_password = raw_password[:-6]
    try:
        # Try to validate the token and give a clue if something
        # is wrong.
        int(provided_totp)
    except ValueError:
        raise CredentialsForbidden('Invalid/missing TOTP token.')

    if not provided_password:
        raise CredentialsForbidden('TOTP token provided without password.')

    return provided_password, provided_totp


def verify_TOTP(secret, provided_totp, memory):
    """
    Check that the `provided_totp` code is currently valid for the shared
    TOTP `secret`.

    It uses `memory` to prevent replay attacks.
    `memory` should be initialized with something like
    `memory = deque(maxlen=3)`.
    """
    totp = pyotp.TOTP(secret)
    if not totp.verify(provided_totp, valid_window=1):
        if totp.verify(provided_totp, valid_window=20):
            # It may be that the codes are out of sync, up to 10 minutes.
            raise CredentialsForbidden(
                'Wrong TOTP / one time verification code. '
                'It is possible that the clocks are out of sync.'
                )

        raise CredentialsForbidden(
            'Wrong TOTP / one time verification code.')

    if provided_totp in memory:
        raise CredentialsForbidden('TOTP token used more than once.')
    memory.append(provided_totp)


# Expose it as a singleton.
password_handler = PasswordHandler()
