# Copyright (c) 2014 Adi Roiban.
# See LICENSE for details.
"""
Configuration handling for authentication.
"""
from __future__ import absolute_import, unicode_literals

from zope.interface import implementer

from chevah.server.commons.constant import (
    CONFIGURATION_SECTION_AUTHENTICATIONS,
    )
from chevah.server.commons.interface import (
    IAnonymousAuthenticationConfiguration,
    IApplicationAuthenticationConfiguration,
    IAuthenticationsConfiguration,
    IAuthenticationConfiguration,
    IDenyUsernameAuthenticationConfiguration,
    ILocalFileAuthenticationConfiguration,
    IIPTimeBanAuthenticationConfiguration,
    )
from chevah.server.configuration.option import (
    EntityConfiguration,
    LocalPathConfigurationMixin,
    TypeOptionsMixin,
    )
from chevah.server.configuration.set import SetWithTypeConfiguration
from chevah.server.configuration.model import (
    EmptyProperty,
    WritableBoolean,
    WritableInteger,
    WritableString,
    WritableCSVUnique,
    WritableCSVUniqueInsensitive,
    )


@implementer(IAuthenticationsConfiguration)
class AuthenticationsConfiguration(SetWithTypeConfiguration):
    """
    See: `IAuthenticationsConfiguration`.
    """
    SECTION_PREFIX = CONFIGURATION_SECTION_AUTHENTICATIONS
    __type_interface__ = IAuthenticationConfiguration


class _AuthenticationConfigurationBase(EntityConfiguration, TypeOptionsMixin):
    """
    Shared attributes by all authentication configurations.
    """
    SECTION_PREFIX = CONFIGURATION_SECTION_AUTHENTICATIONS


@implementer(ILocalFileAuthenticationConfiguration)
class LocalFileAuthenticationConfiguration(
        _AuthenticationConfigurationBase, LocalPathConfigurationMixin):
    """
    Configuration for the local file authentication.

    For now, the local file authentication runnable is managing the loading
    and saving of the identities stored in the local file.
    """
    external_management = WritableBoolean('external_management')

    def __init__(self, parent, uuid):
        super(LocalFileAuthenticationConfiguration, self).__init__(
            parent, uuid)
        self.unloadIdentity()

    def unloadIdentity(self):
        """
        Make sure we have an empty identity database.
        """
        self.groups = self._getEmptyProperty()
        self.accounts = self._getEmptyProperty()
        self._roles = self._getEmptyProperty()
        self._administrators = self._getEmptyProperty()

    def _getEmptyProperty(self):
        """
        Create a new
        """
        return EmptyProperty(self, 'stopped %s.' % (self.name))

    @property
    def administrators(self):
        return self._administrators

    @property
    def roles(self):
        return self._roles


@implementer(IDenyUsernameAuthenticationConfiguration)
class DenyUsernameAuthenticationConfiguration(
        _AuthenticationConfigurationBase):
    """
    Configuration for the deny username authentication.
    """
    usernames = WritableCSVUniqueInsensitive('usernames')


@implementer(IIPTimeBanAuthenticationConfiguration)
class IPTimeBanAuthenticationConfiguration(
        _AuthenticationConfigurationBase):
    """
    Configuration for the IP time based bad authentication.
    """
    ban_interval = WritableInteger('ban_interval', default=3600)
    ban_after_count = WritableInteger('ban_after_count', default=5)


class _AllowedGroupsMixin(object):
    """
    Shared configuration for methods allowing certain groups.
    """
    allowed_groups = WritableCSVUnique('allowed_groups')


@implementer(IApplicationAuthenticationConfiguration)
class ApplicationAuthenticationConfiguration(
        _AuthenticationConfigurationBase, _AllowedGroupsMixin):
    """
    Configuration for application authentication.
    """


@implementer(IAnonymousAuthenticationConfiguration)
class AnonymousAuthenticationConfiguration(_AuthenticationConfigurationBase):
    """
    Configuration for an anonymous authentication.
    """
    anonymous_account_uuid = WritableString('anonymous_account_uuid')
