# Copyright (c) 2011-2014 Adi Roiban.
# See LICENSE for details.
"""
Authentication as defined by the Twisted Cred framework.

This only checks that credentials for an account are valid and passes
the authentication response to the realm for detailed authorization.

The authentication process does some part of authorization, as the
authorization request includes peer IP or requested protocol or service ID,
and the authentication server might reject valid credentials, if they
are not authorized for a certain IP/protocol/service.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

import base64
import binascii
from collections import deque
import time
import six

from datetime import datetime
from ordered_set import OrderedSet
from twisted.cred.checkers import ICredentialsChecker
from twisted.internet import defer, reactor
from zope.interface import implementer
from ipaddress import ip_address, ip_network

from chevah_compat import (
    HasImpersonatedAvatar,
    local_filesystem,
    process_capabilities,
    system_users,
    )
from chevah_compat.exceptions import (
    ChangeUserException,
    )
from chevah_keycert.exceptions import KeyCertException
from chevah_keycert.ssh import Key
from chevah.server import force_unicode
from chevah.server.authentication.base import (
    get_TOTP_secret,
    password_handler,
    verify_TOTP,
    split_TOTP_password,
    )
from chevah.server.commons import UTC, x509
from chevah.server.commons.constant import (
    CREDENTIALS_MAX_LENGTH,
    CONFIGURATION_DEFAULT_AUTHENTICATION,
    TYPE_NAME,
    AUTH_PROCESS_TIMEOUT,
    )
from chevah.server.commons.exception import (
    CredentialsForbidden,
    InvalidAuthentication,
    InsufficientAuthentication,
    ServerError,
    ServerException,
    )
from chevah.server.commons.interface import (
    IAdministratorConfiguration,
    IAS2AnonymousCredentials,
    IRoleConfiguration,
    ILocalManagerCredentials,
    IOSAccountConfiguration,
    IOSAdministratorConfiguration,
    IPasswordCredentials,
    ISSHKeyCredentials,
    ISSLCertificateCredentials,
    )
from chevah.server.configuration.model import (
    is_inherited,
    get_relaxed_datetime,
    )
from chevah.server.identity.avatar import (
    expand_group_path,
    PeerAvatar,
    _AvatarMixin,
    )


@implementer(ICredentialsChecker)
class ServiceCredentialChecker(object):
    """
    Credential checker for a service using any of the supported pluggable
    authentication methods.

    Emit the following signals on the handled credentials:
    * accepted - when the credentials were accepted
    * rejected - when the credentials were rejected

    If SFTPPlus Web Admin is enabled users are first authenticated against
    the web login and fall back to local system authentication.
    """

    _scheduler = reactor

    credentialInterfaces = (
        IPasswordCredentials,
        ISSHKeyCredentials,
        ISSLCertificateCredentials,
        IAS2AnonymousCredentials,
        )

    def __init__(self, service_configuration):
        self._service_configuration = service_configuration
        # This is the cache of currently ongoing authentication sessions
        # that require multiple credential types or advanced validation.
        self._auth_sessions = {}

    @property
    def main_configuration(self):
        return self.root.configuration

    @property
    def service_configuration(self):
        return self._service_configuration

    @property
    def root(self):
        """
        Root of the objects hierarchy.
        """
        return self._service_configuration.root

    def requestAvatarId(self, credentials):
        """
        Check that credentials are valid.

        Returns an account configuration.
        """
        deferred = self._requestAvatarId(credentials)
        deferred.addCallback(self._cbCredentialsAccepted, credentials)
        deferred.addErrback(self._ebCredentialsRejected, credentials)
        return deferred

    def _cbCredentialsAccepted(self, result, credentials):
        """
        Called when credentials were accepted.
        """
        credentials.notify('accepted')
        return result

    def _ebCredentialsRejected(self, failure, credentials):
        """
        Called when credentials were not accepted.
        """
        credentials.notify('rejected')

        if failure.check(InsufficientAuthentication, ServerError):
            # These are expected errors that should be handled by
            # higher level code.
            return failure

        # Treat any non ServerError as internal server error.
        error_details = force_unicode(failure.getBriefTraceback())
        raise ServerError('20059', data={
            'name': credentials.username,
            'details': error_details,
            },
            peer=credentials.peer,
            )

    @defer.inlineCallbacks
    def _requestAvatarId(self, credentials):
        """
        Check credentials according to current server configuration

        It is called by `requestAvatarId` to wrap internal server errors.

        On success return the account configuration that can be used by the
        Realm to generate a full avatar.

        Raise an exceptions on failure.
        """
        data = {
            'username': credentials.username,
            'credentials_type': credentials.kind_name,
            'method_type': '',
            'method_name': '',
            }
        avatar = PeerAvatar(
            peer=credentials.peer,
            component=credentials.creator,
            )

        # FIXME:4590:
        # This should be moved as a RuntimeError for
        # Credential initialization as protocols should not
        # read long usernames/passwords/certificates/SSH keys.
        try:
            self._validateCredentials(credentials)
        except CredentialsForbidden as error:
            # The credentails are invalid, before trying any authentication
            # method.
            # Fill some default data.
            data['method_type'] = 'credentials sanity check'
            data['method_name'] = ''
            data['details'] = error.message
            data['response'] = error.response
            raise ServerError('20136', avatar=avatar, data=data)

        authentications = self._getAuthentications(credentials)
        try:
            for method_uuid in authentications:
                # Fill some default data.
                data['method_type'] = 'unknown'
                data['method_name'] = method_uuid

                result = yield self._authenticate(
                    method_uuid, credentials, data)

                if result is None:
                    # Continue with next method as credentials were not
                    # accepted but neither forbidden.
                    continue

                account_configuration, groups, ignored_groups = (
                    self._getValidatedConfiguration(credentials, result))

                ignored_groups_description = ''
                if ignored_groups:
                    ignored_group_names = [g.name for g in ignored_groups]
                    ignored_groups_description = 'Ignored: %s.' % (
                        ', '.join(ignored_group_names))

                # Stop looking for other methods as credentials were
                # authenticated
                # and we have an account configuration.
                data['account_uuid'] = account_configuration.uuid
                data['account_name'] = account_configuration.name
                # FIXME:2585:
                # Should be replaced with `type`.
                # Here while the cleanup is not yet done.
                data['account_type'] = account_configuration.kind_name
                group_names = [group.name for group in groups]
                data['group_name'] = ', '.join(group_names)
                data['group_names'] = group_names
                data['ignored_groups'] = ignored_groups_description
                # FIXME:5247:
                # For now, this only includes the last accepted credentials.
                self.root.emitEvent('20137', avatar=avatar, data=data)
                defer.returnValue((
                    account_configuration, groups, credentials))

        except InvalidAuthentication as error:
            # FIXME:2583:
            # For now exceptions are converted into errors.
            data['details'] = error.message
            raise ServerError('20142', avatar=avatar, data=data)

        except InsufficientAuthentication as error:
            # Log the error and forward it to the higher level protocol.
            data['required_credentials'] = error.remaining_credentials
            self.root.emitEvent('20126', avatar=avatar, data=data)
            raise error

        except CredentialsForbidden as error:
            data['details'] = error.message
            data['response'] = error.response
            raise ServerError('20136', avatar=avatar, data=data)

        # Tried all method so we fail as there is no method left to
        # check the credentials.
        data = {
            'username': credentials.username,
            'credentials_type': credentials.kind_name,
            }
        raise ServerError('20021', avatar=avatar, data=data)

    def _validateCredentials(self, credentials):
        """
        Check that credentials are valid before sending them to the
        authentication methods.
        """
        if not credentials.username:
            raise CredentialsForbidden('Username can\'t be empty.')

        # FIXME:4472:
        # Also check public SSH key and SSL certificate
        # But doing those checks here is too late.
        # They should be done at protocol level.
        if len(credentials.username) > CREDENTIALS_MAX_LENGTH:
            raise CredentialsForbidden('Username too long.')

        if (
            IPasswordCredentials.providedBy(credentials)
            and credentials.password
            and len(credentials.password) > CREDENTIALS_MAX_LENGTH
                ):
            raise CredentialsForbidden('Password too long.')

    def _getAuthentications(self, credentials):
        """
        Return the list of authentication to be used for credentials.

        The result is never empty.
        """
        # Try service's authentications.
        authentications = (
            self.service_configuration.parent.authentications)

        if authentications:
            return authentications

        # Fall back to server's global authentications.
        if ILocalManagerCredentials.providedBy(credentials):
            # For manager we don't have global auth
            # so we fallback to a simple list.
            # This should never happen in production as the manager should
            # normaly alwasy have a list.
            return (CONFIGURATION_DEFAULT_AUTHENTICATION,)

        # Even when no auth method are explicitly defined,
        # the configuration layer will return the default auth UUID.
        return self.main_configuration.server.authentications

    def _authenticate(self, method_uuid, credentials, data):
        """
        Try to validate the `credentials` using the auth `method_uuid`.
        """
        authentications = self.main_configuration.root.authentications
        method = authentications.getSection(method_uuid)
        if not method:
            return defer.fail(InvalidAuthentication(
                'No authentication method with UUID "%s".' % (method_uuid,)))

        data['method_type'] = method.type
        data['method_name'] = method.name

        if not method.operating:
            return defer.fail(InvalidAuthentication(
                'The authentication method is not active.'))

        if ILocalManagerCredentials.providedBy(credentials):
            return method.authenticateAdministrator(credentials)
        else:
            return method.authenticate(credentials)

    def _getValidatedConfiguration(self, credentials, result):
        """
        Validate the account/administrator and group/role configuration.

        This is called after the credentials were successfully validated.

        Return a tuple with valid (account/administrator, groups/roles)

        Raise an exception if the configuration returned by the
        authentication method is not valid.
        """
        account, groups = result

        if ILocalManagerCredentials.providedBy(credentials):
            group_type = 'Role'
            account_type = 'Administrator'
            groups_uuid = account.roles
            get_section = self.main_configuration.roles.getSection

        else:
            group_type = 'Group'
            account_type = 'Account'
            groups_uuid = account.group
            get_section = self.main_configuration.groups.getSection

        if not account.enabled:
            message = '%s \"%s\" not enabled.' % (account_type, account.name,)
            raise CredentialsForbidden(message)

        if not groups:
            # If the authentication method has not resolved the groups/roles,
            # we try to resolve them here.
            groups = []
            for group_uuid in groups_uuid:
                group = get_section(group_uuid)
                if not group:
                    message = '%s does not exist with UUID: %s.' % (
                        group_type, group_uuid,)
                    raise InvalidAuthentication(message)
                groups.append(group)

        for group in groups:
            if not group.enabled:
                message = '%s \"%s\" not enabled.' % (group_type, group.name,)
                raise CredentialsForbidden(message)

        # For source address, we validate first at the account level.
        _check_source_ip_access(credentials.peer, account, account_type)
        # Then at group level.
        groups, ignored = _filter_associated_groups_or_roles(
            credentials.peer, groups, group_type)

        self._checkEnoughCredentials(credentials, account, groups)

        return account, tuple(groups), ignored

    def _checkEnoughCredentials(self, credentials, account, groups):
        """
        Check that the credentials for the account are enough to
        authenticate it.

        Raise `InsufficientAuthentication` containing the list of
        credentials types that still need to be validated.

        It also cleans the expired authentication processes for which the
        required credentials were not provide.
        """
        if not groups:
            raise CredentialsForbidden(
                'No group/role allowed to be associated with the user.')

        group = groups[0]
        if ILocalManagerCredentials.providedBy(credentials):
            # Only check for non administration authentication requests.
            return

        # Resolve account configuration.
        required_credentials = account.required_credentials
        if is_inherited(required_credentials):
            required_credentials = group.required_credentials

        if not required_credentials:
            # We should not hit this at runtime as the configuration should
            # make sure we always have a value.
            # Since this is security code, we double check.
            raise InvalidAuthentication(
                'Account configured without any explicit '
                '"required_credentials".')

        if TYPE_NAME.ANY in required_credentials:
            # Any credentials are accepted for this account.
            return

        # FIXME:5246:
        # This is vulnerable to IP spoofing attacks.
        # Try to use credentials.peer.uuid instead.
        auth_id = '%s-%s-%s-%s' % (
            credentials.username,
            credentials.creator.uuid,
            credentials.peer.host,
            credentials.peer.port,
            )
        now = self._scheduler.seconds()

        # We are going to modify the dict as we traverse it,
        # this is why a snapshot is used.
        # We check for existing matching session and also do a cleanup.
        found_session = None
        for session_id in list(self._auth_sessions.keys()):
            still_required, timestamp = self._auth_sessions[session_id]

            if now - timestamp > AUTH_PROCESS_TIMEOUT:
                # Previous session has expired.
                del self._auth_sessions[session_id]
                continue

            if session_id != auth_id:
                continue

            # We remove the existing session as later it will be replaced
            # with an updated one based on latest credentials.
            del self._auth_sessions[session_id]
            found_session = still_required

        currently_required = found_session
        if not currently_required:
            # This is the first request from the auth session.
            # Start with the full list.
            # Convert any value to a list as we will mutate it later.
            currently_required = list(required_credentials)

        credentials_type = credentials.type
        if credentials_type == TYPE_NAME.PASSWORD_BASIC_AUTH:
            credentials_type = TYPE_NAME.PASSWORD

        try:
            currently_required.remove(credentials_type)
        except ValueError:
            """
            Newly authenticated credentials are not required.
            Keep the same list.
            """

        if not currently_required:
            # No extra credentials required.
            return

        # FIXME:5247:
        # Find a way to keep track of previously accepted credentials so that
        # we can include them in single audit event.
        self._auth_sessions[auth_id] = (currently_required, now)
        raise InsufficientAuthentication(currently_required)


def check_account_validity(me, groups):
    """
    Check that `me` is a valid account.

    Raises `CredentialsForbidden` on invalid accounts.
    """
    if not me.enabled:
        raise CredentialsForbidden('Account is disabled.')

    for group in groups:
        if not group.enabled:
            if IRoleConfiguration.providedBy(group):
                message = 'Role "%s" is disabled.' % (group.name,)
            else:
                message = 'Group "%s" is disabled.' % (group.name,)
            raise CredentialsForbidden(message)

    expire = _get_expire_datetime(me)
    if expire:
        now = datetime.utcnow().replace(tzinfo=UTC)
        if expire < now:
            raise CredentialsForbidden('Account has expired.')


def _get_expire_datetime(me):
    """
    Return the date and time when this account expires.

    Return `None` if there is no configured date.
    """
    if IAdministratorConfiguration.providedBy(me):
        # For now, admin accounts, never expire.
        return None

    return get_relaxed_datetime(me.expire_datetime, 'Account expiration')


# List of actions allowed for source IP filter.
_ALLOWED_ACTIONS = [
    TYPE_NAME.ALLOW,
    TYPE_NAME.DENY,
    ]


def _check_source_ip_access(address, account, account_type):
    """
    Check if the source_ip is allowed based on account/admin configuration.

    `account` is an account or administrator.

    Raise an exception is `address` is not allowed.
    """
    # Check direct IP filtering.
    origin = '{}s/{}'.format(account_type.lower(), account.uuid)
    if _is_address_allowed(
        address,
        account.source_ip_filter,
        origin=origin,
            ):
        return

    raise CredentialsForbidden(
        'Source IP denied by direct "source_ip_filter" configuration.')


def _filter_associated_groups_or_roles(address, groups_or_roles, group_type):
    """
    Return a tuple with the list of groups associated or ignored for account,
    when connected from source `address`.

    This should be called after the source IP access deny was checked.

    `account` is an account or an administrator.
    `group` is list of groups or roles.
    """
    associated = []
    ignored = []
    for group in groups_or_roles:

        access_rules = group.source_ip_filter
        if not access_rules:
            # Group associated from any source IP.
            associated.append(group)
            continue

        origin = "{}s/{}".format(group_type.lower(), group.uuid)
        action = _get_action(address, access_rules, origin)

        if action == TYPE_NAME.ALLOW:
            # Source IP explicitly allowed to be associated.
            associated.append(group)
            continue

        ignored.append(group)

    # Return the groups left after the source_ip filtering.
    return (associated, ignored)


def _is_address_allowed(address, access_rules, origin):
    """
    Retrun `True` if `address` is allowed based on the list `access_rules`.

    If `access_rules` is empty, the address is allowed by default.
    """
    if not access_rules:
        return True

    action = _get_action(address, access_rules, origin)

    if not action:
        # No match -> not allowed.
        return False

    if action in [TYPE_NAME.DENY]:
        # Explicit deny match.
        return False

    # All other rules are to allow access.
    return True


def _get_action(address, access_rules, origin):
    """
    Return the first action from the `access_rules` based on the `address`.
    """
    host = address.host
    if six.PY2:
        host = host.decode('utf-8')
    current_address = ip_address(host)

    for rule in access_rules:
        try:
            action, target = rule
        except ValueError:
            raise InvalidAuthentication(
                'Invalid "source_ip_filter" rule for {}: "{}".'.format(
                    origin, force_unicode(rule))
                )

        action = action.lower()
        if action not in _ALLOWED_ACTIONS:
            raise InvalidAuthentication(
                'Invalid "source_ip_filter" action for {}: "{}".'.format(
                    origin, force_unicode(action))
                )

        try:
            # See if we have a direct IP address.
            target = [ip_address(target)]
        except ValueError:
            # See if we have an IP range.
            try:
                target = ip_network(target)
            except ValueError:
                # Not an address or network.
                raise InvalidAuthentication(
                    'Invalid "source_ip_filter" address for {}: "{}".'.format(
                        origin, force_unicode(target))
                    )

        if current_address in target:
            return action

        # Nothing more to check.
        # Continue with the next filter value.

    # No matched action
    return None


class CredentialsValidator(HasImpersonatedAvatar, _AvatarMixin):
    """
    Shared code to check credentials for our own authentication methods.

    It received 'me' as the configuration of an account or administrator.

    `groups` is a list of group/role configurations that were already
    check that they exist.
    We might want to refactor to just receive a list of raw configured
    group/role UUID and resolve them here.
    """

    # Here to help with testing.
    _now = time.time

    def __init__(
            self, me, groups, pam_usage='', pam_service='login'):
        super(CredentialsValidator, self).__init__()
        self._me = me
        self._groups = groups
        self._pam_service = pam_service
        self._osPasswordAuthenticationMethod = (
            self._getOSPasswordAuthentication(pam_usage))

    @property
    def name(self):
        return self._me.name

    @property
    def use_impersonation(self):
        """
        See: HasImpersonatedAvatar

        Return True if avatar operations should be impersonated.
        """
        return self._isOSAccount()

    def _getOSPasswordAuthentication(self, pam_usage):
        """
        Return the authentication method used to validated password against
        the OS.
        """
        if process_capabilities.os_name == 'windows':
            # On Windows PAM is not supported
            return self._validateWithoutPAM

        if not pam_usage:
            # PAM is not enabled form configuration.
            return self._validateWithoutPAM

        if pam_usage == 'fallback':
            return self._validateWithPAMFallback
        elif pam_usage == 'exclusive':
            return self._validateWithPAMOnly
        else:
            raise CredentialsForbidden('Unknown pam_usage configuration.')

    def _getSSHAuthorizedKeys(self):
        """
        Return the list of keys as text.
        """
        keys = OrderedSet()
        keys.update(self._getAuthorizedKeysFromPath())
        keys.update(self._getAuthorizedKeysFromDirectConfiguration())
        return list(keys)

    def _getAuthorizedKeysFromPath(self):
        """
        Return a list of authorized keys stored on local file.
        """
        path = self._ssh_authorized_keys_path
        if not path:
            return []

        segments = local_filesystem.getSegmentsFromRealPath(path)
        if local_filesystem.isFolder(segments):
            return self._getAuthorizedKeysFromDirectory(segments)
        else:
            return self._getAuthorizedKeysFromFile(segments)

    def _getAuthorizedKeysFromFile(self, segments, size=None):
        """
        Read OpenSSH public key from a single file.
        """
        keys = []
        stream = None
        try:
            stream = self._openFileForReading(segments, size=size)
            keys.extend(self._getAuthorizedKeysFromContent(
                force_unicode(stream.read())))
        finally:
            # The coverate is 100%, but there is a bug in branch reporting.
            if stream:  # noqa:cover
                stream.close()
        return keys

    def _getAuthorizedKeysFromDirectory(self, segments):
        """
        Read SSH public key from files found in directory.
        """
        keys = []
        for member in local_filesystem.iterateFolderContent(segments):
            keys.extend(self._getAuthorizedKeysFromFile(
                segments + [member.name], size=member.size))
        return keys

    def _getAuthorizedKeysFromDirectConfiguration(self):
        """
        Return a list of SSH keys as directly configured for an account.
        """
        return self._getAuthorizedKeysFromContent(
            self._me.ssh_authorized_keys_content)

    def _getAuthorizedKeysFromContent(self, content):
        """
        Return a list of SSH keys from raw configuration content.
        """
        lines = []
        content = content.strip()
        if not content:
            return lines

        if content.startswith('----'):
            # Try to load a single SSH key as X509 certificate
            # or SSH2 public key.
            try:
                key = Key.fromString(content)
                if not key.isPublic():
                    raise InvalidAuthentication(
                        'Private key found for account authentication. '
                        'Only public keys are allowed.'
                        )
                return [key.toString(type='openssh').decode('ascii')]
            except KeyCertException as error:
                raise InvalidAuthentication(
                    'Invalid configuration for ssh_authorized_keys_content. '
                    '%s' % (force_unicode(error),))

        for line in content.split('\n'):
            line = line.strip()
            if not line:
                # Ignore empty lines.
                continue
            lines.append(line)
        return lines

    @property
    def _ssh_authorized_keys_path(self):
        """
        Return the path to the authorizes ssh keys file used by this
        account.
        """
        path = self._me.ssh_authorized_keys_path

        if is_inherited(path):
            # FIXME:5480:
            # See if we should inherit from multiple groups.
            # Most probably leave it for another ticket.
            group_path = self._groups[0].ssh_authorized_keys_path
            if group_path:
                path = expand_group_path(group_path, self._me)
            else:
                # SSH keys are disabled at the group level.
                path = None

        return path

    def _openFileForReading(self, segments, size=None):
        """
        Open a file for reading.

        This is here to help with reading SSH keys by local accounts who
        require to change the process user id.

        We can not just pass the avatar to the filesystem since the avatar
        may be locked in the fs.
        """
        with self.getImpersonationContext():
            try:
                if size is None:
                    attributes = local_filesystem.getAttributes(segments)
                    size = attributes.size

                if size > 100 * 1024:
                    raise InvalidAuthentication(
                        'Failed to read the file. File is too big.')

                new_file = local_filesystem.openFileForReading(segments)
            except (IOError, OSError) as io_error:
                path = local_filesystem.getRealPathFromSegments(segments)
                message = force_unicode(io_error)
                raise InvalidAuthentication(
                    'Failed to read authorized SSH keys file \"%s\". %s' % (
                        path, message))

        return new_file

    def _isOSAccount(self):
        """
        True if this is an OS account, False otherwise.
        """
        if IOSAccountConfiguration.providedBy(self._me):
            return True
        elif IOSAdministratorConfiguration.providedBy(self._me):
            return True
        else:
            return False

    def validate(self, credentials):
        """
        Check credentials.
        """
        try:
            return self._validate(credentials)
        except ServerException as error:
            raise InvalidAuthentication(error.message)
        except ChangeUserException as error:
            raise CredentialsForbidden(
                'Unless SFTPPlus process is started as root, '
                'the OS account cannot be authenticated. %s' % (
                    error.message,))

    def _validate(self, credentials):
        """
        Return True if credentials are valid.

        It raises various exception to signal the failure reason.
        """
        # FIXME:5732:
        # This should be used in ServiceCredentialChecker to
        # validate other methods.
        check_account_validity(self._me, self._groups)

        if self._me.name != credentials.username:
            raise InvalidAuthentication('These credentials are not for me.')

        if IPasswordCredentials.providedBy(credentials):
            return self._validatePassword(credentials)

        if ISSHKeyCredentials.providedBy(credentials):
            return self._validateSSHKey(credentials)

        if ISSLCertificateCredentials.providedBy(credentials):
            return self._validateSSLCertificate(credentials)

        if IAS2AnonymousCredentials.providedBy(credentials):
            return self._validateAS2Anonymous(credentials)

        raise InvalidAuthentication('Unsupported credential type.')

    def _validatePassword(self, credentials):
        """
        Check credentials based on username and password.
        """
        # Stop if credentials have no password.
        if not credentials.password:
            raise CredentialsForbidden('No password was provided.')

        if self._isOSAccount():
            return self._validateOSPassword(credentials)
        else:
            return self._validateApplicationPassword(credentials)

    def _validateApplicationPassword(self, credentials):
        """
        Validate password for application accounts.
        """
        if not self._isPasswordValid():
            raise CredentialsForbidden('Password has expired.')

        account_password = self._me.password

        # Stop is account has no password.
        if not account_password:
            raise CredentialsForbidden('Password is disabled.')

        if not password_handler.identify(account_password):
            raise InvalidAuthentication(
                'Unknown format for the stored password.')

        provided_password = credentials.password
        account_totp = None
        provided_totp = None
        if self._me.multi_factor_authentication:
            # Account requires TOTP authentication.
            account_totp = get_TOTP_secret(
                self._me.multi_factor_authentication)
            provided_password, provided_totp = split_TOTP_password(
                provided_password)

        try:
            is_valid = password_handler.verify(
                provided_password, account_password)
        except ValueError as error:
            # We shouldn't end up here as we identify the password first.
            raise InvalidAuthentication(
                'Stored hashed password is not valid. %s' % (
                    force_unicode(error),))

        if not is_valid:
            raise CredentialsForbidden('Wrong password.')

        # We have a valid password.
        if not account_totp:
            # All good. No need to validate TOTP.
            return True

        if self._me.last_totps is None:
            self._me.last_totps = deque(maxlen=3)

        verify_TOTP(
            secret=account_totp,
            provided_totp=provided_totp,
            memory=self._me.last_totps,
            )

        return True

    def _isPasswordValid(self):
        """
        Return True if the password stored in the app config for this
        account is still valid.
        """
        duration = self._getOwnValueOrInherited('password_lifetime')

        if not duration:
            # Account had no limit for password.
            return True

        last_update = self._me.last_password_update
        if not last_update:
            # Password was not yet reset.
            # Consider it invalid.
            return False

        # one day = 86400 = 60 * 60 * 24
        age = (self._now() - last_update) / 86400
        return age < duration

    def _validateOSPassword(self, credentials):
        """
        Checks password against OS.
        """
        if not system_users.userExists(credentials.username):
            return False

        return self._osPasswordAuthenticationMethod(credentials)

    def _validateWithoutPAM(self, credentials):
        """
        Don't use PAM to check credentials.
        """
        passwords_match, token = (
            system_users.authenticateWithUsernameAndPassword(
                username=credentials.username,
                password=credentials.password,
                ))

        # Find a better way for passing the token from validator to
        # service avatar creation.
        credentials.token = token
        return passwords_match

    def _validateWithPAMOnly(self, credentials):
        """
        Only use PAM to check credentials.
        """
        credentials.token = None
        return system_users.pamWithUsernameAndPassword(
            username=credentials.username,
            password=credentials.password,
            service=self._pam_service,
            )

    def _validateWithPAMFallback(self, credentials):
        """
        Try standard OS authentication and then continue with PAM.
        """
        result = self._validateWithoutPAM(credentials)

        if result is None:
            result = self._validateWithPAMOnly(credentials)

        return result

    def _validateSSHKey(self, credentials):
        """
        Check that ssh credentials are valid for the requested account.

        Return True if key is valid.

        Raise an exception if SSH key is not valid.
        """
        if not credentials.key_data:
            raise InvalidAuthentication(
                'SSH key credentials without public key.')

        if (
            self.use_impersonation and
            process_capabilities.os_name == 'windows'
                ):
            raise CredentialsForbidden(
                'SSH key authentication not available for Windows accounts.')

        authorized_keys = self._getSSHAuthorizedKeys()
        if not authorized_keys:
            raise CredentialsForbidden('SSH key authentication disabled.')

        for line in authorized_keys:
            l2 = line.split()
            if len(l2) < 2:
                raise InvalidAuthentication(
                    'No algorithm description for key: %s' % line)
            try:
                key_data = credentials.key_data
                if base64.decodebytes(l2[1].encode('ascii')) != key_data:
                    continue
            except binascii.Error:
                raise InvalidAuthentication(
                    'Bad format for public key: %s' % line)

            return True

        raise CredentialsForbidden('Unauthorized SSH key.')

    def _validateSSLCertificate(self, credentials):
        """
        Check SSL credentials.
        """
        if not self._getOwnValueOrInherited(
                'allow_certificate_authentication'):
            raise CredentialsForbidden('Certificate authentication disabled.')

        if not credentials.certificate:
            raise CredentialsForbidden('No certificate was provided.')

        subject = credentials.certificate.get_subject()

        if not subject:
            raise CredentialsForbidden('Certificate without subject.')

        common_name = x509.get_common_name_from_subject(subject)

        if not common_name:
            raise CredentialsForbidden('Certificate without CN.')

        if self._me.name == common_name:
            return True
        else:
            raise CredentialsForbidden('Certificate CN not for this user.')

    def _validateAS2Anonymous(self, credentials):
        """
        AS2 credentials are valid when account/group allows AS2
        request don't require HTTP authentication.
        """
        auth_required = self._me.as2_require_http_authentication

        if is_inherited(auth_required):
            # FIXME:5480:
            # see if we should have this from multiple groups.
            # most probably only the primary group...but we need to document
            # it
            auth_required = self._groups[0].as2_require_http_authentication

        if auth_required:
            raise CredentialsForbidden(
                'AS2 messages without HTTP authentication are not allowed.')

        return True
