# Copyright (c) 2022 Adi Roiban.
# See LICENSE for details.
"""
Authentication via Entra ID HTTP API.

The OAuth2 requests are not using any 3rd party library.
The JSON Web Token are validates using the same library as lets-encrypt(josepy).

For securing info see RFC 6749 Security considerations
https://datatracker.ietf.org/doc/html/rfc6749#section-10

This uses OpenID connect with OAuth2 for authentication:
https://docs.microsoft.com/en-us/azure/active-directory/develop/v2-oauth2-auth-code-flow

OpenID docs:
https://login.microsoftonline.com/common/v2.0/.well-known/openid-configuration
https://docs.microsoft.com/en-us/azure/active-directory/develop/v2-protocols-oidc

Graph API for user configuration:
https://docs.microsoft.com/en-us/graph/api/user-list-memberof?view=graph-rest-1.0&tabs=http
"""
from __future__ import unicode_literals
from __future__ import absolute_import

from base64 import b64encode, urlsafe_b64encode
from binascii import hexlify
from random import choice
from string import ascii_letters
from time import time

from cryptography.hazmat.primitives import hashes
from josepy.jws import JWS
from lazr.delegates import delegate_to
from twisted.internet import defer
from twisted.python.randbytes import secureRandom
from six.moves.urllib.parse import urlencode
from six.moves.urllib.parse import parse_qs, parse_qsl, urlparse, urlunparse
from zope.interface import implementer

from chevah.server import json, force_unicode
from chevah.server.authentication.configuration import (
    _AuthenticationConfigurationBase,
    )
from chevah.server.commons.constant import (
    TYPE_NAME,
    )
from chevah.server.commons.credential import (
    OIDCCredentials,
    )
from chevah.server.commons.exception import (
    InvalidAuthentication,
    ServerException,
    )
from chevah.server.commons.interface import (
    IOIDCAuthentication,
    IApplicationAccountConfiguration,
    IEntraIDAuthenticationConfiguration,
    ILocalManagerCredentials,
    )
from chevah.server.commons.ssl import (
    SSLClientMixin,
    SSLConfigurationSectionMixin,
    )
from chevah.server.commons.runnable import Runnable
from chevah.server.configuration.model import (
    DefaultOnly,
    WritableLines,
    WritableString,
    WritableStringCaseInsensitive,
    WritableCSV,
    )
from chevah.server.configuration.option import PlainPasswordConfigurationMixin
from chevah.server.identity.model import (
    ExternalAccountConfigurationAbstract,
    ExternalAdministratorConfigurationAbstract,
    )
from chevah.server.proxy.http import proxify_http_connect_agent
from six.moves import range

# The API scopes that we ask after the authorization.
# This is not needed when only base groups are used.
# but for now we ask it all the time to simplify our code.
#
# We use GroupMember.Read.All as it should be enabled by default for
# conditional access and should give us access to GRAPH/memberOf
# https://learn.microsoft.com/en-us/azure/active-directory/conditional-access/concept-conditional-access-cloud-apps#all-cloud-apps
_GRAPH_SCOPES = ('GroupMember.Read.All',)
_BASE_SCOPES = (
    'openid',  # For id_token request
    'profile',  # For getting the preferred_username from the id_token
    ) + _GRAPH_SCOPES

# Number a second to allow for a authentication request to finalize.
# A big timeout is needed as the user might delay entering the credentials on
# Entra ID, for example MFA validation via a phone prompt.
_NONCE_TIMEOUT = 120
# The list of supported group association configuration.
_SUPPORTED_GROUP_ASSOCIATIONS = [
    TYPE_NAME.BASE_GROUPS,
    TYPE_NAME.BASE_AND_AZURE_GROUPS,
    ]


class EntraIDCredentials(OIDCCredentials):
    """
    Just a marker for OAuth2 credentials generated via Entra ID for an account.
    """


@delegate_to(ILocalManagerCredentials, context='_context')
@implementer(ILocalManagerCredentials)
class EntraIDAdministratorCredentials(object):
    """
    Credentials for the administrator, when authenticating via Entra ID.

    It is a normal admin credentials, with a few extra attributes that are
    only used as part of the Entra ID auth process.
    They should not be part of the generic interface.
    """
    def __init__(self, context):
        self._context = context

    @property
    def kind_name(self):
        return self.type

    @property
    def method(self):
        """
        The ID authentication method that created these credentials.
        """
        return self._context.method

    @property
    def token(self):
        """
        The OAuth token created during Entra ID authentication.
        """
        return self._context.token

    @token.setter
    def token(self, value):
        self._context.token = value

    @property
    def request(self):
        """
        The HTTP request from which the credentials were generated.
        """
        return self._context.request


@implementer(IApplicationAccountConfiguration)
class EntraIDAccountConfiguration(ExternalAccountConfigurationAbstract):
    """
    Account configuration for an Entra ID.
    """

    @property
    def kind_name(self):
        """
        See: `ExternalAccountConfigurationAbstract`.
        """
        return TYPE_NAME.ENTRA_ID


class EntraIDAdministratorConfiguration(
        ExternalAdministratorConfigurationAbstract):
    """
    Administrator configured using the Entra ID method.
    """

    @property
    def kind_name(self):
        """
        See: `ExternalAccountConfigurationAbstract`.
        """
        return TYPE_NAME.ENTRA_ID


@implementer(IEntraIDAuthenticationConfiguration)
class EntraIDAuthenticationConfiguration(
    _AuthenticationConfigurationBase,
    PlainPasswordConfigurationMixin,
    SSLConfigurationSectionMixin,
        ):
    """
    Configuration for Entra ID authentication.

    For SSL, we use default configuration and hardcoded to MS CA.
    """
    application_id = WritableString('application_id')
    directory_id = WritableString('directory_id')
    group_association = WritableStringCaseInsensitive(
        'group_association', default=TYPE_NAME.BASE_GROUPS)
    base_groups = WritableCSV('base_groups')
    base_roles = WritableCSV('base_roles')
    remove_username_suffix = WritableLines('remove_username_suffix')
    proxy = WritableString('proxy')
    api_scopes = WritableLines('api_scopes')
    ssl_certificate_authority = DefaultOnly('${MICROSOFT_IT_CA}')


@implementer(IOIDCAuthentication)
class EntraIDAuthentication(Runnable, SSLClientMixin):
    """
    Do authentication for accounts for which the password is provided
    by the operating system and which will be mapped to OS accounts.
    """
    PROPERTIES_REQUIRING_RESTART = (
        'directory_id',
        'proxy',
        )

    # Graph API to get group membership for current user.
    _URL_MEMBER_OF = 'https://graph.microsoft.com/v1.0/me/memberOf'

    # These are the URL markers that we use to detect that a request is for us.
    _response_argument = 'response-not-yet-initialized'
    _request_argument = 'request-not-yet-initialized'

    @defer.inlineCallbacks
    def _onStart(self, avatar=None):
        """
        See: Runnable.
        """
        from chevah.server.http.client import NonPersistentAgent

        self._resetState()

        if not self._configuration.directory_id:
            raise ServerException('"directory_id" configuration is requried.')

        if not self._configuration.application_id:
            raise ServerException('"application_id" configuration is requried.')

        if self._configuration.api_scopes and not self._configuration.password:
            raise ServerException(
                'The "password" configuration is required '
                'when extra "api_scopes" are configured.'
                )

        # Call it for validation.
        self._getGroupAssociation()

        # The URL argument used to trigger the OAuth2 authorization request.
        self._request_argument = 'oauth2-request-%s' % (self.uuid,)
        # The URL argument used to receive the OAuth2 authorization response.
        # This is what needs to be configured on the OAuth2 server,
        # So  we try to keep it simple and explicit.
        self._response_argument = 'redirect-%s' % (self.uuid,)

        ssl_context = yield self._initStartedContextFactory(remote_servers=[
            ('login.microsoftonline.com', 443),
            ('graph.microsoft.com', 443)
            ])

        self._agent = NonPersistentAgent(
            event_emitter=self.emitEvent,
            ssl_context=ssl_context.getContext(),
            )

        proxy = self.configuration.proxy
        if proxy:
            proxify_http_connect_agent(
                self._agent,
                proxy=proxy,
                context_factory=ssl_context,
                )

        try:
            openid_config = yield self._getOpenIDConfiguration()
        except Exception as error:
            raise ServerException(
                'Connect error at init: %s' % (force_unicode(error),))

        self._authorization_endpoint = openid_config['authorization_endpoint']
        self._token_endpoint = openid_config['token_endpoint']

    def _onStop(self, avatar=None):
        """
        See: Runnable.
        """
        self._resetState()

    def _resetState(self):
        """
        Reset the internal state of the auth method.
        """
        self._request_argument = 'entra-id-NOT-YET-INITIALIZED'
        self._response_argument = 'redirect-NOT-YET-INITIALIZED'
        self._authorization_endpoint = ''
        self._token_endpoint = ''
        self._agent = None
        self._nonces = {}

    @property
    def request_argument(self):
        """
        Return the URL parameter to trigger the OAuth2 for this method.
        """
        return self._request_argument

    def getAuthorizationURL(self, request):
        """
        Called by the login page to trigger the OAuth2 autorization request.

        Return an URL that can be used on the login page for the
        authorization code flow.

        Return `None` if the request is not for an authorization request.
        """
        if self._request_argument.encode('utf-8') not in request.uri:
            # This is not a request for us.
            return

        state, challenge, nonce, redirect_uri = self._makeNonce(request)
        # Get the initial URL
        sch, net, path, par, query, fra = urlparse(self._authorization_endpoint)

        scope = ' '.join(_BASE_SCOPES + self._configuration.api_scopes)

        # We default to the implicit grant flow as this will not require any
        # client secret.
        # The access is short lived, but we only need access at login time,
        # so we are fine.
        # https://docs.microsoft.com/en-us/azure/active-directory/develop/v2-oauth2-implicit-grant-flow
        response_type = 'id_token token'
        if self._configuration.api_scopes:
            # For extended API access we need to standard auth code flow.
            # https://docs.microsoft.com/en-us/azure/active-directory/develop/v2-oauth2-auth-code-flow
            # We still ask for the OpenID Connect id_token.
            response_type = 'id_token code'

        code_challenge = _get_code_challenge(challenge)

        # Extend the base URL with our extra query arguments.
        params = [
            ('response_type', response_type),
            ('client_id', self._configuration.application_id),
            ('redirect_uri', redirect_uri),
            ('scope', scope),
            # Receive the tokens as a POST request.
            # Is more secure as the code should not be visible in the URL.
            ('response_mode', 'form_post'),
            # Always ask to select the account.
            ('prompt', 'select_account'),
            ('code_challenge_method', 'S256'),
            ('state', state),
            ('nonce', nonce),
            ('code_challenge', code_challenge),
            ]
        queryparams = parse_qsl(query, keep_blank_values=True)
        queryparams.extend(params)
        query = urlencode(queryparams)

        return urlunparse((sch, net, path, par, query, fra))

    def getCredentials(self, request):
        """
        Called by the login page to retrieve the credentials for this request.

        This parses the response from the OIDC server and tries to get
        the username associated with this request.

        Try to do as fewer validatins here.
        The main goal is to extract the username.
        The actual validations will be done insde the `authenticate` method.

        Raise `ServerException` when failing to get the credentials.

        Return `None` if this is not an OIDC request.
        """
        if self._response_argument.encode('utf-8') not in request.uri:
            # Not for us.
            return None

        content_type = request.getHeader('content-type')
        if not content_type:
            raise ServerException(
                'Missing "content-type" header in OIDC response.')

        if content_type.lower() != TYPE_NAME.APPLICATION_FORM_URL_ENCODED:
            raise ServerException(
                'Unsupported "content-type" value in OIDC response. '
                '"{}".'.format(
                    content_type))

        # Since we request a `form_post` response, the value is in the body.
        request.content.seek(0)
        oidc_raw_response = request.content.read().decode('ascii')

        result = parse_qs(oidc_raw_response)
        error = result.get('error', [''])[0]
        if error:
            details = result.get(
                'error_description',
                ['Error received from Entra ID OIDC.']
                )[0]
            raise ServerException(
                'Error in Entra ID OIDC auth response. '
                '[{}] {}'.format(error, details))

        id_token_raw = result.get('id_token', [''])[0]
        if not id_token_raw:
            raise ServerException(
                'Missing "id_token" value from Entra ID OIDC server.')

        # Just extract the token without the signature validation.
        # The signature is validated later, when the credentials are validated.
        # Here we only want to get an username for the credentials.

        try:
            id_token = JWS.from_compact(id_token_raw.encode('ascii'))
            id_token_payload = json.loads(id_token.payload)
        except Exception:
            raise ServerException(
                'Invalid "id_token" value received from '
                'the Entra ID OIDC server.'
                )

        username = id_token_payload.get('preferred_username', '')
        if not username:
            raise ServerException(
                'Missing "preferred_username" in Entra ID OIDC response.')

        return EntraIDCredentials(
            method=self.uuid,
            username=username,
            request=request,
            # Here we hijack the token field and we temporary store the whole
            # OIDC response.
            # One option is to create an internal mapping between
            # EntraIDCredentials instance ID and the token.
            # Then in authenticate we can use the retrieved credentials
            # to obtain the token.
            # We will need some cleanp for this mapping.
            token=result,
            )

    def getAdministratorCredentials(self, request):
        """
        Called by the admin login page to retrieve the admin credentials for
        this request.
        """
        # The credentials are the same.
        # but we wrap them into an adapter to recognize them as being created
        # by admin service and not by the files service.
        user_credentials = self.getCredentials(request)
        if user_credentials is None:
            return None

        return EntraIDAdministratorCredentials(user_credentials)

    @defer.inlineCallbacks
    def authenticate(self, credentials):
        """
        See: `IAuthentication`.

        This is part2 of the authentication.

        By the time this is called, the user is already authenticated.
        We now need to check if the user is also authorized.

        It is called after the OAuth2 server has approved the application
        request and we were able to get the username.
        """
        if not isinstance(credentials, EntraIDCredentials):
            # We only handle Entra ID credentials.
            return

        if credentials.method != self.uuid:
            # We only handle our own credentials.
            return

        initial_token = credentials.token
        oauth_info = yield self._validateReceivedOICD(
            credentials, initial_token)

        groups = yield self._getGroups(oauth_info['authorize'])

        uuid, account_name = self._getUUIDAndName(credentials)

        # Just reset the token as a start to reduce the cases in which we
        # make the OAuth2 info available.
        credentials.token = None
        if self._configuration.api_scopes:
            # Update the credentials with the latest OAuth2 info if extra
            # API scopes are configured.
            # This means that SFTPPlus Entra ID is configured to work with
            # extensions.
            credentials.token = oauth_info

        account = EntraIDAccountConfiguration(
            parent=self.root.configuration.accounts,
            name=account_name,
            external={
                'group': [g.uuid for g in groups],
                'uuid': uuid,
                }
            )
        # Return the account and the full group configuration.
        defer.returnValue((account, groups))

    @defer.inlineCallbacks
    def authenticateAdministrator(self, credentials):
        """
        See: `IAuthentication`
        """
        if not isinstance(credentials, EntraIDAdministratorCredentials):
            # We only handle Entra ID admin credentials.
            return

        if credentials.method != self.uuid:
            # We only handle our own credentials.
            return

        # Retrieve the token from the intial credentials step.
        initial_token = credentials.token
        credentials.token = None

        yield self._validateReceivedOICD(credentials, initial_token)

        roles = self._getRoles(credentials)

        uuid, name = self._getUUIDAndName(credentials)

        administrator = EntraIDAdministratorConfiguration(
            parent=self.root.configuration.administrators,
            uuid=uuid,
            external={
                'name': name,
                'roles': tuple([r.uuid for r in roles]),
                }
            )
        # Return the admin and the full roles configuration.
        defer.returnValue((administrator, roles))

    def _getDiscoveryPage(self):
        """
        Here to help with testing to make it easier to inject a custom
        discovery response.
        """
        discover_url = (
            'https://login.microsoftonline.com/{}/v2.0/'
            '.well-known/openid-configuration'.format(
                self._configuration.directory_id))
        return self._agent.get(url=discover_url, headers={})

    @defer.inlineCallbacks
    def _getOpenIDConfiguration(self):
        """
        Return the valid OpenID configuration.
        """
        response = yield self._getDiscoveryPage()
        openid_config = yield self._agent.readBody(response)

        try:
            openid_config = json.loads(openid_config)
        except Exception:
            raise ServerException(
                'Invalid OpenID configuration format. %s' % (
                    force_unicode(openid_config[:100],)))

        if not isinstance(openid_config, dict):
            raise ServerException(
                'Bad OpenID configuration received. %r' % (
                    force_unicode(openid_config[:100]),)
                )

        error = openid_config.get('error', '')
        if error:
            details = openid_config.get(
                'error_description',
                'Failed to get Entra ID OpenID configuration.')
            raise ServerException(
                'Failed at OpenID discovery. [{}] {}'.format(error, details))

        if 'authorization_endpoint' not in openid_config:
            raise ServerException(
                'Failed at OpenID discovery. '
                'No "authorization_endpoint" found.')

        defer.returnValue(openid_config)

    def _makeNonce(self, request):
        """
        Return the OAuth2 random number as nonce value.

        This value is associated with a state and a challenge.

        The authorization request received by Azure ID is not signed in any way.
        Anyone can ask Azure for a request and redirect to us.

        To prevent replay attacks, unsolicited requests,
        and make sure the request originates from us we use the `nonce` which
        should be very hard for anyone to guess.
        And basically, if there wasn't any auth triggered via our app,
        no `nonce` value will be accepted.

        To prevent CSRF attacks, the `state` is used.

        It also cleans the older nonces.
        """
        now = time()
        nonce = hexlify(secureRandom(16)).decode('ascii')
        state = self._getState(request)
        # This needs to be 43 characters long.
        challenge = ''.join(choice(ascii_letters) for i in range(43))

        redirect_uri = self._getRedirectURL(request)

        self._nonces[nonce] = {
            'time': now,
            'state': state,
            'challenge': challenge,
            'redirect_uri': redirect_uri,
            }

        # Look for any older nonces and remove any expired ones.
        for key, value in list(self._nonces.items()):
            if now - value['time'] > _NONCE_TIMEOUT:
                del self._nonces[key]

        return state, challenge, nonce, redirect_uri

    def _getState(self, request):
        """
        Get the state associated with the request.

        This is to prevent CSRF:
        https://datatracker.ietf.org/doc/html/rfc6749#section-10.12
        """
        # We can't use cookies which is a pain.
        #
        # Use more than source IP for state.
        # Include user agent but might not work as some browser might have
        # random user agent for privacy.
        #
        # For Entra ID this is not a big security issue as Entra ID only
        # allows pre-approved redirect URLs.
        # Also, in our UI we use the actual username associated with the token
        # so if someone will manage to inject another token, the user should
        # see that username.
        # client.host is `bytes/str` on Py2 and `unicode/str` in Py3.
        state_seed = request.client.host.encode('utf-8')

        sha256_hasher = hashes.Hash(hashes.SHA256())
        sha256_hasher.update(state_seed)
        state = b64encode(sha256_hasher.finalize())
        return state.decode('ascii')

    def _getRedirectURL(self, request):
        """
        Return the URL where the OAuth2 server can redirect the response.

        We use the `request` as the same auth method can be used in multiple
        HTTP services.
        """
        # If the request has any referrer, use that one to help with
        # sites available via multiple names.
        # We assume users want to return to the same page, as the one used to
        # initiate the authentication.
        site_url = request.requestHeaders.getRawHeaders('referer')
        if site_url:
            # Twisted has support for multiple headers.
            # We only care about the first value.
            site_url = site_url[0]
        else:
            # If we don't have a referred, fallback to raw request URL.
            site_url = request.prePathURL().decode('utf-8')

        # We only care about the base URL, and not the URL query params.
        url_parts = urlparse(site_url)

        url_template = '{url.scheme}://{url.hostname}{path}/'
        if url_parts.port:
            url_template = '{url.scheme}://{url.hostname}:{url.port}{path}/'
        site_url = url_template.format(
            url=url_parts, path=request.site.base_url_path)
        return '{}?{}' .format(site_url, self._response_argument)

    def _getGroupAssociation(self):
        """
        Retrun and validate the current configured group association.
        """
        group_association = self._configuration.group_association
        if group_association not in _SUPPORTED_GROUP_ASSOCIATIONS:
            raise ServerException(
                'Unknown "group_association" configuration: "%s".' % (
                    group_association,))

        if group_association == TYPE_NAME.BASE_GROUPS:
            if not self._configuration.base_groups:
                raise ServerException(
                    'Missing value for "base_groups" configuration.')

        identity_groups = self.root.configuration.groups
        for group_uuid in self._configuration.base_groups:
            group = identity_groups.getSection(group_uuid)
            if not group:
                raise ServerException(
                    'Unknown group with UUID in "base_groups": "%s".' % (
                        group_uuid,))

        return group_association

    @defer.inlineCallbacks
    def _validateReceivedOICD(self, credentials, initial_token):
        """
        Make sure the received OpenID Connect request is valid.

        This is called as part of the `authenticate()` step, after we have
        received the response from OAuth server.

        Returns the acess_token for these credentials that can be used for the
        OAuth2 enabled MS API.
        """
        # This is the first token that we get from the Entra ID.
        # It might contain only the id token and auth code.
        # Later it might be extended with the access token.
        result = initial_token
        # This is our standard way of passing the original request via the
        # credentials for auth purposes.
        request = credentials.request

        state = result.get('state', [''])[0]
        if not state:
            raise InvalidAuthentication(
                'Missing "state" value from Entra ID server response.')

        id_token_raw = result.get('id_token', [''])[0]
        if not id_token_raw:
            raise InvalidAuthentication(
                'Missing "id_token" value from Entra ID server response.')

        # FIXME:5787:
        # Check signature for the id_token.
        try:
            id_token = JWS.from_compact(id_token_raw.encode('ascii'))
            id_token_payload = json.loads(id_token.payload)
        except Exception:
            raise InvalidAuthentication(
                'Invalid "id_token" value in Entra ID server response.')

        nonce = id_token_payload.get('nonce', '')
        if not nonce:
            raise InvalidAuthentication(
                'Missing "nonce" value from Entra ID server response.')

        expected_nonce = self._nonces.pop(nonce, None)
        if not expected_nonce:
            raise InvalidAuthentication(
                'Unknown nonce received. '
                'Possible replay attack or unsolicited request.'
                )

        if (time() - expected_nonce['time']) > 60:
            raise InvalidAuthentication('Late response from Entra ID server.')

        if state != expected_nonce['state']:
            raise InvalidAuthentication(
                'Missmatched stored "state". Possible CSRF attack.')

        if state != self._getState(request):
            raise InvalidAuthentication(
                'Missmatched received "state". Possible CSRF attack.')

        redirect_uri = expected_nonce['redirect_uri']

        auth_code = result.get('code', [''])[0]
        if auth_code:
            # We have the explicit auth code flow.
            # So we need to request the access_token
            result = yield self._redeemGraphAccessCode(
                auth_code,
                expected_nonce['challenge'],
                redirect_uri,
                request,
                )
        else:
            # In implict flow we can have multiple access token.
            # We only need the first one.
            # We rewrite the result to have single values to make the
            # result compatible with the auth code flow.
            result['access_token'] = result.get('access_token', [''])[0]
            result['token_type'] = result.get('token_type', [''])[0]

        token_type = result.get('token_type', '')
        if not token_type:
            raise InvalidAuthentication(
                'Missing "token_type" value from Entra ID server response.')

        # FIXME:5787:
        # Check signature for the access_token.
        # We know that access_token already exists and is valid JSON.
        access_token_raw = result['access_token']
        access_token = JWS.from_compact(access_token_raw.encode('ascii'))
        access_token_payload = json.loads(access_token.payload)
        appid = access_token_payload.get('appid', 'no appid')
        if appid != self._configuration.application_id:
            raise InvalidAuthentication(
                'Entra ID server response for another appid: "{}".'.format(
                    appid))

        # This is the information that we make available to extensions.
        defer.returnValue({
            'authorize': result,
            'directory_id': self._configuration.directory_id,
            'client_id': self._configuration.application_id,
            'client_secret': self._configuration.password,
            'redirect_uri': redirect_uri,
            })

    @defer.inlineCallbacks
    def _getGroups(self, token):
        """
        Return the list of groups for credentials.

        It fails if not groups could be retrieved.
        """
        groups = yield self._getAssociatedGroups(token)

        if not groups:
            raise InvalidAuthentication(
                'Not a member of any of the allowed groups.')

        defer.returnValue(groups)

    def _getAssociatedGroups(self, token):
        """
        Return the list of groups associated for authenticated account.
        """
        group_association = self._getGroupAssociation()

        base_groups = self._getBaseGroups()

        if group_association == TYPE_NAME.BASE_GROUPS:
            # Only the base groups.
            return defer.succeed(base_groups)

        # The group_association is already validated.
        # If we don't have any other group_association option it means that
        # we have Entra ID groups.
        return self._getEntraIDGroups(base_groups, token)

    def _getBaseGroups(self):
        """
        Return the list of base group configurations for this method.
        """
        identity_groups = self.root.configuration.groups
        groups = []
        for group_uuid in self._configuration.base_groups:
            group = identity_groups.getSection(group_uuid)
            if not group:  # noqa:cover
                # This should not happen as we have previously validated then.
                raise RuntimeError('Unknown group for Entra ID.')
            groups.append(group)

        return groups

    @defer.inlineCallbacks
    def _redeemGraphAccessCode(self, code, challenge, redirect_uri, request):
        """
        Return an access code to the Graph API.
        """
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'Accept': 'application/json',
            }

        token_params = {
            'client_id': self._configuration.application_id,
            'code_verifier': challenge,
            'client_secret': self._configuration.password,
            'code': code,
            'grant_type': 'authorization_code',
            'scope': ' '.join(_GRAPH_SCOPES),
            'redirect_uri': redirect_uri,
            }

        response = yield self._agent.post(
            url=self._token_endpoint,
            headers=headers,
            body=urlencode(token_params).encode("utf-8")
            )
        body = yield self._agent.readBody(response)

        try:
            result = json.loads(body)
        except Exception:
            raise InvalidAuthentication(
                'Failed to get a valid access_token response. %s' % (
                    force_unicode(body, trim=100),))

        error = result.get('error', [''])[0]
        if error:
            details = result.get(
                'error_description',
                'Error received from access_token request.',
                )
            raise InvalidAuthentication(
                'Error in Entra ID access_token response. '
                '[{}] {}'.format(error, details))

        defer.returnValue(result)

    @defer.inlineCallbacks
    def _getEntraIDGroups(self, base_groups, token):
        """
        Return the list of groups associated with this account in Entra ID,
        adding them to the `base_groups`.
        """
        authorization = '{} {}'.format(
            token.get('token_type'),
            token.get('access_token'),
            )
        headers = {'authorization': authorization}
        response = yield self._agent.get(
            url=self._URL_MEMBER_OF,
            headers=headers,
            )
        body = yield self._agent.readBody(response)

        try:
            result = json.loads(body)
        except Exception:
            raise InvalidAuthentication(
                'Failed to get a valid list of Entra ID groups. %s' % (
                    force_unicode(body, trim=100),))

        try:
            azure_groups = result['value']
        except Exception:
            raise InvalidAuthentication('No groups found in Entra ID. %s' % (
                force_unicode(body, trim=100),))

        identity_groups = self.root.configuration.groups
        groups = base_groups[:]
        # This code is not 100% solid, as MS API might not return a list of
        # objects.
        # But we don't do any JSON schema validation now.
        # Any errors should trigger an ISE error and we can look at it later.
        for ad_group in azure_groups:
            display_name = ad_group.get('displayName', '')
            if not display_name:
                raise InvalidAuthentication('No name found for Entra ID group.')

            group = identity_groups.getSectionByName(display_name)
            if not group:
                # Entra ID group not found inside SFTPPlus.
                # Just ignore it.
                continue

            if group in groups:
                # Entra ID group already part of base groups.
                continue

            groups.append(group)

        defer.returnValue(groups)

    def _getUUIDAndName(self, credentials):
        """
        Return the internal account name associate with the Entra ID account.
        """
        uuid = credentials.username
        name = uuid
        for suffix in self._configuration.remove_username_suffix:
            if name.endswith(suffix):
                name = name[:len(suffix) * -1]
                break

        return uuid, name

    def _getRoles(self, credentials):
        """
        Return the list of roles for credentials.

        It fails if no roles could be retrieved.
        """
        # For now we have a fixed set of roles.
        # We don't get a dynamic set of roles based on Entra ID.
        # This might be added later, to match the account authentication.
        roles = self._getBaseRoles()
        if not roles:
            raise InvalidAuthentication(
                'Not a member of any of the allowed roles.')

        return roles

    def _getBaseRoles(self):
        """
        Return the list of base roles configurations for this method.
        """
        all_roles = self.root.configuration.roles
        roles = []
        for role_uuid in self._configuration.base_roles:
            role = all_roles.getSection(role_uuid)
            if not role:
                raise InvalidAuthentication(
                    'Unknown role configured for Entra ID.')
            roles.append(role)

        return roles


def _get_code_challenge(plain_text_code):
    """
    Return the hashed challenge used in the OAuth2 authorized request.
    """
    sha256_hasher = hashes.Hash(hashes.SHA256())
    sha256_hasher.update(plain_text_code.encode('ascii'))
    result = urlsafe_b64encode(sha256_hasher.finalize())
    return result.replace(b'=', b'')
