# Copyright (c) 2014 Adi Roiban.
# See LICENSE for details.
"""
Authentication done over HTTP or HTTPS.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

from twisted.internet import defer
from twisted.web import http
from zope.interface import implementer

from chevah.server import force_unicode, json
from chevah.server.authentication.configuration import (
    _AuthenticationConfigurationBase,
    )
from chevah.server.commons.constant import ADDRESS_UNKNOWN, TYPE_NAME
from chevah.server.commons.credential import PasswordCredentials
from chevah.server.commons.exception import (
    CredentialsForbidden,
    InvalidAuthentication,
    ServerException,
    )
from chevah.server.configuration.option import (
    UsernamePasswordPublicOptionsMixin,
    )
from chevah.server.configuration.model import (
    WritableBoolean,
    WritableLines,
    WritableNumber,
    WritableCSVUnique,
    WritableString,
    )
from chevah.server.commons.interface import (
    IAuthentication,
    IApplicationAccountConfiguration,
    IHTTPAuthenticationConfiguration,
    )
from chevah.server.commons.ssl import (
    SSLClientMixin,
    SSLConfigurationSectionMixin,
    )
from chevah.server.commons.runnable import Runnable
from chevah.server.http.client import (
    extract_configured_header,
    RedundantAgent,
    )
from chevah.server.identity.model import (
    ExternalAccountConfigurationAbstract,
    )
from chevah.server.proxy.http import proxify_http_connect_agent
import six


@implementer(IHTTPAuthenticationConfiguration)
class HTTPAuthenticationConfiguration(
        _AuthenticationConfigurationBase,
        SSLConfigurationSectionMixin,
        UsernamePasswordPublicOptionsMixin,
        ):
    """
    Configuration for a HTTP authentication.
    """

    url = WritableCSVUnique('url')
    proxy = WritableString('proxy')
    headers = WritableLines('headers')
    test_at_start = WritableBoolean('test_at_start', default=True)
    timeout = WritableNumber(
        'timeout', default=120, allow_negative=False, allow_zero=False)


@implementer(IApplicationAccountConfiguration)
class HTTPAccountConfiguration(ExternalAccountConfigurationAbstract):
    """
    Account configured using HTTP request.
    """

    @property
    def kind_name(self):
        """
        See: `ExternalAccountConfigurationAbstract`.
        """
        return TYPE_NAME.HTTP


@implementer(IAuthentication)
class HTTPAuthentication(Runnable, SSLClientMixin):
    """
    Do authentication over HTTP.
    """
    PROPERTIES_REQUIRING_RESTART = (
        SSLConfigurationSectionMixin.PROPERTIES_REQUIRING_RESTART +
        (
            'proxy',
            'url',
            'username',
            'password_update',
            'timeout',
            ))

    # Keys expected in the root of response.
    _root_keys = {'account'}

    def __init__(self, parent, configuration):
        super(HTTPAuthentication, self).__init__(parent, configuration)
        self._agent = None
        self._server_public_identity = self.root.public_identity
        self._urls = None
        self._headers = {}

    def _onStart(self, avatar=None):
        """
        See: Runnable.
        """
        for line in self._configuration.headers:
            key, value = extract_configured_header(line)
            self._headers[key] = [value]

        self._urls = self.configuration.url
        deferred = self._initStartedContextFactory(urls=self._urls)

        def eb_got_response(failure):
            """
            Called when failing to get any kind of response.
            """
            # It is converted into a InvalidAuthentication to allow general
            # authentication exception handling to decide what to do.
            message = 'Connect error at init: %s' % (
                force_unicode(failure.value),)
            raise ServerException(message)

        def cb_got_context_factory(result):
            """
            Called once the client SSL context was set up.
            """
            agent_credentials = PasswordCredentials(
                username=self._configuration.username,
                password=self._configuration.password,
                peer=ADDRESS_UNKNOWN,
                creator=self,
                )
            self._agent = RedundantAgent(
                urls=self._urls,
                event_emitter=self.emitEvent,
                ssl_context=result.getContext(),
                credentials=agent_credentials,
                timeout=self._configuration.timeout,
                )

            proxy = self.configuration.proxy
            if proxy:
                proxify_http_connect_agent(
                    self._agent,
                    proxy=proxy,
                    context_factory=self._context_factory,
                    )

            if not self.configuration.test_at_start:
                # No need to test the remote server.
                return

            payload_credentials = PasswordCredentials(
                'test-pass',
                username='test-init',
                peer=ADDRESS_UNKNOWN,
                creator=self,
                )
            request = {
                'credentials': payload_credentials,
                'server': self._server_public_identity
                }

            deferred = self._agent.postJSONWithRedundancy(
                body=request, headers=self._headers)

            deferred.addErrback(eb_got_response)
            return deferred

        deferred.addCallback(cb_got_context_factory)
        return deferred

    def _onStop(self, avatar=None):
        """
        See: Runnable.
        """
        self._urls = None
        self._headers = {}
        if not self._agent:
            # Somehow the start fails before creating the agent.
            return
        return self._agent.closePersistentConnections()

    def authenticate(self, credentials):
        """
        See: `IAuthentication`
        """

        def eb_got_response(failure):
            """
            Called when failing to get any kind of response.
            """
            # It is converted into a InvalidAuthentication to allow general
            # authentication exception handling to decide what to do.
            message = 'Connect error: %s' % force_unicode(failure.value,)
            raise InvalidAuthentication(message)

        @defer.inlineCallbacks
        def cb_got_response(response):
            """
            Called when request response was received.
            """
            try:
                body = yield self._agent.readBody(response)
            except Exception as error:
                """
                Error while reading body.

                The server might have closed the connection while reading or
                server headers advertised the content length, but not
                all data was received, or the body deliver timed out.
                """
                # The error details will look ugly but I don't know a better
                # "cost effective" method to show the error details.
                raise InvalidAuthentication(
                    'Failure while receiving response body. %s' % (
                        force_unicode(error)))

            if response.code == http.UNAUTHORIZED:
                # Not authorized, but might be authorized by other methods.
                defer.returnValue(None)

            if response.code == http.FORBIDDEN:
                # Not authorized, stop looking for other methods.
                body = force_unicode(body)
                message = 'HTTP %s %s' % (
                    force_unicode(response.code),
                    force_unicode(response.phrase),
                    )
                public_response = {}
                try:
                    body = json.loads(body)
                    # See if we have a more detailed error message.
                    # Of keep the current message.
                    message = body.get('message', message)
                    # See if we have a custom public response,
                    # or default to empty.
                    public_response = body.get('public_response', {})
                except Exception:
                    """
                    Maybe not a JSON body.
                    This is ok. We will just keep the body as text.
                    """

                response = {
                    'code': force_unicode(response.code),
                    'phrase': force_unicode(response.phrase),
                    'message': message,
                    'body': body,
                    'public_response': public_response,
                    }
                raise CredentialsForbidden(message, response=response)

            if response.code == http.OK:
                # Authorized with configuration.
                result = yield self._gotBody(credentials, body)
                defer.returnValue(result)

            elif response.code == http.NO_CONTENT:
                # Authorized without explicit configuration.
                result = yield self._gotBody(credentials, None)
                defer.returnValue(result)

            # Unknown response.
            raise InvalidAuthentication('Unknown response %s:%s' % (
                response.code, force_unicode(response.phrase)))

        request = {
            'credentials': credentials,
            'server': self._server_public_identity
            }

        deferred = self._agent.postJSONWithRedundancy(
            body=request, headers=self._headers)

        deferred.addCallbacks(cb_got_response, eb_got_response)
        return deferred

    def _gotBody(self, credentials, result):
        """
        Called when the body of the request was received.
        """
        if result is None:
            # Short response without content.
            external = {}
        else:
            # Content expected.
            if not result:
                raise InvalidAuthentication('No response body.')

            try:
                body = json.loads(result)
            except ValueError:
                raise InvalidAuthentication('Invalid JSON body.')

            root_keys = list(body.keys())

            if not root_keys:
                raise InvalidAuthentication('Empty JSON response.')

            if set(root_keys) != self._root_keys:
                raise InvalidAuthentication(
                    'Invalid root keys: %s.' % ', '.join(root_keys))
            external = body['account']

        # In version prior to 4.23.0 `home_folder_structure` was a list of list.
        # Now it should be a list of strings.
        # Here we add backward compatibility for older APIs.
        legacy_home_folder_structure = external.get('home_folder_structure', [])
        home_folder_structure = []
        for member in legacy_home_folder_structure:
            if isinstance(member, six.string_types):
                home_folder_structure.append(member)
                continue
            home_folder_structure.append(member[0])

        if home_folder_structure:
            # Update only if we had a structure in the original message.
            external['home_folder_structure'] = home_folder_structure

        account_configuration = HTTPAccountConfiguration(
            parent=self.root.configuration.accounts,
            name=credentials.username,
            external=external,
            )
        return account_configuration, ()

    def authenticateAdministrator(self, credentials):
        """
        See: `IAuthentication`
        """
        return defer.fail(InvalidAuthentication(
            'Administrators are not yet supported.'))
