# Copyright (c) 2015 Adi Roiban.
# See LICENSE for details.
"""
Authentication which ban source IP after a number of failures.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

from zope.interface import implementer

from twisted.internet.task import LoopingCall

from chevah.server.commons.exception import CredentialsForbidden
from chevah.server.commons.interface import IAuthentication
from chevah.server.commons.runnable import Runnable


@implementer(IAuthentication)
class IPTimeBanAuthentication(Runnable):
    """
    Simple authentication which bans source IP after a number of consecutive
    failures.
    """
    PROPERTIES_REQUIRING_RESTART = ()

    # Dict with mapping of ip -> failure_count, ban_lift_timestamp
    _failures = None
    # Loop used to clean the failures memory.
    _clear_loop = None

    def _onStart(self, avatar=None):
        """
        See: Runnable.
        """
        self._failures = {}
        self._clear_loop = LoopingCall(self._lpClearFailures)
        self._clear_loop.clock = self._scheduler
        # Run the cleanup every 10 minutes.
        self._clear_loop.start(10 * 60, now=False)

    def _onStop(self, avatar=None):
        """
        See: Runnable.
        """
        # Maybe we failed to start so we don't have yet a loop.
        if self._clear_loop:
            self._clear_loop.stop()

        self._clear_loop = None
        self._failures = {}

    def authenticate(self, credentials):
        """
        See: `IAuthentication`
        """
        if self._shouldReject(credentials):
            raise CredentialsForbidden('Source IP failed for too many times.')

        # Credentials are subscribed last to avoid calling the reject hook
        # if we have already rejected them.
        credentials.subscribe('accepted', self._onAccepted)
        credentials.subscribe('rejected', self._onRejected)
        return None

    def authenticateAdministrator(self, credentials):
        """
        See: `IAuthentication`
        """
        return self.authenticate(credentials)

    def _shouldReject(self, credentials):
        """
        Check if the credentials should be allowed.
        """
        now = self._scheduler.seconds()
        host = credentials.peer.host
        suspect = self._failures.get(host, None)
        if not suspect:
            # Source IP has no recent failures.
            return False

        count, expire = suspect

        if expire < now:
            # Source IP ban has expired.
            return False

        if count < self.configuration.ban_after_count:
            # Source IP can still fail more.
            return False

        # Refresh the ban.
        ban_lift_timestamp = now + self.configuration.ban_interval
        self._failures[host] = (count, ban_lift_timestamp)

        return True

    def _onAccepted(self, signal):
        """
        Called when credentials were accepted.
        """
        credentials = signal.source
        self._unsubscribe(credentials)
        # Removes the cached failures... if any.
        self._failures.pop(credentials.peer.host, None)

    def _onRejected(self, signal):
        """
        Called when credentials were rejected.
        """
        credentials = signal.source
        host = credentials.peer.host

        self._unsubscribe(credentials)

        suspect = self._failures.get(host, None)
        if suspect:
            failure_count = suspect[0]
        else:
            failure_count = 0

        failure_count += 1

        ban_lift_timestamp = (
            self._scheduler.seconds() + self.configuration.ban_interval)

        # Record the number of failed attempts and when the ban can be lifted.
        self._failures[host] = (failure_count, ban_lift_timestamp)

    def _lpClearFailures(self):
        """
        Called in a loop to clean failures from memory.
        """
        now = self._scheduler.seconds()
        unexpired_failures = {}
        for host, memory in self._failures.items():
            count, ban_lift = memory

            if ban_lift <= now:
                continue

            unexpired_failures[host] = (count, ban_lift)

        self._failures = unexpired_failures

    def _unsubscribe(self, credentials):
        """
        Unsubscribe our hooks from the credentials.
        """
        credentials.unsubscribe('accepted', self._onAccepted)
        credentials.unsubscribe('rejected', self._onRejected)
