# -*- test-case-name: normal.authentication.test_ldap -*-
# Copyright (c) 2016 Adi Roiban.
# See LICENSE for details.
"""
Authentication based on a remote LDAP server.
"""
from __future__ import absolute_import, unicode_literals

from ldaptor.protocols.ldap.ldapclient import LDAPClient
from ldaptor.protocols.ldap.ldapsyntax import LDAPEntry
from ldaptor.protocols.ldap.ldaperrors import LDAPInvalidCredentials
from ldaptor.protocols.pureldap import (
    LDAP_SCOPE_baseObject,
    LDAP_SCOPE_wholeSubtree,
    )
from OpenSSL import SSL
from twisted.internet import defer
from twisted.internet.endpoints import (
    connectProtocol,
    SSL4ClientEndpoint,
    TCP4ClientEndpoint,
    )
from twisted.internet.error import ConnectError, ConnectionClosed
from zope.interface import implementer

from chevah.server import force_unicode
from chevah.server.authentication.base import (
    apply_group_mapping,
    check_password_credentials_validity,
    )
from chevah.server.authentication.configuration import (
    _AuthenticationConfigurationBase,
    )
from chevah.server.commons.constant import (
    CONNECT_TIMEOUT,
    BIND_DN_TYPE_NAME,
    TYPE_NAME,
    )
from chevah.server.commons.exception import (
    CredentialsForbidden,
    InvalidAuthentication,
    ServerException,
    )
from chevah.server.commons.interface import (
    IApplicationAccountConfiguration,
    IAuthentication,
    ILDAPAuthenticationConfiguration,
    )
from chevah.server.commons.match import get_matched_groups
from chevah.server.commons.runnable import Runnable
from chevah.server.commons.public_format import public_format
from chevah.server.commons.ssl import (
    SSLClientMixin,
    SSLConfigurationSectionMixin,
    )
from chevah.server.configuration.model import (
    WritableBoolean,
    WritableLines,
    WritableStringCaseInsensitive,
    WritableCSVLinesUnique,
    WritableCSVUnique,
    WritableString,
    )
from chevah.server.configuration.option import (
    AddressPortOptionsMixin,
    )
from chevah.server.extension import load_extension_object
from chevah.server.extension.auth_ldap_noop import AuthLDAPNoop
from chevah.server.identity.model import (
    ExternalAccountConfigurationAbstract,
    ExternalAdministratorConfigurationAbstract,
    )


@implementer(IApplicationAccountConfiguration)
class LDAPApplicationAccountConfiguration(
        ExternalAccountConfigurationAbstract):
    """
    Account configured using the LDAP method.
    """

    @property
    def kind_name(self):
        """
        See: `ExternalAccountConfigurationAbstract`.
        """
        return TYPE_NAME.LDAP


class LDAPApplicationAdministratorConfiguration(
        ExternalAdministratorConfigurationAbstract):
    """
    Administrator configured using the LDAP method.
    """

    @property
    def kind_name(self):
        """
        See: `ExternalAccountConfigurationAbstract`.
        """
        return TYPE_NAME.LDAP + '-administrator'


@implementer(ILDAPAuthenticationConfiguration)
class LDAPAuthenticationConfiguration(
        _AuthenticationConfigurationBase,
        AddressPortOptionsMixin,
        SSLConfigurationSectionMixin,
        ):
    """
    Configuration for the LDAP authentication.
    """

    PORT_DEFAULT = 389

    secure_connection = WritableBoolean('secure_connection', default=False)
    bind_dn_type = WritableStringCaseInsensitive('bind_dn_type')
    # This is called bind_dn for legacy reasons.
    # It's more of a list of search base DNs.
    bind_dn = WritableLines('bind_dn', default=())
    home_folder_attribute = WritableCSVUnique('home_folder_attribute')
    email_attribute = WritableString('email_attribute', default='')
    username_attribute = WritableString('username_attribute', default='cn')
    extension_entry_point = WritableString(
        'extension_entry_point', default='')
    extension_configuration = WritableString(
        'extension_configuration', default='')
    username_suffix = WritableString('username_suffix')
    search_filter = WritableString('search_filter')
    manager_search_filter = WritableString('manager_search_filter')
    group_mapping = WritableCSVLinesUnique('group_mapping')


@implementer(IAuthentication)
class LDAPAuthentication(Runnable, SSLClientMixin):
    """
    Authenticate against a remote LDAP server.

    This is not using persistent LDAP connections.
    Each authentication will create a new LDAP connection.
    """

    PROPERTIES_REQUIRING_RESTART = (
        SSLConfigurationSectionMixin.PROPERTIES_REQUIRING_RESTART + (
            'bind_dn',
            'bind_dn_type',
            'extension_entry_point',
            'group_mapping',
            )
        )

    _bind_dn_type = None
    # List of base DN in which to try the authentication.
    _base_dns = ()
    # List of list of group mapping definition. Already validated.
    _group_mapping = None
    # Set of LDAP non-duplicate attributes used in defining the group mapping.
    _group_mapping_attributes = None
    # TLS context factory.
    _context_factory = None

    # Reference to the extension object
    _extension = None

    def _onStart(self, avatar=None):
        """
        See: Runnable.
        """
        # Configuration can be changed at any time, and at start we
        # don't have a password, but at least try to connect to check
        # that the server is available and do a simple configuration
        # validation. In this way we try to prevent late errors raised
        # due to miss-configuration.
        self._setConfiguration()

        def eb_anonymous_bind(failure):
            """
            Called after the anonymous bind failed.
            """
            if failure.check(LDAPInvalidCredentials):
                # It's OK to get the BIND request rejected during startup,
                # as we are trying an anonymous bind.
                return

            if failure.check(ConnectionClosed):
                # Connection close (done or lost) during initial bind exchange.
                # Maybe not an LDAP server.
                server_type = 'LDAP'
                if self._configuration.secure_connection:
                    server_type = 'LDAPS'

                raise ServerException(
                    'Disconnected during the initial LDAP message. '
                    'Maybe not an %s server.' % (server_type,)
                    )

            failure.trap(SSL.Error)
            raise ServerException('Invalid TLS configuration. %s' % (
                force_unicode(failure.value)),)

        def on_connect(client):
            """
            Called after TCP connection is made.
            """
            # FIXME:6505:
            # Implement handshake interface.
            # With the updated, this should be called only after the handshake.
            #
            # This is called before the handshake is done.
            # Do an anonymous BIND request to trigger some traffic
            # that will finalize the handshake and catch any
            # initial TLS errors.
            deferred = client.bind(b'', b'')
            deferred.addErrback(eb_anonymous_bind)
            return deferred

        # We try to connect but the remote server might not be LDAP.
        # It can behave in many ways.
        # Some may close the connection when they receive LDAP data
        # which is not their known protocol.
        # Others might wait for more data to be sent by us.
        # This would cause the LDAP to stall on start.
        return self._connect(on_connect=on_connect)

    def _onStop(self, avatar=None):
        """
        See: Runnable.
        """
        # We conditionally call stop as the start might fail before the
        # extension is created.
        if self._extension:
            self._extension.stop()

    @defer.inlineCallbacks
    def authenticate(self, credentials):
        """
        See: `IAuthentication`
        """
        name = credentials.username
        home_folder_attribute = self.configuration.home_folder_attribute
        email_attribute = self.configuration.email_attribute

        yield check_password_credentials_validity('LDAP', credentials)

        search_filter = self.configuration.search_filter.strip()

        attributes = self._group_mapping_attributes[:]
        if home_folder_attribute:
            attributes.append(home_folder_attribute[0])
        if email_attribute:
            attributes.append(email_attribute)

        last_error = None
        for base_dn in self._base_dns:

            try:
                entry = yield self._getLDAPEntry(
                    credentials=credentials,
                    search_filter=search_filter,
                    attributes=attributes,
                    base_dn=base_dn,
                    )
                result = yield self._getAccount(
                    entry, name, home_folder_attribute, email_attribute)
                defer.returnValue(result)
            except Exception as error:
                last_error = error
                # Try the next base DN.
                continue

        # We know from the initial configuration validation that we have
        # at least one base_dn.
        # So if we are here, it means that the authentication failed.
        self._raiseConvertedError(last_error)

    @defer.inlineCallbacks
    def authenticateAdministrator(self, credentials):
        """
        See: `IAuthentication`
        """
        search_filter = self.configuration.manager_search_filter
        if not search_filter:
            raise InvalidAuthentication(
                'No configured "manager_search_filter".')

        yield check_password_credentials_validity('LDAP', credentials)

        last_error = None
        for base_dn in self._base_dns:

            try:
                yield self._getLDAPEntry(
                    credentials=credentials,
                    search_filter=search_filter,
                    attributes=[],
                    base_dn=base_dn,
                    )
                # For administrator we don't use any LDAP value,
                # only check that it can be authenticated.
                administrator = LDAPApplicationAdministratorConfiguration(
                    parent=self.root.configuration.administrators,
                    uuid=credentials.username,
                    )
                # In the future, we might have multiple roles for
                # administrators authenticated via LDAP.
                role = self.root.configuration.roles.getSection(
                    administrator.roles[0])
                defer.returnValue((administrator, (role,)))

            except Exception as error:
                last_error = error
                # Try the next base DN.
                continue

        # We know from the initial configuration validation that we have
        # at least one base_dn.
        # So if we are here, it means that the authentication failed.
        self._raiseConvertedError(last_error)

    def _raiseConvertedError(self, error):
        """
        Called as the last error handling for the authentication process.
        """
        if isinstance(error, (CredentialsForbidden, InvalidAuthentication)):
            raise error

        if isinstance(error, ServerException):
            raise InvalidAuthentication(error.message)

        # Wrap any error into an InvalidAuthentication... but this
        # is more like an ISE.
        raise InvalidAuthentication(
            'LDAP authentication failed with an internal error. %s' % (
                force_unicode(error)))

    def _getBindDN(self, credentials, base_dn):
        """
        Return the fully resolved DN for `credentials` to be used for the
        BIND operation.
        """
        if (
            self._bind_dn_type == TYPE_NAME.ABSOLUTE
            or self._bind_dn_type == TYPE_NAME.DIRECT_USERNAME
                ):
            return credentials.username + self.configuration.username_suffix

        if self._bind_dn_type == TYPE_NAME.RELATIVE:
            return '%s%s,%s' % (
                credentials.username,
                self.configuration.username_suffix,
                base_dn,
                )

        if self._bind_dn_type == TYPE_NAME.PARENT:
            return '%s=%s%s,%s' % (
                self.configuration.username_attribute,
                credentials.username,
                self.configuration.username_suffix,
                base_dn,
                )

    def _getLDAPEntry(self, credentials, search_filter, attributes, base_dn):
        """
        Return the authenticated LDAP entry for credentials.
        """
        if not search_filter:
            search_filter = None

        self._extension.updateCredentials(credentials)
        extension_attibutes = self._extension.getExtraAttributes()

        bind_dn = self._getBindDN(credentials, base_dn)

        # objectclass is always added so that we will always get some
        # results.
        attributes = ['objectclass'] + extension_attibutes + attributes

        def on_connect(client):
            """
            Called when we are connected.
            """
            deferred = self._bind(client, bind_dn, credentials.password)
            deferred.addCallback(cb_bind, client)
            deferred.addCallback(
                self._extension.augmentEntry,
                credentials, client,
                )
            return deferred

        def cb_bind(ignored, client):
            """
            Called when connection was authenticated.
            """
            return self._searchLDAPEntry(
                client, base_dn, bind_dn, attributes, search_filter)

        return self._connect(on_connect=on_connect)

    @defer.inlineCallbacks
    def _searchLDAPEntry(
            self, client, base_dn, bind_dn, attributes, search_filter):
        """
        Return the LDAP entry matched by `search_filter`.
        """
        if self._bind_dn_type == TYPE_NAME.DIRECT_USERNAME:
            # We need to filter based on the provided username.
            username = bind_dn
            username_attribute = self.configuration.username_attribute.lower()
            if (username_attribute == 'samaccountname') and ('\\' in bind_dn):
                username = bind_dn.split('\\', 1)[1]
            username_filter = '(%s=%s)' % (
                self.configuration.username_attribute, username)
            # Since we don't know the DN for the account,
            # we will get the account configuration from the base DN.
            scope = LDAP_SCOPE_wholeSubtree
        else:
            # No need to filter for username as we know the exact DN.
            username_filter = None
            # Try to get the account's configuration from the same LDAP
            # entry as the one used for the bind.
            base_dn = bind_dn
            scope = LDAP_SCOPE_baseObject

        o = LDAPEntry(client, base_dn)

        if username_filter:
            if search_filter:
                # We need to amend the base filter.
                search_filter = '(&%s%s)' % (username_filter, search_filter)
            else:
                # Just use the username.
                search_filter = username_filter

        # Serialized the input for LDAP
        if search_filter:
            search_filter = search_filter
        else:
            search_filter = None

        try:
            results = yield o.search(
                attributes=attributes,
                scope=scope,
                filterText=search_filter,
                )
        except Exception as error:
            # Under normal operations we should not hit this branch as
            # the base_dn is the same base_dn and since bind operation
            # succeed, we should have a base_dn.
            raise InvalidAuthentication(
                'LDAP failed to get entry for "%s": %s' % (
                    base_dn, force_unicode(error)))

        if len(results) != 1:
            # Most probably we can get 0 results, hence the user is filtered.
            # It is unlikely that we got more than 1 results as it means that
            # LDAP has duplicate users.
            raise CredentialsForbidden(
                'LDAP has returned %d entries for "%s" at %s with '
                'search filter %s.' % (
                    len(results), attributes, base_dn, search_filter))

        defer.returnValue(results[0])

    def _getAccount(self, entry, name, home_folder_attribute, email_attribute):
        """
        Called when we got the LDAP entry.

        Return the LDAP account configuration from the LDAP entry.
        """
        external = {}
        self._resolveHomeFolderPath(entry, home_folder_attribute, external)
        external['email'] = self._resolveEmail(entry, email_attribute)
        external = apply_group_mapping(self._group_mapping, external, entry)

        account = LDAPApplicationAccountConfiguration(
            parent=self.root.configuration.accounts,
            name=name,
            external=external,
            )
        # We return no explicit group configuration and let the common code
        # to resolve the group based on the configured account group.
        # Here we could set a group instance when the group-uuid configuration
        # would be also stored in LDAP.
        return account, ()

    def _resolveHomeFolderPath(self, entry, home_folder_attribute, external):
        """
        Extract the home folder path from LDAP `entry` if required and store
        it into `external`.
        """
        if not home_folder_attribute:
            return

        # Extract home path from entry.
        home_folder_path = entry.get(home_folder_attribute[0])
        if not home_folder_path:
            # A LDAP home path was requested, but was not provided.
            raise InvalidAuthentication(
                'LDAP entry has not returned a value for "%s".' % (
                    home_folder_attribute[0],))

        if len(home_folder_path) != 1:
            raise InvalidAuthentication(
                'LDAP entry does not have a single home folder path.')

        # LDAP attributes are a set and we know it has a single member,
        # as this should have been checked above.
        home_path = list(home_folder_path)[0].decode('utf-8')

        if len(home_folder_attribute) > 1:
            template = home_folder_attribute[1]
            key = home_folder_attribute[0].replace('-', '_').lower()
            substitutions = {key: home_path}
            home_path = public_format(template, **substitutions)

        external['home_folder_path'] = home_path

    def _resolveEmail(self, entry, email_attribute):
        """
        Return the email fond in `entry`

        Return empty string if no email is defined.
        """
        if not email_attribute:
            return ''

        # Extract home path from entry.
        attributes = entry.get(email_attribute)
        if attributes is None:
            # No email attribute found,
            return ''

        emails = list(attributes)
        if not emails:  # noqa:cover
            # Email attribute found but has no value.
            # This should not happen as if has no value,
            # the empty string is returned.
            return ''

        # We sort the emails, so that we return consistent results.
        return sorted(emails)[0].decode('utf-8')

    def _setConfiguration(self):
        """
        Check that configuration is valid and expand it to be used in the
        authentication requests.

        Raise a ServerException with details if not valid.
        """
        bind_dn_type = self.configuration.bind_dn_type
        base_dns = self.configuration.bind_dn

        if bind_dn_type not in list(BIND_DN_TYPE_NAME.keys()):
            raise ServerException('Unknown value for "bind_dn_type".')
        self._bind_dn_type = bind_dn_type

        if bind_dn_type == TYPE_NAME.ABSOLUTE:
            # Base DN is ignored for absolute bind.
            # We still set one value to make sure the BIND is requested once.
            self._base_dns = ['']
        else:
            if not base_dns:
                raise ServerException('No value provided for "bind_dn".')
            self._base_dns = base_dns

        self._group_mapping = self.configuration.group_mapping
        self._group_mapping_attributes = set([])

        if self._group_mapping:
            for rule in self._group_mapping[1:]:
                if len(rule) < 3:
                    raise ServerException(
                        'The group mapping should have at least 3 elements.')
                self._group_mapping_attributes.add(rule[0])

                # Check that expression is valid.
                get_matched_groups(
                    value='', rule=rule[1], usage='group_mapping')

        self._group_mapping_attributes = list(self._group_mapping_attributes)

        # Load the configured extension or a default NOOP extension.
        # We use the NOOP extension to make sure the interface is up to
        # date and we don't accidentally update the interface without
        # documenting the changes.
        self._extension = load_extension_object(
            self._configuration.extension_entry_point,
            default=AuthLDAPNoop,
            )(self._configuration.extension_configuration)

    def _bind(self, client, bind_dn, password):
        """
        Call the BIND operation using `password`.
        """
        def eb_bind(failure):
            if failure.check(LDAPInvalidCredentials):
                # Normal rejection for our credentials.
                raise CredentialsForbidden(
                    'LDAP BIND forbidden as "%s".' % (bind_dn,))
            else:
                # Something went wrong.
                raise InvalidAuthentication('LDAP BIND failed as "%s". %s' % (
                    bind_dn, force_unicode(failure.value)))

        deferred = client.bind(
            bind_dn.encode('utf-8'), password.encode('utf-8'))
        deferred.addErrback(eb_bind)
        return deferred

    def _connect(self, on_connect):
        """
        Connect to the remote LDAP server, calling `on_connect` with the LDAP
        client and disconnection when `on_connect` is done.

        `on_connect` should return a deferred.
        """
        def cb_connect(client):
            """
            Called when TCP connection was successfully made,
            but before TLS handshake is done.
            """
            deferred = on_connect(client)
            deferred.addBoth(bb_connect_done, client)
            return deferred

        def bb_connect_done(result, client):
            """
            Called after all LDAP operations to clean the client.
            """
            if not client.connected:
                return result

            client.unbind()
            # A hack to be able to wait for client disconnection.
            # But this is not always working as our custom TLS shutdown
            # might still be waiting for remote peer to close the TLS session.
            # See https://github.com/twisted/ldaptor/issues/225
            deferred = client.bind(b'', b'')
            deferred.addBoth(lambda _: result)
            return deferred

        def eb_connect(failure):
            """
            Called when connection failed.
            """
            failure.trap(ConnectError)
            raise ServerException(
                'Failed to connect. %s' % (force_unicode(failure.value),))

        host = self.configuration.address
        port = self.configuration.port

        def get_tcp_endpoint():
            # FIXME:5192:
            # Add a custom connect timeout.
            return defer.succeed(TCP4ClientEndpoint(
                reactor=self._scheduler,
                host=host,
                port=port,
                timeout=CONNECT_TIMEOUT,
                ))

        @defer.inlineCallbacks
        def get_ssl_endpoint():
            context_factory = yield self._initStartedContextFactory(
                remote_servers=[(
                    self.configuration.address, self.configuration.port)])
            # FIXME:5192:
            # Add a custom timeout.
            # FIXME:6505:
            # Implement handshake interface.
            endpoint = SSL4ClientEndpoint(
                reactor=self._scheduler,
                host=host,
                port=port,
                sslContextFactory=context_factory,
                timeout=CONNECT_TIMEOUT,
                )
            defer.returnValue(endpoint)

        # FIXME:5189:
        # Use a generic endpoint that can resolve hostname to IPv6.
        if self._configuration.secure_connection:
            deferred = get_ssl_endpoint()
        else:
            deferred = get_tcp_endpoint()

        deferred.addCallback(
            lambda endpoint: connectProtocol(endpoint, LDAPClient()))

        # FIXME:5192:
        # Add a timeout in the case after connect, LDAP operation
        # takes a long time.
        deferred.addCallbacks(cb_connect, eb_connect)
        return deferred
