# Copyright (c) 2015 Adi Roiban.
# See LICENSE for details.
"""
Authentication done for application accounts.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

from twisted.internet.task import LoopingCall
from zope.interface import implementer


from chevah.server.authentication.credential_checker import (
    CredentialsValidator,
    )
from chevah.server.commons.constant import CONFIGURATION_DEFAULTS
from chevah.server.commons.exception import (
    CredentialsForbidden,
    InvalidAuthentication,
    ServerException,
    )
from chevah.server.commons.interface import ILocalFileAuthentication
from chevah.server.commons.runnable import Runnable
from chevah.server.configuration.file import FileConfigurationProxy
from chevah.server.identity.model import (
    AccountsConfiguration,
    GroupsConfiguration,
    )


@implementer(ILocalFileAuthentication)
class LocalFileAuthentication(Runnable):
    """
    Do authentication for application accounts with passwords stored in a
    dedicated local file.

    Once loaded it will update the configuration options with the loaded
    identities.
    """
    PROPERTIES_REQUIRING_RESTART = ('path', 'external_management')

    # Loop reloading the config.
    _update_loop = None
    # Path to the configuration file.
    _path = None
    # Low level configuration file.
    _proxy = None

    @property
    def accounts(self):
        return self.configuration.accounts

    @property
    def groups(self):
        return self.configuration.groups

    @property
    def administrators(self):
        return self.configuration.administrators

    @property
    def roles(self):
        return self.configuration.roles

    def _onStart(self, avatar=None):
        """
        See: Runnable.
        """
        if not self.configuration.path:
            raise ServerException('The configured path cannot be empty.')

        self._path = self.configuration.path

        self._lcLoadConfiguration()

        if self.configuration.external_management:
            self._update_loop = LoopingCall(self._lcLoadConfiguration)
            self._update_loop.clock = self._scheduler
            deferred = self._update_loop.start(5 * 60, now=False)
            deferred.addErrback(self._ebLoadConfiguration)

    def _onStop(self, avatar=None):
        """
        See: Runnable.
        """
        # Make sure we will not touch the config file after we are stop.
        self._proxy = None
        self._path = None

        if self._update_loop and self._update_loop.running:
            self._update_loop.stop()
        self._update_loop = None

        # In the ideal case, the identity would be only stored on the
        # runnable, but the current Web Manager has no support for
        # managing the configuration for a runnable.
        self.configuration.unloadIdentity()

    def _lcLoadConfiguration(self):
        """
        Called each time we need to reload the configuration.
        """
        self._proxy = FileConfigurationProxy(
            configuration_path=self._path, defaults=CONFIGURATION_DEFAULTS)
        self._proxy.load()

        # In the ideal case, the identity would be only stored on the
        # runnable, but the current Web Manager has no support for
        # managing the configuration for a runnable.
        self.configuration.accounts = AccountsConfiguration(self)
        self.configuration.groups = GroupsConfiguration(self)

        notifiers = [
            self.configuration.accounts,
            self.configuration.groups,
            self.configuration.roles,
            self.configuration.administrators,
            ]
        for target in notifiers:
            target.subscribe('update', self._onConfigurationChange)
            target.subscribe('create', self._onConfigurationChange)
            target.subscribe('delete', self._onConfigurationChange)

    def _ebLoadConfiguration(self, failure):
        """
        Called when we fail to load the configuration.
        """
        self.fail(failure.value)

    def _onConfigurationChange(self, signal):
        """
        Called when the configuration was changed.
        """
        self._proxy.save()

    def authenticate(self, credentials):
        """
        See: `IAuthentication`
        """
        account_configuration = (
            self.configuration.accounts.getApplicationAccountConfiguration(
                credentials.username))

        if account_configuration is None:
            # User does not exist.
            return None

        groups = []
        for group_uuid in account_configuration.group:
            group = self.groups.getSection(group_uuid)
            groups.append(group)
            if not group:
                raise InvalidAuthentication(
                    'Group does not exist: %s' % (group_uuid,))

        validator = CredentialsValidator(
            me=account_configuration,
            groups=groups,
            )
        result = validator.validate(credentials)

        if not result:
            raise CredentialsForbidden('Credentials denied')

        return account_configuration, groups

    def authenticateAdministrator(self, credentials):
        """
        See: `IAuthentication`
        """
        raise InvalidAuthentication(
            'Local file administrators are not yet supported.')
