# Copyright (c) 2015 Adi Roiban.
# See LICENSE for details.
"""
Authentication done for application accounts.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

from zope.interface import implementer

from chevah_compat import system_users, process_capabilities

from chevah.server.authentication.configuration import (
    _AuthenticationConfigurationBase,
    _AllowedGroupsMixin,
    )
from chevah.server.authentication.credential_checker import (
    CredentialsValidator,
    )
from chevah.server.commons.constant import (
    CONFIGURATION_ALL_OS_GROUPS_PLACEHOLDER,
    CONFIGURATION_DEFAULT_GROUP,
    CONFIGURATION_DEFAULT_ROLE,
    CONFIGURATION_INHERIT,
    OS_GROUP_ASSOCIATION,
    OS_ROLE_ASSOCIATION,
    TYPE_NAME,
    )
from chevah.server.commons.exception import (
    CredentialsForbidden,
    InvalidAuthentication,
    ServerException,
    )
from chevah.server.commons.interface import (
    IAuthentication,
    IOSAccountConfiguration,
    IOSAuthenticationConfiguration,
    IPasswordCredentials,
    )
from chevah.server.commons.runnable import Runnable
from chevah.server.configuration.model import (
    WritableString,
    WritableStringCaseInsensitive,
    WritableCSVUnique,
    WritableStringOrEmpty,
    WritableCSV,
    )

from chevah.server.identity.model import (
    ExternalAccountConfigurationAbstract,
    OSAdministratorConfiguration,
    )


@implementer(IOSAccountConfiguration)
class MissingOSAccountConfiguration(ExternalAccountConfigurationAbstract):
    """
    Account configuration for an operating system account not found
    in the configuration database.
    """

    @property
    def kind_name(self):
        """
        Human readable name for the type of this account configuration.
        """
        return TYPE_NAME.OS_EXTERNAL

    @property
    def type(self):
        """
        Type is always OS.
        """
        return TYPE_NAME.OS

    @property
    def lock_in_home_folder(self):
        """
        OS account might not be locked in home folders.
        """
        return CONFIGURATION_INHERIT[0]


def _get_OSAccountConfiguration(accounts, name, group_uuids):
    """
    Return configuration for OS account with `name`.
    """
    if not system_users.userExists(name):
        # OS user doesn't exist so we don't have any configuration.
        return None

    account = accounts.getSectionByName(name)
    if not account:
        # Account exists in the OS, but not in the configuration.
        return MissingOSAccountConfiguration(
            parent=accounts,
            name=name,
            external={'group': group_uuids},
            )

    if not IOSAccountConfiguration.providedBy(account):
        # The name found in the configuration is not an OS account.
        # The authentication fails, to avoid confusing an OS account
        # with an application account having the same name but being very
        # different accounts.
        return None

    return account


@implementer(IOSAuthenticationConfiguration)
class OSAuthenticationConfiguration(
        _AuthenticationConfigurationBase, _AllowedGroupsMixin):
    """
    Configuration for OS authentication.
    """

    pam_usage = WritableStringOrEmpty('pam_usage')
    pam_service = WritableString('pam_service')
    group_association = WritableString(
        'group_association', default=TYPE_NAME.BASE_GROUPS)
    base_groups = WritableCSV('base_groups')
    manager_allowed_groups = WritableCSVUnique(
        'manager_allowed_groups')
    role_association = WritableStringCaseInsensitive(
        'role_association', default=TYPE_NAME.BASE_ROLES)
    base_roles = WritableCSV('base_roles')

    def _migrateINI(self):
        """
        Handle converting from older version configuration to the latest one.

        This is only when initialied from .INI
        """
        # Update the configuration from single group association used on 4.22
        # and before to to multiple group association.
        valid_group_associations = [
            k.lower() for k in OS_GROUP_ASSOCIATION.keys()]
        group_associations_or_uuid = self.group_association

        if group_associations_or_uuid.lower() not in valid_group_associations:
            # We have a legacy default group UUID.
            self.base_groups = [group_associations_or_uuid]
            self.group_association = TYPE_NAME.BASE_GROUPS

        self.group_association = self.group_association.lower()

        if not self._proxy.hasOption(self._section_name, 'base_groups'):
            self.base_groups = [CONFIGURATION_DEFAULT_GROUP]

        # Update from 4.26.2 or older in which there were no explicit
        # base_roles.
        # If they are missing, go with backward compatible behaviour in
        # which the default role is always associated.
        if not self._proxy.hasOption(self._section_name, 'base_roles'):
            self.base_roles = [CONFIGURATION_DEFAULT_ROLE]


@implementer(IAuthentication)
class OSAuthentication(Runnable):
    """
    Do authentication for accounts for which the password is provided
    by the operating system and which will be mapped to OS accounts.
    """
    PROPERTIES_REQUIRING_RESTART = (
        'allowed_groups',
        'manager_allowed_groups',
        'base_groups',
        'base_roles',
        'group_association',
        'role_association',
        )

    # UUID for the group that is associated with the authenticated accounts.
    # Set to None when the group should be associated based on OS group name.
    _group_uuid = None
    # List of groups allowed for non-admin account authentication.
    _allowed_groups = None
    # List of roles that are associated to all admins
    _base_roles = ()

    def _onStart(self, avatar=None):
        """
        See: Runnable.
        """
        if process_capabilities.os_name == 'osx':
            raise ServerException('OS accounts are not available on macOS.')

        self._validateConfiguration()

    def _validateConfiguration(self):
        """
        Check that configuration is valid.
        """
        self._allowed_groups = self._configuration.allowed_groups
        self._manager_allowed_groups = (
            self._configuration.manager_allowed_groups)

        if (
            not self._allowed_groups
            and not self._manager_allowed_groups
                ):
            raise ServerException(
                'No configured "allowed_groups" or "manager_allowed_groups".')

        self._validateAccountGroups()
        self._validateAdministratorConfiguration()

    def _validateAccountGroups(self):
        """
        Check that allowed_groups and base_groups have a valid configuration.
        """
        if not self._allowed_groups:
            # Nothing to check since accounts are not configured.
            return

        self._group_association = self._configuration.group_association

        if self._allowAnyGroup() and len(self._allowed_groups) > 1:
            raise ServerException(
                'Any group marker "%s" '
                "can't be mixed with other group names." % (
                    CONFIGURATION_ALL_OS_GROUPS_PLACEHOLDER,))

        # Check that configured groups that we are expected to associate
        # with exists.
        if (
            self._allowAnyGroup()
            and self._group_association != TYPE_NAME.BASE_GROUPS
                ):
            # We are configured to associate account with a group name,
            # but no explicit accepted group names are configured.
            # This is rejected, as an OS user can be member of multiple
            # groups and we don't know for which group should we enable
            # SFTPPlus group association.
            raise ServerException(
                'You can\'t configure "group_association" while accepting '
                'accounts from any OS group.')

        self._pre_auth_group_uuids = self._getPreAuthGroupUUIDs(
            self._group_association)
        if not self._pre_auth_group_uuids:
            # FIXME:5239:
            # See how to handle pre-authentication group association.
            # This affects allowed_IPs and ssh keys validation in
            # CredentialsValidator.
            self._pre_auth_group_uuids = (CONFIGURATION_DEFAULT_GROUP,)

    def _validateAdministratorConfiguration(self):
        """
        Check that configuration for admin authentication is valid.
        """
        if not self._manager_allowed_groups:
            # Nothing to check since admin auth is not enabled.
            return

        role_association = self._configuration.role_association
        if role_association not in list(OS_ROLE_ASSOCIATION.keys()):
            raise ServerException(
                'Unknown "role_association": %s.' % (role_association,))
        self._role_association = role_association

        roles = self.root.configuration.roles
        base_roles = []
        for role_uuid in self._configuration.base_roles:
            role = roles.getSection(role_uuid)
            if not role:
                raise ServerException(
                    'Unknown role UUID configured for "base_roles": %s.' % (
                        role_uuid,))
            base_roles.append(role)

        self._base_roles = tuple(base_roles)

        if (
            self._role_association == TYPE_NAME.BASE_ROLES
            and not self._base_roles
                ):
            raise ServerException(
                'Administrators authentications configured with '
                '"role_association = base-roles" but no roles defined '
                'in "base_roles".')

        if self._role_association == TYPE_NAME.BASE_AND_OS_GROUPS:
            for group_name in self._manager_allowed_groups:
                role = self.root.configuration.roles.getSectionByName(
                    group_name)
                if not role:
                    raise ServerException(
                        'Configured OS group "%s" for '
                        '"manager_allowed_groups" doesn\'t exist.' % (
                            group_name))

    def _getPreAuthGroupUUIDs(self, group_association):
        """
        Validate and define the default groups associated with the
        pre-authenticated accounts.
        """
        groups = self.root.configuration.groups

        def check_group_existence(group, group_name, config_name):
            """
            Helper to try to prevent configuration error that might be raise
            later during runtime for a group that don't exist.
            """
            if group:
                return
            raise ServerException(
                'Unknown group "%s" defined for "%s" '
                'configuration.' % (group_name, config_name))

        if group_association == TYPE_NAME.GROUP_NAME_WITH_DEFAULT:
            # Is hard to know what groups to allow at this point
            # as anything not found goes to default.
            return ()

        if group_association in [
            TYPE_NAME.GROUP_NAME,
            TYPE_NAME.BASE_AND_OS_GROUPS,
                ]:

            for group_name in self._allowed_groups:
                check_group_existence(
                    groups.getSectionByName(group_name),
                    group_name,
                    'allowed_groups',
                    )

            if group_association == TYPE_NAME.BASE_AND_OS_GROUPS:
                # We start with base groups and add more groups later.
                return self._configuration.base_groups

            # We have legacy `group-name`
            return ()

        # Simple base groups only.
        if not self._configuration.base_groups:
            raise ServerException('No "base_groups" configured.')

        for group_uuid in self._configuration.base_groups:
            check_group_existence(
                groups.getSection(group_uuid), group_uuid, 'base_groups')

        return self._configuration.base_groups

    def _onStop(self, avatar=None):
        """
        See: Runnable.
        """
        self._pre_auth_group_uuids = []
        self._allowed_groups = None
        self._group_association = None
        self._role_association = None
        self._base_roles = ()

    def authenticate(self, credentials):
        """
        See: `IAuthentication`
        """
        if not self._allowed_groups:
            raise CredentialsForbidden('No configured "allowed_groups".')

        # Get OS configuration for pre-autentication.
        account_configuration = _get_OSAccountConfiguration(
            self.root.configuration.accounts,
            credentials.username,
            self._pre_auth_group_uuids)
        if account_configuration is None:
            # User does not exist in the application.
            # Allow other methods to check it.
            return None

        groups = self._validateOSUser(account_configuration, credentials)

        if self._allowAnyGroup():
            # We don't need to check OS group membership.
            return account_configuration, groups

        # We need to check the OS group membership.
        matched_groups = self._getOSGroups(credentials, self._allowed_groups)

        if not matched_groups:
            raise CredentialsForbidden(
                'User is not a member of an allowed group.')

        if not isinstance(account_configuration, MissingOSAccountConfiguration):
            # For OS accounts in the configuration file
            # we return what is configured.
            return account_configuration, groups

        if self._group_association == TYPE_NAME.BASE_AND_OS_GROUPS:
            # We have simple base groups association.
            groups = []
            for group_uuid in self._configuration.base_groups:
                # The base_groups should always exist.
                # If you try to delete them after this OS auth method starts,
                # you will get an error, via the generic configuration
                # dependency protection mechanism.
                # It will detect the group.uuid is used here.
                groups.append(
                    self.root.configuration.groups.getSection(group_uuid))

            for os_group_name in matched_groups:
                extra_group = self.root.configuration.groups.getSectionByName(
                    os_group_name)
                if not extra_group:
                    # This should normally not happen.
                    # At method startup we check that configured group names
                    # exist.
                    raise InvalidAuthentication(
                        'User associated with a group that '
                        'doesn\'t have a matching SFTPPlus group: "%s"' % (
                            os_group_name,))
                if extra_group.uuid in self._configuration.base_groups:
                    # Group already added via base groups.
                    continue
                groups.append(extra_group)

            # If we are here, it means we have an OS account not defined in
            # our config. Update the account config to match the groups.
            account_configuration._group = [g.uuid for g in groups]
            return account_configuration, groups

        if self._group_association == TYPE_NAME.GROUP_NAME:
            associated_group = self.root.configuration.groups.getSectionByName(
                matched_groups[0])
            if not associated_group:
                # This should normally not happen.
                # At method startup we check that configured group names exist.
                raise InvalidAuthentication(
                    'User associated with a group '
                    'that doesn\'t have a matching SFTPPlus group: "%s"' % (
                        matched_groups[0],))
            # If we are here, it means we have an OS account not defined in
            # our config. Update the account config to match the groups.
            account_configuration._group = [associated_group.uuid]
            return account_configuration, (associated_group,)

        if self._group_association == TYPE_NAME.GROUP_NAME_WITH_DEFAULT:
            associated_group = self.root.configuration.groups.getSectionByName(
                matched_groups[0])

            if not associated_group:
                associated_group = self.root.configuration.groups.getSection(
                    CONFIGURATION_DEFAULT_GROUP)

            # If we are here, it means we have an OS account not defined in
            # our config. Update the account config to match the groups.
            account_configuration._group = [associated_group.uuid]
            return account_configuration, (associated_group,)

        # We have simple base groups association.
        return account_configuration, groups

    def authenticateAdministrator(self, credentials):
        """
        See: `IAuthentication`
        """
        if not self._manager_allowed_groups:
            raise CredentialsForbidden(
                'No configured "manager_allowed_groups".')

        account_name = credentials.username
        if not system_users.userExists(account_name):
            return None

        pre_auth_roles = self._base_roles
        if not pre_auth_roles:
            # When no base_roles are defined, we need to set some roles,
            # as this is required by the shared CredentialsValidator code.
            # We get the default role as it is needed by the shared auth code.
            # It should be always available.
            default_role = self.root.configuration.roles.getSection(
                CONFIGURATION_DEFAULT_ROLE)
            pre_auth_roles = (default_role,)

        administrator = OSAdministratorConfiguration(
            parent=self.root.configuration.administrators,
            uuid=account_name,
            )

        validator = CredentialsValidator(
            me=administrator,
            groups=pre_auth_roles,
            )
        result = validator.validate(credentials)

        if not result:
            raise CredentialsForbidden('OS Manager credentials denied.')

        # We need to check the OS group membership.
        matched_groups = self._getOSGroups(
            credentials, self._manager_allowed_groups)
        if not matched_groups:
            raise CredentialsForbidden(
                'Administrator is not a member of an allowed group.')

        roles = list(self._base_roles)

        if self._role_association == TYPE_NAME.BASE_AND_OS_GROUPS:
            for os_group_name in matched_groups:
                extra_role = self.root.configuration.roles.getSectionByName(
                    os_group_name)
                if not extra_role:
                    # This should normally not happen.
                    # At method startup we check that configured group names
                    # exist.
                    raise InvalidAuthentication(
                        'Administrator associated with a group '
                        'that doesn\'t have a matching SFTPPlus role: "%s"' % (
                            os_group_name,))
                if extra_role.uuid in self._configuration.base_roles:
                    # Role already added via base roles.
                    continue
                roles.append(extra_role)

        return administrator, tuple(roles)

    def _allowAnyGroup(self):
        """
        Return `True` if OS account from any group are allowed for
        non-admin authentication.
        """
        return CONFIGURATION_ALL_OS_GROUPS_PLACEHOLDER in self._allowed_groups

    def _getOSGroups(self, credentials, target_groups):
        """
        Return the list of OS group names to which the username is member of,
        based on the list that we expect to have access to SFTPPlus via
        `target_groups`.
        """
        token = None
        if IPasswordCredentials.providedBy(credentials):
            # If we got a result, the credentials were updated with a token.
            token = credentials.token

        result = []
        for os_group in target_groups:
            matched_group = system_users.getGroupForUser(
                username=credentials.username,
                token=token,
                groups=[os_group],
                )
            if not matched_group:
                continue
            result.append(os_group)
        return result

    def _validateOSUser(self, account_configuration, credentials):
        """
        Check that `credentials` are accepted by the OS and our configuration.
        """
        groups = []
        for group_uuid in account_configuration.group:
            group = self.root.configuration.groups.getSection(group_uuid)
            groups.append(group)
            if not group:
                raise InvalidAuthentication(
                    'Group does not exist: %s' % (group_uuid,))

        validator = CredentialsValidator(
            me=account_configuration,
            groups=groups,
            pam_usage=self._configuration.pam_usage,
            pam_service=self._configuration.pam_service,
            )
        result = validator.validate(credentials)

        if not result:
            raise CredentialsForbidden('Credentials denied')

        return groups
