# Copyright (c) 2017 Adi Roiban.
# See LICENSE for details.
"""
Code for handling authorization on the server-side.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from chevah.server.commons.constant import (
    PERMISSION,
    )
from chevah.server.commons.exception import AuthorizationException
from chevah.server.commons.match import get_matched_groups


class AuthorizationAvatarAdapter(object):
    """
    Avatar used for the authorization process.

    All path should be passed in absolute format

    This tries to implement something based on Windows permissions, reusing
    as much as possible of the design and keeping things simple:
    https://msdn.microsoft.com/en-us/library/bb727008.aspx

    See public documentation for more details.
    """

    def __init__(self, avatar):
        permissions = avatar.permissions
        self._global_permissions = permissions[0]
        self._path_permissions = permissions[1:]

    def _check(self, targets, path):
        """
        Raise an exception if none of the `target` permissions is
        allowed for path.
        """
        target_permissions = self._getPathPermissions(path)

        if target_permissions is None:
            target_permissions = self._global_permissions

        if PERMISSION.DENY_FULL in target_permissions:
            # Deny full-control takes precedence over everything.
            raise AuthorizationException(
                'Denied due to "%s" for "%s"' % (PERMISSION.DENY_FULL, path))

        if PERMISSION.ALLOW_FULL in target_permissions:
            # full-control shadows any permission.
            return

        for target in targets:
            if target in target_permissions:
                # We have one permissions which matches the path.
                return

        raise AuthorizationException('Need "%s" permission for "%s".' % (
            ' or '.join(targets), path))

    def _getPathPermissions(self, path):
        """
        Return the list of permissions for `path`.

        Return `None` when no permissions could be found for `path`.

        The returned value is just the permissions, without the path
        matching expression.
        """
        for rule in self._path_permissions:
            if get_matched_groups(
                value=path,
                rule=rule[0],
                usage='permissions',
                    ):
                return rule[1:]

        # No match found.
        return None

    def canFull(self, path):
        """
        Check if has full permissions for `path`.
        """
        self._check([PERMISSION.ALLOW_FULL], path)

    def canRead(self, path):
        """
        Check if has permissions to read file at `path`.
        """
        self._check([PERMISSION.ALLOW_READ], path)

    def canWrite(self, path):
        """
        Check if has permissions to write/amend/create file at `path`.
        """
        self._check([PERMISSION.ALLOW_WRITE], path)

    def canList(self, path):
        """
        Check if has permissions to list folder at `path`.
        """
        self._check([PERMISSION.ALLOW_LIST], path)

    def canTraverse(self, path):
        """
        Check if has permissions to traverse folder at `path`.

        If one can list, it can also traverse.
        """
        self._check(
            [PERMISSION.ALLOW_LIST, PERMISSION.ALLOW_TRAVERSE], path)

    def canRename(self, from_path, to_path=None):
        """
        Check if has permissions to rename filer or folder from
        `from_path` to `to_path`.

        Both source and destination are required.
        """
        self._check([PERMISSION.ALLOW_RENAME], from_path)
        self._check([PERMISSION.ALLOW_RENAME], to_path)

    def canDeleteFile(self, path):
        """
        Check if has permissions to delete file at `path`.
        """
        self._check([PERMISSION.ALLOW_DELETE_FILE], path)

    def canDeleteFolder(self, path):
        """
        Check if has permissions to delete folder at `path`.
        """
        self._check([PERMISSION.ALLOW_DELETE_FOLDER], path)

    def canCreateFolder(self, path):
        """
        Check if has permissions to create folder at `path`.
        """
        self._check([PERMISSION.ALLOW_CREATE_FOLDER], path)

    def canGetAttributes(self, path):
        """
        Check if has permissions to get permissions for `path`.
        """
        self._check(
            [PERMISSION.ALLOW_READ, PERMISSION.ALLOW_LIST], path)

    def canSetAttributes(self, path):
        """
        Check if has permissions to set permissions for `path`.
        """
        self._check([PERMISSION.ALLOW_SET_ATTRIBUTES], path)
