# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
Module containing code used by the admin-commands.(sh|bat) script.
"""
from argparse import ArgumentParser
from getpass import getpass
from io import BytesIO, StringIO
import os
import re
import signal
import sys
import uuid
import webbrowser

from unidecode import unidecode

from chevah_compat import local_filesystem, process_capabilities
from chevah_keycert.exceptions import KeyCertException
from chevah_keycert.ssh import (
    generate_ssh_key as crypto_generate_ssh_key,
    generate_ssh_key_parser,
    Key,
    )
from chevah_keycert.ssl import (
    generate_and_store_csr,
    generate_csr_parser,
    generate_self_signed_parser,
    generate_ssl_self_signed_certificate,
    )
from chevah.server import force_unicode, MODULE_PATH
from chevah.server.authentication.base import password_handler
from chevah.server.commons import OPENSSL_ED25519_SUPPORTED
from chevah.server.commons.constant import (
    CONFIGURATION_FOLDER_NAME,
    CONFIGURATION_PID_FILE,
    CONFIGURATION_SERVER_FILE,
    DEFAULT_KEY_SIZE,
    DEFAULT_KEY_TYPE,
    PASSWORD_CRYPT_ROUNDS,
    PRODUCT_NAME,
    PRODUCT_VERSION,
    TYPE_NAME,
    )
from chevah.server.commons.configuration import ApplicationConfiguration
from chevah.server.commons.exception import ServerError, ServerException
from chevah.server.commons.info import (
    WINDOWS_SERVICE_NAME,
    WINDOWS_DISPLAY_NAME,
    )

from chevah.server.commons.process import ChevahServerProcess
from chevah.server.commons.jinja import ChevahJinjaEnvironment
from chevah.server.openpgp.util import generate_pgp_key
import six

if os.name == 'nt':
    import win32service
else:
    win32service = None


# We have a shared log file for the supervisor process and the
# SFTPPlus process. 1MB max without any backup.
_SUPERVISORD_INI = """
[supervisord]
logfile=%(base)s/log/server-startup-errors.log
logfile_maxbytes=100KB
logfile_backups=0
loglevel=critical
pidfile=%(pid)s
nodaemon=false
umask=022


[rpcinterface:supervisor]
supervisor.rpcinterface_factory =
  supervisor.rpcinterface:make_main_rpcinterface

[eventlistener:exit_process]
command=%(base)s/bin/supervisor-stop.sh %(pid)s
autorestart=false
events=PROCESS_STATE_STOPPED, PROCESS_STATE_EXITED, PROCESS_STATE_FATAL

[program:sftpplus-service]
directory=%(base)s
umask=022
autorestart=false
command=
  %(base)s/bin/python -m chevah.server.__init__
    start-in-foreground --config %(config)s

redirect_stderr=true
stdout_logfile=%(base)s/log/server-startup-errors.log
stdout_logfile_maxbytes=100KB
stdout_logfile_backups=0

stopwaitsecs=10
stopasgroup=true
killasgroup=true
"""


class UnicodeArgumentParser(ArgumentParser):
    """
    An argument parser which can handle internal Unicode messages.
    """

    def _print_message(self, message, file=None):
        """
        Called write a message.

        This overwrites the upstream code so that we can have control over
        the way data is written.
        """
        if not message:
            return

        if file is None:
            file = sys.stderr

        if six.PY2 and isinstance(message, six.text_type):
            message = message.encode('utf-8')

        file.write(message)


class ServerCommandParser(UnicodeArgumentParser):
    """
    Argument parser used by SFTPPlus' command line.
    """

    def error(self, message):
        """
        Called when parser got into errors.
        """
        # When no arguments are provided, show full help.
        if len(sys.argv) == 1:
            self.print_help()

        # Default action of sending error to stdout.
        if six.PY2 and process_capabilities.os_family != 'posix':
            # On Windows we get IOError if we try to send Unicode characters
            # even if they are UTF-8 encoded so we transliterate them to
            # plain ASCII.
            message = unidecode(message)

        output = 'error: %s\n' % message

        if six.PY2:
            output = output.encode('utf-8')

        sys.stdout.write(output)
        sys.exit(2)


def _multi_line(value):
    """
    Return the value as multi-line ini value.
    """
    if isinstance(value, six.binary_type):
        value = value.decode('utf-8')

    result = []
    for line in value.splitlines():
        result.append('  ' + line)
    return '\n'.join(result)


class Commands(object):
    """
    Main container for tasks available via admin-commands.
    """

    _PASSWORD_CRYPT_ROUNDS = PASSWORD_CRYPT_ROUNDS

    # Here to help with automated testing.
    _win32service = win32service
    _os_name = os.name
    _os_getcwd = os.getcwd

    def __init__(self):
        self._server_module_segments = None
        script_name = 'admin-commands'

        self._parser = ServerCommandParser(prog=script_name)
        self._subparsers = self._parser.add_subparsers(
            help='Available commands', dest='action')
        self._subparsers.required = False

        self._addGenericOptions()
        self._addUnixOptions()

    def executeSubcommand(self, options):
        """
        Called when default action should be performed for `options`.
        """
        action = options.action
        handler_name = 'subcommand_%s' % (action.replace('-', '_'))
        handler = getattr(self, handler_name)
        try:
            exit_code, message = handler(options)
        except Exception as error:
            exit_code = 1
            message = 'Failed to execute the command. %s' % (error,)

        # This is the exit point.
        message = message + os.linesep
        if six.PY2:
            message = message.encode('utf-8')

        sys.stdout.write(message)
        sys.exit(exit_code)

    def _addGenericOptions(self):
        """
        Add options available for all systems.
        """
        self._parser.add_argument(
            '--version', action='version', version=PRODUCT_VERSION)

        self._configuration_file = ArgumentParser(add_help=False)
        self._configuration_file.add_argument(
            '-c', '--config',
            action='store', dest='config',
            metavar="CONFIG", default=CONFIGURATION_SERVER_FILE,
            help='Read configuration data from CONFIG. Default: %s' % (
                CONFIGURATION_SERVER_FILE),
            )

        self._subparsers.add_parser(
            'debug',
            help=(
                'Start SFTPPlus in foreground. '
                'Logs are sent to standard output.'),
            parents=[self._configuration_file],
            )

        self._subparsers.add_parser(
            'start-in-foreground',
            help='Start SFTPPlus as a foreground process.',
            parents=[self._configuration_file],
            )

        self._subparsers.add_parser(
            'manager',
            help='Opens the local manager page in the default browser.',
            parents=[self._configuration_file],
            )

        self._subparsers.add_parser(
            'documentation',
            help='Opens the documentation page in the default browser.',
            parents=[self._configuration_file],
            )

        private_key = ArgumentParser(add_help=False)
        private_key.add_argument(
            '--key-size',
            action='store', type=int, dest='key_size',
            metavar="NUMBER", default=DEFAULT_KEY_SIZE,
            help='Generate a RSA or DSA key of size SIZE. Default %d.' % (
                    (DEFAULT_KEY_SIZE)),
            )
        private_key.add_argument(
            '--key-type',
            action='store', dest='key_type',
            metavar="[rsa|dsa]", default=DEFAULT_KEY_TYPE,
            help='Generate a DSA or RSA key. Default %s.' % (
                    (DEFAULT_KEY_TYPE)),
            )
        private_key.add_argument(
            '--key-file',
            action='store', dest='key_file',
            metavar="PATH", default=None,
            help=(
                'Store the keys pair in FILE and FILE.pub. Default id_TYPE.'),
            )

        generate_ssh_key_parser(
            self._subparsers, 'generate-ssh-key',
            default_key_type=DEFAULT_KEY_TYPE,
            )

        generate_csr_parser(
            self._subparsers, 'generate-csr',
            default_key_size=DEFAULT_KEY_SIZE,
            )

        generate_self_signed_parser(
            self._subparsers, 'generate-self-signed',
            default_key_size=DEFAULT_KEY_SIZE,
            )

        generate_pgp_key = self._subparsers.add_parser(
            'generate-pgp-key',
            help='Generate a new PGP key.',
            parents=[private_key],
            )
        generate_pgp_key.add_argument(
            '--email', action='store',
            default='no-email@example.com',
            metavar='NAME@DOMAIN.TLD',
            help='Email associated with the new PGP key.',
            )
        generate_pgp_key.add_argument(
            '--name', action='store',
            default='Autogenerated Key',
            metavar='TEXT',
            help='Name associated with the new PGP key.',
            )

        generate_pgp_key.add_argument(
            '--comment', action='store',
            default='Created by SFTPPlus',
            metavar='TEXT',
            help='Comment associated with the new PGP key.',
            )

        generate_pgp_key.add_argument(
            '--password', action='store',
            default='',
            metavar='TEXT',
            help='Passphrase used to encrypt the new PGP private key.',
            )

        generate_password = self._subparsers.add_parser(
            'generate-password',
            help=(
                'Create an encrypted password used for application accounts. '
                'To be specified in command line or read interactively.'),
            )
        generate_password.add_argument('password', nargs='?')
        generate_password.add_argument(
            '--format', action='store',
            default=TYPE_NAME.CRYPT_SHA512,
            metavar='crypt-sha512|crypt-sha256|pbkdf2-sha512|pbkdf2-sha256',
            help='Format used to generate the hashed password.',
            )

        verbose = ArgumentParser(add_help=False)
        verbose.add_argument(
            '-v', '--verbose', action='store_true', dest='verbose',
            default=False,
            help='Show more info while executing a command.',
            )

        local_manager = ArgumentParser(add_help=False)
        local_manager.add_argument(
            '--init-admin',
            action='store', dest='admin_name',
            metavar="NAME", default='admin',
            help='Name of the initial administrator account. Default: admin.',
            )
        local_manager.add_argument(
            '--init-password',
            action='store', dest='admin_password',
            metavar="PLAIN-TEXT-PASSWORD", required=True,
            help='Password for the initial administrator account.',
            )
        local_manager.add_argument(
            '--local-admin-access',
            action='store_true', dest='local_admin_access', default=False,
            help=(
                'Only allow management access from localhost. '
                'When not set, it will allow access from any host.'
                ),
            )
        local_manager.add_argument(
            '--service-os-user',
            action='store',
            dest='service_os_user',
            metavar="OS_USER",
            default='sftpplus',
            help=(
                'Name of the OS user dedicated to the SFTPPlus process. '
                ),
            )
        verbose.add_argument(
            '--drop-user', action='store_true', dest='drop_user',
            default=False,
            help=(
                'Whether to start SFTPPlus as root and then drop permissions '
                'to the service user.'
                'Default: Not start as root.'
                ),
            )

        initialize = self._subparsers.add_parser(
            'initialize',
            help='Initialize SFTPPlus configuration.',
            parents=[verbose, local_manager],
            )
        initialize.add_argument(
            '--key-size',
            action='store', type=int, dest='key_size',
            metavar="SIZE", default=DEFAULT_KEY_SIZE,
            help='Generate a RSA or DSA key of size SIZE. Default %d.' % (
                    (DEFAULT_KEY_SIZE)),
            )

        self._subparsers.add_parser(
            'migrate',
            help='Update SFTPPlus configuration for current version.',
            parents=[verbose, local_manager, private_key],
            )

        self._subparsers.add_parser(
            'generate-uuid',
            help='Generate UUID.',
            description='Generate UUID.',
            )

        self._subparsers.add_parser(
            'validate',
            help='Validate SFTPPlus configuration.',
            parents=[self._configuration_file]
            )

        help_parser = self._subparsers.add_parser(
            'help',
            help='Show help message for an action.',
            )
        help_parser.add_argument('help_action', nargs='?')

    def _addUnixOptions(self):
        '''Add Unix specific options.'''
        if process_capabilities.os_family != 'posix':
            return

        start = self._subparsers.add_parser(
            'start',
            help=(
                'Start SFTPPlus as a background '
                '(daemon) process.'),
            parents=[self._configuration_file],
            )
        start.add_argument(
            '-p', '--pid-file',
            action='store', dest='pid',
            metavar="FILE", default=CONFIGURATION_PID_FILE,
            help=('Store server process ID into FILE. Default: %s' % (
                CONFIGURATION_PID_FILE)),
            )

    def parseArguments(self, args=None):
        """
        Parse command line arguments.

        Return a namespace with parsed options.
        """
        if args is None:
            args = sys.argv[1:]

        # Convert all arguments to Unicode as this is an entry point.
        unicode_args = []
        for arg in args:
            if not isinstance(arg, six.text_type):
                arg = arg.decode('utf-8')
            unicode_args.append(arg)

        namespace = self._parser.parse_args(unicode_args)

        if namespace.action is None:
            # On Py2 the parser will raise this error.
            # Here on py3 we raise this to keep the same behaviour.
            self._parser.print_usage()
            self._parser.error('too few arguments')
            # We shouldn't hit this as Parser.error() should exit.
            return

        return namespace

    @property
    def server_module_segments(self):
        if self._server_module_segments is None:
            segments = local_filesystem.getSegmentsFromRealPath(MODULE_PATH)
            self._server_module_segments = segments
        return self._server_module_segments[:]

    def help(self):
        '''Print help message on standard output.'''
        self._parser.print_help()

    def subcommand_manager(self, options):
        """
        Does nothing as the command is handled via slim_server.
        """

    def subcommand_documentation(self, options):
        """
        Does nothing as the command is handled via slim_server.
        """

    def subcommand_generate_password(self, options):
        '''
        Launch the generation of a password.
        '''
        if not options.password:
            password_1 = getpass(b'Enter password   (not echoed):')
            password_2 = getpass(b'Confirm password (not echoed):')
            if password_1 != password_2:
                return (1, 'Passwords do not match.')
            options.password = password_1

        hash = password_handler.encrypt(
            options.password,
            rounds=self._PASSWORD_CRYPT_ROUNDS,
            format=options.format,
            )

        return (0, hash)

    def subcommand_start(self, options):
        """
        Called to start the Unix daemon.
        """
        from chevah.slim_server.pool import spawn_pool
        from supervisor.options import ServerOptions
        from supervisor.supervisord import go
        from chevah.server.commons.process import setproctitle

        class SupervisorOptions(ServerOptions):
            """
            Modified configuration for supervisord to not start a control
            server.
            """
            first = True
            test = False

            def openhttpservers(self, supervisord):
                """
                We don't run the HTTP server, as we try to use supervisord
                in an "embedded" mode.
                """
                return

        process_name = 'sftpplus-service-supervisor'
        setproctitle(process_name)

        # We can stop the processed, as they will be started again in
        # the forked process.
        spawn_pool.stop()

        config_content = _SUPERVISORD_INI % {
            'base': local_filesystem.getAbsoluteRealPath(''),
            'config': local_filesystem.getAbsoluteRealPath(options.config),
            'pid': local_filesystem.getAbsoluteRealPath(options.pid),
            }

        config = SupervisorOptions()
        if six.PY2:
            config.configfile = BytesIO(config_content.encode('utf-8'))
        else:
            config.configfile = StringIO(config_content)
        config.realize(args=[], progname=process_name)
        go(config)
        config.close_httpservers()
        config.close_logger()

        # This is not called as this is called in the fork.
        return (0, 'Done')  # noqa:cover

    def subcommand_start_in_foreground(self, options):
        """
        Caled to start the process without a fork.
        """
        return self._startProcess(options)

    def subcommand_debug(self, options):   # noqa:cover
        """
        Launch SFTPPlus in debug mode.
        """
        # We don't have automated tests for this
        # but this is alwasy called during development.
        out = 'DEBUG: %s started in debug mode.\n' % (PRODUCT_NAME,)
        if six.PY2:
            out = out.encode('utf-8')
        print(out)
        return self._startProcess(options, debug=True)

    def _startProcess(self, options, debug=False):

        # We might send signals multiple times, but only want to stop
        # once.
        self._process_stopping = False

        def stop_process(signum, frame):
            """
            Signal handler for keyboard interrupt and process termination.
            """
            if self._process_stopping:
                sig_names = {
                    0: b"SIG_DFL",
                    1: b"SIG_IGN",
                    2: b"SIGINT",
                    4: b"SIGILL",
                    8: b"SIGFPE",
                    11: b"SIGSEGV",
                    15: b"SIGTERM",
                    21: b"SIGBREAK",
                    22: b"SIGABRT",
                    23: b"NSIG",
                    }
                print(b'Already stopping. Ignoring %s' % sig_names[signum])
                return
            self._process_stopping = True
            process.stop()

        def exit(exit_code):
            """
            Hook the exit process to detect if we fail.
            """
            if exit_code != 0 and not debug:
                # Add process to internal function closure.
                process_closure = process
                process_closure.event_dispatcher.addStandardErrorHandler()
                # Use debug event 20000 to trigger the cached events.
                process_closure.event_dispatcher.emit(
                    '20000',
                    data={'message': (
                        'Failed to start SFTPPlus. '
                        'Check previous messages for the error details.'
                        )},
                    )
            os._exit(exit_code)

        # Register same handler for multiple signals.
        signal.signal(signal.SIGINT, stop_process)
        signal.signal(signal.SIGTERM, stop_process)

        if not debug:
            process = ChevahServerProcess(exit_callable=exit)
        else:
            process = ChevahServerProcess()
            # We add default log handlers as soon as possible.
            process.event_dispatcher.addStandardOutputHandler()

        process.configure(configuration_path=options.config)
        # This part is not called on configuration errors.
        process.start()
        # This part not called in production, since start() will trigger
        # the event look.
        return (0, '')  # noqa:cover

    def subcommand_initialize(self, options):
        """
        Initialize the server.
        """
        return self._initialize(options)

    def subcommand_migrate(self, options):
        """
        Only try to create configuration if not exists.
        """
        return self._initialize(options, migrate=True)

    def _initialize(self, options, migrate=False):
        """
        Create a new configuration file and write it to disk.

        Creates initial configuration files.
        Creates initial SSH keys.
        Creates self signed SSL certificate.

        Fails if configuration already exists.
        It can fail by raising an exception
        or returning a tuple with a non-zero first value.
        """
        server_ini_segments = _get_configuration_target('server.ini')
        if local_filesystem.exists(server_ini_segments):
            if not migrate:
                # This is a request for a new install,
                # but we already have the files.
                # Just fail.
                return (
                    1,
                    'File "configuration/server.ini" exists. '
                    'Not overwriting.',
                    )

            # We are doing a migration for an existing install.
            return self._migrate()

        admin_name, service_os_user = _validate_initialize_input(options)
        # Run the process as OS user
        # and don't switch to [server] -> account by default.
        server_account = ''
        os_process_user = service_os_user

        if options.drop_user:
            # To drop user, we start as root and then drop the user inside
            # SFTPPlus config.
            if process_capabilities.os_name in ['windows', 'osx']:
                # We only support OS user drop on Linux and AIX.
                raise ServerException('Drop user not supported on this OS.')

            server_account = service_os_user
            os_process_user = 'root'

        jinja_environment, data = self._getScriptsData(
            os_process_user=os_process_user,
            server_account=server_account,
            )

        code, message = self._generateScripts(jinja_environment, data)
        if code != 0:
            # There was an error generating the scripts.
            return (code, message)

        password = password_handler.encrypt(
            options.admin_password, rounds=self._PASSWORD_CRYPT_ROUNDS)

        rsa_key = Key.generate(
            key_type='rsa',
            key_size=options.key_size).toString(type='openssh_v1')
        ecdsa_sha2_nistp256_key = Key.generate(
            key_type='ecdsa', key_size=256).toString(type='openssh_v1')
        ecdsa_sha2_nistp384_key = Key.generate(
            key_type='ecdsa', key_size=384).toString(type='openssh_v1')
        ecdsa_sha2_nistp521_key = Key.generate(
            key_type='ecdsa', key_size=521).toString(type='openssh_v1')
        if OPENSSL_ED25519_SUPPORTED:
            ed25519_key = Key.generate(key_type='ed25519').toString(
                type='openssh_v1')
        else:
            ed25519_key = ''

        options.common_name = 'localhost'
        options.alternative_name = 'DNS:localhost,IP:127.0.0.1'
        options.key_usage = (
            'digital-signature, non-repudiation, key-encipherment,'
            'key-agreement, server-authentication, client-authentication'
            )
        ssl_certificate, ssl_key = generate_ssl_self_signed_certificate(
            options)

        data.update({
            'admin': {
                'name': admin_name,
                'password': password,
                },
            'rsa_key': _multi_line(rsa_key),
            'ecdsa_sha2_nistp256_key': _multi_line(ecdsa_sha2_nistp256_key),
            'ecdsa_sha2_nistp384_key': _multi_line(ecdsa_sha2_nistp384_key),
            'ecdsa_sha2_nistp521_key': _multi_line(ecdsa_sha2_nistp521_key),
            'ed25519_key': _multi_line(ed25519_key),
            'ssl_certificate': _multi_line(ssl_certificate),
            'ssl_key': _multi_line(ssl_key),
            })

        if options.local_admin_access:
            data['manager']['address'] = '127.0.0.1'

        stream = None
        try:
            stream = local_filesystem.openFileForWriting(server_ini_segments)
            template = jinja_environment.get_template('server.ini.j2')
            stream.write(template.render(**data).encode('utf-8'))
        except Exception as error:
            if stream:
                stream.close()
            return (1, 'Failed to write configuration. {}'.format(
                force_unicode(error)))

        return (0, 'SFTPPlus configuration successfully initialized.')

    def _migrate(self):
        """
        Called when the configuration file exists, and the migration is
        requested.

        Note that the configuration has separate migration process.
        This is for migrating non-configuration parts.
        """
        if self._os_name != 'nt':
            # On Linux and macOS, when we migrate, we leave everything as it is,
            # without recreating the init scripts.
            return (
                0,
                'File "configuration/server.ini" already exists.',
                )

        # This is a process to migrate.
        # On Windows we update only the service install
        # and uninstall script to help with Windows service migration.
        # On Windows, the scripts are not writing the service
        # account and password.
        jinja_environment, data = self._getScriptsData(
            os_process_user='not-used-on-win',
            server_account='not-used-on-win',
            )

        exit_code, message = self._generateScripts(jinja_environment, data)

        legacy_win_service_migration_message = self._migrateLegacyWinService(
            data)
        return exit_code, message + legacy_win_service_migration_message

    def _migrateLegacyWinService(self, data):
        """
        If needed, migrate the legacy PythonService.exe windows service to
        latest shawl, while keeping the configured username and password.
        """
        win32service = self._win32service
        service_manager = win32service.OpenSCManager(
            None,  # machineName
            None,  # dbName
            win32service.SC_MANAGER_ALL_ACCESS,  # desiredAccess
            )
        service = win32service.OpenService(
            service_manager,  # scHandle
            data['windows']['service']['id'],  # name
            win32service.SERVICE_ALL_ACCESS,  # desiredAccess
            )
        service_config = win32service.QueryServiceConfig(service)

        if 'pythonservice.exe' not in service_config[3].lower():
            # Nothing to migrate.
            return ' Not using legacy Windows service.'

        # Make sure base dir
        base_dir = self._os_getcwd()
        if not base_dir.startswith('\\\\?\\'):
            base_dir = '\\\\?\\' + base_dir
        base_dir.rstrip('\\')
        log_dir = os.path.join(base_dir, 'log')
        service_path = os.path.join(base_dir, data['windows']['service']['exe'])
        worker_path = os.path.join(base_dir, 'lib\\sftpplus-worker.exe')
        binary_path = (
            '{} run --name "{}" --stop-timeout 30000 --no-restart '
            '--cwd "{}" --log-dir "{}" -- '
            '"{}" -m chevah.server.__init__ start-in-foreground '
            '--config configuration\\server.ini'.format(
                service_path,
                data['windows']['service']['id'],
                base_dir,
                log_dir,
                worker_path,
                )
            )

        win32service.ChangeServiceConfig(
            service,
            win32service.SERVICE_NO_CHANGE,  # serviceType
            win32service.SERVICE_NO_CHANGE,  # startType
            win32service.SERVICE_NO_CHANGE,  # errorControl
            binary_path,  # binaryFile
            None,  # loadOrderGroup
            False,  # bFetchTag
            [],  # serviceDeps
            None,  # acctName
            None,  # password
            data['windows']['service']['name'],
            )
        return ' Legacy Windows service migrated to sftpplus-service-shawl.'

    def _getScriptsData(self, os_process_user, server_account):
        """
        Return the jinja environment and data that can be used to generate
        the scripts based on jinja templates.
        """
        jinja_environment = ChevahJinjaEnvironment(
            source=['configuration'],
            # we are generating a .INI here and we should be OK
            # not to have HTML escape codes.
            autoescape=True,
            )
        install_dir = local_filesystem.getRealPathFromSegments(
            local_filesystem.installation_segments)
        data = {
            'manager': {
                'address': '0.0.0.0',
                },
            'install_dir': install_dir,
            # Value set inside init scripts.
            'service_os_user': os_process_user,
            # Value set inside configuration.ini
            'server_account': server_account,
            'windows': {
                'service': {
                    'id': WINDOWS_SERVICE_NAME,
                    'name': WINDOWS_DISPLAY_NAME,
                    'exe': 'lib\\Scripts\\sftpplus-service-shawl.exe',
                    'base_dir': install_dir,
                    },
                },
            }
        return jinja_environment, data

    def _generateScripts(self, jinja_environment, data):
        """
        Write the scripts used to manage the process.
        """
        if self._os_name == 'nt':
            # Windows scripts.
            templates = {
                'service-install.bat.j2': ['service-install.bat'],
                'service-uninstall.bat.j2': ['service-uninstall.bat'],
                'service-start.bat.j2': ['service-start.bat'],
                }
        else:
            # Unix scripts.
            templates = {
                'sftpplus-mft.plist.j2': ['bin', 'sftpplus-mft.plist'],
                'sftpplus-mft.service.j2': ['bin', 'sftpplus-mft.service'],
                'sftpplus-mft.openrc.sh.j2': ['bin', 'sftpplus-mft.openrc.sh'],
                'sftpplus-mft.sysv.sh.j2': ['bin', 'sftpplus-mft.sysv.sh'],
                'sftpplus-mft.sysv.instances.sh.j2':
                    ['bin', 'sftpplus-mft.sysv.instances.sh'],
                }

        for key in sorted(templates.keys()):
            value = templates[key]
            stream = None
            destination_segments = (
                local_filesystem.installation_segments + value)
            try:
                stream = local_filesystem.openFileForWriting(
                    destination_segments)
                template = jinja_environment.get_template(key)
                result = template.render(**data)
                if process_capabilities.os_family == 'nt':
                    # Use Windows newlines.
                    result = result.replace('/n', '/r/n')

                stream.write(result.encode('utf-8'))
            except Exception as error:
                return (
                    1,
                    'Failed to write service file {}. {}'.format(
                        '/'.join(value), error)
                    )
            finally:
                if stream:
                    stream.close()

        return 0, 'Scripts generated with success.'

    def subcommand_generate_uuid(self, options):
        """
        Generate UUID version 4.
        """
        uuid_object = uuid.uuid4()

        return (0, six.text_type(uuid_object))

    def subcommand_validate(self, options):
        """
        Validate SFTPPlus configuration.
        """
        process = ChevahServerProcess()
        try:
            configuration = ApplicationConfiguration(
                configuration_path=options.config, parent=process)
            process.configuration = configuration
        except ServerException as error:
            return (1, error.message)
        except ServerError as error:
            exit_code = 1
            message = (
                'Could not load the configuration file. Error code '
                'is "%s". See documentation for more details.'
                ) % (error.id)
            return (exit_code, message)

        try:
            configuration.validateConfiguration()
        except ServerException as error:
            return (1, error.message)
        except ServerError as error:
            exit_code = 1
            message = (
                'Configuration is not valid. Error code '
                'is "%s". See documentation for more details.'
                ) % (error.id)
            return (exit_code, message)

        exit_code = 0
        message = 'Configuration is valid.'
        return (exit_code, message)

    def subcommand_help(self, options):
        """
        Called for `help` subcommand.
        """
        if options.help_action:
            sys.argv = [sys.argv[0], options.help_action, '-h']
        else:
            sys.argv = [sys.argv[0], '-h']

        self.parseArguments()
        # This is never reached as help will exit from argparse.
        return (0, '')

    def subcommand_generate_ssh_key(self, options):
        """
        Launch the generation of SSH key.
        """
        exit_code, message, key = crypto_generate_ssh_key(options)
        return exit_code, message

    def subcommand_generate_pgp_key(self, options):
        """
        Launch the generation of PGP key.
        """

        private_name = options.key_file
        if not private_name:
            private_name = 'pgp-key'
        public_name = private_name + '.pub'

        private_path = local_filesystem.getEncodedPath(private_name)
        public_path = local_filesystem.getEncodedPath(public_name)

        if os.path.exists(private_path):
            return (1, 'Private file already exists. Not overwriting.')

        if os.path.exists(public_path):
            return (1, 'Public file already exists. Not overwriting.')

        # Generate the file only if we can write the private path.
        try:
            result = generate_pgp_key(options)
        except ServerException as error:
            return (1, error.message)

        try:
            with open(private_path, 'wb') as stream:
                stream.write(result['private'].encode('utf-8'))

            with open(public_path, 'wb') as stream:
                stream.write(result['public'].encode('utf-8'))

            return (0, 'PGP key generated at "%s".' % private_name)
        except EnvironmentError as error:
            return (1, 'Failed to write PGP key. %s' % (force_unicode(error)))

    def subcommand_generate_csr(self, options):
        """
        Called for `generate-csr` subcommand.
        """
        try:
            if process_capabilities.os_family == 'nt':
                options.key_file = local_filesystem.getEncodedPath(
                    options.key_file)
            generate_and_store_csr(options)
            return (0, 'Key and CSR generated.')
        except KeyCertException as error:
            return (1, error.message)

    def subcommand_generate_self_signed(self, options):
        """
        Called for `generate-self-signed` subcommand.
        """
        try:
            ssl_certificate, ssl_key = generate_ssl_self_signed_certificate(
                options)
            return (0, ssl_certificate + ssl_key)
        except KeyCertException as error:
            return (1, error.message)


def _get_configuration_target(filename):
    """
    Return segments for `filename` as target inside the configuration dir.
    """
    segments = local_filesystem.installation_segments
    segments.append(CONFIGURATION_FOLDER_NAME)
    segments.append(filename)
    return segments


def _validate_initialize_input(options):
    """
    Check that usernames don't contain spaces or special characters.
    """
    # Can't start with number.
    # Can only contains ascii and numbers and _ and - and dot.
    valid_admin_re = '^[A-Za-z._][A-Za-z0-9 ._-]*$'
    valid_os_re = '^[A-Za-z._][A-Za-z0-9._-]*$'
    admin_name = options.admin_name.strip()
    service_os_user = options.service_os_user.strip()
    if not admin_name:
        raise ServerException('You need to provide an admin name.')

    if not service_os_user:
        raise ServerException('You need to provide an OS user name.')

    if len(admin_name) > 128:
        raise ServerException('Admin name is too long.')

    if not re.match(valid_admin_re, admin_name, re.UNICODE):
        raise ServerException(
            'Administrator username contains invalid characters.')

    if not re.match(valid_os_re, service_os_user):
        raise ServerException(
            'Operating system username contains invalid characters.')

    _validate_service_os_user(service_os_user)

    return (admin_name, service_os_user)


def _validate_service_os_user(service_os_user):
    """
    Extra validation for `service_os_user`,
    other than the generic name validation.
    """
    if process_capabilities.os_family != 'posix':
        # Only on Unix we do extra validation for now.
        return

    # We do the import after we check that we are not on Windows.
    import pwd
    import grp

    try:
        pwd.getpwnam(service_os_user)
        raise ServerException(
            'OS username "{}" already exists.\n'
            'Only dedicated accounts are supported by the installer.'.format(
                service_os_user),
            )
    except KeyError:
        """
        All ok as we expect this to fail when searching for user.
        """

    try:
        grp.getgrnam(service_os_user)
        raise ServerException(
            'OS group "{}" already exists.\n'
            'Only dedicated groups are supported by the installer.'.format(
                service_os_user),
            )
    except KeyError:
        """
        All ok as we expect this to fail when searching for group.
        """


def _start_browser_at_url(url, timeout=10):
    """
    Open the default browser at the specified `url` and wait for it to
    open at most `timeout` seconds.
    """
    exit_code = 0
    message = ''

    # There is someone listening on the URL so we can open the browser.
    result = webbrowser.open_new_tab(url)
    message = 'Opening the local manager from: %s' % (url,)
    if not result:
        exit_code = 2
        message = (
            'Failed to open the web browser at "%s". '
            'Please open the web page manually.'
            ) % (url,)

    return (exit_code, message)


def admin_commands():
    """
    Call the execute the administrative command using the command line
    arguments.
    """
    commands = Commands()
    options = commands.parseArguments()
    commands.executeSubcommand(options)
