# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
Common components used by Chevah Server.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from zope.component import getGlobalSiteManager
from zope.interface import implementer

from chevah.server.authentication.application import ApplicationAuthentication
from chevah.server.authentication.entra_id import (
    EntraIDAuthentication,
    EntraIDAuthenticationConfiguration,
    )
from chevah.server.authentication.configuration import (
    AnonymousAuthenticationConfiguration,
    ApplicationAuthenticationConfiguration,
    DenyUsernameAuthenticationConfiguration,
    IPTimeBanAuthenticationConfiguration,
    LocalFileAuthenticationConfiguration,
    )
from chevah.server.authentication.anonymous import AnonymousAuthentication
from chevah.server.authentication.deny_username import (
    DenyUsernameAuthentication,
    )
from chevah.server.authentication.http import (
    HTTPAuthentication,
    HTTPAuthenticationConfiguration,
    )
from chevah.server.authentication.ip_time_ban import IPTimeBanAuthentication
from chevah.server.authentication.local_file import LocalFileAuthentication
from chevah.server.authentication.ldap import (
    LDAPAuthentication,
    LDAPAuthenticationConfiguration,
    )
from chevah.server.authentication.os import (
    OSAuthentication,
    OSAuthenticationConfiguration,
    )
from chevah.server.authentication.radius import (
    RADIUSAuthenticationConfiguration,
    RADIUSAuthentication,
    )

from chevah.server.identity.configuration import (
    ApplicationAccountConfiguration,
    OSAccountConfiguration,
    ApplicationAdministratorConfiguration,
    GroupConfiguration,
    IGroupConfiguration,
    IRoleConfiguration,
    RoleConfiguration,
    )
from chevah.server.commons.constant import (
    CONFIGURATION_SECTION_AUTHENTICATIONS,
    _PROTOCOL_NAME,
    TYPE_NAME,
    )
from chevah.server.commons.interface import (
    IAccountConfiguration,
    IAuthentication,
    IAdministratorConfiguration,
    IAuthenticationConfiguration,
    IEventHandler,
    IEventHandlerConfiguration,
    ILocation,
    ILocationConfiguration,
    IServiceConfiguration,
    ITransferConfiguration,
    )
from chevah.server.commons.service_configuration import (
    ServiceConfigurationSection,
    )
from chevah.server.commons.runnable import DeprecatedRunnable
from chevah.server.configuration.option import (
    EntityConfiguration,
    TypeOptionsMixin,
    )
from chevah.server.event.configuration import (
    DatabaseEventHandlerConfiguration,
    DeprecatedEventHandlerConfiguration,
    ExtensionEventHandlerConfiguration,
    CreateArchiveEventHandlerConfiguration,
    ExtractArchiveEventHandlerConfiguration,
    LocalFileEventHandlerConfiguration,
    OpenPGPEventHandlerConfiguration,
    StandardStreamEventHandlerConfiguration,
    SyslogEventHandlerConfiguration,
    WindowsEventLogEventHandlerConfiguration,
    )

from chevah.server.event.account_interaction import (
    AccountInteractionEventHandler,
    AccountInteractionEventHandlerConfiguration,
    )
from chevah.server.event.deprecated import DeprecatedEventHandler
from chevah.server.event.email_sender import (
    EmailSenderEventHandler,
    EmailSenderEventHandlerConfiguration,
    )
from chevah.server.event.extension import ExtensionEventHandler
from chevah.server.event.create_archive import CreateArchiveEventHandler
from chevah.server.event.extract_archive import ExtractArchiveEventHandler
from chevah.server.event.external_executable import (
    ExternalExecutableEventHandlerConfiguration,
    ExternalExecutableEventHandler,
    )
from chevah.server.event.file_digest import (
    FileDigestEventHandler,
    FileDigestEventHandlerConfiguration,
    )
from chevah.server.event.file_dispatcher import (
    FileDispatcherEventHandler,
    FileDispatcherEventHandlerConfiguration,
    )
from chevah.server.event.http import (
    HTTPEventHandler,
    HTTPEventHandlerConfiguration,
    )
from chevah.server.event.local_file import LocalFileEventHandler
from chevah.server.event.openpgp import OpenPGPEventHandler
from chevah.server.event.rabbitmq import (
    RabbitMQEventHandler,
    RabbitMQEventHandlerConfiguration,
    )
from chevah.server.event.standard_stream import StandardStreamEventHandler
from chevah.server.event.syslog import SyslogEventHandler
from chevah.server.event.windows_eventlog import WindowsEventLogEventHandler
from chevah.server.event.database import DatabaseEventHandler

from chevah.server.location.configuration import (
    FilesystemLocationConfiguration,

    SFTPLocationConfiguration,
    )
from chevah.server.location.as2 import AS2Location, AS2LocationConfiguration
from chevah.server.location.filesystem import FilesystemLocation
from chevah.server.location.ftp import (
    FTPLocation,
    FTPSELocation,
    FTPSILocation,
    FTPLocationConfiguration,
    FTPSELocationConfiguration,
    FTPSILocationConfiguration,
    )
from chevah.server.location.http_pull import (
    HTTPPullLocation,
    HTTPPullLocationConfiguration,
    )
from chevah.server.location.sftp import SFTPLocation
from chevah.server.location.smb import SMBLocation, SMBLocationConfiguration
from chevah.server.location.webdav import (
    WebDAVSLocation,
    WebDAVSLocationConfiguration,
    )
from chevah.server.location.azure_blob import (
    AzureBLOBLocation,
    AzureBLOBLocationConfiguration,
    )
from chevah.server.location.azure_file import (
    AzureFileLocation,
    AzureFileLocationConfiguration,
    )

from chevah.server.resource.component import register_resources

from chevah.server.transfer.configuration import (
    TransferConfiguration,
    )


# Export this as part of interfaces module.
registry = getGlobalSiteManager()


@implementer(IAuthenticationConfiguration)
class DeprecatedAuthenticationConfiguration(
        EntityConfiguration, TypeOptionsMixin):
    """
    Placeholder configuration for a deprecated component.
    """

    SECTION_PREFIX = CONFIGURATION_SECTION_AUTHENTICATIONS

    deprecated = True

    @property
    def enabled(self):
        """
        Deprecated components are always disabled.
        """
        return False

    @enabled.setter
    def enabled(self, value):
        """
        Does nothing but we allow writing in order not to raise an error
        if someone tries to configure it.
        """


@implementer(IAuthentication)
class DeprecatedAuthentication(DeprecatedRunnable):
    """
    An authentication method that is no longer supported.
    """


def register():
    """
    Register all adapters and utilities.
    """

    register_resources()

    # Utilities
    registry.registerUtility(
        ApplicationAccountConfiguration,
        IAccountConfiguration,
        name=TYPE_NAME.APPLICATION,
        )

    registry.registerUtility(
        OSAccountConfiguration,
        IAccountConfiguration,
        name=TYPE_NAME.OS,
        )

    registry.registerUtility(
        ApplicationAdministratorConfiguration, IAdministratorConfiguration)
    registry.registerUtility(RoleConfiguration, IRoleConfiguration)

    registry.registerUtility(GroupConfiguration, IGroupConfiguration)

    registry.registerUtility(TransferConfiguration, ITransferConfiguration)

    # FIXME:2599:
    # This should be updated to register utilities with names
    # for each service type.
    for name in _PROTOCOL_NAME.keys():
        registry.registerUtility(
            ServiceConfigurationSection,
            IServiceConfiguration,
            name=name,
            )

    # Local filesystem location - configuration and runnable.
    registry.registerUtility(
        FilesystemLocationConfiguration,
        ILocationConfiguration,
        name=TYPE_NAME.FILESYSTEM,
        )
    registry.registerUtility(
        FilesystemLocation,
        ILocation,
        name=TYPE_NAME.FILESYSTEM,
        )
    # SFTP location - configuration and runnable.
    registry.registerUtility(
        SFTPLocationConfiguration,
        ILocationConfiguration,
        name=TYPE_NAME.SFTP,
        )
    registry.registerUtility(
        SFTPLocation,
        ILocation,
        name=TYPE_NAME.SFTP,
        )
    # FTP location - configuration and runnable.
    registry.registerUtility(
        FTPLocationConfiguration,
        ILocationConfiguration,
        name=TYPE_NAME.FTP,
        )
    registry.registerUtility(
        FTPLocation,
        ILocation,
        name=TYPE_NAME.FTP,
        )
    # FTPSE location - configuration and runnable.
    registry.registerUtility(
        FTPSELocationConfiguration,
        ILocationConfiguration,
        name=TYPE_NAME.FTPSE,
        )
    registry.registerUtility(
        FTPSELocation,
        ILocation,
        name=TYPE_NAME.FTPSE,
        )
    # FTPSI location - configuration and runnable.
    registry.registerUtility(
        FTPSILocationConfiguration,
        ILocationConfiguration,
        name=TYPE_NAME.FTPSI,
        )
    registry.registerUtility(
        FTPSILocation,
        ILocation,
        name=TYPE_NAME.FTPSI,
        )
    # WebDAVs location - configuration and runnable.
    registry.registerUtility(
        WebDAVSLocationConfiguration,
        ILocationConfiguration,
        name=TYPE_NAME.WEBDAVS,
        )
    registry.registerUtility(
        WebDAVSLocation,
        ILocation,
        name=TYPE_NAME.WEBDAVS,
        )
    # HTTP Pull location - configuration and runnable.
    registry.registerUtility(
        HTTPPullLocationConfiguration,
        ILocationConfiguration,
        name=TYPE_NAME.HTTP_PULL,
        )
    registry.registerUtility(
        HTTPPullLocation,
        ILocation,
        name=TYPE_NAME.HTTP_PULL,
        )
    # AS2 location - configuration and runnable.
    registry.registerUtility(
        AS2LocationConfiguration,
        ILocationConfiguration,
        name=TYPE_NAME.AS2,
        )
    registry.registerUtility(
        AS2Location,
        ILocation,
        name=TYPE_NAME.AS2,
        )
    # Azure BLOB location - configuration and runnable.
    registry.registerUtility(
        AzureBLOBLocationConfiguration,
        ILocationConfiguration,
        name=TYPE_NAME.AZURE_BLOB,
        )
    registry.registerUtility(
        AzureBLOBLocation,
        ILocation,
        name=TYPE_NAME.AZURE_BLOB,
        )
    # Azure File location - configuration and runnable.
    registry.registerUtility(
        AzureFileLocationConfiguration,
        ILocationConfiguration,
        name=TYPE_NAME.AZURE_FILE,
        )
    registry.registerUtility(
        AzureFileLocation,
        ILocation,
        name=TYPE_NAME.AZURE_FILE,
        )

    # SMB location - configuration and runnable.
    registry.registerUtility(
        SMBLocationConfiguration,
        ILocationConfiguration,
        name=TYPE_NAME.SMB,
        )
    registry.registerUtility(
        SMBLocation,
        ILocation,
        name=TYPE_NAME.SMB,
        )

    # HTTP authentication - configuration and runnable.
    registry.registerUtility(
        HTTPAuthenticationConfiguration,
        IAuthenticationConfiguration,
        name=TYPE_NAME.HTTP,
        )
    registry.registerUtility(
        HTTPAuthentication,
        IAuthentication,
        name=TYPE_NAME.HTTP,
        )

    # Deprecated SFTPPlus Legacy authentication - configuration and runnable.
    registry.registerUtility(
        DeprecatedAuthenticationConfiguration,
        IAuthenticationConfiguration,
        name=TYPE_NAME.LEGACY_WEBADMIN,
        )
    registry.registerUtility(
        DeprecatedAuthentication,
        IAuthentication,
        name=TYPE_NAME.LEGACY_WEBADMIN,
        )

    # LDAP authentication - configuration and runnable.
    registry.registerUtility(
        LDAPAuthenticationConfiguration,
        IAuthenticationConfiguration,
        name=TYPE_NAME.LDAP,
        )
    registry.registerUtility(
        LDAPAuthentication,
        IAuthentication,
        name=TYPE_NAME.LDAP,
        )
    # RADIUS authentication - configuration and runnable.
    registry.registerUtility(
        RADIUSAuthenticationConfiguration,
        IAuthenticationConfiguration,
        name=TYPE_NAME.RADIUS,
        )
    registry.registerUtility(
        RADIUSAuthentication,
        IAuthentication,
        name=TYPE_NAME.RADIUS,
        )
    # Local file authentication - configuration and runnable.
    registry.registerUtility(
        LocalFileAuthenticationConfiguration,
        IAuthenticationConfiguration,
        name=TYPE_NAME.LOCAL_FILE,
        )
    registry.registerUtility(
        LocalFileAuthentication,
        IAuthentication,
        name=TYPE_NAME.LOCAL_FILE,
        )
    # Deny username authentication - configuration and runnable.
    registry.registerUtility(
        DenyUsernameAuthenticationConfiguration,
        IAuthenticationConfiguration,
        name=TYPE_NAME.DENY_USERNAME,
        )
    registry.registerUtility(
        DenyUsernameAuthentication,
        IAuthentication,
        name=TYPE_NAME.DENY_USERNAME,
        )
    # IP time ban authentication - configuration and runnable.
    registry.registerUtility(
        IPTimeBanAuthenticationConfiguration,
        IAuthenticationConfiguration,
        name=TYPE_NAME.IP_TIME_BAN,
        )
    registry.registerUtility(
        IPTimeBanAuthentication,
        IAuthentication,
        name=TYPE_NAME.IP_TIME_BAN,
        )
    # Anonymous accounts authentication - config and runnable.
    registry.registerUtility(
        AnonymousAuthenticationConfiguration,
        IAuthenticationConfiguration,
        name=TYPE_NAME.ANONYMOUS,
        )
    registry.registerUtility(
        AnonymousAuthentication,
        IAuthentication,
        name=TYPE_NAME.ANONYMOUS,
        )
    # Application accounts authentication - config and runnable.
    registry.registerUtility(
        ApplicationAuthenticationConfiguration,
        IAuthenticationConfiguration,
        name=TYPE_NAME.APPLICATION,
        )
    registry.registerUtility(
        ApplicationAuthentication,
        IAuthentication,
        name=TYPE_NAME.APPLICATION,
        )
    # Entra ID authentication - config and runnable.
    registry.registerUtility(
        EntraIDAuthenticationConfiguration,
        IAuthenticationConfiguration,
        name=TYPE_NAME.ENTRA_ID,
        )
    registry.registerUtility(
        EntraIDAuthentication,
        IAuthentication,
        name=TYPE_NAME.ENTRA_ID,
        )
    # OS accounts authentication - config and runnable.
    registry.registerUtility(
        OSAuthenticationConfiguration,
        IAuthenticationConfiguration,
        name=TYPE_NAME.OS,
        )
    registry.registerUtility(
        OSAuthentication,
        IAuthentication,
        name=TYPE_NAME.OS,
        )

    # Amend content event handler was deprecated in v4.
    registry.registerUtility(
        DeprecatedEventHandlerConfiguration,
        IEventHandlerConfiguration,
        name=TYPE_NAME.AMEND_CONTENT,
        )
    registry.registerUtility(
        DeprecatedEventHandler,
        IEventHandler,
        name=TYPE_NAME.AMEND_CONTENT,
        )
    # Database event handler - configuration and runnable.
    registry.registerUtility(
        DatabaseEventHandlerConfiguration,
        IEventHandlerConfiguration,
        name=TYPE_NAME.DATABASE,
        )
    registry.registerUtility(
        DatabaseEventHandler,
        IEventHandler,
        name=TYPE_NAME.DATABASE,
        )
    # Account activity handler was deprecated in v4.
    registry.registerUtility(
        DeprecatedEventHandlerConfiguration,
        IEventHandlerConfiguration,
        name=TYPE_NAME.ACCOUNT_ACTIVITY,
        )
    registry.registerUtility(
        DeprecatedEventHandler,
        IEventHandler,
        name=TYPE_NAME.ACCOUNT_ACTIVITY,
        )
    # Account interaction event handler - configuration and runnable.
    registry.registerUtility(
        AccountInteractionEventHandlerConfiguration,
        IEventHandlerConfiguration,
        name=TYPE_NAME.ACCOUNT_INTERACTION,
        )
    registry.registerUtility(
        AccountInteractionEventHandler,
        IEventHandler,
        name=TYPE_NAME.ACCOUNT_INTERACTION,
        )
    # Digital signature validation was deprecated in v4.
    registry.registerUtility(
        DeprecatedEventHandlerConfiguration,
        IEventHandlerConfiguration,
        name=TYPE_NAME.DIGITAL_SIGNATURE_VALIDATION,
        )
    registry.registerUtility(
        DeprecatedEventHandler,
        IEventHandler,
        name=TYPE_NAME.DIGITAL_SIGNATURE_VALIDATION,
        )
    # Email sender event handler - configuration and runnable.
    registry.registerUtility(
        EmailSenderEventHandlerConfiguration,
        IEventHandlerConfiguration,
        name=TYPE_NAME.EMAIL_SENDER,
        )
    registry.registerUtility(
        EmailSenderEventHandler,
        IEventHandler,
        name=TYPE_NAME.EMAIL_SENDER,
        )
    # Extension event handler - configuration and runnable.
    registry.registerUtility(
        ExtensionEventHandlerConfiguration,
        IEventHandlerConfiguration,
        name=TYPE_NAME.EXTENSION,
        )
    registry.registerUtility(
        ExtensionEventHandler,
        IEventHandler,
        name=TYPE_NAME.EXTENSION,
        )
    # Create archive event handler - configuration and runnable.
    registry.registerUtility(
        CreateArchiveEventHandlerConfiguration,
        IEventHandlerConfiguration,
        name=TYPE_NAME.CREATE_ARCHIVE,
        )
    registry.registerUtility(
        CreateArchiveEventHandler,
        IEventHandler,
        name=TYPE_NAME.CREATE_ARCHIVE,
        )

    # Extract archive event handler - configuration and runnable.
    registry.registerUtility(
        ExtractArchiveEventHandlerConfiguration,
        IEventHandlerConfiguration,
        name=TYPE_NAME.EXTRACT_ARCHIVE,
        )
    registry.registerUtility(
        ExtractArchiveEventHandler,
        IEventHandler,
        name=TYPE_NAME.EXTRACT_ARCHIVE,
        )
    # OpenPGP event handler - configuration and runnable.
    registry.registerUtility(
        OpenPGPEventHandlerConfiguration,
        IEventHandlerConfiguration,
        name=TYPE_NAME.OPENPGP,
        )
    registry.registerUtility(
        OpenPGPEventHandler,
        IEventHandler,
        name=TYPE_NAME.OPENPGP,
        )
    # External executable event handler - configuration and runnable.
    registry.registerUtility(
        ExternalExecutableEventHandlerConfiguration,
        IEventHandlerConfiguration,
        name=TYPE_NAME.EXTERNAL_EXECUTABLE,
        )
    registry.registerUtility(
        ExternalExecutableEventHandler,
        IEventHandler,
        name=TYPE_NAME.EXTERNAL_EXECUTABLE,
        )
    # HTTP event handler - configuration and runnable.
    registry.registerUtility(
        HTTPEventHandlerConfiguration,
        IEventHandlerConfiguration,
        name=TYPE_NAME.HTTP,
        )
    registry.registerUtility(
        HTTPEventHandler,
        IEventHandler,
        name=TYPE_NAME.HTTP,
        )
    # RabbitMQ event handler - configuration and runnable.
    registry.registerUtility(
        RabbitMQEventHandlerConfiguration,
        IEventHandlerConfiguration,
        name=TYPE_NAME.RABBITMQ,
        )
    registry.registerUtility(
        RabbitMQEventHandler,
        IEventHandler,
        name=TYPE_NAME.RABBITMQ,
        )
    # File digest event handler - configuration and runnable.
    registry.registerUtility(
        FileDigestEventHandlerConfiguration,
        IEventHandlerConfiguration,
        name=TYPE_NAME.FILE_DIGEST,
        )
    registry.registerUtility(
        FileDigestEventHandler,
        IEventHandler,
        name=TYPE_NAME.FILE_DIGEST,
        )
    # File dispatcher event handler - configuration and runnable.
    registry.registerUtility(
        FileDispatcherEventHandlerConfiguration,
        IEventHandlerConfiguration,
        name=TYPE_NAME.FILE_DISPATCHER,
        )
    registry.registerUtility(
        FileDispatcherEventHandler,
        IEventHandler,
        name=TYPE_NAME.FILE_DISPATCHER,
        )
    # Local file event handler - configuration and runnable.
    registry.registerUtility(
        LocalFileEventHandlerConfiguration,
        IEventHandlerConfiguration,
        name=TYPE_NAME.LOCAL_FILE,
        )
    registry.registerUtility(
        LocalFileEventHandler,
        IEventHandler,
        name=TYPE_NAME.LOCAL_FILE,
        )

    # Standard stream event handler - configuration and runnable.
    registry.registerUtility(
        StandardStreamEventHandlerConfiguration,
        IEventHandlerConfiguration,
        name=TYPE_NAME.STANDARD_STREAM,
        )
    registry.registerUtility(
        StandardStreamEventHandler,
        IEventHandler,
        name=TYPE_NAME.STANDARD_STREAM,
        )

    # Syslog event handler - configuration and runnable.
    registry.registerUtility(
        SyslogEventHandlerConfiguration,
        IEventHandlerConfiguration,
        name=TYPE_NAME.SYSLOG,
        )
    registry.registerUtility(
        SyslogEventHandler,
        IEventHandler,
        name=TYPE_NAME.SYSLOG,
        )
    # Windows Event Log event handler - configuration and runnable.
    registry.registerUtility(
        WindowsEventLogEventHandlerConfiguration,
        IEventHandlerConfiguration,
        name=TYPE_NAME.WINDOWS_EVENTLOG,
        )
    registry.registerUtility(
        WindowsEventLogEventHandler,
        IEventHandler,
        name=TYPE_NAME.WINDOWS_EVENTLOG,
        )
