# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
Configuration shared by all services.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
from six.moves.urllib.parse import parse_qs, urlparse
import os

from twisted.internet import reactor
from zope.interface import implementer

from chevah.server.authentication.configuration import (
    AuthenticationsConfiguration,
    )
from chevah.server.commons.constant import (
    CONFIGURATION_DEFAULTS,
    CONFIGURATION_DEFAULT_AUTHENTICATION,
    CONFIGURATION_DEFAULT_GROUP,
    CONFIGURATION_SECTION_AUTHENTICATIONS,
    CONFIGURATION_SECTION_RESOURCES,
    CONFIGURATION_SECTION_SERVER,
    CONFIGURATION_SECTION_SERVICES,
    DEFAULT_EMAIL_CLIENT_UUID,
    DEFAULT_LETS_ENCRYPT_UUID,
    DEFAULT_ANALYTICS_UUID,
    DEFAULT_SQLITE_DB_UUID,
    DEFAULT_UUIDS,
    EVENTS_DB,
    PRODUCT_NAME,
    TYPE_NAME,
    )
from chevah.server.commons.exception import ServerException
from chevah.server.commons.interface import (
    IProtocolOptions,
    IApplicationConfiguration,
    IServerOptionsSection,
    )
from chevah.server.configuration.file import FileConfigurationProxy
from chevah.server.configuration.model import (
    PropertyMixin,
    WritableInteger,
    WritableString,
    WritableStringOrEmpty,
    WritableStringCaseInsensitive,
    WritableCSVUnique,
    is_inherited,
    )
from chevah.server.commons.observer import Signal
from chevah.server.configuration.option import (
    EntityConfiguration,
    )
from chevah.server.commons.service_configuration import (
    ServicesConfigurationSection,
    )
from chevah.server.event.configuration import EventHandlersConfiguration
from chevah.server.event.model import EventsDefinition
from chevah.server.identity.model import (
    AccountsConfiguration,
    AdministratorsConfiguration,
    GroupsConfiguration,
    RolesConfiguration,
    )
from chevah.server.transfer.configuration import TransfersConfiguration
from chevah.server.resource.configuration import ResourcesConfiguration
import six


_DEFAULT_PORTS = {
    'amqp': 5672,
    'amqps': 5671,
    'ftp': 21,
    'ftpsi': 990,
    'ftpse': 21,
    'sftp': 22,
    'webdavs': 443,
    'http': 80,
    'https': 443,
    'smb': 445,
    }


def parse_url(raw_url):
    """
    Parse the `raw_url` and return a dict with its parts.
    """
    if '://' not in raw_url:
        raise ServerException('No protocol specified in the URL.')

    url = urlparse(raw_url)
    protocol = url.scheme
    netloc = url.netloc

    if not netloc and protocol not in ['file', 'unix']:
        raise ServerException('No hostname specified in the URL.')

    try:
        port = url.port
    except ValueError:
        raise ServerException('Invalid port number for "%s".' % (raw_url,))

    if not port:
        port = _DEFAULT_PORTS.get(protocol, 0)

    username = url.username if url.username else ''

    return {
        'protocol': protocol,
        'location': url.netloc,
        'address': url.hostname,
        'port': port,
        'username': username,
        'path':  url.path,
        'query': parse_qs(url.query)
        }


@implementer(IApplicationConfiguration)
class ApplicationConfiguration(PropertyMixin):
    """
    Main configuration class for the server.

    Signals:
    `save` - called when configuration is saved.
    """

    _DEFAULT_SECTIONS = [CONFIGURATION_SECTION_SERVER]

    def __init__(
        self, configuration_path=None, configuration_file=None,
        defaults=None, parent=None,
            ):
        """
        Initialize the configuration object.
        """
        if defaults is None:
            defaults = CONFIGURATION_DEFAULTS

        if (
            not (
                (configuration_path and not configuration_file) or
                (not configuration_path and configuration_file))
                ):
            raise RuntimeError(
                'Configuration objects must be initialized with either '
                'a configuration path or a configuration file stream.')

        self._proxy = None
        self._parent = parent

        self.server = None
        self.events = None

        try:
            # Load local main configuration file.
            self._proxy = FileConfigurationProxy(
                configuration_path=configuration_path,
                configuration_file=configuration_file,
                defaults=defaults)
            self._proxy.load()
        except ServerException as error:
            # Add more details to the existing error.
            message = (
                'Failed to load the main configuration file '
                'for %(path)s. %(details)s') % {
                    'path': configuration_path,
                    'details': error.message,
                    }
            raise ServerException(message)

        # Create default section if missing from loaded configuration file.
        for section in self._DEFAULT_SECTIONS:
            if not self._proxy.hasSection(section):
                self._proxy.addSection(section)

        self.server = ServerConfigurationSection(self)

        self.authentications = AuthenticationsConfiguration(self)
        self.event_handlers = EventHandlersConfiguration(self)
        self.services = ServicesConfigurationSection(self)
        self.transfers = TransfersConfiguration(self)

        self.accounts = AccountsConfiguration(self)
        self.groups = GroupsConfiguration(self)
        self.administrators = AdministratorsConfiguration(self)
        self.roles = RolesConfiguration(self)

        # Avoid a circular import that is observed only on Windows.
        from chevah.server.location.configuration import LocationsConfiguration
        self.locations = LocationsConfiguration(self)
        self.resources = ResourcesConfiguration(self)
        self.node_variables = NodeConfiguration(self)

    @property
    def parent(self):
        """
        The parent of this node.
        """
        return self._parent

    @parent.setter
    def parent(self, value):
        self._parent = value

    @property
    def reactor(self):
        """
        The parent of this node.
        """
        parent_reactor = self.parent.reactor
        if parent_reactor:
            return parent_reactor
        else:
            return reactor

    def save(self):
        """
        Persist the configuration.
        """
        self._proxy.save()
        self.notify('save')

    def validateConfiguration(self):
        """
        Checks that current configuration is in a consistent state
        and can be used to operate the process.

        Also creates any default component or section that is required
        for minimal process operation.

        Raises an error when the configuration is in an unusable state.
        Returns if everything is OK.
        """

        # This needs to be called first, as we are changing the type of
        # an authentication method.
        have_migrated_auths = self._migrateAuthentications()

        have_migrated_components = self._createDefaulComponents()
        have_migrated_handlers = self._migrate3to4EventHandlers()
        have_migrated_accounts = self._migrateAccounts()

        have_migrated_roles = self._migrateRoles()
        have_migrated_monitors = self._migrateMonitors()

        # FIXME:6065:
        # Update the implementation to save configuration
        # when things are migrated.
        self._migrateSourceIPFilter()
        self._migrateManagerAuthentications()

        if (
            have_migrated_components
            or have_migrated_handlers
            or have_migrated_accounts
            or have_migrated_roles
            or have_migrated_monitors
            or have_migrated_auths
                ):
            self._proxy.save()

        # Request reading all the configuration tree.
        # This should raise errors that otherwise would be raised later at
        # runtime.
        self.getProperty()

    def loadExternalConfiguration(self):
        """
        Load the content of all external configurations.
        """
        self.events = EventsDefinition()
        # Also update the constants.
        EVENTS_DB.update(self.events.data)

    def findAllReferences(self, uuid):
        """
        Return a list of tuples with the full section name (with UUID) and
        the name of configurations referring to `uuid`.
        """
        result = []

        def has_reference(uuid, value):
            """
            Return true if `value` contains a reference to `target`.
            """
            if isinstance(value, (str, six.text_type)) and value == uuid:
                return True

            if isinstance(value, (list, tuple)):
                for candinate in value:
                    result = has_reference(uuid, candinate)
                    if result:
                        return True

            return False

        def find_reference(target):
            for section in target.getAll():
                find_reference(target=section)

            for name in target.getPublicAttributeNames():
                if name == 'uuid':
                    # We don't look at UUID,
                    # since references should not be used in uuid.
                    # `uuid` is for self identification and reference,
                    # not for referencing other components.
                    continue
                value = target.getValue(name)

                if has_reference(uuid, value):

                    if IApplicationConfiguration.providedBy(target.parent):
                        # Parent is the root and it has no SECTION_PREFIX
                        parent_name = 'server'
                        conflict_source = target
                    else:
                        # Look how the parent is named in its parent and use
                        # that name.
                        if IProtocolOptions.providedBy(target):
                            # Services has configuration/configuration
                            conflict_source = target.parent
                            parent = target.parent.parent
                        else:
                            conflict_source = target
                            parent = target.parent

                        parent_parent = parent.parent

                        for child in parent_parent.getPublicSectionNames():
                            if parent is parent_parent.getSection(child):
                                parent_name = child
                                break

                    target_reference = '%s/%s' % (
                        parent_name, conflict_source.uuid)
                    result.append((conflict_source.name, target_reference))

        find_reference(target=self)

        return result

    def _createDefaulComponents(self):
        """
        Create default components.
        """
        authentication = self._createDefaulAuthentication()
        group = self._createDefaulGroup()
        db = self._createDefaulDB()
        email = self._createDefaultEmail()
        monitor = self._createAnalyticsResource()
        lets_encrypt = self._createLetsEncryptClient()

        return (
            authentication
            or group
            or db
            or email
            or monitor
            or lets_encrypt
            )

    def _migrateSourceIPFilter(self):
        """
        Auto-migrate the source_ip_filter which was single line CSV in
        version 4.21.0 and now is multi line space separated.
        """
        for account in self.accounts.getAll():
            self._migrateSourceIPFilterFor(account)

        for group in self.groups.getAll():
            self._migrateSourceIPFilterFor(group)

        for admin in self.administrators.getAll():
            self._migrateSourceIPFilterFor(admin)

        for role in self.roles.getAll():
            self._migrateSourceIPFilterFor(role)

    def _migrateSourceIPFilterFor(self, target):
        """
        Migrate source_ip_filter from single CSV to multi line space separated
        value.
        """
        lines = target.source_ip_filter

        if len(lines) != 1:
            # Legacy is a single row,
            # It looks this is not a legacy value, so don't bother with the
            # migration.
            return

        values = lines[0]

        if is_inherited(values):
            # This is an inherited value.
            # Remove inheritance as it's implicit for admins/accounts and
            # not valid for groups/roles.
            target.source_ip_filter = []
            return

        if len(values) == 1:
            # We have a single legacy IP address.
            target.source_ip_filter = [[TYPE_NAME.ALLOW, values[0]]]
            return

        should_migrate = False
        for value in values:
            if ',' in value:
                should_migrate = True
                break

        if not should_migrate:
            return

        migrated = []
        for value in values:
            value = value.strip(',')
            if not value:
                continue
            migrated.append([TYPE_NAME.ALLOW, value])

        target.source_ip_filter = migrated

    # FIXME:6521:
    # This can be removed in V6.
    # If this is a big issue, it can also be removed later in v5.
    def _migrateManagerAuthentications(self):
        """
        Move [server] -> manager_authentications to the default manager
        service.
        """
        to_migrate = self.server.getToMigrateAuthentication()
        if not to_migrate:
            return

        for service in self.services.getAll():
            if service.type not in [
                    TYPE_NAME.LOCAL_MANAGER, TYPE_NAME.LOCAL_MANAGER_UNSECURED]:
                # Not a manager service.
                # Nothing to migration.
                continue

            if service.authentications:
                # We already have custom authentications
                continue

            service.authentications = to_migrate

    def _createDefaulGroup(self):
        """
        Create default group if missing.
        """
        have_default = False
        for group in self.groups.getAll():
            if group.uuid == CONFIGURATION_DEFAULT_GROUP:
                have_default = True
                break

        if have_default:
            return

        self._proxy.addSection('groups/%s' % CONFIGURATION_DEFAULT_GROUP)
        group = self.groups.getSection(
            CONFIGURATION_DEFAULT_GROUP)
        group.name = 'DEFAULT_GROUP'
        group.destination = (
            'This is automatically associated to all accounts '
            '(application or operating system).'
            )
        group.enabled = True
        group.create_home_folder = False
        group.home_folder_path = '${OS_HOME}'
        group.allow_own_password_change = False

        self.groups.notify(
            'create',
            Signal(
                self.groups,
                uuid=CONFIGURATION_DEFAULT_GROUP,
                ),
            )
        return True

    def _createDefaulAuthentication(self):
        """
        Create default authentication if is missing.

        This is backward compatible with the old `app-uuid`.
        """
        have_default_auth = False
        for auth in self.authentications.getAll():
            if auth.uuid in ['app-uuid', CONFIGURATION_DEFAULT_AUTHENTICATION]:
                have_default_auth = True
                break

        if have_default_auth:
            return

        self._proxy.addSection(
            'authentications/%s' % CONFIGURATION_DEFAULT_AUTHENTICATION)
        self._proxy.setString(
            'authentications/%s' % CONFIGURATION_DEFAULT_AUTHENTICATION,
            'type',
            TYPE_NAME.APPLICATION,
            )
        auth_config = self.authentications.getSection(
            CONFIGURATION_DEFAULT_AUTHENTICATION)
        auth_config.name = 'SFTPPlus Accounts and Administrators'
        auth_config.destination = (
            'This authentication method allows authentication of accounts '
            'and administrators defined in STPPlus\'s own configuration file.'
            )
        auth_config.enabled = True
        self.authentications.notify(
            'create',
            Signal(
                self.authentications,
                uuid=CONFIGURATION_DEFAULT_AUTHENTICATION,
                ),
            )
        return True

    def _createDefaulDB(self):
        """
        Create default DB resource if missing.
        """
        have_default = False
        for resource in self.resources.getAll():
            if resource.uuid == DEFAULT_SQLITE_DB_UUID:
                have_default = True
                break

        if have_default:
            return False

        self._proxy.addSection('resources/%s' % DEFAULT_SQLITE_DB_UUID)
        self._proxy.setString(
            'resources/%s' % DEFAULT_SQLITE_DB_UUID,
            'type',
            TYPE_NAME.SQLITE,
            )
        db = self.resources.getSection(DEFAULT_SQLITE_DB_UUID)
        db.name = 'Embedded DB'
        db.destination = 'Embedded database using sqlite.'
        db.path = 'configuration/cache.db3'

        self.resources.notify(
            'create',
            Signal(
                self.resources,
                uuid=DEFAULT_SQLITE_DB_UUID,
                ),
            )
        return True

    def _createDefaultEmail(self):
        """
        Create default email resource if missing.

        Or migrate the first email client, if create under different UUID.
        """
        have_default = False
        configuration = {
            'name': 'Email-Client',
            'description': (
                'SMTP client configuration used to send email. '
                'Update the address and port to match your infrastructure.'
                ),
            'email_from_address': 'sftpplus@example.com',
            'email_to_recipients': ['admin-team@example.com'],
            'address': 'smtp.example.com',
            'port': 25,
            'username': '',
            'password_update': '',
            }

        for resource in self.resources.getAll():
            if resource.uuid == DEFAULT_EMAIL_CLIENT_UUID:
                have_default = True
                break
            if resource.type == TYPE_NAME.EMAIL_CLIENT:
                # Copy configuration from existing resource.
                for key in list(configuration.keys()):
                    if key == 'password_update':
                        resource_key = 'password'
                    else:
                        resource_key = key
                    configuration[key] = getattr(resource, resource_key)
                resource.name = resource.name + ' (IGNORED)'

        if have_default:
            return False

        self._proxy.addSection('resources/%s' % DEFAULT_EMAIL_CLIENT_UUID)
        self._proxy.setString(
            'resources/%s' % DEFAULT_EMAIL_CLIENT_UUID,
            'type',
            TYPE_NAME.EMAIL_CLIENT,
            )
        email = self.resources.getSection(DEFAULT_EMAIL_CLIENT_UUID)
        for key, value in configuration.items():
            setattr(email, key, value)

        self.resources.notify(
            'create',
            Signal(
                self.resources,
                uuid=DEFAULT_EMAIL_CLIENT_UUID,
                ),
            )
        return True

    def _createAnalyticsResource(self):
        """
        Create analytics resource if missing.

        Or migrate the first process-monitor, if create under different UUID.
        """
        have_default = False
        previous_uuid = None
        configuration = {
            'enabled': True,
            'name': 'Analytics Engine',
            'description': 'Monitor own activity and resource usage.',
            'monitor_interval': 600,
            'memory_resident_trigger': 0,
            'memory_virtual_trigger': 0,
            'connection_count_trigger': 0,
            'file_count_trigger': 0,
            'thread_count_trigger': 0,
            }
        migration_keys = [
            'name',
            'description',
            'monitor_interval',
            'memory_resident_trigger',
            'memory_virtual_trigger',
            'connection_count_trigger',
            'file_count_trigger',
            'thread_count_trigger',
            ]

        for resource in self.resources.getAll():
            if resource.uuid == DEFAULT_ANALYTICS_UUID:
                have_default = True
                break
            if resource.type == TYPE_NAME.PROCESS_MONITOR:
                # Copy configuration from existing resource.
                for key in migration_keys:
                    configuration[key] = getattr(resource, key)
                previous_uuid = resource.uuid

        if have_default:
            return False

        self._proxy.addSection('resources/%s' % DEFAULT_ANALYTICS_UUID)
        self._proxy.setString(
            'resources/%s' % DEFAULT_ANALYTICS_UUID,
            'type',
            TYPE_NAME.ANALYTICS,
            )
        analytics = self.resources.getSection(DEFAULT_ANALYTICS_UUID)
        for key, value in configuration.items():
            if key == 'name':
                # We get duplicated named during the migration since
                # we first create the duplicate migrated resource,
                # and later delete the older one.
                analytics._setName(value)
                continue
            setattr(analytics, key, value)

        self.resources.notify(
            'create',
            Signal(
                self.resources,
                uuid=DEFAULT_ANALYTICS_UUID,
                ),
            )
        if previous_uuid:
            self.resources.deleteProperty(previous_uuid)
        return True

    def _createLetsEncryptClient(self):
        """
        Create lets encrypt resource if missing.

        Or migrate the first lets encrypt, if create under different UUID.
        """
        have_default = False
        previous_uuid = None
        configuration = {
            'enabled': False,
            'name': 'Lets-Encrypt-Client',
            'description': (
                'Component handling the generation and renewal of '
                'Let\'s Encrypt certificates'
                ),
            'address': '0.0.0.0',
            'port': 80,
            'acme_url': 'https://acme-v02.api.letsencrypt.org/directory',
            }

        for resource in self.resources.getAll():
            if resource.uuid == DEFAULT_LETS_ENCRYPT_UUID:
                have_default = True
                break
            if resource.type == TYPE_NAME.LETS_ENCRYPT:
                # Copy configuration from existing resource.
                for key in list(configuration.keys()):
                    configuration[key] = getattr(resource, key)
                previous_uuid = resource.uuid

        if have_default:
            return False

        self._proxy.addSection('resources/%s' % DEFAULT_LETS_ENCRYPT_UUID)
        self._proxy.setString(
            'resources/%s' % DEFAULT_LETS_ENCRYPT_UUID,
            'type',
            TYPE_NAME.LETS_ENCRYPT,
            )

        resource = self.resources.getSection(DEFAULT_LETS_ENCRYPT_UUID)
        for key, value in configuration.items():
            if key == 'name':
                # We get duplicated named during the migration since
                # we first create the duplicate migrated resource,
                # and later delete the older one.
                resource._setName(value)
                continue
            setattr(resource, key, value)

        self.resources.notify(
            'create',
            Signal(
                self.resources,
                uuid=DEFAULT_LETS_ENCRYPT_UUID,
                ),
            )
        if previous_uuid:
            self.resources.deleteProperty(previous_uuid)
        return True

    def _migrate3to4EventHandlers(self):
        """
        Update the event handlers from v3 configuration.
        """
        default_handler = None
        for handler in self.event_handlers.getAll():
            if handler.type != TYPE_NAME.DATABASE:
                # For now, only migrate the DB event handler.
                continue

            if not default_handler:
                default_handler = handler

            if handler.uuid != 'default-database':
                # Not the default DB handler.
                continue

            if not handler.path:
                handler.path = 'log/server.db3'
                return True
            else:
                # Already has a path.
                # Nothing to migrate.
                return False

        if not default_handler:
            # No database handler defined in this configuration.
            # Nothing to migrate.
            return False

        # Here we don't have the default database handler, but another
        # DB handler.
        if default_handler.path:
            # Default handler already has a path.
            # Nothing to migrate.
            return False

        default_handler.path = 'log/server.db3'
        default_handler.name = '[Auto Migrated] ' + default_handler.name
        return True

    def _migrateAuthentications(self):
        """
        Update authentication methods.
        """
        have_migration = False
        for section_name in self._proxy.sections:
            if not section_name.startswith(
                    CONFIGURATION_SECTION_AUTHENTICATIONS):
                continue
            auth_type = self._proxy.getString(section_name, 'type').lower()
            if auth_type != TYPE_NAME.AZURE_AD:
                # Not a legacy Azure AD
                continue

            # Set the type to the new Entra ID.
            self._proxy.setString(section_name, 'type', TYPE_NAME.ENTRA_ID)
            have_migration = True

        return have_migration

    def _migrateAccounts(self):
        """
        Update account configuration from a previous version
        to latest configuration version.

        Returns True if there were changes to accounts configuration.
        """
        have_migration = False
        for account in self.accounts.getAll():
            # Last password update timestamp was moved to the cache file.
            last_password_update = account.last_password_update_v3
            if not last_password_update:
                continue
            account.last_password_update = last_password_update
            have_migration = True
        return have_migration

    def _migrateRoles(self):
        """
        Update roles configuration from a previous version
        to latest configuration version.

        Returns True if there were changes to roles configuration.
        """
        def migrate_target(target):
            runnable_family = [
                'services',
                'authentications',
                'resources',
                'transfers',
                'locations',
                'event_handlers',
                ]
            target = target.strip('/')
            for family in runnable_family:
                if target.startswith(family):
                    return True, 'operation/' + target

            return False, target

        have_migration = False
        for role in self.roles.getAll():
            # Check if permissions should be prefixed.
            migrated_permissions = []
            have_migrate_permissions = False
            for permission in role.permissions:
                have_migrated, migrated_target = migrate_target(
                    permission[0])
                permission = (migrated_target,) + permission[1:]
                have_migrate_permissions = (
                    have_migrate_permissions or have_migrated)
                migrated_permissions.append(permission)

            if have_migrate_permissions:
                role.permissions = migrated_permissions
                have_migration = True

        return have_migration

    def _migrateMonitors(self):
        """
        Update monitor configuration from a previous version
        in which they were services, to the current version in they are
        resources.

        Returns True if there were changes to monitors configuration.
        """
        have_migration = False
        for section_name in self._proxy.sections:
            if not section_name.startswith(CONFIGURATION_SECTION_SERVICES):
                continue
            service_type = self._proxy.getString(section_name, 'type')
            if service_type != TYPE_NAME.MONITOR:
                continue
            uuid = section_name[len(CONFIGURATION_SECTION_SERVICES):]
            have_migration = True
            monitor_section = CONFIGURATION_SECTION_RESOURCES + uuid
            self._proxy.renameSection(section_name, monitor_section)

        return have_migration


@implementer(IServerOptionsSection)
class ServerConfigurationSection(EntityConfiguration):
    """
    Configurations for the server section.

    It also contains the 'enabled' property for consistency, but right now
    it does nothing.
    """
    _section_name = 'server'

    # We overwrite name from default INI handling as for server section
    # we don't need to enforce unique names.
    name = WritableString('name', default=PRODUCT_NAME)
    account = WritableString('account', default='')
    authentications = WritableCSVUnique(
        'authentications', default=(CONFIGURATION_DEFAULT_AUTHENTICATION,))
    password_minimum_strength = WritableInteger(
        'password_minimum_strength', default=4)
    password_minimum_length = WritableInteger(
        'password_minimum_length', default=8)
    password_history = WritableInteger('password_history', default=0)
    password_hashing_scheme = WritableStringCaseInsensitive(
        'password_hashing_scheme')
    product_key = WritableStringCaseInsensitive('product_key', default='')
    ssl_certificate = WritableStringOrEmpty('ssl_certificate')
    ssl_key = WritableStringOrEmpty('ssl_key')

    def __init__(self, parent):
        self._parent = parent
        self._normalize()

    def _normalize(self):
        """
        This is here to convert from raw low-level configuration value to
        simplified values that are used inside the app.

        This is some sort of automatic/transport recovery from invalid options.
        """
        if self._proxy.isDisabledValue(self.account):
            # Backward compatibilliy.
            self.account = ''

        if os.name == 'nt' and self.account != '':
            # On Windows, we don't have a separate service account.
            self.account = ''

        if self.product_key not in ['', 'ask', 'client-only', 'server-only']:
            # For now, any uknown valule is handled as 'ask'
            self.product_key = 'ask'

    def getToMigrateAuthentication(self):
        """
        Return the list of legacy `manager_authentications` options
        and clear the list.

        Here to help with the migration of the general service configuration.
        """
        if not self._proxy.hasOption(
                self._section_name, 'manager_authentications'):
            return []

        value = self._proxy.getString(
            self._section_name, 'manager_authentications')
        self._proxy.removeOption(
            self._section_name, 'manager_authentications')
        return [v.strip() for v in value.split(',') if v.strip()]

    @property
    def default_uuids(self):
        """
        See: `IServerOptionsSection`.

        This is here as a helper for components that shouldn't be removed.
        """
        return DEFAULT_UUIDS

    @property
    def _proxy(self):
        """
        Here to help reducing circular references.
        """
        return self._parent._proxy

    @property
    def uuid(self):
        """
        UUID for the server.
        """
        return self._proxy.getString(self._section_name, 'uuid')

    def _setUUID(self, value):
        """
        Helper for testing since server UUID can not be change.
        """
        return self._proxy.setString(self._section_name, 'uuid', value)

    @property
    def umask(self):
        """
        See: `IServerOptionsSection`.
        """
        umask_string = self._proxy.getString(self._section_name, 'umask')
        self._validateUmaskString(umask_string)
        return umask_string

    @property
    def umask_as_int(self):
        """
        See: `IServerOptionsSection`.
        """
        return int(self.umask, 8)

    @umask.setter
    def umask(self, value):
        self._validateUmaskString(value)
        self._proxy.setString(self._section_name, 'umask', value)

    def _validateUmaskString(self, value):
        """
        Raise an configuration error if umask value as string is not valid.
        """
        try:
            umask_int = int(value, 8)
        except ValueError:
            raise ServerException('Not a valid octal value for umask.')

        if umask_int < 0 or umask_int > int('777', 8):
            raise ServerException('Not a valid umask value.')

    @property
    def ssh_host_private_keys(self):
        """
        Return the direct `ssh_host_private_keys` or fallback to deprecated
        RSA and DSA keys.
        """
        value = self._proxy.getString(
            self._section_name, 'ssh_host_private_keys')

        if value:
            # We have a direct configuration.
            return value

        # FIXME:6521:
        # Should be removed in v6.
        rsa_key = self._proxy.getString(self._section_name, 'rsa_private_key')
        dsa_key = self._proxy.getString(self._section_name, 'dsa_private_key')

        keys = []
        if rsa_key:
            keys.append(rsa_key)
        if dsa_key:
            keys.append(dsa_key)

        return '\n\n'.join(keys)

    @ssh_host_private_keys.setter
    def ssh_host_private_keys(self, value):
        """
        Set value and remove deprecated RSA and DSA
        """
        self._proxy.setStringOrNone(
            self._section_name, 'ssh_host_private_keys', value)

        # FIXME:6521:
        # Should be removed in v6.
        self._proxy.removeOption(self._section_name, 'rsa_private_key')
        self._proxy.removeOption(
            self._section_name, 'rsa_private_key_password')
        self._proxy.removeOption(self._section_name, 'dsa_private_key')
        self._proxy.removeOption(
            self._section_name, 'dsa_private_key_password')

    def setPropertyFromSync(self, source):
        """
        See `ISyncConfiguration`.
        """
        for name in self.getSyncAttributeNames():

            if name not in source:
                continue

            if name == 'uuid':
                # UUID has special handling.
                self._setUUID(source['uuid'])
                continue

            # FIXME:3480:
            # Emit event on configuration changes.
            self.setValue(name, source[name])

        return self


class NodeConfiguration(PropertyMixin):
    """
    Configurations for separate node values.

    This is the public JSON API for the node variables.
    The goal is to present the node variables as similar as possible to other
    API configuration options.
    """
    # To differenciate between real sections that are using the '/' separator
    # the node variables are using a different separator when generating
    # the keys
    _KEY_SEPARATOR = '|'

    def __init__(self, parent):
        self._parent = parent

    def _getJSONModel(self):
        """
        Return the data model for the node variables used by the JSON-RPC API.
        """
        result = {}
        raw_structure = self._parent._proxy.getNodeConfiguration()
        for section, component in raw_structure.items():
            for key, value in component.items():
                # We avoid using the '/' separator in the key name to reduce
                # the confusion on the UI.
                # Otherwise you might think that you have an embedded
                # strucure for a key like 'services/ftp-uuid/ftps_enabled'.
                # This is why we use a key like 'services|ftp-uuid|ftps_enabled'
                variable_id = (
                    section.replace('/', self._KEY_SEPARATOR) +
                    self._KEY_SEPARATOR +
                    key)
                # The embedded dict is needed here to make it easier to
                # integrate with angular ngModel.
                # Otherise, just setting the value would have been enough.
                # In the future we might refactor it, but for now is easiest
                # to make this changes on the backend.
                result[variable_id] = {
                    'section': section,
                    'key': key,
                    'value': value,
                    }
        return result

    def getFilteredProperty(self, getter, property_path):
        """
        Return the full dictionary of node variables.

        This is kind of a private API, and getProperty should be used
        for the main call.

        `getter` is ignored as everything is returned for both configuration
        and sync operations.

        This is a single level dictionary in which each variable has
        an unique ID.
        """
        return self._getJSONModel()

    def _setProperty(self, property_path, value, notify):
        """
        See `IProperty`.

        The node_variables are update via `PATH/value` and not via `PATH`
        to make it easier to handle this via the UI.
        Also `PATH/value` is what we tell the UI to handle via
        the getFilteredProperty read method.
        """
        if not property_path.endswith('/value'):
            raise ServerException('Invalid request to set node variable.')

        # The path received from LM will always be SECTION/KEY/value
        variable_path = property_path[:-6]
        # Revert the path separator used in the read part.
        variable_path = variable_path.replace(self._KEY_SEPARATOR, '/')
        section, key = variable_path.rsplit('/', 1)
        self._parent._proxy.setNodeVariable(section, key, value)

    def deleteProperty(self, property_path):
        """"
        See: `PropertyMixin`.

        This deletes the property and also emit a signal.
        """
        try:
            section, key = property_path.rsplit(self._KEY_SEPARATOR, 1)
            section = section.replace(self._KEY_SEPARATOR, '/')
        except ValueError:
            raise ServerException(
                'Invalid node_instance delete request '
                'received for "{}".'.format(property_path,))

        self._parent._proxy.removeNodeVariable(section, key)

    def resetINI(self, raw_ini):
        """
        Reset the configuration based on raw ini format text.
        """
        self._parent._proxy.resetINI(raw_ini)
