# Copyright (c) 2011 Adi Roiban.
# See LICENSE for details.
# -*- coding: utf-8 -*-
"""
Shared constants.
"""
# Make sure everything is imported as private.
from __future__ import unicode_literals as _unicode_literals
from __future__ import absolute_import as _absolute_import

# Import things as private method to not export it in JSON.
from bunch import Bunch as _Bunch
from twisted.internet.address import IPv4Address as _IPv4Address

from chevah.server.build import EXPIRATION_TIMESTAMP
from chevah.server.commons import OPENSSL_ED25519_SUPPORTED


def _update_without_overwrite(source, update):
    """
    Update `source` dictionary raising an error if `update` dictionary
    contains keys that would overwrite values in source.
    """
    for key, value in update.items():
        if key in source and value != source[key]:
            raise RuntimeError(
                'Trying to overwrite existing value for key "%s".' % (
                    key))
        source[key] = value


# Fake usage to make lint happy. as we don't use __all__.
OPENSSL_ED25519_SUPPORTED

# Values for each account type
ACCOUNT_TYPE_APPLICATION_MANAGER = 4
ACCOUNT_TYPE_OS_MANAGER = 6

# Address used when we don't know the local or remote address.
ADDRESS_UNKNOWN = _IPv4Address(type='TCP', host='0.0.0.0', port=0)

PRODUCT_NAME = 'SFTPPlus'
# Make sure main branch (and any non release branch has .dev0 release)
# so that the PR specific version can work.
PRODUCT_VERSION = '5.2.0'
PRODUCT_URL = 'https://www.sftpplus.com'
PRODUCT_PUBLISHER = 'ProAtria Ltd.'

if EXPIRATION_TIMESTAMP:
    PRODUCT_NAME += ' TRIAL'
    PRODUCT_VERSION += '+trial'

SERVICE_NAME = 'sftpplus-service'

PASSWORD_CRYPT_ROUNDS = 80000
# The limit is at 128 + extra character for one time password concatenation.
CREDENTIALS_MAX_LENGTH = 150
# Number of seconds in which an initiated authentication process should
# finalize.
AUTH_PROCESS_TIMEOUT = 60

# Define the port range that the server is allowed to listen.
PORT_MIN = 1
PORT_MAX = 65535

# Unix mode used when opening a file. Umask applied on top of it.
DEFAULT_FILE_MODE = 0o666

# Timeout in seconds to wait for a connection to be initialized.
CONNECT_TIMEOUT = 15

# Number of seconds to wait for unfinished handshake SSL connections
# before forcing to close it.
SSL_CONNECTION_CLOSE_TIMEOUT = 5

# Maximum allowed attachment size.
MAX_ALLOWED_ATTACHMENT_SIZE = 5 * 1024 * 1024

# Maximum size for received data to be kept in memory.
# 100 KB
MAXIMUM_IN_MEMORY_RECEIVED_DATA = 100 * 1024

# List of SSL/TLS methods supported by the server.
SSL_SUPPORTED_METHODS = ['sslv3', 'tlsv1', 'tlsv1.1', 'tlsv1.2', 'tlsv1.3']
# List of SSL/TLS methods displayed in the UI.
SSL_METHODS = [
    'secure', 'all', 'tlsv1.3', 'tlsv1.2', 'tlsv1.1', 'tlsv1.0', 'sslv3']

CONDITION = _Bunch(
    ALWAYS='always',
    FAILURE='failure',
    NO='no',
    SUCCESS='success',
    YES='yes',
    SKIP='skip',
    )

PERMISSION = _Bunch(
    ALLOW_CREATE_FOLDER='allow-create-folder',
    ALLOW_DELETE_FOLDER='allow-delete-folder',
    ALLOW_DELETE_FILE='allow-delete-file',
    ALLOW_FULL='allow-full-control',
    ALLOW_LIST='allow-list',
    ALLOW_READ='allow-read',
    ALLOW_RENAME='allow-rename',
    ALLOW_TRAVERSE='allow-traverse',
    ALLOW_WRITE='allow-write',
    ALLOW_SET_ATTRIBUTES='allow-set-attributes',
    DENY_FULL='deny-full-control',
    )

# The list of all available permissions.
_PERMISSIONS_ALLOW = [
    PERMISSION.ALLOW_CREATE_FOLDER,
    PERMISSION.ALLOW_DELETE_FOLDER,
    PERMISSION.ALLOW_DELETE_FILE,
    PERMISSION.ALLOW_FULL,
    PERMISSION.ALLOW_LIST,
    PERMISSION.ALLOW_READ,
    PERMISSION.ALLOW_RENAME,
    PERMISSION.ALLOW_TRAVERSE,
    PERMISSION.ALLOW_WRITE,
    PERMISSION.ALLOW_SET_ATTRIBUTES,
    ]

_PERMISSIONS_DENY = [
    PERMISSION.DENY_FULL,
    ]

PERMISSIONS = _PERMISSIONS_ALLOW + _PERMISSIONS_DENY

ROLE_PERMISSION = _Bunch(
    ALL='all',
    READ='read',
    UPDATE='update',
    DELETE='delete',
    CREATE='create',
    DENY='deny',
    SYNC='sync',
    )

ROLE_PERMISSIONS = [
    ROLE_PERMISSION.ALL,
    ROLE_PERMISSION.READ,
    ROLE_PERMISSION.UPDATE,
    ROLE_PERMISSION.CREATE,
    ROLE_PERMISSION.DELETE,
    ROLE_PERMISSION.DENY,
    ROLE_PERMISSION.SYNC,
    ]

CONFIGURATION_DISABLED_VALUES = ['disabled', 'none', 'disable']
CONFIGURATION_DISABLED_VALUE = CONFIGURATION_DISABLED_VALUES[0]

CONFIGURATION_INHERIT = ['inherit', 'inherited']

TYPE_NAME = _Bunch(
    ABSOLUTE='absolute',
    ACCOUNT_ACTIVITY='account-activity',  # Deprecated in v4.
    ACCOUNT_INTERACTION='account-interaction',
    AES128='aes128',
    AES192='aes192',
    AES256='aes256',
    ALL='all',
    ALLOW='allow',
    AMEND_CONTENT='amend-content',  # Deprecated in v4
    ANALYTICS='analytics',
    ANONYMOUS='anonymous',
    ANY='any',
    ATTACHMENT='attachment',
    APPLICATION='application',
    APPLICATION_FORM_URL_ENCODED='application/x-www-form-urlencoded',
    APPLICATION_JSON='application/json',
    APPLICATION_OCTET_STREAM='application/octet-stream',
    APPLICATION_ZIP='application/zip',
    AS2='as2',
    AS2_ANONYMOUS='as2-anonymous',
    AUTHENTICATION='authentication',
    AZURE_AD='azure-ad',
    ENTRA_ID='entra-id',
    AZURE_BLOB='azure-blob',
    AZURE_FILE='azure-file',
    BASE64='base64',
    BASE_GROUPS='base-groups',
    BASE_ROLES='base-roles',
    BASE_AND_AZURE_GROUPS='base-and-azure-groups',
    BASE_AND_OS_GROUPS='base-and-os-groups',
    # Used as authentication method.
    BASIC_AUTHENTICATION_SCHEME='basic-authentication-scheme',
    BSD='bsd',
    COPY='copy',
    CRL_DISTRIBUTION_POINTS='crl-distribution-points',
    CHAP='chap',
    CUSTOM='custom',
    CREATE_ARCHIVE='create-archive',
    CRYPT_SHA256='crypt-sha256',
    CRYPT_SHA512='crypt-sha512',
    DATABASE='database',
    DENY='deny',
    DENY_USERNAME='deny-username',
    DES3='3des',
    # Deprecated in v4.
    DIGITAL_SIGNATURE_VALIDATION='digital-signature-validation',
    DIRECT_USERNAME='direct-username',
    DISABLE_IDENTITY_SECURITY='disable-identity-security',
    EMAIL_CLIENT='email-client',
    EMAIL_SENDER='email-sender',
    ENCODING='encoding',
    ENCRYPTED='encrypted',
    EVENT_HANDLER='event-handler',
    EXACT='exact',
    EXTENSION='extension',
    EXTERNAL_EXECUTABLE='external-executable',
    EXTRACT_ARCHIVE='extract-archive',
    FILESYSTEM='filesystem',
    FILE_DIGEST='file-digest',
    FILE_DISPATCHER='file-dispatcher',
    FIXED_PATH='fixed-path',
    FTP='ftp',
    FTPS='ftps',
    FTPSE='ftpse',
    FTPSI='ftpsi',
    GET='get',
    GNU='gnu',
    # GROUP_NAME and GROUP_NAME_WITH_DEFAULT are the legacy option in which a
    # single group is supported.
    GROUP_NAME='group-name',
    GROUP_NAME_WITH_DEFAULT='group-name-with-default',
    GZIP='gzip',
    HTTP='http',
    HTTP_REDIRECT='http-redirect',
    HTTP_POST='http-post',
    HTTP_PULL='http-pull',
    HTTPS='https',
    IP_TIME_BAN='ip-time-ban',
    JSON='json',
    LEGACY_WEBADMIN='legacy-webadmin',
    LETS_ENCRYPT='lets-encrypt',
    LDAP='ldap',
    # os_name used in chevah-compat for Linux.
    LINUX='linux',
    LOCAL_FILE='local-file',
    LOCAL_MANAGER='manager',
    LOCAL_MANAGER_UNSECURED='_manager_unsecured',
    LOCATION='location',
    LOG_HANDLER='log handler',
    LOWERCASE='lowercase',
    MD5='md5',
    MONITOR='monitor',
    MOVE='move',
    MOZILLA_CA_ROOTS='${MOZILLA_CA_ROOTS}',
    MS_CHAP_V1='ms-chap-v1',
    MS_CHAP_V2='ms-chap-v2',
    NOT_DEFINED='no-type-defined',
    NO_ACTION='no-action',
    OIDC='oidc',
    ONLY_FILENAME='only-filename',
    OPENPGP='openpgp',
    OS='os',
    # os_name used in chevah-compat for macOS.
    OSX='osx',
    OS_EXTERNAL='os not in config',
    PAP='pap',
    PARENT='parent',
    PASSWORD='password',
    # Used as credential types.
    PASSWORD_BASIC_AUTH='password-basic-auth',
    PBKDF2_SHA256='pbkdf2-sha256',
    PBKDF2_SHA512='pbkdf2-sha512',
    POST='post',
    PROCESS_MONITOR='process-monitor',  # Deprecated since v4.
    PUT='put',
    RADIUS='radius',
    RABBITMQ='rabbitmq',
    RSA_SHA2_256='rsa-sha2-256',
    RSA_SHA2_512='rsa-sha2-512',
    RELATIVE='relative',
    REPLACE_SEPARATOR='replace-separator',
    RESOURCE='resource',
    SERVICE='service',
    SFTP='sftp',
    SEED_FILE='seed-file',
    SET_ON_FIRST_CONNECTION='set-on-first-connection',
    PIN_PUBLIC_KEY='pin-public-key',
    SHA1='sha1',
    SHA224='sha224',
    SHA256='sha256',
    SHA348='sha348',
    SHA512='sha512',
    SHAREPOINT_ONLINE='sharepoint-online',
    SIGNED='signed',
    SIGNED_AND_ENCRYPTED='signed-and-encrypted',
    SMB='smb',
    SOAP='soap',
    SQLITE='sqlite',
    SSH='ssh',
    SSH_DSS='ssh-dss',
    SSH_KEY='ssh-key',
    SSH_PRIVATE_KEY='ssh-private-key',
    SSH_RSA='ssh-rsa',
    SSL_CERTIFICATE='ssl-certificate',
    STANDARD_STREAM='standard-stream',
    SYNC_SIGNED='sync-signed',
    SYNC_UNSIGNED='sync-unsigned',
    SYSLOG='syslog',
    TEMPORARY_PREFIX='temporary-prefix',
    TEMPORARY_SUFFIX='temporary-suffix',
    TEXT='text',
    TEXT_HTML='text/html',
    TIMESTAMP_ALWAYS='timestamp-always',
    TRANSFER='transfer',
    TRANSFER_JOB='transfer job',
    TRANSFORM='transform',
    UI_GEN_1='ui-gen-1',
    UI_GEN_2='ui-gen-2',
    WEBDAVS='webdavs',
    # os_name used in chevah-compat for Windows.
    WINDOWS='windows',
    WINDOWS_EVENTLOG='windows-eventlog',
    XML='xml',
    ZIP='zip',
    _MEMORY=':memory:',
    )

OVERWRITE_RULE = _Bunch(
    FAIL='fail',
    OVERWRITE='overwrite',
    TIMESTAMP_ALWAYS=TYPE_NAME.TIMESTAMP_ALWAYS,
    TIMESTAMP_TO_EXISTING_FILE='timestamp-to-existing-file',
    TIMESTAMP_TO_NEW_FILE='timestamp-to-new-file',
    DISABLE='disable',
    SKIP='skip',
    )

# FIXME:2669:
# Use a single prefix for all tables.
# Database table prefixes.
TABLE_PREFIX = _Bunch(
    LOGGER='chevah_',
    REPORTS='chevah_reports_',
    TRANSFER_MEMORY='chevah_transfer_memory_',
    )
# Table name used for database log output.
LOG_TABLE_NAME = '%slog_entries' % TABLE_PREFIX.LOGGER
# Table name used for last login monitor.
LAST_LOGIN_TABLE_NAME = '%saccount' % TABLE_PREFIX.REPORTS


HTTP_METHOD_NAMES = {
    TYPE_NAME.GET: 'HTTP GET',
    TYPE_NAME.POST: 'HTTP POST',
    TYPE_NAME.PUT: 'HTTP PUT',
    }

AUTHENTICATION_NAMES = {
    TYPE_NAME.ANONYMOUS: 'Anonymous Account',
    TYPE_NAME.APPLICATION: 'Application Accounts',
    TYPE_NAME.ENTRA_ID: 'Microsoft Entra ID',
    TYPE_NAME.DENY_USERNAME: 'Deny Usernames',
    TYPE_NAME.HTTP: 'HTTP Request',
    TYPE_NAME.IP_TIME_BAN: 'Source IP Address Temporary Ban',
    TYPE_NAME.LDAP: 'LDAP bind',
    TYPE_NAME.RADIUS: 'RADIUS',
    TYPE_NAME.LOCAL_FILE: 'Local file',
    TYPE_NAME.OS: 'Operating System Accounts',
    }

EVENT_HANDLER_NAMES = {
    TYPE_NAME.ACCOUNT_INTERACTION: 'Monitor inter-account file operations',
    TYPE_NAME.DATABASE: 'Store in embedded database',
    TYPE_NAME.EMAIL_SENDER: 'Send as email',
    TYPE_NAME.EXTENSION: 'Handle via an extension',
    TYPE_NAME.CREATE_ARCHIVE: 'Create/Compress archive',
    TYPE_NAME.EXTRACT_ARCHIVE: 'Extract/Uncompress archive',
    TYPE_NAME.EXTERNAL_EXECUTABLE: 'Call external executable',
    TYPE_NAME.FILE_DIGEST: 'Generate file digest message',
    TYPE_NAME.FILE_DISPATCHER: 'Dispatch file',
    TYPE_NAME.HTTP: 'Send as HTTP request',
    TYPE_NAME.LOCAL_FILE: 'Store in local file',
    TYPE_NAME.OPENPGP: 'PGP Encrypt/Decrypt',
    TYPE_NAME.RABBITMQ: 'Publish to RabbitMQ AMQP',
    TYPE_NAME.SYSLOG: 'Send to Syslog server',
    TYPE_NAME.STANDARD_STREAM: 'Send to Standard Output',
    TYPE_NAME.WINDOWS_EVENTLOG: 'Send to Windows EventLog',
    }

HTTP_EVENT_HANDLER_FORMAT_NAME = {
    TYPE_NAME.JSON: 'JSON',
    TYPE_NAME.CUSTOM: 'Custom template format',
    TYPE_NAME.LEGACY_WEBADMIN: 'Legacy SFTPPlus WebAdmin',
    TYPE_NAME.SOAP: 'XML SOAP Envelope',
    TYPE_NAME.XML: 'XML Document',
    }

LOCATION_NAMES = {
    TYPE_NAME.FTP: 'FTP (unsecured)',
    TYPE_NAME.FTPSE: 'FTPS Explicit',
    TYPE_NAME.FTPSI: 'FTPS Implicit',
    TYPE_NAME.HTTP_PULL: 'HTTPS pull',
    TYPE_NAME.SFTP: 'SFTP (over SSH)',
    TYPE_NAME.SMB: 'SMB (Windows Share)',
    TYPE_NAME.AS2: 'AS2 send',
    TYPE_NAME.AZURE_FILE: 'Azure File Service',
    TYPE_NAME.WEBDAVS: 'Secure WebDAV (over HTTPS)',
    TYPE_NAME.FILESYSTEM: 'Local filesystem',
    }

_PROTOCOL_NAME = {
    TYPE_NAME.LOCAL_MANAGER: 'Local Management',
    TYPE_NAME.LOCAL_MANAGER_UNSECURED: 'Unsecure Management',
    TYPE_NAME.FTP: 'FTP and Explicit FTPS',
    TYPE_NAME.FTPS: 'FTP and Explicit FTPS',
    TYPE_NAME.FTPSI: 'Implicit FTPS',
    TYPE_NAME.HTTP: 'HTTP (unsecured)',
    TYPE_NAME.HTTPS: 'HTTPS',
    TYPE_NAME.HTTP_REDIRECT: 'HTTP redirection site',
    TYPE_NAME.SSH: 'SSH (SFTP/SCP)',
    TYPE_NAME.MONITOR: 'Local filesystem monitor',
    }

SERVICE_NAMES = {
    TYPE_NAME.LOCAL_MANAGER: 'Local Management',
    TYPE_NAME.LOCAL_MANAGER_UNSECURED: 'Unsecure Management',
    TYPE_NAME.FTP: 'FTP and Explicit FTPS',
    TYPE_NAME.FTPSI: 'Implicit FTPS',
    TYPE_NAME.HTTP: 'HTTP (unsecured)',
    TYPE_NAME.HTTPS: 'HTTPS',
    TYPE_NAME.HTTP_REDIRECT: 'HTTP redirection site',
    TYPE_NAME.SSH: 'SSH (SFTP/SCP)',
    }

PASSWORD_HASHING_SCHEME_NAME = {
    TYPE_NAME.CRYPT_SHA256: 'Crypt SHA-256',
    TYPE_NAME.CRYPT_SHA512: 'Crypt SHA-512',
    TYPE_NAME.PBKDF2_SHA256: 'PBKDF2 SHA-256',
    TYPE_NAME.PBKDF2_SHA512: 'PBKDF2 SHA-512',
    }

AS2_SEND_SECURITY_NAME = {
    TYPE_NAME.SIGNED_AND_ENCRYPTED: 'Signed and encrypted',
    TYPE_NAME.SIGNED: 'Only signed',
    TYPE_NAME.ENCRYPTED: 'Only encrypted',
    CONFIGURATION_DISABLED_VALUE: 'No signature nor encryption',
    }

AS2_MDN_RECEIPT_NAME = {
    TYPE_NAME.SYNC_SIGNED: 'Synchronous signed',
    TYPE_NAME.SYNC_UNSIGNED: 'Synchronous unsigned',
    CONFIGURATION_DISABLED_VALUE: 'No MDN receipt request.',
    }

DIGEST_NAMES = {
    TYPE_NAME.MD5: 'MD5',
    TYPE_NAME.SHA1: 'SHA1',
    TYPE_NAME.SHA224: 'SHA224',
    TYPE_NAME.SHA256: 'SHA256',
    TYPE_NAME.SHA348: 'SHA348',
    TYPE_NAME.SHA512: 'SHA512',
    }

AS2_ENCRYPTION_NAME = {
    TYPE_NAME.AES128: 'AES 128bit CBC',
    TYPE_NAME.AES192: 'AES 192bit CBC',
    TYPE_NAME.AES256: 'AES 256bit CBC',
    TYPE_NAME.DES3: '3DES CBC',
    }

DIGEST_FORMATS = {
    TYPE_NAME.BASE64: 'Base64 only encoding',
    TYPE_NAME.GNU: 'GNU',
    TYPE_NAME.BSD: 'BSD tag'
    }

EXECUTABLE_OUTPUT_FORMAT_NAME = {
    TYPE_NAME.TEXT: 'Plain text',
    TYPE_NAME.JSON: 'JSON object',
    }

# These values can't be changes as they will break backward compatibility.
CONFIGURATION_DEFAULT_GROUP = 'DEFAULT_GROUP'
CONFIGURATION_DEFAULT_ROLE = 'DEFAULT-ROLE'
CONFIGURATION_DEFAULT_ADMINISTRATOR = 'DEFAULT-ADMINISTRATOR-UUID'
CONFIGURATION_DEFAULT_AUTHENTICATION = 'DEFAULT-AUTHENTICATION'

DEFAULT_EMAIL_CLIENT_UUID = 'DEFAULT-EMAIL-CLIENT'
DEFAULT_LETS_ENCRYPT_UUID = 'DEFAULT-LETS-ENCRYPT'
DEFAULT_LOCAL_FILESYSTEM_UUID = 'DEFAULT-LOCAL-FILESYSTEM'
DEFAULT_LOCAL_MANAGER_UUID = 'DEFAULT-MANAGER'
DEFAULT_ANALYTICS_UUID = 'DEFAULT-ANALYTICS'
DEFAULT_SQLITE_DB_UUID = 'DEFAULT-SQLITE'

# UUIDS that are expected to be present in every configuration that
# should not be removed.
DEFAULT_UUIDS = (
    CONFIGURATION_DEFAULT_GROUP,
    CONFIGURATION_DEFAULT_ROLE,
    CONFIGURATION_DEFAULT_ADMINISTRATOR,
    CONFIGURATION_DEFAULT_AUTHENTICATION,
    DEFAULT_EMAIL_CLIENT_UUID,
    DEFAULT_LETS_ENCRYPT_UUID,
    DEFAULT_LOCAL_FILESYSTEM_UUID,
    DEFAULT_LOCAL_MANAGER_UUID,
    DEFAULT_ANALYTICS_UUID,
    DEFAULT_SQLITE_DB_UUID,
    )

EMAIL_ASSOCIATED_FILES_NAME = {
    '': 'Disabled',
    TYPE_NAME.ATTACHMENT: 'Attachment',
    }

BIND_DN_TYPE_NAME = {
    TYPE_NAME.PARENT: 'Username generated from Bind DN',
    TYPE_NAME.ABSOLUTE: 'Absolute DN',
    TYPE_NAME.RELATIVE: 'Append to Bind DN',
    TYPE_NAME.DIRECT_USERNAME: 'Username as Bind DN',
    }

PUBLIC_PROTOCOLS = [
    TYPE_NAME.LOCAL_MANAGER,
    TYPE_NAME.FTP,
    TYPE_NAME.FTPSI,
    TYPE_NAME.HTTP,
    TYPE_NAME.HTTP_REDIRECT,
    TYPE_NAME.HTTPS,
    TYPE_NAME.SSH,
    ]

SSL_PROTOCOLS = [
    TYPE_NAME.HTTPS,
    TYPE_NAME.LOCAL_MANAGER,
    TYPE_NAME.FTPSI,
    ]


# Self-signed certificate returned for the domains for which we don't yet
# have a certificate.
# openssl req -x509 -newkey rsa:3072 -nodes \
#   -keyout key.pem -out cert.pem -days 7300
# It has CN = "loading-lets-encrypt"
_PEMS_PLACEHOLDER = """
-----BEGIN CERTIFICATE-----
MIIEOTCCAqGgAwIBAgIUaDe5a342kZKuSFvcU2l4CGmTV/owDQYJKoZIhvcNAQEL
BQAwLDELMAkGA1UEBhMCR0IxHTAbBgNVBAMMFGxvYWRpbmctbGV0cy1lbmNyeXB0
MB4XDTIyMDgyNjE0MDEyNloXDTQyMDgyMTE0MDEyNlowLDELMAkGA1UEBhMCR0Ix
HTAbBgNVBAMMFGxvYWRpbmctbGV0cy1lbmNyeXB0MIIBojANBgkqhkiG9w0BAQEF
AAOCAY8AMIIBigKCAYEApn3Vd7x3mpZ3Veo/kYoph2L8bjRyuoUqApNUR7Us+QJb
UO6dCel8t5H0yXPbqHelr4EgZob6paHTMjlCznhaoo6gy+zs0C4HLIDG7+viYow3
05eWmlaYQt5F7WGdlx7lcJYv1+BFWRlp+Wl9fFfaZP/9wXHZHUKfHakcytj2H/Ar
hO648nD7i1k6UQ2tDAjiQS5VlCoZOOhEuj6EOt1myDEmoWsELlSuzGNWJbi/TmyR
zheOdIVmbVu63Suy14+qS4v+2aaIFIopwbNDyJQI0qUIjLkMzx8Xt6Ph33jJ0OQ5
wVkGNtv6OkOFrc+puOLiyykFqWmTkTYTICvbyHnznDRXv6qtzwqBJwzY+plIzhe5
1mxPDYiqnuK/eAOoN3kAdHQW2xbbgWeL/pWf+2Z4pmHy3InsCdvpLcmUVL+aQ366
As+v5a8hLHytns35R/UitdDRvrymKeJsawfkvobe+cGZVFQ39e6fA1P25KDXtTuS
b+CS4W+HEkESQaCFQXw3AgMBAAGjUzBRMB0GA1UdDgQWBBQNgFwCC4NtrJO1Amw/
C5Ij5sHfNDAfBgNVHSMEGDAWgBQNgFwCC4NtrJO1Amw/C5Ij5sHfNDAPBgNVHRMB
Af8EBTADAQH/MA0GCSqGSIb3DQEBCwUAA4IBgQA4tfzUSGbatOKJ3DGBZJtcFD9Z
hMf0sDpycvnWPsw6nF6NpxmcZ7Mx55V8ofZlUk8jPraOagtCRLJWD+9IQFY0heF3
9pWl6Bqeg1FjPBvjFJ8B6nXjRJ69d2nhGGr0ZfolKV3SC8pMdROeN5hQ6zs0R0t7
k7H9zL3LKjrRILLuM5Ojq0J1HGXKrGnVtRzKtZm5wQM7qXW3TzTbkyLQhOPfrfnI
z9l6Mpt8clcCd+adifPeJunlAqRIKOtwUZX1SpWmxcUudl7SGQq81NSXtK37JT29
przgysrpD1+TEeWjrc8GYgMEmk+mewqZGz7vLzi/+dIaYqChD1tOX8JVVvHnPWaj
iniroC1VWhOR+1Lf4SMc0L+1oIBIIoa2HU19Li+sUX5eAsS89Q50OzTMgfuhhVVf
DwPAqPfrhn211TaxUzaV554Ddt/uBJi91cj3GSwlRuUqv5ZHJdhO8MlGnCZFn58m
VYZ49m6pD/mEBYDeSd7Wyzm0TmrK92ZjbzN+uYU=
-----END CERTIFICATE-----
-----BEGIN PRIVATE KEY-----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-----END PRIVATE KEY-----
"""

NO_AUTHENTICATION_PROTOCOLS = [
    TYPE_NAME.MONITOR,
    TYPE_NAME.HTTP_REDIRECT,
    ]

STATIC_FOLDER_NAME = 'static'
HTTP_PRIVATE_CHILD_NAME = '__chsps__'
HTTP_PUBLIC_CHILD_NAME = 'public'
# Name of the public path configuration when when should be mounted as root.
HTTP_PUBLIC_ROOT = 'ROOT'
HTTP_REST_CHILD_NAME = 'home'

# /__chsps__/login
HTTP_REST_LOGIN_NAME = 'login'
# /__chsps__/logout
HTTP_REST_LOGOUT_NAME = 'logout'

# /home/__chsps__/password-update
HTTP_REST_PASSWORD_UPDATE_NAME = 'password-update'

HTTP_STATIC_FOLDER_NAME = 'http_service'
LIBS_CHILD_NAME = 'libs'
LOCAL_MANAGER_FOLDER_NAME = 'manager_service'
CONSTANTS_JS_NAME = 'constants.js'
JS_COMMONS_CHILD_NAME = 'js_commons'
BUILD_CHILD_NAME = 'build'

# List of languages available for translation.
HTTP_LANGUAGES = [
    'en',  # The base language that is not using translations.
    'es',
    'ro',
    'de',
    ]

CONFIGURATION_FOLDER_NAME = 'configuration'

CONFIGURATION_SERVER_FILE = 'configuration/server.ini'
CONFIGURATION_PID_FILE = 'run/server.pid'

CONFIGURATION_USERNAME_PLACEHOLDER = '${USER}'
CONFIGURATION_DEFAULT_OWNER_PLACEHOLDER = '${DEFAULT_OS_USER}'
CONFIGURATION_DEFAULT_GROUP_PLACEHOLDER = '${DEFAULT_OS_GROUP}'
CONFIGURATION_HOME_PLACEHOLDER = '${OS_HOME}'
CONFIGURATION_SHARED_PLACEHOLDER = '${SHARED}'
CONFIGURATION_ALL_OS_GROUPS_PLACEHOLDER = '${ALL_OS_GROUPS}'
CONFIGURATION_CONTINUE_PLACEHOLDER = '${CONTINUE}'

CONFIGURATION_SECTION_ACCOUNTS = 'accounts/'
CONFIGURATION_SECTION_ADMINISTRATORS = 'administrators/'
CONFIGURATION_SECTION_AUTHENTICATIONS = 'authentications/'
CONFIGURATION_SECTION_DATABASES = 'databases/'
CONFIGURATION_SECTION_EVENT_HANDLERS = 'event-handlers/'
CONFIGURATION_SECTION_GROUPS = 'groups/'
CONFIGURATION_SECTION_LOCATIONS = 'locations/'
CONFIGURATION_SECTION_RESOURCES = 'resources/'
CONFIGURATION_SECTION_ROLES = 'roles/'
CONFIGURATION_SECTION_SERVER = 'server'
CONFIGURATION_SECTION_SERVICES = 'services/'
CONFIGURATION_SECTION_TRANSFERS = 'transfers/'


class _inherit_marker(object):
    """
    This should not hit the public API, so it should not be JSON serializable.
    """


# Internal object to mark that a config value is inherited.
# This should be removed in future code.
_INHERIT_MARKER = _inherit_marker()


# Logger constants.
LOGGER_NAME = 'ChevahLogger'

CONFIGURATION_ALL_LOG_GROUPS = 'all'

CONFIGURATION_DEFAULT_TYPE = 'no-type-defined'

CONFIGURATION_ACCOUNT_DEFAULTS = {
    'password': '',
    'home_folder_path': 'inherit',
    'create_home_folder': False,
    'create_home_folder_owner': CONFIGURATION_DEFAULT_OWNER_PLACEHOLDER,
    'create_home_folder_group': CONFIGURATION_DEFAULT_GROUP_PLACEHOLDER,
    'lock_in_home_folder': True,
    'ssh_authorized_keys_path': 'inherit',
    'ssh_authorized_keys_content': '',
    'required_credentials': '',
    'source_ip_filter': '',
    'allow_certificate_authentication': True,
    'as2_require_http_authentication': '',
    'as2_certificates': '',
    'as2_async_mdn_ca': 'Disabled',
    'email': '',
    'expire_datetime': '',
    'amend_write_name': '',
    }

# This is the legacy value used for configuration placeholders.
CONFIGURATION_SSL_DEFAULTS = {
    'ssl_certificate': '',
    'ssl_key': '',
    'ssl_key_password': '',
    'ssl_certificate_authority': '',
    'ssl_certificate_revocation_list': '',
    'ssl_certificate_revocation_list_refresh': '',
    'ssl_cipher_list': '',
    'ssl_allowed_methods': '',
    }

SSL_DEFAULTS = {
    'ssl_cipher_list': 'secure',
    'ssl_allowed_methods': ('secure',),
    'ssl_certificate': '',
    'ssl_key': '',
    'ssl_certificate_authority': '',
    'ssl_certificate_revocation_list': '',
    'ssl_certificate_revocation_list_refresh': 0,
    }

LOCATION_CONNECTION_DEFAULTS = {
    'idle_connection_keepalive_interval': 0,
    'idle_connection_timeout': 300,
    'connection_retry_count': 12,
    'connection_retry_interval': 300,
    }

LOCATION_AS2_DEFAULTS = {
    'as2_send_security': TYPE_NAME.SIGNED_AND_ENCRYPTED,
    'as2_mdn_receipt': TYPE_NAME.SYNC_SIGNED,
    'as2_encryption_algorithm': TYPE_NAME.AES256,
    'as2_signature_algorithm': TYPE_NAME.SHA256,
    'as2_content_type': TYPE_NAME.APPLICATION_OCTET_STREAM,
    }

CONFIGURATION_DEFAULTS = {
    'enabled': True,
    'name': '',
    'description': '',
    'uuid': '',
    'type': CONFIGURATION_DEFAULT_TYPE,
    'created': '',

    # This is outside of default SSL as is a server-only value.
    'ssl_domains': '',

    'banner': PRODUCT_NAME,
    'debug': False,

    # process section.
    'account': '',
    'product_key': '',
    'umask': '022',
    'password_minimum_strength': '',
    'password_minimum_length': '',
    'password_history': '',
    'password_hashing_scheme': 'crypt-sha512',

    # FTP Service
    'passive_port_range': '0 - 0',
    'passive_address': '',
    'active_source_port': '',
    'enable_password_authentication': True,
    'enable_ssl_certificate_authentication': True,
    'idle_data_connection_timeout': '',

    'ftps_explicit_enabled': '',
    'ftps_force_secured_command_channel': '',
    'ftps_require_session_reuse': '',
    'ignore_ascii_data_type': False,
    'ascii_data_type_as_default': False,

    # SSH service
    'sftp': True,
    'scp': False,
    'ssh_host_private_keys': '',

    # FIXME:6521:
    # Should be removed in v6.
    'rsa_private_key': '',
    'rsa_private_key_password': '',
    'dsa_private_key': '',
    'dsa_private_key_password': '',

    'dh_prime_size': '1024, ideal',
    'ignore_create_permissions': False,

    # Generic.
    'address': '127.0.0.1',
    'idle_connection_timeout': '',
    'maximum_concurrent_connections': '',

    # `authentications` is shared between server and service configuration,
    # so we can't add a default value here as they have different defaults.
    'authentications': '',
    'multi_factor_authentication': '',
    'components': '',
    'database': '',
    'auto_delete': '',
    'port': '',
    'protocol': CONFIGURATION_DEFAULT_TYPE,
    'password': '',
    'path': '',
    'rotate_at_size': 0,
    'rotate_count': 0,
    'rotate_each': 'Disabled',
    'rotate_on': '',
    'rotate_external': False,
    'entry_content': (
        '{timestamp.iso_8601_local} {id} {component.uuid} {account.name} '
        '{account.peer.address}:{account.peer.port} {message}'),
    'structured_fields': '',
    'expected_response': '',
    'extra_data': '',
    'source_addresses': '',
    'target': '',
    'groups': '',
    'url': '',
    'test_at_start': '',
    'public_error': '',
    'username': '',
    'usernames': '',
    'data_filter': '',
    'home_folder_structure': '',
    'virtual_folders': 'inherit',
    'permissions': '',
    'proxy': '',
    'public_account_uuid': '',
    'public_name': '',

    # AS2 server-side.
    'as2_pending_path': '',
    'as2_receive_name': '',
    'as2_receive_path': '',
    'as2_receive_key': '',
    'as2_receive_certificate': '',
    'as2_default_filename': '',
    'as2_no_mdn_success_text': '',
    # As2 client-side
    'as2_partner_identifier': '',
    'as2_partner_certificates': '',
    'as2_own_identifier': '',
    'as2_own_certificate': '',
    'as2_own_private_key': '',
    'as2_send_security': '',
    'as2_mdn_receipt': '',
    'as2_encryption_algorithm': '',
    'as2_signature_algorithm': '',
    'as2_content_type': '',

    'request_method': '',
    'request_body': '',
    'request_headers': '',
    'response_code': '',
    'response_file_name': '',

    'headers': '',
    'metadata_fields': '',
    'accepted_origins': '',
    'base_url_path': '',
    'client_forwarded_header': '',
    'redirect_url': '',
    'theme_path': '',
    'ui_version': '',
    'login_footnotes': '',
    'languages': '',
    'triggers': '',
    'announce_session_authentication': True,
    'domains': '',
    'acme_url': 'https://acme-v02.api.letsencrypt.org/directory',
    'contact_email': '',
    # Private persistence.
    '_account_key': '',
    '_storage': '',

    # Authentication methods.
    'allowed_groups': '',
    'manager_allowed_groups': '',
    'group_association': '',
    'base_groups': '',
    'base_roles': '',
    'role_association': '',
    'anonymous_account_uuid': '',
    'ban_interval': '',
    'ban_after_count': '',
    'external_management': True,
    'pam_usage': 'fallback',
    'pam_service': 'login',
    'bind_dn': '',
    'bind_dn_type': TYPE_NAME.PARENT,
    'group_mapping': '',
    'home_folder_attribute': 'homeDirectory',
    'email_attribute': '',
    'search_filter': '',
    'manager_search_filter': '',
    'username_attribute': '',
    'username_suffix': '',
    'secure_connection': '',
    'continue_authentication': '',
    'authentication_method': '',
    'authentication_type': '',
    'allow_insecure_http': '',
    'nas_port': '',
    'extension_entry_point': '',
    'extension_configuration': '',
    'application_id': '',
    'directory_id': '',
    'remove_username_suffix': '',
    'api_scopes': '',

    # Event handlers.
    'http_content_type': 'json',
    'table_name': LOG_TABLE_NAME,
    'entry_point': '',
    'delete_source_on_success': True,
    'delete_source_parent_delay': '',
    'digest_algorithm': '',
    'output_format': '',
    'passphrase': '',
    'encryption_cipher': 'aes128',
    'encryption_extension': '.pgp',
    'encryption_public_keys': '',
    'decryption_private_keys': '',
    'executable_path': 'no/path/defined',
    'executable_arguments': '{data.real_path}',
    'executable_output_format': '',
    'concurrent_processes': '',
    'configuration': '',
    'exchange': '',
    'routing_key': '',
    'payload_template': '',

    # Database event handlers.
    'queue_size': '',

    'source_attribute': 'real_path',
    'dispatch_attribute': 'real_path',
    'dispatch_delay': '',
    'dispatch_rules': '',
    'fallback_rule': '',
    'matching_expressions': '',
    'create_destination_folder': '',

    'email_from_address': 'SFTPPlus <no-reply@sftpplus.example.com>',
    'email_to_recipients': '',
    'email_cc': '',
    'email_bcc': '',
    'email_subject': '[{id}] [{component.name}] New event from SFTPPlus',
    'email_body': '[{timestamp.cwa_14051}] {message}{LF}{LF}{data_json}',
    'email_associated_files': '',

    'monitor_interval': '',
    'memory_resident_trigger': '',
    'memory_virtual_trigger': '',
    'connection_count_trigger': '',
    'file_count_trigger': '',
    'thread_count_trigger': '',

    'signer_certificate_path': 'no-path-defined',

    # Locations.
    'idle_connection_keepalive_interval': '',

    # FTP(S) locations.
    'acct': '',
    'ftps_reuse_session': '',
    'ignore_passive_address': '',

    # For backward compatibility. Replaced by file_age_notification.
    'warn_non_modified_files_interval': 0,
    # We use -1 to detect that value is not present in the config file.
    # and fallbackc to warn_non_modified_files_interval.
    'file_age_notification': -1,
    'file_age_auto_delete': '',
    # Once WritableCSVUniqueInsensitive has support for default we can
    # leave this empty.
    'monitored_operations': 'all',

    # Transfers.
    'source_uuid': '',
    'destination_uuid': '',
    'source_path': '',
    'destination_path': '',
    'destination_temporary_suffix': '',
    'destination_fallback_path': '',
    'recursive': False,
    'changes_poll_interval': '',
    'stable_interval': '',
    'batch_interval': '',
    'batch_marker_path': '',
    'retry_count': '',
    'retry_wait': '',
    'retry_increase': '',
    'transfer_memory_duration': '',
    'minimum_transfer_count': '',
    'minimum_transfer_interval': '',
    'ignore_duplicate_paths': False,

    'source_filter': '',
    'source_filter_age': '',

    'timeout': '',
    'executable_timeout': '',
    'execute_timeout': '',
    'execute_before': '',
    'execute_after_success': '',
    'execute_after_failure': '',
    'execute_on_destination_before': '',
    'execute_on_destination_after_success': '',
    'execute_on_destination_after_failure': '',
    'archive_success_path': '',
    'archive_failure_path': '',
    'archive_retention_period': '',
    'archive_format': '',
    'archive_name': '',
    'schedule': '',
    'overwrite_rule': OVERWRITE_RULE.FAIL,
    'destination_path_actions': '',
    'destination_content_actions': '',

    'ssh_server_identity': '',
    # Alias for v3 backward compatibility configuration.
    'ssh_server_fingerprint': 'not-defined-yet',

    'ssh_private_key': '',
    'ssh_cipher_list': 'secure',
    'ssh_delete_delay': '',

    'require_encryption': '',

    'file_write_size': '',
    'disable_on_inactivity': '',
    'password_lifetime': '',
    'allow_own_password_change': False,
    'roles': '',
    'group': '',
    'node_variables': '',

    # Runnables.
    'fail_after_errors': '',
    'connection_retry_count': '',
    'connection_retry_interval': '',
    }

CA_PLACEHOLDER = {
    # All Microsoft and Azure certs.
    # https://learn.microsoft.com/en-us/azure/security/fundamentals/azure-ca-details
    # Legacy site
    # https://www.microsoft.com/pki/mscorp/cps/default.htm
    # We used to include intermediary CA, but we should only include the
    # root CAs. Similar to what `certifi` does.
    '${MICROSOFT_IT_CA}': 'microsoft-it.pem',
    TYPE_NAME.MOZILLA_CA_ROOTS: 'mozilla-ca-roots.pem',
    # https://letsencrypt.org/certificates/
    '${LETS_ENCRYPT_X3_CA}': 'lets-encrypt-authority-x3.pem',
    # https://ssl-ccp.godaddy.com/repository
    '${GO_DADDY_G2_G1}': 'gd_bundle-g2-g1.crt',
    }

# SHAREPOINT_CA is a deprecated CA which now is the same as MS IT CAs.
CA_PLACEHOLDER['${SHAREPOINT_CA}'] = CA_PLACEHOLDER['${MICROSOFT_IT_CA}']

CRL_PLACEHOLDER = {
    '${MICROSOFT_IT_CRL}': [
        'http://crl3.digicert.com/ssca-sha2-g6.crl',
        ]
    }

# Backward compatibility.
CRL_PLACEHOLDER['${SHAREPOINT_CRL}'] = CRL_PLACEHOLDER['${MICROSOFT_IT_CRL}']

# Only accounts defaults are added, since they include the values for
# accounts.
_update_without_overwrite(
    CONFIGURATION_DEFAULTS, CONFIGURATION_ACCOUNT_DEFAULTS)
_update_without_overwrite(
    CONFIGURATION_DEFAULTS, CONFIGURATION_SSL_DEFAULTS)


SORT = _Bunch(
    ASCENDING='asc',
    DESCENDING='desc',
    )

# RSA and DSA keys.
DEFAULT_KEY_SIZE = 3072
DEFAULT_KEY_TYPE = 'rsa'

# Make sure each state has a unique description.
STATE = _Bunch(
    STARTED='Started',
    STOPPED='Stopped',
    CONNECTING='Connecting',
    DISCONNECTED='Disconnected',
    DISCONNECTING='Disconnecting',
    DONE='Done',
    FAILING='Failing',
    FAILED_AT_START='Failed at start',
    FAILED_AT_STOP='Failed at stop',
    FAILED_AT_RUN='Failed while running',
    STARTING='Starting',
    STOPPING='Stopping',
    SUSPENDED='Suspended',
    RESTARTING='Restarting',
    SOURCE_FAILED='Source has failed',
    SOURCE_STOPPED='Source is stopped',
    DESTINATION_FAILED='Destination has failed',
    )

# List of states for which a runnable is stopped.
STATES_STOP = [
    STATE.STOPPED,  # Stop with success.
    STATE.FAILED_AT_START,
    STATE.FAILED_AT_STOP,
    STATE.FAILED_AT_RUN,
    ]

EVENT_TYPE = _Bunch(
    CREATED='created',  # Newly created members.
    DELETED='deleted',  # Removed members.
    EXISTING='existing',  # Existing members when watcher starts.
    MODIFIED='modified',  # Existing members which are changes.
    MOVED='moved',  # Existing members which are moved / renamed.
    READ='read',  # Existing member read/downloaded.
    )


# List of all supported operations for the account interaction event handler.
ACCOUNT_INTERACTION_OPERATIONS = (
    TYPE_NAME.ALL,
    EVENT_TYPE.CREATED,
    EVENT_TYPE.DELETED,
    EVENT_TYPE.MOVED,
    EVENT_TYPE.READ,
    EVENT_TYPE.MODIFIED,
    )

# List of all supported monitored operation configuration values.
# See services/monitor.py for the list of supported operations.
MONITORED_OPERATIONS = [
    TYPE_NAME.ALL,
    'disabled',
    'exist',
    'delete',
    'create',
    'rename',
    'modify',
    ]

TRANSFER_ARCHIVE_FORMATS = {
    TYPE_NAME.TIMESTAMP_ALWAYS: 'Always timestamp in mirrored directories',
    TYPE_NAME.EXACT: 'In mirrored directories',
    }

# Available configuration options for `group_association` for OS auth.
OS_GROUP_ASSOCIATION = {
    TYPE_NAME.BASE_GROUPS: 'Only base groups',
    TYPE_NAME.BASE_AND_OS_GROUPS: 'Base groups and OS groups',
    TYPE_NAME.GROUP_NAME: 'First OS group match only',
    TYPE_NAME.GROUP_NAME_WITH_DEFAULT: 'First OS group or default group',
    }

# Available configuration options for `group_association` for OS auth.
OS_ROLE_ASSOCIATION = {
    TYPE_NAME.BASE_ROLES: 'Only base roles',
    TYPE_NAME.BASE_AND_OS_GROUPS: 'Base roles and OS groups',
    }

# Available configuration options for `group_association` for Entra ID auth.
ENTRA_ID_GROUP_ASSOCIATION = {
    TYPE_NAME.BASE_GROUPS: 'Only base groups',
    TYPE_NAME.BASE_AND_AZURE_GROUPS: 'Base groups and Azure groups',
    }

# Error codes shared by the local manager between backend and fronted.
LM_ERROR = _Bunch(
    SESSION_NOT_VALID='session-not-valid',
    OIDC_NOT_ENABLED='oidc-not-enabled',
    OIDC_NO_URL='oid-no-url',
    OIDC_GET_CREDENTIALS='oid-get-credentials',
    CREDENTIALS_DENIED='credentials-denied',
    )

# Text for the error message for errors shared between frontend and backend.
LM_MESSAGE = {
    LM_ERROR.SESSION_NOT_VALID: (
        'Session is no longer valid. It has probably expired.'),
    LM_ERROR.OIDC_NOT_ENABLED: (
        'OpenID Connect auth method is not enabled.'),
    LM_ERROR.OIDC_NO_URL: 'No URL configured for OpenID Connect.',
    LM_ERROR.OIDC_GET_CREDENTIALS: (
        'Failed to get OpenID Connect credentials.'),
    LM_ERROR.CREDENTIALS_DENIED: 'Credentials denied',
    }

# These are empty lists.
# Later will be populated once SSL, SSH, and other code is loaded.
# Implemented in this way to avoid circular import.
SSL_CIPHERS = []
SSH_CIPHERS = []
EVENTS_DB = {
    'events': {},
    'groups': {},
    }
