# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
Module for authentication related classes.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
import base64

from OpenSSL import crypto
from twisted.internet.address import IPv4Address, IPv6Address
from zope.interface import implementer

from chevah.server.commons.constant import TYPE_NAME
from chevah.server.commons.interface import (
    IAS2AnonymousCredentials,
    ICredentials,
    IEventCreator,
    IFTPPasswordCredentials,
    IFTPSPasswordCredentials,
    IFTPSSSLCertificateCredentials,
    IHTTPPasswordCredentials,
    IHTTPSPasswordCredentials,
    IHTTPSSSLCertificateCredentials,
    IPasswordCredentials,
    ILocalManagerCredentials,
    ISSHKeyCredentials,
    ISSHPasswordCredentials,
    ISSHPrivateKeyCredentials,
    ISSLCertificateCredentials,
    )
from chevah.server.commons.observer import ObserverMixin
import six


@implementer(ICredentials)
class CredentialsAbstract(ObserverMixin):
    """
    Base class for all credentials used in the server.
    """
    def __init__(self, username, peer, creator):
        if not username:
            # Just use an empty name and this will be rejected later.
            username = ''
        if not isinstance(username, six.text_type):
            raise RuntimeError('Only Unicode usernames are supported.')
        if not isinstance(peer, (IPv4Address, IPv6Address)):
            raise RuntimeError('Only IPv4 and IPv6 addresses are supported.')
        if not IEventCreator.providedBy(creator):
            raise RuntimeError(
                'Credentials with invalid creator %r.' % (creator,))
        self._username = username
        self._peer = peer
        self._creator = creator

    @property
    def type(self):
        """
        See: `ICredentials`.
        """
        raise NotImplementedError('type not implemented')

    @property
    def content(self):
        """
        See: `ICredentials`.
        """
        raise NotImplementedError('content not implemented')

    @property
    def username(self):
        """
        See: `ICredentials`.
        """
        return self._username

    @property
    def peer(self):
        """
        See: `ICredentials`.
        """
        return self._peer

    @property
    def creator(self):
        """
        See: `ICredentials`.
        """
        return self._creator

    @property
    def kind_name(self):
        """
        See: `ICredentials`.
        """
        # Default to type.
        return self.type


@implementer(IAS2AnonymousCredentials)
class AS2AnonymousCredentials(CredentialsAbstract):
    """
    Credentials for AS2 request without an HTTP authentication header.

    Initialed with username as the raw value of AS2-From hTTP header.
    """

    def __init__(self, username, peer, creator):
        username = username.strip()
        username = username.strip('"')
        username = username.strip()
        super(AS2AnonymousCredentials, self).__init__(
            username=username, peer=peer, creator=creator)

    @property
    def type(self):
        """
        See: `ICredentials`.
        """
        return TYPE_NAME.AS2_ANONYMOUS

    @property
    def content(self):
        """
        See: `ICredentials`.
        """
        return None


@implementer(IPasswordCredentials)
class PasswordCredentials(CredentialsAbstract):
    """
    Credentials based on password.
    """

    def __init__(self, password=None, token=None, *args, **kwargs):
        super(PasswordCredentials, self).__init__(*args, **kwargs)
        self.token = token
        if password:
            if not isinstance(password, six.text_type):
                raise RuntimeError('Only Unicode passwords are supported.')
        self.password = password

    @property
    def type(self):
        """
        See: `ICredentials`.
        """
        return TYPE_NAME.PASSWORD

    @property
    def content(self):
        """
        See: `ICredentials`.
        """
        return self.password


class _HTTPCredentialsAbstract(PasswordCredentials):
    """
    Shared code for both HTTP and HTTPS credentials.
    """

    def __init__(self, request=None, *args, **kwargs):
        self.request = request
        kwargs['peer'] = _get_peer(request)
        kwargs['creator'] = request.site
        super(_HTTPCredentialsAbstract, self).__init__(*args, **kwargs)


def _get_peer(request):
    """
    Return the source peer as advertised by the request.
    """
    header_name = request.site.client_forwarded_header
    if not header_name:
        # Return the current TCP connection source.
        return request.getAvatarPeer()

    header_value = request.requestHeaders.getRawHeaders(header_name)

    if not header_value:
        # This might be a direct connection.
        # Fallback to TCP connection.
        return request.getAvatarPeer()

    # Only take into consideration the first header.
    header_value = header_value[0]

    try:
        peer = _parse_forwarded_header(header_name, header_value)
    except ValueError:
        # Ignore invalid headers.
        peer = None

    if not peer:
        # No peer could be extracted from the header.
        # Fallback to TCP connection.
        return request.getAvatarPeer()

    return peer


def _parse_forwarded_header(name, value):
    """
    Return a peer by parsing the `value` for header `name`.

    The first address found, is the address of the original client.
    The rest are intermediary proxies.

    X-Forwarded-For: <client>, <proxy1>, <proxy2>
    """
    if name != 'forwarded':
        # We have a non-standard header.
        # Try our best to extract the address and port.
        # See https://github.com/eclipse/jetty.project/issues/3630
        # for discussion on non-standard header values.
        peer = value.split(',')[0].strip()
        port = 0
        if '.' in peer and '::' not in peer:  # Exclude IPv4-to-IPv6 mapped.
            address_class = IPv4Address
            if ':' in peer:
                # Have custom port.
                peer, port = peer.rsplit(':', 1)
        else:
            address_class = IPv6Address
            if '[' in peer:
                # Some IPv6 literals.
                if ':' in peer:
                    # Try to see if we have custom port.
                    address, port = peer.rsplit(':', 1)
                    if ']' in port:
                        # Nope. We have only [IP:V:6] literal without port.
                        port = 0
                    else:
                        peer = address

            peer = peer.strip('[]')

        port = int(port)
        return address_class(type='TCP', host=peer, port=port)

    # Handle standard RFC 7238 section 4 header.
    # for=192.0.2.60:1245;proto=http;by=203.0.113.43
    # For=192.0.2.60:1245,"[2001:db8:cafe::17]:4711";proto=http
    fors = [
        v.strip().lower() for v in value.split(';')
        if v.strip().startswith('for=')
        ]

    if not fors:
        raise ValueError('Malformed "Forwarded" header value.')

    peer = [v.strip()[4:] for v in fors[0].split(',')][0]
    port = 0

    if '"' in peer:
        # Have IPv6 as they are always quoted
        peer = peer.strip('"')
        if not peer.endswith(']'):
            # We have a port.
            peer, port = peer.rsplit(':', 1)
        peer = peer.strip('[]')

        port = int(port)
        return IPv6Address(type='TCP', host=peer, port=port)

    # Have IPv4.
    if ':' in peer:
        # Have custom port.
        peer, port = peer.rsplit(':', 1)

    port = int(port)
    return IPv4Address(type='TCP', host=peer, port=port)


@implementer(IHTTPPasswordCredentials)
class HTTPPasswordCredentials(_HTTPCredentialsAbstract):
    """
    Marker class for password based credentials.

    HTTP Basic Auth credentials are username and password based.
    """


@implementer(IHTTPSPasswordCredentials)
class HTTPSPasswordCredentials(_HTTPCredentialsAbstract):
    """
    Marker class for password based credentials.

    HTTPS Basic Auth credentials are username and password based.
    """


class OIDCCredentials(_HTTPCredentialsAbstract):
    """
    Credentials for OpenID Connect over HTTPS.
    """

    def __init__(self, method, *args, **kwargs):
        self._method_uuid = method
        super(OIDCCredentials, self).__init__(*args, **kwargs)

    @property
    def method(self):
        """
        UUID of the method that has created these credentials.
        """
        return self._method_uuid

    @property
    def type(self):
        """
        See: `ICredentials`.
        """
        return TYPE_NAME.OIDC


@implementer(IFTPPasswordCredentials)
class FTPPasswordCredentials(PasswordCredentials):
    """
    Marker class for password based credentials used with FTP.
    """


@implementer(IFTPSPasswordCredentials)
class FTPSPasswordCredentials(PasswordCredentials):
    """
    Marker class for password based credentials used with FTPS.
    """


@implementer(ISSHPasswordCredentials)
class SSHPasswordCredentials(PasswordCredentials):
    """
    Marker class for password based credentials used with SSH.
    """


@implementer(ISSHKeyCredentials)
class SSHKeyCredentials(CredentialsAbstract):
    """
    An SSH key based credentials for public key and its signed data.

    `key_data` is raw blob bytes.
    The content is returned as text in base64 encoding of that public key.
    """

    def __init__(
        self,
        key_algorithm=None, key_data=None, key_signed_data=None,
        key_signature=None,
        *args, **kwargs
            ):
        super(SSHKeyCredentials, self).__init__(*args, **kwargs)
        if key_algorithm is None:
            # Just have some placeholder.
            key_algorithm = b'no-key-alg'

        if isinstance(key_algorithm, six.binary_type):
            key_algorithm = key_algorithm.decode('utf-8')

        self.key_algorithm = key_algorithm
        self.key_data = key_data
        self.key_signed_data = key_signed_data
        self.key_signature = key_signature

    @property
    def type(self):
        """
        See: `ICredentials`.
        """
        return TYPE_NAME.SSH_KEY

    @property
    def kind_name(self):
        """
        See: `ICredentials`.
        """
        return self.type + '-' + self.key_algorithm

    @property
    def content(self):
        """
        See: `ICredentials`.
        """
        if not self.key_data:
            return None
        return base64.b64encode(self.key_data).decode('ascii')


@implementer(ISSHPrivateKeyCredentials)
class SSHPrivateKeyCredentials(CredentialsAbstract):
    """
    An SSH private key based credentials.
    """

    def __init__(self, key=None, *args, **kwargs):
        super(SSHPrivateKeyCredentials, self).__init__(*args, **kwargs)
        self._key = key

    @property
    def type(self):
        """
        See: `ICredentials`.
        """
        return TYPE_NAME.SSH_PRIVATE_KEY

    def getPrivateKey(self):
        """
        See: `ISSHPrivateKeyCredentials`.
        """
        return self._key

    def getPublicKey(self):
        """
        See: `ISSHPrivateKeyCredentials`.
        """
        return self._key.public()

    @property
    def content(self):
        """
        See: `ICredentials`.

        For now, private key is not exposed as generic content.
        """
        return None


@implementer(ISSLCertificateCredentials)
class SSLCertificateCredentials(CredentialsAbstract):
    """
    A SSL certificate key based credentials.
    """

    def __init__(self, certificate, *args, **kwargs):
        super(SSLCertificateCredentials, self).__init__(*args, **kwargs)
        self.certificate = certificate
        self._certificate_pem = None
        if self.certificate:
            try:
                self._certificate_pem = crypto.dump_certificate(
                    crypto.FILETYPE_PEM, self.certificate).decode('ascii')
            except Exception:
                self._certificate_pem = 'INVALID'

    @property
    def type(self):
        """
        See: `ICredentials`.
        """
        return TYPE_NAME.SSL_CERTIFICATE

    @property
    def content(self):
        """
        See: `ICredentials`.
        """
        return self._certificate_pem


@implementer(IFTPSSSLCertificateCredentials)
class FTPSSSLCertificateCredentials(SSLCertificateCredentials):
    """
    A SSL certificate key based credentials for FTPS.
    """


@implementer(IHTTPSSSLCertificateCredentials)
class HTTPSSSLCertificateCredentials(SSLCertificateCredentials):
    """
    A SSL certificate key based credentials for HTTPS.
    """


@implementer(ILocalManagerCredentials)
class LocalManagerCredentials(ObserverMixin):
    """
    Credentials used by the local manager service.
    """

    def __init__(self, username, password, request, creator):
        self.username = username
        self.password = password
        self.peer = _get_peer(request)
        self.creator = creator

    @property
    def type(self):
        """
        See: ICredentials.
        """
        return TYPE_NAME.PASSWORD

    @property
    def kind_name(self):
        """
        See: ICredentials.
        """
        return self.type

    @property
    def content(self):
        """
        See: ICredentials.
        """
        return self.password
