# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
Exceptions raised by the server.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
import traceback

from twisted.internet import error

from chevah.server import force_unicode


class ServerError(Exception):
    """
    Generic error used by Chevah components.

    It has an ID and data attached and is very tightly coupled with an event.

    Errors are exceptions that could not be handled.

    In most cases the human readable error message is generated from the
    events definition and the event's data.
    In case the error is raised before the event definition is loaded the
    data dict will contain the non-interpolated text as the `message` member.
    """

    def __init__(self, message_id='0', avatar=None, peer=None, data=None):
        """
        Holds the message id and its data.
        """
        if data is None:
            data = {}

        self.id = message_id
        self.data = data

        if avatar:
            self.avatar = avatar
        elif peer:
            # We should raise an error if peer and avatar are both present
            # as this might get into a conflict.
            # But for now, we will ignore, as it didn't got us into any big
            # trouble.
            # Prevent circular import.
            from chevah.server.identity.avatar import PeerAvatar
            self.avatar = PeerAvatar(peer)
        else:
            # Prevent circular import.
            from chevah.server.identity.avatar import ProcessAvatar
            self.avatar = ProcessAvatar()

    @property
    def _repr(self):
        """
        Helper to see object's representation.
        """
        return '%s\n%s' % (self.id, self.data)

    def __repr__(self):
        # return self._repr     # Use during development.
        raise RuntimeError(
            'Use _repr if you need debug serialization.\n%s\n%s' % (
                self._repr, ''.join(traceback.format_stack())))

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        """
        Simple equality for ServerErrors.

        ServerErrors are equal as long as they have the same id and data.
        """
        id_result = self.id == other.id
        data_result = self.data == other.data
        return id_result and data_result


class EmittedError(ServerError):
    """
    An exception which wraps a ServerError and signals that an event
    was emitted for the wrapped exceptions.

    This is a proxy/delegate to ServerError.

    It inherits from ServerError so that it looks like a ServerError.

    Once created, an EmittedError should not be changed as the error was
    already emitted and could generate conflicts in the event state.
    """

    def __init__(self, error):
        # Ugly access to __dict__ to avoid recursive call.
        self.__dict__['_context'] = error

    def __getattr__(self, attrib):
        return getattr(self._context, attrib)

    def __setattr__(self, attrib, value):
        if attrib == '__traceback__':
            # This is called by Twisted when the failure is cleared.
            # Basically to set traceback to None.
            self._context.__traceback__ = value
            return
        raise AttributeError('This is a read-only exception.')


class NoSuchAttributeError(ServerError):
    """
    Error raised when the configuration does not have a requested attribute.
    """

    def __init__(self, message=''):
        self.id = '20152'
        self.data = {'name': message}
        # Prevent circular import.
        from chevah.server.identity.avatar import ProcessAvatar
        self.avatar = ProcessAvatar()


class NoSuchSectionError(ServerError):
    """
    Error raised when the configuration does not have a requested section.
    """

    def __init__(self, message=''):
        self.id = '20153'
        self.data = {'section_name': message}
        # Prevent circular import.
        from chevah.server.identity.avatar import ProcessAvatar
        self.avatar = ProcessAvatar()


class CreateNotSupportedError(ServerError):
    """
    Error raised when trying to create a property for which create
    is not supported.
    """

    def __init__(self, message=''):
        self.id = '20154'
        self.data = {'section_name': message}
        # Prevent circular import.
        from chevah.server.identity.avatar import ProcessAvatar
        self.avatar = ProcessAvatar()


class DeleteNotSupportedError(ServerError):
    """
    Error raised when trying to delete a property for which delete
    is not supported.
    """

    def __init__(self, message=''):
        self.id = '20155'
        self.data = {'name': message}
        # Prevent circular import.
        from chevah.server.identity.avatar import ProcessAvatar
        self.avatar = ProcessAvatar()


class _BaseException(Exception):
    """
    Common code for exceptions used by chevah/server.
    """
    message = 'No details.'

    def __init__(self, message='Not yet defined.'):
        # We can't make `message` as a required argument since when
        # serializing the exception in multi-processing, the exception is
        # initialized without the arguments and then the message is set.
        if not message:
            # Exceptions should not be created without a message.
            raise RuntimeError(
                'ServerException created without a message. %s' % (
                    traceback.format_stack(),))
        self.message = force_unicode(message)
        super(_BaseException, self).__init__(message)

    def __repr__(self):
        return '%d:%s' % (id(self), self.message)

    def __str__(self):
        return self.message

    def __eq__(self, other):
        return self.message == other.message

    def __ne__(self, other):
        return self.message != other.message


class ServerException(_BaseException):
    """
    Generic exceptions raised by the server.
    """


class AbortFileException(ServerException):
    """
    Raised when a transfer job should be aborted for a single file
    and no longer retried.
    """
    def __init__(self, message):
        message += ' Retry aborted.'
        super(AbortFileException, self).__init__(message)


class ConnectException(ServerException):
    """
    Raised when failing to connect.
    """
    def __init__(self, message):
        message = 'Remote connection error. ' + message
        super(ConnectException, self).__init__(message)


class HTTPClientException(ServerException):
    """
    Generic exception raised by the HTTP client side.
    """
    def __init__(self, message, path, code, phrase):
        phrase = force_unicode(phrase)
        message = '%s (%s) %s:%s' % (message, path, code, phrase)
        super(HTTPClientException, self).__init__(message=message)
        self.path = path
        self.code = code
        self.phrase = phrase


class ConfigurationException(ServerException):
    """
    Exception raised by the configuration layer.
    """


class AuthorizationException(ServerException):
    """
    Exceptions raised when an account does not have permissions to perform an
    operation.
    """


class ConfigurationDeleteInUseException(Exception):
    """
    Exceptions raised when deleting a configuration is blocked.

    Is initialized with the `uuid` of the configuration which was trying to be
    deleted and a list of other configuration names still using a reference.
    """

    def __init__(self, uuid, usage):
        self.uuid = uuid
        self.usage = usage


class BotException(ServerException):
    """
    Error used in system tests to signal an error in bot operation.
    """

    def __init__(self, message=None):
        if message is None:
            message = 'No details.'
        self.message = message


class FTPBotException(BotException):
    """
    An error from FTP client containing error code and response message.
    """
    def __init__(self, message):
        self.message = message
        parts = message.split(' ', 1)
        if len(parts) == 1:
            self.code = '000'
            self.response = parts[0]
        else:
            self.code = parts[0]
            self.response = parts[1]


class PermissionException(ServerException):
    """
    The action was not permitted.
    """


class SSLException(Exception):
    """
    Error at the SSL/TLS layer raised by the test bots.
    """

    def __init__(self, error):
        self.message = '%s\n%s' % (
            force_unicode(error), force_unicode(traceback.format_exc()))


class InvalidAuthentication(_BaseException):
    """
    Raised when credentials could not be authenticated due to reasons which
    are independent of the credentials content.
    """


class CredentialsForbidden(_BaseException):
    """
    Raised when credentials were forbidden and should not be tried by
    other methods.

    `response` is a dict with extra info received while the credentials were
    rejected.
    """

    def __init__(self, message='Not yet defined.', response=None):
        super(CredentialsForbidden, self).__init__(message=message)
        self._response = response

    @property
    def response(self):
        return self._response


class InsufficientAuthentication(_BaseException):
    """
    Raised when the current credentials were accepsted but more credentials
    are requried to fully authenitcate the account.
    """
    def __init__(self, remaining_credentials):
        message = 'Insufficient credentials provided. Still required: %s.' % (
            ', '.join(remaining_credentials),)
        self.remaining_credentials = remaining_credentials
        super(InsufficientAuthentication, self).__init__(message=message)


class ConnectionAborted(error.ConnectionAborted):
    """
    Connection was aborted by the server.

    This fix the docstring of Twisted error.
    """

    def __init__(self, message='Connection was aborted by the server.'):
        self.message = message

    def __str__(self):
        return self.message


class HTTPRequestRejectedException(Exception):
    """
    Exception raised when the HTTP request is rejected to signal the
    HTTP protocol to stop processing the request, stop reading or writing on
    the transport as most probably it is already closed.

    Before raising this exceptions code should log or handle the error.
    """
