# Copyright (c) 2014 Adi Roiban.
# See LICENSE for details.
"""
Shared code by protocol factories.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from twisted.internet import reactor

from twisted.internet.protocol import Factory
from twisted.protocols import policies
from zope.interface import implementer

from chevah.server.proxy.proxy_protocol import proxify_v2

from chevah.server.commons.interface import (
    IServiceFactory,
    ISSLServiceFactory,
    )


@implementer(IServiceFactory)
class ServiceFactoryBase(policies.LimitTotalConnectionsFactory, object):
    """
    Base code for a service factory.

    This should be set as first inherited class so make sure build and stop
    protocols are overwritten.

    This is a mixed of Twisted and own code so variables / methods name
    conventions are mixed.
    """
    # LimitTotalConnectionsFactory variables.
    connectionCount = 0
    overflowProtocol = None

    #: Protocol used when failing to build the normal protocol.
    internalErrorProtocol = None

    # Use to scheduler calls.
    _scheduler = reactor
    # Use to identify which service manager has created this factory so that
    # we can emit events on behalf of the runnable.
    _creator = None

    # Use to help with testing and get hold of a server-side connection.
    _last_protocol = None

    def __init__(self, service_configuration, on_error):
        self._service_configuration = service_configuration
        self._onError = on_error
        super(ServiceFactoryBase, self).__init__()

    def emitEvent(self, event_id, avatar=None, peer=None, data=None):
        """
        Emit an event for this service factory.
        """
        # Delegate to the process event emitter and signal the event creator.
        self._service_configuration.root.emitEvent(
            event_id=event_id,
            avatar=avatar,
            peer=peer,
            data=data,
            creator=self,
            )

    def buildProtocol(self, addr):
        """
        See: `IServiceFactory`.
        """
        wrapped_protocol = None

        def get_debug(wrapped_protocol):
            """
            Return the debug callback for a protocol.
            """
            return lambda details: wrapped_protocol.emitEvent(
                '20000', data={'message': details})

        try:
            new_protocol = (
                policies.LimitTotalConnectionsFactory.buildProtocol(
                    self, addr))

            wrapped_protocol = new_protocol.wrappedProtocol
            wrapped_protocol.idle_connection_timeout = (
                self.idle_connection_timeout)

            if not self._service_configuration.parent.use_ssl:
                # Enable direct Proxy Protocol V2 for non TLS ports.
                # For SSL protocols, the proxy is enabled at the TLS layer.
                proxify_v2(wrapped_protocol)

            if self._service_configuration.parent.debug:
                wrapped_protocol.debug = get_debug(wrapped_protocol)
            else:
                wrapped_protocol.debug = lambda _: None
            self._configureProtocol(wrapped_protocol)

        except Exception:
            if not self.internalErrorProtocol:
                raise

            import traceback
            details = traceback.format_exc()
            new_protocol = self.internalErrorProtocol()
            new_protocol.factory = self
            new_protocol.details = details

        self._last_protocol = new_protocol
        return new_protocol

    def _configureProtocol(self, protocol):
        """
        Called after protocol was created to allow adding instance specific
        configurations.
        """
        pass

    @property
    def root(self):
        """
        The root of the object hierarchy.
        """
        return self._service_configuration.root

    @property
    def service_configuration(self):
        """
        See: `IServiceFactory`.
        """
        return self._service_configuration

    @property
    def type(self):
        """
        See: `IServiceFactory`.
        """
        return self._service_configuration.parent.type

    @property
    def uuid(self):
        """
        See: `IServiceFactory`.
        """
        return self._service_configuration.parent.uuid

    @property
    def name(self):
        """
        See: `IServiceFactory`.
        """
        return self._service_configuration.parent.name

    @property
    def timeOut(self):
        """
        Alias for idle_connection_timeout.

        Obsolete:
        The protocol should use factory.timeOut but rather us its own
        self.timeOut.
        """
        # This is mainly here due to ugly FTP code, but we should clean
        # it someday.
        return self.idle_connection_timeout

    @property
    def idle_connection_timeout(self):
        """
        See: `IServiceFactory`.
        """
        value = self.service_configuration.idle_connection_timeout
        if not value:
            return None
        else:
            return value

    @property
    def connectionLimit(self):
        """
        See: `IServiceFactory`.

        As required by L{policies.LimitTotalConnectionsFactory}

        When set to 0, returns None to comply implement disabled semantic
        as required by L{policies.LimitTotalConnectionsFactory}.
        """
        value = self.service_configuration.maximum_concurrent_connections
        if not value:
            return None
        else:
            return value

    def startFactory(self):
        """
        See: `Factory`.
        """
        super(ServiceFactoryBase, self).startFactory()
        # FIXME:3245:
        # For SSL factories this is called TWICE and should be fixed.
        # Once we have a dedicated factory which will call start factory
        # before accessing the context we can alway call _configureSSL
        # here.
        if (
            ISSLServiceFactory.providedBy(self) and
            not self.ssl_context_factory
                ):
            self._configureSSL()

    def stopFactory(self):
        """
        Called before service is stopped.
        """
        # Support for SSL services is added as a mixin and we can not
        # overwrite stopFactory() in the mixin.
        if ISSLServiceFactory.providedBy(self) and self.ssl_context_factory:
            self.ssl_context_factory.stop()
        return super(ServiceFactoryBase, self).stopFactory()


class FactoryForProtocol(Factory):
    """
    A factory which return a single instance of the protocol.

    Note that this takes an instance, while Factory.forProtocol takes a class.
    """

    def __init__(self, instance):
        self._instance = instance

    def buildProtocol(self, address):
        """
        Called after connected to the peer.
        """
        self._instance.factory = self
        return self._instance
