# Copyright (c) 2011 Adi Roiban.
# See LICENSE for details.
"""
Module that returns various information regarding service configuration.

This is also used by the legacy Python base service runner.

This is kept as a private CLI tool.
"""
from __future__ import print_function, unicode_literals
from __future__ import absolute_import

from optparse import OptionParser
import platform
import sys
import six

from chevah.server import native_string
from chevah.server.instance_name import (
    INSTANCE_NAME
    )
from chevah.server.commons.constant import (
    PRODUCT_NAME,
    PRODUCT_VERSION,
    SERVICE_NAME,
    )


# Resolve the Windows Service based on current installation instance.
WINDOWS_SERVICE_NAME = SERVICE_NAME
WINDOWS_DISPLAY_NAME = 'SFTPPlus MFT'
if INSTANCE_NAME != PRODUCT_NAME:
    WINDOWS_SERVICE_NAME = INSTANCE_NAME.replace(' ', '-').lower()
    WINDOWS_DISPLAY_NAME = INSTANCE_NAME


def main():
    '''Main module entry.'''

    parser = OptionParser()

    parser.add_option(
        '--service-name', action='store_true', dest='service_name',
        default=False,
        help='Get Windows service name.',
        )

    parser.add_option(
        '--os-arch', action='store_true', dest='os_arch',
        default=False, help='Get operating system architecture.',
        )
    parser.add_option(
        '--os-name', action='store_true', dest='os_name',
        default=False, help='Get operating system name.',
        )
    parser.add_option(
        '-v', '--version', action='store_true', dest='version', default=False,
        help='Get server version.',
        )

    options = parser.parse_args()[0]

    # The return is called after sys.exit(),
    # since when we test this method, we monkey
    # path sys.exit so that it will not exit the running python program.

    if options.version:
        show_version()
        sys.exit(0)
        return

    if options.os_arch:
        show_os_arch()
        sys.exit(0)
        return

    if options.os_name:
        show_os_name()
        sys.exit(0)
        return

    if options.service_name:
        sys.exit(show_service_name())
        return

    parser.print_help()
    sys.exit(1)
    return


def show_version():
    """
    Ouput the current server version.
    """
    version = PRODUCT_VERSION
    if six.PY2:
        version = version.encode('utf-8')

    print(version, end=native_string(''))


def show_os_arch():
    '''Ouput operating system architecture.'''
    print(platform.architecture()[0], end=native_string(''))


def show_os_name():
    '''Ouput operating system name.'''
    print(platform.system(), end=native_string(''))


def show_service_name():
    """
    Ouput the Windows service name.

    This is used by the Windows install/uninstall scripts.
    """
    name = WINDOWS_SERVICE_NAME
    if six.PY2:
        name = name.encode('utf-8')
    print(name, end=native_string(''))
    return 0


if __name__ == '__main__':
    main()
