# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
Common interfaces used by Chevah products.
"""
from __future__ import absolute_import, unicode_literals

from twisted.conch.interfaces import ISFTPServer as TwistedISFTPServer
from twisted.cred.portal import IRealm
from twisted.web.iweb import IRequest
from twisted.web.resource import IResource
from zope.interface import Attribute, Interface, taggedValue

from chevah_compat.interfaces import (
    IFileSystemAvatar,
    IProcess,
    )
from chevah.server.commons.constant import TYPE_NAME


class SecretSyncAttribute(Attribute):
    """
    An attribute that is exposed only for configuration synchronization.
    """


class PublicAttribute(Attribute):
    """
    A public attribute which can only be read.
    """


class PublicWritableAttribute(PublicAttribute):
    """
    A public attribute which can be read and written.
    """


class PublicSyncAttribute(PublicAttribute):
    """
    A public attribute which can be read and sync.

    It should not be written outside of the sync function.
    """


class PublicSectionAttribute(Attribute):
    """
    A public section attribute. This is a collection of attributes and
    can not be written.
    """


class LocalPublicSectionAttribute(PublicSectionAttribute):
    """
    A public section attribute that should not be synchronized but available
    to the JSON-PRC for local operations.
    """


class IPublicObject(Interface):
    """
    Marker interface for exportable/public objects.
    """


class IObserver(Interface):
    """
    A simple observer/notify implementation.
    """

    def unsubscribe(name=None, callback=None):
        """
        Unsubscribe callbacks.

        If `callback` is None, all callbacks for `name` will be removed.
        If `name` is None all callbacks will be removed.
        """

    def subscribe(name, callback):
        """
        Subscribe the callback to signal with `name`.
        """

    def notify(name, signal=None):
        """
        Trigger all subscribers with name.

        Return a list with results from all subscribers.
        """


class IConfigurationProxy(Interface):
    """
    Interface for configurations objects.

    Returning or setting non-text value should be deprecated.
    The low-level configuration should only handle key-value options without
    types.

    The types should be handled at the higher level.
    """

    def save():
        """
        Persist the configuration..
        """

    def hasSection(section):
        '''Returns True if `section` exists.'''

    def addSection(section):
        '''Add `section` to the configuration.'''

    def hasOption(section, option):
        """
        Returns True if `section` contains `option` as a custom value.
        """

    def definesOption(section, options):
        """
        Returns True if `section` directly defines `option`.
        """

    def removeSection(section):
        """
        Remove `section` from configuration.
        """

    def renameSection(section, new_name):
        """
        Rename`section` to new_name.
        """

    def removeOption(section, option):
        """
        Remove `option` from `section`.
        """

    def getString(section, option):
        """
        Return the String value for `option` from `section`.

        Trailing and leading quotes are removed.
        """

    def getRawString(section, option):
        """
        Return the String value for `option` from `section` not modified.
        """

    def setString(section, option, value):
        '''Set the String value for `option` from `section`.'''

    def getInteger(section, option):
        """
        Return the Integer value for `option` from `section`.
        """

    def getIntegerOrNone(section, option):
        """
        Return the Integer value for `option` from `section` or the special
        `None` value.
        """

    def setInteger(section, option, value):
        """
        Set value as integer number.
        """

    def setIntegerOrNone(section, option, value):
        """
        Set value as integer number or the special `None` value.
        """

    def getBoolean(section, option):
        '''Return the Boolean value for `option` from `section`.'''

    def getBooleanOrInherit(section, option):
        """
        Return the Boolean value for `option` from `section` or the
        INHERIT value.
        """

    def setBoolean(section, option, value):
        '''Set the Boolean value for `option` from `section`.'''

    def setBooleanOrInherit(section, option, value):
        """
        Set the Boolean value for `option` from `section` or the INHERIT
        value.
        """

    def getFloat(section, option):
        """
        Return the value for `option` from `section` as floating number.
        """

    def setFloat(section, option, value):
        """
        Set value as floating number.
        """


class INodeConfiguration(Interface):
    """
    Interface for node variable configuration values.

    This is designed to be make it easy to adapt the JSON-RPC IProperty API
    to the low-level configuration.
    """

    def resetINI(content):
        """
        Overwrite the internal state.
        """

    def getNodeConfiguration():
        """
        Return a dictionary with the full node configuration.

        Each section is a key of the root dictionary.
        Each section is another dictionary of keys and values.
        """

    def setNodeVariable(section, option, value):
        """
        Set the value for a node variable.

        Will automatically crate the section, if section doesn't exist.
        """

    def removeNodeVariable(section, option):
        """
        Remove the value for `option` from `section`.

        When section is left empty, this also removes the section.
        """


class IProperty(Interface):
    """
    Private interface use to share attributed related to properties
    handling.

    This defined a CRUD interface for managing configuration:
    * traverse
    * create
    * get (read)
    * set (update)
    * delete

    All methods receive a `property_path` is in the format
    `section/property` or `section/subsection`.

    Configuration values are passed as primitive data
    (string/integer/None/boolean) and complex objects are not supported.
    Primitive data can be grouped in arrays or dictionaries.

    The class implementing this interface should also define one of the
    following properties to allow exporting them:
    * PublicAttribute
    * PublicWritableAttribute
    * PublicSectionAttribute

    It will emit an 'update' event after a batch configuration was changed,
    but also emit individual 'update-PROPERTY_NAME' signals for each
    configuration option which was changed.
    """

    parent = Attribute(
        'The parent of these properties. None if this is root.')

    root = Attribute(
        'The root of the property tree.')

    def traversePath(property_path):
        """
        Parse property_path and return head and tail.

        None is returned for head for tail if they don't exists.
        """

    def getPublicAttributeNames():
        """
        Return the list of all names of public read attributes.
        """

    def getPublicWritableAttributeNames():
        """
        Return the list of all names of public writable attributes.
        """

    def getSyncAttributeNames():
        """
        Return the list of all names of attributes that can be written
        as part of the sync operation.
        """

    def getPublicSectionNames():
        """
        Return the list of all names for public sections.
        """

    def getAll():
        """
        Return all available public sections.
        """

    def getValue(name):
        """
        Return the value of attribute with `name`.

        WARNING. This is the low-level method that is not protected.
        Most of the time you should use `getProperty`

        Raise NoSuchAttributeError if attribute does not exist.
        """

    def setValue(name, value):
        """
        Set value for attribute with name.

        WARNING. This is the low-level method that is not protected.
        Most of the time you should use `setProperty`

        Raise NoSuchAttributeError when attribute does not exist.

        Emit a signal with `name` and with `new_value` and `old_value`

        Also emits the `update` signal after all setAttributes are settled.
        """

    def getSection(name):
        """
        Return the section with `name`.

        Return `None` if section does not exist.
        """

    def createProperty(property_path, value):
        """
        Add property denoted by `property_path`.
        """

    def getFilteredProperty(getter, property_path):
        """
        Return all the values of the properties with names returned by
        `getter`.
        """

    def getProperty(property_path=None):
        """
        Return a dictionary for public properties matching property_path.

        When property_path is None returns all property members as key-value.
        {
            property1_name: property1_value,
            property2_name: property2_value,
            sub_section1: { prop1: value2 },
        }

        When a property_path is specified, it will return only the properties
        for that path.

        For example for `sec1/sec2` it will return:
        {
            'prop1': value1,
            'prop2': value2,
        }

        For example for `sec1/sec2/prop1` it will return `value1`.
        """

    def getPropertyForSync(property_path):
        """
        Return a dictionary with the configuration options that should be
        synchronized.
        """

    def setProperty(property_path, value):
        """
        Set property value denoted by `property_path` via the public API.

        This is a method directly exosed by our JSON-RPC API.
        """

    def setPropertyWithoutNotification(property_path, value):
        """
        Set property value denoted by `property_path` without emitting
        any signals.
        """

    def deleteProperty(property_path):
        """
        Delete property denoted by `property_path`.
        """


class IEventCreator(IPublicObject):
    """
    Something emitting events.

    Usually a runnable.
    """
    type = PublicAttribute('Type of this thing as text.')
    uuid = PublicAttribute('UUID of this thing.')

    name = PublicAttribute('Human readable name of this component.')

    def emitEvent(event_id, data, avatar=None):
        """
        Emit an event as generated by this component.

        This method is designed as a proxy which adds the event creator.
        """


class IProcessRunnables(IProperty):
    """
    The collection of all runnables for this process.
    """
    status = PublicSectionAttribute('Server status.')

    services = PublicSectionAttribute('Service management.')
    authentications = PublicSectionAttribute('Authentications management.')

    transfers = PublicSectionAttribute('Transfers management.')
    locations = PublicSectionAttribute('Locations management.')

    event_handlers = PublicSectionAttribute('Event handlers.')
    resources = PublicSectionAttribute('Resources management.')


class IServerProcess(IProcess, IProperty):
    """
    Objects exported by the main process.
    """

    public_identity = Attribute(
        'Delegation for the IEventCreator for this process.')
    configuration = PublicSectionAttribute('Main configuration.')

    # FIXME:6641:
    # Should be renamed to `operation` and check role permissions.
    runnables = PublicSectionAttribute('Collection of runnables')

    # FIXME:6521:
    # To be removed in v6 they were moved to `runnables`.
    authentications = PublicSectionAttribute('Authentications management.')
    event_handlers = PublicSectionAttribute('Event handlers.')
    status = PublicSectionAttribute('Server status.')
    services = PublicSectionAttribute('Service management.')
    transfers = PublicSectionAttribute('Transfers management.')
    locations = PublicSectionAttribute('Locations management.')
    resources = PublicSectionAttribute('Resources management.')


class _IIdentityOptions(IProperty):
    """
    Configuration section provided by the identity service.
    """

    accounts = PublicSectionAttribute("Section with all accounts.")
    administrators = PublicSectionAttribute(
        "Section with all administrators.")
    groups = PublicSectionAttribute("Section with all groups.")
    roles = PublicSectionAttribute("Section with all roles.")


class IApplicationConfiguration(IProperty, _IIdentityOptions):
    """
    Root configuration object.
    """

    def __init__(configuration_path, configuration_file):
        """
        Initialize the application configuration.

        configuration_path and configuration_file are mutually exclusive.
        """

    authentications = PublicSectionAttribute('All authentications.')
    event_handlers = PublicSectionAttribute('All event handlers.')
    services = PublicSectionAttribute(
        'Group of all available services for this server.')
    server = PublicSectionAttribute(
        'Configuration for general server configuration.')
    transfers = PublicSectionAttribute('All transfers.')
    locations = PublicSectionAttribute('All locations.')
    resources = PublicSectionAttribute('All resources configurations.')
    node_variables = LocalPublicSectionAttribute(
        'All node specific configurations.')

    def validateConfiguration():
        """
        Check that current configuration is valid.

        Raise an error when configuration is not valid.
        """


class ISyncConfiguration(Interface):
    """
    Methods shared to sync the configuration.
    """

    def setPropertyFromSync(source):
        """
        Update the configuration using the source value which is a dict that
        contains a UUID.
        """


class ISetConfiguration(IProperty, ISyncConfiguration):
    """
    Configuration section implementing CRUD interface.
    """

    def getSectionByName(name):
        """
        Return the configuration section with `name`, or `None` if
        no section was found.
        """

    def onChildCreate(section_name, value, uuid):
        """
        Called when start creating a new child property.

        Return a new dictionary based on `value`.

        This can be use for validating or filtering of input data
        for createProperty.
        """

    def beforeChildDelete(uuid):
        """
        Called before deleting a child property with `uuid`.

        Raise `ConfigurationDeleteInUseException` if child shouldn't be
        deleted.
        """


class IAccountsConfiguration(ISetConfiguration):
    """
    All accounts provided by the identity service.
    """

    def getApplicationAccountConfiguration(account_name):
        """
        Return configuration for an application account.
        """


class IEntityConfiguration(Interface):
    """
    An uniquely identifiable configuration object.
    """

    enabled = PublicWritableAttribute(
        'True when entity is launched/enabled at startup.')

    uuid = PublicSyncAttribute(
        """
        Unique identifier used by this entity.
        """)

    name = PublicWritableAttribute(
        """
        Name of the entity.

        This is an human readable string used for referencing this entity.

        The characters are restricted to alpha-numeric values without spaces.
        """)

    description = PublicWritableAttribute(
        """
        Description for the entity.

        This can be a long text with comments or details about the entity.
        """)

    def _migrateINI():
        """
        Called at init time to do any configuration migration for this
        component.

        It should be used only for migrating the low-level .ini format
        representation of the configuration of this component.

        This is not called when creating using createProperty().
        """


class IServerOptionsSection(IEntityConfiguration, ISyncConfiguration):
    """
    Options for server process.
    """

    enabled = Attribute(
        """
        Always returns True.

        This is here to comply with Entity interface but it is read only.
        """)

    password_minimum_strength = PublicWritableAttribute(
        'Minimum strength of the password.')
    password_minimum_length = PublicWritableAttribute(
        'Minimum length of the password.')
    password_history = PublicWritableAttribute(
        'The number of unique new passwords that must be associated '
        'with a user account before an old password can be reused.'
        )
    password_hashing_scheme = PublicWritableAttribute(
        'Hash used for stored passwords.')

    authentications = PublicWritableAttribute(
        'List with default authentications for the file transfer services.')

    account = PublicWritableAttribute(
        'Account under which the server should execute')

    product_key = PublicWritableAttribute(
        'Product key for this installation.')

    umask = PublicWritableAttribute(
        """
        umask used for creating new files and folders.

        Stored as string.
        Ignored on Windows.
        """)

    umask_as_int = Attribute('Return umask value as integer.')

    ssl_certificate = PublicWritableAttribute(
        'Default certificate used by TLS services.')
    ssl_key = PublicWritableAttribute('Default key used by TLS services.')

    ssh_host_private_keys = PublicWritableAttribute(
        'Default host keys used by SSH services.')

    # Export defaults UUID in the config to prevent these components to be
    # removed.
    default_uuids = PublicAttribute(
        'List with UUID which should not be removed.')


class IServerStatus(IProperty):
    """
    Status for server process.
    """
    uuid = PublicAttribute(
        'UUID of this server as found in the configuration.')
    name = PublicAttribute(
        'Name of this server as found in the configuration.')

    start_time = PublicAttribute('Timestamp when server was started.')

    restart_required = PublicAttribute(
        """
        Sequence of configuration options due to which the process requires
        restart.
        """)

    startup_configuration = PublicAttribute(
        """
        Dict with configuration values at server startup.
        """)

    os_family = PublicAttribute(
        """
        Name of the OS family under which the server is running.
        """)

    os_name = PublicAttribute(
        """
        Name of the OS under which the server is running.

        This is not the distribution.
        """)

    can_impersonate_local_accounts = PublicAttribute(
        'True when server has the OS permissions to authenticate and run '
        'under local OS accounts')

    can_create_local_home_folders = PublicAttribute(
        'True when server has the OS permissions to create home '
        'folders for OS accounts')

    can_get_local_home_folders = PublicAttribute(
        'True when server has the OS capabilities to find the path to home '
        'folders for OS accounts')

    openssl_version = PublicAttribute(
        'Version of the OpenSSL library used by the process.')

    instance_name = PublicAttribute('The name of this installation.')
    install_path = PublicAttribute('The path of this installation.')
    cwd = PublicAttribute('The current working directory.')
    python_version = PublicAttribute('Version of python used by the process.')
    os_user = PublicAttribute(
        'Name of the OS account under which the main SFTPPlus process runs.')

    windows_legacy_service = PublicAttribute(
        'True if process was started using  the legacy Windows service.')


class IRunnablesManager(IProperty):
    """
    Manager of runnable of same type.

    All members should implement IRunnable.
    """

    def getStarted():
        """
        Return all runnables that are started.
        """

    def start():
        """
        Start all enabled runnables.
        """

    def stop():
        """
        Stop all runnables.

        Return a deferred which fires when all runnables are stopped.
        """

    def restart():
        """
        Restart all runnables that are currently running.

        Return a deferred which fires when all runnables are restarted.
        """


class IServicesManager(IRunnablesManager):
    """
    Marker interface for manager for available services.
    """


class IRunnable(IProperty, IEventCreator, IObserver):
    """
    Persistent object which can be started / stopped and has a state.
    """

    PROPERTIES_REQUIRING_RESTART = Attribute(
        """
        Properties which when changed a restart is required.
        """
        )

    scheduler = Attribute('Reactor used to handle time based tasks.')

    family = PublicAttribute(
        """
        Human readable family of this instance.

        Runnables of different types can be of the same family.
        """)

    enabled = Attribute('True if runnable is enabled at startup.')

    state = PublicAttribute('The current state of the service.')
    started = PublicAttribute(
        """
        Return `True` when runnable was started.

        It was successfully started or in the process of being started
        and not yet stopped or failed.
        """)
    operating = PublicAttribute(
        """
        Return `True` is runnable is fully operating and ready to execute
        the requested actions.
        """)
    failed = PublicAttribute(
        """
        True if the component has failed or is failing right now.
        """)

    start_time = PublicAttribute(
        'Unix timestamp when service was started.')

    stop_time = PublicAttribute(
        'Unix timestamp when service was stopped.')

    last_error = PublicAttribute(
        """
        Last error that caused the service to stop.

        None if service is not enabled or is running.
        """)

    configuration = Attribute('Configuration attached to this runnable.')

    restart_required = PublicAttribute(
        'Sequence of configuration options due to which it requires restart.')

    startup_configuration = PublicAttribute(
        """
        Dict with configuration values at service startup.
        """)

    def start(avatar=None):
        """
        Start the runnable.

        Returns a deferred which is fired when object is started.

        Deferred will return `True` on successful start and `False` otherwise.
        On failure it sets the `last_error` attribute.

        `avatar` if the one requesting the action.
        """

    def stop(avatar=None):
        """
        Stop the runnable.

        Returns a deferred which is fired when object is stopped.

        Deferred will return `True` on successful stop and `False` otherwise.
        On failure it sets the `last_error` attribute.

        `avatar` if the one requesting the action.
        """

    def restart(avatar=None):
        """
        Restart the runnable.

        Returns a deferred which is fired when object is restarted. If the
        runnable is not running it will be started.

        Deferred always succeed and on failure it sets the last_error.

        `avatar` if the one requesting the action.
        """

    def fail(error, avatar=None):
        """
        Stop the runnable due to unrecoverable error encountered while
        running.

        Returns a deferred which is fired when the runnable is stopped.

        Deferred will always succeed and set `last_error` to specified error.
        """


class IServiceManager(IRunnable):
    """
    Marker interface for a service.
    """
    taggedValue('family', TYPE_NAME.SERVICE)


class IProtocolOptions(IProperty):
    """
    Configuration for a protocol.
    """
    PROPERTIES_REQUIRING_RESTART = Attribute(
        """
        List of properties for which services requires a restart
        in order to apply them.

        Example: 'port', 'configuration/ftps_enabled'.
        """
        )


class IAddressPortPublicOptions(Interface):
    """
    Configuration having an IP address and port.

    The port number can't be 0.
    A value of 0, means that the configuration has an invalid value.
    """
    address = PublicWritableAttribute('Host address.')
    port = PublicWritableAttribute('Port number.')


class IURLOptions(Interface):
    """
    Configuration having an URL configuration.

    This is not compatible with IAddressPortPublicOptions and should not be
    mixed.
    """
    url = PublicWritableAttribute(
        """
        URL for the remote service.
        """
        )

    # These are here for event handling to allow creating an avatar.
    # They mirror the URL, so they should be read-only.
    # They are also used to get the remote server certificate and CA chain.
    address = Attribute('IP address.')
    port = Attribute('Port number.')


class IURLFallbackOptions(Interface):
    """
    Configuration having multiple URLs configuration.

    """
    url = PublicWritableAttribute(
        """
        List of URLs to the remote service.
        """
        )


class _IProxyOptions(Interface):
    """
    Configuration having the proxy options.
    """
    proxy = PublicWritableAttribute(
        'URI like configuration for the proxy to be used.')


class IPasswordOptions(Interface):
    """
    Configuration entities which have a password option.
    """
    password = SecretSyncAttribute('Read-only access to raw password.')
    password_update = PublicWritableAttribute(
        'Write-only field for updating the password with clear text input.')
    # Read-only for the API as this is set internally together with the
    # password update.
    last_password_update = PublicAttribute(
        'Timestamp for last password change.')
    last_passwords = PublicAttribute(
        'List of last passwords used for this account.')


class IUsernamePasswordPublicOptions(IPasswordOptions):
    """
    Configuration have username and plain text password.
    """
    username = PublicWritableAttribute('Username used for this section.')


class IAutoFailOptions(Interface):
    """
    Configuration having an number of errors after which it fails.
    """
    fail_after_errors = PublicWritableAttribute(
        'Number of errors after which the runnable will fail.')


class ITypeOptions(Interface):
    """
    Configuration options for sections containing type which can not be
    changed at runtime.
    """

    type = PublicSyncAttribute('Type of this configuration.')


class IServiceConfiguration(
        IEntityConfiguration,
        IAddressPortPublicOptions,
        ITypeOptions,
        ):
    """
    A section from the main configuration file describing a service.

    This is the generic sections used to launch all services.
    """

    authentications = PublicWritableAttribute(
        'List with UUIDs of the authentication methods for this service.')

    debug = PublicWritableAttribute(
        'True when service should emit debug messages.')

    configuration = PublicSectionAttribute(
        '`IProtocolOptions` for this service.')

    use_ssl = Attribute(
        'True if protocol should start as a listener over SSL')

    https_redirect = Attribute(
        'True if protocol should redirect HTTP to HTTPS over the same port.')


class IServiceFactory(IEventCreator):
    """
    Factory for creating protocol instances for new connections.
    """
    connectionLimit = Attribute(
        'Total number of allowed concurrent connections. 0 to disable.')
    connectionCount = Attribute('Current number of active connections.')
    overflowProtocol = Attribute(
        'Protocol to use when connection count is exceeded.')

    internalErrorProtocol = Attribute(
        'Protocol used when failing to build the normal protocol.')

    service_configuration = Attribute('Configuration for this service.')
    idle_connection_timeout = Attribute(
        'Number of seconds after which idle protocols are disconnected.')

    def buildProtocol(addr):
        """
        Create a new instance of protocol for peer.

        If there are too many connections is return an instance of
        `overflowProtocol`

        When protocol initialization fails it return an instance of
        `internalErrorProtocol`.
        """


class IWebSite(IServiceFactory):
    """
    An HTTP or HTTPS generic factory aka site.
    """

    resource = Attribute('Root resource for this site.')
    headers = Attribute('List of headers set for all responses.')
    accepted_origins = Attribute('List of accepted origins.')
    base_url_path = Attribute(
        """
        The base URL used when the HTTP requests are via a reverse proxy.
        This needs to be an absolute URL without a trailing slash.
        """)
    client_forwarded_header = Attribute(
        'Name of the header from which to extract client source address.')


class IFileTransferSite(IWebSite):
    """
    The HTTP or HTTPS service providing file transfer capabilities.
    """
    public_error = Attribute('If low-level login errors should be public.')
    triggers = Attribute('List of triggers for this site. For all users.')
    metadata_configuration = Attribute(
        'Parsed configuration for metadata header.')
    as2_complete = Attribute(
        'Segments inside user home folder where to store AS2 received files.')
    as2_pending = Attribute(
        'Segments inside user home folder where to store temporary AS2 files.')
    as2_key = Attribute('Private key for decryption and MDN signing.')
    as2_certificate_asn1 = Attribute('Certificate ASN1 value for MDN signing.')
    as2_certificate_id = Attribute('Certificate ID for MDN signing.')


class ISSLServiceFactory(IServiceFactory):
    """
    A service factory that support SSL/TLS.
    """

    ssl_context_factory = Attribute(
        'SSL context factory used by all SSL sessions.')

    def _configureSSL():
        """
        Prepare SSL/TLS context.

        Will raise a ServerException if initial configuration fails.
        Will call self._onError if SSL related things fail at runtime.
        """


class ISSLContextFactory(Interface):
    """
    A factory generating SSL contexts.
    """

    def start():
        """
        Return a deferred which is fired when the initialization is done and
        the context can be used.
        """

    def stop():
        """
        Called after context is no longer used.
        """


class ISSLOptions(Interface):
    """
    Configuration for components implementing SSL/TLS transport.
    """

    ssl_domains = PublicWritableAttribute(
        'List of domains used when generating the certificate.')

    ssl_certificate_value = PublicAttribute(
        """
        The actual SSL certificate used by the service.
        Might be the global one.
        """)
    ssl_key_value = PublicAttribute('The actual SSK key used by the service.')

    ssl_certificate = PublicWritableAttribute(
        'Path to the SSL certificate for the component.')
    ssl_key = PublicWritableAttribute(
        'Path to the private SSL key for the component.')

    ssl_key_password = Attribute(
        'Password used for storing the encrypted SSL key.')
    ssl_key_password_update = PublicWritableAttribute(
        'Write-only access to password used for storing the SSL key.')
    ssl_certificate_authority = PublicWritableAttribute(
        'Path to the certificate authority\'s certificate.')
    ssl_certificate_revocation_list = PublicWritableAttribute(
        'Path to a file containing certificate revocation list.')
    ssl_certificate_revocation_list_refresh = PublicWritableAttribute(
        'Number of seconds after which local CRL is read again.')
    ssl_cipher_list = PublicWritableAttribute(
        'List of allowed ciphers, using OpenSSL format.')
    ssl_allowed_methods = PublicWritableAttribute(
        'List of allowed SSL methods, separated by commas.')


class IConnectionOptions(Interface):
    """
    Configuration section with connection options.
    """

    idle_connection_timeout = PublicWritableAttribute(
        'Number of seconds after an idle peer is disconnected.')
    maximum_concurrent_connections = PublicWritableAttribute(
        'Maximum number of concurrent connections allowed for this service.')


class IFTPServiceOptions(IProtocolOptions, ISSLOptions):
    '''`Service` section for ftp configuration.'''

    banner = PublicWritableAttribute('Text sent at client connection.')
    passive_port_range = PublicWritableAttribute(
        'Tuple with minimum and maximum port used for passive connections.')
    passive_address = PublicWritableAttribute(
        """
        String describing public address advertised to client for passive
        connections.
        """)
    active_source_port = PublicWritableAttribute(
        'Port number used as source port for active connections.')
    enable_password_authentication = PublicWritableAttribute(
        'True if service allows password authentication.')
    enable_ssl_certificate_authentication = PublicWritableAttribute(
        'True if service allows SSL certificate authentication.')
    ftps_explicit_enabled = PublicWritableAttribute(
        'True if explicit FTPS is enabled.')
    ftps_force_secured_command_channel = PublicWritableAttribute(
        'True when commands should only be sent over secured channel.')
    ftps_require_session_reuse = PublicWritableAttribute(
        """
        True when data channel should use the same TLS session as
        previously negotiated over the command channel.
        """)
    idle_data_connection_timeout = PublicWritableAttribute(
        """
        Number of seconds after which idle data transport channel
        is disconnected.
        """)
    ignore_ascii_data_type = PublicWritableAttribute(
        """
        When enabled, ASCII transfers are ignored and data is transfered in
        IMAGE type.
        """)
    ascii_data_type_as_default = PublicWritableAttribute(
        """
        When enabled, ASCII is the default transfer mode.
        """)


class IFTPSIServiceOptions(IFTPServiceOptions):
    """
    Marker interface for FTPS Implicit service.
    """


class ILocationWatchConfiguration(Interface):
    """
    Configuration options for watcher on a location.
    """
    source_uuid = PublicWritableAttribute(
        """
        UUID of the watch location.

        `None` when default local filesystem is used.
        """)

    source_path = PublicWritableAttribute(
        'Path which should be watched.')

    source_filter = PublicWritableAttribute(
        'Glob or regular expression used to filter source files names.')

    recursive = PublicWritableAttribute(
        '`True` when the watch should be recursive.')

    changes_poll_interval = PublicWritableAttribute(
        'Number in seconds after which a new snapshot is taken.')

    stable_interval = PublicWritableAttribute(
        """
        Number in seconds after which changes in a path are considered stable.
        """)

    retry_count = PublicWritableAttribute(
        """
        Number of times to retry failed files.
        """)


class IMonitorConfiguration(ILocationWatchConfiguration):
    """
    Configuration for monitor resource.
    """

    file_age_notification = PublicWritableAttribute(
        """
        Age in seconds after which an event is triggered for files which are
        not modified in that time.
        """)
    file_age_auto_delete = PublicWritableAttribute(
        """
        Age in seconds after which files which are not modified are deleted.
        """)
    monitored_operations = PublicWritableAttribute(
        """
        List of file operations for which to emit an event.
        """)


class _ISSHEncryptionOptions(Interface):
    """
    Shared options for ssh encryption used by both client and server.
    """
    ssh_cipher_list = PublicWritableAttribute('List of allowed ciphers.')


class ISSHServiceOptions(IProtocolOptions, _ISSHEncryptionOptions):
    """
    Service section for SSH configuration.
    """
    sftp = PublicWritableAttribute(
        'True when SFTP is enabled.')
    scp = PublicWritableAttribute(
        'True when SCP is enabled.')
    banner = PublicWritableAttribute(
        'Text set at client connection.')

    ssh_host_keys = PublicAttribute(
        'The actual host keys used by the service. Might be the global one.')

    ssh_host_private_keys = PublicWritableAttribute(
        'Host private keys explicitly set for this service.')

    dh_prime_size = PublicWritableAttribute(
        'Minimum and preferred size of the prime number used for the DH'
        'key exchange.')
    ignore_create_permissions = PublicWritableAttribute(
        """
        When `True`, ignore permission and attributes
        for file open requests.
        """)


class _IHTTPBaseServiceOptions(IProtocolOptions):
    """
    Configuration options available for all HTTP based services, including
    the local manager.
    """
    headers = PublicWritableAttribute(
        'List of raw headers set for all responses.')
    accepted_origins = PublicWritableAttribute('List of accepted origins.')
    client_forwarded_header = PublicWritableAttribute(
        'Name of the header from which to extract client source address.')
    base_url_path = PublicWritableAttribute(
        'URL prefix used when the service is accessed via a reverse-proxy.')


class ILocalManagerServiceOptions(_IHTTPBaseServiceOptions):
    """
    Service section for local manager configuration.
    """


class IHTTPServiceOptions(_IHTTPBaseServiceOptions):
    '''`Service` section for http configuration.'''

    public_account_uuid = PublicWritableAttribute(
        'UUID to account which is used for the public files.')
    public_name = PublicWritableAttribute(
        'URL fragment name for the public site.')
    public_error = PublicWritableAttribute(
        'True if the authentication method rejection error should be public.')
    as2_receive_name = PublicWritableAttribute(
        'URL fragment name for the AS2 site.')
    as2_receive_path = PublicWritableAttribute(
        'Path inside user home folder where to store valid received AS2 files.')
    as2_pending_path = PublicWritableAttribute(
        """
        Path inside user home folder where to store temporary received files,
        before they are validated for AS2.
        """)
    as2_default_filename = PublicWritableAttribute(
        'Name used to store the received data when the AS2 request '
        'is missing the filename.'
        )
    as2_no_mdn_success_text = PublicWritableAttribute(
        'Text present in the response on success when no MDN was requested.')
    as2_receive_key = PublicWritableAttribute(
        'Private key used to sign and decrypt AS2 files.')
    as2_receive_certificate = PublicWritableAttribute(
        'Certificate used to sign the MDN for received AS2 files.')
    announce_session_authentication = PublicWritableAttribute(
        'Flag to hide the session method from www-authenticate header..')
    theme_path = PublicWritableAttribute(
        'Absolute path to the theme files.')
    ui_version = PublicWritableAttribute(
        'The version of the UI to be used.')
    login_footnotes = PublicWritableAttribute(
        'Custom footnote to open a dialog with more info from the login page.')
    languages = PublicWritableAttribute(
        'List of available languages.')
    triggers = PublicWritableAttribute(
        'List of custom actions available to clients .')
    metadata_fields = PublicWritableAttribute(
        """
        Name of header from which to extract metadata fields to associate
        with a file.
        """)


class IHTTPRedirectServiceOptions(_IHTTPBaseServiceOptions):
    """
    Configuration for HTTP redirect service.
    """

    redirect_url = PublicWritableAttribute(
        'URL where requests are redirected.')


class IHTTPSServiceOptions(IHTTPServiceOptions, ISSLOptions):
    '''`Service` section for https configuration.'''


class ILocalManagerAvatarSeed(Interface):
    """
    Avatar seed for the Web Manager service.
    """
    account_type = Attribute('Type of the authenticated credentials.')
    credentials = Attribute('Credentials that were authenticated.')
    roles = Attribute(
        "Roles based on which the administrator was authenticated.")
    uuid = Attribute('UUID of the authenticated manager.')


class IServiceRealm(IRealm):
    """
    Realm for a service.
    """

    main_configuration = Attribute('Main configuration for the application.')
    service_configuration = Attribute('Configuration for the service.')


class _IAuthenticationOptions(Interface):
    """
    Shared configuration options for accounts, group, administrators and roles
    used in the authentication process.
    """
    source_ip_filter = PublicWritableAttribute(
        'List of source IPs rules for this entity.')
    allow_certificate_authentication = PublicWritableAttribute(
        'True if SSL certificate authentication is allowed.')


class IAccountAndGroupOptions(IEntityConfiguration, _IAuthenticationOptions):
    """
    Configuration options for an account or a group.

    These are just the properties as they are persisted
    in the configuration.
    """

    kind_name = PublicAttribute('Human readable of the type of this entity.')

    lock_in_home_folder = PublicWritableAttribute('')
    home_folder_path = PublicWritableAttribute('')

    create_home_folder = PublicWritableAttribute('')
    create_home_folder_owner = PublicWritableAttribute('')
    create_home_folder_group = PublicWritableAttribute('')

    required_credentials = PublicWritableAttribute(
        'List of credentials that are required to authenticate the account.')
    ssh_authorized_keys_path = PublicWritableAttribute(
        'Path to file used for storing ssh keys.')

    permissions = PublicWritableAttribute(
        'List of permissions for accounts from this group.')
    amend_write_name = PublicWritableAttribute(
        'What to do to have unique uploads.')

    virtual_folders = PublicWritableAttribute(
        'List of mapping from virtual path to real paths.')

    as2_require_http_authentication = PublicWritableAttribute(
        "Whether AS2 request need to be authenticated at the HTTP layer")

    password_lifetime = PublicWritableAttribute(
        'Number of days for which the password is valid.')
    disable_on_inactivity = PublicWritableAttribute(
        'Number of days after which an account is disabled if not active.')
    file_write_size = PublicWritableAttribute(
        'Maximum bytes allowed for a single write operation.')


class IAccountConfiguration(IAccountAndGroupOptions, ITypeOptions, IProperty):
    """
    Base properties for an account configuration.
    """
    group = PublicWritableAttribute('Group to which this account belongs.')

    # This is a text value, not a list.
    email = PublicWritableAttribute(
        'Email or comma-separated emails associated with this account.')

    # Authentication related attributes should be moved in a separate
    # interface.
    ssh_authorized_keys_content = PublicWritableAttribute(
        'List of public ssh keys, delimited by new line, each key on a line.')
    created = PublicAttribute(
        """
        Date and time as ISO text when the account was created.

        This is the raw value.
        We have the datetime object in the IServiceAvatar.
        """)
    expire_datetime = PublicWritableAttribute(
        """
        Date and time as ISO string after which the account is no longer
        enabled.
        """)
    multi_factor_authentication = PublicWritableAttribute(
        'MFA URL configuration')

    as2_certificates = PublicWritableAttribute(
        """
        One or more SSL certificates in PEM format used to validate the
        signature for the received files.
        """)

    as2_async_mdn_ca = PublicWritableAttribute(
        """
        One or more CA certificates in PEM format used to validate the
        async MDN HTTPS outgoing connection.
        """)


class ILocalPathOptions(Interface):
    """
    Configuration entities with a local path.
    """

    path = PublicWritableAttribute('Path to local native filesystem.')


class IApplicationAccountConfiguration(
        IAccountConfiguration, IPasswordOptions):
    """
    Application accounts will access the filesystem using the service account.
    """


class IOSAccountConfiguration(IAccountConfiguration):
    """
    Marker interface for operating system accounts.

    For OS accounts the filesystem access is impersonated.
    """


class IExternalAccountConfiguration(IAccountConfiguration):
    """
    Marker interface for an account not defined in configuration.

    These accounts can not be persisted and can not be managed
    (created/removed/changed) by SFPPPlus.
    They can only be used.

    This can be an OS or Application account.
    """


class IGroupConfiguration(IAccountAndGroupOptions):
    """
    Properties for a group configuration.
    """
    home_folder_structure = PublicWritableAttribute(
        'List of folders to be created inside the home folder.')
    allow_own_password_change = PublicWritableAttribute(
        'True if accounts can change their own password.')
    ssh_delete_delay = PublicWritableAttribute(
        'Number of seconds after which the delete operation is performed.')


class IAdministratorConfiguration(
        IEntityConfiguration, _IAuthenticationOptions):
    """
    Properties for an administrator.
    """
    kind_name = PublicAttribute('Human readable of the type of this entity.')
    roles = PublicWritableAttribute(
        'List of UUID for the roles to which this administrator belongs to.')
    multi_factor_authentication = PublicWritableAttribute(
        'MFA URL configuration')
    node_variables = PublicWritableAttribute(
        """
        List of (section, value) configuration for node variables
        for this remote instance.
        """)


class IApplicationAdministratorConfiguration(
        IAdministratorConfiguration, IPasswordOptions):
    """
    Properties for an administrator as stored in the application.
    """
    multi_factor_authentication = PublicWritableAttribute(
        'MFA URL configuration')


class IExternalAdministratorConfiguration(
        IApplicationAdministratorConfiguration):
    """
    Marker interface for external administrators.
    """


class IOSAdministratorConfiguration(IExternalAdministratorConfiguration):
    """
    Marker interface for OS administrators.
    """


class IRoleConfiguration(IEntityConfiguration, _IAuthenticationOptions):
    """
    Properties for a group of administrators, known as a role.
    """

    permissions = PublicWritableAttribute(
        'List of permissions for this role.')


class ITransfersConfiguration(ISetConfiguration):
    """
    Marker interface for the container holding all configured transfers.
    """


class ITransferConfiguration(
        IEntityConfiguration,
        ILocationWatchConfiguration,
        ):
    """
    Configurations available to all transfer types.
    """

    destination_uuid = PublicWritableAttribute(
        """
        UUID of the destination location.

        None when default local filesystem is used.
        """)

    destination_path = PublicWritableAttribute(
        'Path to destination folder of this transfer.')

    destination_temporary_suffix = PublicWritableAttribute(
        'Use a temporary suffix when transferring file to destination.')

    destination_fallback_path = PublicWritableAttribute(
        'Used when transfer to normal destination fails.')

    batch_interval = PublicWritableAttribute(
        'Seconds to wait for files to be part of a batch.')

    batch_marker_path = PublicWritableAttribute(
        'Matching expression for a file name which will trigger the batch.')

    execute_timeout = PublicWritableAttribute(
        """
        Number of seconds after which the external commands execute during
        a transfer are automatically closed.
        """)

    execute_before = PublicWritableAttribute(
        'External command to execute before performing a transfer.')

    execute_after_success = PublicWritableAttribute(
        'External command to execute after a successful transfer.')

    execute_after_failure = PublicWritableAttribute(
        'External command to execute after a failed transfer.')

    execute_on_destination_before = PublicWritableAttribute(
        """
        List of commands to execute on destination location before performing
        a transfer.
        """)

    execute_on_destination_after_success = PublicWritableAttribute(
        """
        List of commands to execute on destination location after a successful
        transfer.
        """)

    execute_on_destination_after_failure = PublicWritableAttribute(
        """
        List of commands to execute on destination location after a failed
        transfer.
        """)

    archive_success_path = PublicWritableAttribute(
        """
        Path of a local folder where a copy all each successfully transferred
        file is archived.
        """)

    archive_failure_path = PublicWritableAttribute(
        """
        Path of a local folder where a copy all each failed file is archived.
        """)

    archive_retention_period = PublicWritableAttribute(
        """
        Number of days after which older archived files are automatically
        removed.
        """)

    archive_format = PublicWritableAttribute(
        """
        The way in which files are archived.
        """)

    retry_wait = PublicWritableAttribute(
        """
        Number of seconds to wait before a retry.
        """)

    schedule = PublicWritableAttribute(
        """
        List of actions (start/stop/run) planned for this transfer.
        """)

    overwrite_rule = PublicWritableAttribute(
        """
        Rule use for overwriting existing remote file.
        """)

    destination_path_actions = PublicWritableAttribute(
        """
        Set of actions to be applied on the destination path.
        """)

    destination_content_actions = PublicWritableAttribute(
        """
        Rules used to transform the content of the transferred file on the fly.
        """)

    transfer_memory_duration = PublicWritableAttribute(
        """
        For how many days to keep a record of transferred files.
        """)
    ignore_duplicate_paths = PublicWritableAttribute(
        """
        True when a file with the same path and name should not be
        successfully transferred twice.
        """)
    minimum_transfer_count = PublicWritableAttribute(
        """
        Minimum number of files expected to be successfully transferred
        in a time interval.
        """)
    minimum_transfer_interval = PublicWritableAttribute(
        """
        A time interval used to count the minimum transferred files.
        """)

    source_filter_age = PublicWritableAttribute(
        """
        Age of files (in seconds) after which source files are ignored.
        """)

    delete_source_on_success = PublicWritableAttribute(
        """
        If this should be a copy or a move transfer.
        """)

    delete_source_parent_delay = PublicWritableAttribute(
        """
        Delay after which the parent directory will also be removed.
        """)


class IAuthenticationsConfiguration(ISetConfiguration):
    """
    Marker interface for the container holding all configured authentication
    methods.
    """


class IAuthenticationConfiguration(IEntityConfiguration, ITypeOptions):
    """
    Shared configurations for a all authentication methods.

    This is also a marker to select configuration type.
    """


class IApplicationAuthenticationConfiguration(IAuthenticationConfiguration):
    """
    Configuration for application accounts authentication.
    """
    allowed_groups = PublicWritableAttribute('List of allowed group UUIDs.')


class IOSAuthenticationConfiguration(IAuthenticationConfiguration):
    """
    Configuration for OS accounts authentication.
    """

    pam_usage = PublicWritableAttribute(
        'Operation mode for PAM authentication.')
    pam_service = PublicWritableAttribute(
        'Name of the PAM service used for authentication.')
    allowed_groups = PublicWritableAttribute('List of allowed OS groups.')
    base_groups = PublicWritableAttribute(
        'List of SFTPPlus groups associated with any OS user.')
    group_association = PublicWritableAttribute(
        'Method used to associate authenticated users to SFTPPlus groups.')
    manager_allowed_groups = PublicWritableAttribute(
        'List of OS groups of which users are allowed as administrators.')
    base_roles = PublicWritableAttribute(
        'List of SFTPPlus roles associated with any administrators.')
    role_association = PublicWritableAttribute(
        """
        Method used to associate authenticated administrators
        to SFTPPlus roles.
        """)


class IEntraIDAuthenticationConfiguration(
    IAuthenticationConfiguration,
    _IProxyOptions,
    IPasswordOptions,
        ):
    """
    Configuration for Entra ID authentication.

    The shared secret is protected using the password option.
    """
    application_id = PublicWritableAttribute('Entra ID app ID.')
    directory_id = PublicWritableAttribute('Entra ID tenant for this APP.')
    group_association = PublicWritableAttribute(
        'Group to which authenticated accounts are associated.')
    base_groups = PublicWritableAttribute(
        'List of groups to be associated with Entra ID users.')
    base_roles = PublicWritableAttribute(
        'List of roles to be associated with Entra ID administrators.')
    remove_username_suffix = PublicWritableAttribute(
        'List of suffixed removed from users.')
    api_scopes = PublicWritableAttribute(
        'List of extra API scopes request on authentication.')


class IHTTPAuthenticationConfiguration(
        IAuthenticationConfiguration,
        IURLFallbackOptions,
        _IProxyOptions,
        ISSLOptions,
        IUsernamePasswordPublicOptions,
        ):
    """
    Configuration for HTTP authentication.
    """
    headers = PublicWritableAttribute(
        """
        List of headers set for the outgoing request.
        """)
    test_at_start = PublicWritableAttribute(
        'True if the authentication method should be tested at start.')
    timeout = PublicWritableAttribute("""
        Duration, in seconds, to wait for a response from the server.
        """)


class IDenyUsernameAuthenticationConfiguration(IAuthenticationConfiguration):
    """
    Configuration for deny username authentication.
    """

    usernames = PublicWritableAttribute('List of denied usernames.')


class IIPTimeBanAuthenticationConfiguration(IAuthenticationConfiguration):
    """
    Configuration for source IP time based ban authentication.
    """

    ban_interval = PublicWritableAttribute(
        'Number of seconds for which the ban is in effect once activated.')
    ban_after_count = PublicWritableAttribute(
        'Number of failures after which the ban activates.')


class IAnonymousAuthenticationConfiguration(IAuthenticationConfiguration):
    """
    Configurations for the anonymous authentication.
    """

    anonymous_account_uuid = PublicWritableAttribute(
        'UUID of the account associated with this account.')


class ILDAPAuthenticationConfiguration(
        IAuthenticationConfiguration, IAddressPortPublicOptions):
    """
    Configurations for the LDAP authentication.
    """

    secure_connection = PublicWritableAttribute(
        'True when LDAPS should be enabled.')
    bind_dn = PublicWritableAttribute('DN used to bind the LDAP connection.')
    bind_dn_type = PublicWritableAttribute("""
        DN used to bind the LDAP connection.
        """)
    home_folder_attribute = PublicWritableAttribute("""
        LDAP attribute name used to retrieve the home folder path associated
        with the authenticated account.
        """)
    email_attribute = PublicWritableAttribute("""
        LDAP attribute name used to retrieve the email for the authenticated
        account.
        """)
    username_attribute = PublicWritableAttribute("""
        LDAP attribute name used to associate the username requested for
        authentication.
        """)
    username_suffix = PublicWritableAttribute("""
        Value added after the username.
        """)
    search_filter = PublicWritableAttribute("""
        LDAP search filter expression used to filter the entries accepted as
        users for the file transfer services.
        """)
    manager_search_filter = PublicWritableAttribute("""
        LDAP search filter expression used to filter the entries accepted as
        administrators for the Web Manager.
        """)
    group_mapping = PublicWritableAttribute("""
        List of mapping from LDAP entry name and value to internal group UUID.
        """)
    extension_entry_point = PublicWritableAttribute("""
        Path/address to the extension to be used as part of the authentication.
        """)
    extension_configuration = PublicWritableAttribute("""
        JSON value with the configuration for the extension.
        """)


class IRADIUSAuthenticationConfiguration(
    IAuthenticationConfiguration,
    IAddressPortPublicOptions,
    IPasswordOptions,
        ):
    """
    Configurations for the RADIUS authentication.
    """

    continue_authentication = PublicWritableAttribute(
        'Whether to try other authentication methods on failure.')
    group_mapping = PublicWritableAttribute("""
        List of mapping from RADIUS attributes to internal group UUID.
        """)
    timeout = PublicWritableAttribute("""
        Duration, in seconds, to wait for a response from the server.
        """)
    authentication_type = PublicWritableAttribute(
        'The Auth-Type to use for the access request.')
    debug = PublicWritableAttribute('True to enable debugging.')
    nas_port = PublicWritableAttribute(
        'NAS Port over which the request is made.')


class ILocalFileAuthenticationConfiguration(
    IAuthenticationConfiguration, ILocalPathOptions, _IIdentityOptions,
        ):
    """
    Configuration for local file authentication.
    """

    external_management = PublicWritableAttribute("""
        True when the configuration is managed by an external process.
        """)


class IAuthenticationsManager(IRunnablesManager):
    """
    Marker interface for authentication manager.
    """


class IAuthentication(IRunnable):
    """
    Interface used by all authentication methods.

    It is very similar to twisted.cred.checkers.ICredentialsChecker but instead
    of avatar_id, it returns account configuration.
    """
    taggedValue('family', TYPE_NAME.AUTHENTICATION)

    def authenticate(credentials):
        """
        Authenticate requested `credentials`.

        Return a deferred which fires with:
        * Tuple[`IAccountConfiguration`, Sequence[`IGroupConfiguration`]] when
          credentials were accepted.
        * `None` when credentials were not accepted but can continue with
          other methods.

        Deferred will fail with:
        * `CredentialsForbidden` when credentials are denied and should
          not try other methods.
        """

    def authenticateAdministrator(credentials):
        """
        Authenticate requested `credentials` as administrator.

        Return a deferred which fires with:
        * Tuple[
            `IOSAdministratorConfiguration`,
            Sequence[`IRoleConfiguration`]
            ]
          when credentials were accepted.
        * `None` when credentials were not accepted but can continue with
          other methods.

        Deferred might fail with:
        * `CredentialsForbidden` when credentials are denied and should
          not try other methods.
        """


class IOIDCAuthentication(IAuthentication):
    """
    Handle authentication over OpenID Connect.
    """

    request_argument = Attribute(
        'The argument to use for the URL when requesting to trigger OIDC.')

    def getAuthorizationURL(request):
        """
        Return the URL from where users can request the OIDC authentication
        from their provider/server.
        """

    def getCredentials(request):
        """
        Called when we got an authorization code and we can retrieve the
        credentials.
        """


class ILocalFileAuthentication(IAuthentication, _IIdentityOptions):
    """
    The local file authentication runnable.
    """


class IEventHandlersConfiguration(ISetConfiguration):
    """
    Marker interface for the container holding all configured event handlers.
    """


class IEventHandlerConfiguration(
        IEntityConfiguration, ITypeOptions, IAutoFailOptions):
    """
    Shared configurations for all event handle.

    This is also a marker to select configuration type.
    """


class _IFilterOptionsMixin(Interface):
    """
    Shared filter options.
    """

    target = PublicWritableAttribute(
        'Comma separated list of event IDs on which this handler is active.')
    groups = PublicWritableAttribute(
        'Comma separated list of event groups on which this handler is active.'
        )
    usernames = PublicWritableAttribute(
        'Comma separated list of usernames on which this handler will '
        'react.'
        )
    components = PublicWritableAttribute(
        'Comma separated list of component UUIDs on which this handler will '
        'react.'
        )
    data_filter = PublicWritableAttribute(
        'List of comma separated name and matching expression.')
    source_addresses = PublicWritableAttribute(
        'Filter based on source addresses.')


class IFormattedEvent(Interface):
    """
    Shared option for formatting the event message.
    """
    entry_content = PublicWritableAttribute(
        'Format string to use when serializing/representing an event as '
        'human readable text.')
    structured_fields = PublicWritableAttribute(
        'Fields to use when serializing an event to a structured format.')


class _ISourceEventAndDestinationPathOptions(Interface):
    """
    Configuration options for source in the event and destination on the
    local filesystem.
    """
    destination_path = PublicWritableAttribute(
        'Local path where to store the modified copy.')
    source_attribute = PublicWritableAttribute(
        'Name of the data attribute from which the path is extracted.')
    delete_source_on_success = PublicWritableAttribute(
        'Whether to delete the source file on successful operation.')
    overwrite_rule = PublicWritableAttribute(
        """
        Rule use for overwriting existing destination file.
        """)


class ICreateArchiveEventHandlerConfiguration(
    IEventHandlerConfiguration,
    _ISourceEventAndDestinationPathOptions,
    _IFilterOptionsMixin,
        ):
    """
    Configuration for create archive event handler.
    """

    archive_format = PublicWritableAttribute(
        'Type of the archive to be created.')
    archive_name = PublicWritableAttribute(
        'Name of the archive to be created.')


class IExtractArchiveEventHandlerConfiguration(
    IEventHandlerConfiguration,
    IPasswordOptions,
    _ISourceEventAndDestinationPathOptions,
    _IFilterOptionsMixin,
        ):
    """
    Configuration for extract archive event handler.
    """


class IOpenPGPEventHandlerConfiguration(
    IEventHandlerConfiguration,
    IPasswordOptions,
    _ISourceEventAndDestinationPathOptions,
    _IFilterOptionsMixin,
        ):
    """
    Configuration for extract archive event handler.
    """
    encryption_cipher = PublicWritableAttribute(
        'Cipher used for encryption.')
    encryption_extension = PublicWritableAttribute(
        'Extenion used for the encrypted files.')
    encryption_public_keys = PublicWritableAttribute(
        'Public key used for file encryption.')
    decryption_private_keys = PublicWritableAttribute(
        'Private key used for file decryption.')


class IExternalExecutableEventHandlerConfiguration(
    IEventHandlerConfiguration,
    _IFilterOptionsMixin,
        ):
    """
    Configuration for extract archive event handler.
    """
    executable_path = PublicWritableAttribute(
        'Path on local filesystem to the executable.')
    executable_arguments = PublicWritableAttribute(
        'Arguments with which the executable is called.')
    executable_timeout = PublicWritableAttribute(
        'Number of seconds after which to close the process execution.')
    executable_output_format = PublicWritableAttribute(
        'The expected format for the command output.')
    concurrent_processes = PublicWritableAttribute(
        'Number of processes to handle in the same time.')


class IExtensionEventHandlerConfiguration(
        IEventHandlerConfiguration, _IFilterOptionsMixin):
    """
    Configuration for extension event handler.
    """
    entry_point = PublicWritableAttribute('Dotted path to a Python object.')
    configuration = PublicWritableAttribute(
        'Configuration in free text format.')


class _IRetryOptions(Interface):
    """
    A component that will retry with an increasing delay.
    """
    retry_count = PublicWritableAttribute(
        """
        Number of times to retry failed connections.
        """)

    retry_increase = PublicWritableAttribute(
        """
        Number of seconds to add to each wait period before retrying.
        """)


class IHTTPEventHandlerConfiguration(
    IEventHandlerConfiguration,
    _IFilterOptionsMixin,
    IURLFallbackOptions,
    IUsernamePasswordPublicOptions,
    ISSLOptions,
    _IRetryOptions,
        ):
    """
    Configuration for HTTP event handler.
    """
    request_method = PublicWritableAttribute(
        'HTTP method used for the request.')
    http_content_type = PublicWritableAttribute(
        'Format used to send the event.')

    headers = PublicWritableAttribute('List of headers set for all responses.')

    expected_response = PublicWritableAttribute(
        'Matching expression for the expected response content.')

    extra_data = PublicWritableAttribute('Custom data to be sent.')
    timeout = PublicWritableAttribute(
        'Number of seconds to wait for server response.')


class IAccountInteractionEventHandlerConfiguration(
    IEventHandlerConfiguration,
    _IFilterOptionsMixin,
        ):
    """
    Configuration for the account interaction event handler.
    """
    monitored_operations = PublicWritableAttribute(
        """
        List of file operations for which to emit an event.
        """)


class IRabbitMQEventHandlerConfiguration(
    IEventHandlerConfiguration,
    _IFilterOptionsMixin,
    IUsernamePasswordPublicOptions,
    ISSLOptions,
    _IRetryOptions,
    IURLOptions,
        ):
    """
    Configuration for RabbitMQ AMQP event handler.
    """

    exchange = PublicWritableAttribute(
        "Exchange where the events should be published.")
    routing_key = PublicWritableAttribute(
        "Routing key used when publishing the event to the exchange.")
    payload_template = PublicWritableAttribute(
        "Template of the message to be published.")


class ILocalFileEventHandlerConfiguration(
        IEventHandlerConfiguration, _IFilterOptionsMixin, IFormattedEvent):
    """
    Configuration for local-file event handler.
    """

    path = PublicWritableAttribute(
        'Path to file where logs are stored.')

    rotate_external = PublicWritableAttribute(
        'Should be enabled when an external log rotation is enabled. '
        'Yes | No')

    rotate_at_size = PublicWritableAttribute(
        'Trigger rotation when file reaches this size. 0 | Disabled')

    rotate_on = PublicWritableAttribute(
        '4 day-of-month | 02:30 time-of-day | Disabled')

    rotate_count = PublicWritableAttribute(
        'Defines whether to keep all rotated log files, to rotate the log '
        'file in place or to keep the only certain amount of rotated '
        'log files. Disabled | 0 | -1 | 3')


class IWindowsEventLogEventHandlerConfiguration(
        IEventHandlerConfiguration, _IFilterOptionsMixin):
    """
    Configuration for Windows EventLog handler.
    """


class IDatabaseEventHandlerConfiguration(
        IEventHandlerConfiguration, _IFilterOptionsMixin):
    """
    Configuration for database event handler.
    """

    path = PublicWritableAttribute('Path to the SQLite file.')
    auto_delete = PublicWritableAttribute(
        'Number of days after which to delete older events.')


class IEmailSenderEventHandlerConfiguration(
        IEventHandlerConfiguration, _IFilterOptionsMixin):
    """
    Configuration for email sender event handler.
    """

    email_to_recipients = PublicWritableAttribute(
        'email_to_recipients')
    email_cc = PublicWritableAttribute(
        'email_to_recipients')
    email_bcc = PublicWritableAttribute(
        'email_to_recipients')
    email_subject = PublicWritableAttribute('email_subject')
    email_body = PublicWritableAttribute('email_body')
    email_associated_files = PublicWritableAttribute('email_associated_files')


class IFileDigestEventHandlerConfiguration(
        IEventHandlerConfiguration,
        _IFilterOptionsMixin,
        ):
    """
    Configuration for file digest handler.
    """
    source_attribute = PublicWritableAttribute(
        'Name of the data attribute from which the source path is extracted.'
        )
    digest_algorithm = PublicWritableAttribute(
        'Digest algorithm to be used.'
        )
    output_format = PublicWritableAttribute(
        'Format used to represent the computed digest'
        )
    # Here to help with shared code.
    delete_source_on_success = Attribute('Always false')
    destination_path = Attribute('Alwasy empty')


class IFileDispatcherEventHandlerConfiguration(
        IEventHandlerConfiguration, _IFilterOptionsMixin):
    """
    Configuration for file dispatcher event handler.
    """

    dispatch_attribute = PublicWritableAttribute(
        'Name of the data attribute from which the path is extracted.')
    dispatch_delay = PublicWritableAttribute(
        'Number of seconds to delay the processing of the event.')
    destination_temporary_suffix = PublicWritableAttribute(
        'Use a temporary suffix when copying files to destination.')
    overwrite_rule = PublicWritableAttribute(
        "Rule use for overwriting existing destination file.")
    dispatch_rules = PublicWritableAttribute('Rules for dispatching.')
    fallback_rule = PublicWritableAttribute('Rule for dispatching fallback.')
    create_destination_folder = PublicWritableAttribute(
        'The way in which missing destination folders are created.')
    matching_expressions = PublicWritableAttribute(
        'List of regular expression rules used to generate the destination.')
    retry_count = PublicWritableAttribute(
        """
        Number of times to retry failed files.
        """)
    retry_wait = PublicWritableAttribute(
        """
        Number of seconds to wait before a retry.
        """)
    executable_timeout = PublicWritableAttribute(
        'Timestamp for the external executable.')


class IStandardStreamEventHandlerConfiguration(
        IEventHandlerConfiguration,
        _IFilterOptionsMixin,
        ):
    """
    Configuration for Standard Stream handler.
    """
    entry_content = PublicWritableAttribute(
        'Format string to use when serializing/representing an event as '
        'human readable text.',
        )


class ISyslogEventHandlerConfiguration(
        IEventHandlerConfiguration,
        _IFilterOptionsMixin,
        IURLOptions,
        ):
    """
    Configuration for Syslog event handler.
    """


class IEventHandlersManager(IRunnablesManager):
    """
    Marker interface for event handlers manager.
    """


class IEventHandler(IRunnable):
    """
    Interface used by all event handlers.
    """
    taggedValue('family', TYPE_NAME.EVENT_HANDLER)

    def handle(event):
        """
        Called when an event was emitted and needs to be handled.
        """


class ILogHandlerConfiguration(
        IEntityConfiguration, IAutoFailOptions, ITypeOptions):
    """
    Generic configuration for a log handler.
    """


class IPublicAvatar(IPublicObject):
    """
    A public (exportable) avatar.
    """
    name = PublicAttribute('Name of this avatar.')
    uuid = PublicAttribute('UUID of this avatar.')
    peer = PublicAttribute('The remote peer associated to this avatar.')
    component = Attribute('The component which has created this avatar.')

    # For now, the email is only public for server-side accounts, and not
    # for administrators or peer or job avatars.
    # Note that the low-level configuration is just a string value,
    # for the `email` attribute.
    email = PublicAttribute('Primary email of this avatar.')
    emails = PublicAttribute(
        'All email of this avatar. '
        'As a list of (email_name, email_address) tuples.')


class IExtensionAvatar(IPublicAvatar):
    """
    Avatar available to the API extensions.
    """
    token = PublicAttribute(
        'OS or OAuth2 token obtained during the authentication.')


class IServiceAvatar(IPublicAvatar, IFileSystemAvatar):
    """
    An avatar used by a service.

    Here 'service' only refers to user's services and excluded administration
    and other internal services.
    """

    created = Attribute('Date and time when the account was created.')

    home_folder_path = Attribute('Path to home folder for this account.')
    root_folder_path = Attribute(
        'Path to folder which is the root of the avatar\'s filesystem.')

    lock_in_home_folder = Attribute(
        'True if account should be locked in the home folder.')

    create_home_folder = Attribute(
        'True if missing home folder should be created for this account.')
    create_home_folder_owner = Attribute(
        'Name of owner for the newly created home folder.')
    create_home_folder_group = Attribute(
        'Name of group for the newly created home folder.')

    required_folders = Attribute(
        'List of path with directories which should exit for this account.'
        )

    virtual_folders = Attribute(
        'List of tuples with mapping from virtual segments to real paths.')

    permissions = Attribute('List of permissions for this accounts.')
    file_write_size = Attribute(
        'Maximum size in bytes for a single uploaded file.')

    def getAmendedUploadSegments(segments):
        """
        Return the segments as they should be used during the upload process
        and a boolean to signal if the segments were amended.

        Return a copy of the input segments when nothing should be changed.

        Return (segments, was_amended)
        """

    can_change_password = Attribute('If it can change its password.')

    def updatePassword(current, new, event_emitter):
        """
        Update avatar's password.

        Return `True` if password was successfully changed.
        """

    def isMember(group_uuid):
        """
        Return `True` is this is a member of group with uuid.
        """

    def getAS2Certificates():
        """
        Return the list of certificates used to validate the signature for
        AS2 message received for this account.
        """

    def getAS2MDNCA():
        """
        Return the path to CA used for async MDN response.

        Return empty string if remote peer should not be validated.

        Raise an error if async MDN is disabled.
        """


class ICredentials(IPublicObject, IObserver):
    """
    Hold credentials as provides by clients while authentication using
    various services.

    Provides attributes shared by all credential types.
    """
    username = PublicAttribute(
        '''
        Username for which the authentication is requested.
        ''')

    peer = PublicAttribute(
        '''
        IP address and port number for the remote peer requesting
        authentication.
        ''')

    type = PublicAttribute(
        """
        Human readable name for the type of these credentials.

        Should be a single word. Example:
        * password
        * ssh-key
        * ssl-certificate
        """)

    kind_name = Attribute(
        """
        Human readable name for the details of these credentials.

        For example, SSH fingerprint or certificate serial.
        """)

    content = PublicAttribute("String serialization of credentials content.")

    creator = PublicAttribute(
        "IEventCreator which created this instance.")


class IPasswordCredentials(ICredentials):
    """
    Credentials based on password.
    """
    # Password should not be public.
    password = Attribute(
        '''
        Password associated with the username.
        ''')


class ISSHKeyCredentials(ICredentials):
    """
    Credentials based on SSH public key.

    This is received by the server from a client.
    """

    key_algorithm = Attribute('Algorithm for SSH key.')
    key_data = Attribute('Content of SSH key.')
    key_signed_data = Attribute('Signed data for SSH key.')
    key_signature = Attribute('Signature for SSH key.')


class ISSHPrivateKeyCredentials(ICredentials):
    """
    Credentials based on SSH private key.

    This is used by the client to generate a public key credentials to be
    sent to the server.
    """

    def getPrivateKey():
        """
        Private part of the chevah_keycert.ssh.Key associated with the
        credentials.
        """

    def getPublicKey():
        """
        Public part of the chevah_keycert.ssh.Key associated with the
        credentials.
        """


class ISSLCertificateCredentials(ICredentials):
    """
    Credentials base on SSL certificate.
    """

    certificate = Attribute(
        '''
        pyOpenSSL certificate object.
        ''')


class IHTTPPasswordCredentials(IPasswordCredentials):
    """
    Password based credentials generated by HTTP service.
    """

    request = Attribute(
        "The IRequest from which the authentication was requested.")


class IHTTPSPasswordCredentials(IHTTPPasswordCredentials):
    """
    Password based credentials generated by HTTPS service.
    """


class IAS2AnonymousCredentials(ICredentials):
    """
    Credentials from AS2 request without HTTP authentication.
    """


class IFTPPasswordCredentials(IPasswordCredentials):
    '''Marker interface for a password based credentials obtained via FTP.'''


class IFTPSPasswordCredentials(IPasswordCredentials):
    '''Marker interface for a password based credentials obtained via FTPS.'''


class ISSHPasswordCredentials(IPasswordCredentials):
    '''Marker interface for a password based credentials obtained via SSH.'''


class IFTPSSSLCertificateCredentials(ISSLCertificateCredentials):
    '''Interface for SSL certificate based credentials used for FTPS.'''


class IHTTPSSSLCertificateCredentials(ISSLCertificateCredentials):
    '''Interface for SSL certificate based credentials used for HTTPS.'''


class ILocalManagerCredentials(IPasswordCredentials):
    """
    Credentials used for administration.
    """
    password = Attribute('Password used at authentication.')


class ILocalManagerAvatar(IPublicAvatar):
    """
    An avatar used by local manager.
    """

    roles_uuid = PublicAttribute('List with UUID of the associated roles.')


class IApplicationLocalManagerAvatar(ILocalManagerAvatar):
    """
    Marker interface for local manager authenticated based on application
    administrator.
    """


class IOSLocalManagerAvatar(ILocalManagerAvatar):
    """
    Marker interface for local manager authenticated based on OS
    administrator.
    """


class IConfigurationAvatar(Interface):
    """
    Configuration data for an account.
    """

    name = Attribute('Name of the account.')
    description = Attribute('Description of the account.')
    create_home_folder = Attribute(
        'True if server should try to create missing home folders.')
    home_folder_path = Attribute('Path to user files.')


class IServerShell(Interface):
    """
    A shell for accessing the files on the server-side on behalf of
    an account.

    It is initialized with an avatar and will receive the path arguments
    as visible by the client-side.
    """

    def getRealPath(path):
        """
        Return local filesystem path associated with `path`.

        `path` is received as UTF-8.

        It tries its best to get a path, but never raises an error. As a last
        resort it return `path` as Unicode.
        """

    def getPath(segments):
        """
        Return the path as visible to client session.
        """

    def getAttributes(path):
        """
        Retrieve the attributes for path.

        This is like list, except it return information for the path itself,
        not for children.
        """

    def makeDirectory(path):
        """
        Create a folder
        """

    def removeDirectory(path):
        """
        Remove a folder.
        """

    def removeFile(path):
        """
        Remove a file.
        """

    def rename(from_path, to_path):
        """
        Rename a file or directory.
        """

    def openForReading(path):
        """
        Return the file descriptor opened for read only mode.
        """

    def openForWriting(path):
        """
        Return the IServerShellfile descriptor opened for write only mode.
        """


class IHTTPShell(IServerShell):
    """
    Filesystem operation for the HTTP server-side protocol.
    """

    def exists(path):
        """
        Return true if path exists.

        It does not emit any event.

        On any errors it returns False.
        """

    def makeDirectories(base, relative):
        """
        Create all folders for `relative` path, starting with `base`.
        """

    def openDirectory(segments):
        """
        Return an iterator with attributes of all members.
        """


class IFTPShell(IServerShell):
    """
    A shell for the FTP server-side session.

    Inspired by Twisted IFTPShell.

    All methods should return a deferred.
    """

    working_segments = Attribute('Segments for current working directory.')

    def setWorkingPath(path):
        """
        Set current working path.

        It supports absolute paths or relative to current working path.
        """

    def getWorkingPath():
        """
        Return public path for current working folder.
        """

    def getAbsolutePath(path):
        """
        Return absolute path associated with `path`.

        `path` is received as UTF-8.

        It tries its best to get a path, but never raises an error. As a last
        resort it return `path` as Unicode.
        """

    def openForAppending(path):
        """
        Return the file descriptor opened for writing in
        appending and create mode.
        """

    def setAttributes(path, attributes):
        """
        Set attributes for path.

        `attributes` is a dict.
        """

    def getMembersName(path, glob=None):
        """
        Retrieve the folder names at bytes `path`.

        It returns an iterator with batches of names in a list.

        The returned members are filtered based on the globbing bytes `glob`.
        """

    def list(path, glob=None):
        """
        Retrieve a list of members with their attributes for the given
        bytes `path`.

        The returned members are filtered based on the globbing bytes `glob`.

        If the path represents a non-directory, the result list should
        have only one entry with information about that non-directory.
        Otherwise, the result list should have an element for each
        child of the directory.

        It returns a tuple of (own_attributes, batch_iterator)
        """


class IRESTFolder(Interface):
    """
    For now this is a marker interface for a REST folder resource.
    """


class IChevahSession(Interface):
    """
    Persistent storage for every session.
    """

    resource = Attribute(
        """
        Resource which is already authenticated for the current session.

        None when the session was not authenticated.
        """)


class IUploadStream(Interface):
    """
    A stream object which receives uploaded data.
    """

    closed = Attribute('True when stream is closed.')
    name = Attribute('Name of the stream.')

    def write(data):
        """
        Write received `data`.
        """

    def close():
        """
        Close the stream.
        """

    def interrupt(error):
        """
        Close the stream before receiving all data.
        """


class IChevahTimeStamp(IPublicObject):
    """
    An timestamp object.
    """
    timestamp = PublicAttribute('Local unix timestamp with milliseconds.')
    cwa_14051 = PublicAttribute(
        'Local CEN CWA 14051 format string representation of timestamp.')
    iso_8601 = PublicAttribute(
        'UTC ISO 8601 format string representation of timestamp.')
    iso_8601_fractional = PublicAttribute(
        'UTC ISO 8601 format string representation of timestamp with '
        'fractional seconds.')
    iso_8601_local = PublicAttribute(
        'Local ISO 8601 format string representation of timestamp.')
    iso_8601_basic = PublicAttribute(
        'UTC ISO 8601 basic format string representation.')
    iso_8601_compact = PublicAttribute(
        'Local ISO 8601 basic format string representation without timezone.')


class IEvent(IPublicObject):
    """
    An event as raised by a service.
    """

    id = PublicAttribute('ID/Type of this event.')
    uuid = PublicAttribute('Unique ID of this event.')
    account = PublicAttribute(
        'Local or remote account associated with this event.')
    data = PublicAttribute('Dictionary with data for this event.')

    timestamp = PublicAttribute(
        'ChevahTimeStamp with the time this'
        'event was created')
    message = PublicAttribute('The event message.')

    server = PublicAttribute('The server emitting this event.')
    # FIXME:3231:
    # We should have only `component` as now it is a proxy to `creator`
    # The same for `created` as it holds the same information as `timestamp`
    component = PublicAttribute('A public proxy to the event creator.')
    creator = Attribute('IEventCreator which created this event.')
    created = PublicAttribute(
        'Unix timestamp with milliseconds with time at which this event was'
        'generated.')
    data_json = PublicAttribute('JSON string serialization of event data.')


class IEventDefinition(Interface):
    """Configuration definition for an Event."""

    id = Attribute('Name of the group.')
    message = Attribute('Text logged for this event.')
    description = Attribute('Detailed description')
    version_added = Attribute('Version when this event was added.')
    version_removed = Attribute('Version when this event was removed.')


class IEventGroupDefinition(IPublicObject):
    """Configuration definition for an EventGroup."""

    name = PublicAttribute('Name of the group.')
    description = PublicAttribute('Group description')


class IEventsDefinition(Interface):
    """Stores the definitions for events and event groups.

    Definitions are the static rules for all events.
    """

    def load():
        """Load configured events, event groups and rules.

        Raises an error if the configuration is not valid.
        """

    def eventHasGroup(id, name):
        """
        Returns True if the event `id` has group `name`, False otherwise.
        """

    def getEventDefinition(id):
        """
        Return the EventDefinition for `id`.

        Raises an error if EventDefinition was not found.
        """

    def getAllEventDefinitions():
        """
        Return a dictionary with all EventDefinition.

        The dictionary is keyed based on event id.
        """

    def getAllEventDefinitionsPadded():
        """
        Return a dictionary with all EventDefinition.

        The dictionary is keyed based on padded event id.
        """

    def getEventGroupDefinition(name):
        """
        Return the EventGroupDefinition for `name`.

        Raises an exception if EventGroupDefinition was not found.
        """

    def getAllEventGroupDefinitions():
        """
        Return a dictionary with all EventGroupDefinition.

        The dictionary is keyed based on group name.
        """

    def generateDocumentation(template):
        """
        Return a string with events' documentation.

        Documentation is formatted as ReStructuredText using Jinja2
        `template`.
        """


class IIdleTimeoutProtocol(Interface):
    """
    A protocol that times out if idle.
    """

    timeOut = Attribute(
        'Number of seconds ofter which the idle connection is disconnected')

    def callLater(period, func):
        """
        Wrapper around L{reactor.callLater} for test purpose.
        """

    def resetTimeout():
        """
        Reset the timeout count down.

        If the connection has already timed out, then do nothing.  If the
        timeout has been cancelled (probably using C{setTimeout(None)}), also
        do nothing.

        It's often a good idea to call this when the protocol has received
        some meaningful input from the other end of the connection.  "I've got
        some data, they're still there, reset the timeout".
        """

    def setTimeout(period):
        """
        Change the timeout period

        @type period: C{int} or C{NoneType}
        @param period: The period, in seconds, to change the timeout to, or
        C{None} to disable the timeout.
        """

    def timeoutConnection():
        """
        Called when the connection times out.

        Override to define behavior other than dropping the connection.
        """


class ISFTPServer(TwistedISFTPServer):
    """
    A SFTP sub-system which is also reused by SCP protocol.

    See: `twisted.conch.interfaces.ISFTPServer`.
    """

    def start(service_name):
        """
        Called when the shell can start operating.
        """


class IChevahRequest(IRequest):
    """
    Extended request used in Chevah project.
    """
    # FIXME:6502:
    # https://github.com/twisted/twisted/issues/10191
    # We need to see how to send it upstream.
    maximumBodyLength = Attribute(
        """
        Set the maximum accepted body length.

        Clients sending more data are disconnected.

        Setting value `0` or None will allow the request to receive an
        unlimited body.
        """)

    # FIXME:6502:
    # https://github.com/twisted/twisted/issues/10190
    # We need to see how to send it upstream.
    def headersReceived(command, path, version):
        """
        Called by channel when all headers has been received.

        When a non-continue (non 100) response is returned, this method should
        set all the required headers for the response.
        """


class IChevahResource(IResource):
    """
    Extended resource used by Chevah project.
    """

    def headersReceived(request):
        """
        Called by request when all headers has been received.

        Should return _CONTINUE if the request should continue receiving
        the body.

        Any other response code is returned to client and connection is
        closed.

        Response can be a simple HTTP_CODE or (HTTP_CODE, CUSTOM_MESSAGE)
        """


class IResourceRunnable(IRunnable):
    """
    A resource which acts as a client to provide the end-user functionality.
    """

    avatar = Attribute('The client/credentials for this resource.')

    def heartbeat():
        """
        Called to signal that a command was requested for the location.

        Return a deferred which is fired when locations is connected.
        """

    def callWhenConnected(call, *args, **kwargs):
        """
        Call when runnable is connected and ready to accept commands.

        If it is not connected, it will try to connect.

        Return a deferred which is fired with the result of the call.
        """


class IWatchHandler(Interface):
    """
    A handler watching the source directory for a transfer.
    """
    running = Attribute('True if watch snapshots are still scheduled.')

    def start(self):
        """
        Start watching the path.
        """

    def stop(self):
        """
        Stop watching the path.

        This can be called at any time, event before start.
        It should not raise an error if called while not running.

        Returns the base path that was unwatched.
        The help with the audit.
        """


class ILocation(IResourceRunnable):
    """
    Local or remote location used by an ITransfer.

    All method support only absolute paths.
    """
    taggedValue('family', TYPE_NAME.LOCATION)

    def watch(configuration, handler, on_error, on_snapshot):
        """
        Watch path defined by `configuration` and use `handler` to
        handle changes.

        When watch fails, it will call `on_error` with the exception as
        argument.

        `on_snapshot` is called for each time a new snapshot is made by the
        watch.

        Return a `IWatchHandler` object representing the watched path and
        which can be used to unwatch the path.
        """

    def unwatch(watched_handler):
        """
        Stop watching `path` for changes.
        """

    def joinPath(*parts):
        """
        Join one or more path components, resolving any relative parts and
        converting separators.

        This also handles URL paths.
        """

    def splitPath(path):
        """
        Split the path into a (dirname, basename).
        The basename part will never contain a slash; if path ends in a slash,
        basename will be empty.
        """

    def relativePath(base, path):
        """
        Return a relative file path to `path` from the `base` directory.
        """

    def exists(path):
        """
        `True` when path exists, `False` otherwise.

        Return a deferred that fires with the result.
        """

    def rename(source_path, destination_path):
        """
        Rename file or folder inside the location.

        Return a deferred that fires when operation is done.
        """

    def deleteFile(path):
        """
        Delete file at path.

        Return a deferred that fires when operation is done.
        """

    def deleteFolder(path, recursive=False):
        """
        Delete folder at path.

        Return a deferred that fires when operation is done.
        """

    def createFolder(path):
        """
        Create a folder at path, without creating parents.

        Return a deferred that fires when operation is done.
        """

    def getFolderMemberAttributes(path):
        """
        Return a deferred that fires with a dictionary of
        `name` -> `IFileAttributes`
        for files contained in the given path.
        """

    def getAttributes(path):
        """
        Return a deferred that fires with the `IFileAttributes` for `path`.
        """

    def getStreamForReading(path):
        """
        Return an deferred with an `ITransferStream` opened for
        reading at the start of the stream.
        """

    def getStreamForWriting(path, length=None):
        """
        Return a deferred with an `ITransferStream` opened for writing,

        `length` is the expected data to be written and some location
        implementation might ignore it.

        Creates the file if it does not exist, otherwise removes existing
        content.
        """

    def getStreamForAppending(path):
        """
        Return a deferred with an `ITransferStream` opened for appending.

        Creates the file if it does not exist.
        """


class ILocalLocation(ILocation):
    """
    A location on local filesystem.
    """

    def touch(path):
        """
        Create a new file at path or update its modified date.

        Return a deferred that fires when operation is done.
        """

    def copyFile(source_path, destination_path, overwrite=False):
        """
        Return a deferred which fires when file was copied from `source_path`
        to destination path.

        If `destination_path` is a folder, the file will be copied at
        `destination_path/SOURCE_FILENAME`.

        If `destination_path` already exists and `overwrite` is not `true`,
        copy will fail.
        """


class ILocationShell(Interface):
    """
    Low level client shell used by a location to perform file operations.

    The client is already TCP connected as location takes care of networking.

    The shell should do protocol authentication or channel setup.
    """

    avatar = Attribute('Client avatar for this location.')

    def authenticate():
        """
        Return a deferred which fires when the client is authenticated.
        """

    def callOnConnectionLost(callback):
        """
        Call `callback` when connection is lost.
        """


class ITransferStream(Interface):
    """
    A file/stream which is used to read/write data.

    All the methods return a deferred.

    No other method can be called after it is closed.
    """

    location = Attribute('ILocation associated with this stream.')
    path = Attribute('Path in the ILocation associated with this stream.')
    mode = Attribute('Description of the mode in which stream was opened.')

    def close():
        """"
        Close the stream.
        """

    def read(length):
        """
        Return data of maximum `length`.
        """

    def write(data):
        """
        Write data.
        """


class ILocationsManager(IRunnablesManager):
    """
    Marker interface for transfers manager.
    """


class ITransfersManager(IRunnablesManager):
    """
    Marker interface for transfers manager.
    """


class ITransfer(IRunnable):
    """
    Marker interface for a transfer.
    """
    taggedValue('family', TYPE_NAME.TRANSFER)


class ITransferTransformer(Interface):
    """
    Converts streaming data between various formats.

    It is designed to be as generic as possible and allow handling things
    like text conversion, PGP encrypt/decrypt, or TAR.GZIP compress/decompress
    """

    name = Attribute('Short description for this type of transformation.')

    def reset():
        """
        Reinitialize the state, in order to start a new transform.
        """

    def consume(data):
        """
        Receives a chunk of bytes data and returns the transformed bytes chunk.

        It might return empty string if more data is needed to create a new
        transformed chunk.

        The returned chunk might be of a different size.
        """

    def finalize():
        """
        Return any remaining transformed data.

        It can return empty bytes.

        It can raise an error if the remaining data can't be fully
        transformed.
        """


class ITransferJob(IRunnable):
    """
    Interface for a transfer job.
    """
    taggedValue('family', TYPE_NAME.TRANSFER_JOB)

    transformer = Attribute(
        """
        ITransferTransformer

        Handles on-the fly data conversion for the transfer.
        """)


class IResourcesConfiguration(ISetConfiguration):
    """
    Marker interface for the configuration of all the resources.
    """


class IResourceConfiguration(IEntityConfiguration, ITypeOptions):
    """
    Configurations for all resources.

    This is also a marker to select the configuration type.
    """


class IEmailClientConfiguration(
    IResourceConfiguration,
    ISSLOptions,
    IUsernamePasswordPublicOptions,
        ):
    """
    Configuration for email client.
    """
    # Here enable -> launch at startup
    # email
    enabled = PublicAttribute('Emails are auto-started when needed.')
    email_from_address = PublicWritableAttribute(
        'Address used in the From field.')
    email_to_recipients = PublicWritableAttribute(
        'Address or addresses used in the To field by default.')


class ILetsEncryptClientConfiguration(
        IResourceConfiguration, IAddressPortPublicOptions):
    """
    Section for Let's Encrypt configuration.
    """

    acme_url = PublicWritableAttribute('URL to ACME server directory.')
    contact_email = PublicWritableAttribute('Contact email addresses.')
    redirect_url = PublicWritableAttribute(
        'URL where to redirect other requests.')
    debug = PublicWritableAttribute('Whether to debug the HTTP server.')
    use_ssl = Attribute('From service configuration.')
    _account_key = Attribute('Private key as PEM used for the certificate.')
    _storage = Attribute(
        'Certificates as concatenated PEM managed by this service')


class IEmailClientResource(IRunnable):
    """
    A resource which sends email as a client.
    """
    taggedValue('family', TYPE_NAME.RESOURCE)

    def send(recipients, subject, body, attachments=None):
        """
        Send email with `subject`, `body` and (optionally) `attachments`
        to `recipients`. `attachments` is a dictionary each key of
        which contains full path to a local file and each value
        contains file-like object (e.g. `io.BytesIO`). This file-like
        object contains the contents of the file.

        Return a deferred which is fired when all requested recipients have
        received the message.
        """

    # FIXME:5893:
    # This should be moved into the runnable interface.
    # With an implementation that by default does not fail, but returns
    # a message that testing is not yet supported.
    def check(configuration):
        """
        Check that the `configuration` can be used to start and use this
        runnable.,

        `configuration` is a dict raw format, similar to what we have for
        the JSON-RPC input.

        Return a text with the description of result of the check.
        The text is intended to be displaye to the person that has
        triggered the test.

        Raises an ServerException when failing to validate the configuration.
        """


class ISQLiteConfiguration(
        IResourceConfiguration, ILocalPathOptions):
    """
    Configuration for the embedded SQLite resource.
    """
    # Here enable -> launch at startup
    enabled = PublicAttribute('Embedded db is auto-started when needed.')


class ISQLiteResource(IRunnable):
    """
    A resource which provides access to SQLite DB.
    """
    taggedValue('family', TYPE_NAME.RESOURCE)


class IProcessMonitorConfiguration(IResourceConfiguration):
    """
    Deprecated section for resource monitor configuration.

    Still here in order to not crash v3 to v4 migration and recognize the
    component.

    It contains all the v3 values to allow simple migration
    using high-level code.

    Such a configuration should be automatically removed after start.
    """
    monitor_interval = PublicWritableAttribute(
        'Seconds at which to run the monitor.')
    memory_resident_trigger = PublicWritableAttribute(
        'Trigger for resident memory.')
    memory_virtual_trigger = PublicWritableAttribute(
        'Trigger for virtual memory.')
    connection_count_trigger = PublicWritableAttribute(
        'Trigger for number of connections.')
    file_count_trigger = PublicWritableAttribute(
        'Trigger for number of files.')
    thread_count_trigger = PublicWritableAttribute(
        'Trigger for number of threads.')


class IProcessMonitorResource(IRunnable):
    """
    Deprecated resource which monitors the self OS process.

    It was replaced by the analytics engine.

    Still here in oder to not crash v3 to v4 migration.
    """
    taggedValue('family', TYPE_NAME.RESOURCE)


class IAnalyticsConfiguration(IResourceConfiguration):
    """
    Configuration section containing settings for analytics resources.
    """

    monitor_interval = PublicWritableAttribute(
        'Seconds at which to run the monitor.')
    memory_resident_trigger = PublicWritableAttribute(
        'Trigger for resident memory.')
    memory_virtual_trigger = PublicWritableAttribute(
        'Trigger for virtual memory.')
    connection_count_trigger = PublicWritableAttribute(
        'Trigger for number of connections.')
    file_count_trigger = PublicWritableAttribute(
        'Trigger for number of files.')
    thread_count_trigger = PublicWritableAttribute(
        'Trigger for number of threads.')


class IAnalyticsResource(IRunnable):
    """
    Resource which monitors and records own activity
    """
    taggedValue('family', TYPE_NAME.RESOURCE)


class ILetsEncryptClientResource(IRunnable):
    """
    A resource which handles Let's Encrypt certificates.
    """
    taggedValue('family', TYPE_NAME.RESOURCE)

    def getCertificatePEMs(name, sans=None):
        """
        Return a PEM serialized of all the certificate objects for `name`.

        This include cert itself, key and cert chain.

        `name` is the CN of the certificate.
        `sans` is a list of subjectAlternativeNames used when generating
        the certificate.

        This will always return a certificate.
        If no certificate is yet found, it will return a self-signed one.
        """


class IMonitorResource(IRunnable):
    """
    Service of type monitor.
    """
    taggedValue('family', TYPE_NAME.RESOURCE)


class ILocationsConfiguration(ISetConfiguration):
    """
    Marker interface for the container holding all configured locations.
    """


class ILocationConfiguration(IResourceConfiguration):
    """
    Shared configurations for a all locations.

    This is also a marker to select configuration type.
    """

    enabled = PublicAttribute(
        """
        Locations don't have the launch at startup config.
        Here to comply with the shared code.
        """
        )

    idle_connection_timeout = PublicWritableAttribute(
        'Number of seconds after which an idle location is disconnected.')

    idle_connection_keepalive_interval = PublicWritableAttribute(
        """
        Number of seconds of inactivity after which to send a keep alive
        command.
        """)
    connection_retry_count = PublicWritableAttribute(
        'Number of times to try to connect.')
    connection_retry_interval = PublicWritableAttribute(
        """
        Number of seconds after which to re-try a previously failed
        connection.
        """)


class ILocationWithAddressConfiguration(
        ILocationConfiguration, IAddressPortPublicOptions):
    """
    Location that is configured via an address.

    All configurations start with a public address:port that is used for
    connecting to the remote server and event handling.
    """


class ILocationWithURLConfiguration(ILocationConfiguration, IURLOptions):
    """
    Location that is configured via an URL.

    Locations that have URLs also export a private read-only address:port.
    """


class IFilesystemLocationConfiguration(ILocationWithAddressConfiguration):
    """
    Configurations for a local filesystem location.
    """

    username = PublicAttribute('Username used for this location.')


class ISFTPLocationConfiguration(
        ILocationWithAddressConfiguration,
        IUsernamePasswordPublicOptions,
        _ISSHEncryptionOptions,
        _IProxyOptions,
        ):
    """
    Configurations for an SFTP location.
    """
    ssh_private_key = PublicWritableAttribute(
        'Path to private SSH key used to authenticate the location.')

    ssh_server_identity = PublicWritableAttribute(
        'Remote server identity. Fingerprint or full public SSH key.')


class IFTPLocationConfiguration(
        ILocationWithAddressConfiguration,
        IUsernamePasswordPublicOptions,
        ):
    """
    Configurations for an FTP location.
    """
    debug = PublicWritableAttribute(
        'True when location should emit debug events.')
    ignore_passive_address = PublicWritableAttribute(
        'True when the PASV response IP should be ignored.')
    acct = PublicWritableAttribute(
        'Value sent via ACCT if requestes by the server.')


class _IFTPSLocationOptions(Interface):
    """
    Shared configuration options for Explicit and implicit FTPS.
    """
    ftps_reuse_session = PublicWritableAttribute(
        """
        Whether data connection should use the same session
        as the command connection.
        """)


class IFTPSILocationConfiguration(
        IFTPLocationConfiguration, _IFTPSLocationOptions, ISSLOptions):
    """
    Configurations for an Implicit FTPS location.
    """


class IFTPSELocationConfiguration(
        IFTPLocationConfiguration, _IFTPSLocationOptions, ISSLOptions):
    """
    Configurations for an explicit FTPS location.
    """


class _IHTTPServerLocationConfiguration(
    ILocationWithURLConfiguration,
    IUsernamePasswordPublicOptions,
    ISSLOptions,
        ):
    """
    Shared configuration options for a location interacting with an HTTP server.
    """


class IHTTPPullLocationConfiguration(_IHTTPServerLocationConfiguration):
    """
    Configurations for pulling files from a remote HTTP(S) server.
    """
    request_method = PublicWritableAttribute(
        'HTTP method used for the request.')
    request_body = PublicWritableAttribute(
        'HTTP content of the request.')
    request_headers = PublicWritableAttribute(
        'HTTP headers used for the request.')
    response_code = PublicWritableAttribute(
        'Matching expression for the expected success HTTP codes.')
    response_file_name = PublicWritableAttribute(
        'Name of the file used to store the response body.')


class IWebDAVSLocationConfiguration(
    _IHTTPServerLocationConfiguration,
    _IProxyOptions,
        ):
    """
    Configurations for a WebDAV location over HTTPS.
    """

    authentication_method = PublicWritableAttribute(
        'Authentication method used by this location.')
    allow_insecure_http = PublicWritableAttribute(
        'Whether to use HTTP (insecure plaintext) rather than HTTPS.')


class IAS2LocationConfiguration(_IHTTPServerLocationConfiguration):
    """
    Configurations for a AS2 location over HTTP or HTTPS.
    """
    as2_own_identifier = PublicWritableAttribute(
        'Identifier for our own organization.')
    as2_own_certificate = PublicWritableAttribute(
        'Certificate used to sign send files.')
    as2_own_private_key = PublicWritableAttribute(
        'Private key used to sign send files.')

    as2_partner_identifier = PublicWritableAttribute(
        'Identifier for the remote partner.')
    as2_partner_certificates = PublicWritableAttribute(
        'Certificates in PEM format of the remote partner.')

    as2_send_security = PublicWritableAttribute(
        'Defined the way files are sent.')
    as2_mdn_receipt = PublicWritableAttribute(
        'Defined the way MDN receipt is requested.')
    as2_encryption_algorithm = PublicWritableAttribute(
        'Algorithm used to encrypt the files.')
    as2_signature_algorithm = PublicWritableAttribute(
        'Algorithm used to create the digest signature.')

    as2_content_type = PublicWritableAttribute(
        'Content type header for the sent files.')


class _IAzureStorageLocationConfiguartion(
        ILocationWithAddressConfiguration,
        IUsernamePasswordPublicOptions,
        ISSLOptions,
        ):
    """
    Shared location configurations for a File and BLOB service
    of an Azure Storage account. Over HTTPS.
    """
    # Azure has read-only address, since the URL if fixed based on the username.
    address = Attribute('IP address.')
    port = Attribute('Port number.')


class IAzureBLOBLocationConfiguration(_IAzureStorageLocationConfiguartion):
    """
    Configurations for a Azure BLOB Service location over HTTPS.
    """


class IAzureFileLocationConfiguration(_IAzureStorageLocationConfiguartion):
    """
    Configurations for a Azure File Service location over HTTPS.
    """


class ISMBLocationConfiguration(
        ILocationWithAddressConfiguration,
        IUsernamePasswordPublicOptions,
        ):
    """
    Configurations for an SMB location.
    """
    require_encryption = PublicWritableAttribute(
        """
        If it should reject servers that don't support encryption.
        """
        )
