# Copyright (c) 2014 Adi Roiban.
# See LICENSE for details.
"""
Support for templating system.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import os
import six

from bunch import Bunch
from jinja2 import Environment, PackageLoader, StrictUndefined

from chevah.server.commons.constant import (
    HTTP_STATIC_FOLDER_NAME,
    STATIC_FOLDER_NAME,
    )


class ChevahJinjaEnvironment(Environment):
    """
    A Jinja environment based on package loader.

    This can can be used for RST or HTML or anything else.
    """
    _package = 'chevah.server.%s' % (STATIC_FOLDER_NAME)

    def __init__(self, source, autoescape=True):
        """
        It is initialized with templates from `folder` located inside
        static folder.
        """
        folder = os.path.join(*source)
        loader = PackageLoader(self._package, folder)
        super(ChevahJinjaEnvironment, self).__init__(
            loader=loader,
            undefined=StrictUndefined,
            autoescape=autoescape,
            auto_reload=True,
            extensions=['jinja2.ext.i18n'],
            )


class JinjaEnvironment(ChevahJinjaEnvironment):
    """
    A Jinja environment based on package loader used for HTML
    rendering.
    """
    _package = 'chevah.server.%s' % (STATIC_FOLDER_NAME)

    def __init__(self, *args):
        """
        It is initialized with templates from `folder` located inside
        static folder.
        """
        super(JinjaEnvironment, self).__init__(source=args)


#: Global default Jinja environment with templates from HTTP service.
default_environment = JinjaEnvironment(HTTP_STATIC_FOLDER_NAME, 'jinja')


class ViewModel(object):
    """
    A container for view data.

    It behaves like a read-only dict and is passed as Jinja template context
    data.

    It should contains as much view logic as possible and delegate as few
    logic as possible to the templating system.

    It will transform internal raw data, into primitive data to be inserted
    in the template.

    Data can be exported to view as members, properties or methods.

    `pack` method can be used to hold dict data in object like structure so
    that in templates we don't have to worry about dictionaries.
    """

    def __getitem__(self, name):
        """
        Implement read-only dict.

        Jinja will only request members exported in `keys` so we don't check
        again private members here.
        """
        return getattr(self, name)

    def keys(self):
        """
        Return all public members.

        This is called by Jinja to fill template data.
        """
        result = []
        for name in dir(self):
            if name.startswith('_'):
                continue
            if name in ['keys', 'pack']:
                continue
            if six.PY2:
                name = name.decode('utf-8')
            result.append(name)
        return result

    def pack(self, *args, **kwargs):
        """
        Pack arguments / dict into an object to simplify access from view.
        """
        return Bunch(*args, **kwargs)
