# Copyright (c) 2015 Adi Roiban.
# See LICENSE for details.
"""
Glob and regex matching utilities.
"""
from __future__ import absolute_import, unicode_literals

import fnmatch
import re
from datetime import datetime, timedelta

from twisted.internet import reactor

from chevah.server import force_unicode
from chevah.server.commons.exception import ServerException
from chevah.server.commons.public_format import datetime_format


def get_glob_rule(raw_rule):
    """
    Extract the glob from from the raw_rule.

    `raw_rule` can be a regular expression of other expression type and in
    this case `None` is returned.

    It can return an empty string.
    """
    if (
        raw_rule.startswith('m/')
        or raw_rule.startswith('e/')
        or raw_rule.startswith('t/')
            ):
        return None

    if raw_rule.startswith('g/'):
        # Still globbing but with explicit marker.
        rule = raw_rule[2:-1]
        return rule

    return raw_rule


def get_matched_groups(value, rule, usage, clock=reactor):
    """
    Return a tuple of matching groups if `value` is accepted by the expression
    defined in `rule`.

    Even when no groups are defined in the matching rule, on a match it will
    always have the input value as the first group.

    Return the empty tuple when no match is found.

    It can be used on both globbing or regex.

    When globbing is used it will return a single group, containing the full
    matching expression.

    Raise `ServerException` on any errors.
    """
    if not rule:
        return (value,)

    # FIXME:6229:
    # Use the get_glob_rule here once we no longer need to special
    # exception for file dispatcher.
    if not (
            rule.startswith('m/')
            or rule.startswith('g/')
            or rule.startswith('e/')
            or rule.startswith('t/')
            ):
        return _glob_matching(value, rule)

    if rule.startswith('g/'):
        # Still globbing but with explicit marker.
        rule = rule[2:-1]
        result = _glob_matching(value, rule)
        if not result:
            return result
        # FIXME:6229:
        # Globbing doesn't support grouping.
        # We fake a grouping result as workaround for enabling advance
        # rules in file-dispatcher event handler while keeping backward
        # compatibility. I hope this will not byte us in the future.
        #
        # We should look at removing this difference in future major
        # upgrades.
        return (result[0], result[0])

    if rule.startswith('t/'):
        now = datetime.fromtimestamp(clock.seconds())
        try:
            rule = _parse_datetime_configuration(rule, now)
            return _glob_matching(value, rule)
        except ServerException as error:
            raise ServerException(
                'Invalid datetime expression for "%s": %s' % (
                    usage, error.message))

    # We have regex filtering.
    rule_type = rule[0]
    try:
        rule, flags = _parse_regex_configuration(rule)

        match = re.search(rule, value, flags=flags)
    except (TypeError, re.error, ServerException) as error:
        message = 'Invalid regex for "%s": %s' % (
            usage,
            force_unicode(error),
            )
        raise ServerException(message)

    if rule_type == 'e':
        # We have exclusion rule and groups don't make sense.
        if not match:
            return (value,)
        else:
            return ()

    if not match:
        return ()

    return (value,) + match.groups()


def _glob_matching(value, rule):
    """
    Do globbing filtering, case insensitive.

    it supports multiple ORed expression using the | character
    """
    if not rule.startswith('!'):
        return _glob_any(value, rule)

    # We have a negation match
    rule = rule[1:].strip()
    if _glob_any(value, rule):
        # If the expression is a match, for a negation to return empty
        # result
        return ()

    # We didn't had a match, so since this is a negation we have a match.
    return (value,)


def _glob_any(value, rule):
    """
    Return a match value matches at least one single rule separated by
    pipe (|) character.
    """
    target = value.lower()
    parts = rule.lower().split('|')

    for expression in parts:
        if fnmatch.fnmatch(target, expression.strip()):
            return (value,)

    # No match.
    return ()


def _parse_regex_configuration(value):
    """
    Parse regex configuration option and return a tuple of (rule, flags).
    """
    # Do all matches in Unicode.
    flags = re.UNICODE

    # Remove leading m/ marker.
    rule = value[2:]

    # Remove trailing / marker and get trailing options ex /i.
    parts = rule.rsplit('/', 1)
    if len(parts) != 2:
        raise ServerException('Trailing marker not found.')
    rule = parts[0]
    raw_flags = parts[1]

    if not raw_flags:
        pass
    elif raw_flags == 'i':
        flags |= re.IGNORECASE
    else:
        raise ServerException('Unknown flags "%s".' % raw_flags)

    return (rule, flags)


def _parse_datetime_configuration(value, now):
    """
    Resolve the datetime matching rule.
    """
    parts = value[2:].rsplit('/', 1)
    if len(parts) != 2:
        raise ServerException('Trailing marker not found.')

    rule = parts[0]
    raw_flags = parts[1].lower()

    target = now

    if not raw_flags:
        pass
    elif raw_flags == '-1d':
        # Need to match for yesterday.
        target = target - timedelta(days=1)
    else:
        raise ServerException('Unknown flags "%s".' % raw_flags)

    return datetime_format(rule, target)
