# Copyright (c) 2011 Adi Roiban.
# See LICENSE for details.
"""
Observer implementation for Chevah project.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from zope.interface import implementer

from chevah.server.commons.interface import IObserver


class Signal(object):
    """
    A signal triggered by the observer.

    It contains a source together with additional arguments provided by the
    triggering call.
    """

    def __init__(self, source, **kwargs):
        self.source = source
        self._attributes = kwargs
        for key, value in kwargs.items():
            setattr(self, key, value)

    def getAllAttributes(self):
        """
        Return a dictionary with all signal attributes.
        """
        return self._attributes.copy()


@implementer(IObserver)
class ObserverMixin(object):
    """
    A base class for implementing observers.

    The internal subscribers database is kept by `_subscribers`.
    Since this class is designed to be used as an mixin the `_subscribers`
    member is generated on-demand.
    """
    _subscribers = None

    def unsubscribe(self, name=None, callback=None):
        """
        Unsubscribe callbacks.

        If `callback` is None, all callbacks for `name` will be removed.
        If `name` is None all callbacks will be removed.
        """
        if not self._subscribers:
            # No subscribers defined yet, so nothing to do.
            return

        if name is None:
            self._subscribers = {}
        else:
            if callback is None:
                self._subscribers[name] = []
            else:
                try:
                    self._subscribers[name].remove(callback)
                except (KeyError, ValueError):
                    # No callback for `name` or `callback` not subscribed yet.
                    pass

    def subscribe(self, name, callback):
        """
        Subscribe the callback to signal with `name`.
        """
        if self._subscribers is None:
            # Initialize subscribers.
            self._subscribers = {}

        if name in self._subscribers:
            # Append the callback to the existing list.
            if callback not in self._subscribers[name]:
                # Avoid adding the same callback multiple times.
                self._subscribers[name].append(callback)
        else:
            # Create a new list with the initial callback.
            self._subscribers[name] = [callback]

    def notify(self, name, signal=None):
        """
        Trigger all subscribers with name.

        Return a list with results from all subscribers.
        """
        if signal is None:
            signal = Signal(source=self)

        result = []
        if not self._subscribers:
            # No subscribers at all yet.
            return result

        if name not in self._subscribers:
            # No subscriber for signal name.
            return result

        # Make a copy of the subscribers as the callback might modify
        # the list.
        for callback in list(self._subscribers[name]):
            result.append(callback(signal))

        return result

    def hasSubscribers(self, name):
        """
        Return True if there are subscribers for event with `name`.
        """
        if not self._subscribers:
            return False

        subscribers = self._subscribers.get(name, None)
        if subscribers:
            return True

        return False
