# Copyright (c) 2010 Adi Roiban.
# See LICENSE for details.
"""
A process manages the running components of a product.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from datetime import timedelta
import os
import traceback

from twisted.internet import defer, reactor
from zope.interface import implementer

try:
    from setproctitle import setproctitle
    setproctitle
except ImportError:
    def setproctitle(name):
        """
        Does nothing.
        """

from chevah_compat import (
    process_capabilities,
    system_users,
    )
from chevah_compat.exceptions import ChangeUserException
from chevah.slim_server.pool import spawn_pool
from chevah.server import (
    build,
    emit as _emit,
    emit_error_details,
    event_dispatcher,
    force_unicode,
    )
from chevah.server.authentication.manager import AuthenticationsManager
from chevah.server.commons.configuration import ApplicationConfiguration
from chevah.server.commons.constant import (
    PRODUCT_NAME,
    PRODUCT_VERSION,
    SERVICE_NAME,
    )

from chevah.server.commons.exception import (
    ServerError,
    ServerException,
    )
from chevah.server.commons.interface import (
    IEventCreator,
    IProcessRunnables,
    IServerProcess,
    IRunnablesManager,
    )
from chevah.server.commons.status import ServerStatus
from chevah.server.commons.service import ServicesManager
from chevah.server.configuration.model import PropertyMixin
from chevah.server.event.handler import EventHandlersManager
from chevah.server.location.manager import LocationsManager
from chevah.server.process.worker import ChevahWorkers

from chevah.server.resource.manager import ResourcesManager
from chevah.server.transfer.transfer import TransfersManager


# Singleton worker pool.
global_workers = ChevahWorkers(name=SERVICE_NAME + '-worker')


@implementer(IEventCreator)
class ProcessPublicIdentity(PropertyMixin):
    """
    Export only the public attributes of a process to be handled by an
    event handler.
    """
    def __init__(self, process):
        self._parent = process
        self._configuration = None

    def setConfiguration(self, configuration):
        self._configuration = configuration

    @property
    def type(self):
        """
        See: `IEventCreator`.
        """
        return 'main-process'

    @property
    def uuid(self):
        """
        The UUID of this process/server instance.
        """
        if not self._configuration:
            return 'SERVER-UUID-NOT-YET-DEFINED'
        return self._configuration.server.uuid

    @property
    def name(self):
        """
        The name of this process/server instance.
        """
        if not self._configuration:
            return 'SERVER-NAME-NOT-YET-DEFINED'
        return self._configuration.server.name

    def emitEvent(self, event_id, data=None, avatar=None):
        """
        This is here just to comply with the interface.

        In the future we might want to emit all process events from here
        instead of emitting from the process instance.
        """
        self._parent.emitEvent(
            event_id, data=data, avatar=avatar, creator=self)


@implementer(IProcessRunnables)
class ProcessRunnables(PropertyMixin):
    """
    The collection with the status of all runnables for the process,
    include the status for the process itself.
    """
    def __init__(self, parent):
        self._parent = parent
        self._event_handlers = None
        self._resources = None
        self._services = None
        self._authentications = None
        self._locations = None
        self._transfers = None
        self._status = None

        # List of tuple with member name, start callable.
        # Start callable is called with process as first argument.
        # Start databases first so that DB log handlers can make use of them.
        self._members = [
            # Event handlers need to be started first as they should handle
            # the events as soon as possible.
            ('_event_handlers', EventHandlersManager),
            ('_resources', ResourcesManager),
            ('_services', ServicesManager),
            ('_authentications', AuthenticationsManager),
            # Locations needs to be started first as they are used by the
            # transfers.
            ('_locations', LocationsManager),
            ('_transfers', TransfersManager),
            ('_status', ServerStatus),
            ]

    @property
    def event_handlers(self):
        """
        See IProcessRunnables
        """
        return self._event_handlers

    @property
    def resources(self):
        """
        See IProcessRunnables
        """
        return self._resources

    @property
    def services(self):
        """
        See IProcessRunnables
        """
        return self._services

    @property
    def authentications(self):
        """
        See IProcessRunnables
        """
        return self._authentications

    @property
    def locations(self):
        """
        See IProcessRunnables
        """
        return self._locations

    @property
    def transfers(self):
        """
        See IProcessRunnables
        """
        return self._transfers

    @property
    def status(self):
        """
        See IProcessRunnables
        """
        return self._status

    def configure(self):
        """
        Initialize all runnable managers, but don't start the runnables.
        """
        for name, init_method in self._members:
            # FIXME:6603:
            # Set the real parent here.
            setattr(self, name, init_method(self.parent))

    @defer.inlineCallbacks
    def start(self):
        """
        Start all runnable managers.
        """
        for name, _ in self._members:
            manager = getattr(self, name)

            if not IRunnablesManager.providedBy(manager):
                continue

            yield manager.start()

    @defer.inlineCallbacks
    def stop(self):
        """
        Stop all runnable managers.
        """
        for name, _ in reversed(self._members):
            manager = getattr(self, name)

            if not manager or not IRunnablesManager.providedBy(manager):
                continue

            yield manager.stop()


@implementer(IServerProcess)
class ChevahServerProcess(PropertyMixin):
    """
    Main class that starts the server process and keeps it configured.
    """

    reactor = reactor
    _os = os  # Here to help with testing.

    # We have singleton event handlers and workers to help with tests.
    # During the tests we create a lot of ChevahServerProcess objects
    # and create and stopping these objects for each process will take a
    # lot of time.
    event_dispatcher = event_dispatcher
    workers = global_workers
    spawn_pool = spawn_pool

    def __init__(self, exit_callable=None):
        """
        See `IProcess`.
        """
        self._parent = None
        self.configuration = None
        self.runnables = ProcessRunnables(self)
        self.public_identity = ProcessPublicIdentity(self)

        #: A callable that will be triggered when terminating and takes care
        #: of process termination.
        self._exit_callable = self._os_exit
        if exit_callable:
            self._exit_callable = exit_callable

    @property
    def event_handlers(self):
        """
        FIXME:6521:
        Just a redirection to the new resource runnables.
        Should be removed in v6
        """
        return self.runnables.event_handlers

    @property
    def resources(self):
        """
        FIXME:6521:
        Just a redirection to the new resource runnables.
        Should be removed in v6
        """
        return self.runnables.resources

    @property
    def services(self):
        """
        FIXME:6521:
        Just a redirection to the new resource runnables.
        Should be removed in v6
        """
        return self.runnables.services

    @property
    def authentications(self):
        """
        FIXME:6521:
        Just a redirection to the new resource runnables.
        Should be removed in v6
        """
        return self.runnables.authentications

    @property
    def locations(self):
        """
        FIXME:6521:
        Just a redirection to the new resource runnables.
        Should be removed in v6
        """
        return self.runnables.locations

    @property
    def transfers(self):
        """
        FIXME:6521:
        Just a redirection to the new resource runnables.
        Should be removed in v6
        """
        return self.runnables.transfers

    @property
    def status(self):
        """
        FIXME:6521:
        Just a redirection to the new resource runnables.
        Should be removed in v6
        """
        return self.runnables.status

    def emitEvent(
            self, event_id, peer=None, avatar=None, data=None, creator=None):
        """
        See: IEventCreator
        """
        if creator is None:
            creator = self.public_identity
        _emit(event_id, creator=creator, peer=peer, avatar=avatar, data=data)

    def configure(self, configuration_path=None, configuration=None):
        """
        See `IProcess`.
        """
        # It will start with a few thread and can go up to that many
        # threads.
        reactor.suggestThreadPoolSize(40)

        try:
            if configuration is None:
                configuration = ApplicationConfiguration(
                    configuration_path, parent=self)

            self.configuration = configuration

            self.configuration.loadExternalConfiguration()

            # Configure the event handler as soon as possible.
            self.event_dispatcher.configure(
                definitions=configuration.events,
                configuration=configuration,
                )

            self.public_identity.setConfiguration(configuration)

            # This will also migrate the configuration.
            # It's important to do all the configuration migration,
            # before loading the runnables.
            self.configuration.validateConfiguration()
            self.runnables.configure()

            self._setProcessTitle()

            self._validateDemoExpiration()

            # On Unix daemon self.stop will not call sys.exit so
            # we need to signal in other way that this method
            # was successful.
            return True

        except Exception as error:
            data = {
                # Add the traceback, as this is critical low-level error.
                'tb': traceback.format_exc(),
                # We might not have the event details loaded, so we inject the
                # message here.
                # Make sure to keep it in sync with the `events.json` file.`
                'message': 'Fail to configure the process. %(details)s %(tb)s',
                }

            emit_error_details(
                event_id='20010',
                error=error,
                data=data,
                emitter=self.emitEvent,
                )

            self.stop(1)

    def _validateDemoExpiration(self):
        """
        Checks that the demo build expiration date is not reached.

        Emits the relevant events and raises an exception if the date expired.
        If the date is not in the past, it will schedule a second validation
        an hour after the date.
        """
        if build.EXPIRATION_TIMESTAMP == 0:
            # This version has no limitations.
            # Nothing to check.
            return

        expires_in = build.EXPIRATION_TIMESTAMP - self.reactor.seconds()
        days_left = timedelta(seconds=expires_in).days

        if days_left < 0:
            details = 'The demo has expired.'
        else:
            details = 'The demo expires in %s day(s).' % (days_left)

        self.emitEvent('20186', data={'details': details})

        if days_left < 0:
            raise ServerException(details)

        # Schedule the call to validate the metadata an hour after expiration.
        self.reactor.callLater(
            expires_in + 60 * 60,
            self._validateDemoExpirationCallback,
            )

    def _validateDemoExpirationCallback(self):
        """
        It is used to scheduled a later validation of the demo expiration date.

        In case of an exception, this method will emit the relevant event and
        will stop the process.
        """
        try:
            self._validateDemoExpiration()
        except ServerException:
            self.stop(1)

    def _setProcessTitle(self):
        """
        Set process title.

        Helper for testing.
        """
        setproctitle(SERVICE_NAME)

    def _getProductInfo(self):
        # We might now have the the status object initialized.
        # for example if we fail to start the process.
        instance_name = PRODUCT_NAME
        if self.status:
            instance_name = self.status.instance_name

        data = {
            'instance_name': instance_name,
            'product_version': PRODUCT_VERSION,
            }
        return data

    def start(self):
        """
        See `IProcess`.

        Execution will block after this call as it starts the reactor.
        """

        def on_proccess_start():
            """
            Called when the process should start.
            """
            self.emitEvent('20001', data=self._getProductInfo())

            self._informCapabilities()
            self._umask(self.configuration.server.umask_as_int)

        def eb_exit_on_errors(failure):
            """
            Emit error and stop process.
            """
            if not failure.check(ServerError):
                traceback = force_unicode(failure.getBriefTraceback())
            else:
                traceback = ''

            emit_error_details(
                event_id='20004',
                error=failure.value,
                data={'traceback': traceback},
                emitter=self.emitEvent,
                )
            self.stop(1)
            return False

        def cbProcessStarted(result):
            """
            Emit an event when all components have started.
            """
            self.emitEvent('20181', data=self._getProductInfo())

        deferred = defer.maybeDeferred(on_proccess_start)
        deferred.addCallback(self._cbDropPrivileges)
        deferred.addCallback(self._cbStartAll)
        deferred.addCallback(cbProcessStarted)
        deferred.addErrback(eb_exit_on_errors)

        self._run()
        # _Run is a blocking call and this return is here to help with
        # testing where _run is mocked.
        return deferred

    def _umask(self, mask):
        """
        Set process umask.

        This is defined to help with testing.
        """
        os.umask(mask)

    def _run(self):
        """
        Method created to help with testing.
        """
        # Start the reactor main loop.
        # After this call the server will enter reactor's main loop and
        # no other code will be executed this line.
        self.reactor.run(installSignalHandlers=False)

    def stop(self, exit_code=0):
        """
        See `IProcess`.
        """
        if self.configuration and self.configuration.server:
            self.configuration.server.unsubscribe('update')
        data = self._getProductInfo()
        # Process we might fail to load event definitions,
        # so then we pass a message as a fallback.
        data['message'] = 'Stopping %(instance_name)s - %(product_version)s.'
        self.emitEvent('20002', data=data)
        self._exit(exit_code)

    def _os_exit(self, exit_code=0):
        """
        Terminates process. Helper method for Windows service implementation.

        Using os._exit() to avoid re-raising SystemExit error.
        """
        os._exit(exit_code)

    def _exit(self, exit_code=0):
        """
        Helper method for testing.

        When reactor is running add a hook to be called after reactor stop.

        This is mocked in tests since we can not addSystemEventTrigger
        multiple times for the same reactor.
        """
        if self.reactor.running:
            # Make sure to keep in sync these steps with the else part.
            self.reactor.addSystemEventTrigger(
                'before', 'shutdown', self._stopAll)
            self.reactor.addSystemEventTrigger(
                'after', 'shutdown', self._shutdown, exit_code)
            self.reactor.stop()
        else:
            # Process failed to start before starting the reactor.
            # When reactor is stopped we do the shutdown in synchronous mode.
            self._stopAll()
            self._shutdown(exit_code)

    def _shutdown(self, exit_code):
        """
        Perform last steps just before exiting the process.
        """
        # FIXME:2252:
        # Replace with single call to event_dispatcher.shutdown() once we fix
        # the event_dispatcher shutting down.
        self.event_dispatcher.removeConfiguration()
        self._exit_callable(exit_code)

    @defer.inlineCallbacks
    def _cbStartAll(self, result):
        """
        Call start on all runnable managers and.

        It is called as a callback.
        """
        self.spawn_pool.start()
        yield self.workers.start()
        yield self.runnables.start()

    @defer.inlineCallbacks
    def _stopAll(self):
        """
        Call stop on all runnable managers and sub-process.
        """
        yield self.runnables.stop()
        yield self.workers.stop()
        self.spawn_pool.stop()

    def _informCapabilities(self):
        """
        Emit events for available / missing capabilities.
        """
        data = {
            'cryptography_library_version': self.status.openssl_version,
            'process_privileges': (
                process_capabilities.getCurrentPrivilegesDescription()),
            # force_unicode can be removed once we no longer need py2.
            'cwd': self.status.cwd,
            'python_version': self.status.python_version,
            'windows_legacy_service': '',
            # We log the actual username at start time.
            # Most of the time, it should be the same.
            # The process.status.os_name can also be use to check this
            # value at any time.
            'os_user': self.status.os_user,
            }
        if self.status.windows_legacy_service:
            data['windows_legacy_service'] = (
                'Legacy Windows service detected. '
                'Updating the Windows service is required. '
                )

        self.emitEvent('20072', data=data)
        if not self.status.can_impersonate_local_accounts:
            self.emitEvent('20070')
        if not self.status.can_get_local_home_folders:
            self.emitEvent('20074')
        if not self.status.can_create_local_home_folders:
            self.emitEvent('20073')

    def _cbDropPrivileges(self, result):
        """
        Drop process privileges to service account.

        If service account is disable, it will try to run as "SUDO_USER".
        It sudo users fails, it will just give up trying to drop privileges.
        """
        if os.name == 'nt':
            # We don't try to drop privileges on Windows.
            return result

        account = self._getServiceAccount()
        if not account:
            if self._os.getuid() == 0 or self._os.geteuid() == 0:
                raise ServerError('20008')
            # Emit a warning that the server is running under the same account
            # and return the previous result.
            self.emitEvent('20069')
            return result

        try:
            system_users.dropPrivileges(username=account)
        except ChangeUserException as error:
            text = error.message
            data = {
                'account': account,
                'details': text,
                }
            raise ServerError('20007', data=data)

        # Emit an info that account was switched and return previous result.
        self.emitEvent('20071', data={'account_name': account})
        return result

    def _getServiceAccount(self):
        """
        Return the account name under which to execute the process.

        First try the `[server] account` configuration option.

        If disabled:
          * check if we are running under SUDO and return the initial account
            who called "SUDO".

        If all fails, return None and continue as current user.
        """
        account = self.configuration.server.account
        # Return the configured value if it was not disabled.
        if account:
            return account

        # If no service account was explicitly configured,
        # check if application was launched using "sudo" and return
        # the initial account who launched "sudo".
        try:
            account = os.environ['SUDO_USER']
        except KeyError:
            # Just ignore if could not get 'SUDO_USER'.
            pass

        # Just give up and return an empty account.
        # Process should log an warning.
        return ''
