# Copyright (c) 2017 Adi Roiban.
# See LICENSE for details.
"""
The standard string format() API is not safe when the format string is provided
by the user. This module provides a public_format() which protects against
misuse and/or information leak.

Based on http://lucumr.pocoo.org/2016/12/29/careful-with-str-format/.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
from string import Formatter
import numbers
import os
import re

from bunch import Bunch

from chevah.server.commons.exception import ServerException
import six

# This is a necessary API but it's undocumented and moved around
# between Python releases.
# It returns a tuple with the name of the first field and an iterator of tuples
# in the format (is_attribute, field).
#
# For example: 'object.attr[2]'._formatter_field_name_split() will
# return:
# ('object', <iterator>), and the iterator contains:
# ((True, 'attr'), (False, 2)).
try:
    from _string import formatter_field_name_split
except ImportError:
    # This can be removed once we are only on Py3.
    def formatter_field_name_split(s):
        return s._formatter_field_name_split()


class _PublicFormatter(Formatter):
    """
    A formatter implementation not allowing format_spec or conversion
    syntax and also protects access to private attributes.
    """
    def convert_field(self, value, conversion):
        """
        Raises ValueError when trying to use the formatter conversion syntax.

        See: `Formatter`.
        """
        if conversion:
            raise ValueError('Conversion syntax is not allowed.')

        return super(_PublicFormatter, self).convert_field(value, conversion)

    def format_field(self, value, format_spec):
        """
        Raises ValueError when trying to use the format_spec syntax.

        See: `Formatter`.
        """
        if format_spec:
            raise ValueError(
                'Format specification syntax is not allowed.')

        result = super(_PublicFormatter, self).format_field(value, format_spec)
        if isinstance(result, bytes):
            return result.decode('utf-8')
        else:
            return result

    def get_field(self, field_name, args, kwargs):
        """
        Raises ValueError if trying to access a private attribute or
        a method.

        See: `Formatter`.
        """
        field, attributes = formatter_field_name_split(field_name)
        try:
            # str._formatter_field_name_split can't handle { 0 }.
            # It only works with {0}.. so we add support here.
            field = int(field)
        except ValueError:
            pass

        if not isinstance(field, numbers.Number):
            field = field.lower().strip()

        obj = self.get_value(field, args, kwargs)
        for is_attr, attr in attributes:

            if is_attr:
                if attr.startswith('_'):
                    raise ValueError(
                        'Private access: %s.' % (field_name,))
                attr = attr.lower().strip()
                obj = getattr(obj, attr)
            else:
                if not isinstance(attr, six.integer_types):
                    raise ValueError(
                        'Index access: %s.' % (field_name,))
                obj = obj[int(attr)]

            if callable(obj):
                raise ValueError(
                    'Callable access: %s.' % (field_name,))

        return obj, field


def public_format(format_string, *args, **kwargs):
    """
    Formats the string filtering access to unwanted attributes and not allowing
    padding or type conversions.

    `format_string` is the template to be resolved.

    This is the low-level API that we use to implement string expansion
    based on user input.
    It has no custom syntax, other than standard Python syntax for string
    formatting.

    It exists to make sure we don't have security issues and we restrict
    what can be done using standard string formatting.
    See https://lucumr.pocoo.org/2016/12/29/careful-with-str-format/

    Returns Unicode text.
    """
    if not format_string:
        return ''

    formatter = _PublicFormatter()
    try:
        return formatter.vformat(format_string, args, kwargs)
    except (KeyError, AttributeError) as error:
        raise ServerException('Invalid variable in "%s": "%s"' % (
            format_string, error.args[0]))
    except (ValueError, IndexError) as error:
        raise ServerException('Invalid format in "%s": %s' % (
            format_string, error.args[0]))


def isidentifier(name):
    """
    A function equivalent to the str.isidentifier method on Py3.

    From:
    http://python-future.org/_modules/future/utils.html
    """
    name_re = re.compile(r"[a-zA-Z_][a-zA-Z0-9_]*$")
    for identifier in name.split('.'):
        if not bool(name_re.match(identifier)):
            return False
    return True


class NoStrBunch(Bunch):
    """
    A Bunch which can't be directly serialized so you can only get the
    end nodes.
    """

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        raise ValueError('Access to inner value denied.')


def datetime_format(rule, when, extra=None):
    """
    Expand the `rule` based on `when` datetime.

    To be human friendly it will auto translate to strptime format based on
    OS version.

    year:
        decimal
        no_century
    month:
        decimal
        padded
        name
        abbreviated
    day:
        padded
        decimal
        of_year_padded
        name
        abbreviated
    hour:
        padded_24
        decimal_24
        padded_12
        decimal_12
        am_pm
    minute:
        padded
        decimal
    second:
        padded
        decimal
    """

    translation = Bunch(
        year=Bunch(
            decimal='%Y',
            no_century='%y',
            ),
        month=Bunch(
            decimal='%-m',
            padded='%m',
            name='%B',
            abbreviated='%b',
            ),
        day=Bunch(
            decimal='%-d',
            padded='%d',
            of_year_padded='%j',
            name='%A',
            abbreviated='%a',
            ),
        hour=Bunch(
            decimal_24='%-H',
            padded_24='%H',
            decimal_12='%-I',
            padded_12='%I',
            am_pm='%p',
            ),
        minute=Bunch(
            decimal='%-M',
            padded='%M',
            ),
        second=Bunch(
            decimal='%-S',
            padded='%S',
            ),
        )

    if extra:
        translation.update(extra)

    if os.name == 'nt':
        # Use special decimal formating for Windows.
        translation.month.decimal = '%#m'
        translation.day.decimal = '%#d'
        translation.hour.decimal_24 = '%#H'
        translation.hour.decimal_12 = '%#I'
        translation.minute.decimal = '%#M'
        translation.second.decimal = '%#S'

    translated = public_format(rule, **translation)
    # FIXME:6193:
    # We should look at a different option,
    # other than strftime that is not locale dependent.
    if six.PY2:
        return when.strftime(translated.encode('utf-8')).decode('utf-8')
    else:
        return when.strftime(translated)
