# Copyright (c) 2011-2014 Adi Roiban.
# See LICENSE for details.
"""
Batch processing queues.
"""
from __future__ import unicode_literals

from __future__ import absolute_import
from twisted.internet import defer, reactor
from zope.interface import Interface


class MaxCapacityReachedException(Exception):
    """
    Queue capacity is exceeded.
    """

    def __init__(self, capacity):
        self.message = 'Maximum queue capacity of %d reached.' % (capacity,)
        super(MaxCapacityReachedException, self).__init__(self.message)


class IBatchProcessor(Interface):
    """
    Batch processor used by `BatchProcessQueue`.
    """

    def process(batch):
        """
        Process batch.
        """


class BatchProcessQueue(object):
    """
    A queue that processes it's items sequentially, in batches of
    maximum the pre-configured size.
    """

    #: Number of items in the queue after which the queue will refuse to
    # accept new items/requests.
    CAPACITY = 1024
    #: Maximum number of items in a batch. When queue has more items, the
    # remaining items will be delayed for the next batch
    MAX_BATCH_SIZE = 512

    def __init__(
            self,
            processor,
            capacity=CAPACITY,
            max_batch_size=MAX_BATCH_SIZE,
            ):
        # This is the queue of the items which are pending to be processed.
        self._buffer = []
        self._processor = processor
        self._capacity = capacity
        self._max_batch_size = max_batch_size
        #: Keeps track of in-progress batch and fires upon it's
        # completion regardless of success/failure.
        self._process_deferred = None
        self._running = False

    def start(self):
        """
        Start(restart) queue processing.
        """
        self._resume()
        return defer.succeed(None)

    @property
    def running(self):
        """
        `True` if queue is started, `False` otherwise.
        """
        return self._running

    def _pause(self):
        """
        Pause queue processing.
        """
        self._running = False

    def _resume(self):
        """
        Resume queue processing.
        """
        self._running = True
        self._tryProcessing()

    def stop(self):
        """
        Pause the queue and cancels the current in-progress batch deferred.

        It returns the canceled deferred to wait for any cleanup.
        """
        self._pause()
        self._buffer = []

        in_progress = defer.succeed(None)
        if self._process_deferred:
            in_progress, self._process_deferred = self._process_deferred, None
            in_progress.cancel()

        return in_progress

    def append(self, item):
        """
        Append item to queue.

        Raises an error if queue is full.
        """
        if len(self._buffer) >= self._capacity:
            # FIXME:2447:
            # Implement support for eviction policies.
            raise MaxCapacityReachedException(self._capacity)

        self._buffer.append(item)
        self._tryProcessing()

    def _clear(self):
        """
        Remove all items from the queue.
        """
        self._buffer = []

    @property
    def size(self):
        """
        Return the number of items in the queue.
        """
        return len(self._buffer)

    @defer.inlineCallbacks
    def _process(self):
        """
        Split queue in batches and delegate processing to `processor`.
        """
        def eb_process(failure, batch):
            """
            Called when the pending processing failed.
            """
            if failure.check(defer.CancelledError):
                # Forward the cancellation.
                return failure

            # FIXME:2446:
            # Implement failure reconciliation policies.
            # For now, ignore the errors.
            return None

        while self._running and self._buffer:
            batch = self._buffer[:self._max_batch_size]
            del self._buffer[:self._max_batch_size]

            deferred = defer.maybeDeferred(self._processor.process, batch)
            deferred.addErrback(eb_process, batch)
            yield deferred

    def _tryProcessing(self):
        """
        Trigger queue processing if not already in-progress and queue is
        running.
        """
        # Ignore if no entries to process.
        if not self._buffer:
            return

        # Ignore if already processing the queue.
        if self._process_deferred:
            return

        # Ignore if not running.
        if not self.running:
            return

        def bb_process(result_or_failure):
            """
            Clear processing in-progress flag on both success and failure.
            """
            self._process_deferred = None
            return result_or_failure

        def cb_process(result):
            """
            Continue processing.
            """
            self._tryProcessing()
            return result

        deferred = self._process()
        self._process_deferred = deferred
        deferred.addBoth(bb_process)
        deferred.addCallback(cb_process)


class SingleProcessQueue(object):
    """
    Handle a single job at a time.

    All other jobs are queued.

    Nothing is processed unless the queue is started or stopped.

    created -> stopped (empty queue)

    stopped (empty queue)   --add-job--> stopped (queued jobs)
                            --stop--> stopped (empty queue)
                            --start--> running (empty queue)

    stopped (queued jobs)   --stop--> cancel jobs
                            --add-job--> stopped (queued jobs)
                            --start--> running (queued jobs)

    running (empty queue)   --start--> running (empty queue)
                            --stop--> stopped (empty queue)
                            --add-job--> running (queued jobs)

    running (queued jobs)   --start--> running (queued jobs)
                            --stop--> cancel jobs
                            --add-job--> running (queued jobs)
                            --job-done--> running (queued jobs)
                            --all-jobs-done --> running (empty queue)

    cancel jobs -> stopped (empty queue)
    """

    def __init__(self):
        self._started = False
        self._queue = []
        # Ongoing job.
        self._process_deferred = None

    def queue(self, _callback, *args, **kwargs):
        """
        Add the request to the processing queue.

        `callback` might return a deferred.

        Returns a deferred that is called when the operation is done.
        """
        deferred = defer.Deferred()
        values = (deferred, _callback, args, kwargs)
        self._queue.append(values)
        if self._started:
            self._process()
        return deferred

    def start(self):
        """
        Start processing any queued jobs.
        """
        self._started = True
        self._process()

    def stop(self):
        """
        Clear the queue and cancel any pending operations.
        """
        self._started = False
        if not self._process_deferred:
            return

        # We clear the queue first and then clear the pending deferred,
        # as otherwise if we cancel the pending deferred the queue will be
        # automatically triggered.
        while self._queue:
            values = self._queue.pop(0)
            deferred = values[0]
            deferred.cancel()

        self._process_deferred, deferred = None, self._process_deferred
        deferred.cancel()

    @defer.inlineCallbacks
    def _process(self):
        """
        Trigger the execution of the next jobs from the queue.
        Wait for that job to finalize and trigger the next job.

        This method should never fail as there is no extra code to handle
        it's failures.

        Any failure from the is forwarded to the deferred associated with
        the job.
        """
        # Ignore if already processing the queue.
        if self._process_deferred:
            return

        if not self._queue:
            # Nothing queued or suspended.
            return

        job_deferred, callback, args, kwargs = self._queue.pop(0)
        try:
            self._process_deferred = callback(*args, **kwargs)
            result = yield self._process_deferred
            job_deferred.callback(result)

        except Exception as error:
            # Bubble the error and ignore the error from the current stack.
            job_deferred.errback(error)

        finally:
            self._process_deferred = None
            reactor.callLater(0, self._process)
