# Copyright (c) 2011-2014 Adi Roiban.
# See LICENSE for details.
"""
Realm as defined by the Twisted Cred framework.

The code started as designed for Twisted Cred, but it evolved into something
of its own.

This handles the post-authentication checks process and the authorization
checks or preparing the authorization object (avatar).
"""
from __future__ import unicode_literals
from __future__ import absolute_import
import errno

from twisted.internet import defer
from zope.interface import implementer

from chevah_compat import (
    DefaultAvatar,
    LocalFilesystem,
    local_filesystem,
    SuperAvatar,
    )
from chevah_compat.exceptions import CompatError
from chevah.server import emit, force_unicode
from chevah.server.commons.constant import TYPE_NAME
from chevah.server.commons.exception import ServerError, ServerException
from chevah.server.commons.interface import (
    IPasswordCredentials,
    IOSAccountConfiguration,
    IServiceAvatar,
    IServiceRealm,
    )
from chevah.server.identity.avatar import ServiceAvatar


@implementer(IServiceRealm)
class RealmAbstract(object):
    """
    Realm for Chevah services.

    It transforms an authenticated account configuration to an avatar object
    which is uses by each service to identify each users and get user specific
    data and behavior.
    """

    def __init__(self, service_configuration):
        self._service_configuration = service_configuration

    def _requestAvatar(self, authentication_result, mind=None, *interfaces):
        """
        Return a deferred that returns the avatar based on
        authentication_result.
        """
        raise NotImplementedError('_requestAvatar not implemented')

    @property
    def service_configuration(self):
        return self._service_configuration

    @property
    def main_configuration(self):
        return self._service_configuration.root.configuration

    @property
    def _analytics(self):
        """
        Return the analytics resource.
        """
        return self._service_configuration.root.resources.getAnalytics()

    def requestAvatar(self, authentication_result, mind=None, *interfaces):
        """
        Return a deferred that returns the avatar for credentials.
        """
        _, _, credentials = authentication_result

        def wrap_internal_error(failure):
            '''Wrap internal server errors into ServerError.'''
            if failure.check(ServerError):
                return failure

            data = {
                'username': credentials.username,
                'details': force_unicode(failure.getBriefTraceback()),
                }
            raise ServerError('20058', data=data, peer=credentials.peer)

        def format_result(result):
            """
            Get avatar result and format result as required by the realm
            interface.
            """
            (interface, avatar) = result
            return (
                interface,
                avatar,
                getattr(avatar, 'logout', lambda: None),
                )

        deferred = defer.maybeDeferred(
            self._requestAvatar, authentication_result)
        deferred.addCallback(format_result)
        deferred.addErrback(wrap_internal_error)
        return deferred


class ServiceRealm(RealmAbstract):
    """
    Realm for services.
    """

    # Here for dependency injection during tests.
    LocalFilesystem = LocalFilesystem

    def emitEvent(self, event_id, avatar=None, data=None):
        """
        Emit an event for this realm.
        """
        emit(event_id, avatar=avatar, data=data)

    @defer.inlineCallbacks
    def _requestAvatar(self, authentication_result, mind=None, *interfaces):
        """
        Return a deferred that returns the avatar based on
        authentication_result.
        """
        account_configuration, groups_configuration, credentials = (
            authentication_result)

        token = self._getToken(account_configuration, credentials)

        try:
            avatar = ServiceAvatar(
                account_configuration=account_configuration,
                groups=groups_configuration,
                peer=credentials.peer,
                token=token,
                component=credentials.creator,
                )

            # Then check the home folder.
            filesystem = self._getFilesystem(avatar)
            self._verifyHomeFolder(filesystem, avatar)
            # Check that later a shell can use filesystem for avatar.
            self.LocalFilesystem(avatar=avatar)
            # Now we can create the structure.
            self._createRequiredFolders(filesystem, avatar)
            yield self._analytics.accountAuthenticated(avatar)

            defer.returnValue((IServiceAvatar, avatar))

        except ServerError as error:
            # Bubble the known error.
            raise error

        except (ServerException, OSError, CompatError) as error:
            # This is a known error for this code,
            # so we convert it into a ServerError to be handled at higher
            # level.
            data = {
                'username': account_configuration.name,
                'details': force_unicode(error),
                }
            raise ServerError('20031', peer=credentials.peer, data=data)

    def _getToken(self, account_configuration, credentials):
        """
        Return the token for this account configuration.
        """
        if credentials.type == TYPE_NAME.OIDC:
            # For OpenID connect credentials we propagate the token,
            # in the case this is needed for extensions.
            return credentials.token

        if not IOSAccountConfiguration.providedBy(account_configuration):
            return None

        if not IPasswordCredentials.providedBy(credentials):
            return None

        return getattr(credentials, 'token', None)

    def _verifyHomeFolder(self, filesystem, avatar):
        """
        Public method and helper for callback testing.
        """
        home_segments = local_filesystem.getSegmentsFromRealPath(
            avatar.home_folder_path)

        try:
            if local_filesystem.isFolder(home_segments):
                # Home folder exists no need to create it.
                return

            if local_filesystem.exists(home_segments):
                # Fail if folder exists but is not a folder.
                raise ServerException('Path is not a folder.')

        except OSError as error:
            raise ServerException(
                'Failed to check home folder existence. %s' % (
                    force_unicode(error)))

        if not avatar.create_home_folder:
            raise ServerException(
                'Home folder does not exist at "%s".' % (
                    avatar.home_folder_path))

        self._attemptHomeFolderCreation(filesystem, avatar)

    def _attemptHomeFolderCreation(self, filesystem, avatar):
        """
        Try to create home folder for this avatar.

        This is an transactional operation and folder is deleted in case there
        were errors while creating and configuring it.
        """
        home_folder_path = avatar.home_folder_path
        home_segments = filesystem.getSegmentsFromRealPath(home_folder_path)

        try:
            self._createFolder(
                avatar=avatar,
                filesystem=filesystem,
                path=home_folder_path
                )
        except Exception as initial_error:
            try:
                try:
                    filesystem.deleteFolder(
                        segments=home_segments, recursive=False)
                except OSError as os_error:
                    if os_error.errno != errno.ENOENT:
                        raise ServerException(
                            'Failed to revert partially created home folder. '
                            '%s' % (force_unicode(str(os_error))))
            except Exception as current_error:
                # We are already in an exception and we want to throw the
                # original exception, not this secondary one.
                # Most probably this should be a critical security error, as
                # the home folder might be in an inconsistent state.
                # We emit an error and hope that event handler will decide
                # to quit.
                data = {
                    'home_folder_path': avatar.home_folder_path,
                    'details': force_unicode(str(current_error)),
                    }
                # We emit the event as the root process.
                self.emitEvent('20075', avatar=avatar, data=data)

            raise initial_error

    def _getFilesystem(self, avatar):
        """
        Return a default filesystem which can be used to handle files
        for `avatar`.
        """
        if avatar.use_impersonation:
            # We allow SuperAvatar to get our own token.
            # Only used on Windows.
            fs_avatar = SuperAvatar()
            fs_avatar.token = avatar.token
            filesystem = self.LocalFilesystem(avatar=fs_avatar)
        else:
            filesystem = self.LocalFilesystem(avatar=DefaultAvatar())
        return filesystem

    def _createFolder(self, avatar, filesystem, path):
        """
        Create a folder with the required owner and group.
        """
        segments = filesystem.getSegmentsFromRealPath(path)

        try:
            filesystem.createFolder(
                segments=segments, recursive=False)
        except OSError as error:
            raise ServerException(
                'Failed to create missing path in home "%s". %s' % (
                    path, force_unicode(error)))

        self._setOwnerAndGroup(
            avatar=avatar, filesystem=filesystem, segments=segments)

        owner = avatar.create_home_folder_owner
        if not owner:
            owner = 'Default'
        group = avatar.create_home_folder_group
        if not group:
            group = 'Default'

        data = {
            'path': path,
            'owner':  owner,
            'group': group,
            }
        # Emit the event as the base process.
        self.service_configuration.root.emitEvent(
            '20042', avatar=avatar, data=data)

    def _setOwnerAndGroup(self, avatar, filesystem, segments):
        """
        Helper for setting owner and group for a path specified by segments.
        """
        try:
            if avatar.create_home_folder_owner:
                filesystem.setOwner(
                    segments=segments, owner=avatar.create_home_folder_owner)

            if avatar.create_home_folder_group:
                filesystem.addGroup(
                    segments=segments, group=avatar.create_home_folder_group)
        except CompatError as error:
            raise ServerException(
                'Failed to set owner/group as "%s/%s" for path "%s". %s' % (
                    avatar.create_home_folder_owner,
                    avatar.create_home_folder_group,
                    filesystem.getRealPathFromSegments(segments),
                    error.message,
                    )
                )

    def _createRequiredFolders(self, filesystem, avatar):
        """
        Create the folders required for this avatar after login.
        """
        required_paths = avatar.required_folders

        if not required_paths:
            return

        for path in required_paths:
            segments = filesystem.getSegmentsFromRealPath(path)
            if filesystem.isFolder(segments):
                # Already exists.
                continue

            # This will try to create the folder, if a folder is not there.
            # It will fail if a file with the same name exists.
            self._createFolder(
                avatar=avatar,
                filesystem=filesystem,
                path=path,
                )
