# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
from __future__ import absolute_import
from __future__ import unicode_literals
from time import time
import traceback
import random

from bunch import Bunch
from zope.component import queryUtility
from zope.interface import implementer
from twisted.internet import defer, reactor
from twisted.python.failure import Failure

from chevah.server import force_unicode
from chevah.server.commons.constant import STATE
from chevah.server.commons.exception import ServerError, ServerException
from chevah.server.commons.interface import IRunnable, IRunnablesManager
from chevah.server.commons.observer import Signal
from chevah.server.configuration.model import PropertyMixin


@implementer(IRunnable)
class Runnable(PropertyMixin):
    """
    See: `IRunnable`.

    Keeps weak references to both parent and configuration.

    Specialized classes SHOULD implement the methods:
    * _onStart - called to start the specialized runnable
    * _onStop - called to stop the specialized runnable

    Classed MAY implement the methods:
    * _afterStart - called after runnable was started
    * _afterStop - called after runnable was stopped

    Provides support to auto fail when a pre-configured number of failures is
    reached. To enable it call:
    * _cbActionSucceeded each time an operation is successful.
    * _ebActionFailed each time an operation fails.
    _ebActionFailed should be called after _cbActionSucceeded as when
    auto fail is reached the _ebActionFailed will succeed with the
    result from fail() action.

    A runnable can be in one of the following main states:
    * STOPPED - not yet started or manually stopped with success.
    * STARTING - in process of starting to be operational
    * STOPPING - in process of stopping the runnable.
    * FAILING - in process of stopping due to a failure.
    * STARTED - up and running and in full operational capabilities
      should be renamed to RUNNING at some point.
    * SUSPENDED - started but paused based on configuration or admin action.
    * FAILED_AT_RUN, FAILED_AT_START, FAILED_AT_STOP - not operational due to a
        failure

    If a specialized runnable add more states it should also update the list of
    started/operating states.

    Signals:
    * start - after a successful start
    * stop - after a successful stop, or on failures
    """

    # States in which the runnable is started, is going to be started and is
    # not going to be stopped or fail.
    _STARTED_STATES = [
        STATE.STARTED,
        STATE.SUSPENDED,
        STATE.STARTING,
        STATE.DISCONNECTED,
        STATE.SOURCE_STOPPED,
        ]

    # States in which the runnable is operating and ready and able to process
    # the requested actions.
    _OPERATING_STATES = [
        STATE.STARTED,
        ]

    _state = STATE.STOPPED
    _start_time = None
    _stop_time = None
    _startup_configuration = None

    _last_error = None

    _failure_count = 0

    # Clock used to schedule calls.
    _scheduler = reactor
    # Random module use to get random values.
    _random = random

    #: Set with names of configuration options which when changed,
    #: require the restart of the runnable.
    PROPERTIES_REQUIRING_RESTART = None

    def __init__(self, parent, configuration):
        self._parent = parent
        self._configuration = configuration
        self.__reset()

    def __reset(self):
        """
        Called when internal state should be reset.
        """
        self._failure_count = 0
        self._setLastError(None)
        self._state = STATE.STOPPED

    @property
    def configuration(self):
        """
        See: `IRunnable`.
        """
        return self._configuration

    def _onConfigurationFinalize(self, weak_reference):
        """
        Called when configuration is about to be finalized.

        This is called when configuration is removed.
        When configuration is removed the runnable is stopped, but the
        runnable might require various configuration values during the
        shutdown. For example `name` or `type`
        """
        if not self._startup_configuration:
            # We cannot replace the configuration.
            self._configuration = None
        else:
            # Use the version we got when starting. It might not be as
            # accurate, but it should be something.
            self._configuration = Bunch(**self._startup_configuration)

    @property
    def scheduler(self):
        """
        See: `IRunnable`.
        """
        return self._scheduler

    @property
    def uuid(self):
        """
        See: `IRunnable`.
        """
        return self._getKnownConfiguration('uuid', '')

    @property
    def enabled(self):
        """
        See: `IRunnable`.
        """
        return self._getKnownConfiguration('enabled', False)

    @property
    def state(self):
        """
        See: `IRunnable`.
        """
        return self._state

    @property
    def start_time(self):
        """
        See: `IRunnable`.
        """
        return self._start_time

    @property
    def stop_time(self):
        """
        See: `IRunnable`.
        """
        return self._stop_time

    @property
    def name(self):
        """
        See: `IRunnable`.

        Return runnable name for configuration.
        """
        return self._getKnownConfiguration('name', '')

    @property
    def short_description(self):
        """
        The short description of this runnable.
        """
        return ''

    @property
    def type(self):
        """
        See: `IRunnable`.
        """
        return self._getKnownConfiguration('type', '')

    @property
    def _kind(self):
        """
        Human readable type to show in the events.

        Some runnables don't have a type.
        """
        if not self.type:
            return ''
        return ' of type "%s"' % (self.type)

    @property
    def family(self):
        """
        See: `IRunnable`.
        """
        for interface in self.__implemented__.flattened():
            try:
                return interface.getTaggedValue('family')
            except KeyError:
                pass
        raise RuntimeError('No family defined.')

    @property
    def startup_configuration(self):
        """
        See: `IRunnable`.
        """
        return self._startup_configuration

    @property
    def restart_required(self):
        """
        See `IRunnable`.
        """
        result = []

        if not self.operating:
            # Runnable is not operating so there is no need to restart.
            # Either stopped, failed or in the process of starting.
            return result

        for member_name in self.PROPERTIES_REQUIRING_RESTART:
            startup_value = self._getStartupConfiguration(member_name)
            current_value = self._configuration.getProperty(member_name)
            if current_value != startup_value:
                result.append(member_name)

        return result

    @property
    def last_error(self):
        """
        See: `IRunnable`.

        Return `ServerError` instance or `None` if not set.
        """
        return self._last_error

    @property
    def _max_failures(self):
        """
        See: `IRunnable`.
        """
        return self._configuration.fail_after_errors

    def _setLastError(self, error):
        """
        Will set `_last_error` to the specified error if it's an instance
        of `ServerError` or `None`.

        It will raise an exception otherwise.
        """
        if error is None:
            self._last_error = None
            return

        if not isinstance(error, ServerError):
            message = '%s is not an instance of ServerError' % (error,)
            raise RuntimeError(message)

        self._last_error = error

    @property
    def started(self):
        """
        See: `IRunnable`.
        """
        return self.state in self._STARTED_STATES

    @property
    def operating(self):
        """
        See: `IRunnable`.
        """
        return self.state in self._OPERATING_STATES

    @property
    def failed(self):
        """
        See: `IRunnable`.
        """
        return self.state in [
            STATE.FAILING,
            STATE.FAILED_AT_START,
            STATE.FAILED_AT_RUN,
            STATE.FAILED_AT_STOP,
            ]

    def _getStartupConfiguration(self, property_name):
        """
        Return startup configuration for value specified as `property_name`.

        It is here to help with integration of ServiceManager code.
        """
        if not self._startup_configuration:
            return None

        return self._startup_configuration.get(property_name, None)

    def _getKnownConfiguration(self, property_name, default=''):
        """
        Try to get the configuration value for `property_name`
        Either from current configuration or from startup configuration.
        """
        try:
            result = getattr(self._configuration, property_name)
        except Exception:
            # We have a reference to the configuration,
            # but the low-level INI file no longer has the section,
            # since it was just removed.
            result = self._getStartupConfiguration(property_name)

        if result is None:
            result = default

        return result

    def _getEventData(self, avatar=None):
        """
        Return default data for all events emitted by this runnable.
        """
        return {
            'name': self.name,
            'avatar': avatar,
            'family': self.family,
            'kind': self._kind,
            'type': self.type,
            }

    def start(self, avatar=None):
        """
        See: `IRunnable`.
        """
        if self.started:
            return defer.succeed(False)

        # FIXME:3428:
        # Emit an event at the beginning of the start process.

        def cb_start(result):
            if not self.started:
                # Start failed.
                return False

            self.__reset()
            self._state = STATE.STARTED
            self._start_time = time()
            self._startup_configuration = self._configuration.getProperty()
            self.emitEvent(
                '20156',
                avatar=avatar,
                data={
                    'name': self.name,
                    'family': self.family,
                    'kind': self._kind,
                    'type': self.type,
                    'description': self.short_description,
                    },
                )
            self.notify('start', Signal(self))
            return True

        def eb_start(failure):
            """
            Call as final step in start errback chain.
            """
            deferred = self.__terminate(
                avatar=avatar,
                error=failure.value,
                success_state=STATE.FAILED_AT_START,
                failure_state=STATE.FAILED_AT_START,
                )
            # Even if stop succeed, we still return False to signal
            # that start has failed.
            deferred.addCallback(lambda _: False)
            return deferred

        # FIXME:3312:
        # Don't start if in the process stopping or failing.
        self._state = STATE.STARTING
        deferred = defer.maybeDeferred(self._onStart, avatar)
        deferred.addCallback(cb_start)
        deferred.addErrback(
            self.__ebConvertUnknownError, '20158', '20184')
        deferred.addErrback(self.__ebEmitFailure, avatar)
        deferred.addErrback(eb_start)

        def cb_after_start(result):
            """
            Called when _afterStart step should be executed.
            """
            if result is not True:
                # Start was not successful.
                return result

            # Only call when start was successful.
            deferred = defer.maybeDeferred(self._afterStart, avatar)
            # Ignore the result from after start and pass the previous result.
            deferred.addCallback(lambda _: result)
            return deferred

        def eb_after_start(failure):
            """
            Called when after start has failed.
            """
            deferred = self.fail(error=failure.value, avatar=avatar)
            deferred.addCallback(lambda _: False)
            return deferred

        deferred.addCallback(cb_after_start)
        deferred.addErrback(eb_after_start)

        return deferred

    def stop(self, avatar=None):
        """
        See `IRunnable`.
        """
        if not self.started or self.state == STATE.STOPPING:
            return defer.succeed(False)

        self._state = STATE.STOPPING
        return self.__terminate(
            avatar=avatar,
            error=None,
            success_state=STATE.STOPPED,
            failure_state=STATE.FAILED_AT_STOP,
            )

    def fail(self, error, avatar=None):
        """
        See: `IRunnable`.
        """
        return self._fail(
            error=error,
            avatar=avatar,
            emit_event=True,
            state=STATE.FAILED_AT_RUN
            )

    def restart(self, avatar=None):
        """
        See `IRunnable`.
        """
        if not self.started:
            return self.start(avatar=avatar)

        deferred = self.stop(avatar=avatar)
        deferred.addCallback(lambda _: self.start(avatar=avatar))
        return deferred

    def _fail(self, error, avatar, emit_event, state=STATE.FAILED_AT_RUN):
        """
        When `emit_event` is False, an event dedicated to the error is not
        emitted. This can be used when you want to emit the error in
        another place.
        """
        # Since we fail, we set the state as soon as possible so that
        # runnable will not process new requests while we still work to
        # stop it.
        self._state = STATE.FAILING
        # Failure might be requested with a non-ServeError, so
        # we convert it here into the generic ServerError.
        error = self.__convertUnknownError(error, '20165')

        if emit_event:
            self.__emitFailure(Failure(error), avatar=avatar)

        return self.__terminate(
            avatar=avatar,
            error=error,
            success_state=state,
            failure_state=state,
            )

    def __terminate(self, avatar, error, success_state, failure_state):
        """
        Low level implementation to stop the runnable. Shared both by stop
        and fail.

        `error` can be None in which case this is a normal stop call and not
        a call from the failure.

        `success_state` is the final state set after stop was successful.
        `failure_state` is the final state set after stop has failed.

        Stop runnable and update `last_error`.
        """
        def cb_stop(result):
            self._setLastError(error)
            self.__stopReached(state=success_state, avatar=avatar)
            return True

        def eb_stop(failure):
            """
            Called when failed to stop.
            """
            # If this is a stop during a failure, we ignore the stop error
            # and forward the original failure.
            if error is None:
                last_error = failure.value
            else:
                last_error = error
            self._setLastError(last_error)
            self.__stopReached(state=failure_state, avatar=avatar)
            return False

        deferred = defer.maybeDeferred(self._onStop, avatar)
        deferred.addCallback(cb_stop)
        deferred.addErrback(
            self.__ebConvertUnknownError, '20159', '20185')
        deferred.addErrback(self.__ebEmitFailure, avatar)
        deferred.addErrback(eb_stop)
        return deferred

    def emitEvent(self, event_id, data=None, avatar=None, peer=None):
        """
        Helper to prevent circular imports.
        """
        # FIXME:3231:
        # This should delegate to the root process.
        # Prevent circular imports.
        from chevah.server import _emit

        if data is None:
            data = {}
        data = data.copy()
        # FIXME:2635:
        # Here we replace the avatar from data, with the avatar passed
        # from the argument.
        # so 2635 is already fixed for runnables,
        # But we should move this inside _emit().
        data.update(self._getEventData())

        _emit(
            event_id=event_id,
            creator=self,
            avatar=avatar,
            data=data,
            peer=peer,
            )

    def _onStart(self, avatar=None):
        """
        Do the specific start.
        """
        raise NotImplementedError('_onStart not implemented')

    def _onStop(self, avatar=None):
        """
        Do the specific stop.

        Called on runnable should stop. Either on normal stop or on failure.
        """
        raise NotImplementedError('_onStop not implemented')

    def _afterStart(self, avatar=None):
        """
        Called after successful start.
        """
        pass

    def _afterStop(self, avatar=None):
        """
        Called after runnable was stopped.
        """
        pass

    def __ebConvertUnknownError(
            self,
            failure,
            known_failure_event_id,
            critical_failure_event_id):
        """
        Adapter from a failure to the exception based error converter.
        """
        traceback = force_unicode(failure.getBriefTraceback())

        if failure.check(ServerException):
            # Not a critical error.
            event_id = known_failure_event_id
        else:
            # It is an unknown error, so we treat it as
            # a critical-failure event.
            event_id = critical_failure_event_id

        raise self.__convertUnknownError(failure.value, event_id, tb=traceback)

    def __convertUnknownError(self, error, event_id, tb=None):
        """
        Convert an unknown error/exception into a ServerError.

        `traceback` is optional and help with debugging. For now is a simple
        string but in the future we might want to expand it to auto-get the
        current traceback.
        """
        if isinstance(error, ServerError):
            return error

        data = self._getEventData()

        if isinstance(error, defer.CancelledError):
            data['details'] = 'Operation canceled.'
        else:
            data['details'] = force_unicode(error)

        if not tb:
            tb = force_unicode(traceback.format_exc())
        data['tb'] = tb

        return ServerError(event_id, data=data)

    def __emitFailure(self, failure, avatar):
        """
        Called when a failure was received and we want to emit an event for it.

        Failure needs to be a ServerError.
        """
        # Prevent circular import.
        from chevah.server import emitFromError

        # FIXME:3231:
        # Update this to a more visible method and way to update error data.
        # Emit failure event before stop event.
        failure.value.data['name'] = self.name
        emitFromError(failure.value, avatar=avatar, emitter=self.emitEvent)

    def __ebEmitFailure(self, failure, avatar):
        """
        Errback wrapper for _emitFailure.
        """
        self.__emitFailure(failure, avatar)
        return failure

    def __stopReached(self, state, avatar):
        """
        Helper called when stop state was reached.
        """
        self._stop_time = time()
        self._state = state
        self.emitEvent(
            '20157',
            avatar=avatar,
            data={
                'name': self.name,
                'family': self.family,
                'kind': self._kind,
                'type': self.type,
                'reason': state,
                },
            )
        self.notify('stop', Signal(self))
        self._startup_configuration = None
        self._afterStop(avatar=avatar)

    def _cbActionSucceeded(self, result):
        """
        Called when the main runnable operation succeeded.
        """
        self._failure_count = 0
        return result

    def _ebActionFailed(self, failure):
        """
        Called when the main runnable operation failed.
        """
        # Do not trigger automatic failure of the runnable when maximum
        # failures set to 0.
        if self._max_failures == 0:
            return failure

        if self._failure_count >= self._max_failures:
            # Max failures reached by a previous actions.
            # We reach this when many actions were queues and they fail.
            return failure

        self._failure_count += 1
        if self._failure_count < self._max_failures:
            return failure

        self.emitEvent('20066', data=self._getEventData())
        # Since the failure is forwarded we don't emit it here.
        deferred = self._fail(
            error=failure.value,
            avatar=None,
            emit_event=False
            )
        # Wait for runnable to fail and then propagate the initial
        # failure.
        deferred.addBoth(lambda _: failure)
        return deferred


class DeprecatedRunnable(Runnable):
    """
    Placeholder a runnable which should no longer be used.
    """
    PROPERTIES_REQUIRING_RESTART = ()

    def _onStart(self, avatar=None):
        """
        See: `Runnable`.

        It always fails.
        """
        raise ServerException(
            'Component of type "%s" is no longer supported. '
            'See the upgrade documentation.' % (
                self.configuration.type,))

    def _onStop(self, avatar=None):
        """
        See: `Runnable`.

        Does nothing.
        """


@implementer(IRunnablesManager)
class RunnablesManager(PropertyMixin):
    """
    Manager of runnables of same type.
    """

    def __init__(self, parent):
        self._parent = parent
        self._runnables = {}
        self._state = STATE.STOPPED
        self._initializeRunnables()

    @property
    def state(self):
        return self._state

    def _initializeRunnables(self):
        """
        Called at initialization to setup runnables from this manager.
        """
        raise NotImplementedError('_initializeRunnables not implemented')

    def _addRunnableFromConfiguration(self, configuration):
        """
        Create a runnable from `configuration` and add it to be managed.
        """
        runnable = self._createRunnable(configuration)
        self._runnables[configuration.uuid] = runnable

    def _createRunnable(self, configuration):
        """
        Create a new runnable from `configuration`.
        """
        raise NotImplementedError('_createRunnable not implemented')

    def getPublicSectionNames(self):
        """
        See `IProperty`.
        """
        return list(self._runnables.keys())

    def getSection(self, name):
        """
        See `IProperty`.
        """
        return self._runnables.get(name, None)

    def getStarted(self):
        """
        See: `IRunnablesManager`.
        """
        result = []
        for runnable in self.getAll():
            if runnable.started:
                result.append(runnable)
        return result

    def start(self):
        """
        See: IRunnablesManager.
        """
        self._state = STATE.STARTING
        deferreds = []
        for runnable in self.getAll():
            if not runnable.enabled:
                continue
            deferreds.append(runnable.start())

        deferred = defer.DeferredList(
            deferreds, fireOnOneErrback=True, consumeErrors=True)

        def cb_start(result):
            self._state = STATE.STARTED
            return result

        def eb_start(failure):
            self._state = STATE.FAILED_AT_START
            return failure

        deferred.addCallbacks(cb_start, eb_start)
        return deferred

    def stop(self):
        """
        See: IRunnablesManager.
        """
        self._state = STATE.STOPPING
        deferreds = []
        for runnable in self.getStarted():
            deferred = runnable.stop()
            deferreds.append(deferred)

        deferred = defer.DeferredList(
            deferreds,
            fireOnOneErrback=True,
            consumeErrors=True,
            )

        def cb_stop(result):
            self._state = STATE.STOPPED
            return result

        def eb_stop(failure):
            self._state = STATE.FAILED_AT_STOP
            return failure

        deferred.addCallbacks(cb_stop, eb_stop)
        return deferred

    def restart(self):
        """
        See: IRunnablesManager.
        """
        deferreds = []
        for runnable in self.getStarted():
            deferreds.append(runnable.restart())

        deferred = defer.DeferredList(
            deferreds, fireOnOneErrback=True, consumeErrors=True)
        return deferred


class RunnablesManagerWithConfigurationSet(RunnablesManager):
    """
    A manager which keeps runnables in sync with a configuration set.
    """

    def _initializeRunnables(self):
        """
        See RunnablesManager.
        """
        self._runnables_configuration = self._getRunnablesConfiguration()
        self._runnables_configuration.subscribe(
            'delete', self._onRunnableDelete)
        self._runnables_configuration.subscribe(
            'create', self._onRunnableCreate)
        self._synchronizeRunnables()

    def _synchronizeRunnables(self):
        """
        Update status object with latest runnables.
        """
        configured_runnables = (
            self._runnables_configuration.getPublicSectionNames())

        # Check for newly added runnables.
        for runnable_uuid in configured_runnables:
            if self.getSection(runnable_uuid):
                # Already exists.
                continue

            configuration = self._runnables_configuration.getSection(
                runnable_uuid)
            self._addRunnableFromConfiguration(configuration)

        # Check for removed runnables.
        for runnable_uuid in self.getPublicSectionNames():
            if runnable_uuid in configured_runnables:
                continue
            self._removeRunnable(runnable_uuid)

    def _getRunnablesConfiguration(self):
        """
        Return the configuration set for managed runnables.

        Return None if runnable manager should not take care of keeping
        runnables in sync with the configuration or there is no configuration
        for runnables.
        """
        raise NotImplementedError('_getRunnablesConfiguration not implemented')

    def _onRunnableCreate(self, signal):
        """
        Called when a runnable is created in the configuration.
        """
        self._synchronizeRunnables()

    def _removeRunnable(self, uuid):
        """
        Remove runnable from manager cache.
        """
        self._runnables.pop(uuid, None)

    def _onRunnableDelete(self, signal):
        """
        Called when a runnable is removed from the configuration.
        """
        # We get the runnable before deleting it so that we can stop it.
        runnable = self.getSection(signal.uuid)
        # In case we don't have a runnable, it means that it was
        # deleted before reaching the manager.
        if not runnable:
            return defer.succeed(None)

        self._synchronizeRunnables()
        return runnable.stop()


class RunnablesManagerWithConfigurationType(
        RunnablesManagerWithConfigurationSet):
    """
    A manager which keeps runnables in sync with a configuration set and
    create runnables based on configuration type.

    Type is defined using __type_interface__.
    """

    def _createRunnable(self, configuration):
        """
        Create a new runnable from `configuration`.
        """
        type_class = queryUtility(
            self.__type_interface__, name=configuration.type, default=None)

        # We should not hit this in production since the configuration
        # type should have already been validated by the configuration
        # runnable. We don't need to raise a custom error again.
        # If there no location class, it means that when developing a new
        # runnable, the developer has forget to register an utility for it.
        if not type_class:
            raise RuntimeError(
                'No utility registered for runnable with type %s' % (
                    configuration.type))

        return type_class(parent=self, configuration=configuration)
