# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
Services Management.
"""
from __future__ import absolute_import
from __future__ import unicode_literals
import errno
import socket

from twisted.internet import defer, reactor
from twisted.internet.error import CannotListenError
from twisted.internet.endpoints import TCP4ServerEndpoint
from zope.interface import implementer

from chevah_compat import process_capabilities, system_users
from chevah_compat.exceptions import ChangeUserException
from chevah.server import force_unicode
from chevah.server.commons.exception import ServerException
from chevah.server.commons.interface import (
    IServiceManager,
    IServicesManager,
    )
from chevah.server.commons.constant import (
    STATE,
    TYPE_NAME,
    )
from chevah.server.commons.ssl import ChevahTLSMemoryBIOFactory
from chevah.server.commons.runnable import (
    Runnable,
    RunnablesManagerWithConfigurationSet,
    )
from chevah.server.ftp.factory import FTPFactory, FTPSIFactory
from chevah.server.ssh.factory import SSHServerFactory
from chevah.server.http.factory import (
    HTTPFactory,
    HTTPRedirectFactory,
    HTTPSFactory,
    )
from chevah.server.manager.factory import (
    LocalManagerHTTPFactory,
    LocalManagerHTTPSFactory,
    )

SERVICE_FACTORY = {
    TYPE_NAME.FTP: FTPFactory,
    TYPE_NAME.FTPSI: FTPSIFactory,
    TYPE_NAME.HTTP: HTTPFactory,
    TYPE_NAME.HTTP_REDIRECT: HTTPRedirectFactory,
    TYPE_NAME.HTTPS: HTTPSFactory,
    TYPE_NAME.SSH: SSHServerFactory,
    TYPE_NAME.LOCAL_MANAGER: LocalManagerHTTPSFactory,
    TYPE_NAME.LOCAL_MANAGER_UNSECURED: LocalManagerHTTPFactory,
    }


@implementer(IServicesManager)
class ServicesManager(RunnablesManagerWithConfigurationSet):
    """
    Aggregation of status for all active service.
    """

    def _getRunnablesConfiguration(self):
        """
        See: `RunnablesManagerWithConfigurationSet`.
        """
        return self.parent.configuration.services

    def _createRunnable(self, configuration):
        """
        See: `RunnablesManagerWithConfigurationSet`.
        """
        return ServiceManager(parent=self, configuration=configuration)


@implementer(IServiceManager)
class ServiceManager(Runnable):
    """
    Representation of the service as a runnable.
    """

    def __init__(self, parent, configuration):
        super(ServiceManager, self).__init__(parent, configuration)
        self._port = None

    def __repr__(self):
        result = "%d:%s: %s %s" % (
            id(self),
            self.__class__,
            self.name,
            self.state,
            )
        return result

    @property
    def short_description(self):
        return 'Listening on %s:%s.' % (
            self._configuration.address, self._configuration.port)

    def _onStart(self, avatar=None):
        """
        Start the factory for `service_configuration`.
        """
        service_configuration = self._configuration
        factory = None
        try:
            factory_class = SERVICE_FACTORY[service_configuration.type]
        except KeyError:
            # We should not reach this point as a service can not be
            # initialized with an unknown protocol.
            raise RuntimeError(
                'There is no factory for protocol '
                '"%s" of service "%s"' % (
                    service_configuration.type,
                    service_configuration.name))

        factory = factory_class(
            service_configuration=service_configuration.configuration,
            on_error=self.fail,
            )

        # Make sure we have the same clock between the manager and the factory.
        factory._scheduler = self._scheduler
        factory._creator = self

        def cb_set_port(result):
            self._port = result
            if self.state != STATE.STARTING:
                return self._onStop(avatar=avatar)

        def eb_bind_error(failure):
            failure.trap(CannotListenError)
            raise ServerException(force_unicode(failure.value))

        endpoint, factory = self._getEndpoint(service_configuration, factory)
        deferred = listen(endpoint, factory, service_configuration.port)
        deferred.addCallback(cb_set_port)
        deferred.addErrback(eb_bind_error)
        return deferred

    def _getEndpoint(self, service_configuration, factory):
        """
        Return the endpoint for `service` configuration.
        """
        if service_configuration.use_ssl:
            # FIXME:3245:
            # We force the factory start so that the SSL context is created.
            # We might want to have a special endpoint which will get the
            # context only after startFactory is called.
            factory.startFactory()
            factory = ChevahTLSMemoryBIOFactory(
                contextFactory=factory.ssl_context_factory,
                isClient=False,
                wrappedFactory=factory,
                )
            factory.https_redirect = service_configuration.https_redirect

        endpoint = TCP4ServerEndpoint(
            reactor=reactor,
            interface=service_configuration.address,
            port=service_configuration.port,
            )
        return endpoint, factory

    def _onStop(self, avatar=None):
        """
        See: Runnable
        """
        if not self._port:
            # Nothing to stop yet.
            return

        port, self._port = self._port, None
        return port.stopListening()

    @property
    def PROPERTIES_REQUIRING_RESTART(self):
        """
        See: `IServiceManager`.
        """
        # FIXME:2599:
        # This should be obtained from the runnable itself and not from the
        # configuration.
        return self._configuration.PROPERTIES_REQUIRING_RESTART

    def _getStartupConfiguration(self, property_name):
        """
        Return startup configuration for value specified as `property_name`.

        Ex: port of configuration/port.
        """
        if property_name.startswith('configuration/'):
            member_name = property_name.split('/')[1]
            return self._startup_configuration['configuration'][member_name]
        else:
            return self._startup_configuration[property_name]


def listen(endpoint, factory, port):
    """
    Low level code to listen on a port.

    Return a deferred which fires when listening is done.
    """
    def eb_listen_normal(failure):
        """
        Called when when we failed to listen as normal user.
        """
        if process_capabilities.os_family != 'posix':
            # On non-Unix/Linux we can listed on any port, without extra
            # permission.
            return failure

        if (port > 1023 or port == 0):
            # On Unix/Linux anyone can listen on ports above 1024.
            # Port 0 is used for testing to listen on a random free port.
            return failure

        if not failure.check(CannotListenError):
            # Not a listening error.
            return failure

        error = failure.value.socketError
        if not (
                isinstance(error, socket.error) and
                error.errno == errno.EACCES):
            # Not a permission error.
            return failure

        deferred = defer.maybeDeferred(listen_root)
        deferred.addErrback(eb_convert_ChangeUserException)
        return deferred

    def listen_root():
        """
        Try to listen as root.
        """
        with system_users.executeAsUser(username='root'):
            deferred = endpoint.listen(factory)
            if not deferred.called:
                return defer.fail(RuntimeError(
                    'Internal Server Error. '
                    'Failed to start the service as privileged account.'))
            return deferred

    def eb_convert_ChangeUserException(failure):
        """
        Called when we fail to listen as root.
        """
        # https://github.com/chevah/compat/issues/615
        # system_users.executeAsUser should raise a CompatError
        # instead of ChangeUserException.
        failure.trap(ChangeUserException)
        raise ServerException(
            'Failed to listen on port below 1024. '
            'Missing privileges.'
            )

    # First try to listen as current user.
    # and then see how to handle the errors.
    deferred = endpoint.listen(factory)
    deferred.addErrback(eb_listen_normal)
    return deferred
