# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
Generic configurations for a service.
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from zope.interface import implementer

from chevah.server.commons.constant import (
    CONFIGURATION_SECTION_SERVICES,
    SSL_PROTOCOLS,
    TYPE_NAME,
    )
from chevah.server.commons.interface import (
    IServiceConfiguration,
    IConnectionOptions,
    )
from chevah.server.configuration.model import (
    PropertyMixin,
    WritableBoolean,
    WritableInteger,
    WritableNumber,
    WritableCSVUnique,
    )
from chevah.server.configuration.option import (
    AddressPortOptionsMixin,
    EntityConfiguration,
    TypeOptionsMixin,
    )
from chevah.server.configuration.set import SetWithTypeConfiguration
import six


# FIXME:2599:
# Once configuration is flattened, this should be removed and use the
# standard SetWithTypeConfiguration for services.
class ServicesConfigurationSection(SetWithTypeConfiguration):
    """
    Group of all services available for the server.
    """

    SECTION_PREFIX = CONFIGURATION_SECTION_SERVICES
    __type_interface__ = IServiceConfiguration

    def onChildCreate(self, section_name, value, uuid):
        """
        Add sub-configurations.
        """
        result = super(ServicesConfigurationSection, self).onChildCreate(
            section_name, value, uuid)

        configuration = value.get('configuration', {})
        service = self.getSection(uuid)
        for key_name, key_value in configuration.items():
            service.setPropertyWithoutNotification(
                'configuration/' + key_name, key_value)

        # Delete `configuration` as we have already set it.
        result.pop('configuration', None)

        return result

    def _getTypeName(self, uuid):
        """
        Return a string containing type name for section with `uuid`.

        Return None if there is no ection with `uuid`.
        """
        section = '%s%s' % (self.SECTION_PREFIX, uuid)

        if not self._proxy.hasSection(section):
            return None

        # FIXME:2599:
        # For now, we migrate this at runtime.
        # This should be moved to be done at startup time.
        self._migrateType(section)

        return super(ServicesConfigurationSection, self)._getTypeName(uuid)

    def _migrateType(self, section):
        """
        Do service specific `protocol` option to generic `type` option.

        Also do any other protocol related migration required.
        """
        if (
            not self._proxy.definesOption(section, 'type')
            and self._proxy.definesOption(section, 'protocol')
                ):
            value = self._proxy.getString(section, 'protocol')
            self._proxy.setString(section, 'type', value)

        # Remove legacy protocol option as it was converted to `type`.
        if self._proxy.definesOption(section, 'protocol'):
            self._proxy.removeOption(section, 'protocol')

        # Convert old `sftp` type name to new `ssh` type name.
        value = self._proxy.getString(section, 'type')
        if value.lower() == TYPE_NAME.SFTP:
            self._proxy.setString(section, 'type', TYPE_NAME.SSH)

    # FIXME:2599:
    # This should be removed once configuration is flat.
    def setPropertyFromSync(self, value):
        """
        See `ISetConfiguration`.
        """
        # Do the sync for the base.
        result = super(ServicesConfigurationSection, self).setPropertyFromSync(
            value)

        service_configuration = value.get('configuration', None)

        if not service_configuration:
            # No service specific configuration was requested.
            return result

        # Do the sync for the embedded configuration.
        for name in result.configuration.getSyncAttributeNames():
            if name.endswith('password_update'):
                # FIXME:2229:
                # See how to set other passwords or secrets and in a generic
                # way.
                # For now, other secrets are not synced.
                continue

            if name not in service_configuration:
                # Configuration attribute not requested for sync.
                continue

            # FIXME:3480:
            # Emit event on configuration changes.
            result.configuration.setValue(
                name, service_configuration[name])

        return result


@implementer(IServiceConfiguration)
class ServiceConfigurationSection(
        EntityConfiguration, AddressPortOptionsMixin, TypeOptionsMixin):
    """
    Configurations for a service section.

    This represent a network service that is running on the server.
    Each service has an enable status together with address and port.

    It also has a `configuration` member with protocol specific
    configurations.
    """

    authentications = WritableCSVUnique('authentications')
    debug = WritableBoolean('debug', default=False)

    def __init__(self, parent, uuid):
        self._parent = parent
        self._section_name = CONFIGURATION_SECTION_SERVICES + uuid
        self.configuration = self._loadServiceConfiguration()

    def __repr__(self):
        result = (
            'Service configuration for section "%s - %s", '
            'type "%s" - %s:%d' % (
                self.uuid, self.name, self.type, self.address, self.port))
        return result

    @property
    def parent(self):
        """
        See: `IProperty`.
        """
        return self._parent

    @property
    def _proxy(self):
        return self._parent._proxy

    def _loadServiceConfiguration(self):
        """
        Load service configuration.
        """

        # Since this is the main configuration class, it should not import
        # from children.
        # The fix is to use a registry, but it is too complicated for now.
        from chevah.server.manager.configuration import (
            LocalManagerServiceConfigurationSection,
            )
        from chevah.server.ftp.configuration import (
            FTPServiceConfigurationSection,
            FTPSIServiceConfiguration,
            )
        from chevah.server.http.configuration import (
            HTTPServiceConfigurationSection,
            HTTPRedirectServiceConfigurationSection,
            HTTPSServiceConfigurationSection,
            )
        from chevah.server.ssh.configuration import (
            SSHServiceConfigurationSection
            )

        SERVICE_CONFIGURATION = {
            TYPE_NAME.FTP: (FTPServiceConfigurationSection, 10021),
            TYPE_NAME.FTPSI: (FTPSIServiceConfiguration, 10990),
            TYPE_NAME.HTTP: (HTTPServiceConfigurationSection, 18080),
            TYPE_NAME.HTTP_REDIRECT: (
                HTTPRedirectServiceConfigurationSection, 80),
            TYPE_NAME.HTTPS: (HTTPSServiceConfigurationSection, 10443),
            TYPE_NAME.LOCAL_MANAGER: (
                LocalManagerServiceConfigurationSection, 10020),
            TYPE_NAME.LOCAL_MANAGER_UNSECURED: (
                LocalManagerServiceConfigurationSection, 10019),
            TYPE_NAME.SSH: (SSHServiceConfigurationSection, 10022),
            }

        klass, default_port = SERVICE_CONFIGURATION.get(self.type)

        self.PORT_DEFAULT = default_port
        return klass(parent=self, section_name=self._section_name)

    @property
    def uuid(self):
        '''
        UUID of the service.

        This is an unique string use internally for referencing this service.

        The characters are restricted to alpha-numeric values.

        uuid can not be changed.
        '''
        value = six.text_type(
            self._section_name[len(CONFIGURATION_SECTION_SERVICES):])
        return value

    @property
    def use_ssl(self):
        """
        See `IServiceConfiguration`.
        """
        if self.type in SSL_PROTOCOLS:
            return True
        else:
            return False

    @property
    def https_redirect(self):
        """
        See `IServiceConfiguration`.
        """
        return self.type in [
            TYPE_NAME.HTTPS, TYPE_NAME.LOCAL_MANAGER]

    @property
    def PROPERTIES_REQUIRING_RESTART(self):
        """
        See `IServiceConfiguration`.
        """
        result = ['port', 'address', 'type']
        service_specific = self.configuration.PROPERTIES_REQUIRING_RESTART
        for name in service_specific:
            result.append('configuration/%s' % (name,))
        return result


@implementer(IConnectionOptions)
class ConfigurationWithConnections(PropertyMixin):
    """
    Mixing class for configurations related to connections.
    """

    def __init__(self, parent, section_name):
        self._parent = parent
        self._section_name = section_name

    @property
    def _proxy(self):
        return self._parent._proxy

    idle_connection_timeout = WritableNumber(
        'idle_connection_timeout',
        default=300,
        allow_negative=False,
        allow_zero=False,
        )
    maximum_concurrent_connections = WritableInteger(
        'maximum_concurrent_connections', default=10000)
