# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
Common mixins used by Chevah products.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from twisted.internet.error import AlreadyCalled, AlreadyCancelled
from twisted.protocols.policies import TimeoutMixin

from chevah.server import emit
from chevah.server.identity.avatar import ProcessAvatar


class RemoteConnectionMixin(object):
    """
    Remote connection session information mixin.
    """
    #: This is here to help with testing.
    _emit = emit

    def emitEvent(self, event_id, data=None, avatar=None, creator=None):
        """
        Emits an event with the specified id, avatar and data attributes.
        """
        if self.avatar is None:
            # This should not usually happen as we should have a PeerAvatar
            # once the connection is made.
            avatar = ProcessAvatar()
        else:
            avatar = self.avatar

        kwargs = {
            'avatar': avatar,
            'data': data,
            }
        self._emit(event_id, **kwargs)


def emit_account_login_paths(
        id, avatar, emitter, path, real_path=None, data=None):
    """
    Emit the event informing the path configuration for a successfully
    authenticated avatar.
    """
    if data is None:
        data = {}

    if real_path is None:
        real_path = avatar.home_folder_path

    data.update({
        'real_path': real_path,
        'virtual_path': path,
        })
    emitter(id, avatar=avatar, data=data)


def patched_resetTimeout(self):
    """
    Reset the timeout count down.

    This is here to handle our threaded reactor tests.

    If the connection has already timed out, then do nothing.  If the
    timeout has been cancelled (probably using C{setTimeout(None)}), also
    do nothing.

    It's often a good idea to call this when the protocol has received
    some meaningful input from the other end of the connection.  "I've got
    some data, they're still there, reset the timeout".
    """
    try:
        original_resetTimeout(self)
    except (AlreadyCalled, AlreadyCancelled):  # noqa:cover
        # Timeout already canceled.
        # This only happens when using the threaded reactor.
        return


# We need to call original methods as patched versions don't have access to
# the hidden members.
original_resetTimeout = TimeoutMixin.resetTimeout
TimeoutMixin.resetTimeout = patched_resetTimeout
