# Copyright (c) 2010 Adi Roiban.
# See LICENSE for details.
'''SSL commons support for services.'''
from __future__ import absolute_import
from __future__ import unicode_literals
from binascii import hexlify
from datetime import datetime, timedelta
from six.moves.urllib.parse import urlparse
import os
import random
import time

from cryptography.hazmat.primitives import serialization
from OpenSSL import SSL, crypto
from service_identity import CertificateError, VerificationError
from service_identity.hazmat import DNS_ID, verify_service_identity
from service_identity.pyopenssl import extract_patterns
from twisted.internet import defer, endpoints, reactor, task
from twisted.internet.error import AlreadyCalled, AlreadyCancelled
from twisted.internet.interfaces import ISSLTransport, IHandshakeListener
from twisted.internet.main import CONNECTION_LOST
from twisted.internet.protocol import Protocol
from twisted.internet.ssl import (
    ClientContextFactory,
    DefaultOpenSSLContextFactory,
    )
from twisted.protocols.policies import ProtocolWrapper
from twisted.protocols.tls import TLSMemoryBIOFactory, TLSMemoryBIOProtocol
from twisted.python.failure import Failure
from twisted.python.randbytes import secureRandom
from zope.interface import (
    noLongerProvides,
    implementer,
    )
import pem


from chevah_compat import local_filesystem, process_capabilities
from chevah.server import (
    emit,
    emit_error_details,
    force_unicode,
    MODULE_PATH,
    native_string,
    )
from chevah.server.helpers import add_timeout
from chevah.server.commons import x509, UTC
from chevah.server.commons.constant import (
    CA_PLACEHOLDER,
    CONFIGURATION_DISABLED_VALUES,
    CRL_PLACEHOLDER,
    SSL_CONNECTION_CLOSE_TIMEOUT,
    SSL_DEFAULTS,
    SSL_CIPHERS,
    STATE,
    TYPE_NAME,
    _PEMS_PLACEHOLDER,
    )
from chevah.server.commons.exception import ServerError, ServerException
from chevah.server.commons.interface import (
    ISSLOptions,
    ISSLContextFactory,
    ISSLServiceFactory,
    )
from chevah.server.commons.session import RemoteConnectionMixin
from chevah.server.configuration.model import (
    PropertyMixin,
    WritableInteger,
    WritableString,
    WritableStringOrEmpty,
    WritableCSVUnique,
    WritableCSVUniqueInsensitive,
    _convert_ordered_sequence_to_non_duplicate_string,
    _convert_string_to_non_duplicate_tuple,
    )
from chevah.server.event.model import ChevahTimeStamp
from chevah.server.proxy.proxy_protocol import proxify_v2
import six


# Some constants that might not be exported by pyOpenSSL.
# They should have the same name as OpenSSL C code to make it easier
# to search.
OP_NO_TLSv1_3 = 0x20000000
SSL_OP_NO_RENEGOTIATION = 0x40000000
SSL_OP_CIPHER_SERVER_PREFERENCE = 0x400000

X509_V_FLAG_PARTIAL_CHAIN = 0x80000

# Number of seconds after which to retry loading the CRL.
_CRL_SOFT_RELOAD_DELAY = 4 * 60 * 60


def patch_shutdownTLS(self):
    """
    Initiate, or reply to, the shutdown handshake of the TLS layer.
    """
    def try_shutdown():
        if not self._tlsConnection:
            self.transport.loseConnection()
            return

        if self._lostTLSConnection:
            # Connection was already lost.
            # Nothing more to do.
            self.transport.loseConnection()
            return

        if self._SSL_CONNECTION_CLOSE_TIMEOUT <= 0:
            self.transport.loseConnection()
            return

        shutdown_start = getattr(self, '__shutdown_start', None)
        if shutdown_start is None:
            self.__shutdown_start = time.time()

        waited = time.time() - self.__shutdown_start
        if waited > self._SSL_CONNECTION_CLOSE_TIMEOUT:
            emit('10021', peer=self.getPeer())
            self.transport.loseConnection()
            return

        state = self._tlsConnection.get_state_string()
        if state in [b'error']:
            self.transport.loseConnection()
            return

        if b'negotiation finished successfully' not in state:
            # Negotiation not yet ready.
            self._try_shutdown_call = reactor.callLater(0.01, try_shutdown)
            return

        try:
            success = self._tlsConnection.shutdown()
        except SSL.WantReadError:
            if not getattr(self.wrappedProtocol, '_force_close', False):
                self._flushReceiveBIO()
                self._try_shutdown_call = reactor.callLater(0.01, try_shutdown)
                return
            # If we force the connection, we fake the fact that TLS shutdown
            # is finalized.
            success = True
        except SSL.SysCallError as error:
            if error.args == (-1, "Unexpected EOF"):
                # Connection was already closed.
                success = True
            else:
                self.transport.loseConnection()
                raise

        self._flushSendBIO()

        if not success:
            self._try_shutdown_call = reactor.callLater(0.01, try_shutdown)
            return

        if self._loseConnectionOnTLSShutdown:
            # This is here to not lose the connection if we just shut down
            # TLS, for example for FTP CCC command.
            self.transport.loseConnection()

    try_shutdown()


def patch_tlsShutdownFinished(self, reason):
    """
    Called when TLS connection has gone away; tell underlying transport to
    disconnect.

    This is patching the method to signal when tls was stopped and to prevent
    closing the connection.
    """
    if reason is not None:
        # Squash an EOF in violation of the TLS protocol into
        # ConnectionLost, so that applications which might run over
        # multiple protocols can recognize its type.
        if tuple(reason.value.args[:2]) == (-1, 'Unexpected EOF'):
            reason = Failure(CONNECTION_LOST)

        if getattr(self.wrappedProtocol, '_force_close', False):
            reason = None

    if self._reason is None:
        self._reason = reason
    self._lostTLSConnection = True
    deferred, self._tlsShutdownDeferred = self._tlsShutdownDeferred, None
    if deferred:
        deferred.callback(None)

    if not self._loseConnectionOnTLSShutdown:
        return

    # Before closing the connection,
    # make sure we send anything left in the buffer.
    start_time = time.time()
    while self._flushSendBIO():
        # There is still data on the buffer.
        # Wait and try to send more later.
        time.sleep(0.01)
        if time.time() - start_time > 5:
            # Give up if we still have data after 5 seconds.
            break

    # Using loseConnection causes the application protocol's
    # connectionLost method to be invoked non-reentrantly, which is always
    # a nice feature. However, for error cases (reason != None) we might
    # want to use abortConnection when it becomes available. The
    # loseConnection call is basically tested by test_handshakeFailure.
    # At least one side will need to do it or the test never finishes.
    self.transport.loseConnection()


def _getEndProtocol(wrappedProtocol):
    """
    Try to get the final wrapped protocol.
    """
    if isinstance(wrappedProtocol, ProtocolWrapper):
        return _getEndProtocol(wrappedProtocol.wrappedProtocol)
    return wrappedProtocol


def patch_checkHandshakeStatus(self):
    """
    Ask OpenSSL to proceed with a handshake in progress.

    Initially, this just sends the ClientHello; after some bytes have been
    stuffed in to the C{Connection} object by C{dataReceived}, it will then
    respond to any C{Certificate} or C{KeyExchange} messages.

    This is patching to add support for nested wrapped protocols.
    """
    # The connection might already be aborted (eg. by a callback during
    # connection setup), so don't even bother trying to handshake in that
    # case.
    if self._aborted:
        return
    try:
        self._tlsConnection.do_handshake()
    except SSL.WantReadError:
        self._flushSendBIO()
    except SSL.Error:
        fail = Failure()
        self._tlsShutdownFinished(fail)
    else:
        self._handshakeDone = True
        end_protocol = _getEndProtocol(self.wrappedProtocol)
        if IHandshakeListener.providedBy(end_protocol):
            end_protocol.handshakeCompleted()


# Here to allow patching the value to speed up the test.
_SSL_CONNECTION_CLOSE_TIMEOUT = SSL_CONNECTION_CLOSE_TIMEOUT
TLSMemoryBIOProtocol._SSL_CONNECTION_CLOSE_TIMEOUT = (
    SSL_CONNECTION_CLOSE_TIMEOUT)
TLSMemoryBIOProtocol._try_shutdown_call = None
TLSMemoryBIOProtocol._tlsShutdownDeferred = None
TLSMemoryBIOProtocol._loseConnectionOnTLSShutdown = True
TLSMemoryBIOProtocol._shutdownTLS = patch_shutdownTLS
TLSMemoryBIOProtocol._tlsShutdownFinished = patch_tlsShutdownFinished
TLSMemoryBIOProtocol._checkHandshakeStatus = patch_checkHandshakeStatus


def stop_tls(transport):
    """
    Reverts the changes done by C{startTLS} without closing the
    connection and waiting for the remote peer to shutdown the connection.
    """
    # If we have a producer, unregister it, and then re-register it below
    # once we've switched off the TLS mode, so it gets hooked up correctly:
    producer, streaming = None, None
    if transport.producer is not None:
        producer, streaming = transport.producer, transport.streamingProducer
        transport.unregisterProducer()

    deferred = defer.Deferred()
    transport.protocol._tlsShutdownDeferred = deferred
    transport.protocol._loseConnectionOnTLSShutdown = False
    transport.protocol._tlsConnection.shutdown()
    transport.protocol._flushSendBIO()

    transport.protocol = transport._clean_protocol

    del transport.getHandle
    del transport.getPeerCertificate

    # Mark the transport as insecure.
    noLongerProvides(transport, ISSLTransport)
    transport.TLS = False

    # Restore producer if necessary:
    if producer:
        transport.registerProducer(producer, streaming)
    return deferred


# Marker for the state in which the CRL was not yet loaded.
_CRL_NOT_YET_LOADED = object()


@implementer(ISSLContextFactory)
class OpenSSLContextFactory(DefaultOpenSSLContextFactory):
    """
    OpenSSL context factory for a client or server side to be used in
    Twisted related code.

    Errors are signaled using on_error callback. After on_error is called the
    context factory should be considered stopped as it auto-stops.
    """

    # For now, the state is private only to help with testing,
    # It starts as suspended so that we can recognize when it was stopped.
    _state = STATE.SUSPENDED

    # For more information about the format for this options see OpenSSL
    # cipher(1).
    # http://www.openssl.org/docs/apps/ciphers.html#CIPHER_LIST_FORMAT
    # ATTENTION: keep it in sync with ssl_cipher_list configuration
    # documentation
    _SECURE_CIPHERS = (
        'HIGH:!PSK:!RSP:!eNULL:!aNULL'
        ':!RC4:!MD5:!DES:!3DES:!aDH:!kDH:!DSS'
        )

    def __init__(
        self,
        on_error,
        event_emitter,
        certificate=None,
        key=None,
        ca=None,
        crl_value=[],
        crl_refresh=86400,
        cipher_list='',
        allowed_methods='',
        key_password=None,
        remote_names=None,
        scheduler=None,
            ):
        # We set up these helpers as soon as possible so that they can be used
        # in init.
        self._on_error_callback = on_error
        self._emitEvent = event_emitter

        self.key_password = key_password
        self.certificate = certificate
        self.key = key
        self.ca = ca

        self.crl_value = crl_value
        # Handle legacy "Disabled" value.
        for value in crl_value:
            if value.lower() in CONFIGURATION_DISABLED_VALUES:
                self.crl_value = []
                break

        # Always use default values, even for an explicit empty value.
        if not cipher_list:
            cipher_list = SSL_DEFAULTS['ssl_cipher_list']
        if not allowed_methods:
            allowed_methods = SSL_DEFAULTS['ssl_allowed_methods']

        if 'secure@seclevel=0' in cipher_list.lower():
            if (
                SSL.OPENSSL_VERSION_NUMBER > 0x10101000
                and (six.PY3 or process_capabilities.os_family != 'nt')
                    ):
                # Only use SECLEVEL where is supported.
                cipher_list = self._SECURE_CIPHERS + '@SECLEVEL=0'
            else:
                cipher_list = self._SECURE_CIPHERS
        elif 'secure' in cipher_list.lower():
            cipher_list = self._SECURE_CIPHERS

        self.cipher_list = cipher_list

        self.allowed_methods = allowed_methods
        if 'sslv3' in self.allowed_methods:
            self._emitEvent(
                '20139', data={'configuration': self.allowed_methods})

        # Name of the remote server which should match the presented
        # certificate.
        # In plan text and in objects used by the validation library.
        self._remote_names = remote_names
        if not self._remote_names:
            self._remote_names = []
        self._remote_ids = []

        # Deferred associated with adding the CRL.
        # It is called when the CRL was added.
        self._crl_add_deferred = defer.Deferred()

        # pyOpenSSL certificate for this context local-side.
        self._certificate = None

        # The pinned public keys accepted for remote connections.
        self._pinned_keys = []

        self._emitEvent = event_emitter

        self._crl_validator = CRLValidator(
            scheduler=scheduler,
            update_interval=crl_refresh,
            event_emitter=event_emitter,
            )
        try:
            self._makeContext()
        except SSL.Error as ssl_error:
            raise ServerError('20032', data={
                'cert': certificate,
                'key': key,
                'ca': self.ca,
                'crl': self.crl_value,
                'details': force_unicode(ssl_error),
                })

    def start(self):
        """
        See: ISSLContextFactory.
        """
        self._state = STATE.STARTING

        def cb_started(result):
            self._state = STATE.STARTED
            return self

        # Trigger the possible loading of the CRL.
        # We save the deferred, as during loading it can fail and this will
        # remove the deferred.
        deferred = self._crl_add_deferred

        deferred.addCallback(cb_started)
        deferred.callback(None)
        return deferred

    def stop(self):
        """
        See: ISSLContextFactory.
        """
        self._state = STATE.STOPPED
        self._crl_validator.stop()

        self._crl_add_deferred = None

    def _onError(self, error):
        """
        Called when something went wrong.
        """
        self._on_error_callback(error)
        self.stop()

    def _onValidatePeerCertificate(
        self, connection, certificate, error_code, verify_depth, preverify_ok,
            ):
        """
        SSL Context callback to check the peer's certificate chain.

        The certificate chain is checked starting with the deepest nesting
        level (the root CA certificate) and worked upward to the peer's
        certificate.

        At each level signatures and issuer attributes are checked and
        `preverify_ok` is `1` when the certificate is valid.

        If this method returns 0/False, the verification process is immediately
        stopped with "verification failed" state.
        """
        message = _get_message_for_verify(error_code)
        try:
            if verify_depth != 0:
                # Not at the end level of the peer's certificate.
                if preverify_ok:
                    # Intermediate cert pre-validation is ok.
                    # We don't check the CRL for intermediate certificates.
                    return True
                else:
                    # Intermediate cert pre-validation failed.
                    raise ServerException(
                        'Intermediate certificate not accepted. %s - %s.' % (
                            error_code, message))

            if certificate.has_expired():
                raise ServerException('Expired certificate.')

            if not preverify_ok:
                # Initial context validation has failed and we don't know
                # the reason.
                raise ServerException('Certificate not accepted. %s - %s.' % (
                    error_code, message))

            if self.crl_value:
                # Only validate if context is a using CRL validation.
                self._crl_validator.validate(peer_certificate=certificate)

            _validate_hostname(
                certificate, self._remote_ids, self._remote_names)

        except ServerException as error:
            cert_subject = x509.get_subject(
                certificate.get_subject().get_components(),
                'certificate during TLS peer validation',
                fail=False,
                )

            data = {
                'subject': cert_subject,
                'details': error.message,
                }

            # FIXME:2902:
            # For client-side connections that are refused due to CA
            # show the configured CA vs remote CA.
            self._emitEvent('20040', data=data)
            return False

        return True

    def _onValidatePinnedKey(
        self, connection, certificate, error_code, verify_depth, preverify_ok,
            ):
        """
        SSL Context callback to check the peer's certificate chain.

        If this method returns 0/False, the verification process is immediately
        stopped with "verification failed" state.
        """
        if verify_depth > 0:
            # We ignore any intermediary or root certificate.
            return True

        for pinned_key in self._pinned_keys:
            if pinned_key == certificate.get_pubkey().to_cryptography_key():
                # We have a pinned public key.
                # All ok
                return True

        cert_subject = x509.get_subject(
            certificate.get_subject().get_components(),
            'ssl_certificate_authority',
            fail=False,
            )
        data = {
            'subject': cert_subject,
            'details': (
                'Certificate not found in the list of pinned public keys.'
                ),
            }
        self._emitEvent('20040', data=data)
        return False

    def _makeContext(self):
        # We start by allowing all SSL/TLS methods, including SSLv2 and then
        # disable them using SSL_OP_NO_*
        context = SSL.Context(SSL.SSLv23_METHOD)

        # We go with the default OpenSSL options.
        # See: http://www.openssl.org/docs/ssl/SSL_CTX_set_options.html
        context.set_options(SSL.OP_ALL)
        # FIXME:6432:
        # For now, we don't do caching on client-side as this breaks
        # FTPS session resumption on TLS 1.3
        context.set_session_cache_mode(SSL.SESS_CACHE_SERVER)

        if not self._remote_names:
            # For server-side connections, we disable renegotiation,
            # to prevent Dos attacks.
            # Renegotiation is not supported in TLS 1.3 and is replaced
            # by SSL_key_update.
            # For client-side, renegotiation is OK as we trust the server
            # will issue renegotiation in good faith.
            context.set_options(SSL_OP_NO_RENEGOTIATION)
            context.set_options(SSL_OP_CIPHER_SERVER_PREFERENCE)
            dh_param_path = os.path.join(
                MODULE_PATH, 'static', 'crypto', 'dhparam.pem')
            try:
                context.load_tmp_dh(dh_param_path)
            except SSL.Error as error:  # noqa:cover
                # This should never happen as the file is part of our
                # source code.
                raise ServerException(
                    'Failed to load DH parameters files. %s' % (
                        force_unicode(error,)))

        # For both TLS 1.2 and TLS 1.3 we want to use the
        # stateful session ID.
        # We don't want to use the stateless tickets,
        # as pyOpenSSL has no API for handling them.
        context.set_options(SSL.OP_NO_TICKET)

        try:
            context.set_cipher_list(str(self.cipher_list))
        except SSL.Error as error:
            if not error.args[0]:
                # On Ubuntu with OpenSSL 1.1.1 We don't get any error here.
                error.args = (
                    'Invalid "ssl_cipher_list" configuration: no cipher match',
                    )
            raise error

        # Certificates should be loaded after we set the context options
        # and ciphers.
        self._setCertificate(context)
        self._setCA(context)
        self._setCRL(context)

        blocked_methods = self._convertToBlockedMethods(self.allowed_methods)
        for no_method in blocked_methods:
            context.set_options(no_method)

        # This was copied from upstream twisted.internet._sslverify:
        # https://github.com/twisted/twisted/blob/7c84d1d92ca895059ed6c4251cac62e5badd36ef/src/twisted/internet/_sslverify.py#L1667
        # This is only used for server-side.
        sessionIDContext = hexlify(secureRandom(7))
        context.set_session_id(sessionIDContext)

        self._context = context

    def _setCertificate(self, context):
        """
        Configure the context with a local certificate and its key and
        ca chain.
        """
        if not self.certificate:
            return

        # They all contain pyOpenSSL objects.
        certificate = None
        key = None
        chain = []

        # Start with PKCS12 format.
        cryptography_certificate, cryptography_key, cryptography_chain = (
            load_PKCS12(path=self.certificate, password=self.key_password))
        if cryptography_certificate:
            certificate = crypto.X509.from_cryptography(
                cryptography_certificate)
            key = crypto.PKey.from_cryptography_key(cryptography_key)
            chain = []
            for ca_cert in cryptography_chain:
                chain.append(crypto.X509.from_cryptography(ca_cert))

        if key and self.key:
            raise ServerException(
                'Configured PKCS12 "ssl_certificate" contains a private key '
                'which conflicts with the "ssl_key" configuration.'
                )

        if not certificate:
            # Try PEM format.
            certificate, key, chain = read_PEM_configuration(
                self.certificate, self.key, password=self.key_password)

        if not certificate:
            raise ServerException(
                'Configured "ssl_certificate" does not '
                'contain a certificate.'
                )

        if not key:
            raise ServerException(
                'Missing private key. '
                'Add it to either "ssl_key" or "ssl_certificate".'
                )

        context.use_certificate(certificate)
        self._certificate = certificate

        for member in chain:
            context.add_extra_chain_cert(member)

        try:
            context.use_privatekey(key)
        except SSL.Error as error:
            # Maybe it is not associated with the certificate.
            raise ServerException(
                'Failed to use the private key. %s' % (force_unicode(error),))

    def _setCA(self, context):
        """
        Configure the certificate authorities.
        """
        if (
            not self.ca
            or self.ca.lower() == TYPE_NAME.DISABLE_IDENTITY_SECURITY
                ):
            # CA check is requested to be disable.
            # * for server-side we have an empty value for backward
            # compatibility.
            # * for client-side we have an explicit disable value.
            return

        configuration = self.ca.strip()
        ca_expansion = CA_PLACEHOLDER.get(configuration, None)
        if ca_expansion:
            # We have a CA placeholder so we resolve it to the full path.
            configuration = os.path.join(
                MODULE_PATH, 'static', 'ca', ca_expansion)

        ca_store = context.get_cert_store()
        # See flags here: https://www.openssl.org/docs/manmaster/man3/
        #   X509_VERIFY_PARAM_set_flags.html#VERIFICATION-FLAGS
        # With partial chain we can accept self issued cert and also
        # pin a certificate.
        ca_store.set_flags(X509_V_FLAG_PARTIAL_CHAIN)

        pem_objects = _load_PEM(configuration, 'ssl_certificate_authority')
        ca_cert = None
        for ca_cert in _get_certificates(pem_objects, ca_expansion):
            ca_store.add_cert(ca_cert)

        pinned_keys = _get_public_keys(pem_objects)

        if not ca_cert and not pinned_keys:
            raise ServerException(
                'Configured "ssl_certificate_authority" '
                'has no valid certificates or pinned public keys.'
                )

        if ca_cert:
            # Check that we have a DNS, as we can't validate IP.
            # Parsing the name now will raise and error at start and not
            # at runtime.
            self._remote_ids = []
            try:
                for name in self._remote_names:
                    if not name:
                        raise ServerException(
                            'Missing value for one of the remote names.')
                    self._remote_ids.append(DNS_ID(name))

                # Keep the raw names for error messages.
                self._remote_names = ', '.join(self._remote_names)
            except ValueError:
                raise ServerException(
                    'Remote peer validation is supported only for FQDN: '
                    '{}.'.format(', '.join(self._remote_names)))
            except TypeError:
                raise ServerException('Invalid values for remote names.')

            context.set_verify(
                SSL.VERIFY_PEER | SSL.VERIFY_FAIL_IF_NO_PEER_CERT |
                SSL.VERIFY_CLIENT_ONCE,
                self._onValidatePeerCertificate,
                )

        if pinned_keys:
            if ca_cert:
                raise ServerException(
                    'Configured "ssl_certificate_authority" '
                    'does not support mixed certificates and pinned public '
                    'keys.'
                    )
            context.set_verify(
                SSL.VERIFY_PEER | SSL.VERIFY_FAIL_IF_NO_PEER_CERT |
                SSL.VERIFY_CLIENT_ONCE,
                self._onValidatePinnedKey,
                )
            self._pinned_keys = pinned_keys

        self.ca = configuration

    def _setCRL(self, context):
        """
        Configure the CRL.
        """
        if not self.ca or not self.crl_value:
            # No CA or CRL is configured don't set up the validator.
            # This is useless, but should be explicit that CRL should not be
            # used.
            self._crl_validator.stop()
            return

        # See if we need to resolve placeholders.
        expanded_crl = []
        for crl_value in self.crl_value:
            # Try to see if we can expand it, and if not, default to
            # current value
            crl_value = CRL_PLACEHOLDER.get(crl_value.strip(), [crl_value])
            expanded_crl.extend(crl_value)

        self.crl_value = expanded_crl

        # Set up the async loading of CRL
        self._crl_add_deferred.addCallback(
            lambda _: self._crl_validator.configure(
                ca=self.ca,
                crls=self.crl_value,
                certificate=self._certificate,
                ))

        def cb_crl_add(result):
            self._crl_validator.start()
            return True

        def eb_crl_add(failure):
            # FIXME:5166:
            # Here we trigger 2 errors. On for start and another for the
            # onError callback.
            # See how to refactor this so that we don't trigger 2 errors.
            self._onError(failure.value)
            raise ServerException('Failed to load the CRL.')

        self._crl_add_deferred.addCallback(cb_crl_add)
        self._crl_add_deferred.addErrback(eb_crl_add)

    def _convertToBlockedMethods(self, allowed_methods):
        """
        Convert the list of allowed methods to a list of SSL options
        to be applied on the SSL_CONTEXT as OP_NO_METHOD.

        This is a bit strange, since we configure the list of allowed methods,
        while OpenSSL is configured by listing the blocked SSL methods.

        It will start with a list that will block all methods and then
        remove from the list the allowed methods.
        """
        ssl_options = [
            SSL.OP_NO_SSLv2,
            SSL.OP_NO_SSLv3,
            SSL.OP_NO_TLSv1,
            SSL.OP_NO_TLSv1_1,
            SSL.OP_NO_TLSv1_2,
            OP_NO_TLSv1_3,
            ]

        for method in allowed_methods:
            if method == 'sslv3':
                ssl_options.remove(SSL.OP_NO_SSLv3)
            elif method in ['tlsv1', 'tlsv1.0']:
                ssl_options.remove(SSL.OP_NO_TLSv1)
            elif method == 'tlsv1.1':
                ssl_options.remove(SSL.OP_NO_TLSv1_1)
            elif method == 'tlsv1.2':
                ssl_options.remove(SSL.OP_NO_TLSv1_2)
            elif method == 'tlsv1.3':
                ssl_options.remove(OP_NO_TLSv1_3)
            elif method == 'all':
                # We don't disable anything.
                return []
            elif method == 'secure':
                # Disable what are considered insecure.
                return [
                    SSL.OP_NO_SSLv2,
                    SSL.OP_NO_SSLv3,
                    SSL.OP_NO_TLSv1,
                    SSL.OP_NO_TLSv1_1,
                    ]
            else:
                raise ServerException(
                    'Unknown SSL/TLS method "%s".' % (method,))
        return ssl_options


def _get_certificates(pem_objects, ca_expansion):
    """
    Return a list with pyOpenSSL certificates extracted from PEM instances.
    """
    result = []
    for member in pem_objects:
        if not isinstance(member, pem.Certificate):
            # We ignore any PEM objects which are not certificates.
            continue

        try:
            ca_cert = crypto.load_certificate(
                crypto.FILETYPE_PEM, member.as_bytes())
        except crypto.Error as error:
            raise ServerException(
                'Failed to parse the chain in ssl_certificate_authority. '
                '%s' % (error,))

        if ca_cert.has_expired():
            # The current configured CA certificate is expired.
            if ca_expansion:
                # But this is configured from our placeholders,
                # so we will ignore it so that older version of SFTPPlus
                # will still work.
                continue

            cert_subject = x509.get_subject(
                ca_cert.get_subject().get_components(),
                'ssl_certificate_authority',
                fail=False,
                )
            raise ServerException(
                'Certificate expired on "{}" configured for '
                '"ssl_certificate_authority": {}.'.format(
                    force_unicode(ca_cert.get_notAfter()),
                    cert_subject))

        # CA cert has passed all validations.
        result.append(ca_cert)

    return result


def _get_public_keys(pem_objects):
    """
    Return a list with cryptography public keys extracted from PEM instances.
    """
    result = []
    for member in pem_objects:
        if not isinstance(member, pem.PublicKey):
            # We ignore any PEM objects which are not certificates.
            continue

        try:
            public_key = crypto.load_publickey(
                crypto.FILETYPE_PEM, member.as_bytes())
        except crypto.Error as error:
            raise ServerException(
                'Failed to parse the public key in ssl_certificate_authority. '
                '%s' % (error,))
        result.append(public_key.to_cryptography_key())
    return result


class ChevahTLSMemoryBIOProtocol(TLSMemoryBIOProtocol, object):
    """
    TLSMemoryBIOProtocol used by Chevah services.

    The main diff from upstream is trying to close the TLS connection in a
    nice way:

    * wait for start handshake to finalized.
    * wait for shutdown handshake to finalize.
    """
    def loseConnection(self):
        """
        Monkey patching for TLSMemoryBIOProtocol to wait for handshake to end
        (and not abort it) before closing the connection.

        Send a TLS close alert and close the underlying connection.
        """
        if self.disconnecting or not self.connected:
            return

        # If connection setup has not finished, OpenSSL 1.0.2f+ will not shut
        # down the connection until we write some data to the connection which
        # allows the handshake to complete. However, since no data should be
        # written after loseConnection, this means we might be stuck forever
        # waiting for shutdown to complete.
        # But maybe we just need to wait a bit more for the peer to send its
        # handshake.
        self.disconnecting = True
        self._shutdownTLS()

    def connectionLost(self, reason):
        """
        Handle the possible repetition of calls to this method (due to either
        the underlying transport going away or due to an error at the TLS
        layer) and make sure the base implementation only gets invoked once.

        Patch TLSMemoryBIOProtocol to set _lostTLSConnection before flushing.
        """
        if self._try_shutdown_call:
            call, self._try_shutdown_call = self._try_shutdown_call, None
            try:
                call.cancel()
            except (  # noqa:cover
                AlreadyCancelled, AlreadyCalled, AttributeError
                    ):
                """
                Do nothing at this can happen with threaded reactor.
                """

        if not self._lostTLSConnection:
            # Tell the TLS connection that it's not going to get any more data
            # and give it a chance to finish reading.
            self._tlsConnection.bio_shutdown()
            self._lostTLSConnection = True
            self._flushReceiveBIO()

        reason = self._reason or reason
        self._reason = None
        ProtocolWrapper.connectionLost(self, reason)


def _redirectfy_https(target_protocol):
    """
    Wrap the protocol and if first received data is HTTP redirect with a
    Host header, redirect to HTTPS
    """

    protocol_receive = target_protocol.dataReceived

    def get_http_host(data):
        """
        Return the value of the Host header found inside `data` HTTP
        raw request.
        """
        parts = data.split(b'\r\n')
        if len(parts) == 1:
            return b''

        for part in parts:
            if not part.lower().startswith(b'host'):
                continue

            segments = part.split(b':', 1)
            if len(segments) != 2:
                return b''

            return segments[1].strip()

        return b''

    def first_ssl_receive(data):
        """
        Called when initial connection data is received over SSL.

        It checks if an HTTP request is made over HTTPS port, instead of an
        SSL handshake.
        """
        # We only want to handle the first package.
        target_protocol.dataReceived = protocol_receive

        host = get_http_host(data)

        if not host:
            # Just forward the data to the wrapped protocol.
            return protocol_receive(data)

        target_protocol.transport.write(
            b"HTTP/1.1 301 Moved Permanently\r\n"
            b"Location: https://%s\r\n"
            b"Cache-Control: no-store, no-cache\r\n"
            b"Connection: close\r\n"
            b"Content-Length: 8\r\n"
            b"\r\n"
            b"redirect" % (host,))
        target_protocol.transport.loseConnection()
        return b''

    target_protocol.dataReceived = first_ssl_receive


class ChevahTLSMemoryBIOFactory(TLSMemoryBIOFactory, object):
    '''Adapter for ChevahTLSMemoryProtocol to TLSMemoryBIOFactory.'''

    protocol = ChevahTLSMemoryBIOProtocol
    https_redirect = 0
    # Use to help with testing and get hold of a server-side connection.
    _last_protocol = None

    def buildProtocol(self, addr):
        protocol = self.protocol(self, self.wrappedFactory.buildProtocol(addr))
        self._last_protocol = protocol
        if self.https_redirect:
            _redirectfy_https(protocol)
        proxify_v2(protocol)
        return protocol


@implementer(ISSLOptions)
class SSLConfigurationSectionMixin(PropertyMixin):
    """
    Configuration options for services using SSL/TLS.
    """

    _ssl_prefix = 'ssl'

    PROPERTIES_REQUIRING_RESTART = (
        'ssl_domains',
        'ssl_certificate_value',  # Possible the global value.
        'ssl_key_value',  # Possible the global value.
        'ssl_certificate_authority',
        'ssl_certificate_revocation_list',
        'ssl_certificate_revocation_list_refresh',
        'ssl_cipher_list',
        'ssl_allowed_methods',
        )

    ssl_certificate = WritableStringOrEmpty('ssl_certificate')
    ssl_key = WritableStringOrEmpty('ssl_key')

    @property
    def ssl_certificate_value(self):
        """
        The local one, of the global one.
        """
        result = self.ssl_certificate
        if result:
            return result
        return self.root.configuration.server.ssl_certificate

    @property
    def ssl_key_value(self):
        """
        The local one, of the global one.
        """
        result = self.ssl_key

        if not result:
            result = ''

        if self.ssl_certificate:
            return result

        return self.root.configuration.server.ssl_key

    # Flag to indicate when the password was updated to help with knowing
    # when a components needs a restart.
    _ssl_key_password_updated = False

    @property
    def ssl_key_password(self):
        """
        Get ssl_key_password.
        """
        self._ssl_key_password_updated = False
        value = self._proxy.getStringOrNone(
            self._section_name, 'ssl_key_password')
        if value is None or value.strip() == '':
            return ''
        else:
            return value

    @ssl_key_password.setter
    def ssl_key_password(self, value):
        """
        Set password.
        """
        raise RuntimeError(
            'Use ssl_key_password_update for changing the password.'
            )

    @property
    def ssl_key_password_update(self):
        """
        This is a write-only attribute.
        """
        if self._ssl_key_password_updated:
            return 'restart-required'

        return ''

    @ssl_key_password_update.setter
    def ssl_key_password_update(self, value):
        """
        Set ssl_key_password as clear text value.
        """
        self._ssl_key_password_updated = True
        return self._proxy.setStringOrNone(
            self._section_name, 'ssl_key_password', value)

    ssl_domains = WritableCSVUniqueInsensitive('ssl_domains')
    ssl_certificate_authority = WritableStringOrEmpty(
        'ssl_certificate_authority')
    ssl_certificate_revocation_list = WritableCSVUnique(
        'ssl_certificate_revocation_list')
    ssl_certificate_revocation_list_refresh = WritableInteger(
        'ssl_certificate_revocation_list_refresh', default=0)
    ssl_cipher_list = WritableString(
        'ssl_cipher_list', default=SSL_DEFAULTS['ssl_cipher_list'])

    @property
    def ssl_allowed_methods(self):
        """
        Read the comma separated list with backward compatibility for
        space separated list.
        """
        value = self._proxy.getString(
            self._section_name, 'ssl_allowed_methods')

        if not value:
            # We have the default.
            return SSL_DEFAULTS['ssl_allowed_methods']

        value = value.strip().lower()

        if 'secure' in value:
            return ('secure',)

        if 'all' in value:
            return ('all',)

        # We might have the backward compatible space separation.
        # Do a brutal replacement of spaces with comma and try to parse
        # again. If it works it ain't stupid :)
        value = value.replace(' ', ',')
        return _convert_string_to_non_duplicate_tuple(value)

    @ssl_allowed_methods.setter
    def ssl_allowed_methods(self, value):
        """
        The value should be always set as a list.

        We should simplify the writing code.
        Any default handling should be done in the value read part.
        """
        value = _convert_ordered_sequence_to_non_duplicate_string(value)
        self._proxy.setString(
            self._section_name, 'ssl_allowed_methods', value.lower())


@implementer(IHandshakeListener)
class SSLConnectionMixin(RemoteConnectionMixin):
    """
    Shared code by protocols running over SSL/TLS.
    """

    # How long to wait before raising an error if the TLS handshake was not
    # finalized.
    _TLS_HANDSHAKE_TIMEOUT = 30
    # Called when TLS handshake is done or is aborted.
    _tls_handshake_completed = None
    # Event ID emitted when TLS handshake was done with success.
    _TLS_OK_EVENT = None
    # Event ID emitted when TLS handshake failed.
    _TLS_FAIL_EVENT = None

    def _getTLSConnection(self):
        """
        Return the SSL/TLS connection.

        Return `None` if connection is plain.
        """
        try:
            # On SSL/TLS self.transport is the SSL transport
            return self.transport._tlsConnection
        except AttributeError:
            # On STARTTLS we have a wrapper to allow both TLS and not TLS
            # transports.
            # Usually it is at self.transport.protocol, but when the connection
            # is close, self.transport.protocol is set to None and the
            # reference is lost.
            # self.transport.getPeerCertificate is not lost and we can use it
            # to recover.
            # See the mangling at twisted.internet._newtls.startTLS
            try:
                transport = self.transport.getPeerCertificate.__self__
                return transport._tlsConnection
            except AttributeError:
                return None

    def _getPeerCertificate(self):
        """
        Return the peer certificate if remote peer is on a SSL connection
        and has sent us a certificate.

        Return `None` otherwise.

        Calling this before the SSL handshake is done will always return None.
        """
        try:
            return self.transport.getPeerCertificate()
        except AttributeError:
            # Not a TLS transport.
            return None

    def _getPeerCertificateDescription(self, certificate=None):
        """
        Return the human readable representation of the peer certificate.

        This should never fail as it is used for audit purposes.
        """
        if not certificate:
            certificate = self._getPeerCertificate()
        if not certificate:
            return 'no certificate'

        cert_subject = x509.get_subject(
            certificate.get_subject().get_components(),
            'certificate during TLS handshake',
            fail=False,
            )

        return '(%x) %s' % (
            certificate.get_serial_number(),  cert_subject)

    def _getEncryptionDescription(self):
        """
        Return the description of the encryption used by the current
        connection.
        """
        connection = self._getTLSConnection()

        if connection is None:
            return 'no encryption'

        try:
            encryption = '%s %s' % (
                force_unicode(connection.get_protocol_version_name()),
                force_unicode(connection.get_cipher_name()),
                )
        except Exception:
            encryption = 'unknown encryption'

        return encryption

    def handshakeCompleted(self):
        """
        See: IHandshakeListener.
        """
        self._tls_handshake_completed.callback(None)

    def waitHandshakeCompleted(self):
        """
        Returns the deferred that is called when handshake is completed.
        """
        # Allow to receive extra data, as this should be the TLS
        # handshake.
        return self._tls_handshake_completed

    def beforeTLSHandshake(self, channel_type):
        """
        Called to prepare the state before a TLS handshake.
        """
        self._tls_handshake_completed = defer.Deferred()
        add_timeout(
            self._scheduler,
            self._tls_handshake_completed,
            self._TLS_HANDSHAKE_TIMEOUT,
            message='Stalled TLS handshake for {}.'.format(channel_type)
            )

        def cb_handshake_done(result):
            """
            Called when TLS handshake was done.
            """
            self.emitEvent(self._TLS_OK_EVENT, data={
                'channel_type': channel_type,
                'encryption': self._getEncryptionDescription(),
                'certificate': self._getPeerCertificateDescription(),
                })
            return result

        def eb_handshake(failure):
            # With OpenSSL 1.1.1, startTLS can fail right away.
            # Most of the time it fail when receiving the USER
            # command, that is usually sent by the client after AUTH.
            self.emitEvent(self._TLS_FAIL_EVENT, data={
                'details': force_unicode(failure.value),
                'channel_type': channel_type,
                })
            return failure

        self._tls_handshake_completed.addCallback(cb_handshake_done)
        self._tls_handshake_completed.addErrback(eb_handshake)

    def _tlsConnectionLost(self, failure):
        """
        Called when connection is lost.

        THis is here to inform if the TLS handshake was started but not
        completed.
        """
        if not self._tls_handshake_completed:
            return
        if self._tls_handshake_completed.called:
            return

        # Once the connection is lost, there shouldn't be any other
        # process waiting for the deferred, so it's safe to ignore the error,
        # that we are triggering here.
        # We push the error here to trigger emitting the events.
        self._tls_handshake_completed.addErrback(lambda f: None)
        self._tls_handshake_completed.errback(failure)


@implementer(ISSLServiceFactory)
class SSLServiceFactoryMixin(object):
    """
    Mixin for common code used by factories with support for SSL/TLS.

    It should be mixed with a Runnable.

    The factories should define the _onError callback which is called with
    an exception for SSL related errors.
    """

    ssl_enabled = True
    ssl_context_factory = None

    def _configureSSL(self):
        """
        See: `ISSLServiceFactory`

        Remove an existing context if ssl is not (longer) enabled.
        """

        if not self.ssl_enabled:
            # Don't initialize the SSL context if SSL is not enabled.
            self.ssl_context_factory = None
            return

        if self.ssl_context_factory is not None:
            raise RuntimeError(
                'SSL/TLS context factory was already initialized.')

        service_config = self.service_configuration

        if service_config.ssl_domains:
            name = service_config.ssl_domains[0]
            sans = list(service_config.ssl_domains[1:])
            # Since nobody waits for this method call,
            # Let's Encrypt will always return a certificate.
            # A self-signed one is returned while the real Let's Encrypt is
            # being generated.
            # We don't save this configuration changes are they are done
            # by the lets-encrypt service.
            pems = self._getLetEncryptService().getCertificatePEMs(name, sans)
            service_config.ssl_certificate = pems
            service_config.ssl_key = ''

        ssl_certificate = service_config.ssl_certificate_value
        ssl_key = service_config.ssl_key_value
        ssl_ca = service_config.ssl_certificate_authority
        crl_value = service_config.ssl_certificate_revocation_list

        if not ssl_certificate:
            raise ServerException(
                '"ssl_certificate" option must be specified.')

        if ssl_ca and ssl_ca.lower() == TYPE_NAME.SET_ON_FIRST_CONNECTION:
            raise ServerException(
                'Option "set-on-first-connection" is not supported for '
                'server-side "ssl_certificate_authority" configuration.'
                )

        if ssl_ca and ssl_ca.lower() == TYPE_NAME.PIN_PUBLIC_KEY:
            raise ServerException(
                'Option "pin-public-key" is not supported for '
                'server-side "ssl_certificate_authority" configuration.'
                )

        self.ssl_context_factory = OpenSSLContextFactory(
            certificate=ssl_certificate,
            key=ssl_key,
            key_password=service_config.ssl_key_password,
            ca=ssl_ca,
            crl_value=crl_value,
            crl_refresh=service_config.ssl_certificate_revocation_list_refresh,
            allowed_methods=service_config.ssl_allowed_methods,
            cipher_list=service_config.ssl_cipher_list,
            on_error=self._onError,
            event_emitter=self.emitEvent,
            scheduler=self._scheduler,
            )
        return self.ssl_context_factory.start()

    def _getLetEncryptService(self):
        """
        Return the first service of type Let's Encrypt.
        """
        return self.root.resources.getLetsEncryptClient()


class SSLClientMixin(object):
    """
    Mixin for common code used by client-side protocols using SSL/TLS.

    It should be mixed with a Runnable having a configuration based on
    `SSLConfigurationSectionMixin`.

    The code using it should define the _onError callback which is called with
    an exception for SSL related errors.
    """
    # The SSL context factory which is currently loaded.
    _context_factory = None

    def _onError(self, error):
        """
        Called when one of the sub-components got an error.
        """
        self.fail(error)

    @defer.inlineCallbacks
    def _initStartedContextFactory(self, remote_servers=None, urls=None):
        """
        Return a deferred which is called when the new SSL context factory
        which is ready to be used for Twisted connections.

        `remote_servers` is the list of (FQDN, port) tuple to which this client
        will accept connection, after their identity is validated.

        `urls` is a list of URLS from which the remote names are extracted.
        """
        if remote_servers and urls:
            raise RuntimeError('Need only remote_servers OR urls.')

        remote_servers = remote_servers
        if not remote_servers:
            remote_servers = []

        only_http = False
        if urls:
            only_http = True
            for url in urls:
                url = urlparse(url)
                hostname = url.hostname
                if not hostname:
                    raise ServerException(
                        'Configured URL has invalid format '
                        'or is missing the hostname.'
                        )
                if url.scheme.lower() != TYPE_NAME.HTTP:
                    only_http = False

                remote_servers.append((hostname, url.port))

        configuration = self.configuration
        if not configuration.ssl_certificate_authority and not only_http:
            # Default is an empty string, as a shared value between client and
            # server.
            # But for client, we do a run-time default
            # If we only have
            configuration.ssl_certificate_authority = (
                TYPE_NAME.SET_ON_FIRST_CONNECTION)

        yield self._setRemoteServerCAOnConnection(remote_servers)
        yield self._pinPublicKeyOnConnection(remote_servers)

        remote_names = [n[0] for n in remote_servers]
        context = OpenSSLContextFactory(
            certificate=configuration.ssl_certificate,
            key=configuration.ssl_key,
            key_password=configuration.ssl_key_password,
            ca=configuration.ssl_certificate_authority,
            crl_value=configuration.ssl_certificate_revocation_list,
            crl_refresh=(
                configuration.ssl_certificate_revocation_list_refresh),
            allowed_methods=configuration.ssl_allowed_methods,
            cipher_list=configuration.ssl_cipher_list,
            remote_names=remote_names,
            on_error=self._onError,
            event_emitter=self.emitEvent,
            scheduler=self._scheduler,
            )
        result = yield context.start()
        self._context_factory = result
        defer.returnValue(result)

    @defer.inlineCallbacks
    def _setRemoteServerCAOnConnection(self, remote_names):
        """
        See if we need to set the CA chain on first connection.
        """
        if (
            self.configuration.ssl_certificate_authority.lower() !=
            TYPE_NAME.SET_ON_FIRST_CONNECTION
                ):
            # No need to auto-set.
            return

        if len(remote_names) != 1:
            # Later, we might add support for multiple URLs.
            raise ServerException(
                'Value "set-on-first-connection" is only supported when '
                'configured to connect to a single remote server.'
                )

        address = remote_names[0][0]
        port = remote_names[0][1]
        certificates = yield get_remote_certificates(address, port)

        if len(certificates) < 2:
            raise ServerException(
                'Failed to get remote CA chain. '
                'Maybe self-signed certificate or server is not configured '
                'to send the CA chain.')

        result = []
        # The first certificate, is the server cert.
        # We only care about about intermediate and root CAs.
        now = ChevahTimeStamp(time.time())
        for certificate in certificates[1:]:
            # FIXME:1955:
            # Only accept CA chain.
            # If server presents other certificates,
            # raise an error.
            pem_cert = crypto.dump_certificate(
                crypto.FILETYPE_PEM, certificate).strip().decode('ascii')

            subject = x509.get_subject(
                certificate.get_subject().get_components(),
                error_location='set-on-first-connection',
                )
            config_content = (
                'Remote certificate\n'
                '{}\n'
                'auto-set on {} using set-on-first-connection option.\n\n'
                '{}'.format(subject, now.cwa_14051, pem_cert))

            result.append(config_content)

        pem_content = '\n'.join(result)

        self.configuration.ssl_certificate_authority = pem_content
        self.configuration._proxy.save()

    @defer.inlineCallbacks
    def _pinPublicKeyOnConnection(self, remote_names):
        """
        See if we need to pin the public key on first connection.
        """
        if (
            self.configuration.ssl_certificate_authority !=
            TYPE_NAME.PIN_PUBLIC_KEY
                ):
            # No need to auto-set.
            return

        if len(remote_names) != 1:
            raise ServerException(
                'Value "pin-public-key" is only supported when '
                'configured to connect to a single remote server.'
                )

        address = remote_names[0][0]
        port = remote_names[0][1]
        certificates = yield get_remote_certificates(address, port)

        # We only set the first certificate.
        server_certificate = certificates[0]
        pem_public_key = crypto.dump_publickey(
            crypto.FILETYPE_PEM, server_certificate.get_pubkey()
            ).strip().decode('ascii')

        now = ChevahTimeStamp(time.time())
        subject = x509.get_subject(
            server_certificate.get_subject().get_components(),
            error_location='pin-public-key',
            )
        config_content = (
            'Public key pinning\n'
            '{}\n'
            'auto-set on {} using pin-public-key option.\n\n'
            '{}'.format(subject, now.cwa_14051, pem_public_key))
        self.configuration.ssl_certificate_authority = config_content
        self.configuration._proxy.save()


class CRLValidator(object):
    """
    Helper for validating certificates against CRL.

    For now it only support a single CRL signed by a single CA.

    It provides the `CRLValidator.validate(certificate)` method which can be
    used as a callback for `SSLContext.set_verify`.
    """
    _random = random

    def __init__(self, scheduler, update_interval, event_emitter):
        if not scheduler:
            scheduler = reactor
        self._scheduler = scheduler
        self._emitEvent = event_emitter
        # Time in seconds after which the CRL is loaded.
        # 0 for auto-reload.
        self._update_interval = update_interval

        # Here to avoid circular import. HTTP client also uses SSL.
        from chevah.server.http.client import NonPersistentAgent
        # HTTP client used to get CRL over HTTP.
        self._http_agent = NonPersistentAgent(event_emitter=event_emitter)

        self._resetBeforeStartState()

    def _resetBeforeStartState(self):
        """
        Set the validator state as before starting.
        """

        # Deferred attached to the next update for the CRL location.
        self._next_update_deferreds = {}

        # Mapping from CA subject to CA cert used to validate CRL signature.
        self._ca_certs = {}

        self._running = False
        # Mapping between CA subject and list of revoked ids.
        self._revoked_serials = {}
        # Mapping between CRL location to issuer.
        self._crl_location_to_issuer = {}
        # True when validator should auto-load CRL from CDP.
        self._use_cdp = False

    def start(self):
        """
        Called after CRL validator was configured and can be used.
        """
        if not self._ca_certs:
            raise ServerException('CA was not set to validate the CRL.')

        if not self._revoked_serials:
            raise ServerException('No CRL was configured.')

        self._running = True

    def stop(self):
        """
        Called when CRL validator should no longer be used.
        """
        for location, deferred in self._next_update_deferreds.items():
            if not deferred.called:
                try:
                    deferred.cancel()
                except AlreadyCancelled:  # noqa:cover
                    """
                    Already cancelled.
                    This can happen in the threaded reactor.
                    """

        self._resetBeforeStartState()

    @property
    def running(self):
        """
        `True` when the validator should be used.
        """
        return self._running

    @defer.inlineCallbacks
    def configure(self, ca, crls, certificate):
        """
        Configure based on CA PEM file stored at `ca` path and load CRLs
        from `crl` list.

        Return a deferred which is fired when the CRL is configured.
        """
        self._addCAFromPath(path=ca)

        if TYPE_NAME.CRL_DISTRIBUTION_POINTS in crls:
            deferred = self._configureCDP(start_certificate=certificate)
        else:
            deferred = self._addCRLs(crls=crls)

        # Wait for initial CRL cache to be loaded.
        yield deferred

    def _addCAFromPath(self, path):
        """
        Add CA certificates stored as PEM file on local filesystem to
        validate the CRL.
        """
        if self._ca_certs:
            # For now we don't support multiple calls to this method since
            # we don't need it.
            # This can be change if we somehow need it.
            raise ServerException(
                'Only a single CA file can be loaded for CRL validation.')

        ca_segments = local_filesystem.getSegmentsFromRealPath(path)

        if local_filesystem.isFolder(ca_segments):
            raise ServerException(
                'CRL can only be used when CA certificates are stored '
                'in a single file.'
                )

        ca_file = None
        try:
            ca_file = local_filesystem.openFileForReading(ca_segments)
            pem_objects = pem.parse(ca_file.read())
        finally:
            if ca_file:
                ca_file.close()

        ca_certs = {}
        for pem_object in pem_objects:
            if not isinstance(pem_object, pem.Certificate):
                continue
            cert = crypto.load_certificate(
                crypto.FILETYPE_PEM, pem_object.as_bytes())

            subject = x509.get_subject(
                cert.get_subject().get_components(),
                'CA loaded from %s' % (path,),
                )
            ca_certs[subject] = cert

        if not ca_certs:
            raise ServerException(
                'CA file used to validated the CRL has no certificates.')

        self._ca_certs.update(ca_certs)

    @defer.inlineCallbacks
    def _addCRLs(self, crls):
        """
        Add a list of CRLs to validate against.

        Returns a deferred which is fired when the CRLs were loaded,
        successfully or not.

        On failure, only the error from the first CRL is raised.

        If you want to catch errors from all CRLs call this separately with
        a CLR at a time.
        """
        for crl in crls:
            try:
                location = 'NOT-YET-DEFINED'
                get_method, location = self._getCRLMethod(
                    crl_configuration=crl)

                issuer = self._crl_location_to_issuer.get(location, None)
                if issuer:
                    # Location is already loaded. Nothing new to do.
                    continue
                yield self._updateCRL(get_method, location)

            except Exception as error:
                raise self._getCRLLoadErrorWithInvalidation(error, location)

    def _configureCDP(self, start_certificate):
        """
        Configure the validator to use CRL advertised by CDP.

        FIXME:3583:
        Allow enabling CDP without a start certificate.

        An initial CRL is extracted from the CDP of the `start_certificate`
        to cache the most probable CDP in use.

        Returns a deferred which is fired when the initial CRL was loaded or
        is errbacked on errors.
        """
        if not start_certificate:
            # FIXME:4382:
            # This should be removed once we have CDP for the client side.
            error = self._getCRLLoadErrorWithInvalidation(
                error=ServerException(
                    'CDP cannot be used without a certificate.'
                    ),
                location='crl-distribution-points',
                )
            return defer.fail(error)

        self._use_cdp = True
        crls = x509.get_CRLs_from_CDP_extension(certificate=start_certificate)
        if not crls:
            error = self._getCRLLoadErrorWithInvalidation(
                error=ServerException(
                    'No CDP extension defined for the local certificate.'),
                location='crl-distribution-points',
                )
            return defer.fail(error)

        # For now we only use the first CRL and the others are ignored.
        return self._addCRLs(crls=crls[:1])

    def validate(self, peer_certificate):
        """
        Check if certificate is valid and not part of a CRL.

        Raise ServerException with details if certificate is not valid.
        """
        if not self._running:
            raise ServerException('CRL validation was not yet started.')

        issuer = x509.get_subject(
            peer_certificate.get_issuer().get_components(),
            'peer certificate'
            )
        revoked_is = self._revoked_serials.get(issuer, _CRL_NOT_YET_LOADED)

        if self._use_cdp:
            # When CDP is used, clients should always use a certificate which
            # is advertising a CDP.
            crls = x509.get_CRLs_from_CDP_extension(peer_certificate)
            if not crls:
                raise ServerException(
                    'CDP is in usage and peer certificate '
                    'has no CDP extension.'
                    )

        if revoked_is is _CRL_NOT_YET_LOADED:

            if not self._use_cdp:
                # CDP are not used and CRL for this issuer is not yet
                # loaded.
                raise ServerException(
                    'CRL not yet loaded for %s.' % (issuer,))

            # We are using CDP, and we don't have a CRL yet.
            # So load the CRL if is not already loaded.
            load_message = self._loadCDP(
                crls=crls, peer_certificate=peer_certificate)
            raise ServerException(load_message)

        # We have a revocation list for this issuer.
        if peer_certificate.get_serial_number() in revoked_is:
            raise ServerException('Revoked certificate.')

    def _loadCDP(self, crls, peer_certificate):
        """
        Load a CRL at runtime, as advertised in the CDP extension.

        Return a human readable message describing the action which was
        taken.
        """
        # For now, we only use the first CRL and the others are ignored.
        crls = crls[:1]
        if (
            crls[0] in self._next_update_deferreds and
            not self._next_update_deferreds[crls[0]].called
                ):
            # CRL is in the processes of being updated so until it is
            # loaded, we fail.
            return 'CDP/CRL for the peer certificate is not loaded yet.'

        # Make sure we have a subject for the peer certificate.
        peer_subject = x509.get_subject(
            peer_certificate.get_subject().get_components(),
            'peer_certificate during CDP/CRL',
            )

        # No CRL yet for this peer certificate, so we trigger their loading.
        deferred = self._addCRLs(crls=crls)
        deferred.addErrback(lambda failure: emit_error_details(
            error=failure.value,
            event_id='20178',
            data={'peer_subject': peer_subject},
            emitter=self._emitEvent,
            ))
        self._cdp_auto_load_deferred = deferred

        # We fail this validation as we don't have yet a CRL.
        return 'CDP/CRL for the peer certificate is being loaded.'

    def _now(self):
        """
        Return the current datetime in UTC.
        """
        result = datetime.utcfromtimestamp(self._scheduler.seconds())
        return result.replace(tzinfo=UTC)

    def _getCRLLoadErrorWithInvalidation(self, error, location):
        """
        Return a ServerError associated with CRL load operation and invalidate
        the currently loaded CRL.
        """
        self._invalidateCRL(location)

        data = {
            'uri': location,
            'details': force_unicode(error),
            }
        return ServerError('20036', data=data)

    def _invalidateCRL(self, location):
        """
        Invalidate the CRL which is already loaded from location.
        """
        try:
            issuer = self._crl_location_to_issuer[location]
            self._revoked_serials[issuer] = _CRL_NOT_YET_LOADED
        except KeyError:
            # It might be that we were not able to load the location so we
            # don't know its issuer.
            pass

    def _getCRLMethod(self, crl_configuration):
        """
        Return a tuple of
        (get_method, location) for the single `crl_configuration`.
        """
        get_method = None
        location = crl_configuration

        if not location:
            return defer.fail(ServerException(
                'CRL location can not be empty.'))

        url = urlparse(crl_configuration)
        if crl_configuration[1:3] == ':\\':
            # Windows absolute path without a scheme.
            get_method = self._loadCRLFromPath
        elif url.scheme == '':
            # No scheme is fallback to local path.
            get_method = self._loadCRLFromPath
        elif url.scheme == 'file':
            get_method = self._loadCRLFromPath
            if url.netloc:
                # we have a relative path or a Windows path and the final
                # location is url.netloc + url.path
                if not url.path:
                    # We have a windows path or a relative path comprised of
                    # a single segment.
                    # FIXME:3480:
                    # Might need to store it as absolute path.
                    location = url.netloc
                else:
                    # We have Unix relative path.
                    # FIXME:3480:
                    # Might need to store it as absolute path.
                    location = url.netloc + url.path
            else:
                # We have absolute unix path.
                location = url.path
        elif url.scheme == 'http':
            get_method = self._loadCRLFromURL
        else:
            raise ServerException(
                'Unknown method for loading CRL: %s' % (url.scheme,))

        return get_method, location

    @defer.inlineCallbacks
    def _updateCRL(self, get_method, location):
        """
        Load/reload revoked certificates from the configured CRL.
        """
        deferred = defer.maybeDeferred(get_method, location)
        # Let other know that  we are loading a CRL for this location.
        # It will be later overwritten by the schedule call, but in
        # case someone else want to know if there is already a deferred for
        # this location, it will check via this dict.
        self._next_update_deferreds[location] = deferred
        crl_data = yield deferred

        if not crl_data:
            raise ServerException('Empty CRL at the configured location.')

        crl_details = x509.crl_load(crl_data)

        issuer = crl_details['issuer']
        try:
            ca_cert = self._ca_certs[issuer]
        except KeyError:
            raise ServerException(
                'CRL signed by a CA which is not trusted: %s' % (issuer,))

        try:
            crypto.verify(
                ca_cert,
                crl_details['signature'],
                crl_details['to-be-signed'],
                native_string(crl_details['digest-algorithm']),
                )
        except crypto.Error as error:
            raise ServerException('CRL with invalid signature: %s' % (error,))

        new_revoked_certificates = []
        revoked_certificates = crl_details['revoked']
        for revoked in revoked_certificates:
            new_revoked_certificates.append(revoked['serial'])

        existing_issuer = self._crl_location_to_issuer.get(location, issuer)
        if existing_issuer != issuer:
            self._revoked_serials[existing_issuer] = _CRL_NOT_YET_LOADED
            raise ServerException(
                'Reseting current CRL. '
                'Issuer of the CRL at %s was changed from "%s" to "%s".' % (
                    location, existing_issuer, issuer))

        # Check to see that the current issuer is not already loaded from
        # another location.
        for other_location, other_issuer in (
                list(self._crl_location_to_issuer.items())
                ):
            if other_issuer == issuer and other_location != location:
                # For now, we only support a single CRL per CA.
                raise ServerException(
                    'CRL for %s is already loaded from %s. '
                    'It was tried to also be loaded from %s.' % (
                        issuer, other_location, location))

        self._revoked_serials[issuer] = new_revoked_certificates
        self._crl_location_to_issuer[location] = issuer

        next_update = crl_details['next-update']
        delay = self._getCRLUpdateDelay(crl_details)
        self._scheduleCRLNextUpdate(delay, get_method, location, next_update)

        update_datetime = self._now() + timedelta(seconds=delay)

        self._emitEvent('20037', data={
            'uri': location,
            'issuer': issuer,
            'count': len(new_revoked_certificates),
            'update_datetime': update_datetime,
            'update_seconds': delay,
            'next_publish': crl_details['next-publish'],
            'next_update': next_update,
            })

    def _loadCRLFromPath(self, path):
        """
        Load the raw CRL from a local filesystem.
        """
        crl_segments = local_filesystem.getSegmentsFromRealPath(path)

        if local_filesystem.isFolder(crl_segments):
            raise ServerException(
                'Loading CRL from directories is not supported.')

        crl_file = local_filesystem.openFileForReading(crl_segments)
        try:
            crl_data = crl_file.read()
        finally:
            crl_file.close()

        return crl_data

    def _loadCRLFromURL(self, url):
        """
        Load the raw CRL over HTTP.
        """
        # HTTP request should be done via a redirection agent.
        deferred = self._http_agent.request(b'GET', url.encode('utf-8'))

        def cb_got_response(response):
            if response.code != 200:
                self._http_agent.readBodyWithIgnore(response)
                raise ServerException(
                    'Failed to retrieve the CRL at %s over HTTP: %s %s' % (
                        url, response.code, force_unicode(response.phrase)))
            return self._http_agent.readBody(response)

        deferred.addCallback(cb_got_response)
        return deferred

    def _getCRLUpdateDelay(self, crl_details):
        """
        Return the seconds after which the CRL should be updated.
        """
        next_update = crl_details['next-update']
        next_publish = crl_details['next-publish']

        if next_update:
            now = self._now()
            crl_delta = next_update - now
            next_load = crl_delta.total_seconds()
            if next_load <= 0:
                raise ServerException(
                    'CRL had expired on UTC %s. Now is %s' % (
                        next_update.isoformat(), now.isoformat()))

            if next_publish:
                if next_publish > next_update:
                    raise ServerException(
                        'CRL next publish at %s UTC is after '
                        'next update at %s UTC.' % (
                            next_publish.isoformat(), next_update.isoformat()))

                # We have the CRL Next Publish extension, so use that value.
                crl_delta = next_publish - now
                next_publish_load = crl_delta.total_seconds()
                if next_publish_load <= 0:
                    # We are pass the next publish time.
                    # Retry using the soft retry or next update,
                    # which comes first.
                    next_load = min(next_load, _CRL_SOFT_RELOAD_DELAY)
                else:
                    next_load = next_publish_load

        if self._update_interval:
            # We have a pre-configured refresh rate.
            next_load = self._update_interval
        else:
            # If the current CRL had expired, there is no next update to
            # schedule.
            if not next_update:
                raise ServerException('CRL has no nextUpdate field.')

            # Use the refresh value advertised in the certificate to
            # compute the next refresh time.
            # Refresh it with a random delay of up to 60 seconds so what we
            # don't hit it at the exact refresh time.
            next_load = next_load + self._random.randint(0, 60)

        return next_load

    def _scheduleCRLNextUpdate(self, delay, get_method, location, next_update):
        """
        Called to schedule the next update of the CRL.
        """
        def eb_update(failure, location):
            if failure.check(defer.CancelledError):
                # We are stopping so this is a normal way to signal.
                return

            if next_update <= self._now():
                # The CRL is no longer valid.
                self._invalidateCRL(location)
                next_load = _CRL_SOFT_RELOAD_DELAY
            else:
                # Retry after the soft reload or at the final next update,
                # which is sooner, so that we don't run with a loaded CRL
                # past the Next Update date.
                crl_delta = next_update - self._now()
                # Delay an extra 1 second so that on fast system with low
                # clock resolution we don't fail to often.
                next_load = crl_delta.total_seconds() + 1
                next_load = min(next_load, _CRL_SOFT_RELOAD_DELAY)

            # Try reloading the CRL.
            self._scheduleCRLNextUpdate(
                delay, get_method, location, next_update)
            self._emitEvent('20038', data={
                'uri': location,
                'next_load': next_load,
                'details': force_unicode(failure.value),
                })

        deferred = task.deferLater(
            clock=self._scheduler,
            delay=delay,
            callable=self._updateCRL,
            get_method=get_method,
            location=location,
            )
        deferred.addErrback(eb_update, location)
        self._next_update_deferreds[location] = deferred


def _get_message_for_verify(code):
    """
    Return the message for certificate verification process.

    From OpenSSL source code:
    * include/openssl/x509_vfy.h
    * crypto/x509/x509_txt.c
    """
    messages = {
        0: 'OK',
        1: 'Unspecified certificate verification error',
        2: 'Unable to get issuer certificate',
        3: 'Unable to get certificate CRL',
        4: 'Unable to decrypt certificate\'s signature',
        5: 'Unable to decrypt CRL\'s signature',
        6: 'Unable to decode issuer public key',
        7: 'Certificate signature failure',
        8: 'CRL signature failure',
        9: 'Certificate is not yet valid',
        10: 'Certificate has expired',
        11: 'CRL is not yet valid',
        12: 'CRL has expired',
        13: 'Format error in certificate\'s notBefore field',
        14: 'Format error in certificate\'s notAfter field',
        15: 'Format error in CRL\'s lastUpdate field',
        16: 'Format error in CRL\'s nextUpdate field',
        17: 'Out of memory',
        18: 'Self-signed certificate',
        19: 'Self-signed certificate in certificate chain',
        20: 'Unable to get local issuer certificate',
        21: 'Unable to verify the first certificate',
        22: 'Certificate chain too long',
        23: 'Certificate revoked',
        24: 'Invalid CA certificate',
        25: 'Path length constraint exceeded',
        26: 'Unsupported certificate purpose',
        27: 'Certificate not trusted',
        28: 'Certificate rejected',
        29: 'Subject issuer mismatch',
        30: 'Authority and subject key identifier mismatch',
        31: 'Authority and issuer serial number mismatch',
        32: 'Key usage does not include certificate signing',
        33: 'Unable to get CRL issuer certificate',
        34: 'Unhandled critical extension',
        35: 'Key usage does not include CRL signing',
        36: 'Unhandled critical CRL extension',
        37: 'Invalid non-CA certificate (has CA markings)',
        38: 'Proxy path length constraint exceeded',
        39: 'Key usage does not include digital signature',
        40: 'Proxy certificates not allowed, please set the appropriate flag',
        41: 'Invalid or inconsistent certificate extension',
        42: 'Invalid or inconsistent certificate policy extension',
        43: 'No explicit policy',
        44: 'Different CRL scope',
        45: 'Unsupported extension feature',
        46: 'RFC 3779 resource not subset of parent\'s resources',
        47: 'Permitted subtree violation',
        48: 'Excluded subtree violation',
        49: 'Name constraints minimum and maximum not supported',
        50: 'Application verification failure',
        51: 'Unsupported name constraint type',
        52: 'Unsupported or invalid name constraint syntax',
        53: 'Unsupported or invalid name syntax',
        54: 'CRL path validation error',
        55: 'Path Loop',
        56: 'Suite B: Certificate version invalid',
        57: 'Suite B: Invalid public key algorithm',
        58: 'Suite B: Invalid ECC curve',
        59: 'Suite B: Invalid signature algorithm',
        60: 'Suite B: Curve not allowed for this LOS',
        61: 'Suite B: Cannot sign P-384 with P-256',
        62: 'Hostname mismatch',
        63: 'Email address mismatch',
        64: 'IP address mismatch',
        65: 'No matching DANE TLSA records',
        66: 'EE certificate key too weak',
        67: 'CA certificate key too weak',
        68: 'CA signature digest algorithm too weak',
        69: 'Invalid certificate verification context',
        70: 'Issuer certificate lookup error',
        71: 'Certificate Transparency required, but no valid SCTs found',
        72: 'Proxy subject name violation',
        73: 'OCSP verification needed',
        74: 'OCSP verification failed',
        75: 'OCSP unknown cert',
        }

    return messages.get(
        code, 'Unknown certificate verification error')


def read_PEM_configuration(
    certificate, key,
    password=None,
    certificate_option='ssl_certificate', key_option='ssl_key',
        ):
    """
    Read the PEM objects from configuration.
    """
    cert_configuration = certificate
    key_configuration = key
    certificate = None
    key = None
    chain = []

    pem_objects = _load_PEM(cert_configuration, certificate_option)
    for member in pem_objects:
        if not certificate and isinstance(member, pem.Certificate):
            # First member is our own certificate
            certificate = member
            continue

        if isinstance(member, pem.Certificate):
            # Remaining certificates are part of the chain.
            try:
                cert = crypto.load_certificate(
                    crypto.FILETYPE_PEM, member.as_bytes())
            except crypto.Error as error:
                raise ServerException(
                    'Failed to parse the chain in "%s". %s' % (
                        certificate_option, error))
            chain.append(cert)
            continue

        if isinstance(member, pem.Key):
            if key is not None:
                raise ServerException(
                    'Configured "%s" contains more than '
                    'one private key.' % (certificate_option,)
                    )
            key = member
            continue

    if key and key_configuration:
        raise ServerException(
            'Configured "%s" already contains a '
            'private key which conflicts with the "%s" '
            'configuration.' % (certificate_option, key_option)
            )

    # Convert cert from PEM to pyOpenSSL instance.
    if certificate:
        try:
            certificate = crypto.load_certificate(
                crypto.FILETYPE_PEM, certificate.as_bytes())
        except crypto.Error as error:
            raise ServerException(
                'Failed to parse "%s". %s' % (certificate_option, error))

    if not key and key_configuration:
        # We have a separate key configuration path.
        pem_objects = _load_PEM(key_configuration, key_option)
        for member in pem_objects:
            if isinstance(member, pem.Key):
                if key is not None:
                    raise ServerException(
                        'Configured "%s" has more than one '
                        'private key.' % (key_option,)
                        )
                key = member
                continue

        if key is None:
            raise ServerException(
                'Configured "%s" has no private key.' % (key_option,))

    if key:
        # Convert key from PEM to pyOpenSSL instance.
        if password:
            password = password.encode('utf-8')
        else:
            password = b''  # noqa:bandit

        try:
            key = crypto.load_privatekey(
                crypto.FILETYPE_PEM, key.as_bytes(), password)
        except crypto.Error as error:
            raise ServerException(
                'Failed to parse "%s". %s' % (key_option, six.text_type(error)))

    return (certificate, key, chain)


def _load_PEM(configuration, kind):
    """
    Return the PEM object from `configuration`.

    Configuration can be a path or a text.
    """
    def is_path(value):
        """
        Return True if `value` looks like a valid path.

        This is just a simple heiristic to detect invalid configurations.
        """
        if '\n' in value:
            return False

        return True

    configuration = configuration.strip()

    # See if configuration is text.
    pem_objects = pem.parse(configuration.encode('utf-8'))

    if not pem_objects:
        # See if configuration is a path.
        if not is_path(configuration):
            # It doesn't look like we have a path configured.
            # Just return nothing and we will fail later.
            return []
        segments = local_filesystem.getSegmentsFromRealPath(configuration)
        cert_file = None
        try:
            # See if we have a file.
            cert_file = local_filesystem.openFileForReading(segments)
            pem_objects = pem.parse(cert_file.read())
        except EnvironmentError as error:
            raise ServerException('Failed to load %s. %s' % (kind, error))
        finally:
            if cert_file:
                cert_file.close()

    return pem_objects


def load_PKCS12(password, path=None, data=None, source='ssl_certificate'):
    """
    Read the PKCS12 objects from path or PEM data.

    This also reads modern PFX https://en.wikipedia.org/wiki/PKCS_12

    Raise ServerException on error.

    Returns a tuple with certificate, private key and certificate chain,
    in `cryptography` format.
    `chain` is a list or `cryptography` certificates.
    """
    certificate = None
    chain = []

    if path and data:
        raise RuntimeError('Only path or data can be used')

    if path:
        _, extension = os.path.splitext(path.lower())
        if extension not in ['.p12', '.pfx']:
            return (None, None, [])

        segments = local_filesystem.getSegmentsFromRealPath(path)
        stream = None
        try:
            stream = local_filesystem.openFileForReading(segments)
            data = stream.read()
        except EnvironmentError as error:
            raise ServerException(
                'Failed to read the PKCS#12 / PFX file from '
                '%s. %s' % (source, force_unicode(error),))
        finally:
            if stream:
                stream.close()
    elif isinstance(data, str):
        data = data.encode('ascii')

    if password:
        password = password.encode('utf-8')
    else:
        password = b''  # noqa:bandit

    try:
        p12 = serialization.pkcs12.load_pkcs12(data, password)
    except ValueError as error:
        raise ServerException(
            'Failed to parse PKCS#12 / PFX file from '
            '%s. %s' % (source, force_unicode(error)))

    # Unwrap the P12Certificate so that we have normal Certificate instances.
    certificate = p12.cert.certificate
    for ca_cert in p12.additional_certs:
        chain.append(ca_cert.certificate)

    return (certificate, p12.key, chain)


def dump_certificate(certificate):
    """
    Return the PEM serialization.
    """
    try:
        if isinstance(certificate, crypto.X509):
            return crypto.dump_certificate(
                crypto.FILETYPE_PEM, certificate).strip().decode('ascii')

        return certificate.public_bytes(
            serialization.Encoding.PEM).decode('ascii')

    except (AttributeError, ValueError, TypeError):
        raise ServerException('Failed to convert to PEM certificate.')


def dump_key(key):
    """
    Return the PEM serialization of a `cryptography` key.
    """
    try:
        return key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption()
            ).decode('ascii')
    except (AttributeError, ValueError, TypeError):
        raise ServerException('Failed to convert to PEM key.')


def _validate_hostname(certificate, remote_ids, remote_names):
    """
    Check that certificate is issued for one of the accepted `remote_id`.
    """
    if not remote_ids:
        return

    try:
        actual = extract_patterns(certificate)
        last_error = None
        for candidate in remote_ids:
            try:
                return verify_service_identity(
                    cert_patterns=actual,
                    obligatory_ids=[candidate],
                    optional_ids=[],
                    )
            except VerificationError as error:
                # Allow to continue with another accepted name
                # and just record the error to have something to report
                # if everything fails
                last_error = error
            except ValueError:
                # See: https://github.com/pyca/service-identity/issues/22
                raise ServerException(
                    'Only fully qualified domain names (FQDN) '
                    'can be validated.')

        # We have not found any match.
        raise last_error

    except VerificationError:
        # Try to extract the actual names from the DNSPatterns.
        actual_names = []
        for name in actual:
            name = name.pattern
            if isinstance(name, six.binary_type):
                name = name.decode('utf-8')
            else:
                name = str(name)

            actual_names.append(name)
        actual_names = ', '.join(actual_names)
        raise ServerException(
            'Hostname verification failed for "%s". '
            'Remote identity is: %s' % (
                remote_names, actual_names))
    except CertificateError as error:
        # service_identity uses repr() in the error message.
        # we remove the bytes marker to have nicer message.
        details = force_unicode(error.args[0]).replace("b'", "'")
        raise ServerException(
            'Hostname verification failed for "%s". '
            'Remote certificate is invalid. %s' % (
                remote_names, details))


@defer.inlineCallbacks
def get_remote_certificates(address, port):
    """
    Connect to TLS server at `address:port` and retrieve the certificates chain
    obtained during TLS handshake.

    It returns a list of pyOpenSSL certificates.
    """
    # We use a very permissive context that should allow
    # any server.
    # FIXME:2093:
    # For some servers, we also need to send our certificate.
    # For now, this use case fails.
    # We will add a certificate+key to the context factory later.
    endpoint = endpoints.SSL4ClientEndpoint(
        reactor, address, port, ClientContextFactory())

    protocol = _GetCAChain(address)
    try:
        yield endpoints.connectProtocol(endpoint, protocol)
    except Exception as error:
        raise ServerException(
            'Failed to get CA chain. '
            'Failed to connect to remote server. {}'.format(
                force_unicode(error)))

    try:
        result = yield protocol.done
    except Exception as error:
        raise ServerException(
            'Failed to get CA chain. {}'.format(
                force_unicode(error)))
    defer.returnValue(result)


@implementer(IHandshakeListener)
class _GetCAChain(Protocol):
    """
    A client-side protocol used to extract the peer CA chain.

    It returns the CA chain as PEM via the self.done deferred.
    """
    def __init__(self, hostname):
        self._hostname = hostname.encode('idna')
        self.done = defer.Deferred()

    def connectionMade(self):
        """
        We assume this is called as soon a the connection is made
        and before the TLS handshake request is made.
        """
        self.transport._tlsConnection.set_tlsext_host_name(
            self._hostname)

    def connectionLost(self, reason):
        """
        Called when connection is lost.
        """
        if self.done.called:
            # Certificates already retrieved,
            return

        # Inform that we failed to get the certificates.
        self.done.errback(reason.value)

    def handshakeCompleted(self):
        """
        See: IHandshakeListener.
        """
        try:
            cert_chain = self.transport._tlsConnection.get_peer_cert_chain()
            self.done.callback(cert_chain)
        finally:
            # We always disconnect , as there is no much to do here.
            self.transport.abortConnection()


def _get_available_ssl_ciphers():
    """
    Return a list with available OpenSSL ciphers.
    """
    # Try to setup our server-side context with as many ciphers
    # as possible.
    factory = OpenSSLContextFactory(
        certificate=_PEMS_PLACEHOLDER,
        cipher_list='ALL',
        on_error=lambda e: None,  # No error is expected here.
        event_emitter=lambda: None,  # This should not emit anything.
        allowed_methods=['all'],
        )

    # We fake a connection, so that we can use pyopenssl API to
    # get the available ciphers.
    connection = SSL.Connection(factory._context,  None)
    return connection.get_cipher_list()


# Update the constants value, but keep the reference.
del SSL_CIPHERS[:]
SSL_CIPHERS.extend(['secure', 'ALL'])
SSL_CIPHERS.extend(_get_available_ssl_ciphers())
