# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
Status for server's process.
"""
import os
import sys
from time import time

from OpenSSL import SSL
from zope.interface import implementer

from chevah_compat import local_filesystem, process_capabilities
from chevah.server import force_unicode
from chevah.server.commons.interface import (
    IServerStatus,
    )
from chevah.server.configuration.model import PropertyMixin
from chevah.server.instance_name import INSTANCE_NAME


@implementer(IServerStatus)
class ServerStatus(PropertyMixin):
    """
    Aggregates status for the server process.

    This represents the runnable which is configured via the `server`
    configuration section.
    """
    # Here to help with testing.
    _os = os

    def __init__(self, parent):
        self._configuration = parent.configuration
        # This is set by the legacy pywin32 service manager,
        # In the case in which the process is started in that way.
        self.windows_legacy_service = False
        self._start_time = time()
        self._can_impersonate_local_accounts = (
            process_capabilities.impersonate_local_account)
        self._can_create_local_home_folders = (
            process_capabilities.create_home_folder)
        self._can_get_local_home_folders = (
            process_capabilities.get_home_folder)
        self._startup_configuration = {
            'server/umask': self._configuration.server.umask,
            'server/account': self._configuration.server.account,
            }
        self._os_family = process_capabilities.os_family
        if self._os_family == 'posix':
            self._os_family = 'Unix'
        if self._os_family == 'nt':
            self._os_family = 'Windows'

        self._os_name = process_capabilities.os_name

        self._install_path = local_filesystem.getRealPathFromSegments(
            local_filesystem.installation_segments)

        self._openssl_version = "%s (%s)" % (
            force_unicode(SSL.SSLeay_version(SSL.SSLEAY_VERSION)),
            force_unicode(SSL.OPENSSL_VERSION_NUMBER),
            )

    @property
    def parent(self):
        """
        This has the root.
        """
        return None

    @property
    def uuid(self):
        """
        See IServerStatus.
        """
        return self._configuration.server.uuid

    @property
    def name(self):
        """
        See IServerStatus.
        """
        return self._configuration.server.name

    @property
    def start_time(self):
        """
        Get the configuration start time in seconds since epoch.
        """
        return self._start_time

    @property
    def restart_required(self):
        """
        See: `IServerStatus`.
        """
        # It would be nice to refactor to reuse more code from a runnable
        # so that t can share the startup configuration and restart required.
        # It should have configuration.server as its main configuration.
        latest_umask = self._configuration.server.umask
        latest_account = self._configuration.server.account
        result = []

        if latest_umask != self._startup_configuration['server/umask']:
            result.append('server/umask')

        if latest_account != self._startup_configuration['server/account']:
            result.append('server/account')

        return result

    @property
    def startup_configuration(self):
        """
        See: `IServerStatus`.
        """
        return self._startup_configuration

    @property
    def os_family(self):
        """
        See: `IServerStatus`.
        """
        return self._os_family

    @property
    def os_name(self):
        """
        See: `IServerStatus`.
        """
        return self._os_name

    @property
    def can_impersonate_local_accounts(self):
        """
        True if server can impersonate local accounts.
        """
        return self._can_impersonate_local_accounts

    @property
    def can_create_local_home_folders(self):
        """
        True if server can create local home folders.
        """
        return self._can_create_local_home_folders

    @property
    def can_get_local_home_folders(self):
        """
        True if server can retrieve local home folder paths for any account.
        """
        return self._can_get_local_home_folders

    @property
    def openssl_version(self):
        """
        See: `IServerStatus`.
        """
        return self._openssl_version

    @property
    def cwd(self):
        """
        See: `IServerStatus`.
        """
        return os.getcwd()

    @property
    def install_path(self):
        """
        See: `IServerStatus`.
        """
        return self._install_path

    @property
    def instance_name(self):
        """
        See: `IServerStatus`.
        """
        return INSTANCE_NAME

    @property
    def python_version(self):
        """
        See: `IServerStatus`.
        """
        # We only want the Python version,
        # without the compile environment.
        return sys.version.splitlines()[0]

    @property
    def os_user(self):
        """
        See IServerStatus.
        """
        if os.name == 'nt':
            return _get_windows_os_user()
        return _get_posix_os_user(self._os)


def _get_posix_os_user(module_os):
    """
    Return the name of the current effective user on posix.
    """
    import pwd
    try:
        return pwd.getpwuid(module_os.geteuid()).pw_name
    except Exception:
        return 'FAIL-TO-GET-OS-USER'


def _get_windows_os_user():
    """
    Return the name of the current user on Windows.
    """
    import win32api
    try:
        return win32api.GetUserNameEx(win32api.NameSamCompatible)
    except Exception:
        return 'FAIL-TO-GET-OS-USER'
