# Copyright (c) 2018 Adi Roiban.
# See LICENSE for details.
"""
Common code for streamed transfers.
"""
from __future__ import unicode_literals

from __future__ import absolute_import
from twisted.internet import defer

from chevah.server.commons.exception import ServerException


class UploadProducer(object):
    """
    A producer sending data to a remote TCP connection.

    The write method will wait for the TCP connection to be available for
    sending.

    This implements part of ITransferStream.
    """

    def __init__(self, consumer):
        self._consumer = consumer
        # Deferred used to signal that write operation should be paused.
        self._pause_write = None
        # Temporary storage for data which can't be yet written.
        # `None` is used to signal that no write is waiting.
        self._pause_buffer = None

    @property
    def mode(self):
        """
        Always in writing.
        """
        return 'writing'

    def read(self, length):
        """
        See: `ITransferStream`.
        """
        return defer.fail(ServerException('Only opened for writing.'))

    def write(self, data):
        """
        See: `ITransferStream`.
        """
        if not self._consumer:
            return defer.fail(ServerException('Write after stop.'))

        # FIXME:3979:
        # The chunk is sent to the consumer and it will process it.
        # Once we have a new interface this will be removed,
        # as this API is forced.
        # The consume will use only pause/resume production method to signal
        # that no more data should be written.
        if self._pause_write:
            if self._pause_buffer is not None:
                # Here we ensure that with our current API design the write
                # operation don't overalap.
                # A write call must ensure the previous write call was
                # finalized.
                return defer.fail(ServerException(
                    'Write request made '
                    'while previous write was not finalized.'
                    ))
            self._pause_buffer = data
            return self._pause_write
        else:
            return defer.maybeDeferred(self._consumer.write, data)

    def stopProducing(self):
        """
        See: IPushProducer

        This is usually called on errors, or when the remote connection
        is closed.
        For example when we wrote past the advertised length.
        """
        self._consumer = None

        if self._pause_write:
            # Discard any pending data.
            self._pause_buffer = None
            # Signal that write is done.
            pause, self._pause_write = self._pause_write, None
            pause.errback(ServerException(
                'Received request to stop the upload.'))

    def pauseProducing(self):
        """
        See: IPushProducer
        """
        if self._pause_write:
            # Already paused.
            return

        self._pause_write = defer.Deferred()

    def resumeProducing(self):
        """
        See: IPushProducer
        """
        # Write operations can continue.
        pause, self._pause_write = self._pause_write, None
        pause_buffer, self._pause_buffer = self._pause_buffer, None

        if pause_buffer:
            # pause_buffer can be None so we only write if we have a buffer.
            self.write(pause_buffer)

        if pause:
            # One might call it without calling pauseProduction first, so we
            # are flexible here.
            pause.callback(None)


class DelayedWrite(object):
    """
    Will create the file in binary mode only on first write.
    """
    def __init__(self, path):
        self._stream = None
        self._path = path

    def write(self, data):
        if not self._stream:
            self._stream = open(self._path, 'wb')
        self._stream.write(data)

    def close(self):
        if not self._stream:
            return
        self._stream.close()
        self._stream = None

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()
        return False
