# Copyright (c) 2016 Adi Roiban.
# See LICENSE for details.
"""
Handling of X.509 files.
"""
from __future__ import absolute_import
from __future__ import unicode_literals
import datetime
from binascii import a2b_base64
from collections import OrderedDict

from OpenSSL import crypto
from asn1crypto.core import Void
from asn1crypto.crl import CertificateList
from asn1crypto import x509 as asn1crypto_x509

from chevah.server import force_unicode
from chevah.server.commons import UTC
from chevah.server.commons.exception import ServerException
import six
from six.moves import range

OID_to_name = {
    # DN oids.
    (2, 5, 4, 3): 'CN',  # common name
    (2, 5, 4, 6): 'C',  # country
    (2, 5, 4, 7): 'L',  # locality
    (2, 5, 4, 8): 'ST',  # stateOrProvince
    (2, 5, 4, 10): 'O',  # organization
    (2, 5, 4, 11): 'OU',  # organizationalUnit
    (0, 9, 2342, 19200300, 100, 1, 25): 'DC',  # domainComponent
    (1, 2, 840, 113549, 1, 9, 1): 'emailAddress',

    # Digest algorithms.
    # See: http://www.alvestrand.no/objectid/1.2.840.113549.1.1.html
    (1, 2, 840, 113549, 1, 1, 4): 'md5',
    (1, 2, 840, 113549, 1, 1, 5): 'sha1',
    (1, 2, 840, 113549, 1, 1, 11): 'sha256',
    (1, 2, 840, 113549, 1, 1, 13): 'sha512',
    }

name_to_OID = {value: key for key, value in OID_to_name.items()}


def _dotted_to_tuple(dotted_oid):
    """
    Return a tuple to be used with OID_to_name mapper.
    """
    return tuple(int(n) for n in dotted_oid.split('.'))


def _decode_time(obj, time_name):
    """
    Decode x.509 time objects.

    See RFC 5280.

    It supports UTCTime only in GMT.
    """
    if isinstance(obj, Void):
        return None

    time_obj = obj.parse()

    time_formats = {
        # 4.1.2.5.1. UTCTime.
        23: "%y%m%d%H%M%SZ",
        # 4.1.2.5.2. GeneralizedTime
        # ASN1 defines multiple formats for formats, RFC 5280 mandates a single
        # format.
        24: "%Y%m%d%H%M%SZ",
        }
    time_format = time_formats[time_obj.tag]
    return datetime.datetime.strptime(
        time_obj.contents.decode('ascii'), time_format).replace(tzinfo=UTC)


def _decode_rdn_components(rdn_sequence):
    """
    Return a list with name and value of RelativeDistinguishedName.
    """
    parts = []
    for component in rdn_sequence:
        component = component[0]
        comp_type = _dotted_to_tuple(component['type'].dotted)
        value = component['value'].native

        name = OID_to_name[comp_type]
        parts.append((name, value))

    return parts


def crl_parse(crl_der):
    """
    Return a dictionary with the CRL attributes of the CRL passed as DER.

    RFC 5280

    CertificateList  ::=  SEQUENCE  {
        tbsCertList          TBSCertList,
        signatureAlgorithm   AlgorithmIdentifier,
        signatureValue       BIT STRING  }

    TBSCertList  ::=  SEQUENCE  {
        version                 Version OPTIONAL,
                                     -- if present, MUST be v2
        signature               AlgorithmIdentifier,
        issuer                  Name,
        thisUpdate              Time,
        nextUpdate              Time OPTIONAL,
        revokedCertificates     SEQUENCE OF SEQUENCE  {
             userCertificate         CertificateSerialNumber,
             revocationDate          Time,
             crlEntryExtensions      Extensions OPTIONAL
                                      -- if present, version MUST be v2
                                  }  OPTIONAL,
        crlExtensions           [0]  EXPLICIT Extensions OPTIONAL
                                      -- if present, version MUST be v2
                                  }

    Version  ::=  INTEGER  {  v1(0), v2(1), v3(2)  }

    CertificateSerialNumber  ::=  INTEGER

    Time ::= CHOICE {
        utcTime        UTCTime,
        generalTime    GeneralizedTime }
    """
    try:
        crl_asn1 = CertificateList.load(crl_der)
    except ValueError as error:
        raise ServerException(
            'Failed to decode CRL: {}'.format(force_unicode(error)))

    to_be_signed = crl_asn1['tbs_cert_list']

    issuer = to_be_signed['issuer']
    this_update = to_be_signed['this_update']
    next_update = to_be_signed['next_update']

    signature_value = crl_asn1['signature']

    signature_algorithm = crl_asn1['signature_algorithm']
    algorithm = _dotted_to_tuple(signature_algorithm['algorithm'].dotted)

    try:
        digest_name = OID_to_name[algorithm]
    except KeyError:
        raise ServerException(
            'CRL signature algorithm not supported: %s' % (algorithm,))

    revoked = to_be_signed['revoked_certificates']
    revoked_certificates = _parse_revoked(revoked)

    parsed_extensions = to_be_signed['crl_extensions']
    extensions = _parse_extensions(parsed_extensions)

    # Look for next publish extension and expose it as root attribute.
    next_publish = None
    for extension in extensions:
        if extension['name'] != 'CRL Next Publish':
            continue
        next_publish = extension['content']['nextPublish']
    return {
        'to-be-signed': crl_asn1[0].dump(),
        'issuer': get_subject(
            _decode_rdn_components(issuer.parse()),
            'loaded CRL issuer',
            ),
        'signature': signature_value.native,
        'digest-algorithm': digest_name,
        'this-update': _decode_time(this_update, 'this-update'),
        'next-update': _decode_time(next_update, 'next-update'),
        'next-publish': next_publish,
        'revoked': revoked_certificates,
        'extensions': extensions,
        }


def _parse_revoked(revoked):
    """
    Return the list with certificates in the revocation list.
    """
    if not revoked:
        return []

    revoked_certificates = []
    for revoked_certificate in revoked:
        serial = revoked_certificate['user_certificate']
        revoked_certificates.append({
            'serial': int(serial),
            })

    return revoked_certificates


def _parse_extensions(parsed_extensions):
    """
    Return the list of extensions as serialized in ASN.1 `parsed_extensions`
    sequence.
    """
    if not parsed_extensions:
        return []

    def oid_to_name(oid):
        """
        Convert the CRL extension OID to a human readable name.

        Return the same `oid` if extension is not known.
        """

        mapping = {
            # Microsoft CertSrv Infrastructure
            # https://support.microsoft.com/en-us/kb/287547
            '1.3.6.1.4.1.311.21.4': 'CRL Next Publish',
            # Standard extensions.
            # As defined in RFC 5280.
            '2.5.29.9': 'Subject Directory Attributes',
            '2.5.29.14': 'Subject Key Identifier',
            '2.5.29.15': 'Key Usage',
            '2.5.29.16': 'Private Key Usage Period',
            '2.5.29.17': 'Subject Alternative Name',
            '2.5.29.18': 'Issuer Alternative Name',
            '2.5.29.19': 'Basic Constraints',
            '2.5.29.20': 'CRL Number',
            '2.5.29.21': 'Revocation Reason',
            '2.5.29.24': 'Invalidity Date',
            '2.5.29.27': 'Delta CRL Indicator',
            '2.5.29.28': 'Issuing Distribution Point',
            '2.5.29.29': 'Certificate Issuer',
            '2.5.29.30': 'Name Constraints',
            '2.5.29.31': 'CRL Distribution Points',
            '2.5.29.32': 'Certificate Policies',
            '2.5.29.33': 'Policy Mappings',
            '2.5.29.35': 'Authority Key Identifier',
            '2.5.29.36': 'Policy Constraints',
            '2.5.29.37': 'Extended Key Usage',
            '2.5.29.46': 'Freshest CRL (a.k.a. Delta CRL Distribution Point)',
            '2.5.29.54': 'Inhibit anyPolicy',
            }
        return mapping.get(oid, oid)

    def parse_extension(oid, value):
        """
        Try to parse the CRL extension, returning an empty dict if the
        extension  could not be parsed.
        """
        result = {}

        if oid == '1.3.6.1.4.1.311.21.4':
            # https://tools.ietf.org/html/draft-deacon-lightweight-ocsp-profile-00#page-12
            result['nextPublish'] = _decode_time(value, 'next-publish')
            return result

        if oid == '2.5.29.20':
            result['crlNumber'] = value.parse().native
            return result

        return result

    extensions = []
    for extension in parsed_extensions:
        oid = six.text_type(extension['extn_id'])
        value = extension['extn_value']
        extensions.append({
            'oid': oid,
            'name': oid_to_name(oid),
            'critical': bool(extension['critical']),
            'value': value,
            'content': parse_extension(oid, value)
            })

    return extensions


def crl_load(crl_data):
    """
    Parse the CRL with input in DER or PEM.
    """
    if not crl_data:
        raise ServerException('No data for CRL.')

    lines = crl_data.strip().splitlines()
    if b'BEGIN X509 CRL' in lines[0]:
        # We have a PEM data
        # Now convert to DER.
        content = []
        for line in lines[1:]:
            if b'END X509 CRL' in line:
                # End of content.
                break
            content.append(line)

        crl_der = pem_lines_to_der(content)
    else:
        crl_der = crl_data

    return crl_parse(crl_der)


def get_subject(components, error_location, fail=True):
    """
    Return subject components as string in OpenSSL text format.
    """
    if not components:
        message = 'Missing X509 subject in %s.' % (error_location,)
        if fail:
            raise ServerException(message)
        return message

    parts = []
    for name, value in components:
        parts.append('/%s=%s' % (force_unicode(name), force_unicode(value)))

    return ''.join(parts)


def get_CRLs_from_CDP_extension(certificate):
    """
    Return a list of CRL location as advertised in the x509 extension for
    `certificate`
    """
    result = []

    for index in range(0, certificate.get_extension_count()):
        extension = certificate.get_extension(index)
        if extension.get_short_name() != b'crlDistributionPoints':
            continue

        data = extension.get_data()
        cdp = asn1crypto_x509.CRLDistributionPoints.load(data)

        for distribution_point in cdp:
            dp = distribution_point['distribution_point']
            dp_name = dp.parse()[0]
            if dp_name.parse().tag != 6:
                # For now only uniformResourceIdentifier are supported.
                # Entries like the following are ingored
                # * DNS:cdp.example.com
                # * IP:192.168.0.0/255.255.0.0
                continue
            result.append(dp_name.native)

    return result


def get_common_name_from_subject(subject):
    """
    Return the Common Name value from the X509 OpenSSL subject.
    """
    try:
        return subject.commonName
    except crypto.Error as error:
        if error.args[0][0][1] != 'ASN1_mbstring_copy':
            raise error
        # On older versions of OpenSSL we ca not decode newer x509
        # certificates using the commoneName attribute.
        # This happens on Solaris 10 for example.
        components = subject.get_components()
        for key, value in components:
            if key != b'CN':
                continue
            return value.decode('utf-8')


def pem_lines_to_der(lines):
    """
    Convert the BASE64 PEM lines to ASN.1 serialization.

    Lines should be passed without the armor.
    """
    content = []
    for line in lines:
        # Accumulate content without newlines or spaces.
        content.append(line.strip(b'\n\t\r '))
    return a2b_base64(b''.join(content))


class PreOrderedName(asn1crypto_x509.Name):
    """
    This is the copy of upstream, but will not re-order the attribute
    if they are already passed as an ordered dict.
    """

    @classmethod
    def build(cls, name_dict, use_printable=False):
        """
        Creates a Name object from a dict of unicode string keys and values.
        The keys should be from NameType._map, or a dotted-integer OID unicode
        string.

        :param name_dict:
            A dict of name information, e.g. {"common_name": "Will Bond",
            "country_name": "US", "organization": "Codex Non Sufficit LC"}

        :param use_printable:
            A bool - if PrintableString should be used for encoding instead of
            UTF8String. This is for backwards compatibility with old software.

        :return:
            An x509.Name object
        """

        rdns = []
        if not use_printable:
            encoding_name = 'utf8_string'
            encoding_class = asn1crypto_x509.UTF8String
        else:
            encoding_name = 'printable_string'
            encoding_class = asn1crypto_x509.PrintableString

        # This condition is our custom code.
        if not isinstance(name_dict, OrderedDict):
            # Sort the attributes according to NameType.preferred_order
            name_dict = OrderedDict(
                sorted(
                    list(name_dict.items()),
                    key=lambda item: asn1crypto_x509.NameType.preferred_ordinal(
                        item[0])
                    )
                )

        for attribute_name, attribute_value in name_dict.items():
            attribute_name = asn1crypto_x509.NameType.map(attribute_name)
            if attribute_name == 'email_address':
                value = asn1crypto_x509.EmailAddress(attribute_value)
            elif attribute_name == 'domain_component':
                value = asn1crypto_x509.DNSName(attribute_value)
            elif attribute_name in {
                    'dn_qualifier', 'country_name', 'serial_number'}:
                value = asn1crypto_x509.DirectoryString(
                    name='printable_string',
                    value=asn1crypto_x509.PrintableString(attribute_value)
                    )
            else:
                value = asn1crypto_x509.DirectoryString(
                    name=encoding_name,
                    value=encoding_class(attribute_value)
                    )

            rdns.append(asn1crypto_x509.RelativeDistinguishedName([
                asn1crypto_x509.NameTypeAndValue({
                    'type': attribute_name,
                    'value': value
                    })
                ]))

        return cls(name='', value=asn1crypto_x509.RDNSequence(rdns))
