# Copyright (c) 2019 Adi Roiban.
# See LICENSE for details.
"""
Code for handling ZIP archives.
"""
from __future__ import unicode_literals, absolute_import

import os
import ntpath
import struct
import zipfile
import zlib
import time
import six

from chevah.server.commons.exception import ServerException


class ChevahZip(zipfile.ZipFile):
    """
    ZipFile which handles unicode paths.
    """

    def getTargetPath(self, member_path, base):
        """
        Return the (is_file, path) for `member_path` as if it would be
        extracted in path `base`.

        The returned `path` is the normalized OS path.

        `is_file` is None when member_path is a root folder.

        It raises an exception if zip file has malicious paths.
        """
        base = os.path.normpath(base)

        # Force path to be relative.
        arcname = member_path.lstrip('/')

        if not arcname:
            # This member is just the root.
            return None, base

        for part in arcname.strip('/').split('/'):
            part = part.strip()
            if not part:
                raise ServerException('ZIP with empty member name.')

        # build the destination pathname, replacing
        # forward slashes to platform specific separators.
        arcname = arcname.strip().replace('/', os.path.sep)

        # We might have malicious zips like:
        # /c:/abs/win
        #
        # interpret absolute pathname as relative, remove drive letter or
        # UNC path, redundant separators, "." and ".." components.
        drive, arcname = ntpath.splitdrive(arcname)

        if drive:
            raise ServerException('ZIP archive has absolute paths.')

        targetpath = os.path.normpath(os.path.join(base, arcname))

        if not targetpath.startswith(base):
            raise ServerException(
                'ZIP archive has members with path outside of the base path.')

        if member_path[-1] == '/':
            return False, targetpath

        return True, targetpath


class StreamWriterWithPosition(object):
    """
    Adapter for a write-only stream which knows its current position.

    A stream can't be rewind/seek or flushed.
    """
    def __init__(self, stream):
        self._stream = stream
        self._pos = 0

    def write(self, data):
        if self._stream.finished:
            return
        self._stream.write(data)
        self._pos += len(data)

    def tell(self):
        return self._pos

    def flush(self):
        pass


class ZipStreamWriter(zipfile.ZipFile):
    """
    A zip file which can write files from and to a stream.

    See https://users.cs.jmu.edu/buchhofp/forensics/formats/pkzip.html
    """
    _current_info = None
    _compressor = None

    def writeHeader(self, path, modified=None):
        """
        Write the header for a file which is going to be written in chunks.

        `path` is the relative path as Unicode in the archive.
        """
        if not modified:
            modified = time.time()

        # Path should be relative.
        path = path.lstrip('/')

        filename = path
        if six.PY2:
            filename = filename.encode('utf-8')

        zinfo = zipfile.ZipInfo(
            filename=filename,
            date_time=time.localtime(modified)[:6],
            )
        zinfo.file_size = 0
        zinfo.compress_size = 0

        zinfo.header_offset = self.fp.tell()

        zinfo.compress_type = self.compression

        self._compressor = None
        if zinfo.compress_type == zipfile.ZIP_DEFLATED:
            self._compressor = zlib.compressobj(
                zlib.Z_DEFAULT_COMPRESSION,  # level
                zlib.DEFLATED,  # method
                -15,  # wbits
                9,  # memlevel
                zlib.Z_DEFAULT_STRATEGY,  # strategy,
                )

        if zinfo.filename[-1] == '/':
            # Set spacial permissions for dirs.
            zinfo.external_attr = 0o40775 << 16   # drwxrwxr-x
            zinfo.external_attr |= 0x10           # MS-DOS directory flag
        else:
            # For Files
            zinfo.external_attr = 0o600 << 16     # ?rw-------

        self._writecheck(zinfo)
        self._didModify = True

        # For stream content we don't know the final size and CR32, so we
        # use a final data description.
        # The data descriptor is only present if bit 3 of the bit flag field is
        # set.
        # In this case, the CRC-32, compressed size,
        # and uncompressed size fields in the local header are set to zero.
        # The data descriptor field is byte aligned and immediately follows
        # the file data.
        zinfo.flag_bits = 0x08
        zinfo.CRC = 0

        self.fp.write(zinfo.FileHeader(zip64=True))
        self._current_info = zinfo

    def writeContentChunk(self, data):
        """
        Write a chunk of bytes `data` to the currently opened file.
        """
        chunk_size = len(data)
        self._current_info.file_size += chunk_size
        self._current_info.CRC = (
            zlib.crc32(data, self._current_info.CRC) & 0xffffffff)

        if self._compressor:
            data = self._compressor.compress(data)
            chunk_size = len(data)

        self._current_info.compress_size += chunk_size
        self.fp.write(data)

    def writeFooter(self):
        """
        Write the footer (data descriptor)
        """
        if self._compressor:
            data = self._compressor.flush()
            self._current_info.compress_size += len(data)
            self.fp.write(data)

        zip64 = (
            self._current_info.file_size > zipfile.ZIP64_LIMIT or
            self._current_info.compress_size > zipfile.ZIP64_LIMIT
            )
        # Write CRC and file sizes after the file data
        fmt = '<LQQ' if zip64 else '<LLL'
        self.fp.write(struct.pack(
            fmt,
            self._current_info.CRC,
            self._current_info.compress_size,
            self._current_info.file_size,
            ))

        self.fp.flush()

        self.filelist.append(self._current_info)
        self.NameToInfo[self._current_info.filename] = self._current_info
        self._current_info = None
        # This is needed for Py3
        self.start_dir = self.fp.tell()
