# Copyright (c) 2011 Adi Roiban.
# See LICENSE for details.
"""
Module for configuration loaded from local files.

This is a configuration file on steroids.

It has a JSON attached file where metadata configuration can be stored.
For example to store the datatime when a configuration option was updated.

It has a local "node" auxiliary configuration file that can be used to
overwrite the options from the main `.ini` file.
This is used as some sort of "templating" in which you have the main
configuration file as the template, and the node file with inputs for the
template.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

import codecs
import os
import six
from io import BytesIO, StringIO

from lazr.delegates import delegate_to
from tinydb import operations, TinyDB, where
from tinydb.storages import MemoryStorage, JSONStorage
from zope.interface import implementer

from chevah_compat import local_filesystem
from chevah.server import force_unicode, json
from chevah.server.commons.constant import (
    CONFIGURATION_DISABLED_VALUE,
    CONFIGURATION_DISABLED_VALUES,
    CONFIGURATION_INHERIT,
    )
from chevah.server.commons.exception import (
    ConfigurationException,
    ServerError,
    ServerException,
    )
from chevah.server.commons.interface import (
    IConfigurationProxy,
    INodeConfiguration,
    )
from chevah.slim_server.raw_config_parser import (
    ChevahRawConfigParser,
    configparser,
    )


class ChevahJSONStorage(JSONStorage):
    """
    Store the data in a JSON file.
    """

    def __init__(self, path, create_dirs=False, encoding=None, **kwargs):
        """
        Create a new instance.

        Also creates the storage file, if it doesn't exist.

        :param path: Where to store the JSON data.
        :type path: str
        """
        super(JSONStorage, self).__init__()
        self._kwargs = kwargs
        self._path = path
        self._encoded_path = local_filesystem.getEncodedPath(path)
        self._cache = None

    def read(self):
        """
        Read from cache or from disk.
        """
        if self._cache is not None:
            return self._cache

        try:
            if not os.path.exists(self._encoded_path):
                with open(self._encoded_path, 'ab'):
                    """
                    Attempt to create the file.
                    """
        except EnvironmentError:
            """
            We can not create a cache file.
            Allow the application to proceed, but saving will fail later.
            """
            self._cache = {}
            return self._cache

        try:
            with open(self._encoded_path, 'rb') as stream:
                content = stream.read().strip()
                self._cache = None
                if content:
                    self._cache = json.loads(content)

        except EnvironmentError:
            message = (
                'Failed to read the configuration cache '
                'at %s' % (self._path,))
            raise ServerException(message)

        return self._cache

    def write(self, data):
        """
        Write in memory.
        """
        self._cache = data

    def save(self):
        """
        Write the data to disk.
        """
        try:
            with open(self._encoded_path, 'wb') as stream:
                stream.write(json.dumps(
                    self._cache, **self._kwargs).encode('utf-8  '))
        except EnvironmentError:
            message = 'Failed to write the configuration cache at %s' % (
                self._path,)
            raise ServerException(message)

    def close(self):
        """
        Write the data to disk and reset the buffer in memory.
        """
        self.save()
        self._cache = None


def _utf8_wrap(stream):
    """
    Wrap the bytes `stream` into UTF-8 coder/decoder.
    """
    info = codecs.lookup('utf8')
    utf8_stream = codecs.StreamReaderWriter(
        stream, info.streamreader, info.streamwriter, 'replace')
    utf8_stream.encoding = 'utf8'
    return utf8_stream


class ChevahTinyDB(TinyDB):
    """
    Modified TinyDB to have the same behaviour on Py2 and Py3.
    """
    DEFAULT_TABLE = '_default'


@implementer(IConfigurationProxy)
@delegate_to(INodeConfiguration, context='_node_config')
class FileConfigurationProxy(object):
    """
    Config parser for Chevah projects.

    `configuration_file` is as stream that is already opened in text mode.
    configuration_file is used only for testing, as in PROD it receives
    a path.
    """

    def __init__(self, configuration_path=None, configuration_file=None,
                 defaults=None):
        raw_defaults = None
        if defaults:
            raw_defaults = {}
            for key, value in defaults.items():
                raw_defaults[key] = json.dumps(value)

        self._db = None
        self._raw_config = ChevahRawConfigParser(raw_defaults)
        self._node_config = NodeConfigurationFile()
        self._configuration_path = configuration_path
        self._configuration_file = configuration_file

        if not configuration_file and not configuration_path:
            raise RuntimeError('You must specify a path or a file.')

    def load(self):
        """
        Load configuration from input file.
        """
        config_stream = None
        try:
            if self._configuration_file:
                # We are running from a stream.
                # No need to handle exceptions.
                self._db = ChevahTinyDB(storage=MemoryStorage)
                config_stream = self._configuration_file
            else:
                config_stream = self._openConfiguration()
                self._db = self._getCacheDB()

        except Exception:
            # Cleanup the main config if the cache loading fails.
            if config_stream:
                config_stream.close()
            raise

        try:
            self._raw_config.read_file(config_stream)
        except (configparser.ParsingError, AttributeError) as error:
            message = 'Could not parse the configuration file. %s' % (
                force_unicode(error),)
            raise ServerException(message)
        finally:
            config_stream.close()

        self._db.storage.read()

        # This can raise an error if node variables file exists
        # but can't be read or had invalid content.
        self._node_config.load(self._configuration_path)

    def _openConfiguration(self):
        """
        Open the configuration from file and return a stream of Unicode.
        """
        configuration_segments = local_filesystem.getSegmentsFromRealPath(
            self._configuration_path)
        if not local_filesystem.isFile(configuration_segments):
            message = (
                'Configuration file "%s" does not exist.' % (
                    self._configuration_path,))
            raise ServerException(message)

        try:
            configuration_file = _utf8_wrap(
                local_filesystem.openFileForReading(
                    configuration_segments))

        except EnvironmentError:
            message = (
                'Server process could not read the configuration file '
                'at %s' % (
                    self._configuration_path,))
            raise ServerException(message)

        return configuration_file

    def _getCacheDB(self):
        """
        Return the cache DB for the currently configured file.
        """
        configuration_segments = local_filesystem.getSegmentsFromRealPath(
            self._configuration_path)

        db_segments = (
            configuration_segments[:-1]
            + [configuration_segments[-1] + '.cache']
            )
        db_path = local_filesystem.getRealPathFromSegments(db_segments)
        db = ChevahTinyDB(db_path, storage=ChevahJSONStorage)

        return db

    def save(self):
        """
        Store the configuration into file.

        By default, it stores the data, in the same *path* from
        which it was loaded from.

        To prevent file corruption, it saves to a temporary file first,
        and then does a rename operation.
        """
        if self._db is None:
            raise ServerException(
                'Configuration must be loaded before a save.')

        self._db.storage.close()
        if self._configuration_file:
            # If we have a testing configuration, just save it
            # without creating the TMP file.
            _write_and_close(
                self._raw_config, store_file=self._configuration_file)
            return

        self._node_config.save()

        real_segments = local_filesystem.getSegmentsFromRealPath(
            self._configuration_path)
        tmp_segments = real_segments[:]
        tmp_segments[-1] = tmp_segments[-1] + '.tmp'
        store_file = local_filesystem.openFileForWriting(tmp_segments)

        _write_and_close(
            self._raw_config, store_file=store_file)
        # We delete the file first to work around windows problems.
        local_filesystem.deleteFile(real_segments)
        local_filesystem.rename(tmp_segments, real_segments)

    def get(self, section, option):
        """
        Always fails as the low level methods should not be called.
        """
        raise NotImplementedError('Use getString instead of get.')

    def getboolean(self, section, option):
        """
        Always fails as the low level methods should not be called.
        """
        raise NotImplementedError('Use getBoolean instead of getboolean.')

    def getfloat(self, section, option):
        """
        Always fails as the low level methods should not be called.
        """
        raise NotImplementedError('Use getfloat instead of getFloat.')

    def hasSection(self, section):
        """
        See `IConfigurationProxy`.
        """
        return self._raw_config.has_section(section)

    def addSection(self, section):
        """
        See `IConfigurationProxy`.
        """
        self._raw_config.add_section(section)

    def removeSection(self, section):
        """
        See `IConfigurationProxy`.
        """
        self._db.remove(where('section') == section)
        return self._raw_config.remove_section(section)

    def renameSection(self, source, destination):
        """
        See `IConfigurationProxy`.
        """
        defaults = self._raw_config.defaults()
        self.addSection(destination)
        for option, value in self._raw_config.items(source):
            if value == defaults[option]:
                # The default values are listed as items
                # and we get all defaults, not only the ones
                # directly configured to the section.
                #
                # Directly defined values are returned with double quotes,
                # so we can tell which is actual missing value,
                # and which is set.
                continue
            self._raw_config.set(destination, option, value)
        self.removeSection(source)

    def removeOption(self, section, option):
        """
        See `IConfigurationProxy`.
        """
        return self._raw_config.remove_option(section, option)

    def hasOption(self, section, option):
        """
        See `IConfigurationProxy'.
        """
        return self._raw_config.has_option(section, option)

    def definesOption(self, section, option):
        """
        See: `IConfigurationProxy'.
        """
        if not self.hasSection(section):
            return False

        return option in self._raw_config._sections[section]

    @property
    def sections(self):
        '''See `IConfigurationProxy`.'''
        return self._raw_config.sections()

    def _get(self, method_name, section, option, type_name=''):
        """
        Helper to get a value for a specific type.
        """
        node_value = self._node_config.get(method_name, section, option)
        if node_value is not None:
            return node_value

        config_method = getattr(self._raw_config, method_name)
        try:
            return config_method(section, option)
        except ValueError as error:
            raise ServerError('20120', data={
                'type': type_name,
                'option': option,
                'section': section,
                'details': force_unicode(error),
                })
        except configparser.NoOptionError:
            # If we ended up here it means that we have not defined a default
            # value for this option.
            error_message = (
                'Configuration file does not have any option "%(option)s" '
                'in section "%(section)s". You must add a default value.' % {
                    'option': option,
                    'section': section,
                    })
            raise ConfigurationException(error_message.encode('utf-8'))
        except configparser.NoSectionError:
            # If we ended up here it means that we have not created the
            # default sections if they were missing.
            error_message = (
                'Configuration file does not define section "%(section)s" '
                'for storing the option "%(option)s". You must create this '
                'section.' % {
                    'option': option,
                    'section': section,
                    })
            raise ConfigurationException(error_message.encode('utf-8'))
        raise RuntimeError(
            'All exceptions should have been previously catch for '
            'section:%s option:%s.' % (section, option))

    def _set(self, converter, section, option, value, type_name=''):
        """
        Helper to set a value for a specific type.
        """
        try:
            converted_value = converter(value)
        except (ValueError, TypeError) as error:
            raise ServerError(
                '20121',
                data={
                    'type': type_name,
                    'value': force_unicode(value),
                    'option': option,
                    'section': section,
                    'details': force_unicode(error),
                    },
                )
        self._raw_config.set(section, option, six.text_type(converted_value))

    def getString(self, section, option):
        """
        See `IConfigurationProxy`.
        """
        value = self._get('get', section, option, 'string')
        if type(value) is not six.text_type:
            value = value.decode('utf-8')
        if value.startswith("'") and value.endswith("'"):
            return value[1:-1]
        if value.startswith('"') and value.endswith('"'):
            return value[1:-1]
        return value

    def getRawString(self, section, option):
        '''See `IConfigurationProxy`.'''
        value = self._get('get', section, option, 'string')
        if type(value) is not six.text_type:
            value = value.decode('utf-8')
        return value

    def getStringOrNone(self, section, option):
        '''See `IConfigurationProxy`.'''
        value = self.getString(section, option)

        if not value:
            return None
        elif self.isDisabledValue(value):
            return None
        else:
            return value

    def isDisabledValue(self, value):
        """
        Return True if value is one of the disabled values.
        """
        if value.lower() in CONFIGURATION_DISABLED_VALUES:
            return True
        else:
            return False

    def getStringOrInherit(self, section, option):
        value = self.getString(section, option)
        if value.lower() in CONFIGURATION_INHERIT:
            return CONFIGURATION_INHERIT[0]
        else:
            return value

    def setString(self, section, option, value):
        """
        See `IConfigurationProxy`.
        """
        value = value.strip(' \t')
        return self._raw_config.set(section, option, value)

    def setStringOrNone(self, section, option, value):
        '''See `IConfigurationProxy`.'''
        if value is None:
            value = CONFIGURATION_DISABLED_VALUE
        return self.setString(section, option, value)

    def setStringOrInherit(self, section, option, value):
        '''See `IConfigurationProxy`.'''
        if value.lower() is CONFIGURATION_INHERIT:
            value = CONFIGURATION_INHERIT[0]
        self.setString(section, option, value)

    def getInteger(self, section, option):
        '''See `IConfigurationProxy`.'''
        return self._get('getint', section, option, 'integer number')

    def getIntegerOrNone(self, section, option):
        '''See `IConfigurationProxy`.'''
        value = self.getString(section, option)
        if self.isDisabledValue(value):
            return None
        else:
            return self.getInteger(section, option)

    def getIntegerOrInherit(self, section, option):
        '''See `IConfigurationProxy`.'''
        value = self.getString(section, option)
        if value.lower() in CONFIGURATION_INHERIT:
            return CONFIGURATION_INHERIT[0]
        else:
            return self.getInteger(section, option)

    def setInteger(self, section, option, value):
        '''See `IConfigurationProxy`.'''
        return self._set(
            int, section, option, value, type_name='integer')

    def setIntegerOrNone(self, section, option, value):
        '''See `IConfigurationProxy`.'''
        if value is None:
            value = CONFIGURATION_DISABLED_VALUE
            return self._raw_config.set(section, option, six.text_type(value))
        else:
            return self.setInteger(section, option, value)

    def getBoolean(self, section, option):
        '''See `IConfigurationProxy`.'''
        return self._get('getboolean', section, option, 'boolean')

    def getBooleanOrInherit(self, section, option):
        """
        See `IConfigurationProxy`.
        """
        value = self.getString(section, option)
        if value.lower() in CONFIGURATION_INHERIT:
            return CONFIGURATION_INHERIT[0]

        return self.getBoolean(section, option)

    def _booleanConverter(self, input):
        """
        Convert input value to a boolean.
        """
        if input in [True, False]:
            return input

        if input in [1, '1']:
            return True

        if input in [0, '0']:
            return False

        try:
            input_lower = input.lower()
        except Exception:
            raise ValueError('Not a boolean value: %s' % (input))

        if input_lower in ['yes', 'true']:
            return True
        if input_lower in ['no', 'false']:
            return False

        raise ValueError('Not a boolean value: %s' % (input))

    def setBoolean(self, section, option, value):
        '''See `IConfigurationProxy`.'''
        return self._set(
            self._booleanConverter,
            section,
            option,
            value,
            type_name='boolean',
            )

    def setBooleanOrInherit(self, section, option, value):
        '''See `IConfigurationProxy`.'''
        if (
            isinstance(value, six.string_types) and
                value.lower() in CONFIGURATION_INHERIT
                ):
            self.setString(
                section, option, CONFIGURATION_INHERIT[0])
        else:
            self.setBoolean(section, option, value)

    def getFloat(self, section, option):
        """
        See `IConfigurationProxy`.
        """
        return self._get('getfloat', section, option, 'floating number')

    def setFloat(self, section, option, value):
        """
        See `IConfigurationProxy`.
        """
        return self._set(float, section, option, value, 'floating number')

    def getCachedValue(self, section, option):
        """
        Return the value from cache, or None if the value doesn't exist.
        """
        docs = self._db.search(where('section') == section)
        if not docs:
            return None
        return docs[0].get(option, None)

    def setCachedValue(self, section, option, value):
        """
        Set the value to cache.
        """
        docs = self._db.search(where('section') == section)
        if not docs:
            new_doc = {'section': section}
            new_doc[option] = value
            self._db.insert(new_doc)
            return

        self._db.update(
            operations.set(option, value),
            where('section') == section,
            )


def _write_and_close(raw_config, store_file):
    """
    Write serialized raw INI configuration to a file stream
    and close the stream at the end.
    """
    try:
        for section in raw_config._sections:
            result = []
            result.append(('[%s]\n' % section))
            for (key, value) in raw_config._sections[section].items():
                if key != '__name__':
                    result.append(('%s = %s\n' % (
                        key, value.replace('\n', '\n\t'))
                        ))
            result.append('\n')
            result = ''.join(result)

            if not isinstance(store_file, StringIO):
                result = result.encode('utf-8')
            store_file.write(result)
    finally:
        store_file.close()


# Filename extensions used to store the node variables associated with
# the main configuration file.
_NODE_EXTENSION = '.override'


@implementer(INodeConfiguration)
class NodeConfigurationFile(object):
    """
    This is here to extract the node variables part from the main configuration
    class.

    It also provides high-level API that is not related to .INI format
    via `INodeConfiguration`.
    """

    def __init__(self):
        # Store any possible load time errors.
        self._error = ''
        self._node_segments = None
        self._raw_config = ChevahRawConfigParser()

    def load(self, configuration_path):
        """
        Load the associated node variables for the main `configuration_path`.
        """

        if not configuration_path:
            # This is called during unit testing and those unit tests
            # should not use node variables.
            return

        configuration_segments = local_filesystem.getSegmentsFromRealPath(
            configuration_path)

        node_filename = configuration_segments[-1]
        try:
            node_filename, extension = node_filename.rsplit('.', 1)
            extension = '.' + extension
        except ValueError:
            extension = ''

        node_filename = node_filename + _NODE_EXTENSION + extension

        configuration_segments[-1] = node_filename
        self._node_segments = configuration_segments

        if not local_filesystem.isFile(self._node_segments):
            # If no file, we just return an empty stream.
            return BytesIO()

        try:
            node_stream = local_filesystem.openFileForReading(
                self._node_segments)

        except EnvironmentError:
            path = local_filesystem.getRealPathFromSegments(
                self._node_segments)
            message = (
                'Server process could not read the node configuration file '
                'at %s' % (path,))
            raise ServerException(message)

        self._loadINI(node_stream)

    def save(self):
        """
        Save the in-memory configuration changes to a file.
        """
        if not self._node_segments:
            # Called during unit testing.
            # Nothing to save.
            return

        if not self._raw_config.sections():
            # Nothing to save.
            # Delete any already existing file.
            if local_filesystem.exists(self._node_segments):
                local_filesystem.deleteFile(self._node_segments)
            return

        store_file = local_filesystem.openFileForWriting(self._node_segments)
        _write_and_close(self._raw_config, store_file=store_file)

    def resetINI(self, content):
        """
        Update the internal state based on new `content`.

        See: INodeConfiguration
        """
        node_stream = BytesIO(content.encode('utf-8'))
        self._loadINI(node_stream)

    def _loadINI(self, node_stream):
        """
        Loads the stream in INI format and closes the file.
        """
        self._raw_config = ChevahRawConfigParser()
        try:
            self._raw_config.read_file(_utf8_wrap(node_stream))
        except (configparser.ParsingError, AttributeError) as error:
            message = (
                'Could not parse the node configuration file. %s' % (
                    force_unicode(error),))
            raise ServerException(message)
        finally:
            node_stream.close()

    def get(self, method_name, section, option):
        """
        Return the stored value for `section` and `option` with the
        type value converted based on `method_name`

        `method` name is to allow number or boolean conversion.
        For the long term we should refactor the low-level file configuration
        to only handle text.

        Returns `None` if there is no value stored.
        """
        node_method = getattr(self._raw_config, method_name)
        try:
            return node_method(section, option)
        except (
            configparser.NoSectionError,
            configparser.NoOptionError,
                ):
            """
            We don't have it in the node configuration.
            """
            return None

    def getNodeConfiguration(self):
        """
        Returns a dictionary with the current node variables.

        This is designed to be used at an intermediary level.

        See: INodeConfiguration
        """
        result = {}
        for section in self._raw_config.sections():
            result[section] = {}
            for option in self._raw_config.options(section):
                result[section][option] = self._raw_config.get(
                    section, option)
        return result

    def setNodeVariable(self, section, option, value):
        """
        Set the value to the separate node configuration file.

        See: INodeConfiguration
        """
        if not self._raw_config.has_section(section):
            self._raw_config.add_section(section)

        self._raw_config.set(section, option, value)

    def removeNodeVariable(self, section, option):
        """
        Remove the value for `option` from `section`.

        When section is left empty, this also removes the section.

        See: INodeConfiguration
        """
        try:
            self._raw_config.remove_option(section, option)
        except configparser.NoSectionError:
            return

        # Remove the whole section if no other options are there.
        if not self._raw_config.options(section):
            self._raw_config.remove_section(section)
