# Copyright (c) 2011-2013 Adi Roiban.
# See LICENSE for details.
"""
Properties and sections module.

Sections are group of properties and could also contain other sections.
Sections are exposing a read / write CRUD interface.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
import collections
import re

from dateutil.tz import tzoffset
import dateutil.parser
import six
from zope.interface import implementer

from chevah.server.commons.constant import (
    CONFIGURATION_INHERIT,
    _INHERIT_MARKER,
    )
from chevah.server.commons.csv import from_csv_row
from chevah.server.commons.exception import (
    CreateNotSupportedError,
    DeleteNotSupportedError,
    NoSuchAttributeError,
    NoSuchSectionError,
    ServerError,
    ServerException,
    )
from chevah.server.commons.interface import (
    IProperty,
    PublicAttribute,
    PublicSectionAttribute,
    PublicSyncAttribute,
    PublicWritableAttribute,
    SecretSyncAttribute,
    )
from chevah.server.commons.observer import ObserverMixin, Signal


def is_inherited(value):
    """
    Return True if value is inherited.

    It works with primary values or single-level sequences.
    """

    def check_single_value(value):
        """
        Return True if `value` is inherited marker.
        """
        if value is _INHERIT_MARKER:
            return True

        if value is None:
            return False

        if not isinstance(value, str) and not isinstance(value, six.text_type):
            return False

        if value.strip().lower() in CONFIGURATION_INHERIT:
            return True

        return False

    if check_single_value(value):
        # We know that is inherited.
        return True

    # Continue with checking a sequence of values.
    if isinstance(value, collections.Sequence):
        for member in value:
            if check_single_value(member):
                return True

    return False


def is_inherited_list_of_lists(target):
    """
    Return `True` if a sequence of sequence contains the inherit value.

    If 'inherit' marker is mixed with other values, it is ignored.
    """
    if not target:
        return False

    inherited = False

    for candidate in target:
        if not candidate:
            # For now, ignore empty lines.
            continue

        if is_inherited(candidate[0]):
            inherited = True
            continue

        return False

    return inherited


def get_attribute_names(instance, kind):
    """
    Iterate over all interfaces and return the name of the attributes
    that are instances of `kind`, where `kind` can be a class or tuple of
    classes.

    Interface can inherit other interfaces and multiple interface can
    overwrite each others attributes.

    !!!WARNING!!! don't overly rely on interface inheritance, as zope
    interface doesn't keep a precise order of inheritance.
    """
    result = []

    interfaces = list(instance.__implemented__.__iro__)
    interfaces.reverse()

    for interface in interfaces:
        for name, description in interface.namesAndDescriptions():

            if six.PY2:
                name = name.decode('utf-8')

            if name in result and not isinstance(description, kind):
                result.remove(name)
                continue

            if isinstance(description, kind):
                result.append(name)

    # We sort to have deterministic order of attributes.
    return sorted(result)


def get_relaxed_datetime(raw_value, usage):
    """
    Returns a datetime from and ISO date or ISO datetime text raw_value.

    Returns None if no raw_value was provided.
    """
    if not raw_value:
        return None

    if not re.match(r'^\d\d\d\d-\d\d-\d\d.*', raw_value):
        raise ServerException(
            '%s date and time needs year, month, and day.' % (usage,))

    try:
        datetime = dateutil.parser.parse(raw_value)
    except ValueError:
        raise ServerException(
            '%s date and time format is not a valid '
            'ISO 8601 format.' % (usage,)
            )

    if datetime.tzinfo is None:
        # Force local timezone so that we always have zone aware values.
        datetime = datetime.replace(tzinfo=tzoffset(None, 0))

    return datetime


def WritableString(name, default=None):
    """
    String configuration which can be written.
    """
    def getter(self):
        result = self._proxy.getString(self._section_name, name)
        if not result and default is not None:
            return default
        return result

    def setter(self, value):
        return self._proxy.setString(self._section_name, name, value)

    return property(getter, setter)


def WritableRawString(name):
    """
    String configuration which can be written.
    """
    def getter(self):
        return self._proxy.getRawString(self._section_name, name)

    def setter(self, value):
        return self._proxy.setString(self._section_name, name, value)

    return property(getter, setter)


def WritableLines(name, default=None):
    """
    String configuration which store values on multiple lines.

    Each line is an element from an list.

    Empty lines are ignore.
    """
    def getter(self):
        result = self._proxy.getString(self._section_name, name).strip()
        if not result and default is not None:
            return default

        return tuple([
            line.strip() for line in result.splitlines() if line.strip()])

    def setter(self, value):
        serialization = []
        for line in value:
            line = line.strip()
            if not line:
                continue
            serialization.append(line)

        return self._proxy.setString(
            self._section_name, name, '\n'.join(serialization))

    return property(getter, setter)


def WritableSSVLines(name, default=()):
    """
    Configuration stored as multiple lines.
    Each lines containing multiple space-separated values.
    """
    def getter(self):
        raw_value = self._proxy.getString(self._section_name, name).strip()
        if not raw_value:
            return default

        result = []
        for line in raw_value.splitlines():
            line = line.strip()
            if not line:
                continue
            result.append(tuple([
                p.strip() for p in re.split('[\\t ]', line) if p.strip()]))

        return tuple(result)

    def setter(self, value):
        serialization = []
        for line in value:
            line = ' '.join(line).strip()
            if not line:
                continue
            serialization.append(line)

        return self._proxy.setString(
            self._section_name, name, '\n'.join(serialization))

    return property(getter, setter)


def WritableStringOrEmpty(name, default='', inherit=False):
    """
    String configuration which can be written and can be disabled as empty,
    or disabled via the explicit `disabled` value.

    WARNING: Using this in the configuration is considered bad design.
             Try to use `WritableString`.
             This is here mainly to support backward compatibility.
             It should not be used for new configuration options.
    """
    def getter(self):
        value = self._proxy.getString(self._section_name, name)
        if self._proxy.isDisabledValue(value):
            return ''

        if not value:
            return default

        if inherit and is_inherited(value):
            return CONFIGURATION_INHERIT[0]

        if not inherit and is_inherited(value):
            # FIXME:5494:
            # This should be removed once we no longer have inherit as default
            # placeholder.
            return default

        return value

    def setter(self, value):
        return self._proxy.setStringOrNone(self._section_name, name, value)

    return property(getter, setter)


def WritableStringCaseInsensitive(name, default=None):
    """
    String configuration which can be written and always in lowercase.
    """
    if isinstance(default, dict):
        default = default[name]

    def getter(self):
        result = self._proxy.getString(self._section_name, name).lower()
        if not result and default is not None:
            return default
        return result

    def setter(self, value):
        return self._proxy.setString(self._section_name, name, value.lower())

    return property(getter, setter)


def WritableStringOrEmptyCaseInsensitive(name):
    """
    String configuration which can be empty when disabled and is
    always in lowercase.
    """
    def getter(self):
        value = self._proxy.getStringOrNone(self._section_name, name)
        if value is None:
            return ''
        else:
            return value.lower()

    def setter(self, value):
        if value is not None:
            value = value.lower()
        return self._proxy.setStringOrNone(
            self._section_name, name, value)

    return property(getter, setter)


def WritableCSV(name):
    """
    A sequence of strings serialized as comma separated values.

    Note that duplicate items are kept.

    Items shouldn't contain comma or leading/trailing spaces.
    """
    def getter(self):
        value = self._proxy.getString(self._section_name, name)
        return _convert_string_to_tuple(value)

    def setter(self, value):
        value = _convert_ordered_sequence_to_string(value)
        self._proxy.setString(self._section_name, name, value)

    return property(getter, setter)


def WritableCSVInsensitive(name, default=(), inherit=True):
    """
    A sequence of strings serialized as comma separated values in
    lower case.

    When `inherit` is False, setting this to an inherited value will use
    the default value.

    `default` values should be all lower case as no extra logic is done.

    Items shouldn't contain comma or leading/trailing spaces.
    """
    def getter(self):
        value = self._proxy.getString(self._section_name, name)

        result = _convert_string_to_tuple(value.lower())
        if not result:
            return default

        if not inherit and is_inherited(result):
            # FIXME:5494:
            # This should be removed once we no longer have inherit as default
            # placeholder.
            return default

        return result

    def setter(self, value):
        if not inherit and is_inherited(value):
            # FIXME:5494:
            # This should be removed once we no longer have inherit as default
            # placeholder.
            value = default

        value = _convert_ordered_sequence_to_string(value)
        self._proxy.setString(self._section_name, name, value.lower())

    return property(getter, setter)


def WritableCSVUnique(name, default=(), inherit=True):
    """
    A sequence of strings serialized as comma separated values.

    Items should not contain comma or leading/trailing spaces.

    When set to an inherit configuration value, but inheritance is not
    supported it will return the default value.

    Duplicate values are ignored.
    """
    def getter(self):
        value = self._proxy.getString(self._section_name, name)

        result = _convert_string_to_non_duplicate_tuple(value)

        if not result:
            return default

        if not inherit and is_inherited(result):
            # FIXME:5494:
            # This should be removed once we no longer have inherit as default
            # placeholder.
            return default

        return result

    def setter(self, value):

        if not inherit and is_inherited(value):
            # FIXME:5494:
            # This should be removed once we no longer have inherit as default
            # placeholder.
            value = default

        value = _convert_ordered_sequence_to_non_duplicate_string(value)
        self._proxy.setString(self._section_name, name, value)

    return property(getter, setter)


def WritableCSVUniqueInsensitive(name):
    """
    A sequence of strings serialized as comma separated values, ignoring cases.

    All items are stored as lower case.

    Duplicate values are ignored.
    """
    def getter(self):
        value = self._proxy.getString(self._section_name, name)
        return _convert_string_to_non_duplicate_tuple(value.lower())

    def setter(self, value):
        value = _convert_ordered_sequence_to_non_duplicate_string(
            value).lower()
        self._proxy.setString(self._section_name, name, value)

    return property(getter, setter)


def WritableCSVLines(name, default=(), inherit=True):
    """
    A sequence of sequences of strings serialized as comma separated values,
    each on its own line.

    Items should not contain comma or leading/trailing spaces.
    """
    def getter(self):
        result = []
        lines = self._proxy.getString(self._section_name, name)

        for line in lines.splitlines():
            if not line:
                continue

            values = _convert_string_to_tuple(line)

            if not inherit and is_inherited(values):
                # FIXME:5494:
                # This should be removed once we no longer have inherit as
                # default placeholder.
                return default

            result.append(values)

        if not result:
            return default

        return tuple(result)

    def setter(self, value):
        result = []
        for sequence in value:
            if not sequence:
                continue
            result.append(_convert_ordered_sequence_to_string(sequence))
        self._proxy.setString(self._section_name, name, '\n'.join(result))

    return property(getter, setter)


def WritableCSVLinesUnique(name, default=()):
    """
    A sequence of sequences of strings serialized as comma separated values,
    each on its own line.

    Items should not contain comma or leading/trailing spaces.

    Duplicate values are ignored.
    """
    def getter(self):
        result = []
        lines = self._proxy.getString(self._section_name, name)

        for line in lines.splitlines():
            if not line:
                continue
            result.append(_convert_string_to_non_duplicate_tuple(line))

        if not result:
            return default

        return tuple(result)

    def setter(self, value):
        result = []
        for sequence in value:
            if not sequence:
                continue
            result.append(_convert_ordered_sequence_to_non_duplicate_string(
                sequence))
        self._proxy.setString(self._section_name, name, '\n'.join(result))

    return property(getter, setter)


def DefaultOnly(default):
    """
    A property that only returns the default value.

    It allows writing, but any write is ignored.

    Warning. Before using this, consider defining the interfaces in such a
    way that a write operation is not neeed.

    This exists to simplify the node sync code so that we don't have to
    add extra information for readonly public attributes as part of the
    interface.
    """
    def getter(self):
        return default

    def setter(self, value):
        """
        Setting is ignored.
        """

    return property(getter, setter)


def _string_to_tuple(value, allow_duplicates, allow_empty):
    """
    Return a tuple representation of string breaking at commas.

    If `allow_duplicates` is False duplicate values are ignored.
    """
    if not value:
        # In case of empty string, returns an empty tuple.
        return ()

    value = value.replace('\r', '').replace('\n', '')
    values = []
    for item in from_csv_row(value):
        item = item.strip()
        if not allow_empty and not item.strip():
            # Ignore empty.
            continue
        if not allow_duplicates and item in values:
            # Ignore duplicate items.
            continue
        values.append(item)

    return tuple(values)


def _convert_string_to_non_duplicate_tuple(value):
    """
    Return a tuple representation of string breaking at commas.

    Duplicate values are ignored.
    """
    return _string_to_tuple(value, allow_duplicates=False, allow_empty=False)


def _convert_string_to_tuple(value):
    """
    Return a tuple representation of string break at commas.
    """
    return _string_to_tuple(value, allow_duplicates=True, allow_empty=True)


def _ordered_sequence_to_string(value, allow_duplicates, allow_empty):
    """
    Return the string serialization of `value` sequence.

    Elements are separated by comma.

    If `allow_duplicates` is False duplicate values are ignored.
    """
    if isinstance(value, set):
        raise ServerException('Sets are not ordered sequences.')

    filtered_values = []
    for item in value:
        item = six.text_type(item).strip()
        if not allow_empty and not item:
            # Ignore empty.
            continue
        if ',' in item:
            item = '"%s"' % (item.replace('"', '""'),)
        if not allow_duplicates and item in filtered_values:
            # Ignore duplicates.
            continue
        filtered_values.append(item)

    return ', '.join(filtered_values)


def _convert_ordered_sequence_to_string(value):
    """
    Return the string serialization of `value` sequence.

    Elements are separated by comma and duplicate values are kept.
    """
    return _ordered_sequence_to_string(
        value, allow_duplicates=True, allow_empty=True)


def _convert_ordered_sequence_to_non_duplicate_string(value):
    """
    Return the string serialization of `value` sequence.

    Elements are separated by comma and duplicate values are ignored.
    """
    return _ordered_sequence_to_string(
        value, allow_duplicates=False, allow_empty=True)


def WritableBoolean(name, default=True):
    """
    Boolean configuration which can be written.
    """
    def getter(self):
        try:
            return self._proxy.getBoolean(self._section_name, name)
        except ServerError as error:
            if error.id == '20120':
                return default
            raise error

    def setter(self, value):
        return self._proxy.setBoolean(self._section_name, name, value)

    return property(getter, setter)


def WritableBooleanWithInherit(name, default=CONFIGURATION_INHERIT[0]):
    """
    Boolean configuration which can be inherited.
    """
    def getter(self):

        result = self._proxy.getString(self._section_name, name)
        if not result:
            return default

        try:
            return self._proxy.getBooleanOrInherit(self._section_name, name)
        except ServerError as error:
            if error.id == '20120':
                return default
            raise error

    def setter(self, value):
        return self._proxy.setBooleanOrInherit(self._section_name, name, value)

    return property(getter, setter)


def WritableInteger(name, default):
    """
    Integer number configuration which can be written.

    `default` can be default value or a dict from which the default is retrieved
    based on `name`.

    FIXME:2378:
    Using it should be avoided and instead use WritableNumber.
    There is also a lot of duplicate code.
    """
    if isinstance(default, dict):
        default = default[name]

    def getter(self):
        result = self._proxy.getString(self._section_name, name)
        if not result:
            return default

        try:
            result = int(result.split('.', 1)[0])
        except ValueError:
            self.root.emitEvent('20190', data={
                'invalid_value': result,
                'section': self._section_name,
                'option': name,
                'default': default,
                'details': 'Not an integer number.',
                })
            result = default
            self._proxy.setString(self._section_name, name, str(result))

        return result

    def setter(self, value):
        # FIXME:5707:
        # We should validate the value and raise an error for invalid values.
        return self._proxy.setString(self._section_name, name, str(value))

    return property(getter, setter)


def WritableIntegerWithInherit(name, default=0):
    """
    Integer number configuration which can be written and can be inherited.

    When  value is empty, the default value is used.

    It can read or set as an Inherited value.
    """
    def getter(self):
        value = self._proxy.getString(self._section_name, name)
        if is_inherited(value):
            return CONFIGURATION_INHERIT[0]

        if value == '':
            return default

        try:
            return int(value)
        except Exception:
            # value is not number.
            # Just fallback to default.
            return default

    def setter(self, value):
        if is_inherited(value):
            return self._proxy.setString(
                self._section_name, name, CONFIGURATION_INHERIT[0])

        if value == '':
            value = default

        return self._proxy.setInteger(self._section_name, name, value)

    return property(getter, setter)


def WritableNumber(name, default, allow_negative=True, allow_zero=True):
    """
    Number configuration which can be written.

    When `allow_zero` is false and configured value is `0`,
    the default value is used.

    A number is an integer or a float. It can be negative.

    When `default` is a dict, the value is returned from `default[name]`.
    """
    if isinstance(default, dict):
        default = default[name]

    def getter(self):
        result = self._proxy.getString(self._section_name, name)
        if not result:
            return default

        try:
            result = float(result)
        except ValueError:
            self.root.emitEvent('20190', data={
                'invalid_value': result,
                'section': self._section_name,
                'option': name,
                'default': default,
                'details': 'Not a number.',
                })
            result = default
            self._proxy.setFloat(self._section_name, name, result)

        if not allow_zero and result == 0:
            return default

        if not allow_negative and result < 0:
            return default

        return result

    def setter(self, value):
        # FIXME:5707:
        # We should fail when trying to set an invalid float.
        # Also fail on large number as the precision is lost.
        return self._proxy.setFloat(self._section_name, name, value)

    return property(getter, setter)


def CachedValue(name, default):
    """
    Value stored in the configuration cache.
    """
    def getter(self):
        result = self._proxy.getCachedValue(self._section_name, name)
        if not result:
            return default
        return result

    def setter(self, value):
        return self._proxy.setCachedValue(self._section_name, name, value)

    return property(getter, setter)


@implementer(IProperty)
class PropertyMixin(ObserverMixin):
    """
    Mixin for object exporting public properties.

    See `IProperty`.

    Emits the following signals:

    * update-PROPERTY_NAME - right after a property was changed
    * update - with a delay after all properties were changed
    """

    # Used for waiting for all updates to be done and raise a single signal.
    _UPDATE_DELAY = 0.5
    __update_delayed_call = None
    _parent = None

    @property
    def parent(self):
        """
        See: `IProperty`.
        """
        return self._parent

    @property
    def root(self):
        """
        See: `IProperty`.
        """
        root = self
        parent = self.parent
        while parent:
            root = parent
            parent = root.parent
        return root

    @property
    def reactor(self):
        """
        See: `IProperty`.
        """
        return self.parent.reactor

    def getPublicAttributeNames(self):
        """
        Return the list of all names of public attributes.
        """
        return get_attribute_names(self, PublicAttribute)

    def getPublicWritableAttributeNames(self):
        """
        Return the list of all names of public attributes.
        """
        return get_attribute_names(self, PublicWritableAttribute)

    def getSyncAttributeNames(self):
        """
        Return the list of all names of attributes that can be written
        as part of the sync operation.
        """
        return get_attribute_names(self, (
            PublicWritableAttribute, SecretSyncAttribute, PublicSyncAttribute))

    def getPublicSectionNames(self):
        """
        Return the list of all names for public sections.
        """
        return get_attribute_names(self, PublicSectionAttribute)

    def getAll(self):
        """
        See: `IProperty`.
        """
        result = []
        for uuid in self.getPublicSectionNames():
            section = self.getSection(uuid)
            result.append(section)
        return result

    def getValue(self, name):
        """
        Return the value of attribute with `name`.

        Raise MissingAttributeError if leaf does not exist.
        """
        try:
            return getattr(self, name)
        except AttributeError:
            raise NoSuchAttributeError(name)

    def setValue(self, name, value):
        """
        Set value for attribute with name.
        """
        old_value = self._setValueWithoutNotification(name, value)
        signal_name = 'update-%s' % (name)
        self.notify(
            signal_name, Signal(
                source=self, new_value=value, old_value=old_value))
        self._delayedNotifyUpdate()

    def _setValueWithoutNotification(self, name, value):
        """
        Set value for attribute with name returning the old value.
        """
        old_value = None
        try:
            old_value = getattr(self, name)
        except AttributeError:
            raise NoSuchAttributeError(name)
        except ServerError:
            # If the old value is invalid, we just ignore that error and
            # hopefully we will set a new valid value.
            pass

        setattr(self, name, value)
        return old_value

    def _delayedNotifyUpdate(self):
        """
        Wait `_UPDATE_DELAY` before emitting the `update` signal to buffer
        all pending requests.EmptyProperty
        """
        def notify_update():
            """
            Emit the update signal.
            """
            self.notify('update', Signal(source=self))
            self.__update_delayed_call = None

        if self._UPDATE_DELAY == 0:
            # When delay is not set, just call right away.
            # Here to help testing.
            notify_update()
        elif (
            not self.__update_delayed_call or
            not self.__update_delayed_call.active()
                ):
            self.__update_delayed_call = self.reactor.callLater(
                self._UPDATE_DELAY, notify_update)
        else:
            self.__update_delayed_call.reset(self._UPDATE_DELAY)

    def getSection(self, name):
        """
        Return the branch with `name`.

        Raise NoSuchSectionError if branch does not exist.
        """
        try:
            return getattr(self, name)
        except AttributeError:
            raise NoSuchSectionError(name)

    def traversePath(self, property_path):
        """
        Parse property_path and return head and tail.
        """
        if property_path:
            try:
                head, tail = property_path.split('/', 1)
            except ValueError:
                head = property_path
                tail = None
        else:
            head = None
            tail = None

        if head == '':
            head = None

        if tail == '':
            tail = None

        return (head, tail)

    def getProperty(self, property_path=None):
        """
        See `IProperty`.
        """
        def public_getter(target):
            return get_attribute_names(target, PublicAttribute)

        return self.getFilteredProperty(public_getter, property_path)

    def getPropertyForSync(self, property_path=None):
        """
        Return the configuration values that can be used for synchronization.
        """
        def sync_getter(target):
            return get_attribute_names(target, (
                PublicWritableAttribute,
                PublicSyncAttribute,
                SecretSyncAttribute),
                )

        return self.getFilteredProperty(sync_getter, property_path)

    def getFilteredProperty(self, getter, property_path):
        """
        See `IProperty`.
        """
        head, tail = self.traversePath(property_path)

        result = {}

        # Look for direct attributes.
        for name in getter(self):
            value = self.getValue(name)

            # Return direct property.
            if head and head == name:
                return value

            result.update({name: value})

        # Then for sections.
        for name in self.getPublicSectionNames():
            value = self.getSection(name)
            section_properties = value.getFilteredProperty(getter, tail)

            # Return direct subsections.
            if head and head == name:
                return section_properties

            result.update({name: section_properties})

        return result

    def setProperty(self, property_path, value):
        """
        See `IProperty`.
        """
        return self._setProperty(property_path, value, notify=True)

    def setPropertyWithoutNotification(self,  property_path, value):
        """
        See `IProperty`.
        """
        return self._setProperty(property_path, value, notify=False)

    def _setProperty(self, property_path, value, notify=True):
        """
        Low level implementation for setting property.
        """
        head, tail = self.traversePath(property_path)

        if tail is None:
            for name in self.getPublicWritableAttributeNames():
                if name != head:
                    continue

                if notify:
                    self.setValue(head, value)
                else:
                    self._setValueWithoutNotification(head, value)
                return

            # If we are here, it means that property was not found or it is
            # not writable.
            # We raise not found error for a better security.
            raise NoSuchAttributeError(head)
        else:
            for name in self.getPublicSectionNames():
                if name != head:
                    continue

                section = self.getSection(name)
                section._setProperty(tail, value, notify)
                # If we are here it means that we set a child, so we inform
                # the parent about the upate.
                self._delayedNotifyUpdate()
                # We stop the search here.
                return

            # If we are here, it means that section was not found.
            raise NoSuchSectionError(head)

    def deleteProperty(self, property_path):
        """
        See `IProperty`.
        """
        if property_path is None:
            raise DeleteNotSupportedError()

        head, tail = self.traversePath(property_path)

        # Right now only sections support delete operation.
        for name in self.getPublicSectionNames():
            if name != head:
                continue

            section = self.getSection(name)
            return section.deleteProperty(tail)

        # If we are here, it means that section was not found.
        raise NoSuchSectionError(head)

    def createProperty(self, property_path, value):
        """
        See `IProperty`.
        """
        if property_path is None:
            raise CreateNotSupportedError()

        head, tail = self.traversePath(property_path)

        # Right now only sections support add operations.
        for name in self.getPublicSectionNames():
            if name != head:
                continue

            section = self.getSection(name)
            return section.createProperty(tail, value)

        # If we are here, it means that section was not found.
        raise NoSuchSectionError(head)


class EmptyProperty(PropertyMixin):
    """
    A property which has only empty direct values and no sub-members.
    """

    def __init__(self, parent, create_message):
        self._parent = parent
        self._proxy = parent._proxy
        self._create_message = create_message

    def getProperty(self, path):
        return {}

    def createProperty(self, property_path, value):
        """
        See `IProperty`.
        """
        raise CreateNotSupportedError(self._create_message)
