# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
Common configuration options.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
import time

from zope.interface import implementer
from zxcvbn import zxcvbn

from chevah.server import force_unicode
from chevah.server.authentication.base import password_handler
from chevah.server.commons.constant import (
    DEFAULT_LOCAL_FILESYSTEM_UUID,
    PORT_MAX,
    PORT_MIN,
    )
from chevah.server.commons.exception import (
    ConfigurationException,
    ServerError,
    ServerException,
    )
from chevah.server.commons.interface import (
    IAddressPortPublicOptions,
    IAutoFailOptions,
    IEntityConfiguration,
    ILocationWatchConfiguration,
    IPasswordOptions,
    IUsernamePasswordPublicOptions,
    ITypeOptions,
    IURLOptions,
    )
from chevah.server.commons.observer import Signal
from chevah.server.configuration.model import (
    CachedValue,
    PropertyMixin,
    WritableBoolean,
    WritableInteger,
    WritableNumber,
    WritableString,
    WritableStringOrEmpty,
    )
import six


@implementer(IEntityConfiguration)
class EntityConfiguration(PropertyMixin):
    """
    Common attributes for a .INI configuration section.

    Must define the following class members:
    * SECTION_PREFIX = 'logs/'
    """
    def __init__(self, parent, uuid):
        self._uuid = six.text_type(uuid)
        self._section_name = self.SECTION_PREFIX + uuid
        self._parent = parent
        self._proxy = parent._proxy
        self._migrateINI()

    def __repr__(self):
        return '%d:%s: %s - %s' % (
            id(self), self.__class__, self.name, self.uuid)

    @property
    def enabled(self):
        """
        See: `IEntityConfiguration`.
        """
        return self._proxy.getBoolean(self._section_name, 'enabled')

    @enabled.setter
    def enabled(self, value):
        return self._proxy.setBoolean(
            self._section_name, 'enabled', value)

    @property
    def uuid(self):
        """
        See: `IEntityConfiguration`.
        """
        return self._uuid

    @property
    def name(self):
        """
        See: `IEntityConfiguration`.
        """
        result = self._proxy.getString(self._section_name, 'name')
        return result

    @name.setter
    def name(self, value):
        value = value.strip()
        for component in self.parent.getAll():
            if component.uuid == self.uuid:
                # This is us.
                # So it's ok if somehow we set the same name.
                continue
            if component.name.lower() == value.lower():
                raise ConfigurationException(
                    'Name "{}" already exists '
                    'for component with UUID "{}".'.format(
                        value, component.uuid)
                    )
        self._setName(value)

    def _setName(self, value):
        """
        Set the name value without any validation.
        """
        return self._proxy.setString(self._section_name, 'name', value)

    @property
    def description(self):
        """
        See: `IEntityConfiguration`.
        """
        return self._proxy.getString(self._section_name, 'description')

    @description.setter
    def description(self, value):
        return self._proxy.setString(
            self._section_name, 'description', value)

    def _updateWithNotify(self, setter, name, value):
        """
        Update configuration and notify changes.

        Revert configuration on error.
        """
        initial_value = getattr(self, name)
        setter(self._section_name, name, value)
        current_value = getattr(self, name)

        signal = Signal(
            self, initial_value=initial_value, current_value=current_value)
        try:
            self.notify(name, signal)
        except Exception:
            setter(
                self._section_name, name, initial_value)
            raise

    def _migrateINI(self):
        """
        Standard place to migrate any configuration related to this component.
        """


@implementer(IPasswordOptions)
class PlainPasswordConfigurationMixin(object):
    """
    Configuration entities with a password which is stored in plain text.
    """

    # Since the password is hidden to the public API is hard for a
    # runnable to know when a restart is required.
    # To help with that, we do a simple heuristic,
    # assuming that as long as password was updated but not read,
    # a restart is required.
    _password_updated = False

    _now = time.time

    last_password_update = CachedValue('last_password_update', default=0)

    @property
    def last_passwords(self):
        """
        Password history is not used for plain password, but is here to have
        the same interface.
        """
        return ()

    @property
    def password(self):
        """
        Get password.
        """
        self._password_updated = False
        value = self._proxy.getStringOrNone(self._section_name, 'password')
        if value is None or value.strip() == '':
            return ''
        else:
            return value

    @password.setter
    def password(self, value):
        """
        Set password.
        """
        raise RuntimeError(
            'Use password_update or in tests _setRawPassword() for '
            'changing the password.'
            )

    def _setRawPassword(self, value):
        """
        Helper to set raw password value and bypass hashing in tests.
        """
        return self._proxy.setStringOrNone(
            self._section_name, 'password', value)

    @property
    def password_update(self):
        """
        This is a write-only attribute.
        """
        if self._password_updated:
            return 'restart-required'

        return ''

    @password_update.setter
    def password_update(self, value):
        """
        Set password as clear text value.
        """
        self._password_updated = True
        try:
            return self._proxy.setStringOrNone(
                self._section_name, 'password', value)
        finally:
            self.last_password_update = self._now()


@implementer(IUsernamePasswordPublicOptions)
class UsernamePasswordPublicOptionsMixin(PlainPasswordConfigurationMixin):
    """
    Configuration for username with password in plain text.
    """
    username = WritableStringOrEmpty('username')


class PasswordConfigurationMixin(PlainPasswordConfigurationMixin):
    """
    Configuration entities with a password which is stored as a hash crypt.
    """

    _now = time.time

    multi_factor_authentication = WritableString('multi_factor_authentication')
    last_password_update = CachedValue('last_password_update', default=0)
    last_passwords = CachedValue('last_passwords', default=())

    # List with last TOTP accepted for this account.
    # This is not exposed to the public API.
    last_totps = None

    @property
    def password_update(self):
        """
        This is a write-only attribute.
        """
        return ''

    @password_update.setter
    def password_update(self, value):
        """
        Set password by hashing the clear text value.

        Setting to empty string, string with only spaces or `disabled` values,
        will disable the password.
        """
        raw_value = None

        def get_error(details):
            """
            Return the error for failing to set the password.
            """
            return ServerError('20064', data={'details': details})

        def check_history(plain, new_encrypted, history_length):
            """
            Check that password was not used in the past.

            Returns the new password history data.
            """
            if history_length < 1:
                return []

            current_password = []
            if self.password:
                current_password = [self.password]

            history = current_password + list(self.last_passwords)
            for existing in history:
                try:
                    if password_handler.verify(plain, existing):
                        raise get_error('Password already used in the past.')
                except ValueError:
                    """
                    We shouldn't end up here as cache should only have
                    valid values.
                    But we just ignore it.
                    """

            return [new_encrypted] + history[:history_length - 1]

        if (
            not value
            or not value.strip()
            or self._proxy.isDisabledValue(value)
                ):
            # Password set as disabled.
            self._proxy.setStringOrNone(self._section_name, 'password', None)
            self.last_password_update = self._now()
            return

        server_section = self.root.configuration.server
        minimum_length = server_section.password_minimum_length
        minimum_strength = server_section.password_minimum_strength

        if len(value) < minimum_length:
            raise get_error(
                'Password is shorter than %s.' % (minimum_length,))

        current_strength = zxcvbn(value, user_inputs=self._getUserInputs())
        if current_strength['score'] < minimum_strength:
            raise get_error(
                'Password is weaker than %s.' % (minimum_strength,))

        raw_value = password_handler.encrypt(
            value,
            rounds=self.parent._PASSWORD_CRYPT_ROUNDS,
            format=server_section.password_hashing_scheme,
            )
        new_history = check_history(
            value, raw_value, server_section.password_history)

        self._proxy.setStringOrNone(
            self._section_name, 'password', raw_value)
        self.last_passwords = new_history
        self.last_password_update = self._now()

    def _getUserInputs(self):
        """
        Return a list of values which are extracted from user data.
        """
        result = [
            'SFTPPlus',
            'sftpplus',
            'ftp',
            'ftps',
            'sftp',
            'FTP',
            'FTPS',
            'SFTP',
            self.name,
            self.uuid,
            ]
        for token in self.description.split(' '):
            if not token:
                continue
            token = token.strip('.?!-+_')
            result.append(token)

        return result

    @property
    def last_password_update_v3(self):
        """
        Backward compatible value to help with migration.

        It only allows reading once, after that the value is removed from the
        configuration.
        """
        if not self._proxy.hasOption(
                self._section_name, 'last_password_update'):
            return 0

        current = self._proxy.getString(
            self._section_name, 'last_password_update')

        try:
            current = current.strip()
            if not current:
                return 0

            return float(current)
        finally:
            self._proxy.removeOption(
                self._section_name, 'last_password_update')


class LocalPathConfigurationMixin(object):
    """
    See: `ILocalPathOptions`.
    """

    @property
    def path(self):
        """
        See: `ILocalPathOptions`.
        """
        return self._proxy.getStringOrNone(self._section_name, 'path')

    @path.setter
    def path(self, value):
        return self._updateWithNotify(
            setter=self._proxy.setStringOrNone,
            name='path',
            value=value,
            )


@implementer(IAddressPortPublicOptions)
class AddressPortOptionsMixin(object):
    """
    See: `IAddressPortPublicOptions`.
    """
    #: Minimum port number allowed.
    PORT_MIN = PORT_MIN
    #: Maximum port number allowed.
    PORT_MAX = PORT_MAX

    #: Default port value when it is not defined.
    PORT_DEFAULT = 1

    address = WritableString('address')

    def _validatePort(self, port):
        """
        Return a valid integer number for `port`.

        Emit an event on errors.
        """
        if not port:
            port = self.PORT_DEFAULT

        if not isinstance(port, int):
            try:
                port = int(port)
            except ValueError:
                # FIXME:5707:
                # Instead of using default value,
                # We should return a number, but that is a marker for "invalid"
                # value.
                # High-level code can check it and show an error.
                self.root.emitEvent('20190', data={
                    'invalid_value': force_unicode(port),
                    'section': self._section_name,
                    'option': 'port',
                    'default': self.PORT_DEFAULT,
                    'details': 'Port value must be an integer.',
                    })
                port = self.PORT_DEFAULT
                # Set the value to not emit other events on each future read.
                self._proxy.setString(self._section_name, 'port', str(port))

        if port < 1:
            # Don't emit extra errors.
            port = self.PORT_DEFAULT

        if port > 65535:
            # Don't emit extra errors.
            port = self.PORT_DEFAULT

        return port

    @property
    def port(self):
        """
        Port number used by the service for listening for new
        connections.

        This need to be an `Integer` between 1 - 65535. It is recommend to
        choose a value between 1 - 49152.
        """
        port = self._proxy.getString(self._section_name, 'port')
        return self._validatePort(port)

    @port.setter
    def port(self, value):
        # FIXME:5707:
        # here we should raise an error for invalid values.
        value = self._validatePort(value)
        self._proxy.setString(self._section_name, 'port', str(value))


@implementer(IURLOptions)
class URLOptions(object):
    """
    Shared configuration for an URL with parsed address and port.
    """

    url = WritableString('url')

    @property
    def address(self):
        """
        Mirrors the URL configuration.
        """
        try:
            from chevah.server.commons.configuration import parse_url
            return parse_url(self.url)['address']
        except ServerException:
            return 'no-address-yet'

    @property
    def port(self):
        """
        Mirrors the URL configuration.
        """
        try:
            from chevah.server.commons.configuration import parse_url
            return parse_url(self.url)['port']
        except ServerException:
            return 0


@implementer(ITypeOptions)
class TypeOptionsMixin(object):
    """
    See: `ITypeOptions`.
    """
    @property
    def type(self):
        """
        See: ITypeOptions.
        """
        return self._proxy.getString(self._section_name, 'type').lower()


@implementer(ILocationWatchConfiguration)
class LocationWatchConfigurationMixin(object):
    """
    See: `ILocationWatchConfiguration`.
    """
    source_uuid = WritableString(
        'source_uuid', default=DEFAULT_LOCAL_FILESYSTEM_UUID)
    source_path = WritableStringOrEmpty('source_path')
    changes_poll_interval = WritableNumber(
        'changes_poll_interval',
        default=10,
        allow_negative=False,
        allow_zero=False,
        )
    stable_interval = WritableNumber(
        'stable_interval',
        default=11,
        allow_negative=False,
        allow_zero=False,
        )
    source_filter = WritableStringOrEmpty('source_filter')
    recursive = WritableBoolean('recursive')


@implementer(IAutoFailOptions)
class AutoFailOptionsMixin(object):
    """
    Properties for objects exporting Runnable options.
    """
    fail_after_errors = WritableInteger('fail_after_errors', default=10)


class RetryIncreaseOptionsMixin(object):
    """
    See: `_IRetryOptions`.
    """
    retry_count = WritableInteger('retry_count', default=2)
    retry_increase = WritableNumber('retry_increase', default=10)
