# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
A set configuration is an object containing a collection of other
configuration options.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from zope.component import queryUtility
from zope.interface import implementer

from chevah.server.commons.constant import CONFIGURATION_DEFAULT_TYPE
from chevah.server.commons.exception import (
    ConfigurationDeleteInUseException,
    ServerError,
    ServerException,
    )
from chevah.server.commons.interface import ISetConfiguration
from chevah.server.commons.observer import Signal
from chevah.server.configuration.model import (
    PropertyMixin,
    )
import six


@implementer(ISetConfiguration)
class SetConfiguration(PropertyMixin):
    """
    Shared code by all sections holding a set of configuration options.

    All set members implement `__type_interface__`.

    Child sections are cached to allow registering to events.

    Must define the following class members:
    * SECTION_PREFIX = 'logs/'

    Emit the following signals:

    * delete - Signal(uuid)
    * create - Signal(uuid)
    """

    #: The interfaces used to find the configuration class for type.
    __type_interface__ = None

    def __init__(self, parent):
        self._parent = parent
        self._proxy = parent._proxy
        #: Dict used to help getting the same object for the same uuid.
        self._get_cache = {}

    def _getUUID(self):
        """
        See: `ISetConfiguration`.
        """
        # We import uuid here so that we can use uuid variable name in the
        # of this configuration file.
        import uuid
        return six.text_type(uuid.uuid4())

    def getSection(self, name):
        """
        See: `IProperty`.

        Return a new section based on __type_interface__.
        """
        cached_section = self._get_cache.get(name, None)
        if not cached_section:
            new_section = self._getSectionWithoutCache(name)
            if new_section:
                self._get_cache[name] = new_section
            return new_section
        else:
            return cached_section

    def _getSectionWithoutCache(self, name):
        """
        Return a new section instance.
        """
        available_uuids = self.getPublicSectionNames()
        if name not in available_uuids:
            return None

        section_class = queryUtility(self.__type_interface__, default=None)

        if section_class is None:
            raise RuntimeError(
                'No configuration class found for type interface %s.' % (
                    self.__type_interface__,))

        return section_class(parent=self, uuid=name)

    def getPublicSectionNames(self):
        """
        See: `PropertyMixin`.

        Return a list with names of all public sections.
        """
        result = []
        for section in self._proxy.sections:
            if not section.startswith(self.SECTION_PREFIX):
                continue
            uuid = section[len(self.SECTION_PREFIX):]
            result.append(uuid)
        return result

    def deleteProperty(self, property_path):
        """"
        See: `PropertyMixin`.

        This deletes the property and also emit a signal.
        """
        head, tail = self.traversePath(property_path)

        if not head:
            raise RuntimeError(
                '%s No child requested for delete operation.' % (
                    self.SECTION_PREFIX))

        if tail:
            section = self.getSection(head)
            if not section:
                raise RuntimeError(
                    '%s Can not delete a child for %s.' % (
                        self.SECTION_PREFIX, property_path))

            return section.deleteProperty(tail)

        uuid = self._deleteDirectProperty(head)
        self.notify('delete', Signal(self, uuid=uuid))
        return uuid

    def _deleteDirectProperty(self, uuid):
        """
        Checks if there are no references to property specified in
        property_path before deleting it.
        """
        try:
            self._checkReferenceByUUID(uuid)
        except ConfigurationDeleteInUseException as error:
            data = {
                "uuid": error.uuid,
                "usage": ', '.join(error.usage),
                }
            raise ServerError('20108', data=data)

        self.beforeChildDelete(uuid)

        return self._forceDeleteProperty(uuid)

    def _forceDeleteProperty(self, uuid):
        """
        Delete property specified in property_path.
        """
        self._get_cache.pop(uuid, None)
        section_name = self.SECTION_PREFIX + uuid
        section_existed = self._proxy.removeSection(section_name)
        if not section_existed:
            raise ServerException(
                'Property %s does not exist.' % section_name)
        return uuid

    def beforeChildDelete(self, uuid):
        """
        See `ISetConfiguration`.
        """
        pass

    def _checkReferenceByUUID(self, uuid):
        """
        Raise an error if `uuid` is referenced by any configuration.

        Self references are ignored.
        """
        references = self.root.configuration.findAllReferences(uuid)
        if references:
            in_use_by = []
            for reference_uuid, section in references:
                if reference_uuid == uuid:
                    # Ignore self reference
                    continue
                in_use_by.append('%s (%s)' % (reference_uuid, section))
            if in_use_by:
                raise ConfigurationDeleteInUseException(uuid, in_use_by)

    def createProperty(self, property_path, value):
        """
        See: `PropertyMixin`.

        It will create a new property and try to apply all attributes defined
        in `value`.

        Return configuration object for the new property.
        """
        head, tail = self.traversePath(property_path)

        if head:
            section = self.getSection(head)
            if section:
                return section.createProperty(tail, value)

        if head:
            raise RuntimeError(
                '%s Can not create. Bad property_path %s.' % (
                    self.SECTION_PREFIX, property_path))

        uuid = value.pop('uuid', self._getUUID())

        try:
            section = self._createSection(uuid, value)
            self.notify('create', Signal(self, uuid=uuid))
            return section
        except Exception:
            # We force delete the property and re-raise the error.
            self._forceDeleteProperty(uuid)
            raise

    def _createSection(self, uuid, value):
        """
        Create a section in the low level persistence.
        """
        section_name = '%s%s' % (self.SECTION_PREFIX, uuid)

        self._proxy.addSection(section_name)
        filtered_value = self.onChildCreate(section_name, value, uuid)
        new_section = self.getSection(uuid)
        # To help with testing.
        new_section._UPDATE_DELAY = self._UPDATE_DELAY

        for key_name, key_value in filtered_value.items():
            new_section.setPropertyWithoutNotification(key_name, key_value)

        return new_section

    def onChildCreate(self, section_name, value, uuid=None):
        """
        Prepare section just after it was created.
        """
        return value.copy()

    def getSectionByName(self, name):
        """
        See: `ISetConfiguration`.
        """
        for uuid in self.getPublicSectionNames():
            section = self.getSection(uuid)
            if section.name == name:
                return section
        return None

    def setPropertyFromSync(self, value):
        """
        See `ISetConfiguration`.
        """
        value = value.copy()
        uuid = value.pop('uuid', None)
        type_value = value.pop('type', None)

        if not uuid:
            raise ServerException('Source sync value has no UUID.')

        component = self.getSection(uuid)

        if not component:
            # Targeted component doesn't exist.
            # Start by creating a minimum component
            # and we will update it later.
            seed_value = {'uuid': uuid}
            if type_value:
                seed_value['type'] = type_value

            # FIXME:3480:
            # Emit event on new component.
            # Might need to be moved to node_sync.py as we don't want to
            # emit events from this low-level code.
            component = self.createProperty('', seed_value)

        for name in component.getSyncAttributeNames():
            if name.endswith('password_update'):
                # FIXME:6416:
                # See how to set other passwords or secrets and in a generic
                # way.
                # For now, other secrets are not synced.
                continue

            if name not in value:
                # Don't update if source is missing this attribute.
                continue

            # _setRawPassword will update both hashed and plain text
            # password.
            if name == 'password':
                # FIXME:6416:
                # Add notification
                # See how to synchronize password.
                component._setRawPassword(value['password'])
                continue

            # FIXME:3480:
            # Emit event on configuration changes.
            component.setValue(name, value[name])
        return component


class SetWithTypeConfiguration(SetConfiguration):
    """
    A section for which members can have different types.
    """

    #: Type used when configuration has not defined an explicit type.
    __DEFAULT_TYPE__ = None

    def _getSectionWithoutCache(self, name):
        """
        See `SetConfiguration`.
        """
        type_name = self._getTypeName(uuid=name)

        # Section not found.
        if not type_name:
            return None

        configuration_class = queryUtility(
            self.__type_interface__, name=type_name, default=None)

        if not configuration_class:
            data = {
                'uuid': name,
                'type': type_name,
                }
            raise ServerError('20091', data=data)

        return configuration_class(parent=self, uuid=name)

    def _getTypeName(self, uuid):
        """
        Return a string containing type name for section with `uuid`.
        """
        section = '%s%s' % (self.SECTION_PREFIX, uuid)
        if not self._proxy.hasSection(section):
            return None

        value = self._proxy.getString(section, 'type').lower().strip()

        if value == CONFIGURATION_DEFAULT_TYPE and self.__DEFAULT_TYPE__:
            return self.__DEFAULT_TYPE__

        return value

    def onChildCreate(self, section_name, value, uuid):
        """
        Add type to raw section.
        """
        type_name = value.get('type', None)

        if not type_name:
            type_name = CONFIGURATION_DEFAULT_TYPE

        self._proxy.setString(section_name, 'type', type_name)

        # Make sure `type` is not set as part of the standard attributes.
        result = value.copy()
        result.pop('type', None)
        return result
