# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
Database entities definitions.
"""

from __future__ import absolute_import
from __future__ import unicode_literals
from chevah.server.commons.exception import ServerException


class InvalidColumnDefinitionException(Exception):
    """
    Invalid column definition.
    """

    def __init__(self, column, reason):
        args = (column.name, reason)
        self.message = 'Invalid column definition for column "%s". %s.' % args
        super(InvalidColumnDefinitionException, self).__init__(self.message)


class InvalidIndexDefinitionException(ServerException):
    """
    Invalid index definition.
    """

    def __init__(self, index, reason):
        args = (index.name, reason)
        message = 'Invalid index definition for "%s". %s.' % args
        super(InvalidIndexDefinitionException, self).__init__(message)


class DBTableColumn(object):
    """
    Represents a table column and maintains basic information about it:
    name, type, size and description.
    """

    def __init__(
            self, name, kind, size=None, description='', primary_key=False,
            auto_increment=False):
        self._name = name
        self._type = kind
        self._size = size
        self._description = description
        self._primary_key = primary_key
        self._auto_increment = auto_increment

        if self._auto_increment and not self._primary_key:
            reason = 'Auto-increment is only allowed for primary keys'
            raise InvalidColumnDefinitionException(self, reason)

    @property
    def name(self):
        """
        Column name.
        """
        return self._name

    @property
    def type(self):
        """
        Column SQL data type.
        """
        return self._type

    @property
    def size(self):
        """
        Column data size, if set, otherwise None.
        """
        return self._size

    @property
    def description(self):
        """
        Column description, human readable free text.
        """
        return self._description

    @description.setter
    def description(self, value):
        """
        Set the column description text.
        """
        if not value:
            self._description = ''
        else:
            self._description = value

    @property
    def primary_key(self):
        """
        `True` if the column is a primary key.
        """
        return self._primary_key

    @property
    def auto_increment(self):
        """
        `True` if column has auto increment enabled.

        Columns with `auto_increment` set are ignored during insert.
        """
        return self._auto_increment

    def isText(self):
        """
        Return `True` if column type is text.
        """
        if self.type.upper() in ['TEXT', 'VARCHAR', 'CHAR']:
            return True
        else:
            return False


class DBFunctionColumn(DBTableColumn):
    """
    Marker for column that is a function.

    Represents either a function name or expression column used in SQL
    queries.
    """

    def __init__(self, function, name, kind):
        self.function = function
        super(DBFunctionColumn, self).__init__(name, kind)


class DBTableColumns(object):
    """
    Columns definitions for a database table. Columns must have unique names.
    """

    def __init__(self):
        self._columns = []
        self._primary_key = None

    def add(self, column):
        """
        Add a column definition to the list.

        Raise an `InvalidColumnDefinitionException` if more than one column
        configured as primary key.
        """
        if column.primary_key and self._primary_key:
            reason = 'Only one primary key allowed'
            raise InvalidColumnDefinitionException(column, reason)

        self._columns.append(column)

        if column.primary_key:
            self._primary_key = column

    @property
    def names(self):
        """
        Return the list with the names of all columns.
        """
        result = [column.name for column in self._columns]
        return result

    @property
    def column_count(self):
        """
        Return the number of columns defined.
        """
        return len(self._columns)

    @property
    def primary_key(self):
        """
        Return the `DBTableColumn` instance that is the primary key.
        """
        return self._primary_key

    def hasColumn(self, name):
        """
        Returns `True` if columns with `name` exists in this columns
        definition.
        """
        for candidate in self._columns:
            if name == candidate.name:
                return True
        return False

    def column(self, name):
        """
        Returns the column definition instance for the column with the
        specified `name`.

        Raises an exception if no column with the specified name is found.
        """
        for candidate in self._columns:
            if name == candidate.name:
                return candidate

        raise KeyError('Column with name "%s" cannot be found.' % name)

    def __iter__(self):
        """
        Return an iterable list of columns.
        """
        for item in self._columns:
            yield item


class DBTableIndex(object):
    """
    Portable representation for an SQL table index.
    """

    ASCENDING = 'ASC'
    DESCENDING = 'DESC'

    def __init__(self, name):
        self._name = name
        self._columns = []

    @property
    def name(self):
        """
        Index name.
        """
        return self._name

    @property
    def definition(self):
        """
        List containing column/direction pair(s).
        """
        result = []
        result.extend(self._columns)
        return result

    @property
    def columns(self):
        """
        Return a list containing column names that are members of this
        index definition.
        """
        result = []
        for column, direction in self._columns:
            result.append(column.name)
        return result

    def addColumn(self, column, direction=ASCENDING):
        """
        Add `column` as member of the index with the specified sorting
        `direction`.
        """
        for candidate, _ in self._columns:
            if candidate.name != column.name:
                continue

            reason = 'A column with name "%s" already present' % column.name
            raise InvalidIndexDefinitionException(self, reason)

        pair = (column, direction)
        self._columns.append(pair)

        return self


class DBTableIndexes(object):
    """
    Index definitions for a database table. Indexes need to have unique
    names.
    """

    def __init__(self):
        self._indexes = []

    def addIndex(self, index):
        """
        Add index definition.

        Raise an exception if an index with the same name already present.
        """
        if index.name in self.names:
            reason = 'An index with the same name already defined'
            raise InvalidIndexDefinitionException(index, reason)

        self._indexes.append(index)

    @property
    def names(self):
        """
        Return the list with the names of all indexes.
        """
        result = [index.name for index in self._indexes]
        return result

    def index(self, name):
        """
        Return index definition with specified `name`.

        Raises an exception if no index with the specified name is found.
        """
        for candidate in self._indexes:
            if name == candidate.name:
                return candidate

        raise KeyError('Index with name "%s" cannot be found.' % name)


class DBTable(object):
    """
    Portable representation for an SQL table.
    """

    def __init__(self, name, columns=None, indexes=None):
        self._name = name

        if columns is None:
            columns = DBTableColumns()

        if indexes is None:
            indexes = DBTableIndexes()

        self._columns = columns
        self._indexes = indexes

    @property
    def name(self):
        """
        Returns the table name.
        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def columns(self):
        """
        Returns a `DBTableColumns` instance containing the table schema
        definition.
        """
        return self._columns

    @property
    def indexes(self):
        """
        Returns a `DBTableIndexes` instance containing index definitions.
        """
        return self._indexes
