# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
Low level database handling.
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from time import time
import gc

from twisted.internet import defer, reactor, threads
from twisted.python.threadpool import ThreadPool
import six

try:
    from pysqlite2 import dbapi2
    from pysqlite2.dbapi2 import Error as SQLiteError
except ImportError:
    from sqlite3 import dbapi2
    from sqlite3.dbapi2 import Error as SQLiteError

from chevah_compat import local_filesystem

from chevah.server import force_unicode
from chevah.server.commons.constant import (
    SORT,
    )
from chevah.server.commons.exception import ServerException

from chevah.server.db.entity import (
    DBFunctionColumn,
    DBTableColumn,
    DBTableColumns,
    )

# List with all interaction calls, used for debugging.
call_history = []

# Flag to enable db debugging.
_DEBUG_ENABLED = False


def enable_debug():
    """
    Configure this module for debugging.
    """
    global _DEBUG_ENABLED
    _DEBUG_ENABLED = True


class DBInteractionSQLite(object):
    """
    High-level database operations controller.

    It required the implementation to define a `configuration` member, which
    has the `path`.

    It does not keep track of failures, nor does it fail after the maximum
    number of failures is reached.
    """
    # SQLite3 connection.
    _conn = None
    # The threadpool, with a single thread in which sqlite operates.
    _pool = None
    _reactor = reactor

    def _startDB(self, avatar=None):
        """
        Start the DB.
        """
        if self._conn:
            return defer.succeed(False)

        if not self.configuration.path:
            return defer.fail(ServerException('No configured "path".'))

        def connect():
            try:
                self._conn = dbapi2.connect(
                    local_filesystem.getEncodedPath(self.configuration.path))
                self._conn.execute('SELECT * FROM `sqlite_master`')
            except SQLiteError as error:
                raise ServerException(
                    'SQLite error: {}'.format(force_unicode(error)))
            return True

        self._pool = ThreadPool(
            minthreads=1, maxthreads=1,
            name=b'SQlite' + self.name.encode('ascii', errors='replace'),
            )
        self._pool.start()
        return threads.deferToThreadPool(
            self._reactor,
            self._pool,
            connect,
            )

    def _stopDB(self, avatar=None):
        """
        Stop the DB connection.
        """
        if not self._conn:
            if self._pool:
                self._pool.stop()
                self._pool = None
            return defer.succeed(False)

        conn, self._conn = self._conn, None

        def close(conn):
            conn.commit()
            conn.close()
            # The low-level DB file is not closed right away.
            # This is an issue during testing as the file is still kept
            # open and the test clean fails.
            gc.collect(0)

        deferred = threads.deferToThreadPool(
            self._reactor,
            self._pool,
            close, conn)
        deferred.addCallback(lambda _: self._pool.stop())
        return deferred

    def _runInteraction(self, interaction, action):
        """
        Execute a low level DB interaction, making sure the DB is auto-started
        as the DB is considered a resource.
        """
        def execute():
            """
            Called from sqlite thread.
            """
            if not self._conn:
                raise ServerException(
                    'DB %s operation while stopped.' % (action,))

            try:
                cursor = self._conn.cursor()
                result = interaction(cursor)
                cursor.close()
                self._conn.commit()
                return result
            except BaseException as error:
                try:
                    self._conn.rollback()
                except BaseException:
                    """
                    Ignore rollback errors.
                    """

                raise error

        def cbStarted(result):
            """
            Called when DB start operation is done.

            This is also called when it fails to start.
            """
            if not self.operating:
                raise ServerException('Failed to auto-start the DB.')

            return threads.deferToThreadPool(
                self._reactor, self._pool, execute)

        def ebConvertError(failure):
            """
            Convert low level DB error into ServerError.
            """
            failure.trap(SQLiteError, ServerException)

            details = force_unicode(failure.value)
            message = (
                'Failed to perform %(action)s in db \"%(database_name)s\". '
                '%(details)s'
                ) % {
                    'database_name': self.name,
                    'action': action,
                    'details': details,
                    }
            raise ServerException(message)

        def bbStopCounter(result_or_failure, call_debug):
            """
            Called to stop debug counter for current interaction.
            """
            call_debug['end'] = time()
            call_debug['duration'] = call_debug['end'] - call_debug['start']
            return result_or_failure

        if _DEBUG_ENABLED:
            call_debug = {
                'action': action,
                'interaction': str(interaction),
                'start': time(),
                'end': None,
                'duration': -1,
                }
            call_history.append(call_debug)

        deferred = self.start()
        deferred.addCallback(cbStarted)

        if _DEBUG_ENABLED:
            deferred.addBoth(bbStopCounter, call_debug)

        deferred.addErrback(ebConvertError)
        return deferred

    def _getIndexes(self, txn, table):
        """
        Get the current indexes present for `table` in the
        persisted DB.
        """
        table_name = table.name
        query_get_indexes = self._getTableIndexesQuery(table_name)
        txn.execute(query_get_indexes)
        indexes = txn.fetchall()

        if not indexes:
            return {}

        result = {}
        for index in indexes:
            name = self._adaptIndex(table_name, index)
            query_get_index = self._getTableIndexQuery(table_name, name)
            txn.execute(query_get_index)

            for line in txn.fetchall():
                entry = self._adaptIndexDefinition(table_name, name, line)
                if entry['name'] in result:
                    existing = result[entry['name']]
                    existing.append(entry['column'])
                else:
                    result[entry['name']] = [entry['column']]

        return result

    def _getColumns(self, txn, table):
        """
        Get the current columns present for `table` in the
        persisted DB.
        """
        table_name = table.name
        query_get_schema = self._getTableSchemaQuery(table_name)
        txn.execute(query_get_schema)
        results = txn.fetchall()
        return self._adaptColumns(results)

    def _createTable(self, txn, name, columns):
        """
        Create a new table.

        Helper to share some code with the tests.
        """
        values = (
            self._escapeTableName(name),
            self._getColumnsDefinition(columns)
            )
        query_create = 'CREATE TABLE %s (%s)' % values
        txn.execute(query_create)

    def _indexExists(self, txn, table_name, index_name):
        query_exists = self._getIndexExistsQuery(table_name, index_name)
        txn.execute(query_exists)
        rows = txn.fetchall()
        if len(rows) < 1:
            return False

        candidate = rows[0][0]
        candidate_name = self._getUniqueIndexName(table_name, candidate)
        unique_name = self._getUniqueIndexName(table_name, index_name)

        return candidate_name == unique_name

    def _createIndexesIfMissing(self, txn, table_name, indexes):
        """
        Add all `indexes` to `table_name`.
        """
        for name in indexes.names:
            index = indexes.index(name)
            if self._indexExists(txn, table_name, index.name):
                continue

            columns = []
            for column, direction in index.definition:
                entry = '%s %s' % (
                    self._escapeColumnName(column),
                    direction,
                    )
                columns.append(entry)

            index_definition = ','.join(columns)

            values = (
                self._escapeIndexName(table_name, index.name),
                self._escapeTableName(table_name),
                index_definition
                )
            query_create = 'CREATE INDEX %s ON %s (%s)' % values
            txn.execute(query_create)

    def migrateTable(self, table_instance):
        """
        Creates the corresponding database table if not already existing and
        checks its schema against the column definitions.

        It automatically creates missing columns or indexes.

        Raises an error if the existing columns don't match the
        definition for`table_instance`.
        """
        table_name = table_instance.name
        table_columns = table_instance.columns

        def tableExists(txn):
            query_exists = self._getTableExistsQuery(table_name)
            txn.execute(query_exists)
            result = txn.fetchall()
            if len(result) != 1:
                return False

            item = result[0]
            return table_name in item

        def _do(txn):
            """
            Create table and indexes, validate them using a single transaction
            for improved speed and cohesion -- when either operation fails the
            database is left in a clean state.
            """
            if not tableExists(txn):
                self._createTable(txn, table_name, table_columns)

            candidates = self._getColumns(txn, table_instance)

            self._validateExistingColumns(table_columns, candidates)
            self._createMissingColumns(txn, table_instance, candidates)

            self._createIndexesIfMissing(
                txn, table_instance.name, table_instance.indexes)

            candidates = self._getIndexes(txn, table_instance)
            self._validateIndexes(candidates, table_instance.indexes)

        action = 'migrate table "%s"' % table_name
        deferred = self._runInteraction(_do, action)
        deferred.addCallback(lambda _: table_instance)
        return deferred

    def deleteTable(self, table_name):
        """
        Delete table `table_name` if it exists.

        Raise an error if the table does not exist.
        """
        def deleteTable(txn):
            query_drop = 'DROP TABLE %s' % self._escapeTableName(table_name)
            txn.execute(query_drop)
            return True

        action = 'delete table "%s"' % table_name
        return self._runInteraction(deleteTable, action)

    def insertRow(self, table, values):
        """
        Insert a row into the table with the specified values.

        * table: an instance of DBTable class
        * values: a dictionary containing values for each non-id column
        """
        statement = self._getInsertQuery(table)
        arguments = self._getInsertValues(table, values)

        def insert(txn):
            txn.execute(statement, arguments)
            return True

        action = 'insert single row into "%s"' % table.name
        return self._runInteraction(insert, action)

    def insertRows(self, table, rows):
        """
        Insert multiple rows into the table.

        * table: an instance of DBTable class
        * rows: a sequence of dictionaries containing values for each non-id
                column
        """
        statement = self._getInsertQuery(table)
        arguments = []
        for values in rows:
            arguments.append(self._getInsertValues(table, values))

        def insert_many(txn):
            txn.executemany(statement, arguments)
            return True

        action = 'insert multiple rows into "%s"' % table.name
        return self._runInteraction(insert_many, action)

    def updateRows(self, table, values, criteria=None):
        """
        Update table values for row(s) which match the specified `criteria`.
        """
        if not criteria:
            criteria = []

        update_statement = self._getUpdateQuery(table, values)
        update_arguments = self._getUpdateValues(table, values)

        where_clause, where_arguments = self._getWhereClause(
            table.columns, criteria)

        arguments = []
        arguments.extend(update_arguments)
        arguments.extend(where_arguments)

        statement = '%s%s;' % (update_statement, where_clause)

        def update(txn):
            txn.execute(statement, arguments)
            return True

        action = 'update row(s) of "%s"' % table.name
        return self._runInteraction(update, action)

    def deleteRows(self, table, criteria):
        """
        Delete rows matching `criteria` from specified `table`.
        """
        if not criteria:
            message = 'Deletion of rows without a criteria is forbidden'
            raise ServerException(message)

        delete_statement = self._getDeleteQuery(table)
        where_clause, where_arguments = self._getWhereClause(
            table.columns, criteria)
        statement = '%s%s;' % (delete_statement, where_clause)

        arguments = []
        arguments.extend(where_arguments)

        def delete(txn):
            result = txn.execute(statement, arguments)
            return result.rowcount

        action = 'delete row(s) from "%s"' % table.name
        return self._runInteraction(delete, action)

    def countRows(self, table, criteria=None):
        """
        Return number of rows from `table` that match the `criteria`.

        You should avoid using it it production since the can have
        performance issues for huge tables and complex criteria.
        """
        if not criteria:
            criteria = []

        column_count = DBFunctionColumn(
            function='COUNT(*)',
            name='count_all',
            kind='INT',
            )
        statement, arguments = self._getSelectQuery(
            table=table, criteria=criteria, limit=1, columns=[column_count])

        def count(txn):
            txn.execute(statement, arguments)
            result = txn.fetchone()
            return result[0]

        action = 'count rows from "%s"' % table.name
        return self._runInteraction(count, action)

    def selectRows(
            self, table, criteria=None, limit=None, offset=None, orderBy=None,
            columns=None):
        """
        Return rows from `table` with values matching `criteria` and ordered
        according to `orderBy`.

        Result is a list of dictionaries where every dictionary represents a
        table row: key is column name and value is actual table row value.

        If `limits` are specified it will return at most `limits` rows. A
        start `offset` row can be also specified.

        Criteria is a list of list of tuples in the following form:
            (column, comparison, value)

        Order by is a list of tuples in the following form:
            (column, TABLE_SORT_ORDER.ASCENDING) for ascending order
            (column, TABLE_SORT_ORDER.DESCENDING) for descending order

        `columns` is a list of `DBTableColumns` to return.
        """
        if not columns:
            columns = table.columns

        if not criteria:
            criteria = []

        statement, arguments = self._getSelectQuery(
            table=table,
            columns=columns,
            criteria=criteria,
            limit=limit,
            offset=offset,
            orderBy=orderBy,
            )

        def select(txn):
            txn.execute(statement, arguments)
            rows = txn.fetchall()

            if not rows:
                return []

            result = []
            for row in rows:
                values = list(row)
                entry = {}
                for column in columns:
                    value = values.pop(0)
                    entry[column.name] = value
                result.append(entry)

            return result

        action = 'select rows from "%s"' % table.name  # noqa
        return self._runInteraction(select, action)

    def _getSelectQuery(
            self, table, columns, criteria, limit=None, offset=None,
            orderBy=None):
        """
        Returns a tuple with the parametrized SQL query as a string and the
        list of arguments for the query.
        """
        column_names = [self._escapeColumnName(column) for column in columns]
        columns_list = ','.join(column_names)

        where_clause, arguments = self._getWhereClause(
            table.columns, criteria)
        order_clause = self._getOrderClause(table.columns, orderBy)
        limit_clause = self._getLimitClause(limit, offset)

        query = (
            'SELECT %(columns)s '
            'FROM %(table)s'
            '%(where)s'
            '%(order)s'
            '%(limits)s;'
            )
        values = {
            'columns': columns_list,
            'table': self._escapeTableName(table.name),
            'where': where_clause,
            'limits': limit_clause,
            'order': order_clause,
            }
        return query % values, arguments

    @staticmethod
    def _resolveColumn(columns, column):
        """
        Check that column is a member of `columns` and return it.

        When `column` is a string (column name) it will look for the
        corresponding column instance in `columns` and return it.

        Raises an error if unable to resolve column or specified `column` is
        not a member of `columns`.
        """
        error_message = 'Unknown column "%s" used.'
        if isinstance(column, six.string_types):
            try:
                column = columns.column(column)
            except KeyError:
                raise ServerException(error_message % column)
        elif column not in columns:
            raise ServerException(error_message % column.name)

        return column

    def _getWhereClause(self, columns, criteria):
        """
        Returns a tuple containing the WHERE clause for SELECT/UPDATE/DELETE
        queries and a list of actual values for the clause parameters.

        SQL escaping for the values is handled by the DB-API database driver
        at lower level.

        criteria is a list of lists.
        The values from the first level list are ANDed.
        The values from second level list are ORed.
        """
        if not criteria:
            return '', []

        and_conditions = []
        values = []
        for and_rule in criteria:
            or_conditions = []
            for candidate, comparison, value in and_rule:
                column_instance = self._resolveColumn(columns, candidate)
                sql_condition, sql_value = self._getWhereCondition(
                    column_instance, comparison, value)
                condition = '%s %s' % (
                    self._escapeColumnName(column_instance), sql_condition)
                or_conditions.append(condition)
                values.append(sql_value)
            and_conditions.append('(' + ' OR '.join(or_conditions) + ')')

        compound_conditions = ' AND '.join(and_conditions)
        return ' WHERE %s' % compound_conditions, values

    def _getWhereCondition(self, column, comparison, value):
        """
        Returns a tuple with the SQL comparison as a string and value expanded
        for SQL comparison.
        """
        if comparison == 'contains' and column.isText():
            comparison = 'LIKE'
            value = '%%%s%%' % value
        elif comparison == '!contains' and column.isText():
            comparison = 'NOT LIKE'
            value = '%%%s%%' % value
        elif comparison in ['=', '!=', '<', '>', '<=', '>=']:
            pass
        else:
            message = 'Comparison not supported: {} {}.'.format(
                column.name, comparison)
            raise ServerException(message)

        placeholder = comparison + ' ?'

        return placeholder, value

    def _getLimitClause(self, limit, offset=None):
        """
        Returns the LIMIT clause used for SELECT queries.
        """
        if not limit:
            if offset:
                message = 'Invalid use of offset without limit.'
                raise ServerException(message)
            return ''

        if not offset:
            # Enforce integer to prevent SQL injection/abuse.
            if not isinstance(limit, int):
                message = 'Limit value must be an integer: "%s".' % limit
                raise ServerException(message)

            return ' LIMIT %d' % limit

        # Enforce integer to prevent SQL injection/abuse.
        if not isinstance(offset, int):
            message = 'Offset value must be an integer: "%s".' % offset
            raise ServerException(message)

        return ' LIMIT %d OFFSET %d' % (limit, offset)

    def _getOrderClause(self, columns, orderBy):
        """
        Returns the ORDER BY sql clause as string.

        Order by criteria is a list of tuples in the following form:
            (column, TABLE_SORT_ORDER.ASCENDING) - for ascending order
            (column, TABLE_SORT_ORDER.DESCENDING) - for descending order

        If `orderBy` is `None` or an empty list no ordering of results will
        be used.
        """
        if not orderBy:
            return ''

        items = []
        for item in orderBy:
            if isinstance(item, six.string_types):
                candidate = item
                direction = SORT.ASCENDING
            else:
                try:
                    candidate, direction = item
                except TypeError:
                    candidate = item
                    direction = SORT.ASCENDING

            column_instance = self._resolveColumn(columns, candidate)
            escaped_column_name = self._escapeColumnName(column_instance)

            if direction == SORT.ASCENDING:
                items.append('%s ASC' % escaped_column_name)
            elif direction == SORT.DESCENDING:
                items.append('%s DESC' % escaped_column_name)
            else:
                message = 'Invalid direction "%s" used for orderBy.'
                raise ServerException(message % direction)

        return ' ORDER BY %s' % ','.join(items)

    def _getInsertQuery(self, table):
        """
        Return the SQL query for inserting a row into the specified table.
        """
        columns = self._filterColumns(table.columns)
        columns_names = [self._escapeColumnName(column) for column in columns]
        columns_list = ','.join(columns_names)

        values = (
            self._escapeTableName(table.name),
            columns_list,
            self._getInsertPlaceholders(columns),
            )

        return 'INSERT INTO %s (%s) VALUES (%s)' % values  # noqa

    def _getUpdateQuery(self, table, values):
        """
        Return the SQL query for updating a row of the specified table.
        """
        columns = self._filterColumns(table.columns)
        columns_with_values = []
        for column in columns:
            if column.name in list(values.keys()):
                columns_with_values.append(column)

        values = (
            self._escapeTableName(table.name),
            self._getUpdatePlaceholders(columns_with_values),
            )

        return 'UPDATE %s SET %s' % values  # noqa

    def _getDeleteQuery(self, table):
        """
        Return the SQL query for deleting rows from the specified table.
        """
        return 'DELETE FROM %s' % self._escapeTableName(table.name)  # noqa

    @staticmethod
    def _filterColumns(columns, skip_auto_incremented=True):
        """
        Return columns matching the filtering criteria.
        """
        result = []
        for column in columns:
            if column.auto_increment and skip_auto_incremented:
                continue

            result.append(column)

        return result

    def _getSQLValuePlaceholder(self):
        """
        Return the SQL string placeholder for INSERT/UPDATE queries.
        """
        return '?'

    def _getInsertPlaceholders(self, columns):
        """
        Return the SQL string for INSERT VALUES query.
        """
        return ','.join([self._getSQLValuePlaceholder()] * len(columns))

    def _getUpdatePlaceholders(self, columns):
        """
        Return the SQL string for UPDATE query.
        """
        items = []
        sql_value_placeholder = self._getSQLValuePlaceholder()

        for column in columns:
            escaped_column_name = self._escapeColumnName(column)
            values = (escaped_column_name, sql_value_placeholder)
            items.append('%s = %s' % values)

        return ','.join(items)

    def _getInsertValues(self, table, values):
        """
        Returns a list with the escaped values for INSERT query.
        """
        insert_columns = self._filterColumns(table.columns)
        return self._getEscapedValues(
            columns=insert_columns, values=values, require_all_columns=True)

    def _getUpdateValues(self, table, values):
        """
        Returns a list with the escaped values for UPDATE query.
        """
        update_columns = self._filterColumns(table.columns)
        return self._getEscapedValues(
            columns=update_columns, values=values, require_all_columns=False)

    def _getEscapedValues(self, columns, values, require_all_columns=True):
        """
        Return a list with the values ordered according to the table
        structure.

        Raises a `ServerException` if values are missing and
        `require_all_columns` is `True`.
        """
        row = []
        for column in columns:
            try:
                value = values[column.name]
                row.append(value)
            except KeyError:
                if not require_all_columns:
                    continue

                message = (
                    'Missing value for mandatory column "%s" in: %s.') % (
                        column.name, values)
                raise ServerException(message)

        return row

    def _escapeIndexName(self, table_name, index_name):
        """
        Return SQL escaped index name.
        """
        return '`%s`' % self._getUniqueIndexName(table_name, index_name)

    def _validateExistingColumns(self, candidates, columns):
        """
        Checks that the existing database `columns` attribute matches or are
        compatible with the expected `candidates` argument.

        Raises an `ServerException` if there are type or size
        mismatches or if there are missing columns.
        """
        for column in columns:
            name = column.name
            try:
                candidate = candidates.column(name)
            except KeyError:
                message = (
                    'Column "%s" found in the DB '
                    'but not in the table definition.' % name)
                raise ServerException(message)

            if column.type != candidate.type:
                message = 'Column "%s" type mismatch.' % name
                raise ServerException(message)
            if column.size != candidate.size:
                message = 'Column "%s" size mismatch.' % name
                raise ServerException(message)

    def _createMissingColumns(self, txn, table, existing_columns):
        """
        Create any column not found in the existing DB but defined in `table`.
        """
        # SQLite does not support adding multiple columns to a table
        # using a single statement.
        for column in table.columns:
            if existing_columns.hasColumn(column.name):
                continue
            if column.primary_key:
                # The column may not have a PRIMARY KEY or UNIQUE constraint.
                # https://www.sqlite.org/lang_altertable.html
                raise ServerException(
                    'Primary key columns can only be added '
                    'when the table is created.')

            query = (
                'ALTER TABLE %s ADD COLUMN %s'
                ) % (
                    self._escapeTableName(table.name),
                    self._getColumnDefinition(column)
                    )
            txn.execute(query)

    def _validateIndexes(self, candidates, indexes):
        """
        Check that the index definitions are compatible with the `candidates`
        dictionary.

        Raises an `ServerException` if there are type or size
        mismatches or if there are missing indexes.
        """
        for index_name in indexes.names:
            index = indexes.index(index_name)
            try:
                existing = candidates[index_name]
                if not existing:
                    raise KeyError()
            except KeyError:
                message = 'Missing index "%s"' % index_name
                raise ServerException(message)

            for column_name in index.columns:
                if column_name not in existing:
                    values = (column_name, index_name)
                    message = 'Missing column "%s" for index "%s"' % values
                    raise ServerException(message)

    def _parseColumnType(self, sql_type):
        """
        Parses the SQL type/size definition and returns a tuple containing
        standardized values.
        """
        if '(' in sql_type:
            parts = sql_type.split('(')
            kind = parts[0].upper()
            size_str = parts[1][:-1]
            try:
                size = int(size_str)
            except ValueError:
                message = 'Invalid int value %s' % size_str
                raise ServerException(message)
        else:
            kind = sql_type.upper()
            size = None

        if size and size <= 0:
            raise ServerException('Invalid size value %s' % size)

        return (kind, size)

    def _adaptColumns(self, raw_columns):
        """
        Return the `DBTableColumns` structure from raw SQL column data.
        """
        if not raw_columns:
            raw_columns = []

        result = DBTableColumns()

        for line in raw_columns:
            column = self._adaptColumn(line)
            result.add(column)

        return result

    def _adaptColumn(self, raw_sql_line):
        """
        See: `DBInteraction`.
        """
        # Raw SQL column data:
        # (index, 'name', 'type(size)', bool, bool, bool)
        order, name, sql_type, null, default, indexed = raw_sql_line

        kind, size = self._parseColumnType(sql_type)

        return DBTableColumn(name=name, kind=kind, size=size)

    def _adaptIndex(self, table_name, raw_sql_line):
        """
        See: `DBInteraction`.
        """
        # Raw SQL column data:
        # (seq, 'name', unique)
        suffix = '_%s' % table_name
        index_name = raw_sql_line[1].replace(suffix, '')
        return index_name

    def _adaptIndexDefinition(self, table_name, name, raw_sql_line):
        """
        See: `DBInteraction`.
        """
        # Raw SQL column data:
        # (seqno, cid, 'name')
        suffix = '_%s' % table_name
        index_name = name.replace(suffix, '')
        return {
            'name': index_name,
            'table': table_name,
            'column': raw_sql_line[2],
            }

    def _escapeTableName(self, table_name):
        """
        See: `DBInteraction`.

        http://www.sqlite.org/lang_keywords.html
        """
        return '`%s`' % table_name

    def _escapeColumnName(self, column):
        """
        See: `DBInteraction`.

        http://www.sqlite.org/lang_keywords.html
        """
        if isinstance(column, DBFunctionColumn):
            return column.function

        return '`%s`' % column.name

    def _getUniqueIndexName(self, table_name, index_name):
        """
        See: `DBInteraction`.

        SQLite stores all indexes in a single table regardless of the table
        they were defined on.

        In order to avoid name clashes between two identical named indexes (on
        different tables) we use the table name as a suffix. This will
        guarantee database level uniqueness as you cannot have two tables with
        the same name.
        """
        suffix = '_%s' % table_name

        if index_name.endswith(suffix):
            return index_name

        return '%s%s' % (index_name, suffix)

    def _getColumnsDefinition(self, columns):
        """
        Generate SQLite specific column definitions.
        """
        items = []
        for column in columns:
            items.append(self._getColumnDefinition(column))
        result = ','.join(items)
        return result

    def _getColumnDefinition(self, column):
        """
        Generate SQLite for a single column.
        """
        modifiers = []
        if column.primary_key:
            modifiers.append('PRIMARY KEY')
            if column.auto_increment:
                modifiers.append('AUTOINCREMENT')

        if modifiers:
            attributes = ' %s' % (' '.join(modifiers))
        else:
            attributes = ''

        if column.size:
            values = (
                self._escapeColumnName(column),
                column.type,
                column.size,
                attributes,
                )
            return '%s %s(%d)%s' % values

        values = (
            self._escapeColumnName(column),
            column.type,
            attributes,
            )
        return '%s %s%s' % values

    def _getTableSchemaQuery(self, table_name):
        """
        Return SQLite query for obtaining table schema.
        """
        return 'PRAGMA table_info(%s)' % self._escapeTableName(table_name)

    def _getTableIndexesQuery(self, table_name):
        """
        Return SQLite query for obtaining table index.
        """
        query = 'PRAGMA index_list(%s)'
        return query % self._escapeTableName(table_name)

    def _getTableIndexQuery(self, table_name, index_name):
        """
        Return SQLite query for obtaining table indexes.
        """
        query = 'PRAGMA index_info(%s)'
        return query % self._escapeIndexName(table_name, index_name)

    def _getTableExistsQuery(self, table_name):
        """
        Return SQLite query for checking table existence.

        Quotes (single or double) are escaped by using two of them.
        https://www.sqlite.org/faq.html#q14
        """
        query = """SELECT `name` FROM `sqlite_master`
            WHERE type = "table" AND name = "%s"
        """
        return query % table_name.replace('"', '""')

    def _getIndexExistsQuery(self, table_name, index_name):
        """
        Return SQLite query for checking table existence.
        """
        query = """
        SELECT `name` FROM `sqlite_master`
            WHERE type = "index" AND name = "%s"
        """
        return query % self._getUniqueIndexName(table_name, index_name)
