# Copyright (c) 2022 Adi Roiban.
# See LICENSE for details.
"""
Event handler for detectiong file interaction between accounts.

It creates a map with all the base paths know to the SFTPPlus configuration.

When a file is modified inside those paths, it emits an event with information
about the other accounts associated with those files.

It support all the file transfer protocols as well as the monitor service.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
import posixpath

from zope.interface import implementer

from chevah.server.commons.constant import (
    ACCOUNT_INTERACTION_OPERATIONS,
    DEFAULT_LOCAL_FILESYSTEM_UUID,
    EVENT_TYPE,
    TYPE_NAME,
    )
from chevah.server.commons.exception import ServerException
from chevah.server.commons.interface import (
    IAccountInteractionEventHandlerConfiguration,
    )
from chevah.server.configuration.model import (
    WritableCSVUnique,
    )
from chevah.server.event.configuration import (
    _EventHandlerConfigurationBase,
    FilterOptionsMixin,
    )
from chevah.server.event.handler import EventHandlerWithFilter
from chevah.server.identity.avatar import (
    resolve_home_folder,
    get_virtual_folders,
    )


_EVENTS = {
    # FTP
    '10078': EVENT_TYPE.CREATED,  # upload
    '10069': EVENT_TYPE.READ,     # download
    '10075': EVENT_TYPE.MOVED,    # rename
    '10044': EVENT_TYPE.DELETED,  # delete file
    # Service monitor
    '20169': EVENT_TYPE.CREATED,   # created
    '20166': EVENT_TYPE.MODIFIED,  # modified
    '20167': EVENT_TYPE.MOVED,     # moved
    '20171': EVENT_TYPE.DELETED,   # delete file
    # SFTP
    '30068': EVENT_TYPE.CREATED,  # upload or download or modified
    '30067': EVENT_TYPE.READ,     # download
    '30023': EVENT_TYPE.DELETED,  # delete file
    '30026': EVENT_TYPE.MOVED,    # rename
    # HTTP
    '40017': EVENT_TYPE.CREATED,  # upload
    '40039': EVENT_TYPE.READ,     # download
    '40012': EVENT_TYPE.DELETED,  # delete file
    }


@implementer(IAccountInteractionEventHandlerConfiguration)
class AccountInteractionEventHandlerConfiguration(
    _EventHandlerConfigurationBase,
    FilterOptionsMixin,
        ):
    """
    Configuration for the account interaction event handler.
    """
    monitored_operations = WritableCSVUnique(
        'monitored_operations')


class AccountInteractionEventHandler(EventHandlerWithFilter):
    """
    Detect file operation between accounts.
    """

    PROPERTIES_REQUIRING_RESTART = ('monitored_operations',)

    # The list of base paths and the associated account UUID and ccp.
    _base_paths = None
    # Configured operations.
    _monitored_operations = ()

    def _onStart(self, avatar=None):
        """
        See: `Runnable`.
        """
        self._monitored_operations = self._configuration.monitored_operations
        for operation in self._monitored_operations:
            if operation not in ACCOUNT_INTERACTION_OPERATIONS:
                raise ServerException(
                    'Unknown operation configured for '
                    'monitored_operations: "{}".'.format(operation))

        self._location = self.root.locations.getSection(
            DEFAULT_LOCAL_FILESYSTEM_UUID)
        self._base_paths = self._getBasePaths()
        self.root.configuration.subscribe('save', self._onConfigurationSave)

    def _onStop(self, avatar=None):
        """
        See: `Runnable`.
        """
        self._monitored_operations = ()
        self.root.configuration.unsubscribe('save', self._onConfigurationSave)
        self._base_paths = {}

    def _onEvent(self, event):
        """
        See: EventHandlerWithFilterMixin.
        """
        event_id = event.id
        if event_id not in _EVENTS:
            return

        if not self._base_paths:
            # The paths not yet loaded.
            return

        target_path = _get_normalized(event.data['real_path'])
        trigger_account_uuid = event.account.uuid
        operation = _EVENTS[event_id]

        if (
            self._monitored_operations
            and TYPE_NAME.ALL not in self._monitored_operations
            and operation not in self._monitored_operations
                ):
            # Operation ignored.
            return

        account_interactions = []
        for base_path, interactions in self._base_paths.items():
            if not target_path.startswith(base_path):
                continue
            for interaction in interactions:
                account_uuid, account_base = interaction

                member_path = _get_normalized(self._location.relativePath(
                    base_path, target_path))
                account_path = posixpath.join(account_base, member_path)
                account_interactions.append((account_uuid, account_path))

        emitted_accounts = []
        for account_uuid, account_path in account_interactions:
            if account_uuid == trigger_account_uuid:
                # We don't want to trigger for files operated by the account
                # itself.
                continue

            if account_uuid in emitted_accounts:
                # We have already emitted an event for this account.
                # Account might have access to this file via multiple groups.
                # We only emit it once.
                continue
            emitted_accounts.append(account_uuid)

            account = self.root.configuration.accounts.getSection(account_uuid)
            if not account:
                # Maybe account was already removed.
                # This should never happen.
                continue

            if not account.enabled:
                continue

            data = {
                'operation': operation,
                # Info for the associated/interacted other account.
                'account_uuid': account_uuid,
                'account_name': account.name,
                'account_path': account_path,
                # Real file.
                'real_path': target_path,
                }
            self.emitEvent('20193', avatar=event.account, data=data)

    def _onConfigurationSave(self, signal):
        """
        Called when the main configuration is updated.
        """
        self._base_paths = self._getBasePaths()

    def _getBasePaths(self):
        """
        Return the base path mapping to account configurations.

        It returns a dict with base paths as keys and content a dict with a list
        of tuples of (account_uuid, chrooted_base_path).

        chrooted_base_path can be `/` for the base home folder or the name
        of a configured virtual path.
        """
        result = {}
        for account in self.root.configuration.accounts.getAll():

            groups = []
            for group_uuid in account.group:
                group = self.root.configuration.groups.getSection(group_uuid)
                if not group:  # noqa:cover
                    # Most probably the group is removed while
                    # a file interaction is was triggered.
                    # Ignore as most probably the group and user are now
                    # removed.
                    continue

                groups.append(group)

            home_path = resolve_home_folder(account, groups[0])
            self._addBasePaths(result, home_path, '/', account)

            # Get the virtual paths, direct or from all groups.
            virtual_folders = get_virtual_folders(account, groups)
            for parts in virtual_folders.values():
                virtual_base = '/' + '/'.join(parts[0])
                virtual_path = parts[1]
                self._addBasePaths(result, virtual_path, virtual_base, account)

        return result

    def _addBasePaths(self, result, real_path, account_path,  account):
        """
        Add `path` to `result` for `account`.

        `real_path` is the path as visible to the OS. It can be Unix or
        Windows.

        `account_path` is the path as visible to the account.
        Always in Unix format.
        """
        path_info = (account.uuid, account_path)

        normalized_path = _get_normalized(real_path)

        if normalized_path in result:
            result[normalized_path].append(path_info)
            return

        result[normalized_path] = [path_info]


def _get_normalized(path):
    """
    Retrun the normalized versionn of path.

    Always use paths with Unix forward slash separators, event for Windows.
    Any trailing slash is removed.
    """
    # Start by moving to Unix separators.
    result = path.replace('\\', '/').rstrip('/')

    # Remove any 2 separator to single one.
    result = result.replace('//', '/')
    # Remove any 2 separator left to single one
    # and with that we should got all of them.
    result = result.replace('//', '/')

    return result
