"""
Code shared by multiple event handlers.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from twisted.internet import defer

from chevah.server import force_unicode
from chevah.server.commons.constant import (
    OVERWRITE_RULE,
    )
from chevah.server.commons.exception import (
    ServerException,
    ServerError,
    )
import six


class _ArchiveMemberException(ServerException):
    """
    Error raised when failing to extract an archive member.
    """
    def __init__(self, message, member_path):
        self.message = '%s %s' % (message, member_path)
        self.member_path = member_path


class SkipHandling(Exception):
    """
    Error raised when destination exists and handling should be skipped.
    """


class CheckDestinationMixin(object):
    """
    Shared code for checking the destination file.
    """

    @defer.inlineCallbacks
    def _checkDestination(
            self, source_path, destination_path, operation, pre_exist=None):
        """
        Check the status of destination path.

        Raise ServerException is handler should stop and not overwrite the
        destination.
        """
        overwrite_rule = self.configuration.overwrite_rule
        if overwrite_rule == OVERWRITE_RULE.DISABLE:
            # No need to check destination.
            return

        exists = pre_exist
        if exists is None:
            exists = yield self._location.exists(
                destination_path, with_event=False)

        if not exists:
            return

        if overwrite_rule == OVERWRITE_RULE.SKIP:
            self.emitEvent(
                '20123',
                data={
                    'operation': operation,
                    'path': source_path,
                    'real_path': source_path,
                    'destination_path': destination_path,
                    },
                )
            raise SkipHandling()

        if overwrite_rule == OVERWRITE_RULE.OVERWRITE:
            self.emitEvent(
                '20188',
                data={
                    'operation': operation,
                    'path': source_path,
                    'real_path': source_path,
                    'destination_path': destination_path,
                    },
                )
            return

        # Fails as fallback rule.
        raise _ArchiveMemberException(
            'Destination exists.', destination_path)


class SourceAttributeMixin(object):
    """
    Shared code for event handlers having a source_attribute.

    Derived classes need to implement `_getHandler(source_path)`
    """

    # Size of data for which we do one read at a time.
    _CHUNK_SIZE = 16000

    def _getHandler(self, source_name):
        """
        Return a tuple of (callable, operation_name)

        `operation_name` is a text describing the operation.

        `callable` will be called with:
        * source_path
        * source_name
        * destination_base
        * operation - same value as returned by _getHandler.
        * event - the original event for which this handler was triggered.
        """
        raise NotImplementedError('SourceAttributeMixin._getHandler')

    def _getSourcePaths(self, event):
        """
        Return a lists of tuple of (source_path, source_parent, source_name)

        Raise an exception if path for the event is not a file.
        """
        source_attribute = self.configuration.source_attribute
        if not source_attribute:
            raise ServerException('No source attribute was defined.')

        sources = event.data.get(source_attribute.lower(), None)
        if not sources:
            raise ServerException(
                'Handler configured for an event which is missing or has an '
                'empty source attribute.'
                )

        if isinstance(sources, six.string_types):
            sources = [sources]

        result = []
        for source_path in sources:
            source_parent, source_name = self._location.splitPath(source_path)
            if not source_name:
                raise ServerException(
                    'Event does not has an associated file path.')
            result.append((source_path, source_parent, source_name))

        return result

    @defer.inlineCallbacks
    def _onEvent(self, event):
        """
        See: EventHandlerWithFilterMixin.
        """
        for member_data in self._getSourcePaths(event):
            yield self._onMemberEvent(member_data, event)

    def _onMemberEvent(self, member_data, event):
        """
        Handle for single path from the members associated to the event.
        """
        _, _, source_name = member_data

        handler, operation = self._getHandler(source_name)

        return self._handleMember(member_data, handler, operation, event)

    def _handleMember(self, member_data, handler, operation, event):
        """
        Perform `handler` operation on the source file.

        It resolved the destination path.
        """
        source_path, source_parent, source_name = member_data

        destination_base = self.configuration.destination_path
        if not destination_base:
            destination_base = source_parent

        deferred = handler(
            source_path=source_path,
            source_name=source_name,
            destination_base=destination_base,
            event=event,
            operation=operation,
            )
        deferred.addCallback(self._cbDeleteSource, source_path)
        deferred.addCallback(self._cbEmitSuccess, source_path, operation, event)
        deferred.addErrback(self._ebEmitFailure, source_path, operation, event)
        return deferred

    def _cbDeleteSource(self, result, source_path):
        """
        Called when file was successfully extracted.
        """
        if not self.configuration.delete_source_on_success:
            return result

        # FIXME:6292:
        # Add here the avatar of the original caller.
        defer = self._location.deleteFile(source_path)
        defer.addCallback(lambda _: result)
        return defer

    def _cbEmitSuccess(self, destination_path, source_path, operation, event):
        """
        Called when file was successfully handled.
        """
        self.emitEvent(
            '20051',
            avatar=event.account,
            data={
                'operation': operation,
                'path': destination_path,
                'real_path': destination_path,
                'source_path': source_path,
                },
            )

    def _ebEmitFailure(self, failure, source_path, operation, event):
        """
        Called when failed to handle the file.
        """
        if failure.check(SkipHandling):
            # Don't trigger an error on skip.
            return

        member_path = ''
        if failure.check(ServerError):
            details = failure.value.data['details']
        elif failure.check(_ArchiveMemberException):
            member_path = failure.value.member_path
            details = failure.value.message
        else:
            details = force_unicode(failure.value)

        self.emitEvent(
            '20052',
            avatar=event.account,
            data={
                'operation': operation,
                'path': source_path,
                'real_path': source_path,
                'member_path': member_path,
                'details': details,
                },
            )
        return failure
