# Copyright (c) 2014 Adi Roiban.
# See LICENSE for details.
"""
Configuration for events and events handlers.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
import re
import time

from zope.interface import implementer

from chevah.server.commons.constant import (
    CONFIGURATION_SECTION_EVENT_HANDLERS,
    TYPE_NAME,
    )
from chevah.server.commons.exception import ServerError
from chevah.server.commons.interface import (
    IDatabaseEventHandlerConfiguration,
    IEventHandlersConfiguration,
    IEventHandlerConfiguration,
    IExtensionEventHandlerConfiguration,
    ICreateArchiveEventHandlerConfiguration,
    IExtractArchiveEventHandlerConfiguration,
    ILocalFileEventHandlerConfiguration,
    IOpenPGPEventHandlerConfiguration,
    IStandardStreamEventHandlerConfiguration,
    ISyslogEventHandlerConfiguration,
    IWindowsEventLogEventHandlerConfiguration,
    )
from chevah.server.configuration.option import (
    AutoFailOptionsMixin,
    EntityConfiguration,
    LocalPathConfigurationMixin,
    PlainPasswordConfigurationMixin,
    TypeOptionsMixin,
    URLOptions,
    )
from chevah.server.configuration.set import SetWithTypeConfiguration
from chevah.server.configuration.model import (
    WritableBoolean,
    WritableInteger,
    WritableString,
    WritableStringCaseInsensitive,
    WritableCSV,
    WritableCSVLines,
    WritableCSVUnique,
    WritableCSVUniqueInsensitive,
    )
import six


@implementer(IEventHandlersConfiguration)
class EventHandlersConfiguration(SetWithTypeConfiguration):
    """
    See: `IEventHandlersConfiguration`.
    """
    SECTION_PREFIX = CONFIGURATION_SECTION_EVENT_HANDLERS
    __type_interface__ = IEventHandlerConfiguration


@implementer(IEventHandlerConfiguration)
class _EventHandlerConfigurationBase(
        EntityConfiguration, TypeOptionsMixin, AutoFailOptionsMixin):
    """
    Shared attributes by all EventHandler configurations.
    """

    SECTION_PREFIX = CONFIGURATION_SECTION_EVENT_HANDLERS


class FilterOptionsMixin(object):
    """
    Configuration option for filtering by target and account.
    """
    target = WritableCSVUniqueInsensitive('target')
    groups = WritableCSVUniqueInsensitive('groups')
    usernames = WritableCSVUnique('usernames')
    components = WritableCSVUnique('components')
    data_filter = WritableCSVLines('data_filter')
    source_addresses = WritableCSVUnique('source_addresses')


class DeprecatedEventHandlerConfiguration(
    _EventHandlerConfigurationBase,
    FilterOptionsMixin,
        ):
    """
    Placeholder configuration for an event handler which was deprecated.
    """

    @property
    def enabled(self):
        """
        Deprecated components are always disabled.
        """
        return False

    @enabled.setter
    def enabled(self, value):
        """
        Does nothing but we allow writing in order not to raise an error
        if someone tries to configure it.
        """


@implementer(ILocalFileEventHandlerConfiguration)
class LocalFileEventHandlerConfiguration(
        _EventHandlerConfigurationBase,
        LocalPathConfigurationMixin,
        FilterOptionsMixin,
        ):
    """
    All configurations for a File handler, as they are persisted.
    """
    entry_content = WritableString('entry_content')
    structured_fields = WritableCSV('structured_fields')

    @property
    def rotate_external(self):
        """
        Return rotate_external.
        """
        return self._proxy.getBoolean(
            self._section_name, 'rotate_external')

    @rotate_external.setter
    def rotate_external(self, value):
        self._updateWithNotify(
            setter=self._proxy.setBoolean,
            name='rotate_external',
            value=value,
            )

    @property
    def rotate_count(self):
        """
        Return rotate_count.
        """
        value = self._proxy.getIntegerOrNone(
            self._section_name, 'rotate_count')
        if value is None:
            value = 0
        return value

    @rotate_count.setter
    def rotate_count(self, value):
        self._updateWithNotify(
            setter=self._proxy.setIntegerOrNone,
            name='rotate_count',
            value=value,
            )

    @property
    def rotate_at_size(self):
        """
        Return rotate_at_size.
        """
        value = self._proxy.getIntegerOrNone(
            self._section_name, 'rotate_at_size')
        if value is None:
            value = 0
        return value

    @rotate_at_size.setter
    def rotate_at_size(self, value):
        self._updateWithNotify(
            setter=self._proxy.setIntegerOrNone,
            name='rotate_at_size',
            value=value,
            )

    @property
    def rotate_on(self):
        """
        Returns a tuple of (calendar_value, calendar_type).
        """
        value = self._proxy.getStringOrNone(self._section_name, 'rotate_on')

        if not value:
            # See if we have backward `rotate_each` and translate it to a
            # default `rotate_on`.
            if self._proxy.getStringOrNone(
                    self._section_name, 'rotate_each'):
                value = '00:00 time-of-day'

        return self._rotateOnToMachineReadable(value)

    @rotate_on.setter
    def rotate_on(self, value):
        if not value:
            update_value = ''
        elif (isinstance(value, six.string_types) and
                self._proxy.isDisabledValue(value)
              ):
            update_value = ''
        else:
            update_value = self._rotateOnToHumanReadable(value)

        self._updateWithNotify(
            setter=self._proxy.setString,
            name='rotate_on',
            value=update_value,
            )

    def _rotateOnToMachineReadable(self, value):
        """
        Return the machine readable format for `value`.

        Inside the configuration file, the value is stored as a human
        readable format like::

            4 day-of-month | 02:30 time-of-day | Disabled

        When reading the value, it will return::

            (4, 'day-of-month') | (150, 'time-of-day') | None

        The time value for time-of-day is stored in seconds since the start
        of the day.
        """

        if not value:
            return ''

        tokens = re.split(r'\s+', value)

        if len(tokens) != 2:
            raise self._rotateOnError('Got: "{}"'.format(value))

        calendar_type = tokens[1].lower()
        if calendar_type == 'day-of-month':
            return (self._getCalendarDay(tokens), calendar_type)
        elif calendar_type == 'time-of-day':
            return (self._getTimeOfDayToSeconds(tokens), calendar_type)
        else:
            raise self._rotateOnError(
                'Unknown calendar type. Got: "{}"'.format(tokens[1]))

    def _rotateOnToHumanReadable(self, value):
        """
        Return the human readable representation for rotate_on
        tuple provided as `value'.

        (2, 'day-of-month') returns '2 day-of-month'
        (9000, 'time-of-day') returns '02:30 time-of-day'.
        """
        calendar_type = value[1]
        if calendar_type == 'day-of-month':
            return '{} {}'.format(self._getCalendarDay(value), calendar_type)
        elif calendar_type == 'time-of-day':
            return '{} {}'.format(
                self._getTimeOfDayFromSeconds(value), calendar_type)
        else:
            raise self._rotateOnError(
                'Unknown calendar type. Got: "{}"'.format(calendar_type))

    def _getTimeOfDayFromSeconds(self, value):
        """
        Retrieve a string for the time of day from the given `value`
        in seconds.
        """
        try:
            seconds = int(value[0])
        except ValueError:
            msg = 'Time should be an integer in seconds. Got: "{}"'
            msg = msg.format(value[0])
            raise self._rotateOnError(msg)
        minutes, _ = divmod(seconds, 60)
        hours, minutes = divmod(minutes, 60)
        return "{:02d}:{:02d}".format(hours, minutes)

    def _getTimeOfDayToSeconds(self, value):
        """
        Retrieve a time of day in seconds from the given string `value`.
        """
        try:
            time_tuple = time.strptime(value[0], '%H:%M')
        except ValueError:
            msg = 'Time format is invalid. Got: "{}"'
            msg = msg.format(value[0])
            raise self._rotateOnError(msg)
        seconds = (time_tuple.tm_hour * 60 + time_tuple.tm_min) * 60

        return seconds

    def _getCalendarDay(self, value):
        """
        Retrieve a calendar day from the given `value`.
        """
        try:
            day = int(value[0])
        except ValueError:
            msg = 'Calendar day is not an integer. Got: "{}"'
            msg = msg.format(value[0])
            raise self._rotateOnError(msg)
        if day < 0 or day > 31:
            raise self._rotateOnError(
                'Invalid calendar day. Got: "{}"'.format(day))
        return day

    def _rotateOnError(self, details):
        """
        Raise `ServerError` with given `details` about the fact that
        `rotate_on` has invalid value.
        """
        details = 'Invalid value for "rotate_on". {}'.format(details)
        return self._rotateError(details)

    def _rotateError(self, details):
        """
        Raise `ServerError` with given `details`.
        """
        return ServerError('20143', data={'details': details})


class _SourceEventAndDestinationPathOptions(object):
    """
    Options for event handlers having the source in the event and destination
    a local path.
    """

    source_attribute = WritableString('source_attribute')
    destination_path = WritableString('destination_path')
    delete_source_on_success = WritableBoolean('delete_source_on_success')
    overwrite_rule = WritableString('overwrite_rule')


@implementer(ICreateArchiveEventHandlerConfiguration)
class CreateArchiveEventHandlerConfiguration(
    _EventHandlerConfigurationBase,
    _SourceEventAndDestinationPathOptions,
    FilterOptionsMixin,
        ):
    """
    Configuration for an event handler create archive file.
    """
    archive_format = WritableStringCaseInsensitive(
        'archive_format', default=TYPE_NAME.ZIP)
    archive_name = WritableCSVUnique(
        'archive_name',
        default=('{fullname}{archive_extension}',),
        inherit=False,
        )


@implementer(IExtractArchiveEventHandlerConfiguration)
class ExtractArchiveEventHandlerConfiguration(
    _EventHandlerConfigurationBase,
    _SourceEventAndDestinationPathOptions,
    PlainPasswordConfigurationMixin,
    FilterOptionsMixin,
        ):
    """
    Configuration for an event handler extracting archive file.
    """


@implementer(IOpenPGPEventHandlerConfiguration)
class OpenPGPEventHandlerConfiguration(
    _EventHandlerConfigurationBase,
    _SourceEventAndDestinationPathOptions,
    PlainPasswordConfigurationMixin,
    FilterOptionsMixin,
        ):
    """
    Configuration for an event handler doing OpenPGP operations on file.
    """
    encryption_cipher = WritableStringCaseInsensitive(
        'encryption_cipher', default='aes128')
    encryption_extension = WritableString(
        'encryption_extension', default='.pgp')
    encryption_public_keys = WritableString('encryption_public_keys')
    decryption_private_keys = WritableString('decryption_private_keys')


@implementer(IExtensionEventHandlerConfiguration)
class ExtensionEventHandlerConfiguration(
        _EventHandlerConfigurationBase, FilterOptionsMixin):
    """
    Configuration for an event handler delegating to an extension.
    """
    entry_point = WritableString('entry_point')
    configuration = WritableString('configuration')


@implementer(IStandardStreamEventHandlerConfiguration)
class StandardStreamEventHandlerConfiguration(
    _EventHandlerConfigurationBase,
    FilterOptionsMixin,
        ):
    """
    Configurations for the standard stream handler.
    """
    entry_content = WritableString('entry_content')


@implementer(ISyslogEventHandlerConfiguration)
class SyslogEventHandlerConfiguration(
        _EventHandlerConfigurationBase,
        FilterOptionsMixin,
        URLOptions,
        ):
    """
    All configurations for an syslog handler, as they are persisted.
    """

    @property
    def url(self):
        """
        URL is implemented with backward compatible "path" value.
        """
        value = self._proxy.getString(self._section_name, 'url')
        if value:
            return value

        # This should be removed in v5.
        path = self._proxy.getStringOrNone(self._section_name, 'path')
        if path:
            return 'file://' + path

        address = self._proxy.getStringOrNone(
            self._section_name, 'address')
        port = self._proxy.getString(self._section_name, 'port')
        if address and address != '127.0.0.1':
            return 'udp://%s:%s' % (address, port)

        # Nothing configured, and no backward settings found.
        return ''

    @url.setter
    def url(self, value):
        return self._proxy.setString(self._section_name, 'url', value)


@implementer(IWindowsEventLogEventHandlerConfiguration)
class WindowsEventLogEventHandlerConfiguration(
        _EventHandlerConfigurationBase, FilterOptionsMixin):
    """
    Configuration for a Windows EventLog event handler.
    """


@implementer(IDatabaseEventHandlerConfiguration)
class DatabaseEventHandlerConfiguration(
        _EventHandlerConfigurationBase, FilterOptionsMixin):
    """
    Configuration for a database event handler.
    """
    path = WritableString('path')
    auto_delete = WritableInteger('auto_delete', default=0)
    queue_size = WritableInteger('queue_size', default=10240)
