# Copyright (c) 2020 Adi Roiban.
# See LICENSE for details.
"""
Event handler creating archives based on files attached to an event.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
from contextlib import closing
from datetime import datetime
import gzip
import zipfile

from bunch import Bunch

from twisted.internet import defer

from chevah.server.commons.constant import (
    DEFAULT_LOCAL_FILESYSTEM_UUID,
    TYPE_NAME,
    )
from chevah.server.commons.exception import ServerException
from chevah.server.commons.zip import ZipStreamWriter
from chevah.server.event.common import CheckDestinationMixin
from chevah.server.event.handler import EventHandlerWithFilter
from chevah.server.event.extract_archive import (
    SourceAttributeMixin,
    )
from chevah.server.commons.public_format import datetime_format


class CreateArchiveEventHandler(
        EventHandlerWithFilter, SourceAttributeMixin, CheckDestinationMixin):
    """
    Create archives based on the files attached to an event.
    """

    PROPERTIES_REQUIRING_RESTART = (
        'archive_format',
        )

    def _onStart(self, avatar=None):
        """
        See: `Runnable`.
        """
        self._location = self.root.locations.getSection(
            DEFAULT_LOCAL_FILESYSTEM_UUID)

        archive_format = self.configuration.archive_format

        if archive_format == TYPE_NAME.GZIP:
            self._onEvent = self._onEventGZIP
        elif archive_format == TYPE_NAME.ZIP:
            self._onEvent = self._onEventZIP
        else:
            raise ServerException(
                'Archive format not supported: "%s".' % (archive_format,))

        try:
            self._getArchiveName('test_name.TEST', 1, '.ARCHIVE_EXT')
        except ServerException as error:
            raise ServerException(
                'Invalid configuration for "archive_name": %s' % (
                    error.message,))

    def _onStop(self, avatar=None):
        """
        See: `Runnable`.
        """
        self._location = None

    @defer.inlineCallbacks
    def _onEventGZIP(self, event):
        """
        Handle event as GZIP archive.

        If event has multiple members, each file is archived
        as a separate file.
        """
        for member_data in self._getSourcePaths(event):
            yield self._handleMember(
                member_data, self._createGZIP, 'create GZIP', event,
                )

    @defer.inlineCallbacks
    def _createGZIP(
            self, source_path, source_name, destination_base, event, operation):
        """
        Compress the source file as GZIP.
        """
        _, _, destination_path = self._getArhivePaths(
            [(source_path, destination_base, source_name)], extension='.gz')

        yield self._checkDestination(source_path, destination_path, operation)

        source = None
        destination = None
        try:
            source = yield self._location.getStreamForReading(source_path)

            destination = yield self._location.getStreamForWriting(
                destination_path)
            destination_stream = destination._context._stream

            with closing(
                    gzip.GzipFile(mode='wb', fileobj=destination_stream)
                    ) as archive:

                while True:
                    chunk = yield source.read(self._CHUNK_SIZE)
                    if not chunk:
                        break

                    archive.write(chunk)
        finally:
            if source:
                yield source.close()
            if destination:
                yield destination.close()

        defer.returnValue(destination_path)

    def _onEventZIP(self, event):
        """
        Handle event as ZIP archive.

        If event has multiple members, all files is archived in the same file.
        """
        operation = 'create ZIP'
        members_data = self._getSourcePaths(event)

        source_base_path, source_path, destination_path = self._getArhivePaths(
            members_data, extension='.zip')

        defered = self._handleZIP(
            members_data,
            source_base_path, source_path, destination_path,
            operation,
            )
        defered.addCallback(self._cbEmitSuccess, source_path, operation, event)
        defered.addErrback(self._ebEmitFailure, source_path, operation, event)
        return defered

    @defer.inlineCallbacks
    def _handleZIP(
        self, members_data,
        source_base_path, source_path, destination_path,
        operation,
            ):
        """
        Handle the creation of the ZIP archive.
        """
        yield self._checkDestination(source_path, destination_path, operation)

        destination = None
        try:
            destination = yield self._location.getStreamForWriting(
                destination_path)
            destination_stream = destination._context._stream

            try:
                with closing(
                        ZipStreamWriter(
                            file=destination_stream,
                            mode='w',
                            compression=zipfile.ZIP_STORED,
                            allowZip64=True,
                            )
                        ) as archive:

                    yield self._writeZIP(
                        archive, members_data, source_base_path)
            except Exception as error:
                yield destination.close()
                destination = None
                yield self._location.deleteFile(destination_path)
                raise error

        finally:
            if destination:
                yield destination.close()

        defer.returnValue(destination_path)

    def _getArhivePaths(self, members_data, extension):
        """
        Return a tuple with (source_base_path, source_path, destination path)
        for the archive to be created.

        When no explicit destination path is configured it will use the
        parent of the first member.
        """
        source_path, source_base_path, source_name = members_data[0]

        if len(members_data) > 1:
            source_path = source_base_path

        destination_base = self.configuration.destination_path
        if not destination_base:
            destination_base = source_base_path

        destination_name = self._getArchiveName(
            source_name, len(members_data), extension)

        destination_path = self._location.joinPath(
            destination_base, destination_name)

        return source_base_path, source_path, destination_path

    def _getArchiveName(self, source_name, members_size, extension):
        """
        Return the name of the archive to be created, including the extension.
        """
        archive_name = self.configuration.archive_name
        if len(archive_name) > 2:
            raise ServerException(
                'Too many values for "archive_name" configuration.')

        now = datetime.fromtimestamp(self._scheduler.seconds())
        extra = Bunch(
            fullname=source_name,
            archive_extension=extension,
            )

        if len(archive_name) < 2:
            # We have a single configuration for single and multi file.
            return datetime_format(archive_name[0], now, extra)

        if members_size == 1:
            # Archive with a single member.
            return datetime_format(archive_name[0], now, extra)

        # Use name for multi file archive.
        return datetime_format(archive_name[1], now, extra)

    @defer.inlineCallbacks
    def _writeZIP(self, archive, members_data, source_base_path):
        for source_path, _, source_name in members_data:

            source = None
            try:
                source = yield self._location.getStreamForReading(
                    source_path)

                try:
                    # We store the relative path in the archive.
                    member_path = self._location.relativePath(
                        source_base_path, source_path)
                except ServerException:
                    # Just use the filename in root when there is no
                    # relation.
                    member_path = source_name

                # Start new ZIP member.
                archive.writeHeader(path=member_path)

                while True:
                    chunk = yield source.read(self._CHUNK_SIZE)
                    archive.writeContentChunk(chunk)
                    if not chunk:
                        break

            finally:
                if source:
                    # Finalize the ZIP member.
                    archive.writeFooter()
                    yield source.close()

            yield self._cbDeleteSource(None, source_path)
