# Copyright (c) 2015 Adi Roiban.
# See LICENSE for details.
"""
Event handler that stores events in a database.
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from twisted.internet import defer
from twisted.internet.task import LoopingCall
from zope.interface import implementer

from chevah.server import json, force_unicode
from chevah.server.commons.constant import LOG_TABLE_NAME, STATE
from chevah.server.commons.exception import ServerException
from chevah.server.commons.interface import IEventHandler
from chevah.server.commons.queue import (
    BatchProcessQueue,
    IBatchProcessor,
    MaxCapacityReachedException,
    )

from chevah.server.db.entity import DBTable, DBTableColumn, DBTableIndex
from chevah.server.db.interaction import DBInteractionSQLite
from chevah.server.event.handler import EventHandlerWithFilter
from chevah.server.event.stdlib_logging import EventLogRecordFormatterAdapter


class LogEntryDBTable(DBTable):
    """
    Log entry as persisted in the database.
    """

    def __init__(self, name):
        super(LogEntryDBTable, self).__init__(name)

        # Columns definitions.
        self.columns.add(DBTableColumn(
            'id',
            'INTEGER',
            primary_key=True,
            auto_increment=True))
        self.columns.add(DBTableColumn(
            'timestamp',
            'DOUBLE',
            description='Timestamp of log event creation.',
            ))
        self.columns.add(DBTableColumn(
            'server_name',
            'VARCHAR',
            size=180,
            description='Name of the server generating the entry.',
            ))
        self.columns.add(DBTableColumn(
            'server_uuid',
            'VARCHAR',
            size=180,
            description='UUID of the server generating the entry.',
            ))
        self.columns.add(DBTableColumn(
            'client_peer',
            'VARCHAR',
            size=30,
            description='IP/port of client peer.',
            ))
        self.columns.add(DBTableColumn(
            'event_id',
            'VARCHAR',
            size=80,
            description='Unique ID event.',
            ))
        self.columns.add(DBTableColumn(
            'message',
            'TEXT',
            description='Message associated with the event.',
            ))
        self.columns.add(DBTableColumn(
            'data',
            'TEXT',
            description='JSON of event data argument(s).',
            ))
        self.columns.add(DBTableColumn(
            'user_name',
            'VARCHAR',
            size=180,
            description='Name of the user (optional).',
            ))
        self.columns.add(DBTableColumn(
            'service_name',
            'VARCHAR',
            size=180,
            description='Name of the service (optional).',
            ))
        self.columns.add(DBTableColumn(
            'file_name',
            'TEXT',
            description='Name of the filename (optional).',
            ))

        # Index definitions.
        idx_timestamp = DBTableIndex('idx_timestamp')
        idx_timestamp.addColumn(self.columns.column('timestamp'))
        self.indexes.addIndex(idx_timestamp)

        idx_event = DBTableIndex('idx_event')
        idx_event.addColumn(self.columns.column('event_id'))
        self.indexes.addIndex(idx_event)

        idx_peer = DBTableIndex('idx_peer')
        idx_peer.addColumn(self.columns.column('client_peer'))
        self.indexes.addIndex(idx_peer)

        idx_user = DBTableIndex('idx_user')
        idx_user.addColumn(self.columns.column('user_name'))
        self.indexes.addIndex(idx_user)

    @staticmethod
    def getRowValuesFromEvent(event):
        """
        Return a dictionary with `DBLogEntry` compatible keys and values
        from the `Event` instance.
        """
        peer = event.account.peer
        client_peer = EventLogRecordFormatterAdapter.formatPeer(peer)

        data = event.data.copy()
        # Rename internal 'avatar' to public 'account' in data member.
        data['account'] = data.pop('avatar', 'unknown')

        # FIXME:5517:
        # Besides filename,
        # also look what to do with the other attributes from here.
        # We need to also store the component that generated it.
        # Maybe just component UUID to minimize the DB.
        service_name = data.get('service', 'unknown')
        file_name = data.get('file_path', 'unknown')
        server_name = data.get('server_name', 'unknown')
        server_uuid = data.get('server_uuid', 'unknown')

        return {
            'event_id': event.id,
            'timestamp': event.created,
            'message': event.message,
            'user_name': event.account.name,
            'client_peer': client_peer,
            'service_name': service_name,
            'file_name': file_name,
            'server_name': server_name,
            'server_uuid': server_uuid,
            # FIXME:2666:
            # Change json.dumps to always return unicode.
            'data': force_unicode(json.dumps(data)),
            }


@implementer(IEventHandler, IBatchProcessor)
class DatabaseEventHandler(DBInteractionSQLite, EventHandlerWithFilter):
    """
    An event handler for storing events inside a database table.

    This is an event handler and a DB interaction.
    """

    PROPERTIES_REQUIRING_RESTART = ('path',)

    _TABLE_NAME = LOG_TABLE_NAME
    # Every 6 hours to spread the load.
    _AUTO_DELETE_INTERVAL = 4 * 60 * 60

    _table = None
    # Call which will periodically schedule the next auto delete.
    _loop_auto_delete = None

    def __init__(self, parent, configuration):
        # Once we migrate to go we should have a dedicated interaction.
        # This is left her to make it easier to identify DB operations.
        self._interaction = self
        super(DatabaseEventHandler, self).__init__(parent, configuration)

        self._definitions = self.root.event_dispatcher.definitions

    def _onStart(self, avatar=None):
        """
        See: `Runnable`.
        """
        self._queue = BatchProcessQueue(
            processor=self, capacity=self.configuration.queue_size)

        # Auto delete is a looping call which should never fail.
        # It is started when migration is done for the first time
        # and continues to run, even when the DB is stopped, as long
        # as the handler is started.
        self._loop_auto_delete = LoopingCall(self._clAutoDelete)
        self._loop_auto_delete.clock = self._scheduler

        # FIXME:5949:
        # We need to have it called `_table` to allow access from JSON-RPC.
        self._table = LogEntryDBTable(name=self._TABLE_NAME)

        def cb_db_started(result):
            """
            Called when DB was started.
            """
            # We need to hurry setting the STARTED state, as otherwise
            # the interaction can't be used... and we can't try to
            # create the table.
            self._state = STATE.STARTED
            return self._interaction.migrateTable(self._table)

        def cb_db_ready(result):
            """
            Called when DB is ready for operations.
            """
            self._loop_auto_delete.start(
                self._AUTO_DELETE_INTERVAL, now=True)
            return self._queue.start()

        deferred = self._startDB()
        deferred.addCallback(cb_db_started)
        deferred.addCallback(self._cbCheckPermissions)
        deferred.addCallback(cb_db_ready)
        return deferred

    def _onStop(self, avatar=None):
        """
        See: `Runnable`.

        Stop queue and clear leftover events.
        """
        self._table = None

        if self._loop_auto_delete and self._loop_auto_delete.running:
            self._loop_auto_delete.stop()
        self._loop_auto_delete = None

        deferred = self._queue.stop()
        deferred.addBoth(lambda _: self._stopDB())
        return deferred

    @defer.inlineCallbacks
    def _cbCheckPermissions(self, ignored):
        """
        Check INSERT/UPDATE/DELETE permissions for configured database and
        table.
        """
        row = {
            'event_id': '100',
            'timestamp': 0,
            'server_name': '',
            'server_uuid': '',
            'client_peer': '',
            'message': '',
            'data': '',
            'user_name': '',
            'service_name': '',
            'file_name': '',
            }
        yield self._interaction.insertRow(self._table, row)
        values = {'timestamp': 1}
        criteria = [
            [(self._table.columns.column('event_id').name, '=', '100')],
            [(self._table.columns.column('server_uuid').name, '=', '')],
            ]
        yield self._interaction.updateRows(self._table, values, criteria)
        yield self._interaction.deleteRows(self._table, criteria)

    def _onEvent(self, event):
        """
        See: `EventHandlerMixin`.
        """
        try:
            self._queue.append(event)
        except MaxCapacityReachedException as error:
            return defer.fail(ServerException(error.message))

        return defer.succeed(None)

    def process(self, batch):
        """
        See: `IBatchProcessor`.
        """
        rows = []
        for event in batch:
            values = LogEntryDBTable.getRowValuesFromEvent(event)
            rows.append(values)

        deferred = self._interaction.insertRows(self._table, rows)
        return deferred

    def _clAutoDelete(self):
        """
        Called regularly to do the auto delete.
        """
        auto_delete = self.configuration.auto_delete
        if auto_delete <= 0:
            # Nothing to auto-delete for now.
            return

        if not self._interaction.operating:
            # Database is not yet started.
            return

        def eb_auto_delete(failure):
            """
            Called when failed to delete older events.
            """
            self.emitEvent(
                '20062', data={'details': force_unicode(failure.value)})

        max_age = self._scheduler.seconds() - auto_delete * 24 * 60 * 60

        criteria = [[
            (self._table.columns.column('timestamp'), '<', max_age),
            ]]
        deferred = self._interaction.deleteRows(self._table, criteria)
        deferred.addErrback(eb_auto_delete)
        # To help with testing to wait for the delete action.
        self._auto_delete_deferred = deferred
        return deferred
