# Copyright (c) 2015 Adi Roiban.
# See LICENSE for details.
"""
Event handler sending a copy of the event as email.
"""
from io import BytesIO
from email.utils import getaddresses

from zope.interface import implementer

from chevah_compat import local_filesystem
from chevah.server.commons.constant import (
    EMAIL_ASSOCIATED_FILES_NAME,
    MAX_ALLOWED_ATTACHMENT_SIZE,
    TYPE_NAME,
    )
from chevah.server.commons.exception import ServerException
from chevah.server.commons.interface import (
    IEmailSenderEventHandlerConfiguration,
    )
from chevah.server.event.configuration import (
    _EventHandlerConfigurationBase,
    FilterOptionsMixin,
    )
from chevah.server.event.handler import EventHandlerWithFilter
from chevah.server.configuration.model import (
    WritableCSVUnique,
    WritableString,
    WritableStringOrEmptyCaseInsensitive
    )


@implementer(IEmailSenderEventHandlerConfiguration)
class EmailSenderEventHandlerConfiguration(
        _EventHandlerConfigurationBase, FilterOptionsMixin):
    """
    Configuration for an email sender event handler.
    """
    email_to_recipients = WritableCSVUnique(
        'email_to_recipients')
    email_cc = WritableCSVUnique('email_cc')
    email_bcc = WritableCSVUnique('email_bcc')
    email_subject = WritableString('email_subject')
    email_body = WritableString('email_body')
    email_associated_files = WritableStringOrEmptyCaseInsensitive(
        'email_associated_files')


class EmailSenderEventHandler(EventHandlerWithFilter):
    """
    Sends events over SMTP as emails.
    """
    PROPERTIES_REQUIRING_RESTART = (
        'email_associated_files',
        )

    # For dependency injection during testing.
    _filesystem = local_filesystem
    # Email client used to s end the files.
    _client = None
    # How to handle the associated file in the email.
    _associated_files = None

    def _onStart(self, avatar=None):
        """
        See: `Runnable`.
        """
        self._client = self.root.resources.getEmailClient()

        email_subject = self.configuration.email_subject
        if not email_subject:
            raise ServerException(
                'Emails need a configured subject.')

        associated_files = self.configuration.email_associated_files
        if associated_files not in EMAIL_ASSOCIATED_FILES_NAME:
            raise ServerException(
                'Invalid value for "email_associated_files": %s' %
                (associated_files,)
                )
        self._associated_files = associated_files

    def _onStop(self, avatar=None):
        """
        See: `Runnable`.
        """
        self._client = None
        self._associated_files = None

    def _onEvent(self, event):
        """
        See: `EventHandlerWithFilterMixin`.
        """
        email_subject = self.configuration.email_subject
        subject = event.renderFromTemplate(email_subject, 'email_subject')
        email_body = self._configuration.email_body
        body = event.renderFromTemplate(email_body, 'email_body')

        # If no `to` is resolved, the resource will send the email to
        # default recipient.
        to = self._resolveEmails(self.configuration.email_to_recipients, event)
        cc = self._resolveEmails(self.configuration.email_cc, event)
        bcc = self._resolveEmails(self.configuration.email_bcc, event)

        return self._client.send(
            to=to,
            cc=cc,
            bcc=bcc,
            subject=subject,
            body=body,
            attachments=self._getAttachment(event),
            )

    def _resolveEmails(self, emails_or_uuid, event):
        """
        Return a list with emails or resolved emails from UUIDS.

        * Account UUIDs resolve to their configured emails if any.
        * Group UUIDs resolve to their associated accounts' emails.
        """
        result = []
        for value in emails_or_uuid:
            result.extend(self._resolveEmail(value, event))
        # We use a set to reduce duplicates.
        return tuple(sorted(set(result)))

    def _resolveEmail(self, email_or_uuid, event):
        """
        Return a list with resolved emails for a single email or uuid.

        When we have group UUID, it returns all the emails from that group.
        """
        # Try to see if we have a template.
        # We use `email` as generic error reporting since this is use form
        # many other fields.
        rendered_email = event.renderFromTemplate(
            email_or_uuid, 'email_address')

        # Not an account or a group.
        if not rendered_email:
            # Maybe configured with a template that has no result.
            # Just skip.
            return []

        group = self.root.configuration.groups.getSection(rendered_email)
        if group:
            result = []
            for account in self.root.configuration.accounts.getAll():
                if group.uuid not in account.group:
                    continue
                result.extend(_get_account_emails(account))

            # If group is empty, or no user has email configured, we might
            # return an empty list.
            return result

        account = self.root.configuration.accounts.getSection(rendered_email)
        if account:
            return _get_account_emails(account)

        # Just return the raw value or the transformed one.
        # If we are here, we have no UUID.
        return [rendered_email]

    def _getAttachment(self, event):
        """
        Return a dictionary with an attachment associated with the given
        `event`.
        """
        if (self._associated_files != TYPE_NAME.ATTACHMENT):
            # No need to attach anything.
            return {}

        real_path = event.data.get('real_path')
        if not real_path:
            # No file path is attached to this event.
            return {}

        segments = self._filesystem.getSegmentsFromRealPath(real_path)

        filesize = self._filesystem.getFileSize(segments)
        if filesize > MAX_ALLOWED_ATTACHMENT_SIZE:
            message = 'File size %s larger than limit %s for \"%s\"' % (
                filesize, MAX_ALLOWED_ATTACHMENT_SIZE, real_path)
            raise ServerException(message)

        filename = segments[-1]
        # FIXME:3666:
        # The file can be huge and this could be cause
        # of a crash. For now we have limitation on file size, but
        # the better solution is to implement non-blocking email
        # sendinassociated_filesg. This method should return just file-like
        # objects which are not opened yet. And then the sending
        # code can open, read chuck, close them on demand.
        with self._filesystem.openFileForReading(segments) as f:
            attachment = {
                filename: BytesIO(f.read()),
                }

        return attachment

    def _isOwnEvent(self, event):
        """
        See: `EventHandlerMixin`.
        """
        # FIXME:3696:
        # Remove once support for resources is implemented.

        # Treat events emitted by the email client as they were my own.
        if self._client is not None and event.creator is self._client:
            return True

        return super(EmailSenderEventHandler, self)._isOwnEvent(event)


def _get_account_emails(account):
    """
    Return the list of emails for account.
    """
    result = []
    emails = getaddresses([account.email])
    for _, email in emails:
        result.append(email)

    return result
