# Copyright (c) 2017 Adi Roiban.
# See LICENSE for details.
"""
Event handler which delegated the handling to an extension
"""
from __future__ import unicode_literals
from __future__ import absolute_import
import types

from bunch import Bunch
from lazr.delegates import delegate_to
from twisted.internet import defer, reactor
from twisted.internet.task import deferLater
from zope.interface import implementer

from chevah.server import json
from chevah.server.commons.exception import ServerException
from chevah.server.commons.interface import (
    IEventCreator,
    IExtensionAvatar,
    IPublicAvatar,
    )
from chevah.server.event.handler import EventHandlerWithFilter
from chevah.server.extension import load_extension_object
import six


@implementer(IExtensionAvatar)
@delegate_to(IPublicAvatar, context='_context')
class ExtensionAvatar(object):
    def __init__(self, context):
        self._context = context

    @property
    def token(self):
        return getattr(self._context, 'token', '')

    def __eq__(self, candidate):
        """
        All `ExtensionAvatar` instances are equal if they have the same
        public values.
        """
        if not IPublicAvatar.providedBy(candidate):
            return False

        return (
            self.uuid == candidate.uuid
            and self.name == candidate.name
            and self.peer == candidate.peer
            and self.component == candidate.component
            and self.token == candidate.token
            )

    def __ne__(self, candidate):
        """
        See: `__eq__`.
        """
        return not self.__eq__(candidate)


@implementer(IEventCreator)
class ExtensionParent(object):
    """
    A wrapper for the information available to the extension.
    """
    type = 'extension'

    def __init__(self, uuid, name, configuration, event_emitter):
        self._uuid = uuid
        self._name = name
        self._configuration = configuration
        self._event_emitter = event_emitter

    @property
    def parent(self):
        return None

    @property
    def root(self):
        return self

    @property
    def uuid(self):
        return self._uuid

    @property
    def name(self):
        return self._name

    @property
    def configuration(self):
        return self._configuration

    def emitEvent(self, event_id, data, avatar=None):
        """
        Guard emitting events from the extensions.

        `avatar` is ignored as it is always emitted to the avatar
        associated to the event handler.
        """
        try:
            # FIXME:2635:
            # Make sure the emit_data has no custom avatar
            # so that we don't allow the extension to overwrite the avatar.
            # This can be removed once data['avatar'] is always ignored.
            del data['avatar']
        except KeyError:
            """
            All ok. No avatar in extension data.
            """

        self._event_emitter(event_id, data=data)

    def delay(self, seconds, callback, *args, **kwargs):
        """
        Helper to provide a non-Twisted API to extension users.
        """
        return deferLater(reactor, seconds, callback, *args, **kwargs)


class ExtensionEventHandler(EventHandlerWithFilter):
    """
    Delegate the event handler to an extension.
    """
    PROPERTIES_REQUIRING_RESTART = (
        'entry_point',
        'configuration',
        )

    _extension = None

    def _onStart(self, avatar=None):
        """
        See: `Runnable`.

        Check that configuration is valid.
        """
        entry_point = self.configuration.entry_point
        # The extension is re-loaded again in the worker process.
        # We also load it here for main process functionality like
        # configuration parsing and start/stop calls.
        extension = load_extension_object(entry_point)

        self._extension = extension()

        if not isinstance(self._extension.handle, types.FunctionType):
            raise ServerException(
                '"handle()" need to be a @staticmethod.')

        self._entry_point = entry_point

        return self._extension.onStart(ExtensionParent(
            uuid=self.uuid,
            name=self.name,
            configuration=self.configuration.configuration,
            event_emitter=self.emitEvent
            ))

    def _onStop(self, avatar=None):
        """
        See: `Runnable`.
        """
        if self._extension:
            deferred = defer.maybeDeferred(self._extension.onStop)
        else:
            deferred = defer.succeed(None)
        self._extension = None
        self._entry_point = None
        return deferred

    @defer.inlineCallbacks
    def _onEvent(self, event):
        """
        See: EventHandlerWithFilterMixin.
        """
        # We sanitize the event to make sure only the public attributes
        # are available.
        private_event = event
        public_event = Bunch.fromDict(json.loads(json.dumps(private_event)))
        public_event['account'] = ExtensionAvatar(context=private_event.account)
        # Make sure we don't accidentally use the private event.
        event = public_event

        result = yield self._extension.getConfiguration(public_event)

        if result is None:
            # This event should be skipped without any extra event emitting.
            return

        try:
            configuration, events = result
        except ValueError:
            # FIXME:6521:
            # This was changed in v4.
            # We keep it for a while.
            # This should be removed in version 6.
            configuration = result
            events = []

        self._emitExtensionEvents(events, avatar=private_event.account)

        if configuration is None:
            # This event should be skipped.
            return

        result = yield self.root.workers.handleEvent(
            self._entry_point,
            json.dumps(public_event),
            configuration,
            )

        self._emitExtensionEvents(result, avatar=private_event.account)

    def _emitExtensionEvents(self, events, avatar):
        """
        Emit the events generated by the extension.
        """
        if not events:
            events = []

        if isinstance(events, six.string_types):
            # We have a simple text event.
            events = [{'event_id': '20200', 'message': events}]

        if isinstance(events, dict):
            # We have a single dict event.
            events = [events]

        for emit_data in events:

            if isinstance(emit_data, tuple):
                # See if we have a legacy tuple event emitting rule.
                event_id, emit_data = emit_data
                emit_data['event_id'] = event_id

            event_id = emit_data.get('event_id', '20200')
            emit_data['message'] = emit_data.get(
                'message', 'No message returned from the extension.')
            # We emit the extension event using the avatar from the original
            # event, to be able to audit the avatar that triggered this event.
            try:
                # FIXME:2635:
                # Make sure the emit_data has no custom avatar
                # so that we don't allow the extension to overwrite the avatar.
                # This can be removed once data['avatar'] is always ignored.
                del emit_data['avatar']
            except KeyError:
                """
                All ok. No avatar in extension data.
                """
            self.emitEvent(event_id, avatar=avatar, data=emit_data)
