# Copyright (c) 2019 Adi Roiban.
# See LICENSE for details.
"""
Event handler which calls external executables.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

from twisted.internet import defer
from zope.interface import implementer

from chevah.server import force_unicode, json
from chevah.server.commons.constant import (
    EXECUTABLE_OUTPUT_FORMAT_NAME,
    TYPE_NAME,
    )
from chevah.server.commons.exception import ServerException
from chevah.server.commons.interface import (
    IExternalExecutableEventHandlerConfiguration,
    )
from chevah.server.configuration.model import (
    WritableString,
    WritableStringCaseInsensitive,
    WritableRawString,
    WritableNumber,
    WritableInteger,
    )
from chevah.server.event.configuration import (
    _EventHandlerConfigurationBase,
    FilterOptionsMixin,
    )
from chevah.server.process import shlex3
from chevah.server.process.spawn import defer_spawn
from chevah.server.event.handler import EventHandlerWithFilter


@implementer(IExternalExecutableEventHandlerConfiguration)
class ExternalExecutableEventHandlerConfiguration(
    _EventHandlerConfigurationBase,
    FilterOptionsMixin,
        ):
    """
    Configuration for an event handler calling external executables.
    """
    executable_path = WritableString('executable_path')
    executable_arguments = WritableRawString('executable_arguments')
    executable_output_format = WritableStringCaseInsensitive(
        'executable_output_format', default=TYPE_NAME.TEXT)
    executable_timeout = WritableNumber('executable_timeout', default=30)
    concurrent_processes = WritableInteger('concurrent_processes', default=2)


class ExternalExecutableEventHandler(EventHandlerWithFilter):
    """
    Call an external executable with event data.
    """
    PROPERTIES_REQUIRING_RESTART = (
        'concurrent_processes',
        'executable_timeout',
        'executable_output_format',
        )
    _timeout = 0

    def _onStart(self, avatar=None):
        """
        See: `Runnable`.
        """
        if not self._configuration.executable_path:
            raise ServerException('"executable_path" is not configured.')

        self._timeout = self._configuration.executable_timeout
        if self._timeout < 1:
            raise ServerException(
                '"executable_timeout" must be at least 1 second.')

        self._concurrency = self._configuration.concurrent_processes
        if self._concurrency < 1:
            raise ServerException(
                '"concurrent_processes" must be at least 1.')

        self._executable_output_format = (
            self._configuration.executable_output_format)
        if self._executable_output_format not in EXECUTABLE_OUTPUT_FORMAT_NAME:
            raise ServerException(
                'Unknown value for '
                '"executable_output_format": "{}".'.format(
                    self._executable_output_format
                    ))

        self._inititateState()

    def _inititateState(self):
        """
        Set up the initial internal state.
        """
        self._queue = []
        self._active_count = 0

    def _onStop(self, avatar=None):
        """
        See: `Runnable`.
        """
        self._inititateState()

    def _onEvent(self, event):
        """
        See: EventHandlerWithFilterMixin.
        """
        executable = self._configuration.executable_path
        if not executable:
            raise ServerException('"executable_path" is not configured.')

        raw_arguments = event.renderFromTemplate(
            self._configuration.executable_arguments,
            'executable_arguments'
            )
        raw_command = '%s %s' % (executable, raw_arguments)

        command = [executable] + shlex3.split(raw_arguments, posix=True)
        environment = {
            'EVENT': event.id,
            'HANDLER_UUID': self._configuration.uuid,
            'TIMESTAMP': force_unicode(event.timestamp.timestamp),
            'PEER': '%s:%s' % (
                event.account.peer.host, event.account.peer.port),
            'USER': event.account.name,
            'COMPONENT_UUID': event.component.uuid,
            }

        for key, value in event.data.items():
            # FIXME:2635:
            # Data should not contain the `avatar` as now we have a dedicated
            # `account` member
            if key == 'avatar':
                continue
            environment['DATA_' + key] = force_unicode(value)

        deferred = defer.Deferred()
        self._queue.append((deferred, command, environment, raw_command))
        self._continue()
        return deferred

    def _continue(self):
        """
        Continue the execution of an event.
        """
        if self._active_count >= self._concurrency:
            # No more room to execute anything.
            # Pause and will resume once a slot is done.
            return

        try:
            wait_deferred, arguments, environment, command = self._queue.pop(0)
        except IndexError:
            # No more tasks.
            return

        self._active_count += 1

        # FIXME:5251:
        # Add active deferred to a list and cancel when handler stops.
        deferred = defer_spawn(
            arguments,
            environment=environment,
            timeout=self._timeout,
            )
        deferred.addCallback(self._cbSpawn, wait_deferred, command)
        deferred.addErrback(self._ebSpawn, wait_deferred, command)

    def _cbSpawn(self, result, wait_deferred, raw_command):
        """
        Called after spawn was finalized with success.
        """
        exit_code, output, error = result

        def sanitize_log(full_text):
            """
            Return a value that can be displayed in the logs.
            """
            result = full_text[:200]
            result = result.replace('\r\n', ' ')
            return result.replace('\n', ' ')

        raw_output = output
        if self._executable_output_format == TYPE_NAME.JSON:
            try:
                output = json.loads(output)
            except Exception:
                raise ServerException(
                    'Invalid JSON output. Output: {}'.format(
                        sanitize_log(raw_output)))

        data = {
            'command': raw_command,
            'exit_code': exit_code,
            'output': output,
            'error': error,
            'output_log': sanitize_log(raw_output),
            'error_log': sanitize_log(error),
            }

        self.emitEvent('20053', data=data)

        self._active_count -= 1
        self._continue()

        wait_deferred.callback(None)

    def _ebSpawn(self, failure, wait_deferred, raw_command):
        """
        Called after spawn failed.
        """
        details = force_unicode(failure.value)
        self.emitEvent('20054', data={
            'command': raw_command,
            'tb': force_unicode(failure.getBriefTraceback()),
            'details': details,
            })
        wait_deferred.errback(
            ServerException(
                'Failed to execute external command. %s' % (details,)))

        self._active_count -= 1
        self._continue()
