# Copyright (c) 2018 Adi Roiban.
# See LICENSE for details.
"""
Event handler which extracts/uncompress files.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

from contextlib import closing
import gzip
import tarfile
import zipfile
import six
try:
    import bz2
except ImportError:
    bz2 = None

from twisted.internet import defer

from chevah.server.commons.constant import (
    DEFAULT_LOCAL_FILESYSTEM_UUID,
    )
from chevah.server.commons.exception import (
    ServerException,
    )
from chevah.server.event.handler import EventHandlerWithFilter
from chevah.server.location.commons import (
    create_destination_parents,
    get_target_path,
    )
from chevah.server.event.common import (
    CheckDestinationMixin,
    SkipHandling,
    SourceAttributeMixin,
    )


class ExtractArchiveEventHandler(
        EventHandlerWithFilter, SourceAttributeMixin, CheckDestinationMixin):
    """
    Extract the source file associated with the event.
    """
    PROPERTIES_REQUIRING_RESTART = (
        'password_update',
        )

    def _onStart(self, avatar=None):
        """
        See: `Runnable`.
        """
        self._password = b''
        if self._configuration.password:
            self._password = self._configuration.password.encode('utf-8')
        self._location = self.root.locations.getSection(
            DEFAULT_LOCAL_FILESYSTEM_UUID)

    def _onStop(self, avatar=None):
        """
        See: `Runnable`.
        """
        self._password = None
        self._location = None

    def _getHandler(self, source_name):
        """
        See: SourceAttributeMixin.
        """
        lower_name = source_name.lower()

        if lower_name.endswith('.tar'):
            return (self._handleExtractTAR, 'extract TAR')

        if lower_name.endswith('.tar.gz'):
            return (self._handleExtractTARGZ, 'extract TAR GZIP')

        if bz2 and lower_name.endswith('.tar.bz2'):
            return (self._handleExtractTARBZ2, 'extract TAR BZ2')

        if lower_name.endswith('.gz'):
            return (self._handleExtractGZIP, 'extract GZIP')

        if lower_name.endswith('.zip'):
            return (self._handleExtractZIP, 'extract ZIP')

        raise ServerException(
            'Format not supported for "%s".' % (source_name,))

    def _emitMember(self, operation, source_path, destination_path):
        """
        Emit event for handling of a member of the archive.
        """
        self.emitEvent(
            '20187',
            data={
                'operation': operation,
                'path': destination_path,
                'real_path': destination_path,
                'source_path': source_path,
                },
            )

    @defer.inlineCallbacks
    def _handleExtractGZIP(
            self, source_path, source_name, destination_base, event, operation):
        """
        Extract simple GZIP file.

        Since a GZIP contains a single file, it will not emit separate
        events the extracted file.

        There is only the final event emitted when the whole archive is
        extracted.
        """
        # Remove .gz extension.
        destination_name = source_name[:-3]
        destination_path = self._location.joinPath(
            destination_base, destination_name)

        yield self._checkDestination(source_path, destination_path, operation)

        source = None
        destination = None
        try:
            source = yield self._location.getStreamForReading(source_path)
            source_stream = source._context._stream

            with closing(
                    gzip.GzipFile(mode='rb', fileobj=source_stream)
                    ) as archive:

                # Do one read before opening the destination as a way
                # to make sure source is looks like a valid GZPIP.
                chunk = archive.read(self._CHUNK_SIZE)

                destination = yield self._location.getStreamForWriting(
                    destination_path)
                yield destination.write(chunk)

                while True:
                    chunk = archive.read(self._CHUNK_SIZE)
                    if not chunk:
                        break

                    yield destination.write(chunk)
        finally:
            if source:
                yield source.close()
            if destination:
                yield destination.close()

        defer.returnValue(destination_path)

    @defer.inlineCallbacks
    def _handleExtractZIP(
            self, source_path, source_name, destination_base, event, operation):
        """
        Extract ZIP archive.

        It extract it in recursive way.
        Will emit an event for each extracted files.
        """
        @defer.inlineCallbacks
        def extract_file(archive, member, path):
            """
            `path` is filesystem encoded path.
            """
            with archive.open(member, pwd=self._password) as source:
                yield self._writeFile(path, destination_base, source)

        source = None
        try:
            source = yield self._location.getStreamForReading(source_path)
            source_stream = source._context._stream

            with closing(
                    zipfile.ZipFile(file=source_stream, mode='r')
                    ) as archive:
                for member in archive.infolist():
                    # Filename is already decode if Archive has the right
                    # flag bits.
                    destination_path = get_target_path(
                        member.filename, destination_base)

                    if destination_path is None:
                        # We have a root path.
                        continue

                    is_file = member.filename[-1] != '/'

                    yield self._extractFile(
                        archive, member, is_file,
                        destination_path, source_path,
                        operation,
                        extract_callback=extract_file,
                        )
        finally:
            if source:
                yield source.close()

        defer.returnValue(destination_base)

    def _handleExtractTARGZ(
            self, source_path, source_name, destination_base, event, operation):
        """
        Extract TAR.GZ  file.
        """
        return self._handleExtractTAR(
            source_path, source_name, destination_base, event,
            operation='extract TAR GZIP')

    def _handleExtractTARBZ2(
            self, source_path, source_name, destination_base, event, operation):
        """
        Extract TAR.BZ2 file.
        """
        return self._handleExtractTAR(
            source_path, source_name, destination_base, event,
            operation='extract TAR BZ2')

    @defer.inlineCallbacks
    def _extractFile(
            self, archive, member, is_file,
            destination_path, source_path, operation, extract_callback
            ):
        """
        Shared code to extract a file from an archive.
        """
        exists = yield self._location.exists(
            destination_path, with_event=False)

        if not is_file:
            # We are requested to extract a directory.
            if exists:
                # Directory already exists.
                # All file.
                return

            # We have a directory with doesn't exist.
            yield self._location.createFolder(destination_path)
            self._emitMember(
                operation=operation + ' folder member',
                source_path=source_path,
                destination_path=destination_path
                )
            return

        try:
            yield self._checkDestination(
                source_path, destination_path,
                operation=operation,
                pre_exist=exists,
                )
        except SkipHandling:
            return

        yield extract_callback(archive, member, destination_path)
        self._emitMember(
            operation=operation + ' file member',
            source_path=source_path,
            destination_path=destination_path
            )

    @defer.inlineCallbacks
    def _writeFile(self, destination_path, destination_base, source_stream):
        """
        Write a local file.
        """
        try:
            target = yield self._location.getStreamForWriting(destination_path)
        except Exception:
            relative_path = self._location.relativePath(
                destination_base, destination_path)
            yield create_destination_parents(
                location=self._location,
                base_path=destination_base,
                target_path=relative_path)
            target = yield self._location.getStreamForWriting(destination_path)

        try:
            while True:
                chunk = source_stream.read(self._CHUNK_SIZE)
                if not chunk:
                    break
                yield target.write(chunk)
        finally:
            yield target.close()

    @defer.inlineCallbacks
    def _handleExtractTAR(
        self, source_path, source_name, destination_base,
        event, operation='extract TAR'
            ):
        """
        Extract TAR, TAR.GZ or TAR.BZ2 file.
        """
        @defer.inlineCallbacks
        def extract_file(archive, member, path):
            """
            `path` is filesystem encoded path.
            """
            source = archive.extractfile(member)
            yield self._writeFile(path, destination_base, source)

        source = None
        try:
            source = yield self._location.getStreamForReading(source_path)
            source_stream = source._context._stream

            with closing(
                    tarfile.TarFile.open(
                        fileobj=source_stream,
                        mode='r',
                        dereference=True,
                        encoding='utf-8',
                        )
                    ) as archive:
                for member in archive.getmembers():

                    is_file = member.isreg()
                    if not (is_file or member.isdir()):
                        raise ServerException(
                            'Only TAR archives consisting of regular files '
                            'and directories are supported.'
                            )

                    member_name = member.name
                    if six.PY2:
                        member_name = member_name.decode('utf-8')
                    destination_path = get_target_path(
                        member_name, destination_base)
                    if destination_path is None:
                        # We have a root path.
                        continue

                    yield self._extractFile(
                        archive, member, is_file,
                        destination_path, source_path,
                        operation,
                        extract_callback=extract_file,
                        )
        finally:
            if source:
                yield source.close()

        defer.returnValue(destination_base)
