# Copyright (c) 2023 Adi Roiban.
# See LICENSE for details.
"""
Event handler for generating the checksum of a file.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

import base64
import binascii

from cryptography.hazmat.primitives import hashes
from twisted.internet import defer
from zope.interface import implementer

from chevah.server.commons import DEFAULT_BACKEND
from chevah.server.commons.constant import (
    DEFAULT_LOCAL_FILESYSTEM_UUID,
    DIGEST_FORMATS,
    TYPE_NAME,
    )
from chevah.server.event.common import SourceAttributeMixin
from chevah.server.commons.exception import ServerException
from chevah.server.commons.interface import IFileDigestEventHandlerConfiguration
from chevah.server.configuration.model import (
    WritableStringCaseInsensitive,
    )
from chevah.server.event.configuration import (
    _EventHandlerConfigurationBase,
    FilterOptionsMixin,
    )
from chevah.server.event.handler import EventHandlerWithFilter


_SUPPORTED_DIGEST = {
    'md5': hashes.MD5,
    'sha1': hashes.SHA1,
    'sha224': hashes.SHA224,
    'sha256': hashes.SHA256,
    'sha384': hashes.SHA384,
    'sha512': hashes.SHA512,
    }


@implementer(IFileDigestEventHandlerConfiguration)
class FileDigestEventHandlerConfiguration(
    _EventHandlerConfigurationBase,
    FilterOptionsMixin,
        ):
    """
    Configuration for the file digest event handler.
    """
    @property
    def delete_source_on_success(self):
        """
        We don't want to delete the source file for now.
        Here to be able to use the shared code from SourceAttributeMixin.
        """
        return False

    @property
    def destination_path(self):
        """
        No destination for now.
        Defined here to be able to reuse shared code.
        """
        return ''

    source_attribute = WritableStringCaseInsensitive(
        'source_attribute', default='real_path')
    digest_algorithm = WritableStringCaseInsensitive(
        'digest_algorithm', default='sha512')
    output_format = WritableStringCaseInsensitive(
        'output_format', default=TYPE_NAME.GNU)


class FileDigestEventHandler(EventHandlerWithFilter, SourceAttributeMixin):
    """
    Compute the message digest of a file.

    For now, this only uses the local filesystem, but it is designed to make it
    easy to use it with any other location.
    """
    PROPERTIES_REQUIRING_RESTART = ()

    def _onStart(self, avatar=None):
        """
        See: `Runnable`.
        """
        # Just check that configuration is valid at start.
        self._getAlgorithm()
        output_format = self._configuration.output_format
        if output_format not in list(DIGEST_FORMATS.keys()):
            raise ServerException(
                'Unknown output_format: "{}".'.format(output_format))
        self._location = self.root.locations.getSection(
            DEFAULT_LOCAL_FILESYSTEM_UUID)

    def _onStop(self, avatar=None):
        """
        See: `Runnable`.
        """
        self._location = None

    def _getHandler(self, source_name):
        """
        See SourceAttributeMixin
        We only have a single handler.
        """
        return (self._handlePath, 'digest')

    @defer.inlineCallbacks
    def _handlePath(
            self, source_path, source_name, destination_base, event, operation):
        """
        This is the handler for `SourceAttributeMixin._handleMember()`.
        """
        digest = yield self._getDigest(source_path)
        output_value = self._getOutput(digest, source_name)
        self.emitEvent(
            '20194',
            avatar=event.account,
            data={
                'output': output_value,
                'path': source_path,
                }
            )
        # To comply with the shared code, we return the targeted path.
        # In the future we might write to a checksum file and we can then
        # return that path.
        defer.returnValue(source_path)

    def _getOutput(self, digest, source_name):
        """
        Return the format used for the output.
        """
        output_format = self._configuration.output_format
        raw_digest = digest.finalize()
        base64_digest = base64.b64encode(raw_digest).decode('ascii')
        hex_digest = binascii.hexlify(raw_digest).decode('ascii')

        if output_format == TYPE_NAME.BASE64:
            return base64_digest

        if output_format == TYPE_NAME.GNU:
            return "{} {}".format(hex_digest, source_name)

        alg_name = self._configuration.digest_algorithm.upper()
        if output_format == TYPE_NAME.BSD:
            return "{} ({}) = {}".format(alg_name, source_name, hex_digest)

        raise ServerException(
            'Unknown output_format: "{}".'.format(output_format))

    def _getAlgorithm(self):
        """
        Return the algorithm instance.
        """
        name = self._configuration.digest_algorithm
        digest_class = _SUPPORTED_DIGEST.get(name, None)
        if not digest_class:
            raise ServerException('Digest "{}" is not supported.'.format(name))
        return hashes.Hash(
            digest_class(), backend=DEFAULT_BACKEND)

    @defer.inlineCallbacks
    def _getDigest(self, source_path):
        """
        Return the digest computed for the content of file.
        """
        digest = self._getAlgorithm()
        stream = yield self._location.getStreamForReading(path=source_path)
        try:
            while True:
                data = yield stream.read(self._CHUNK_SIZE)
                if not data:
                    break
                digest.update(data)
        finally:
            yield stream.close()

        defer.returnValue(digest)
