# Copyright (c) 2015 Adi Roiban.
# See LICENSE for details.
"""
Event handler for dispatching file to directories based on a matching
expression.
"""
from __future__ import unicode_literals

from __future__ import absolute_import
from twisted.internet import defer
from zope.interface import implementer

from chevah.server import emit_error_details, emit_from_error, force_unicode
from chevah.server.commons.constant import DEFAULT_LOCAL_FILESYSTEM_UUID
from chevah.server.commons.exception import ServerException, ServerError
from chevah.server.commons.interface import (
    IFileDispatcherEventHandlerConfiguration,
    )
from chevah.server.commons.match import get_matched_groups
from chevah.server.commons.public_format import Bunch, public_format
from chevah.server.configuration.model import (
    WritableInteger,
    WritableNumber,
    WritableString,
    WritableStringCaseInsensitive,
    WritableCSVLines,
    WritableCSVLinesUnique,
    WritableCSVUnique,
    WritableStringOrEmptyCaseInsensitive,
    )
from chevah.server.event.common import (
    CheckDestinationMixin,
    SkipHandling,
    SourceAttributeMixin,
    )
from chevah.server.event.configuration import (
    _EventHandlerConfigurationBase,
    FilterOptionsMixin,
    )
from chevah.server.event.handler import EventHandlerWithFilter
from chevah.server.location.commons import (
    get_file_name_with_timestamp,
    get_file_name_with_unixtime,
    )
from chevah.server.process.spawn import defer_spawn
import six
from six.moves import range


@implementer(IFileDispatcherEventHandlerConfiguration)
class FileDispatcherEventHandlerConfiguration(
        _EventHandlerConfigurationBase, FilterOptionsMixin):
    """
    Configuration for the file dispatcher event handler.
    """

    dispatch_attribute = WritableStringCaseInsensitive('dispatch_attribute')
    dispatch_delay = WritableNumber('dispatch_delay', default=0)
    dispatch_rules = WritableCSVLinesUnique('dispatch_rules')
    fallback_rule = WritableCSVUnique('fallback_rule')
    create_destination_folder = WritableStringOrEmptyCaseInsensitive(
        'create_destination_folder')
    matching_expressions = WritableCSVLines('matching_expressions')
    retry_count = WritableInteger('retry_count', default=0)
    retry_wait = WritableNumber('retry_wait', default=60)
    destination_temporary_suffix = WritableString(
        'destination_temporary_suffix')
    overwrite_rule = WritableString('overwrite_rule')
    executable_timeout = WritableNumber('executable_timeout', default=30)


class FileDispatcherEventHandler(
        EventHandlerWithFilter, SourceAttributeMixin, CheckDestinationMixin):
    """
    Dispatch a source file to one or multiple directories based on one or
    multiple rules.

    It will call _execute_MODE-NAME(filename, filepath, destination) for
    the rule matching the path associated with the event.
    """
    PROPERTIES_REQUIRING_RESTART = (
        'create_destination_folder',
        'dispatch_attribute',
        'dispatch_rules',
        'fallback_rule',
        'matching_expressions',
        )

    _SUPPORTED_MODES = [
        'copy',
        'delete',
        'ignore',
        'move',
        'move-with-timestamp',
        'rename',
        'rename-prepend-unixtime',
        'execute',
        ]

    _NO_DESTINATION_MODES = [
        'delete',
        'ignore',
        ]

    _CREATE_DESTINATION_MODES = [
        '',  # Empty value, means to not create destination.
        'parent',
        ]

    def _onStart(self, avatar=None):
        """
        See: `Runnable`.
        """
        self._location = self.root.locations.getSection(
            DEFAULT_LOCAL_FILESYSTEM_UUID)

        dispatch_attribute = self.configuration.dispatch_attribute
        if not dispatch_attribute:
            raise ServerException('No dispatch attribute was defined.')

        create_destination = self.configuration.create_destination_folder
        if create_destination not in self._CREATE_DESTINATION_MODES:
            raise ServerException(
                'Unknown value for create_destination_folder: "%s".' % (
                    create_destination,))

        # Validate rules at start time to prevent late errors due to
        # invalid configuration.
        rules = self.configuration.dispatch_rules
        if not rules:
            raise ServerException('You need to specify at least one rule.')

        for rule in rules:
            if len(rule) < 2:
                raise ServerException(
                    'Configuration should include at least the dispatching '
                    'mode and the matching rule.'
                    )

            if rule[0] not in self._SUPPORTED_MODES:
                raise ServerException(
                    'Dispatch mode not supported: %s.' % (rule[0],))

            if rule[0] in self._NO_DESTINATION_MODES and len(rule) != 2:
                raise ServerException(
                    'Configured dispatch mode only needs the file matching '
                    'rule. The destination paths is not required.'
                    )

            if rule[0] not in self._NO_DESTINATION_MODES and len(rule) < 3:
                raise ServerException(
                    'Configuration should include the dispatching mode, '
                    'the matching rule and at least one destination path.'
                    )

            # Call acceptance test trigger configuration errors.
            get_matched_groups(
                value='', rule=rule[1], usage='dispatch_rules')

        self._matching_expressions = self._configuration.matching_expressions
        for rule in self._matching_expressions:
            if len(rule) != 2:
                raise ServerException(
                    'Invalid "matching_expressions" configured.')
            # Call acceptance test trigger configuration errors.
            get_matched_groups(
                value='',
                rule=rule[1],
                usage='matching_expressions')

        fallback_rule = self.configuration.fallback_rule
        if fallback_rule:

            if fallback_rule[0] not in self._SUPPORTED_MODES:
                raise ServerException(
                    'Fallback dispatch mode not supported: %s.' % (
                        fallback_rule[0],))

            if (
                fallback_rule[0] in self._NO_DESTINATION_MODES and
                len(fallback_rule) != 1
                    ):
                raise ServerException(
                    'The configured fallback mode doesn\'t need a '
                    'destination path.'
                    )

            if (
                fallback_rule[0] not in self._NO_DESTINATION_MODES and
                len(fallback_rule) < 2
                    ):
                raise ServerException(
                    'Fallback configuration should include the dispatching '
                    'mode and at least one destination path.'
                    )

        self._dispatch_attribute = dispatch_attribute
        self._rules = rules
        self._fallback_rule = fallback_rule
        self._create_destination_folder = create_destination

    def _onStop(self, avatar=None):
        """
        See: `Runnable`.
        """
        self._location = None
        self._rules = None
        self._fallback_rule = None
        self._create_destination_folder = None

    def _onEvent(self, event):
        """
        See: EventHandlerWithFilterMixin.
        """
        dispatch_path = event.data.get(self._dispatch_attribute, None)
        if not dispatch_path:
            raise ServerException(
                'Handler configured for an event which is missing the '
                'dispatch attribute.'
                )

        if isinstance(dispatch_path, six.string_types):
            # Transform single paths to a list.
            dispatch_path = [dispatch_path]

        deferred = defer.Deferred()
        if not self._configuration.dispatch_delay:
            # Handle right away.
            deferred.callback(True)
            file_handler = self._cbHandleFile
        else:
            self._scheduler.callLater(
                self._configuration.dispatch_delay,
                deferred.callback, None,
                )
            file_handler = self._cbHandleFileAfterDelay

        # We can have multiple paths associated with a single event, and
        # we dispatch each of them.
        # The execution is chained so that we don't call them in parallel
        # as it will use more resource and will make testing complicated as
        # we will not know which will be executed first.

        for path in dispatch_path:
            deferred.addCallback(file_handler, path, event)
        return deferred

    def _cbHandleFileAfterDelay(self, ignored, real_path, event):
        """
        Dispatching a single file after the initial delay.
        """
        def cb_exists(result):
            """
            Called when we got the result from source file exist operation.
            """
            if not result:
                # Source no longer exists.
                self.emitEvent(
                    '20124',
                    avatar=event.account,
                    data={'real_path': real_path, 'path': real_path},
                    )
                return

            return self._cbHandleFile(ignored, real_path, event)

        return self._location.exists(
            real_path, with_event=False).addCallback(cb_exists)

    def _cbHandleFile(self, ignored, real_path, event):
        """
        Check the dispatching for a single file.
        """
        for rule in self._rules:
            dispatch_mode = rule[0]
            dispatch_rule = rule[1]
            destinations = rule[2:]
            # Matching is done for the full path.
            match_groups = get_matched_groups(
                value=real_path,
                rule=dispatch_rule,
                usage='dispatch_rules',
                clock=self._scheduler,
                )
            if not match_groups:
                # No matching rule.
                continue

            # Dispatch as normal rule.
            data = {
                'path': real_path,
                'destinations': ', '.join(destinations),
                'mode': dispatch_mode,
                }
            deferred = self._doDispatchAction(
                source_path=real_path,
                configured_destinations=destinations,
                dispatch_mode=dispatch_mode,
                destination_substitutions=match_groups,
                event=event,
                )
            deferred.addCallback(self._cbDispatchFile, '20130', event, data)
            deferred.addErrback(self._ebDispatchFile, '20125', event, data)
            # Will stop at first matching rule.
            return deferred

        # If we are here it means that file was not matched so we look for
        # fallback dispatching.

        if not self._fallback_rule:
            # No need for fallback
            return defer.succeed(None)

        # Dispatch as fallback.
        dispatch_mode = self._fallback_rule[0]
        destinations = self._fallback_rule[1:]
        data = {
            'path': real_path,
            'destinations': ', '.join(destinations),
            'mode': dispatch_mode,
            }
        deferred = self._doDispatchAction(
            source_path=real_path,
            configured_destinations=destinations,
            dispatch_mode=dispatch_mode,
            event=event,
            )
        deferred.addCallback(self._cbDispatchFile, '20109', event, data)
        deferred.addErrback(self._ebDispatchFile, '20110', event, data)
        return deferred

    def _doDispatchAction(
        self, source_path, configured_destinations, dispatch_mode, event,
        destination_substitutions=(),
            ):
        """
        Handle the dispatching action for a single file from `source_path`
        to the `configured_destinations` lists.

        `configured_destinations` is list of paths as found in the
        configuration. Some might need to be resolved to full paths.

        `destination_substitutions` is a tuple of positional elements that
        can be used for creating the destination path.

        Return the list of destination paths which were dispatched.
        The methods might change the destination, and this is why we use
        the returned destinations.
        """

        # We want the substitution to raise an error and not return a deferred
        # so that we will only have the generic event handling event error,
        # without the dispatch action event error.
        substitutions = self._getSubstitution(
            destination_substitutions, event)
        destination_paths = self._resolveDestinations(
            source_path=source_path,
            destinations=configured_destinations,
            substitutions=substitutions,
            dispatch_mode=dispatch_mode,
            )
        return self._doWithRetry(dispatch_mode, source_path, destination_paths)

    @defer.inlineCallbacks
    def _doWithRetry(self, dispatch_mode, source_path, destination_paths):
        """
        Execute `method` and retry on errors.
        """
        method_name = '_execute_%s' % dispatch_mode.replace('-', '_')
        method = getattr(self, method_name, None)

        retry_wait = self._configuration.retry_wait
        retry_count = self._configuration.retry_count
        for retry in range(retry_count + 1):
            error = None
            try:
                result = yield method(source_path, destination_paths)
                defer.returnValue(result)

            except SkipHandling as skip_error:
                # Stop retrying on skip.
                error = skip_error
                break

            except Exception as catch_all_error:
                error = catch_all_error
                if retry < retry_count:
                    # Emit failure to know that action will be retried.
                    data = {
                        'mode': dispatch_mode,
                        'path': source_path,
                        'real_path': source_path,
                        'count': retry_count - retry,
                        'wait': retry_wait,
                        }
                    emit_error_details(
                        error,
                        '20146',
                        data=data,
                        emitter=self.emitEvent,
                        )

                if retry_wait <= 0:
                    # Retry right away.
                    continue

                deferred = defer.Deferred()
                self._scheduler.callLater(
                    self._configuration.retry_wait, deferred.callback, None)
                yield deferred

        # Raise the last error, when no retry was successful.
        raise error

    def _getSubstitution(self, destination_substitutions, event):
        """
        Return the format substitution data available for the file.
        """
        event_data = event.data
        data = event.asDict()

        if not self._matching_expressions:
            return destination_substitutions, data

        for name, expression in self._matching_expressions:
            attribute = event_data.get(name, None)
            if not attribute:
                raise ServerException(
                    'Handler configured for an event which is missing the '
                    'matching expression attribute.'
                    )

            matches = get_matched_groups(
                value=attribute,
                rule=expression,
                usage='source_rules',
                clock=self._scheduler,
                )

            if len(matches) < 1:
                # If we have not match.
                raise ServerException(
                    'Handler configured for an event which didn\'t match the '
                    '"%s" matching expression attribute.' % (name,)
                    )
            if len(matches) < 2:
                # We have a match but without explicit groups..
                raise ServerException(
                    'Handler configured with "matching_expressions" without '
                    'groups for "%s".' % (name,)
                    )
            data[name] = {}
            for i in range(len(matches)):
                data[name][str(i)] = matches[i]
                data[name][str(i) + '_upper'] = matches[i].upper()
                data[name][str(i) + '_lower'] = matches[i].lower()

        return destination_substitutions, data

    def _resolveDestinations(
            self, source_path, destinations, substitutions, dispatch_mode):
        """
        This will transform the `destinations` into full path to files

        Return a list of fully resolved paths.
        """
        result = []

        _, name = self._location.splitPath(source_path)
        for base_destination in destinations:
            # For backward compatibility, only do substitution when
            # the matching rule uses groups.
            # We always use substitution for newer actions.
            if (
                base_destination[-1] in ['/', '\\']
                and dispatch_mode not in ['execute']
                    ):
                # This is a path denoting a directory.
                # We don't need to substitute the file name.
                have_file_substitution = False
            else:
                have_file_substitution = True

            try:
                resolved_destination = public_format(
                    base_destination,
                    *substitutions[0],
                    **Bunch.fromDict(substitutions[1])
                    )
            except Exception as error:
                # String formatting can raise many error types:
                # IndexError, KeyError, AttributeError, TypeError
                # This is why we catch everything here.
                raise ServerException(
                    'destination "%s" '
                    'configured with wrong expression. %s' % (
                        base_destination, error))

            if len(substitutions[0]) < 2 and dispatch_mode not in ['execute']:
                # FIXME:6229:
                # This change was done in v4.
                # This can be remove with a major version update.
                # Backward compatibility for the case when no matching
                # rule is defined.
                # For backward compatibility the configured path is always the
                # base parent directory.
                destination_path = self._location.joinPath(
                    base_destination, name)
            elif have_file_substitution:
                # No need to generate the file name as it is generated by
                # the substitution rule.
                # We still normalize the path.
                destination_path = self._location.joinPath(
                    resolved_destination)
            else:
                # Append the name to the file if destination is a
                # directory.
                destination_path = self._location.joinPath(
                    resolved_destination, name)

            result.append(destination_path)

        return result

    @defer.inlineCallbacks
    def _createDestinationForFile(self, destination_file):
        """
        Check to see if we need to create the path up to `destination_file`.
        """
        if not self._create_destination_folder:
            defer.returnValue(None)

        # We only care about parent path.
        parent, _ = self._location.splitPath(destination_file)

        result = yield self._location.exists(parent, with_event=False)

        if result:
            # Already exists.
            defer.returnValue(None)

        yield self._location.createFolder(parent)

    @defer.inlineCallbacks
    def _execute_delete(self, source_path, destination_paths):
        """
        Execute the `delete` dispatch method.
        `destination_paths` are ignored and the start code should take care of
        validating the configuration so that they are not passed.
        """
        yield self._location.deleteFile(source_path)

        # Delete has no destination.
        defer.returnValue(['no destination for delete'])

    def _execute_ignore(self, source_path, destination_paths):
        """
        Execute the "ignore" action which does nothing.
        """
        return defer.succeed(['no destination for ignore'])

    @defer.inlineCallbacks
    def _execute_copy(self, source_path, destination_paths):
        """
        Execute the `copy` dispatch method.

        Return a new list with the full destination paths.
        """
        # We use a new list just to make sure that the result cant be
        # mutated without affecting the input values.
        result = []
        destination_temporary_suffix = (
            self._configuration.destination_temporary_suffix)

        for destination in destination_paths:

            try:
                yield self._checkDestination(
                    source_path, destination, 'dispatch copy')
            except SkipHandling:
                continue

            yield self._createDestinationForFile(destination)
            copy_destination = destination + destination_temporary_suffix

            yield self._location.copyFile(
                source_path=source_path,
                destination_path=copy_destination,
                overwrite=True,
                )

            if destination_temporary_suffix:
                yield self._location.rename(
                    source_path=copy_destination,
                    destination_path=destination,
                    )

            result.append(destination)

        if not result:
            # All destination files were skipped.
            # Skip so that we will not trigger the success event.
            raise SkipHandling()

        defer.returnValue(result)

    @defer.inlineCallbacks
    def _execute_move(self, source_path, destination_paths):
        """
        Execute the `move` dispatch method.

        Return a new list with the full destination paths.
        """
        result = yield self._execute_copy(source_path, destination_paths)
        yield self._location.deleteFile(source_path)
        defer.returnValue(result)

    def _execute_move_with_timestamp(self, source_path, destination_paths):
        """
        Execute the `move-with-timestamp` dispatch method.

        This will amend the destination paths with timestamp and then
        use the normal move code.
        """
        amended_destinations = []

        for destination in destination_paths:
            parent, name = self._location.splitPath(destination)
            amended_name = get_file_name_with_timestamp(
                name=name,
                seconds=self._scheduler.seconds(),
                random_mark=self._random.randint(0, 100),
                )
            amended_destinations.append(
                self._location.joinPath(parent, amended_name))

        return self._execute_move(source_path, amended_destinations)

    @defer.inlineCallbacks
    def _execute_rename_prepend_unixtime(self, source_path, destination_paths):
        """
        Execute the `rename-prepend-unixtime` dispatch method
        """
        if len(destination_paths) != 1:
            raise ServerException(
                'rename-prepend-unixtime needs a single destination path.')

        destination = destination_paths[0]

        parent, name = self._location.splitPath(destination)

        amended_name = get_file_name_with_unixtime(
            name=name,
            seconds=self._scheduler.seconds(),
            )

        # Re-create the destination path with timestamp name.
        destination_path = self._location.joinPath(parent, amended_name)

        yield self._checkDestination(
            source_path, destination_path,
            'dispatch rename-prepend-unixtime',
            )

        yield self._createDestinationForFile(destination_path)

        yield self._location.rename(
            source_path=source_path, destination_path=destination_path)

        defer.returnValue([destination_path])

    @defer.inlineCallbacks
    def _execute_rename(self, source_path, destination_paths):
        """
        Execute the `rename` dispatch method
        """
        if len(destination_paths) != 1:
            raise ServerException(
                'rename dispatch action needs a single destination path.')

        destination_path = destination_paths[0]

        yield self._checkDestination(
            source_path, destination_path, 'dispatch rename')

        yield self._createDestinationForFile(destination_path)

        yield self._location.rename(
            source_path=source_path, destination_path=destination_path)

        defer.returnValue([destination_path])

    def _cbDispatchFile(self, result, event_id, event, data):
        """
        Called when whole dispatch was successful.

        `result` is the list of destinations where the source was dispatched.
        """
        # It only handles real paths, so we can duplicate it here.
        data['real_path'] = data['path']
        # The event starts with only the destination paths, but if we are
        # here it means that all files were dispatched and we know the exact
        # destination.
        data['destinations'] = ', '.join(result)
        # Also add the list object for API usage and the first element
        # for convenient usage.
        data['destination_paths'] = result
        data['destination_path'] = result[0]

        self.emitEvent(event_id, avatar=event.account, data=data)

    def _ebDispatchFile(self, failure, event_id, event, data):
        """
        Called when any dispatch step fails.
        """
        if failure.check(SkipHandling):
            # Ignore the skip error.
            return

        # It only handled real paths, so we can duplicate it here.
        data['real_path'] = data['path']

        if failure.check(ServerError):
            # Emit the original error, if we have an error and not just
            # an exception.
            if not failure.value.avatar:
                failure.value.avatar = event.account

            emit_from_error(
                failure.value, emitter=self.emitEvent)

            # Add some details to the event dedicated to the event dispatcher.
            data['details'] = failure.value.data.get(
                'details', 'See previous event.')
        else:
            data['details'] = force_unicode(failure.value)

        # Bubble the failure so that the event handler can auto stop
        # on too many errors.
        raise ServerError(event_id, avatar=event.account, data=data)

    @defer.inlineCallbacks
    def _execute_execute(self, source_path, destination_paths):
        """
        Execute the `execute` dispatch method.

        `destination_paths` is a list of executable and its arguments.
        """
        command = destination_paths
        (exit_code, output, error) = yield defer_spawn(
            command,
            timeout=self._configuration.executable_timeout,
            environment={
                'SOURCE_PATH': source_path,
                'ARGUMENTS': ' '.join(command[1:]),
                },
            )

        if exit_code != 0:
            raise ServerException(
                'Failed to execute "%s". [%s] %s %s' % (
                    ' '.join(command), exit_code, output, error))

        defer.returnValue([command[0]])
