# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
Events are the basic mechanism for signaling that something specific took
place inside the server/client.

The products have an **event definitions** database.
The event definition is the basic configuration for each event.
Multiple event definitions are grouped using EventGroupDefinition.

The **event definition** is the static part of the event, generated by the
product at startup time.

The **event** itself is the dynamic part, generated by the product at runtime.
The products will **emit** an event.

An event is emitted using emit(ID, DATA).
ID is the unique event definition id. It will be looked in the database to
retrieve a message and other configurations.
DATA is a dictionary with key:values that will be applied for this event.

Code should not call emit() directly but rather use a wrapper which will
insert the event creator.

If the event database is not available (for example at product initialization)
an event can be emitted using emit(ID, MESSAGE). In this case, the event
will be emitted using default configuration.

Various handlers can be attached to the emitted events to trigger specific
actions.
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from collections import deque
import sys
import traceback

from twisted.conch.ssh.filetransfer import SFTPError
from twisted.internet import defer
from twisted.protocols.ftp import FTPCmdError
from twisted.python.failure import Failure
from zope.interface import implementer
import six

from chevah.server import force_unicode, get_path_handling_module
from chevah.server.commons.constant import STATE
from chevah.server.commons.exception import (
    EmittedError,
    ServerError,
    ServerException,
    )
from chevah.server.commons.interface import (
    IEventCreator,
    IEventHandler,
    IEventHandlersManager,
    )
from chevah.server.commons.observer import ObserverMixin, Signal
from chevah.server.commons.match import get_matched_groups
from chevah.server.commons.runnable import (
    Runnable,
    RunnablesManagerWithConfigurationType,
    )
from chevah.server.event.model import (
    Event,
    EventFailedInterpolation,
    EventNotFound,
    StringFormatter,
    )
from chevah.server.identity.avatar import PeerAvatar, ProcessAvatar


@implementer(IEventCreator)
class ProcessComponent(object):
    """
    Place holder class for when an event is emitted without a creator.
    """
    # FIXME:4464:
    # This class is only necessary because we still emit some events
    # without a creator.
    def __init__(self):
        self.name = 'Process'
        self.uuid = 'Process'
        self.type = 'process-component'

    def emitEvent(self, event_id, data, avatar=None):
        """
        This should never be called by an instance of this class as this
        is only used as a place holder for events without a creator.
        """
        raise RuntimeError(
            'ProcessComponent instances should not call this method.')


class EventDispatcher(ObserverMixin):
    """
    It creates an event and then decide what to do with it.

    Signals:

    * emit(event) - Called when an event is dispatched.
    """

    def __init__(self):
        self._definitions = None
        self._configuration = None
        self._initializePendingEventsQueue()

    def _initializePendingEventsQueue(self):
        """
        Create a new queue.
        """
        # Event emitted before we have any event handler subscribed.
        self._pending_events = deque(maxlen=200)

    def configure(self, definitions, configuration):
        """
        Configure the events handler.
        """
        self._definitions = definitions
        self._configuration = configuration

    def removeConfiguration(self):
        """
        Remove all configuration for the handler.
        """
        self._definitions = None
        self._configuration = None

    @property
    def configured(self):
        """
        Return true if handler was configured.
        """
        if self._definitions and self._configuration:
            return True
        else:
            return False

    @property
    def definitions(self):
        """
        Return the EventDefinition and EventGroupDefinition database.
        """
        if not self._definitions:
            raise RuntimeError('Event definitions are not yet defined.')
        return self._definitions

    def emit(self, *args, **kwargs):
        """
        Creates an event, resolves its message and then emits it.

        This should be the only method to trigger events.
        """
        return self.__emit(*args, **kwargs)

    # FIXME:4464:
    # Make creator compulsory.
    def __emit(
            self, event_id, creator=None, avatar=None, peer=None, data=None):
        """
        Low level emit which allows overwriting the message.
        """
        if (peer is not None and avatar is not None):
            raise RuntimeError('You should not pass both peer and avatar.')

        if avatar is None:
            if peer is None:
                avatar = ProcessAvatar()
            else:
                avatar = PeerAvatar(peer)

        if not creator and avatar.component:
            # Fall back to the component associated with the avatar.
            # It is possible to pass the avatar between components
            # and or those events we should have the same avatar, but
            # different components.
            creator = avatar.component

        if creator is None:
            # FIXME:4464:
            # Some events are still emitted without a creator.
            # We need to update them and fail if we don't have a creator.
            creator = ProcessComponent()

        if data is None:
            data = {}
        data = data.copy()

        # FIXME:2635:
        # Data should not contain the `avatar` as now we have a dedicated
        # `account` member
        # We sync data['avatar'] with our avatar.
        if 'avatar' in data and data['avatar']:
            avatar = data.get('avatar')
        data['avatar'] = avatar

        self._updateFilePath(data)

        if self._configuration:
            server = self._configuration.root.public_identity
        else:
            # This is most probably for the error events emitted before
            # we were able to read the configuration.
            from chevah.server.commons.process import ProcessPublicIdentity
            server = ProcessPublicIdentity(self)

        if self.configured:
            try:
                message = self._resolveEventMessage(event_id, data)
            except ServerError:
                # FIXME:5545:
                # Decide how to handle unknown events.
                # Maybe use a custom exception.
                # For now, we just raise a new event and stop.
                return self._emitEvent(EventNotFound(
                    event_id=event_id,
                    data=data,
                    creator=creator,
                    account=avatar,
                    server=server,
                    traceback=traceback.format_stack(),
                    message_resolver=self._resolveEventMessage,
                    ))
            except (TypeError, ValueError, KeyError) as error:
                # FIXME:5545:
                # Decide how to handle interpolation errors.
                # Raise a new event informing the error.
                return self._emitEvent(EventFailedInterpolation(
                    event_id=event_id,
                    bad_data=data,
                    creator=creator,
                    account=avatar,
                    server=server,
                    error=error,
                    message_resolver=self._resolveEventMessage
                    ))
        else:
            message = data.get('message', '%r')
            message = message % data
            message = force_unicode(message)

        event = Event(
            id=event_id,
            account=avatar,
            creator=creator,
            message=message,
            data=data,
            server=server,
            )
        return self._emitEvent(event)

    def emitFromError(
            self, error, avatar=None, peer=None, creator=None, emitter=None):
        """
        Creates an event using the error id and data attributes, then
        emits it using `emitter`.

        Return the emitted error.
        """
        if isinstance(error, EmittedError):
            # Don't re-emit the error.
            return error

        if not isinstance(error, ServerError):
            message = (
                'Invalid error type: only CompatError/ServerError '
                'are supported. %s %s' % (error, traceback.format_exc())
                )
            raise RuntimeError(message)

        if avatar is None:
            if peer:
                avatar = PeerAvatar(peer)
            else:
                avatar = error.avatar

        if emitter is None:
            # FIXME:3231:
            # Emitter should be mandatory and we could turn this code into
            # a function.
            self.__emit(
                error.id,
                avatar=avatar,
                creator=creator,
                data=error.data,
                )
        else:
            if creator is not None:
                message = 'This should not be called with an creator. %s' % (
                    traceback.format_exc(),)
                raise RuntimeError(message)

            emitter(
                error.id,
                avatar=avatar,
                data=error.data,
                )
        return EmittedError(error)

    def _emitEvent(self, event):
        """
        Low level function to trigger the event.

        Returns a deferred that fires when actions were finalized.
        """
        if not self.hasSubscribers('emit'):
            # We don't have any event handlers attached.
            self._pending_events.append(event)
            return defer.succeed(None)

        if self._pending_events:
            # Reset the current queue and re-send all the
            # queued events.
            # FIXME:3570:
            # Instead of sending only to the first event handler.
            # we should send the pending events to all the event handlers
            # that are expected to be started at startup.
            events = self._pending_events
            self._initializePendingEventsQueue()
            for past_event in events:
                self.notify('emit', Signal(source=self, event=past_event))

        result = self.notify('emit', Signal(source=self, event=event))
        return defer.succeed(result)

    def _resolveEventMessage(self, event_id, data):
        """
        Return the text message for an event.
        """
        event_definition = self.definitions.getEventDefinition(event_id)

        if event_definition:
            message = event_definition.message
        elif 'message' in data:
            message = data['message']
        else:
            data = data.get('details', '')
            message = event_id + ' ' + data
        text = message % data

        return text

    def addStandardOutputHandler(self):
        """
        Add default handler for standard output.

        This is used only for debugging and shell code and it will not be
        unsubscribed.
        """
        self.subscribe(
            'emit', lambda signal: _handle_to_stream(signal, sys.stdout))

    def addStandardErrorHandler(self):
        """
        Add handler for standard error output.

        This is used only for debugging and shell code and it will not be
        unsubscribed.
        """
        self.subscribe(
            'emit', lambda signal: _handle_to_stream(signal, sys.stderr))

    def _updateFilePath(self, data):
        """
        Update directory and file name for path/real_path data member.
        """
        def split(path):
            module = get_path_handling_module(path)
            dir_name, file_name = module.split(path)
            file_base, file_extension = module.splitext(file_name)
            if not file_extension:
                # On Python 2 the empty extension is bytes.
                file_extension = ''
            return dir_name, file_name, file_base, file_extension

        def clean_path(path):
            r"""
            Internally we might have Windows unicode paths,
            but for end users, we want to expose the normal Windows path.

            Raw path:   \\?\C:\some\path
            Clean path: C:\some\path
            """
            if path.startswith('\\\\?\\'):
                path = path[4:]

            return path

        real_path = data.get('real_path', None)
        if real_path:
            data['real_path_raw'] = real_path
            real_path = clean_path(real_path)
            data['real_path'] = real_path
            dir_name, file_name, file_base, file_extension = split(real_path)
            data['real_file_name'] = file_name
            data['real_directory_name'] = dir_name
            data['real_file_base'] = file_base
            data['real_file_extension'] = file_extension

        path = data.get('path', None)
        if path:
            data['path_raw'] = path
            path = clean_path(path)
            data['path'] = path
            dir_name, file_name, file_base, file_extension = split(path)
            data['file_name'] = file_name
            data['directory_name'] = dir_name
            data['file_base'] = file_base
            data['file_extension'] = file_extension

        source_path = data.get('source_path', None)
        if source_path:
            data['source_path_raw'] = source_path
            source_path = clean_path(source_path)
            data['source_path'] = source_path
            dir_name, file_name, file_base, file_extension = split(source_path)
            data['source_file_name'] = file_name
            data['source_directory_name'] = dir_name
            data['source_file_base'] = file_base
            data['source_file_extension'] = file_extension

        destination_path = data.get('destination_path', None)
        if destination_path:
            data['destination_path_raw'] = destination_path
            destination_path = clean_path(destination_path)
            data['destination_path'] = destination_path
            dir_name, file_name, file_base, file_extension = split(
                destination_path)
            data['destination_file_name'] = file_name
            data['destination_directory_name'] = dir_name
            data['destination_file_base'] = file_base
            data['destination_file_extension'] = file_extension


class EventHandlerMixin(object):
    """
    Provides `IEventHandler.handle` implementation with own events failure
    filtering.

    It breaks `handle` method into the `_isFiltered` and `_onEvent` abstract
    methods.

    It must be combined with `Runnable` or it's specializations.
    """

    def _isOwnEvent(self, event):
        """
        Return `True` if event was emitted by the handler itself, False
        otherwise.
        """
        return event.creator is self

    def handle(self, event):
        """
        Called when a new event was emitted.
        """
        def process(event):
            """
            Called as a maybeDeferred in order to have general and uniform
            error handling.
            """
            if self._isOwnEvent(event):
                # We ignore event generated by the same instance to prevent
                # circular loop.
                return

            if self._isFiltered(event):
                # We ignore filtered events.
                return

            deferred = defer.maybeDeferred(self._onEvent, event)
            deferred.addCallback(self._cbActionSucceeded)
            return deferred

        def eb_handle_error(failure):
            """
            Called when failing to handle the event.

            Can be called even after the handler was stopped due to async
            processing of the deferred(s).
            """
            failed_or_stopped = [
                STATE.FAILING, STATE.FAILED_AT_RUN,
                STATE.STOPPING, STATE.STOPPED,
                ]
            if self.state in failed_or_stopped:
                # Do not emit errors if handler is not running any more.
                # This is needed as some events are processed via deferreds
                # and the failure will occur later, when:
                #   a) handler failed due to maximum failures limit,
                #   b) product is shutting down and handler is stopped
                return None

            target_path = event.data.get('path', '')
            if not target_path:
                target_path = event.data.get(
                    'real_path', 'no-associated-path')

            data = {
                'id': event.id,
                'name': self.name,
                'family': self.family,
                'target_path': target_path,
                'details': 'See previous event.',
                'tb': '',
                }
            if failure.check(ServerError):
                # Emit now the error.
                # Prevent circular imports.
                from chevah.server import emit_from_error
                emit_from_error(failure.value, emitter=self.emitEvent)
            elif failure.check(ServerException):
                data['details'] = failure.value.message
            else:
                # FIXME:3708:
                # `EmailSenderEventHandler might throw an
                # exception (such as `IOError`, `OSError`) with utf-8 file
                # name and the name is not correctly displayed.
                # Add details for unknown event.
                data['details'] = force_unicode(failure.value)
                data['tb'] = force_unicode(failure.getBriefTraceback())

            # Generic event for failing to handle an event.
            self.emitEvent(
                '20174',
                avatar=event.account,
                data=data,
                )
            # This error is ignored since the handler will auto-fail with the
            # last error when maximum number of failures is reached.
            return None

        deferred = defer.maybeDeferred(process, event)
        deferred.addErrback(self._ebActionFailed)
        deferred.addErrback(eb_handle_error)
        return deferred


@implementer(IEventHandler)
class EventHandlerWithFilter(Runnable, EventHandlerMixin):
    """
    Event handler with configuration options for filtering events.
    """
    def __init__(self, *args, **kwargs):
        super(EventHandlerWithFilter, self).__init__(*args, **kwargs)
        self._definitions = self.root.event_dispatcher.definitions

    def _isFiltered(self, event):
        """
        See: `EventHandlerMixin.handle`.

        Return `True` if the event should not be handled.
        """
        result = self._isFilteredSet(
            configuration=self.configuration.target,
            actual=event.id,
            )
        if result:
            return True

        result = self._isFilteredSet(
            configuration=self.configuration.usernames,
            actual=event.account.name,
            )
        if result:
            return True

        result = self._isFilteredSet(
            configuration=self.configuration.components,
            actual=event.creator.uuid,
            )
        if result:
            return True

        result = self._isFilteredSet(
            configuration=self.configuration.source_addresses,
            actual=event.account.peer.host,
            )
        if result:
            return True

        result = self._isFilteredByData(event)
        if result:
            return True

        result = self._isFilteredByGroups(event)
        if result:
            return True

        # Not filtered by any rule.
        return False

    def _isFilteredSet(self, configuration, actual):
        """
        Return true if `actual` value is filtered by the `configuration`
        which is a list of filtered values.
        """
        if not configuration:
            return False

        # FIXME:3663:
        # We should prevent running the handler if the target configuration
        # is being used to both allow and ignore events.
        allowed = []
        ignored = []
        for target in configuration:
            if target.startswith('!'):
                ignored.append(target[1:].strip())
            else:
                allowed.append(target)

        if actual in ignored:
            # This should be ignored so we filter it.
            return True

        if allowed and actual not in allowed:
            # We only filter if there is a configuration.
            return True

        return False

    def _isFilteredByData(self, event):
        """
        Return `True` if event should be ignored based on data attribute
        configuration.
        """
        data_filter = self.configuration.data_filter
        if not data_filter:
            return False

        for rule in data_filter:

            try:
                name = rule[0]
                expression = rule[1]
            except IndexError:
                raise ServerException(
                    'Invalid data_filter format. '
                    'Needs a name and an expression.'
                    )

            data_member = name.lower()
            try:
                # We convert to text so that we can do the match.
                # This should always work, as it will fall back to repr()
                # and any object should have that.
                # If it fails, we should see a generic error or an ISE.
                value = force_unicode(event.data[data_member])
            except KeyError:
                # We don't want to ignore a missing data, because otherwise
                # the events can get alwasy ignored and customers will not
                # know why.
                raise ServerException(
                    'Missing filtered data "{}" from event.'.format(
                        data_member))

            by_data = get_matched_groups(
                value=value, rule=expression, usage='data_filter')

            if not by_data:
                # No need to do further checks.
                return True

        # All rules are a match.
        # Accept this filter.
        return False

    def _isFilteredByGroups(self, event):
        """
        Return `True` if event should be ignored based on group configuration.

        Event is ignored if none of its groups is allowed.
        Event is ignored if any of its groups is ignored.
        """
        groups = self.configuration.groups
        if not groups:
            return False

        definition = self._definitions.getEventDefinition(event.id)
        event_groups = [g.name.lower() for g in definition.groups]

        # FIXME:3663:
        # We should prevent running the handler if the target configuration
        # is being used to both allow and ignore events.
        allowed = []
        ignored = []
        for target in groups:
            if target.startswith('!'):
                ignored.append(target[1:])
            else:
                allowed.append(target)

        for group in event_groups:
            if group in ignored:
                return True

        if not allowed:
            return False

        for group in event_groups:
            if group in allowed:
                return False

        return True


@implementer(IEventHandlersManager)
class EventHandlersManager(RunnablesManagerWithConfigurationType):
    """
    Management of available event handlers.

    It also forwards emitted events to started handlers.
    """
    __type_interface__ = IEventHandler

    def _getRunnablesConfiguration(self):
        """
        See: `RunnablesManagerWithConfigurationSet`.
        """
        return self.parent.configuration.event_handlers

    def start(self):
        """
        See: Runnable.
        """
        self.root.event_dispatcher.subscribe('emit', self.__handle)

        return super(EventHandlersManager, self).start()

    def stop(self, avatar=None):
        """
        See: Runnable.
        """
        self.root.event_dispatcher.unsubscribe('emit', self.__handle)
        return super(EventHandlersManager, self).stop()

    def __handle(self, signal):
        """
        Called when a new signal is notified.
        """
        event = signal.event
        for handler in self.getStarted():
            handler.handle(event)


def get_exception_details(exception):
    """
    Return the string representation of error.
    """
    if isinstance(exception, ServerError):
        details = exception.data.get('details', 'See previous event.')
        return '%s: %s' % (exception.id, details)

    if isinstance(exception, SFTPError):
        result = '%s %s' % (
            exception.code, force_unicode(exception.message))
        return result

    if isinstance(exception, FTPCmdError):
        result = exception.response()
        return force_unicode(result)

    # See if we have an exception, wrapping a failure.
    if exception.args:
        possible_failure = exception.args[0]
        if isinstance(possible_failure, Failure):
            exception = possible_failure.value

    return force_unicode(exception)


def emit_error_details(error, event_id, data, avatar=None, emitter=None):
    """
    Emit 'event_id' with `data` augmented with `error` details using `emitter`.

    When `error` is a `ServerError` it is emitted in the addition of the
    `event_id`.
    """
    # FIXME:5545:
    # This is poor man's exception handling since we don't have a
    # standard format for errors.
    # IOError can be both UTF-8 encoded or Unicode.
    if emitter is None:
        # Prevent circular import.
        from chevah.server import emit
        emitter = emit

    if isinstance(error, ServerError):
        # Prevent circular import.
        from chevah.server import emit_from_error
        emit_from_error(error, avatar=avatar, emitter=emitter)

    data['details'] = get_exception_details(error)
    if avatar:
        emitter(event_id, data=data, avatar=avatar)
    else:
        emitter(event_id, data=data)


def _handle_to_stream(signal, stream):
    """
    Called when a new event is triggered as a signal.
    """
    from chevah.server.event.stdlib_logging import (
        EventLogRecord,
        EventLogRecordFormatterAdapter,
        )
    formatter = EventLogRecordFormatterAdapter(StringFormatter())
    output = formatter.format(EventLogRecord(signal.event))
    if six.PY2:
        output = output.encode('utf-8')
    stream.write(output)
    stream.flush()
