# Copyright (c) 2014 Adi Roiban.
# See LICENSE for details.
"""
Event handler triggering HTTP POST hooks.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from six.moves.urllib.parse import urlencode
from defusedxml import minidom

from dict2xml import dict2xml
from twisted.internet import defer
from twisted.web import http
from zope.interface import implementer

from chevah.server import force_unicode, json
from chevah.server.commons.constant import (
    ADDRESS_UNKNOWN,
    HTTP_EVENT_HANDLER_FORMAT_NAME,
    HTTP_METHOD_NAMES,
    TYPE_NAME,
    )
from chevah.server.commons.credential import PasswordCredentials
from chevah.server.commons.exception import ServerException
from chevah.server.commons.interface import (
    IHTTPEventHandlerConfiguration,
    )
from chevah.server.commons.match import get_matched_groups
from chevah.server.commons.ssl import (
    SSLClientMixin,
    SSLConfigurationSectionMixin,
    )
from chevah.server.configuration.model import (
    WritableLines,
    WritableNumber,
    WritableString,
    WritableStringCaseInsensitive,
    WritableCSVUnique,
    )
from chevah.server.configuration.option import (
    UsernamePasswordPublicOptionsMixin,
    RetryIncreaseOptionsMixin,
    )
from chevah.server.event.configuration import (
    _EventHandlerConfigurationBase,
    FilterOptionsMixin,
    )
from chevah.server.event.handler import EventHandlerWithFilter
from chevah.server.event.model import StringFormatter
from chevah.server.event.stdlib_logging import EventLogRecordFormatterAdapter
from chevah.server.http.client import (
    extract_configured_header,
    RedundantAgent,
    )
import six

_SOAP_TEMPLATE = """<?xml version="1.0"?>

<soap:Envelope
xmlns:soap="http://www.w3.org/2003/05/soap-envelope/"
soap:encodingStyle="http://www.w3.org/2003/05/soap-encoding">

<soap:Body xmlns:m="http://www.sftpplus.com/event">
  <m:Events>
    <m:Event>
%s
    </m:Event>
  </m:Events>
</soap:Body>

</soap:Envelope>
"""


@implementer(IHTTPEventHandlerConfiguration)
class HTTPEventHandlerConfiguration(
    _EventHandlerConfigurationBase,
    FilterOptionsMixin,
    RetryIncreaseOptionsMixin,
    SSLConfigurationSectionMixin,
    UsernamePasswordPublicOptionsMixin,
        ):
    """
    Configuration for a HTTP EventHandler.
    """

    url = WritableCSVUnique('url')
    request_method = WritableStringCaseInsensitive(
        'request_method', default='post')
    http_content_type = WritableString('http_content_type')
    expected_response = WritableString('expected_response')
    extra_data = WritableString('extra_data')
    headers = WritableLines('headers')
    timeout = WritableNumber(
        'timeout', default=120, allow_negative=False, allow_zero=False)


class HTTPEventHandler(EventHandlerWithFilter, SSLClientMixin):
    """
    Sends events over HTTP as JSON POST requests.

    For each supported format you will need to define a `_format_*` method and
    update the HTTP_EVENT_HANDLER_FORMAT_NAME constant.
    """
    PROPERTIES_REQUIRING_RESTART = (
        'expected_response',
        'extra_data',
        'headers',
        'http_content_type',
        'password_update',
        'url',
        'username',
        'timeout',
        ) + SSLConfigurationSectionMixin.PROPERTIES_REQUIRING_RESTART

    # This is here to help testing.
    _Agent = RedundantAgent

    def __init__(self, parent, configuration):
        super(HTTPEventHandler, self).__init__(parent, configuration)
        self._log_formatter = EventLogRecordFormatterAdapter(StringFormatter())
        self._reset()

    def _reset(self):
        """
        Reset the internal state.
        """
        self._agent = None
        self._format_method = None
        self._extra_data = ''
        self._urls = None
        self._expected_response = ''
        self._headers = {}

    def _onStart(self, avatar=None):
        """
        See: `Runnable`.
        """
        self._format_method = self._getFormatMethod()
        format_type = self._configuration.http_content_type.lower()
        self._extra_data = self._configuration.extra_data

        request_method = self._configuration.request_method
        if request_method not in list(HTTP_METHOD_NAMES.keys()):
            raise ServerException('Unknown "request_method" type: {}'.format(
                request_method
                ))

        if format_type == TYPE_NAME.JSON:
            if self._extra_data:
                try:
                    self._extra_data = json.loads(self._extra_data)
                except ValueError:
                    raise ServerException(
                        'Invalid JSON format in "extra_data".')
            else:
                self._extra_data = {}

            if not isinstance(self._extra_data, dict):
                raise ServerException(
                    '"extra_data" needs to be a JSON object.')

        for line in self._configuration.headers:
            key, value = extract_configured_header(line)
            self._headers[key] = [value]

        # Validated the value.
        self._expected_response = self._configuration.expected_response
        get_matched_groups(
            '', self._expected_response, 'expected_response')

        self._urls = self._configuration.url
        deferred = self._initStartedContextFactory(urls=self._urls)

        def cb_got_context_factory(result):
            """
            Called once the client SSL context was set up.
            """
            credentials = PasswordCredentials(
                username=self._configuration.username,
                password=self._configuration.password,
                peer=ADDRESS_UNKNOWN,
                creator=self,
                )

            self._agent = self._Agent(
                urls=self._urls,
                event_emitter=self.emitEvent,
                ssl_context=result.getContext(),
                credentials=credentials,
                timeout=self._configuration.timeout,
                )

        deferred.addCallback(cb_got_context_factory)
        return deferred

    def _getFormatMethod(self):
        """
        Return the method used to format the events.
        """
        format_method_name = self.configuration.http_content_type.lower()
        if format_method_name not in HTTP_EVENT_HANDLER_FORMAT_NAME:
            raise ServerException(
                'Unknown format: %s' % (format_method_name,))
        format_method_name = '_format_%s' % (
            format_method_name.replace('-', '_'),)
        format_method = getattr(self, format_method_name, None)
        if not format_method:
            raise ServerException(
                'Format not implemented: %s' % format_method_name)
        return format_method

    def _onStop(self, avatar=None):
        """
        See: `Runnable`.
        """
        agent = self._agent
        self._reset()

        if not agent:
            return

        if self._context_factory:
            self._context_factory.stop()

        return agent.closePersistentConnections()

    @defer.inlineCallbacks
    def _onEvent(self, event):
        """
        See: EventHandlerWithFilterMixin.
        """
        request_method = self._configuration.request_method
        if request_method not in list(HTTP_METHOD_NAMES.keys()):
            raise ServerException('Unknown "request_method" type: {}'.format(
                request_method
                ))

        content_type, body = self._format_method(event)
        headers = {
            'Content-Type': [content_type],
            }
        headers.update(self._headers)

        agent_request_method = self._agent.postWithRedundancy
        if request_method == TYPE_NAME.GET:
            agent_request_method = self._agent.getWithRedundancy

        if request_method == TYPE_NAME.PUT:
            agent_request_method = self._agent.putWithRedundancy

        configuration = self.configuration
        deferred = agent_request_method(
            headers=headers,
            body=body,
            retry_count=configuration.retry_count,
            retry_increase=configuration.retry_increase,
            )
        response = yield deferred
        if response.code not in [http.OK, http.NO_CONTENT]:
            # All unknown response code is considered an error.
            yield self._agent.readBodyWithIgnore(response)
            raise ServerException('Remote server returned %s:%s' % (
                response.code, force_unicode(response.phrase)))

        # All good. Consume the body.
        content = yield self._agent.readBody(response)
        content = force_unicode(content)
        groups = get_matched_groups(
            content, self._expected_response, 'expected_response')
        if not groups:
            raise ServerException(
                'Expected expression "%s" not found in the response '
                'from the remote server. Server returned: %s' % (
                    self._expected_response, content))

        # Preserve data from the original event to help with chained
        # event handlers.
        data = event.data.copy()
        data['event_id'] = event.id
        data['target_path'] = data.get('path', 'no-associated-path')
        self.emitEvent('20189', avatar=event.account, data=data)

    def _format_custom(self, event):
        """
        Format the event as JSON.
        """
        content_type = b'text/plain; charset=utf-8'
        result = self._resolveEvent(event, self._extra_data)
        return content_type, result.encode('utf-8')

    def _format_json(self, event):
        """
        Format the event as JSON.
        """
        content_type = b'application/json; charset=utf-8'
        body = {
            'events': [event],
            }
        extra_data = self._resolveEvent(event, self._extra_data)
        body.update(extra_data)
        return content_type, json.dumps(body).encode('utf-8')

    def _format_soap(self, event):
        """
        Format the event as JSON.
        """
        content_type = b'application/soap+xml; charset=utf-8'
        public_event = event.asDict()
        del public_event['data_json']

        xml_events = _SOAP_TEMPLATE % dict2xml(public_event, wrap='')

        dom = minidom.parseString(xml_events.encode('utf-8'))
        result = '\n'.join([
            line for line in dom.toprettyxml(indent='  ').split('\n')
            if line.strip()
            ])
        return content_type, result.encode('utf-8')

    def _format_xml(self, event):
        """
        Format the event as JSON.
        """
        content_type = b'text/xml; charset=utf-8'
        public_event = event.asDict()

        del public_event['data_json']
        result = dict2xml({'event': public_event}, wrap='events')
        return content_type, result.encode('utf-8')

    def _resolveEvent(self, event, extra_data):
        """
        Walk the `extra_data` and resolve the variables based on the event's
        data.
        """
        result = {}

        if isinstance(extra_data, six.string_types):
            return event.renderFromTemplate(extra_data, 'extra_data')

        if not isinstance(extra_data, dict):
            return extra_data

        for key, value in extra_data.items():
            key = event.renderFromTemplate(key, 'extra_data')
            if isinstance(value, list):
                value = [self._resolveEvent(event, v) for v in value]
            else:
                value = self._resolveEvent(event, value)

            result[key] = value

        return result

    def _format_legacy_webadmin(self, event):
        """
        Format the event as URL encoded for the legacy SFTPPlus WebAdmin.
        """
        content_type = 'application/x-www-form-urlencoded'
        body = {
            'MSGID': event.id.encode('utf-8'),
            'MSGTEXT': event.message.encode('utf-8'),
            'USERNAME': event.account.name.encode('utf-8'),
            'REMOTE_ADDR': self._log_formatter.formatPeer(event.account.peer),
            'SERVER': self.root.configuration.server.name.encode('utf-8'),
            # Default string conversion of float truncates the decimals, so
            # we go for repr().
            'TIMESTAMP': repr(event.timestamp.timestamp),
            'NG': 'TRUE',
            }

        return content_type, urlencode(body)
