# Copyright (c) 2015 Adi Roiban.
# See LICENSE for details.
"""
Event handler sending events to a local file.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from calendar import monthrange
from datetime import datetime
from dateutil import relativedelta
from stat import ST_DEV, ST_INO
import logging
import os
import re
import socket
import time
import traceback

from twisted.internet import defer, reactor, error, task

from chevah_compat import local_filesystem
from chevah.server import force_unicode
from chevah.server.commons.exception import ServerException
from chevah.server.commons.public_format import NoStrBunch, public_format
from chevah.server.event.stdlib_logging import EventHandlerWithLogger
import six
from six.moves import range


class FileHandler(logging.FileHandler):
    """
    A handler class which writes formatted logging records to disk files.
    """

    def __init__(self, filename, mode='ab'):
        # We keep the state to know if we are writing to an existing file or
        # a new file, so we can include a header if it is configured to do so.
        self._is_new_file = False
        self.file_header = b''
        self.base_path = filename
        if isinstance(self.base_path, six.binary_type):
            self.base_path = self.base_path.decode('utf-8')
        logging.FileHandler.__init__(self, filename, mode, None, delay=True)

    def open(self):
        """
        Called to open the stream.
        """
        self.stream = self._open()

    def _open(self):
        """
        See: `logging.FileHandler`
        """
        self._is_new_file = not os.path.exists(self.baseFilename)
        return super(FileHandler, self)._open()

    def handleError(self, record):
        """
        See: `logging.FileHandler`

        Overrides raising the error so it is handled by the high level
        event handler.
        """
        raise

    def emit(self, record):
        """
        Emit a record.

        If a formatter is specified, it is used to format the record.
        The record is then written to the stream. If
        exception information is present, it is formatted using
        traceback.print_exception and appended to the stream.  If the stream
        has an 'encoding' attribute, it is used to determine how to do the
        output to the stream.
        """
        if self.stream is None:
            # An error occured opening the stream, probably when rotating,
            # so we try to open it again.
            self.stream = self._open()

        stream = self.stream
        if self._is_new_file and self.file_header:
            stream.write(self.file_header.encode('utf-8'))
            self._is_new_file = False

        msg = self.format(record)

        stream.write(msg.encode('utf-8'))
        self.flush()

    def entryLength(self, record):
        """
        Determine entry length in bytes.
        """
        msg = self.format(record)
        encoding = self.encoding
        if not encoding:
            encoding = 'utf-8'
        return len(msg.encode(encoding))


class WatchedFileHandler(FileHandler):
    """
    A handler for logging to a file, which watches the file
    to see if it has changed while in use. This can happen because of
    usage of programs such as newsyslog and logrotate which perform
    log file rotation. This handler, intended for use under Unix,
    watches the file to see if it has changed since the last emit.
    (A file has changed if its device or inode have changed.)
    If it has changed, the old file stream is closed, and the file
    opened to get a new stream.

    This handler is not appropriate for use under Windows, because
    under Windows open files cannot be moved or renamed - logging
    opens the files with exclusive locks - and so there is no need
    for such a handler. Furthermore, ST_INO is not supported under
    Windows; stat always returns zero for this value.

    This handler is based on a suggestion and patch by Chad J.
    Schroeder.
    """

    def open(self):
        """
        Called when the stream should be opened.
        """
        super(WatchedFileHandler, self).open()
        if not os.path.exists(self.baseFilename):
            self.dev, self.ino = -1, -1
        else:
            stat = os.stat(self.baseFilename)
            self.dev, self.ino = stat[ST_DEV], stat[ST_INO]

    def emit(self, record):
        """
        Emit a record.

        First check if the underlying file has changed, and if it
        has, close the old stream and reopen the file to get the
        current stream.
        """
        if not os.path.exists(self.baseFilename):
            stat = None
            changed = 1
        else:
            stat = os.stat(self.baseFilename)
            changed = (stat[ST_DEV] != self.dev) or (stat[ST_INO] != self.ino)
        if changed and self.stream is not None:
            self.stream.flush()
            self.stream.close()
            self.stream = self._open()
            if stat is None:
                stat = os.stat(self.baseFilename)
            self.dev, self.ino = stat[ST_DEV], stat[ST_INO]
        FileHandler.emit(self, record)


class RotatingHandlerAbstract(FileHandler):
    """
    Base class for handlers that rotate log files at a certain
    point.
    """
    def _shouldRollover(self, record):
        """
        Determine if rollover should occur.
        """
        raise NotImplementedError()

    def _doRollover(self):
        """
        Do a rollover.
        """
        raise NotImplementedError()

    def __init__(self, filename, mode, event_emitter):
        """
        Use the specified filename for streamed logging
        """
        FileHandler.__init__(self, filename, mode)
        self.mode = mode
        self._emitEvent = event_emitter

    def emit(self, record):
        """
        Emit a record. Output the record to the file.
        """
        if self._shouldRollover(record):
            self._doRollover()
        FileHandler.emit(self, record)

    def _emitRotationEvent(self, previous_path, current_path):
        """
        Emit event about the fact that the file `previous_path` was
        successfully rotated as `current_path`.
        """
        try:
            path = local_filesystem.getEncodedPath(current_path)
            size = os.path.getsize(path)
            size = force_unicode(size)
        except Exception:
            # We can't get the size.
            # But we don't want to fail here.
            size = 'unknown'

        data = {
            'source_path': previous_path,
            'path': current_path,
            'real_path': current_path,
            'size': size,
            }
        self._emitEvent('20175', data=data)

    def _emitRotationEventInPlace(self):
        """
        Emit event about the fact that the log file `path` was
        successfully rotated in place.
        """
        data = {
            'path': self.base_path,
            'real_path': self.base_path,
            }
        self._emitEvent('20176', data=data)

    def _emitRotationEventError(self, details):
        """
        Emit the event about an error happened during log file
        rotation.
        """
        data = {
            'details': details,
            'tb': force_unicode(traceback.format_exc()),
            }
        self._emitEvent('20183', data=data)


class _TimeRotatingFileHandler(RotatingHandlerAbstract):
    """
    Base class for timed rotating file handlers.
    """

    def __init__(
            self, filename, event_emitter, on_rollover_failed,
            rotate_count=0, clock=reactor
            ):
        """
        `filename` contains a path to the file which needs to be
        rotated.

        When the file is rotated appropriate event is emitted using
        `event_emitter`.

        If there is a failure during rotation `on_rollover_failed` is called.

        `rotate_count` equals to `-1` rotates a file in
        place. `rotate_count` equals to `0` keeps all rotated
        files. `rotate_count` greater than `0` keeps only specified
        amount of rotated files.

        `clock` is mainly used for testing purposes. It helps us to
        make sure that all objects are in sync and are using the same
        clock.
        """
        RotatingHandlerAbstract.__init__(
            self, filename=filename, mode='ab', event_emitter=event_emitter)
        self.rotateCount = rotate_count
        self._clock = clock

        # The rolloverDeferred stores the scheduled call to do the rollover
        # and reschedule the next one.
        self._rolloverDeferred = None

        # This is used so we can keep track of failures in the scheduled
        # rollover call.
        self._ebRolloverFailed = on_rollover_failed

    def _scheduleRollover(self):
        """
        Schedules the next rollover.
        """
        self._rolloverDeferred = task.deferLater(
            self._clock,
            max(0, self.rolloverAt - self._clock.seconds()),
            self._doRolloverAndReschedule,
            )

        def eb_cancel(failure):
            """
            Do nothing when rotation is canceled, as it is cancelled on stop.
            """
            failure.trap(defer.CancelledError)

        self._rolloverDeferred.addErrback(eb_cancel)

        # Will call ebRolloverFailed if the rollover fails.
        self._rolloverDeferred.addErrback(self._ebRolloverFailed)

    def open(self):
        """
        Called to open the stream.
        """
        super(_TimeRotatingFileHandler, self).open()
        self._scheduleRollover()

    def close(self):
        """
        Called to close the stream.
        """
        try:
            self._rolloverDeferred.cancel()
        except (AttributeError, error.AlreadyCancelled, error.AlreadyCalled):
            # This was never opened properly, probably failed while starting
            # the handler, or for some reason this was just called but
            # not rescheduled yet. We ignore it and continue with closing
            # the handler.
            pass
        super(_TimeRotatingFileHandler, self).close()

    def emit(self, record):
        """
        Emit a record. Output the record to the file.
        """
        FileHandler.emit(self, record)

    def _getRolledFilename(self):
        """
        See `_RotatingFileHandlerAbstract`.
        """
        time_tuple = datetime.fromtimestamp(self.rolloverAt)
        name, extension = os.path.splitext(self.base_path)
        return name + '.' + time_tuple.strftime(self.suffix) + extension

    def _doRollover(self):
        """
        Do a rollover.
        """
        self._doRolloverBasefile()
        self.rolloverAt = self._computeNextRollover(self._clock.seconds())

    def _doRolloverAndReschedule(self):
        """
        Do the rollover and schedule the next one.
        """
        self._doRollover()
        self._scheduleRollover()

    def _doRolloverBasefile(self):
        """
        Do rollover for the base file `self.baseFilename`.

        Rename it and keep only certain amount (defined by
        `rotateCount`) of rotated files.

        Event is emitted after the rollover.
        """
        if self.stream:
            self.stream.close()
            self.stream = None

        if self.rotateCount >= 0:
            try:
                renamed_file = self._renameBaseFile()
            except Exception as error:
                details = 'Cannot rename "{}": {}'.format(
                    self.base_path, force_unicode(error))
                self._emitRotationEventError(details=details)
            else:
                self._emitRotationEvent(
                    previous_path=self.base_path, current_path=renamed_file)
                if self.rotateCount > 0:
                    self._removeRotatedFiles()
            self.mode = 'ab'
            self.stream = self._open()
        else:
            # The log file is being rotated in place (i.e. the current
            # log file is being reset and no rotated file is being
            # created).
            self.mode = 'wb'
            self.stream = self._open()
            self._emitRotationEventInPlace()

    def _renameBaseFile(self):
        """
        Rename the base file and return its new name.
        """
        renamed_base_file = self._getRolledFilename()
        rename_base_path = local_filesystem.getEncodedPath(renamed_base_file)
        if os.path.exists(rename_base_path):
            os.remove(rename_base_path)
        os.rename(self.baseFilename, rename_base_path)

        return renamed_base_file

    def _removeRotatedFiles(self):
        """
        Remove the oldest rotated log files.
        """
        for f in self._getFilesToDelete():
            try:
                local_filesystem.deleteFile(f)
            except Exception as error:
                path = local_filesystem.getRealPathFromSegments(f)
                details = 'Cannot remove "{}": {}'.format(
                    path, force_unicode(error))
                self._emitRotationEventError(details=details)

    def _shouldRollover(self, record):
        """
        Overrided to not do anything since the rollover is clock based.
        """

    def _getFilesToDelete(self):
        """
        Get a list with the files to be deleted after the rotation.
        """
        dir_name, file_name = os.path.split(self.base_path)
        dir_segments = local_filesystem.getSegmentsFromRealPath(dir_name)
        result = []

        base_name, base_extension = os.path.splitext(file_name)

        for target in local_filesystem.getFolderContent(dir_segments):
            if not target.startswith(base_name):
                # Does not look like a log file.
                continue

            if base_extension:
                target_name, target_extension = os.path.splitext(target)
            else:
                target_name = target
                target_extension = ''

            rotated_name_part = target_name[len(base_name) + 1:]

            if not rotated_name_part:
                # Same name, but different extension.
                continue

            if not self.extMatch.match(rotated_name_part):
                # Looks like a rotated file, but is not following our rules.
                # Maybe an older file based on a different method.
                continue

            if base_extension != target_extension:
                # Looks like the rotated file, has the rotated patter but
                # is with different extension.
                continue

            result.append(dir_segments + [target])

        result.sort()
        if len(result) < self.rotateCount:
            result = []
        else:
            result = result[:len(result) - self.rotateCount]
        return result


class DailyRotatingFileHandler(_TimeRotatingFileHandler):
    """
    Daily rotating file handler that provides facilities to rotate a
    file on a certain time of the day.
    """
    def __init__(
            self,
            filename,
            event_emitter,
            on_rollover_failed,
            day_offset=0,
            rotate_count=0,
            clock=reactor,
            ):
        """
        The `filename` is rotated every `day_offset` (given in seconds
        from the start of the day, from 0 to 86400) of each day.

        See `_TimeRotatingFileHandler` for more details.
        """
        super(DailyRotatingFileHandler, self).__init__(
            filename=filename,
            event_emitter=event_emitter,
            on_rollover_failed=on_rollover_failed,
            rotate_count=rotate_count,
            clock=clock,
            )
        self.day_offset = day_offset
        # Can't set time less than 00:00 or more than 23:59.
        if self.day_offset < 0 or self.day_offset >= 86400:
            msg = 'Invalid time specified for daily rollover: {}'
            raise ValueError(msg.format(self.day_offset))
        self.rotateCount = rotate_count
        self.rolloverAt = self._computeNextRollover(self._clock.seconds())
        self.suffix = '%Y-%m-%dT%H%M%S.%f'
        self.extMatch = re.compile(
            r'^\d{4}-\d{2}-\d{2}T\d{2}\d{2}\d{2}\.\d+$')

    def _computeNextRollover(self, timestamp):
        """
        Compute next rollover date on which a file would be rotated.
        """
        current_date = datetime.fromtimestamp(timestamp)

        minute, _ = divmod(self.day_offset, 60)
        hour, minute = divmod(minute, 60)
        rollover_at = current_date.replace(second=0, microsecond=0)

        while True:
            if rollover_at.hour != hour:
                # Advance to the next hour and zero minutes.
                rollover_at += relativedelta.relativedelta(
                    hours=1, minutes=-rollover_at.minute)
            else:
                rollover_at += relativedelta.relativedelta(minutes=1)
            if rollover_at.hour == hour and rollover_at.minute == minute:
                break
        return time.mktime(rollover_at.timetuple())


class MonthlyRotatingFileHandler(_TimeRotatingFileHandler):
    """
    Monthly rotating file handler that provides facilities to rotate a
    file on a certain day of a month.
    """
    def __init__(
            self,
            filename,
            event_emitter,
            on_rollover_failed,
            day=1,
            rotate_count=0,
            clock=reactor,
            ):
        """
        Create an instance of `MonthlyRotatingFileHandler`.

        The `filename` is rotated every `day` of each month.
        """
        super(MonthlyRotatingFileHandler, self).__init__(
            filename=filename,
            event_emitter=event_emitter,
            on_rollover_failed=on_rollover_failed,
            rotate_count=rotate_count,
            clock=clock,
            )
        self.day = day
        if self.day < 1 or self.day > 31:
            msg = (
                'Invalid day of the month specified for '
                'monthly rollover: {}'
                )
            raise ValueError(msg.format(self.day))
        self.rotateCount = rotate_count
        self.rolloverAt = self._computeNextRollover(self._clock.seconds())
        self.suffix = '%Y-%m-%d'
        self.extMatch = re.compile(r'^\d{4}-\d{2}-\d{2}$')

    def _computeNextRollover(self, timestamp):
        """
        Compute next rollover date on which a file would be rotated.
        """
        current_date = datetime.fromtimestamp(timestamp).date()

        if self.day > 28:
            # February 29 is not valid date for a non-leap
            # year. April 31 is not valid date as well. So before
            # usage of `monthDay` make sure that it is valid
            # regarding the month when rollover would occur
            # (either in the current or in the next month).
            rollover_at = current_date
            real_day_of_month = self._getRealDayOfMonth(
                self.day, rollover_at.year, rollover_at.month)
            if rollover_at.day >= real_day_of_month:
                # Next rollover would occur in the next month.
                rollover_at += relativedelta.relativedelta(months=1)
            real_day_of_month = self._getRealDayOfMonth(
                self.day, rollover_at.year, rollover_at.month)
        else:
            real_day_of_month = self.day

        rollover_at = current_date
        if rollover_at.day != real_day_of_month:
            while rollover_at.day != real_day_of_month:
                rollover_at += relativedelta.relativedelta(days=1)
        else:
            rollover_at += relativedelta.relativedelta(months=1)

        return time.mktime(rollover_at.timetuple())

    def _getRealDayOfMonth(self, day_of_month, year, month):
        """
        Return the day of the month adjusting it to the last
        day of the month in a case `day_of_month` is bigger
        than the actual days of given `month`.
        """
        days_in_month = monthrange(year, month)[1]
        if day_of_month > days_in_month:
            day_of_month = days_in_month
        return day_of_month


class SizeRotatingFileHandler(RotatingHandlerAbstract):
    """
    Handler for logging to a set of files, which switches from one file
    to the next when the current file reaches a certain size.
    """

    def __init__(
            self,
            filename,
            event_emitter,
            mode='ab',
            max_bytes=0,
            rotate_count=0,
            ):
        """
        Open the specified file and use it as the stream for logging.

        By default, the file grows indefinitely. You can specify
        particular values of `max_bytes` and `rotate_count` to allow
        the file to rollover at a predetermined size.

        Rollover occurs whenever the current log file is nearly
        `max_bytes` in length.

        If `rotate_count` is >= 0, the system will successively create
        new files with the same pathname as the base file, but with
        extensions ".1", ".2" etc. appended to it. For example, with a
        `rotate_count` of 5 and a base file name of "app.log", you
        would get "app.log", "app.log.1", "app.log.2", ... through to
        "app.log.5". The file being written to is always "app.log" -
        when it gets filled up, it is closed and renamed to
        "app.log.1", and if files "app.log.1", "app.log.2" etc. exist,
        then they are renamed to "app.log.2", "app.log.3"
        etc. respectively. When rollover is done, and `rotate_count` >
        0, no more than `rotate_count` - files are kept - the oldest
        ones are deleted. In case of `rotate_count` equals to zero,
        all rotated files are kept.

        If `rotate_count` equals to `-1` rollover in place is
        performed. The current log file is reset and no rotated files
        are created.

        If `max_bytes` is zero, rollover never occurs.
        """
        # If rotation/rollover is wanted, it doesn't make sense to use another
        # mode. If for example 'w' were specified, then if there were multiple
        # runs of the calling application, the logs from previous runs would
        # be lost if the 'w' is respected, because the log file would be
        # reset on each run.
        if max_bytes > 0:
            mode = 'ab'
        RotatingHandlerAbstract.__init__(self, filename, mode, event_emitter)
        self.maxBytes = max_bytes
        self.rotateCount = rotate_count

    def _doRollover(self):
        """
        Do a rollover, as described in __init__().
        """
        if self.stream:
            self.stream.close()
            self.stream = None
        destination_name = None
        if self.rotateCount >= 0:
            if self.rotateCount == 0:
                # In case of all rotated files are kept, find the
                # number of the oldest rotated file.
                oldest_rotated_file_num = 1
                while True:
                    rotated_file = "%s.%d" % (
                        self.base_path, oldest_rotated_file_num, )
                    rotated_path = local_filesystem.getEncodedPath(rotated_file)
                    if not os.path.exists(rotated_path):
                        break
                    oldest_rotated_file_num += 1
            else:
                oldest_rotated_file_num = self.rotateCount
            # Update already rotated files.
            for i in range(oldest_rotated_file_num - 1, 0, -1):
                source_name = "%s.%d" % (self.base_path, i)
                destination_name = "%s.%d" % (self.base_path, i + 1)

                source_path = local_filesystem.getEncodedPath(source_name)
                destination_path = local_filesystem.getEncodedPath(
                    destination_name)

                if os.path.exists(source_path):
                    if os.path.exists(destination_path):
                        os.remove(destination_path)
                    os.rename(source_path, destination_path)

            # Do the rotation of source file.
            destination_name = '%s.1' % (self.base_path,)
            destination_path = local_filesystem.getEncodedPath(
                destination_name)
            if os.path.exists(destination_path):
                os.remove(destination_path)
            os.rename(self.baseFilename, destination_path)

        self.mode = 'wb'
        self.stream = self._open()
        if destination_name:
            self._emitRotationEvent(
                previous_path=self.base_path,
                current_path=destination_name,
                )
        else:
            self._emitRotationEventInPlace()

    def _shouldRollover(self, record):
        """
        Determine if rollover should occur.

        Basically, see if the supplied record would cause the file to
        exceed the size limit we have.
        """
        if self.stream is None:
            # delay was set...
            self.stream = self._open()
        if self.maxBytes > 0:
            # are we rolling over?
            length = FileHandler.entryLength(self, record)
            self.stream.seek(0, 2)
            # Due to non-posix-compliant Windows
            if self.stream.tell() + length >= self.maxBytes:
                return 1
        return 0


class LocalFileEventHandler(EventHandlerWithLogger):
    """
    Store events to a local file.
    """
    PROPERTIES_REQUIRING_RESTART = (
        'path',
        'rotate_count',
        'rotate_at_size',
        'rotate_on',
        'rotate_external',
        'entry_content',
        'structured_fields',
        )

    def _getLogHandler(self):
        """
        See: `EventHandlerWithLogger`.
        """
        path = self.configuration.path
        if not path:
            raise ServerException('Missing configuration for "path".')

        path_variables = NoStrBunch(
            host=NoStrBunch(
                name=socket.gethostname(),
                fqdn=socket.getfqdn(),
                ),
            )
        try:
            path = public_format(path, **path_variables)
        except Exception as error:
            raise ServerException('Invalid "path" configuration: %s' % error)

        path = os.path.abspath(local_filesystem.getEncodedPath(path))

        rotate_count = self.configuration.rotate_count
        bytes = self.configuration.rotate_at_size
        rotate_on_config = self.configuration.rotate_on

        # When rotate_on is configured as 0 day-of-month it should
        # be disabled. This is due to how the Web Manager UI works.
        rotate_on = rotate_on_config and not (
            rotate_on_config[1] == 'day-of-month' and rotate_on_config[0] == 0)

        handler = None
        if self.configuration.rotate_external:
            handler = WatchedFileHandler(path)
        elif rotate_on:
            day_or_time, calendar_type = rotate_on_config
            if calendar_type == 'day-of-month':
                handler = MonthlyRotatingFileHandler(
                    path,
                    event_emitter=self.emitEvent,
                    on_rollover_failed=self._ebActionFailed,
                    day=day_or_time,
                    rotate_count=rotate_count,
                    clock=self._scheduler,
                    )
            elif calendar_type == 'time-of-day':
                handler = DailyRotatingFileHandler(
                    path,
                    event_emitter=self.emitEvent,
                    on_rollover_failed=self._ebActionFailed,
                    day_offset=day_or_time,
                    rotate_count=rotate_count,
                    clock=self._scheduler,
                    )
        elif bytes:
            handler = SizeRotatingFileHandler(
                path,
                event_emitter=self.emitEvent,
                max_bytes=bytes,
                rotate_count=rotate_count,
                )
        else:
            handler = FileHandler(path)

        return handler
