# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
Events are the basic mechanism for signaling that something specific took
place inside the server/client.

The products have an **event definitions** database.
The event definition is the basic configuration for each event.
Multiple event definitions are grouped using EventGroupDefinition.

The **event definition** is the static part of the event, generated by the
product at startup time.

The **event** itself is the dynamic part, generated by the product at runtime.
The products will **emit** an event.

An event is emitted using emit(ID, DATA).
ID is the unique event definition id. It will be looked in the database to
retrieve a message and other configurations.
DATA is a dictionary with key:values that will be applied for this event.

If the event database is not available (for example at product initialization)
an event can be emitted using emit(ID, MESSAGE). In this case, the event
will be emitted using default configuration.
"""
from __future__ import unicode_literals

from __future__ import absolute_import
import os
import time
import uuid
from logging import Formatter

from bunch import Bunch
from ordered_set import OrderedSet
from zope.interface import implementer

from chevah.server import MODULE_PATH, force_unicode, json
from chevah.server.commons.constant import STATIC_FOLDER_NAME
from chevah.server.commons.csv import to_csv_row
from chevah.server.commons.exception import ServerError, ServerException
from chevah.server.commons.interface import (
    IEvent,
    IEventDefinition,
    IEventGroupDefinition,
    IEventsDefinition,
    IChevahTimeStamp,
    )
from chevah.server.commons.constant import CONFIGURATION_DEFAULTS
from chevah.server.commons.jinja import ChevahJinjaEnvironment
from chevah.server.commons.public_format import public_format, isidentifier
from chevah.server.extension import load_extension_object
from chevah.server.json.file import JSONFile
import six

CWA_14051_FORMAT = '%Y-%m-%d %H:%M:%S'
ISO_8601_EXTENDED = '%Y-%m-%dT%H:%M:%S'
ISO_8601_BASIC = '%Y%m%dT%H%M%S'


def _get_utc(timestamp):
    """
    Return the UTC time structure.
    """
    target = time.localtime(timestamp)
    if target.tm_isdst:
        offset = time.altzone
    else:
        offset = time.timezone
    return time.localtime(timestamp + offset)


@implementer(IChevahTimeStamp)
class ChevahTimeStamp(object):
    """
    Object representing a local timestamp.
    """
    def __init__(self, timestamp):
        self._timestamp = timestamp

    @property
    def timestamp(self):
        """
        Local Unix timestamp with milliseconds.
        """
        return self._timestamp

    @property
    def cwa_14051(self):
        """
        Local CEN CWA 14051 format string representation of timestamp.
        """
        return six.text_type(time.strftime(
            CWA_14051_FORMAT, time.localtime(self.timestamp)))

    @property
    def iso_8601(self):
        """
        UTC ISO 8601 format string representation without fractional seconds.
        """
        return six.text_type(time.strftime(
            ISO_8601_EXTENDED, _get_utc(self.timestamp)) + 'Z')

    @property
    def iso_8601_basic(self):
        """
        UTC ISO 8601 format string representation with only the T delimited
        between date and time.
        """
        return six.text_type(time.strftime(
            ISO_8601_BASIC, _get_utc(self.timestamp)) + 'Z')

    @property
    def iso_8601_compact(self):
        """
        Local ISO 8601 format string representation without timezone
        designator (unqualified) and with only the T delimited between date
        and time.
        """
        return six.text_type(time.strftime(
            ISO_8601_BASIC, time.localtime(self.timestamp)))

    @property
    def iso_8601_fractional(self):
        """
        UTC ISO 8601 format string representation with fractional seconds.
        """
        fractional = str(self.timestamp - int(self.timestamp))[1:]
        return six.text_type(time.strftime(
            ISO_8601_EXTENDED, _get_utc(self.timestamp)) + fractional + 'Z')

    @property
    def iso_8601_local(self):
        """
        Local ISO 8601 format string representation without fractional seconds.
        """
        target = time.localtime(self.timestamp)
        if target.tm_isdst:
            offset = time.altzone
        else:
            offset = time.timezone

        hour_offset = offset / 3600.0
        minute_offset = (abs(offset) % 3600) / 60.0

        if offset > 0:
            utc_offset = '-%02i:%02i' % (hour_offset, minute_offset)
        else:
            utc_offset = '+%02i:%02i' % (hour_offset * -1, minute_offset)

        return six.text_type(time.strftime(
            ISO_8601_EXTENDED + utc_offset, target))


@implementer(IEventGroupDefinition)
class EventGroupDefinition(object):
    """
    The definition for a group of events.

    Events will be attached (tagged) to various groups.
    The EventDefinition <-> EventGroupDefinition association is done in the
    EventDefinition.

    EventGroupDefinition has the following fields:
     * name
     * description
    """
    def __init__(self, name, description=None):
        self._name = name
        if description is None:
            description = 'No description.'
        self._description = description

    def __repr__(self):
        return 'EventGroupDefinition(name=%s)' % (self.name)

    @property
    def name(self):
        """
        Return the name of this group.
        """
        return self._name

    @property
    def description(self):
        """
        Return the description of this group.
        """
        return self._description


@implementer(IEventDefinition)
class EventDefinition(object):
    """
    The definition for a single event.
    """

    def __init__(
        self, id, message, groups=None, description=None,
        version_added=None, version_removed=None, data=None
            ):
        self._id = id
        self.message = message
        if groups is None:
            groups = []
        self.groups = groups
        self.description = description
        self.version_added = version_added
        self.version_removed = version_removed
        if data is None:
            data = {}
        self.data = data

    def __repr__(self):
        groups = [group.name for group in self.groups]
        return 'EventDefinition(id=%s, message=%s, groups=%s)' % (
            self.id, self.message, ', '.join(groups))

    @property
    def id(self):
        """
        The raw id of this event.
        """
        return self._id

    @property
    def id_padded(self):
        """
        The padded is as used in documentation and logs.
        """
        if len(self._id) >= 5:
            return self._id
        else:
            return ('0' * (5 - len(self._id))) + self._id

    @property
    def group_names(self):
        """
        Return a set containing names for the attached groups.
        """
        result = OrderedSet([group.name for group in self.groups])
        return result


@implementer(IEvent)
class Event(object):
    """
    An event as raised by a service.
    """

    def __init__(self, id, creator, account, data, server, message=''):
        self.created = time.time()
        # This is called `id` but is more like the event type.
        self.id = id
        # This is the real unique id.
        self._uuid = six.text_type(uuid.uuid4())
        self.data = data
        self.account = account
        self.creator = creator
        # This is the text serialization of the event, already resolved.
        self.message = message
        self.server = server

    @property
    def uuid(self):
        """
        The unique ID of the event.
        """
        return self._uuid

    @property
    def component(self):
        """
        The component that emitted this event.
        """
        return self.creator

    @property
    def data_json(self):
        """
        Return the string representation of event data.
        """
        return json.dumps(self.data, indent=4, sort_keys=True)

    @property
    def timestamp(self):
        """
        A ChevahTimeStamp object containing the time the event was emitted.
        """
        return ChevahTimeStamp(self.created)

    def asText(self):
        """
        Return the string representation of this event.
        """
        # FIXME:3174:
        # code should be moved into here.
        from chevah.server.event.stdlib_logging import (
            EventLogRecordFormatterAdapter,
            StringFormatter,
            )
        formatter = EventLogRecordFormatterAdapter(StringFormatter())

        timestamp = six.text_type(
            time.strftime(
                CWA_14051_FORMAT,
                time.localtime(self.created),
                ))

        return '%s %s %s %s %s %s' % (
            self.id,
            timestamp,
            formatter.formatService(self.creator),
            formatter.formatAvatar(self.account),
            formatter.formatPeer(self.account.peer),
            self.message,
            )

    def asDict(self):
        """
        Return the whole event as a dict.
        """
        return json.loads(json.dumps(self))

    def _renderAsText(self, template):
        """
        Render from given template.
        """
        event_dict = self.asDict()
        # We add {CR} and {LF} to the dict so users can customize the
        # end line char.
        event_dict['cr'] = '\r'
        event_dict['lf'] = '\n'

        return public_format(template, **Bunch.fromDict(event_dict))

    def renderFromTemplate(self, template, option_name):
        """
        Try to render the given `template`. The template could contain in
        itself variables such as `{id}`, `{message}`, etc., which are
        replaced by the attributes of the current event.

        Raises ServerException if the template contains bad values.
        """
        try:
            return self._renderAsText(template=template)
        except ServerException as error:
            raise ServerException(
                'Error in "%s" configuration: %s' % (
                    option_name, error.message))

    def __repr__(self):
        try:
            creator = json.dumps(self.creator)
        except Exception:
            creator = repr(self.creator)

        return '%s Event:0x%x\na-%s\n\n%s\n\nc-%s)\n---------------\n\n' % (
            self.id,
            id(self),
            json.dumps(self.account),
            json.dumps(self.data),
            creator,
            )


class EventNotFound(Event):
    """
    Event raised when an initial event was not found in the database.
    """
    def __init__(
        self, event_id, data, creator, account, message_resolver,
        server, traceback=''
            ):
        data = data.copy()
        data.update({
            'id': event_id,
            'details': traceback,
            })
        message = message_resolver('20144', data)
        super(EventNotFound, self).__init__(
            id='20144',
            creator=creator,
            account=account,
            data=data,
            message=message,
            server=server,
            )


class EventFailedInterpolation(Event):
    """
    Event raised what an initial event message interpolation failed.
    """
    def __init__(
        self, event_id, bad_data, creator, account, error, message_resolver,
        server,
            ):
        data = {
            'id': event_id,
            'bad_data': bad_data,
            'details': force_unicode(error),
            }
        message = message_resolver('20145', data)
        super(EventFailedInterpolation, self).__init__(
            id='20145',
            creator=creator,
            account=account,
            data=data,
            message=message,
            server=server,
            )


@implementer(IEventsDefinition)
class EventsDefinition(JSONFile):
    """
    Manages the definitions for Event and EventGroups.

    EventGroups and EventDefinitions are stored in a configuration file.
    """

    def __init__(self, path=None, file=None, auto_load=True):

        if path is None and file is None:
            path = os.path.join(
                MODULE_PATH, STATIC_FOLDER_NAME,
                'events', 'events.json')

        super(EventsDefinition, self).__init__(path=path, file=file)

        self._group_definitions = {}
        self._event_definitions = {}

        if auto_load:
            self.load()

    def load(self):
        """
        See `IEventsDefinition`.
        """
        super(EventsDefinition, self).load()

        # Initialize empty file or if no group or events are defined.
        if 'groups' not in self.data:
            self.data['groups'] = {}
        if 'events' not in self.data:
            self.data['events'] = {}

        # First we must load the EventGroups and later EventDefinitions
        # as EventDefinitions depends on EventGroups.
        self._loadEventGroupDefinitions()
        self._loadEventDefinitions()

    def eventHasGroup(self, event_id, group_name):
        """
        See `IEventsDefinition`.
        """
        if event_id not in self._event_definitions:
            return False

        return group_name in self.getEventDefinition(event_id).group_names

    def _loadEventGroupDefinitions(self):
        """
        Load all configured EventGroups.
        """
        for group_id, group_data in self.data['groups'].items():

            name = group_id
            description = self.getValueOrNone(group_data, 'description')

            event_group = EventGroupDefinition(
                name=name, description=description)

            self._group_definitions[name] = event_group

    def _loadEventDefinitions(self):
        """
        Load all configured EventDefinitions.
        """
        for event_id, event_data in self.data['events'].items():
            self._loadEventDefinition(
                event_id=event_id, event_data=event_data)

    def _loadEventDefinition(self, event_id, event_data):
        """
        Load a configured EventDefinition.
        """
        message = event_data['message']
        description = self.getValueOrNone(event_data, 'description')

        groups = []
        for group_name in event_data['groups']:
            group_name = group_name.strip()
            group = self.getEventGroupDefinition(name=group_name)
            groups.append(group)

        version_added = self.getValueOrNone(event_data, 'version_added')
        version_removed = self.getValueOrNone(event_data, 'version_removed')
        data = event_data['data']

        event_definition = EventDefinition(
            id=event_id,
            message=message,
            description=description,
            groups=groups,
            version_added=version_added,
            version_removed=version_removed,
            data=data,
            )

        self._event_definitions[event_definition.id] = event_definition

    def getEventDefinition(self, id):
        """
        See `IEventsDefinition`.
        """
        try:
            event_definition = self._event_definitions[id]
        except KeyError:
            data = {
                'message': 'No EventDefinition with id "%(id)s"',
                'id': id,
                }
            raise ServerError('20150', data)
        else:
            return event_definition

    def getAllEventDefinitions(self):
        """
        See `IEventsDefinition`.
        """
        return self._event_definitions.copy()

    def getAllEventDefinitionsPadded(self):
        """
        See `IEventsDefinition`.
        """
        result = {}
        for event_id, event in self._event_definitions.items():
            result[event.id_padded] = event
        return result

    def getEventGroupDefinition(self, name):
        """
        See `IEventsDefinition`.
        """
        try:
            event_group = self._group_definitions[name]
        except KeyError:
            data = {
                'name': name,
                }
            raise ServerError('20151', data=data)
        else:
            return event_group

    def getAllEventGroupDefinitions(self):
        """
        See `IEventsDefinition`.
        """
        result = self._group_definitions.copy()
        try:
            del result['__last_group__']
        except KeyError:
            """
            We might have an empty DB.
            """

        return result

    def generateDocumentation(self, template):
        """
        See `IEventsDefinition`.
        """
        return template.render(
            groups=self.getAllEventGroupDefinitions(),
            events=self.getAllEventDefinitionsPadded(),
            )

    def renderDocumentationForEvents(self):
        """
        Return a string containing the documentation for the events.

        Documentation is formatted in ReStructuredText.
        """
        jinja_environment = ChevahJinjaEnvironment(
            source=['documentation'],
            # Since we produce RST, it's OK not to autoescape by default.
            autoescape=False,
            )
        template = jinja_environment.get_template('events/events.j2')
        return self.generateDocumentation(template=template)

    def renderDocumentationForEventGroups(self):
        """
        Return a string containing the documentation for the event groups.

        Documentation is formatted in ReStructuredText.
        """
        jinja_environment = ChevahJinjaEnvironment(
            source=['documentation'],
            # Since we product RST, is ok not to autoescape by default.
            autoescape=False,
            )
        template = jinja_environment.get_template('events/groups.j2')
        return self.generateDocumentation(template=template)


class StringFormatter(Formatter):
    """
    Serializes an `EventLogRecord` to its Unicode representation.
    """
    def __init__(self, fmt=None):
        """
        See: `logging.Formatter`
        """

        if not fmt:
            fmt = CONFIGURATION_DEFAULTS['entry_content']

        kwargs = {'fmt': fmt}

        if six.PY3:
            # This can be removed once we only support Py3.
            kwargs['style'] = '{'
            # This will need a refactor, since for python extension,
            # we pass invalid format.
            kwargs['validate'] = False

        super(StringFormatter, self).__init__(**kwargs)

        # Load extension or fallback to default behaviour.
        self.format = load_extension_object(fmt, default=self._format_text)

    def _format_text(self, event):
        """
        Return a custom string representation of `EventLogRecord`.
        """
        msg = event.renderFromTemplate(self._fmt, 'entry_content')
        if not msg.endswith('\n'):
            msg += os.linesep
        return msg


class CSVFormatter(Formatter):
    """
    Serializes an `EventLogRecord` to it's CSV representation.
    """

    def __init__(self, fmt):
        """
        See: `logging.Formatter`

        `fmt` is a list of `EventLogRecord` field names.
        """
        format_fields = []
        field_names = []
        for field in fmt:
            if not field.startswith('{') and not field.endswith('}'):
                # Fields can be configured with or without braces.
                # If not configured as {field} we need to add the braces.
                field_name = field
                format_field = '{{{}}}'.format(field)
            else:
                field_name = field[1:-1]
                format_field = field

            if not isidentifier(field_name):
                raise ServerException(
                    'Invalid field name: "%s".' % field)

            field_names.append(field_name)
            format_fields.append(format_field)

        self.header = to_csv_row(field_names).upper()

        kwargs = {'fmt': ''}
        if six.PY3:
            # This can be removed once we only support Py3.
            kwargs['style'] = '{'
        super(CSVFormatter, self).__init__(**kwargs)
        self._fmt = format_fields

    def format(self, event):
        """
        Return a custom CSV representation of `EventLogRecord`.
        """
        fields = []
        for field in self._fmt:
            field_value = event.renderFromTemplate(
                field,
                'structured_fields'
                )
            fields.append(field_value)
        return to_csv_row(fields)
