# Copyright (c) 2018 Adi Roiban.
# See LICENSE for details.
"""
Event handler which encrypts/decrypts PGP files.
"""
from __future__ import unicode_literals

from __future__ import absolute_import
from twisted.internet import defer
from twisted.internet.threads import deferToThread

from chevah_compat import local_filesystem
from chevah.server import force_unicode
from chevah.server.commons.constant import DEFAULT_LOCAL_FILESYSTEM_UUID
from chevah.server.commons.exception import ServerException
from chevah.server.commons.stream import DelayedWrite
from chevah.server.event.common import (
    CheckDestinationMixin,
    SourceAttributeMixin,
    )
from chevah.server.event.handler import EventHandlerWithFilter
from chevah.server.openpgp.wrapper import ChevahGPG


class OpenPGPEventHandler(
        EventHandlerWithFilter, SourceAttributeMixin, CheckDestinationMixin):
    """
    Handle the source file using OpenPGP.

    Each handler has separate public and private keyrings.
    They are removed when handler stops.
    """
    PROPERTIES_REQUIRING_RESTART = (
        'encryption_cipher',
        'encryption_extension',
        'password_update',
        'encryption_public_keys',
        'decryption_private_keys',
        )

    # List of GPG Ciphers:
    # IDEA, 3DES, CAST5, BLOWFISH, AES, AES192, AES256, TWOFISH,
    # CAMELLIA128, CAMELLIA192, CAMELLIA256
    # We only support some of them for encryption.
    _SUPPORTED_ENCRYPTIONS = {
        'aes': b'AES',
        'aes128': b'AES',
        'aes192': b'AES192',
        'aes256': b'AES256',
        'cast5': b'CAST5',
        '3des': b'3DES',
        }

    # The main GPG object.
    _gpg = None
    # This triggers symmetric.
    _encryption_cipher = None
    _encryption_extension = None
    _passphrase = None
    # This triggers asymmetric.
    _encryption_recipients = None

    def _onStart(self, avatar=None):
        """
        See: `Runnable`.
        """
        self._resetState()

        self._passphrase = self.configuration.password
        if self._passphrase:
            self._passphrase = self._passphrase.encode('utf-8')
        else:
            # When no password is set, the GPG wrappners needs `None``.
            self._passphrase = None

        self._encryption_extension = self.configuration.encryption_extension
        if not self._encryption_extension:
            raise ServerException(
                'You need to configure "encryption_extension".')

        if self._passphrase:
            encryption_cipher = self.configuration.encryption_cipher
            try:
                self._encryption_cipher = self._SUPPORTED_ENCRYPTIONS[
                    encryption_cipher]
            except KeyError:
                raise ServerException(
                    'encryption_cipher "%s" is not supported.' % (
                        encryption_cipher))

        public_keys = self.configuration.encryption_public_keys
        private_keys = self.configuration.decryption_private_keys

        self._location = self.root.locations.getSection(
            DEFAULT_LOCAL_FILESYSTEM_UUID)

        try:
            self._gpg = ChevahGPG(uuid=self.configuration.uuid)
        except Exception as error:
            raise ServerException(
                'Failed to initialize the OpenPGP environment. %s' % (error,))

        if not self._gpg.is_valid_passphrase(self._passphrase):
            raise ServerException(
                '"passphrase" contains characters which are not supported.')

        if not (self._passphrase or public_keys or private_keys):
            raise ServerException('A non empty "passphrase" is required.')

        if (self._passphrase and (public_keys or private_keys)):
            raise ServerException(
                'Both symmetric and asymmetric encryption are not supported '
                'at the same time.'
                )

        def check_pgp_import(results, kind):
            """
            Check the result of an PGP import operation and raise an
            exception on errors.
            """
            for response in results:
                have_error = response.get('problem', None)
                if not have_error:
                    continue
                message = response.get(
                    'text', 'Unknown private key import error.')
                raise ServerException(
                    'Failed to import the %s keys. %s' % (
                        kind, message))

        if public_keys:
            result = self._gpg.import_keys(public_keys)
            check_pgp_import(result.results, 'public encryption')
            if not result.fingerprints:
                raise ServerException(
                    'Invalid "encryption_public_keys" configuration.')
            self._encryption_recipients = result.fingerprints
            self._gpg.trust_keys(result.fingerprints, 'TRUST_ULTIMATE')
            # Check if we have any private keys.
            result = self._gpg.list_keys(secret=True)
            if result.fingerprints:
                raise ServerException(
                    'Private keys detected inside the '
                    '"encryption_public_keys" configuration.'
                    )

        if private_keys:
            result = self._gpg.import_keys(private_keys)
            check_pgp_import(result.results, 'private decryption')
            self._gpg.trust_keys(result.fingerprints, 'TRUST_ULTIMATE')

    def _onStop(self, avatar=None):
        """
        See: `Runnable`.
        """
        if self._gpg:
            self._gpg.stop()
        self._resetState()

    def _resetState(self):
        """
        Reset the internal state.
        """
        self._location = None
        self._gpg = None
        self._passphrase = None
        self._encryption_recipients = None
        self._encryption_recipients = None

    def _getHandler(self, source_name):
        """
        See: SourceAttributeMixin.
        """
        extension = source_name[-4:].lower()
        if extension.endswith('.pgp') or extension.endswith('.gpg'):
            handler = self._handleDecrypt
            if self._passphrase:
                operation = 'PGP symmetric decryption'
            else:
                operation = 'PGP asymmetric decryption'
        else:
            handler = self._handleEncrypt
            if self._passphrase:
                operation = 'PGP symmetric encryption'
            else:
                operation = 'PGP asymmetric encryption for %s' % (
                    ','.join(self._encryption_recipients))

        return (handler, operation)

    @defer.inlineCallbacks
    def _handleEncrypt(
            self, source_path, source_name, destination_base, event, operation):
        """
        Encrypt file using PGP.
        """
        destination_path = self._location.joinPath(
            destination_base, source_name + self._encryption_extension)

        yield self._checkDestination(source_path, destination_path, operation)

        def gpg_encrypt():
            path = local_filesystem.getEncodedPath(source_path)
            dest = local_filesystem.getEncodedPath(destination_path)
            with open(path, 'rb') as source:
                with DelayedWrite(dest) as output:
                    return self._gpg.encrypt_file(
                        source,
                        recipients=self._encryption_recipients,
                        passphrase=self._passphrase,
                        symmetric=self._encryption_cipher,
                        armor=False,
                        output=output,
                        )

        result = yield deferToThread(gpg_encrypt)
        if not result.ok:
            errors = result.stderr.splitlines()[1:]
            details = force_unicode(' '.join(errors))
            raise ServerException('Failed to encrypt: %s' % (details))

        defer.returnValue(destination_path)

    @defer.inlineCallbacks
    def _handleDecrypt(
            self, source_path, source_name, destination_base, event, operation):
        """
        Decrypt file using PGP.
        """
        # Remove .pgp/.gpg extension.
        destination_name = source_name[:-4]
        destination_path = self._location.joinPath(
            destination_base, destination_name)

        yield self._checkDestination(source_path, destination_path, operation)

        def gpg_decrypt():
            path = local_filesystem.getEncodedPath(source_path)
            dest = local_filesystem.getEncodedPath(destination_path)
            with open(path, 'rb') as source:
                with DelayedWrite(dest) as output:
                    return self._gpg.decrypt_file(
                        source,
                        passphrase=self._passphrase,
                        output=output,
                        )

        result = yield deferToThread(gpg_decrypt)
        if not result.ok:
            raise ServerException(
                'Failed to decrypt: %s' % (self._formatError(result)))

        defer.returnValue(destination_path)

    def _formatError(self, raw):
        """
        Return a easy to read error.
        """
        try:
            error = raw.stderr
        except AttributeError:
            return force_unicode(raw.status)

        result = []
        for line in error.splitlines()[1:]:
            line = force_unicode(line)
            if 'gpg: keyring' in line:
                continue
            if '[GNUPG:]' in line:
                continue
            result.append(line[5:])

        return ' '.join(result)
