# Copyright (c) 2021 Adi Roiban.
# See LICENSE for details.
"""
Event handler for publishing an event over AMQP 0-9-1 with RabbitMQ.

RabbitMQ also supports MQTT
"""
from __future__ import unicode_literals

from __future__ import absolute_import
import pika
from pika import spec
from pika.adapters import twisted_connection
from pika.exceptions import AMQPConnectionError, ChannelClosed
from twisted.internet import defer, endpoints
from zope.interface import implementer

from chevah_compat import process_capabilities

from chevah.server import force_unicode
from chevah.server.commons.configuration import parse_url
from chevah.server.commons.exception import ServerException
from chevah.server.commons.interface import IRabbitMQEventHandlerConfiguration
from chevah.server.commons.ssl import (
    SSLClientMixin,
    SSLConfigurationSectionMixin,
    )
from chevah.server.configuration.model import (
    WritableString,
    )
from chevah.server.configuration.option import (
    RetryIncreaseOptionsMixin,
    UsernamePasswordPublicOptionsMixin,
    URLOptions,
    )
from chevah.server.event.configuration import (
    _EventHandlerConfigurationBase,
    FilterOptionsMixin,
    )
from chevah.server.event.handler import EventHandlerWithFilter


_PREFETCH_COUNT = 2
# FIXME:2672:
# Add configurable timeout.
_CONNECT_TIMEOUT = 30


@implementer(IRabbitMQEventHandlerConfiguration)
class RabbitMQEventHandlerConfiguration(
    _EventHandlerConfigurationBase,
    FilterOptionsMixin,
    SSLConfigurationSectionMixin,
    UsernamePasswordPublicOptionsMixin,
    RetryIncreaseOptionsMixin,
    URLOptions,
        ):
    """
    Configuration for the file dispatcher event handler.
    """
    exchange = WritableString("exchange")
    routing_key = WritableString("routing_key")
    payload_template = WritableString('payload_template')


class RabbitMQEventHandler(EventHandlerWithFilter, SSLClientMixin):
    """
    Send an event to RabbitMQ using AMQP 0-9-1.
    """
    PROPERTIES_REQUIRING_RESTART = (
        'url',
        'username',
        'password_update',
        ) + SSLConfigurationSectionMixin.PROPERTIES_REQUIRING_RESTART

    _protocol = None

    @defer.inlineCallbacks
    def _onStart(self, avatar=None):
        """
        See: `Runnable`.
        """
        url = parse_url(self.configuration.url)
        self._protocol = self._getProtocol(url)
        endpoint = yield self._getEndpoint(url)

        try:
            yield endpoints.connectProtocol(endpoint, self._protocol)
        except Exception as error:
            raise ServerException(
                'Failed to connect to the RabbitMQ AMQP server. %s' % (
                    force_unicode(error,)))

        try:
            yield self._protocol.ready
        except AMQPConnectionError:
            # Pika is not returning any details to the handshake process.
            raise ServerException(
                'Failed to authenticate to RabbitMQ AMQP server. '
                'Check credentials and virtual host configuration.'
                )
        self._protocol.closed.addCallback(self._cbClientClose)

    def _onStop(self, avatar=None):
        """
        See: `Runnable`.
        """
        if not self._protocol:
            # Might stop after a failure at initialization.
            return

        protocol, self._protocol = self._protocol, None

        if not protocol.connected:
            return

        return protocol.close()

    def _cbClientClose(self, error):
        """
        Called when AMQP client connection is lost.

        Called for normal close operation or on errors.
        """
        if not self._protocol:
            # This might be an normal stop of the handler.
            return

        # FIXME:5271:
        # Retry with a delay... and stop after a number of disconnections
        # Might be disconnected from the server.
        self.fail(error)

    def _onEvent(self, event):
        """
        See: EventHandlerWithFilterMixin.
        """
        template = self._configuration.payload_template
        message = event.renderFromTemplate(template, 'payload_template')

        if not message:
            raise ServerException(
                'RabbitMQ AMQP publish event handler '
                'configured without a payload.'
                )

        return self._protocol.sendMessage(
            self._configuration.exchange,
            self._configuration.routing_key,
            message,
            )

    def _getProtocol(self, url):
        """
        Return the configured AMQP protocol for this event handler.
        """
        virtual_host = url['path'].strip('/')

        client_properties = {
            'product': 'SFTPPlus RabbitMQ AMQP Publisher',
            'platform': process_capabilities.os_name,
            # Information is there to help with testing.
            'information': self._configuration.name,
            # Have own event handler version + pika version.
            'version': '0.1+p' + pika.__version__,
            # capabilities copied from pika.connection.Connection
            'capabilities': {
                'authentication_failure_close': True,
                'basic.nack': True,
                'connection.blocked': True,
                'consumer_cancel_notify': True,
                'publisher_confirms': True
                },
            }

        parameters = pika.ConnectionParameters(
            # FIXME:6474:
            # Also add the custom port here.
            host=url['address'],
            virtual_host=virtual_host,
            client_properties=client_properties,
            credentials=pika.PlainCredentials(
                self.configuration.username,
                self.configuration.password,
                ),
            )
        return PikaProtocol(parameters)

    @defer.inlineCallbacks
    def _getEndpoint(self, url):
        """
        Return the endpoint use to connect to the server.
        """
        protocol = url['protocol'].lower()
        host = url['address']
        port = url['port']

        if protocol == 'amqp':
            defer.returnValue(endpoints.TCP4ClientEndpoint(
                reactor=self._scheduler,
                host=host,
                port=port,
                timeout=_CONNECT_TIMEOUT,
                ))

        if protocol == 'amqps':
            context_factory = yield self._initStartedContextFactory(
                remote_servers=[(host, port)])
            defer.returnValue(endpoints.SSL4ClientEndpoint(
                reactor=self._scheduler,
                host=host,
                port=port,
                sslContextFactory=context_factory,
                timeout=_CONNECT_TIMEOUT,
                ))

        raise ServerException('Unknown RabbitMQ protocol: "%s".' % (protocol,))


class PikaProtocol(twisted_connection.TwistedProtocolConnection):
    """
    A thin wrapper around Pika Twisted protocol to simplify the publish
    operation.
    """
    connected = False

    # Deferred used to recreate a closed channeled.
    # Here to help with testing.
    _recreate_deferred = None

    @defer.inlineCallbacks
    def connectionReady(self):
        """
        Called by Pika after AMQP handshake is done.
        """
        # Every AMQP communication must be done over a channel.
        self._channel = yield self.channel()
        yield self._channel.basic_qos(prefetch_count=_PREFETCH_COUNT)
        self.connected = True
        yield self._channel.confirm_delivery()

    @defer.inlineCallbacks
    def sendMessage(self, exchange, routing_key, payload):
        """
        Send a single message.

        Raise an exception when failing to send.
        """
        def eb_channel_recreate(failure):
            """
            Called when channel was closed on error and we fail to
            recreate it.
            """
            # Stop the connection as we can't operate without a channel.
            return self.close()

        properties = spec.BasicProperties(
            delivery_mode=2,  # Make message persistent.
            )
        try:
            yield self._channel.basic_publish(
                exchange=exchange,
                routing_key=routing_key,
                body=payload,
                properties=properties,
                )
        except ChannelClosed:

            # Since the channel was closed, recreate it.
            self._recreate_deferred = self.connectionReady()
            self._recreate_deferred.addErrback(eb_channel_recreate)

            # FIXME:5271:
            # Add support to retry failed messages.
            # Forward the exception to let the event handler know
            raise
