# Copyright (c) 2015 Adi Roiban.
# See LICENSE for details.
"""
Support for sending events to the Python's stdlib logging system.
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from logging import (
    _removeHandlerRef,
    INFO,
    LogRecord,
    )

from bunch import Bunch
from twisted.internet import defer

from chevah.server.commons.constant import (
    LOGGER_NAME,
    )
from chevah.server.commons.interface import IFormattedEvent
from chevah.server.event.handler import EventHandlerWithFilter
from chevah.server.event.model import (
    CSVFormatter,
    StringFormatter,
    )


class EventLogRecord(LogRecord, object):
    """
    An entry that will be received by the stdlib log handlers.
    """

    def __init__(self, event):
        super(EventLogRecord, self).__init__(
            LOGGER_NAME,
            INFO,
            '(unknown file)',
            0,
            event.message,
            None,
            None,
            '(unknown function)',
            )

        self._event = event

    @property
    def event(self):
        """
        Event for this record.
        """
        return self._event

    def __repr__(self):
        result = 'LogEntry(%s, %s, %s, avatar=%s, peer=%s)' % (
            self.event.timestamp.cwa_14051, self.event.id, self.msg,
            self.event.account, self.event.account.peer)
        return result


class EventHandlerWithLogger(EventHandlerWithFilter):
    """
    An event handler which sends the event to the Python sdtlib logger.

    Resolves the `Runnable` `_onStart` and `_onStop` abstract method.
    Resolves the `EventHandlerMixin` `_onEvent` abstract method.
    Requires the `_getLogHandler` abstract method.
    """

    def __init__(self, parent, configuration):
        super(EventHandlerWithLogger, self).__init__(parent, configuration)
        self._handler = None

    def _getLogHandler(self):
        """
        Called to get the Python stdlib handler.
        """
        raise NotImplementedError('_getLogHandler not implemented.')

    def _onStart(self, avatar=None):
        """
        See: `Runnable`.
        """
        self._handler = self._getLogHandler()
        if IFormattedEvent.providedBy(self.configuration):
            # We use a cached configuration to reduce the hits we made to the
            # configuration object.
            configuration = Bunch(
                entry_content=self.configuration.entry_content,
                structured_fields=self.configuration.structured_fields)
            if configuration.structured_fields:
                csv_formatter = CSVFormatter(configuration.structured_fields)
                self._handler.file_header = csv_formatter.header
                self._handler.formatter = EventLogRecordFormatterAdapter(
                    csv_formatter)
            else:
                self._handler.formatter = EventLogRecordFormatterAdapter(
                    StringFormatter(configuration.entry_content))
        else:
            self._handler.formatter = EventLogRecordFormatterAdapter(
                StringFormatter())

        # We are reusing the default Python log handler, but we don't want
        # to use their cleanup code.
        _removeHandlerRef(self._handler)
        return defer.maybeDeferred(self._handler.open)

    def _onStop(self, avatar=None):
        """
        See: `Runnable`.
        """
        if not self._handler:
            return

        self._handler.close()
        self._handler = None

    def _onEvent(self, event):
        """
        See: EventHandlerWithFilterMixin.
        """
        record = EventLogRecord(event)
        # handler.emit will never return a deferred, but we use this to
        # convert exceptions into failed deferred.
        # It goes directly for `emit` method instead of the standard `handle`
        # as we don't care about filtering and mutex.
        deferred = defer.maybeDeferred(self._handler.emit, record)
        return deferred


class EventLogRecordFormatterAdapter(object):
    """
    A class for formatting EventLogRecord using the specified formatter.
    """
    def __init__(self, formatter):
        self._formatter = formatter

    def format(self, record):
        """
        Return a serialized EventLogRecord.
        """
        return self._formatter.format(record.event)

    @staticmethod
    def formatPeer(peer):
        """
        Return host and port address of the peer if not `None`, or u'None'
        otherwise.
        """
        if peer:
            return '%s:%d' % (peer.host, peer.port)

        return 'None'

    @staticmethod
    def formatAvatar(avatar):
        """
        Return the avatar's name if not `None`, or u'None' otherwise.
        """
        if avatar:
            return avatar.name

        return 'None'

    @staticmethod
    def formatService(creator):
        """
        Returns service name or u'None` if the service is `None` or an empty
        string.
        """
        # FIXME:3231:
        # Add tests once we have real creator.
        # Also consider having the creator name as UUID are not human
        # friendly.
        if creator:
            return creator.uuid

        return 'None'
