# Copyright (c) 2014 Adi Roiban.
# See LICENSE for details.
"""
Event handler sending events to a Syslog server.
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from logging import Handler
from logging.handlers import SysLogHandler as PythonSyslogLogHandler
from six.moves.urllib.parse import urlparse
import socket
import sys
import time

from twisted.internet import defer, threads

from chevah_compat import local_filesystem, process_capabilities
from chevah.server.commons.exception import ServerException
from chevah.server.event.stdlib_logging import (
    EventHandlerWithLogger,
    EventLogRecordFormatterAdapter,
    )
import six


class SyslogLogHandler(PythonSyslogLogHandler):
    """
    Patched version of Python SyslogLogHandler.
    """

    def __init__(
        self,
        address=('localhost', 514),
        facility=PythonSyslogLogHandler.LOG_USER,
        socktype=None,
        process_name=None,
            ):
        """
        Initialize a handler.

        If address is specified as a string, a UNIX socket is used. To log to a
        local syslogd, "SysLogHandler(address="/dev/log")" can be used.
        If facility is not specified, LOG_USER is used. If socktype is
        specified as socket.SOCK_DGRAM or socket.SOCK_STREAM, that specific
        socket type will be used. For Unix sockets, you can also specify a
        socktype of None, in which case socket.SOCK_DGRAM will be used, falling
        back to socket.SOCK_STREAM.
        """
        Handler.__init__(self)

        self.address = address
        self.facility = facility
        self.socktype = socktype
        self._process_name = process_name
        self._hostname = socket.gethostname()
        self._connected = False
        self._connecting = False

        # Deferred for the current reconnection. Used to help with testing.
        self._connect_deferred = None

    def format(self, record):
        """
        Return a string representation of `EventLogRecord` formatted to
        conform RFC 3164.
        https://www.ietf.org/rfc/rfc3164.txt

        The format is `<PRI>TIMESTAMP HOSTNAME PROCESS_ID: MSG`.

        <PRI> field is not included here as it is added later by the
        Python stdlib SyslogLogHandler.

        If the resulting text is bigger than 1020 bytes, it is truncated,
        as it is specified by the RFC.

        Process ID comes from the server name configuration.
        """
        header = '%s %s %s ' % (
            time.strftime(
                '%b %d',
                time.localtime(record.event.created)
                ).replace(' 0', '  '),  # Removes zero-padding (see RFC 3164)
            time.strftime(
                '%H:%M:%S',
                time.localtime(record.event.created)
                ),
            self._hostname,
            )
        msg = '%s: %s %s %s %s' % (
            self._process_name,
            record.event.id,
            EventLogRecordFormatterAdapter.formatAvatar(record.event.account),
            EventLogRecordFormatterAdapter.formatPeer(
                record.event.account.peer
                ),
            record.event.message,
            )
        return (header + msg)[:1020]

    def open(self):
        """
        Called to open the stream.
        """
        self._connecting = True

        def cb_open_done(result):
            self._connecting = False
            self._connected = True

        def eb_open(failure):
            self._connecting = False
            return failure

        def bb_open(result_or_failure):
            self._connect_deferred = None
            return result_or_failure

        self._connect_deferred = defer.maybeDeferred(self._open)
        self._connect_deferred.addCallback(cb_open_done)
        self._connect_deferred.addErrback(eb_open)
        # The deferred might be already called and in this case the following
        # call will set it to None and we will have nothing to return.
        result = self._connect_deferred
        self._connect_deferred.addBoth(bb_open)
        return result

    def _open(self):
        """
        Low level code handling the socket opening.
        """
        if isinstance(self.address, six.string_types):
            self.unixsocket = 1
            self._connect_unixsocket(self.address)
            # Unix socket should always be created without blocking.
            return

        self.unixsocket = 0
        if self.socktype is None:
            self.socktype = socket.SOCK_DGRAM

        self.socket = socket.socket(socket.AF_INET, self.socktype)

        if self.socktype == socket.SOCK_STREAM:
            # Connect might block if address fails to resolve or if the
            # connection to address fails to be established.
            return threads.deferToThread(self.socket.connect, self.address)
        else:
            # No need to wait for non-stream connections.
            return

    def close(self):
        """
        See: `ILogHandler`.

        This is a patch from Python 3.3 as 2.7 was only closing Unix sockets.
        """
        self._connected = False
        if not self.socket:
            return
        try:
            self.socket.close()
            Handler.close(self)
        finally:
            self.socket = None

    def handleError(self, record):
        """
        Will just raise the error so that it will be handled by
        the log handler adapter.
        """
        # FIXME:2975:
        # This should be re-implemented with a sound error handling.
        def reconnect():
            deferred = self.open()
            # Ignore if we are not able to reconnected.
            deferred.addErrback(lambda _: None)

        if self._connecting:
            # Ignore any errors received while we connect.
            return

        if not self._connected:
            # We are not connected, so just re-raise the error and retry
            # to connect.
            reconnect()
            raise

        # We are connected. See if we should try to reconnect.
        # On py2 we get socket.error
        # On py3 we get BrokenPipeError we use OSError to get py2 compat.
        error_class, error, traceback = sys.exc_info()
        if isinstance(error, (socket.error, OSError)):
            self._connected = False
            reconnect()

        # Bubble the error.
        raise


class SyslogEventHandler(EventHandlerWithLogger):
    """
    Sends events to a Syslog server.
    """
    PROPERTIES_REQUIRING_RESTART = ('url',)

    # Here to help with testing. Holds the address for the syslog as
    # parsed from the configuration.
    _server_address = None

    def _getLogHandler(self):
        """
        See: `EventHandlerWithLogger`.
        """
        # FIXME:2975:
        # For now it is implemented on the cheap by reusing the Python stdlib.
        socket_type, self._server_address = self._parseConfiguration()
        handler = SyslogLogHandler(
            address=self._server_address,
            socktype=socket_type,
            # FIXME:2228:
            # Facility and priority are for now hard coded.
            facility=PythonSyslogLogHandler.LOG_DAEMON,
            process_name=self.root.configuration.server.name,
            )
        return handler

    def _parseConfiguration(self):
        """
        Return a server:port tuple or path Unix socket to
        configured Syslog server.
        """
        address = None
        socket_type = socket.SOCK_DGRAM
        raw_url = self._configuration.url
        url = urlparse(raw_url)
        scheme = url.scheme
        if scheme == 'file':
            if process_capabilities.os_family == 'nt':
                raise ServerException(
                    'The "file://PATH" value is not supported '
                    'for the "url" configuration.'
                    )
            address = local_filesystem.getEncodedPath(url.path)
        elif scheme == 'tcp':
            socket_type = socket.SOCK_STREAM
            address = _split_address(url.netloc, default_port=601)
        elif scheme == 'udp':
            socket_type = socket.SOCK_DGRAM
            address = _split_address(url.netloc, default_port=514)
        else:
            raise ServerException(
                'Invalid URL in configuration: "%s".' % (raw_url,))

        return (socket_type, address)


def _split_address(address, default_port):
    """
    Return a tuple of (host, port) from `address`.
    """
    parts = address.split(':', 1)
    if len(parts) < 2:
        return (address, default_port)

    try:
        port = int(parts[1])
    except ValueError:
        raise ServerException('Invalid port definition in URL configuration.')

    return (parts[0], port)
