# Copyright (c) 2019 Adi Roiban.
# See LICENSE for details.
"""
Extensions modifying the content of a file.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
import os
from shutil import copyfile

from chevah_compat import local_filesystem

from chevah.server import force_unicode, json
from chevah.server.commons.exception import ServerException


class RemoveLastLine(object):
    """
    Remove the last line of the source file and create a copy in a configurable
    destination path.

    Configuration defined as JSON string.
    {
        "destination_path": "/destination/directory",
        "source_attribute": "real_path"
    }
    """

    _CHUNK_SIZE = 16000

    def onStart(self, parent):
        """
        Does nothing.
        """
        self._configuration = parent.configuration

    def onStop(self):
        """
        Does nothing.
        """
        self._configuration = None

    def getConfiguration(self, event):
        """
        Does nothing.
        """
        return self._configuration

    @staticmethod
    def handle(event, configuration):
        """
        See: EventHandlerWithFilterMixin.
        """
        configuration = json.loads(configuration)
        destination_base = configuration['destination_path']
        if not destination_base:
            raise ServerException('No destination path was defined.')

        source_path, _, source_name = _get_source_path(
            event, configuration['source_attribute'])

        destination_path = os.path.join(destination_base, source_name)

        data = {
            'source_path': source_path,
            'path': destination_path,
            'real_path': destination_path,
            }

        try:
            _amend_file(source_path, destination_path)
            return [('20087', data)]
        except Exception as error:
            data['details'] = force_unicode(error)
            return [('20088', data)]


def _get_source_path(event, source_attribute):
    """
    Return a tuple of (source_path, source_parent, source_name)

    Raise an exception if path for the event is not a file.
    """
    if not source_attribute:
        raise ServerException('No source attribute was defined.')

    source_path = event.data.get(source_attribute.lower(), None)
    if not source_path:
        raise ServerException(
            'Handler configured for an event which is missing the '
            'source attribute.'
            )

    source_parent, source_name = os.path.split(source_path)
    if not source_name:
        raise ServerException(
            'Event does not has an associated file path.')

    return source_path, source_parent, source_name


def _amend_file(source_path, destination_path):
    """
    Do the actual file changes.
    """
    source_path = local_filesystem.getEncodedPath(source_path)
    destination_path = local_filesystem.getEncodedPath(destination_path)

    # We start by copying the file to destination.
    # This should validate the source and destination.
    if os.path.exists(destination_path):
        raise ServerException('Destination already exists.')

    copyfile(source_path, destination_path)

    # Search the last line.
    with open(destination_path, 'rb') as stream:
        stream.seek(0, os.SEEK_END)
        size = stream.tell()

        # Read each chunk in the file, backwards
        # We search for a newline character.
        # If we find a new line, we look for a companion '\r' and then
        # exit the search.
        cursor = size
        while cursor > 0:
            cursor = max(0, cursor - RemoveLastLine._CHUNK_SIZE)
            stream.seek(cursor, os.SEEK_SET)
            chunk = stream.read(RemoveLastLine._CHUNK_SIZE)

            newline_pos = chunk.rfind(b'\n')
            if newline_pos == -1:
                # Read another chunk
                continue

            # Continue the process,
            # as we found a new line.

            if newline_pos == 0 and cursor > 0:
                # At the start of the chunk.
                # Read another character to see if we have a Windows
                # new line.
                stream.seek(cursor - 1, os.SEEK_SET)
                previous_char = stream.read(1)
                if previous_char == b'\r':
                    cursor = cursor - 1
            else:
                # After the start of the chunk.
                if (
                    newline_pos > 0
                    and chunk[newline_pos - 1: newline_pos] == b'\r'
                        ):
                    # We have a `\r\n` so new line is one char before.
                    newline_pos = newline_pos - 1

                cursor = cursor + newline_pos

            # We don't need to look further.
            cursor = cursor + 1
            break

        if cursor == 0:
            # We are at the start, so the file has no new lines.
            return

    with open(destination_path, 'ab') as stream:
        # Truncate the file before the last newline marker.
        stream.seek(cursor - 1, os.SEEK_SET)
        stream.truncate()
