# Copyright (c) 2016 Adi Roiban.
# See LICENSE for details.
"""
Event handler for digitally signed files.
"""
from __future__ import unicode_literals

from __future__ import absolute_import
from base64 import b64decode
import calendar
import dateutil.parser

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric.utils import Prehashed
from OpenSSL import crypto

from twisted.internet import defer, reactor

from chevah_compat import local_filesystem
from chevah.server import force_unicode, json
from chevah.server.commons.exception import ServerException
from chevah.server.commons.line_processor import get_normalized_lines

# Cached used by the workers to keep verifying protocols.
_PUBLIC_KEY_CACHE = {}


class ValidateCSV_RSASSA_PSS(object):
    """
    Validate signature of a CSV file.

    RFC 3447 - RSASSA-PSS

    Configuration defined as JSON string.
    {
        "signer_certificate_path": "/path/to/certificate",
        "ssl_certificate_authority": ["/path/to/CA/certificate"],
        "ssl_certificate_revocation_list": ["REVACATION-LIST-CONFIGURATION"],
        "ssl_certificate_revocation_list_refresh": 0,
    }

    The `extension can be configured to check if files have a valid signature.

    For now, it only handles signatures designed for comma separated values
    text files.
    The signature is distributed as trailing content separated by comma.

    The signed file is in the format::

        SIGNED_CONTENT,SIGNATURE

    It uses the SHA-256 hash function.

    The signature validation process is based on the
    RSA Digital Signature Algorithm PKCS#1 v2.1 also known as RSASSA-PSS and
    documented in `RFC 3447 <https://tools.ietf.org/html/rfc3447>`_.

    New lines in the ``SIGNED_CONTENT`` are normalized to Unix new
    lines ``LF / \n / 0xA``.

    If you have a different signature distribution method, please get in
    touch with us.

    When the configured certificate has expired or was revoked the handler will
    continue to operate and will consider all files as having an invalid
    signature.

    Dedicated events are emitted for valid and invalid signatures.

    ..  note::
        The Digital Signature Validation event handler will only check files
        with lines smaller than 16,000 characters.
        If the file contains longer lines, it will not be validated.
        This is done to prevent handling accidental binary files.

    """

    _scheduler = reactor

    def __init__(self):
        # Public key of the validation certificate.
        self._public_pem = None
        # CRL validation object used by this handler.
        self._crl_validator = None
        # Human readable representation of the certificate used to validate the
        # signature.
        self._signer_marker = 'not-loaded-yet'
        # Configuration for this extension.
        self._options = {}
        self._parent = None

    def onStart(self, parent):
        """
        Called when the extension starts.
        """
        self._parent = parent
        self._options = json.loads(parent.configuration)
        certificate_path = self._options['signer_certificate_path']

        if not certificate_path:
            return defer.fail(ServerException(
                'You need to configure the certificate used for validation.'
                ))

        certificate_path = self._options['signer_certificate_path']
        segments = local_filesystem.getSegmentsFromRealPath(certificate_path)

        certificate_stream = None
        try:
            certificate_stream = local_filesystem.openFileForReading(segments)
            certificate_content = certificate_stream.read()
        except EnvironmentError:
            return defer.fail(ServerException(
                'Failed to read the certificate.'))
        finally:
            if certificate_stream:
                certificate_stream.close()

        try:
            certificate = crypto.load_certificate(
                crypto.FILETYPE_PEM, certificate_content)

            public_key = certificate.get_pubkey()
            if public_key.type() != crypto.TYPE_RSA:
                return defer.fail(ServerException(
                    'Only RSA keys are supported.'))

            self._public_pem = public_key.to_cryptography_key().public_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PublicFormat.SubjectPublicKeyInfo
                )

            # Data appended to each content before signature validation.
            issuer_components = []
            for i_key, i_value in certificate.get_issuer().get_components():
                issuer_components.append(
                    '%s=%s' % (force_unicode(i_key), force_unicode(i_value)))
            subject_components = []
            for i_key, i_value in certificate.get_subject().get_components():
                subject_components.append(
                    '%s=%s' % (force_unicode(i_key), force_unicode(i_value)))
            certificate_details = 'Issuer:%s Subject:%s Serial:%s' % (
                ','.join(issuer_components),
                ','.join(subject_components),
                certificate.get_serial_number(),
                )
        except Exception as error:
            return defer.fail(ServerException(
                'Failed to parse the certificate. %s' % (error,)))

        not_before = certificate.get_notBefore()
        not_after = certificate.get_notAfter()
        if not not_before or not not_after:
            return defer.fail(ServerException(
                'Certificate without validity interval is not supported.'))

        try:
            not_before = dateutil.parser.parse(not_before)
            not_after = dateutil.parser.parse(not_after)
        except ValueError:
            return defer.fail(ServerException(
                'Certificate validity dates with invalid format.'))

        not_before = calendar.timegm(not_before.utctimetuple())
        not_after = calendar.timegm(not_after.utctimetuple())
        now = self._scheduler.seconds()
        if not_before > now:
            return defer.fail(
                ServerException('Certificate is not yet valid.'))
        if not_after < now:
            return defer.fail(ServerException('Certificate is expired.'))

        self._not_after = not_after
        self._signer_marker = certificate_details.encode('utf-8')
        self._certificate = certificate

    def onStop(self):
        """
        See: `Runnable`.
        """
        if self._crl_validator:
            self._crl_validator.stop()
        self._crl_validator = None

        self._public_pem = None
        self._certificate = None
        self._signer_marker = None

    def getConfiguration(self, event):
        """
        Return the ID and PEM of the certificate.
        """
        if self._crl_validator:
            try:
                self._crl_validator.validate(
                    peer_certificate=self._certificate)
            except ServerException as error:
                # Certificate is now revoked.
                raise ServerException('Certificate is no longer valid. %s' % (
                    force_unicode(error),))

        if self._not_after < self._scheduler.seconds():
            # Certificate is no longer valid.
            raise ServerException('Certificate is now  expired.')

        real_path = event.data.get('real_path', None)
        if not real_path:
            real_path = event.data.get('path', None)
        if not real_path:
            # No path. Nothing to do.
            raise ServerException(
                'Handler configured for an event which has no path.')

        return (self._signer_marker, self._public_pem, real_path)

    @staticmethod
    def handle(event, configuration):
        """
        See: EventHandlerWithFilterMixin.

        Called in the worker process.
        """
        signer_marker, public_key, real_path = configuration

        key = _PUBLIC_KEY_CACHE.get(signer_marker, None)
        if not key:
            key = serialization.load_pem_public_key(public_key)
            _PUBLIC_KEY_CACHE[signer_marker] = key

        # Digital signature is only handling real paths for now.
        data = {
            'path': real_path,
            'real_path': real_path,
            }

        try:
            _check_file_signature(real_path, key, signer_marker)
            return [('20173', data)]
        except Exception as error:
            data['details'] = force_unicode(error)
            return [('20177', data)]


def _check_file_signature(path, key, signer_marker):
    """
    Check that file at path has a valid signature.

    This should be called from a thread as it can take a long time to
    read the file and to compute the signature.

    Raise `ServerException` if signature can not be validated.

    Called in the worker process.
    """
    segments = local_filesystem.getSegmentsFromRealPath(path)
    hash_sha256 = hashes.SHA256()
    sha256_hasher = hashes.Hash(hash_sha256)
    previous_line = None
    data_stream = None
    try:
        data_stream = local_filesystem.openFileForReading(segments)
        for line in get_normalized_lines(data_stream):
            if not previous_line:
                # First iteration.
                previous_line = line
                continue

            sha256_hasher.update(previous_line)
            previous_line = line

    except EnvironmentError:
        raise ServerException('Fail to read the source file.')

    finally:
        if data_stream:
            data_stream.close()

    if not previous_line:
        raise ServerException('File looks empty.')

    # Whole file was read.
    last_line = line.rstrip(b'\n')

    try:
        last_data, signature = last_line.rsplit(b',', 1)
    except ValueError:
        raise ServerException('No signature found on the last line.')

    try:
        signature = b64decode(signature)
    except Exception:
        raise ServerException('Signature with invalid encoding.')

    sha256_hasher.update(last_data)

    digest = sha256_hasher.finalize()
    try:
        return key.verify(
            signature,
            digest,
            padding.PSS(
                mgf=padding.MGF1(hashes.SHA256()),
                salt_length=hash_sha256.digest_size,
                ),
            Prehashed(hash_sha256)
            )
    except InvalidSignature:
        raise ServerException(
            'Invalid signature. Validating using %s' % (
                signer_marker.decode('ascii'),))
