# -*- coding: UTF_8 -*-
# Keep the above first line as it is.
#
# Place this file in `extension/file_mover.py`
# inside SFTPPlus installation folder.
#
# This is designed to be used inside OpenShift cluster, on Linux,
# with a specific folder structure.
# The destination base path for each user is required to be created
# outside of this script.
#
# For the event handler set
# type = extension
# entry_point = python:file_mover.FileMover
#
# Emitted events:
# * 20174 - for critical errors
# * 20200 - for normal log messages
#
# This is designed to work on Linux.
#
# You can run the automated tests for this extension using
# $ cd SFTPPLUS_INSTALL_PATH
# $ LANG=C ./bin/python extension/file_mover.py -v
#
# To run a single test
# $ ./bin/python extension/file_mover.py TestFileMover.test_handle_ok
#
# When tests are failing you might need to clean up the test data:
# $ rm -rf /tmp/test_source/ ; rm -rf /tmp/test_destination/
#
from __future__ import unicode_literals
from datetime import datetime
import json
import os
import platform
import fnmatch
import shutil
import unittest

from twisted.internet import defer, reactor
from twisted.internet.task import Clock, deferLater

REQUIRED_SFTPPLUS_VERSION = (5, 0, 0)
EXTENSION_VERSION = 7
# Changelog
#
# Version 7 - 2024-04-05 (adi.roiban@proatria.com)
#  When the extension starts, add a log with current extension version.
# Version 6 - 2024-03-19 (adi.roiban@proatria.com)
#  Update for SFTPPlus 5.0.0
# Version 5 - 2023-09-26 (adi.roiban@proatria.com)
#  Export source and destination path to the successful event.
# Version 4 - 2023-09-26 (adi.roiban@proatria.com)
#  Add support for keeping the file extension for PGP/GPG files.
# Version 3 - 2022-06-03
#  Add support for UTF-8 encoded unicode file names.
# Version 2
#  Add milliseconds to timestamp.
# Version 1
#  Initial implementation
#

# Delay when no explicit configuration is defined.
DEFAULT_DELAY = 5
# Base path where to move the files when no explicit configuration is defined.
DEFAULT_TARGET_ROOT = '/tmp/'

# Permissions to be set on the destination file.
DESTINATION_PERMISSIONS = 0o664

# Message emitted when a file is ignored.
MSG_INFO_SOURCE_IGNORED = '[INFO] {source} ignored.'
# Message emitted as debug when a file is scheduled to be processed.
MSG_DEBUG_START = (
    '[DEBUG] Handling will start for {source} after {delay} seconds.')
# Message emitted when a file is uploaded outside of the USER/in/ directory.
MSG_WARN_NOT_USER_IN_FILE = '[WARN] {source} not an $USER/in/ file.'
# Message emitted if the source file no longer exists by the time
# the file is processed.
MSG_WARN_SOURCE_MISSING = (
    '[WARN] Source no longer exists after delay: {source}')
# Message emitted when the destination directory doesn't exists.
MSG_WARN_NO_TARGET = '[WARN] No TARGET: {path}'
# Message emitted when failing to do the move operation.
MSG_WARN_FAIL_MOVE = (
    '[WARN] Fail to move {source} to {destination}. {error}')
# Message emitted when the move was successful.
MSG_INFO_OK = '[INFO] {source} moved to {destination}'

# List of glob expression for files that are ignored when no explicit
# configuration is defined.
DEFAULT_IGNORE_PATHS = (
    '*.lock',
    '*.partial',
    '*.filepart',
    '*.tmp',
    '*.TEMP',
    '*.TMP',

    )

# List of events that can be handled, and the base service directory
# for each event.
KNOWN_EVENTS = {
    '30017': 'ssh-ftp',
    '30026': 'ssh-ftp',
    '30068': 'ssh-ftp',
    '10075': 'ftp',
    '10078': 'ftp',
    }


###############################################################################
# Implementation code below
###############################################################################

class FileMover(object):
    """
    An event handler designed to move the source files into destination
    based on very specific rules.
    """
    # This is here to help with automated tests and allow faking time passing.
    _scheduler = reactor

    def __init__(self):
        """
        Called when the event handler is initialized.
        """
        self._configuration = None
        self._delay = 0
        self._debug = False
        self._ignore_paths = []
        self._target_root = ''

    def onStart(self, parent):
        """
        Called in the main process when the event handler starts.

        `parent.configuration` is the Unicode string defined in the
        extension configuration option.
        """
        self._parent = parent
        self._configuration = parent.configuration

        # Load JSON configuration.
        if not self._configuration:
            # If no configuration defined, init an empty JSON that
            # can be parsed.
            self._configuration = '{}'

        try:
            json_config = json.loads(self._configuration)
        except Exception as error:
            raise RuntimeError(
                'Failed to load the configuration. {error}'.format(
                    error=error))

        self._delay = int(json_config.get('delay', DEFAULT_DELAY))
        self._debug = bool(json_config.get('debug', False))
        self._ignore_paths = json_config.get(
            'ignore_paths', DEFAULT_IGNORE_PATHS)
        self._target_root = json_config.get('target_root', DEFAULT_TARGET_ROOT)

        # Normalize the configuration to be passed to workers as plain JSON
        self._configuration = json.dumps({
            'debug': self._debug,
            'base': self._target_root,
            },
            indent=-1,
            separators=(',', ':'),
            )
        message = 'File mover extension started. Version {}.'.format(
            EXTENSION_VERSION)
        self._parent.emitEvent('20200', data={'message': message})

    def onStop(self):
        """
        Called in the main process when the event handler stops.
        """
        self._configuration = None
        self._debug = False

    def getConfiguration(self, event):
        """
        Called to validate the event.

        Return the configuration for this event as a string.
        Return `None` if the event handling should be skipped/ignored.
        """
        if event.id not in KNOWN_EVENTS.keys():
            raise RuntimeError("Extension can't handle this event.")

        user = event.account.name
        source = event.data['real_path']

        # See if we have an ignored file.
        for glob in self._ignore_paths:
            if fnmatch.fnmatch(source.lower(), glob.lower()):
                # Always inform about ignored files,
                # event when debug is disabled.
                message = MSG_INFO_SOURCE_IGNORED.format(source=source)
                self._parent.emitEvent('20200', data={'message': message})
                # Return `None` to inform the API that we don't want to
                # handle this event.
                return None

        # Destination is any subdirectory of /SOME_PATH/USER/in/
        parts = source.split('/{}/in/'.format(user), 1)

        if len(parts) != 2:
            message = MSG_WARN_NOT_USER_IN_FILE.format(
                source=source)
            self._parent.emitEvent('20200', data={'message': message})
            return None

        if self._debug:
            # Ask the handler to emit an event at the end.
            self._parent.emitEvent(
                '20200',
                data={'message': MSG_DEBUG_START.format(
                    source=event.data['real_path'],
                    delay=self._delay,
                    )})

        if self._delay <= 0:
            return self._configuration

        return deferLater(
            self._scheduler, self._delay, self._checkSourceProcess, source)

    def _checkSourceProcess(self, source):
        """
        See if source should still be processed.
        """
        if not os.path.exists(encode_path(source)):
            # No need to handle the file as source no longer exists.
            message = MSG_WARN_SOURCE_MISSING.format(source=source)
            self._parent.emitEvent('20200', {'message': message})
            return None

        return self._configuration

    @staticmethod
    def handle(event, configuration, _clock=reactor):
        """
        Called when we should handle the event.
        """
        # Extract info from configuration.
        configuration = json.loads(configuration)
        target_root = configuration['base']
        # Extract info from event.
        source = event.data['real_path']
        user = event.account.name

        # Destination is a subdirectory of /SOME_PATH/USER/in/
        parts = source.split('/{}/in/'.format(user), 1)
        destination_subpath, filename = os.path.split(parts[1])

        # Any subpath separator is replaced with hyphen.
        # And any space with underscore.
        destination_subpath = destination_subpath.replace('/', '-')
        destination_subpath = destination_subpath.replace(' ', '_')

        # Any space in the filename is replaced with underscore.
        filename = filename.replace(' ', '_')

        basename, extension = os.path.splitext(filename)
        have_pgp = False
        if extension in ['.pgp', '.gpg']:
            have_pgp = True
            # We have a PGP file, so extension is composite.
            basename, original_extension = os.path.splitext(basename)
            extension = original_extension + extension

        # Generate timestamp in modified local ISO8601 format
        # Use datetime.utcfromtimestamp() for UTC
        timestamp = datetime.fromtimestamp(_clock.seconds()).isoformat()
        timestamp = timestamp.replace(':', '')
        timestamp = timestamp.replace('-', '')
        timestamp = timestamp.replace('T', '.')

        # Insert timestamp before the extension
        filename = "{filename}_{timestamp}".format(
            filename=filename,
            timestamp=timestamp,
            )
        if have_pgp:
            filename = "{basename}_{timestamp}{extension}".format(
                basename=basename,
                timestamp=timestamp,
                extension=extension,
                )

        service_directory = KNOWN_EVENTS[event.id]

        destination_parent = os.path.join(
            target_root,
            service_directory,
            user,
            destination_subpath,
            )

        full_destination = os.path.join(
            destination_parent,
            filename,
            )

        if not os.path.exists(encode_path(destination_parent)):
            # A single text can be returned and is used to emit an event.
            return MSG_WARN_NO_TARGET.format(path=destination_parent)

        try:
            # Do the renaming, or a copy+delete if destination is on a separate
            # filesystem.
            # For security reason we also set the source permissions before
            # the move.
            os.chmod(encode_path(source), DESTINATION_PERMISSIONS)
            shutil.move(encode_path(source), encode_path(full_destination))
            os.chmod(encode_path(full_destination), DESTINATION_PERMISSIONS)
        except Exception as error:
            return MSG_WARN_FAIL_MOVE.format(
                source=source,
                destination=full_destination,
                error=str(error),
                )

        return [{
            'message': MSG_INFO_OK.format(
                source=source, destination=full_destination),
            'source': source,
            'destination': full_destination,
            'real_path': full_destination,
            }]


def encode_path(path):
    """
    Encode path to be handled by the OS.
    """
    return path.encode('utf-8')


###############################################################################
# Testing code below
###############################################################################


class DummyParent(object):
    """
    A parent for the event handler that has
    a configuration and will record the emitted events.

    Default configuration has no delay.
    """
    def __init__(self, configuration='{"delay": 0}'):
        self.configuration = configuration
        self.events = []

    def emitEvent(self, event_id, data):
        self.events.append([event_id, data])


class DummyAccount(object):
    """
    A simple account representation to be used in tests.
    """
    def __init__(self, name):
        self.name = name


class DummyEvent(object):
    """
    A simple event used during testing that is associated with a path.
    """
    def __init__(
        self, event_id, path='/tmp/some/file.txt', username='some-user',
            ):
        self.id = event_id
        self.account = DummyAccount(name=username)
        self.data = {
            'real_path': path,
            }


# The test contain many similar checks.
# The duplicate checks are not refactored into a common function to make
# the code easier to read.
# This is why the copy/paste detection is disabled.
# jscpd:ignore-start
@unittest.skipIf(platform.system() != 'Linux', 'Only on Linux.')
class TestFileMover(unittest.TestCase):
    """
    Unit tests for the file mover script.

    All the tests will try to move files inside '/tmp' folder.

    * onStart is called once at startup
    * getConfiguration is called from the main thread before each event
    * handle is called from sub-process for each event that is not ignored
    """
    def assertEvents(self, events, parent):
        """
        Check that events were emitted.
        """
        expected = []
        for event in events:
            expected.append(['20200', {'message': event}])
        self.assertEqual(expected, parent.events)
        parent.events = []

    def assertNoResult(self, deferred):
        """
        Raise an exception if deferred has a result.
        """
        if not hasattr(deferred, 'result'):
            return

        if isinstance(deferred.result, defer.Deferred):
            return

        raise AssertionError(
            'Deferred has a result: {}'.format(deferred.result))

    def getStartedExtension(self, parent):
        """
        Return an extension that is already started and has a rigged clock
        that we can use to fake time passing.
        """
        sut = FileMover()
        sut._scheduler = Clock()
        sut.onStart(parent)

        self.assertEvents([
            'File mover extension started. Version {}.'.format(
                EXTENSION_VERSION)
            ],
            parent)

        return sut

    def test_start_empty_configuration(self):
        """
        It can be started with an empty configuration and will have
        default configuration.
        """
        parent = DummyParent(configuration='')
        sut = FileMover()

        sut.onStart(parent)

        # Debug is disabled by default.
        self.assertFalse(sut._debug)
        # There is a delay of 5 seconds by default.
        self.assertEqual(5, sut._delay)
        # Files are moved to TMP by default.
        self.assertEqual('/tmp/', sut._target_root)

    def test_start_invalid_configuration(self):
        """
        It fails to start with a configuration that is not a valid JSON format.
        """
        parent = DummyParent(configuration='not-json format')
        sut = FileMover()

        with self.assertRaises(RuntimeError) as context:
            sut.onStart(parent)

        self.assertEqual(
            'Failed to load the configuration. '
            'Expecting value: line 1 column 1 (char 0)',
            context.exception.args[0])

    def test_getConfiguration_unknown_event_id(self):
        """
        It fails if called for an event with an ID
        that is not supported by the handler.
        """
        parent = DummyParent()
        event = DummyEvent('1245')
        sut = FileMover()
        sut.onStart(parent)

        with self.assertRaises(RuntimeError) as context:
            sut.getConfiguration(event)

        self.assertEqual(
            "Extension can't handle this event.",
            context.exception.args[0])

    def test_getConfiguration_ignored_path(self):
        """
        It will skip handling, and emit an event for the ignored files
        when debug is enabled.

        This is also a test to check the configuration for "ignore_paths",
        that will match the source path in a case-insensitive way.
        """
        parent = DummyParent("""{
"debug": true,
"ignore_paths": ["*.other", "*.txt"]
        }""")
        event = DummyEvent('10075', '/some/user/file.TXT')
        sut = self.getStartedExtension(parent)

        result = sut.getConfiguration(event)

        self.assertIsNone(result)
        self.assertEvents(['[INFO] /some/user/file.TXT ignored.'], parent)

        # Unicode names are also handled
        event = DummyEvent('10075', '/some/user/another.other')
        sut = self.getStartedExtension(parent)
        result = sut.getConfiguration(event)
        self.assertIsNone(result)
        self.assertEvents(['[INFO] /some/user/another.other ignored.'], parent)

    def test_getConfiguration_not_user_path(self):
        """
        It will emit an event and not move the path if the source path
        is not matching the */USERNAME/in/* pattern.

        This is also a test to check that event 30068 is accepted.
        """
        parent = DummyParent()
        event = DummyEvent('30068', '/some/not-ignored.pdf')
        sut = self.getStartedExtension(parent)

        result = sut.getConfiguration(event)

        self.assertIsNone(result)
        self.assertEvents(
            ['[WARN] /some/not-ignored.pdf not an $USER/in/ file.'],
            parent)

    def test_getConfiguration_delay_execution_check(self):
        """
        It will delay returning the configuration for the event.

        If after the delay, the source file no longer exists, the processing
        is skipped.

        This is also a test to check the configuration for "delay"
        and the "debug" info for starting to execute the event for path.
        """
        parent = DummyParent("""{
"delay": 3,
"debug": true,
"ignore_paths": ["some-other-glob"]
        }""")
        event = DummyEvent('10078', '/some/some-user/in/not-ignored.pdf')
        sut = self.getStartedExtension(parent)

        deferred = sut.getConfiguration(event)

        # No configuration returned yet.
        self.assertNoResult(deferred)
        # After 1 second, still no returned.
        sut._scheduler.advance(1)
        self.assertNoResult(deferred)

        # The result is available after total of 3 seconds and will
        # trigger the processing of the event.
        sut._scheduler.advance(2)
        self.assertIsNone(deferred.result)

        # A debug event is emitted
        self.assertEvents([
            '[DEBUG] Handling will start for '
            '/some/some-user/in/not-ignored.pdf after 3 seconds.',
            '[WARN] Source no longer exists after delay: '
            '/some/some-user/in/not-ignored.pdf',
            ], parent)

    def test_handle_not_destination_parent(self):
        """
        It returns an error message and not move the path if the base
        destination doesn't exist.

        This is also a test to check that event 10078 is moved into the
        destination `ftp` directory.

        And this checks that parent path separators are replaced with '-'
        and spaces in path are replaced with '_'.
        """
        parent = DummyParent()
        event = DummyEvent(
            '10078',
            path='/source/johnD/in/base path/sub-folder/other/text file.txt',
            username='johnD',
            )
        sut = self.getStartedExtension(parent)
        configuration = sut.getConfiguration(event)

        result = sut.handle(event, configuration)

        # Files is not handled and an error message returned to be emitted.
        self.assertEqual(
            '[WARN] No TARGET: /tmp/ftp/johnD/base_path-sub-folder-other',
            result)

    def test_handle_fail_missing_source(self):
        """
        It returns an error message when source doesn't exist

        This is also a test to check that event 30026 got files moved into the
        destination `ssh-ftp` directory.
        """
        os.makedirs('/tmp/ssh-ftp/johnD/test_missing_source'.encode('utf-8'))
        self.addCleanup(os.rmdir, '/tmp/ssh-ftp')
        self.addCleanup(os.rmdir, '/tmp/ssh-ftp/johnD')
        self.addCleanup(
            os.rmdir, '/tmp/ssh-ftp/johnD/test_missing_source'.encode('utf-8'))

        parent = DummyParent()
        event = DummyEvent(
            '30026',
            path='/source/johnD/in/test_missing_source/text file.txt',
            username='johnD',
            )
        sut = self.getStartedExtension(parent)
        configuration = sut.getConfiguration(event)

        result = sut.handle(event, configuration)

        # The source and destination paths are found in the message.
        # We check chunks as the destination filename contains a timestamp.
        self.assertTrue(result.startswith(
            '[WARN] Fail to move /source/johnD/in/test_missing_source/'
            'text file.txt to /tmp/ssh-ftp/johnD/test_missing_source/'
            'text_file.txt_20'),
            result)
        # The low level OS error is present in the message.
        self.assertIn('[Errno 2] No such file or directory', result)

    def test_handle_ok(self):
        """
        It will move the source path to the destination with a timestamp,
        that is added at the end, without preserving the original file
        extension.

        sample.txt -> sample.txt_TIMESTAMP

        The source and destination path are made available to the emitted
        event, and can be used by other events or structured logs.

        This is also a test to check that event 30017 got files moved into the
        destination `ssh-ftp` directory.
        """
        source_file = '/tmp/test_source/johnD/in/text file.txt'
        destination_file = (
            '/tmp/test_destination/ssh-ftp/'
            'johnD/text_file.txt_19700101.000003'
            )
        # Create source directory and file and setup cleanup.
        self.addCleanup(os.rmdir, '/tmp/test_source')
        self.addCleanup(os.rmdir, '/tmp/test_source/johnD')
        self.addCleanup(os.rmdir, '/tmp/test_source/johnD/in')
        try:
            os.makedirs('/tmp/test_source/johnD/in')
        except Exception:
            "Is OK if the dir exists."

        with open(source_file, 'wb') as stream:
            stream.write(b'some test content')

        # Create destination directory.
        self.addCleanup(os.rmdir, '/tmp/test_destination')
        self.addCleanup(os.rmdir, '/tmp/test_destination/ssh-ftp')
        self.addCleanup(os.rmdir, '/tmp/test_destination/ssh-ftp/johnD')
        self.addCleanup(os.remove, destination_file)
        os.makedirs('/tmp/test_destination/ssh-ftp/johnD/')

        # Check we have source but not destination.
        self.assertTrue(os.path.exists(source_file))
        self.assertFalse(os.path.exists(destination_file))

        parent = DummyParent(
            '{"debug": true, "delay": 3, '
            '"target_root": "/tmp/test_destination"}'
            )
        event = DummyEvent(
            '30026',
            path=source_file,
            username='johnD',
            )
        sut = self.getStartedExtension(parent)
        deferred = sut.getConfiguration(event)

        # No configuration returned yet.
        self.assertNoResult(deferred)
        # After the delay, we got the configuration in a compressed form.
        # Only the configuration values needed for the handler are present.
        sut._scheduler.advance(3)
        configuration = deferred.result
        self.assertEqual(
            '{\n"debug":true,\n"base":"/tmp/test_destination"\n}',
            configuration)

        result = sut.handle(event, configuration, _clock=sut._scheduler)

        message = (
            '[INFO] /tmp/test_source/johnD/in/text file.txt '
            'moved to /tmp/test_destination/ssh-ftp/johnD/'
            'text_file.txt_19700101.000003'
            )
        self.assertEqual([{
            'message': message,
            'source': '/tmp/test_source/johnD/in/text file.txt',
            'destination': (
                '/tmp/test_destination/ssh-ftp/johnD/'
                'text_file.txt_19700101.000003'
                ),
            'real_path': (
                '/tmp/test_destination/ssh-ftp/johnD/'
                'text_file.txt_19700101.000003'
                ),
            }],
            result)

        # Source is moved to destination.
        self.assertFalse(os.path.exists(source_file))
        self.assertTrue(os.path.exists(destination_file))
        dest_stat = os.stat(destination_file)
        # The extra stats are for the file and no sticky or other
        # permission.
        self.assertEqual(0o100664, dest_stat.st_mode)

    def test_handle_ok_pgp(self):
        """
        It will move the PGP source path to the destination with a timestamp,
        while preserving the .pgp extension and also the original,
        non-encrypted file extension.

        sample.txt.pgp -> sample_TIMESTAMP.txt.pgp

        The source and destination path are made available to the emitted
        event, and can be used by other events or structured logs.

        This is also a test to check that event 30017 got files moved into the
        destination `ssh-ftp` directory.
        """
        source_file = '/tmp/test_source/johnD/in/text file.txt.pgp'
        destination_file = (
            '/tmp/test_destination/ssh-ftp/johnD/'
            'text_file_19700101.000003.txt.pgp'
            )
        # Create source directory and file and setup cleanup.
        self.addCleanup(os.rmdir, '/tmp/test_source')
        self.addCleanup(os.rmdir, '/tmp/test_source/johnD')
        self.addCleanup(os.rmdir, '/tmp/test_source/johnD/in')
        try:
            os.makedirs('/tmp/test_source/johnD/in')
        except Exception:
            "Is OK if the dir exists."

        with open(source_file, 'wb') as stream:
            stream.write(b'some test content')

        # Create destination directory.
        self.addCleanup(os.rmdir, '/tmp/test_destination')
        self.addCleanup(os.rmdir, '/tmp/test_destination/ssh-ftp')
        self.addCleanup(os.rmdir, '/tmp/test_destination/ssh-ftp/johnD')
        self.addCleanup(os.remove, destination_file)
        os.makedirs('/tmp/test_destination/ssh-ftp/johnD/')

        # Check we have source but not destination.
        self.assertTrue(os.path.exists(source_file))
        self.assertFalse(os.path.exists(destination_file))

        parent = DummyParent(
            '{"debug": true, "delay": 3, '
            '"target_root": "/tmp/test_destination"}'
            )
        event = DummyEvent(
            '30026',
            path=source_file,
            username='johnD',
            )
        sut = self.getStartedExtension(parent)
        deferred = sut.getConfiguration(event)

        # No configuration returned yet.
        self.assertNoResult(deferred)
        # After the delay, we got the configuration in a compressed form.
        # Only the configuration values needed for the handler are present.
        sut._scheduler.advance(3)
        configuration = deferred.result
        self.assertEqual(
            '{\n"debug":true,\n"base":"/tmp/test_destination"\n}',
            configuration)

        result = sut.handle(event, configuration, _clock=sut._scheduler)

        message = (
            '[INFO] /tmp/test_source/johnD/in/text file.txt.pgp '
            'moved to /tmp/test_destination/ssh-ftp/johnD/'
            'text_file_19700101.000003.txt.pgp'
            )
        self.assertEqual([{
            'message': message,
            'source': '/tmp/test_source/johnD/in/text file.txt.pgp',
            'destination': (
                '/tmp/test_destination/ssh-ftp/johnD/'
                'text_file_19700101.000003.txt.pgp'
                ),
            'real_path': (
                '/tmp/test_destination/ssh-ftp/johnD/'
                'text_file_19700101.000003.txt.pgp'
                ),
            }],
            result)

        # Source is moved to destination.
        self.assertFalse(os.path.exists(source_file))
        self.assertTrue(os.path.exists(destination_file))
        dest_stat = os.stat(destination_file)
        # The extra stats are for the file and no sticky or other
        # permission.
        self.assertEqual(0o100664, dest_stat.st_mode)


# jscpd:ignore-end
# Helper to run the automated tests.
if __name__ == '__main__':
    unittest.main()
