# -*- test-case-name: normal.authentication.test_ldap_extension_mfa -*-
# Copyright (c) 2021 Adi Roiban.
# See LICENSE for details.
"""
Extensions to add custom MFA to SFTPPlus LDAP authentication.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
from collections import deque

from chevah.server import json
from chevah.server.commons.exception import ServerException

from chevah.server.authentication.base import (
    get_TOTP_secret,
    split_TOTP_password,
    verify_TOTP,
    )


class AugmentedTOTP(object):
    """
    Allows augmenting a standard LDAP server with TOTP authentication.

    The LDAP does the normal password authentication and stores the TOTP
    secret.

    SFTPPlus validates that the provided TOTP from the user matches the
    secret stored in LDAP.
    """

    def __init__(self, configuration):
        self._configuration = json.loads(configuration)
        # Mapping from username to the TOTP that was stripped from the
        # password before the LDAP bind.
        self._provided_totps = {}
        # Mapping from username to the last accepted TOTP codes.
        # Required to implement replay protection.
        self._totp_memory = {}

    def stop(self):
        """
        Called when extension is no longer used.
        """

    def updateCredentials(self, credentials):
        """
        Called before BIND to allow mutating the credentials.
        """
        password, provided_totp = split_TOTP_password(credentials.password)

        # The LDAP bind is done using the base password.
        credentials.password = password
        # We save the TOTP for later validation.
        self._provided_totps[credentials.username] = provided_totp

    def getExtraAttributes(self):
        """
        Return the list of extra LDAP attributes to be used during the entry
        search operation.
        """
        return [self._configuration['mfa_attribute']]

    def augmentEntry(self, entry, credentials, ldap_client):
        """
        Called after BIND was successful and we got the LDAP entry for the
        account.
        """
        attribute = entry.get(self._configuration['mfa_attribute'])
        if not attribute:
            raise ServerException(
                'LDAP user has no TOTP shared secret attribute.')

        mfa_url = list(attribute)[0].decode('utf-8')
        secret = get_TOTP_secret(mfa_url)

        provided_totp = self._provided_totps[credentials.username]
        verify_TOTP(
            secret=secret,
            provided_totp=provided_totp,
            memory=self._getTOTPMemory(credentials.username),
            )
        return entry

    def _getTOTPMemory(self, key):
        """
        Return the TOTP memory for `key` LDAP credentials username.
        """
        memory = self._totp_memory.get(key, None)
        if memory is None:
            memory = deque(maxlen=3)
            self._totp_memory[key] = memory
        return memory
