# Copyright (c) 2022 Adi Roiban.
# See LICENSE for details.
"""
Event handler for managing Purview labels.

Version 3

You can manage the labels for an organization from here
https://compliance.microsoft.com/informationprotection?viewid=sensitivitylabels

It will check that an uploaded file is placed into a directory having
the same name as the associated label.

The label is checked case-insensitive.

Dedicated events are emitted for valid and invalid labels.
A separate event is emitted when a label is applied by SFTPPlus.


Release history
===============

* Version 4 - 2024-03-15 - Update for SFTPPlus version 5.1.0
* Version 3 - 2022-09-21 - Fix hardcoded application id.
                           Fix debugging errors.
* Version 2 - 2022-09-08 - The directory_id, client_id, client_secret, and
                           redirect_id configurations are longer needed.
                         - Add debug option.
                         - Update documentation.
* Version 1 - 2022-09-07 - Initial release.


Configuration
=============

On Windows the C++ redistributables need to be installed.
Get them from: https://aka.ms/vs/17/release/vc_redist.x64.exe

This event handler needs to be active for the following events:
* 20137 - login - to get the user details
* 40017 - file uploaded with success

The following Azure API permissions are required and should be configured
as extra `api_scopes` in the Entra ID configuration.
* offline_access
* InformationProtectionPolicy.Read
* https://psor.o365syncservice.com/UnifiedPolicy.User.Read

The following extension configuration is required

configuration: {
  "success_path": "Download",
  "error_path": "Error",
  "debug": "yes",
  "missing_label_action": "auto-label"
  }

Test scenarios.
==============

Test data:
 * An PDF file that is not labeled.
 * An PDF file that is not labeled and has no extension
 * An PDF file that is labeled

After each test, success or error, check that the upload directory is empty.
The uploaded files should always be moved to either Download or Error and
no extra files left in the upload directory.

* Upload the PDF without label and see that it is labeled and moved to the
  success path.
* Upload the same PDF without label twice and see that it is labeled and moved
  to the success path with a timestamp. The existing file is left in place.
* Upload the PDF without extension and see that it is not labeled
  and moved to the success path.
  Upload the same file twice and see that second file has a timestamp.
* Upload the PDF with a label in the expected directory and see that
  it is not modified and moved to the success path.
  Upload the same file twice, and see that the second file is moved with a
  timestamp.
* Upload the PDF with a label in a wrong directory.
  The file is removed from the upload directory.
  A text file is created inside the error path with the error details.

"""
from time import time
from urllib.parse import urlencode
import os
import traceback

from twisted.internet import defer, reactor

from chevah.server import json, force_unicode
from chevah.server.commons.exception import ServerException
from chevah.server.commons.constant import TYPE_NAME
from chevah.server.commons.ssl import SSLClientMixin
from chevah.server.http.client import NonPersistentAgent
from chevah.server.process.spawn import _spawn


# Path to Windows installation.
# Use the know where to search for DLLs.
_WINDOWS_PATH = 'C:\\Windows'

# Text added before the extension by the API for the modified files.
_SDK_MODIFIED_MARKER = '_modified'

# Keep configuration for every authenticated user so that we can have it
# later when a file is uploaded.
# This is only access from the main process.
_ACCOUNTS_CACHE = {}

_LOGIN_EVENT_ID = '40007'
_UPLOAD_EVENT_ID = '40017'


class AzureSSLConfiguration(object):
    """
    A simple SSL configuration that has the Azure CA.
    """
    ssl_certificate = ''
    ssl_key = ''
    ssl_key_password = ''
    ssl_certificate_authority = '${MICROSOFT_IT_CA}'
    ssl_certificate_revocation_list = ''
    ssl_certificate_revocation_list_refresh = 0
    ssl_allowed_methods = ['secure']
    ssl_cipher_list = 'secure'


class PurviewLabeler(SSLClientMixin):
    """
    Check file labels based on purview SDK and auto-apply labels on files
    without a label.
    """

    _scheduler = reactor

    # They are defined on the class to simplify testing.
    _URL_GRAPH_BASE = 'https://graph.microsoft.com'
    _URL_LIST_LABELS = (
        'https://graph.microsoft.com/beta/me/'
        'security/informationProtection/sensitivityLabels'
        )

    def __init__(self):
        self._resetState()

    def _resetState(self):
        """
        Called at init and stop to reset the internal state.
        """
        # Dict with the extension sub-configuration, already parsed from JSON.
        self._configuration = {}
        # This is the dummy component configuration that we used to share
        # SSL code.
        self.configuration = AzureSSLConfiguration()

    @defer.inlineCallbacks
    def onStart(self, parent):
        """
        Called when the extension starts.
        """
        self.emitEvent = parent.emitEvent

        # We only connect to the Graph API but also to login to refresh tokens.
        ssl_context = yield self._initStartedContextFactory(remote_servers=[
            ('login.microsoftonline.com', 443),
            ('graph.microsoft.com', 443),
            ])

        self._agent = NonPersistentAgent(
            event_emitter=self.emitEvent,
            ssl_context=ssl_context.getContext(),
            )

        # Just to a simple test that we can call the labeler.
        self._configuration = json.loads(parent.configuration)
        _labeler(
            ['-h'],
            client_id='not-used-for-start-test', token='not-used',
            )

    def onStop(self):
        """
        See: `Runnable`.
        """
        self._resetState()

    @defer.inlineCallbacks
    def getConfiguration(self, event):
        """
        Called before sending the event to the sub-process.
        """
        if event.id not in [_LOGIN_EVENT_ID, _UPLOAD_EVENT_ID]:
            return

        # The info required to make another OAuth2 token request
        # is available on the account token.
        # Is not only the raw `access_token` but the full info.
        oauth2 = event.account.token
        now = self._scheduler.seconds()

        # Delete any old sessions.
        valid_interval = 3600 * 24
        for cached_value in list(_ACCOUNTS_CACHE.values()):
            if now - cached_value['created'] < valid_interval:
                continue
            del _ACCOUNTS_CACHE[cached_value.uuid]

        if event.id == _LOGIN_EVENT_ID:
            if event.account._context._me.kind_name != TYPE_NAME.ENTRA_ID:
                # Not an Entra ID user.
                return

            labels = yield self._getAvailableLabels(oauth2)
            mip_token = yield self._getMIPToken(oauth2)

            _ACCOUNTS_CACHE[event.account.uuid] = {
                'labels': labels,
                'mip_token': mip_token,
                # Defined here to help with cleanup.
                'uuid': event.account.uuid,
                'created': now,
                }
            # No need to do extra processing of the event.
            return

        # We are left with the file upload event.
        account_config = _ACCOUNTS_CACHE.get(event.account.uuid, {})
        if not account_config:
            raise ServerException('No Purview session found for user.')

        serialized_configuration = json.dumps({
            't': account_config['mip_token'],
            'c': self._configuration,
            'cid': oauth2['client_id'],
            'l': account_config['labels'],
            })
        defer.returnValue(serialized_configuration)

    @staticmethod
    def handle(event, configuration):
        """
        See: EventHandlerWithFilterMixin.

        Called in the worker process.

        This is only called on a HTTP file upload event.
        The pre-call filtering is done in `getConfiguration`.
        """
        debug = ''
        try:
            configuration = json.loads(configuration)
            debug = configuration['c'].get('debug', '')
            return _handle_file_upload(event, configuration)
        except Exception as error:
            # Reporting errors from the sub-process is complicated.
            # so when debug is enables with return the traceback as a normal
            # success message.
            if debug != 'yes':
                raise

            return [{
                'event_id': '20202',
                'message': (
                    'Failed to handle the labeling for the uploaded file.'),
                'details': force_unicode(error),
                'tb': force_unicode(traceback.format_exc()),
                }]

    @defer.inlineCallbacks
    def _getAvailableLabels(self, oauth2_info):
        """
        Return a dict of labels for the current user, indexed by label name.
        """
        # The result is in `value` with a list of labels.
        # {'description': '', 'parent': None, 'isAppliable': True, 'color': '',
        # 'sensitivity': 0, 'tooltip': 'some-text',
        # 'contentFormats': ['file', 'email'], 'hasProtection': False,
        # 'id': '58e16b63-45bb-4cc6-b8f8-21f8bfabbf6b', 'isActive': True,
        # 'name': 'official'}

        # We can use the initial token as it should have access to Graph API.
        token = oauth2_info['authorize']
        authorization = '{} {}'.format(
            token['token_type'],
            token['access_token'],
            )
        headers = {'authorization': authorization}
        response = yield self._agent.get(
            url=self._URL_LIST_LABELS,
            headers=headers,
            )
        body = yield self._agent.readBody(response)
        response_result = json.loads(body)
        result = {}
        for label in response_result['value']:
            result[label['name']] = {
                'id': label['id'],
                }

        if not result:
            raise ServerException('No available labels found for user.')

        defer.returnValue(result)

    @defer.inlineCallbacks
    def _getMIPToken(self, oauth2_info):
        """
        Return a token that can be used for MIP api.
        """
        token = oauth2_info['authorize']
        redirect_uri = oauth2_info['redirect_uri']
        client_id = oauth2_info['client_id']
        client_secret = oauth2_info['client_secret']
        directory_id = oauth2_info['directory_id']

        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'Accept': 'application/json',
            }
        token_params = {
            'client_id': client_id,
            'client_secret': client_secret,
            'refresh_token': token['refresh_token'],
            'grant_type': 'refresh_token',
            'scope': 'https://psor.o365syncservice.com/UnifiedPolicy.User.Read',
            'redirect_uri': redirect_uri,
            }

        token_endpoint = (
            'https://login.microsoftonline.com/{}/oauth2/v2.0/token'
            ).format(directory_id)
        response = yield self._agent.post(
            url=token_endpoint,
            headers=headers,
            body=urlencode(token_params).encode("utf-8")
            )
        body = yield self._agent.readBody(response)
        try:
            result = json.loads(body)
        except Exception:
            raise ServerException(
                'Failed to get a valid access_token response. %r' % (
                    body[:100],))

        error = result.get('error', [''])[0]
        if error:
            details = result.get(
                'error_description',
                'Error received from access_token request.',
                )
            raise ServerException(
                'Error in Entra ID access_token response. '
                '[{}] {}'.format(error, details))

        # Update the current token so that it will have the latest refresh
        # value.
        # Dont't update the access token as the latest access token is only
        # for us.
        # We leave the original in place.
        access_token = result.pop('access_token')
        token.update(result)
        defer.returnValue(access_token)


def _labeler(arguments, client_id, token):
    """
    Call the external labelers.
    """
    # Path to the "file_sample" executable from the MIP SDK.
    # This is the sample/demo file that is provided to demonstrate the File API.
    mip_sdk_file_sample_path = None

    # Start with the common environment for both Linux and Windows.
    environment = {}  # For now common env is empty.

    if os.name == 'nt':
        mip_sdk_file_sample_path = "lib/purview/file_sample.exe"
        environment['SystemRoot'] = _WINDOWS_PATH
    else:
        # Linux debug lib path.
        mip_sdk_file_sample_path = "bins/debug/x86_64/file_sample"
        mip_sdk_library_path = "bins/debug/x86_64/core/"
        environment['LD_LIBRARY_PATH'] = mip_sdk_library_path

    command = [
        mip_sdk_file_sample_path,
        '--scctoken={}'.format(token),
        '--protectiontoken={}'.format(token),
        '--protectionbaseurl=https://api.aadrm.com',
        '--policybaseurl=https://dataservice.protection.outlook.com',
        '--clientid={}'.format(client_id),
        ]
    command.extend(arguments)

    exit_code, out, err = _spawn(
        command=command, timeout=60, environment=environment)

    out = force_unicode(out)
    err = force_unicode(err)

    if exit_code not in [0, -6]:
        error = ServerException(
            'Failure on label helper "{}". {} {} {}'.format(
                arguments, exit_code, out, err))
        error.output = out
        raise error

    return out


def _get_label(source_path, client_id, token):
    """
    Return a dict with label name and label id associated to `source_path`.

    Returns `None` if file has no label.
    """
    out = _labeler(['-g', '-f', source_path], client_id=client_id, token=token)

    if out.startswith('File is labeled'):
        # Try to extract and valiate the label.
        # > File is labeled
        # > File is labeled as: official
        # > Id: 58e16b63-45bb-4cc6-b8f8-21f8bfabbf6b
        # > Set time: 1657124430
        # > Privileged: False
        result = {}
        for line in out.splitlines():
            try:
                key, value = line.split(':', 1)
            except ValueError:
                continue

            key = key.lower()
            value = value.strip()
            if key == 'file is labeled as':
                result['name'] = value
                continue
            if key == 'id':
                result['id'] = value
                continue

        # Return the extracted label info.
        return result

    # No label found.
    return None


def _validate_label(actual, expected, configuration):
    """
    Check that the `actual` label attached to the file, is the `expected`
    label matching the upload directory.

    Raise an error on invalid labels.
    Return the
    """
    if actual['name'].lower() != expected.lower():
        raise ServerException(
            'Uploaded file labeled as "{}". '
            'Expecting label "{}".'.format(
                actual['name'], expected))
    label_info = configuration['l'].get(expected, {})
    if not label_info:
        available_labels = ', '.join(list(configuration['l'].keys()))
        raise ServerException(
            'Uploaded file with label "{}" that is not available to user. '
            'Available labels: {}'.format(actual['name'], available_labels)
            )
    if label_info['id'] != actual['id']:
        raise ServerException(
            'Uploaded file with label "{}" has a different label ID. '
            'Got {}. Expecting {}.'.format(
                actual['name'], actual['id'], label_info['id']))


def _get_label_id(configuration, label_name):
    """
    Return the ID for `label_name`.

    Raise an exception if `label_name` is not found.
    """
    label_info = configuration['l'].get(label_name, {})
    if not label_info:
        available_labels = ', '.join(list(configuration['l'].keys()))
        raise ServerException(
            'Label "{}" is not available to user. '
            'Available labels: {}'.format(
                label_name, available_labels))

    return label_info['id']


def _handle_file_upload(event, configuration):
    """
    Handle the labels for an uploaded file.

    This is only called for HTTP file upload events.

    Called in the worker process.
    """
    source_path = event.data['real_path']
    client_id = configuration['cid']

    parent_path, file_name = os.path.split(source_path)
    base_path, label_name = os.path.split(parent_path)
    file_base, file_extension = os.path.splitext(file_name)
    modified_file_name = file_base + _SDK_MODIFIED_MARKER + file_extension
    modified_path = os.path.join(parent_path, modified_file_name)
    error_file_name = (
        file_base + '_' + str(time()) + file_extension + '.txt')
    error_path = os.path.join(
        base_path, configuration['c']['error_path'], error_file_name)
    success_path = os.path.join(
        base_path, configuration['c']['success_path'], file_name)

    try:
        source_label = _get_label(
            source_path,
            client_id=client_id,
            token=configuration['t'],
            )

        if source_label:
            _validate_label(
                actual=source_label,
                expected=label_name,
                configuration=configuration,
                )

            # All ok. Move labeled file.
            destination_path = _move_without_overwrite(
                source_path, success_path)

            return [{
                'message': (
                    'Label "{}" was validated for "{}" and '
                    'file moved to "{}".'.format(
                        label_name, source_path, destination_path)),
                'real_path': destination_path,
                'label_name': label_name,
                'label_id': source_label['id'],
                }]

        # No label found. Try to apply the label.
        # > File is neither labeled nor protected
        label_id = _get_label_id(configuration, label_name)
        try:
            _labeler(
                ['-s', label_id, '-f', event.data['real_path']],
                client_id=client_id,
                token=configuration['t'],
                )
        except ServerException as error:
            unknown_format = (
                "Something bad happened: "
                "Default format doesn't support writing properties "
                "without protection, CorrelationId"
                )
            if unknown_format in error.output:
                # We accept any file that can't be labeled.
                destination_path = _move_without_overwrite(
                    source_path, success_path)
                # An empty modified file is left so we clean it.
                os.unlink(modified_path)
                return [{
                    'message': (
                        'File was accepted as "{}" without a label '
                        'and moved to {}.'.format(
                            label_name, destination_path)),
                    'destination_path': destination_path,
                    'real_path': destination_path,
                    'label_name': label_name,
                    }]

            # Unknown error.
            raise error

        # All ok. An exce[tion is raise if applying the label fails.
        # > New file created
        os.unlink(source_path)
        destination_path = _move_without_overwrite(
            modified_path, success_path)

        return [{
            'message': (
                'Label "{}" was succesfully auto-applied to "{}".'.format(
                    label_name, destination_path)),
            'destination_path': destination_path,
            'real_path': destination_path,
            'label_name': label_name,
            'label_id': label_id,
            }]

    except ServerException as error:
        # Failed to apply the label.
        _handle_error(error, error_path, source_path, modified_path)


def _handle_error(error, error_path, source_path, modified_path):
    """
    Handle the error of labeling a file.
    """
    try:
        # Create an error placeholder.
        with open(error_path, 'w') as stream:
            stream.write(error.message)
        raise error
    finally:
        # Delete the original file.
        try:
            os.unlink(source_path)
        except Exception:
            pass
        # The SDK might create an empty modified file.
        try:
            os.unlink(modified_path)
        except Exception:
            pass


def _move_without_overwrite(source, destination):
    """
    Try to move `source` to `destination`.

    If destination exits, it will generate the file with a timestamp.

    Returns the destination path.
    """
    destination_path = destination

    if os.path.exists(destination_path):
        file_base, file_extension = os.path.splitext(destination_path)
        timestamp = str(time())
        destination_path = "{}_{}{}".format(
            file_base, timestamp, file_extension)

    os.rename(source, destination_path)
    return destination_path
