# Copyright (c) 2017 Adi Roiban.
# See LICENSE for details.
"""
Event handler extensions to store logs into custom formats.

These extensions are not documented as part of the public documentation,
but we keep it as part of the product distribution.

The code from here is treated as core product, fully tested.
Public documentation is the only exception.
"""
from __future__ import absolute_import, unicode_literals

import csv
import os
import six

from chevah_compat import local_filesystem

from chevah.server import event_dispatcher, force_unicode
from chevah.server.commons.csv import to_csv_row
from chevah.server.commons.exception import ServerError, ServerException
from chevah.server.location.configuration import (
    DefaultFilesystemLocationConfiguration,
    )
from chevah.server.location.filesystem import FilesystemLocation


def format_event(event):
    """
    Return a custom string representation of `EventLogRecord` to be used for
    the real time monthly report.

    The name should be kept for backward compatibility.

    It will generate a report like:

    DATE/TIME, ACCOUNT_NAME, PATH, TRANSFER_TYPE, OPERATION_STATUS,
    2017-06-28 09:20:58, 1234, user1, /parent/file_T1.csv, file-upload, success
    2017-06-28 09:20:58, 4567, john4, /file_T000032.csv, file-download, failure

    It is designed to be used with an event handler catching these events.

    For FTP:
    10013, 10040, 10041, 10044, 10045, 10053, 10054, 10069, 10070, 10072,
    10073, 10075, 10076, 10078, 10079, 10095,

    For SFTP:
    30023, 30024, 30026, 30027, 30029, 30030, 30032, 30033, 30044, 30045,
    30046, 30047, 30067, 30068
    """
    try:
        definition = event_dispatcher.definitions.getEventDefinition(event.id)
        groups = definition.group_names
        if 'success' in groups:
            status = 'success'
        elif 'failure' in groups:
            status = 'failure'
        else:
            status = 'unknown'

        operation = 'unknown'
        if 'file-operation' in groups:
            for group in groups:  # pragma: no branch
                if 'operation-' in group:
                    operation = group
                    break

    except ServerError:
        status = 'unknown'
        operation = 'unknown'

    path = event.data.get('path', 'no-associated-path')

    fields = [
        event.timestamp.cwa_14051,
        event.id,
        event.account.name,
        path,
        operation,
        status,
        ]

    return to_csv_row(fields)


class GenerateRetriesReport(object):
    """
    Event handler for generating the monthly operational audit report
    after the log file is rotated.

    It will only handle the event with ID 20175 as this is the one used
    for rotating a file.

    It will only work when the log file is formatted using the `format_event`
    extension.

    It will only look for uploads and downloads operations and count the
    number of time the same file is uploaded / downloaded by the same user.
    """

    _REPORT_HEADERS = [
        'DATE/TIME',
        'ACCOUNT',
        'PATH',
        'TRANSFER_TYPE',
        'OPERATION_STATUS',
        'RETRIES',
        ]

    def __init__(self):
        """
        Called when the event handler is started.
        """
        self.parent = None
        # We have a custom location with no audited events.
        self._location = FilesystemLocation(
            parent=None,
            configuration=DefaultFilesystemLocationConfiguration(parent=self),
            )

    def onStart(self, parent):
        """
        Called in main thread when the extension starts.

        Nothing to do here for now.
        """

    def onStop(self):
        """
        Called in main thread when the extension stops.

        Nothing to do here for now.
        """

    def getConfiguration(self, event):
        """
        Called in main thread when an event is trigerred.

        Nothing to do here for now.
        """
        return ''

    @staticmethod
    def handle(event, configuration):
        """
        Called when we should handle the event
        """
        if event.id != '20175':
            raise ServerException('Handler only works for event 20175.')

        path = event.data.get('path', None)
        if not path:
            raise ServerException('Event has no associated path.')

        encoded_path = local_filesystem.getEncodedPath(path)

        base_path, report_name = os.path.split(encoded_path)

        retries_name = (
            local_filesystem.getEncodedPath('operational-audit-') +
            report_name
            )

        output_path = os.path.join(base_path, retries_name)

        kwargs = {}
        read_mode = 'rb'
        write_mode = 'wb'
        if six.PY3:
            kwargs['newline'] = ''
            read_mode = 'r'
            write_mode = 'w'

        with open(encoded_path, read_mode, **kwargs) as stream:
            report = _generate_report(stream)

        with open(output_path, write_mode, **kwargs) as stream:
            _write_report(stream, report)

        data = {
            'path': force_unicode(output_path),
            'real_path': force_unicode(output_path),
            }
        return [('20039', data)]


def _generate_report(source_stream):
    """
    Called the generate a dict with the report.
    """
    result = {}
    csv_file = csv.reader(source_stream)
    for row in csv_file:
        if len(row) != 6:
            raise ServerException('Source report has invalid format.')
        operation = row[4].strip()
        if operation not in ['operation-read', 'operation-write']:
            # Ignore any now upload or download.
            continue

        # Key is 'user, path'.
        key = (row[2], row[3])

        previous_entry = result.get(key, None)
        if previous_entry:
            count = previous_entry[-1] + 1
            previous_entry[-1] = count
        else:
            # Event ID is ignored and retries count is put last.
            result[key] = [
                row[0],
                row[2],
                row[3],
                operation,
                row[5],
                0,
                ]

    return result


def _write_report(destination_stream, report):
    """
    Write the generated report.
    """
    csv_file = csv.writer(destination_stream)

    csv_file.writerow(GenerateRetriesReport._REPORT_HEADERS)
    for _, rows in report.items():
        csv_file.writerow(rows)
