# Copyright (c) 2015 Adi Roiban.
# See LICENSE for details.
"""
FTP client side protocol.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

from datetime import datetime
import re
import six
import time

from OpenSSL import SSL
from OpenSSL._util import lib as openssl_lib
from twisted.internet import defer
from twisted.internet.error import ConnectError, TimeoutError
from twisted.internet.interfaces import IConsumer
from twisted.internet.protocol import Protocol
from twisted.protocols import ftp
from twisted.protocols.basic import LineReceiver
from zope.interface import implementer

from chevah_compat import local_filesystem, FileAttributes
from chevah.server import add_timeout, force_unicode, native_string
from chevah.server.identity.avatar import ClientAvatar
from chevah.server.commons.constant import ADDRESS_UNKNOWN
from chevah.server.commons.credential import FTPPasswordCredentials
from chevah.server.commons.exception import (
    ServerError,
    ServerException,
    )
from chevah.server.commons.factory import FactoryForProtocol
from chevah.server.commons.interface import ILocationShell, ITransferStream
from chevah.server.commons.ssl import SSLConnectionMixin
from chevah.server.commons.stream import UploadProducer
from chevah.server.location.commons import OnConnectionLostMixin

# Second to wait for command which need to be fast.
# Most operations will be interrupted after `idle_connection_timeout`
# but for some, we don't want to wait that long.
_SLOW_TIMEOUT = 15


# FIXME:6366:
# Can be renamed to FTPCommand once we no longer use Twisted code.
class ChevahFTPCommand(object):
    """
    FTP command as scheduled to be executed by the client.

    This is based on Twisted FTPCommand but was refactored as we plan
    to stop using the Twisted FTP client code.
    """

    def __init__(self, text, data_deferred=None):
        self.text = text
        self.deferred = defer.Deferred()
        # The command is sent only after `data_deferred` was called.
        self.data_deferred = data_deferred

    def fail(self, failure):
        """
        This is from Twisted, but we just call FTPCommand.deferred.errback()
        instead.

        FIXME:6366:
        Can be removed once we no longer use Twisted code.
        """
        raise RuntimeError('ChevahFTPCommand.fail should not be called.')


class FTPConnectionMixin(SSLConnectionMixin):
    """
    Shared code for generating FTP connection events.
    """
    # FTP client-side command and data channel have the same event IDs.
    _TLS_OK_EVENT = '10101'
    _TLS_FAIL_EVENT = '10107'

    @property
    def avatar(self):
        """
        See: ILocationShell.
        """
        # FIXME:4950:
        # Look at getting the avatar from the location.
        return self._avatar


@implementer(ILocationShell)
class FTPClientProtocol(
        ftp.FTPClient, OnConnectionLostMixin, FTPConnectionMixin):
    """
    Protocol for FTP client side.

    This also provides the foundation for FTPS Explicit and Implicit.
    But this class only implements FTP.

    `location` acts as a factory for this protocol.
    The data connection are initiated via the `dtp_endpoint_factory` which
    is a callable which should return a client endpoint.

    This is both a protocol as it does the client-server talk, but is also a
    shell since it provides client-side commands.

    Commands should be handled as text/unicode.
    Then encoding is done at the very last time, just before sending the
    command to the wire.

    `nextDeferred` attribute works more like the 'ongoingCommandDeferred'.
    It should be `None` when no command is executed or waiting its response.
    """

    _scheduler = None
    # Here to help with testing.
    # The minimum time to wait for a FTP command response.
    _MINIMUM_TIMEOUT = 5

    def __init__(
            self, location, dtp_endpoint_factory, debug=None):
        # We don't call FTPClient as we don't want to authenticate the
        # connection on initialization.
        ftp.FTPClientBasic.__init__(self)
        self._scheduler = location.scheduler

        # Protocol for the current data connection.
        self._dtp = None

        # FIXME:3036:
        # For now, only passive connections are supported.
        self.passive = True

        self._location = location
        self._server_address = location.configuration.address
        # Callable used to get the client endpoint needed to do
        # passive connections.
        self._getPassiveEndpoint = dtp_endpoint_factory

        # Flag to signal a multi-line response.
        self._is_multi_line_reply = False
        # Response lines received for the last command.
        self._response_lines = []

        # Initialized with an unknown peer.
        self._updateCredentials(peer=ADDRESS_UNKNOWN)
        # Called when command protocol is disconnected.
        # The disconnection might be delayed, especially for SSL
        # connection for which no data is transferred and handshake
        # is not done.
        self.disconnect_deferred = defer.Deferred()

        # If the operation has failed,
        # and we should not process further commands.
        self._failed = False

        self._tls_handshake_completed = defer.succeed(None)

        # Cache of server features
        self._features = None
        if debug:
            self._debug = debug

    def _debug(self, details):
        """
        By default, don't do anything on debug.
        """

    def emitEvent(self, event_id, data=None):
        """
        Helper to emit an event for this FTP location.
        """
        self._location.emitEvent(event_id, data=data)

    def _updateCredentials(self, peer):
        """
        Update credentials used by this protocol and keep the avatar in
        sync with the credentials.
        """
        # FIXME:4950:
        # See how to update the credentials and peer of the existing avatar.
        self._credentials = FTPPasswordCredentials(
            username=self._location.configuration.username,
            password=self._location.configuration.password,
            creator=self._location,
            peer=peer,
            )
        self._avatar = ClientAvatar(self._credentials)

    def escapePath(self, path):
        """
        Returns a FTP escaped path (replace newlines with nulls).
        """
        return path.replace('\n', '\0')

    def storeFile(self, path, offset=0):
        """
        API not supported.
        """
        raise NotImplementedError('use getStreamForWriting')

    def retrieveFile(self, path, protocol, offset=0):
        """
        API not supported.
        """
        raise NotImplementedError('use getStreamForReading')

    def connectionMade(self):
        """
        Called when we have a transport and we are connected to the server.

        This is called before the TLS handshake is done.
        """
        self._updateCredentials(peer=self.transport.getPeer())
        self.emitEvent('10102')

    def connectionLost(self, reason):
        """
        Called when connection to server was lost.
        """
        self._tlsConnectionLost(reason)
        error = reason.value
        if self.connected:
            data = {}
            data['certificate'] = self._getPeerCertificateDescription()
            data['encryption'] = self._getEncryptionDescription()
            data['reason'] = force_unicode(error)
            # Not yet updated the disconnected state.
            self.emitEvent('10103', data=data)
            self.disconnect_deferred.callback(None)

        # This is inherited from BasicProtocol but due to Twisted strange
        # design, this is not auto-updated.
        self.connected = 0
        # We fails as way to cancel all pending commands.
        self.fail(error)
        # Inform that we lost the connection.
        self._onConnectionLost(error)

    def lineReceived(self, line):
        """
        Called when server has sent us a command response line.
        """
        # Some bad servers might send multi-line with non-standard delimiters.
        if b'\n' in line:
            for part in line.splitlines():
                self.lineReceived(part)
            return

        line = force_unicode(line)
        self._debug('< %s' % (line,))

        # Add line to the current response
        self._response_lines.append(line)

        # The first line of a multi-line response start with 3 digits followed
        # by hyphen,
        is_multi_line = re.match(r'\d{3}-', line)
        if is_multi_line:
            # We will receive a multi-line response.
            self._is_multi_line_reply = True
            return

        # The last line of response starts with 3 digits followed by a
        # space.
        is_last_line = re.match(r'\d{3} ', line)
        if not is_last_line and self._is_multi_line_reply:
            # Continue if this isn't the last line of a response and we
            # are waiting a multi-line response.
            return

        # We no longer have the multi-line as we have reached the last line.
        self._is_multi_line_reply = False

        if self.nextDeferred is None:
            # No response was expected or the connection was already lost.
            self.fail(ftp.UnexpectedResponse(self._response_lines))
            return

        # Reset the response before triggering the callback as it might
        # trigger a new command right away.
        response_lines = self._response_lines
        self._response_lines = []

        code = line[0:3]

        if code[0] == '1':
            # Inform preliminary result and send the future deferred.
            # Don't do any further process as we are waiting for the final
            # result at the end of the data connection.
            data_end_deferred = defer.Deferred()
            current = self.nextDeferred
            self.nextDeferred = data_end_deferred
            current.callback((response_lines, data_end_deferred))
            return

        if code[0] in ('2', '3'):
            # Success
            self.nextDeferred.callback(response_lines)
        elif code[0] in ('4', '5'):
            self.nextDeferred.errback(ftp.CommandFailed(response_lines))
        else:
            # This shouldn't happen unless something screwed up.
            self.nextDeferred.errback(ftp.BadResponse(response_lines))

        # After the response was received from the server, see if we need to
        # execute more commands.
        self._processNextCommand()

    def queueStringCommand(self, command, data_deferred=None):
        """
        Add a FTP command to the queue and return a deferred that is called
        when a result is received for the command.

        When data_deferred is not set, the command is expected to return
        a response before the idle_connection_timeout.
        Otherwise the command deferred fails with a timeout.
        """
        ftp_command = ChevahFTPCommand(command, data_deferred=data_deferred)
        self._queueCommand(ftp_command)
        if not data_deferred:
            # FIXME:6411:
            # See if we can move the timeout computation part
            # at the location configuration level.
            timeout_interval = max(
                self._location.configuration.idle_connection_timeout,
                self._MINIMUM_TIMEOUT,
                )
            add_timeout(
                self._scheduler,
                ftp_command.deferred,
                timeout_interval,
                message='No response for {}.'.format(command)
                )
        return ftp_command.deferred

    def _queueCommand(self, command):
        """
        Add an ChevahFTPCommand object to the queue and execute it when
        possible.
        """
        if self._failed and command.text != 'QUIT':
            # We have already failed.
            # Fail the command right away.
            # QUIT is a special case as on a normal close, the QUIT command
            # is called via the fail().
            command.deferred.errback(
                ftp.ConnectionLost('Connection is closing.'))
            return

        self.actionQueue.append(command)
        if (
            len(self.actionQueue) == 1 and
            self.transport is not None and
            self.nextDeferred is None
                ):
            # It's the only thing in the queue, and we are connected and we
            # aren't waiting for a response of an earlier command.
            self._processNextCommand()

    def sendNextCommand(self, line):
        """
        This public method from upstream should not be used.
        """
        raise RuntimeError('Use _processNextCommand instead.')

    def _processNextCommand(self):
        """
        Processes the next command in the queue, if there is any.

        This is called in an indirect loop and might be called even if there
        are no command to be processed. For example as the last iteration in
        the loop.
        """
        next_command = self.popCommandQueue()
        if next_command is None:
            # No command to execute.
            # Mark that previous command was done.
            self.nextDeferred = None
            return

        self.nextDeferred = next_command.deferred
        if next_command.data_deferred:
            deferred = next_command.data_deferred
        else:
            deferred = defer.succeed(next_command)

        def cb_send_line(ignored):
            if next_command.deferred.called:
                # By the time we need to send it, it might be already
                # triggered due to a failure.
                # Try to send the following command.
                self._processNextCommand()
                return

            self._sendLine(next_command.text)

        deferred.addCallback(cb_send_line)
        # In case there are any errors, we mute them as nobody is waiting here
        # and delayed commands should handle this in a separate way.
        deferred.addErrback(lambda failure: None)

    def sendLine(self, line):
        """
        This is here due to LineReceiver API, but code shouldn't call it.
        """
        raise RuntimeError('Use queueStringCommand instead.')

    def queueCommand(self, command):
        """
        We no longer support this Twisted API.
        """
        raise RuntimeError('Use queueStringCommand instead.')

    def _sendLine(self, line):
        """
        Sends a text line to the command channel.
        """
        self._debug('> %s' % (force_unicode(line),))

        if not self.transport or not self.transport.connected:
            # This is not expected, as when the connection is lost,
            # any pending command is cancelled.
            # We are supposed to send a command but we don't have a connected
            # transport.
            # Fail the command as at this point we don't have a timeout and we
            # don't support waiting for the transport to be connected.
            deferred = self.nextDeferred
            self.nextDeferred = None
            deferred.errback(ftp.ConnectionLost(
                'Connection was already lost.',
                ftp.FTPError('Connection already lost'),
                ))
            return

        # We bypass our own or parent's sendLine methods.
        LineReceiver.sendLine(self, line.encode('utf-8'))

    @defer.inlineCallbacks
    def _receiveFromDataConnection(self, command, protocol):
        """
        Send a command that is expected to receive data from the
        server.
        """
        initial_lines, data_end_deferred = yield self._openDataConnection(
            command, protocol)

        @defer.inlineCallbacks
        def cb_wait_dtp_close(final_lines):
            """
            Called when over the command channel we have received the
            final response from the data command.
            """
            # Wait for the data connection to be closed.
            deferred = protocol.disconnect_deferred
            add_timeout(self._scheduler, deferred, _SLOW_TIMEOUT)
            yield deferred
            defer.returnValue(final_lines)

        # We got a response from the command channel that the transfer is
        # done, but also wait for the connection to be closed.
        data_end_deferred.addCallback(cb_wait_dtp_close)

        # If we fail to get the final response,
        # close the data channel.
        def eb_end_error(failure):
            """
            Make sure we inform the data protocol on errors.
            """
            if protocol.transport.connected:
                protocol.transport.loseConnection()
            return failure

        data_end_deferred.addErrback(eb_end_error)

        defer.returnValue((initial_lines, data_end_deferred))

    @defer.inlineCallbacks
    def _sendToDataConnection(self, command, path, mode):
        """
        Send FTP command string to data connection and return a consumer
        protocol that can handle that data that we want to upload.

        Return a tuple of (initial_lines, end_lines_deferred)
        """
        consumer_protocol = SenderProtocol(self._avatar, self._scheduler)
        _, data_end_deferred = yield self._openDataConnection(
            command, consumer_protocol)

        def eb_end_error(failure):
            """
            Make sure we inform the data protocol on errors.
            """
            if consumer_protocol.transport.connected:
                # Somehow the server sent its response before
                # we got to send all our data.
                consumer_protocol.transport.loseConnection()
            return failure

        data_end_deferred.addErrback(eb_end_error)

        result = FTPUploadedFile(
            fail=self.fail,
            path=path,
            location=self._location,
            consumer=consumer_protocol,
            end=data_end_deferred,
            mode=mode,
            scheduler=self._scheduler,
            )

        defer.returnValue(result)

    @defer.inlineCallbacks
    def _openDataConnection(self, command, protocol):
        """
        Called to established a data connection needed by `command` and
        over which `protocol` will communicate.

        This method returns a deferred that will fire with a tuple of
        (initial_lines, end_command_deferred).

        It will fail if the data connection was established,
        but the command issued for the data connection did not return a
        confirmation that it can proceed.

        `initial_lines` are the lines received over the command channel as
        an initial response to the command.

        `end_deferred` is a deferred that can be used to wait for the end of the
        data connection.
        """

        # The commands attached to this PASV transfer needs to send next,
        # as otherwise we might interleave multiple passive requests.
        # And due to RFC 3659, the REST command might come before a
        # RERT or STOR, so we need to issue multiple command inside a PASV
        # window.
        # Ex valid: PASV -> REST -> RETR
        # Ex invalid REST -> PASV -> RETR
        pasv_deferred = self.queueStringCommand('PASV')
        # We queue the data command right away so that no other command
        # will interfere with this PASV connection.
        # Hold sending the next command until we establish the data
        # connection.
        pasv_done = defer.Deferred()
        data_deferred = self.queueStringCommand(
            command, data_deferred=pasv_done)

        pasv_lines = None
        try:
            pasv_lines = yield pasv_deferred
            host, port = ftp.decodeHostPort(pasv_lines[-1][4:])
            if self._location.configuration.ignore_passive_address:
                # PASV host IP is ignored based on configuration.
                host = self._server_address
        except Exception as error:
            # We failed parsing the PASV response.
            # Release the data command.
            pasv_done.callback(None)
            try:
                # By calling the command here and calling _processNextCommand
                # later, we trigger the ignoring of this command.
                data_deferred.callback([''])
            except defer.AlreadyCalledError:
                """
                Maybe the client is quiting so this might be already
                called.
                """
            # Ignore any error for the data command,
            # as we will forward the PASV error.
            data_deferred.addErrback(lambda _: None)

            self._processNextCommand()

            if pasv_lines is None:
                raise error
            raise ftp.CommandFailed(pasv_lines)

        endpoint = self._getPassiveEndpoint(self, host, port)
        if endpoint.use_ssl:
            protocol.beforeTLSHandshake('data')
        try:
            self._dtp = yield endpoint.connect(FactoryForProtocol(protocol))
            remote_peer = self._dtp.transport.getPeer()
            self.emitEvent('10099', data={
                'host': remote_peer.host,
                'port': str(remote_peer.port),
                'command': command,
                'fqdn': host,
                })
        except ConnectError as error:
            raise ftp.ConnectionLost(
                'Failed to connect to passive channel at "%s:%s". %s' % (
                    host, port, force_unicode(error),))

        self._preDTPHandshake()

        # We are successfully connected to the DTP.
        # We can issue our data command.
        pasv_done.callback(None)

        dtp_failed = False
        try:
            yield protocol.waitHandshakeCompleted()
        except defer.CancelledError:
            raise
        except Exception:
            """
            We ignore any error to setup the TLS on the data channel.
            We will send the command in order to receive the error over the
            command channel.

            This can happen if the server finds out that the requested
            command fails, so it closes the DTP right away, without waiting
            for the DTP tls handshake.
            """
            # Take a note that we failed,
            # In the case the command channel will not observe the failure,
            # we will force a failure on the client-side.
            dtp_failed = True

        try:
            result = yield data_deferred

            try:
                response_lines, data_end_deferred = result
            except Exception:
                # Something went very wrong.
                # It looks like we got a success response,
                # instead of of the initial partial response.
                raise ftp.CommandFailed(result)

            # The DTP failed, but somehow the server still thinks we are ok.
            if dtp_failed:
                # It's ok to ignore `data_end_deferred` as at this point
                # it should not have anything attached to it.
                response = ['542 DTP setup failed.'] + response_lines
                raise ftp.CommandFailed(response)

            self._postDTPHandshake()

        except Exception:
            # Close the data connection and this will
            # trigger the protocol closing.
            # If we are here, the data connection was already made.
            try:
                protocol.transport.loseConnection()
            except defer.AlreadyCalledError:  # noqa:cover
                """
                All the protocols have a waiting deferred, which
                might be already called if the connection was already closed.
                """
            raise

        defer.returnValue((response_lines, data_end_deferred))

    @defer.inlineCallbacks
    def fail(self, error):
        """
        Trigger the client failure with `error`.

        This is also used to stop the client.
        `fail` name is used since it's used by Twisted code.

        This will disconnect the client and try to be nice by sending the
        QUIT command.

        This is called from the timeout process, so it has its own timeout
        handling.
        """
        if self._failed:
            # We're already failing.
            return

        self._failed = True

        # We send the error first, as otherwise disconnecting will call
        # _fail with normal disconnection error and we will not be able to
        # _fail with the requested error.
        self._fail(error)

        try:
            if self.connected:
                # If we are connected, send the QUIT command to the server.
                yield self.queueStringCommand('QUIT')
        except Exception:
            """
            We ignore any failure cause by the QUIT command.
            """
        finally:
            self.transport.loseConnection()

    def _fail(self, error):
        """
        Fail all queued deferreds with a connection lost.
        """
        # First close the data port.
        if self._dtp and self._dtp.transport.connected:
            self._dtp.transport.loseConnection()
        fail_wrap_error = ftp.ConnectionLost('FTP fail', error)

        # Let the current and pending command know that we have failed.
        if self.nextDeferred:
            # This is current/ongoing command.
            deferred, self.nextDeferred = self.nextDeferred, None
            try:
                deferred.errback(fail_wrap_error)
            except defer.AlreadyCalledError:
                """
                Maybe we are disconnecting, but have already
                received an error response from the server.
                Ignore the error.
                """

        # Fail all the queued commands.
        next_command = self.popCommandQueue()
        while next_command:
            try:
                next_command.deferred.errback(fail_wrap_error)
            except defer.AlreadyCalledError:
                """
                Maybe the queued command was cancelled as a side effect of the
                current command.
                Ignore the error.
                """
            next_command = self.popCommandQueue()

    @defer.inlineCallbacks
    def authenticate(self):
        """
        Called when we should authenticate.

        On success it returns the low level FTP client which for now is the
        protocol itself.
        """
        try:
            # For implicit FTPS Wait for any TLS handshake to be done,
            # before starting the authentication.
            # The TLS handshake might fail.
            yield self.waitHandshakeCompleted()
            yield self._beforeAuthentication()
            yield self._authenticateWithUsernameAndPassword()
            self.emitEvent('10106')
        except SSL.Error as error:
            # We don't want to trigger an ISE for TLS errors.
            raise ServerException(force_unicode(error))
        except ftp.CommandFailed as error:
            details = force_unicode(error)
            if six.PY2:
                # Try to have similar text representation in py2 and py3.
                details = details.replace("u'", "'")
            raise ServerError('10104', data={
                'name': self._location.name,
                'details': details,
                },
                avatar=self.avatar,
                )

        defer.returnValue(self)

    def _beforeAuthentication(self):
        """
        Called before the command connection is authenticated.
        For plain FTP there is nothing to do.
        """
        return defer.succeed(None)

    @defer.inlineCallbacks
    def _authenticateWithUsernameAndPassword(self):
        """
        Low level authentication without error handling.
        """
        lines = yield self.queueStringCommand(
            'USER %s' % (self._credentials.username,))

        if _is_auth_accepted(lines):
            # USER is accepted.
            # No need to send PASS or ACCT.
            defer.returnValue(self)

        if _need_ACCT(lines):
            lines = yield self._sendACCT()

        if _is_auth_accepted(lines):
            # USER + possible ACCT accepted.
            # No need to send PASS.
            defer.returnValue(self)

        if _need_PASS(lines):
            lines = yield self._sendPASS()

        if _is_auth_accepted(lines):
            # PASS accepted and is enough.
            defer.returnValue(self)

        if _need_ACCT(lines):
            lines = yield self._sendACCT()

        if _is_auth_accepted(lines):
            # No need to try other methods.
            defer.returnValue(self)

        # If we are here, it means that auth was not accepted.
        # There is response code 336 but this is designed for interactive
        # usage. For example for TOTP code.
        # This is why for now, we fail for 336.
        # We can look into 336 together with the first customer.
        raise ftp.CommandFailed(lines)

    def _sendPASS(self):
        """
        Send the PASS command.
        """
        password = self._credentials.password
        if not password:
            password = ''  # noqa:bandit
        return self.queueStringCommand('PASS %s' % (password,))

    def _sendACCT(self):
        """
        Send the ACCT command
        """
        acct = self._location.configuration.acct
        if not acct:
            raise ftp.CommandFailed(
                'The remote server requests extra ACCT account information. '
                'Set a value to "acct" configuration.'
                )
        return self.queueStringCommand('ACCT %s' % (acct,))

    def getFeatures(self):
        """
        Return a list if server FTP features.
        """
        if self._features is not None:
            return defer.succeed(self._features)

        def cb_got_feat(raw_features):
            """
            Called when we got the features from the server.
            """
            result = []
            for part in raw_features[1:-1]:
                result.append(force_unicode(part.strip()))
            self._features = tuple(result)
            return self._features

        deferred = self.queueStringCommand('FEAT')
        # If FEAT fails, we just ignore and consider that no features are
        # available.
        deferred.addErrback(lambda _: ['ignored', 'ignored'])
        deferred.addCallback(cb_got_feat)
        return deferred

    def _hasFeature(self, name):
        """
        Return True if `name` is in the cached features.
        """
        if not self._features:
            return False

        for feature in self._features:
            if feature.startswith(name):
                return True

        return False

    @defer.inlineCallbacks
    def list(self, path):
        """
        Return a list of members of `path`.
        """
        def path_joiner(name):
            """
            Delegate path joining to the associated location.
            """
            return self._location.joinPath(path, name)

        parse_protocol = FTPListProtocol(
            path_joiner,
            avatar=self.avatar,
            scheduler=self._scheduler,
            )
        command = 'LIST %s' % (self.escapePath(path),)

        # We ignore the first lines received on the command channel,
        # as if they are an error, the call will raise an exception.
        # Any success is accepted.
        # And if we care about them, we can enable the debug for the location.
        _, end = yield self._receiveFromDataConnection(command, parse_protocol)

        # Wait for the data connection to be receive the final response.
        # Also ignore the response, as it is returned only on success.
        yield end

        # FIXME:4104:
        # Return iterator instead of a list for performance
        # reasons, but this should be piped directly into the
        # parse protocol, so that the lines for server are
        # forwarded as soon as they are received.
        defer.returnValue(parse_protocol.entries)

    def exists(self, path):
        """
        Return `True` if path exists.

        It uses the LIST command as documented in RFC 959.
        """
        if self._hasFeature('MLST'):
            return self._existsMLST(path)
        else:
            return self._existsLIST(path)

    def _existsLIST(self, path):
        """
        Check the existence of a file using the LIST command.
        """
        dirname, basename = self._location.splitPath(path)
        if not basename:
            # We don't have a basename when path end with a slash, signaling
            # a folder path.
            dirname, basename = self._location.splitPath(dirname)

        deferred = self.list(dirname)

        def cb_list(members):
            """
            Called when we got a successful result from LIST.
            """
            if not basename:
                # We have the root folder,we can list it, so it should exist.
                return True

            names = [m.name for m in members]
            return basename in names

        deferred.addCallback(cb_list)
        deferred.addErrback(self._ebExists)
        return deferred

    def _existsMLST(self, path):
        """
        Check the existence of a file using the MLST command.
        """

        deferred = self.queueStringCommand('MLST ' + path)

        def cb_MLST(result):
            """
            Called when we got a response from MLST.
            """
            # We don't care about the attributes.
            # If we are here, path exists.
            return True

        deferred.addCallback(cb_MLST)
        deferred.addErrback(self._ebExists)
        return deferred

    def _ebExists(self, failure):
        """
        Called when the command use to check path existence fails.
        """
        failure.trap(ftp.CommandFailed)
        for response in failure.value.args[0]:
            if response.startswith('550'):
                return False

        return failure

    @defer.inlineCallbacks
    def getStreamForReading(self, path):
        """
        Return a stream to the data connection for a download request for
        `path` together with a deferred for the responses from the command
        channel.
        """
        stream_protocol = ReceiverProtocol(self._avatar, self._scheduler)
        # We ignore the lines received from control channel, as they are
        # returned only on success.
        # On failure, this call will raise an exception.
        command = 'RETR ' + path
        _, end = yield self._receiveFromDataConnection(command, stream_protocol)

        result = FTPDownloadedFile(
            fail=self.fail,
            path=path,
            location=self._location,
            data_producer=stream_protocol,
            end=end,
            scheduler=self._scheduler,
            )
        defer.returnValue(result)

    def getStreamForWriting(self, path, length=None):
        """
        Return a stream to the data connection for an upload request for
        `path` together with a deferred for the responses from the command
        channel.

        `length` is ignored as for FTP we don't need to announce how much we
        will write.
        """
        return self._sendToDataConnection(
            'STOR ' + self.escapePath(path), path=path, mode='writing')

    def getStreamForAppending(self, path):
        """
        Return a stream to the data connection for an append request for
        `path` together with a deferred for the responses from the command
        channel.
        """
        return self._sendToDataConnection(
            'APPE ' + self.escapePath(path), path=path, mode='appending')

    def _preDTPHandshake(self):
        """
        Called after DTP was connected, and before the TLS handshake.

        For FTP does nothing.
        """

    def _postDTPHandshake(self):
        """
        Called after DTP was connected, and before the TLS handshake.

        For FTP does nothing.
        """
        return None


class _FTPSClientProtocol(FTPClientProtocol):
    """
    Shared code by explicit and implicit FTP client side.

    All data connection are sharing the same context factory so that
    we can re-use the command connection session as a way to make sure
    we are connecting to the same remote application.
    """

    def __init__(
        self, location, dtp_endpoint_factory, ssl_context_factory,
        debug=None,
            ):
        super(_FTPSClientProtocol, self).__init__(
            location, dtp_endpoint_factory,
            debug=debug,
            )
        self._ssl_context_factory = ssl_context_factory
        self._ftps_reuse_session = location.configuration.ftps_reuse_session

    @defer.inlineCallbacks
    def _cbAfterTLSHandshake(self, ignored=None):
        """
        Called after the TLS handshake is done to inform the FTP server
        of our security setup.

        This is both a callback, and normal function to help share the code
        with implicit/explicit.

        For explicit, this is called as a method.
        For implicit, this is called as a callback.
        """
        # After a successful authentication FTPS clients (implicit or
        # explicit) needs to set up the security level.
        # As in RFC 2228 and 4217.
        yield self.queueStringCommand('PBSZ 0')
        yield self.queueStringCommand('PROT P')

    def _preDTPHandshake(self):
        """
        Called after DTP was connected, and before the TLS handshake.
        """
        # If we don't have a TLS connection over data command, it means
        # we are on FTP or FTPS with CCC.
        command_tls = self._getTLSConnection()
        if not self._ftps_reuse_session:
            return

        # We just set the same session for data connect as the
        # command connection, before we call the TLS handshake.
        # If the server doesn't recognized the session,
        # it will negotiate a new session.
        # After the handshake we check that we have the same session.
        data_tls = self._dtp.transport.protocol._tlsConnection
        # Setting the session only makes sense for client-operations.
        # The server will reuse the session, if asked by the client.
        data_tls.set_session(command_tls.get_session())

    def _postDTPHandshake(self):
        """
        Called after the DTP TLS handshake connection was done.
        """
        command_tls = self._getTLSConnection()
        # After the handshake we check that we have connected
        # to the same session.
        data_tls = self._dtp.transport.protocol._tlsConnection

        # Check that we have the same server certificate.
        data_cert = data_tls.get_peer_certificate()
        command_cert = command_tls.get_peer_certificate()

        try:
            self._checkSameCertificates(command_cert, data_cert)
        except ServerException as error:
            raise ServerException(
                'Data connection made using a different certificate. {} '
                'Command: {} - Data: {}'.format(
                    error.message,
                    self._getPeerCertificateDescription(command_cert),
                    self._getPeerCertificateDescription(data_cert),
                    ))

        if not self._ftps_reuse_session:
            # We are not doing more checks.
            return

        # See
        # https://www.openssl.org/docs/man1.1.1/man3/SSL_session_reused.html
        # FIXME:6481:
        # Try to use pyOpenSSL API.
        if openssl_lib.SSL_session_reused(data_tls._ssl) == 0:
            raise ServerException(
                'FTPS server has not reused the TLS session.')

    def _checkSameCertificates(self, command_cert, data_cert):
        """
        Raise an exception if command_cert and data_cert are not the same.

        Digest check is the final test.
        We run a few extra test to be able to provide better reporting.
        """
        if data_cert.get_subject() != command_cert.get_subject():
            raise ServerException('Different subject.')

        data_digest = data_cert.digest(native_string('sha256'))
        command_digest = command_cert.digest(native_string('sha256'))
        if data_digest != command_digest:
            raise ServerException('Different digest.')


class FTPSEClientProtocol(_FTPSClientProtocol):
    """
    Protocol for explicit FTP client side.
    """

    @defer.inlineCallbacks
    def _beforeAuthentication(self):
        """
        See: FTPClientProtocol.
        """
        try:
            # As in RFC 4217 we use `AUTH TLS` hoping that all FTP servers
            # support it an they don't ask for `AUTH TLS`.
            yield self.queueStringCommand('AUTH TLS')
        except ftp.CommandFailed as error:
            # Add a hint that we failed at AUTH TLS as the server might not
            # understand this command and just reply with a generic error.
            # This might not be the best idea as we are making it to look
            # like the inserted error came from the server, but at this points
            # it too complex to introduce another exception type.
            error.args[0].insert(0, '500 AUTH TLS rejected.')
            raise error

        self.beforeTLSHandshake('command')
        # We don't check for a `234` code and we will continue for any
        # 2XX or 3XX code
        # After STARTTLS, the transport will buffer all data received
        # while the handshake is not done.
        # It would be nice to wait here... but unless we are on
        # Twisted 16.4 and use IHandshakeListener.
        # So basically the client will block on the next command.
        # We would introduce here a NOOP like command (ex FEAT) but as
        # this is the low level client, we should not run commands which
        # were not explicitly requested and are not related to the current
        # operation.
        try:
            self.transport.startTLS(self._ssl_context_factory)
            yield self.waitHandshakeCompleted()
        except defer.CancelledError:
            raise
        except Exception as error:
            raise ftp.ConnectionLost(force_unicode(error))

        yield self._cbAfterTLSHandshake()


class FTPSIClientProtocol(_FTPSClientProtocol):
    """
    Protocol for Implicit FTP client side.
    """

    def connectionMade(self):
        """
        Called when we have a transport and we are connected to the server.

        This is called before the TLS handshake is done.
        For FTPSI, the handshake is triggered right away at connection time.
        """
        super(FTPSIClientProtocol, self).connectionMade()
        self.beforeTLSHandshake('command')

        # We send PBSZ/PROT after TLS handshake.
        # Also, reporting any TLS handshake errors is deferred until the
        # client authentication is requested.
        # We don't have other API to report an error that happened after the
        # connection but before the authentication.
        deferred = self.waitHandshakeCompleted()
        deferred.addCallback(self._cbAfterTLSHandshake)


class ReceiverProtocol(Protocol, FTPConnectionMixin):
    """
    Protocol used to receive file data from the data channel using a pull
    consumer.

    As a protocol, it is created only when a new connection is connected.

    It emulates the FTP data channel push producer into a pull producer.
    """

    def __init__(self, avatar, scheduler):
        # Called when the data protocol is disconnected.
        self.disconnect_deferred = defer.Deferred()
        # Called after the transport has produced the amount of data that
        # was requested by the read operation.
        self._read_deferred = None
        # The size of data requested by the current read command.
        self._read_length = 0
        # Data received so far and not consumed yet.
        self._buffer = b''
        # Connection state of this protocol as defined by BasicProtocol in
        # Twisted.
        self.connected = False
        self._avatar = avatar
        self._scheduler = scheduler
        self._tls_handshake_completed = defer.succeed(None)

    def connectionMade(self):
        """
        Called when data channel was opened.
        """
        self.connected = True

    def connectionLost(self, reason):
        """
        Called when data channel was closed.
        """
        self.connected = False
        self._tlsConnectionLost(reason)
        if self._read_deferred:
            # Connection was lost while we were waiting to read more data.
            self.dataReceived(b'')

        if not self.disconnect_deferred.called:
            self.disconnect_deferred.callback(None)

    def read(self, length):
        """
        Called to read up to length.

        It might return less than `length`.

        Return a deferred which is fired when request data is received.
        """
        if self._read_deferred:
            return defer.fail(
                RuntimeError('A read request was already made.'))

        if not self.connected:
            data = self._buffer
            self._buffer = b''
            return defer.succeed(data)

        if length <= len(self._buffer):
            # We already have the requested data.
            data = self._buffer[:length]
            self._buffer = self._buffer[length:]
            return defer.succeed(data)

        # We don't have all the requested data.
        # Ask the transport to produce more and wait for it.
        self._read_deferred = defer.Deferred()
        self._read_length = length
        self.transport.resumeProducing()
        return self._read_deferred

    def dataReceived(self, data):
        """
        Called when data was received from the data channel.
        """
        self._buffer += data

        if not self._read_deferred:
            # Data was received before we made a read request.
            # Just pause to make sure we don't receive any more data.
            self.transport.pauseProducing()
            return

        current_data = self._buffer[:self._read_length]
        self._buffer = self._buffer[self._read_length:]

        self.transport.pauseProducing()

        # Triggering the deferred might call another read right away, so
        # we clear the state now.
        deferred = self._read_deferred
        self._read_deferred = None

        # Let the deferred who wait for our read to know that data was
        # received.
        deferred.callback(current_data)


class SenderProtocol(Protocol, FTPConnectionMixin):
    """
    Helper to send data over the data connection.
    """

    def __init__(self, avatar, scheduler):
        self.disconnect_deferred = defer.Deferred()
        self._scheduler = scheduler
        self._avatar = avatar
        self._tls_handshake_completed = defer.succeed(None)

    def dataReceived(self, data):  # noqa:cover
        raise RuntimeError(
            "Received data from the server on a "
            "send-only data-connection"
            )

    def connectionLost(self, reason):
        self._tlsConnectionLost(reason)
        self.connected = False
        self.disconnect_deferred.callback('connection done')

    def write(self, data):
        if not self.connected:
            raise ServerException('FTP write called while not connected.')
        self.transport.write(data)

    def registerProducer(self, producer, streaming):
        """
        Register the given producer with our transport.
        """
        self.transport.registerProducer(producer, streaming)

    def unregisterProducer(self):
        """
        Unregister the previously registered producer.
        """
        self.transport.unregisterProducer()

    def finish(self):
        self.transport.loseConnection()


class _FTPTransferStreamMixin(object):
    """
    Common code for both sender and receiver FTP data streams.
    """

    @defer.inlineCallbacks
    def _handleStreamClose(self, end_deferred):
        """
        Attach the deferred for handling the close operation.
        """
        add_timeout(self._scheduler, end_deferred, _SLOW_TIMEOUT)

        try:
            result = yield end_deferred
            """
            Called when connection was closed and we got a response from the
            command channel.
            """
            # Now that the connection is closed check the command response.
            # The FTP connection will return a huge deferred list only after
            # Data connection is closed.
            # and we need to extract the command channel responses.
            if not result:
                raise ServerException('No response for data channel closure.')

            for response in result:
                if response.startswith('5'):
                    raise ServerException(response)

        except ftp.CommandFailed as error:
            """
            Called when a non-success message is returned when the stream is
            closed by closing the data connection.
            """
            for response in error.args[0]:
                if response.startswith('426'):
                    # 426 Connection closed; transfer aborted.
                    # Close is successful as the connection is closed, even
                    # if the transfer failed.
                    return

            raise
        except TimeoutError:
            # Trigger the data channel to close.
            self._fail(ServerException('Fail to close data connection.'))
            # Forward the error to the close operation.
            raise


@implementer(ITransferStream)
class FTPUploadedFile(UploadProducer, _FTPTransferStreamMixin):
    """
    Stream used to upload a file to the FTP server.
    """

    def __init__(
        self, fail, path, location, consumer, end, scheduler,
        mode='writing',
            ):
        super(FTPUploadedFile, self).__init__(consumer)
        self._fail = fail
        self._consumer = consumer
        self._consumer.registerProducer(self, streaming=True)
        self._end = end
        self.location = location
        self.path = path
        self._mode = mode

        # Temporary storage for data which can't be yet written.
        # `None` is used to signal that no write is waiting.
        self._pause_buffer = None
        self._scheduler = scheduler

    @property
    def mode(self):
        """
        See: `ITransferStream`.
        """
        return self._mode

    def close(self):
        """"
        See: `ITransferStream`.
        """
        if not self._consumer:
            # Data connection was already closed.
            return defer.fail(ServerException(
                'Connection lost before sending all FTP data.'))

        if self._consumer.disconnect_deferred.called and not self._end.called:
            # Data connection might not even be established.
            # This is a failure, as from here should be the only successful
            # place from where the upload data connection should be closed.
            self._end.errback(ServerException(
                'Connection lost before sending all FTP data.'))
        self._consumer.unregisterProducer()
        self._consumer.finish()
        # Close local file and wait for the data channel to close.
        defered = self._handleStreamClose(self._end)
        return defered


@implementer(ITransferStream, IConsumer)
class FTPDownloadedFile(_FTPTransferStreamMixin):
    """
    A pull consumer for FTP data channel.

    This requires that the data channel to be converted to a pull producer.
    """

    mode = 'reading'

    def __init__(
            self, fail, path, location, data_producer, end, scheduler):
        self._fail = fail
        self._filesystem = local_filesystem
        self._producer = data_producer
        self._end = end
        self.path = path
        self.location = location
        self._scheduler = scheduler

    def close(self):
        """
        See: `ITransferStream`.
        """
        self._producer.transport.loseConnection()
        # `_end` is called when the server sends the final response for the
        # data command.
        # This means that the operation was done.
        return self._handleStreamClose(self._end)

    def write(self, data):
        """
        See: `ITransferStream`.
        """
        return defer.fail(ServerException('Only opened for reading.'))

    def read(self, length):
        """
        See: `ITransferStream`.
        """
        return self._producer.read(length)


class FTPListProtocol(LineReceiver, FTPConnectionMixin):
    r"""
    Parser for UNIX (bin\ls) folder style listings.

    It supports both recent and old (more than 6 moths) date formats.
    All date/time values are assuming UTC timezone.

    Inspiration from `twisted.protocols.ftp.FTPFileListProtocol`.
    """

    #: Valid entry regular expression pattern.
    _pattern = re.compile(
        br'^(?P<type>.)(?P<perms>.{9})\s+(?P<links>\d*)\s*'
        br'(?P<owner>\S+)\s+(?P<group>\S+)\s+(?P<size>\d+)\s+'
        br'(?P<date>...\s+\d+\s+[\d:]+)\s+(?P<name>.+?)'
        br'( -> (?P<target>[^\r]*))?\r?$'
        )
    #: Old date (> 6 months old) regular expression pattern.
    _old_date_pattern = re.compile(
        br'^(?P<month>[\w]{3})\s+(?P<day>[\d]{1,2})\s+'
        br'(?P<year>[\d]{4})$'
        )
    #: Recent date (<= 6 months old) regular expression
    #: pattern.
    _recent_date_pattern = re.compile(
        br'^(?P<month>[\w]{3})\s+(?P<day>[\d]{1,2})'
        br'\s+(?P<hour>[\d]{2}):(?P<minute>[\d]{2})$'
        )
    #: Function called to build full path for each element.
    _path_builder = None

    # We go for the simple Unix delimiter and not the FTP delimiter to also
    # handle servers sending just Unix newlines.
    # The second part of this is in lineReceived.
    delimiter = b'\n'

    # The parser will do its best to extract a date, but on certain conditions
    # it will return these placeholder to help identify what went wrong.
    # The changes of having real files with these dates are low
    _TIMESTAMP_UNDERFLOW = 0
    _TIMESTAMP_NOT_FOUND = 1

    def __init__(self, path_builder, avatar, scheduler):
        super(FTPListProtocol, self).__init__()
        self._path_builder = path_builder
        self._entries = []
        self._unknown = []
        self.connected = False
        self._avatar = avatar
        self._scheduler = scheduler
        # Called when connection is lost and no more data will be received.
        self.disconnect_deferred = defer.Deferred()
        # Called when TLS handshake is done.
        self._tls_handshake_completed = defer.succeed(None)

    @property
    def entries(self):
        """
        List of IFileAttributes for parsed entries
        """
        return self._entries

    @property
    def unknown(self):
        """
        List of lines that could not be parsed
        """
        return self._unknown

    def connectionMade(self):
        """
        Called when the data connection was made.
        """
        self.connected = True

    def connectionLost(self, reason=None):
        """
        Called when data channel was closed.
        """
        self.connected = False
        self._tlsConnectionLost(reason)
        self.disconnect_deferred.callback(None)

    def lineReceived(self, line):
        """
        See: `LineReceiver`
        """
        # Some servers might send multi-line with non-standard delimiters.
        # The first part of this is in `cls.delimiter`
        if b'\r' in line:
            for part in line.split(b'\r'):
                self._parseLine(part)
            return

        self._parseLine(line)

    def rawDataReceived(self, data):
        """
        See: `LineReceiver`
        """
        raise RuntimeError('FTP List protocol should not receive raw data.')

    def _parseLine(self, line):
        """
        Called when a line was received from the remote server.

        This method is expected to populate entries and unknown
        instance variables.
        """
        if not line:
            # Ignore empty lines.
            return

        # Line parsing is done in bytes,
        # this means that we can in theory support any encoding.
        match = self._pattern.match(line)
        if match is None:
            self._unknown.append(line)
            return

        values = match.groupdict()
        # Fix spaces in file/link names.
        values['name'] = values['name'].replace(br'\ ', b' ')

        is_link = False
        type = values['type']
        if type == b'd':
            is_file = False
            is_folder = True
        elif type == b'l':
            is_file = True
            is_folder = False
            is_link = True
        elif type == b'-':
            is_file = True
            is_folder = False
        else:
            # Non file/folder entries are not supported.
            self._unknown.append(line)
            return

        if values['target']:
            values['target'] = values['target'].replace(br'\ ', b' ')
            if not is_link:
                # This is not a link, but it has a name similar to link.
                # We recreate the original name.
                values['name'] = values['name'] + b' -> ' + values['target']

        timestamp = self._parseDate(values['date'])

        # FIXME:4015:
        # We assume the client want the line as UTF-8 encoded.
        name = force_unicode(values['name'])
        # Must be numerical.
        size = int(values['size'])
        owner = force_unicode(values['owner'])
        group = force_unicode(values['group'])
        path = self._path_builder(name)

        entry = FileAttributes(
            name=name,
            path=path,
            size=size,
            is_file=is_file,
            is_folder=is_folder,
            is_link=is_link,
            modified=timestamp,
            mode=0,
            hardlinks=1,
            owner=owner,
            group=group,
            )

        self.entries.append(entry)

    def _parseDate(self, candidate):
        """
        Parses the date/time information.

        Supported formats:
        * Jan 29 12:02
        * May 24  2016

        When the value cannot be parsed it will return 1.
        """
        # FIXME:4016:
        # For now we parse it as it would be local timezone,
        # but it can be wrong as server might be in a different timezone
        # and server shows its local timezone.

        match = self._recent_date_pattern.match(candidate)
        if match is None:
            # It does not look like a time for this year.
            match = self._old_date_pattern.match(candidate)
            if match is None:
                # It is not a patter for a past year.
                # Return a placeholder.
                return self._TIMESTAMP_NOT_FOUND

            values = match.groupdict()
            month = self._translateMonth(values['month'])
            month_day = int(values['day'])
            year = int(values['year'])
            values = (
                year, month, month_day, 0, 0, 0, 0, 0, -1)
        else:
            values = match.groupdict()
            month = self._translateMonth(values['month'])
            month_day = int(values['day'])
            hour = int(values['hour'])
            minute = int(values['minute'])

            today = datetime.today()
            values = (
                today.year,
                month,
                month_day,
                hour,
                minute,
                0,
                0,
                0,
                -1
                )

        if values[0] < 1001:
            # Just some code to get the same behaviour on py2 and py3.
            # Should be removed once we are only on Py3.
            return self._TIMESTAMP_UNDERFLOW

        try:
            return time.mktime(values)
        except (OverflowError, ValueError):
            # On Py2 we get underflow,
            # But py3 will be happy with any time.
            return self._TIMESTAMP_UNDERFLOW

    def _translateMonth(self, name):
        """
        Translates month name into it's numeric code.
        It only supports English month names.

        In case the month name cannot be parsed it will
        return 1 (for January).
        """
        month_name = name.lower().strip()
        # FIXME:4017:
        # Add support for localized month names.
        # https://tools.ietf.org/html/rfc2640#page-3
        month = 1
        if month_name == b'jan':
            month = 1
        elif month_name == b'feb':
            month = 2
        elif month_name == b'mar':
            month = 3
        elif month_name == b'apr':
            month = 4
        elif month_name == b'may':
            month = 5
        elif month_name == b'jun':
            month = 6
        elif month_name == b'jul':
            month = 7
        elif month_name == b'aug':
            month = 8
        elif month_name == b'sep':
            month = 9
        elif month_name == b'oct':
            month = 10
        elif month_name == b'nov':
            month = 11
        elif month_name == b'dec':
            month = 12

        return month


def _is_code(lines, code):
    """
    Check that response from lines, is the response `code.`
    """
    if lines[0].startswith(code + ' ') and len(lines) == 1:
        # A single line message.
        return True

    if len(lines) == 1:
        # Not the expected code and we have a single line message.
        return False

    if lines[0].startswith(code + '-') and lines[-1].startswith(code + ' '):
        # A multi line message.
        return True

    return False


def _is_auth_accepted(lines):
    """
    Check for response `lines` and return True if user is successfully
    authenticated and no more commands are needed.

    Line starting with `CODE-custom message` are multi line responses.
    """
    if _is_code(lines, '230'):
        # Sufficient credentials provided to grant the user access.
        return True

    if _is_code(lines, '232'):
        # Similar to 230. Used for FTPS.
        # The server is willing to allow the user access based on
        # previously exchanged security data.
        return True

    return False


def _need_ACCT(lines):
    """
    Return `True` if ACCT is needed.
    """
    return _is_code(lines, '332')


def _need_PASS(lines):
    """
    Return `True` if PASS is needed.
    """
    return _is_code(lines, '331')
