# Copyright (c) 2011 Adi Roiban.
# See LICENSE for details.
"""
Configuration for FTP service.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
from zope.interface import implementer

from chevah.server.commons.interface import (
    IFTPServiceOptions,
    IFTPSIServiceOptions,
    )
from chevah.server.commons.exception import ServerError
from chevah.server.commons.observer import Signal
from chevah.server.commons.service_configuration import (
    ConfigurationWithConnections,
    )
from chevah.server.commons.ssl import (
    SSLConfigurationSectionMixin,
    )
from chevah.server.configuration.model import (
    WritableBoolean,
    WritableInteger,
    WritableNumber,
    WritableString,
    WritableStringOrEmpty,
    )


@implementer(IFTPServiceOptions)
class FTPServiceConfigurationSection(
        ConfigurationWithConnections, SSLConfigurationSectionMixin):
    """
    Service configuration section options within the FTP service.
    """
    PROPERTIES_REQUIRING_RESTART = (
        # For now, FTP service has no extra needs-restart configuration.
        ) + SSLConfigurationSectionMixin.PROPERTIES_REQUIRING_RESTART

    banner = WritableString('banner')
    ftps_explicit_enabled = WritableBoolean(
        'ftps_explicit_enabled', default=True)
    ftps_force_secured_command_channel = WritableBoolean(
        'ftps_force_secured_command_channel', default=True)
    ftps_require_session_reuse = WritableBoolean(
        'ftps_require_session_reuse', default=True)
    enable_password_authentication = WritableBoolean(
        'enable_password_authentication')
    enable_ssl_certificate_authentication = WritableBoolean(
        'enable_ssl_certificate_authentication')
    idle_data_connection_timeout = WritableNumber(
        'idle_data_connection_timeout',
        default=30,
        allow_negative=False,
        allow_zero=False,
        )
    passive_address = WritableStringOrEmpty('passive_address')
    active_source_port = WritableInteger('active_source_port', default=0)
    ignore_ascii_data_type = WritableBoolean('ignore_ascii_data_type')
    ascii_data_type_as_default = WritableBoolean('ascii_data_type_as_default')

    def _parse_passive_port_range(self, raw_range):
        """
        Parse port range string and return a tuple or min and max port.
        """
        try:
            if type(raw_range) is tuple or type(raw_range) is list:
                ports = raw_range
            else:
                # For now it only supports "MIN - MAX" format.
                ports = raw_range.split('-')
                ports = (ports[0].strip(), ports[1].strip())

            min_port = int(ports[0])
            max_port = int(ports[1])
        except (ValueError, IndexError, TypeError):
            details = (
                'Passive port range should be specified by integer numbers.'
                )
            raise ServerError('20047', data={'details': details})

        return (min_port, max_port)

    @property
    def passive_port_range(self):
        '''Return the passive port range.'''
        raw_range = self._proxy.getString(
            self._section_name, 'passive_port_range')

        try:
            min_port, max_port = self._parse_passive_port_range(raw_range)
            return (min_port, max_port)
        except ServerError:
            # If the existing configuration is invalid, return a placeholder
            # which should trigger an error later.
            return (70001, 70000)

    @passive_port_range.setter
    def passive_port_range(self, value):
        """
        Set passive port range.
        """
        initial_value = self.passive_port_range

        value = self._parse_passive_port_range(value)
        min_port = value[0]
        max_port = value[1]

        raw_range = '%d - %d' % (min_port, max_port)
        self._proxy.setString(
            self._section_name,
            'passive_port_range',
            raw_range)

        self.notify(
            'passive_port_range',
            Signal(
                self, initial_value=initial_value, current_value=value))


@implementer(IFTPSIServiceOptions)
class FTPSIServiceConfiguration(FTPServiceConfigurationSection):
    """
    Service configuration section options within the FTPS Implicit service.

    Just a marker interface for now.
    """
