# Copyright (c) 2010 Adi Roiban.
# See LICENSE for details.
"""
Chevah FTP factory module.
"""

from __future__ import unicode_literals

from __future__ import absolute_import
from twisted.cred.portal import Portal

from chevah.server.authentication.credential_checker import (
    ServiceCredentialChecker,
    )
from chevah.server.commons.exception import ServerException
from chevah.server.commons.factory import ServiceFactoryBase
from chevah.server.commons.realm import ServiceRealm
from chevah.server.commons.ssl import SSLServiceFactoryMixin
from chevah.server.ftp.protocol import (
    FTPInternalErrorProtocol,
    FTPProtocol,
    FTPSIProtocol,
    FTPOverflowProtocol,
    )


class FTPFactory(ServiceFactoryBase, SSLServiceFactoryMixin):
    """
    Factory for creating FTP protocol instances.
    """

    protocol = FTPProtocol
    overflowProtocol = FTPOverflowProtocol
    internalErrorProtocol = FTPInternalErrorProtocol

    allowAnonymous = False
    userAnonymous = 'anonymous'

    def __init__(self, service_configuration, on_error):
        super(FTPFactory, self).__init__(service_configuration, on_error)
        self.ssl_context_factory = None
        self._validateConfiguration()

        credential_checker = [
            ServiceCredentialChecker(self.service_configuration)]
        realm = ServiceRealm(self.service_configuration)
        self.portal = Portal(realm, credential_checker)

    def _configureProtocol(self, protocol):
        """
        See: `ServiceFactoryBase`.
        """
        protocol.portal = self.portal

    def startFactory(self):
        """
        See: `Factory`.
        """
        super(FTPFactory, self).startFactory()
        self._registerConfigurationChanges()

    def stopFactory(self):
        """
        Called before service is stopped.
        """
        self._deregisterConfigurationChanges()
        return super(FTPFactory, self).stopFactory()

    @property
    def passive_port_range(self):
        """
        Return the port range to use for passive connections.
        """
        min_port = self.service_configuration.passive_port_range[0]
        max_port = self.service_configuration.passive_port_range[1]

        if min_port >= 70000:
            raise ServerException(
                'Invalid value configured for "passive_port_range".')

        if max_port - min_port < 0:
            raise ServerException(
                'Starting port in the range should be smaller than ending '
                'port.')

        return min_port, max_port

    @property
    def passive_address(self):
        """
        Address used in PASV response.
        """
        return self.service_configuration.passive_address

    @property
    def listen_address(self):
        """
        Address used to listen for incoming command connections.
        """
        return self.service_configuration.parent.address

    @property
    def active_source_port(self):
        """
        Source port used in PORT connections.
        """
        return self.service_configuration.active_source_port

    @property
    def ssl_enabled(self):
        """
        See ISSLServiceFactory
        """
        return self.service_configuration.ftps_explicit_enabled

    @property
    def ftps_force_secured_command_channel(self):
        return self.service_configuration.ftps_force_secured_command_channel

    @property
    def ftps_require_session_reuse(self):
        return self.service_configuration.ftps_require_session_reuse

    @property
    def enable_password_authentication(self):
        return self.service_configuration.enable_password_authentication

    @property
    def enable_ssl_certificate_authentication(self):
        # Workaround for long line.
        conf = self.service_configuration
        return conf.enable_ssl_certificate_authentication

    @property
    def ignore_ascii_data_type(self):
        return self.service_configuration.ignore_ascii_data_type

    @property
    def ascii_data_type_as_default(self):
        return self.service_configuration.ascii_data_type_as_default

    @property
    def idle_data_connection_timeout(self):
        """
        Timeout for data connection.

        See `idle_connection_timeout` for main control connection timeout.
        """
        return self.service_configuration.idle_data_connection_timeout

    @property
    def welcome_message(self):
        """
        Text displayed by FTP service at client connection.
        """
        return self.service_configuration.banner

    def _validateConfiguration(self):
        """
        Check that the current configuration is valid.
        """
        if (
            not self.enable_password_authentication and
            not self.enable_ssl_certificate_authentication
                ):
            # No authentication method enabled.
            raise ServerException(
                'No credentials type enabled. '
                'Enable password or SSL certificate based authentication.'
                )

        if (
            not self.enable_password_authentication and
            not self.ssl_enabled
                ):
            # Password authentication not enabled.
            raise ServerException(
                'Password-based authentication must be enabled when '
                'FTPS is not enabled.'
                )

        self.passive_port_range

    def _registerConfigurationChanges(self):
        """
        Register all configuration changes affecting the FTP service.

        This is factored in a function to allow overwriting in the implict
        FTPS service.
        """
        self.service_configuration.subscribe(
            'update-ftps_explicit_enabled', self._onFTPExplicitEnabledSet)

    def _deregisterConfigurationChanges(self):
        """
        Register all configuration changes affecting the FTP service.
        """
        self.service_configuration.unsubscribe(
            'update-ftps_explicit_enabled', self._onFTPExplicitEnabledSet)

    def _onFTPExplicitEnabledSet(self, event):
        """
        Action performed when ftps_explicit_enabled is changed.
        """
        human_readable_change = 'unknown'
        if event.new_value:
            # (re)enable explicit FTPS.
            self.ssl_context_factory = None
            self._configureSSL()
            human_readable_change = '(re)enabled'
        else:
            # disable FTPS.
            self.ssl_context_factory = None
            human_readable_change = 'disabled'

        self.emitEvent('10093', data={
            'service': self.service_configuration.parent.name,
            'state': human_readable_change,
            })


class FTPSIFactory(FTPFactory):
    """
    Factory for FTPS Implicit.
    """

    protocol = FTPSIProtocol

    def _registerConfigurationChanges(self):
        """
        Register all configuration changes affecting the FTPSI service.
        """
        return

    @property
    def ssl_enabled(self):
        """
        See ISSLServiceFactory.
        SSL is always enabled.
        """
        return True

    @property
    def ftps_force_secured_command_channel(self):
        return True
