# Copyright (c) 2010 Adi Roiban.
# See LICENSE for details.
"""
FTP server side protocol.

One instance of FTP protocol is created by the FTP Factory for each client
session.

Since ASCII is a subset of UTF-8, UTF-8 support is enabled by default.
Main RFCs:
 * http://tools.ietf.org/html/rfc959
 * http://tools.ietf.org/html/rfc3659
 * http://tools.ietf.org/html/rfc4217

Whenever possible file transfer related events should be emitted via the
FTP shell and not the protocol.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
from calendar import timegm
from datetime import datetime
from random import shuffle
import os
import posixpath
import stat

from OpenSSL import SSL
from OpenSSL._util import lib as openssl_lib

from twisted.internet import address, error as internet_error, reactor, defer
from twisted.internet.tcp import BaseClient
from twisted.python.failure import Failure
from twisted.python.filepath import Permissions
from twisted.protocols import ftp, policies
from twisted.protocols.basic import LineReceiver
from twisted.protocols.ftp import (
    AuthorizationError,
    decodeHostPort,
    FTP,
    RESPONSE,
    )
from zope.interface import implementer
import six
from six.moves import range

from chevah_compat import process_capabilities, system_users
from chevah.server import emitFromError, force_unicode, native_string
from chevah.server.commons.credential import (
    FTPPasswordCredentials,
    FTPSPasswordCredentials,
    FTPSSSLCertificateCredentials,
    )
from chevah.server.commons.exception import (
    ConnectionAborted,
    InsufficientAuthentication,
    ServerError,
    ServerException,
    )
from chevah.server.identity.avatar import PeerAvatar, ProcessAvatar
from chevah.server.commons.interface import IIdleTimeoutProtocol
from chevah.server.commons.ssl import (
    ChevahTLSMemoryBIOFactory,
    SSLConnectionMixin,
    )
from chevah.server.event.handler import emit_error_details
from chevah.server.ftp.shell import FTPShell, NLST_ON_FILE_MARKER
from chevah.server.proxy.proxy_protocol import proxify_v2


# constants
# response codes
# Keep response codes sorted ascending by code and sub code.
PBSZ_OK = '200.100'
PROT_OK_CLEAR = '200.101'
PROT_OK_PRIVATE = '200.102'
OPTS_UTF8_OK = '200.103'
ALREADY_AUTH = '200.104'
EPSV_ALL_OK = '200.105'
ENTERING_EPRT_MODE = '200.106'
CCC_OK = '200.107'
OPTS_MLST_OK = '200.108'
ALLO_NOOP = '202.100'
FEAT_OK = '211.100'
SET_MTIME_OK = '213.100'
SET_MFMT_OK = '213.101'
HELP_NOOP = '214.100'
USR_LOGGED_IN_SSL_PROCEED = '230.100'
AUTH_PROCEED = '234.100'
MLST_OK = '250.100'
REST_OK = '350.100'

CANT_OPEN_PASSIVE_DATA_CNX = '425.100'
CANT_OPEN_ACTIVE_DATA_CNX = '425.101'

OPTS_NOT_IMPLEMENTED = '501.100'
OPTS_MLST_INVALID = '501.101'
PROT_UNRECOGNIZED = '502.101'
NO_AUTH_METHOD_ENABLED = '502.103'
AUTH_NOT_ENABLED = '502.104'
AUTH_ALREADY = '503.100'
PBSZ_BAD = '503.101'
PROT_NOT_ALLOWED = '503.102'
PROT_DO_PBSZ_FIRST = '503.103'
EPSV_ALL_ACTIVE = '503.104'
AUTH_BAD = '504.100'
INTERNAL_ERROR = '505.100'
PROT_UNSUPORTED = '521.100'
SSL_COMMAND_REQUIRED = '521.101'
SSL_DATA_REQUIRED = '521.102'
SSL_CERT_REQUIRED = '521.103'
EPSV_UNKNOWN_PROTOCOL = '522.100'
SSL_REUSE_REQUIRED = '522.101'
CCC_ALREADY = '533.100'
CCC_NOT_ALLOWED = '534.100'
PSWD_FAIL = '537.100'
SET_MTIME_BAD = '550.100'
AUTH_FAIL = '551.100'

RESPONSE[PBSZ_OK] = '200 PBSZ=0 successful.'
RESPONSE[PROT_OK_CLEAR] = '200 Protection set to Clear.'
RESPONSE[PROT_OK_PRIVATE] = '200 Protection set to Private.'
RESPONSE[OPTS_UTF8_OK] = '200 Always in UTF8 mode.'
RESPONSE[ALREADY_AUTH] = '200 User is already authenticated.'
RESPONSE[EPSV_ALL_OK] = '200 EPSV ALL OK.'
RESPONSE[ENTERING_EPRT_MODE] = '200 EPRT OK.'
RESPONSE[CCC_OK] = '200 Clear Command Channel OK.'
RESPONSE[OPTS_MLST_OK] = '200 MLST OPTS %s'
RESPONSE[ALLO_NOOP] = '202 Command not implemented, superfluous at this site.'
RESPONSE[SET_MTIME_OK] = '213 File modification time set.'
RESPONSE[SET_MFMT_OK] = '213 Modify=%s; %s'
RESPONSE[HELP_NOOP] = '214 Contact support for any help.'
RESPONSE[USR_LOGGED_IN_SSL_PROCEED] = (
    '230 User logged in, authorized by security data exchange')
RESPONSE[AUTH_PROCEED] = '234 Proceed with negotiation.'
RESPONSE[FEAT_OK] = '211-Features:%s211 End'

RESPONSE[MLST_OK] = '250-Listing %s250 End'

RESPONSE[REST_OK] = '350 Restart position accepted (%s).'

RESPONSE[CANT_OPEN_PASSIVE_DATA_CNX] = (
    '425 Cannot open passive data connection.')
RESPONSE[CANT_OPEN_ACTIVE_DATA_CNX] = '425 Cannot open active data connection.'

RESPONSE[OPTS_NOT_IMPLEMENTED] = '501 Option not implemented.'
RESPONSE[OPTS_MLST_INVALID] = '501 Invalid MLST options'
RESPONSE[PROT_UNRECOGNIZED] = '502 Unrecognized PROT type (use C or P).'
RESPONSE[AUTH_NOT_ENABLED] = '502 FTP Security Extensions not enabled.'
RESPONSE[AUTH_ALREADY] = '503 Already using TLS.'
RESPONSE[PBSZ_BAD] = '503 PBSZ not allowed on unsecured control connection.'
RESPONSE[PROT_NOT_ALLOWED] = (
    '503 PROT not allowed on unsecured control connection.')
RESPONSE[PROT_DO_PBSZ_FIRST] = (
    '503 You must issue the PBSZ command prior to PROT.')
RESPONSE[EPSV_ALL_ACTIVE] = '503 EPSV ALL was enabled.'
RESPONSE[AUTH_BAD] = '504 Unknown AUTH type.'
RESPONSE[INTERNAL_ERROR] = '505 Internal server error.'
RESPONSE[PROT_UNSUPORTED] = '521 PROT %s unsupported (use C or P).'
RESPONSE[EPSV_UNKNOWN_PROTOCOL] = (
    '522 Network protocol not supported, use (%s).')
RESPONSE[CCC_ALREADY] = '533 Command channel is already cleared.'
RESPONSE[CCC_NOT_ALLOWED] = '534 Clear Command Channel not allowed.'
RESPONSE[SSL_COMMAND_REQUIRED] = '521 SSL/TLS required on the control channel.'
RESPONSE[SSL_DATA_REQUIRED] = '521 SSL/TLS required on the data channel.'
RESPONSE[SSL_CERT_REQUIRED] = (
    '521 Users are required to authenticate using a SSL certificates.')
RESPONSE[SSL_REUSE_REQUIRED] = (
    '522 SSL connection failed: session reuse required.')
RESPONSE[PSWD_FAIL] = '537 Changing own password failed.'
RESPONSE[NO_AUTH_METHOD_ENABLED] = (
    '502 No authentication method enabled on the server.')
RESPONSE[SET_MTIME_BAD] = '550 Could not set file modification time. %s'

# This is defined in Twisted as
# '550 Cannot rmd, %s is not a directory'
# But we use it for generic error.
RESPONSE[ftp.IS_NOT_A_DIR] = (
    '550 Requested action not taken: %s is not a directory.')
# In Twisted this is defined as `Transfer Complete.` but some clients are
# broken and look for lower case.
RESPONSE[ftp.TXFR_COMPLETE_OK] = '226 Transfer complete.'
# In Twisted this is `550 %s: Permission denied.`
RESPONSE[ftp.PERMISSION_DENIED] = '553 %s: Permission denied.'


# FIXME:6185:
# Some exceptions in twisted still have bytes on Py2,
# but this should be fine as we migrate to Py3.
# Convert everyting to text/unicode.
for key, value in list(RESPONSE.items()):
    key = six.text_type(key)
    value = six.text_type(value)
    RESPONSE[key] = value
    for member in list(dir(ftp)):
        value = six.text_type(getattr(ftp, member))
        if value not in RESPONSE:
            continue
        setattr(ftp, member, value)


def patched_collectSocketDetails(self):
    """
    Clean up references to the socket and its file descriptor.

    Upstream Twisted code assumes that at cleanup socket or fileno are always
    set.
    """
    try:
        del self.socket
    except AttributeError:
        pass

    try:
        del self.fileno
    except AttributeError:
        pass


BaseClient._collectSocketDetails = patched_collectSocketDetails


class SSLReuseRequiredError(ftp.FTPCmdError):
    """
    This is an error raised when TLS reuse is required, but data connection
    is not reusing the TLS session.
    """
    def __init__(self):
        self.errorCode = SSL_REUSE_REQUIRED
        super(SSLReuseRequiredError, self).__init__()


@implementer(IIdleTimeoutProtocol)
class FTPProtocol(FTP, SSLConnectionMixin):
    """
    Extending twisted.protocols.ftp to implement explicit FTPS.

    It also fixed a few bugs for Chevah.
    """

    # FTP server-side command and data channel have the same event IDs.
    _TLS_OK_EVENT = '10100'
    _TLS_FAIL_EVENT = '10088'

    PUBLIC_COMMANDS = ['CLNT', 'FEAT', 'HELP', 'LANG', 'OPTS', 'SYST', 'QUIT']
    _PUBLIC_FTPS_COMMANDS = ['AUTH', 'CCC', 'PBSZ', 'PROT']

    _reactor = reactor
    _scheduler = reactor

    def __init__(self):
        self.factory = None
        self.shell = None
        self.state = self.UNAUTH
        self.logout = None
        self._pbsz = False
        self._protected_data_requested = False
        self._protected_command_requested = False
        self._avatar = ProcessAvatar()
        self._epsv_all_active = False
        # Value of the last REST command.
        self._transfer_offset = 0
        self.binary = True
        # Set by DTPFactory when client connects.
        self.dtpInstance = None
        # Helper for testing.
        self._dtp_stop_deferred = None

        # By default we return everything.
        self._mlsx_selected_facts = self._MLSx_DEFAULT_FACTS[:]
        self._tls_handshake_completed = None

    def debug(self, message):
        """
        Do nothing by default.

        This is overwritten by the factory.
        """

    @property
    def avatar(self):
        """
        See: `RemoteConnectionMixin.avatar`.
        """
        return self._avatar

    @property
    def workingDirectory(self):
        """
        Return segments for current working directory.

        This is here to help with migrating filesystem state inside FTP shell.

        Twisted FTP protocol still use it, but only for reading.
        All writing was replaced with Chevah code.
        """
        return self.shell.working_segments

    @property
    def is_ftps_command_active(self):
        '''Return `True` if current command connection is using SSL.'''
        return self._protected_command_requested

    @property
    def have_valid_authentication_channel(self):
        """
        Return True if command channel is prepared for sending
        authentication requests.
        """
        if not self.factory.ftps_force_secured_command_channel:
            return True

        return self.is_ftps_command_active

    @property
    def dtpTimeout(self):
        """
        Raise a warning when dtpTimeout is used.
        """
        raise RuntimeError('Please use idle_data_connection_timeout.')

    @property
    def idle_data_connection_timeout(self):
        """
        Timeout for data connection.
        """
        return self.factory.idle_data_connection_timeout

    @property
    def _timeoutCall(self):
        """
        Return the timeout call.

        Used only for testing.
        """
        return self._TimeoutMixin__timeoutCall

    def connectionMade(self):
        """
        Called when a new command connection was made.
        """
        self._avatar = PeerAvatar(
            self.transport.getPeer(),
            component=self.factory,
            )
        self.emitEvent('10033')

        self._resetSSLStatus()
        self.state = self.UNAUTH
        self.setTimeout(self.factory.idle_connection_timeout)

        # Update configurations as provided by the factory.
        self.binary = not self.factory.ascii_data_type_as_default

        if self.factory.ignore_ascii_data_type:
            # If ASCII should be ignored, binary should always be the
            # default.
            self.binary = True

        if self._tls_handshake_completed is None:
            self.reply(ftp.WELCOME_MSG, self.factory.welcome_message)

    def replyIfConnected(self, key, *args):
        """
        Send a reply to command channel, ignoring the message when
        channel is disconnected.
        """
        have_text = True
        for a in args:
            if isinstance(a, six.binary_type):
                have_text = False
        if have_text:
            message = RESPONSE[key] % args
        else:
            # Maybe we have some legacy Py2 code in  which args are bytes.
            message = RESPONSE[key].encode('utf-8') % args
        self.sendLineIfConnected(message)

    def sendLineIfConnected(self, message):
        """
        Send a response on command channel, ignoring the message when
        channel is disconnected as there is no longer a client to receive
        our message.
        """
        if not self.transport:
            return

        self.sendLine(message)

    def sendLine(self, line):
        """
        Send a line to data channel.

        If line is I{unicode}, it will be converted to I{str}.
        """
        self.debug('> %s' % (force_unicode(line),))

        if isinstance(line, six.text_type):
            line = line.encode('utf-8')
        self.transport.write(line + b'\r\n')

    def _resetSSLStatus(self):
        '''Reset SSL/TLS related status.'''
        self._pbsz = False
        self._protected_data_requested = False
        self._protected_command_requested = False

    def connectionLost(self, reason):
        """
        Called when the current command connection was lost.
        """
        self.connected = False

        self._tlsConnectionLost(reason)

        data = {
            'details': force_unicode(reason.value),
            'encryption': self._getEncryptionDescription(),
            'certificate': self._getPeerCertificateDescription(),
            }
        if (isinstance(reason, Failure) and
                isinstance(reason.value, SSL.Error)):
            family = reason.value.args[0][0][1]
            if family == native_string('SSL3_GET_CLIENT_CERTIFICATE'):
                # Emit a dedicated event to help investigation this
                # common issue.
                self.emitEvent('10014', data=data)

        if reason.check(internet_error.ConnectionDone):
            self.emitEvent('10034', data=data)
        else:
            # Connection closed with a failure.
            self.emitEvent('10042', data=data)

        # This will cause the DTP to be closed after the command channel.
        self.cleanupDTP()

        self.setTimeout(None)
        if hasattr(self.shell, 'logout') and self.shell.logout is not None:
            self.shell.logout()

        # We don't reset the avatar as it will be used when DTP is closed
        # a bit later.
        self.shell = None
        self.transport = None

    def timeoutConnection(self):
        """
        Called when command connection times out.
        """
        self.emitEvent('10084')
        self.transport.loseConnection()

    def rawDataReceived(self, data):
        """
        FTP protocol should not use raw data.
        """
        raise RuntimeError('FTP protocol should not receive raw data.')

    def lineReceived(self, line):
        """
        Re-implement lineReceived so that we can implement logging
        for internal server errors.

        It returns a deferred for debugging.
        """
        self.resetTimeout()
        # We don't want to receive other commands before processing the
        # current command.
        self.pauseProducing()

        def eb_process_failed(failure, line):
            """
            Called when processing the previous FTP command has failed.

            `line` is already force converted to Unicode.
            """
            self.cleanupDTP()
            if failure.check(ftp.CmdNotImplementedError):
                # Extra logging for unknown commands... but preventing flooding
                # the event for long commands or trying to decode random data.
                data = {'command': line[:100]}
                self.emitEvent('10019', data=data)

            if failure.check(ftp.FTPCmdError):
                self.sendLineIfConnected(failure.value.response())
                return

            if failure.check(TypeError):
                message = failure.value.args[0]
                if 'takes at least' in message:
                    # This is Py2 error.
                    error_message = 'Too few arguments for'
                elif 'required positional argument' in message:
                    # This is PY3 error.
                    error_message = 'Too few arguments for'
                elif message.find('takes exactly') != -1:
                    error_message = 'Invalid arguments for'
                else:
                    return failure

                self.replyIfConnected(
                    ftp.SYNTAX_ERR, '%s %s.' % (error_message, cmd,))
                return

            # This is an unknown error so we just forward it.
            return failure

        def eb_internal_error(failure, line):
            """
            Emit an internal error event.
            """
            error_details = force_unicode(failure.getBriefTraceback())
            data = {
                'line': line,
                'details': force_unicode(error_details),
                }
            self.emitEvent('10016', data=data)
            self.replyIfConnected(INTERNAL_ERROR)

        def cb_process_succeeded(result):
            """
            Called when previous command was successfully processed.
            """
            if result is None:
                # Nothing to send back.
                return

            if isinstance(result, tuple):
                return self.replyIfConnected(*result)

            return self.replyIfConnected(result)

        def bb_command_processed(result_or_failure, command):
            """
            Called after current command was processed, successfully or not.
            """
            # Ask for more FTP commands only if we are still connected.
            if self.transport and self.connected and not self.disconnected:
                self.resumeProducing()
            return result_or_failure

        line = force_unicode(line)
        spaceIndex = line.find(' ')
        if spaceIndex != -1:
            cmd = line[:spaceIndex]
            args = (line[spaceIndex + 1:],)
        else:
            cmd = line
            args = ()

        d = defer.maybeDeferred(self.processCommand, cmd, *args)
        d.addCallback(cb_process_succeeded)
        d.addErrback(eb_process_failed, line)
        d.addBoth(bb_command_processed, cmd)
        d.addErrback(eb_internal_error, line)
        return d

    def processCommand(self, cmd, *params):
        """
        Process FTP commands `cmd` which is Unicode..

        If FTPS is enabled and secure connections are forced, make sure
        we have a secure channel before executing commands.
        """
        cmd = cmd.upper()
        self.debug('< %r %s' % (cmd, force_unicode(' '.join(params))))

        if cmd in self.PUBLIC_COMMANDS:
            method = getattr(self, "ftp_" + cmd, None)
            if method is not None:
                return method(*params)

        if self.factory.ssl_enabled:
            if cmd in self._PUBLIC_FTPS_COMMANDS:
                method = getattr(self, "ftp_" + cmd, None)
                return method(*params)

            if cmd in ('USER', 'PASS'):
                if not self.have_valid_authentication_channel:
                    self.emitEvent('10035')
                    return SSL_COMMAND_REQUIRED
            elif cmd in ('PASV', 'EPSV', 'PORT', 'EPRT'):
                if (
                    self.factory.ftps_force_secured_command_channel and
                    not self._protected_data_requested
                        ):
                    self.emitEvent('10036')
                    return SSL_DATA_REQUIRED

        # Handle RNFR -> RNTO order.
        if self.state == self.AUTHED and cmd == 'RNTO':
            return ftp.BAD_CMD_SEQ, 'RNFR required before RNTO.'

        if self.state == self.RENAMING:
            if cmd == 'RNTO':
                return self.ftp_RNTO(*params)
            else:
                return ftp.BAD_CMD_SEQ, "RNTO required after RNFR"

        if self.state == self.INAUTH:
            if cmd == 'PASS':
                return self.ftp_PASS(*params)
            else:
                return ftp.BAD_CMD_SEQ, "PASS required after USER"

        if self.state == self.AUTHED:
            method = getattr(self, "ftp_" + cmd, None)
            if method is not None:
                return method(*params)
            return defer.fail(ftp.CmdNotImplementedError(cmd))

        # If we reach here we are not authenticated and no public command
        # was requested.
        # Basically this is UNAUTH.
        if cmd == 'PASS':
            return ftp.BAD_CMD_SEQ, "USER required before PASS"

        if cmd == 'USER':
            return self.ftp_USER(*params)

        # Finally inform that session is not authenticated.
        return ftp.NOT_LOGGED_IN

    def ftp_FEAT(self):
        """
        Advertise the features supported by the server.

        http://tools.ietf.org/html/rfc2389

        FTP implementations which support FEAT MUST include in the response
        to the FEAT command all properly documented FTP extensions beyond
        those commands and mechanisms described in RFC 959.
        """
        mlst_feat = []
        for fact in self._MLSx_FACTS:
            if fact in self._mlsx_selected_facts:
                mlst_feat.append(fact + '*')
            else:
                mlst_feat.append(fact)
        # Separated by `;` and ending with `;`.
        mlst = '%s;' % ';'.join(mlst_feat)

        # FEAT and OPTS are not advertised in the FEAT command.
        features = [
            '',  # Force a new line at the beginning of features.
            'AUTH SSL',
            'AUTH TLS',
            'CCC',
            'CLNT',
            'EPRT',
            'EPSV',
            'LANG EN*',
            'MDTM',
            'MFMT',
            'MLST %s' % (mlst,),
            'PASV',
            'PBSZ',
            'PROT',
            'REST STREAM',
            'SITE IDLE',
            'SITE UTIME',
            'SIZE',
            'TVFS',
            'TYPE A;I;U',
            'UTF8',
            ]

        if process_capabilities.os_family == 'posix':
            features.append('SITE CHMOD')

        # Each multi-line start with a space.
        ftp_delimiter = self.delimiter.decode('ascii')
        feat_delimiter = ftp_delimiter + ' '

        response = feat_delimiter.join(features) + ftp_delimiter

        return self.reply(FEAT_OK, response)

    def ftp_ALLO(self, ignored):
        """
        ALLOCATE (ALLO)

        This command may be required by some servers to reserve
        sufficient storage to accommodate the new file to be
        transferred.  The argument shall be a decimal integer
        representing the number of bytes (using the logical byte
        size) of storage to be reserved for the file.  For files
        sent with record or page structure a maximum record or page
        size (in logical bytes) might also be necessary; this is
        indicated by a decimal integer in a second argument field of
        the command.  This second argument is optional, but when
        present should be separated from the first by the three
        Telnet characters <SP> R <SP>.  This command shall be
        followed by a STORe or APPEnd command.  The ALLO command
        should be treated as a NOOP (no operation) by those servers
        which do not require that the maximum size of the file be
        declared beforehand, and those servers interested in only
        the maximum record or page size should accept a dummy value
        in the first argument and ignore it.
        """
        # We don't support any pre-transfer allocation..
        return ALLO_NOOP

    @defer.inlineCallbacks
    def ftp_AUTH(self, mode):
        """
        Set up secure control channel.

        http://tools.ietf.org/html/rfc4217
        """
        arg = mode.upper().strip()
        if self.is_ftps_command_active:
            defer.returnValue(AUTH_ALREADY)

        # Start by not having TLS.
        self._protected_command_requested = False

        if not self.factory.ssl_context_factory:
            defer.returnValue(AUTH_NOT_ENABLED)

        if arg not in ('TLS', 'TLS-C', 'SSL', 'TLS-P'):
            self._protected_command_requested = False
            defer.returnValue(AUTH_BAD)

        self.beforeTLSHandshake('command')
        # From RFC-4217: "As the SSL/TLS protocols self-negotiate
        # their levels, there is no need to distinguish between SSL
        # and TLS in the application layer".

        self.emitEvent('10024')
        # Inform the client that TLS will start.
        # This is send before the actual switch to TLS.
        self.reply(AUTH_PROCEED)
        try:
            self.transport.startTLS(self.factory.ssl_context_factory)
            # Receiving data over FTP command is paused while we handle
            # the current command.
            # Resume receiving to receive the TLS handshake.
            self.resumeProducing()
            yield self.waitHandshakeCompleted()

        except Exception as error:
            self._tlsConnectionLost(Failure(error))
            self.transport.loseConnection()
            # No need to send any response, since we just disconnect on
            # this type of errors.
            return

        self._protected_command_requested = True
        # We don't send any extra response, since the response was sent
        # before starting to do the switch.
        return

    def ftp_CCC(self):
        '''
        When a server receives the CCC command, it should behave as follows:

        If the server does not accept CCC commands (or does not understand
        them), then a 500 reply should be sent.

        Otherwise, if the control connection is not protected with TLS,
        then a 533 reply should be sent.

        Otherwise, if the server does not wish to allow the control
        connection to be cleared at this time, then a 534 reply should be
        sent.

        Otherwise, the server is accepting the CCC command and should do
        the following:

         o  Send a 200 reply.

         o  Shutdown the TLS session on the socket and leave it open.

         o  Continue the control connection in plaintext, expecting the
            next command from the client to be in plaintext.

         o  Not accept any more PBSZ or PROT commands.  All subsequent
            data transfers must be protected with the current PROT
            settings.
        '''
        if not self.is_ftps_command_active:
            # Already cleared.
            self.emitEvent('10086')
            return CCC_ALREADY

        # CCC is not supported since version 5.
        self.emitEvent('10087')
        return (CCC_NOT_ALLOWED,)

    def ftp_CDUP(self):
        """
        CHANGE TO PARENT DIRECTORY (CDUP)

        This command is a special case of CWD, and is included to
        simplify the implementation of programs for transferring
        directory trees between operating systems having different
        syntaxes for naming the parent directory.  The reply codes
        shall be identical to the reply codes of CWD.
        """
        # Rely on ftp_CWD to simplify the code.
        return self.ftp_CWD('..')

    def ftp_XCUP(self):
        """
        Obsolete command for CDUP from RFC-775
        """
        return self.ftp_CDUP()

    def ftp_CLNT(self, name):
        """
        Client will send its name.

        We just ignore it.
        """
        return ftp.CMD_OK

    def ftp_HELP(self, ignored=''):
        """
        HELP (HELP)

        This command shall cause the server to send helpful
        information regarding its implementation status over the
        control connection to the user.  The command may take an
        argument (e.g., any command name) and return more specific
        information as a response.  The reply is type 211 or 214.
        It is suggested that HELP be allowed before entering a USER
        command. The server may use this reply to specify
        site-dependent parameters, e.g., in response to HELP SITE.
        """
        return HELP_NOOP

    def ftp_CWD(self, path):
        """
        CHANGE WORKING DIRECTORY (CWD)

        This command allows the user to work with a different
        directory or dataset for file storage or retrieval without
        altering his login or accounting information.  Transfer
        parameters are similarly unchanged.  The argument is a
        pathname specifying a directory or other system dependent
        file group designator.
        """
        defer_result = self.shell.setWorkingPath(path)

        def cb_changed(result):
            return (ftp.REQ_FILE_ACTN_COMPLETED_OK,)
        defer_result.addCallback(cb_changed)
        return defer_result

    def ftp_XCWD(self, path):
        """
        Obsolete command for CWD from RFC-775
        """
        return self.ftp_CWD(path)

    def ftp_DELE(self, path):
        """
        DELETE (DELE)

        This command causes the file specified in the pathname to be
        deleted at the server site.
        """
        deferred = self.shell.removeFile(path)
        deferred.addCallback(lambda ign: (ftp.REQ_FILE_ACTN_COMPLETED_OK,))
        return deferred

    def ftp_EPRT(self, address):
        """
        Request for a extended active connection.

        The EPRT command allows for the specification of an extended address
        for the data connection.  The extended address MUST consist of the
        network protocol as well as the network and transport addresses.  The
        format of EPRT is:

        The following are sample EPRT commands:
            EPRT |1|132.235.1.2|6275|
            EPRT |2|1080::8:800:200C:417A|5282|

        The FTP protocol specifies to actually initiate the data channel only
        when requested as result of a data transfer command.
        Here we don't follow the RFC and try to initiate the data channel
        right away.
        """
        if self._epsv_all_active:
            return (EPSV_ALL_ACTIVE,)

        try:
            separator = address[0]
            address_parts = address.split(separator)
            protocol = address_parts[1]
            ip = address_parts[2]
            port = int(address_parts[3])
        except Exception:
            return self._badDataAddress(address, 'passive')

        if protocol not in ['1', '2']:
            # IPv4 and IPv6 are supported.
            return (EPSV_UNKNOWN_PROTOCOL, '1')

        data = {
            'protocol': protocol,
            'address': ip,
            'port': force_unicode(port),
            'mode': 'EPRT',
            }
        self.emitEvent('10062', data=data)

        return self._startActiveDTP(ip=ip, port=port, use_extended=True)

    def ftp_EPSV(self, protocol=None):
        """
        Request for a extended passive connection.

        When the EPSV command is issued with no argument, the server will
        choose the network protocol for the data connection based on the
        protocol used for the control connection.  However, in the case of
        proxy FTP, this protocol might not be appropriate for communication
        between the two servers.  Therefore, the client needs to be able to
        request a specific protocol.  If the server returns a protocol that
        is not supported by the host that will be connecting to the port, the
        client MUST issue an ABOR (abort) command to allow the server to
        close down the listening connection.  The client can then send an
        EPSV command requesting the use of a specific network protocol, as
        follows:

            EPSV<space><net-prt>

        If the requested protocol is supported by the server, it SHOULD use
        the protocol.  If not, the server MUST return the 522 error messages

        The EPSV command can be used with the argument "ALL" to
        inform Network Address Translators that the EPRT command (as well as
        other data commands) will no longer be used.  An example of this
        command follows:

        EPSV ALL

        Upon receipt of an EPSV ALL command, the server MUST reject all data
        connection setup commands other than EPSV - EPRT, PORT, PASV, et

        http://tools.ietf.org/html/rfc2428

        The FTP protocol specifies to actually initiate the data channel only
        when requested as result of a data transfer command.
        Here we don't follow the RFC and try to initiate the data channel
        right away.
        """
        if protocol:
            if protocol.lower() == 'all':
                # Protocol might be the special `ALL` value.
                self._epsv_all_active = True
                return (EPSV_ALL_OK,)

            if protocol not in ['1', '2']:
                # Only IPV4 and IPV6 is supported.
                return (EPSV_UNKNOWN_PROTOCOL, '1')

        return self._startPassiveDTP(use_extended=True)

    def _startPassiveDTP(self, use_extended):
        """
        Start a passive data transport channel.
        """
        ftp_mode = 'PASV'
        if use_extended:
            ftp_mode = 'EPSV'

        # If we have a DTP port set up, remove it as start again
        # the DTP connection. This is similar to vsftpd implementation.
        self.cleanupDTP()
        self.dtpFactory = DTPFactory(
            pi=self,
            reactor=self._reactor,
            use_ssl=self._protected_data_requested,
            )
        try:
            self.dtpPort = self._getPassivePort(self.dtpFactory)
        except ServerException as error:
            self.cleanupDTP()
            data = {
                'mode': ftp_mode,
                'details': error.message,
                }
            self.emitEvent('10015', data=data)
            return (CANT_OPEN_PASSIVE_DATA_CNX,)

        our_peer = self.dtpPort.getHost()
        port = our_peer.port

        if use_extended:
            mode = ftp.ENTERING_EPSV_MODE
            encoded_port = force_unicode(port)
            host = our_peer.host
        else:
            mode = ftp.ENTERING_PASV_MODE
            host = self._getPassiveAddress()
            encoded_port = ftp.encodeHostPort(host, port)

        # Inform the client what port (and maybe address) to use for passive
        # connection.
        self.reply(mode, encoded_port)

        data = {
            'address': force_unicode(host),
            'port': force_unicode(port),
            'mode': ftp_mode,
            }
        self.emitEvent('10022', data=data)

        def cb_stop_listening(result):
            """
            When we are in passive mode, we need to stop the server for
            listening on further passive connections.
            """
            self._dtp_stop_deferred = self.dtpPort.stopListening()

        self.dtpFactory.deferred.addCallback(cb_stop_listening)

        def eb_pasv_failure(failure):
            """
            Handle errors while initializing the data connection.

            These are usually timeout errors.

            We have already informed the client that connection
            was initialized so there is no way to inform the client about
            the failure since FTP protocol support a single response per
            command.
            """
            self.cleanupDTP()
            emit_error_details(
                failure.value, '10023', data=data, emitter=self.emitEvent)

        self.dtpFactory.deferred.addErrback(eb_pasv_failure)
        return self.dtpFactory.deferred

    def _getPassivePort(self, dtp_factory):
        """
        Return a port for passive access, using L{self.passive_port_range}
        attribute.

        Raise an error if no port could be obtained.
        """
        # This will raise an exception for an invalid configuration.
        port_start, port_end = self.factory.passive_port_range

        # Try some random port usage.
        # Might makes it harder to exploit any future vulnerability.
        #
        # Very simple method for now.
        # We go through all the range, but by the time we try all,
        # some of the ports that were previously tested, might be now free.
        port_range = list(range(port_start, port_end + 1))
        shuffle(port_range)

        error = None
        for port_number in port_range:
            try:
                if dtp_factory.use_ssl:
                    factory = ChevahTLSMemoryBIOFactory(
                        contextFactory=self.factory.ssl_context_factory,
                        isClient=False,
                        wrappedFactory=dtp_factory)
                else:
                    factory = dtp_factory

                return self._reactor.listenTCP(
                    port_number,
                    factory,
                    interface=self.factory.listen_address,
                    )
            except internet_error.CannotListenError as listen_error:
                error = listen_error

        # If we can not obtain a port, just return None.
        raise ServerException('Ran out of ports in range %s - %s. %s' % (
            port_start, port_end, force_unicode(error)))

    def _getPassiveAddress(self):
        """
        Return the address advertised in PASV response.
        """
        if not self.factory.passive_address:
            return self.transport.getHost().host
        else:
            return force_unicode(self.factory.passive_address)

    def _startActiveDTP(self, ip, port, use_extended):
        """
        Initialize an active data transport channel.
        """
        # If we have a DTP port set up, remove it and start with a clean DTP.
        # This is similar to vsftpd implementation.
        self.cleanupDTP()

        # # This is a client-side connection, no need for Proxy Protocol v2.
        self.dtpFactory = DTPFactory(
            pi=self,
            peerHost=self._avatar.peer.host,
            use_ssl=self._protected_data_requested,
            use_pp2=False,
            )

        if self.dtpFactory.use_ssl:
            # It is strange, but for active connection the SSL
            # layer is still seen as a server.
            factory = ChevahTLSMemoryBIOFactory(
                contextFactory=self.factory.ssl_context_factory,
                isClient=False,
                wrappedFactory=self.dtpFactory,
                )
        else:
            factory = self.dtpFactory

        if use_extended:
            mode = ENTERING_EPRT_MODE
        else:
            mode = ftp.ENTERING_PORT_MODE

        source_address = self.factory.listen_address
        source_port = self.factory.active_source_port

        # Initiated event data with some preliminary info.
        data = {
            'address': ip,
            'port': force_unicode(port),
            'source_address': force_unicode(source_address),
            'source_port': force_unicode(source_port),
            }

        def cb_port_created(port):
            self.dtpPort = port
            # Chain the DTP deferred.
            return self.dtpFactory.deferred

        def cb_connected(ignored, mode):
            """
            Called when the data channel is connected.
            """
            # Update event data with the actual values of the socket.
            source_host = self.dtpPort.transport.getHost()
            data['source_address'] = force_unicode(source_host.host)
            data['source_port'] = force_unicode(source_host.port)
            self.emitEvent('10063', data=data)
            return (mode,)

        def eb_connection_failed(failure):
            """
            Called when failed to connect to client port.
            """
            self.cleanupDTP()
            emit_error_details(
                failure.value, '10064', data=data, emitter=self.emitEvent)
            return (CANT_OPEN_ACTIVE_DATA_CNX,)

        # Connect from a deferred wrapper to have uniform error handling.
        deferred = defer.maybeDeferred(
            self.__connect,
            ip, port, factory, source_address, source_port)

        deferred.addCallback(cb_port_created)
        # Set up DTP deferred handlers.
        deferred.addCallback(cb_connected, mode)
        deferred.addErrback(eb_connection_failed)
        return deferred

    def __connect(self, ip, port, factory, source_address, source_port):
        """
        Try to connect `factory` to destination ip:port as
        source_address:source_port.
        """
        if source_port == 0:
            # Listen on any port as non-root
            return self._reactor.connectTCP(ip, port, factory, timeout=30)

        bind_address = address.ReusableIPv4Address(
            type='TCP', host=source_address, port=source_port)

        if (process_capabilities.os_family != 'posix' or source_port > 1023):
            # We don't need to listen as root.
            return self._reactor.connectTCP(
                ip, port, factory, timeout=30, bindAddress=bind_address)

        # We are on Unix and required a privileged port.
        with system_users.executeAsUser(username='root'):
            return self._reactor.connectTCP(
                ip, port, factory, timeout=30, bindAddress=bind_address)

    def cleanupDTP(self):
        """
        Clean and close DTP channel.

        It is also called from DTP channel itself, on close.
        """
        if not self.dtpFactory:
            # No data connection yet. Nothing to clean.
            return

        dtp_factory = self.dtpFactory
        self.dtpFactory = None

        old_dtpPort = self.dtpPort
        self.dtpPort = None

        if old_dtpPort:
            if ftp.interfaces.IListeningPort.providedBy(old_dtpPort):
                if old_dtpPort.connected and not old_dtpPort.disconnecting:
                    self._dtp_stop_deferred = old_dtpPort.stopListening()

                if self._isDTPConnected():
                    self.dtpInstance.transport.abortConnection()

            elif ftp.interfaces.IConnector.providedBy(old_dtpPort):
                self._dtp_stop_deferred = old_dtpPort.disconnect()
            else:
                raise RuntimeError(
                    'dtpPort should be an IListeningPort or IConnector, '
                    'instead is %r' % (old_dtpPort,))

        dtp_factory.stopFactory()
        self.dtpInstance = None

    def _isDTPConnected(self):
        """
        Return `True` when DTP exists and is already connected.
        """
        return self.dtpInstance and self.dtpInstance.connected

    def _checkDTP(self, command):
        """
        Check that DTP was already requested.
        """
        if not self._isDTPConnected():
            raise ftp.BadCmdSequenceError(
                'PORT or PASV required before {}'.format(command))

    @defer.inlineCallbacks
    def _waitDTP(self, command):
        """
        Return a deferred that is fired when DTP is ready.
        For FTPS, this is called after TLS handshake.

        Return a failed deferred on errors.
        """
        # Some clients will initiate the TLS handshake only after we
        # inform that the data connection is ready.
        self.reply(ftp.DATA_CNX_ALREADY_OPEN_START_XFR)

        command_tls = self._getTLSConnection()
        if not command_tls:
            # We don't have TLS for main command.
            # So nothing to validate.
            # Maybe we are plain FTP or CCC was requested.
            return

        if not self.factory.ftps_require_session_reuse:
            # No need to enforce session reuse.
            return

        # Wait for data command handshake to be ready.
        yield self.dtpInstance.waitHandshakeCompleted()

        data_tls = self.dtpInstance._getTLSConnection()
        # See
        # https://www.openssl.org/docs/man1.1.1/man3/SSL_session_reused.html
        # FIXME:6481:
        # Try to use pyOpenSSL API.
        if openssl_lib.SSL_session_reused(data_tls._ssl) == 1:
            # All ok.
            return

        # We send the response over the command channel,
        # we close the data channel,
        # and raise the error, so normal data command handling is
        # aborted.
        self.emitEvent('10108', data={'command': command})
        # Since we raise an error, the common code should close the DTP
        # connection.
        # We close it here to make sure we don't accidentally transfer
        # any data.
        self.dtpInstance.transport.loseConnection()
        raise SSLReuseRequiredError()

    def ftp_LANG(self, code=None):
        """
        FTP command set to allow server-FTP process to determine in which
        language to present server greetings and the textual part of command
        responses.

        If a LANG command without a parameter is issued the server's default
        language will be used.
        """
        # For now we accept any language, while keeping EN as the current
        # language.
        # Well behaved clients should not ask non-EN language as EN is the
        # only language advertised in the FEAT response.
        return ftp.CMD_OK

    def ftp_MDTM(self, arguments):
        """
        File Modification Time (MDTM)

        The FTP command, MODIFICATION TIME (MDTM), can be used to determine
        when a file in the server NVFS was last modified.  This command has
        existed in many FTP servers for many years, as an adjunct to the REST
        command for STREAM mode, thus is widely available.  However, where
        supported, the "modify" fact that can be provided in the result from
        the new MLST command is recommended as a superior alternative.

        http://tools.ietf.org/html/rfc3659

        It also support setting the modification time.
        """
        parts = arguments.split(' ', 1)
        parsed_time = None
        if len(parts) == 1:
            # No spaces in the arguments. Just a simple path.
            path = arguments
        else:
            # Might have a request to set the time.
            parsed_time = self._parseFTPTime(parts[0])
            if parsed_time:
                path = parts[1]
            else:
                # Not a request to set the path.
                path = arguments

        if parsed_time:
            # Set the time.
            return self._setModificationTime(time=parsed_time, path=path)

        # Continue with returning the date.
        deferred = self.shell.getAttributes(path)

        def cb_stat(attributes):
            modified = self._getRFC3659DateTime(attributes.modified)
            return (ftp.FILE_STATUS, modified)

        deferred.addCallback(cb_stat)
        return deferred

    def _parseFTPTime(self, input):
        """
        Try to parse the FTP time at `input` and return the parsed datetime.
        Return `None` if time could not be parsed.
        """
        # Check without milliseconds.
        try:
            return datetime.strptime(input, '%Y%m%d%H%M%S')
        except ValueError:
            pass

        # Check with milliseconds.
        try:
            return datetime.strptime(input, '%Y%m%d%H%M%S.%f')
        except ValueError:
            pass

        # Failed to parse.
        return None

    def _setModificationTime(self, time, path, response=None):
        """
        Update the file at `path`, using the UTC `time`.
        """
        if response is None:
            # Use the default response format.
            response = (SET_MTIME_OK,)

        timestamp = timegm(time.timetuple()) + time.microsecond / 1e6

        deferred = self.shell.setAttributes(
            path, {'mtime': timestamp, 'atime': timestamp})

        deferred.addCallback(lambda _: response)
        return deferred

    def _setModificationTimeFromInput(self, input, path, response=None):
        """
        Update the file at `path`, using the `input` time as FTP time string.
        """
        parsed_time = self._parseFTPTime(input)
        if not parsed_time:
            return (SET_MTIME_BAD, 'Failed to parse the time.')

        return self._setModificationTime(
            time=parsed_time, path=path, response=response)

    def ftp_MFMT(self, arguments):
        """
        Modify Fact: Modification Time

        Used to modify the last modification time of an object in the NVFS.

        https://tools.ietf.org/html/draft-somers-ftp-mfxx-04
        """
        parts = arguments.split(' ', 1)
        if len(parts) != 2:
            return ftp.SYNTAX_ERR_IN_ARGS, 'MFMT requires 2 arguments'

        return self._setModificationTimeFromInput(
            input=parts[0],
            path=parts[1],
            # Use the specific response for the MFTM command.
            response=(SET_MFMT_OK, parts[0], parts[1]),
            )

    def ftp_MKD(self, path):
        """
        MAKE DIRECTORY (MKD)

        This command causes the directory specified in the pathname
        to be created as a directory (if the pathname is absolute)
        or as a subdirectory of the current working directory (if
        the pathname is relative).
        """
        deferred = self.shell.makeDirectory(path)
        deferred.addCallback(lambda _: (
            ftp.MKD_REPLY, self.shell.getAbsolutePath(path))
            )
        return deferred

    def ftp_XMKD(self, path):
        """
        Obsolete command for MKD from RFC-775
        """
        return self.ftp_MKD(path)

    def ftp_MODE(self, mode):
        """
        TRANSFER MODE (MODE)

        The argument is a single Telnet character code specifying
        the data transfer modes described in the Section on
        Transmission Modes.

        The following codes are assigned for transfer modes:
           S - Stream
           B - Block
           C - Compressed
        The default transfer mode is Stream.

        Not implemented yet.
        """
        mode = mode.lower()
        if mode == 's':
            return ftp.CMD_OK

        return (ftp.CMD_NOT_IMPLMNTD_FOR_PARAM, mode)

    def _parseListingPath(self, path):
        """
        Return the path and globbing expression used for directory listing.
        """
        path = path.strip()

        def parse_globbing(path):
            """
            Helper for checking if a `path` received from FTP command contains
            a wildcard Unix expression.
            """
            path_parts = path.rsplit('/', 1)
            if len(path_parts) == 1:
                path_candidate = ''
                glob_candidate = path
            else:
                path_candidate = path_parts[0] + '/'
                glob_candidate = path_parts[1]

            if (
                '*' in glob_candidate or
                '?' in glob_candidate or
                ('[' in glob_candidate and ']' in glob_candidate)
                    ):
                path = path_candidate
                glob = glob_candidate
            else:
                # Not a globbing expression.
                glob = None

            return path, glob

        # Various clients send arguments like -L or -al... and combinations.
        # We just ignore them.
        if path.lower() in ['-a', '-l', '-al', '-la']:
            # We have arguments as the single requested path.
            path = ''
        else:
            # We have arguments and a real path. Extract the path from the
            # raw request.
            first_possible_argument = path[:4].lower()
            if first_possible_argument in ['-al ', '-la ']:
                path = path[4:].strip()

            first_possible_argument = path[:3].lower()
            if first_possible_argument in ['-a ', '-l ']:
                path = path[3:].strip()

        path, glob = parse_globbing(path)

        return path, glob

    @defer.inlineCallbacks
    def ftp_LIST(self, path='.'):
        """
        LIST (LIST)

        This command causes a list to be sent from the server to the
        passive DTP.  If the pathname specifies a directory or other
        group of files, the server should transfer a list of files
        in the specified directory.  If the pathname specifies a
        file then the server should send current information on the
        file.  A null argument implies the user's current working or
        default directory.  The data transfer is over the data
        connection in type ASCII or type EBCDIC.  (The user must
        ensure that the TYPE is appropriately ASCII or EBCDIC).
        Since the information on a file may vary widely from system
        to system, this information may be hard to use automatically
        in a program, but may be quite useful to a human user.
        """
        self._checkDTP('LIST')

        def convert_to_FTP_attributes(attributes):
            """
            Return a list with FTP attributes to be sent to the data channel.
            """
            result = []

            for key in (
                'size', 'is_folder', 'mode', 'hardlinks',
                'modified', 'uid', 'gid'
                    ):
                value = getattr(attributes, key)

                # Convert the value into something expected by the FTP
                # protocol.
                if key in ['uid', 'gid']:
                    value = six.text_type(value)
                elif key == 'mode':
                    value = Permissions(value)

                result.append(value)

            return result

        path, glob = self._parseListingPath(path)

        _, iterator = yield self.shell.list(path, glob)

        yield self._waitDTP('LIST')

        for deferred in iterator:
            # Results are returned in batches.
            results = yield deferred
            for attributes in results:
                self.dtpInstance.sendListResponse(
                    self._encodeName(attributes.name),
                    convert_to_FTP_attributes(attributes),
                    )

        self.dtpInstance.transport.loseConnection()
        defer.returnValue((ftp.TXFR_COMPLETE_OK,))

    @defer.inlineCallbacks
    def ftp_NLST(self, path=''):
        """
        NAME LIST (NLST)

        This command causes a directory listing to be sent from
        server to user site.  The pathname should specify a
        directory or other system-specific file group descriptor; a
        null argument implies the current directory.  The server
        will return a stream of names of files and no other
        information.  The data will be transferred in ASCII or
        EBCDIC type over the data connection as valid pathname
        strings separated by <CRLF> or <NL>.  (Again the user must
        ensure that the TYPE is correct.)  This command is intended
        to return information that can be used by a program to
        further process the files automatically.  For example, in
        the implementation of a "multiple get" function.
        """
        self._checkDTP('NLST')

        yield self._waitDTP('NLST')

        path, glob = self._parseListingPath(path)
        try:
            iterator = yield self.shell.getMembersName(path, glob)
        except Exception:
            """
            RFC 959 specifies that an NLST request may only return directory
            listings. Thus, send nothing and just close the connection.

            @type results: L{Failure}
            @param results: The L{Failure} wrapping a L{FileNotFoundError}
                that occurred while trying to list the contents of a
                nonexistent directory.

            @rtype: C{tuple}
            @returns: A C{tuple} containing the status code for a successful
                transfer.
            """
            self.dtpInstance.transport.loseConnection()
            defer.returnValue((ftp.TXFR_COMPLETE_OK,))

        for deferred in iterator:
            # Names are returned in batches.
            names = yield deferred
            for name in names:
                if name is NLST_ON_FILE_MARKER:
                    # The path itself is returned to signal that the
                    # request is for a file.
                    member = path
                else:
                    member = posixpath.join(path, name)

                self.dtpInstance.sendLine(member.encode('utf-8'))

        self.dtpInstance.transport.loseConnection()
        defer.returnValue((ftp.TXFR_COMPLETE_OK,))

    # All available MLST fact supported by the server.
    _MLSx_FACTS = [
        'Size',
        'Type',
        'Perm',
        'Modify',
        'Unique',
        ]

    # For now we don't support any non-standard fact, so we enable everything
    # by default.
    _MLSx_DEFAULT_FACTS = _MLSx_FACTS

    def _getMLsxType(self, attributes, current=False):
        """
        Return the type as required for MLSx commands.

        Five values are possible for the type fact:

        file         -- a file entry
        cdir         -- the listed directory
        pdir         -- a parent directory (not implemented)
        dir          -- a directory or sub-directory
        OS.name=type -- an OS or file system dependent file type
                       (not implemented)
        """
        if attributes.is_link:
            return 'OS.unix=slink'
        elif attributes.is_folder:
            if current:
                return 'cdir'
            else:
                return 'dir'
        else:
            return 'file'

    def _getMLsxPermissions(self, attributes):
        """
        Return the permissions as required for MLSx commands.

        Return empty string when it has no permissions

        There are ten permission indicators currently defined.
        Many are meaningful only when used with a particular type of object.

        a - file - append is allowed
        c - dir - files can be created
        d - all - can be deleted
        e - dir - can traverse
        f - all - can rename
        l - dir - can list
        m - dir - new directories can be created
        r - file - can download
        w - file - can upload
        """
        mode = attributes.mode
        if not mode:
            return ''

        permissions = []

        if mode & stat.S_IRUSR:
            can_read = True
        else:
            can_read = False

        if mode & stat.S_IWUSR:
            can_write = True
        else:
            can_write = False

        if attributes.is_folder:
            if can_read:
                permissions.extend(['el'])
            if can_write:
                permissions.extend(['dcm'])
        else:
            if can_read:
                permissions.extend(['r'])
            if can_write:
                permissions.extend(['wad'])

        return ''.join(permissions)

    def _getRFC3659DateTime(self, timestamp):
        """
        Return the date and time as specified in RFC 3569.

        Time is returned in UTC as YYYYMMDDHHMMSS.sssss

        `timestamp` is local unix time.

        See: https://tools.ietf.org/html/rfc3659#section-2.3
        """
        utc_date_time = datetime.utcfromtimestamp(timestamp)
        return force_unicode(utc_date_time.strftime('%Y%m%d%H%M%S.%f'))

    def _getMLSxLine(self, attributes, current=False, only_name=False):
        """
        Return the line as required for MLSx commands.
        """
        facts = []
        for fact in self._mlsx_selected_facts:
            if fact == 'Type':
                facts.append('Type=%s;' % self._getMLsxType(
                    attributes, current=current))

            if fact == 'Perm':
                facts.append(
                    'Perm=%s;' % self._getMLsxPermissions(attributes))

            if fact == 'Size' and not attributes.is_folder:
                # Only add size to non-folders, as exemplified in the RFC.
                facts.append('Size=%s;' % attributes.size)

            if fact == 'Modify':
                modified = self._getRFC3659DateTime(attributes.modified)
                facts.append('Modify=%s;' % modified)

            if fact == 'Unique':
                facts.append('Unique=%s;' % attributes.node_id)

        if only_name:
            path = attributes.name
        else:
            path = attributes.path

        return ('%(facts)s %(path)s') % {
            'facts': ''.join(facts),
            'path': path,
            }

    def ftp_MLST(self, path=''):
        """
        Listings for Machine Processing of a path.

        MLST provides data about exactly the object named on its command line,
        and no others.

        If no argument is given MLST must return a listing giving information
        about the current working directory itself.

        Supported facts:
        * size
        * type
        * perm
        * modify
        * unique

        Unlike similar commands used to retrieve information,
        the response is sent over the command channel.
        Therefore, no data connection is required to issue this command.

        C> PWD
        S> 257 "/tmp" is current directory.
        C> MLst cap60.pl198.tar.gz
        S> 250- Listing cap60.pl198.tar.gz
        S>  Type=file;Size=1024990;Perm=r; /tmp/cap60.pl198.tar.gz
        S> 250 End
        """
        deferred = self.shell.getAttributes(path)

        def cb_got_attributes(attributes):
            """
            Called when we got the attributes for path.
            """
            response = '%(name)s\r\n%(response)s\r\n' % {
                'name': attributes.name,
                'response': self._getMLSxLine(attributes),
                }
            return (MLST_OK, response)

        deferred.addCallback(cb_got_attributes)
        return deferred

    @defer.inlineCallbacks
    def ftp_MLSD(self, path=''):
        """
        Listings for Machine Processing if a directory.

        Giving a pathname that exists but is not a directory as the argument
        to a MLSD command generates a 501 reply.

        If no argument is given then MLSD must return a listing of the
        contents of the current working directory,

        C> MLSD tmp
        S> 150 BINARY connection open for MLSD tmp
        D> Type=cdir;Modify=19981107085215;Perm=el; /tmp
        D> Type=pdir;Modify=19990112030508;Perm=el; ..
        D> Type=file;Size=25730;Modify=19940728095854;Perm=; capmux.tar.z
        D> Type=file;Size=1830;Modify=19940916055648;Perm=r; hatch.c
        S> 226 MLSD completed

        If a directory name is given then MLSD must return a listing of the
        contents of the named directory, otherwise it issues a 501 reply, and
        does not open a data connection.
        """
        self._checkDTP('MLSD')

        result = yield self.shell.list(path, accept_file=False)
        own_attributes, iterator = result

        yield self._waitDTP('MLSD')

        self.dtpInstance.sendLine(
            self._getMLSxLine(own_attributes, current=True))

        for deferred in iterator:
            # Results are returned in batches.
            results = yield deferred
            for attributes in results:
                self.dtpInstance.sendLine(
                    self._getMLSxLine(attributes, only_name=True))

        self.dtpInstance.transport.loseConnection()
        defer.returnValue((ftp.TXFR_COMPLETE_OK,))

    def ftp_NOOP(self):
        """
        NOOP (NOOP)

        This command does not affect any parameters or previously
        entered commands. It specifies no action other than that the
        server send an OK reply.
        """
        return (ftp.CMD_OK,)

    def ftp_OPTS(self, option):
        """
        Handle OPTS command.

        http://tools.ietf.org/html/rfc2389

        Response codes:
        * opts-good  - "200" SP response-message CRLF
        * opts-bad-temporary - "451" SP response-message CRLF
        * opts-bad-permanent - "501" SP response-message CRLF
        """
        # Handle UTF-8 option as described in
        # http://tools.ietf.org/html/draft-ietf-ftpext-utf-8-option-00
        if option.lower().startswith('utf8'):
            # We will always use UTF-8 so just look for the option name and
            # ignore the option
            return OPTS_UTF8_OK

        # For MLST OPT we are more permissive and will ignore fact that
        # we don't support or which are in wrong format.
        if option.lower().startswith('mlst'):
            return self._ftp_OPTS_MLST(option[5:])

        return OPTS_NOT_IMPLEMENTED

    def _ftp_OPTS_MLST(self, facts):
        """
        Called when OPTS MLST is called.
        """
        requested_facts = facts.lower()

        selected_facts = []
        for fact in self._MLSx_FACTS:
            if fact.lower() + ';' not in requested_facts:
                continue
            selected_facts.append(fact)

        if not selected_facts:
            return OPTS_MLST_INVALID

        self._mlsx_selected_facts = selected_facts
        active_fact = '%s;' % ';'.join(self._mlsx_selected_facts)
        return (OPTS_MLST_OK, active_fact)

    def ftp_PASV(self):
        """
        PASSIVE (PASV)

        This command requests the server-DTP to "listen" on a data
        port (which is not its default data port) and to wait for a
        connection rather than initiate one upon receipt of a
        transfer command.  The response to this command includes the
        host and port address this server is listening on.

        The FTP protocol specifies to actually initiate the data channel only
        when requested as result of a data transfer command.
        Here we don't follow the RFC and try to initiate the data channel
        right away.
        """
        if self._epsv_all_active:
            return (EPSV_ALL_ACTIVE,)

        return self._startPassiveDTP(use_extended=False)

    def ftp_PBSZ(self, line):
        '''Negotiate size of buffer for secure data transfer.

        For TLS/SSL the only valid value for the parameter is '0'.
        Any other value is accepted but ignored.

        http://tools.ietf.org/html/rfc4217
        '''
        if not self.is_ftps_command_active:
            return PBSZ_BAD
        else:
            self._pbsz = True
            return PBSZ_OK

    def ftp_PORT(self, address):
        """
        DATA PORT (PORT)

        The argument is a HOST-PORT specification for the data port
        to be used in data connection.  There are defaults for both
        the user and server data ports, and under normal
        circumstances this command and its reply are not needed.  If
        this command is used, the argument is the concatenation of a
        32-bit internet host address and a 16-bit TCP port address.
        This address information is broken into 8-bit fields and the
        value of each field is transmitted as a decimal number (in
        character string representation).  The fields are separated
        by commas.  A port command would be:
           PORT h1,h2,h3,h4,p1,p2
        where h1 is the high order 8 bits of the internet host
        address.

        The FTP protocol specifies to actually initiate the data channel only
        when requested as result of a data transfer command.
        Here we don't follow the RFC and try to initiate the data channel
        right away.
        """
        if self._epsv_all_active:
            return (EPSV_ALL_ACTIVE,)

        try:
            ip, port = decodeHostPort(address)
        except Exception:
            return self._badDataAddress(address, 'active')

        data = {
            'address': ip,
            'protocol': 1,
            'port': force_unicode(port),
            'mode': 'PORT',
            }
        self.emitEvent('10062', data=data)

        return self._startActiveDTP(ip=ip, port=port, use_extended=False)

    def _badDataAddress(self, address, kind):
        """
        Helper to raise event from a single place and returning the same
        error code.
        """
        self.emitEvent('10026', data={'address': address, 'kind': kind})
        if kind == 'active':
            return (CANT_OPEN_ACTIVE_DATA_CNX,)
        return (CANT_OPEN_PASSIVE_DATA_CNX,)

    def ftp_PROT(self, line):
        '''Data Connection Security Negotiation the

          The command defined in [RFC-2228] to negotiate data connection
          security is the PROT command.  As defined, there are four values
          that the PROT command parameter can take.
                'C' - Clear - neither Integrity nor Privacy
                'S' - Safe - Integrity without Privacy
                'E' - Confidential - Privacy without Integrity
                'P' - Private - Integrity and Privacy

          As TLS negotiation encompasses (and exceeds) the Safe /
          Confidential / Private distinction, only Private (use TLS) and
          Clear (don't use TLS) are used.

          For TLS, the data connection can have one of two security levels.
                1) Clear (requested by 'PROT C')
                2) Private (requested by 'PROT P')

        http://tools.ietf.org/html/rfc4217
        '''
        arg = line.upper()
        if not self.is_ftps_command_active:
            return PROT_NOT_ALLOWED
        elif not self._pbsz:
            return PROT_DO_PBSZ_FIRST
        elif arg == 'C':
            self._protected_data_requested = False
            return PROT_OK_CLEAR
        elif arg == 'P':
            self._protected_data_requested = True
            return PROT_OK_PRIVATE
        elif arg in ('S', 'E'):
            return PROT_UNSUPORTED, arg
        else:
            return PROT_UNRECOGNIZED

    def ftp_PWD(self):
        '''PRINT WORKING DIRECTORY (PWD)

        This command causes the name of the current working
        directory to be returned in the reply.
        '''
        self.emitEvent('10065')
        return (ftp.PWD_REPLY, self.shell.getWorkingPath())

    def ftp_XPWD(self):
        """
        Obsolete command for PDW from RFC-775
        """
        return self.ftp_PWD()

    def ftp_QUIT(self):
        """
        LOGOUT (QUIT)

        This command terminates a USER and if file transfer is not
        in progress, the server closes the control connection.  If
        file transfer is in progress, the connection will remain
        open for result response and the server will then close it.
        If the user-process is transferring files for several USERs
        but does not wish to close and then reopen connections for
        each, then the REIN command should be used instead of QUIT.

        An unexpected close on the control connection will cause the
        server to take the effective action of an abort (ABOR) and a
        logout (QUIT).
        """
        self.emitEvent('10066')
        self._resetSSLStatus()
        self.reply(ftp.GOODBYE_MSG)
        self.transport.loseConnection()
        self.disconnected = True

    def ftp_REST(self, offset):
        """
        Restart of Interrupted Transfer (REST)

        To avoid having to resend the entire file if the file is only
        partially transferred, both sides need some way to agree on where in
        the data stream to restart the data transfer.

        https://tools.ietf.org/html/rfc3659
        """
        try:
            offset = int(offset)
        except ValueError:
            self._transfer_offset = 0
            return (ftp.SYNTAX_ERR_IN_ARGS, 'Offset not a number.')

        if offset < 0:
            self._transfer_offset = 0
            return (ftp.SYNTAX_ERR_IN_ARGS, 'Offset should be positive.')

        self._transfer_offset = offset
        return (REST_OK, offset)

    @defer.inlineCallbacks
    def ftp_RETR(self, path):
        """
        RETRIEVE (RETR)

        This command causes the server-DTP to transfer a copy of the
        file, specified in the pathname, to the server- or user-DTP
        at the other end of the data connection.  The status and
        contents of the file at the server site shall be unaffected.
        """
        self._checkDTP('RETR')

        # At least AIX FTP client has issue with parsing the dir listing in
        # binary mode and sends an extra '\r' on the mget request.
        path = path.rstrip('\r')

        # FIXME:5038:
        # Call normal openForReading and then do a seek.
        deferred = self.shell.openForReading(
            path, offset=self._transfer_offset)

        public_path = self.shell.getAbsolutePath(path)

        # See if we can open the file with success.
        try:
            source_file = yield deferred

        except ftp.FTPCmdError as error:
            # Failed to open with an FTP error.
            # Return an FTP error and don't forget to reset the resume
            # request.
            self._transfer_offset = 0
            # The force unicode from here can be removed once we no longer
            # need py2.
            defer.returnValue((force_unicode(error.errorCode), public_path))

        except Exception:
            # Generic error.
            # Return an FTP error and don't forget to reset the resume
            # request.
            self._transfer_offset = 0
            defer.returnValue((ftp.REQ_ACTN_NOT_TAKEN, public_path))

        # File open with success.
        yield self._waitDTP('RETR')

        if not self.binary:
            cons = ToASCIIConsumerWrapper(self.dtpInstance)
        else:
            cons = self.dtpInstance

        # Start sending data.
        try:
            yield source_file.send(cons)
            # All transfer done.
            defer.returnValue((ftp.TXFR_COMPLETE_OK,))
        except Exception:
            # Emitting then event on errors is handled by the shell.
            # We just return the FTP protocol response here.
            defer.returnValue((ftp.CNX_CLOSED_TXFR_ABORTED,))
        finally:
            # After each transfer, make sure to reset a previous
            # resume request.
            self._transfer_offset = 0

    def ftp_RMD(self, path):
        """
        REMOVE DIRECTORY (RMD)

        This command causes the directory specified in the pathname
        to be removed as a directory (if the pathname is absolute)
        or as a subdirectory of the current working directory (if
        the pathname is relative).
        """
        deferred = self.shell.removeDirectory(path)
        deferred.addCallback(lambda ign: (ftp.REQ_FILE_ACTN_COMPLETED_OK,))
        return deferred

    def ftp_XRMD(self, path):
        """
        Obsolete command for RMD from RFC-775
        """
        return self.ftp_RMD(path)

    def ftp_RNFR(self, from_name):
        """
        RENAME FROM (RNFR)

        We only log RNTO. This is here just for documentation.

        This command specifies the old pathname of the file which is
        to be renamed.  This command must be immediately followed by
        a "rename to" command specifying the new file pathname
        """
        if not from_name:
            return (ftp.FILENAME_NOT_ALLOWED,)

        self._rename_from_name = from_name
        self.state = self.RENAMING
        return (ftp.REQ_FILE_ACTN_PENDING_FURTHER_INFO,)

    def ftp_RNTO(self, to_name):
        """
        RENAME TO (RNTO)

        This command specifies the new pathname of the file
        specified in the immediately preceding "rename from"
        command.  Together the two commands cause a file to be
        renamed.
        """
        from_name = self._rename_from_name
        del self._rename_from_name
        self.state = self.AUTHED

        if not to_name:
            return (ftp.FILENAME_NOT_ALLOWED,)

        deferred = self.shell.rename(from_name, to_name)
        deferred.addCallback(lambda _: (ftp.REQ_FILE_ACTN_COMPLETED_OK,))
        return deferred

    def ftp_SITE(self, parameters):
        """
        SITE PARAMETERS (SITE)

        This command is used by the server to provide services
        specific to his system that are essential to file transfer
        but not sufficiently universal to be included as commands in
        the protocol.  The nature of these services and the
        specification of their syntax can be stated in a reply to
        the HELP SITE command.

        It will delegate to ftp_SITE_* methods.
        """
        try:
            command, arguments = parameters.split(' ', 1)
            arguments = [arguments]
        except ValueError:
            command = parameters
            arguments = []

        method = getattr(self, "ftp_SITE_" + command.upper(), None)
        if method is None:
            return defer.fail(
                ftp.CmdNotImplementedError('SITE %s' % parameters))

        return method(*arguments)

    def ftp_SITE_CHMOD(self, arguments):
        """
        SITE CHMOD MODE PATH

        Only CHMOD is supported, and only on Unix.
        """
        if process_capabilities.os_family != 'posix':
            return defer.fail(ftp.CmdNotImplementedError(
                'SITE CHMOD is only supported on Unix/Linux.'))
        try:
            mode, path = arguments.split(' ', 1)
        except ValueError:
            return defer.fail(ftp.CmdNotImplementedForArgError(
                'SITE CHMOD requires MODE and PATH parameters.'))

        try:
            mode = int(mode, 8)
        except ValueError:
            return defer.fail(ftp.CmdNotImplementedForArgError(
                'SITE CHMOD requires MODE in octal format.'))

        try:
            self.shell.setAttributes(path, {'mode': mode})
            return (ftp.CMD_OK,)
        except Exception:
            return (ftp.REQ_ACTN_NOT_TAKEN, 'Failed to change mode.')

    def ftp_SITE_IDLE(self):
        """
        Alias for NOOP.
        """
        return self.ftp_NOOP()

    def ftp_SITE_UTIME(self, arguments):
        """
        SITE UTIME utc_date_time some path
        """
        parts = arguments.split(' ', 1)
        if len(parts) != 2:
            return ftp.SYNTAX_ERR_IN_ARGS, 'SITE UTIME requires 2 arguments'

        return self._setModificationTimeFromInput(
            input=parts[0], path=parts[1])

    def ftp_SITE_PSWD(self, arguments=None):
        """
        Allow an user to change its password.

        Implemented in a similar way to Serv-U / Titan FTP.

        Returns:
        * 200 - The command was successful.
        * 501 - Syntax error in parameters or arguments.
        * 537 - Access denied.

        `arguments` are required, but marked as option to fit with FTP SITE
        framework.

        SITE is not allowed for users which are not authenticated.
        """
        current = None
        new = None

        if not arguments:
            return (ftp.SYNTAX_ERR_IN_ARGS, 'SITE PSWD requires 2 arguments')

        arguments = arguments.strip()
        if ' ' not in arguments:
            return (ftp.SYNTAX_ERR_IN_ARGS, 'SITE PSWD requires 2 arguments')

        if '" "' in arguments:
            # We have quoted password.
            if not arguments.startswith('"') or not arguments.endswith('"'):
                return (
                    ftp.SYNTAX_ERR_IN_ARGS, 'Invalid quotation for arguments')

            # Remove first and last quotes.
            arguments = arguments.strip('"')
            parts = arguments.split('" "')
        else:
            # We don't have quoted arguments.
            parts = arguments.split(' ')

        if len(parts) != 2:
            return (
                ftp.SYNTAX_ERR_IN_ARGS, 'Invalid delimiter in arguments')

        current = parts[0]
        new = parts[1]

        if self.avatar.updatePassword(
                current, new, event_emitter=self.emitEvent):
            return ftp.CMD_OK
        else:
            # Failed to update password.
            # No event is emitted and the avatar will emit the event.
            return PSWD_FAIL

    def ftp_SIZE(self, path):
        """
        File SIZE

        The FTP command, SIZE OF FILE (SIZE), is used to obtain the transfer
        size of a file from the server-FTP process.  This is the exact number
        of octets (8 bit bytes) that would be transmitted over the data
        connection should that file be transmitted.  This value will change
        depending on the current STRUcture, MODE, and TYPE of the data
        connection or of a data connection that would be created were one
        created now.  Thus, the result of the SIZE command is dependent on
        the currently established STRU, MODE, and TYPE parameters.

        The SIZE command returns how many octets would be transferred if the
        file were to be transferred using the current transfer structure,
        mode, and type.  This command is normally used in conjunction with
        the RESTART (REST) command when STORing a file to a remote server in
        STREAM mode, to determine the restart point.  The server-PI might
        need to read the partially transferred file, do any appropriate
        conversion, and count the number of octets that would be generated
        when sending the file in order to correctly respond to this command.
        Estimates of the file transfer size MUST NOT be returned; only
        precise information is acceptable.

        http://tools.ietf.org/html/rfc3659
        """
        deferred = self.shell.getAttributes(path)

        def cb_got_attributes(attributes):
            """
            Called when we got the attributes for path.
            """
            return (ftp.FILE_STATUS, attributes.size)

        deferred.addCallback(cb_got_attributes)
        return deferred

    def ftp_STOR(self, path):
        """
        STORE (STOR)

        This command causes the server-DTP to accept the data
        transferred via the data connection and to store the data as
        a file at the server site.  If the file specified in the
        pathname exists at the server site, then its contents shall
        be replaced by the data being transferred.  A new file is
        created at the server site if the file specified in the
        pathname does not already exist.
        """
        return self._doUpload(path, self.shell.openForWriting)

    def ftp_APPE(self, path):
        """
         APPEND (with create) (APPE)

        This command causes the server-DTP to accept the data
        transferred via the data connection and to store the data in
        a file at the server site.  If the file specified in the
        pathname exists at the server site, then the data shall be
        appended to that file; otherwise the file specified in the
        pathname shall be created at the server site.
        """
        if self._transfer_offset > 0:
            # When we have a REST offset, APPE will behave like STOR, as
            # documented in the RFC
            return self.ftp_STOR(path)

        return self._doUpload(path, self._openForAppending)

    # FIXME:5038:
    # The offset argument should be removed.
    # It is not used here and is just to make the shared code happy.
    def _openForAppending(self, path, offset=0):
        """
        Adapter for calling the appending mode with the _doUpload shared
        code.
        """
        return self.shell.openForAppending(path)

    @defer.inlineCallbacks
    def _doUpload(self, path, open_callback):
        """
        Process the upload requests. STOR or APPE.
        """
        # The DTP is already started,
        # as otherwise wouldn't have reached this point.
        self._checkDTP('STOR/APPE')

        try:
            # FIXME:5038:
            # The offset should be remove and instead do an explicit seek
            # after the file is opened.
            target_file = yield open_callback(
                path, offset=self._transfer_offset)
            try:
                # Inform the client that we are ready.
                yield self._waitDTP('STOR/APPE')

                consumer = target_file.getConsumer()
                if not self.binary:
                    consumer = FromASCIIConsumerWrapper(consumer)

                yield self.dtpInstance.registerConsumer(consumer)
                target_file.close()
                defer.returnValue((ftp.TXFR_COMPLETE_OK,))

            except ftp.FTPCmdError as error:
                # Most probably the TLS session was not reused.
                target_file.fail(error)
                defer.returnValue((error.errorCode,) + error.errorMessage)

            except Exception as error:
                # Called when something happend during the transfer.
                target_file.fail(error)
                defer.returnValue((ftp.CNX_CLOSED_TXFR_ABORTED,))
        except Exception as error:
            """
            Called when failed to open the file.

            For known errors return the FTP error code.
            For all other, return an action not taken error.
            """
            if isinstance(error, ftp.FTPCmdError):
                public_path = self.shell.getAbsolutePath(path)
                # The force unicode from here can be removed once we no longer
                # need py2.
                defer.returnValue((force_unicode(error.errorCode), public_path))

            defer.returnValue((
                ftp.REQ_ACTN_NOT_TAKEN, 'Fail to open the file.'))
        finally:
            self._transfer_offset = 0

    def ftp_STRU(self, structure):
        """
        FILE STRUCTURE (STRU)

        The argument is a single Telnet character code specifying
        file structure described in the Section on Data
        Representation and Storage.

        The following codes are assigned for structure:
           F - File (no record structure)
           R - Record structure
           P - Page structure

        Only the file mode is supported..
        """
        structure = structure.lower()
        if structure == 'f':
            return ftp.CMD_OK

        return (ftp.CMD_NOT_IMPLMNTD_FOR_PARAM, structure)

    def ftp_SYST(self):
        """
        SYSTEM (SYSTE)

        This command is used to find out the type of operating
        system at the server.  The reply shall have as its first
        word one of the system names listed in the current version
        of the Assigned Numbers document http://tools.ietf.org/html/rfc943

        Since RFC 943 up to 1700 are obsolete and the new iana.org does
        not keep a record of SYSTEM NAMES

        For security reasons, we alway lie that we have a generic Unix system.

        The Unix system should also help the client in building correct paths.
        """
        return ftp.NAME_SYS_TYPE

    def ftp_TYPE(self, type):
        r"""
        REPRESENTATION TYPE (TYPE)

        The argument specifies the representation type as described
        in the Section on Data Representation and Storage.  Several
        types take a second parameter.  The first parameter is
        denoted by a single Telnet character, as is the second
        Format parameter for ASCII and EBCDIC; the second parameter
        for local byte is a decimal integer to indicate Bytesize.
        The parameters are separated by a <SP> (Space, ASCII code 32).

        The types ASCII and EBCDIC also take a second (optional)
        parameter; this is to indicate what kind of vertical format
         control, if any, is associated with a file.

        The following codes are assigned for type:
           A - ASCII   \    / N - Non-print
           U - Unicode |-><-| T - Telnet format effectors
           E - EBCDIC  /    \ C - Carriage Control (ASA)

           I - Image
           L <byte size> - Local byte Byte size
        """
        type = type.upper()
        requested_type = type[:1]
        mode = type[1:].strip()
        if requested_type:
            f = getattr(self, 'type_' + requested_type, None)
            if f is not None:
                return f(mode)
            return self.type_UNKNOWN(requested_type)

        return (ftp.SYNTAX_ERR,)

    def _doTypeChange(self, binary, type):
        """
        Called when type should be changed.
        """
        if self.factory.ignore_ascii_data_type and not binary:
            self.binary = True
            self.emitEvent('10082', data={'type': type})
        else:
            self.binary = binary
            self.emitEvent('10081', data={'type': type})
        return (ftp.TYPE_SET_OK, type)

    def type_A(self, code):
        """
        Called for text type.

        We are always in Unicode mode.
        """
        if code == '' or code == 'N':
            return self._doTypeChange(binary=False, type='A N')
        else:
            return defer.fail(ftp.CmdArgSyntaxError(code))

    def type_I(self, code):
        """
        Called for binary type.
        """
        if code == '':
            return self._doTypeChange(binary=True, type='I')
        else:
            return defer.fail(ftp.CmdArgSyntaxError(code))

    def type_UNKNOWN(self, code):
        """
        Called when an unsupported type is requested.
        """
        self.emitEvent('10080', data={'type': code})
        return defer.fail(ftp.CmdNotImplementedForArgError(code))

    def type_U(self, code):
        """
        TYPE U is equivalent with A as we are always in Unicode mode.

        UTF-8 RFC http://tools.ietf.org/html/draft-klensin-ftp-typeu-00
        """
        if code == '' or code == 'N':
            return self._doTypeChange(binary=False, type='U N')
        else:
            return defer.fail(ftp.CmdArgSyntaxError(code))

    def ftp_USER(self, username):
        """
        USER NAME (USER)

        The argument field is a Telnet string identifying the user.
        The user identification is that which is required by the
        server for access to its file system.  This command will
        normally be the first command transmitted by the user after
        the control connections are made (some servers may require
        this).  Additional identification information in the form of
        a password and/or an account command may also be required by
        some servers.  Servers may allow a new USER command to be
        entered at any point in order to change the access control
        and/or accounting information.  This has the effect of
        flushing any user, password, and account information already
        supplied and beginning the login sequence again.  All
        transfer parameters are unchanged and any file transfer in
        progress is completed under the old access control
        parameters.

        If password based authentication is enabled, the peer SSL certificate
        validation will be optional.
        """
        if not username:
            return defer.fail(ftp.CmdSyntaxError('USER requires an argument'))

        # FIXME:4974:
        # This should be configurable and disabled by default.
        # For now it's enabled by default to help make our tests green.
        try:
            self.transport.transport._force_close = True
        except AttributeError:
            """
            Most probably we are over unsecured FTP and there is no need to
            force close.
            """

        def proceed_with_password_authentication():
            """
            Prepare the session for next PASS command.
            """
            self._user = username
            self.state = self.INAUTH
            return (ftp.USR_NAME_OK_NEED_PASS, username)

        def proceed_with_ssl_certificate_authentication(peer_certificate):
            """
            Attempt to authenticate based on SSL certificate.
            """
            self._user = username
            self.state = self.INAUTH

            def eb_ssl_certificate_error(failure):
                """
                Called when the certificate was rejected.
                """
                if self.factory.enable_password_authentication:
                    return proceed_with_password_authentication()
                else:
                    return self._ebAuthenticationError(failure)

            def cb_ssl_authentication_done(result):
                """
                Called when the certificate was accepted.
                """
                (interface, avatar, logout) = result
                self._setupSession(avatar, logout)
                # Delete temporary _user variable used between USER and PASS
                # commands.
                del self._user
                return USR_LOGGED_IN_SSL_PROCEED

            credentials = FTPSSSLCertificateCredentials(
                username=username,
                certificate=peer_certificate,
                peer=self.avatar.peer,
                creator=self.factory,
                )
            deferred_login = self.factory.portal.login(credentials, None)
            deferred_login.addCallbacks(
                cb_ssl_authentication_done, eb_ssl_certificate_error)
            return deferred_login

        def inform_ssl_authentication_required():
            '''On SSL auth failure, try password authentication if enabled or
            inform client that SSL authentication is required.'''
            self.emitEvent('10028', data={'username': username})
            return SSL_CERT_REQUIRED

        if self.factory.enable_ssl_certificate_authentication:
            peer_certificate = self._getPeerCertificate()
            if peer_certificate is None:
                if self.factory.enable_password_authentication:
                    return proceed_with_password_authentication()
                else:
                    return inform_ssl_authentication_required()
            else:
                return proceed_with_ssl_certificate_authentication(
                    peer_certificate)
        elif self.factory.enable_password_authentication:
            return proceed_with_password_authentication()
        else:
            self.state = self.UNAUTH
            self.emitEvent('10027')
            return NO_AUTH_METHOD_ENABLED

    def ftp_PASS(self, password):
        """
        Second part of login. Get the password the peer wants to
        authenticate with.
        """
        if self.state == self.AUTHED:
            return ALREADY_AUTH
        username = self._user

        def eb_password_error(failure):
            """
            Called when username and password credentials are rejected.
            """
            return self._ebAuthenticationError(failure)

        def cb_password_authentication_done(result):
            '''Callback for handling finalized login.'''
            (interface, avatar, logout) = result
            self._setupSession(avatar, logout)
            # Delete temporary _user variable used between USER and PASS
            # commands.
            del self._user
            return ftp.USR_LOGGED_IN_PROCEED

        if self.is_ftps_command_active:
            credentials_class = FTPSPasswordCredentials
        else:
            credentials_class = FTPPasswordCredentials

        credentials = credentials_class(
            username=username,
            password=password,
            peer=self.avatar.peer,
            creator=self.factory,
            )

        deferred_login = self.factory.portal.login(credentials, None)
        deferred_login.addCallbacks(
            cb_password_authentication_done, eb_password_error)
        return deferred_login

    def _ebAuthenticationError(self, failure):
        """
        Callback when authentication attempt failed.
        """
        self.state = self.UNAUTH
        del self._user

        if failure.check(InsufficientAuthentication):
            # FIXME:5254:
            # Add support for certificate AND password credentials.
            # For now this is handled as a normal authentication failure.
            # Nothing is logged as the higher level authentication code
            # has already emitted an event for this failure.
            raise AuthorizationError()

        failure.trap(ServerError)

        emitFromError(failure.value, emitter=self.emitEvent)
        raise AuthorizationError()

    def _setupSession(self, avatar, logout):
        """
        Prepare the FTP session after successful login.
        """
        try:
            self.shell = FTPShell(avatar)
        except Exception as error:
            # We fail to initialize the FTP session,
            # even if we have valid credentials.
            # Most probably the account is not configured ok.
            emit_error_details(error, event_id='10089', data={}, avatar=avatar)
            raise AuthorizationError()
        else:
            self._avatar = avatar

        self.logout = logout
        self.state = self.AUTHED
        self.shell.start()


class FTPOverflowProtocol(LineReceiver, SSLConnectionMixin):
    """
    FTP mini-protocol for when there are too many connections.
    """

    _avatar = None

    @property
    def avatar(self):
        """
        See: `RemoteConnectionMixin.avatar`.
        """
        return self._avatar

    def connectionMade(self):
        self._avatar = PeerAvatar(
            self.transport.getPeer(),
            component=self.factory,
            )
        self.emitEvent('10091')
        self.sendLine(RESPONSE[ftp.TOO_MANY_CONNECTIONS].encode('utf-8'))
        self.transport.loseConnection()


class FTPInternalErrorProtocol(LineReceiver, SSLConnectionMixin):
    """
    FTP mini-protocol for when an internal server error occurred while
    initializing the actual protocol.
    """

    _avatar = None
    details = 'Internal server error.'

    @property
    def avatar(self):
        """
        See: `RemoteConnectionMixin.avatar`.
        """
        return self._avatar

    def connectionMade(self):
        self._avatar = PeerAvatar(
            self.transport.getPeer(),
            component=self.factory,
            )
        self.emitEvent('10092', data={'details': self.details})
        response = RESPONSE[ftp.REQ_ACTN_NOT_TAKEN] % (
            "Internal server error.",)
        self.sendLine(response.encode('utf-8'))
        self.transport.loseConnection()


class FTPSIProtocol(FTPProtocol):
    """
    Implicit FTPS protocol.
    """

    def connectionMade(self):
        """
        For implicit FTPS, TLS handshake is triggered right away
        at connection time, automatically.
        """
        def cb_handshake_done(result):
            """
            Called when TLS handshake is done.
            """
            self.reply(ftp.WELCOME_MSG, self.factory.welcome_message)
            return result

        def eb_handshake_failed(failure):
            # Since handshake fails, we want to disable the
            # warning for connections closed before TLS handshake.
            self.transport.protocol._SSL_CONNECTION_CLOSE_TIMEOUT = 0
            # Make sure we disconnect on any handshake failure.
            # The failure should have already emitted an event.
            # So nothing else to do here.
            self.transport.loseConnection()

        # The TLS handshake is triggered implicity when connection is made.
        # Attach error handling to TLS handshake deferred.
        self.beforeTLSHandshake('command')
        tls_handshake = self.waitHandshakeCompleted()
        tls_handshake.addCallback(cb_handshake_done)
        tls_handshake.addErrback(eb_handshake_failed)

        super(FTPSIProtocol, self).connectionMade()

    @property
    def is_ftps_command_active(self):
        """
        Command channel is always secured in FTPSI.
        """
        return True

    def have_valid_authentication_channel(self):
        """
        Data channel is always secured in FTPSI.
        """
        return True


class _DataWriteAfterClose(ftp.FTPCmdError):
    """
    Raised when requesting a write after the connection was closed.
    """

    errorCode = ftp.CNX_CLOSED_TXFR_ABORTED


@implementer(IIdleTimeoutProtocol)
class DTPProtocol(ftp.DTP, policies.TimeoutMixin, SSLConnectionMixin):
    """
    This is the protocol used over the data channel.
    """
    connected = False

    # FTP server-side command and data channel have the same event IDs.
    _TLS_OK_EVENT = '10100'
    _TLS_FAIL_EVENT = '10088'

    def __init__(self, dtp_factory):
        """
        :param dtp_factory: Reference to parent DTP factory.
        """
        self.factory = dtp_factory
        self._scheduler = dtp_factory._reactor
        self._host = None

        self._read_size = 0
        self._wrote_size = 0
        self._start_time = 0
        # Using self.transport.loseConnection() we can not set a status why
        # connection was closed and Twisted will just assume that connection
        # was closed in the "clean" way.
        # This is why we use this flag to check that connection was
        # closed due to a timeout.
        self._timeout_called = None
        # When we don't have FTPS, we don't have handshake,
        # so we pretend handshake is always done.
        self._tls_handshake_completed = defer.succeed(None)
        if dtp_factory.use_ssl:
            # For FTPS, set the deferred that will be later called when
            # TLS is done.
            self.beforeTLSHandshake('data')

    @property
    def avatar(self):
        """
        See: `RemoteConnectionMixin.avatar`.
        """
        return self.factory.pi._avatar

    @property
    def _timeoutCall(self):
        """
        Return the timeout call.

        Used only for testing.
        """
        return self._TimeoutMixin__timeoutCall

    @property
    def host(self):
        """
        Address of the local connection.
        """
        return self._host

    def connectionMade(self):
        """
        Called when data channel is connected.
        """
        self._host = self.transport.getHost()
        self.connected = True
        self.factory.deferred.callback(None)
        self._buffer = []
        self._read_size = 0
        self._wrote_size = 0
        self._start_time = self._scheduler.seconds()
        self.setTimeout(self.factory.idle_data_connection_timeout)

        data = {
            "host_address": force_unicode(self.host.host),
            "host_port": force_unicode(self.host.port),
            "peer_address": force_unicode(self.avatar.peer.host),
            "peer_port": force_unicode(self.avatar.peer.port),
            }
        self.emitEvent('10055', data=data)

    def connectionLost(self, reason):
        """
        Called when data channel was closed.

        This might be closed after the command channel is closed.
        """
        self._tlsConnectionLost(reason)
        duration = self._scheduler.seconds() - self._start_time
        speed = 0
        if duration > 0:
            speed = (self._read_size + self._wrote_size) / duration

        data = {
            "host_address": force_unicode(self.host.host),
            "host_port": force_unicode(self.host.port),
            "peer_address": force_unicode(self.avatar.peer.host),
            "peer_port": force_unicode(self.avatar.peer.port),
            "received": force_unicode(self._read_size),
            "sent": force_unicode(self._wrote_size),
            "speed": speed,
            "duration": duration,
            }
        self.setTimeout(None)
        self.connected = False

        if not reason.check(internet_error.ConnectionDone):
            # Connection not closed in a clean way.
            if reason.check(SSL.Error):
                reason = Failure(internet_error.ConnectionLost(
                    "SSL Error " + force_unicode(reason.value)))
            elif reason.check(internet_error.ConnectionAborted):
                # Fix docstring of Twisted ConnectionAborted
                reason = Failure(ConnectionAborted())
            elif reason.check(internet_error.ConnectionLost):
                # FIXME:3748:
                # Maybe we don't always want to ignore this error.
                # Some FTPS clients are just loosing the connection at the
                # end of a transfer.
                reason = Failure(
                    internet_error.ConnectionDone('Connection lost'))
        else:
            # Connection was closed in a clean way
            if self._timeout_called:
                reason = Failure(
                    internet_error.ConnectionLost("Data connection timeout"))

        # FIXME:4102:
        # For now _onConnLost is only used when we receive data.
        if self._onConnLost:
            if reason.check(internet_error.ConnectionDone):
                self._onConnLost.callback(None)
            else:
                self._onConnLost.errback(reason)

        data['details'] = force_unicode(reason.value)
        # The certificate is included in the connection lost message
        # as for connectionMade, the handshake might not be finalized
        # and we don't have a certificate.
        data['certificate'] = self._getPeerCertificateDescription()
        data['encryption'] = self._getEncryptionDescription()

        if reason.check(internet_error.ConnectionDone):
            self.emitEvent('10030', data=data)
        else:
            self.emitEvent('10031', data=data)

    def timeoutConnection(self):
        """
        Called when data connection times out.
        """
        data = {
            "host_address": force_unicode(self.host.host),
            "host_port": force_unicode(self.host.port),
            "peer_address": force_unicode(self.avatar.peer.host),
            "peer_port": force_unicode(self.avatar.peer.port),
            }
        self.emitEvent('10032', data=data)
        # setTimeout should be called later by loseConnection... but
        # we can also do it now.
        self.setTimeout(None)
        # loseConnection will call .closeConnection(ConnectionDone),
        # but we want to sent a signal to .closeConnection that the
        # connection was not closed in a clean manner.
        self._timeout_called = True
        self.transport.loseConnection()

    def sendLine(self, line):
        """
        Send a line to data channel.

        @type  line: I{str} or I{unicode}
        @param line: The line to be sent.

        If line is I{unicode}, it will be converted to I{str}.
        """
        if isinstance(line, six.text_type):
            line = line.encode('utf-8')
        self.write(line + b'\r\n')

    def resetTimeout(self):
        """
        Reset timeout for data and control protocols.
        """
        self.factory.pi.resetTimeout()
        super(DTPProtocol, self).resetTimeout()

    def write(self, data):
        self.resetTimeout()
        self._wrote_size += len(data)

        if self.connected:
            return self.transport.write(data)

        # FIXME:4102:
        # Let the producer know that this is dead.
        # When we send data, the connection might be closed by the other
        # side.
        # The command channel is not informed about this, so next time the
        # command channel tries to send something, we raise this error.
        raise _DataWriteAfterClose()

    def dataReceived(self, bytes):
        """
        Called when we receive data over the data channel.
        """
        self.resetTimeout()
        self._read_size += len(bytes)
        if self._cons is not None:
            self._conswrite(bytes)
            return

        if self._buffer is None:
            # Over FTPS, when we disconnect an upload,
            # there might still be another packet pending.
            # We just ignore it.
            return

        # We have received data, but we don't yet have a consumer.
        self._buffer.append(bytes)

    def registerConsumer(self, consumer):
        """
        Called when we have an upload/push request.
        """
        self._cons = consumer
        self._cons.registerProducer(self, True)

        for chunk in self._buffer:
            self._conswrite(chunk)

        # The buffer is set to None to make sure we will no longer receive
        # data after disconnection.
        self._buffer = None
        if self.connected:
            self._onConnLost = d = defer.Deferred()
            d.addBoth(self._unregConsumer)
            return d
        else:
            # Connection already lost.
            # We might have already received all data.
            self._cons.unregisterProducer()
            self._cons = None
            return defer.succeed(None)

    def _unregConsumer(self, ignored):
        """
        Called at the end of an upload/receive transfer.

        This is legacy code design from upstream Twisted.
        """
        self._cons.unregisterProducer()
        self._cons = None
        del self._onConnLost
        return ignored

    def _conswrite(self, bytes):
        """
        Called when data is recevied and a consumer is already registred.

        This is legacy code design from upstream Twisted.
        """
        try:
            self._cons.write(bytes)
        except BaseException:
            self._onConnLost.errback()

    def registerProducer(self, producer, streaming):
        """
        Called when we need to download/send transfer.
        """
        return self.transport.registerProducer(producer, streaming)

    def unregisterProducer(self):
        """
        Called when no more data needs to be sent.
        """
        self.transport.unregisterProducer()
        self.transport.loseConnection()


class DTPFactory(ftp.DTPFactory, object):
    """
    The DTPFactory used for listening to incoming data connection.

    It is designed to accept a single connection, and to stop listening
    for new connections once the first connection was closed.
    One DTPFactory is initialized for each data connection and will have
    a single DTPProtocol.

    A `factory` and `protocol` was used to integrate with general Twisted
    conventions for handling connections.
    """

    protocol = DTPProtocol

    def __init__(self, pi, use_ssl, peerHost=None, reactor=None, use_pp2=True):
        super(DTPFactory, self).__init__(
            pi=pi, peerHost=peerHost, reactor=reactor)
        self._use_ssl = use_ssl
        self._use_pp2 = use_pp2
        self.setTimeout(self.idle_data_connection_timeout)

    @property
    def use_ssl(self):
        return self._use_ssl

    def buildProtocol(self, addr):
        """
        Called when a new TCP connection is made.
        """
        if self._state is not self._IN_PROGRESS:
            return None
        self._state = self._FINISHED

        self.cancelTimeout()
        p = self.protocol(dtp_factory=self)
        if not self._use_ssl and self._use_pp2:
            # We proxify if pp2 is requested and the factory is not
            # used via a TLS/SSL wrapper.
            proxify_v2(p)
        self.pi.dtpInstance = p
        return p

    def stopFactory(self):
        """
        This is called when TCP port stops to listen for new connections.
        """
        self.cancelTimeout()

    def setTimeout(self, seconds):
        """
        Set the timeout for waiting for the DTP connection to be made.
        """
        if not seconds:
            return

        self.delayedCall = self._reactor.callLater(
            seconds, self.timeoutFactory)

    def timeoutFactory(self):
        """
        Called when initial data connections is not made.
        """
        if self._state is not self._IN_PROGRESS:
            return
        self._state = self._FAILED
        deferred = self.deferred
        self.deferred = None
        deferred.errback(
            ftp.PortConnectionError("DTPFactory timeout"))

    @property
    def idle_data_connection_timeout(self):
        """
        Data channel idle timeout value
        """
        return self.pi.idle_data_connection_timeout


class FTPConsumerWrapper(object):
    """
    Base code for FTP newline wrappers.

    Specialized classes should define the `write` method.
    """
    def __init__(self, context):
        self._context = context
        self._lockback = b''
        self.registerProducer = self._context.registerProducer

    def unregisterProducer(self):
        """
        Stop writing data, when converting FTP newlines to Unix newlines.
        """
        if self._lockback:
            # We might have one last byte.
            self._context.write(self._lockback)
            self._lockback = b''
        self._context.unregisterProducer()

    def _getChunk(self, data):
        """
        Return a chunk which don't have a partial delimiter as last line.
        """
        last_character = data[-1:]
        if last_character == b'\r':
            # We might get a delimiter between chunks.
            chuck = self._lockback + data[:-1]
            self._lockback = last_character
        else:
            chuck = self._lockback + data
            self._lockback = b''
        return chuck


class FromASCIIConsumerWrapper(FTPConsumerWrapper):
    """
    A wrapper around a consumer to convert FTP newlines into local
    system new lines.
    """
    def __init__(self, context):
        super(FromASCIIConsumerWrapper, self).__init__(context)
        if os.linesep == native_string("\r\n"):
            # No need to convert new lines as local system has the same
            # delimiter.
            self.write = self._context.write
        else:
            # We assume that local system delimiter is '\n'
            self.write = self._writeWithFTPToUnixConversion

    def _writeWithFTPToUnixConversion(self, bytes):
        """
        Write the data, converting FTP newlines to Unix newlines.
        """
        data = self._getChunk(bytes)
        return self._context.write(data.replace(b'\r\n', b'\n'))


class ToASCIIConsumerWrapper(FTPConsumerWrapper):
    """
    A wrapper around a consumer to convert system lines to
    FTP newlines (CRLF).
    """
    def __init__(self, context):
        super(ToASCIIConsumerWrapper, self).__init__(context)

    def write(self, data):
        """
        Write the data, converting any new lines to FTP newlines.
        """
        chunk = self._getChunk(data)
        output = chunk.replace(b'\r\n', b'\n').replace(b'\n', b'\r\n')
        return self._context.write(output)
