# Copyright (c) 2011 Adi Roiban.
# See LICENSE for details.
"""
Module for hosting the Chevah FTP filesystem access.

This is a high level module doing low level filesystem access.
Errors should be logged in upper levels.
This module should only raise exceptions.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
import errno
import fnmatch
import os

import six
from zope.interface import implementer
from twisted.internet import defer, reactor
from twisted.internet.task import deferLater
from twisted.internet.interfaces import IConsumer
from twisted.protocols import ftp
from twisted.protocols.ftp import (
    errnoToFailure,
    FileNotFoundError,
    )
from twisted.python.failure import Failure

from chevah_compat import LocalFilesystem

from chevah.server import force_unicode
from chevah.server.commons.authorization import AuthorizationAvatarAdapter
from chevah.server.commons.constant import DEFAULT_FILE_MODE
from chevah.server.commons.exception import (
    AuthorizationException,
    ServerException,
    )
from chevah.server.commons.interface import IFTPShell
from chevah.server.commons.session import (
    emit_account_login_paths,
    RemoteConnectionMixin,
    )
from chevah.server.event.handler import emit_error_details

# A market to signal that NLST was requested for a file.
NLST_ON_FILE_MARKER = object()


def to_segments(cwd, path):
    """
    Normalize a path, as represented by a list of strings each
    representing one segment of the path.

    This is the place where segments are created and we make sure
    they are unicode.
    """
    if not isinstance(path, six.text_type):
        path = path.decode('utf-8')

    path_encoded = path.encode('utf-8')

    if path.startswith('/'):
        segs = []
    else:
        segs = cwd[:]

    for s in path.split('/'):
        if s == '.' or s == '':
            continue
        elif s == '..':
            if segs:
                segs.pop()
            else:
                raise ftp.InvalidPath(cwd, path_encoded)
        elif '\0' in s or '/' in s:
            raise ftp.InvalidPath(cwd, path_encoded)
        else:
            segs.append(s)
    return segs


ftp.toSegments = to_segments


class FTPErrorMixin(object):
    """
    Shared code for handling FTP protocol error.
    """

    def _ebHandleFailure(self, failure, event_id, data):
        """
        Emit detailed error and convert it into a FTP failure.

        FIXME:3649:
        This should be moved into the FTP protocol as we should have a
        protocol independent shell which can be reused by multiple protocols.
        """
        error = failure.value
        emit_error_details(
            error=error,
            event_id=event_id,
            data=data,
            emitter=self.emitEvent,
            )

        if isinstance(error, ftp.FTPCmdError):
            return defer.fail(error)

        if isinstance(error, (OSError, IOError)):
            return errno_to_failure(error.errno, data['path'])

        if isinstance(error, AuthorizationException):
            error = ftp.PermissionDeniedError(data['path'])
            return defer.fail(error)

        # The error should be already emitted in the dedicated shell
        # method. Here we are just converting to not get an ISE error
        # in the protocol handling.
        error = ftp.FTPCmdError('Storage operation failure.')
        error.errorCode = ftp.REQ_ACTN_NOT_TAKEN
        return defer.fail(error)

    def _cbEmitSuccess(self, result, event_id, data):
        """
        Emit the event and forward the result.
        """
        self.emitEvent(event_id, data=data)
        return result


def errno_to_failure(error, path):
    """
    Map C{OSError} and C{IOError} to standard FTP errors.
    """
    if error == errno.ENOENT:
        return defer.fail(FileNotFoundError(path))

    if error == errno.EACCES or error == errno.EPERM:
        return defer.fail(ftp.PermissionDeniedError(path))

    if error == errno.ENOTDIR:
        return defer.fail(ftp.IsNotADirectoryError(path))

    if error == errno.EEXIST:
        return defer.fail(ftp.FileExistsError(path))

    if error == errno.EISDIR:
        return defer.fail(ftp.IsADirectoryError(path))

    # Fallback to permission error since we don't know what kind of error
    # this is.
    return defer.fail(ftp.PermissionDeniedError(path))


@implementer(IFTPShell)
class FTPShell(RemoteConnectionMixin, FTPErrorMixin):
    """
    Adapter for ServiceAvatar into a IFTPShell.

    This is mostly a wrapper around compat.LocalFilesystem.

    This is a modified version of twisted FTPShell in order to fix bugs and
    adapt to the needs of Chevah FTP Server.

    All methods receive bytes / undecoded / raw data as input.

    All methods should return deferreds.

    FIXME:3649:
    All deferred failures should be of type FTPCmdError as otherwise it
    will generate an ISE error.

    Actions are impersonated to reflect account associated with this shell
    session.

    Permissions are checked after resolving the requested path, and emitting
    the initial informational event, so that we have valid path and
    Unicode paths.
    """
    _scheduler = reactor

    def __init__(self, avatar):
        self._avatar = avatar
        self._authorization = AuthorizationAvatarAdapter(avatar)
        self._filesystem = LocalFilesystem(avatar=avatar)
        self._working_segments = self._filesystem.home_segments

    def start(self):
        """
        Called when the shell can start operating.
        """
        start_path = self._filesystem.getPath(self._working_segments)
        emit_account_login_paths(
            id='10059',
            avatar=self._avatar,
            emitter=self.emitEvent,
            path=start_path,
            )

    def emitEvent(self, event_id, data=None, avatar=None):
        """
        Emit an event for this shell.
        """
        if data is None:
            data = {}

        path = data.get('path', None)
        if path:
            data['real_path'] = self.getRealPath(path)
        super(FTPShell, self).emitEvent(event_id, data=data)

    @property
    def avatar(self):
        """
        See: RemoteConnectionMixin.
        """
        return self._avatar

    @property
    def working_segments(self):
        """
        See: IFTPShell.
        """
        return self._working_segments

    def getAbsolutePath(self, path):
        """
        See: IFTPShell.
        """
        path = force_unicode(path)
        try:
            segments = self._getSegments(path)
        except FileNotFoundError:
            return path

        return self._filesystem.getPath(segments)

    def getRealPath(self, path):
        """
        See: IFTPShell.
        """
        path = force_unicode(path)
        try:
            segments = self._getSegments(path)
        except FileNotFoundError:
            return 'no real path/associated'

        return self._filesystem.getRealPathFromSegments(segments)

    def _getSegments(self, path):
        """
        Return segments associated with path.

        Raise error if path is invalid.
        """
        try:
            return ftp.toSegments(self._working_segments, path)
        except ftp.InvalidPath:
            raise FileNotFoundError(path)

    def makeDirectory(self, path):
        """
        See `IFTPShell`.
        """
        data = {'path': self.getAbsolutePath(path)}
        self.emitEvent('10052', data=data)

        deferred = defer.maybeDeferred(
            self._authorization.canCreateFolder, data['path'])
        deferred.addCallback(
            self._cbHandlePath,
            self._filesystem.createFolder, path, recursive=False)
        deferred.addCallback(self._cbEmitSuccess, '10053', data=data)
        deferred.addErrback(self._ebHandleFailure, '10054', data=data)
        return deferred

    def removeDirectory(self, path):
        """
        See `IFTPShell`.
        """
        data = {'path': self.getAbsolutePath(path)}
        self.emitEvent('10071', data=data)

        deferred = defer.maybeDeferred(
            self._authorization.canDeleteFolder, data['path'])
        deferred.addCallback(
            self._cbHandlePath,
            self._filesystem.deleteFolder, path, recursive=False)
        deferred.addCallback(self._cbEmitSuccess, '10072', data=data)
        deferred.addErrback(self._ebHandleFailure, '10073', data=data)
        return deferred

    def rename(self, from_path, to_path):
        """
        See `IFTPShell`.
        """
        data = {
            'path': self.getAbsolutePath(to_path),
            'from': self.getAbsolutePath(from_path),
            'to': self.getAbsolutePath(to_path),
            }
        self.emitEvent('10074', data=data)

        deferred = defer.maybeDeferred(
            self._authorization.canRename, data['from'], data['to'])
        deferred.addCallback(self._cbRename, from_path, to_path)
        deferred.addCallback(self._cbEmitSuccess, '10075', data=data)
        deferred.addErrback(self._ebHandleFailure, '10076', data=data)
        return deferred

    def _cbRename(self, result, from_path, to_path):
        """
        Wrapper for catching all rename errors.

        Path are received in utf-8 as requested by the client.
        """
        try:
            from_segments = ftp.toSegments(self.working_segments, from_path)
        except ftp.InvalidPath:
            return defer.fail(FileNotFoundError(from_path))

        try:
            to_segments = ftp.toSegments(self.working_segments, to_path)
        except ftp.InvalidPath:
            return defer.fail(FileNotFoundError(to_path))

        try:
            result = self._filesystem.rename(from_segments, to_segments)
        except (IOError, OSError) as e:
            return errnoToFailure(e.errno, from_path)
        except Exception:
            return defer.fail()

        return defer.succeed(result)

    def removeFile(self, path):
        """
        See `IFTPShell`.
        """
        data = {'path': self.getAbsolutePath(path)}
        self.emitEvent('10043', data=data)

        deferred = defer.maybeDeferred(
            self._authorization.canDeleteFile, data['path'])
        deferred.addCallback(
            self._cbHandlePath,
            self._filesystem.deleteFile, path)
        deferred.addCallback(self._cbEmitSuccess, '10044', data=data)
        deferred.addErrback(self._ebHandleFailure, '10045', data=data)
        return deferred

    def openForReading(self, path, offset=0):
        """
        See `IFTPShell`.
        """
        data = {
            'path': self.getAbsolutePath(path),
            'offset': offset,
            }
        self.emitEvent('10068', data=data)

        def cb_open(result):
            self._seekFile(result, offset)
            return result

        deferred = defer.maybeDeferred(
            self._authorization.canRead, data['path'])
        deferred.addCallback(
            self._cbHandlePath,
            self._filesystem.openFileForReading, path)
        deferred.addCallback(cb_open)
        deferred.addCallback(self._cbEmitSuccess, '10094', data=data)
        deferred.addErrback(self._ebHandleFailure, '10095', data=data)
        deferred.addCallback(lambda fd: _FTPDownload(
            fd, self.emitEvent, data, self._scheduler))
        return deferred

    def openForWriting(self, path, offset=0):
        """
        See `IFTPShell`.
        """
        # Start with emitting an event to inform about the request.

        path = self.getAbsolutePath(path)
        # Data is passed as an input/output variable to share with error
        # handling.
        data = {'path': path}

        deferred = defer.succeed(None)
        deferred.addCallback(lambda _: _FTPFileWriter(
            shell=self,
            segments=self._getSegments(path),
            data=data,
            offset=offset,
            scheduler=self._scheduler,
            ))
        deferred.addErrback(self._ebHandleFailure, '10013', data=data)
        return deferred

    def _seekFile(self, file, offset):
        """
        Set cursor of `file` at `offset` making sure we don't go over the
        existing file size.
        """
        if offset > 0:
            file.seek(0, os.SEEK_END)
            max_offset = file.tell()
            if offset > max_offset:
                # While the OS might accept it,
                # it can create sparse files.
                raise ServerException(
                    'Seeking over the current file size is not supported.')

        file.seek(offset, os.SEEK_SET)

    def openForAppending(self, path):
        """
        See `IFTPShell`.
        """
        path = self.getAbsolutePath(path)
        # Data is passed as an input/output variable to share with error
        # handling.
        data = {'path': path}

        deferred = defer.succeed(None)
        deferred.addCallback(lambda _: _FTPFileWriter(
            shell=self,
            segments=self._getSegments(path),
            data=data,
            append=True,
            scheduler=self._scheduler,
            ))
        deferred.addErrback(self._ebHandleFailure, '10041', data=data)
        return deferred

    def setWorkingPath(self, path):
        """
        See `IFTPShell`.
        """
        data = {'path': self.getAbsolutePath(path)}

        self.emitEvent('10037', data=data)

        deferred = defer.maybeDeferred(
            self._authorization.canTraverse, data['path'])
        deferred.addCallback(
            self._cbHandlePath,
            self._filesystem.getAttributes, path
            )

        def cb_got_attributes(attributes):
            """
            Called when we have received the attributes.
            """
            if not attributes.is_folder:
                raise ftp.IsNotADirectoryError(path)

            segments = self._getSegments(path)
            self._working_segments = segments

        deferred.addCallback(cb_got_attributes)
        deferred.addCallback(self._cbEmitSuccess, '10038', data=data)
        deferred.addErrback(self._ebHandleFailure, '10039', data=data)
        return deferred

    def getWorkingPath(self):
        """
        See `IFTPShell`.
        """
        return self._filesystem.getPath(self.working_segments)

    def getPath(self, segments):
        """
        See `IServerShell`.
        """
        return self._filesystem.getPath(segments)

    def _startListing(self, path, glob, operation):
        """
        Called when starting to execute a listing operation, returning the
        actual path which is called by the list, without arguments.
        """
        path = self.getAbsolutePath(path)
        data = {
            'path': path,
            'glob': glob,
            'operation': operation,
            }
        self.emitEvent('10046', data=data)
        return path, glob, data

    def getMembersName(self, path, glob=None):
        """
        See `IFTPShell`

        Return an iterator which contains a single marker, when this
        was requested for a file.

        This will emit the same event ID as `list' in order to reuse the
        code.
        """
        path, glob, data = self._startListing(path, glob, 'only names')

        def cb_authorized(result):
            """
            Called when we are authorized the perform the operations.
            """
            segments = self._getSegments(path)

            if glob or path.endswith('/') or path.endswith('\\'):
                # We have a request for a path or for glob.
                # So list right away and don't try the operation for a file.
                return do_list(segments)

            # This might be a file.
            attributes = self._filesystem.getAttributes(segments)
            if attributes.is_folder:
                # Is a folder, so do normal folder listing.
                return do_list(segments)

            # This is a file.
            return _FTPFileNameIterator(
                segments=segments,
                filesystem='ignored',
                glob=None,
                event_emitter=self.emitEvent,
                scheduler=self._scheduler,
                event_data=data,
                )

        def do_list(segments):
            """
            Called when we can start listing.
            """
            return _FTPFolderNamesIterator(
                segments=segments,
                filesystem=self._filesystem,
                glob=glob,
                event_emitter=self.emitEvent,
                scheduler=self._scheduler,
                event_data=data,
                )

        deferred = defer.maybeDeferred(self._authorization.canList, path)
        deferred.addCallback(cb_authorized)
        deferred.addErrback(self._ebHandleFailure, '10048', data=data)
        return deferred

    def list(self, path, glob=None, accept_file=True):
        """
        See `IFTPShell`.

        Return the tuple of (own_attribtes, batch_attribtes)

        Batch attributes are list of files at given `path` together with their
        attributes.
        Path for members is absolute virtual path.

        Path for own_attributes is full real path.

        If we don't have permissions to get details for a file, unknown name
        and zero values are returned.

        This will emit the same event ID as `getMembersName'.
        """
        path, glob, data = self._startListing(path, glob, 'all attributes')

        def cb_authorized_traversal(result):
            """
            Called when we are authorized the perform the traversal operation.
            """
            segments = self._getSegments(path)

            list_authorization_error = None
            show_only_folders = True
            try:
                self._authorization.canList(path)
                show_only_folders = False
            except AuthorizationException as error:
                list_authorization_error = error
                # We can traverse but can't list, so we only show folders.
                show_only_folders = True

            attributes = self._filesystem.getAttributes(segments)
            attributes.path = self._filesystem.getPath(segments)

            if glob or path.endswith('/') or path.endswith('\\'):
                # We have a request for a path or for glob.
                # So list right away and don't try the operation for a file.
                return attributes, do_list(segments, show_only_folders)

            # This might be a file.
            if attributes.is_folder:
                # Is a folder, so do normal folder listing.
                return attributes, do_list(segments, show_only_folders)

            if not accept_file:
                raise OSError(
                    errno.ENOTDIR,
                    'Not a directory',
                    path.encode('utf-8'),
                    )

            # This is a file.
            if show_only_folders:
                # We are not allowed to touch the files.
                raise list_authorization_error

            return attributes, _FTPFileAttributesIterator(
                # We use the already retrieved attributes.
                attributes=attributes,
                segments=segments,
                filesystem=self._filesystem,
                glob=None,
                event_emitter=self.emitEvent,
                scheduler=self._scheduler,
                event_data=data,
                )

        def do_list(segments, show_only_folders):
            """
            Called when we can start listing.
            """
            return _FTPFolderAttributesIterator(
                segments=segments,
                filesystem=self._filesystem,
                glob=glob,
                event_emitter=self.emitEvent,
                scheduler=self._scheduler,
                event_data=data,
                only_folders=show_only_folders
                )

        deferred = defer.maybeDeferred(self._authorization.canTraverse, path)
        deferred.addCallback(cb_authorized_traversal)
        deferred.addErrback(self._ebHandleFailure, '10048', data=data)
        return deferred

    def getAttributes(self, path):
        """
        See `IFTPShell`.

        It hides the real path and only shows the virtual full path.
        """
        data = {
            'path': self.getAbsolutePath(path),
            }
        self.emitEvent('10049', data=data)

        def cb_got_attributes(result):
            """
            Called when we got the attributes.
            """
            result.path = data['path']
            return result

        deferred = defer.maybeDeferred(
            self._authorization.canGetAttributes, data['path'])
        deferred.addCallback(
            self._cbHandlePath,
            self._filesystem.getAttributes, path)
        deferred.addCallback(cb_got_attributes)
        deferred.addCallback(self._cbEmitSuccess, '10050', data=data)
        deferred.addErrback(self._ebHandleFailure, '10051', data=data)

        return deferred

    def setAttributes(self, path, attributes):
        """
        Set attributes for path.
        """
        data = {
            'path': self.getAbsolutePath(path),
            'attributes': attributes,
            }
        self.emitEvent('10096', data=data)

        deferred = defer.maybeDeferred(
            self._authorization.canSetAttributes, data['path'])
        deferred.addCallback(
            self._cbHandlePath,
            self._filesystem.setAttributes, path, attributes)
        deferred.addCallback(self._cbEmitSuccess, '10097', data=data)
        deferred.addErrback(self._ebHandleFailure, '10098', data=data)

        return deferred

    def _handleFileSegments(self, handler_callback, segments):
        """
        Return a deferred with the result of handler_callback(`segments`)
        considering `segments` should be for a file.
        """
    def _cbHandlePath(self, result, handler_callback, path, *args, **kwargs):
        """
        Call handler_callback(`segments`),
        where segments are resolved from `path`.
        """
        segments = self._getSegments(path)
        return handler_callback(segments, *args, **kwargs)


class _FTPFolderIteratorAbstract(FTPErrorMixin):
    """
    Iterates over the content of the folder at `segments` using `filesystem`,
    returning the results in batches of deferred.

    It will filter members not matching `glob`.

    This is here since we don't have yet a consumer/producer design for NLST
    and LIST.
    """

    _batch_size = 500

    def __init__(
        self, segments, filesystem, event_emitter, scheduler,
        glob, event_data, only_folders=False,
            ):
        self.emitEvent = event_emitter
        self._filesystem = filesystem
        self._segments = segments
        self._scheduler = scheduler
        self._event_data = event_data
        self._glob = glob
        self._show_only_folders = only_folders
        self._folderIterator = self._getFolderIterator()

    def _getFolderIterator(self):
        """
        Return the iterator for this folder.
        """
        return self._filesystem.iterateFolderContent(self._segments)

    def _handleMember(self, candidate):
        """
        Called to return the processed result for `candidate`
        """
        raise NotImplementedError('_handleMember not implemented.')

    def __iter__(self):
        """
        We are our own iterator.
        """
        return self

    def __next__(self):
        """
        Iterate the next batch.
        """
        try:
            batch = self._next()
            deferred = deferLater(self._scheduler, 0, lambda: batch)
        except StopIteration:
            # Bubble the end of the iteration, without any audit as the
            # event should have already been emitted.
            raise
        except Exception as error:
            deferred = defer.fail(error)

        deferred.addErrback(
            self._ebHandleFailure, '10048', data=self._event_data)
        return deferred

    next = __next__

    def _next(self):
        """
        Return the next batch or fail with an exception
        """
        batch = []
        count = 0

        def add_candidate(count):
            attributes = next(self._folderIterator)
            if self._glob and not fnmatch.fnmatch(attributes.name, self._glob):
                # Candidate should be ignored.
                return count

            result = self._handleMember(attributes)
            if not result:
                # Candidate should be ignored.
                return count

            batch.append(result)

            updated_count = count + 1
            return updated_count

        try:
            count = add_candidate(count)
        except StopIteration:
            # In the case in which folder iteration is done, this will
            # bubble to final stop iteration.
            self.emitEvent('10047', data=self._event_data)
            raise

        while count < self._batch_size:
            try:
                count = add_candidate(count)
            except StopIteration:
                # We are done with folder listing so we send the current
                # batch.
                # The Iterator will be stop at the next call.
                break

        return batch


class _FTPFolderAttributesIterator(_FTPFolderIteratorAbstract):
    """
    Iterator of deferred for batches with member names and attributes
    """

    def _handleMember(self, attributes):
        """
        For LIST we have a tuple of name and attributes.

        When failing to get the attributes, it returns a
        placeholder attributes.
        """
        member_segments = self._segments + [attributes.name]
        # FIXME:4055:
        # Should be removed once we have a valid getAttributes.
        attributes.path = self._filesystem.getPath(member_segments)

        if self._show_only_folders and not attributes.is_folder:
            # Hide the member as is not a folder.
            return None

        return attributes


class _FTPFileAttributesIterator(_FTPFolderIteratorAbstract):
    """
    Iterator which is used when the LIST command is requested for a file.
    """

    def __init__(self, attributes, *args, **kwargs):
        self._attributes = attributes
        super(_FTPFileAttributesIterator, self).__init__(*args, **kwargs)

    def _getFolderIterator(self):
        """
        Return the iterator for this folder.
        """
        return iter([self._attributes])

    def _handleMember(self, candidate):
        """
        The candidate are the attributes.
        """
        attributes = candidate
        attributes.path = self._filesystem.getPath(self._segments)
        return attributes


class _FTPFolderNamesIterator(_FTPFolderIteratorAbstract):
    """
    Iterator of deferreds for batches with members name.
    """

    def _handleMember(self, attributes):
        """
        For NLST we just add the name including the requested parent path.
        """
        return attributes.name


class _FTPFileNameIterator(_FTPFolderNamesIterator):
    """
    Iterator which is used when the NLST command is requested for a file.
    """

    def _getFolderIterator(self):
        """
        Return the iterator for this folder.

        It should return anything, is important to have an iterator with a
        single value.
        """
        return iter(['.'])

    def _handleMember(self, attributes):
        """
        For NLST on a file we return the marker.
        """
        return NLST_ON_FILE_MARKER


@implementer(ftp.IReadFile)
class _FTPDownload(object):
    """
    A producer for the file content to be read over FTP and consumed
    by an FTP DTP instance.
    """

    _CHUNK_SIZE = 32 * 1024

    def __init__(self, stream, event_emitter, event_data, scheduler):
        self._stream = stream
        self._emitEvent = event_emitter
        self._event_data = event_data
        self._scheduler = scheduler
        self._read_total = 0
        self._start_time = scheduler.seconds()
        self._consumer = None

        self._paused = False
        self._end = defer.Deferred()
        self._end.addBoth(self._close)

    def _produce(self):
        """
        Send/write file content to the consumer, as much as we can.
        """
        while True:
            if self._paused:
                # Consumer can't take more data.
                return

            try:
                chunk = self._stream.read(self._CHUNK_SIZE)
            except Exception as error:
                self._end.errback(ServerException(
                    'Failed while reading the source file. %s' % (
                        force_unicode(error,))))
                return

            if not chunk:
                # No more data to read.
                self._end.callback(None)
                return

            self._read_total += len(chunk)
            self._consumer.write(chunk)

    def pauseProducing(self):
        """
        Called when the sent socket can't receive more data from us.
        """
        self._paused = True

    def resumeProducing(self):
        """
        Called when the sent socket is ready to receive more data.
        """
        if self._end.called:
            # We should not end up here as the transfer was done.
            return

        self._paused = False
        self._produce()

    def stopProducing(self):
        """
        This is called by the DTP connection when we should no longer
        send data.
        """
        if self._end.called:
            # All good.
            return

        self._end.errback(ServerException(
            'Data connection was lost before sending all data.'))

    def _close(self, passthrough):
        """
        Called when transfer is done. With success or failure.
        """
        self._stream.close()

        data = self._event_data.copy()
        data.update(self._getStats())

        if isinstance(passthrough, Failure):
            emit_error_details(
                error=passthrough.value,
                event_id='10070',
                data=data,
                emitter=self._emitEvent,
                )
        else:
            self._emitEvent('10069', data=data)

        if self._consumer:
            # This will also disconnect the consumer.
            self._consumer.unregisterProducer()

        return passthrough

    def send(self, consumer):
        """
        Send the content of our file to `consumer`.

        Returns a deferred that is called when all data was sent.
        The deferred fails if transfer.
        """
        self._consumer = consumer
        self._consumer.registerProducer(self, streaming=True)

        # The consumer might or might not resume the transfer
        # during registration.
        # We resume, just to be sure.
        self.resumeProducing()

        return self._end

    def _getStats(self):
        """
        Return a dictionary with the stats for the read operation so far.
        """
        read_speed = 0
        duration = self._scheduler.seconds() - self._start_time

        if duration > 0:
            read_speed = self._read_total / duration

        return {
            'total': self._read_total,
            'speed': read_speed,
            'duration': duration,
            }


@implementer(IConsumer)
class _FTPUpload(object):
    """
    A consumer that writes data to a file received over FTP.
    """
    def __init__(self, stream, scheduler, pre_write_callback):
        self._stream = stream
        self._scheduler = scheduler
        self._start_time = scheduler.seconds()
        self._end_time = None
        self._write_total = 0
        self._pre_write_callback = pre_write_callback

    def registerProducer(self, producer, streaming):
        self.producer = producer
        assert streaming

    def unregisterProducer(self):
        self.producer = None
        self._end_time = self._scheduler.seconds()
        self._stream.close()

    def write(self, bytes):
        self._pre_write_callback(self._write_total, len(bytes))
        self._stream.write(bytes)
        self._write_total += len(bytes)

    def getStats(self):
        """
        Return a dictionary with the stats for the transfer.
        """
        duration = 0
        write_speed = 0

        if self._start_time and self._end_time:
            duration = self._end_time - self._start_time

        if duration > 0:
            write_speed = self._write_total / duration

        return {
            'total': self._write_total,
            'speed': write_speed,
            'duration': duration,
            }


@implementer(ftp.IWriteFile)
class _FTPFileWriter(object):
    """
    Handles opening the file for writing.

    This is a delegation of FTPShell.openForWriting.

    The local file is opened at init so that we can test that we can receive
    data.

    .receive() is called to register a consumer that will write the file.
    This is all very complicated
    """

    def __init__(
            self, shell, segments, data, scheduler, append=False, offset=0):
        self._shell = shell
        self._data = data
        self._scheduler = scheduler

        filesystem = self._shell._filesystem
        avatar = self._shell._avatar
        if append:
            # We don't want to truncate existing files,
            # but we do want to create if does not exist.
            extra_flags = (
                filesystem.OPEN_APPEND |
                filesystem.OPEN_CREATE
                )
        elif offset == 0:
            # No seek is required, so we start a new file.
            # Truncate the file if exists, or create one.
            extra_flags = (
                filesystem.OPEN_CREATE |
                filesystem.OPEN_TRUNCATE
                )
        else:
            # We are open for updating, so we don't want to create the file,
            # or truncate it.
            extra_flags = 0

        requested_path = filesystem.getPath(segments)
        amended_segments, was_amended = avatar.getAmendedUploadSegments(
            segments)
        path = filesystem.getPath(amended_segments)

        if was_amended:
            # We always want to create a new file and fail if file
            # exists.
            extra_flags = (
                filesystem.OPEN_CREATE |
                filesystem.OPEN_EXCLUSIVE
                )

        flags = filesystem.OPEN_WRITE_ONLY | extra_flags

        self._data.update({
            'path': path,
            'requested_path': requested_path,
            'offset': offset,
            })

        if append and not was_amended:
            # Append event only when we don't have a name change, as for
            # name change an unique write operation is used.
            info_event_id = '10025'
            success_open_event_id = '10040'
        else:
            info_event_id = '10077'
            success_open_event_id = '10012'

        self._shell.emitEvent(info_event_id, data=self._data)

        self._shell._authorization.canWrite(data['path'])

        fd = filesystem.openFile(
            amended_segments, flags, DEFAULT_FILE_MODE)
        self._stream = os.fdopen(fd, 'wb')

        self._shell._seekFile(self._stream, offset)
        # FIXME:6587:
        # If we have `amend` we should reduce the current size of the file.
        self._write_limit = avatar.file_write_size
        # File opened and seek with success.
        self._shell.emitEvent(success_open_event_id, data=data)
        self._consumer = _FTPUpload(
            stream=self._stream,
            scheduler=self._scheduler,
            pre_write_callback=self._checkQuota,
            )

    def getConsumer(self):
        """
        Called when the transfer should start.
        """
        return self._consumer

    def fail(self, error):
        """
        Called by the protocol on failure to receive all the data.

        From our part of writing the received content, all is ok, but
        something happened during the transfer.
        """
        # The upload might fail before we call .receive() and can register
        # a consumer.
        # The whole _FTPUpload design is a mess, but we go with this for now.
        if not self._stream.closed:
            self._consumer.unregisterProducer()

        data = self._data.copy()
        data.update(self._consumer.getStats())
        emit_error_details(
            error=error,
            event_id='10079',
            data=data,
            emitter=self._shell.emitEvent,
            )
        if not self._stream.closed:
            self._stream.close()

    def close(self):
        """
        Called at the end a successful transfer.
        """
        # FileConsumer will close the file object, so nothing special here.
        # We still close the file, in case of strange errors
        # and raise an error so that we are aware of this case.
        if not self._stream.closed:
            self._stream.close()
            return defer.fail(ServerException(
                'File not closed at the end of FTP transfer.'))

        data = self._data.copy()
        data.update(self._consumer.getStats())
        self._shell.emitEvent('10078', data=data)
        return defer.succeed(None)

    def _checkQuota(self, written_size, new_chunk_size):
        """
        Check that we are allowed to write another `size` bytes.
        """
        if not self._write_limit:
            return

        if written_size + new_chunk_size > self._write_limit:
            raise ServerException(
                'Single file write limit is "{}" bytes.'.format(
                    self._write_limit))
