# Copyright (c) 2020 Adi Roiban.
# See LICENSE for details.
"""
Utility code used everywhere.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
import datetime
import time

import dateutil.parser
from OpenSSL import SSL

from twisted.internet.defer import CancelledError

from chevah.server.commons import UTC
import six


def _(string):
    """
    Placeholder for future gettext integration.
    """
    return string


def native_string(string):
    """
    Input should be Unicode.

    On Py3 does nothing.
    On Py2 returns str/bytes.
    """
    if six.PY2:
        string = string.encode('ascii')
    return string


def force_unicode(value, trim=0):
    """
    Decode the `value` to unicode.

    If `trim` is set it will truncate the result.

    It will try to extract the message from an exception.

    In case there are encoding errors when converting the invalid characters
    are replaced.
    """
    import errno

    from chevah_compat.exceptions import CompatException
    from chevah.server.commons.exception import ServerException

    def trim_with_ellipsis(value):
        """
        Return a str that is trimmed if needed.
        """
        if not trim or trim < 1:
            return value
        if len(value) < trim:
            return value

        return value[:trim] + '...'

    def safe_str(value):
        """
        Wrapper for str_or_repr with trim.
        """
        return trim_with_ellipsis(str_or_repr(value))

    def str_or_repr(value):

        if not isinstance(value, six.string_types + (bytes,)):
            # Convert numbers to text.
            try:
                value = six.text_type(value)
            except (TypeError, UnicodeDecodeError):
                # This is only raised on Py2.
                # On Py3 we get TypeError since str() needs to return a valid
                # text.
                try:
                    value = repr(value)
                except (TypeError, UnicodeDecodeError):
                    raise RuntimeError(
                        'Requested object has no text representation.')

        if isinstance(value, six.text_type):
            # Already text.
            return value

        try:
            return value.decode('utf-8')
        except Exception:
            """
            Not UTF-8 encoded value.
            """

        # We fallback to this encoding since this is used by default on
        # some non-EN Windows installs.
        try:
            return value.decode('windows-1252')
        except Exception:
            result = repr(value)
            if six.PY2:
                result = result.decode('ascii')
            return result

    if value is None:
        return 'None'

    if isinstance(value, six.text_type):
        return trim_with_ellipsis(value)

    if isinstance(value, ServerException):
        return trim_with_ellipsis(value.message)

    if isinstance(value, CancelledError):
        return "Operation was cancelled."

    if isinstance(value, CompatException):
        return trim_with_ellipsis(value.message)

    if isinstance(value, SSL.Error):
        try:
            # Try to see if we can extract a nicer view of the
            # error, last stack.
            return safe_str(native_string(' - ').join(value.args[0][-1]))
        except Exception:
            # Fallback to raw arguments.
            details = safe_str(value.args)

    if isinstance(value, ValueError):
        return safe_str(value.args[0])

    if isinstance(value, EnvironmentError) and value.errno:
        # IOError, OSError, WindowsError.
        code = value.errno
        message = value.strerror
        # Convert to Unix message to help with testing.
        if code == errno.ENOENT:
            # On Windows it is:
            # The system cannot find the file specified.
            message = 'No such file or directory'
        if code == errno.EEXIST:
            # On Windows it is:
            # Cannot create a file when that file already exists
            message = 'File exists'
        if code == errno.EBADF:
            # On AIX: Bad file number
            message = 'Bad file descriptor'

        if code and message:
            filename = value.filename
            if getattr(value, 'filename2', ''):
                # For rename or symlink operations on Py3
                # there are 2 paths.
                # On py2, the `filename2` doesn't exist.
                filename = value.filename2
            if filename:
                return "[Errno %s] %s: '%s'" % (
                    code,
                    str_or_repr(message),
                    str_or_repr(filename),
                    )
            return trim_with_ellipsis(
                '[Errno %s] %s.' % (code, str_or_repr(message)))

    if isinstance(value, Exception):
        try:
            details = str(value)
        except (UnicodeDecodeError, UnicodeEncodeError):
            details = value.args[0]
        result = safe_str(details)
        if result:
            return result
        # It looks like the exception __str__ is empty.
        # Fallback to repr.
        return safe_str(repr(value))

    return safe_str(value)


def get_path_handling_module(path):
    """
    Return the module which can be used to handle paths like `path`.

    Il will default to Unix path when `path` has not separator.
    """
    import ntpath
    import posixpath

    # We start with Unix path, as Windows path can not contain a slash
    # but Unix paths can contain backslashes.
    if '/' in path:
        return posixpath

    if '\\' in path:
        return ntpath

    return posixpath


def add_timeout(scheduler, deferred, delay, message=None):
    """
    Attach a timeout to the deferred.
    """
    from twisted.internet.error import TimeoutError

    # This is a hack to work around deferred error wrapping.
    timeout_called = [False]

    if not message:
        message = 'No data exchange.'

    def cancel_deferred(deferred, timeout_called):
        """
        Called when we should trigger the deferred cancellation.
        """
        timeout_called[0] = True
        deferred.cancel()

    def eb_convert_timeout(failure):
        # On cancel we will have a deferred which failed,
        # but the cancel failure can be wrapped in multiple other failures,
        # depending on how the deferred chain was set up before
        # we add the timeout.
        # This is why we need this hack.
        # The error is not always raised during testing.
        if failure.check(TimeoutError):  # noqa:cover
            # This is a timeout from Twisted low level code.
            # Just add more details to the error.
            details = '%s TCP took longer than %s seconds' % (
                message, delay)
            raise TimeoutError(string=details)
        elif timeout_called[0]:
            details = '%s Took longer than %s seconds' % (message, delay)
            raise TimeoutError(string=details)
        else:
            # Bubble the original failure as we don't have a timeout.
            return failure

    def bb_cancel_timeout(result_or_failure, timeout_call):
        """
        Cancel the deferred as the deferred was acted upon.
        """
        if timeout_call.active():
            timeout_call.cancel()
        return result_or_failure

    timeout_call = scheduler.callLater(
        delay, cancel_deferred, deferred, timeout_called)
    deferred.addBoth(bb_cancel_timeout, timeout_call)
    deferred.addErrback(eb_convert_timeout)

    # Here to help identify from where is the timeout added.
    timeout_call._debug_message = message
    return timeout_call


def get_timestamp(dt):
    """
    Convert an ISO 8601 string or a `datetime` object
    to a POSIX timestamp as float.

    Source: https://hg.python.org/cpython/file/3.3/Lib/datetime.py#l1428
    """
    if not isinstance(dt, datetime.datetime):
        dt = dateutil.parser.parse(dt)

    if dt.tzinfo is None:
        return time.mktime((dt.year, dt.month, dt.day,
                            dt.hour, dt.minute, dt.second,
                            -1, -1, -1)) + dt.microsecond / 1e6

    epoch = datetime.datetime(1970, 1, 1, tzinfo=UTC)
    return (dt - epoch).total_seconds()
