# Copyright (c) 2020 Adi Roiban.
# See LICENSE for details.
"""
AS2 file transfer server side / receiving file.

Encapsulation order:
 * Encrypted[ Signed [ Compressed [ File Content ]]] - RFC5402-3.5 PyAS2
 * Encrypted[ Compressed [ Signed [ File Content ]]] - RFC5402-3.6

RFC5652 Cryptographic Message Syntax (CMS).
The way encrypted or compressed or signed messages are formated.

RFC4130 encapsulation (no compression):

No encryption, no signature - Direct HTTP headers and body
  -RFC2616/2045
     -RFC1767/RFC3023 (application/EDIxxxx or /xml)

No encryption, signature - Multipart body
  -RFC2616/2045
    -RFC1847 (multipart/signed)
      -RFC1767/RFC3023 (application/EDIxxxx or /xml)
      -RFC3851 (application/pkcs7-signature)

Encryption, no signature - Mime body with headers and file as payload.
  -RFC2616/2045
    -RFC3851 (application/pkcs7-mime)
      -RFC1767/RFC3023  (application/EDIxxxx or /xml)(encrypted)

Encryption, signature - MIME body with multipart as payload.
  -RFC2616/2045
    -RFC3851 (application/pkcs7-mime)
      -RFC1847 (multipart/signed)(encrypted)
        -RFC1767/RFC3023  (application/EDIxxxx or /xml)(encrypted)
        -RFC3851 (application/pkcs7-signature)(encrypted)

MDN over HTTP, no signature
  -RFC2616/2045 - HTTP headers with HTTP MDN in HTTP body
    -RFC3798 (message/disposition-notification)

MDN over HTTP, signature
  -RFC2616/2045 - Multipart body.
    -RFC1847 (multipart/signed)
     -RFC3798 (message/disposition-notification)
     -RFC3851 (application/pkcs7-signature)

"""
from __future__ import unicode_literals
from __future__ import absolute_import

from collections import OrderedDict
try:
    from cStringIO import StringIO
except ImportError:
    # cStringIO is not available on py2.
    from io import StringIO

from datetime import datetime
from email.generator import Generator
from email.header import decode_header
from email.mime.application import MIMEApplication
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
import base64
import zlib
import random
import uuid
import traceback

from six.moves import map
import six
from asn1crypto import cms, core, algos, x509
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric.utils import Prehashed
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms
from cryptography.hazmat.primitives.ciphers.modes import CBC
from cryptography.hazmat.primitives.padding import PKCS7
from OpenSSL import crypto
from twisted.internet import defer, reactor
from twisted.internet.interfaces import IConsumer
from twisted.python.components import registerAdapter
from twisted.web import http
from twisted.web.server import Session
from zope.interface import implementer

from chevah.server import emit_error_details, force_unicode, native_string
from chevah.server.commons import asn1, DEFAULT_BACKEND, UTC
from chevah.server.commons.constant import TYPE_NAME
from chevah.server.commons.configuration import parse_url
from chevah.server.commons.credential import AS2AnonymousCredentials
from chevah.server.commons.exception import ServerException
from chevah.server.commons.interface import IChevahSession
from chevah.server.commons.i18n import NULL_TRANSLATION
from chevah.server.commons.public_format import public_format
from chevah.server.commons.session import (
    emit_account_login_paths,
    )
from chevah.server.commons.ssl import (
    OpenSSLContextFactory,
    read_PEM_configuration,
    )
from chevah.server.commons.x509 import name_to_OID, PreOrderedName
from chevah.server.event.model import ChevahTimeStamp
from chevah.server.http.client import NonPersistentAgent
from chevah.server.http.form import (
    FormLengthExceededException,
    MultiPartFormData,
    update_headers,
    )
from chevah.server.http.protocol import _CONTINUE, parse_options_header
from chevah.server.http.resource import (
    AuthenticatedResourceWrapper,
    ChevahResource,
    )
from chevah.server.http.rest import RESTShell, _RESTMixin
from chevah.server.location.commons import (
    get_file_name_with_timestamp,
    )


_SUPPORTED_DIGEST = {
    'md5': hashes.MD5,
    'sha1': hashes.SHA1,
    'sha224': hashes.SHA224,
    'sha256': hashes.SHA256,
    'sha384': hashes.SHA384,
    'sha512': hashes.SHA512,
    }

# As received from remote partner.
_SUPPORTED_ENCRYPTIONS = {
    'tripledes_3key': algorithms.TripleDES,
    'aes128_cbc': algorithms.AES,
    'aes192_cbc': algorithms.AES,
    'aes256_cbc': algorithms.AES,
    }


class ChevahMIMEMultipart(MIMEMultipart, object):
    """
    Customization to get the same code working on py2 and py3.
    Can be removed once we no longer need py2.
    """
    def __init__(self, subtype):
        super(ChevahMIMEMultipart, self).__init__(subtype)
        del self['MIME-Version']

    def set_boundary(self, boundary):
        super(ChevahMIMEMultipart, self).set_boundary(boundary)
        if six.PY3:
            return
        result = self['Content-Type']
        del self['Content-Type']
        self['Content-Type'] = result


class IAS2Session(IChevahSession):
    """
    Persistent storage for AS2 requests.
    """


@implementer(IAS2Session)
class AS2Session(object):
    """
    See: `IAS2Session`.
    """

    def __init__(self, session):
        self.resource = None


registerAdapter(AS2Session, Session, IAS2Session)


class AS2ReceiveResource(AuthenticatedResourceWrapper):
    """
    The root resource for receiving AS2 messages.

    It handled the authentication and then delegates to AS2Receiver.
    """
    isLeaf = True

    def __init__(self, service_configuration):
        super(AS2ReceiveResource, self).__init__(
            resource_class=AS2Receiver,
            session_interface=IAS2Session,
            service_configuration=service_configuration,
            translation=NULL_TRANSLATION,
            )

    def _getCredentials(self, request):
        """
        Return non HTTP authentication credentials from `request`.
        """
        credentials = super(AS2ReceiveResource, self)._getCredentials(
            request)

        if credentials:
            # Return the already found credential
            return credentials

        as2_from = request.getHeader('as2-from')
        if as2_from:
            # We have an AS2 credential.
            return AS2AnonymousCredentials(
                username=as2_from,
                peer=request.getAvatarPeer(),
                creator=request.site,
                )

        return None


class AS2Receiver(ChevahResource, _RESTMixin):
    """
    Handled AS2 message receive requests.

    A new resource is initialized after each authentication.

    Unsigned/unencrypted -> Whole file in the request body.

    Signed/unencrypted -> Multipart: first part the file,
        second part the signature.

    Encrypted -> Multipart. A single part.
    """

    _scheduler = reactor
    # Delay between the return the response and sending the async MDN.
    _ASYNC_DELAY = 0.5
    # Random module use to get random values.
    # Here to help with testing.
    _random = random
    _uuid = uuid

    def __init__(self, avatar, translation=None, ui_version=None):
        # There is no UI for AS2 so translations or UI_versions are not used.
        self._avatar = avatar
        self._shell = RESTShell(avatar)

        self._private_key = avatar.component.as2_key
        self._certificate_asn1 = avatar.component.as2_certificate_asn1
        self._certificate_id = avatar.component.as2_certificate_id

        self._reset()
        super(AS2Receiver, self).__init__()

    def _reset(self):
        """
        Called with each new request.
        """
        # We start as root AS2.
        # `_segments` is used by `HasPublicPath` to track the current
        # working file.
        # We start with it being set to the pending path.
        # And later we reset it to the final path.
        self._segments = self._avatar.component.as2_pending
        self._final_segments = self._avatar.component.as2_complete

        # Parse information for the request AS2 specific parameters.
        self._received_data = None

        # Digest algorithm handler for the received data.
        self._digest = None
        # Cached digest value.
        self._digest_value = None

        # The multi-part with the received signature.
        self._received_signature = None

        # The handler for the received file content / payload.
        self._received_file = None

        # The request for this resource.
        self._request = None

        # To know what type of authentication was used.
        # This is not updated for public sites.
        self._credentials_type = TYPE_NAME.NOT_DEFINED

    def start(self, credentials_type):
        """
        Called when the authentication was successful and we can start
        operating.

        We delay the start event up to when all headers are received to
        be able to reject the request on invalid headers.
        """
        self._reset()
        self._credentials_type = credentials_type

    def reset(self):
        """
        Called when the resource is already authenticated and it needs to
        serve a new request.
        """
        self._reset()

    def headersReceived(self, request):
        """
        See: `IChevahResource`.
        """
        self._request = request
        # Start by reading a bit just to parse the headers and be able
        # to open the destination file.
        request.maximumBodyLength = 100 * 1024

        if request.method in [b'HEAD']:
            # Most probably just someone testing our server.
            return _CONTINUE

        try:
            result = self._parseHeaders(request)
        except ServerException as error:
            data = {
                'message_id': 'unknown',
                'as2_from': 'unknown',
                'as2_to': 'unknown',
                'mdn_info': 'unknown',
                'path': 'unknown',
                }
            return self._badHeaders(error.message, data)

        self._received_data = result
        # Update with values from HTTP headers.
        # This might be updated later if filename is in the payload.
        # For example for encrypted or compressed files.
        # FIXME:6082:
        # Here we should have a unique temporary name for each request.
        self._segments = result['segments']
        self._final_segments = result['final_segments']
        # For now we just use the default filename as a hint if site should
        # allow bare requests.
        # In the future we might have a dedicated configuration.
        # But for now we keep it simple.
        allow_empty_headers = bool(request.site._as2_default_filename)

        if request.method not in [b'POST', b'PUT']:
            return self._badHeaders(
                'Only POST or PUT requests are allowed. Got "%s".' % (
                    force_unicode(request.method),),
                result)

        if not result['message_id'] and not allow_empty_headers:
            return self._badHeaders(
                'Missing Message-Id header in request.', result)

        if not result['as2_from'] and not allow_empty_headers:
            return self._badHeaders(
                'Missing AS2-From header in request.', result)

        if not result['as2_to'] and not allow_empty_headers:
            return self._badHeaders(
                'Missing AS2-To header in request.', result)

        async_to = result['mdn']['async_to']
        if async_to and async_to.startswith('mailto:'):
            return self._badHeaders(
                'SMTP MDN is not yet supported.', result)

        if (
            (result['signature'] and not result['is_encrypted'])
            and not result['boundary']
                ):
            return self._badHeaders(
                'Missing multi-part boundary in request.', result)

        for key, value in self._getResponseHeaders(result).items():
            request.setHeader(key, value)

        self._digest = self._getDigest()

        try:
            # This might fail right away when we don't have a signed or
            # encrypted file.
            #
            # For signed and encrypted files we need to read the payload
            # before knowing if all is ok.
            consumer = self._createConsumer(request)
            request.switchConsumer(consumer)

            # So far all looks good.
            return _CONTINUE

        except Exception as error:
            return self._badHeaders(force_unicode(error), result)

    def _badHeaders(self, message, data):
        """
        Emit error event and return HTTP error code when invalid headers
        are received.
        """
        # FIXME:5794:
        # This will close the connection before the requet body is recieved.
        # Maybe we should return an MDN error content here to be AS2 standard.
        data['details'] = message
        # For header errors we should have good error message.
        # No need for traceback.
        # This is here just because 40044 is used for the body handling.
        data['tb'] = ''
        self.emitEvent('40044', data=data)
        return http.BAD_REQUEST

    def _badContent(self, request, error):
        """
        Emit error event and return a bad request page.
        """
        self._received_data['details'] = force_unicode(error)
        self._received_data['tb'] = force_unicode(traceback.format_exc())
        self.emitEvent('40044', data=self._received_data)
        # FIXME:5794:
        # Maybe we should return an MDN error content here to be AS2 standard.
        # We need to check other servers what error code they use.
        request.setResponseCode(http.BAD_REQUEST)
        return b'Failed to process the received AS2 message body.'

    def render_HEAD(self, request):
        """
        For HEAD, the body is ignored anyway.

        Does nothing. The header parsing and URL resource dispatching should
        handle the auth.
        """
        return b''

    def render_PUT(self, request):
        """
        PUT is not a valid AS2 method.

        This is here to allow files received over a very simple PUT API in
        which no file is provided and the whole file is in the body,
        unencrypted and unsigned.
        """
        return self.render_POST(request)

    def render_POST(self, request):
        """
        This is called after all the data was received.
        """
        try:
            return self._renderResponse(request)
        except Exception as error:
            return self._badContent(request, error)

    def _renderResponse(self, request):
        """
        Called after all the content was received to generate the response
        body.

        This also raises an exception for any error found while receiving
        the request body.
        """
        # This might trigger an error if the consumer failed to received
        # the content of the request.
        request.content.unregisterProducer()

        request.setResponseCode(http.CREATED)

        status, details, error = self._getMDNStatus()

        digest_name, digest_value = self._getDigestValue()
        self._received_data['digest'] = base64.b64encode(
            digest_value).decode('ascii')
        if not self._received_data['signature']:
            self._received_data['signature'] = 'Not requested - %s' % (
                digest_name)

        if error:
            # We just log the error on server side.
            # The error details will be sent to the AS2 peer as part of the MDN.
            self._received_data['details'] = error.message
            self.emitEvent('40047', data=self._received_data)
        else:
            # Do the final rename operation and emit the success event.
            try:
                self._shell.rename(self._segments, self._final_segments)
                self._segments = self._final_segments
                self.emitEvent('40046', data=self._received_data)
            except Exception as error:
                # We just emit an error event on the server side.
                # The file was fully received and the AS2 message was valid.
                # There is not much help to propagate this error to the AS2
                # partner, as if they retry, the operation would most probably
                # fail.
                emit_error_details(
                    error,
                    event_id='40047',
                    data=self._received_data,
                    emitter=self.emitEvent,
                    )

        if not self._received_data['mdn']['to']:
            # No need for MDN.
            return request.site._as2_no_mdn_success_text.encode('utf-8')

        if not self._received_data['mdn']['micalg']:
            # We have unsigned MDN request
            root = self._getMDN(status, details)
        else:
            root = self._getSignedMDN(status, details)

        # Prepare the MDN content.
        raw_root = root.as_string()
        if six.PY2:
            raw_root = raw_root.decode('utf-8')

        parts = raw_root.split('\n\n', 1)
        mdn_content_type = root['Content-Type']
        mdn_body = parts[1].replace('\n', '\r\n')

        if self._received_data['mdn']['async_to']:
            # Handle Async MDN.
            self._scheduler.callLater(
                self._ASYNC_DELAY,
                self._handleAsyncMDN,
                content_type=mdn_content_type,
                body=mdn_body,
                )
            return b''

        # Return sync MDN.
        request.setHeader('content-type', mdn_content_type)
        return mdn_body.encode('utf-8')

    def _parseHeaders(self, request):
        """
        Start the AS2 message processing.

        Example of info headers:

        AS2-From: "\"  as2Name  \""
        AS2-To: 0123456780000

        Date: Wed, 31 Jul 2002 13:34:50 GMT
        From: mrAS2@example.com
        AS2-Version: 1.1
        Subject: Test Case
        Message-Id: <200207310834482A70BF63@\"~~foo~~\">

        ---
        Headers for non-encrypted, non signed:

        Content-Disposition: attachment; filename="some-name.xml"
        Content-Type: application/edi-consent

        ---
        Headers for non-encrypted, signed:

        Content-Type: multipart/signed; protocol="application/pkcs7-signature";
            micalg="sha1"; boundary="===============0891120645055798038=="

        ---
        Header for encrypted, unsigned or signed:
        Content-Type: application/pkcs7-mime; name="smime.p7m";
            smime-type="enveloped-data"

        ---
        Header for compressed
        Content-Type: application/pkcs7-mime; name="smime.p7m";
            smime-type="compressed-data"

        ---
        Headers for MDN:

        Disposition-Notification-To: ignored@email.for.MDN
        Disposition-Notification-Options: signed-receipt-protocol=optional,
            pkcs7-signature; signed-receipt-micalg=optional,sha1,md5
        """
        destination_name = self._getDestinationFilename(request)
        result = {
            # Temporary segments, besides having a temporary path, also
            # have a temporary file extensions.
            'segments': self._segments + [destination_name + '.tmp'],
            'final_segments': self._final_segments + [destination_name],
            'boundary': request.getContentBoundary(),
            'mdn': {
                'to': None,
                'async_to': None,
                'micalg': None,
                'micalg_required': False,
                },
            'mdn_info': 'not requested',
            # Signature algorithm.
            'signature': None,
            'is_encrypted': False,
            'is_compressed': False,
            }
        mdn_importance = 'optional'

        headers = request.requestHeaders.getRawHeaders('content-type')
        if headers:
            # We have a content type.
            name, options = parse_options_header(headers[-1])
            name = name.lower()
            if name == 'application/pkcs7-mime':
                smime_type = options.get('smime-type', '').lower()
                if smime_type == 'compressed-data':
                    result['is_compressed'] = True
                else:
                    result['is_encrypted'] = True
            elif name == 'multipart/signed':
                result['signature'] = canonical_algorithm_name(options.get(
                    'micalg', 'no_micalg_defined'))
            else:
                """
                Unsigned / unencrypted file received.
                """

        # The set of standard AS2 headers. These are also included in the
        # event's data.
        header_names = [
            'message-id',
            'as2-from',
            'as2-to',
            'date',
            'from',
            'subject',
            ]

        for name in header_names:
            headers = request.requestHeaders.getRawHeaders(name)
            data_name = name.replace('-', '_')
            if headers:
                value = force_unicode(headers[-1])
            elif name in ['as2-from', 'as2-to', 'message-id']:
                # These are required headers and we don't auto-fill them.
                value = ''
            else:
                value = 'no-%s-provided' % (name,)

            result[data_name] = value

        # Check options for MDN
        # signed-receipt-protocol=optional,pkcs7-signature;
        #     signed-receipt-micalg=optional,sha1,sha-1,md5'
        headers = request.requestHeaders.getRawHeaders(
            'disposition-notification-options')
        if headers:
            # We only support `pkcs7-signature` protocol
            is_required, digest_name = self._getDigestName(headers[-1])
            result['mdn']['micalg'] = digest_name
            result['mdn']['micalg_required'] = is_required
            if is_required:
                mdn_importance = 'required'

        # Just check that headers are present, ignoring the value as it is
        # specified in RFC section 7.3.  Requesting a Signed Receipt.
        headers = request.requestHeaders.getRawHeaders(
            'disposition-notification-to')
        if headers:
            result['mdn']['to'] = force_unicode(headers[-1])
            result['mdn_info'] = 'sync %s %s' % (
                mdn_importance, result['mdn']['micalg'])

        # See if Async MDN is requested.
        headers = request.requestHeaders.getRawHeaders(
            'receipt-delivery-option')
        if headers:
            async_to = force_unicode(headers[-1])
            result['mdn']['async_to'] = async_to
            result['mdn_info'] = 'async %s %s %s' % (
                async_to, mdn_importance, result['mdn']['micalg'])

        return result

    def _getDestinationFilename(self, request):
        """
        Return the destination file name, as could be implied
        only from the HTTP request headers.

        It is possible for the HTTP headers not to have the name, since the
        name is found in the MIME payload headers.
        """
        default_name = request.site._as2_default_filename
        if not default_name:
            default_name = get_file_name_with_timestamp(
                name='as2-received-file',
                seconds=self._scheduler.seconds(),
                random_mark=self._random.randint(0, 100),
                )

        try:
            default_name = public_format(
                default_name,
                account=self.avatar,
                timestamp=ChevahTimeStamp(self._scheduler.seconds()),
                uuid=six.text_type(self._uuid.uuid4())
                )
        except ServerException as error:
            raise ServerException(
                'Invalid "as2_default_filename" configuration. %s' % (
                    error.message,))

        headers = request.requestHeaders.getRawHeaders(b'content-disposition')
        if not headers:
            # No explicit file name header.
            return default_name

        # Re-generate the raw header in order to use shared code for
        # extracting the filename.
        name = _get_filename(_parse_headers(
            b'content-disposition: ' + headers[-1]))
        if not name:
            # No name found in the header where a file name was expected.
            # Return default.
            return default_name

        # We got a file name in the top-level HTTP headers.
        return name

    def _createConsumer(self, request):
        """
        Return the object receiving the raw payload of the request.
        """
        if self._received_data['is_encrypted']:
            # We have an encrypted, signed or unsigned payload.

            # FIXME:5583:
            # Here we should validate and make sure we have a private key.
            content = EncryptedConsumer(
                private_key=self._private_key,
                certificate_id=self._certificate_id,
                open_callback=self._openPart,
                destination_callback=self._getDestinationConsumer,
                received_data=self._received_data,
                )
            return content

        if self._received_data['is_compressed']:
            # We have an compressed: encrypted, non-encrypted,
            # signed or unsigned payload.
            content = CompressedConsumer(
                destination_callback=self._getDestinationConsumer,
                open_callback=self._openPart,
                received_data=self._received_data,
                )
            return content

        if self._received_data['signature']:
            content = MultiPartFormData(
                boundary=self._received_data['boundary'],
                open_callback=self._openPart,
                )
            content.deferred.addErrback(self._ebMultipart, request)
            return content

        # We have unsigned and unencrypted payload,
        return self._getDestinationConsumer()

    def _openPart(self, headers, raw_headers):
        """
        Called at the start of a multipart.
        """
        if not headers:
            return None

        content_type = headers.get('content-type', {'value': ''})
        content_type_value = content_type['value'].strip().lower()

        if not content_type_value:
            return None

        if content_type_value.startswith('application/pkcs7-signature'):
            # We load the signature in memory.
            if self._received_signature:
                raise ServerException('Only a single signature is supported.')

            self._received_signature = ReceivedSignature()
            return self._received_signature

        if content_type_value.startswith('application/pkcs7-mime'):
            # Compressed signed content.
            self._received_data['is_compressed'] = True

            self._digest.update(raw_headers)
            return CompressedConsumer(
                destination_callback=self._getDestinationConsumer,
                open_callback=self._openPart,
                received_data=self._received_data,
                digest_updater=self._digest.update,
                )

        # We have a plain-text signed file, with content inside the multipart.
        return self._getDestinationConsumer(
            _get_filename(headers), raw_headers)

    def _getDestinationConsumer(
            self, name=None, raw_headers=None, update_digest=True):
        """
        Get the consumer used to store the actual file.
        """
        if self._received_file:
            raise ServerException(
                'Receiving of multiple files in the same request '
                'is not supported.'
                )

        if name:
            # Update the destination file name, with the file name found
            # inside the multi-part headers.
            self._segments[-1] = name + '.tmp'
            self._final_segments[-1] = name

        if raw_headers:
            self._digest.update(raw_headers)

        digest = None
        if update_digest:
            digest = self._digest.update

        # Create any parent directories for the temporary AS2 path.
        self._shell.makeDirectories(base=[], relative=self._segments[:-1])

        self._received_file = self._shell.openForWriting(
            self._segments, digest=digest, request=self._request)

        # We are ready to read the whole body.
        self._request.maximumBodyLength = None

        emit_account_login_paths(
            id='40045',
            avatar=self._avatar,
            emitter=self.emitEvent,
            path='/',
            data=self._received_data,
            )
        return self._received_file

    def _ebMultipart(self, failure, request):
        """
        Called when form parser fails.
        """
        failure.check(FormLengthExceededException)

        request.loseConnection(
            code=http.REQUEST_ENTITY_TOO_LARGE,
            message=str(failure.value),
            )

    def _getResponseHeaders(self, data):
        """
        Return a dict with the response headers.
        """
        result = {}

        result['AS2-Version'] = '1.2'
        result['AS2-From'] = data['as2_to']
        result['AS2-To'] = data['as2_from']
        result['Message-ID'] = data['message_id']

        return result

    def _getMDN(self, status, details):
        """
        Return the MDN mime message with `details` as the text part.

        For non-encrypted non-signed MIC is computed over the payload.
        For non-encrypted and signed MIC is computed over the multipart section
            of the file payload including multipart headers.
        """
        root = ChevahMIMEMultipart('report')
        root.set_param(
            native_string('report-type'),
            native_string('disposition-notification'),
            )

        details = details + '\n'
        try:
            details.encode('ascii')
            # Go with ASCII as is easier to read.
            info = MIMEText(details, 'plain', 'us-ascii')
        except UnicodeEncodeError:  # noqa:cover
            # MDN content should always be ASCII.
            # This is here to alert if it's otherwise.
            # If we need to support Unicode MDN text
            # we should use: MIMEText(details, 'plain', 'utf-8')
            raise RuntimeError('Only ASCII MDN is supported.')

        root.attach(info)

        digest_name, digest_value = self._getDigestValue()

        # Make sure we don't have trailing newlines.
        mdn_content = """Reporting-UA: SFTPPlus
Original-Recipient: rfc822; "%(as2_to)s"
Final-Recipient: rfc822; "%(as2_to)s"
Original-Message-ID: %(message_id)s
Received-Content-MIC: %(digest_value)s, %(digest_name)s
Disposition: automatic-action/MDN-sent-automatically; %(status)s
""" % {
            'as2_to': self._received_data['as2_to'],
            'message_id': self._received_data['message_id'],
            'digest_value': base64.b64encode(digest_value).decode('ascii'),
            'digest_name': digest_name,
            'status': status,
            }

        mdn = MIMEBase('message', 'disposition-notification')
        if six.PY2:
            mdn_content = mdn_content.encode('utf-8')
        mdn.set_payload(mdn_content)
        root.attach(mdn)
        return root

    def _getMDNStatus(self):
        """
        Return the status for the MDN header.

        disposition and info can be sent to the remote party.

        error_details is for internal usage.

        Is the tuple of (disposition, info, error)
        """
        info = [
            'The message has been received.',
            ]

        # Add some non common messages to help with testing and might help
        # general troubleshooting for support.
        if self._received_data['is_encrypted']:
            info.append('The message has been received encrypted.')

        if self._received_data['is_compressed']:
            info.append('The message has been received compressed.')

        if (
            not self._received_data['mdn']['micalg']
            or not self._received_signature
                ):
            # Just the MDN without any signature.
            status = 'processed'
            info.append('The integrity was not verified.')
            return status, ' '.join(info), None

        if self._received_signature:
            try:
                signed_algorithm, signed_digest, signer = (
                    self._received_signature.getDigestValue())
            except ServerException as error:
                status = 'failed/Failure: MIC validation failed'
                info.append('The integrity validation failed.')
                return status, ' '.join(info), error

        # FIXME:5424:
        # Add configuration to enforce the received signature.
        # See what error is raised by ArcESB
        if not signed_digest and False:
            message = 'A signature for the payload is required.'
            status = 'failed/Failure: MIC required. %s' % (message)
            info.append('The integrity validation failed.')
            return status, ' '.join(info), ServerException(message)

        received_algorithm, received_digest = self._getDigestValue()

        # If the signature has an explicit digest via the signed attributes
        # do a pre-validation in order to show a better error message.
        # The message might not have an explicit digest. Then this check
        # is skipped and we validate the signature later against the full
        # payload.
        if signed_digest is not None and received_digest != signed_digest:
            message = (
                'Expecting signed digest %s %s '
                'but in the signature attributes we got %s %s.'
                ) % (
                    received_algorithm,
                    base64.b64encode(received_digest).decode('ascii'),
                    signed_algorithm,
                    base64.b64encode(signed_digest).decode('ascii'),
                    )
            status = 'failed/Failure: MIC not valid. %s' % (message,)
            info.append('The integrity was not confirmed in this message.')
            return status, ' '.join(info), ServerException(message)

        account_certificates = self._avatar.getAS2Certificates()
        try:
            self._validateSignature(signer, account_certificates)
        except ServerException as error:
            status = (
                'failed/Failure: signature not valid. '
                'Digest %s matched the received data but the signature '
                'could not be validated.'
                ) % (base64.b64encode(received_digest).decode('ascii'))
            info.append('The integrity was not confirmed in this message.')
            return status, ' '.join(info), error

        # If we are here, all is OK.
        status = 'processed'
        info.append('The integrity is confirmed in this message.')
        return status, ' '.join(info), None

    def _getSignedMDN(self, status, info):
        """
        Return the signed MDN MIME message.
        """
        data = self._received_data
        root = ChevahMIMEMultipart('signed')
        root.set_param('micalg', data['mdn']['micalg'])
        root.set_param('protocol', 'application/pkcs7-signature')

        mdn = self._getMDN(status, info)
        root.attach(mdn)

        fp = StringIO()
        Generator(fp, maxheaderlen=0).flatten(mdn, unixfrom=False)
        data_to_sign = fp.getvalue()
        if six.PY2:
            data_to_sign = data_to_sign.decode('utf-8')
        data_to_sign = data_to_sign.replace('\n', '\r\n')

        mdn_digest = self._getDigest()
        mdn_digest.update(data_to_sign.encode('utf-8'))

        # FIXME:5583:
        # Here we should make sure we have a valid certificate.
        signature = MIMEApplication(
            create_signature(
                mdn_digest.finalize(),
                digest=mdn_digest,
                private_key=self._private_key,
                certificate_asn1=self._certificate_asn1,
                certificate_id=self._certificate_id,
                ),
            'pkcs7-signature')
        signature.set_param('name', 'smime.p7s')
        signature['Content-Disposition'] = 'attachment; filename=smime.p7s'
        root.attach(signature)

        return root

    def _getDigest(self):
        """
        Return the digest method for validating the content.
        """
        digest_class = _SUPPORTED_DIGEST.get(
            self._received_data['mdn']['micalg'], None)

        if not digest_class:
            digest_class = _SUPPORTED_DIGEST['sha256']

        return hashes.Hash(
            digest_class(), backend=DEFAULT_BACKEND)

    def _getDigestName(self, value):
        """
        Return the digest algorithm to be used, as requested in the header

        Return a tuple of (is_required, digest_name).

        signed-receipt-protocol=optional,pkcs7-signature;
            signed-receipt-micalg=optional,sha1,md5

        signed-receipt-protocol=required,pkcs7-signature;
            signed-receipt-micalg=optional,sha1'
        """
        parts = [p.strip().lower() for p in value.split(';') if p.strip()]

        is_required = False
        digest_name = None

        for part in parts:
            if not part.startswith('signed-receipt-micalg'):
                continue

            value = part.split('=', 1)
            if len(value) != 2:
                raise ServerException('Invalid "signed-receipt-micalg" value.')

            values = [v.strip() for v in value[1].split(',') if v.strip()]

            if values[0] == 'required':
                is_required = True

            for name in values[1:]:
                name = canonical_algorithm_name(name)
                if name not in list(_SUPPORTED_DIGEST.keys()):
                    continue
                digest_name = name
                # Stop on first match.
                break

            # Don't search for other options.
            break

        return is_required, digest_name

    def _getDigestValue(self):
        """
        Return a tuple with digest algorithm name
        and the final binary digest value for the received file.
        """
        if not self._digest:
            raise ServerException('No digest was requested.')

        if not self._digest_value:
            self._digest_value = self._digest.finalize()

        name = self._digest.algorithm.name

        return name, self._digest_value

    def _validateSignature(self, signer, certificates):
        """
        Raise an exception if the signer is not among the one
        configured by the account `certificates` or if the signer signature
        is invalid.
        """
        # RSA algorithm and padding for the signature.
        signer_algorithm = canonical_algorithm_name(signer[
            'signature_algorithm']['algorithm'].native)
        # Hash method of the signature.
        digest_algorithm = canonical_algorithm_name(signer[
            'digest_algorithm']['algorithm'].native)
        # As a common option, the signature is done against a set of attributes
        # instead of the whole payload.
        # And the attributes is basically a single attribute called
        # `message_digest`.
        signature_data = signer['signed_attrs'].untag().dump()
        # The signature itself.
        signature_value = signer['signature'].native

        if digest_algorithm not in list(_SUPPORTED_DIGEST.keys()):
            raise ServerException(
                'Signer signature digest not supported: %s' % (
                    digest_algorithm,))

        if not certificates:
            raise ServerException(
                'No "as2_certificates" configured for the account.')

        if signer_algorithm not in 'rsassa_pkcs1v15':
            # FIXME:5421:
            # Add support for rsassa_pss.
            raise ServerException(
                'Only rsassa_pkcs1v15 signature algorithm is supported.')

        signer_serial = signer['sid'].native['serial_number']
        signer_issuer = _get_asn1crypto_subject(
            signer['sid'].native['issuer'])
        signer_id = 'Issued by "%s" with serial %s' % (
            signer_issuer, signer_serial)

        serial_match = False
        configured_ids = []
        for certificate in certificates:
            certificate_serial = certificate.serial_number
            certificate_issuer = certificate.issuer.rfc4514_string()
            certificate_subject = certificate.subject.rfc4514_string()
            configured_ids.append('"%s" issued by "%s" with serial %s' % (
                certificate_subject, certificate_issuer, certificate_serial))

            if certificate_serial == signer_serial:
                # For now, only validate serial number.
                # The probability of having the same serial number from
                # different issuers is low and the step will fail anyway
                # later at the signature validation.
                serial_match = True

        configured_ids = ' | '.join(configured_ids)

        if not serial_match:
            raise ServerException(
                'Signature signed by unknown certificate. '
                'Received: %s. '
                'Configured: %s' % (
                    signer_id, configured_ids)
                )

        for certificate in certificates:
            public_key = certificate.public_key()

            signature_digest = _SUPPORTED_DIGEST[digest_algorithm]()
            if not signature_data:
                # We don't have signed attributes, so we use the whole
                # payload and a pre-computed hash.
                signature_data = self._digest_value
                signature_digest = Prehashed(signature_digest)

            try:
                public_key.verify(
                    signature_value,
                    signature_data,
                    padding.PKCS1v15(),
                    signature_digest,
                    )
                # FIXME:5384:
                # Log which certificate was used for signing.
                # Maybe have a separate event to inform a valid/invalid
                # signature and this event can contain all the signing info.
                # including the Digest value and file size.
                # All good.
                return
            except InvalidSignature:
                """
                Continue with the other certificate.
                """

        raise ServerException(
            'The digest is not signed by any of the trusted certificates. '
            'Received: %s. '
            'Configured: %s' % (
                signer_id, configured_ids)
            )

    def _handleAsyncMDN(self, content_type, body):
        """
        Send an async MDN over HTTP/HTTPS.
        """
        url = self._received_data['mdn']['async_to']

        def cb_get_context(ignored):
            """
            Prepare the HTTPS remote server validation.
            """
            parts = parse_url(url)

            if parts['protocol'] == 'http':
                return None

            if parts['protocol'] != 'https':
                raise ServerException(
                    'Unknown MDN URL received from the AS2 partner.')

            server_name = parts['address']

            service_configuration = (
                self._avatar.component.service_configuration)

            # When we run AS2 server over HTTP, we don't have HTTPS config
            # so we just go without HTTPS checks.
            cipher_list = None
            allowed_methods = None
            if service_configuration.parent.use_ssl:
                cipher_list = service_configuration.ssl_cipher_list
                allowed_methods = service_configuration.ssl_allowed_methods

            context = OpenSSLContextFactory(
                ca=self._avatar.getAS2MDNCA(),
                allowed_methods=allowed_methods,
                cipher_list=cipher_list,
                remote_names=[server_name],
                # It should never fail as we don't use CRL.
                on_error=None,
                event_emitter=self.emitEvent,
                scheduler=self._scheduler,
                )
            deferred = context.start()
            deferred.addCallback(lambda _: context.getContext())
            return deferred

        def cb_post_mdn(context):
            """
            Called when we got the SSL context and we can trigger the
            HTTP request for sending the MDN content.
            """
            agent = NonPersistentAgent(
                event_emitter=self.emitEvent,
                ssl_context=context,
                )

            headers = self._getResponseHeaders(self._received_data)
            headers['content-type'] = content_type
            return agent.post(
                url,
                headers=headers,
                body=body,
                retry_count=5,
                retry_increase=60,
                )

        def eb_async_mdn(failure):
            """
            Called when failing to send the async MDN.
            """
            emit_error_details(
                event_id='40052',
                error=failure.value,
                data=self._received_data,
                emitter=self.emitEvent,
                )

        deferred = defer.succeed(None)
        deferred.addCallback(cb_get_context)
        deferred.addCallback(cb_post_mdn)
        deferred.addErrback(eb_async_mdn)
        return deferred


class ReceivedSignature(object):
    """
    Handle the received signature data inside a multi-part.
    """

    def __init__(self):
        self._raw = []
        self._error = None
        self._closed = False

    @property
    def closed(self):
        return self._closed

    def write(self, data):
        """
        Called when signature data is received.
        """
        self._raw.append(data)

    def close(self):
        """
        Called when all the data was received.
        """
        self._closed = True

    def interupts(self, error):
        """
        Called when there was an error while receiving the data.
        """
        self._error = error

    def getDigestValue(self):
        """
        Return a tuple of digest algorithm name,
        digest value in the signature and
        the signer info.

        Returns `None` as `data_digest` if there are no signed attributes
        and the signature is on the whole payload.

        It does not validates that the digest is properly signed or that is
        matches the received content.
        """
        signature = cms.ContentInfo.load(base64.b64decode(b''.join(self._raw)))
        if signature['content_type'].native != 'signed_data':
            raise ServerException('Not a signed request.')

        signatures = signature['content']['signer_infos']

        if not signatures:
            raise ServerException('Signer not found in the signature part.')

        # Only the first signature is used.
        signer = signatures[0]

        # The digest for the payload as found in the signature.
        data_digest = None
        for attribute in signer['signed_attrs']:
            if attribute.native['type'] != 'message_digest':
                continue
            data_digest = b''.join(attribute.native['values'])
            if data_digest == b'':
                # The code might have worked even without this failure,
                # as the code should have fail later since an empty value
                # would not be equal with the actual digest.
                # But better to fail here with an explicit message.
                # Most of the time, if the client is using signed attributes
                # it would set a value here.
                # This most probably means a bug in the client implementation.
                raise ServerException(
                    'Empty digest found in the signed attributes.')
            break

        digest_algorithm = signer['digest_algorithm']['algorithm'].native

        return (digest_algorithm, data_digest, signer)


def _get_asn1crypto_subject(raw_subject):
    """
    Return the subject text representation in RFC 4514 format
    from the internal OrderdDict used by asn1crypto.

    Example: CN=localhost,OU=Chevah Test,O=Chevah,C=GB
    """
    MAPPING = {
        'country_name': 'C',
        'organization_name': 'O',
        'organizational_unit_name': 'OU',
        'common_name': 'CN',
        'locality_name': 'L',
        'state_or_province_name': 'ST',
        }
    parts = []
    for part_name in raw_subject:
        short_name = MAPPING.get(part_name, None)
        if not short_name:
            # Ignored unsupported parts.
            continue
        parts.append('%s=%s' % (short_name, raw_subject[part_name]))

    return ','.join(parts)


@implementer(IConsumer)
class ContentInfoConsumer(object):
    """
    Shared code for consuming and processing cms.ContentInfo.

    Is the shared code for doing decryption and decompression.
    """
    # List of step used to parse the stream.
    # Should be defined by each subclass
    _steps = None

    # Callback when all the of the payload headers were received.
    _onHeadersReceived = None

    def __init__(self):
        self._producer = None
        self._decoder = asn1.StreamingASN1Decoder()
        # Consumer of the decrypted payload.
        self._consumer = None
        # Any error found while receiving/processing the payload.
        # We need to raise it later when producer is unregistred.
        self._error = None
        # Header of the multi-part payload.
        self._headers = {}
        # Data accumulated while parsing the headers.
        self._headers_buffer = b''

        # Result of the last step.
        self._last_tag = None

        self._plainReceived = self._headersReceived

    def registerProducer(self, producer):
        """
        Signal that we are receiving data from a streamed request.

        Only stream producer is supported.
        """
        self._producer = producer

    def unregisterProducer(self):
        """
        Called when all data was received.

        This is where we can raise any error found during the received data.
        """
        self._producer = None
        if self._consumer:
            self._consumer.write(self._finalize())
            self._consumer.unregisterProducer()
            self._consumer = None

        if self._error:
            raise self._error

    def close(self):
        """
        Called when all the data was received.
        """

    def write(self, data):
        """
        Called by transport when raw content is received
        (encrypted or compressed).
        """
        if self._error:
            # Ignore any received content as we have an error.
            # We need to "consume" the content in order to send an HTTP
            # response back.
            return

        try:
            self._decoder.dataReceived(data)
            self._process()
        except Exception as error:
            self._error = error

    def seek(self, offset, reference_point=0):
        """
        We don't do nothing here, but this is called by the HTTP server
        after all data was received in stream mode.
        """

    def _process(self):
        """
        Called after raw encrypted/encapsulated data was received.
        """
        if not self._steps:
            return self._consumeContent()

        while self._steps:
            try:
                self._last_tag = self._steps[0](self._last_tag)
            except asn1.ASN1WantMore:
                return
            # Step done.
            self._steps.pop(0)

        self._consumeContent()

    def _consumeContent(self):
        """
        Called each time we got raw encrypted data.
        """
        if not self._last_tag:
            try:
                self._last_tag = self._decoder.getTag()
            except asn1.ASN1WantMore:
                # Next chunk not ready
                return

        while self._last_tag.raw != b'\x00\x00':
            encrypted = self._decoder.flush()

            if encrypted is None:
                # Current chunk done.
                # Go to next chunk and prepare to fail to read the full
                # new chunk.
                self._last_tag = None
                try:
                    self._last_tag = self._decoder.getTag()
                except asn1.ASN1WantMore:
                    # Next chunk not ready yet.
                    return

                if self._last_tag.raw == b'\x00\x00':
                    return

                encrypted = self._decoder.flush()

            if not encrypted:
                # No more data available.
                return

            self._plainReceived(self._transform(encrypted))

    def _headersReceived(self, data):
        """
        Called when the received data was decrypted and we receive the
        headers data (or part of the headers).
        """
        self._headers_buffer += data

        if len(self._headers_buffer) < 4:
            # We need more data to parse the headers.
            return

        parts = self._headers_buffer.split(b'\r\n\r\n', 1)

        if len(parts) == 1:
            # We don't have yet the end of header marker.
            return

        # We will no longer be receiving headers. Move to body parsing.
        self._plainReceived = self._bodyReceived

        raw_headers = parts[0] + b'\r\n\r\n'
        headers = _parse_headers(parts[0])
        self._consumer = self._onHeadersReceived(headers, raw_headers)

        # Here we should have request.transport
        # or maybe something from AS2Receiver,
        self._consumer.registerProducer(self)

        # Continue processing the remaining data as body.
        self._plainReceived(parts[1])

    def _bodyReceived(self, data):
        """
        Called when multi-part body data (excluding headers) was decrypted.
        """
        self._consumer.write(data)

    def _getTag(self, ignored):
        """
        Parse the next tag.

        `ignore` argument is passed to fit to the generic tag parsing
        framework that we built.
        """
        return self._decoder.getTag()

    def _consumeContext(self, tag):
        """
        The encrypted data might be constructed from multiple chunks.
        """
        if tag.cls != asn1.Classes.Context:
            # Not a context
            return tag

        if tag.length == 0:
            # A context with unknown length.
            return self._decoder.getTag()

        if tag.type == asn1.Types.Constructed:
            # A constucted context
            return self._decoder.getTag()

        # Just return the context as it should be read.
        return tag


class EncryptedConsumer(ContentInfoConsumer):
    """
    A consumer for the SMIME encrypted payload.

    It will decrypt the data and then delegated it to the multi-part
    consumer.
    """

    def __init__(
        self, private_key, certificate_id,
        open_callback, destination_callback,
        received_data,
            ):
        super(EncryptedConsumer, self).__init__()
        self._received_data = received_data
        # Private key configured to decrypt the received content.
        self._private_key = private_key
        # Certificate assocaited with the private key.
        self._certifificate_id = certificate_id
        # Called before each part of the multi-part body.
        self._open_callback = open_callback
        # Called to get the file on disk where to store decrypted data.
        self._destination_callback = destination_callback

        # Information about the message recipients for which this message
        # was encrypted.
        self._recipient_infos = []

        # Information about the encryption method and key.
        self._encryption_algorithm = None

        # List of steps called to decrypt the payload.
        self._steps = [
            self._getTag,  # The root ContentInfo sequence.
            self._getTag, self._decoder.read,  # ContentType enveloped_data
            self._getTag,  # Any tag for content.
            self._getTag,  # EnvelopedData sequence
            self._getTag, self._decoder.read,  # CMSVersion
            self._getTag,  self._parseRecipientInfos,
            self._getTag,  # EncryptedContentInfo sequence,
            self._getTag, self._decoder.read,  # ContentType
            self._getTag, self._parseEncryptionAlgorithm,
            self._getTag, self._consumeContext,
            ]

    def tell(self):
        """
        We can't tell the size of an encrypted payload and we don't support
        seek/tell.

        This is called by Twisted web framework to do it's own multi-part
        handling but we are not using that code as we do our own handling.

        See: https://twistedmatrix.com/trac/ticket/6928
        """
        return None

    def _transform(self, data):
        """
        Do the decryption transformation.
        """
        result = self._decipher.update(data)
        return self._unpadder.update(result)

    def _finalize(self):
        """
        Do the final decryption transformation.
        """
        result = self._decipher.finalize()
        result = self._unpadder.update(result)
        return result + self._unpadder.finalize()

    def _onHeadersReceived(self, headers, raw_headers):
        """
        Called after all the headers were received for the decrypted payload.

        It should return the consumer for the body.
        """
        content_type = headers.get(
            'content-type',
            {'value': '', 'options': {}},
            )
        content_type_value = content_type['value'].strip().lower()
        if content_type_value.startswith('multipart/'):
            # We have the signed payload, inside the encrypted payload.
            # There should always be an `micalg` value.
            # If missing, the AS2 peer is broken.
            micalg = content_type['options'].get('micalg', b'missing_micalg')
            self._received_data['signature'] = canonical_algorithm_name(micalg)

            boundary = headers['content-type']['options']['boundary']
            return MultiPartFormData(
                boundary=boundary,
                open_callback=self._open_callback,
                )

        if content_type_value.startswith('application/pkcs7-mime'):
            # We have compressed data that is not encrypted.
            # Might be signed.
            self._received_data['is_compressed'] = True
            return CompressedConsumer(
                destination_callback=self._destination_callback,
                open_callback=self._open_callback,
                received_data=self._received_data,
                )

        # Unsigned encrypted body.
        return self._destination_callback(
            _get_filename(headers), raw_headers)

    def _parseRecipientInfos(self, tag):
        """
        First we get the recipients.
        """
        self._recipient_infos = cms.RecipientInfos.load(
            self._decoder.dump(tag))

    def _parseEncryptionAlgorithm(self, tag):
        """
        Last we get the encryption info.
        """
        self._encryption_algorithm = algos.EncryptionAlgorithm.load(
            self._decoder.dump(tag))

        self._decipher, self._unpadder = self._getDecipher()

    def _getDecipher(self):
        """
        Return the cipher used to decrypt the payload.
        """
        key = self._getEncryptionKey()
        algorithm = self._encryption_algorithm.native['algorithm']
        parameters = self._encryption_algorithm.native['parameters']

        if algorithm not in _SUPPORTED_ENCRYPTIONS:
            raise ServerException(
                'Encryption algorithm %s is not supported.' % (algorithm))

        algorithm = _SUPPORTED_ENCRYPTIONS[algorithm](key)

        decipher = Cipher(
            algorithm,
            CBC(parameters),
            backend=DEFAULT_BACKEND,
            ).decryptor()

        unpadder = PKCS7(algorithm.block_size).unpadder()

        return decipher, unpadder

    def _getEncryptionKey(self):
        """
        Return the raw key used to encrypt the content.
        """
        if not self._recipient_infos:
            raise ServerException(
                'Message does not contain recipient_infos.')

        recipient = self._recipient_infos[0]
        encryption_algorithm = recipient.native[
            'key_encryption_algorithm']['algorithm']

        if encryption_algorithm != 'rsaes_pkcs1v15':
            raise ServerException(
                'Only RSA encryption is supported.')

        rid = recipient.native.get('rid', None)
        if rid is None:
            raise ServerException(
                'No encryption certificate found in the payload.')

        serial_number = rid.get('serial_number', None)
        if serial_number is None:
            raise ServerException(
                'Found encryption certificate in the payload '
                'without a serial number.'
                )

        expected_serial = self._certifificate_id.native['serial_number']
        if serial_number != expected_serial:
            raise ServerException(
                'The received message is encryption using an unknown '
                'certificate. Expecting ID: %s Received ID: %s.' % (
                    expected_serial, serial_number)
                )

        try:
            return self._private_key.decrypt(
                recipient.native['encrypted_key'],
                padding.PKCS1v15()
                )
        except ValueError as error:
            raise ServerException(
                'Invalid key used for encryption. %s' % (force_unicode(error)),
                )


class CompressedConsumer(ContentInfoConsumer):
    """
    A consumer for the SMIME compressed payload.

    Also a multi-part consumer.

    It will decompress the data and then delegated it to the enveloped
    handler.

    See https://tools.ietf.org/html/rfc5402
    """

    def __init__(
        self, received_data, open_callback, destination_callback,
        digest_updater=None
            ):
        super(CompressedConsumer, self).__init__()
        # List of steps called to decrypt the payload.
        self._steps = [
            self._getTag,  # The root ContentInfo.
            self._getTag, self._decoder.read,  # ContentType - compressed_data
            self._getTag,  # Any tag - content.
            self._getTag,  # CompressedData sequence
            self._getTag, self._decoder.read,  # CMSVersion
            self._getTag,  self._parseAlgorithm,
            self._getTag,  # EncapsulatedContentInfo sequence,
            self._getTag, self._decoder.read,  # ContentType - data
            self._getTag, self._consumeContext,
            ]
        self._decompressor = zlib.decompressobj()
        self._received_data = received_data
        # Called to get the file on disk where to store decrypted data.
        self._destination_callback = destination_callback
        # Called to handle the signed content, that is compressed.
        self._open_callback = open_callback
        self._closed = False

        # Set to a callback when the raw compressed data is signed.
        # Set to None, when the signature is on the uncompressed data.
        self._digest_updater = digest_updater

    def tell(self):
        """
        We can't tell the size of the compressed payload and we don't support
        seek/tell.

        This can be called by Twisted web framework to do it's own multi-part
        handling but we are not using that code as we do our own handling.

        See: https://twistedmatrix.com/trac/ticket/6928
        """
        return None

    def write(self, data):
        """
        Called by transport when compressed raw content is received.
        """
        # Here we update the digest even on error.
        # On error the digest is ignored.
        # Computing the digest should not be CPU intensive.
        # We prefer to just call super() with all its implementation
        # than duplicating error ignore here.
        if self._digest_updater:
            self._digest_updater(data)

        return super(CompressedConsumer, self).write(data)

    @property
    def closed(self):
        return self._closed

    def close(self):
        """
        Called when all the data was received.
        """
        self._closed = True
        self.unregisterProducer()

    def interupts(self, error):
        """
        Called when there was an error while receiving the data.
        """
        self._error = error

    def _parseAlgorithm(self, tag):
        """
        Called when we got the compression algorithms.
        """
        struct = cms.CompressionAlgorithm().load(self._decoder.dump(tag))
        if struct.native['algorithm'] != 'zlib':
            raise ServerException('Only zlib compression is supported.')

    def _transform(self, data):
        """
        Do the decryption transformation.
        """
        return self._decompressor.decompress(data)

    def _finalize(self):
        """
        Do the final decryption transformation.
        """
        return self._decompressor.flush()

    def _onHeadersReceived(self, headers, raw_headers):
        """
        Called after all the headers were received for the decrypted payload.
        """
        content_type = headers.get(
            'content-type',
            {'value': '', 'options': {}},
            )
        content_type_value = content_type['value'].strip().lower()

        if content_type_value.startswith('multipart/signed'):
            # The signature is compressed.
            # There should always be an `micalg` value.
            # If missing, the AS2 peer is broken.
            micalg = content_type['options'].get('micalg', b'missing_micalg')
            self._received_data['signature'] = canonical_algorithm_name(micalg)
            boundary = content_type['options']['boundary']
            return MultiPartFormData(
                boundary=boundary,
                open_callback=self._open_callback,
                )

        update_digest = True
        if self._digest_updater:
            update_digest = False
            raw_headers = None

        return self._destination_callback(
            _get_filename(headers),
            raw_headers=raw_headers,
            update_digest=update_digest,
            )


def _get_filename(headers):
    """
    Return the name of the file as found in the already parsed `headers`.

    The name can be encoded using MIME encoded-words.

    Returns None if no name is found or if an generic S/MIME placeholder is
    is found.
    """
    disposition = headers.get('content-disposition', None)

    if not disposition:
        return None

    if disposition['value'].lower() != 'attachment':
        return None

    if 'filename' not in disposition['options']:
        return None

    name = ''
    raw_name = disposition['options']['filename']
    try:
        encoded_name, encoding = decode_header(raw_name)[0]
        if not encoding:
            name = force_unicode(encoded_name)
        else:
            name = encoded_name.decode(encoding)
    except UnicodeEncodeError:
        return raw_name

    name = name.strip(' \'\t"')
    if name in ['smime.p7m', 'smime.p7z']:
        return None

    return name


def _parse_headers(data):
    """
    Read raw headers `data` and return the headers as `dict`.
    """
    headers = {}

    # Put all headers, one header per line.
    header_lines = []
    for line in data.split(b'\r\n'):
        if line.startswith(b' ') or line.startswith(b'\t'):
            # Header continuation
            # The multi-line parsing is observed in Seeburger product.
            # For HTTP this is absolete behaviour
            # https://www.rfc-editor.org/rfc/rfc7230#section-3.2.4
            # But Seeburger is using it in the MIME context.
            header_lines[-1] = header_lines[-1] + line.strip()
            continue

        # Have a new header.
        header_lines.append(line)

    # Parse each line up to last line, as the last line might not be
    # complete.
    for line in header_lines:
        update_headers(headers, line)

    return headers


def get_AS2_key_and_certificate(
    component_configuration, root_configuration,
    certificate_option, key_option
        ):
    """
    Return a dict with private key, certificate ID and DER as used for
    CMS signature ASN.1 structure.

    Return a structure without values when no certificate is configured.
    """
    certificate_configuration = getattr(
        component_configuration, certificate_option)
    key_configuration = getattr(component_configuration, key_option)

    result = {
        'key': None,
        'asn1': None,
        'id': None,
        }

    if not certificate_configuration:
        # Use general configuration.
        certificate_configuration = root_configuration.ssl_certificate
        key_configuration = root_configuration.ssl_key
    else:
        if not key_configuration:
            key_configuration = root_configuration.ssl_key

    if not certificate_configuration:
        # For backward compatibly don't use any key and disable AS2.
        return result

    cert, key, chain = read_PEM_configuration(
        certificate_configuration, key_configuration,
        password=None,
        certificate_option=certificate_option,
        key_option=key_option,
        )

    result['asn1'] = x509.Certificate.load(
        crypto.dump_certificate(crypto.FILETYPE_ASN1, cert))

    issuer_dict = OrderedDict()
    for name, value in cert.get_issuer().get_components():
        name = name.decode('utf-8')
        oid = '.'.join(map(str, name_to_OID[name]))
        issuer_dict[oid] = value.decode('utf-8')

    result['id'] = cms.SignerIdentifier({
        "issuer_and_serial_number": cms.IssuerAndSerialNumber({
            "issuer": PreOrderedName.build(issuer_dict),
            "serial_number": cert.get_serial_number(),
            }),
        })

    result['key'] = key.to_cryptography_key()

    return result


# FIXME:5421:
# See why rsassa_pss don't work.
# We don't want to use rsassa_pkcs1v15 as it is less secure.
_SIGNATURE_ALGORITHM = algos.SignedDigestAlgorithm({
    "algorithm": algos.SignedDigestAlgorithmId('rsassa_pkcs1v15')})


class SmimeCapability(core.Sequence):
    """
    Define the possible list of Smime Capability.
    """

    _fields = [
        ("0", core.Any, {"optional": True}),
        ("1", core.Any, {"optional": True}),
        ("2", core.Any, {"optional": True}),
        ("3", core.Any, {"optional": True}),
        ("4", core.Any, {"optional": True}),
        ]


class SmimeCapabilities(core.Sequence):
    """
    Define the Smime Capabilities supported by pyas2.
    """

    _fields = [
        ("0", SmimeCapability),
        ("1", SmimeCapability, {"optional": True}),
        ("2", SmimeCapability, {"optional": True}),
        ("3", SmimeCapability, {"optional": True}),
        ("4", SmimeCapability, {"optional": True}),
        ("5", SmimeCapability, {"optional": True}),
        ]


_SMIME_CAPABILITIES = OrderedDict([
    (
        "0",
        OrderedDict(
            [("0", core.ObjectIdentifier("2.16.840.1.101.3.4.1.42"))]
            ),
        ),
    (
        "1",
        OrderedDict(
            [("0", core.ObjectIdentifier("2.16.840.1.101.3.4.1.2"))]
            ),
        ),
    (
        "2",
        OrderedDict([("0", core.ObjectIdentifier("1.2.840.113549.3.7"))]),
        ),
    (
        "3",
        OrderedDict([
            ("0", core.ObjectIdentifier("1.2.840.113549.3.2")),
            ("1", core.Integer(128)),
            ]),
        ),
    (
        "4",
        OrderedDict([
            ("0", core.ObjectIdentifier("1.2.840.113549.3.4")),
            ("1", core.Integer(128)),
            ]),
        ),
    ])


def create_signature(
        digest_value, digest, certificate_asn1, certificate_id, private_key):
    """
    Return the CMS (RFC5652) PKCS7 (RFC2315) signature payload for the data.
    """
    algo_name = digest.algorithm.name
    if six.PY2:
        algo_name = algo_name.decode('ascii')
    digest_algorithm = algos.DigestAlgorithm({
        "algorithm": algos.DigestAlgorithmId(algo_name)})

    signed_attributes = cms.CMSAttributes([
        cms.CMSAttribute({
            "type": cms.CMSAttributeType("content_type"),
            "values": cms.SetOfContentType([cms.ContentType("data")]),
            }),
        cms.CMSAttribute({
            "type": cms.CMSAttributeType("signing_time"),
            "values": cms.SetOfTime([
                cms.Time({
                    "utc_time": core.UTCTime(
                        datetime.utcnow().replace(tzinfo=UTC))
                    })
                ]),
            }),
        cms.CMSAttribute({
            "type": cms.CMSAttributeType("message_digest"),
            "values": cms.SetOfOctetString(
                [core.OctetString(digest_value)]
                ),
            }),
        cms.CMSAttribute({
            "type": cms.CMSAttributeType("1.2.840.113549.1.9.15"),
            "values": cms.SetOfAny(
                [core.Any(SmimeCapabilities(_SMIME_CAPABILITIES))]
                ),
            }),
        ])

    signature_value = private_key.sign(
        signed_attributes.dump(),
        padding.PKCS1v15(),
        _SUPPORTED_DIGEST[digest.algorithm.name](),
        )

    result = cms.ContentInfo({
        "content_type": cms.ContentType("signed_data"),
        "content": cms.SignedData({
            "version": cms.CMSVersion("v1"),
            "digest_algorithms": cms.DigestAlgorithms([digest_algorithm]),
            "encap_content_info": cms.ContentInfo({
                "content_type": cms.ContentType("data")}),
            "certificates": cms.CertificateSet([
                cms.CertificateChoices({
                    "certificate": certificate_asn1})]),
            "signer_infos": cms.SignerInfos([cms.SignerInfo({
                "version": cms.CMSVersion("v1"),
                "sid": certificate_id,
                "digest_algorithm": digest_algorithm,
                "signed_attrs": signed_attributes,
                "signature_algorithm": _SIGNATURE_ALGORITHM,
                "signature": core.OctetString(signature_value),
                })]),
            }),
        })

    return result.dump()


def canonical_algorithm_name(name):
    """
    Return the internal standardized name for algorithms.
    """
    return name.replace('-', '').lower()
