# Copyright (c) 2020 Adi Roiban.
# See LICENSE for details.
"""
Code for sending files over AS2 via a location.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

import base64
import os
import random
import time
import six
from collections import OrderedDict
from email.header import make_header
from email.message import Message
from uuid import uuid4

from asn1crypto import cms, core, algos
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms
from cryptography.hazmat.primitives.ciphers.modes import CBC
from cryptography.hazmat.primitives.padding import PKCS7
from email import message_from_string
from OpenSSL import crypto
from twisted.internet import defer
from twisted.web import http
import pem

from chevah.server import emit_error_details, force_unicode
from chevah.server.commons import asn1, DEFAULT_BACKEND
from chevah.server.commons.constant import (
    CONFIGURATION_DISABLED_VALUES,
    TYPE_NAME,
    )
from chevah.server.commons.exception import (
    HTTPClientException,
    ServerException,
    )
from chevah.server.commons.ssl import _load_PEM
from chevah.server.commons.x509 import name_to_OID, PreOrderedName
from chevah.server.http.as2 import (
    _SUPPORTED_DIGEST,
    get_AS2_key_and_certificate,
    canonical_algorithm_name,
    create_signature,
    )
from chevah.server.http.client import HTTPUploadedFile
from chevah.server.http.client_shell import _HTTPClientShell
from chevah.server.http.protocol import parse_options_header
from six.moves import map
from six.moves import range

# As found in location configuration.
# All encryptions are CBC.
_SUPPORTED_ENCRYPTIONS = {
    # DES3 has a key size of 192.
    TYPE_NAME.DES3: (algorithms.TripleDES, 24, '1.2.840.113549.3.7'),
    TYPE_NAME.AES128: (algorithms.AES, 16, '2.16.840.1.101.3.4.1.2'),
    TYPE_NAME.AES192: (algorithms.AES, 24, '2.16.840.1.101.3.4.1.22'),
    TYPE_NAME.AES256: (algorithms.AES, 32, '2.16.840.1.101.3.4.1.42'),
    }


class AS2Client(_HTTPClientShell):
    """
    Sending files over AS2 as a client-side partner.
    """

    def __init__(self, location, context_factory, scheduler=None):
        super(AS2Client, self).__init__(
            location, context_factory, scheduler)
        # Use the email resource as the from email to reduce the
        # configuration complexity for the location.
        default_email = location.root.resources.getEmailClient()
        self._from_email = default_email.configuration.email_from_address
        if not self._from_email:
            self._from_email = 'mdn.receipt@over.http.com'

        self._signing_certificate = get_AS2_key_and_certificate(
            component_configuration=location.configuration,
            root_configuration=location.root.configuration.server,
            certificate_option='as2_own_certificate',
            key_option='as2_own_private_key',
            )

        self._partner_certificates = _get_AS2_certificates(
            location.configuration.as2_partner_certificates,
            'as2_partner_certificates')

    def _isAuthenticationSuccessful(self, code):
        """
        Check the auth response from AS2 server.
        """
        if code in [http.UNAUTHORIZED, http.FORBIDDEN]:
            # We got an explicit reject code.
            return False

        # Accept anything as AS2 server might not support HEAD and is hard
        # to check without pushing a file.
        return True

    @staticmethod
    def _getMessageID():
        """
        Return a new unique message id.

        `self` is ignored as it is designed to be used to patch an instance
        function.

        Should be formatted as RFC 2822
        Defined in RFC 4130 as maximum 998, recommended max 255.
        """
        return '%s@%s' % (uuid4().hex, int(time.time() * 100))

    def getStreamForWriting(self, path, length=None):
        """
        Return a deferred with an `ITransferStream` opened for writing,
        """

        if not length:
            return defer.fail(ServerException(
                'AS2 uploads are only supported for fixed length files.'))

        @defer.inlineCallbacks
        def cb_got_response(response, producer):
            """
            Called when we have a response with all headers.

            By the time this is called all the data was sent.
            """
            # FIXME:2014:
            # Make sure we disconnect if the body is too big.

            # The response body might be a MDN.
            body = yield self._agent.readBody(response)
            # Nothing is returned, since the main deferred is inside the
            # producer.
            # The producer will return a success or a failure.
            producer.responseReceived(body, response)

        def eb_request(failure):
            """
            Called when the request fails.
            """
            # Any error is forwarded to the main deferred, depending on
            # the stage we are at.
            if not producer.connect_deferred.called:
                producer.connect_deferred.errback(failure)

            if not producer.close_deferred.called:
                producer.close_deferred.errback(failure)

        producer = _AS2UploadedFile(
            path,
            self._location,
            length,
            self._url,
            self._from_email,
            self._signing_certificate,
            self._partner_certificates,
            message_id_generator=self._getMessageID,
            )
        deferred = self._agent.post(
            url=self._url,
            headers=producer.getHeaders(),
            body=producer,
            )

        deferred.addCallback(cb_got_response, producer)
        deferred.addErrback(eb_request)

        # The request deferred will only be called when all the data was
        # uploaded.
        # We don't wait for that but return a deferred which signal that the
        # connection is ready to receive data.
        return producer.connect_deferred

    def exists(self, path):
        """
        See: ILocation.

        Remote AS2 files never exist so that we can send it without
        worrying that it will fail.
        """
        return defer.succeed(False)


class _AS2UploadedFile(HTTPUploadedFile):
    """
    Stream used to upload a file to the AS2 server.

    This is both an IBodyProducer and ITransferStream.

    It should not be used outside of getStreamForWriting.
    """

    def __init__(
        self, path, location, length, url, from_email,
        signing_certificate, partner_certificates, message_id_generator,
            ):
        # Consumer is initiated a bit later.
        super(_AS2UploadedFile, self).__init__(
            path=path, location=location, length=length)

        self.emitEvent = location.emitEvent
        self._from_email = from_email
        self._url = url

        self._partner_certificates = partner_certificates

        # Cache as2 configuration to make sure it will not change
        # during the transfer
        self._as2_data = self._initAS2Data(
            configuration=location.configuration,
            real_path=path,
            message_id_generator=message_id_generator,
            )

        # This is how SFTPPlus sees the remote path.
        # We don't have a real path in AS2.
        self.path = self._as2_data['remote_path']

        self.connect_deferred.addErrback(self._ebConnect)

        # The client code is waiting on the `close_deferred`
        # Here we setup final validation around this deferred.
        # We set up this here, to have this code called first.
        self.close_deferred.addCallback(self._cbClose)
        self.close_deferred.addErrback(self._ebClose)

        self._as2_writter = _AS2Writter(
            raw_writter=self.write,
            as2_data=self._as2_data,
            payload_length=length,
            signing_certificate=signing_certificate,
            partner_certificates=partner_certificates,
            )
        self.length = self._as2_writter.total_length
        self.write = self._as2_writter.writeContent

    def _initAS2Data(self, configuration, real_path, message_id_generator):
        """
        We cache all the configuration data, to make sure we don't have it
        change while transfer is running.
        """
        _, filename = self.location.splitPath(real_path)

        try:
            # See if we have only ASCII names, so no special encoding
            # is needed.
            filename.encode('ascii')
            smime_filename = filename
        except UnicodeEncodeError:
            # Looks like not ASCII, so we go with UTF-8
            smime_filename = make_header([(filename, 'utf-8')]).encode()
            if six.PY2:
                smime_filename = smime_filename.decode('ascii')

        result = {
            'real_path': real_path,
            'remote_path': '%s %s' % (self._url, filename),
            'content_disposition': 'attachment; filename="%s"' % (
                smime_filename,),

            'as2_from': configuration.as2_own_identifier,
            'as2_to': configuration.as2_partner_identifier,
            'message_id': '<' + message_id_generator() + '>',
            # The payload content type.
            'content_type': configuration.as2_content_type,
            # The HTTP headers content type.
            'http_content_type': configuration.as2_content_type,

            # The algorithm name,
            'digest': configuration.as2_signature_algorithm,
            'digest_value': '',
            'digest_name': '',
            # Started as algorithm name. Updated with digest value.
            'signature': configuration.as2_signature_algorithm,
            'is_encrypted': False,
            'encryption_algorithm': '',
            # Will be set with a value when sent file is signed.
            'signature_boundary': '',
            'signature_content_type': '',
            'mdn_signed': False,
            'mdn_info': 'not requested',
            'mdn_request': '',  # Empty when MDN is not requested.
            'mdn_header': '',  # Empty when MDN should be unsigned.
            }

        self._updateMDNRequest(configuration, result)
        self._updateHTTPContentType(configuration, result)
        return result

    def _updateMDNRequest(self, configuration, result):
        """
        Set the MDN headers for the request `headers`.
        """
        as2_mdn_receipt = configuration.as2_mdn_receipt

        if as2_mdn_receipt in CONFIGURATION_DISABLED_VALUES:
            return

        result['mdn_request'] = self._from_email

        if as2_mdn_receipt == TYPE_NAME.SYNC_UNSIGNED:
            result['mdn_info'] = 'unsigned'
            return

        result['mdn_header'] = (
            'signed-receipt-protocol=optional, pkcs7-signature; '
            'signed-receipt-micalg=optional, %s') % (result['digest'],)
        result['mdn_info'] = 'signed %s' % (result['digest'],)
        result['mdn_signed'] = True

    def _updateHTTPContentType(self, configuration, result):
        """
        Return the content type for the main HTTP request.
        """
        as2_send_security = configuration.as2_send_security

        if as2_send_security in [
                TYPE_NAME.SIGNED, TYPE_NAME.SIGNED_AND_ENCRYPTED]:
            result['signature_boundary'] = _generate_boundary()
            result['signature_content_type'] = (
                'multipart/signed; protocol="application/pkcs7-signature"; '
                'micalg="%s"; boundary="%s"'
                ) % (result['digest'], result['signature_boundary'])

        if as2_send_security in [
                TYPE_NAME.ENCRYPTED, TYPE_NAME.SIGNED_AND_ENCRYPTED]:
            result['is_encrypted'] = True
            result['http_content_type'] = (
                'application/pkcs7-mime; name="smime.p7m"; '
                'smime-type="enveloped-data"'
                )

            encryption_name = configuration.as2_encryption_algorithm
            if encryption_name not in _SUPPORTED_ENCRYPTIONS:
                raise ServerException(
                    'Encryption algorithm is not supported: %s' % (
                        encryption_name,))
            result['encryption_algorithm'] = encryption_name

            return

        if as2_send_security == TYPE_NAME.SIGNED:
            result['http_content_type'] = result['signature_content_type']
            return

        # Fallback to unsigned and unencrypted.
        result['http_content_type'] = configuration.as2_content_type

    def startProducing(self, consumer):
        """
        See: IBodyProducer

        Called when connection is made and we can start writing.
        """
        self.emitEvent('40050', data=self._as2_data)
        self._consumer = consumer
        self._as2_writter.writeStart()
        return super(_AS2UploadedFile, self).startProducing(consumer)

    def getHeaders(self):
        """
        Return the headers sent with each request for AS2 file upload.

        This this always returns the same value since this might be called
        from multiple places.
        """
        result = {
            'Content-Disposition': self._as2_data['content_disposition'],
            'AS2-From': self._as2_data['as2_from'],
            'AS2-To': self._as2_data['as2_to'],
            'AS2-Version': b'1.1',
            'Message-Id': self._as2_data['message_id'],
            'Date': 'Mon, 24 Aug 2020 13:34:50 GMT',
            'Content-Type': self._as2_data['http_content_type']
            }

        if self._as2_data['mdn_request']:
            result['Disposition-Notification-To'] = (
                self._as2_data['mdn_request'])

        if self._as2_data['mdn_header']:
            result['Disposition-Notification-Options'] = (
                self._as2_data['mdn_header'])

        return result

    def _ebConnect(self, failure):
        """
        Called when failed to start the POST request to send data.
        """
        emit_error_details(
            event_id='40048',
            error=failure.value,
            data=self._as2_data,
            emitter=self.emitEvent,
            )

        # Let the transfer know that AS2 transfer failed.
        return failure

    def responseReceived(self, body, response):
        """
        Called when we got the HTTP response after all the data was uploaded
        and all the response body was received.
        """
        # Trigger digest creation.
        self._as2_writter.updatePayloadDigest()

        self._as2_data['signature'] = '%s %s' % (
            self._as2_data['digest'], self._as2_data['digest_value'])

        # The MDN is validate when the file is closed
        self.close_deferred.callback((response, body))

    def _cbClose(self, result):
        """
        Called when the upload was successful.

        This is called via responseReceived so that we can catch any error
        inside the close_deferred.
        """
        response, body = result

        if response.code not in [http.OK, http.CREATED]:
            details = 'Failed to upload. %s' % (body[:200])
            raise HTTPClientException(
                message=details,
                path=self.path,
                code=response.code,
                phrase=response.phrase,
                )

        self._as2_writter.updatePayloadDigest()
        _validate_mdn(
            self._as2_data,
            response.headers.getAllRawHeaders(),
            body,
            self._partner_certificates,
            )

        # If validate has no raised an issue, all is ok.
        self.emitEvent('40051', data=self._as2_data)

    def _ebClose(self, failure):
        """
        Called when failed to end the POST request to send data.

        This is called from multiple places.
        """
        emit_error_details(
            event_id='40049',
            error=failure.value,
            data=self._as2_data,
            emitter=self.emitEvent,
            )

        # Let the transfer know that AS2 transfer failed.
        return failure


class _AS2Writter(object):
    """
    Handled writing the AS2 streamed data.
    """

    def __init__(
        self, raw_writter, as2_data, payload_length,
        signing_certificate, partner_certificates,
            ):
        self._as2_data = as2_data
        self._raw_writter = raw_writter
        self._signing_certificate = signing_certificate
        self._partner_certificates = partner_certificates
        self._total_length = payload_length

        # The size the the actual file data.
        self._payload_length = payload_length
        # Size of the file data received so far.
        self._payload_received = 0
        # The raw digest value for the received data.
        self._payload_digest_value = None

        # Digest tracking the actual signed content.
        # This can be just file content or multi-part header + content.
        self._digest = None

        # Cipher used to encrypt the payload.
        self._cipher = None
        self._padder = None

        # Request content send before the file content for signed requests.
        self._first_boundary = b''
        # Request content send after the file content for signed requests.
        # it will contain the signature.
        self._final_part = b''

        # The extra headers added for the file payload for signed or
        # encrypted requests.
        self._payload_headers = b''

        # ASN.1 data sent before the encrypted data.
        self._encryption_header = b''
        # ASN.1 sent after the encrypted data.
        # Usually just the double zero bites end of content marker for
        # indefinite-lenght.
        self._encryption_footer = b''

        self._digest_class = _SUPPORTED_DIGEST.get(self._as2_data['digest'])
        self._digest = hashes.Hash(
            self._digest_class(), backend=DEFAULT_BACKEND)
        self._initSignatureOverhead()
        self._initEncryptionOverhead()

    @property
    def total_length(self):
        """
        Return the total length of the AS2 stream.
        """
        return self._total_length

    def writeStart(self):
        """
        Called when connection is made and we can start writing.
        """
        if self._encryption_header:
            self._raw_writter(self._encryption_header)

        self._writeMetaData(self._first_boundary)

        # FIXME:6267:
        # If we don't have _payload_headers it means we have a plain
        # request and we need to update based on HTTP headers,
        # in a canonical/normalized way.
        self._digest.update(self._payload_headers)
        self._writeMetaData(self._payload_headers)

    def writeContent(self, data):
        """
        Writer normal content
        """
        # FIXME:2022:
        # See if we get boundary inside data.
        # If this happens, fail and hope the transfer will try again.
        self._digest.update(data)
        self._payload_received += len(data)

        if self._payload_received < self._payload_length:
            # We still need more data to receive.
            data = self._transform(data)
            if not data:
                # We don't yet have a block to encrypt.
                return defer.succeed(None)

            return self._raw_writter(data)

        if self._final_part:
            # Write signature.
            self.updatePayloadDigest()
            self._final_part = self._getSignaturePart()
            data += self._final_part

        # Encrypt the remaining data.
        data = self._transform(data)

        if self._encryption_footer:
            # Write encryption padding and footer.
            data += self._finalize() + self._encryption_footer

        return self._raw_writter(data)

    def _initSignatureOverhead(self):
        """
        Set up the data sent as multipart and the digest used for the payload.
        """
        if not self._as2_data['signature_boundary']:
            return

        if not self._signing_certificate:
            raise ServerException('No certificate')

        self._first_boundary = (
            '--' + self._as2_data['signature_boundary'] + '\r\n'
            ).encode('ascii')

        self._payload_headers = self._getPayloadHeaders()

        # We start with an empty signature, just so that we can know the
        # length of the body.
        self._final_part = self._getSignaturePart()

        self._total_length += len(
            self._first_boundary + self._payload_headers + self._final_part)

    def _initEncryptionOverhead(self):
        """
        Set up the extra data sent as part of encrypted message and prepare
        the encryption cipher and padder.
        """
        if not self._as2_data['is_encrypted']:
            return

        if not self._first_boundary:
            # This is an unsigned upload.
            self._payload_headers = self._getPayloadHeaders()
            self._total_length += len(self._payload_headers)
        else:
            # This is signed and we should add the signature payload header.
            signature_headers = (
                'Content-Type: %(signature_content_type)s\r\n'
                '\r\n' % self._as2_data
                ).encode('utf-8')
            self._first_boundary = signature_headers + self._first_boundary
            self._total_length += len(signature_headers)

        if not self._partner_certificates:
            raise ServerException(
                'No AS2 partner certificate was configured for '
                'sending encrypted files.'
                )
        # We encrypt using the first configured certificate.
        encryption_certificate = self._partner_certificates[0]

        # Encryption algorithm with random symmetric key.
        encrypt_class, key_size, encrypt_id = _SUPPORTED_ENCRYPTIONS[
            self._as2_data['encryption_algorithm']]

        encryption_key = os.urandom(key_size)
        algorithm = encrypt_class(encryption_key)
        iv = os.urandom(algorithm.block_size // 8)

        self._cipher = Cipher(
            algorithm, CBC(iv), backend=DEFAULT_BACKEND).encryptor()

        self._padder = PKCS7(algorithm.block_size).padder()

        encrypted_key = encryption_certificate['key'].encrypt(
            encryption_key, PKCS1v15())

        content_encryption_algorithm = algos.EncryptionAlgorithm({
            'algorithm': encrypt_id,
            'parameters': cms.OctetString(iv),
            })

        recipients_info = cms.RecipientInfos([
            cms.KeyTransRecipientInfo({
                "version": cms.CMSVersion("v0"),
                "rid": encryption_certificate['rid'],
                "key_encryption_algorithm": cms.KeyEncryptionAlgorithm({
                    "algorithm": cms.KeyEncryptionAlgorithmId(
                        'rsaes_pkcs1v15')}),
                "encrypted_key": cms.OctetString(encrypted_key),
                })
            ])

        block_bytes = self._padder.block_size // 8
        padding_size = block_bytes - self._total_length % block_bytes
        self._total_length += padding_size

        self._encryption_header = (
            b'0\x80' +  # ContentInfo header
            cms.ContentType('enveloped_data').dump() +
            b'\xa0\x800\x80' +  # Context EnvelopedData header
            cms.CMSVersion('v0').dump() +
            recipients_info.dump() +
            b'0\x80' +  # EncryptedContentInfo header
            cms.ContentType('data').dump() +
            content_encryption_algorithm.dump() +
            b'\xa0\x80' +  # Payload context header
            asn1.Tag(  # encrypted_content start
                number=asn1.Numbers.OctetString,
                type=asn1.Types.Primitive,
                cls=asn1.Classes.Universal,
                length=self._total_length,
                ).raw
            )

        # We need to have as many end markers as open headers.
        self._encryption_footer = (
            b'\x00\x00'  # Payload context
            b'\x00\x00'  # EncryptedContentInfo
            b'\x00\x00\x00\x00'  # Context EnvelopedData
            b'\x00\x00'  # ContentInfo
            )

        self._total_length += len(
            self._encryption_header + self._encryption_footer)

    def _getPayloadHeaders(self):
        """
        Return the file payload headers sent for signed requests or
        encrypted+unsigned requests.
        """
        return (
            'Content-Type: %(content_type)s\r\n'
            'Content-Transfer-Encoding: binary\r\n'
            'Content-Disposition: %(content_disposition)s\r\n'
            '\r\n' % self._as2_data).encode('utf-8')

    def _transform(self, data):
        """
        Return the encrypted and padded for data.
        """
        if not self._cipher:
            return data

        data = self._padder.update(data)
        return self._cipher.update(data)

    def _finalize(self):
        """
        Do the final encryption transformation.
        """
        result = self._padder.finalize()
        result = self._cipher.update(result)
        return result + self._cipher.finalize()

    def _writeMetaData(self, data):
        """
        Write data to the stream which is not the actual file content
        and which should not be part of the digest, but which should be
        encrypted.
        """
        if not data:
            return

        data = self._transform(data)
        self._raw_writter(data)

    def _getSignaturePart(self):
        """
        Return the multipart with the signature for digests_value.

        It is called twice.
        Once without any data just to compute the size of the signature
        payload.
        The second time to compute the actual transferred payload data.
        """
        digest_value = self._payload_digest_value
        digest = self._digest
        if not digest_value:
            digest = hashes.Hash(
                self._digest_class(), backend=DEFAULT_BACKEND)
            digest_value = digest.finalize()

        signature = create_signature(
            digest_value,
            digest=digest,
            private_key=self._signing_certificate['key'],
            certificate_asn1=self._signing_certificate['asn1'],
            certificate_id=self._signing_certificate['id'],
            )

        part = (
            '\r\n--%(boundary)s\r\n'
            'Content-Type: application/pkcs7-signature; name="smime.p7s"; '
            'smime-type="signed-data"\r\n'
            'Content-Disposition: attachment; filename="smime.p7s"\r\n'
            'Content-Transfer-Encoding: base64\r\n'
            '\r\n'
            '%(signature)s\r\n'
            '--%(boundary)s--\r\n'
            ) % {
                'boundary': self._as2_data['signature_boundary'],
                'signature': base64.b64encode(signature).decode('ascii'),
                }
        return part.encode('utf-8')

    def updatePayloadDigest(self):
        """
        Return a tuple with the digest value and algorithm name for the
        AS2 payload.
        """
        if self._payload_digest_value:
            return

        self._payload_digest_value = self._digest.finalize()

        self._as2_data['digest_value'] = base64.b64encode(
            self._payload_digest_value).decode('ascii')

        self._as2_data['digest_name'] = self._digest.algorithm.name


def _validate_mdn(as2_data, headers, body, partner_certificates):
    """
    Check the response after the upload.

    Raise an error if MDN is not ok.
    """
    if not as2_data['mdn_request']:
        # No MDN requested
        return

    raw_headers = []
    for name, values in headers:
        for value in values:
            header = name.lower() + b": " + value + b"\r\n"
            raw_headers.append(header.decode('utf-8'))
    raw_headers.append('\r\n')

    raw_message = ''.join(raw_headers) + body.decode('utf-8')
    root = message_from_string(raw_message)

    if not root.is_multipart():
        raise ServerException(
            'No MDN found. Response was: %s' % (body[:100]))

    content_type, options = parse_options_header(
        root.get('content-type', ''))
    content_type = content_type.lower()

    boundary = options.get('boundary', '')
    if not boundary:
        raise ServerException('MDN response without boundary.')

    # Start by assuming that we have unsigned MDN.
    report = root
    signature = None

    if content_type.startswith('multipart/signed'):
        # We have signed.
        report = None
        for part in root.get_payload():
            part_type, options = parse_options_header(
                part.get('content-type', ''))
            part_type = part_type.lower()

            if part_type.startswith('multipart/report'):
                if report:
                    raise ServerException(
                        'Response contains more than one MDN.')
                report = part
                continue

            if part_type.startswith('application/pkcs7-signature'):
                if signature:
                    raise ServerException(
                        'Response contains more than one signature.')
                signature = part
                continue

        if not report:
            raise ServerException('No MDN found in the response.')
        if not signature:
            raise ServerException('No signature found in the response.')

    _validateReport(as2_data, report)
    _validateSignature(
        as2_data, signature, boundary, body, partner_certificates)


def _validateReport(as2_data, report):
    """
    Validate the content of the MDN report.

    This does not validate the signature.

    Will return the validated received MIC value (without algorithm).
    Will return None when MDN has not MIC.
    """
    mdn = None
    payload = report.get_payload()
    for part in payload:
        if not isinstance(part, Message):
            raise ServerException(
                'MDN not found. "{}"'.format(force_unicode(payload, trim=100)))

        content_type, options = parse_options_header(
            part.get('content-type', ''))
        content_type = content_type.lower()

        if 'message/disposition-notification' in content_type:
            if mdn:
                raise ServerException(
                    'Response contains more than one '
                    'disposition-notification.'
                    )
            mdn = part
            continue

    if not mdn:
        raise ServerException(
            'Response does not contains the disposition-notification.')

    if not mdn.is_multipart():
        raise ServerException('Received MDN with invalid format.')

    payload = mdn.get_payload()

    if len(payload) != 1:
        raise ServerException('Received MDN with invalid part.')

    payload = payload[0]

    original_recipient = payload['original-recipient']
    final_recipient = payload['final-recipient']
    original_id = payload['original-message-id']
    received_mic = payload['received-content-mic']

    if not original_id:
        raise ServerException('MDN without Original-Message-ID.')
    if as2_data['message_id'] not in original_id:
        raise ServerException(
            'MDN with invalid message ID. Expecting "%s". Got "%s"' % (
                as2_data['message_id'], original_id,))

    if not original_recipient:
        raise ServerException('MDN without Original-Recipient.')
    parts = original_recipient.split(';', 1)
    if len(parts) != 2:
        raise ServerException(
            'MDN with invalid Original-Recipient format: %s.' % (
                original_recipient,))
    original_recipient = parts[1].strip(' "').strip()
    if original_recipient != as2_data['as2_to']:
        raise ServerException(
            'MDN with invalid Original-Recipient. '
            'Expecting "%s". Got "%s"' % (
                as2_data['as2_to'], original_recipient,))

    if not final_recipient:
        raise ServerException('MDN without Final-Recipient.')
    parts = final_recipient.split(';', 1)
    if len(parts) != 2:
        raise ServerException(
            'MDN with invalid Final-Recipient format: %s.' % (
                final_recipient,))
    final_recipient = parts[1].strip(' "').strip()
    if final_recipient != as2_data['as2_to']:
        raise ServerException(
            'MDN with invalid Final-Recipient. '
            'Expecting "%s". Got "%s".' % (
                as2_data['as2_to'], final_recipient,))

    _validateDisposition(payload['disposition'])
    _validateMDNDigest(as2_data, received_mic)


def _validateDisposition(disposition):
    """
    Check that we don't have an error advertised in the MDN.

    RFC4130 Section 7.4.3.  AS2-MDN Fields
    """
    try:
        # we ignore the dispositiont mode.
        # We only look for the errors.
        _, disposition_type = disposition.split(';', 1)
    except ValueError:
        raise ServerException(
            'MDN disposition with invalid format: %s' % (disposition,))

    disposition_type = disposition_type.strip().lower()

    if disposition_type != 'processed':
        raise ServerException('MDN disposition error: %s' % (disposition,))


def _validateMDNDigest(as2_data, received_mic):
    """
    Check that the received MIC matches the file that we sent..
    """
    if not received_mic:
        if not as2_data['mdn_signed']:
            # All good. MDN not received but also not required.
            return

        raise ServerException('MDN without Received-Content-MIC.')

    digest_value = as2_data['digest_value']
    # Normalize the expected algorithm name.
    expected_digest_name = canonical_algorithm_name(as2_data['digest_name'])

    parts = received_mic.split(',')
    if len(parts) != 2:
        raise ServerException(
            'MDN with invalid Received-Content-MIC format: %s' % (
                received_mic,))

    received_mic_value = parts[0].strip()
    # Normalized received algorith name.
    received_mic_algorithm = parts[1].strip().replace('-', '').lower()

    if expected_digest_name != received_mic_algorithm:
        raise ServerException(
            'MDN with invalid Received-Content-MIC algorithm. '
            'Expecting: %s, %s. Got: %s.' % (
                digest_value, expected_digest_name, received_mic))

    if digest_value != received_mic_value:
        raise ServerException(
            'MDN with invalid Received-Content-MIC value. '
            'Expecting: %s, %s. Got: %s.' % (
                digest_value, expected_digest_name, received_mic))


def _validateSignature(
        as2_data, signature, boundary, body, partner_certificates):
    """
    Validate that the raw `signature` is for the MDN from `body`.

    Use any of the configured partner certificates.
    """
    if not as2_data['mdn_signed']:
        # No further validations for unsigned MDNs.
        return

    if not signature:
        raise ServerException('Received unsigned MDN.')

    if not partner_certificates:
        raise ServerException(
            'No partner certificate configured to validate the signed MDN.')

    signature = cms.ContentInfo().load(
        base64.b64decode(signature.get_payload()))

    signed_message = None

    for part in _extract_parts(boundary, body):
        # It would be nice to only check the content-type header
        # and not the whole part body.
        if b'report-type=disposition-notification' in part.lower():
            signed_message = part
            break

        if b"report-type='disposition-notification'" in part.lower():
            signed_message = part
            break

        if b'report-type="disposition-notification"' in part.lower():
            signed_message = part
            break

    if not signed_message:
        # We should not end up where, as previously we are checking
        # for MDN content.
        raise RuntimeError('Response without a MDN.')

    content = signature['content']
    signer_info = content['signer_infos'][0]
    mdn_digest_algorithm = signer_info['digest_algorithm']['algorithm'].native
    signer_id = dict(signer_info['sid'].native)
    signature_algorithm = (
        signer_info['signature_algorithm']['algorithm'].native).lower()
    signature_value = signer_info['signature'].native

    # Mendelson AS2 advertised the rsassa_pkcs1v15 algorithm as sha256_rsa.
    if signature_algorithm not in ['rsassa_pkcs1v15', 'sha256_rsa']:
        # FIXME:5421:
        # Add support for rsassa_pss.
        raise ServerException(
            'Only PKCS#1 v1.5 signature algorithm is supported.')

    digest_class = _SUPPORTED_DIGEST.get(mdn_digest_algorithm, None)
    if not digest_class:
        raise ServerException(
            'MDN signature digest algorithm not supported: %s' % (
                mdn_digest_algorithm,))

    signed_attributes = signer_info['signed_attrs']
    if signed_attributes:
        # The signature is done against the embedded attributes and not
        # against the MDN body.
        data_to_verify = signed_attributes.untag().dump()

        signed_digest = _get_message_digest(signed_attributes)

        received_digest = hashes.Hash(digest_class(), backend=DEFAULT_BACKEND)
        received_digest.update(signed_message)
        received_digest = base64.b64encode(received_digest.finalize())

        # Validate that the signed attribute matches the MDN body digest.
        if signed_digest != received_digest:
            raise ServerException(
                'Signature contains invalid digest value. '
                'Expecting: %s. Got: %s' % (
                    received_digest,
                    signed_digest,
                    ))
    else:
        data_to_verify = signed_message

    for certificate in partner_certificates:
        try:
            certificate['key'].verify(
                signature_value,
                data_to_verify,
                PKCS1v15(),
                digest_class(),
                )
            # All good
            return
        except InvalidSignature:
            continue

    issuer = []
    for key, value in signer_id['issuer'].items():
        issuer.append('%s=%s' % (key, value))
    issuer = ', '.join(issuer)
    raise ServerException(
        'Failed to validate the MDN signed by '
        'serial_number "%s" with issuer "%s".' % (
            signer_id['serial_number'], issuer))


def _get_message_digest(signed_attributes):
    """
    Return the base64 message_digest value from `signed_attrs`.
    """
    for attribute in signed_attributes:
        if attribute['type'].native != 'message_digest':
            continue
        return base64.b64encode(attribute['values'].native[0])

    raise ServerException(
        'Signature signed_attrs does not contain message_digest.')


# Extra printable characters from RFC2046 section 5.1.1 Common Syntax
# We try to avoid them to not break remote servers. Especially ; and :
# b'()+_.,-:;?/'
# We don't use space to make sure we don't randomly put it at the end.
#
# RFC definition:
# boundary := 0*69<bchars> bcharsnospace
# bchars := bcharsnospace / " "
# bcharsnospace := DIGIT / ALPHA / "'" / "(" / ")" /
#                  "+" / "_" / "," / "-" / "." /
#                  "/" / ":" / "=" / "?"
_BOUNDARY_CHARS = (
    'abcdefghijklmnoprstuvwxyz'
    'ABCDEFGHIJKLMNOPRSTUVWXYZ'
    '0123456789'
    '='
    )


def _generate_boundary():
    """
    Return a random 70 character boundary value.
    """
    result = []
    for _ in range(50):
        result.append(random.choice(_BOUNDARY_CHARS))  # noqa:bandit
    return ('--WebKitFormBoundary' + ''.join(result))


def _extract_parts(boundary, body):
    """
    Extract the parts from multipart body.

    It works with both network newlines or Unix newlines.
    """
    result = []
    for part in body.split(b'--' + boundary.encode('ascii')):
        if not part:
            # First part with boundary.
            continue
        if part.startswith(b'--'):
            # Last part with boundary
            continue
        if part.startswith(b'\r\n'):
            part = part[2:-2]
        elif part.startswith(b'\n'):
            part = part[1:-1]
        else:
            # This might be the first part before the boundary.
            continue

        if not part:
            continue
        result.append(part)

    return result


def _get_AS2_certificates(value, certificate_option):
    """
    Get the list of certificates data used for ASN.1 CMS encryption
    structure or MDN signature validation.

    Return an empty list when no certificate is configured.
    """
    result = []

    value = value.strip()

    if not value:
        return result

    pem_objects = _load_PEM(value, certificate_option)

    for member in pem_objects:
        if not isinstance(member, pem.Certificate):
            continue

        try:
            cert = crypto.load_certificate(
                crypto.FILETYPE_PEM, member.as_bytes())
        except crypto.Error as error:
            raise ServerException(
                'Failed to parse load A2 certificate "%s". %s' % (
                    certificate_option, error))

        issuer_dict = OrderedDict()
        for name, value in cert.get_issuer().get_components():
            oid = '.'.join(map(str, name_to_OID[name.decode('ascii')]))
            issuer_dict[oid] = value.decode()

        result.append({
            'rid': cms.RecipientIdentifier({
                "issuer_and_serial_number": cms.IssuerAndSerialNumber({
                    "issuer": PreOrderedName.build(issuer_dict),
                    "serial_number": core.Integer(cert.get_serial_number())
                    }),
                }),
            'key': cert.to_cryptography().public_key()
            })

    return result
