# Copyright (c) 2018 Adi Roiban.
# See LICENSE for details.
"""
Azure File Service  API client-side implementation.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

import stat
import six

from defusedxml import minidom
from twisted.internet import defer
from twisted.internet.task import deferLater
from twisted.web import http
from twisted.web.iweb import IBodyProducer
from zope.interface import implementer

from chevah_compat import FileAttributes

from chevah.server.commons.exception import (
    HTTPClientException,
    ServerException,
    )
from chevah.server.helpers import get_timestamp
from chevah.server.commons.interface import ITransferStream
from chevah.server.commons.stream import UploadProducer
from chevah.server.http.azure_storage import _AzureStorageClient
from chevah.server.http.client import (
    get_url_segments,
    HTTPDownloadedFile,
    )


def _get_mode(attributes, is_file):
    """
    Helper to transform data of the Azure Files Attributes field
    into an equivalent Unix mode.

    Example `attributes` strings: "Archive", "ReadOnly | Archive", "Directory"

    API documentation:
    https://docs.microsoft.com/en-us/rest/api/storageservices/create-file#file-system-attributes
    """
    if attributes is None:
        # Shares do not have Azure Attributes.
        # Return them as read-only, for lack of a better mode.
        return stat.S_IFDIR | 0o555

    mode = stat.S_IFREG
    if not is_file:
        # We assume it is a directory instead.
        mode = stat.S_IFDIR | 0o111

    if 'ReadOnly' in attributes:
        mode |= 0o444
    else:
        mode |= 0o666

    return mode


def _get_modified(date):
    """
    Helper to transform an Azure Files timestamp into a Unix timestamp.
    For files and folders, it is an ISO 8601 string.
    For shares, this is a None object.

    Example Azure timestamp: "2021-07-29T10:00:37.7852368Z"
    """
    if date is None:
        # Shares do not have the LastWriteTime XML tag.
        return 0

    return get_timestamp(date)


def _get_node_id(file_id):
    """
    Convert the Azure `file_id` to an int, if it is not None.

    If `file_id` is None, return None. This is the case for Azure Shares.
    """
    if file_id is None:
        return None

    return int(file_id)


def _first_element(list_or_none):
    """
    If the argument is not None, return the first element inside.
    If the argument is None, return None.

    An Azure file share does not have certain properties,
    and these are returned as None from the XML parser.
    """
    if list_or_none is not None:
        return list_or_none[0]


def _parse_folder_listing(body, root_path):
    """
    Parse an XML Azure folder listing into a dictionary of
    `name` -> `FileAttributes`.

    For examples of the body, see unit tests for `getFolderMemberAttributes`
    in `chevah/server/tests/normal/http/test_azure_file.py`.

    FIXME:5057:
    Address the case when NextMarker indicates there are more entries.

    FIXME:5028:
    Create smaller functions to extract specific tags.
    Print out the XML in case of an error.
    """
    result = minidom.parseString(body)
    enumeration_results = result.getElementsByTagName('EnumerationResults')[0]
    entries = enumeration_results.getElementsByTagName('Entries')[0].childNodes

    result = {}

    for entry in entries:
        is_file = (entry.tagName == 'File')
        name = entry.getElementsByTagName('Name')[0].firstChild.data
        path = root_path.rstrip('/') + '/' + name
        if is_file:
            size_element = entry.getElementsByTagName('Content-Length')[0]
            size = int(size_element.firstChild.data)
        else:
            size = 0

        file_id = int(entry.getElementsByTagName('FileId')[0].firstChild.data)
        last_write_time = (
            entry.getElementsByTagName('LastWriteTime')[0].firstChild.data)
        raw_attributes = (
            entry.getElementsByTagName('Attributes')[0].firstChild.data)

        result[name] = FileAttributes(
            name=name,
            path=path,
            size=size,
            is_file=is_file,
            is_folder=not is_file,
            modified=_get_modified(last_write_time),
            mode=_get_mode(raw_attributes, is_file),
            node_id=file_id,
            )
    return result


def _clean_path(path):
    """
    Strip the ending slashes of a path, but preserve the first character.
    """
    if len(path) <= 1:
        return path
    return path[0] + path[1:].rstrip('/')


class AzureFileClient(_AzureStorageClient):
    """
    Client-side shell for Azure File Service.

    API documentation:
    https://docs.microsoft.com/en-us/rest/api/storageservices/file-service-rest-api
    """

    SERVICE_DNS_BASE = 'file.core.windows.net'
    _CLIENT_ID_BASE = 'sftpplus-azure-file'

    @defer.inlineCallbacks
    def exists(self, path):
        """
        See: ILocation.

        `True` when path exists, `False` otherwise.

        Return a deferred that fires with the result or with a failure if the
        existence can't be checked.

        Implemented via getAttributes.
        """
        try:
            yield self.getAttributes(path)
            defer.returnValue(True)
        except HTTPClientException as error:
            # It's OK if not found.
            if error.code == http.NOT_FOUND:
                defer.returnValue(False)

            # Other cases are not OK.
            error.message = error.message.replace(
                'Failed to get path attributes.',
                'Failed to check if path exists.'
                )
            raise error

    @defer.inlineCallbacks
    def _requestHEAD(self, path, **kwargs):
        """
        Send a HEAD request to the given URL, and return the response.

        Discard the HTTP body.
        """
        response = yield self._azureRequest('head', path=path, **kwargs)
        yield self._agent.readBodyWithIgnore(response)
        defer.returnValue(response)

    @defer.inlineCallbacks
    def getAttributes(self, path):
        """
        Return a deferred that fires with the `FileAttributes` for `path`.

        Here are possible requests:
        HEAD /myshare/mydir/myfile
        HEAD /myshare/mydirectory?restype=directory
        HEAD /myshare?restype=share

        API documentation:
        Files:
        https://docs.microsoft.com/en-us/rest/api/storageservices/get-file-properties
        Directories:
        https://docs.microsoft.com/en-us/rest/api/storageservices/get-directory-properties
        Shares:
        https://docs.microsoft.com/en-us/rest/api/storageservices/get-share-properties
        """
        # We don't initially know whether this is a file.
        is_file = None

        path = _clean_path(path)
        share, share_path = self._splitPath(path)
        if share_path == '/':
            # We have a share.
            try:
                share.encode('ascii')
            except UnicodeEncodeError:
                raise ServerException('Invalid share name.')

            path = '/' + share
            query = [['restype', 'share']]
            is_file = False
        else:
            # We might have a directory or a file.
            # Try the directory.
            query = [['restype', 'directory']]

        response = yield self._requestHEAD(path=path, query=query)

        if response.code == http.NOT_FOUND and is_file is None:
            # Share or directory was excluded. Try checking if it's a file.
            response = yield self._requestHEAD(path=path)
            is_file = True
        else:
            is_file = False

        if response.code != http.OK:
            message = self._getErrorDetails(
                response, 'Failed to get path attributes.')
            raise HTTPClientException(
                message,
                path=path,
                code=response.code,
                phrase=response.phrase,
                )

        modified = response.headers.getRawHeaders('x-ms-file-last-write-time')
        file_id = response.headers.getRawHeaders('x-ms-file-id')
        attributes = response.headers.getRawHeaders('x-ms-file-attributes')

        defer.returnValue(
            FileAttributes(
                name=path.strip('/').split('/')[-1],
                path=path,
                size=int(response.headers.getRawHeaders('content-length')[0]),
                is_file=is_file,
                is_folder=not is_file,
                modified=_get_modified(_first_element(modified)),
                mode=_get_mode(_first_element(attributes), is_file=is_file),
                node_id=_get_node_id(_first_element(file_id)),
                )
            )

    @defer.inlineCallbacks
    def getFolderMemberAttributes(self, path):
        """
        See: `ILocation`.

        API documentation:
        https://docs.microsoft.com/en-us/rest/api/storageservices/list-directories-and-files
        """
        path = _clean_path(path)
        query = [
            ['restype', 'directory'],
            ['comp', 'list'],
            ['include', 'Attributes,Timestamps'],
            ]
        response = yield self._azureRequest('get', path=path, query=query)
        body = yield self._agent.readBody(response)

        if response.code != http.OK:
            message = self._getErrorDetails(
                response, 'Failed to get folder member attributes.')
            raise HTTPClientException(
                message,
                path=path,
                code=response.code,
                phrase=response.phrase,
                )

        defer.returnValue(_parse_folder_listing(body, root_path=path))

    def createFolder(self, path):
        """
        Create a folder at path, without creating parents.

        Return a deferred that fires when operation is done.

        API documentation:
        https://docs.microsoft.com/en-us/rest/api/storageservices/create-directory
        """
        path = _clean_path(path)

        def cb_got_response(response):
            """
            Called when we got a response from the original
            """
            # A successful operation returns status code 201 (Created).
            if response.code == http.CREATED:
                # All good
                return self._agent.readBody(response)

            raise HTTPClientException(
                'Failed to create folder.',
                path=path,
                code=response.code,
                phrase=response.phrase,
                )

        deferred = self._azureRequest(
            'put',
            path=path,
            query=[['restype', 'directory']],
            body=b'',
            headers={
                'x-ms-file-permission': ['inherit'],
                'x-ms-file-attributes': ['Directory'],
                'x-ms-file-creation-time': ['now'],
                'x-ms-file-last-write-time': ['now'],
                },
            )
        deferred.addCallback(cb_got_response)
        return deferred

    def deleteFolder(self, path):
        """
        Delete folder at path.

        Return a deferred that fires when operation is done.

        API documentation:
        https://docs.microsoft.com/en-us/rest/api/storageservices/delete-directory
        """
        path = _clean_path(path)

        def cb_got_response(response):
            if response.code == http.ACCEPTED:
                return self._agent.readBody(response)

            message = self._getErrorDetails(
                response, 'Failed to delete folder.')
            raise HTTPClientException(
                message,
                path=path,
                code=response.code,
                phrase=response.phrase,
                )

        deferred = self._azureRequest(
            'delete', path=path, query=[['restype', 'directory']])
        deferred.addCallback(cb_got_response)
        return deferred

    @defer.inlineCallbacks
    def _raiseIfDirectory(self, path, message):
        """
        Check if the Azure Files URL is a directory. If so, raise an error.
        """
        is_dir_response = yield self._azureRequest(
            'head', path=path, query=[['restype', 'directory']])
        self._agent.readBody(is_dir_response)

        if is_dir_response.code == http.OK:
            raise ServerException(message)

    @defer.inlineCallbacks
    def rename(self, source_path, destination_path):
        """
        Rename file from `source_path` to `destination_path`.
        For now, it does not support folders nor recursion within them.

        Return a deferred that fires when operation is done.

        The Azure Files REST API does not have a rename/move operation,
        therefore we copy to the new location, then delete the original.

        API Docs:
        https://docs.microsoft.com/en-us/rest/api/storageservices/copy-file
        """
        source_path = _clean_path(source_path)
        destination_path = _clean_path(destination_path)

        @defer.inlineCallbacks
        def after_copy_started(response):
            """
            Check the response to the copy request.
            If it has a good HTTP status, forward it down the callback chain.
            """
            # We do not expect a useful body.
            # Read it so the persistent connection is freed.
            yield self._agent.readBody(response)

            if response.code == http.ACCEPTED:
                defer.returnValue(response)

            if response.code == http.NOT_FOUND:
                # The response code 404 is returned when
                # moving a directory.
                # Check if this is the case, and send a more explicit message.
                yield self._raiseIfDirectory(
                    path=source_path,
                    message='Renaming a folder is not supported.',
                    )

                # The response code 404 is also returned when moving,
                # and a directory is a destination.
                # Check if this is the case, and send a more explicit message.
                yield self._raiseIfDirectory(
                    path=destination_path,
                    message='Destination already exists and is a directory.',
                    )

            message = self._getErrorDetails(
                response, 'Rename operation failed during copy.')
            raise HTTPClientException(
                message,
                path=source_path,
                code=response.code,
                phrase=response.phrase,
                )

        @defer.inlineCallbacks
        def wait_for_success(response, wait, timeout=None):
            """
            Check the `x-ms-copy-status` of the response.
            If the response is still pending, wait some more
            to allow the server to finish copying, and then try again.
            When the status becomes successful, delete the source file.

            API Docs:
            https://docs.microsoft.com/en-us/rest/api/storageservices/get-file-properties#see-also
            """
            timeout_duration = 3600  # Keep trying for an hour.
            if timeout is None:
                timeout = self._scheduler.seconds() + timeout_duration

            status = response.headers.getRawHeaders('x-ms-copy-status')
            if status == ['success']:
                # If the delete fails, the whole rename fails,
                # and both the source and the destination files remain.
                # We will not try to rollback the copy operation.
                yield self.deleteFile(source_path)
                return

            if status == ['pending']:
                # Wait a bit to allow for the copy to be finalized
                # and then recheck the status.

                # If we waited too long, give up.
                if self._scheduler.seconds() > timeout:
                    raise ServerException(
                        'Rename operation took longer than %s seconds.' % (
                            timeout_duration,))

                # FIXME:5076:
                # When we are waiting at the 300 second cap,
                # issue an event informing the user that we are still waiting,
                # and we are checking every 5 minutes.
                # If we reached the timeout instead,
                # fail and make sure an event appears.
                yield deferLater(self._scheduler, wait)

                # Get the copy status, and address the response recursively.
                updated_response = yield self._azureRequest(
                    'head', path=destination_path)
                self._agent.readBody(updated_response)

                yield wait_for_success(
                    response=updated_response,
                    wait=min(wait * 2, 300),
                    timeout=timeout)
                return

            message = self._getErrorDetails(
                response, 'Rename operation failed during copy.')

            copy_message = response.headers.getRawHeaders(
                'x-ms-copy-status-description')
            if copy_message:
                message = '%s "%s"' % (message, copy_message[0])

            raise HTTPClientException(
                message,
                path=source_path,
                code=response.code,
                phrase=response.phrase,
                )

        # Request the file copy.
        source_url = self._base_url.replace(
            path=get_url_segments(source_path)).to_uri().to_text()
        response = yield self._azureRequest(
            'put',
            path=destination_path,
            body=b'',
            headers={'x-ms-copy-source': [source_url]},
            )

        response = yield after_copy_started(response)
        yield wait_for_success(response, wait=60)

    def getStreamForWriting(self, path, length=None):
        """
        Return a deferred with an `ITransferStream` opened for writing,

        The HTTP upload process make it complicated to handle errors here.

        The HTTP response is sent only after we have sent all the data
        advertised in the request.
        So even if this is going to be an error, we need to let the write
        operation to perform its job so that we can see the result.

        This is why this will succeeds as long as the HTTP connection is made
        and the error is reported in the close operation.
        """
        stream = _AzureUploadedRange(
            location=self._location,
            path=path,
            source_length=length,
            make_request=self._azureRequest,
            read_body=self._agent.readBody,
            get_error_details=self._getErrorDetails,
            scheduler=self._scheduler,
            )

        deferred = stream.create()
        return deferred

    @defer.inlineCallbacks
    def getStreamForReading(self, path):
        """
        Retrieve a file from Azure, and

        API Docs:
        https://docs.microsoft.com/en-us/rest/api/storageservices/get-file
        """
        path = _clean_path(path)
        response = yield self._azureRequest('get', path=path)

        if response.code != http.OK:
            raise HTTPClientException(
                message='Failed to download.',
                path=path,
                code=response.code,
                phrase=response.phrase,
                )

        defer.returnValue(
            HTTPDownloadedFile(
                path=path,
                location=self._location,
                response=response,
                )
            )


@implementer(ITransferStream)
class _AzureUploadedRange(object):
    """
    Stream used to upload a file to the Azure File Service.

    This is ITransferStream delegating the actual upload to IBodyProducer.

    It produced data to be written to the HTTP server.

    The upload is done across multiple request.
    First request to create the file and then subsequent chunk request of a
    maximum of 4MB.

    The `self.length` is reset for each chunk.

    It should not be used outside of getStreamForWriting.

    """

    def __init__(
            self, path, location, source_length,
            make_request, read_body, get_error_details,
            scheduler,
            ):
        self.location = location
        self.path = _clean_path(path)
        self._scheduler = scheduler
        self._consumer = None

        self._azureRequest = make_request
        self._readBody = read_body
        self._getErrorDetails = get_error_details

        self._source_length = source_length

        # Deferred used to signal that write operation should be paused.
        self._pause_write = None

        # This is used by HTTP client IBodyProducer to know how much to
        # write in the current request.
        self.length = 0

        self._cursor = 0

        # This is initialized for each chunk which is written.
        self._chunk_done = defer.succeed(None)

        # Temporary storage for data which can't be yet written.
        # `None` is used to signal that no write is waiting.
        self._pause_buffer = None

    @property
    def mode(self):
        """
        Always in writing.
        """
        return 'writing'

    def create(self):
        """
        This initiates the stream and return itself.

        It fails if the remote stream can't be initiated.

        API:
        https://docs.microsoft.com/en-us/rest/api/storageservices/create-file
        """
        # First we make the request to create the file.
        # When the file exist, the create will still success.
        # Then will start uploading the content in a separate PUT RANGE.
        deferred = self._azureRequest(
            'put',
            path=self.path,
            body=b'',
            headers={
                'x-ms-content-length': [six.text_type(self._source_length)],
                'x-ms-type': ['file'],
                'x-ms-file-permission': ['inherit'],
                'x-ms-file-attributes': ['Archive'],
                'x-ms-file-creation-time': ['now'],
                'x-ms-file-last-write-time': ['now'],
                },
            )

        def cb_create(response):
            """
            Called when we have a response for the create file request.
            """
            if response.code != http.CREATED:
                message = self._getErrorDetails(
                    response, 'Failed to open for create.')
                raise HTTPClientException(
                    message=message,
                    path=self.path,
                    code=response.code,
                    phrase=response.phrase,
                    )

            # The response might have some body, so we will try to read it.
            return self._readBody(response)

        def cb_clear(ignored):
            """
            Called once a file is created to clear all the existing ranges.
            """
            if self._source_length == 0:
                return defer.succeed(self)

            deferred = self._azureRequest(
                'put',
                path=self.path,
                query=[['comp', 'range']],
                body=b'',
                headers={
                    'range': ['bytes=0-{}'.format(self._source_length - 1)],
                    'x-ms-write': ['Clear'],
                    },
                )
            deferred.addCallback(cb_cleared)
            return deferred

        def cb_cleared(response):
            """
            Called when got a response from the file clearing command.
            """
            if response.code != http.CREATED:
                message = self._getErrorDetails(
                    response, 'Failed to clear the created file.')
                raise HTTPClientException(
                    message=message,
                    path=self.path,
                    code=response.code,
                    phrase=response.phrase,
                    )

            # The response might have some body, so we will try to read it.
            deferred = self._readBody(response)
            # Once the request to create the file is done,
            # we can wait to receive data and forward it to the server.
            deferred.addCallback(lambda _: self)
            return deferred

        deferred.addCallback(cb_create)
        deferred.addCallback(cb_clear)
        return deferred

    def close(self):
        """"
        See: `ITransferStream`.
        """
        # Wait for result from previous write.
        return self._chunk_done

    def read(self, length):
        """
        See: `ITransferStream`.
        """
        return defer.fail(ServerException('Only opened for writing.'))

    def write(self, data):
        """
        See: `ITransferStream`.
        """
        # FIXME:3979:
        # The chunk is sent to the consumer and it will process it.
        # Once we have a new interface this will be removed,
        # as this API is forced.
        # The consume will use only pause/resume production method to signal
        # that no more data should be written.

        chunk_length = len(data)
        if chunk_length > 4194304:
            # Azure File API only accept request of up to 4MB
            return defer.fail(ServerException('Write chunk too big.'))

        end_range = self._cursor + chunk_length - 1
        self.length = chunk_length
        self._chunk = data

        self._chunk_done = defer.Deferred()

        producer = AzureFileProducer(chunk_length)

        # Now we put the actual content.
        deferred = self._azureRequest(
            'put',
            path=self.path,
            query=[['comp', 'range']],
            body=producer,
            headers={
                'range': ['bytes=%s-%s' % (self._cursor, end_range)],
                'x-ms-write': ['Update'],
                },
            )

        def cb_got_range(response, chunk_length):
            """
            Called when we have a response with all headers for the PUT RANGE.

            By the time this is called all the data was sent.
            """
            if response.code != http.CREATED:
                # All data is already written so, will signal the failure
                # at the close operation.
                message = self._getErrorDetails(response, 'Failed to upload.')
                raise HTTPClientException(
                    message=message,
                    path=self.path,
                    code=response.code,
                    phrase=response.phrase,
                    )

            self._cursor += chunk_length

            # The response might have some body, so we will try to read it.
            return self._readBody(response)

        def cb_done(body):
            self._chunk_done.callback(None)

        def eb_range(failure):
            if not self._chunk_done.called:
                self._chunk_done.errback(failure)

        deferred.addCallback(cb_got_range, chunk_length)
        deferred.addCallback(cb_done)
        deferred.addErrback(eb_range)

        def cb_write(ignored, producer):
            producer.write(data)
            # We are not going to write more data.
            producer.done_deferred.callback(None)

        producer.connect_deferred.addCallback(cb_write, producer)

        return self._chunk_done


@implementer(IBodyProducer)
class AzureFileProducer(UploadProducer):
    """
    Stream used to upload a file to the Azure File Service.

    This an IBodyProducer.

    It produced data to be written to the HTTP server.

    The upload is done across multiple request.
    First request to create the file and then subsequent chunk request of a
    maximum of 4MB.

    The `self.length` is reset for each chunk.

    It should not be used outside of getStreamForWriting.

    """

    def __init__(self, length):
        # Consumer is initiated a bit later.
        super(AzureFileProducer, self).__init__(consumer=None)
        self.length = length
        self.connect_deferred = defer.Deferred()

        # This should be called when no more data is to be written.
        # This will put the connection back into the poll.
        self.done_deferred = defer.Deferred()

        self.close_deferred = defer.Deferred()

    def startProducing(self, consumer):
        """
        See: IBodyProducer

        Called when connection is made and we can start writing.
        """
        self._consumer = consumer
        self.connect_deferred.callback(self)
        return self.done_deferred
