# Copyright (c) 2018 Adi Roiban.
# See LICENSE for details.
"""
Shared code for Azure Storage API client-side implementation.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

import hashlib
import hmac
import six.moves.urllib.request
import six.moves.urllib.parse
import six.moves.urllib.error
from base64 import b64encode, b64decode
from hyperlink import URL

from twisted.internet import defer
from twisted.internet.address import IPv4Address
from twisted.internet.error import ConnectionDone
from twisted.web import http
from zope.interface import implementer

from chevah.server import force_unicode
from chevah.server.commons.credential import PasswordCredentials
from chevah.server.commons.interface import ILocationShell
from chevah.server.commons.exception import (
    HTTPClientException,
    ServerException,
    )
from chevah.server.http.client import (
    get_url_path,
    get_url_segments,
    PersistentAgent,
    )
from chevah.server.identity.avatar import ClientAvatar
from chevah.server.location.commons import OnConnectionLostMixin
import six


@implementer(ILocationShell)
class _AzureStorageClient(OnConnectionLostMixin):
    """
    Command code for client-side shell for Azure Storage services over HTTPS.

    The configuration is validated by the location.

    All internal HTTP requests should be done via `_azureRequest` so that the
    authentication header injected with the required Azure signature.
    """

    # These class members need to be defined by each implementation.
    SERVICE_DNS_BASE = None
    _CLIENT_ID_BASE = None

    # No need to be defined by the implementations
    _AZURE_STORAGE_VERSION = '2020-04-08'

    def __init__(self, location, context_factory, scheduler):
        self._agent = PersistentAgent(
            ssl_context=context_factory.getContext(),
            event_emitter=location.emitEvent,
            )

        # The location which is using this shell.
        self._location = location
        self._scheduler = scheduler

        self._hostname = self._location.configuration.address
        self._base_url = URL.from_text('https://{}'.format(self._hostname))
        self._account_name = self._location.configuration.username
        self._base_auth = 'SharedKey ' + self._account_name + ':'
        self._client_request_id = '{}-{}'.format(
            self._CLIENT_ID_BASE, self._location.uuid)
        self._access_key = b64decode(self._location.configuration.password)

        credentials = PasswordCredentials(
            username=self._location.configuration.username,
            password=self._location.configuration.password,
            creator=self._location,
            peer=IPv4Address('TCP', self._hostname, 443)
            )
        self._avatar = ClientAvatar(credentials)

    @property
    def avatar(self):
        """
        See: ILocationShell.
        """
        return self._avatar

    def _azureRequest(
            self, method_name, path, query=(), headers=None, body=None):
        """
        Make an authenticated HTTP request, and retry on authentication errors.

        url is required and is should be just the path.
        When `headers` are provided, the keys should be in lower case.
        """
        method = getattr(self._agent, method_name)

        url = self._base_url.replace(
            path=get_url_segments(path),
            query=query,
            )
        new_headers = self._setSignedHeaders(
            method_name,
            url=url,
            headers=headers,
            body=body,
            )

        kwargs = {}
        if body is not None:
            kwargs['body'] = body

        deferred = method(url=url, headers=new_headers, **kwargs)
        return deferred

    def _setSignedHeaders(self, method_name, url, headers=None, body=None):
        """
        Set the headers in kwargs for a signed Azure shared key request.

        See:
        https://docs.microsoft.com/en-us/rest/api/storageservices/authorize-with-shared-key
        """
        if headers is None:
            headers = {}
        headers = headers.copy()

        # Resolve the length of the request.
        content_length = ''
        headers['content-length'] = ['0']
        if body is not None:
            if isinstance(body, (six.text_type, bytes,)):
                length = str(len(body))
            else:
                length = str(body.length)
            if length != '0':
                content_length = length
                headers['content-length'] = [length]

        range_to_sign = headers.get('range', [''])[0]

        headers['date'] = [
            http.datetimeToString(self._scheduler.seconds()).decode('ascii')]

        # For now, we have a fixed set of x-ms-* headers, so we don't do
        # dynamic generation.
        headers['x-ms-version'] = [self._AZURE_STORAGE_VERSION]
        headers['x-ms-client-request-id'] = [self._client_request_id]

        x_ms_headers = sorted([
            h for h in headers.keys() if h.lower().startswith('x-ms-')
            ])
        canonicalized_headers = []
        for name in x_ms_headers:
            canonicalized_headers.append(
                '%s:%s' % (name.lower(), headers[name][0].strip()))

        canonicalized_headers = '\n'.join(canonicalized_headers)

        # Keys in lower case, sorted in alphabetical order.
        sorted_query = sorted([(k.lower(), v) for k, v in url.query])
        canonicalized_query = ['{}:{}'.format(k, v) for k, v in sorted_query]

        if canonicalized_query:
            canonicalized_query = '\n' + '\n'.join(canonicalized_query)
        else:
            # When we don't have query, don't add anything,
            # not even empty line.
            canonicalized_query = ''

        url_path = get_url_path(url.path)
        canonicalized_resource = URL.from_text(
            '/' + self._account_name + url_path)

        string_to_sign = (
            method_name.upper() + '\n' +  # VERB
            '\n' +  # Content-Encoding
            '\n' +  # Content-Language
            content_length + '\n' +  # Content-Length
            '\n' +  # Content-MD5
            '\n' +  # Content-Type
            headers['date'][0] + '\n' +  # Date
            '\n' +  # If-Modified-Since
            '\n' +  # If-Match
            '\n' +  # If-None-Match
            '\n' +  # If-Unmodified-Since
            range_to_sign + '\n' +  # Range
            canonicalized_headers +
            '\n' +  # Newline not documented.
            canonicalized_resource.to_uri().to_text() +
            canonicalized_query
            ).encode('ascii')

        signature = hmac.new(
            self._access_key,
            string_to_sign,
            digestmod=hashlib.sha256,
            ).digest()

        headers['authorization'] = [
            self._base_auth + b64encode(signature).decode('ascii')]
        return headers

    def _getErrorDetails(self, response, base):
        """
        Return the error message from `response` starting with the
        `base` message.
        """
        details = ''
        error = response.headers.getRawHeaders('x-ms-error-code ')
        if error:
            details = error[0]

        if details:
            return '{} {}'.format(base, details)
        else:
            return base

    def _splitPath(self, path):
        """
        Return a tuple of Unicode container/share-name and name to the path.
        """
        parts = [p.strip() for p in path.split('/') if p.strip()]

        if len(parts) < 1:
            raise ServerException(
                'Path without a base segment (container or share) '
                'is not supported.')

        return (parts[0], '/' + '/'.join(parts[1:]))

    @defer.inlineCallbacks
    def authenticate(self):
        """
        See: `ILocationShell`.
        """
        # As for now we only support shared key, there is no
        # authentication, just a signature.
        # So we just check that credentials are valid by making a
        # generic request to get one container.
        url = self._base_url.replace(
            query=(('restype', 'service'), ('comp', 'properties')))
        headers = self._setSignedHeaders('get', url=url)

        response = yield self._agent.get(url, headers)
        # Always consume the existing full response.
        yield self._agent.readBodyWithIgnore(response)
        if response.code != http.OK:
            raise ServerException(
                'Failed to validate the credentials. %s: %s' % (
                    response.code, force_unicode(response.phrase)))

    def close(self):
        """
        Called when the agent should no longer be used.
        """
        # WebDav is based on HTTP, which is stateless, so there is not
        # much to do here.
        self._onConnectionLost(ConnectionDone('Bye.'))
        return self._agent.closePersistentConnections()

    def _copy(self, source_path, destination_path):
        """
        Duplicate path inside the location.

        For now, this is only here to support the testing

        Return a deferred that fires when operation is done.
        """
        def cb_got_response(response):
            """
            Called when we have a response.
            """
            # In version 2012-02-12 and newer,
            # a successful operation returns status code 202 (Accepted).
            if response.code != http.ACCEPTED:
                raise ServerException(
                    'Failed to copy %s to %s. %s: %s' % (
                        source_path, destination_path,
                        response.code, response.phrase,
                        ))

        source_url = self._base_url.replace(
            path=get_url_segments(source_path)).to_uri().to_text()
        deferred = self._azureRequest(
            'put',
            path=destination_path,
            headers={'x-ms-copy-source': [source_url]},
            body=b'',
            )
        deferred.addCallback(cb_got_response)
        return deferred

    def deleteFile(self, path):
        """
        Delete file at path.

        This is the same URL for Blob and File services.
        """
        def cb_got_response(response):
            """
            Called when we got a response from the original
            """
            # A successful operation returns status code 202 (Accepted).
            if response.code == http.ACCEPTED:
                # All good
                return self._agent.readBody(response)

            message = self._getErrorDetails(
                response, 'Failed to delete file.')
            raise HTTPClientException(
                message,
                path=path,
                code=response.code,
                phrase=response.phrase,
                )

        deferred = self._azureRequest('delete', path=path)
        deferred.addCallback(cb_got_response)
        return deferred
