# Copyright (c) 2014 Adi Roiban.
# See LICENSE for details.
"""
HTTP client part.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

from base64 import b64encode
from io import BytesIO

from lazr.delegates import delegate_to
from hyperlink import URL
from OpenSSL import SSL
from twisted.internet import defer, reactor
from twisted.internet.address import IPv4Address
from twisted.internet.error import ConnectError, ConnectionClosed
from twisted.internet.interfaces import (
    IConsumer,
    IOpenSSLClientConnectionCreator,
    )
from twisted.internet.protocol import Factory, Protocol
from twisted.python.failure import Failure
from twisted.internet.task import deferLater
from twisted.web.client import (
    Agent,
    BrowserLikePolicyForHTTPS,
    FileBodyProducer,
    HTTP11ClientProtocol,
    HTTPConnectionPool,
    PartialDownloadError,
    PotentialDataLoss,
    ResponseDone,
    _WrapperException,
    )
from twisted.web.http import Headers
from twisted.web.iweb import IBodyProducer, UNKNOWN_LENGTH
from zope.interface import implementer

from chevah.server import add_timeout, force_unicode, json
from chevah.server.commons.constant import CONNECT_TIMEOUT
from chevah.server.commons.exception import ServerException
from chevah.server.commons.interface import ITransferStream
from chevah.server.commons.stream import UploadProducer
from chevah.server.commons.ssl import SSLConnectionMixin
from chevah.server.identity.avatar import PeerAvatar
import six
from six.moves import range


def _eb_request(failure):
    """
    Called when request fails and wraps causing error into a
    `ServerException`.
    """
    if failure.check(_WrapperException):
        failure = failure.value.args[0][0]

    error = failure.value

    if isinstance(error, (ConnectionClosed, ConnectError, ServerException)):
        # These are special errors related to connection status.
        error.message = force_unicode(error)
        return failure

    details = force_unicode(failure.value)
    # This is an unknown error, so we convert it into our generic error,
    # while keeping the original traceback.
    message = 'Unknown error during client request: %s' % (details,)
    failure_traceback = failure.getTracebackObject()

    return Failure(
        exc_value=ServerException(message),
        exc_tb=failure_traceback,
        )


def extract_configured_header(line):
    """
    Given a complete HTTP header line, extract the field name and value and
    process the header.
    """
    try:
        key, val = line.split(':', 1)
    except ValueError:
        raise ServerException(
            'The configured "headers" has invalid value: %s.' % (line,))
    return (key.strip().encode('ascii'), val.lstrip().encode('ascii'))


def get_basic_http_header(username, password):
    """
    Create the Basic HTTP "Authorization" header value from given credentials.
    """
    auth_string = b'%s:%s' % (
        username.encode('utf-8'), password.encode('utf-8'))
    auth_value = b'Basic ' + b64encode(auth_string)
    return auth_value


class _ReadBodyProtocolWithTimeout(Protocol):
    """
    Protocol that collects data sent to it and which will time out if
    data is not received in a time interval.
    """
    def __init__(self, reactor, timeout):
        def cancel(deferred):
            """
            Called if the timeout expires.

            This eventually calls `loseConnection`
            on `twisted.internet.tcp.Client`.
            """
            self.transport.stopProducing()

        self._deferred = defer.Deferred(cancel)
        self._received = []
        self._timeout_call = add_timeout(
            reactor, self._deferred, timeout,
            message='Data receiving stalled.')
        self._timeout = timeout

    @property
    def deferred(self):
        """
        Return a deferred that fires when the response was handled.
        """
        return self._deferred

    def dataReceived(self, data):
        """
        Accumulate some more bytes from the response.
        """
        # FIXME:6306:
        # Sometimes, the received data is in memory.
        # For many API requests, we expect the response to be less than 1MB.
        # We should enforce a size for the received data,
        # and allow unlimited size only for special file transfer requeste.
        # In this way, we can precent an memory attack from a malicious server.
        self._received.append(data)
        self._timeout_call.reset(self._timeout)

    def connectionLost(self, reason):
        """
        Deliver the accumulated response bytes to the waiting L{Deferred}, if
        the response body has been completely received without error.
        """
        if self._deferred.called:
            # Nothing to do.
            # We already got the deferred triggered, most probably due to
            # timeout.
            # We will ignore the error from the HTTP protocol parser, since
            # there is nothing we can do with it.
            return

        received = b''.join(self._received)

        if reason.check(ResponseDone):
            # All good.
            self._deferred.callback(received)
        elif reason.check(PotentialDataLoss):
            # When we don't know for sure if all data was received, just ignore
            # the error and try to parse what we have so far.
            # This can happen for HTTP 1.0
            self._deferred.callback(received)
        else:
            self._deferred.errback(reason)


class _ReadIgnoreBodyProtocolWithTimeout(_ReadBodyProtocolWithTimeout):
    """
    Protocol that reads and discards data sent to it
    and which will time out if data is not received in a time interval.
    """
    def dataReceived(self, data):
        """
        Ignore the received data without accumulating it.
        Reset the timeout timer,
        """
        self._timeout_call.reset(self._timeout)


@delegate_to(IConsumer, context='_context')
@implementer(IConsumer)
class _ConsumerTimeoutWrapper(object):
    """
    A consumer which will reset the timeout each time a chunk is
    written.
    """

    def __init__(self, context, timeout_reset, timeout_interval):
        self._context = context
        self._timeout_reset = timeout_reset
        self._timeout_interval = timeout_interval

    def write(self, data):
        """
        Called when new data is produced.
        """
        # Reset the timeout and forward the data and bubble the result.
        self._timeout_reset(self._timeout_interval)
        return self._context.write(data)


@implementer(IBodyProducer)
@delegate_to(IBodyProducer, context='_context')
class _ProducerWithTimeout(object):
    """
    A HTTP body producer which will wrap a consumer which will reset the
    timeout each time a chunk is produced.
    """

    def __init__(self, context, timeout_reset, timeout_interval):
        self._context = context
        self._timeout_reset = timeout_reset
        self._timeout_interval = timeout_interval

    def startProducing(self, consumer):
        """
        Called when we have a consumer where we can send date.
        """
        wrapped_consumer = _ConsumerTimeoutWrapper(
            consumer, self._timeout_reset, self._timeout_interval)
        return self._context.startProducing(wrapped_consumer)


class AgentMixin(object):
    """
    Shared code for all HTTP/1.1 agents.

    Code using this should not call the low level `request` method,
    but one of the specialized `get`, `post` ..etc methods.

    The URL should be passed as Unicode, already encoded.
    """
    _reactor = reactor

    # Identifier used for the User-Agent header added to all the requests.
    _user_agent = b'SFTPPlus/http-agent'

    # Headers sent for every request.
    _general_headers = None

    # Default timeout for waiting to receive all the headers and waiting
    # to receive chunks of body.
    _timeout = 120

    def _getContextFactory(self, ssl_context):
        """
        Return the SSL factory for HTTP.
        """
        if ssl_context is None:
            # Go with default
            return BrowserLikePolicyForHTTPS()
        else:
            return _ChevahPolicyForHTTPS(ssl_context)

    def _setGeneralHeaders(self, key, values):
        """
        Set the general headers with `key` and `values`.
        """
        if self._general_headers is None:
            self._general_headers = {}
        self._general_headers[key] = values

    def _setAuthorization(self, credentials):
        """
        Set the `Authorization` header based on `credentials`.
        """
        if not credentials or not credentials.username:
            return

        # For now only Basic authentication is supported.
        auth_value = get_basic_http_header(
            credentials.username, credentials.password)
        self._setGeneralHeaders(b'Authorization', [auth_value])

    def postJSON(
            self, url, body, headers=None, retry_count=0, retry_increase=0):
        """
        Make a POST request with JSON body.
        """
        if headers is None:
            headers = {}

        headers = headers.copy()
        # Make sure content type is always json.
        headers['Content-Type'] = ['application/json; charset=utf-8']

        return self.post(
            url,
            headers=headers,
            body=json.dumps(body).encode('utf-8'),
            retry_count=retry_count,
            retry_increase=retry_increase,
            )

    def _makeRequest(
            self, method, url, headers,
            body=None, retry_count=0, retry_increase=0,
            ):
        """
        Adapt simple arguments to the low level objects required by the
        agent API.

        `retry_count` can be used only when there is no `body` or body is
        a string.
        """
        headers['User-Agent'] = [self._user_agent]

        if self._general_headers:
            headers.update(self._general_headers)

        if isinstance(body, six.string_types + (bytes,)):
            if isinstance(body, six.text_type):
                content_type = headers.get(
                    'Content-Type', ['application/octet-stream'])[0]
                headers['Content-Type'] = ['{}; charset=utf-8'.format(
                    content_type)]
                body = body.encode('utf-8')
        elif retry_count:
            raise RuntimeError('Retry for non-text body is not supported.')

        encoded_headers = {}
        for key, value in headers.items():
            if isinstance(key, six.text_type):
                key = key.encode('ascii')

            if not isinstance(value, list):
                value = [value]

            if isinstance(value[0], six.text_type):
                value[0] = value[0].encode('utf-8')

            encoded_headers[key] = value

        deferred = self._requestWithRetry(
            method, url, encoded_headers, body, retry_count, retry_increase)
        deferred.addErrback(_eb_request)
        return deferred

    @defer.inlineCallbacks
    def _requestWithRetry(
            self, method, url, headers, body, retry_count, retry_increase):
        """
        Make the request and retry on connection errors or 5XX codes.
        """
        delay = 0
        for i in range(retry_count):
            try:
                response = yield self.__requestOnce(method, url, headers, body)
                code = response.code
                if code < 500 or code >= 600:
                    # This is not a 5XX code. No need to retry.
                    defer.returnValue(response)

                # We don't care about the response as it will be retried,
                # but we need to consume the body.
                yield self._agent.readBodyWithIgnore(response)
                raise Exception('Attempt failed')

            except Exception:
                # Ignore any error and just wait for a retry.
                delay += retry_increase
                yield deferLater(self._reactor, delay, lambda: None)

        # All retried failed. Do one last request and forward any result.
        response = yield self.__requestOnce(method, url, headers, body)
        defer.returnValue(response)

    def __requestOnce(self, method, url, headers, body):
        """
        Make the request and return the result.
        """

        # This is the timeout for getting the headers.
        # We create an early deferred so that we can pass it to the
        # timeout wrapper.
        timeout_call = None

        def timeout_reset(interval):
            if timeout_call:
                timeout_call.reset(interval)

        if body is not None:
            if isinstance(body, six.binary_type):
                body = FileBodyProducer(BytesIO(body))

            # We wrap the producer so that we can reset the timeout each time
            # a chunk is produced.
            # A bigger timeout is used as when data is uploaded, after the
            # last chunk is uploaded the server might need a long time to
            # process the received data and return a response.
            body = _ProducerWithTimeout(body, timeout_reset, self._timeout)

        if isinstance(url, URL):
            encoded_url = url.to_uri().to_text().encode('ascii')
        else:
            # FIXME:6201:
            # See how we want to use this API - Unicode vs URL-encoded.
            # The current UTF is just wrong.
            encoded_url = url.encode('utf-8')

        request_deferred = self.request(
            method=method.encode('ascii'),
            uri=encoded_url,
            headers=Headers(headers),
            bodyProducer=body,
            )
        timeout_call = add_timeout(
            self._reactor, request_deferred, self._timeout,
            message='HTTP request time out before receiving the headers.',
            )
        return request_deferred

    def post(self, url, headers, body, retry_count=0, retry_increase=0):
        """
        Make a POST request with body as string.
        """
        return self._makeRequest(
            method='POST',
            url=url,
            headers=headers,
            body=body,
            retry_count=retry_count,
            retry_increase=retry_increase,
            )

    def put(self, url, headers, body, retry_count=0, retry_increase=0):
        """
        Make a POST request with body as string.
        """
        return self._makeRequest(
            method='PUT',
            url=url,
            headers=headers,
            body=body,
            retry_count=retry_count,
            retry_increase=retry_increase,
            )

    def get(self, url, headers, body=None, retry_count=0, retry_increase=0):
        """
        Make a GET request.

        The body is ignored, this is here to help share some code
        dispatching between GET and PUT/POST.
        """
        return self._makeRequest(
            method='GET',
            url=url,
            headers=headers,
            body=b'',
            retry_count=retry_count,
            retry_increase=retry_increase,
            )

    def head(self, url, headers):
        """
        Make a HEAD request.
        """
        return self._makeRequest(
            method='HEAD',
            url=url,
            headers=headers,
            )

    def delete(self, url, headers):
        """
        Make a DELETE request.
        """
        return self._makeRequest(
            method='DELETE',
            url=url,
            headers=headers,
            )

    def readBody(self, response):
        """
        Get the body of a L{IResponse} and return it as bytes in one go as
        a deferred.

        This is a helper function for clients that don't want to incrementally
        receive the body of an HTTP response.
        """
        protocol = _ReadBodyProtocolWithTimeout(self._reactor, self._timeout)
        response.deliverBody(protocol)
        return protocol.deferred

    def readBodyWithIgnore(self, response):
        """
        Read and discard the received body, to free the persistent connection.
        """
        protocol = _ReadIgnoreBodyProtocolWithTimeout(
            self._reactor, self._timeout)
        response.deliverBody(protocol)
        return protocol.deferred


class AuditedHTTPClientProtocol(HTTP11ClientProtocol, SSLConnectionMixin):
    """
    An HTTP protocol which will emit an event when the transport is connected
    and disconnected.

    Twisted HTTP client protocol handles multiple requests in a single protocol
    and we audit only the connections, not the requests.
    """
    # FIXME:6398:
    # Add dedicated event IDs for http client side.
    _TLS_OK_EVENT = None
    _TLS_FAIL_EVENT = None

    def __init__(self, quiescentCallback, event_emitter):
        self._emit = event_emitter
        self._session_established = False
        # FIXME:6398:
        # If we wait for handshake, default value should be defer.succeed(None)
        self._tls_handshake_completed = defer.Deferred()
        super(AuditedHTTPClientProtocol, self).__init__(quiescentCallback)

    def connectionMade(self):
        """
        Called when we are connected.

        At this point the certificate is not available and this might be the
        connection to the proxy.
        """
        self.avatar = PeerAvatar(self.transport.getPeer())

    def proxyConnectionMade(self, address, port):
        """
        Called when the protocol is proxied and the connection is made to
        the final destination. The connection which is tunneled.
        """
        self.avatar = PeerAvatar(IPv4Address(
            host=address,
            port=port,
            type='TCP',
            ))

    def dataReceived(self, bytes):
        """
        Called when data was received.

        This should be called only once, as after that it will short-circuit
        and always call parent method.
        """
        self._session_established = True
        data = {
            'certificate': self._getPeerCertificateDescription(),
            'encryption': self._getEncryptionDescription(),
            # FIXME:6502:
            # Now it always returns the IP.
            'hostname': force_unicode(self.transport.addr[0]),
            }
        self.emitEvent('40033', data=data)

        # Replace the implementation with parent one.
        dataReceived = super(AuditedHTTPClientProtocol, self).dataReceived
        self.dataReceived = dataReceived
        dataReceived(bytes)

    def connectionLost(self, reason):
        """
        The underlying transport went away.
        """
        data = {
            'session_established': self._session_established,
            # FIXME:5178:
            # Now it always returns the IP.
            'hostname': force_unicode(self.transport.addr[0]),
            }
        self.emitEvent('40032', data=data)
        super(AuditedHTTPClientProtocol, self).connectionLost(reason)


class AuditedHTTPClientFactory(Factory):
    """
    Factory producing audited protocol instances.
    """
    def __init__(self, quiescentCallback, event_emitter):
        self._quiescentCallback = quiescentCallback
        self._event_emitter = event_emitter

    def buildProtocol(self, addr):
        return AuditedHTTPClientProtocol(
            self._quiescentCallback, self._event_emitter)


class AuditedHTTPConnectionPool(HTTPConnectionPool):
    """
    A connection pool which will emit an event when a new connection is made
    """
    _factory = AuditedHTTPClientFactory

    def __init__(self, reactor, event_emitter, persistent=True):
        self._event_emitter = event_emitter
        super(AuditedHTTPConnectionPool, self).__init__(
            reactor=reactor, persistent=persistent)

    def _newConnection(self, key, endpoint):
        """
        Create a new connection.

        This is the upstream Twisted code, but injects the event emitter.
        """
        def quiescentCallback(protocol):
            self._putConnection(key, protocol)

        factory = self._factory(quiescentCallback, self._event_emitter)
        deferred = endpoint.connect(factory)
        return deferred


@implementer(IOpenSSLClientConnectionCreator)
class _ChevahClientTLSOptions(object):
    """
    Client creator for TLS.

    This replaces Twisted implementation so that it will not mess with
    our own info_callback and app_data.
    """

    def __init__(self, hostname, ctx):
        self._hostname = hostname
        self._ctx = ctx

    def clientConnectionForTLS(self, tlsProtocol):
        """
        Create a TLS connection for a client.
        """
        connection = SSL.Connection(self._ctx, None)
        server_name = self._hostname.encode('utf-8')
        try:
            connection.set_tlsext_host_name(server_name)
        except ValueError:
            """
            Older OpenSSL version don't support SNI, but we ignore the
            error.
            """

        return connection


class _ChevahPolicyForHTTPS(BrowserLikePolicyForHTTPS):
    """
    SSL connection creator for web clients based on an already
    created SSL context.
    """
    def __init__(self, ssl_context):
        self._ssl_context = ssl_context

    def creatorForNetloc(self, hostname, port):
        """
        Create a L{client connection creator
        <twisted.internet.interfaces.IOpenSSLClientConnectionCreator>} for a
        given network location.
        """
        return _ChevahClientTLSOptions(
            hostname=hostname.decode("ascii"),
            ctx=self._ssl_context,
            )


class NonPersistentAgent(Agent, AgentMixin):
    """
    HTTP/HTTPS client which closes the connection after each request.

    When `credentials` are set it will make the requests using
    HTTP Basic authentication.
    """

    def __init__(self, event_emitter, ssl_context=None, credentials=None):
        pool = AuditedHTTPConnectionPool(
            self._reactor, event_emitter=event_emitter, persistent=False)
        self.emitEvent = event_emitter
        super(NonPersistentAgent, self).__init__(
            reactor=self._reactor,
            contextFactory=self._getContextFactory(ssl_context),
            connectTimeout=CONNECT_TIMEOUT,
            pool=pool,
            )
        self._setAuthorization(credentials)


class PersistentAgent(Agent, AgentMixin):
    """
    HTTP/HTTPS Client agent with a connection pool.

    When `credentials` are set it will make the requests using
    HTTP Basic authentication.
    """

    def __init__(
        self, event_emitter, pool_size=3, ssl_context=None, credentials=None,
        timeout=120,
            ):
        pool = AuditedHTTPConnectionPool(
            self._reactor, event_emitter=event_emitter, persistent=True)
        pool.retryAutomatically = False
        pool.maxPersistentPerHost = pool_size
        self.emitEvent = event_emitter
        self._timeout = timeout
        super(PersistentAgent, self).__init__(
            reactor=self._reactor,
            contextFactory=self._getContextFactory(ssl_context),
            connectTimeout=timeout,
            pool=pool,
            )
        self._setAuthorization(credentials)

    def closePersistentConnections(self):
        """
        Close persistent connection associated with this agent.
        """
        return self._pool.closeCachedConnections()


class RedundantAgent(object):
    """
    A persistent HTTP agent which makes requests to a list of redundant URLs.

    It likes like an agent, but it has methods for redundant requests against
    a fixed list of URLs.
    """

    # Number of seconds after which the main URL is retried.
    _RETRY_INTERVAL = 5 * 60

    def __init__(
        self, urls, event_emitter, ssl_context=None, credentials=None,
        timeout=120,
            ):
        self._configured_urls = urls
        self._current_urls = list(urls)
        self._agent = PersistentAgent(
            event_emitter=event_emitter,
            ssl_context=ssl_context,
            credentials=credentials,
            timeout=timeout,
            )
        self._restore_callback = None

        # Performance optimization when no redundancy was configured.
        if len(urls) < 2:
            self._requestWithRedundancy = self._requestWithoutRedundancy

    def closePersistentConnections(self):
        """
        Clean up the agent.
        """
        restore_callback, self._restore_callback = self._restore_callback, None
        if restore_callback and restore_callback.active():
            restore_callback.cancel()

        return self._agent.closePersistentConnections()

    @property
    def _reactor(self):
        """
        Delegate to the lower level agent.
        """
        return self._agent._reactor

    @_reactor.setter
    def _reactor(self, value):
        self._agent._reactor = value

    def readBody(self, response):
        """
        Delegate to the lower level agent.
        """
        return self._agent.readBody(response)

    def readBodyWithIgnore(self, response):
        """
        Delegate to the lower level agent.
        """
        return self._agent.readBodyWithIgnore(response)

    @property
    def _getEndpoint(self):
        """
        Delegate to lower level agent.
        """
        return self._agent._getEndpoint

    @_getEndpoint.setter
    def _getEndpoint(self, value):
        self._agent._getEndpoint = value

    def getWithRedundancy(
            self, body=None, headers=None, retry_count=0, retry_increase=0):
        """
        Make a GET which will fallback to a list of URLs.

        For now body is ignored.
        In the future we might used to append to the URL.
        """
        return self._requestWithRedundancy(
            self._agent.get, body=b'', headers=headers,
            retry_count=retry_count, retry_increase=retry_increase)

    def postWithRedundancy(
            self, body, headers=None, retry_count=0, retry_increase=0):
        """
        Make a POST which will fallback to a list of URLs.
        """
        return self._requestWithRedundancy(
            self._agent.post, body, headers,
            retry_count=retry_count, retry_increase=retry_increase)

    def putWithRedundancy(
            self, body, headers=None, retry_count=0, retry_increase=0):
        """
        Make a PUT which will fallback to a list of URLs.
        """
        return self._requestWithRedundancy(
            self._agent.put, body, headers,
            retry_count=retry_count, retry_increase=retry_increase)

    def postJSONWithRedundancy(self, body, headers=None):
        """
        Make a JSON post to multiple URLs.
        """
        return self._requestWithRedundancy(
            self._agent.postJSON, body, headers)

    @defer.inlineCallbacks
    def _requestWithRedundancy(
        self, handler, body,
        headers=None, retry_count=0, retry_increase=0,
            ):
        """
        Make a request to the first URL from the list of urls and
        try the remaining URL on error.

        It will fail with the response of the last URL.
        """
        response = None
        error = None
        for url in list(self._current_urls):
            try:
                response = yield handler(
                    url,
                    body=body,
                    headers=headers,
                    retry_count=retry_count,
                    retry_increase=retry_increase,
                    )

                if response.code >= 500 and response.code < 600:
                    # Convert normal HTTP error responses into critical errors.
                    raise ServerException('Server responded with %s: %s' % (
                        force_unicode(response.code),
                        force_unicode(response.phrase)))

            except Exception as catchall_error:
                # Since this failed, put it back in the list.
                self._current_urls.remove(url)
                self._current_urls.append(url)
                error = catchall_error
                self._agent.emitEvent(
                    '40042',
                    data={
                        'url': url,
                        'retry_interval': force_unicode(self._RETRY_INTERVAL),
                        'details': force_unicode(error),
                        },
                    )
                if self._restore_callback is not None:
                    self._restore_callback = self._agent._reactor.callLater(
                        self._RETRY_INTERVAL, self._recoverURLS)

            else:
                # We got a valid response.
                defer.returnValue(response)

        raise error

    def _requestWithoutRedundancy(
        self, handler, body,
        headers=None, retry_count=0, retry_increase=0,
            ):
        """
        Helper with performance when redundancy is not configured.
        """
        return handler(
            self._configured_urls[0],
            body=body,
            headers=headers,
            retry_count=retry_count,
            retry_increase=retry_increase,
            )

    def _recoverURLS(self):
        """
        Called to restore the configured order or URLs.
        """
        self._current_urls = list(self._configured_urls)
        self._restore_callback = None


@implementer(ITransferStream)
class HTTPDownloadedFile(Protocol):
    """
    A pull consumer for HTTP download.

    This is both an ITransferStream and IProtocol for Response.deliverBody.

    It should not be used outside of getStreamForReading.

    For fast downloads and short files we might receive all the data fast,
    so the read call is made after the data is received and connection is lost.
    """
    mode = 'reading'

    def __init__(self, path, location, response):
        self.path = path
        self.location = location
        self._buffer = []
        self._close_deferred = defer.Deferred()
        self._read_deferred = None
        self._closed = False
        self._connected = True

        # Make sure `self._connect` is set before this call
        # as `deliverBody` needs and will update it.
        response.deliverBody(self)

    def dataReceived(self, data):
        """
        See: IProtocol
        """
        if self._closed:
            self.transport.stopProducing()
            return

        if self._read_deferred:
            # There is someone waiting for us. Send the data right away.
            # And don't pause.
            full_data = b''.join(self._buffer) + data
            self._buffer = []
            read_deferred, self._read_deferred = self._read_deferred, None
            read_deferred.callback(full_data)
            return

        # Nobody is waiting for data to be consumed, so we pause for now.
        self.transport.pauseProducing()
        self._buffer.append(data)

    def connectionLost(self, reason):
        """
        See: IProtocol
        """
        self._connected = False

        if self._read_deferred:
            # There is someone waiting for us.
            data = b''.join(self._buffer)
            read_deferred, self._read_deferred = self._read_deferred, None
            read_deferred.callback(data)

        if reason.check(ResponseDone):
            self._close_deferred.callback(None)
        elif reason.check(PotentialDataLoss):
            self._close_deferred.errback(
                PartialDownloadError(self.status, self.message, ''))
        else:
            self._close_deferred.errback(reason)

    def close(self):
        """
        See: `ITransferStream`.
        """
        self._closed = True

        if self._connected:
            self.transport.stopProducing()

        # FIXME:3979:
        # By the time this is called, the connection should have already been
        # closed. Under normal flow, a connection lost without any error
        # I hope that this will be removed soon once we refactor the
        # ITransferStream.
        return self._close_deferred

    def write(self, data):
        """
        See: `ITransferStream`.
        """
        return defer.fail(ServerException('Only opened for reading.'))

    def read(self, length):
        """
        See: `ITransferStream`.
        """
        if self._closed:
            return defer.fail(ServerException(
                'Read called after stream was closed.'))

        # Always signal the transport that it can produce more data as we have
        # someone interested in the data.
        self.transport.resumeProducing()

        if not self._buffer and self._connected:
            # No data yet, but we are still in business.
            # So wait for data.
            self._read_deferred = defer.Deferred()
            return self._read_deferred

        # FIXME:3979:
        # Here we should check if no data was received yet and wait for data.
        # I hope that this will be removed soon once we refactor the
        # ITransferStream.
        data = b''.join(self._buffer)
        self._buffer = []
        # Return data from any buffer.
        return defer.succeed(data)


@implementer(ITransferStream, IBodyProducer)
class HTTPUploadedFile(UploadProducer):
    """
    Stream used to upload a file to the HTTP server.

    This is both an IBodyProducer and ITransferStream.

    It should not be used outside of getStreamForWriting.
    """

    def __init__(self, path, location, length):
        # Consumer is initiated a bit later.
        super(HTTPUploadedFile, self).__init__(consumer=None)
        self.location = location
        self.path = path
        self._consumer = None
        if length is None:
            length = UNKNOWN_LENGTH
        self.length = length

        # Called when HTTP connection request was made.
        self.connect_deferred = defer.Deferred()

        # Called we no longer have data to send.
        self.done_deferred = defer.Deferred()

        # Called when we got a response from the remote server.
        self.close_deferred = defer.Deferred()

    @defer.inlineCallbacks
    def write(self, data):
        """
        See: `ITransferStream`.

        The Twisted HTTP client API write() method doesn't return anything.
        The actual write might be delayed, for example if there is a need for
        a TLS refresh.
        """
        # FIXME:3979:
        # With this code we will no longer fail if write operations are
        # overlapped, since an overlapping write will wait for the previous
        # one to finalized
        if self._pause_write:
            yield self._pause_write
        yield super(HTTPUploadedFile, self).write(data)

    def close(self):
        """"
        See: `ITransferStream`.
        """
        if self._pause_write:
            # `close` called while a write is pending.
            # Release the write operation with an error.
            self._pause_write.errback(
                ServerException('Close called while still writing.'))

        # Calling done_deferred will cause the consumer to stop waiting for
        # us to produce.
        # This will end the HTTP request and will call stopProducing.
        if self._consumer:
            self.done_deferred.callback(None)
        else:
            self.done_deferred.errback(ServerException('Transfer failed.'))
        # Wait for close result.
        return self.close_deferred

    def startProducing(self, consumer):
        """
        See: IBodyProducer

        Called when connection is made and we can start writing.
        """
        self._consumer = consumer
        self.connect_deferred.callback(self)
        return self.done_deferred


def get_url_segments(path):
    """
    Return a sequence of path parts to help prepare the URL.

    Desiged to be used with hyperlink.URL.replace(path=THIS_RESULT)
    """
    if not path:
        return ()
    # All these characters are denied by Windows.
    # The following values are accepted by SharePoint Online
    # < (less than)
    # > (greater than)
    # : (colon)
    # " (double quote)
    # \ (backslash)
    # | (vertical bar or pipe)
    # The following characters are denied by hyperlink
    # ? (question mark)
    # # - accepted by Windows, but rejected by hyperlink
    # * (asterisk)
    return path.lstrip('/').split('/')


def get_url_path(segments):
    """
    """
    return '/' + '/'.join(segments)
