# Copyright (c) 2022 Adi Roiban.
# See LICENSE for details.
"""
Shared code for locations interacting with HTTP servers, as client-side.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

from twisted.internet import defer
from twisted.internet.address import IPv4Address
from twisted.internet.error import ConnectionDone
from zope.interface import implementer

from chevah.server.commons.credential import PasswordCredentials
from chevah.server.commons.exception import (
    HTTPClientException,
    ServerException,
    )
from chevah.server.commons.interface import ILocationShell
from chevah.server.location.commons import OnConnectionLostMixin
from chevah.server.http.client import PersistentAgent
from chevah.server.identity.avatar import ClientAvatar


@implementer(ILocationShell)
class _HTTPClientShell(OnConnectionLostMixin):
    """
    Shared code for client shell used by HTTP based locations.

    No actual file transfer operation is implemented here.
    Each client should implement their own way of dealing with the transfer.
    """

    def __init__(self, location, context_factory, scheduler=None):
        # The location which is using this shell.
        self._location = location
        self._scheduler = scheduler

        peer = IPv4Address(
            'TCP',
            self._location.configuration.address,
            self._location.configuration.port,
            )
        password_credentials = PasswordCredentials(
            username=self._location.configuration.username,
            password=self._location.configuration.password,
            creator=self._location,
            peer=peer,
            )
        self._avatar = ClientAvatar(password_credentials)

        self._agent = PersistentAgent(
            ssl_context=context_factory.getContext(),
            event_emitter=location.emitEvent,
            credentials=password_credentials,
            )

        self._url = self._location.configuration.url

    @property
    def avatar(self):
        """
        See: ILocationShell.
        """
        return self._avatar

    @defer.inlineCallbacks
    def authenticate(self):
        """
        See: `ILocationShell`.

        Just make a head request to validate the credentials.
        """
        # FIXME:5839:
        # Maybe the server has no support for HEAD.
        # In that case we might just do a simple TCP connection to see if
        # the server is alive.
        response = yield self._agent.head(self._url, {})
        # Consume the body, just in case, to consume the connection and
        # put it back into the pool.
        # HEAD should not return any body.
        yield self._agent.readBodyWithIgnore(response)
        if not self._isAuthenticationSuccessful(response.code):
            details = (
                'Failed to authenticate to server. Unexpected response code.')
            raise HTTPClientException(
                message=details,
                path='/',
                code=response.code,
                phrase=response.phrase,
                )
        defer.returnValue(self)

    def _isAuthenticationSuccessful(self, code):
        """
        Return True if http `code` and response `body` is a valid auth response
        for the HTTP server.
        """
        raise NotImplementedError(
            '_isAuthenticationSuccessful not implemented.')

    def close(self):
        """
        Called when the agent should no longer be used.
        """
        # WebDav is based on HTTP, which is stateless, so there is not
        # much to do here.
        self._onConnectionLost(ConnectionDone('Bye.'))
        return self._agent.closePersistentConnections()

    def rename(self, source_path, destination_path):
        """
        See: ILocation.
        """
        return defer.fail(ServerException(
            'Renaming remote file on %s is not supported.' % (
                self._location.type,)))

    def deleteFile(self, path):
        """
        See: ILocation.
        """
        return defer.fail(ServerException(
            'Deleting remote files on %s is not supported.' % (
                self._location.type,)))

    def deleteFolder(self, path, recursive=False):
        """
        See: ILocation.
        """
        return defer.fail(ServerException(
            'Deleting remote folders on %s is not supported.' % (
                self._location.type,)))

    def createFolder(self, path):
        """
        See: ILocation.
        """
        return defer.fail(ServerException(
            'Creating remote folders on %s is not supported.' % (
                self._location.type,)))

    def getAttributes(self, path):
        """
        See: ILocation.
        """
        return defer.fail(ServerException(
            'Getting attributes on %s is not supported.' % (
                self._location.type,)))

    def getFolderMemberAttributes(self, path):
        """
        See: ILocation.
        """
        return defer.fail(ServerException(
            'Directory listing on %s is not supported.' % (
                self._location.type,)))

    def getStreamForReading(self, path):
        """
        See: ILocation.
        """
        return defer.fail(ServerException(
            'Reading/receiving files from %s is not supported.' % (
                self._location.type,)))

    def getStreamForWriting(self, path, length=None):
        """
        See: ILocation.
        """
        return defer.fail(ServerException(
            'Writing/sending a file to %s is not supported.' % (
                self._location.type,)))

    def getStreamForAppending(self, path):
        """
        See: ILocation.
        """
        return defer.fail(ServerException(
            'Appending to a file on %s is not supported.' % (
                self._location.type,)))
