# Copyright (c) 2011 Adi Roiban.
# See LICENSE for details.
"""
Configuration for HTTP and HTTPS services.
"""

from __future__ import unicode_literals
from __future__ import absolute_import

from zope.interface import implementer

from chevah.server.commons.constant import (
    HTTP_PUBLIC_CHILD_NAME,
    TYPE_NAME,
    )
from chevah.server.commons.exception import ServerException
from chevah.server.commons.interface import (
    IHTTPServiceOptions,
    IHTTPRedirectServiceOptions,
    IHTTPSServiceOptions,
    )
from chevah.server.commons.service_configuration import (
    ConfigurationWithConnections,
    )
from chevah.server.commons.ssl import (
    SSLConfigurationSectionMixin,
    )
from chevah.server.configuration.model import (
    WritableBoolean,
    WritableString,
    WritableStringCaseInsensitive,
    WritableStringOrEmpty,
    WritableCSVLinesUnique,
    WritableCSVLines,
    WritableCSVUniqueInsensitive,
    )


@implementer(IHTTPServiceOptions)
class HTTPServiceConfigurationSection(ConfigurationWithConnections):
    """
    Service configuration section options within HTTP service.
    """
    PROPERTIES_REQUIRING_RESTART = (
        'accepted_origins',
        'announce_session_authentication',
        'as2_default_filename',
        'as2_no_mdn_success_text',
        'as2_receive_name',
        'as2_receive_path',
        'as2_pending_path',
        'as2_receive_key',
        'as2_receive_certificate',
        'client_forwarded_header',
        'headers',
        'metadata_fields',
        'public_account_uuid',
        'public_name',
        'theme_path',
        'ui_version',
        'login_footnotes',
        'languages',
        'triggers',
        )

    # FIXME:6254:
    # For now, the HTTP file server has no configurable path.
    # This is here to comply with the interface.
    base_url_path = ''

    headers = WritableCSVLinesUnique('headers')
    accepted_origins = WritableCSVUniqueInsensitive(
        'accepted_origins')
    client_forwarded_header = WritableStringCaseInsensitive(
        'client_forwarded_header', default='')
    announce_session_authentication = WritableBoolean(
        'announce_session_authentication')
    public_account_uuid = WritableStringOrEmpty('public_account_uuid')
    public_name = WritableString('public_name', default=HTTP_PUBLIC_CHILD_NAME)

    public_error = WritableBoolean('public_error', default=False)

    as2_receive_name = WritableString(
        'as2_receive_name', default='as2receive')
    as2_receive_path = WritableString('as2_receive_path', default='')
    as2_pending_path = WritableString(
        'as2_pending_path', default='_as2_pending')
    as2_receive_key = WritableString('as2_receive_key', default='')
    as2_receive_certificate = WritableString(
        'as2_receive_certificate', default='')
    as2_default_filename = WritableString('as2_default_filename', default='')
    as2_no_mdn_success_text = WritableString(
        'as2_no_mdn_success_text', default='')

    theme_path = WritableStringOrEmpty('theme_path')
    ui_version = WritableStringCaseInsensitive(
        'ui_version', default=TYPE_NAME.UI_GEN_2)
    languages = WritableCSVUniqueInsensitive('languages')
    triggers = WritableCSVLinesUnique('triggers')
    login_footnotes = WritableString('login_footnotes', default='')

    metadata_fields = WritableCSVLines(
        'metadata_fields', inherit=False, default=())


@implementer(IHTTPRedirectServiceOptions)
class HTTPRedirectServiceConfigurationSection(HTTPServiceConfigurationSection):
    """
    Service configuration section options within HTTP service.
    """

    PROPERTIES_REQUIRING_RESTART = (
        'accepted_origins',
        'headers',
        'redirect_url',
        )

    redirect_url = WritableString('redirect_url')

    @property
    def client_forwarded_header(self):
        """
        Always return empty as for redirect site we don't need to do
        authentication and get the proxied source address.
        """
        return ''

    @client_forwarded_header.setter
    def client_forwarded_header(self, value):
        """
        Does nothing. Here to comply with the interface.
        """


@implementer(IHTTPSServiceOptions)
class HTTPSServiceConfigurationSection(
        HTTPServiceConfigurationSection, SSLConfigurationSectionMixin):
    """
    Service configuration section options within HTTPS service.
    """

    @property
    def PROPERTIES_REQUIRING_RESTART(self):
        """
        See: `IProtocolOptions`.
        """
        # This code is complicated due to selective multiple inheritance
        # together with property decorator inheritance.
        # In a normal inheritance, only of of the
        # super properties_requiring_restart method would have been called.
        # Here we need to merge both.
        result = []
        result.extend(
            HTTPServiceConfigurationSection.PROPERTIES_REQUIRING_RESTART)
        result.extend(
            SSLConfigurationSectionMixin.PROPERTIES_REQUIRING_RESTART)
        return tuple(result)


_VALID_UI_VERSIONS = (
    TYPE_NAME.UI_GEN_1,
    TYPE_NAME.UI_GEN_2,
    )


def get_ui_version(ui_version):
    """
    Return the UI version to be used.

    Raise an exception for invalid configurations.
    """
    if ui_version not in _VALID_UI_VERSIONS:
        raise ServerException(
            'Invalid ui_version value: "%s".' % (ui_version,))

    return ui_version
