# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
Credentials used by the HTTP service.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

import binascii

from twisted.cred import error
from twisted.web.iweb import ICredentialFactory
from zope.interface import implementer

from chevah.server import json
from chevah.server.commons.constant import TYPE_NAME
from chevah.server.commons.credential import (
    HTTPPasswordCredentials,
    HTTPSPasswordCredentials,
    )
import six


class HTTPSBasicAuthCredentials(HTTPSPasswordCredentials):
    """
    Credentials for Basic Auth over HTTPS
    """

    @property
    def type(self):
        """
        See: `ICredentials`.
        """
        return TYPE_NAME.PASSWORD_BASIC_AUTH


class HTTPBasicAuthCredentials(HTTPPasswordCredentials):
    """
    Credentials for Basic Auth over HTTP
    """

    @property
    def type(self):
        """
        See: `ICredentials`.
        """
        return TYPE_NAME.PASSWORD_BASIC_AUTH


def _credentials_class_for(request):
    """
    Return the credentials class for `request`.
    """
    if request.isSecure():
        return HTTPSPasswordCredentials
    else:
        return HTTPPasswordCredentials


@implementer(ICredentialFactory)
class HTTPBasicAuthCredentialFactory(object):
    """
    Generator of credentials for HTTP Basic Auth.

    This should comply with the interface and raise the Twisted
    error.LoginFailed exception.
    """

    scheme = b'basic'

    def __init__(self, realm):
        self.realm = realm

    def getChallenge(self, request):
        """
        Return a challenge including the HTTP authentication realm with which
        this factory was created.
        """
        return {'realm': self.realm}

    def decode(self, response, request):
        """
        Parse the base64-encoded, colon-separated username and password into a
        L{credentials.UsernamePassword} instance.
        """

        def credentials_class_for(request):
            """
            Return the credentials class for `request`.
            """
            if request.isSecure():
                return HTTPSBasicAuthCredentials
            else:
                return HTTPBasicAuthCredentials

        try:
            creds = binascii.a2b_base64(response + b'===')
        except binascii.Error:
            raise error.LoginFailed('Bad base64 encoding.')

        credentials_class = credentials_class_for(request)

        creds = creds.split(b':', 1)
        if len(creds) == 2:
            try:
                return credentials_class(
                    username=creds[0].decode('utf-8'),
                    password=creds[1].decode('utf-8'),
                    request=request,
                    )
            except UnicodeError:
                raise error.LoginFailed('Bad UTF-8 encoding.')
        else:
            raise error.LoginFailed('Bad format.')


@implementer(ICredentialFactory)
class SessionAuthCredentialFactory(object):
    """
    Generates credentials for the `session` type of authorization header.

    When `hide` is True, it will not advertise its presence as part of the
    www-athenticate challenge.

    See get_login_page_credential for the credentials from the login page.
    """
    scheme = b'session'

    def __init__(self, get_session, hide=False):
        self._getSession = get_session
        self._hide = hide

    def getChallenge(self, request):
        """
        Return a challenge including the HTTP authentication realm with which
        this factory was created.
        """
        from chevah.server.http.resource import HIDE_CHALLENGE
        if self._hide:
            return HIDE_CHALLENGE
        return {}

    def decode(self, response, request):
        """
        Return the credentials for the session attached to the request which
        has the session id in `response`.
        """
        target_session = request.site.sessions.get(response, None)
        if not target_session:
            raise error.LoginFailed('Requested session does not exist.')

        # We don't set the cookies as this is non-browser API.
        request._setSession(target_session)
        # Now we can get the wrapped session.
        session = self._getSession(request)

        if not session.resource:
            raise error.LoginFailed('Non-authenticated session in header.')

        credentials_class = _credentials_class_for(request)
        return credentials_class(  # noqa:bandit
            username=session.resource.avatar.name,
            password='',
            request=request,
            )


def get_login_page_credential(request):
    """
    Generated credentials for the login done via the dedicated login page.

    Returns `None` when credentials are missing or are in wrong format.

    See SessionAuthCredentialFactory for credentials from HTTP headers.
    """
    content_type, _ = request.getContentType()
    if (
        content_type and
        content_type.lower() == 'application/x-www-form-urlencoded'
            ):
        username = request.args.get(b'username', [b''])[0]
        if not username:
            return None

        password = request.args.get(b'password', [None])[0]
        if password is None:
            return None

        mfa_password = request.args.get(b'mfa_password', [None])[0]
        if mfa_password:
            password += mfa_password

        username = username.decode('utf-8')
        password = password.decode('utf-8')

    else:
        content = request.content.getvalue()
        try:
            result = json.loads(content.decode('utf-8'))
        except Exception:
            # Failed to read the JSON content.
            return None

        username = result.get('username', '')
        if not username:
            return None
        if not isinstance(username, six.text_type):
            return None

        # We allow empty password, to support anonymous authentication.
        password = result.get('password', None)
        if not isinstance(password, six.text_type):
            return None

    credentials_class = _credentials_class_for(request)
    return credentials_class(
        username=username,
        password=password,
        request=request
        )
