# Copyright (c) 2011 Adi Roiban.
# See LICENSE for details.
"""
Code for HTTP file transfer service.

Some code from here is reused in Web Manager service.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

import os
import posixpath
import traceback

from bunch import Bunch
from twisted.internet import defer
from twisted.web import http, server
from zope.interface import implementer

from chevah.server import MODULE_PATH
from chevah.server.commons.constant import (
    BUILD_CHILD_NAME,
    CONSTANTS_JS_NAME,
    HTTP_STATIC_FOLDER_NAME,
    HTTP_PRIVATE_CHILD_NAME,
    HTTP_PUBLIC_ROOT,
    HTTP_REST_CHILD_NAME,
    HTTP_REST_LOGIN_NAME,
    HTTP_REST_LOGOUT_NAME,
    JS_COMMONS_CHILD_NAME,
    LIBS_CHILD_NAME,
    STATIC_FOLDER_NAME,
    )
from chevah.server.commons.exception import ServerException
from chevah.server.commons.factory import ServiceFactoryBase
from chevah.server.commons.interface import IWebSite, IFileTransferSite
from chevah.server.commons.ssl import SSLServiceFactoryMixin
from chevah.server.http.protocol import (
    HTTPInternalErrorProtocol,
    HTTPOverflowProtocol,
    HTTPProtocol,
    HTTPSProtocol,
    )
from chevah.server.http.as2 import (
    AS2ReceiveResource,
    get_AS2_key_and_certificate,
    )
from chevah.server.http.request import ChevahRequest
from chevah.server.http.resource import (
    ConstantsJSResource,
    FaviconResource,
    InternalErrorResource,
    RedirectResource,
    RootResource,
    StaticFilesystemResource,
    StaticTextResource,
    )
from chevah.server.http.rest import (
    RESTPublicRoot,
    RESTLoginResource,
    RESTLogoutResource,
    RESTRootResource,
    RESTRootRedirectResource,
    )
from chevah.server.commons.i18n import get_translation, NULL_TRANSLATION

# FIXME:5693:
# We can remove this once we no longer have weblibs,
# since all was migrated to node and npm.
from chevah.weblibs.bootstrap import MODULE_PATH as WEBLIBS_MODULE_PATH


class ChevahHTTPRequest(ChevahRequest):
    """
    A request designed for HTTP connections.
    """
    session_cookie_force_secure = False
    session_cookie_basename = 'CHEVAH_HTTP_SESSION'


class ChevahHTTPSRequest(ChevahRequest):
    """
    A request designed for HTTPS connections.

    It will always force secured session cookies.
    """
    session_cookie_basename = 'CHEVAH_HTTPS_SESSION'


class SiteWithoutLog(server.Site):
    """
    Twisted.web code implements log handling in the Site but we don't want
    that.

    This class is here to break inheritance chain.
    """

    def __init__(self):
        """
        Overwrite default initialization as the site is started without a
        root resource.
        """

    def startFactory(self):
        pass

    def stopFactory(self):
        pass

    def log(self, request):
        pass


class ChevahSession(server.Session):
    """
    The generic session used for the HTTP session.

    Each service or resource can then create separate session adaptor.
    This should not contain application or page specific code.
    """
    def __init__(self, site, uid, timeout, reactor):
        self.sessionTimeout = timeout
        server.Session.__init__(self, site, uid, reactor)


# Make sure ServiceFactoryBase is inherited before SiteWithoutLog as the later
# breaks the inheritance chain for startFactory and stopFactory.
@implementer(IWebSite)
class HTTPFactoryBase(ServiceFactoryBase, SiteWithoutLog):
    """
    Base factory for HTTP/HTTPS connections.

    It is initialized with a static resource for `root_path` and resources
    for constants and web libraries.

    It is designed to work with a ChevahRequest as the `getResourceFor`
    can handle deferreds.
    """
    requestFactory = ChevahHTTPRequest

    protocol = HTTPProtocol
    overflowProtocol = HTTPOverflowProtocol
    internalErrorProtocol = HTTPInternalErrorProtocol

    # In production and in most tests we disconnect the client right away.
    # For testing we want a bit of delay as otherwise the client might
    # fail with a connection error, and we don't reach  the point at which
    # the client checks the HTTP response for our server.
    # Most of the time, the HTTP client during tests can send their buffer
    # quite fast so it's not disconnected in the middle of the send.
    # When this is 0, there is no delay and the code is executed right away
    # in a blocking way.
    _LOSE_CONNECTION_WITH_DELAY = False

    def __init__(self, service_configuration, on_error):
        super(HTTPFactoryBase, self).__init__(service_configuration, on_error)
        self.sessions = {}
        self._translation = NULL_TRANSLATION
        self._accepted_origins = self._service_configuration.accepted_origins
        self._base_url_path = posixpath.normpath(
            '/' + self._service_configuration.base_url_path
            ).replace('//', '/').rstrip('/')
        self._headers = self._getHeaders()
        self._client_forwarded_header = (
            self._service_configuration.client_forwarded_header)

        # Get the resource after resolving all the configuration.
        self._resource = self._getRootResource()

        # We put favicon.ico for various browsers.
        self._resource.putChild(b'favicon.ico', FaviconResource())

    @property
    def resource(self):
        """
        See: `IWebSite`.
        """
        return self._resource

    @property
    def headers(self):
        """
        See: `IWebSite`.
        """
        return self._headers

    @property
    def accepted_origins(self):
        """
        See: `IWebSite`.
        """
        return self._accepted_origins

    @property
    def base_url_path(self):
        """
        See: `IWebSite`.
        """
        return self._base_url_path

    @property
    def client_forwarded_header(self):
        """
        See: `IWebSite`.
        """
        return self._client_forwarded_header

    def sessionFactory(self, site, uid):
        """
        Return a new session for this site.

        `self` and `site` are  the exact same thing.
        This is how Twisted framework is designed to allow custom sessions.
        """
        timeout = self.service_configuration.idle_connection_timeout
        return ChevahSession(
            site, uid, timeout=timeout, reactor=self._scheduler)

    def _configureProtocol(self, protocol):
        """
        See: `ServiceFactoryBase`.
        """
        protocol.requestFactory = self.requestFactory
        protocol._scheduler = self._scheduler

    def _getHeaders(self):
        """
        Return a list with headers used for all responses.
        """
        result = []
        for header in self._service_configuration.headers:
            parts = header[0].split(':', 1)
            if len(parts) != 2:
                raise ServerException('Invalid format for headers.')
            result.append((parts[0].strip(), parts[1].strip()))
        return result

    def _getTriggers(self):
        """
        There are no triggers for general HTTP sites site.
        """
        return []

    def _getRootResource(self):
        """
        Return the root resource of this site.
        """
        raise NotImplementedError('Implement your own root resource.')

    @defer.inlineCallbacks
    def getResourceFor(self, request):
        """
        Traverse the URL of the `request` and return a deferred which fires
        with the resource based on request.

        This is the main entry point for resource dispatching.
        """
        request.site = self
        # sitepath is used by Twisted for some distributed stuff, but we
        # don't use it.
        request.sitepath = []
        # Start with the root.
        resource = self.resource

        try:
            while request.postpath and not resource.isLeaf:
                fragment = request.postpath.pop(0)
                request.prepath.append(fragment)
                resource = yield resource.getChildWithDefault(
                    fragment, request)
        except Exception:
            resource = InternalErrorResource(
                message='Internal error while retrieving resource.',
                details=traceback.format_exc(),
                )

        defer.returnValue(resource)

    def _populateStaticFolder(self, private_path):
        """
        Add JS constant file and libs.
        """
        # Add constants resource.
        private_path.putChild(
            CONSTANTS_JS_NAME.encode('ascii'),
            ConstantsJSResource(factory=self))

        # External JS/CSS/HTML libs provided by Python packages.
        libs_path = os.path.dirname(WEBLIBS_MODULE_PATH)
        if isinstance(libs_path, bytes):
            libs_path = libs_path.decode('utf-8')
        private_path.putChild(
            LIBS_CHILD_NAME.encode('ascii'),
            StaticFilesystemResource(path=libs_path))

        # Shared JS code.
        js_commons_path = os.path.join(
            MODULE_PATH,
            STATIC_FOLDER_NAME,
            JS_COMMONS_CHILD_NAME,
            )
        if isinstance(js_commons_path, bytes):
            js_commons_path = js_commons_path.decode('utf-8')
        private_path.putChild(
            JS_COMMONS_CHILD_NAME.encode('ascii'), StaticFilesystemResource(
                path=js_commons_path))

        #  All the generated content ready for deployment.
        build_path = os.path.join(
            MODULE_PATH,
            STATIC_FOLDER_NAME,
            BUILD_CHILD_NAME,
            )
        if isinstance(build_path, bytes):
            build_path = build_path.decode('utf-8')
        private_path.putChild(
            BUILD_CHILD_NAME.encode('ascii'), StaticFilesystemResource(
                path=build_path))


@implementer(IFileTransferSite)
class HTTPFactory(HTTPFactoryBase):
    """
    Service for file transfers over HTTP.

    Root is an empty resource which has:
    * /home/ - user files
    * /__chsps__/ - static utility files.
    """

    @property
    def public_error(self):
        """
        See: `IFileTransferSite`.
        """
        return self._service_configuration.public_error

    @property
    def triggers(self):
        """
        See: `IFileTransferSite`.
        """
        return self._triggers

    @property
    def metadata_configuration(self):
        """
        See: `IFileTransferSite`.
        """
        return self._metadata_configuration

    @property
    def as2_complete(self):
        """
        See: `IFileTransferSite`.
        """
        return self._as2_complete

    @property
    def as2_pending(self):
        """
        See: `IFileTransferSite`.
        """
        return self._as2_pending

    @property
    def as2_key(self):
        """
        See: `IFileTransferSite`.
        """
        return self._as2_certificate['key']

    @property
    def as2_certificate_asn1(self):
        """
        See: `IFileTransferSite`.
        """
        return self._as2_certificate['asn1']

    @property
    def as2_certificate_id(self):
        """
        See: `IFileTransferSite`.
        """
        return self._as2_certificate['id']

    def _getRootResource(self):
        """
        Return root resource for HTTP/HTTPS service.
        """
        self._triggers = self._getTriggers()
        self._metadata_configuration = _get_metadata_configuration(
            self._service_configuration.metadata_fields)
        # These are used by AS2 receive resources.
        # Instead of having a lot of AS2 related members we might
        # group them into a shared AS2 factory.
        self._as2_complete = _get_segments(
            self._service_configuration.as2_receive_path)
        self._as2_pending = _get_segments(
            self._service_configuration.as2_pending_path)
        self._as2_certificate = get_AS2_key_and_certificate(
            component_configuration=self._service_configuration,
            root_configuration=self.root.configuration.server,
            certificate_option='as2_receive_certificate',
            key_option='as2_receive_key',
            )
        self._as2_default_filename = (
            self._service_configuration.as2_default_filename)
        self._as2_no_mdn_success_text = (
            self._service_configuration.as2_no_mdn_success_text)

        # Load the translation ASAP as it will be used by the resources
        # from the tree.
        try:
            default_language = self._service_configuration.languages[0]
        except IndexError:
            # No language defined, Fallback to EN.
            default_language = 'en'

        self._translation = get_translation(default_language)

        rest_home = RESTRootResource(
            service_configuration=self.service_configuration,
            translation=self._translation,
            )

        # Public files.
        public_account_uuid = self.service_configuration.public_account_uuid
        public_name = self.service_configuration.public_name
        public_folder = None
        if public_account_uuid:
            public_folder = RESTPublicRoot(
                self.service_configuration,
                self._translation,
                )

        root_path = os.path.join(
            MODULE_PATH,
            STATIC_FOLDER_NAME,
            HTTP_STATIC_FOLDER_NAME,
            )
        if isinstance(root_path, bytes):
            root_path = root_path.decode('utf-8')
        private_path = StaticFilesystemResource(path=root_path)

        self._populateStaticFolder(private_path)

        private_path.putChild(
            HTTP_REST_LOGOUT_NAME.encode('ascii'),
            RESTLogoutResource(logout=rest_home._logout),
            )
        private_path.putChild(
            HTTP_REST_LOGIN_NAME.encode('ascii'),
            RESTLoginResource(
                service_configuration=self.service_configuration,
                login=rest_home._login,
                translation=self._translation,
                ),
            )

        self._addThemeResource(private_path)

        root = RootResource(
            root_resource=RESTRootRedirectResource(),
            other_resource=StaticTextResource('', content_type='text/html'),
            )

        if public_folder and public_name == HTTP_PUBLIC_ROOT:
            # Have public path as the root.
            root = public_folder

        if public_folder and public_name != HTTP_PUBLIC_ROOT:
            # /public (based on configuration)
            root.putChild(public_name.encode('utf-8'), public_folder)

        # /home
        root.putChild(HTTP_REST_CHILD_NAME.encode('ascii'), rest_home)
        # /__chsps__
        root.putChild(HTTP_PRIVATE_CHILD_NAME.encode('ascii'), private_path)

        self._addAS2Resource(root)
        return root

    def _addThemeResource(self, private_path):
        """
        Add default or custom theme files.
        """
        theme_path = self.service_configuration.theme_path
        if not theme_path:
            theme_path = os.path.join(
                MODULE_PATH,
                STATIC_FOLDER_NAME,
                HTTP_STATIC_FOLDER_NAME,
                'theme_default'
                )

        theme_folder = StaticFilesystemResource(path=theme_path)
        private_path.putChild(b'theme', theme_folder)

    def _addAS2Resource(self, root):
        """
        Add the AS2 receiving resource if needed.
        """
        if not self._as2_certificate['key']:
            # Don't enable AS2 if we don't have a private key, for
            # backward compatiblity.
            return

        as2_resource = AS2ReceiveResource(
            service_configuration=self.service_configuration)

        # Normalize the path, in the case Windows separators are used.
        path = self.service_configuration.as2_receive_name.replace('\\', '/')
        # Remove any leading or trailing separators as they are not used.
        path = path.strip('/')

        if '/' not in path:
            # We have a direct root AS2 resource:
            root.putChild(path.encode('utf-8'), as2_resource)
            return

        parts = path.split('/')

        # The last part of the path is the name of our AS2 resource.
        name = parts[-1]
        # Start with root as the main parent.
        parent = root
        # Then look into each sub-path for AS2 to see if we can reuse an
        # existing parent or create a new one.
        for parent_name in parts[:-1]:
            new_parent = parent.getChildWithDefault(
                parent_name, 'resource-ignored')
            code = getattr(new_parent, 'code', 'ignored')
            if code == http.NOT_FOUND:
                new_parent = StaticTextResource(
                    'AS2 Parent', content_type='text/html')
                parent.putChild(parent_name.encode('utf-8'), new_parent)
            parent = new_parent

        parent.putChild(name.encode('utf-8'), as2_resource)
        return

    def _getTriggers(self):
        """
        Return the list of triggers for all accounts.
        """
        result = []
        for trigger in self._service_configuration.triggers:
            if len(trigger) < 2:
                raise ServerException('Invalid triggers configuration.')
            result.append(trigger)
        return result


def _get_segments(path):
    """
    Return the list of segments relative to home path.
    """
    path = path.strip('/').split('/')
    return [p.strip() for p in path if p.strip()]


_VALID_METADATA_OPTIONS = (
    'optional',
    'required',
    )


def _get_metadata_configuration(value):
    """
    Return the parsed configuration for metadata.

    Return None if no metadata header is configured.

    Raise an exception for invalid configurations.
    """
    def extract_options(values):
        """
        Return a list of lowercase options for the list of configured values
        for a rule definition.

        The first value from `values` is ignored as it's not expected to be
        an option, but the header name or a field key.
        """
        result = [o.lower() for o in values[1:]]

        if 'optional' in result and 'required' in result:
            raise ServerException(
                '"metadata_fields" can not be both optional and required.')

        if not result:
            # Add default options.
            result = ['optional']

        for option in result:
            if option not in _VALID_METADATA_OPTIONS:
                raise ServerException(
                    'Unknown option configured for "metadata_fields": %s.' % (
                        option,))

        return result

    if not value:
        return None

    header_name = value[0][0].lower()
    header_required = False
    required_keys = []
    optional_keys = []

    # Check header options.
    header_options = extract_options(value[0])
    if 'required' in header_options:
        header_required = True

    # Check header fields options.
    for line in value[1:]:
        key = line[0]
        options = extract_options(line)

        if 'optional' in options:
            optional_keys.append(key)

        if 'required' in options:
            required_keys.append(key)

    if not (required_keys or optional_keys):
        # If we don't have any keys, allow any keys by default.
        # This will allow a simple configuration in which you only
        # configure the header name,
        optional_keys = ['*']

    return Bunch(
        header_name=header_name,
        header_required=header_required,
        required_keys=required_keys,
        optional_keys=optional_keys,
        )


class HTTPRedirectFactory(HTTPFactoryBase):
    """
    Service for HTTP redirection.

    Any request is redirected
    """

    def _getRootResource(self):
        """
        Return root resource for HTTP/HTTPS redirection service.
        """
        redirect_url = self.service_configuration.redirect_url
        if not redirect_url:
            raise ServerException('No value configured for "redirect_url".')
        root = RedirectResource(
            url=self.service_configuration.redirect_url,
            code=http.MOVED_PERMANENTLY,
            )
        root.isLeaf = True
        return root


class HTTPSFactory(HTTPFactory, SSLServiceFactoryMixin):
    """
    Service for file transfers over HTTPS.
    """
    protocol = HTTPSProtocol
    requestFactory = ChevahHTTPSRequest
