# Copyright (c) 2014 Adi Roiban.
# See LICENSE for details.
"""
Handlers for Flow.js requests.

Below is an example of Flow.js POST request:

------WebKitFormBoundaryvfC4N4mYAQoeYmRK
Content-Disposition: form-data; name="flowChunkNumber"

1
------WebKitFormBoundaryvfC4N4mYAQoeYmRK
Content-Disposition: form-data; name="flowChunkSize"

102400
------WebKitFormBoundaryvfC4N4mYAQoeYmRK
Content-Disposition: form-data; name="flowCurrentChunkSize"

668
------WebKitFormBoundaryvfC4N4mYAQoeYmRK
Content-Disposition: form-data; name="flowTotalSize"

668
------WebKitFormBoundaryvfC4N4mYAQoeYmRK
Content-Disposition: form-data; name="flowIdentifier"

668-test-ssh-dsa-1024
------WebKitFormBoundaryvfC4N4mYAQoeYmRK
Content-Disposition: form-data; name="flowFilename"

test-ssh-dsa-1024
------WebKitFormBoundaryvfC4N4mYAQoeYmRK
Content-Disposition: form-data; name="flowRelativePath"

test-ssh-dsa-1024
------WebKitFormBoundaryvfC4N4mYAQoeYmRK
Content-Disposition: form-data; name="flowTotalChunks"

1
------WebKitFormBoundaryvfC4N4mYAQoeYmRK
Content-Disposition: form-data; name="file"; filename="blob"
Content-Type: application/octet-stream


------WebKitFormBoundaryvfC4N4mYAQoeYmRK--
"""
from __future__ import unicode_literals
from __future__ import absolute_import

import uuid

from twisted.web import http

from chevah.server import force_unicode
from chevah.server.http.protocol import _CONTINUE
from chevah.server.http.resource import (
    ChevahResource,
    )


class InMemorySessionFlowJS(ChevahResource):
    """
    Handler for FlowJS which keeps all received data in memory on the session.

    For now it support uploading a single file with a single chunk.

    When file was successfully received, it return the file UUID.
    """

    def headersReceived(self, request):
        """
        See: `IChevahResource`.
        """
        if request.method == b'POST':
            # For now only multipart/form-data format is supported in POST.
            boundary = request.getContentBoundary()
            if not boundary:
                return (http.BAD_REQUEST, b'No form boundary.')
        else:
            # Only POST request are allowed.
            # In the future we would need to also support GET request for
            # supporting resume.
            return (
                http.BAD_REQUEST,
                b'Method %s is not supported.' % request.method,
                )

        # We don't want to create new session for upload request.
        # All upload request should already have been authenticated.
        session = request.getExistingSession()
        if not session:
            return (http.BAD_REQUEST, 'Request is not authorized.')

        # FIXME:2525:
        # For POST requests use a better form parser to reduce memory usage.
        return _CONTINUE

    def render_POST(self, request):
        """
        Process commands received via POST request.
        """
        # Right now, we only care about file field and ignore all other fields.
        content = request.getBinaryArgument('file')
        if not content:
            message = b'No content was sent.'
            request.setResponseCode(http.BAD_REQUEST, message)
            # We need to prefix the error with a marker as fusty-flow will
            # not read the HTTP response code.
            return b'ERROR: ' + message

        # `force_unicode` can be removed once we have no longer need py2
        # and we can just keep str().
        file_uuid = force_unicode(str(uuid.uuid4()))

        # FIXME:2529:
        # Use a self expiring storage so that we don't overwrite uploaded
        # files and allow uploading multiple files.
        session = request.getSession()

        # Over the session we keep text/unicode,
        # but on the wire for HTTP reponse we put bytes.
        session.flowjs = {file_uuid: content}
        return file_uuid.encode('ascii')
