# Copyright (c) 2014 Adi Roiban.
# See LICENSE for details.
"""
Code for handling data submitted via HTTP form.

We should try to keep this code independent of Chevah project as we plan
to send it upstream in Twisted.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

import six
from io import BytesIO

from twisted.internet.defer import Deferred
from twisted.internet.interfaces import IConsumer
from twisted.protocols.basic import LineReceiver
from zope.interface import implementer

from chevah.server.http.protocol import parse_options_header


class FormException(Exception):
    """
    General exception raised by the code from this module.
    """
    def __init__(self, message):
        self.message = message
        super(FormException, self).__init__(message)


class FormLengthExceededException(FormException):
    """
    Marker exception when received form data exceeds length limits.
    """


@implementer(IConsumer)
class MultiPartFormData(LineReceiver, object):
    """
    Accumulate multipart/form-data content delimited by `boundary` and call
    `open_callback` for each part.

    When `open_callback` return None, part content is stored in memory.
    For non-None values part content is sent to the returned stream.

    A single text line in the form request (excluding binary data) as well as
    total length for a text only part should be smaller than MAX_LENGTH.

    This is a partial implementation which only handles headers parting. Form
    data is still parsed using old in-memory code.
    This is why it also provides file object methods.
    """
    _STATE_START = 'start'
    _STATE_CONTENT = 'content_text'
    _STATE_HEADER = 'header'
    _STATE_END = 'end'

    #: State inside form part.
    _state = _STATE_START

    #: Buffer use to check boundary inside a streamed part.
    _stream_boundary_buffer = b''

    #: IUploadStream used for sending part content.
    _stream = None

    def __init__(self, boundary, open_callback, maximum_length=None):
        # High level API will pass the boundary as text,
        # but when look at the payload content, we use bytes.
        boundary = boundary.encode('ascii')
        self._start_boundary = b'--%s' % boundary
        self._end_boundary = b'--%s--' % boundary
        # In raw mode the boundary includes the starting new line delimiter.
        self._raw_boundary = self.delimiter + self._start_boundary

        if maximum_length is not None:
            self.MAX_LENGTH = maximum_length

        self._open_callback = open_callback
        self._deferred = Deferred()

        self._resetInternalState()
        self._resetPartState()

    def close(self):
        """
        Called by twisted.web.http.Request.finish() after all the data was
        received for the request.

        The form is set as the handler for the request content.
        """
        self.unregisterProducer()

    def registerProducer(self, producer, streaming=True):
        """
        Signal that we are receiving data from a streamed request.

        Only stream producer is supported.
        """
        self._resetInternalState()
        self._producer = producer

    def unregisterProducer(self):
        """
        Called after all data was received.
        """
        self._resetInternalState()
        # Deferred might be already called in case of errors.
        if not self._deferred.called:
            self._deferred.callback(None)

    @property
    def deferred(self):
        """
        Deferred which is called when form parsing is done.

        Return `None` on success or a failure on errors.
        """
        return self._deferred

    def _resetInternalState(self):
        """
        Does a reinitialization of global state.
        """
        self._producer = None
        self._state = self._STATE_START
        self._content = BytesIO()
        self._part = BytesIO()
        self._resetPartState()

    def _resetPartState(self):
        """
        Initialize state for parsing a new part.
        """
        self._headers = {}
        self._headers_raw = []

        # If file was not closed yet, then something went wrong and we use
        # a special method on the stream.
        if self._stream and not self._stream.closed:
            self._stream.interrupt(Exception(
                'Unexpected end of form part.'))
        self._stream = None

    def write(self, data):
        """
        Called by transport when content is received.

        Translates into dataReceived, the expected entry method for a Protocol.
        """
        self.dataReceived(data)

    def lineReceived(self, line):
        """
        See: LineReceiver.

        Dispatch form_STATE method.
        """
        state_method = getattr(self, '_form_%s' % self._state)
        state_method(line)

    def rawDataReceived(self, data):
        """
        See: LineReceiver.
        """
        found, raw, text = self._checkBoundary(data)

        if raw:
            self._stream.write(raw)

        if found:
            self._stream.close()
            self.setLineMode(extra=text)

    def sendLine(self, line):
        """
        See: LineReceiver.
        """
        raise RuntimeError('This is a read-only protocol.')

    def lineLengthExceeded(self, line):
        """
        See: LineReceiver.
        """
        self._deferred.errback(
            FormLengthExceededException('Form line too big.'))

    def _checkBoundary(self, data):
        """
        Look for boundary in raw data.

        Return a tuple with (found, raw, text).

        `found` is True if boundary was reached.
        `raw` is raw part from data.
        `text` is plain text part after raw data, including boundary.
        """
        found = False
        raw = b''
        text = b''
        search_buffer = self._stream_boundary_buffer + data

        boundary_position = search_buffer.find(self._raw_boundary)
        if boundary_position > -1:
            found = True
            raw = search_buffer[:boundary_position]
            text = search_buffer[boundary_position:]
        else:
            # Not found.
            if len(search_buffer) <= len(self._raw_boundary):
                # Search buffer is still not big enough.
                # No new raw data is produced.
                self._stream_boundary_buffer = search_buffer
            else:
                # Keep buffer to minimum size.
                buffer_start = len(search_buffer) - len(self._raw_boundary) + 1
                raw = search_buffer[:buffer_start]
                self._stream_boundary_buffer = search_buffer[buffer_start:]

        return (found, raw, text)

    def _form_start(self, line):
        """
        State before any boundary was received

        It can only change into `header` state.
        """
        # Ignore all data until we get the start marker.
        if line != self._start_boundary:
            return

        self._updateLine(line)
        self._state = self._STATE_HEADER

    def _form_end(self, line):
        """
        State after end boundary was received.
        """
        # Ignore data.
        pass

    def _form_header(self, line):
        """
        State while receiving header content.

        It should normally change into `content` state, but on errors
        is reset into `start` state.
        """
        self._updateLine(line)

        if not line:
            # Header ends / content starts with an empty line.
            self._initializeContent()

        elif line == self._start_boundary:
            # A part without end and content.
            self._finalizePart()

        elif line == self._end_boundary:
            # A part without content.
            # Write it as it is.
            self._finalizePart(end=True)

        else:
            update_headers(self._headers, line)
            self._headers_raw.append(line)

    def _initializeContent(self):
        """
        Prepare state for incoming part content.
        """
        self._state = self._STATE_CONTENT
        raw_headers = (
            self.delimiter.join(self._headers_raw) +
            self.delimiter +
            self.delimiter)
        self._stream = self._open_callback(self._headers, raw_headers)

        if self._stream:
            self.setRawMode()
            self._stream_boundary_buffer = b''

    def _form_content_text(self, line):
        """
        State while receiving text content.

        Raw content is received in rawDataReceived().
        It can only go into `start` state.
        """
        self._updateLine(line)

        if line == self._end_boundary:
            self._finalizePart(end=True)
        elif line == self._start_boundary:
            # We got a start marker before end marker.
            self._finalizePart()
        else:
            # Continue reading content.
            pass

    def _updateLine(self, line):
        """
        Update part with line content.
        """
        if (len(self._part.getvalue()) + len(line)) > self.MAX_LENGTH:
            self._deferred.errback(
                FormLengthExceededException('Form part too big.'))
            return

        self._part.write(line + self.delimiter)

    def _finalizePart(self, end=False):
        """
        Write current parsed part.
        """
        self._content.write(self._part.getvalue())
        self._part = BytesIO()

        if end:
            self._state = self._STATE_END
        else:
            self._state = self._STATE_HEADER

        self._resetPartState()

    def seek(self, position, semantic):
        """
        Proxy to internal BytesIO.
        """
        return self._content.seek(position, semantic)

    def tell(self):
        """
        Proxy to internal BytesIO.
        """
        return self._content.tell()

    def read(self):
        """
        Proxy to internal BytesIO.
        """
        return self._content.read()

    def readline(self, limit=None):
        """
        Proxy to internal BytesIO.
        """
        return self._content.readline(limit)


def update_headers(headers, line):
    """
    Update `headers` dict based on a single raw line bytes.

    header_name is lowercase.

    header_content is empty when header line is invalid and does not
    contain a delimiter.

    It can handle Unicode values encoded as various MIME standards.
    """
    content = b''
    parts = line.split(b':', 1)

    name = parts[0].strip().lower().decode('ascii')

    if not name and len(parts) == 1:
        # Empty line.
        return

    if len(parts) == 2:
        content = parts[1]

    if six.PY3:
        try:
            content = content.decode('utf-8')
        except UnicodeDecodeError:
            raise FormException('Form header with invalid encoding.')

    value, options = parse_options_header(content)

    # Once we no longer need py2, we can remove
    # the six.binary_type checks.
    if isinstance(value, six.binary_type):
        value = value.decode('utf-8')

    unicode_options = {}
    for k, v in options.items():
        converted_value = v
        converted_key = k
        try:
            if isinstance(k, six.binary_type):
                converted_key = k.decode('utf-8')
            if isinstance(v, six.binary_type):
                converted_value = v.decode('utf-8')
        except UnicodeDecodeError:
            raise FormException('Form header value with invalid encoding.')

        unicode_options[converted_key] = converted_value

    headers[name] = {'value': value, 'options': unicode_options}
