# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
HTTP protocol handler.

This is code that should have been upstream.
Consider submitting patching to Twisted to allow the level of customization
required for Chevah without having to mess with HTTPChannel.

The main changes are allowing support for
* debug of the request and dump the request.
* 100-Continue - call request.headersReceived()
* header size validation
"""
from __future__ import unicode_literals
from __future__ import absolute_import

import re
import six
from OpenSSL import SSL
from urllib.parse import unquote_to_bytes

from twisted.internet import defer, reactor
from twisted.web.iweb import INonQueuedRequestFactory
from twisted.web.http import HTTPChannel

from chevah.server import force_unicode, native_string
from chevah.server.commons.exception import HTTPRequestRejectedException
from chevah.server.commons.ssl import SSLConnectionMixin
from chevah.server.identity.avatar import PeerAvatar

_CONTINUE = 100
TOO_MANY_REQUESTS = 429


class HTTPProtocol(HTTPChannel, SSLConnectionMixin):
    """
    HTTPProtocol used by HTTP services.
    """
    _path = None
    avatar = None
    _dump = None
    _paused = False

    # Use to scheduler calls.
    _scheduler = reactor

    # True if the connection was closed due to timeout.
    _timeout_reached = False

    # FIXME:6397:
    # Add events for HTTP server-side TLS handshake.
    _TLS_OK_EVENT = None
    _TLS_FAIL_EVENT = None

    @property
    def site(self):
        return self.factory

    def debug(self, details):
        """
        Nothing is debugged by default.
        This is overwritten by the factory.
        """

    @property
    def idle_connection_timeout(self):
        """
        Convert our standard service `idle_connection_timeout` to
        Twisted HTTP protocol timeout.
        """
        timeout = self.timeOut
        if not timeout:
            return 0
        return timeout

    @idle_connection_timeout.setter
    def idle_connection_timeout(self, value):
        self.timeOut = value

    def callLater(self, period, func):
        """
        This is how Twisted TimeoutMixin is implemented to help with testing.
        """
        return self._scheduler.callLater(period, func)

    def connectionMade(self):
        """
        Called when we have a client connection.

        Added support for dumping the request and keeping track of the
        peer for ProxyProtocol.
        """
        if self.site.service_configuration.parent.debug:
            self._dump = open('http.request.dump', 'wb')

        # FIXME:6397:
        # Here we should setup a deferred that is already called for
        # non-HTTPS usage.
        self._tls_handshake_completed = defer.Deferred()

        self.avatar = PeerAvatar(
            self.transport.getPeer(), component=self.factory)

        # We emit the connection creation event a bit later to share the
        # generic peer avatar.
        # When the connection is created, we don't have information about
        # the encryption or the peer certificate.
        # This is emitted before the TLS handshake is initialized.
        self.emitEvent('40053')

        # FIXME:4698:
        # Emit an event on connection made.
        # Share the code with the HTTPS protocol as at this time there is no
        # cert.
        super(HTTPProtocol, self).connectionMade()

    def connectionLost(self, reason):
        """
        Called when the low-level TCP connection was closed.
        """
        if self._dump:
            self._dump.close()

        details = ''

        if isinstance(reason.value, SSL.Error):
            # Check if we have SSL error details and emit an event for
            # known errors.
            has_details = (
                hasattr(reason.value, 'args') and
                len(reason.value.args) > 0 and
                len(reason.value.args[0]) > 0
                )
            error_details = (
                native_string('SSL unknown'),
                native_string('SSL_UNKNOWN'),
                native_string('Unknown message'),
                )
            if has_details:
                error_details = reason.value.args[0][0]

            if error_details[1].lower() in [
                native_string('ssl3_get_client_certificate'),
                native_string('tls_process_client_certificate'),
                    ]:
                details = (
                    'HTTPS client peer did not send a valid certificate. '
                    '%s' % (force_unicode(reason.value),))
            else:
                # Emit the raw SSL error, hoping that it can help with
                # debugging.
                details = (
                    'HTTPS connection closed on the server-side '
                    'due to a failure. %s' % (
                        force_unicode(reason.getErrorMessage()),))

        if self._timeout_reached:
            # A timeout will overwrite any previous error details.
            details = 'Connection timeout.'

        self.emitEvent('40054', data={
            'certificate': self._getPeerCertificateDescription(),
            'encryption': self._getEncryptionDescription(),
            'details': details,
            })
        super(HTTPProtocol, self).connectionLost(reason)

    def timeoutConnection(self):
        """
        Called when the connection times out.

        Override to define behavior other than dropping the connection.
        """
        self._timeout_reached = True
        super(HTTPProtocol, self).timeoutConnection()

    def emitEvent(self, event_id, data=None):
        """
        Emit an event for this HTTP protocol instance.
        """
        # These are events emitted before the request is authenticated.
        self.factory.emitEvent(event_id, avatar=self.avatar, data=data)

    def dataReceived(self, data):
        """
        Called when we receive data from the client.
        """
        try:
            super(HTTPProtocol, self).dataReceived(data)
        except HTTPRequestRejectedException:
            # Ignore the exception as the request was rejected, the
            # transport was already requested to close and is not an ISE.
            # This only catches the synchronous requests.
            # Headers received and render might be async.
            return

    def rawDataReceived(self, data):
        """
        Here to allow dumping the request for debugging.
        """
        if self._dump:
            self._dump.write(data)
        return super(HTTPProtocol, self).rawDataReceived(data)

    def lineReceived(self, line):
        """
        Code imported from HTTPChannel with minimal changes to support
        calling allHeadersReceived and waiting for the returned deferred.

        FIXME:6502:
        Should be kept in sync after each Twisted upgrade.
        """

        # This is custom debug code.
        if self._dump:
            self._dump.write(line + b'\r\n')
        self.debug('< %s' % (force_unicode(line),))

        self.resetTimeout()

        self._receivedHeaderSize += len(line)
        if (self._receivedHeaderSize > self.totalHeadersSize):
            self._respondToBadRequestAndDisconnect()
            return

        if self._HTTPChannel__first_line:
            # if this connection is not persistent, drop any data which
            # the client (illegally) sent after the last request.
            if not self.persistent:
                self.dataReceived = self.lineReceived = lambda *args: None
                return

            # IE sends an extraneous empty line (\r\n) after a POST request;
            # eat up such a line, but only ONCE
            if not line and self._HTTPChannel__first_line == 1:
                self._HTTPChannel__first_line = 2
                return

            # create a new Request object
            if INonQueuedRequestFactory.providedBy(self.requestFactory):
                request = self.requestFactory(self)
            else:
                request = self.requestFactory(self, len(self.requests))
            self.requests.append(request)

            self._HTTPChannel__first_line = 0

            parts = line.split()
            if len(parts) != 3:
                self._respondToBadRequestAndDisconnect()
                return
            command, request, version = parts
            try:
                command.decode("ascii")
            except UnicodeDecodeError:
                self._respondToBadRequestAndDisconnect()
                return

            self._command = command
            self._path = request
            self._version = version
        elif line == b'':
            # End of headers.
            if self._HTTPChannel__header:
                ok = self.headerReceived(self._HTTPChannel__header)
                # If the last header we got is invalid, we MUST NOT proceed
                # with processing. We'll have sent a 400 anyway, so just stop.
                if not ok:
                    return
            self._HTTPChannel__header = b''

            # Change upstream code to continue only after
            # the resource has processed the headers.
            self.setRawMode()
            self._headers_deferred = self._allHeadersReceived()
            self._headers_deferred.addBoth(self._bbClearHeadersDeferred)
            self.debug(
                '< Only first 1500 bytes of body included in debug.\n%r' % (
                    self._buffer[:1500]))

        elif line[0] in b' \t':
            # Continuation of a multi line header.
            self._HTTPChannel__header = (
                self._HTTPChannel__header + b" " + line.lstrip(b" \t"))
        # Regular header line.
        # Processing of header line is delayed to allow accumulating multi
        # line headers.
        else:
            if self._HTTPChannel__header:
                self.headerReceived(self._HTTPChannel__header)
            self._HTTPChannel__header = line

    def _bbClearHeadersDeferred(self, result_or_failure):
        self._headers_deferred = None
        return result_or_failure

    def _allHeadersReceived(self):
        """
        Called when all headers are received.

        Forward the event to the request.

        Returns a deferred.

        This is custom Chevah code.
        """
        request = self.requests[-1]
        self.persistent = self.checkPersistence(request, self._version)

        deferred = defer.maybeDeferred(
            request.headersReceived,
            self._command, self._path, self._version)

        # This is done to avoid pausing/resuming the transport
        # when we already have a result.
        if (
            not deferred.called or
            isinstance(deferred.result, defer.Deferred)
                ):
            # Deferred is not yet resolved so further callbacks will not
            # be called right away so we need to pause the transport.
            self._networkProducer.pauseProducing()
            self._paused = True

        def cb_headers_received(result):
            """
            Called after the request has successfully processed the headers.
            """
            # Write the custom headers.
            for name, value in self.factory.headers:
                request.setHeader(name, value)

            if self.length == 0:
                self.allContentReceived()
                self.setLineMode()

            if self._paused:
                self._networkProducer.resumeProducing()

        def eb_headers_received(failure):
            """
            Called when we fail to process headers.
            """
            if failure.check(HTTPRequestRejectedException):
                # Ignore the exception as the request was rejected and
                # is not an ISE.
                return

            data = {
                'uri': force_unicode(self._path),
                'details': force_unicode(failure.value),
                'tb': failure.getBriefTraceback(),
                }
            # FIXME:6041:
            # Consider moving event 40024, into the generic HTTP close event.
            # Or dedicated event for close with failure.
            self.emitEvent('40024', data=data)
            self.transport.write(
                b'HTTP/1.1 500 Internal Server Error\r\n'
                b'Connection: close\r\n'
                b'Content-Lenght: 0\r\n'
                b'\r\n'
                )
            self.loseConnection()

        deferred.addCallback(cb_headers_received)
        deferred.addErrback(eb_headers_received)
        return deferred

    def _finishRequestBody(self, data):
        """
        Called when the body was received.

        This is not called for requests with empty body.
        """
        if self._headers_deferred:
            deferred = self._headers_deferred
        else:
            deferred = defer.succeed(None)

        def cb_body_done(result):
            """
            Called after headers were processed.
            """
            self.allContentReceived()
            self._dataBuffer.append(data)

        deferred.addCallback(cb_body_done)

    def requestDone(self, request):
        """
        Called by first request in queue when it is done.

        Most of the time this is called after allContentReceived as a
        request is done after all its data was received.
        But for Expect: 100-continue, a request can be done before
        """
        if self._HTTPChannel__first_line != 1:
            # Request was done before receiving all content.
            # Ex. Expect: 100-continue
            # Reset ALL state variables,
            # so we don't interfere with next request.
            # This is a duplicate of the code from
            # HTTPChannel.allContentReceived.
            self.length = 0
            self._receivedHeaderCount = 0
            self._receivedHeaderSize = 0
            self._HTTPChannel__first_line = 1
            self._transferDecoder = None
            del self._command, self._path, self._version

        return HTTPChannel.requestDone(self, request)

    def allContentReceived(self):
        """
        Called when request content was fully received.

        Skip processing if transport is disconnecting.
        """
        if not self.requests:
            return

        return HTTPChannel.allContentReceived(self)

    def _respondToBadRequestAndDisconnect(self):
        """
        This is a quick and dirty way of responding to bad requests.

        As described by HTTP standard we should be patient and accept the
        whole request from the client before sending a polite bad request
        response, even in the case when clients send tons of data.

        @param transport: Transport handling connection to the client.
        @type transport: L{interfaces.ITransport}
        """
        self.transport.write(
            b"HTTP/1.1 400 Bad Request\r\n"
            b"Connection: close\r\n"
            b"Content-Length: 0\r\n"
            b"\r\n"
            )
        # FIXME:6041:
        # We we should have more details why it fails.
        self.loseConnection()


class HTTPSProtocol(HTTPProtocol):
    """
    HTTPSProtocol used by HTTPS services.

    For now, nothing special.
    This is just the protocol.
    The TLS is handled at the transport level, and not at the protocol level.
    """


class HTTPInternalErrorProtocol(HTTPProtocol):
    """
    Protocol to handlers error while creating the proper HTTP protocol.
    """
    # Set by the factory after initialization.
    details = ''
    factory = None

    def connectionMade(self):
        super(HTTPInternalErrorProtocol, self).connectionMade()
        self.avatar = PeerAvatar(
            self.transport.getPeer(), component=self.factory)
        data = {
            'details': self.details,
            }
        self.emitEvent('40020', data=data)
        self.transport.write(
            b'HTTP/1.1 500 Internal Server Error\r\n'
            b'Connection: close\r\n'
            b'Content-Length: 0\r\n'
            b'\r\n'
            )
        # FIXME:6041:
        # We we should have more details why it fails.
        self.loseConnection()


class HTTPOverflowProtocol(HTTPProtocol):
    """
    Protocol to handler to many connections.
    """
    # Set by the factory after initialization.
    factory = None

    def connectionMade(self):
        super(HTTPOverflowProtocol, self).connectionMade()
        response_message = (
            b'Connection limit reached. Try again later.'
            )
        length = six.text_type(len(response_message)).encode('ascii')
        self.transport.write(
            b'HTTP/1.1 503 Service Unavailable\r\n'
            b'Connection: close\r\n'
            b'Content-Length: %s\r\n'
            b'\r\n'
            b'%s' % (length, response_message)
            )
        # FIXME:6041:
        # We we should have more details why it fails.
        self.loseConnection()


# FIXME:6521:
# Below is vendored and simplified code from werkzeug 1.0.1
# Latest 3.0.1 was not handing our invalid filename for AS2,
# so we ended with this copy.
# I hope that in the future we can remove this code and start using
# latest werkzeug again.
_option_header_piece_re = re.compile(
    r"""
    ;\s*,?\s*  # newlines were replaced with commas
    (?P<key>
        "[^"\\]*(?:\\.[^"\\]*)*"  # quoted string
    |
        [^\s;,=*]+  # token
    )
    (?:\*(?P<count>\d+))?  # *1, optional continuation index
    \s*
    (?:  # optionally followed by =value
        (?:  # equals sign, possibly with encoding
            \*\s*=\s*  # * indicates extended notation
            (?:  # optional encoding
                (?P<encoding>[^\s]+?)
                '(?P<language>[^\s]*?)'
            )?
        |
            =\s*  # basic notation
        )
        (?P<value>
            "[^"\\]*(?:\\.[^"\\]*)*"  # quoted string
        |
            [^;,]+  # token
        )?
    )?
    \s*
    """,
    flags=re.VERBOSE,
    )
_option_header_start_mime_type = re.compile(r",\s*([^;,\s]+)([;,]\s*.+)?")
_accept_re = re.compile(
    r"""
    (                       # media-range capturing-parenthesis
      [^\s;,]+              # type/subtype
      (?:[ \t]*;[ \t]*      # ";"
        (?:                 # parameter non-capturing-parenthesis
          [^\s;,q][^\s;,]*  # token that doesn't start with "q"
        |                   # or
          q[^\s;,=][^\s;,]* # token that is more than just "q"
        )
      )*                    # zero or more parameters
    )                       # end of media-range
    (?:[ \t]*;[ \t]*q=      # weight is a "q" parameter
      (\d*(?:\.\d+)?)       # qvalue capturing-parentheses
      [^,]*                 # "extension" accept params: who cares?
    )?                      # accept params are optional
    """,
    re.VERBOSE,
    )


def unquote_header_value(value, is_filename=False):
    r"""Unquotes a header value.  (Reversal of :func:`quote_header_value`).
    This does not use the real unquoting but what browsers are actually
    using for quoting.

    .. versionadded:: 0.5

    :param value: the header value to unquote.
    """
    if not value:
        return value

    if value[0] == value[-1] == '"':
        # this is not the real unquoting, but fixing this so that the
        # RFC is met will result in bugs with internet explorer and
        # probably some other browsers as well.  IE for example is
        # uploading files with "C:\foo\bar.txt" as filename
        value = value[1:-1]

        # if this is a filename and the starting characters look like
        # a UNC path, then just return the value without quotes.  Using the
        # replace sequence below on a UNC path has the effect of turning
        # the leading double slash into a single slash and then
        # _fix_ie_filename() doesn't work correctly.  See #458.
        if is_filename:
            return value

        if value[:2] == "\\\\":
            return value

        return value.replace("\\\\", "\\").replace('\\"', '"')

    return value


def parse_options_header(value):
    """
    This is code copied from werkzeug 1.0.1

    It has special handling for filename, that we need for AS2.

    Parse a ``Content-Type`` like header into a tuple with the content
    type and the options:

    >>> parse_options_header('text/html; charset=utf8')
    ('text/html', {'charset': 'utf8'})

    This should not be used to parse ``Cache-Control`` like headers that use
    a slightly different format.  For these headers use the
    :func:`parse_dict_header` function.

    :param value: the header to parse.
    :param multiple: Whether try to parse and return multiple MIME types
    :return: (mimetype, options) or (mimetype, options, mimetype, options,)
             if multiple=True
    """
    if not value:
        return "", {}

    result = []

    value = "," + value.replace("\n", ",")
    while value:  # pragma: no branch
        match = _option_header_start_mime_type.match(value)
        if not match:
            break
        result.append(match.group(1))  # mimetype
        options = {}
        # Parse options
        rest = match.group(2)
        continued_encoding = None
        while rest:
            optmatch = _option_header_piece_re.match(rest)
            if not optmatch:
                break
            option, count, encoding, language, option_value = optmatch.groups()
            # Continuations don't have to supply the encoding after the
            # first line. If we're in a continuation, track the current
            # encoding to use for subsequent lines. Reset it when the
            # continuation ends.
            if not count:
                continued_encoding = None
            else:
                if not encoding:
                    encoding = continued_encoding
                continued_encoding = encoding
            option = unquote_header_value(option)
            if option_value is not None:
                is_filename = option == "filename"
                option_value = unquote_header_value(
                    option_value, is_filename=is_filename)
                if encoding is not None:
                    option_value = unquote_to_bytes(
                        option_value).decode(encoding)
            if count:
                # Continuations append to the existing value. For
                # simplicity, this ignores the possibility of
                # out-of-order indices, which shouldn't happen anyway.
                options[option] = options.get(option, "") + option_value
            else:
                options[option] = option_value
            rest = rest[optmatch.end():]
        result.append(options)
        return tuple(result)

    return '', {}


def parse_accept_header(value):
    """
    Code vendored from werkzeug 1.0.1.

    Parses an HTTP Accept-* header.  This does not implement a complete
    valid algorithm but one that supports at least value and quality
    extraction.

    Returns a new :class:`Accept` object (basically a list of
    ``(value, quality)``
    tuples sorted by the quality with some additional accessor methods).

    The second parameter can be a subclass of :class:`Accept` that is created
    with the parsed values and returned.
    :param value: the accept header string to be parsed.
    :param cls: the wrapper class for the return value (can be
                         :class:`Accept` or a subclass thereof)
    :return: an instance of `cls`.
    """
    if not value:
        return Accept(None)

    result = []
    for match in _accept_re.finditer(value):
        quality = match.group(2)
        if not quality:
            quality = 1
        else:
            quality = max(min(float(quality), 1), 0)
        result.append((match.group(1), quality))
    return Accept(result)


class Accept(list):
    """
    Code vendored from werkzeug 1.0.1

    An :class:`Accept` object is just a list subclass for lists of
    ``(value, quality)`` tuples.  It is automatically sorted by specificity
    and quality.

    All :class:`Accept` objects work similar to a list but provide extra
    functionality for working with the data.  Containment checks are
    normalized to the rules of that header:

    >>> a = CharsetAccept([('ISO-8859-1', 1), ('utf-8', 0.7)])
    >>> a.best
    'ISO-8859-1'
    >>> 'iso-8859-1' in a
    True
    >>> 'UTF8' in a
    True
    >>> 'utf7' in a
    False

    To get the quality for an item you can use normal item lookup:

    >>> print a['utf-8']
    0.7
    >>> a['utf7']
    0

    .. versionchanged:: 0.5
       :class:`Accept` objects are forced immutable now.

    .. versionchanged:: 1.0.0
       :class:`Accept` internal values are no longer ordered
       alphabetically for equal quality tags. Instead the initial
       order is preserved.

    """

    def __init__(self, values=()):
        if values is None:
            list.__init__(self)
            return

        values = sorted(
            values,
            key=lambda x: (self._specificity(x[0]), x[1]),
            reverse=True,
            )
        list.__init__(self, values)

    def __repr__(self):
        return "%s([%s])" % (
            self.__class__.__name__,
            ", ".join("(%r, %s)" % (x, y) for x, y in self),
            )

    def values(self):
        """
        Iterate over all values.
        """
        for item in self:
            yield item[0]

    def _specificity(self, value):
        """
        Returns a tuple describing the value's specificity.
        """
        return (value != "*",)

    def _value_matches(self, value, item):
        """
        Check if a value matches a given accept item.
        """
        return item == "*" or item.lower() == value.lower()

    def _best_single_match(self, match):
        for client_item, quality in self:
            if self._value_matches(match, client_item):
                # self is sorted by specificity descending, we can exit
                return client_item, quality

    def best_match(self, matches, default=None):
        """Returns the best match from a list of possible matches based
        on the specificity and quality of the client. If two items have the
        same quality and specificity, the one is returned that comes first.

        :param matches: a list of matches to check for
        :param default: the value that is returned if none match
        """
        result = default
        best_quality = -1
        best_specificity = (-1,)
        for server_item in matches:
            match = self._best_single_match(server_item)
            if not match:
                continue
            client_item, quality = match
            specificity = self._specificity(client_item)
            if quality <= 0 or quality < best_quality:
                continue
            # better quality or same quality but more specific => better match
            if quality > best_quality or specificity > best_specificity:
                result = server_item
                best_quality = quality
                best_specificity = specificity
        return result
