# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
Request used by Chevah HTTP service.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

from io import BytesIO
import six
from six.moves.urllib.parse import unquote, urlparse

from twisted.internet.error import AlreadyCalled, AlreadyCancelled
from twisted.python import failure
from twisted.python.compat import intToBytes
from twisted.web import http
from twisted.web.server import Request
from zope.interface import implementer

from chevah.server import force_unicode
from chevah.server.commons.constant import (
    MAXIMUM_IN_MEMORY_RECEIVED_DATA,
    PRODUCT_NAME,
    )
from chevah.server.commons.exception import HTTPRequestRejectedException
from chevah.server.commons.interface import IChevahRequest
from chevah.server.commons.ssl import SSLConnectionMixin
from chevah.server.http.protocol import (
    _CONTINUE,
    parse_accept_header,
    parse_options_header
    )

# Uses as 422 Unprocessable Entity response.
UNPROCESSABLE_ENTITY = 422
http.RESPONSES[UNPROCESSABLE_ENTITY] = b'Unprocessable Entity'


def _get_server_signature():
    """
    Helper for formating server signature for HTTP usage.
    """
    signature = PRODUCT_NAME.replace(' ', '-')
    return signature.encode('utf-8')


def decode_url(url):
    """
    Return the Unicode representation for URL.
    """
    # FIXME:6201:
    # See how do to proper handling of this, in a py3 and py2 compatible way.
    # For now, this is just a quick hack.
    if six.PY3 and isinstance(url, six.binary_type):
        url = url.decode('utf-8')

    result = unquote(url)

    if isinstance(result, six.binary_type):
        result = result.decode('utf-8')

    return result


class URL(object):
    """
    Represents an URL with the following attributes:
    * protocol: https or http
    * host: full hostname and port
    * path: list with path elements
    """
    def __init__(self):
        self.protocol = None
        self.host = None
        self.path = ['']

    def parseFromRequest(self, request):
        """
        Fill url data from request.
        """
        self.host = request.getHeader(b'host')
        if not self.host:
            # Make sure we always have an Host header for consistency.
            # request.host is an IP address and it's always ASCII,
            # with bytes on Py2 and text on Py3.
            self.host = '{}:{}'.format(
                request.host.host,
                request.host.port,
                ).encode('ascii')

        if request.isSecure():
            self.protocol = b'https'
        else:
            self.protocol = b'http'
        self.path = request.uri[1:].split(b'/')

    def _decodeHost(self, encoded_host):
        parts = encoded_host.split(b':', 1)
        port = ''
        if len(parts) == 2:
            port = ':' + parts[1].decode('ascii')
        host = parts[0].decode('idna')
        return host + port

    def isSameSite(self, origin_url, accepted_origins=()):
        """
        Return True if `origin_url` (IDNA bytes) is from the same site
        or from the list of `accepted_origins`.
        """
        url = urlparse(origin_url.lower())
        origin_host = self._decodeHost(url.netloc).lower()
        our_host = self._decodeHost(self.host).lower()

        if origin_host in accepted_origins:
            # For hosts in the accepted origins, don't check the protocol.
            return True

        if self.protocol.lower() != url.scheme:
            return False

        if our_host != origin_host:
            return False

        # We have same protocol and site (including port)
        return True


@implementer(IChevahRequest)
class ChevahRequest(Request, SSLConnectionMixin):
    """
    Request factory which is used together with HTTPFactoryBase.

    `uri` is always in the raw format, as requested by the client.
    `prepath` is url-decoded, but not UTF-8 decoded.
    """
    # We block everything by default, and then allow specific types from
    # same origin.
    _CSP_POLICY = [
        b"default-src 'none'",
        b"img-src 'self' data:",
        b"connect-src 'self'",
        b"script-src 'self'",
        b"style-src 'self'",
        b"font-src 'self' data:",
        ]

    # FIXME:6502:
    # https://github.com/twisted/twisted/issues/6933
    # Check if this code from here was already imported upstream.
    server_signature = _get_server_signature()
    session_cookie_basename = 'CHEVAH_SESSION'
    session_cookie_force_secure = True

    maximumBodyLength = MAXIMUM_IN_MEMORY_RECEIVED_DATA
    _currentBodyLength = 0

    _dump = None

    @property
    def _public_uri(self):
        """
        The text representation of this request.
        This is private since this is not part of the Twisted API.
        For now we will use it as it is and later we can see if we can find
        a better API.
        """
        return decode_url(self.uri)

    @staticmethod
    def _getDateTimeToString(msSinceEpoch=None):
        """
        Called to return the current time as string for HTTP header.

        Here to help with testing.
        """
        return http.datetimeToString(msSinceEpoch=msSinceEpoch)

    def handleContentChunk(self, data):
        """
        Write a chunk of data.

        This method is not intended for users.
        """
        self._currentBodyLength += len(data)
        if (
            self.maximumBodyLength and
            self._currentBodyLength > self.maximumBodyLength
                ):
            self.loseConnection(http.REQUEST_ENTITY_TOO_LARGE)
        else:
            try:
                self.content.write(data)
            except Exception as error:
                self.loseConnection(
                    code=http.SERVICE_UNAVAILABLE,
                    details=force_unicode(error))

    def loseConnection(self, code, message=None, details=None):
        """
        Respond with `code` and force client disconnection.

        `message` is sent to the HTTP client.
        `details` is only sent to the event.
        """
        if not self.getHeader('content-type'):
            # Go with a default set of content-type.
            self.setHeader('content-type', 'text/html')

        self.setHeader('connection', 'close')
        if message:
            message_encoded = message.encode('utf-8')
        else:
            message_encoded = None
        self.setResponseCode(code, message_encoded)

        # Inform that request is done.
        self.finish()

        # Inform of disconnection before we call disconnection so that
        # we still have the avatar.
        size = self._currentBodyLength
        message = self.code_message.decode('utf-8')
        if details:
            message = '%s %s' % (message, details)
        data = {
            'code': code,
            'details': message,
            'uri': self._public_uri,
            'size': size,
            }
        self.emitEvent('40018', data=data)

        # Force client disconnection in any way, even for persistent
        # connections.
        # This is against HTTP protocol, since the client will only read
        # response status *after* it sends all its data, but we can no
        # longer process its request.
        #
        # Behaved clients should have sent Expect: 100-continue in the
        # request handlers and would be stopped from the start.
        #
        # We close with a very small delay to allow
        #
        # This goes for the low level transport for disconnection.
        if self.site._LOSE_CONNECTION_WITH_DELAY:
            self.site._scheduler.callLater(
                0.1, self.transport.loseConnection)
        else:
            self.transport.loseConnection()

        # Lose connection might be called as a side effect by a higher level
        # function, so we raise an error to prevent the following code to
        # execute.
        # Hypothetical example:
        # data = request.readContent() -> this might close the connection.
        # process_content(data)
        raise HTTPRequestRejectedException()

    def headersReceived(self, command, path, version):
        """
        Called by channel when all headers has been received.

        This should not fail as any failure is converted into ISE.

        Returns a deferred.
        """
        self.method = command
        self.clientproto = version
        self._setResourceIndentification(path)
        self._setDefaultHeaders(path)
        self.parseCookies()

        # Cache the client and server information, we'll need this later to be
        # serialized and sent with the request so CGIs will work remotely.
        self.client = self.getAvatarPeer()
        self.host = self.channel.transport.getHost()

        # Add shortcut for site.
        self.site = self.channel.factory

        # Initialize content with an in memory buffer.
        self.content = BytesIO()

        origin = self.getHeader(b'origin')
        referrer = self.getHeader(b'referer')
        if not self._isSameSite(origin, referrer):
            self.loseConnection(
                http.BAD_REQUEST,
                message='Possible CSRF',
                details='Got origin "%s" and referer "%s".' % (
                    force_unicode(origin), force_unicode(referrer)),
                )

        return self._handleEarlyHeaders()

    def _setResourceIndentification(self, path):
        """
        Called after all headers are received to update define
        resource identification members.
        """
        self.uri = path
        x = self.uri.split(b'?', 1)

        if len(x) == 1:
            self.path = self.uri
        else:
            self.path, argstring = x
            self.args = http.parse_qs(argstring, 1)

        self.prepath = []

        # FIXME:6201:
        # Get prepat and postpath alwasy as Unicode..
        if six.PY2:
            self.postpath = [
                unquote(p)
                for p in self.path[1:].split(b'/')
                ]
        else:
            self.postpath = [
                unquote(p.decode('utf-8')).encode('utf-8')
                for p in self.path[1:].split(b'/')
                ]

    def _isSameSite(self, origin, referrer):
        """
        Check standard headers to verify the request is from same source.
        This is a basic protection against CSRF / XSRF.

        Both `origin` and `referrer` are bytes (IDNA).
        If returns `False` when headers are present and are not valid.

        It does not help against spoofed headers, but it is assumed that
        during a normal CSRF the headers can't be spoofed.

        https://www.owasp.org/index.php/Cross-Site_Request_Forgery_(CSRF)_Prevention_Cheat_Sheet
        """
        url = URL()
        url.parseFromRequest(self)
        if origin is None or origin == b'null':
            # Fallback to referrer.
            origin = referrer

        if origin is None:
            # This looks like a direct API or browser request.
            # No Origin and Referrer.
            return True

        if origin.startswith(b'https://login.microsoftonline.com'):
            # FIXME:5620:
            # This is set via `referrer` as `origin` is missing.
            # For now we allow redirection from MS, for Entra ID login.
            # See how we can trust this in a general way.
            # We might need to add an allow list.
            # The login requests will always originate from outside of our site.
            return True

        return url.isSameSite(
            origin_url=origin, accepted_origins=self.site.accepted_origins)

    def _setDefaultHeaders(self, path):
        """
        Called to set general headers to all responses.
        """
        self.setHeader(b'server', self.server_signature)
        self.setHeader(b'date', self._getDateTimeToString())

        csp_policy = b'; '.join(self._CSP_POLICY)
        self.setHeader(b'content-security-policy', csp_policy)
        self.setHeader(b'x-content-security-policy', csp_policy)
        # Extra headers for example for IE (where CSP is not present)
        self.setHeader(b'x-frame-options', b'DENY')
        self.setHeader(b'x-xss-protection', b'1; mode=block')
        self.setHeader(b'x-content-type-options', b'nosniff')

        # See https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/
        #     Referrer-Policy
        self.setHeader(b'referrer-policy', b'same-origin')

    def _handleEarlyHeaders(self):
        """
        Handle response from resources who have received headers before
        final content parsing.
        """
        def split_response(result):
            """
            Return a tuple with (CODE, MESSAGE)

            Result can be either single HTTP response code (in which case
            message is none) or explicit tuple of (CODE, MESSAGE).
            """
            try:
                code, message = result
            except (ValueError, TypeError):
                code = result
                message = None

            return (code, message)

        def cb_got_resource(result):
            """
            Called when we traversed the URL and found the final resource.
            """
            self._resource = result
            return self._resource.headersReceived(self)

        def cb_headers_received(result):
            """
            Called after the resource has processed the headers.
            """
            code, message = split_response(result)
            self._processHeadersReceived(code, message)

        deferred = self.site.getResourceFor(self)
        deferred.addCallback(cb_got_resource)
        deferred.addCallback(cb_headers_received)
        return deferred

    def _processHeadersReceived(self, result, message=None):
        """
        Continue request process based on response from
        resource.headersReceived()
        """
        message_encoded = message
        if message_encoded and isinstance(message_encoded, six.text_type):
            message_encoded = message_encoded.encode('utf-8')
        # Handle HTTP/1.1 'Expect: 100-continue' based on RFC 2616 8.2.3.
        expectContinue = self.requestHeaders.getRawHeaders('expect')
        if expectContinue:
            # This is a behaved client so we don't force client
            # disconnection.
            if expectContinue[0].lower() != '100-continue':
                # Somehow the expect header is not right, so return
                # a failed response.
                self.setResponseCode(
                    http.EXPECTATION_FAILED, message=message_encoded)
                self.setHeader('content-length', '0')
                self.finish()
                return

            if result == _CONTINUE:
                self.transport.write(b"HTTP/1.1 100 Continue\r\n\r\n")
                return

            # Let client know that it is expected not to continue.
            # But don't disconnect it.
            self.setResponseCode(result, message=message_encoded)
            self.setHeader('content-length', '0')
            self.finish()
            return

        # Not HTTP/1.1 or not expect requested.
        if result == _CONTINUE:
            # Just continue as normal.
            return

        if self.method == b'GET':
            # Here we might have a browser, so we try to display
            # a minimal message.
            self.setResponseCode(result, message=message_encoded)
            # Need to set text/html so that the browser will show the
            # page and not try to download it.
            self.setHeader(b'content-type', b'text/html')
            self.setHeader(b'connection', b'close')
            # FIXME:6188:
            # This is not valid HTML, but we try to keep it minimal.
            data = ('%s - %s' % (result, message)).encode('utf-8')
            # And set content length not to trigger chunk encoding.
            self.setHeader('content-length', str(len(data)))
            self.write(data)

        # Close the connection as we don't want more data from the client.
        self.loseConnection(result, message=message)

    def process(self):
        """
        Called when all content was received.
        """
        try:
            self.render(self._resource)
        except Exception:
            self.processingFailed(failure.Failure())

    def _setSession(self, session):
        """
        Patch upstream to allow easy setting of a session.
        """
        # FIXME:5178:
        # https://twistedmatrix.com/trac/ticket/10192
        # Maybe even allow custom session names.
        # There might be a PR in Twisted for custom session names.
        if self.isSecure():
            self._secureSession = session
        else:
            self._insecureSession = session

    def getSession(self, sessionInterface=None):
        """
        Return the session for this request.

        If no session exists, a new session is

        This is the exact copy from Twisted.Web, but use a custom session
        cookie name and check for secured session cookies.
        """
        # Update or delete any expired session.
        self.touchSession()

        # Session management
        if not self.session:
            # FIXME:6502:
            # Send this upstream.
            # https://github.com/twisted/twisted/issues/6933
            cookiename = self.session_cookie_basename.encode('ascii')
            sessionCookie = self.getCookie(cookiename)
            if sessionCookie:
                existing_session = self.getSessionByID(sessionCookie)
                self._setSession(existing_session)
                self.touchSession()

            # If it still hasn't been set, create a new session.
            if not self.session:
                self._setSession(self.site.makeSession())
                self.addCookie(
                    cookiename,
                    self.session.uid,
                    path='/',
                    secure=self.session_cookie_force_secure,
                    # Session cookies are always HTTP only and same site.
                    httpOnly=True,
                    sameSite='Strict',
                    )

        if sessionInterface:
            return self.session.getComponent(sessionInterface)

        return self.session

    def touchSession(self):
        """
        Update the session access expire time.

        If session is already expired, it is removed.
        """
        if not self.session:
            return

        # Update the session access time.
        # This is already fixed upstream but we still need getSession
        # for custom session name
        # https://github.com/twisted/twisted/issues/9288
        try:
            # Refresh the session, to keep it alive.
            self.session.touch()
        except (AlreadyCalled, AlreadyCancelled):
            # Session has already expired.
            self.removeSession()

    def getAvatarPeer(self):
        """
        Return the session avatar associated to this request.
        """
        return self.transport.wrappedProtocol.avatar.peer

    def switchConsumer(self, consumer):
        """
        Set the consumer for the body received by the request.

        Most probably you will want to call it from inside
        Resource.headersReceived().

        This should only be called once for a request.
        """
        previous_content = self.content
        previous_content.seek(0)
        previous_data = previous_content.read()
        previous_content.close()

        self.content = consumer
        self.content.registerProducer(self.transport)
        if previous_data:
            consumer.write(previous_data)

    def emitEvent(self, event_id, data):
        """
        Emit an event associated with this request.
        """
        return self.site.emitEvent(
            event_id,
            data=data,
            avatar=self.transport.wrappedProtocol.avatar,
            )

    def write(self, data):
        """
        Copied from upstream Twisted to allow debug.

        Write some data as a result of an HTTP request.  The first
        time this is called, it writes out response data.

        @type data: C{bytes}
        @param data: Some bytes to be sent as part of the response body.
        """
        if self.finished:
            raise RuntimeError('Request.write called on a request after '
                               'Request.finish was called.')
        if not self.startedWriting:
            self.startedWriting = 1
            version = self.clientproto
            header = []
            header.append(
                version + b" " +
                intToBytes(self.code) + b" " +
                self.code_message + b"\r\n")

            # if we don't have a content length, we send data in
            # chunked mode, so that we can support pipelining in
            # persistent connections.
            if (
                (version == b"HTTP/1.1") and
                (self.responseHeaders.getRawHeaders(b'content-length') is None)
                and self.method != b"HEAD"
                and self.code not in http.NO_BODY_CODES
                    ):
                header.append(b'Transfer-Encoding: chunked\r\n')
                self.chunked = 1

            # Before doing the first write, check to see if a default
            # Content-Type header should be supplied.
            modified = self.code != http.NOT_MODIFIED
            contentType = self.responseHeaders.getRawHeaders(b'content-type')
            if (
                modified and
                contentType is None and
                self.defaultContentType is not None
                    ):
                self.responseHeaders.setRawHeaders(
                    b'content-type', [self.defaultContentType])

            if self.lastModified is not None:
                if not self.responseHeaders.hasHeader(b'last-modified'):
                    self.responseHeaders.setRawHeaders(
                        b'last-modified',
                        [self._getDateTimeToString(self.lastModified)])

            if self.etag is not None:
                self.responseHeaders.setRawHeaders(b'ETag', [self.etag])

            for name, values in self.responseHeaders.getAllRawHeaders():
                for value in values:
                    header.append(name + b": " + value + b"\r\n")

            for cookie in self.cookies:
                header.append(b'Set-Cookie: %s\r\n' % (cookie,))

            header.append(b"\r\n")

            if self.site.service_configuration.parent.debug:
                self._dump = open('http.response.dump', 'wb')

            for line in header:
                if self._dump:
                    self._dump.write(line)
                self.channel.debug('> ' + force_unicode(repr(line)))

            self.transport.writeSequence(header)

            if self._inFakeHead:
                return

            # if this is a "HEAD" request, we shouldn't return any data
            if self.method == b"HEAD":
                self.write = lambda data: None
                self.channel.debug('> No response body for HEAD.')
                return

            # for certain result codes, we should never return any data
            if self.code in http.NO_BODY_CODES:
                self.channel.debug('> No response body for %s.' % (self.code))
                self.write = lambda data: None
                return

            self.channel.debug(
                '> Only first 1500 bytes of body included in debug.\n%r' % (
                    data[:1500]))

        self.sentLength = self.sentLength + len(data)
        if data:
            if self.chunked:
                self.transport.writeSequence(http.toChunk(data))
            else:
                if self._dump:
                    self._dump.write(data)
                self.transport.write(data)

    def finish(self):
        """
        Indicate that all response data has been written to this L{Request}.
        """
        if self._dump:
            self._dump.close()
        return super(ChevahRequest, self).finish()

    def removeSession(self):
        """
        Make sure the request has no session.

        FIXME:5178:
        https://twistedmatrix.com/trac/ticket/10193
        See if we should send the patch upstream.
        """
        if not self.session:
            return

        try:
            self.session.expire()
        except KeyError:
            """
            Already expired.
            """

        # Remove the previous cookie.
        # Will be added next response for the browser to expire it
        # and it this way it will be removed from client side.
        result = []
        for cookie in self.cookies:
            if cookie.startswith(self.session_cookie_basename.encode('ascii')):
                continue
            result.append(cookie)
        self.cookies = result

        self.addCookie(
            self.session_cookie_basename,
            'expired',
            expires='Thu, 01 Jan 1970 00:00:00 GMT',
            path='/',
            secure=self.session_cookie_force_secure,
            # Session cookies are always HTTP only and same site.
            httpOnly=True,
            sameSite='Strict',
            )
        self._setSession(None)

    def getExistingSession(self):
        """
        Return the current session, without creating a new one.

        Return None if session does not exist.
        """
        if self.session:
            # Only return existing session if it is still available on the
            # site. Ex, it has not expired.
            return self.getSessionByID(self.session.uid)

        cookie_name = b"_".join(
            [self.session_cookie_basename.encode('ascii')] + self.sitepath)
        session_id = self.getCookie(cookie_name)

        if not session_id:
            return None

        return self.getSessionByID(session_id)

    def getSessionByID(self, id):
        """
        Return an existing session, or None.
        """
        if not id:
            return None

        try:
            return self.site.getSession(id)
        except KeyError:
            # Session not found.
            pass

        return None

    def setDisableCacheHeaders(self):
        """
        Set headers to inform client not to cache the request.

        See: http://stackoverflow.com/q/49547
        """
        # For HTTP/1.0.
        self.setHeader(
            b'cache-control',
            (b'private, no-cache, no-store, max-age=0, s-maxage=0, '
                b'must-revalidate, proxy-revalidate'),
            )
        # For HTTP/1.0 cache control.
        self.setHeader(b'expires', b'Fri, 01 Jan 1990 00:00:00 GMT')
        self.setHeader(b'pragma', b'no-cache')
        self.setHeader(b'vary', b'*')

    def getContentType(self):
        """
        Return a tuple of (type, options) for the content type of
        current request.

        Type and option keys are lower case.
        """
        header = self.getHeader('content-type')
        if not header:
            return ('', {})

        kind, options = parse_options_header(header)
        options_lower = {}
        for key, value in options.items():
            options_lower[key.lower()] = value

        return (kind.lower(), options_lower)

    def getAccept(self):
        """
        Return an Accept class for the accept header of current request.
        """
        header = self.getHeader('accept')
        return parse_accept_header(header)

    def getContentBoundary(self):
        """
        Return boundary from multipart/form-data content type header.

        Return None if boundary could not be found.

        Return None if boundary could not be found.
        """
        kind, options = self.getContentType()

        if not kind or not kind.startswith('multipart/'):
            return None

        return options.get('boundary', None)

    def getTextArgument(self, name):
        """
        Return the value for a simple text POST field.

        Return None if field is not defined or is an invalid text field.
        """
        value = self.getBinaryArgument(name)
        if not value:
            return None

        try:
            return value.decode('utf-8')
        except UnicodeError:
            pass

        return None

    def getBinaryArgument(self, name):
        """
        Return the bytes value for a simple POST field in binary format.

        Return None if field is not defined.
        """
        if not self.args:
            return None

        value = self.args.get(name.encode('utf-8'), None)
        if not value:
            return None
        return value[0]

    def processingFailed(self, failure):
        """
        Called when request handling failed.
        """
        if failure.check(HTTPRequestRejectedException):
            return self.loseConnection(code=http.EXPECTATION_FAILED)

        public_uri = decode_url(self.uri)
        data = {
            'uri': public_uri,
            'details': force_unicode(failure.value),
            'tb': failure.getBriefTraceback(),
            }
        self.emitEvent('40023', data=data)

        if self.finished:
            return

        body = (
            b'<html><head><title>Processing Failed</title></head><body>'
            b'<b>Processing Failed</b></body></html>')
        self.setResponseCode(http.INTERNAL_SERVER_ERROR)
        self.setHeader('content-type', 'text/html')
        self.setHeader('content-length', str(len(body)))
        self.write(body)
        self.finish()
