# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
General resources for HTTP service.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

import traceback

from twisted.cred import error as cred_error
from twisted.cred.portal import Portal
from twisted.internet import defer
from twisted.python import filepath
from twisted.web import http, static
from twisted.web.resource import (
    ErrorPage,
    IResource,
    Resource,
    _computeAllowedMethods,
    )
from twisted.web.server import NOT_DONE_YET
from zope.interface import implementer
from six.moves.urllib.parse import quote, unquote, quote_plus
from six.moves.urllib.parse import urlsplit, urlunsplit
import six

from chevah_compat import LocalFilesystem
from chevah.server import emit, emitFromError, force_unicode, json
from chevah.server import native_string
from chevah.server.authentication.credential_checker import (
    ServiceCredentialChecker,
    )
from chevah.server.commons.constant import TYPE_NAME, PRODUCT_NAME
from chevah.server.commons.credential import HTTPSSSLCertificateCredentials
from chevah.server.commons.exception import (
    InsufficientAuthentication,
    ServerError,
    )
from chevah.server.commons.interface import IChevahResource
from chevah.server.commons.i18n import NULL_TRANSLATION
from chevah.server.commons.jinja import ViewModel
from chevah.server.commons.realm import ServiceRealm
from chevah.server.http.configuration import get_ui_version
from chevah.server.http.credential import (
    HTTPBasicAuthCredentialFactory,
    SessionAuthCredentialFactory,
    )
from chevah.server.http.protocol import _CONTINUE

# Twisted resource defined the HEAD as GET, but we want explicit definition
# of HEAD
delattr(Resource, 'render_HEAD')


@implementer(IChevahResource)
class ChevahResource(Resource, object):
    """
    Extended resource used in Chevah project.

    No UI should be defined here.

    If the resource is state-full you will need to implement the `reset`
    method which is called between multiple renders of the resource.
    """

    def getChild(self, name, request):
        """
        Classes inheriting from this should implement own child
        traversal.
        """
        return _NoChild()

    def headersReceived(self, request):
        """
        See: `IChevahResource`.

        Always continues.
        """
        return _CONTINUE

    def reset(self):
        """
        Called when the resource is already authenticated and it needs to
        serve a new request.

        FIXME:5513:
        This should be removed once we refactor not to reuse a resource
        after render.

        By default this does nothing as we should have stateless resources.
        """

    def _isJSONRequest(self, request):
        """
        Return True if response for `request` should be in JSON format.
        """
        # We default to HTML, only when content type or accept is JSON.
        content_type, _ = request.getContentType()
        if content_type == TYPE_NAME.APPLICATION_JSON:
            return True

        accept = request.getAccept()
        if accept.best_match([TYPE_NAME.APPLICATION_JSON]):
            return True

        return False

    def render_OPTIONS(self, request):
        """
        Respond to an OPTION request.

        See: https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/OPTIONS
        """
        try:
            methods = self.allowedMethods
        except AttributeError:
            methods = _computeAllowedMethods(self)
            methods.remove(b'OPTIONS')

        request.responseHeaders.setRawHeaders(
            b'Allow', [b' '.join(methods)])
        # Both 200 OK and 204 No Content are permitted status codes,
        # but some browsers incorrectly believe 204 No Content
        # applies to the resource and do not send the subsequent request
        # to fetch it.
        request.setResponseCode(http.OK)
        return b''


class StaticTextResource(ChevahResource):
    """
    A resource returning a static page.
    """

    def __init__(self, content='', content_type='text/plain', code=http.OK):
        self._content = content.encode('utf-8')
        self._content_type = content_type.encode('ascii')
        self._code = code
        super(StaticTextResource, self).__init__()

    def getChild(self, name, request):
        """
        Called when no other resources are attached.
        """
        if name:
            return _NoChild()
        else:
            return self

    def render(self, request):
        """
        Return the same content.
        """
        request.setHeader(b'Content-Type', self._content_type)
        request.setResponseCode(self._code)
        return self._content


class FaviconResource(ChevahResource):
    """
    A resource returning an empty favicon.
    """
    isLeaf = True

    _data = (
        b'\x00\x00\x01\x00\x01\x00\x10\x10\x02\x00\x01\x00\x01\x00\xb0\x00\x00'
        b'\x00\x16\x00\x00\x00(\x00\x00\x00\x10\x00\x00\x00 \x00\x00\x00\x01'
        b'\x00\x01\x00\x00\x00\x00\x00\x80\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xff\xff'
        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xff\x00'
        b'\x00\xff\xff\x00\x00\xff\xff\x00\x00\xff\xff\x00\x00\xff\xff\x00\x00'
        b'\xff\xff\x00\x00\xff\xff\x00\x00\xff\xff\x00\x00\xff\xff\x00\x00\xff'
        b'\xff\x00\x00\xff\xff\x00\x00\xff\xff\x00\x00\xff\xff\x00\x00\xff\xff'
        b'\x00\x00\xff\xff\x00\x00\xff\xff\x00\x00'
        )

    def render(self, request):
        """
        Return nothing.
        """
        request.setHeader('Content-Type', 'image/x-icon')
        return self._data


# FIXME:6201:
# This whole function is a mess and we should stop reinventing the wheel
# and just use the hyperlink library.
def encode_url(url):
    """
    Convert an Unicode (or Unicode encoded) url into an encoded bytes URL.
    """
    # turn string into unicode
    if not isinstance(url, six.text_type):
        url = url.decode('utf-8')

    # parse it
    parsed = urlsplit(url)

    # divide the netloc further
    userpass, at, hostport = parsed.netloc.partition('@')
    user, colon1, pass_ = userpass.partition(':')
    host, colon2, port = hostport.partition(':')

    scheme = parsed.scheme
    user = quote(user)
    pass_ = quote(pass_)
    path = quote_plus(parsed.path.encode('utf-8'), safe=b'/')
    query = quote(unquote(parsed.query).encode('utf-8'), b'=&?/')
    fragment = quote(unquote(parsed.fragment))

    # put it back together
    netloc = ''.join((user, colon1, pass_, at, host, colon2, port))
    return urlunsplit((scheme, netloc, path, query, fragment)).encode('utf-8')


class RedirectResource(ChevahResource):
    """
    Redirects to an URL with minimal UI.

    # There is no UI version.
    # Is should be OK as this should be a page without any UI.

    This is implemented similar to Amazon S3 and I hope other server.
    https://docs.aws.amazon.com/AmazonS3/latest/dev/RESTRedirect.html

    By default we respond with TEMPORARY_REDIRECT so that the redirection
    is done using the same method.

    We don't respond by default with 302 FOUND for example for a POST request,
    the redirect will be done as GET.
    https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/302
    """
    # Is not a leaf to allow traversal.
    isLeaf = False

    def __init__(self, url, code=http.TEMPORARY_REDIRECT):
        super(RedirectResource, self).__init__()
        # FIXME:6201:
        # See when this is called with bytes and see if we can refactor
        # to always call it with Unicode.
        # URL can be an already encoded URL passed as bytes.
        # URL can also be an Unicode, and it is encoded here.
        if isinstance(url, six.text_type):
            url = encode_url(url)
        self._url = url.decode('ascii')
        self._code = code

    @property
    def url(self):
        """
        URL there this resource will redirect.
        """
        return self._url

    def headersReceived(self, request):
        """
        See: `IChevahResource`.
        """
        self._setHeaders(request)
        return self._code

    def getChild(self, name, request):
        """
        We don't do any traversal.
        """
        return self

    def render(self, request):
        """
        Called to get the body for the redirection.
        """
        self._setHeaders(request)
        request.setResponseCode(self._code)
        request.setHeader(b"Content-Type", b"text/html; charset=utf-8")
        content = """
    <html>
        <head>
            <meta http-equiv=\"refresh\" content=\"0;URL=%(url)s\">
        </head>
        <body bgcolor=\"#FFFFFF\" text=\"#000000\">
        %(product_name)s is redirecting your request.
        <a href=\"%(url)s\">To continue, click here
            if you are not automatically redirected.</a>
        </body>
    </html>
    """ % {
            'url': self._url,
            'product_name': PRODUCT_NAME,
            }
        return content.encode('utf-8')

    def _setHeaders(self, request):
        """
        Set default headers.
        """
        request.setHeader(b"location", self._url)


class GenericViewModel(ViewModel):
    """
    View model for generic page.
    """
    def __init__(self, title, message, icon='far fa-file', button=None):
        self.icon_class = icon
        self.title = title
        self.heading = title
        self.message = message
        # When we have a button, we pack it for easy access.
        if button:
            button = self.pack(button)
        self.button = button


class GenericResource(ChevahResource):
    """
    A simple page which can be use for any purpose.

    It supports returning the result in HTML and JSON.

    The HTML is rendered using a Jinja `template` and a ViewModel `page`.
    The JSON is rendered using the `json_body` dict.
    """

    isLeaf = True

    def __init__(self, code, title, message, ui_version=None):
        self.code = code

        if ui_version is None:
            ui_version = TYPE_NAME.UI_GEN_2

        template_path = 'generic-gen2.html.j2'
        if ui_version == TYPE_NAME.UI_GEN_1:
            template_path = 'generic-gen1.html.j2'

        # FIXME:5490:
        # Here we need to get the configured translation.
        self.template = NULL_TRANSLATION.get_template(template_path)
        self.page = GenericViewModel(title=title, message=message)
        self.json_body = {
            'results': [{
                'title': title,
                'message': message,
                }],
            }

    def render(self, request):
        """
        See: `ChevahResource`.
        """
        request.setResponseCode(self.code)

        if request.method == b'HEAD':
            # No need for content for HEAD.
            # And for now we don't set any extra headers, so no need to go
            # further.
            # Content-type header doesn't make sense, as we have no content.
            return b''

        if self._isJSONRequest(request):
            # We have a JSON request.
            request.setHeader(b'content-type', b'application/json')
            return json.dumps(self.json_body).encode('utf-8')
        else:
            # Respond by default with HTML.
            request.setHeader(b'content-type', b'text/html; charset=utf-8')
            return self.template.render(self.page).encode('utf-8')

    def getChild(self, child_name, request):
        return self

    @property
    def children(self):
        return {}


class ResourceError(GenericResource):
    """
    Resource for an error page.

    It will trigger an event.
    """

    def __init__(self, code, title, message, data=None):
        super(ResourceError, self).__init__(code, title, message)
        self.page.icon_class = 'fas fa-exclamation-circle'
        if data is None:
            data = {}
        self.data = data
        # For errors we rename results into errors.
        self.json_body = {'errors': self.json_body['results']}
        self.__emitted = False

    def headersReceived(self, request):
        """
        See: `IChevahResource`.
        """
        # No point to continue as we have an error.
        self._setHeaders(request)
        self.emitEventFor(request)
        return self.code

    def render(self, request):
        """
        We should not have this rendered as it should failed when headers
        are retrieved.
        """
        raise RuntimeError('Direct rendering should not be called.')

    def renderEmbedded(self, request):
        """
        Error resources are not normally rendered as they will trigger
        a disconnection.
        But sometimes a resource is renders as another one.
        """
        self.emitEventFor(request)
        return super(ResourceError, self).render(request)

    def _setHeaders(self, request):
        """
        Called when we need to set the default headers.
        """

    def emitEventFor(self, request):
        """
        Emit an event for this error.
        """
        raise NotImplementedError(
            'This should be implemented by new classes derived '
            'from ResourceError.'
            )


class _NoChild(ResourceError):
    """
    Resource returned as a placeholder when traversal is requested for a
    child which does not implement own traversal.

    Should only be used by base resource class.

    This is an 404 page.
    """
    def __init__(self):
        title = 'No such resource.'
        message = 'The accessed URL is not associated with any resource.'
        super(_NoChild, self).__init__(
            code=http.NOT_FOUND, title=title, message=message, data=None)

    def emitEventFor(self, request):
        """
        Does not emit an event.
        """
        pass


class StaticResourceNotFound(ResourceError):
    """
    Resource returned as a placeholder for a resource that was not found.

    This is the 404 page.
    """
    def __init__(self, title=None, message=None, data=None):
        if title is None:
            title = "Page not found."
        if message is None:
            message = "The accessed page does not exist."
        super(StaticResourceNotFound, self).__init__(
            code=http.NOT_FOUND, title=title, message=message, data=data)

    def emitEventFor(self, request):
        emit(
            '40001',
            peer=request.client,
            creator=request.site,
            data={'uri': request._public_uri},
            )


class StaticForbiddenResource(ResourceError):
    """
    Resource returned as a placeholder for a resource for which access is
    forbidden.

    This is the 403 page.
    """

    def __init__(self, data=None):
        title = 'Permission denied.'
        message = (
            'You just don\'t have permission to access this resource. '
            'If you think you should have access to this resource, '
            'kindly ask the system administrator for permission. '
            'Don\'t try again until your predicament changes.'
            )
        super(StaticForbiddenResource, self).__init__(
            code=http.FORBIDDEN, title=title, message=message, data=data)

    def emitEventFor(self, request):
        """
        We don't log events.
        """
        pass


class InternalErrorResource(ResourceError):
    """
    Resource returned as a placeholder for an internal server error.

    This is the 500 page.
    It will always has the traceback which is only sent to the server-side and
    not visible to the client.

    Client will only see the message.
    """
    def __init__(self, details, title=None, message=None):
        if title is None:
            title = 'Internal Server Error.'
        if message is None:
            message = 'An error occurred while processing the request.'
        super(InternalErrorResource, self).__init__(
            code=http.INTERNAL_SERVER_ERROR,
            title=title,
            message=message,
            )
        self._details = details
        self._tb = force_unicode(traceback.format_exc())

    def emitEventFor(self, request):
        data = {
            'uri': request._public_uri,
            'title': self.page.title,
            'details': self._details,
            'tb': self._tb,
            }
        emit('40003', peer=request.client, data=data, creator=request.site)


# Marker to let the resource know that it should not advertise the
# availability of a method.
HIDE_CHALLENGE = object()


class UnauthorizedResource(ResourceError):
    """
    A resource returned for requests which are not authorized.

    It contains information both for end users and for server-side logs/audit.

    `response` is an dict with details from the login rejection response.
    """
    isLeaf = True

    def __init__(
            self, credential_factories, details, translation, response):
        translation.install(self)

        self._credential_factories = credential_factories
        if not details:
            details = self._(
                'Authentication failed. '
                'Contact your administrator for more details.'
                )
        self._details = details
        if not response:
            response = {}
        self._response = response

        super(UnauthorizedResource, self).__init__(
            code=http.UNAUTHORIZED,
            title=self._('Unauthorized'),
            message=self._(
                'You are not authenticated. '
                'Please try again using valid credentials or '
                'contact the system administrator for more details.'
                ),
            )

    @property
    def details(self):
        """
        Low level information about the authentication failure.
        """
        return self._details

    def getChildWithDefault(self, path, request):
        """
        No traversal from this resource.
        This should not be called, as we are a leaf.
        """
        return self

    def emitEventFor(self, request):
        """
        Render the response page and emit an event.
        """
        data = {
            'uri': request._public_uri,
            'details': self._details,
            }
        emit('40000', peer=request.client, data=data, creator=request.site)

    def _setHeaders(self, request):
        """
        Set the www-authenticate.
        """
        def generateWWWAuthenticate(scheme, challenge):
            """
            `scheme` is bytes while `challenge` is a dict of text.
            """
            scheme = scheme.decode('ascii')
            methods = []
            for key, value in challenge.items():
                header = "%s=%s" % (key, quoteString(value))
                methods.append(header)
            return " ".join([scheme, ", ".join(methods)])

        def quoteString(s):
            return '"%s"' % (s.replace('\\', '\\\\').replace('"', '\\"'),)

        auth_headers = []
        for factory in self._credential_factories:
            challenge = factory.getChallenge(request)
            if challenge is HIDE_CHALLENGE:
                continue
            auth_headers.append(
                generateWWWAuthenticate(factory.scheme, challenge))
        # We set multiple headers.
        request.responseHeaders.setRawHeaders(
            'www-authenticate', auth_headers)


class StaticFilesystemResource(static.File):
    """
    Root resource for HTTP service.

    It handles static files access and by default hides directory listing.
    """

    childNotFound = StaticResourceNotFound()
    forbidden = StaticForbiddenResource()

    def __init__(
            self, path, defaultType=None, *args, **kwargs):
        # Patch path to still have the Unicode representation since
        # FilePath does not support unicode.
        if defaultType is None:
            # This can be removed once we no longer need py2.
            defaultType = native_string('application/octet')
        self._path = path
        path = LocalFilesystem.getEncodedPath(path)
        super(StaticFilesystemResource, self).__init__(
            path, defaultType, *args, **kwargs)
        # See: https://developer.mozilla.org/en-US/docs/
        #       Web/HTTP/Basics_of_HTTP/MIME_types/Complete_list_of_MIME_types
        # On Twisted py2 this is bytes, while on PY3 this is text.
        self.contentTypes[native_string('.js')] = native_string(
            'application/javascript')
        self.contentTypes[native_string('.json')] = native_string(
            'application/json')
        self.contentTypes[native_string('.csv')] = native_string(
            'text/csv')
        self.contentTypes[native_string('.ico')] = native_string(
            'image/x-icon')

    def __repr__(self):
        return 'StaticFilesystemResource(%r)' % (self.path,)

    def headersReceived(self, request):
        """
        See: `IChevahResource`.
        """
        return _CONTINUE

    def directoryListing(self):
        """
        Directory listing is hidden.
        """
        return StaticTextResource('')

    def getChildWithDefault(self, path, request):
        """
        Return the child resource.

        This is called by the twisted.web framework and path is not Unicode.
        """
        if path in self.children:
            return self.children[path]
        else:
            return self.getChild(path, request)

    def getChild(self, path, request):
        """
        If this L{File}'s path refers to a directory, return a L{File}
        referring to the file named C{path} in that directory.

        If C{path} is the empty string, return a L{DirectoryLister} instead.

        The main HTTP service getResourceFor(request) will continue to call
        getChild until there is path segment.

        For a folder the path segments looks like ['rest', 'folder_name', '']
        For a file the path segments looks like ['rest', 'file_name']

        If a ['rest', 'folder_name'] request is made, it will be redirected
        to ['rest', 'folder_name', ''].
        """
        self.restat(reraise=False)

        if not self.isdir():
            # If this is not a directory, then it is file and files
            # have no children.
            return self.childNotFound

        if path:
            try:
                child_path = self.child(path)
            except filepath.InsecurePath:
                return self.childNotFound
        else:
            # No more children in request, so we can just look for default
            # index or list the current directory.
            child_path = self.childSearchPreauth(*self.indexNames)
            if child_path is None:
                return self.directoryListing()

        if not child_path.exists():
            child_path = child_path.siblingExtensionSearch(*self.ignoredExts)
            if child_path is None:
                return self.childNotFound

        # FIXME:5551:
        # This should be removed once we have a pure Unicode
        # filesystem handling.
        new_path = child_path.path
        if not isinstance(new_path, six.text_type):
            new_path = new_path.decode('utf-8')

        return self.createSimilarFile(new_path)

    def render_GET(self, request):
        """
        Begin sending the contents of this L{File} (or a subset of the
        contents, based on the 'range' header) to the given request.

        This is a copy from Twisted code to have a custom ForbiddenResource
        and to catch both OSError and IOError.
        """
        self.restat(False)

        if self.type is None:
            self.type, self.encoding = static.getTypeAndEncoding(
                self.basename(),
                self.contentTypes,
                self.contentEncodings,
                self.defaultType,
                )

        if not self.exists():
            return self.childNotFound.renderEmbedded(request)

        if self.isdir():
            return self.redirect(request)

        request.setHeader('accept-ranges', 'bytes')

        try:
            fileForReading = self.openForReading()
        except (IOError, OSError) as e:
            import errno
            if e.args[0] == errno.EACCES:
                return self.forbidden.renderEmbedded(request)
            else:
                raise

        if (
            request.method == b'HEAD' or
            request.setLastModified(
                self.getModificationTime()) is http.CACHED
                ):
            # File is auto-closed when is garbage collected,
            # but is better to be explicit in case we will return a wrapper.
            fileForReading.close()
            return b''

        producer = self.makeProducer(request, fileForReading)

        producer.start()
        # and make sure the connection doesn't get closed
        return NOT_DONE_YET
    # Head is implemented just as GET.
    render_HEAD = render_GET


# This is what is patched.
# We keep a reference since we want to keep the behaviour,
# We only want to insert an event emitter.
_upstream_ErrorPage_render = ErrorPage.render


def _event_ErrorPage_render(self, request):
    """
    Patch upstream Twisted code to emit an event on error.
    """
    details = '{}: {}'.format(
        force_unicode(self.code), force_unicode(self.brief))
    emit(
        '40057',
        peer=request.client,
        creator=request.site,
        data={
            'uri': request._public_uri,
            'details': details,
            'tb': '\n'.join(traceback.format_stack()),
            },
        )
    return _upstream_ErrorPage_render(self, request)


ErrorPage.render = _event_ErrorPage_render


def clone_StaticFilesystemResource(self, path):
    """
    Path FilePath.clonePath to return our objects which handles Unicode.
    """
    if not isinstance(path, six.text_type):
        path = path.decode('utf-8')
    return StaticFilesystemResource(path)


StaticFilesystemResource.clonePath = clone_StaticFilesystemResource


def _get_contants_js(base_url_path):
    """
    Return the full content for the constants.js file, including the
    JSON serialization for `events`.

    This is here to help generate a constants file close to a real one for the
    JS unit tests.
    """
    # Delay the import to try to troubleshot flaky test.
    from chevah.server.commons import constant
    content = {}
    for key in dir(constant):
        if key.startswith('_'):
            continue
        content[key] = getattr(constant, key)

    content['BASE_URL_PATH'] = base_url_path

    json_string = json.dumps(content)

    node_support = """
if (typeof exports !== 'undefined') {
    // We are in the nodejs.
    exports.CONSTANTS = CONSTANTS
}
"""

    return 'CONSTANTS = %s\n%s' % (json_string, node_support)


class ConstantsJSResource(ChevahResource):
    """
    A simple resource that returns the Python constants as JS file.
    """

    isLeaf = True
    CONTENT_TYPE = 'application/javascript'

    def __init__(self, factory):
        super(ConstantsJSResource, self).__init__()
        self._factory = factory
        self._file_content = _get_contants_js(factory.base_url_path)

    def render_GET(self, request):
        """
        Render the constants as a valid JavaScript file.
        """
        request.setHeader('content-type', self.CONTENT_TYPE)
        return self._file_content.encode('utf-8')


class RootResource(ChevahResource):
    """
    A resource which return a `root_resource` for '/' request or
    delegates to 'other_resource' for all other requests.
    """

    def __init__(self, root_resource, other_resource):
        self._root_resource = root_resource
        self._other_resource = other_resource

    def getChildWithDefault(self, name, request):
        """
        For root we return `root_resource` and delegate traversal to
        `other_resource`.
        """
        if name == b'':
            return self._root_resource
        else:
            return self._other_resource.getChildWithDefault(name, request)

    def putChild(self, name, resource):
        """
        Delegate to `other_resource`.
        """
        return self._other_resource.putChild(name, resource)


class JinjaResource(ChevahResource):
    """
    A simple resource which renders a Jinja2 template.
    """

    isLeaf = True
    CONTENT_TYPE = 'text/html'

    def __init__(self, environment, template, model):
        super(JinjaResource, self).__init__()
        self._environment = environment
        self._template_path = template
        self._model = model

    def render_GET(self, request):
        """
        Return the string serialization of template in blocking mode.
        """
        request.setHeader('content-type', self.CONTENT_TYPE)
        template = self._environment.get_template(self._template_path)
        return template.render(self._model).encode('utf-8')


class AuthenticatedResourceWrapper(ChevahResource):
    """
    Root resource for authenticated access to sub resources.

    It supports the following authentication methods:
    * Get authenticated session from Authorization header (for API)
    * Get authenticated session from cookie (for browser)
    * SSL certificated with username in CN.
    """

    REALM_NAME = 'Web files access'
    isLeaf = False

    def __init__(
        self, resource_class, session_interface, service_configuration,
        translation,
            ):
        credential_checker = [
            ServiceCredentialChecker(
                service_configuration=service_configuration),
            ]
        realm = HTTPRealm(
            resource_class=resource_class,
            service_configuration=service_configuration,
            translation=translation,
            )
        self._translation = translation
        self._SessionInterface = session_interface
        self._portal = Portal(realm, credential_checker)
        # BasicCredentialFactory should receive an ASCII string.
        self._credential_factories = [
            HTTPBasicAuthCredentialFactory(self.REALM_NAME),
            SessionAuthCredentialFactory(
                get_session=self._getSession,
                hide=not service_configuration.announce_session_authentication,
                ),
            ]

    def getChildWithDefault(self, path, request):
        """
        Inspect the Authorization HTTP header, and return a deferred which,
        when fired after successful authentication, will return an authorized
        C{Avatar}. On authentication failure, an C{UnauthorizedResource} will
        be returned, essentially halting further dispatch on the wrapped
        resource and all children
        """
        # Don't consume any segments of the request - this class should be
        # transparent!
        # FIXME:6201:
        # Once we have everying as unicde, this should just work.
        # Now postpath is bytes, and we inject `path` which is Unicode.
        request.postpath.insert(0, request.prepath.pop())
        return self._getResource(request)

    def headersReceived(self, request):
        """
        See: `IChevahResource`.

        This is called only when we have a URL for the root.
        Otherwise getChildWithDefault is called for the URL traversal and
        another resource is selected as the end resource for the request.
        """
        def cb_go_resource(result):
            """
            Called when we got the resource for the request.
            """
            return result.headersReceived(request)

        deferred = self._getResource(request)
        deferred.addCallback(cb_go_resource)
        return deferred

    def render(self, request):
        """
        This is called when this is called for the root URL,
        as otherwise other resource is returned as part of the traversal.
        """
        return request._wrapped_resource.render(request)

    def _getResource(self, request):
        """
        Return a deferred which fires with the resource for the request.
        """
        fetched_resource = getattr(request, '_wrapped_resource', None)
        if fetched_resource is not None:
            return defer.succeed(fetched_resource)

        def cb_got_resource(result):
            """
            Called when the resource was resolved (success or failure).
            """
            # Only authenticate once per request.
            request._wrapped_resource = result
            return result

        def eb_got_resource(failure):
            """
            Called when we fail to get the resource.

            This is a critical, unknown error as a failure to authentication
            will result in returning a dedicated resource.
            """
            failure.trap(ServerError)
            error = failure.value
            emitFromError(error, peer=request.client)
            return InternalErrorResource(
                message='Error while traversing URL.',
                details='See previous logs.',
                )

        deferred = defer.maybeDeferred(
            self._getResourceForAuthentication, request)
        deferred.addErrback(eb_got_resource)
        # Make sure this is called after error, so that we also get the
        # error resource.
        deferred.addCallback(cb_got_resource)
        return deferred

    def _makeUnauthorizedResource(self, details, response=None):
        """
        Return an UnauthorizedResource.

        `response` is extra information from the authorization
        rejection response.

        Here to allow specialization.
        """
        return UnauthorizedResource(
            credential_factories=self._credential_factories,
            details=details,
            translation=self._translation,
            response=response,
            )

    def _getResourceForAuthentication(self, request):
        """
        Return the resource associated with the authorization request.
        It can be a resource returned by the portal or a resource advertising
        the authentication error.

        Inspect the Authorization HTTP header, and return a deferred which,
        when fired after successful authentication, will return an authorized
        C{Avatar}.
        """
        session = request.getExistingSession()
        if session:
            session = self._getSession(request)
            if session.resource:
                # FIXME:5513:
                # Here we should re-initalize the resource.
                # Session already authenticated.
                # Via cookie or a previous HTTPS certificate.
                session.resource.reset()
                return session.resource

        credentials = None
        authorization_header = request.getHeader(b'authorization')
        if authorization_header:
            # Try to get the credentials from the HTTP auth headers.
            factory, respString = self._selectParseHeader(authorization_header)
            if factory is None:
                # We only want to show a small part of the header.
                # Most of the time the important info on the start.
                # We try to avoind logging the full auth token as it can be
                # a security issue.
                header_excerpt = force_unicode(authorization_header, trim=30)
                return self._makeUnauthorizedResource(
                    details='Unknown auth method "{}".'.format(header_excerpt))
            try:
                credentials = factory.decode(respString, request)
                # This is here to help with testing.
                request._credentials = credentials
            except (cred_error.LoginFailed) as error:
                # Failed to decode the credentials.
                details = error.args[0]
                return self._makeUnauthorizedResource(details=details)
            except Exception:
                return InternalErrorResource(
                    message='Error while decoding REST authentication.',
                    details=force_unicode(traceback.format_exc()),
                    )

            session = request.getExistingSession()
            if session:
                session = self._getSession(request)
                if session.resource:
                    # FIXME:5513:
                    # Here we should re-initialize the resource.
                    # Session already authenticated via Auth header.
                    session.resource.reset()
                    return session.resource

        if not credentials:
            credentials = self._getCredentials(request)

        if not credentials:
            # Session is not valid and the request has no credentials.
            return self._makeUnauthorizedResource(
                details='No authorization header.')

        return self._login(credentials, request)

    def _getCredentials(self, request):
        """
        Return non HTTP authentication credentials from `request`.

        Return None if no credentials could be extracted.
        """
        # Try to get SSL credentials, as we don't have in the headers.
        peer_certificate = request._getPeerCertificate()
        if peer_certificate:
            # We don't have an username, so we use the one found in the
            # certificate.
            return HTTPSSSLCertificateCredentials(
                username=force_unicode(peer_certificate.get_subject().CN),
                certificate=peer_certificate,
                peer=request.getAvatarPeer(),
                creator=request.site,
                )

        return None

    def _login(self, credentials, request):
        """
        Get the L{IResource} avatar for the given credentials.

        This is called internally during the URL traversal, but is also
        called by the login page.

        @return: A L{Deferred} which will be called back with an L{IResource}
            avatar or which will errback if authentication fails.
        """
        if request.getExistingSession():
            request.getSession().expire()
            request.removeSession()

        d = self._portal.login(credentials, None, IResource)
        d.addCallback(self._loginSucceeded, credentials, request)
        d.addErrback(self._loginFailed, credentials)
        return d

    def _logout(self, request):
        """
        Called when the current session should explicitly end.
        """
        session = self._getSession(request)
        request.removeSession()
        if session.resource:
            session.resource.emitEvent('40037')
            session.resource = None
        else:
            """
            Logout called without a valid session.
            """

    def _loginSucceeded(self, result, credentials, request):
        """
        Handle login success by wrapping the resulting L{IResource} avatar
        so that the C{logout} callback will be invoked when rendering is
        complete.
        """
        (interface, delegated_resource, logout) = result

        # All REST access start with a folder, which is traversed based on
        # the URL from the request.
        # That folder is the root of the virtual filesystem.
        delegated_resource.start(credentials.type)

        session = self._getSession(request)
        # FIXME:5513:
        # Here we should not store the resource itself, but some resource seed.
        session.resource = delegated_resource

        # FIXME:4869:
        # Emit an event when session expires.
        return delegated_resource

    def _loginFailed(self, failure, credentials):
        """
        Handle login failure by presenting either another challenge (for
        expected authentication/authorization-related failures) or a server
        error page (for anything else).
        """
        if failure.check(cred_error.Unauthorized, cred_error.LoginFailed):
            return self._makeUnauthorizedResource(
                details=failure.value.args[0])

        if failure.check(ServerError):
            # This should be removed after we have custom Portal object.
            emitFromError(failure.value, peer=credentials.peer)
            # No specific message is visible by default, as this can be
            # an internal server error.
            details = ''
            response = None
            if failure.value.id == '20136':
                # This is a known expected error, so we publish the details.
                response = failure.value.data['response']
                details = failure.value.data['details']
            return self._makeUnauthorizedResource(
                details=details,
                response=response,
                )

        if failure.check(InsufficientAuthentication):
            # FIXME:5250:
            # Add support for password and certificate authentication.
            # For now fail if account is configured to require more
            # authentications.
            return self._makeUnauthorizedResource(
                details=failure.value.message)

        return InternalErrorResource(
            message='Error during authentication.',
            details=force_unicode(failure.getBriefTraceback()),
            )

    def _selectParseHeader(self, header):
        """
        Choose an C{ICredentialFactory} from C{_credential_factories}
        suitable to use to decode the given I{Authenticate} header.

        @return: A two-tuple of a factory and the remaining portion of the
            header value to be decoded or a two-tuple of C{None} if no
            factory can decode the header value.
        """
        elements = header.split(b' ')
        scheme = elements[0].lower()
        for fact in self._credential_factories:
            if fact.scheme == scheme:
                return (fact, b' '.join(elements[1:]))
        return (None, None)

    def _getSession(self, request):
        """
        Return the session object for this resource.
        """
        return self._SessionInterface(request.getSession())


class HTTPRealm(ServiceRealm):
    """
    Extend the ServiceRealm by returning the REST resource for the
    service avatar.
    """

    def __init__(self, resource_class, service_configuration, translation):
        self._resource_class = resource_class
        self._translation = translation
        self._ui_version = get_ui_version(service_configuration.ui_version)
        super(HTTPRealm, self).__init__(service_configuration)

    def _cbGetRESTResource(self, result):
        interface, avatar, logout = result
        return (
            interface,
            self._resource_class(
                avatar, self._translation, ui_version=self._ui_version),
            logout)

    def requestAvatar(self, account_configuration, mind=None, *interfaces):
        """
        Return a resource with is already authenticated / associated with
        the avatar.
        """
        if IResource in interfaces:
            deferred = super(HTTPRealm, self).requestAvatar(
                account_configuration, mind, interfaces)
            deferred.addCallback(self._cbGetRESTResource)
            return deferred
        else:
            raise NotImplementedError('requestAvatar not implemented')
