# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
Resource oriented interface for managing files over HTTP.

While this started as a REST API,
it provides support for WebDAV and custom JSON API.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

import datetime
import errno
import ntpath
import os
import six.moves.urllib.request

import six.moves.urllib.error
import zipfile

from hyperlink import URL as HyperURL
from twisted.internet import defer
from twisted.internet.interfaces import IPullProducer
from twisted.python.components import registerAdapter
from twisted.web import http
from twisted.web.error import UnsupportedMethod
from twisted.web.server import NOT_DONE_YET, Session
from zope.interface import Attribute, implementer

from chevah_compat.exceptions import CompatError
from chevah.server import emit, emitFromError, json, force_unicode
from chevah.server.authentication.credential_checker import (
    check_account_validity,
    )
from chevah.server.commons.constant import (
    HTTP_REST_CHILD_NAME,
    HTTP_PRIVATE_CHILD_NAME,
    HTTP_REST_PASSWORD_UPDATE_NAME,
    HTTP_REST_LOGIN_NAME,
    HTTP_REST_LOGOUT_NAME,
    TYPE_NAME,
    )
from chevah.server.commons.exception import (
    AuthorizationException,
    CredentialsForbidden,
    ServerError,
    ServerException,
    )
from chevah.server.commons.interface import (
    IApplicationAccountConfiguration,
    IChevahSession,
    IRESTFolder,
    IUploadStream,
    )
from chevah.server.commons.session import (
    emit_account_login_paths,
    )
from chevah.server.commons.jinja import ViewModel
from chevah.server.commons.zip import (
    StreamWriterWithPosition,
    ZipStreamWriter,
    )
from chevah.server.http.configuration import get_ui_version
from chevah.server.http.credential import (
    get_login_page_credential,
    )
from chevah.server.http.form import (
    FormLengthExceededException,
    MultiPartFormData,
    )
from chevah.server.http.protocol import _CONTINUE
from chevah.server.http.request import UNPROCESSABLE_ENTITY, URL
from chevah.server.http.resource import (
    AuthenticatedResourceWrapper,
    ChevahResource,
    GenericResource,
    InternalErrorResource,
    RedirectResource,
    StaticFilesystemResource,
    StaticForbiddenResource,
    StaticTextResource,
    UnauthorizedResource,
    )
from chevah.server.http.shell import RESTShell
from chevah.server.identity.avatar import ServiceAvatar
import six

_HOME_URL = '/%s/' % (HTTP_REST_CHILD_NAME)
NO_COOKIES_MARKER = b'?without-cookies'


class IRESTSession(IChevahSession):
    """
    Persistent storage for REST requests.
    """
    alerts = Attribute(
        """
        List with dictionaries representing alerts to be listed at next
        page load.

        Each dict contains `type` and `message`, both as Unicode strings.

        `type` is modeled after Twitter Bootstrap alerts.

        [
        {'type': 'error', 'message': 'Some error message.'},
        {'type': 'info', 'message': 'Informational message.'}
        ]
        """)


@implementer(IRESTSession)
class RESTSession(object):
    """
    See: `IRESTSession`.
    """

    def __init__(self, session):
        self.alerts = []
        self.resource = None


registerAdapter(RESTSession, Session, IRESTSession)


class HomeButtonResourceMixin(object):
    """
    Shared code code for resources having a button to REST root resource.
    """

    def addHomeButton(self):
        """
        Update the view model with a button to home.
        """
        self.page.button = self.page.pack(
            text='Go to home page',
            href=_HOME_URL,
            )


class JSONAPIException(Exception):
    """
    Generic exception raised by the JSON API.
    """
    #: Body of the response as JSON.
    body = None
    #: HTTP response code.
    code = None


class JSONAPIBadRequestException(JSONAPIException):
    """
    Exception raised by the REST API when request is badly formatted.
    """

    def __init__(self, message):
        self.body = {'errors': [{
            'message': message,
            }]}
        self.code = http.BAD_REQUEST


class JSONAPIUnprocessableEntityException(JSONAPIException):
    """
    Exceptions raised by the REST API when request is valid but could
    not be processed due to low level errors.
    """

    def __init__(self, errors):
        self.body = {'errors': errors}
        self.code = UNPROCESSABLE_ENTITY


class RESTRootRedirectResource(RedirectResource):
    """
    Handles the redirections for the REST root page.

    Most of the time it just redirects to /home/.

    It also supports short OAuth2 redirect URLs.
    """

    def __init__(self):
        default_redirect = '/{}/'.format(HTTP_REST_CHILD_NAME)
        super(RESTRootRedirectResource, self).__init__(default_redirect)

    def headersReceived(self, request):
        """
        See: `IChevahResource`.
        """
        if request.method == b'POST':
            # For POST we need to do a dynamic redirect to OAuth2 callback URL.
            return _CONTINUE

        # For everything else, just redirect as fast as possible.
        return super(RESTRootRedirectResource, self).headersReceived(request)

    def render(self, request):
        """
        Called when page load is requested to see where to redirect.

        If headers are not accepted, this is not called.
        """
        if request.method == b'POST':
            # We have a OAuth2 redirect callback URL.
            # There should be no other valid reason for the root page to
            # receive a POST request.
            auth_url = '/{}/{}'.format(
                HTTP_PRIVATE_CHILD_NAME,
                HTTP_REST_LOGIN_NAME).encode('ascii') + request.uri
            request.setHeader(b"location", auth_url)
            request.setResponseCode(http.TEMPORARY_REDIRECT)
            return b'Redirecting to our OAuth2 handler.'

        # For everything else, do normal redirect.
        # But we should not hit this as the redirection should happen at the
        # headers level.
        return super(RESTRootRedirectResource, self).render(request)


class RESTNotFound(GenericResource, HomeButtonResourceMixin):
    """
    Resource for when no file or folder could be found at requested URL.

    This does not emit an event, as this is triggered by a shell access
    and the shell will emit its event.

    This is just the UI side of that error.
    """
    def __init__(self, avatar, path, real_path, ui_version=None):
        title = 'Item does not exist.'
        message = (
            'The requested URL is not associated '
            'with any file or folder.'
            )
        super(RESTNotFound, self).__init__(
            code=http.NOT_FOUND,
            title=title,
            message=message,
            ui_version=ui_version,
            )

        self._avatar = avatar
        self._data = {
            'path': path,
            'real_path': real_path,
            }
        self.page.icon_class = 'far fa-file'
        self.addHomeButton()

    def renderEmbedded(self, request):
        """
        Return response page.
        """
        return super(RESTNotFound, self).render(request)

    def render(self, request):
        """
        Fails as this is an embedded resource only.
        """
        raise RuntimeError('RESTNotFound.render should not be called.')


class RESTBadRequest(GenericResource, HomeButtonResourceMixin):
    """
    Resource for a bad request.

    It uses the default generic UI version.
    """
    def __init__(
            self, event_emitter, path, real_path, message, ui_version=None):
        title = 'Bad request.'
        message = message
        super(RESTBadRequest, self).__init__(
            code=http.BAD_REQUEST,
            title=title,
            message=message,
            ui_version=ui_version,
            )

        self.emitEvent = event_emitter
        self._data = {
            'path': path,
            'real_path': real_path,
            'details': message,
            }
        self.page.icon_class = 'far fa-folder'
        self.addHomeButton()

    def renderEmbedded(self, request):
        """
        This is called as both normal resource and embedded resource.
        """
        return self.render(request)

    def render(self, request):
        """
        Return response page and render an event.
        """
        self._data['uri'] = request._public_uri
        self.emitEvent('40019', data=self._data)
        return super(RESTBadRequest, self).render(request)


class RESTInternalErrorResource(
        InternalErrorResource, HomeButtonResourceMixin):
    """
    Just like super, but provides a navigation link to the REST root resource.

    There is no UI version for this page as users should not see it.
    """
    def __init__(self, *args, **kwargs):
        super(RESTInternalErrorResource, self).__init__(*args, **kwargs)
        self.addHomeButton()


class RESTUnauthorizedResource(UnauthorizedResource):
    """
    Just like super, but for HTML content it redirects to the login page.

    But when cookies are disabled, it will trigger the HTTP Basic Auth process.
    """

    def headersReceived(self, request):
        """
        See: `IChevahResource`.
        """
        request.removeSession()
        self._setHeaders(request)
        expect_continue = request.requestHeaders.getRawHeaders(
            'expect', [''])[0].lower()
        if expect_continue == '100-continue':
            # If the client makes an explicit request for `Continue: 100`,
            # then we inform that they can't continue.
            return http.UNAUTHORIZED
        else:
            # Even if this is an error, we want to continue so that the client
            # will see the error message.
            return _CONTINUE

    def render(self, request):
        """
        See: `ChevahResource`.
        """
        self.emitEventFor(request)

        if _is_api_request(request):
            # Start with an empty response.
            # We always advertise a response, even when empty to make the
            # API more consistent and let users know that our API can
            # populate that response.
            self.json_body['errors'][0]['response'] = {}
            if request.site.public_error:
                # Update the generic JSON error with the more detailed
                # error.
                self.json_body['errors'][0]['message'] = self._details
                # We only send the public part of the response to the client.
                response = self._response.get('public_response', {})
                self.json_body['errors'][0]['response'] = response

            request.setResponseCode(http.UNAUTHORIZED)
            request.setHeader(b'content-type', b'application/json')
            return json.dumps(self.json_body).encode('utf-8')

        if request.uri.endswith(b'/' + NO_COOKIES_MARKER):
            # Respond with non-redirecting HTML.
            request.setResponseCode(http.UNAUTHORIZED)
            request.setHeader(b'content-type', b'text/html; charset=utf-8')
            return self.template.render(self.page).encode('utf-8')

        # For HTML we redirect to login page.
        request.responseHeaders.removeHeader('www-authenticate')
        return RedirectResource(
            '/' + HTTP_PRIVATE_CHILD_NAME + '/' + HTTP_REST_LOGIN_NAME,
            http.FOUND).render(request)


class RESTForbiddenResource(
        StaticForbiddenResource, HomeButtonResourceMixin):
    """
    Just like super, but provides a navigation link to the REST root resource.

    In the future we might want to return a page based on UI version.

    For now we already return the default version to keep the code
    simple and hope customers will not complain.
    """
    def __init__(self, *args, **kwargs):
        super(RESTForbiddenResource, self).__init__(*args, **kwargs)
        self.addHomeButton()


class RESTRootResource(AuthenticatedResourceWrapper):
    """
    Root resource for REST file interface.

    It wraps the main REST Shell to handle the login and logout.
    """

    def __init__(self, service_configuration, translation):
        super(RESTRootResource, self).__init__(
            resource_class=RESTFolder,
            session_interface=IRESTSession,
            service_configuration=service_configuration,
            translation=translation,
            )

    def _makeUnauthorizedResource(self, details, response=None):
        """
        Return an RESTUnauthorizedResource.
        """
        return RESTUnauthorizedResource(
            credential_factories=self._credential_factories,
            details=details,
            response=response,
            translation=self._translation,
            )


class HasPublicPath(object):
    """
    Mixin for resources that are exported as a public path.
    """

    @property
    def public_path(self):
        """
        Return the path as displayed to users.
        """
        return self._shell.getPath(self._segments)

    @property
    def public_path_encoded(self):
        """
        Return the UTF-8 encoded path.
        """
        return self.public_path.encode('utf-8')

    @property
    def real_path(self):
        """
        Return the path as located on the filesystem.
        """
        return self._shell.getRealPath(self._segments)


class _RESTMixin(HasPublicPath):
    """
    Shared code used by REST files and folders.
    """

    _avatar = None
    _segments = []
    _shell = None
    _ui_version = None

    def _getSession(self, request):
        """
        Return the REST session associated with the request.
        """
        return IRESTSession(request.getSession())

    @property
    def avatar(self):
        return self._avatar

    def _getRenderMethod(self, render_method):
        """
        Try to render and convert any known error into a dedicated page
        """
        def render(request):
            request.setHeader('server-path', self.public_path_encoded)

            try:
                return render_method(request)
            except AuthorizationException:
                # The Chevah permission don't allow the operation.
                # The error should already be emitted, here we only present
                # the UI side of it.
                return RESTForbiddenResource().renderEmbedded(request)
            except UnsupportedMethod as error:
                # Forward this error to be handled by higher levels.
                raise error
            except Exception as error:
                error_resource = self._getResourceForError(error)
                return error_resource.renderEmbedded(request)

        return render

    def _getResourceForError(self, error):
        """
        Return the resource associated with this error.
        """
        if isinstance(error, CompatError):
            error = OSError(errno.EACCES, error.message)

        if not isinstance(error, OSError):
            # Not an OS error, handle is as ISE.
            return RESTInternalErrorResource(details=force_unicode(error))

        if error.errno == errno.ENOENT:
            return RESTNotFound(
                avatar=self.avatar,
                path=self._shell.getPath(self._segments),
                real_path=self._shell.getRealPath(self._segments),
                ui_version=self._ui_version,
                )
        elif error.errno == errno.ENOTDIR:
            return self._getBadRequest(message='Not a directory.')
        elif error.errno == errno.EACCES:
            return RESTForbiddenResource()
        else:
            # We don't know the OS error, so we convert it into a bad
            # request, as OS error are common and not really ISE.
            return self._getBadRequest(message=force_unicode(error))

    def _getBadRequest(self, message):
        """
        Return a resource for a bad request.
        """
        return RESTBadRequest(
            event_emitter=self.emitEvent,
            path=self._shell.getPath(self._segments),
            real_path=self._shell.getRealPath(self._segments),
            message=message,
            ui_version=self._ui_version,
            )

    def emitEvent(self, event_id, data=None, avatar=None):
        """
        Emit an event for the current web resource.
        """
        if not data:
            data = {}

        new_data = data.copy()
        new_data['path'] = self.public_path
        new_data['real_path'] = self.real_path

        emit(event_id, avatar=self.avatar, data=new_data)

    def _addAlert(self, target, type=None, message=''):
        """
        Add an alert to the target object without duplicates.
        """
        if type is None:
            type = 'error'

        for alert in target.alerts:
            if alert['type'] == type and alert['message'] == message:
                # Don't add duplicate alerts.
                return

        target.alerts.append({'type': type, 'message': message})

    def _getRedirectToFolder(self, request):
        """
        Redirect request to a URI for a folder.
        """
        redirect = HyperURL.fromText(request.uri.decode('utf-8'))
        if redirect.path[-1] != '':
            redirect = redirect.replace(path=list(redirect.path) + [""])

        self.emitEvent(
            '40008',
            data={
                'uri': request._public_uri,
                'redirect_uri': redirect.to_iri().asText(),
                },
            )
        return RedirectResource(redirect.to_uri().asText().encode('ascii'))

    def _setError(self, message):
        """
        Add an error for the page.
        """
        self.json_body = {'errors': [{'message': message}]}
        self._addAlert(target=self.page, message=message)


class RESTLogoutResource(GenericResource, _RESTMixin):
    """
    The simple page used for UX/UI logout process.

    This is not very visible, so for now we don't bother with its translation.

    The actual logout process is delegated to the `logout` callback.

    For browser - HTML content - it redirects to a clean URL
    which does not have credentials,
    to avoid browsers re-using the URL with bad credentials as
    the reference URL for future request.
    """

    def __init__(self, logout):
        self._logout = logout
        super(RESTLogoutResource, self).__init__(
            code=http.OK,
            title='Logout Page',
            message='Logout with success',
            )

    def render(self, request):
        """
        Show logout page and expire the session.
        """
        self._logout(request)

        url = URL()
        url.parseFromRequest(request)
        self._url = '%s://%s/' % (url.protocol, url.host)

        if self._isJSONRequest(request):
            # Show default logout page.
            return super(RESTLogoutResource, self).render(request)
        else:
            # For HTML we use a redirection page.
            return RedirectResource(
                '/' + HTTP_PRIVATE_CHILD_NAME + '/' + HTTP_REST_LOGIN_NAME,
                http.FOUND).render(request)


class LoginViewModel(ViewModel):
    """
    The template filling of the login page.
    """
    def __init__(self, service_name, login_footnotes=''):
        self.alerts = []
        self.NO_COOKIES_MARKER = NO_COOKIES_MARKER
        self.oidc_name = ''
        self.oidc_url = ''
        self.service_name = service_name
        self.login_footnotes = self._getFootnotes(login_footnotes)

    def _getFootnotes(self, raw):
        """
        Parse the footnotes that are configured as a single text.
        """
        result = {
            'button': '',
            'title': '',
            'content': '',
            }

        if not raw:
            return result

        lines = raw.splitlines()
        if len(lines) < 3:
            raise ServerException(
                'Too few lines for the "login_footnotes" configuration.')

        result['button'] = lines[0]
        result['title'] = lines[1]
        result['content'] = '\n'.join(lines[2:])
        return result


class RESTLoginResource(GenericResource, _RESTMixin):
    """
    Handled the UI/UX for session and OpenID connect based authentication.

    The actual login process is delegated to the `login` callback.

    For HTML requests, it shows a login page.

    This resource is persistent across the whole site and is not
    initialized at each login request.
    """
    def __init__(self, service_configuration, login, translation):
        translation.install(self)
        self._login = login
        self._service_configuration = service_configuration
        # If we are on the login page, we are forbidden by default.
        self.code = http.FORBIDDEN

        ui_version = get_ui_version(service_configuration.ui_version)

        if ui_version == TYPE_NAME.UI_GEN_1:
            template = 'login-gen1.html.j2'
        else:
            template = 'login-gen2.html.j2'

        self.template = translation.get_template(template)
        service_name = self._("Web Client")
        if self._service_configuration.theme_path:
            service_name = self._service_configuration.parent.name
        self.page = LoginViewModel(
            service_name=service_name,
            login_footnotes=self._service_configuration.login_footnotes,
            )
        self.json_body = {
            'errors': [{
                'title': self._('Unauthorized'),
                'message': self._('Authentication is required.'),
                }],
            }

    def _getOIDC(self):
        """
        Return the first authentication method that will handle OpenID connect.

        Return None if no OIDC is enabled.
        """
        root = self._service_configuration.root
        auth_methods = self._service_configuration.parent.authentications
        return root.authentications.getOIDC(
            auth_methods,
            default=root.configuration.server.authentications)

    def render(self, request):
        """
        See: `GenericResource`.
        """
        if request.method == b'POST':
            return self.render_POST(request)

        if request.method == b'GET':
            return self.render_GET(request)

        return super(RESTLoginResource, self).render(request)

    def render_GET(self, request):
        """
        Called when user loads the page or as part of OIDC handshake.

        JSON API should only use POST.
        """
        # For multiple OIDC methods, we should support a list here.
        oidc = self._getOIDC()

        if oidc:
            # FIXME:6201:
            # prepath will no longer needed decoding.
            url = './%s?%s' % (
                request.prepath[-1].decode('utf-8'), oidc.request_argument)
            self.page.oidc_url = url
            self.page.oidc_name = oidc.name

            # See if we have an OIDC authorization request
            # and redirect to the OIDC server.
            auth_url = oidc.getAuthorizationURL(request)
            if auth_url:
                request.setHeader(b"location", auth_url)
                request.setResponseCode(http.TEMPORARY_REDIRECT)
                return b'Redirecting to the OpenID Connect provider.'

        return super(RESTLoginResource, self).render(request)

    def render_POST(self, request):
        """
        Process the authentication request from the browser or API.
        """
        self.page.alerts = []

        # Invalidate any existing session.
        # This is to prevent someone having the ID of on old session, which
        # now is re-used with another credentials.
        request.removeSession()

        try:
            credentials = self._getCredentials(request)
        except ServerException as error:
            # We need to mark the string for extraction,
            # but we only want to translate it for the end-user and not for
            # the events.
            details = self._(
                'Failed to get the credentials from the login page.')
            event_details = error.message
            self._setError(details)
            data = {
                'uri': request._public_uri,
                'details': event_details,
                }
            emit(
                '40000',
                peer=request.client,
                data=data,
                creator=request.site,
                )
            return super(RESTLoginResource, self).render(request)

        self._doLogin(credentials, request)
        return NOT_DONE_YET

    def _getCredentials(self, request):
        """
        Get credentials from the request.

        Raise an exception if no credentials are found.
        """
        credentials = None

        oidc = self._getOIDC()
        if oidc:
            # Try to get from OIDC
            credentials = oidc.getCredentials(request)

        if not credentials:
            # Try to get credentials from page.
            credentials = get_login_page_credential(request)

        if not credentials:
            raise ServerException(
                'No credentials or malformed credentials provided.')

        return credentials

    @defer.inlineCallbacks
    def _doLogin(self, credentials, request):
        """
        Called after we have the credentials.

        This should never fail.
        Instead it should display the error as a page.
        """
        # Login should never fail.
        # On error it returns a resource to display the error to the end user.
        # On success it returns a resource to display the user files.
        result = yield self._login(credentials, request)

        response_page = result
        error_message = self._('Authentication failed.')
        if IRESTFolder.providedBy(result):
            # Result is the resource which needs to generate the final
            # login response.
            # I don't know a better way to detect a success login.
            response_page = RESTLoginSuccessResource()
        elif isinstance(result, RESTUnauthorizedResource):
            # This is an expected login failure result.
            if request.site.public_error:
                # Use a more specific error message if service is configured
                # to disclose error details.
                error_message = result.details
            self._setError(error_message)
        else:
            # For any other response, show a generic error.
            self._setError(error_message)

        response = response_page.render(request)
        request.write(response)
        request.finish()


class RESTLoginSuccessResource(GenericResource, _RESTMixin):
    """
    The simple page used for handing a successful login.

    There is no translation for this page as for web end users this is
    just a redirection.

    For browser - HTML content - it redirects to the start page.
    For JSON, it shows the session id.
    """

    def __init__(self):
        super(RESTLoginSuccessResource, self).__init__(
            code=http.OK,
            title='Login success',
            message='Login with success',
            )

    def render(self, request):
        if not self._isJSONRequest(request):
            # On HTML we redirect to home page.
            code = http.FOUND
            if b'?redirect-' in request.uri:
                # For OpenID the cookies is not set in the redirection
                # response.
                # We trigger an HTML redirection so that the cookies are set.
                code = http.OK
            redirect = RedirectResource('/%s/' % (HTTP_REST_CHILD_NAME,), code)
            return redirect.render(request)

        # Don't set any cookies for JSON API.
        # And show the token as JSON.
        request.cookies = []
        try:
            session_id = request.session.uid
            self.json_body = {
                'results': [{'session': session_id.decode('ascii')}]}
        except Exception:
            self.code = http.BAD_REQUEST
            self.json_body = {
                'errors': [{'message': 'Invalid request.'}]}

        return super(RESTLoginSuccessResource, self).render(request)


class PasswordUpdateViewModel(ViewModel):
    """
    The data filling of the password update page.
    """
    def __init__(self):
        self.name = 'NOT KNOWN YET'
        self.home_href = _HOME_URL
        self.alerts = []


class RESTPasswordUpdateResource(ChevahResource, _RESTMixin):
    """
    Handled the UI/UX for password update

    This needs to be used from an already authenticated resource, so that
    the credentials are validated using the normal authentication chain.

    For HTML content it shows a form which does a POST on the same page.
    In the future we might have:
    * GET /password-update?token=1234567890 - to reset without existing one.

    When the password is changed, the existing session is still valid.

    This resource is persistent across the whole site and is not
    initialized at each page access.
    """
    def __init__(self, avatar, translation, ui_version):
        translation.install(self)
        super(RESTPasswordUpdateResource, self).__init__()

        template_path = 'password_update-gen2.html.j2'
        if ui_version == TYPE_NAME.UI_GEN_1:
            template_path = 'password_update-gen1.html.j2'

        self.template = translation.get_template(template_path)
        self.page = PasswordUpdateViewModel()
        self.json_body = {
            'errors': [{
                'title': 'Bad Request',
                'message': 'Invalid Request.',
                }],
            }
        self.code = http.OK
        self._avatar = avatar

    @property
    def public_path(self):
        return '/'

    @property
    def real_path(self):
        return '/'

    def render_GET(self, request):
        """
        For HTML, show the account update form.

        For JSON, invalid request as JSON should go directly for the POST.
        """
        self.page.name = self.avatar.name

        self.json_body = {
            'errors': [{
                'title': 'Bad request.',
                'message': 'Invalid page access.',
                }],
            }

        if self._isJSONRequest(request):
            # For JSON, we set the error code.
            # For HTML/Browser we leave the default OK code.
            self.code = http.BAD_REQUEST

        return self._render(request)

    def render_POST(self, request):
        """
        For HTML, process the request and show to confirmation page.
        If users will hit refresh, form is resubmitted....but is ok.

        For JSON, process the request and return the result.
        """
        self.page.alerts = []
        user_error = None

        try:
            current_password, new_password = self._extractPassword(request)
            if self.avatar.updatePassword(
                current_password, new_password,
                event_emitter=self.emitEvent
                    ):
                self.code = http.OK
                self.json_body = {
                    'results': [{
                        'message': 'Password successfully updated.',
                        }],
                    }
            else:
                # Failed to update password.
                # No event is emitted as the avatar will emit the event.
                # The user will receive a generic message.
                user_error = self._('Failed to update the password.')

        except ServerException as error:
            user_error = error.message

        if user_error:
            self.code = http.BAD_REQUEST
            if not self._isJSONRequest(request):
                # For web browsers we return OK to not trigger extra errors
                # inside the dev console for browsers like Chrome.
                self.code = http.OK
            self._setError(user_error)
        else:
            self.code = http.OK
            message = self._('Password successfully changed.')
            self.json_body = {
                'results': [{'message': message}],
                }
            self.page.alerts.append({'type': 'info', 'message': message})

        return self._render(request)

    def _extractPassword(self, request):
        """
        Extract the password update input from the request.

        Return a tuple of (current_password, new_password)
        Raise exception on errors.
        """
        if self._isJSONRequest(request):
            content = request.content.getvalue()
            try:
                result = json.loads(content.decode('utf-8'))
            except Exception:
                # Failed to read the JSON content.
                raise ServerException('Failed to read JSON input.')

            current_password = result.get('current_password', None)
            if not isinstance(current_password, six.text_type):
                current_password = None

            new_password = result.get('new_password', None)
            if not isinstance(new_password, six.text_type):
                new_password = None
        else:
            current_password = request.args.get(b'current-password', [b''])[0]
            new_password = request.args.get(b'new-password', [b''])[0]
            confirm_password = request.args.get(b'confirm-password', [b''])[0]
            if new_password != confirm_password:
                raise ServerException(self._('New passwords don\'t match.'))
            current_password = current_password.decode('utf-8')
            new_password = new_password.decode('utf-8')

        if not current_password:
            raise ServerException(self._('Missing current password.'))
        if not new_password:
            raise ServerException(self._('Missing new password.'))

        return (current_password, new_password)

    def _render(self, request):
        """
        Shared code for rendering the HTML and JSON page.
        """
        request.setResponseCode(self.code)
        if self._isJSONRequest(request):
            # We have a JSON request.
            request.setHeader(b'content-type', b'application/json')
            return json.dumps(self.json_body).encode('utf-8')
        else:
            # Respond by default with HTML.
            request.setHeader(b'content-type', b'text/html; charset=utf-8')
            return self.template.render(self.page).encode('utf-8')


class RESTFile(StaticFilesystemResource, _RESTMixin):
    """
    A REST HTTP resource which provides access to the files of an account.

    It should be instantiated by the RESTFolder only for file path.
    For GET and HEAD it has custom code to implement file transfer based
    operations.
    """

    def __init__(self, avatar, shell, segments, ui_version):
        path = shell.getRealPath(segments)
        super(RESTFile, self).__init__(path=path)
        self._ui_version = ui_version
        self._avatar = avatar
        self._shell = shell
        self._segments = segments
        # Overwrite the rendering to convert to REST errors.
        self.render = self._getRenderMethod(self.render)

        # For put requests, we open the file when headers are received,
        # but need to send the response after all data was received.
        self._delayed_put_error = None

    def __repr__(self):
        return "RESTFile(%s)" % (self._shell.getRealPath(self._segments))

    def headersReceived(self, request):
        """
        See: `IChevahResource`.
        """
        if request.method != b'PUT':
            # Use default limit for anything other than PUT.
            return _CONTINUE

        try:
            request.switchConsumer(
                self._shell.openForWriting(self._segments, request))

            # PUT request has no limit as is a file upload.
            # We set this only when have successfully opened the file,
            # otherwise it will go with the default lower limit.
            request.maximumBodyLength = None
            return _CONTINUE
        except Exception as error:
            self._delayed_put_error = error
            return http.BAD_REQUEST

    def _render_GET_no_log(self, request):
        """
        Helper to avoid double event emitting when using HEAD.

        Begin sending the contents of this L{File} (or a subset of the
        contents, based on the 'range' header) to the given request.
        """
        self.type = 'application/octet-stream'
        self.encoding = ''
        if six.PY2:
            # This is due to Twisted having different behaviour on Py2 and Py3.
            self.type = self.type.encode('ascii')
            self.encoding = self.encoding.encode('ascii')

        request.setHeader(b'accept-ranges', b'bytes')
        # Make sure the file is not cached.
        request.setDisableCacheHeaders()

        try:
            fileForReading = self._shell.openForReading(self._segments)
        except OSError as error:
            if error.errno == errno.EISDIR:
                return self._getRedirectToFolder(request).render(request)
            else:
                raise

        if request.method == b'HEAD':
            # Emit an extra event to know that a  HEAD request was received.
            self.emitEvent('40005', data={'uri': self.public_path})
            # For head we are not reading the content.
            fileForReading.closeWithoutEvent()
            return b''

        name = os.path.basename(self._path)
        content_disposition = 'attachment; filename="%s"' % (name,)
        # We force the file to be downloaded by the browser.
        request.setHeader(
            b'Content-Disposition', content_disposition.encode('utf-8'))

        # This returns a producer over the full file or only over the
        # requested range.
        raw_producer = self.makeProducer(request, fileForReading)
        # Handle any error
        producer = _PullProducerWithErrorHandling(request, raw_producer)

        producer.start()
        # and make sure the connection doesn't get closed
        return NOT_DONE_YET

    def render_GET(self, request):
        """
        Return the file content as body and extra data as HTTP headers.
        """
        return self._render_GET_no_log(request)

    def render_HEAD(self, request):
        """
        Similar to GET but only send the headers without content.
        """
        return self._render_GET_no_log(request)

    def render_PUT(self, request):
        """
        Triggered by file upload.

        This is call after all data for this request was received.
        """
        if self._delayed_put_error:
            raise self._delayed_put_error

        # Signal that all data was received.
        request.content.unregisterProducer()

        request.setResponseCode(http.CREATED)
        return b''

    def render_DELETE(self, request):
        """
        A PUT will delete the file as for the REST API.

        Return:
        * 204 - when it was successfully deleted.
        """
        self._shell.removeFile(self._segments)
        request.setResponseCode(http.NO_CONTENT)
        return b''


@implementer(IRESTFolder)
class RESTFolder(ChevahResource, _RESTMixin):
    """
    REST Resource for listing folder content.

    server-path header is set to let the client know the virtual path on
    server representing the current page as there might not be a direct
    mapping between web URL and server paths.

    The `render_*` methods are always called on folders as the traversal part
    makes sure to return a RESTFile if have a file.
    """
    def __init__(self, avatar, translation, ui_version):
        self._avatar = avatar

        translation.install(self)
        self._shell = RESTShell(avatar)

        # To know what type of authentication was used.
        # This is not updated for public sites.
        self._credentials_type = TYPE_NAME.NOT_DEFINED

        # Overwrite the rendering to convert to REST errors.
        self.render = self._getRenderMethod(self.render)
        self.reset()
        super(RESTFolder, self).__init__()

        self._ui_version = ui_version
        private_root = StaticTextResource('', content_type='text/html')
        self.putChild(HTTP_PRIVATE_CHILD_NAME.encode('ascii'), private_root)

        private_root.putChild(
            HTTP_REST_PASSWORD_UPDATE_NAME.encode('ascii'),
            RESTPasswordUpdateResource(self._avatar, translation, ui_version),
            )

    def reset(self):
        """
        Called when the resource is already authenticated and it needs to
        server a new request.
        """
        self._segments = []
        self._uploaded_files = []

    def start(self, credentials_type):
        """
        Called when the authentication was successful and we can start
        operating.
        """
        self._credentials_type = credentials_type
        emit_account_login_paths(
            id='40007',
            avatar=self._avatar,
            emitter=self.emitEvent,
            path='/',
            )

    def __repr__(self):
        result = '<RESTFolder %d of %r>' % (
            id(self),
            self._shell.getRealPath(self._segments),
            )
        return result

    def getChild(self, name, request):
        """
        Consume and accumulate segments and return the final resource.

        While request has still segments to traverse, it will return
        itself so that URL traversal code will continue on the same
        resource.

        It should only try to consume the request URL without doing any
        filesystem operations, as they are delegated to the `render_*` methods.
        """
        if name:
            # Accumulate child.
            self._segments.append(name.decode('utf-8'))
        elif name == b'' and request.postpath:
            # We are at what it looks like a folder path. but we
            # have consecutive slashes.
            return self._getBadRequest('Double slash in URL.')
        else:
            # Empty child (folder path) at end of traversal.
            pass

        if name != b'' and not name.strip():
            # We have a fragment containing only spaces.
            return self._getBadRequest('Empty URL fragment.')

        # We still have segments to navigate.
        if request.postpath:
            return self

        # FIXME:6201:
        # last_fragment will be unicode.
        last_fragment = request.prepath and request.prepath[-1]
        if last_fragment == b'':
            # This is a request for a path ending with a slash, so is
            # a folder request.
            return self

        # It looks like the URL is for a file.
        return RESTFile(
            self.avatar, self._shell, self._segments,
            ui_version=self._ui_version,
            )

    def headersReceived(self, request):
        """
        See: `IChevahResource`.
        """
        # FIXME:6201:
        # request.prepath will be unicode.
        if not request.prepath or request.prepath[-1] != b'':
            # When no traversal is required, this is called directly.
            # But we only want to render it for folder path.
            return self._getRedirectToFolder(request).headersReceived(request)

        if request.method != b'POST':
            # Use default content handler and request size.
            return _CONTINUE

        content_type, _ = request.getContentType()
        if content_type == TYPE_NAME.APPLICATION_JSON:
            # For JSON we accept default limits.
            return _CONTINUE

        # For now only multipart/form-data format is supported in POST.
        boundary = request.getContentBoundary()
        if not boundary:
            message = 'Request does not contain a form boundary.'
            self._addAlert(
                target=self._getSession(request),
                message=message,
                )
            return (http.BAD_REQUEST, message)

        self._setPOSTMultipartFormHandler(request, boundary)

        # We accept request of any size. Limit are enforced later by
        # request and form handler.
        return _CONTINUE

    def _setPOSTMultipartFormHandler(self, request, boundary):
        """
        Configure request content to use streamed form handler.
        """
        # Look at implementing a factory to avoid passing the request in this
        # way.
        self._request = request
        request.switchConsumer(MultiPartFormData(
            boundary=boundary, open_callback=self._cbOpenPart))
        request.content.deferred.addErrback(self._ebUploadForm, request)

        # POST Upload request has no limit.
        # Some limits are imposed by form parser.
        request.maximumBodyLength = None

    def _ebUploadForm(self, failure, request):
        """
        Called when form parser fails.
        """
        failure.check(FormLengthExceededException)
        self._addAlert(
            target=self._getSession(request),
            message='Requested action exceeds size limit.',
            )
        # We don't emit an event as it will be emitted by the request.
        # Right now form parser only fails for parts that are too big.
        request.loseConnection(
            code=http.REQUEST_ENTITY_TOO_LARGE,
            message=str(failure.value),
            )

    def _cbOpenPart(self, headers, raw_headers):
        """
        Called when form received a streamed part.

        See: MultiPartFormData
        """
        if not headers:
            return None

        content_disposition = headers.get('content-disposition')
        if content_disposition['value'] != 'form-data':
            return None

        file_path = self._getHeaderOption(
            content_disposition['options'], 'filename')

        if not file_path:
            return None

        relative_segments = self._getRelativeSegments(file_path)
        self._shell.makeDirectories(
            base=self._segments, relative=relative_segments[:-1])
        absolute_segments = self._segments + relative_segments
        new_file = FormUploadFile(
            self.avatar, self._shell, absolute_segments, self._request,
            self._translation,
            )
        self._uploaded_files.append(new_file)

        return new_file

    def _getHeaderOption(self, options, name):
        """
        Return option with `name` as Unicode from header `options`.
        """
        value = options.get(name, None)
        if value is not None:
            value = force_unicode(value)
        return value

    def render(self, request):
        """
        Render the page.
        """
        # Make sure the folder listing is not cached.
        request.setDisableCacheHeaders()

        # FIXME:6201:
        # request.prepath will be unicode.
        if not request.prepath or request.prepath[-1] != b'':
            # When no traversal is required, this is called directly.
            # But we only want to render it for folder path.
            return self._getRedirectToFolder(request).render(request)

        return super(RESTFolder, self).render(request)

    def render_GET(self, request):
        """
        Return the string with the HTML representing folder content.
        """

        if _is_api_request(request):
            return self._processJSON(self._render_GET_json, request)

        accept = request.getAccept()
        if accept.best_match([TYPE_NAME.APPLICATION_ZIP]):
            # Download the whole folder as ZIP.
            return self._render_GET_zip(request)

        # Fallback to HTML rendering.
        request.setHeader(b'content-type', b'text/html; charset=utf-8')
        return self._render_GET_html(request)

    def render_HEAD(self, request):
        """
        Always return empty body with status code:
        * 200 when folder exists.
        * 409 when path exist but is not a folder
        * 404 on all other errors.

        This assumes that the getChild (traversal) code already checked that
        path exists.
        """
        self.emitEvent('40036')

        attributes = self._shell.getAttributes(self._segments)
        if attributes.is_folder:
            request.setResponseCode(http.OK)
        else:
            # Path exists, but is not a folder.
            request.setResponseCode(http.CONFLICT)

        return b''

    def render_PUT(self, request):
        """
        A PUT will create the folder as for the REST API.
        """
        return self._createFolder(request)

    def render_DELETE(self, request):
        """
        A PUT will delete the folder recursive as for the REST API.

        Return:
        * 204 - when it was successfully deleted.
        """
        self._shell.removeDirectory(self._segments)
        request.setResponseCode(http.NO_CONTENT)
        return b''

    def render_MKCOL(self, request):
        """
        MKCOL creates a new collection resource at the location specified by
        the Request-URI as for the WebDAV API.

        See https://tools.ietf.org/html/rfc4918#section-9.3
        """
        return self._createFolder(request)

    def _createFolder(self, request):
        """
        Code shared by REST and WebDAV.

        Return an empty body with code:
         * 201 when created.
        """
        # From WebDAV
        # 409 (Conflict) - A collection cannot be made at the Request-URI until
        # one or more intermediate collections have been created.  The server
        # MUST NOT create those intermediate collections automatically.
        # For REST
        # In the future we might take an optional argument to not raise an
        # error when folder exists and to just return 200.
        self._shell.makeDirectory(self._segments)
        request.setResponseCode(http.CREATED)
        return b''

    def _render_GET_html(self, request):
        """
        Process GET request as HTML.
        """
        # We trigger session creation before writing anything on the wire so
        # that we can attach any alerts.
        request.getSession()
        data = self._getPageData(request=request)
        page = DirectoryListingPage(
            request=request,
            data=data,
            translation=self._translation,
            )
        template_path = 'home-gen2.html.j2'
        if self._ui_version == TYPE_NAME.UI_GEN_1:
            template_path = 'home-gen1.html.j2'

        content = self._translation.get_template(template_path).render(page)
        return content.encode('utf-8')

    def _render_GET_json(self, request):
        """
        Process GET request as json.
        """
        content = self._listFolder()

        result = []
        for member in content:
            result.append({
                'name': member[0],
                'size': member[1],
                'modified': member[2],
                'is_directory': member[3],
                })
        return {'content': result}

    def _render_GET_zip(self, request):
        """
        Process the GET request as ZIP archive.
        """
        producer = ZIPProducer(
            shell=self._shell,
            base=self._segments[:],
            members=[],
            request=request,
            event_emitter=self.emitEvent,
            )
        producer.start()
        return NOT_DONE_YET

    def _listFolder(self):
        """
        Do the low level folder listing.

        Return a list with folder members.
        """
        directory_content = []
        for member in self._shell.openDirectory(self._segments):
            is_folder = member.is_folder
            if is_folder:
                # Folders might have size, but we don't show it.
                size = 0
            else:
                size = member.size

            directory_content.append((
                member.name,
                size,
                member.modified,
                is_folder,
                ))
        return directory_content

    def _getPageData(self, request):
        """
        Return data required for rendering this resource as HTML.

        All directories have 0 size to have common behaviour between Unix
        and Windows.
        """
        service_name = self._("Web Client")
        if request.site.service_configuration.theme_path:
            service_name = (
                request.site.service_configuration.parent.name)
        directory_content = self._listFolder()
        result = {
            'segments': self._segments,
            'content': directory_content,
            'username': self.avatar.name,
            'can_change_password': self.avatar.can_change_password,
            'credentials_type': self._credentials_type,
            'triggers': self._getTriggers(request),
            'service_name': service_name,
            }

        try:
            url = URL()
            url.parseFromRequest(request)
            result['root'] = url.path[0].decode('ascii')
        except UnicodeError:
            self._addAlert(
                target=self._getSession(request),
                message='Invalid root path.',
                )
            result['root'] = ''

        return result

    def render_POST(self, request):
        """
        Process commands received via POST request.
        """
        content_type, _ = request.getContentType()
        if content_type == TYPE_NAME.APPLICATION_JSON:
            return self._processJSON(self._render_POST_json, request)
        else:
            return self._render_POST_multipart(request)

    def _render_POST_multipart(self, request):
        """
        Process commands received via POST request as 'multipart/form-data'.

        It uses Post/Redirect/Get (PRG) pattern when done via a normal browser
        form request and not from the JS upload script.
        """
        # We trigger session creation before writing anything on the wire.
        session = self._getSession(request)
        action = request.getTextArgument('action')

        already_responded = self._handlePOSTMultipartAction(action, request)

        if already_responded:
            # The action has already done its response to the request.
            # No need for extra handling here.
            return NOT_DONE_YET

        if request.getHeader('X-Upload-JS') and session.alerts:
            # We have a JS upload request and we got errors.
            message = session.alerts[0]['message']
            session.alerts = []
            # We got a request from JS and we have errors.
            # Show them right away without redirection.
            request.setResponseCode(http.BAD_REQUEST)
            return _html_escape(message, quote=True).encode('utf-8')
        else:
            # This is a non-JS request from the browser, so we can't
            # directly return the error, but rather redirect to a page
            # which will load the error.
            return RedirectResource(request.uri, http.FOUND).render(request)

    def _processJSON(self, method, request):
        """
        High level process of requests received via 'application/json' using
        `method`.
        """
        request.setHeader(b'content-type', b'application/json')
        try:
            result = method(request)
            if result:
                return json.dumps(result).encode('utf-8')
            else:
                return b''
        except JSONAPIException as error:
            request.setResponseCode(error.code)
            return json.dumps(error.body).encode('utf-8')

    def _render_POST_json(self, request):
        """
        Execute the actions from POST JSON request.

        Raise exceptions on failures.
        """
        try:
            content = json.loads(request.content.read())
        except ValueError:
            raise JSONAPIBadRequestException('Failed to decode JSON content.')

        if not isinstance(content, dict):
            raise JSONAPIBadRequestException('Body must be an object.')

        commands = content.get('commands', None)
        if not commands:
            raise JSONAPIBadRequestException('No "commands" key in request.')
        if not isinstance(commands, list):
            raise JSONAPIBadRequestException('"commands" must be a list.')

        actions = []
        for command in commands:
            if not isinstance(command, dict):
                raise JSONAPIBadRequestException('Command must be an object.')

            command_name = command.get('command', None)
            if not command_name:
                raise JSONAPIBadRequestException('No specified command type.')

            target = command.get('target', None)
            if not target:
                raise JSONAPIBadRequestException(
                    'No target for command: %s' % (command_name,))

            if command_name == 'delete':
                actions.append((self._deleteMember, target))

            elif command_name == 'create-folder':
                actions.append((self._createNewFolderPOST, target))

            elif command_name == 'create-folder-if-missing':
                actions.append((self._createNewFolderIfMissingPOST, target))

            elif self._isClientTrigger(request, command_name):
                actions.append(
                    (self._triggerClientAction, command_name, target))

            else:
                raise JSONAPIBadRequestException(
                    'Unknown command: %s' % (command_name))

        errors = []
        have_errors = False
        for action in actions:
            method = action[0]
            target = action[1:]
            error_message = method(*target)
            if error_message:
                have_errors = True

            target = ' | '.join([force_unicode(t) for t in target])

            errors.append({'target': target, 'message': error_message})

        if have_errors:
            raise JSONAPIUnprocessableEntityException(errors)

    def _handlePOSTMultipartAction(self, action, request):
        """
        Perform the requested multipart POST action.

        Return `True` if the action has already responded to the request,
        besides acting on it.
        """
        if action == 'create-folder':
            return self._createNewFolderMultipart(request)

        if action == 'delete-members':
            return self._deleteMembersMultipart(request)

        if action == 'download-members':
            return self._downloadMembersMultipart(request)

        if action == 'upload-file':
            return self._checkUploadFile(request)

        if self._isClientTrigger(request, action):
            return self._triggerClientActionMultipart(request, action)

        self._addAlert(
            target=self._getSession(request),
            message=self._('You need to explicitly press an action button.'),
            )
        return False

    def _createNewFolderMultipart(self, request):
        """
        Create a new folder from `request'.
        """
        name = request.getTextArgument('new-folder')
        if not name:
            self._addAlert(
                target=self._getSession(request),
                message=self._(
                    'You need to specify a name for the new folder.'),
                )
            return
        error_message = self._createNewFolderPOST(name)
        if error_message:
            self._addAlert(
                target=self._getSession(request),
                message=error_message,
                )

    def _createNewFolderPOST(self, name):
        """
        Perform low level folder create operation.

        Return None on success or an error message on errors.
        """
        return self.__createNewFolderPOST(name, ignore_already_exists=False)

    def _createNewFolderIfMissingPOST(self, name):
        """
        Perform low level folder create operation and don't raise an error
        if it exists.

        Return None on success or an error message on errors.
        """
        return self.__createNewFolderPOST(name, ignore_already_exists=True)

    def __createNewFolderPOST(self, name, ignore_already_exists):
        """
        Shared code used by create-folder and create-folder-if-missing.
        """
        try:
            self._shell.makeDirectory(
                self._segments + [name], ignore_already_exists)
            # All good.
            return None
        except Exception:
            # Error is already emitted by the shell.
            # We just return the message for the UI.
            return self._('Failed to create new folder.')

    def _checkUploadFile(self, request):
        """
        Check upload request.

        Raise alert on errors.

        Content should have already been handled in streamed multipart parser
        and stored to disk.
        """
        if not self._uploaded_files:
            self._addAlert(
                target=self._getSession(request),
                message=self._('You need to specify a file for upload.'),
                )
            return

    def _getRelativeSegments(self, name):
        """
        Return a valid segments for `file_path`, relative to the current
        working directory.

        IE 7 and above has a security option named:
        `Include directory path when uploading files to server`.
        this is why we parse the file name part.
        """
        # Make sure path is always relative,
        # In the case we have some crazy Windows share path or something else.
        name = name.replace('\\', '/')
        name = name.strip('/')
        if ntpath.isabs(name):
            # We use ntpath since this provides an OS independent file handling.
            # chevah_compat or os.path would return paths for current OS and
            # for example os.path.basename('c:\bla\bla') return `c:\bla\bla` on
            # Unix since that is a valid file name.
            return [ntpath.basename(name)]

        return name.split('/')

    def _deleteMembersMultipart(self, request):
        """
        Delete selected folder members from POST multipart `request`.
        """
        raw_members = request.args.get(b'selected-members', [])
        # Decode UTF-8 input and ignore any empty members or bad encoded.
        members = []
        for name in raw_members:
            if not name:
                continue
            try:
                members.append(name.decode('utf-8'))
            except UnicodeError:
                # Names with bad encoding are ignored.
                raw_name = repr(name)
                self.emitEvent(
                    '40028',
                    data={'name': raw_name, 'operation': 'delete'},
                    )
                continue

        if not members:
            self._addAlert(
                target=self._getSession(request),
                message=self._('You need to select members for deletion.'),
                )
            return False

        failed_members = []

        for name in members:
            error_message = self._deleteMember(name)
            if error_message:
                failed_members.append(name)

        if failed_members:
            failed_members = ', '.join(failed_members)
            self._addAlert(
                target=self._getSession(request),
                message=self._('Failed to delete the requested members.'),
                )

    def _deleteMember(self, name):
        """
        Execute low level member removal.
        """
        member_segments = self._segments[:]
        member_segments.append(name)

        try:
            attributes = self._shell.getAttributes(member_segments)
        except Exception:
            # Error is emitted, but we return this for the UI.
            return self._('Failed to delete.')

        if attributes.is_folder:
            try:
                self._shell.removeDirectory(member_segments)
            except Exception:
                # Error is emitted, but we return this for the UI.
                return self._('Failed to delete folder.')
        else:
            try:
                self._shell.removeFile(member_segments)
            except Exception:
                # Error is emitted, but we return this for the UI.
                return self._('Failed to delete file.')

    def _downloadMembersMultipart(self, request):
        """
        Download selected / all folder members from POST multipart `request` as
        a Zip archive.
        """
        raw_members = request.args.get(b'selected-members', [])
        # Decode UTF-8 input and ignore any empty members or bad encoded.
        members = []
        for name in raw_members:
            if not name:
                continue
            try:
                members.append(name.decode('utf-8'))
            except UnicodeError:
                # Names with bad encoding are ignored.
                raw_name = repr(name)
                self.emitEvent(
                    '40028',
                    data={'name': raw_name, 'operation': 'download'},
                    )
                continue

        name = self._getZIPName(members).encode('utf-8')
        # We force the file to be downloaded by the browser.
        request.setHeader(
            b'Content-Disposition', b'attachment; filename="' + name + b'"')
        producer = ZIPProducer(
            request, self._shell, self._segments[:], members,
            event_emitter=self.emitEvent,
            )
        producer.start()
        return True

    def _getZIPName(self, members):
        """
        Return the name of the ZIP file to be used when presenting the file
        to the web browser.

        This tries to do similar things to Google Drive.
        """
        # The marker is ISO 8601 - 20190306T083531Z
        time_marker = (
            datetime.datetime.utcnow().replace(microsecond=0).isoformat() +
            'Z').replace(':', '').replace('-', '')

        if members:
            name = 'sftpplus-download'
        else:
            try:
                name = self._segments[-1]
            except IndexError:
                # This is a full download.
                name = 'sftpplus-download-full'

        return '%s-%s.zip' % (name, time_marker,)

    def _getTriggers(self, request):
        """
        Return the triggers available for this account
        """
        triggers = []
        for trigger in request.site.triggers:
            try:
                if self._avatar.isMember(trigger[2]):
                    triggers.append(trigger)
            except IndexError:
                # This is an action for all groups.
                triggers.append(trigger)

        return triggers

    def _isClientTrigger(self, request, action):
        """
        Return True if action is a valid trigger for the current authenticated
        account.
        """
        for trigger in self._getTriggers(request):
            if action == trigger[0]:
                return True

        return False

    def _triggerClientActionMultipart(self, request, action):
        """
        Trigger `action` as requested by the client via form request.
        """
        members = [
            force_unicode(name)
            for name in request.args.get(b'selected-members', [])
            ]
        error = self._triggerClientAction(action, members)
        if error:
            self._addAlert(target=self._getSession(request), message=error)

    def _triggerClientAction(self, action, members):
        """
        Trigger `action` as requested by the client via form request.

        Return `None` if all is ok.
        Return a text with the error message.
        """
        if not members:
            return 'You need to select at least one file.'

        if isinstance(members, six.string_types):
            return 'You need to provide a list of files.'

        files = [
            self._shell.getRealPath(self._segments + [member])
            for member in members]

        self.emitEvent(
            '40043',
            data={
                'name': action,
                'files': files,
                }
            )

        # All good
        return None


@implementer(IPullProducer)
class _PullProducerWithErrorHandling(object):
    """
    Handle read errors in upstream Twisted static pull producers.

    Upstream Twisted Static pull producers just panic if the read operation
    fails, without calling stop producer or closing the stream.
    """
    def __init__(self, request, wrappedProducer):
        self._context = wrappedProducer
        self._request = request

    def start(self):
        """
        Hijack the producer registration.
        """
        register_producer = self._request.registerProducer
        try:
            self._request.registerProducer = lambda c, p: None
            self._context.start()
            register_producer(self, False)
        finally:
            self._request.registerProducer = register_producer

    def resumeProducing(self):
        """
        Since we are a pull producer as per the link, this is called each time
        the HTTP request needs more data.
        https://docs.twisted.org/en/stable/core/howto/producers.html
        """
        try:
            self._context.resumeProducing()
        except Exception:
            # This fails while reading.
            # If no data was read yet, we can still set the HTTP status.
            self._request.setResponseCode(http.SERVICE_UNAVAILABLE)
            self.stopProducing()
            # We also need to force closing the connection.
            # Since we have the content-length header or chunk encoding
            # the client is waiting for all the content or the termination
            # chunk and keeps the connection open.
            self._request.transport.loseConnection()

    def stopProducing(self):
        """
        Just forward the request.
        """
        self._request.unregisterProducer()

        if not self._request.finished:
            self._request.finish()

        self._context.stopProducing()


@implementer(IPullProducer)
class ZIPProducer(object):
    """
    Send the `members` from `base` segments as a ZIP archive via `request`.

    All the zip files will have `base` as root.

    If any of the requested members failed to be read, an event is emitted
    and the file is skipped and not added to the archive.
    """

    _CHUNK_SIZE = 65536

    def __init__(self, request, shell, base, members, event_emitter):
        """
        Initialize the instance.
        """
        self.emitEvent = event_emitter
        self._request = request
        self._shell = shell

        # The adapter for the request which will write data as ZIP archive.
        self._zip_stream = None

        # The high level members which needs to be included in the archive.
        # This is a tuple of (segments, cached_attributes).
        self._targets = []

        if not members:
            self._targets.append((base, None))
        else:
            for member in members:
                self._targets.append((base + [member], None))

        # The current source stream.
        self._source = None

        def cb_request_done(result):
            self.stopProducing()

        deferred = self._request.notifyFinish()
        deferred.addCallback(cb_request_done)

    def start(self):
        """
        Called when we can start.
        """
        self._zip_stream = ZipStreamWriter(
            file=StreamWriterWithPosition(self._request),
            mode='w',
            compression=zipfile.ZIP_STORED,
            allowZip64=True,
            )

        # Assume that all will go OK... it will be overwritten on errors,
        # but we can only do that before writing anything to the zip file.
        self._request.setResponseCode(http.OK)
        self._request.setHeader(
            b'Content-Type', TYPE_NAME.APPLICATION_ZIP.encode('ascii'))
        self._startNext()

        if not self._source:
            self._request.setResponseCode(http.BAD_REQUEST)
            return self.stopProducing()

        # We implement as a non-streaming (pull) producer
        self._request.registerProducer(self, False)

    def resumeProducing(self):
        """
        Since we are a non-streaming/pull producer, this is called each time
        the HTTP request needs more data.
        """
        if not self._request:
            return

        try:
            self._readNextChunk()
        except Exception as error:
            error_message = force_unicode(error)
            if self._source:
                error_message = 'Failure for source "{}". {}'.format(
                    '/' + '/'.join(self._source._segments),
                    error_message,
                    )
            self.stopProducing(error_message)

    def _readNextChunk(self):
        """
        Allow wrapping any read operation to make it easy to close on any
        failure.
        """
        if self._source:
            # We still have to read from the current file.
            data = self._source.read(self._CHUNK_SIZE)
            if data:
                # We can continue writing the current file.
                # Zip stream is wrapping the request.write().
                # this .write will spin the reactor, calling .doWrite and then
                # .resumeProducing again, so be prepared for a re-entrant call
                self._zip_stream.writeContentChunk(data)
                return

            # We are done writing the current file.
            self._zip_stream.writeFooter()
            self._source.close()
            self._source = None

        if self._targets:
            # Current file is done,
            # but we have other pending files.
            self._startNext()
            return

        # No more files to write.
        self._request.unregisterProducer()
        self.stopProducing()
        return

    def _startNext(self):
        """
        Start writing the next file in line.
        """
        try:
            segments, attributes = self._targets.pop(0)
        except IndexError:
            return

        try:
            new_file_started = self._handleMember(segments, attributes)
        except (EnvironmentError, ServerException):
            # The shell will already emit events on errors.
            new_file_started = False

        if not new_file_started:
            # Continue processing the remaining files.
            self._startNext()

    def _handleMember(self, segments, attributes):
        """
        Add the member to the zip archive.

        Return `True` if a new file was started.
        """
        if not attributes:
            attributes = self._shell.getAttributes(segments)

        if attributes.is_link:
            # We ignore links.
            return False

        if attributes.is_file:
            # Start processing the file.
            self._source = self._shell.openForReading(segments)
            self._zip_stream.writeHeader(
                path=attributes.path,
                modified=attributes.modified,
                )
            return True

        # Folder is added after a listing... so that if listing fails,
        # we don't include it.
        folder_attributes = attributes

        for attributes in self._shell.openDirectory(segments):
            self._targets.append((segments + [attributes.name], attributes))

        if folder_attributes.path == '/':
            # We don't write the root folder.
            return False

        # Write directory entry only after we were able to list it.
        self._zip_stream.writeHeader(
            path=folder_attributes.path + '/',
            modified=attributes.modified,
            )
        self._zip_stream.writeFooter()
        return False

    def stopProducing(self, error_message=None):
        """
        Stop producing data.

        L{IPullProducer.stopProducing} is called when our consumer has died,
        and subclasses also call this method when they are done producing
        data.

        We also call it when request was finished.
        """
        if self._source:
            self._source.close()
            self._source = None

        if self._zip_stream:
            self._zip_stream.close()
            self._zip_stream = None

        if self._request:
            if self._request.producer:
                # ZIP failed to be fully downloaded.
                details = 'Connection closed before sending all the archive.'
                if error_message:
                    details = error_message
                self.emitEvent('40041', data={'details': details})
                self._request.unregisterProducer()

            if not self._request.finished:
                self._request.finish()

            self._request = None


class DirectoryListingPage(ViewModel):
    """
    Fills the content for directory listing page.
    """

    timezone = None

    def __init__(self, request, data, translation):
        translation.install(self)
        self._request = request
        self._data = data
        # l10n: Human readable name for the root folder.
        self.ROOT_NAME = self._('Home')
        # l10n: Units displayed for the size of a file.
        self.SIZE_UNITS = [
            self._('bytes'), self._('KB'), self._('MB'), self._('GB')]

    @property
    def title(self):
        """
        Title of this page.
        """
        return self.me.path

    @property
    def service_name(self):
        """
        Title of this page.
        """
        return self._data['service_name']

    @property
    def me(self):
        """
        Information about current directory.

        Slots:
        * username
        * name
        * path
        * logout_href
        * change_password_href
        * members_count
        """
        if len(self._data['segments']):
            name = self._data['segments'][-1]
        else:
            name = self.ROOT_NAME

        path = '/' + '/'.join(self._data['segments'])

        if (
            self._data['credentials_type'] == TYPE_NAME.PASSWORD_BASIC_AUTH
            or self._data['credentials_type'] == TYPE_NAME.NOT_DEFINED
                ):
            # It looks like cookies are not enabled or we are in public.
            # So logout is not available.
            logout_href = ''
        else:
            # We have a session cookie so we can use the normal
            # login/logout flow based on dedicated pages.
            logout_href = (
                '/' + HTTP_PRIVATE_CHILD_NAME + '/' + HTTP_REST_LOGOUT_NAME)

        if (
            self._data['credentials_type'] == TYPE_NAME.PASSWORD
            and self._data['can_change_password']
                ):
            change_password_href = '/%s/%s/%s' % (
                HTTP_REST_CHILD_NAME,
                HTTP_PRIVATE_CHILD_NAME,
                HTTP_REST_PASSWORD_UPDATE_NAME,
                )
        else:
            change_password_href = ''

        return self.pack({
            'username': self._data['username'],
            'path': path,
            'name': name,
            'logout_href': logout_href,
            'change_password_href': change_password_href,
            'members_count': len(self._data['content']),
            })

    @property
    def parents(self):
        """
        List with parent directories.

        Slots:
        * name
        * href
        """
        result = []
        if not len(self._data['segments']):
            return result

        root = self._data['root']
        href = '/' + root + '/'
        result.append(self.pack({
            'name': self.ROOT_NAME,
            'href': href,
            }))

        for parent in self._data['segments'][:-1]:
            href = href + parent + '/'
            result.append(self.pack({
                'name': parent,
                'href': href,
                }))
        return result

    @property
    def content(self):
        """
        Current directory content.

        Slots:
        * name
        * icon_class
        * size_bytes
        * size_human
        * modified
        """
        result = []

        # Sort by type (file / directory) and then by name in lower case.
        sorted_content = sorted(
            self._data['content'], key=lambda data: (
                -data[3], data[0].lower()))

        for name, size, modified, is_directory in sorted_content:
            href = name
            if is_directory:
                icon_class = 'far fa-folder'
                size_human = ''
                # Folders require a trailing slash.
                href = href + '/'
            else:
                icon_class = 'far fa-file'
                size_human = self._sizeAsHumanReadable(size)

            result.append(self.pack({
                'name': name,
                'href': href,
                'icon_class': icon_class,
                'size_bytes': str(size),
                'size_human': size_human,
                'modified': self._unixTimeAsHumanReadable(modified),
                }))
        return result

    def _sizeAsHumanReadable(self, size):
        """
        Return human readable size.
        """
        for unit in self.SIZE_UNITS:
            if size < 1024.0 and size > -1024.0:
                return "%3.1f %s" % (size, unit)
            size /= 1024.0
        return "%3.1f %s" % (size, 'TB')

    def _unixTimeAsHumanReadable(self, timestamp):
        """
        Return human readable format for Unix timestamp.
        """
        date = datetime.datetime.fromtimestamp(timestamp, tz=self.timezone)
        return force_unicode(date.strftime('%Y-%m-%d %H:%M:%S'))

    @property
    def alerts(self):
        """
        Return list of current alerts and clears alerts from the session.

        Slots:
        * message
        * css_class
        """
        result = []
        session = IRESTSession(self._request.getSession())

        if (
            self._data['segments']
            and not self._request.getCookie(
                self._request.session_cookie_basename.encode('ascii'))
                ):
            # We might not have a valid session as we are not on the start
            # page and we don't have a session cookie.
            result.append(self.pack(
                type='warning',
                message=(
                    'Your browser is blocking cookies. '
                    'Some functionality might not be available.'
                    ),
                ))

        while session.alerts:
            alert = session.alerts.pop()
            result.append(
                self.pack(type=alert['type'], message=alert['message']))

        return result

    @property
    def triggers(self):
        """
        Return the list of trigger buttons to show on the page.
        """
        triggers = self._data['triggers']

        if not triggers:
            return []

        result = {
            'primary': {
                'name': triggers[0][0],
                'css_class': 'btn-' + triggers[0][1],
                'icon_class': self._getTriggerIconClass(triggers[0][1])
                },
            'secondary': []
            }

        for trigger in triggers[1:]:
            result['secondary'].append({
                'name': trigger[0],
                # FIXME:5562:
                # Here we should only return `trigger.type`
                # And let the UI decide what exact CSS class to use.
                # This should also allow us to remove the icon_class
                # as the UI can decide what icon to use.
                'css_class': 'btn-' + trigger[1],
                'icon_class': self._getTriggerIconClass(trigger[1])
                })

        return result

    def _getTriggerIconClass(self, trigger_type):
        """
        Return the icon class for trigger.
        """
        type_to_css_mapping = {
            'primary': 'fas fa-layer-group',
            'info': 'fas fa-info-circle',
            'success': 'fas fa-check-square',
            'warning': 'fas fa-exclamation',
            'danger': 'fas fa-exclamation-triangle',
            'inverse': 'fas fa-minus-square',
            }
        return type_to_css_mapping.get(trigger_type, '')


@implementer(IUploadStream)
class FormUploadFile(_RESTMixin):
    """
    File like object used for receiving file upload data over an HTTP
    form.

    It wraps the stream return by the shell to handle the errors to be
    visible to the browser.
    """

    def __init__(self, avatar, shell, segments, request, translation):
        translation.install(self)
        self._avatar = avatar
        self._shell = shell
        self._segments = segments
        self._request = request
        self._name = segments[-1]
        # Error produced during operation (open/write/etc).
        self._error = None
        self._file = None

        try:
            self._file = self._shell.openForWriting(segments, request)
        except Exception as error:
            self._error = error
            # The error is already emitted by the shell.
            # Here we handle it for the UI.
            self._addAlert(
                target=IRESTSession(self._request.getSession()),
                message=self._('Failed to open remote file for upload.'),
                )
            # If we fail now, the only way to fail is to close the connection,
            # and if we close the connection the browser will try again.
            # So we just discard all the data from browser and when the
            # browser is done, we sent the error.
            self._file = self._FileNull()

    class _FileNull(object):
        """
        A file like object which discard all received data.
        """

        def __init__(self):
            self.name = 'NULL'
            self.closed = False

        def write(self, data):
            pass

        def close(self):
            self.closed = True

    @property
    def name(self):
        """
        See: IUploadStream.
        """
        return self._name

    @property
    def closed(self):
        """
        See: IUploadStream.
        """
        return self._file.closed

    def write(self, data):
        """
        See: IUploadStream.
        """
        try:
            self._file.write(data)
        except Exception as error:
            # Close the current file.
            try:
                self.close()
            except Exception:
                """
                Ignore error close errors, as we will forward the write
                error.
                """

            # Record the error.
            self._error = error
            # For HTTP form, since is done by the browser we need to
            # receive all the request before sending our error code.
            self._addAlert(
                target=IRESTSession(self._request.getSession()),
                message=self._('Failed while storing uploaded file.'),
                )
            # And continue to next form part.
            self._file = self._FileNull()

    def close(self):
        """
        See: IUploadStream.
        """
        self._file.close()

    def interrupt(self, error):
        """
        See: IUploadStream.
        """
        self._file.interrupt(error)


class PublicForbiddenResource(StaticForbiddenResource):
    """
    Resource used when public site is configured but access is forbidden.
    """

    def __init__(self, message):
        data = {'details': message}
        super(PublicForbiddenResource, self).__init__(data=data)

    def emitEventFor(self, request):
        emit(
            '40038',
            peer=request.client,
            creator=request.site,
            data=self.data,
            )


class RESTPublicRoot(ChevahResource):
    """
    The resource for making the files of an account available
    without username and password.
    """

    def __init__(self, service_configuration, translation):
        translation.install(self)
        super(RESTPublicRoot, self).__init__()
        config = service_configuration.root.configuration
        public_account_uuid = service_configuration.public_account_uuid

        self._public_account = config.accounts.getSection(
            public_account_uuid)
        self._public_child = service_configuration.public_name
        if not self._public_account:
            raise ServerException(
                'The configured public account with UUID "%s" '
                'was not found.' % (public_account_uuid,)
                )
        if not IApplicationAccountConfiguration.providedBy(
                self._public_account):
            raise ServerException(
                'The configured public account with UUID "%s" '
                'is not an application account.' % (public_account_uuid,)
                )

        self._groups = []
        for group_uuid in self._public_account.group:
            group = config.groups.getSection(group_uuid)
            if not group:
                raise ServerException(
                    'The configured public account with UUID "%s" '
                    'has no associate group with UUID "%s".' % (
                        public_account_uuid, group_uuid)
                    )
            self._groups.append(group)

        self._ui_version = get_ui_version(
            service_configuration.ui_version)

    def getChild(self, name, request):
        """
        Called when we need to traverse the public path.
        """
        # Public folder might be "mounted" in root and have non-file
        # children.
        child = self.children.get(name, None)
        if child:
            return child

        # For public access we don't want to show the missing cookie warning.
        # Especially since we get a false alarm the very first time the page
        # is accessed.
        session_cookie = request.getCookie(
            request.session_cookie_basename.encode('ascii'))
        if not session_cookie:
            # Twisted received_cookies is populated by parseCookies
            # which uses the `cookie` header.
            request.received_cookies[
                request.session_cookie_basename.encode('ascii')] = b'a'

        public_avatar = ServiceAvatar(
            account_configuration=self._public_account,
            groups=self._groups,
            peer=request.client,
            component=request.site,
            )

        try:
            check_account_validity(
                me=self._public_account,
                groups=self._groups,
                )
        except CredentialsForbidden as error:
            return PublicForbiddenResource(error.message)

        try:
            return RESTFolder(
                public_avatar,
                self._translation,
                ui_version=self._ui_version,
                ).getChild(name, request)
        except ServerError as error:
            emitFromError(error, peer=request.client)
            return PublicForbiddenResource('See previous message.')
        except ServerException as error:
            return PublicForbiddenResource(error.message)

    def render(self, request):
        """
        Called when the public URl is called without an ending slash.

        In that case, getChildWithDefault is not called as no traversal is
        needed.
        """
        return RedirectResource(
            '/' + self._public_child + '/').render(request)


def _is_api_request(request):
    """
    Return `True` if it looks like we don't have a browser request.
    """
    if request.method not in [b'GET', b'POST']:
        # Browser only makes GET and POST requests.
        return True

    accept = request.getAccept()
    if not list(accept.values()):
        # Browsers always set some Accept.
        # So a missing accepts is a sign for API access.
        return True

    if list(accept.values()) == ['*/*']:
        # Curl asks for generic accept.
        # Browsers ask for specific accept.
        return True

    if accept.best_match([TYPE_NAME.APPLICATION_JSON]):
        # We have explicit request for JSON.
        return True

    # For any other accept value, assume this is not API.
    return False


def _html_escape(s, quote=True):
    """
    Replace special characters "&", "<" and ">" to HTML-safe sequences.
    If the optional flag quote is true (the default), the quotation mark
    characters, both double quote (") and single quote (') characters are also
    translated.

    Should be replaced with html.escape once we no longer need py2.
    """
    s = s.replace("&", "&amp;")  # Must be done first!
    s = s.replace("<", "&lt;")
    s = s.replace(">", "&gt;")
    if quote:
        s = s.replace('"', "&quot;")
        s = s.replace('\'', "&#x27;")
    return s
