# Copyright (c) 2017 Adi Roiban.
# See LICENSE for details.
"""
HTTP server-side file access layer.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

import errno
import os

from zope.interface import implementer

from twisted.internet import reactor
from twisted.internet.interfaces import IConsumer

from chevah_compat import LocalFilesystem

from chevah.server import emit, force_unicode
from chevah.server.commons.authorization import AuthorizationAvatarAdapter
from chevah.server.commons.constant import DEFAULT_FILE_MODE
from chevah.server.commons.exception import (
    AuthorizationException,
    ServerException,
    )
from chevah.server.commons.interface import (
    IHTTPShell,
    IUploadStream,
    )
from chevah.server.commons.csv import from_csv_row
from chevah.server.commons.match import get_matched_groups


class _EventEmitterWithFilesystemMixin(object):
    """
    Common code for emitting an event for a file.
    """
    _filesystem = None

    def _emitEventForPath(self, event_id, segments, data=None):
        """
        Emit an event path at `segments`.
        """
        if not data:
            data = {}

        new_data = data.copy()
        new_data['path'] = self._filesystem.getPath(segments)
        new_data['real_path'] = self._filesystem.getRealPathFromSegments(
            segments)
        emit(event_id, avatar=self._filesystem.avatar, data=new_data)


@implementer(IHTTPShell)
class RESTShell(_EventEmitterWithFilesystemMixin):
    """
    HTTP server-side shell for Chevah HTTP REST service.

    The methods are receiving segments, but should receive path.
    """
    _scheduler = reactor

    def __init__(self, avatar):
        self._avatar = avatar
        self._filesystem = LocalFilesystem(avatar=self._avatar)
        self._authorization = AuthorizationAvatarAdapter(avatar)

    def getRealPath(self, segments):
        """
        See: IServerShell.
        """
        return self._filesystem.getRealPathFromSegments(segments)

    def getPath(self, segments):
        """
        See: IServerShell.

        Return the virtual path.
        """
        return self._filesystem.getPath(segments)

    def exists(self, segments):
        """
        See: IHTTPShell.
        """
        try:
            return self._filesystem.getAttributes(segments)
        except Exception:
            return False

    def getAttributes(self, segments):
        """
        See: IServerShell.
        """
        try:
            path = self._filesystem.getPath(segments)
            self._authorization.canGetAttributes(path)

            attributes = self._filesystem.getAttributes(segments)
            attributes.path = self._filesystem.getPath(segments)
            return attributes

        except Exception as error:
            self._emitEventForPath(
                '40002',
                segments=segments,
                data={'details': force_unicode(error)},
                )
            raise error

    def openDirectory(self, segments):
        """
        See IHTTPShell.
        """
        self._emitEventForPath('40006', segments=segments)

        # We delegate to a separate method as otherwise, with a yield,
        # this method is called only when there is a request to iterate.
        return self._iterateFolder(segments)

    def _iterateFolder(self, segments):
        """
        Iterate over folder member and return an iterator generating
        member's attributes.
        """
        try:
            path = self._filesystem.getPath(segments)
            self._authorization.canTraverse(path)
            try:
                self._authorization.canList(path)
                show_only_folders = False
            except AuthorizationException:
                # We can traverse but can't list, so we only show folders.
                show_only_folders = True

            iterator = self._filesystem.iterateFolderContent(segments)
            for attributes in iterator:
                if show_only_folders and not attributes.is_folder:
                    continue

                attributes.path = self._filesystem.getPath(
                    segments + [attributes.name])
                yield attributes
        except Exception as error:
            self._emitEventForPath(
                '40014',
                segments=segments,
                data={'details': force_unicode(error)},
                )
            raise error

        self._emitEventForPath('40034', segments=segments)

    def makeDirectory(self, segments, ignore_already_exists=False):
        """
        See: IServerShell.
        """
        try:
            path = self._filesystem.getPath(segments)
            self._authorization.canCreateFolder(path)

            self._filesystem.createFolder(segments)
            self._emitEventForPath('40010', segments=segments)
            return

        except Exception as error:
            if (
                ignore_already_exists and
                isinstance(error, OSError) and
                error.errno == errno.EEXIST and
                self._filesystem.getAttributes(segments).is_folder
                    ):
                # Path is already a folder.
                # Ignore the error but emit an event to keep track of the
                # request.
                self._emitEventForPath(
                    '40030',
                    segments=segments,
                    data={'details': force_unicode(error)},
                    )
                # New folder was not created as it exists.
                return

            self._emitEventForPath(
                '40011',
                segments=segments,
                data={'details': force_unicode(error)},
                )
            raise error

    def removeDirectory(self, segments):
        """
        See: IServerShell.
        """
        try:
            path = self._filesystem.getPath(segments)
            self._authorization.canDeleteFolder(path)

            self._filesystem.deleteFolder(segments, recursive=True)
            self._emitEventForPath('40026', segments=segments)
        except Exception as error:
            self._emitEventForPath(
                '40027',
                segments=segments,
                data={'details': force_unicode(error)},
                )
            raise error

    def makeDirectories(self, base, relative):
        """
        See IHTTPShell.

        Create all sub-directories starting with `base` for so that relative
        segments exists.
        """
        if not relative:
            # No subfolder to create.
            return

        parent = base
        for part in relative:
            parent = parent + [part]
            if self.exists(parent):
                continue
            self.makeDirectory(parent)

    def removeFile(self, segments):
        """
        See: IServerShell.
        """
        try:
            path = self._filesystem.getPath(segments)
            self._authorization.canDeleteFile(path)

            self._filesystem.deleteFile(segments)
            self._emitEventForPath('40012', segments=segments)
        except Exception as error:
            self._emitEventForPath(
                '40013',
                segments=segments,
                data={'details': force_unicode(error)})
            raise error

    def rename(self, source, destination):
        """
        See: IServerShell.
        """
        data = {
            'from': self._filesystem.getPath(source),
            'to': self._filesystem.getPath(destination),
            }

        try:
            from_path = self._filesystem.getPath(source)
            to_path = self._filesystem.getPath(destination)
            self._authorization.canRename(from_path, to_path)

            self._filesystem.rename(source, destination)
            self._emitEventForPath('40055', segments=destination, data=data)
        except Exception as error:
            data['details'] = force_unicode(error)
            self._emitEventForPath('40056', segments=destination, data=data)
            raise error

    def openForReading(self, segments):
        """
        See: IServerShell.
        """
        return _HTTPFileReader(
            segments,
            filesystem=self._filesystem,
            authorization=self._authorization,
            scheduler=self._scheduler,
            )

    def openForWriting(self, segments, request=None, digest=None):
        """
        See: IServerShell.

        Here `request` is optional to make the interface happy, but it's
        a required argument.
        """
        return _HTTPFileWriter(
            segments,
            filesystem=self._filesystem,
            authorization=self._authorization,
            scheduler=self._scheduler,
            digest=digest,
            request=request,
            )


class _HTTPFileReader(_EventEmitterWithFilesystemMixin):
    """
    Wrapper for reading a local file over HTTP.
    """

    def __init__(self, segments, filesystem, authorization, scheduler):
        self._scheduler = scheduler
        self._stream = None
        self._filesystem = filesystem
        self._segments = segments
        self._read_total = 0
        self._start_time = self._scheduler.seconds()
        self._eof_reached = False
        self._read_error = None
        try:
            path = self._filesystem.getPath(segments)
            authorization.canRead(path)

            self._stream = self._filesystem.openFileForReading(segments)
            self._emitEventForPath(
                '40004',
                segments=segments,
                data={'uri': self._filesystem.getPath(segments)},
                )
        except Exception as error:
            self._emitEventForPath(
                '40035',
                segments=segments,
                data={'details': force_unicode(error)})
            raise error

    @property
    def closed(self):
        """
        See: IUploadStream.
        """
        return not self._stream

    def seek(self, offset):
        """
        Called for partial transfers.
        """
        self._stream.seek(offset)

    def read(self, length):
        """
        Called to return more data.
        """
        try:
            data = self._stream.read(length)
        except Exception as error:
            error_details = force_unicode(error)
            self._emitEventForPath(
                '40058',
                segments=self._segments,
                data={'details': error_details},
                )
            self._read_error = error_details
            self.close()
            raise

        read_size = len(data)
        self._read_total += read_size
        if not data or read_size < length:
            self._eof_reached = True
        return data

    def closeWithoutEvent(self):
        """
        Close the stream without emitting an event.

        This is used by the HEAD operation which closes the file without
        reading it.
        """
        stream, self._stream = self._stream, None
        stream.close()

    def close(self):
        """
        Called when read is done.
        """
        if not self._stream:
            return

        duration = self._scheduler.seconds() - self._start_time
        read_speed = 0
        if duration > 0:
            read_speed = self._read_total / duration

        data = {
            'total_read': self._read_total,
            'read_speed': read_speed,
            'duration': duration,
            }

        self.closeWithoutEvent()
        if self._eof_reached:
            self._emitEventForPath(
                '40039',
                segments=self._segments,
                data=data,
                )
        else:
            details = 'Connection closed before sending all data.'
            if self._read_error:
                details = 'Fail at read. {}'.format(self._read_error)
            data['details'] = details
            self._emitEventForPath(
                '40040',
                segments=self._segments,
                data=data,
                )


@implementer(IUploadStream, IConsumer)
class _HTTPFileWriter(_EventEmitterWithFilesystemMixin):
    """
    Wrapper for writing a file to local filesystem from HTTP.

    This is used by both the form-data and the direct stream request.

    The behavior is a bit different.
    A stream request will upload a single file, and then the request is done
    there is an implicit end of file.
    A form-data, can have multiple files, and the end of file is explicit via
    the form-boundary.
    """

    def __init__(
        self, segments, filesystem, authorization, scheduler, request,
        digest=None,
            ):
        self._scheduler = scheduler
        self._filesystem = filesystem
        # The low level file stream.
        self._stream = None
        # Signal that we have a producer writing data via a stream request.
        self._producer = None
        # Signal that an error occurred during any operations.
        self._error = None
        self._requested_path = filesystem.getPath(segments)

        self._digest_updater = digest
        if not self._digest_updater:
            # Bypass digestion if we don't need one.
            self._digest_updater = lambda _: None

        # Data attached to all events emitted for this file.
        self._event_data = {
            'requested_path': self._requested_path,
            }

        self._segments, was_path_amended = (
            filesystem.avatar.getAmendedUploadSegments(segments))
        self._name = segments[-1]

        # FIXME:2525:
        # Normal open.
        # Since there is no resume yet we always truncate the file.
        # FIXME:6587:
        # If we support resume, consider updating the maximum file size
        # that is allowed.
        flags = (
            filesystem.OPEN_WRITE_ONLY |
            filesystem.OPEN_CREATE |
            filesystem.OPEN_TRUNCATE
            )

        if was_path_amended:
            # File was amended so we don't want to overwrite existing files.
            flags = flags | filesystem.OPEN_EXCLUSIVE

        self._start_time = self._scheduler.seconds()
        self._write_total = 0

        self._write_limit = filesystem.avatar.file_write_size
        self._check_quota = lambda _, ignored: None
        if self._write_limit > 0:
            self._check_quota = self._checkQuota

        try:
            # Start by extracting the metadata from the request.
            self._event_data.update(_get_metadata(request))
            # Then check permissions.
            path = filesystem.getPath(segments)
            authorization.canWrite(path)
            # Then try to open the file.
            fd = filesystem.openFile(self._segments, flags, DEFAULT_FILE_MODE)
            self._stream = os.fdopen(fd, 'wb')

            self._emitEventForPath(
                '40021', segments=self._segments, data=self._event_data)
        except Exception as error:
            self.interrupt(error)
            raise error

    def registerProducer(self, producer):
        """
        Signal that we are receiving data from a streamed request.

        Only stream producer is supported.
        """
        self._producer = producer

    def unregisterProducer(self):
        """
        Called when all data was received.
        """
        self._producer = None
        self.close()

    def tell(self):
        """
        Fake that we are at the start of the stream.

        We don't support tell/seek since this we really do stream handling.
        But Twisted API is expected to handle a BytesIO buffer.
        In an ideal world we could have native steaming API in Twisted,
        but that is hard as other Twisted dev are not really interested into
        this.

        See: https://twistedmatrix.com/trac/ticket/6928
        """
        # self._write_total would have been some closes to reality,
        # but this might confuse Twisted upstream code as we don't pass
        # the whole request to cgi.parse_multipart().
        return 0

    def seek(self, offset, reference_point=0):
        """
        We don't do nothing here, but this is called by the Twisted web server
        after all data was received in the request.
        """

    @property
    def closed(self):
        """
        See: IUploadStream.
        """
        return not self._stream

    @property
    def name(self):
        """
        See: IUploadStream.
        """
        return self._name

    def write(self, data):
        """
        See: IUploadStream.
        """
        try:
            data_len = len(data)

            # These are hooks to the write operation.
            # They might be replaced with noop at runtime,
            # depending on how this is used.
            self._check_quota(self._write_total, data_len)
            self._digest_updater(data)

            self._stream.write(data)
            self._write_total += data_len
        except Exception as error:
            self._emitEventForPath(
                '40059',
                segments=self._segments,
                data={
                    'details': force_unicode(error),
                    },
                )
            self.interrupt(error)
            raise error

    def interrupt(self, error):
        """
        See: IUploadStream.
        """
        self._error = error

        duration = self._scheduler.seconds() - self._start_time
        write_speed = 0
        if duration > 0:
            write_speed = self._write_total / duration

        data = {
            'total_write': self._write_total,
            'write_speed': write_speed,
            'duration': duration,
            'details': force_unicode(error),
            }
        data.update(self._event_data)

        self._emitEventForPath(
            '40022', segments=self._segments, data=data)
        self.unregisterProducer()

    def close(self):
        """
        See: IUploadStream.
        """
        if self._producer:
            # The stream was requested to be closed while we still are waiting
            # for more streamed data. This is an error so we advertise the
            # transfer interruption.
            self._producer = None
            self.interrupt(
                ServerException('Closed while still receiving data.'))
            return

        if not self._stream:
            return

        self._stream.close()
        self._stream = None

        duration = self._scheduler.seconds() - self._start_time
        write_speed = 0
        if duration > 0:
            write_speed = self._write_total / duration

        data = {
            'total_write': self._write_total,
            'write_speed': write_speed,
            'duration': duration,
            }
        data.update(self._event_data)

        if not self._error:
            self._emitEventForPath('40017', segments=self._segments, data=data)

    def _checkQuota(self, written_size, new_chunk_size):
        """
        Check that we are allowed to write another `size` bytes.
        """
        if written_size + new_chunk_size > self._write_limit:
            raise ServerException(
                'Single file write limit is "{}" bytes.'.format(
                    self._write_limit))


def _get_metadata(request):
    """
    Return the REST metadata associated with the request.
    """
    result = {}
    config = request.site.metadata_configuration

    if not config:
        # No metadata extraction configured.
        return result

    values = request.requestHeaders.getRawHeaders(
        config.header_name.encode('UTF-8'))
    if not values:
        if not config.header_required:
            # No header provided and header is not required. All good.
            return result

        raise ServerException('Missing metadata header.')

    # Multiple occurrences of the header are joined into one string as CSV.
    values = ', '.join(force_unicode(h) for h in values)

    remaining_required = config.required_keys[:]

    for item in from_csv_row(values):

        try:
            key, value = item.split('=', 1)
        except ValueError:
            raise ServerException('Invalid metadata: "%s".' % (item,))

        key = key.strip()
        value = value.strip()

        if key in result:
            result[key].append(value)
        else:
            result[key] = [value]

        was_required = _is_required_metadata(config.required_keys, key)

        if was_required:
            try:
                remaining_required.remove(was_required)
            except ValueError:
                """
                This is a required key that was provided multiple times.

                All fine.
                """
        elif not _is_optional_metadata(config.optional_keys, key):
            raise ServerException('Extra metadata provided: "%s".' % (item,))

    if remaining_required:
        # Not all required keys were found in the request.
        raise ServerException(
            'Missing required metadata: %s.' % (
                ', '.join(remaining_required),))

    return result


def _is_required_metadata(required, candidate):
    """
    Returns False if the candidate key is not required.

    Return the matching required key if candidate is required.
    """
    if not required:
        return False

    is_required = False
    for key_expression in required:
        if get_matched_groups(candidate, key_expression, 'metadata_fields'):
            is_required = key_expression

    return is_required


def _is_optional_metadata(optionals, candidate):
    """
    Return true if `candidate` key name is an optional key.
    """
    for key_expression in optionals:
        if get_matched_groups(candidate, key_expression, 'metadata_fields'):
            return True

    return False
