# Copyright (c) 2016 Adi Roiban.
# See LICENSE for details.
"""
WebDAV extension, client and server-side.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

import posixpath
import six.moves.urllib.request
import six.moves.urllib.parse
import six.moves.urllib.error
from datetime import datetime

import dateutil.parser
from defusedxml import ElementTree
from hyperlink import URL
from twisted.internet import defer
from twisted.internet.address import IPv4Address
from twisted.internet.error import ConnectionDone
from twisted.web import http
from zope.interface import implementer

from chevah_compat import FileAttributes
from chevah.server import force_unicode
from chevah.server.commons.constant import ADDRESS_UNKNOWN, TYPE_NAME
from chevah.server.commons.credential import PasswordCredentials
from chevah.server.commons.interface import ILocationShell
from chevah.server.commons.exception import (
    HTTPClientException,
    ServerException,
    )
from chevah.server.http.client import (
    PersistentAgent,
    HTTPDownloadedFile,
    HTTPUploadedFile,
    get_basic_http_header,
    get_url_segments,
    )
from chevah.server.proxy.http import proxify_http_connect_agent
from chevah.server.identity.avatar import ClientAvatar
from chevah.server.location.commons import OnConnectionLostMixin
import six
from six.moves import range

_SAML_TEMPLATE = """
<s:Envelope xmlns:s="http://www.w3.org/2003/05/soap-envelope"
  xmlns:a="http://www.w3.org/2005/08/addressing"
  xmlns:u="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd"
  >
<s:Header>
<a:Action s:mustUnderstand="1">
  http://schemas.xmlsoap.org/ws/2005/02/trust/RST/Issue
</a:Action>
<a:ReplyTo>
<a:Address>http://www.w3.org/2005/08/addressing/anonymous</a:Address>
</a:ReplyTo>
<a:To s:mustUnderstand="1">https://login.microsoftonline.com/extSTS.srf</a:To>
<o:Security
  s:mustUnderstand="1"
  xmlns:o="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd"
  >
<o:UsernameToken><o:Username>%(username)s</o:Username>
<o:Password>%(password)s</o:Password>
</o:UsernameToken>
</o:Security>
</s:Header>

<s:Body>
<t:RequestSecurityToken xmlns:t="http://schemas.xmlsoap.org/ws/2005/02/trust">
<wsp:AppliesTo xmlns:wsp="http://schemas.xmlsoap.org/ws/2004/09/policy">
<a:EndpointReference>
  <a:Address>%(endpoint)s</a:Address>
</a:EndpointReference>
</wsp:AppliesTo>
<t:KeyType>
  http://schemas.xmlsoap.org/ws/2005/05/identity/NoProofKey
</t:KeyType>
<t:RequestType>
  http://schemas.xmlsoap.org/ws/2005/02/trust/Issue
</t:RequestType>
<t:TokenType>urn:oasis:names:tc:SAML:1.0:assertion</t:TokenType>
</t:RequestSecurityToken>
</s:Body>

</s:Envelope>
"""


def _get_error_details(response, base):
    """
    Return the error message from `response` starting with the `base` message.
    """

    # On SharePoint the error details are part of the headers.
    details = ''
    error = response.headers.getRawHeaders('x-msdavext_error')
    if error:
        # The error is as `timestamp; URL_ENCODE_MESSAGE`.
        details = six.moves.urllib.parse.unquote(error[0]).split('; ', 1)[-1]

    if details:
        return '%s %s' % (base, details)
    else:
        return base


class WebDAVAgent(PersistentAgent):
    """
    An HTTP agent which knows about the WebDAV methods.
    """

    def propfind(self, url, headers):
        """
        Make a PROPFIND request.
        """
        return self._makeRequest(
            method='PROPFIND',
            url=url,
            headers=headers,
            )

    def copy(self, url, headers):
        """
        Make a COPY request.
        """
        return self._makeRequest(
            method='COPY',
            url=url,
            headers=headers,
            )

    def move(self, url, headers):
        """
        Make a MOVE request.
        """
        return self._makeRequest(
            method='MOVE',
            url=url,
            headers=headers,
            )

    def mkcol(self, url, headers):
        """
        Make a MKCOL request.
        """
        return self._makeRequest(
            method='MKCOL',
            url=url,
            headers=headers,
            )


@implementer(ILocationShell)
class WebDAVClient(OnConnectionLostMixin):
    """
    Client-side shell for WebDAV.

    For the WebDAV we don't have `path` but `URI/URL/HREF`.
    A `file` is a WebDAV resource and a `folder` is a WebDAV collection.

    All internal HTTP requests should be done via `_webdavRequest` so that the
    authentication cookie/token is injected in the request.
    """
    _LOGIN_SITE = URL.from_text('https://login.microsoftonline.com')

    def __init__(self, location, context_factory, scheduler=None):
        self._agent = WebDAVAgent(
            ssl_context=context_factory.getContext(),
            event_emitter=location.emitEvent,
            )
        if location.configuration.proxy:
            proxify_http_connect_agent(
                self._agent,
                proxy=location.configuration.proxy,
                context_factory=context_factory,
                )

        # The location which is using this shell.
        self._location = location
        self._authenticationMethod = self._get_auth_method()
        # Token used to prove that once we had valid credentials,
        # or the credentials encoded for Basic Auth.
        self._authentication_headers = {}
        self._base_url = None
        # Initialized with an unknown peer.
        self._updateCredentials(peer=ADDRESS_UNKNOWN)
        # Deferred for the ongoing re-authentication.
        # None when no re-authentication is in progress.
        self._reauthentication_deferred = None
        # List of requests to be done after re-authentication.
        self._reauthentication_requests = []

    def _get_auth_method(self):
        """
        Validate and choose the authentication method.
        """
        auth_methods = {
            TYPE_NAME.SHAREPOINT_ONLINE: self._authenticateSharePoint,
            TYPE_NAME.BASIC_AUTHENTICATION_SCHEME: self._authenticateBasicAuth,
            }
        method = self._location.configuration.authentication_method

        try:
            return auth_methods[method]
        except KeyError:
            raise ServerException(
                'Unknown value for "authentication_method": "%s".' % (method,))

    @property
    def avatar(self):
        """
        See: ILocationShell.
        """
        return self._avatar

    def _updateCredentials(self, peer=None):
        """
        Update credentials used by this protocol and keep the avatar in
        sync with the credentials.
        """
        if peer is None:
            peer = IPv4Address(
                'TCP',
                self._location.configuration.address,
                self._location.configuration.port,
                )
        credentials = PasswordCredentials(
            username=self._location.configuration.username,
            password=self._location.configuration.password,
            creator=self._location,
            peer=peer,
            )
        self._avatar = ClientAvatar(credentials)

    def _updateAuthenticationHeaders(self, arguments):
        """
        Return request keyword arguments updated with authentication headers.
        """
        result = arguments.copy()
        headers = result.get('headers', {}).copy()
        headers.update(self._authentication_headers)
        result['headers'] = headers
        return result

    def _webdavRequest(self, method_name, path, **kwargs):
        """
        Make an HTTP request, and retry on authentication errors.

        `cookies` are not supported yet, as it will always inject the
        authentication cookie.
        """
        uri = self._base_url.replace(path=get_url_segments(path))
        http_method = getattr(self._agent, method_name.lower())

        if self._reauthentication_deferred:
            # We have a pending re-authentication process, so we queue
            # this request.
            return self._queueRequestForReAuthentication(
                http_method, uri, kwargs)

        @defer.inlineCallbacks
        def cb_first_request_done_basic(response):
            """
            Called when we got a response.

            We are authenticating using HTTP Basic.
            """
            if response.code in range(200, 300):
                # All good.
                defer.returnValue(response)

            if response.code in range(500, 600):
                # Try again.
                yield self._agent.readBodyWithIgnore(response)
                result = yield self._reauthenticate(http_method, uri, **kwargs)
                defer.returnValue(result)

            # Most likely a client error. Not worth trying again.
            raise ServerException(
                'Failed to authenticate on {}: {} - {}'.format(
                    self._base_url.to_text(),
                    response.code,
                    force_unicode(response.phrase),
                    ))

        @defer.inlineCallbacks
        def cb_first_request_done_SharePoint(response):
            """
            Called when we got a response.

            We are authenticating using SharePoint Online.
            """
            if response.code not in [http.UNAUTHORIZED, http.FORBIDDEN]:
                # All good.
                defer.returnValue(response)

            # It looks like the initial request was rejected.
            # Consume the body, then try again.
            yield self._agent.readBodyWithIgnore(response)
            response = yield self._reauthenticate(http_method, uri, **kwargs)
            defer.returnValue(response)

        auth_method = self._location.configuration.authentication_method
        response_handlers = {
            TYPE_NAME.BASIC_AUTHENTICATION_SCHEME: cb_first_request_done_basic,
            TYPE_NAME.SHAREPOINT_ONLINE: cb_first_request_done_SharePoint,
            }

        # Start a request with a re-authentication in case of error.
        new_kwargs = self._updateAuthenticationHeaders(kwargs)
        deferred = http_method(url=uri, **new_kwargs)
        deferred.addCallback(response_handlers[auth_method])
        return deferred

    def authenticate(self):
        """
        See: `ILocationShell`.
        """
        scheme = TYPE_NAME.HTTPS
        if self._location.configuration.allow_insecure_http:
            scheme = TYPE_NAME.HTTP

        self._base_url = URL.from_text('%s://%s:%s' % (
            scheme,
            self._location.configuration.address,
            self._location.configuration.port,
            ))

        return self._authenticationMethod()

    def _authenticateBasicAuth(self):
        """
        Authenticate using HTTP Basic Auth.
        """
        auth_value = get_basic_http_header(
            self._location.configuration.username,
            self._location.configuration.password
            )
        self._authentication_headers = {'Authorization': [auth_value]}
        return self._webdavRequest('HEAD', path='')

    def _authenticateSharePoint(self):
        """
        Authenticate using the SharePoint Online O365 claims-based auth.
        """
        endpoint = self._base_url.replace(
            path=('_forms', 'default.aspx'),
            query={'wa': 'wsignin1.0'},
            )
        saml = _SAML_TEMPLATE % {
            'username': self._location.configuration.username,
            'password': self._location.configuration.password,
            'endpoint': endpoint.to_uri().to_text(),
            }

        @defer.inlineCallbacks
        def cb_site_available(response):
            """
            Called when the configured site exists.
            """
            if response.code == http.NOT_FOUND:
                raise ServerException('Site not found.')

            yield self._agent.readBodyWithIgnore(response)

            # The whole authentication / redirection dance is described here
            # https://msdn.microsoft.com/en-us/library/office/hh147177.aspx
            # We go directly to the final URL without all these redirections,
            # but this might cause errors if this entry point will change.
            # Get the initial generic token.

            # We don't expect to connect to the login site to often, so there
            # is no need for a persistent connection here.
            login_url = self._LOGIN_SITE.replace(path=('extSTS.srf',))
            response = yield self._agent.post(
                url=login_url,
                headers={b'Connection': [b'close']},
                body=saml.encode('utf-8'),
                )
            defer.returnValue(response)

        def cb_got_login_response(response):
            """
            Called when we have a response from MS login.
            """
            if response.code != http.OK:
                raise ServerException(
                    'Failed to get microsoftonline.com token. {}: {}'.format(
                        response.code, force_unicode(response.phrase)))

            return self._agent.readBody(response)

        def cb_got_token(result):
            """
            Called when we have a token from MS login.
            """
            if not result:
                raise ServerException(
                    'Empty response from microsoftonline.com.')

            root = ElementTree.fromstring(result)
            body = root.find('{http://www.w3.org/2003/05/soap-envelope}Body')
            response = body.find(
                '{http://schemas.xmlsoap.org/ws/2005/02/trust}'
                'RequestSecurityTokenResponse'
                )

            if not response:
                details_xpath = (
                    './/{http://schemas.microsoft.com/Passport/'
                    'SoapServices/SOAPFault}text'
                    )
                details = body.findall(details_xpath)[0].text.strip()
                raise ServerException('Invalid credentials. %s' % (details,))

            requested_token = response.find(
                '{http://schemas.xmlsoap.org/ws/2005/02/trust}'
                'RequestedSecurityToken'
                )
            if requested_token is None:
                raise ServerException(
                    'No RequestedSecurityToken in the response from '
                    'microsoftonline.com.'
                    )

            token_element = requested_token.find(
                '{http://docs.oasis-open.org/wss/2004/01/'
                'oasis-200401-wss-wssecurity-secext-1.0.xsd}'
                'BinarySecurityToken'
                )
            if token_element is None:
                raise ServerException(
                    'No BinarySecurityToken in the response from '
                    'microsoftonline.com.')
            token = token_element.text.strip()

            # Get the cookie for the final endpoint.
            return self._agent.post(
                url=endpoint,
                headers={},
                body=token.encode('utf-8'),
                )

        @defer.inlineCallbacks
        def cb_got_sharepoint_respose(response):
            """
            Called when we have a response from SharePoint Online.
            """
            if response.code not in [http.FOUND, http.OK]:
                raise ServerException(
                    'Failed to get SharePoint Online cookie. {}: {}'.format(
                        response.code, force_unicode(response.phrase)))

            yield self._agent.readBodyWithIgnore(response)

            set_cookies = response.headers.getRawHeaders(
                'set-cookie', default=[])
            cookies = []
            for set_cookie in set_cookies:
                cookies.append(set_cookie.split(';', 1)[0])

            cookie = '; '.join(cookies)
            if not cookie:
                raise ServerException('No FedAuth cookie found.')

            self._authentication_headers = {'cookie': [cookie]}
            self._updateCredentials()

        deferred = self._agent.head(url=endpoint, headers={})
        deferred.addCallback(cb_site_available)
        deferred.addCallback(cb_got_login_response)
        deferred.addCallback(cb_got_token)
        deferred.addCallback(cb_got_sharepoint_respose)
        return deferred

    def _reauthenticate(self, method, uri, **kwargs):
        """
        Re-authenticate the client sessions.

        This makes sure that only a single request will trigger the
        re-authentication.

        All requests made to re-authenticate while the authentication is in
        progress will not trigger a new authentication request, but will
        receive the response from the ongoing authentication.
        """
        def bb_reauthentication(result_or_failure):
            """
            Called when re-authentication got something.
            """
            self._reauthentication_deferred = None
            return result_or_failure

        def cb_reauthentication(result):
            """
            Called when the re-authentication was done.
            """
            self._location.emitEvent('40031')

            requests, self._reauthentication_requests = (
                self._reauthentication_requests, [])

            for deferred in requests:
                deferred.callback(None)

        def eb_reauthentication(failure):
            """
            Called when re-authentication failed.
            """
            # Trigger the failure on all the pending requests.
            requests, self._reauthentication_requests = (
                self._reauthentication_requests, [])
            for deferred in requests:
                deferred.errback(failure)

            self._location.fail(ServerException(
                'Re-authentication failed. %s' % (
                    force_unicode(failure.value,))))

        deferred = self._queueRequestForReAuthentication(method, uri, kwargs)

        if self._reauthentication_deferred is None:
            # Trigger the re-authentication.
            reauthentication = self.authenticate()
            self._reauthentication_deferred = reauthentication
            reauthentication.addBoth(bb_reauthentication)
            reauthentication.addCallback(cb_reauthentication)
            reauthentication.addErrback(eb_reauthentication)

        return deferred

    def _queueRequestForReAuthentication(self, method, uri, kwargs):
        """
        Add a request to be triggered after the re-authentication got a result.
        """
        def cb_do_request(result, method, uri, kwargs):
            new_kwargs = self._updateAuthenticationHeaders(kwargs)
            return method(uri, **new_kwargs)

        deferred = defer.Deferred()
        deferred.addCallback(cb_do_request, method, uri, kwargs)
        self._reauthentication_requests.append(deferred)
        return deferred

    def close(self):
        """
        Called when the agent should no longer be used.
        """
        # WebDav is based on HTTP, which is stateless, so there is not
        # much to do here.
        self._onConnectionLost(ConnectionDone('Bye.'))
        return self._agent.closePersistentConnections()

    @defer.inlineCallbacks
    def exists(self, path):
        """
        `True` when path exists, `False` otherwise.

        Return a deferred that fires with the result or with a failure if the
        existence can't be checked.
        """
        response = yield self._webdavRequest('HEAD', path=path)
        yield self._agent.readBodyWithIgnore(response)

        if response.code in [http.FOUND, http.OK]:
            defer.returnValue(True)
        elif response.code in [http.NOT_FOUND]:
            defer.returnValue(False)
        else:
            raise HTTPClientException(
                'Failed to check if path exists.',
                path=path,
                code=response.code,
                phrase=response.phrase,
                )

    def _copy(self, source_path, destination_path):
        """
        Duplicate path inside the location.

        For now, this is only here to support the testing code.

        It can be a collection or non-collection.

        Return a deferred that fires when operation is done.
        """
        def cb_got_response(response):
            """
            Called when we have a response.
            """
            # CREATED is for duplicate to a new path
            # NO_CONTENT is for duplicate to an existing path.
            if response.code not in [http.CREATED, http.NO_CONTENT]:
                raise ServerException(
                    'Failed to copy {} to {}. {}: {}'.format(
                        source_path, destination_path,
                        response.code, force_unicode(response.phrase),
                        ))

        destination_uri = self._base_url.replace(
            path=get_url_segments(destination_path))
        headers = {
            'destination': [destination_uri.to_uri().to_text()],
            # SharePoint Online server will only treat this as infinity,
            # even if we pass `0` here. This value is here to show a sign
            # in the case in which the implementation is changed.
            'depth': ['0'],
            }
        deferred = self._webdavRequest(
            'COPY',
            path=source_path,
            headers=headers,
            )
        deferred.addCallback(cb_got_response)
        return deferred

    @defer.inlineCallbacks
    def rename(self, source_path, destination_path):
        """
        Rename file or folder inside the location.

        Return a deferred that fires when operation is done.
        """
        destination_url = self._base_url.replace(
            path=get_url_segments(destination_path))
        headers = {
            'destination': [destination_url.to_uri().to_text()]
            }
        response = yield self._webdavRequest(
            'MOVE',
            path=source_path,
            headers=headers,
            )

        # We always ignore the body, as we only care about the HTTP code.
        yield self._agent.readBodyWithIgnore(response)

        # CREATED is for moving to a new path.
        # NO_CONTENT is for moving to an existing path.
        if response.code not in [http.CREATED, http.NO_CONTENT]:
            raise HTTPClientException(
                'Failed to move to "%s".' % (destination_path,),
                path=source_path,
                code=response.code,
                phrase=response.phrase,
                )

    @defer.inlineCallbacks
    def _delete(self, path):
        """
        Delete folder or file at path.

        Return a deferred that fires when operation is done.
        """
        response = yield self._webdavRequest('DELETE', path=path)
        # We alwasy ignore the body, as we only care for the HTTP code.
        yield self._agent.readBodyWithIgnore(response)
        # FIXME:3948:
        # NO_CONTENT is returned even if the path was not previously there.
        # We should raise an error if path does not exist before the
        # operation.
        if response.code != http.NO_CONTENT:
            raise HTTPClientException(
                'Failed to delete.',
                path=path,
                code=response.code,
                phrase=response.phrase,
                )

    def deleteFile(self, path):
        """
        Delete file at path.
        """
        # FIXME:3948:
        # In WebDAV, delete doe not care if we have a resource or a collection.
        # But here we should prevent deleting folders.
        return self._delete(path)

    def deleteFolder(self, path):
        """
        Delete folder at path.

        Return a deferred that fires when operation is done.
        """
        # FIXME:3948:
        # In WebDAV, delete is recursive... but to keep consistency we might
        # want to prevent removing folders with members.
        return self._delete(path)

    @defer.inlineCallbacks
    def createFolder(self, path):
        """
        Create a folder at path, without creating parents.

        Return a deferred that fires when operation is done.
        """
        response = yield self._webdavRequest('MKCOL', path=path)
        # We don't care about the body, only the HTTP code.
        yield self._agent.readBodyWithIgnore(response)
        if response.code != http.CREATED:
            raise HTTPClientException(
                'Failed to create folder.',
                path=path,
                code=response.code,
                phrase=response.phrase,
                )

    def _getPathFromURI(self, uri):
        """
        Return the normalized path as found in `uri`.
        """
        if isinstance(uri, six.binary_type):
            # Can be removed once we no longer need py2.
            uri = uri.decode('ascii')
        path_parts = URL.from_text(uri).to_iri().path
        return posixpath.normpath('/' + '/'.join(path_parts))

    @defer.inlineCallbacks
    def _cbParsePROPFIND(self, response, path, error, noroot=True):
        """
        Parse the PROPFIND response.

        `path` is already normalized and URL encoded, as sent to the request.
        """
        # FIXME:6201:
        # Implement this using hyperlink.
        path = six.moves.urllib.parse.unquote(path)
        # There is a very, very strange bug.
        # it fails with '/+' but works with '+/'
        path = six.moves.urllib.parse.quote(path.encode('utf-8'), safe=b'+/')

        if response.code != http.MULTI_STATUS:
            yield self._agent.readBodyWithIgnore(response)
            details = _get_error_details(response, error)
            raise HTTPClientException(
                message=details,
                path=path,
                code=response.code,
                phrase=response.phrase,
                )

        body = yield self._agent.readBody(response)
        if not body:
            raise ServerException('No content for %s.' % (path,))

        try:
            root = ElementTree.fromstring(body)
        except ElementTree.ParseError as error:
            raise ServerException('Invalid XML response: %s' % (error))

        base_path = self._getPathFromURI(path).lower()

        result = {}
        for member in root.iter('{DAV:}response'):
            href = member.find('{DAV:}href').text.strip()
            member_path = self._getPathFromURI(href)

            # We are asking for `noroot` but some servers might ignore
            # it and still return the root.
            if noroot and member_path.lower() == base_path:
                # The base/root path is returned.
                # No need to include it in the result.
                continue

            attributes = member.find('{DAV:}propstat/{DAV:}prop')
            name = posixpath.basename(member_path)
            # Folders are signaled in SharePoint by the presence of this:
            # <D:isFolder>t</D:isFolder>.
            is_folder = attributes.find('{DAV:}isFolder') is not None
            if is_folder:
                is_file = False
                is_link = False
            else:
                is_file = True
                is_link = False

            modified = dateutil.parser.parse(
                attributes.find('{DAV:}getlastmodified').text.strip())
            # FIXME:3197:
            # Once we have the client shell, check to see what timezone
            # to show.
            utc_naive = (
                modified.replace(tzinfo=None) - modified.utcoffset())
            timestamp = (utc_naive - datetime(1970, 1, 1)).total_seconds()

            # Owner is in this format:
            # i:0#.f|membership|test@proatria.onmicrosoft.com
            owner = attributes.find(
                '{urn:schemas-microsoft-com:office:office}modifiedby')
            if owner is not None:
                owner = owner.text.strip().split('|')[-1]

            size = int(attributes.find('{DAV:}getcontentlength').text)

            if is_file and size == 0:
                # FIXME:2997:
                # Once we support non-SharePoint WebDAV servers, we should
                # check this.

                # On SharePoint Online, files which are not yet fully
                # uploaded or which have failed to be fully uploaded are
                # listed as having size 0 and `resourcetag` at version 1.
                # For now we treat these files like they don't exists.
                #
                # SharePoint Online does not support uploading empty
                # files, so we are safe for now.
                tag = None
                resource_tag = attributes.find(
                    '{http://schemas.microsoft.com/repl/}resourcetag')
                if resource_tag is not None:
                    # `resourcetag` is in the format `rt:some-uuid@00002`.
                    parts = resource_tag.text.split('@', 1)
                    if len(parts) > 1:
                        tag = int(parts[1])
                if tag == 1:
                    continue

            result[name] = FileAttributes(
                name=name,
                path=member_path,
                size=size,
                is_file=is_file,
                is_folder=is_folder,
                is_link=is_link,
                modified=timestamp,
                mode=0,
                hardlinks=1,
                owner=owner,
                )
        defer.returnValue(result)

    def getAttributes(self, path):
        """
        See: ILocation.
        """
        def cb_got_attributes(result):
            """
            Called when we got the attributes for the path.
            """
            # We asked for a depth of 0 so we should have only receive a
            # single element here.
            try:
                _, attributes = result.popitem()
            except KeyError:
                raise ServerException('No attributes returned.')
            return attributes

        headers = {
            b'depth': [b'0'],
            b'content-type': [b'text/xml'],
            }
        deferred = self._webdavRequest(
            'PROPFIND',
            path=path,
            headers=headers,
            )
        deferred.addCallback(
            self._cbParsePROPFIND,
            path=path,
            error='Failed to get path attributes.',
            noroot=False,
            )
        deferred.addCallback(cb_got_attributes)
        return deferred

    def getFolderMemberAttributes(self, path):
        """
        See: ILocation.
        """
        headers = {
            'depth': ['1,noroot'],
            'content-type': ['text/xml'],
            }
        deferred = self._webdavRequest(
            'PROPFIND',
            path=path,
            headers=headers,
            )
        deferred.addCallback(
            self._cbParsePROPFIND,
            path=path,
            error='Failed to get path content.',
            )
        return deferred

    @defer.inlineCallbacks
    def getStreamForReading(self, path):
        """
        Return a deferred with an `ITransferStream` opened for
        reading at the start of the stream.
        """
        response = yield self._webdavRequest('GET', path=path)
        if response.code != http.OK:
            # Consume the pending body to finalize receiving the response.
            yield self._agent.readBodyWithIgnore(response)
            raise HTTPClientException(
                message='Failed to download.',
                path=path,
                code=response.code,
                phrase=response.phrase,
                )

        defer.returnValue(HTTPDownloadedFile(
            path=path,
            location=self._location,
            response=response,
            ))

    def getStreamForWriting(self, path, length=None):
        """
        Return a deferred with an `ITransferStream` opened for writing,

        The HTTP upload process make it complicated to handle errors here.

        The HTTP response is sent only after we have sent all the data
        advertised in the request.
        So even if this is going to be an error, we need to let the write
        operation to perform its job so that we can see the result.

        This is why this will succeeds as long as the HTTP connection is made
        and the error is reported in the close operation.
        """
        @defer.inlineCallbacks
        def cb_got_response(response, producer):
            """
            Called when we have a response with all headers.

            By the time this is called all the data was sent.
            """
            if response.code not in [http.OK, http.CREATED]:
                # All data is already written so, will signal the failure
                # at the close operation.
                details = _get_error_details(response, 'Failed to upload.')
                yield self._agent.readBodyWithIgnore(response)
                producer.close_deferred.errback(HTTPClientException(
                    message=details,
                    path=path,
                    code=response.code,
                    phrase=response.phrase,
                    ))
                return

            # The response might have some body, so we will try to read it.
            body = yield self._agent.readBody(response)
            if body:
                # FIXME:3884:
                # SharePoint Online is stupid and will return 200 when
                # there are strange errors, like when sending in chunk
                # encoding.
                producer.close_deferred.errback(ServerException(
                    'Upload failed due to critical server error (%s). %s' % (
                        path, force_unicode(body, trim=100))))
            else:
                # File can be successfully closed.
                producer.close_deferred.callback(None)

        def eb_request(failure):
            """
            Called when the request fails.
            """
            # Any error is forwarded to the main deferred, depending on
            # the stage we are at.
            if not producer.connect_deferred.called:
                producer.connect_deferred.errback(failure)

            if not producer.close_deferred.called:
                producer.close_deferred.errback(failure)

        producer = HTTPUploadedFile(
            location=self._location,
            path=path,
            length=length,
            )
        deferred = self._webdavRequest(
            'PUT',
            path=path,
            body=producer,
            )
        deferred.addCallback(cb_got_response, producer)
        deferred.addErrback(eb_request)

        # The request deferred will only be called when all the data was
        # uploaded.
        # We don't wait for that but return a deferred which signal that the
        # connection is ready to receive data.
        return producer.connect_deferred

    def getStreamForAppending(self, path):
        """

        Creates the file if it does not exist.
        """
        return defer.fail('Appending is not yet supported in WebDAV.')
