# Copyright (c) 2011 Adi Roiban.
# See LICENSE for details.
"""
An account as it is seen and used by the services after the authentication
process.

No general authentication related stuff should be placed here.

Authorization related stuff should be placed here, like selective access to
virtual folders or permission for different paths.

Accounts can be configured in many ways, but with an avatar they are exposed
using an unified interface.
"""
import os
import posixpath
import re
from email.utils import getaddresses
from collections import OrderedDict
from uuid import uuid4

import pem
from cryptography import x509
from twisted.internet.address import IPv4Address
from zope.interface import implementer

from chevah_compat import (
    HasImpersonatedAvatar,
    local_filesystem,
    process_capabilities,
    system_users,
    )
from chevah_compat.exceptions import CompatError
from chevah.server import force_unicode
from chevah.server.authentication.base import password_handler
from chevah.server.commons import DEFAULT_BACKEND
from chevah.server.commons.constant import (
    ADDRESS_UNKNOWN,
    CONFIGURATION_DEFAULT_OWNER_PLACEHOLDER,
    CONFIGURATION_DEFAULT_GROUP_PLACEHOLDER,
    CONFIGURATION_DISABLED_VALUES,
    CONFIGURATION_HOME_PLACEHOLDER,
    CONFIGURATION_SHARED_PLACEHOLDER,
    CONFIGURATION_USERNAME_PLACEHOLDER,
    PERMISSIONS,
    TYPE_NAME,
    )
from chevah.server.commons.exception import (
    ConfigurationException,
    ServerError,
    ServerException,
    )
from chevah.server.commons.interface import (
    IOSAccountConfiguration,
    IPublicAvatar,
    IServiceAvatar,
    )
from chevah.server.commons.match import get_matched_groups
from chevah.server.configuration.model import (
    get_relaxed_datetime,
    is_inherited,
    is_inherited_list_of_lists,
    )


def _uuid4_stdlib(self):
    """
    Return UUID4 result based on stdlib implementation.
    """
    return force_unicode(uuid4())


class _AvatarMixin(object):
    """
    Shared code by various avatar flavours.
    """

    def _getOwnValueOrInherited(self, name):
        """
        Return the value as configured directly on the account or as
        inherited from the primary/main group.
        """
        own = getattr(self._me, name)
        if not is_inherited(own):
            # Not an inherited value. Just use it.
            return own

        inherited = getattr(self._groups[0], name)
        if is_inherited(inherited):
            # The value from group can't be an inherit marker.
            data = {
                "property_name": name,
                "group_name": self._groups[0].name,
                }
            raise ServerError("20022", data=data)

        return inherited


def resolve_home_folder(account, group):
    """
    Return the home folder path for `account` as resolved using `group`
    inheritance if needed.
    """
    path = account.home_folder_path

    if not path:
        raise ServerException('No value defined for "home_folder_path".')

    if not is_inherited(path):
        # Have a direct, non-inherited path.
        return path

    # Return the inherited value for home_folder_path.
    # Home path is inherited only from the main/primary group.
    path = group.home_folder_path

    if not path:
        # No path defined for the group. This mean that this group
        # require all accounts to define their own path.
        raise ServerException('Group home folder path is disabled.')

    # Try to expand the inherited value.
    # Expansion is skipped if we have a special ${OS_HOME} value for the path.
    # in the group inheritance.
    # The OS_HOME is expanded in OS specific code.
    if path.lower() in [CONFIGURATION_HOME_PLACEHOLDER.lower()]:
        return path

    return expand_group_path(path, account)


def expand_group_path(path, account):
    """
    Return the expanded version of `path` based on `account` username.
    """
    expansion = _get_user_expansion(account)

    if path.upper().startswith(CONFIGURATION_SHARED_PLACEHOLDER):
        return path[len(CONFIGURATION_SHARED_PLACEHOLDER):].strip()

    if path.find(CONFIGURATION_USERNAME_PLACEHOLDER) != -1:
        return path.replace(
            CONFIGURATION_USERNAME_PLACEHOLDER, expansion)

    if path.find(CONFIGURATION_USERNAME_PLACEHOLDER.lower()) != -1:
        return path.replace(
            CONFIGURATION_USERNAME_PLACEHOLDER.lower(), expansion)

    return os.path.join(path, expansion)


def _get_user_expansion(account):
    """
    Return the home folder name for current avatar as it automatically
    created by default by the operating system.
    """
    username = account.name
    # Replace path separators and any characters not supported by
    # Windows with a .
    # Windows path restriction: `\/:*?"<>|`
    username = re.sub('[\\\\/:*?"<>|]', '.', username)

    if not IOSAccountConfiguration.providedBy(account):
        # Not an OS account, so there are no other changes to be done.
        return username

    if process_capabilities.os_name != 'windows':
        # On non-Windows we allow any characters in OS user names and
        # translate them as they are.
        # With the exception of path restriction.
        return username

    # Windows don't allow certain characters in usernames.
    # Windows username restriction: `\/"[]:|<>+=;,?*@`
    # So if we have an `@` for an OS account it means that this is an
    # domain account, and we translate it in a similar way as Windows
    # does.
    if '@' in username:
        parts = username.split('@', 1)
        return parts[0] + '.' + parts[1].upper()

    # Just a normal Windows OS account.
    return username


def get_virtual_folders(account, groups):
    """
    Return an internal representation for the virtual folders.

    This is a dictionary with virtual path as the key and a tuple of
    (virtual_segments, real_path, is_user_placeholder).
    """
    folders = account.virtual_folders

    if is_inherited_list_of_lists(folders):
        folders = []
        for group in groups:
            folders.extend(group.virtual_folders)

    # We try to preserve the order to simplify testing.
    # We also use the dictionary keys to detect duplicate virtual paths.
    result = OrderedDict()
    for parts in folders:
        try:
            virtual_path, real_path = parts
        except ValueError:
            raise ServerException(
                'Virtual folder definition needs a virtual path and '
                'a target path.')

        # Insert the current username in the virtual and real path.
        is_user_placeholder = False
        if CONFIGURATION_USERNAME_PLACEHOLDER in real_path:
            is_user_placeholder = True

        virtual_path = virtual_path.replace(
            CONFIGURATION_USERNAME_PLACEHOLDER, account.name)
        real_path = real_path.replace(
            CONFIGURATION_USERNAME_PLACEHOLDER, account.name)

        # Be more relaxed about path separators as Windows users might
        # be familiar with back slashes.
        virtual_path = virtual_path.replace('\\', '/')
        # Also, if a path is not defined as an absolute path,
        # make it an absolute, as some users might just type the
        # virtual folder name.
        if not virtual_path.startswith('/'):
            virtual_path = '/' + virtual_path

        is_target_absolute = local_filesystem.isAbsolutePath(real_path)
        if not is_target_absolute:
            raise ServerException(
                'Virtual path target needs to be an absolute path '
                '"%s".' % (real_path,))

        virtual_path = posixpath.abspath(virtual_path)
        virtual_segments = [s for s in virtual_path.split('/') if s]

        if virtual_path in result:
            raise ServerException(
                'Duplicate virtual path "%s".' % (virtual_path,))

        result[virtual_path] = (
            virtual_segments, real_path, is_user_placeholder)

    return result


@implementer(IServiceAvatar)
class ServiceAvatar(HasImpersonatedAvatar, _AvatarMixin):
    """
    All configurations for an account as they are used in the
    application after the account was authenticated.

    It looks to the parent or group inherited configuration.
    For account configurations that are inherited a method is provided for
    getting the computed value.

    It does not allow changing of values.
    Values can be changed by updating the `AccountConfiguration` for account
    or for the groups.

    It should not do more than providing high level configuration.

    High level configuration includes resolving all the inheritance, value
    expansion or resolution and providing ready to use configuration values.

    This should not contain any attribute which is used for authentication or
    validation of the identity / authenticity of the account.

    Other operations should be done in specialized avatars or protocol specific
    avatars or protocol helpers.
    """

    # Here to help with dependency injection during tests.
    _uuid4 = _uuid4_stdlib

    def __init__(
        self,
        account_configuration, groups, component, peer=None, token=None,
            ):

        # FIXME:4403:
        # This should be refactored to be non-mutable, that is if config change
        # the instance will not change and should not have references to the
        # config or anything else.
        # Or to allow getting a non-mutable copy of it.
        self._me = account_configuration
        self._groups = groups

        # Here in case the configuration is lost before this instance is
        # discarded and the instance is still used for audit purposes.
        self._initial_name = self._me.name

        self._component = component
        self._peer = peer
        self._token = token

        self._emails = ()
        self._amend_write_name = None
        self._permissions = self._getValidatedPermissions()
        self._as2_certificates = self._getValidatedAS2Certificates()
        self._validateConfiguration()

    def _validateConfiguration(self):
        """
        Make sure avatar had valid configuration.
        """
        self._emails = getaddresses([self._me.email])
        self._amend_write_name = self._getOwnValueOrInherited(
            'amend_write_name')

        if self._amend_write_name and self._amend_write_name != 'uuid-prefix':
            raise ServerException(
                'Unknown amend_write_name "%s".' % (self._amend_write_name,))

    def __repr__(self):
        result = '%d:%s: for "%s"' % (
            id(self),
            self.__class__,
            self.name,
            )
        return result

    @property
    def token(self):
        """
        See :class:`IServiceAvatar`.
        """
        return self._token

    @property
    def use_impersonation(self):
        """
        Return True if avatar operations should be impersonated.
        """
        if IOSAccountConfiguration.providedBy(self._me):
            return True
        else:
            return False

    @property
    def username(self):
        '''Here due to ChevahAvatar.'''
        return self.name

    @property
    def home_folder(self):
        '''Here due to ChevahAvatar.'''
        return self.home_folder_path

    @property
    def root_folder(self):
        '''Here due to ChevahAvatar.'''
        return self.root_folder_path

    @property
    def peer(self):
        '''Return the peer for which this avatar was created.'''
        return self._peer

    @property
    def name(self):
        """
        Return the name of this account.
        """
        try:
            return self._me.name
        except ConfigurationException:
            return self._initial_name

    @property
    def uuid(self):
        """
        Return the UUID of this account.
        """
        return self._me.uuid

    @property
    def email(self):
        """
        Return the email of this account.

        This is here for backward compatibility,
        when we only supported a single address and only the address
        itself, without an associated name.
        """
        if not self.emails:
            return ''

        return self.emails[0][1]

    @property
    def emails(self):
        """
        Return the list of emails of this account.
        """
        return self._emails

    @property
    def created(self):
        """
        Return the creation `datetime` of this account.
        """
        return get_relaxed_datetime(self._me.created, 'Account created')

    @property
    def component(self):
        """
        See: `IServiceAvatar`.
        """
        return self._component

    @property
    def lock_in_home_folder(self):
        """
        Return True if account should be locked in the home folder.

        Default `True`.
        """
        if not IOSAccountConfiguration.providedBy(self._me):
            return True
        return self._getOwnValueOrInherited('lock_in_home_folder')

    @property
    def home_folder_path(self):
        """
        Path to home folder for this account.
        """
        try:
            path = resolve_home_folder(self._me, self._groups[0])
        except ServerException as error:
            raise self._getHomeFolderPathError(error.message)

        if path.lower() == CONFIGURATION_HOME_PLACEHOLDER.lower():
            if IOSAccountConfiguration.providedBy(self._me):
                try:
                    path = system_users.getHomeFolder(
                        username=self.name, token=self.token)
                except CompatError as error:
                    raise self._getHomeFolderPathError(error.message)
            else:
                raise self._getHomeFolderPathError(
                    'Invalid home_folder_path configuration for account. '
                    'OS_HOME folder can only be used for OS accounts.'
                    )

        if not local_filesystem.isAbsolutePath(path):
            raise self._getHomeFolderPathError(
                'Value for "home_folder_path" should be absolute path. '
                'Got "%s".' % (path,))

        return local_filesystem.getAbsoluteRealPath(path)

    def _getHomeFolderPathError(self, details):
        """
        Helper for creating a configuration error when home_folder_path
        could not be retrieved.

        This is here so that we can raise this exception with an unique ID
        from a single place.
        """
        data = {
            'details': details,
            'avatar': self,
            }
        return ServerError('20067', data=data)

    @property
    def root_folder_path(self):
        """
        Return path to home folder if locked in home folder or `None`
        otherwise.
        """
        if self.lock_in_home_folder:
            return self.home_folder_path
        else:
            return None

    @property
    def create_home_folder(self):
        """
        Return true if missing home folder should be created for this
        account.
        """
        return self._getOwnValueOrInherited('create_home_folder')

    @property
    def create_home_folder_owner(self):
        '''Return the name of owner for the newly created home folder.

        Return `None` if the owner should not be changed.
        Always return `None` for non-OS accounts.

        If DEFAULT_OWNER_PLACEHOLDER is present, it will be expanded to
        the name of this account.
        '''
        if not IOSAccountConfiguration.providedBy(self._me):
            return None

        value = self._getOwnValueOrInherited('create_home_folder_owner')

        if value == CONFIGURATION_DEFAULT_OWNER_PLACEHOLDER:
            value = self.name

        return value

    @property
    def create_home_folder_group(self):
        '''Return the name of group for the newly created home folder.

        Return None if the group should not be changed.
        Always return `None` for application accounts.
        '''
        if not IOSAccountConfiguration.providedBy(self._me):
            return None

        value = self._getOwnValueOrInherited('create_home_folder_group')

        if value == CONFIGURATION_DEFAULT_GROUP_PLACEHOLDER:
            # For OS account we get the default value from the system.
            # App accounts should not need to change owner.
            try:
                value = system_users.getPrimaryGroup(username=self.name)
            except CompatError as error:
                data = {
                    'avatar': self,
                    'details': error.message,
                    }
                raise ServerError('20043', data=data)

        return value

    @property
    def required_folders(self):
        """
        See: IServiceAvatar,
        """
        result = []
        result.extend(self._getHomeStructure())
        result.extend(self._getVirtualStructure())
        return result

    def _getHomeStructure(self):
        """
        Return the list of paths that should be created as part of the
        `home_folder_structure` configuration.
        """
        # For now, the home folder structure is only defined on the group.
        raw_structure = self._me.home_folder_structure
        if is_inherited(raw_structure):
            raw_structure = []
            for group in self._groups:
                raw_structure.extend(group.home_folder_structure)

        result = []

        if not raw_structure:
            # Nothing configured,
            return result

        structure = []

        # Do an initial sanitization by normalizing the path to Unix
        # relative paths.
        for path in raw_structure:
            path = path.strip('/\\').strip()
            path = path.replace('\\', '/')
            if not path:
                continue
            structure.append(path)

        auto_created = []
        for path in structure:
            parts = path.split('/')
            if len(parts) == 1:
                continue
            base_dir = ''
            for sub_dir in parts:
                base_dir = base_dir + '/' + sub_dir
                base_dir = base_dir.strip('/')
                if base_dir not in structure:
                    auto_created.append(base_dir)

        structure = auto_created + structure
        home_path = self.home_folder_path

        # The path is configured as relative to the home folder, so we need
        # to expand it.
        for path in structure:
            if local_filesystem.isAbsolutePath(path):
                raise ServerException(
                    'Absolute path "%s" is not allowed as part of the '
                    'home folder structure configuration.' % (path,))

            path = os.path.join(home_path, path)
            path = local_filesystem.getAbsoluteRealPath(path)

            if path in result:
                # Don't add duplicate names.
                continue

            result.append(path)

        return result

    def _getVirtualStructure(self):
        """
        Return the list of paths that should be created as part of the
        virtual folders.

        But only if account is configured to create missing home folders.
        """
        result = []
        if not self.create_home_folder:
            # This only takes into consideration the primary group config.
            return result

        for parts in get_virtual_folders(self._me, self._groups).values():
            is_user_placeholder = parts[2]
            if not is_user_placeholder:
                continue
            real_path = parts[1]
            result.append(real_path)

        return result

    @property
    def virtual_folders(self):
        """
        See: IServiceAvatar.

        Validate the configuration and return the values as used by the
        chevah/compat layer.
        """
        result = get_virtual_folders(self._me, self._groups)
        # For compat we only need the first 2 values.
        return [(v[0], v[1]) for v in result.values()]

    @property
    def file_write_size(self):
        """
        See: IServiceAvatar.
        """
        return self._getOwnValueOrInherited('file_write_size')

    @property
    def permissions(self):
        """
        See: IServiceAvatar.
        """
        return self._permissions

    def _getValidatedPermissions(self):
        """
        Return the permissions which are valid.
        """
        # Start with own permissions.
        permissions = self._me.permissions

        if is_inherited_list_of_lists(permissions):
            # If inherited, start with the permissions of the primary group,
            # including the default permissions.
            permissions = list(self._groups[0].permissions)
            for group in self._groups[1:]:
                # We ignore the default permission rule of secondary groups.
                permissions.extend(group.permissions[1:])

        result = []
        # Validate global permissions.
        global_permissions = []
        for raw_permission in permissions[0]:
            raw_permission = raw_permission.strip().lower()
            if raw_permission not in PERMISSIONS:
                raise ServerException(
                    'Unknown permission "%s"' % (raw_permission,))
            global_permissions.append(raw_permission)

        result.append(global_permissions)

        # Validate per path permissions.
        for path_permissions in permissions[1:]:
            if len(path_permissions) < 2:
                raise ServerException(
                    'Path permissions needs the rule followed by at '
                    'least one permission.',
                    )

            # We get the configuration value and normalized it for internal
            # usage.

            # Expand the username placeholder.
            target_path = path_permissions[0].replace(
                CONFIGURATION_USERNAME_PLACEHOLDER, self.name)

            # Start the rule with the path name.
            sanitized_permissions = [target_path]
            # Check path matching rule.
            get_matched_groups(
                value='', rule=sanitized_permissions[0], usage='permissions')

            for raw_permission in path_permissions[1:]:
                raw_permission = raw_permission.strip().lower()
                if raw_permission not in PERMISSIONS:
                    raise ServerException(
                        'Unknown permission "%s"' % (raw_permission,))
                sanitized_permissions.append(raw_permission)

            result.append(sanitized_permissions)

        return result

    def getAmendedUploadSegments(self, segments):
        """
        See: IServiceAvatar.
        """
        result = segments[:]

        if not self._amend_write_name:
            # Nothing should be changed.
            return (result, False)

        if len(result) < 1:
            raise ServerException(
                'Invalid path requested for amending "%r".' % (result,))

        # For now only prepend the uuid is implemented.
        result[-1] = '%s-%s' % (self._uuid4(), result[-1])

        return (result, True)

    @property
    def can_change_password(self):
        """
        Return True if avatar can change its password.
        """
        return self._getOwnValueOrInherited('allow_own_password_change')

    def updatePassword(self, current, new, event_emitter):
        """
        See: IServiceAvatar.
        """
        emitEvent = event_emitter
        try:
            self._updatePassword(current, new)
            emitEvent('20085')
            return True
        except ServerError as error:
            from chevah.server import emit_from_error
            emit_from_error(error, avatar=self, emitter=event_emitter)
            return False
        except ServerException as error:
            emitEvent('20086', data={'details': error.message})
            return False

    def _updatePassword(self, current, new):
        """
        Low level password update implementation.
        """
        if not self.can_change_password:
            raise ServerException(
                'Account not allowed to update its password.')

        if not current:
            raise ServerException('Current password can\'t be empty.')

        if not new:
            raise ServerException('New password can\'t be empty.')

        account_password = self._me.password
        if not account_password:
            raise ServerException('The password for this account is disabled.')

        try:
            current_is_valid = password_handler.verify(
                current, account_password)
        except ValueError as error:
            raise ServerException(
                'Failed to check current password. %s' % (error,))

        if not current_is_valid:
            raise ServerException('Current password doesn\'t match.')

        if new.lower() in CONFIGURATION_DISABLED_VALUES:
            raise ServerException('User not allowed to disable own password.')

        # Save state before update.
        old_hash = self._me.password
        old_last_password_update = self._me.last_password_update

        # If this fail, the last update is not changed.
        self._me.password_update = new

        try:
            self._me._proxy.save()
        except Exception:
            self._me._setRawPassword(old_hash)
            self._me.last_password_update = old_last_password_update
            raise

    def isMember(self, group_uuid):
        """
        Return `True` if is a member of group with uuid.
        """
        for group in self._groups:
            if group_uuid == group.uuid:
                return True
        return False

    def _getValidatedAS2Certificates(self):
        """
        Return a list with certificate objects.
        """
        result = []
        raw = self._me.as2_certificates
        if not raw:
            return result

        for pem_object in pem.parse(raw.encode('utf-8')):
            if not isinstance(pem_object, pem.Certificate):
                continue
            content = pem_object.as_bytes()

            try:
                cert = x509.load_pem_x509_certificate(content, DEFAULT_BACKEND)
                # Do some validation to check the cert is ok.
                cert.subject.rfc4514_string()
                cert.issuer.rfc4514_string()

            except Exception:
                raise ServerException(
                    'The account is configured with invalid '
                    '"as2_certificates" value. Unable to load certificate.'
                    )
            result.append(cert)

        return result

    def getAS2Certificates(self):
        """
        See IServiceAvatar.
        """
        return self._as2_certificates

    def getAS2MDNCA(self):
        """
        See IServiceAvatar.
        """
        value = self._me.as2_async_mdn_ca.strip()

        if not value:
            raise ServerException(
                'AS2 async MDN responses are disabled for this account.')

        if value.lower() in CONFIGURATION_DISABLED_VALUES:
            # Return an explicit value to disable peer cert.
            return TYPE_NAME.DISABLE_IDENTITY_SECURITY

        return value


@implementer(IPublicAvatar)
class ProcessAvatar(object):
    """
    Avatar representing the server process.
    """

    def __init__(self):
        self._peer = ADDRESS_UNKNOWN

    @property
    def component(self):
        """
        See: `IPublicAvatar`.
        """
        return None

    @property
    def name(self):
        """
        See: `IPublicAvatar`.
        """
        return 'Process'

    @property
    def uuid(self):
        """
        See: `IPublicAvatar`.
        """
        return 'Process'

    @property
    def peer(self):
        """
        See: `IPublicAvatar`.
        """
        return self._peer

    @property
    def email(self):
        """
        No email address implemented for this avatar.
        """
        return ''

    @property
    def emails(self):
        """
        No email address implemented for this avatar.
        """
        return []

    def __eq__(self, candidate):
        """
        All `ProcessAvatar` instances are equal.
        """
        return isinstance(candidate, ProcessAvatar)

    def __ne__(self, candidate):
        """
        See: `__eq__`.
        """
        return not self.__eq__(candidate)


@implementer(IPublicAvatar)
class TransferAvatar(object):
    """
    The avatar used by a transfer.

    This used for the events that are not associated with a job.

    The transfer avatar might be associated with a remote location and for
    those actions it will use the location peer value.
    """

    def __init__(self, uuid, name, component, peer):
        self._uuid = uuid
        self._name = name
        self._peer = peer
        self._component = component

    @property
    def component(self):
        """
        See: `IPublicAvatar`.
        """
        return self._component

    @property
    def name(self):
        """
        See: `IPublicAvatar`.
        """
        return self._name

    @property
    def uuid(self):
        """
        See: `IPublicAvatar`.
        """
        return self._uuid

    @property
    def peer(self):
        """
        See: `IPublicAvatar`.
        """
        return self._peer

    @property
    def email(self):
        """
        No email address implemented for this avatar.
        """
        return ''

    @property
    def emails(self):
        """
        No email address implemented for this avatar.
        """
        return []

    def __eq__(self, candidate):
        """
        All `TransferAvatar` instances are equal if they have the same UUID.
        """
        if not isinstance(candidate, TransferAvatar):
            return False

        return (
            self.uuid == candidate.uuid
            and self.name == candidate.name
            and self.peer == candidate.peer
            and self.component == candidate.component
            )

    def __ne__(self, candidate):
        """
        See: `__eq__`.
        """
        return not self.__eq__(candidate)


@implementer(IPublicAvatar)
class JobAvatar(object):
    """
    The avatar used by a transfer job.
    """

    def __init__(self, uuid, parent):
        self._uuid = uuid
        self._peer = ADDRESS_UNKNOWN
        self._component = parent

    @property
    def component(self):
        """
        See: `IPublicAvatar`.
        """
        return self._component

    @property
    def name(self):
        """
        See: `IPublicAvatar`.
        """
        return self._uuid

    @property
    def uuid(self):
        """
        See: `IPublicAvatar`.
        """
        return self._uuid

    @property
    def peer(self):
        """
        See: `IPublicAvatar`.
        """
        return self._peer

    @property
    def email(self):
        """
        No email address implemented for this avatar.
        """
        return ''

    @property
    def emails(self):
        """
        No email address implemented for this avatar.
        """
        return []

    def __eq__(self, candidate):
        """
        All `JobAvatar` instances are equal.
        """
        if not isinstance(candidate, JobAvatar):
            return False

        return self.uuid == candidate.uuid

    def __ne__(self, candidate):
        """
        See: `__eq__`.
        """
        return not self.__eq__(candidate)


@implementer(IPublicAvatar)
class PeerAvatar(object):
    """
    A remote peer (host/port) avatar.
    """

    def __init__(self, peer, component=None):
        self._peer = peer
        self._component = component

    @property
    def name(self):
        """
        See: `IPublicAvatar`.
        """
        return 'Unknown'

    @property
    def uuid(self):
        """
        See: `IPublicAvatar`.
        """
        return 'Unknown'

    @property
    def component(self):
        """
        See: `IPublicAvatar`.
        """
        return self._component

    @property
    def peer(self):
        """
        See: `IPublicAvatar`.

        IP address and port number of the remote peer.
        """
        return self._peer

    @property
    def email(self):
        """
        No email address implemented for this avatar.
        """
        return ''

    @property
    def emails(self):
        """
        No email address implemented for this avatar.
        """
        return []

    def __eq__(self, candidate):
        """
        Return `True` if candidate is an instance of `PeerAvatar` and the
        name host/port match.
        """
        if not isinstance(candidate, PeerAvatar):
            return False

        return self.peer == candidate.peer

    def __ne__(self, candidate):
        """
        See: `__eq__`.
        """
        return not self.__eq__(candidate)

    def __repr__(self):
        return '%s: %s' % (type(self), self.peer)


@implementer(IPublicAvatar)
class ClientAvatar(PeerAvatar):
    """
    An avatar for a client connected to a remote server.
    """

    def __init__(self, credentials):
        self._credentials = credentials
        configuration = self._credentials.creator.configuration
        self._peer = IPv4Address(
            type='TCP', host=configuration.address, port=configuration.port)

    @property
    def component(self):
        """
        See: `IPublicAvatar`.
        """
        return self._credentials.creator

    @property
    def name(self):
        """
        See: `IPublicAvatar`.
        """
        return self._credentials.username

    @property
    def uuid(self):
        """
        See: `IPublicAvatar`.
        """
        return self._credentials.username

    @property
    def peer(self):
        """
        See: `IPublicAvatar`.

        IP address and port number of the remote server.
        """
        return self._peer

    @property
    def email(self):
        """
        No email address implemented for this avatar.
        """
        return ''

    @property
    def emails(self):
        """
        No email address implemented for this avatar.
        """
        return []

    def __eq__(self, candidate):
        """
        Return `True` if candidate is holding the same values.
        """
        if not isinstance(candidate, ClientAvatar):
            return False

        return (
            self.peer == candidate.peer and
            self.uuid == candidate.uuid
            )

    def __ne__(self, candidate):
        """
        See: `__eq__`.
        """
        return not self.__eq__(candidate)

    def __repr__(self):
        return '%s: %s-%s' % (type(self), self.uuid, self.peer)


class FilesystemAvatar(ClientAvatar):
    """
    The avatar used by a local filesystem.
    """

    def __init__(self, component):
        self._peer = ADDRESS_UNKNOWN
        self._component = component

    @property
    def component(self):
        """
        See: `IPublicAvatar`.
        """
        return self._component

    @property
    def name(self):
        """
        See: `IPublicAvatar`.
        """
        return self._component.name

    @property
    def uuid(self):
        """
        See: `IPublicAvatar`.
        """
        return self._component.uuid

    def __eq__(self, candidate):
        """
        All `JobAvatar` instances are equal.
        """
        if not isinstance(candidate, FilesystemAvatar):
            return False

        return self.uuid == self._component.uuid
