# Copyright (c) 2011 Adi Roiban.
# See LICENSE for details.
"""
Accounts, groups, administrators and roles as defined in the SFTPPlus own
configuration.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
from zope.interface import implementer

from chevah.server.commons.constant import (
    CONFIGURATION_DEFAULT_GROUP,
    CONFIGURATION_DEFAULT_ROLE,
    CONFIGURATION_INHERIT,
    CONFIGURATION_SECTION_ACCOUNTS,
    CONFIGURATION_SECTION_ADMINISTRATORS,
    CONFIGURATION_SECTION_GROUPS,
    CONFIGURATION_SECTION_ROLES,
    ROLE_PERMISSION,
    TYPE_NAME,
    )
from chevah.server.commons.interface import (
    IApplicationAdministratorConfiguration,
    IApplicationAccountConfiguration,
    IGroupConfiguration,
    IOSAccountConfiguration,
    IRoleConfiguration,

    )
from chevah.server.configuration.option import (
    EntityConfiguration,
    PasswordConfigurationMixin,
    )

from chevah.server.configuration.model import (
    DefaultOnly,
    WritableBoolean,
    WritableBooleanWithInherit,
    WritableInteger,
    WritableIntegerWithInherit,
    WritableSSVLines,
    WritableLines,
    WritableNumber,
    WritableString,
    WritableStringOrEmpty,
    WritableStringOrEmptyCaseInsensitive,
    WritableCSVLines,
    WritableCSVLinesUnique,
    WritableCSVUnique,
    WritableCSVInsensitive,
    _convert_ordered_sequence_to_non_duplicate_string,
    _convert_string_to_non_duplicate_tuple,
    )


class _AccountConfigurationAbstract(EntityConfiguration):
    """
    All configurations for an account as they are persisted inside our own
    configuration.

    It provides raw (unexpanded, non-inherited) access to user settings.
    """

    SECTION_PREFIX = CONFIGURATION_SECTION_ACCOUNTS

    @property
    def kind_name(self):
        """
        Human readable name for the type of this account configuration.
        """
        # Default to type.
        return self.type

    @property
    def type(self):
        """
        Generic type: OS or application.
        """
        raise NotImplementedError('kind_name not implemented')

    # FIXME:5585:
    # This should be renamed to `groups` while reading it as `group`.
    group = WritableCSVUnique(
        'group',
        default=(CONFIGURATION_DEFAULT_GROUP,), inherit=False)

    @property
    def create_home_folder(self):
        '''Return True if the home folders are created for users.'''
        return self._proxy.getBooleanOrInherit(
            self._section_name, 'create_home_folder')

    @create_home_folder.setter
    def create_home_folder(self, value):
        '''Set value for create_home_folder.'''
        self._proxy.setBooleanOrInherit(
            self._section_name, 'create_home_folder', value)

    @property
    def create_home_folder_owner(self):
        '''Return default owner for creating home folders.'''
        return self._proxy.getStringOrInherit(
            self._section_name, 'create_home_folder_owner')

    @create_home_folder_owner.setter
    def create_home_folder_owner(self, value):
        '''Set value for create_home_folder_owner.'''
        return self._proxy.setStringOrInherit(
            self._section_name, 'create_home_folder_owner', value)

    @property
    def create_home_folder_group(self):
        '''Return default group for creating home folders.'''
        return self._proxy.getStringOrInherit(
            self._section_name, 'create_home_folder_group')

    @create_home_folder_group.setter
    def create_home_folder_group(self, value):
        '''Set value for create_home_folder_group.'''
        self._proxy.setStringOrInherit(
            self._section_name, 'create_home_folder_group', value)

    home_folder_path = WritableStringOrEmpty('home_folder_path', inherit=True)

    @property
    def home_folder_structure(self):
        """
        For now, the structure is defined only on groups.
        """
        return CONFIGURATION_INHERIT[0]

    @property
    def allow_own_password_change(self):
        """
        For now, allowing to change the password is defined only on groups.
        """
        return CONFIGURATION_INHERIT[0]

    ssh_authorized_keys_path = WritableStringOrEmpty(
        'ssh_authorized_keys_path', inherit=True)

    ssh_authorized_keys_content = WritableStringOrEmpty(
        'ssh_authorized_keys_content')

    source_ip_filter = WritableSSVLines('source_ip_filter')

    @property
    def allow_certificate_authentication(self):
        '''
        Get allow_certificate_authentication.
        '''
        return self._proxy.getBooleanOrInherit(
            self._section_name, 'allow_certificate_authentication')

    @allow_certificate_authentication.setter
    def allow_certificate_authentication(self, value):
        '''
        Set allow_certificate_authentication.
        '''
        self._proxy.setBooleanOrInherit(
            self._section_name, 'allow_certificate_authentication', value)

    expire_datetime = WritableString('expire_datetime', default='')
    created = WritableString('created', default='')
    # There can be multiple emails define.
    # We keep the configuration as text to simplify the UI.
    # Later we might add UI to multiple emails and have this as CSV
    email = WritableString('email', default='')

    as2_certificates = WritableString('as2_certificates', default='')
    as2_async_mdn_ca = WritableString('as2_async_mdn_ca', default='')
    as2_require_http_authentication = WritableBooleanWithInherit(
        'as2_require_http_authentication')

    amend_write_name = WritableStringOrEmptyCaseInsensitive(
        'amend_write_name')

    permissions = WritableCSVLinesUnique(
        'permissions', default=((CONFIGURATION_INHERIT[0],),))

    virtual_folders = WritableCSVLines('virtual_folders')
    # If we want to support something like OpenSSH where 2 SSH keys are
    # required, we need to allow duplicate values.
    required_credentials = WritableCSVInsensitive(
        'required_credentials', default=(CONFIGURATION_INHERIT[0],))

    disable_on_inactivity = WritableIntegerWithInherit(
        'disable_on_inactivity', default=CONFIGURATION_INHERIT[0])

    file_write_size = WritableIntegerWithInherit(
        'file_write_size', default=CONFIGURATION_INHERIT[0])


@implementer(IApplicationAccountConfiguration)
class ApplicationAccountConfiguration(
        _AccountConfigurationAbstract, PasswordConfigurationMixin):
    """
    Configuration for an application account defined in the configuration.
    """

    @property
    def type(self):
        """
        The type is read only.

        It can't be updated at runtime.

        The type value is taken into consideration when a new property is
        created, but this is done via SetWithTypeConfiguration.
        """
        return TYPE_NAME.APPLICATION

    @property
    def lock_in_home_folder(self):
        '''Return True if the users are locked in their home folder.'''
        return True

    @lock_in_home_folder.setter
    def lock_in_home_folder(self, value):
        """
        Does nothing. Here to help with uniform interface.
        """

    password_lifetime = WritableIntegerWithInherit(
        'password_lifetime', default=CONFIGURATION_INHERIT[0])


@implementer(IOSAccountConfiguration)
class OSAccountConfiguration(_AccountConfigurationAbstract):
    """
    Properties for an operating system account defined in the
    configuration.
    """

    @property
    def type(self):
        """
        It's always OS.
        """
        return TYPE_NAME.OS

    @property
    def lock_in_home_folder(self):
        '''Return True if the users are locked in their home folder.'''
        return self._proxy.getBooleanOrInherit(
            self._section_name, 'lock_in_home_folder')

    @lock_in_home_folder.setter
    def lock_in_home_folder(self, value):
        '''Set value for lock users in their home folder.'''
        self._proxy.setBooleanOrInherit(
            self._section_name, 'lock_in_home_folder', value)

    @property
    def allow_own_password_change(self):
        """
        OS accounts can have their password changed.
        """
        return False

    @property
    def last_password_update_v3(self):
        """
        We don't know when external password was updated.
        """
        return 0

    # We don't know when password was changed for OS accounts.
    last_password_update = DefaultOnly(0)
    password_lifetime = DefaultOnly(0)
    # OS accounts don't have MFA.
    multi_factor_authentication = DefaultOnly('')


@implementer(IGroupConfiguration)
class GroupConfiguration(EntityConfiguration):
    """
    Configuration for a group.

    Groups are there to provide default values for account configuration.
    They have almost all properties of an account configuration.

    Groups should not use 'Inherit' values.

    GroupConfiguration is very similar to _AccountConfigurationAbstract, but
    due to the way @properties are (not) inherited in Python, this requires
    a complete rewrite so no inheritance is used here.
    """

    SECTION_PREFIX = CONFIGURATION_SECTION_GROUPS

    @property
    def kind_name(self):
        """
        See `IGroupConfiguration`.
        """
        return 'group'

    home_folder_path = WritableStringOrEmpty('home_folder_path')
    create_home_folder = WritableBoolean('create_home_folder')
    create_home_folder_owner = WritableString('create_home_folder_owner')
    create_home_folder_group = WritableString('create_home_folder_group')
    lock_in_home_folder = WritableBoolean('lock_in_home_folder')
    source_ip_filter = WritableSSVLines('source_ip_filter')
    allow_own_password_change = WritableBoolean('allow_own_password_change')
    password_lifetime = WritableNumber('password_lifetime', default=0)
    disable_on_inactivity = WritableNumber('disable_on_inactivity', default=0)
    file_write_size = WritableInteger('file_write_size', 0)

    ssh_authorized_keys_path = WritableStringOrEmpty(
        'ssh_authorized_keys_path', default='', inherit=False)
    allow_certificate_authentication = WritableBoolean(
        'allow_certificate_authentication')
    home_folder_structure = WritableLines(
        'home_folder_structure')
    virtual_folders = WritableCSVLines(
        'virtual_folders', inherit=False)
    permissions = WritableCSVLinesUnique(
        'permissions', default=(('allow-full-control',),))
    amend_write_name = WritableStringOrEmptyCaseInsensitive(
        'amend_write_name')
    as2_require_http_authentication = WritableBoolean(
        'as2_require_http_authentication', default=True)
    required_credentials = WritableCSVInsensitive(
        'required_credentials', default=(TYPE_NAME.ANY,), inherit=False)
    ssh_delete_delay = WritableNumber(
        'ssh_delete_delay', default=0, allow_negative=False)


@implementer(IApplicationAdministratorConfiguration)
class ApplicationAdministratorConfiguration(
        EntityConfiguration, PasswordConfigurationMixin):
    """
    All configurations for an administrators as they are persisted.
    """

    SECTION_PREFIX = CONFIGURATION_SECTION_ADMINISTRATORS

    source_ip_filter = WritableSSVLines('source_ip_filter')
    node_variables = WritableString('node_variables')

    @property
    def kind_name(self):
        """
        Human readable name of this administrator type.
        """
        return 'application-administrator'

    # For now, administrators don't have certificate authentication.
    # Any value set is ignored.
    allow_certificate_authentication = DefaultOnly(CONFIGURATION_INHERIT[0])

    # Here to share password credential code with app accounts.
    # Not part of the interface or the configuration.
    # For now, password never expired for admins.
    password_lifetime = DefaultOnly(0)

    @property
    def roles(self):
        """
        Return a sequence of roles for this administrator.
        """
        default = (CONFIGURATION_DEFAULT_ROLE,)

        if self._proxy.hasOption(self._section_name, 'role'):
            # Should be removed after V5.
            # This handles legacy configuration that had 'role' and migrates
            # it to roles on first read.
            value = self._proxy.getString(self._section_name, 'role')
            self._proxy.removeOption(self._section_name, 'role')
            self._proxy.setString(self._section_name, 'roles', value)
        else:
            value = self._proxy.getString(self._section_name, 'roles')

        result = _convert_string_to_non_duplicate_tuple(value)

        if not result:
            return default

        return result

    @roles.setter
    def roles(self, value):
        """
        Copy code from WritableCSVUnique.
        """
        if not value:
            value = (CONFIGURATION_DEFAULT_ROLE,)

        value = _convert_ordered_sequence_to_non_duplicate_string(value)
        self._proxy.setString(self._section_name, 'roles', value)


@implementer(IRoleConfiguration)
class RoleConfiguration(EntityConfiguration):
    """
    All configurations for an administrator's role, as they are persisted.
    """

    SECTION_PREFIX = CONFIGURATION_SECTION_ROLES

    source_ip_filter = WritableSSVLines('source_ip_filter')

    permissions = WritableCSVLinesUnique(
        'permissions',
        default=(
            ('sync', ROLE_PERMISSION.DENY),
            ('*', ROLE_PERMISSION.ALL),
            ),
        )

    # For now, roles don't have certificate authentication.
    # We ignore any set operation for now.
    allow_certificate_authentication = DefaultOnly(False)
