# Copyright (c) 2011 Adi Roiban.
# See LICENSE for details.
"""
Identity configuration as available in the high level layer, without knowledge
about how it is serialized in the configuration.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
from datetime import datetime

from zope.interface import implementer

from chevah.server.commons.constant import (
    CONFIGURATION_DEFAULT_GROUP,
    CONFIGURATION_DEFAULT_ROLE,
    CONFIGURATION_INHERIT,
    CONFIGURATION_SECTION_ACCOUNTS,
    CONFIGURATION_SECTION_ADMINISTRATORS,
    CONFIGURATION_SECTION_GROUPS,
    CONFIGURATION_SECTION_ROLES,
    _INHERIT_MARKER,
    PASSWORD_CRYPT_ROUNDS,
    TYPE_NAME,
    )
from chevah.server.commons.exception import (
    ConfigurationDeleteInUseException,
    ServerError,
    )
from chevah.server.commons.interface import (
    IAccountConfiguration,
    IAccountsConfiguration,
    IAdministratorConfiguration,
    IApplicationAccountConfiguration,
    IExternalAccountConfiguration,
    IExternalAdministratorConfiguration,
    IGroupConfiguration,
    IOSAdministratorConfiguration,
    IRoleConfiguration,
    )
from chevah.server.configuration.option import EntityConfiguration
from chevah.server.configuration.set import (
    SetConfiguration,
    SetWithTypeConfiguration,
    )
import six


@implementer(IExternalAccountConfiguration)
class ExternalAccountConfigurationAbstract(EntityConfiguration):
    """
    Configuration for an account which is not defined in the application's own
    configuration.

    This can be both an application account or an OS account.

    It is initialized with a name and an optional external configuration dict.

    It should be not be directly instantiate, but use one of the classes
    inheriting from it.
    """

    # Set with keys allowed in external configuration.
    _valid_keys = set([
        # FIXME:5585:
        # This should be renamed to `groups`.
        # While `group` is still supported for backward compatibility.
        'group',
        'create_home_folder',
        'create_home_folder_owner',
        'create_home_folder_group',
        'home_folder_path',
        'home_folder_structure',
        'virtual_folders',
        'permissions',
        'amend_write_name',
        'uuid',
        'email',
        'as2_certificates',
        'as2_async_mdn_ca',
        'as2_require_http_authentication',
        'required_credentials',
        'expire_datetime',
        'file_write_size',
        # These are only for audit usage.
        'created',
        ])

    def __init__(self, parent, name, external=None):
        self._parent = parent
        if external is None:
            external = {}

        unknown_keys = sorted(set(external.keys()) - self._valid_keys)
        if unknown_keys:
            raise ServerError(
                '20149', data={'details': ', '.join(unknown_keys)})

        self._name = name
        self._uuid = external.get('uuid', name)

        self._group = external.get('group', (CONFIGURATION_DEFAULT_GROUP,))

        if isinstance(self._group, six.string_types):
            # Add support for backward compatible API in which the group
            # is returned as a single string.
            self._group = (self._group,)

        self._home_folder_path = external.get(
            'home_folder_path', CONFIGURATION_INHERIT[0])
        self._create_home_folder = external.get(
            'create_home_folder', CONFIGURATION_INHERIT[0])
        self._create_home_folder_owner = external.get(
            'create_home_folder_owner', CONFIGURATION_INHERIT[0])
        self._create_home_folder_group = external.get(
            'create_home_folder_group', CONFIGURATION_INHERIT[0])
        self._home_folder_structure = external.get(
            'home_folder_structure', _INHERIT_MARKER)
        self._virtual_folders = external.get(
            'virtual_folders', [[CONFIGURATION_INHERIT[0]]])
        self._permissions = external.get(
            'permissions', [[CONFIGURATION_INHERIT[0]]])
        self._amend_write_name = external.get(
            'amend_write_name', CONFIGURATION_INHERIT[0])
        self._email = external.get('email', '')
        self._created = external.get('created', '')
        self._file_write_size = external.get(
            'file_write_size', CONFIGURATION_INHERIT[0])
        self._expire_datetime = external.get('expire_datetime', '')
        self._as2_certificates = external.get('as2_certificates', '')
        self._as2_async_mdn_ca = external.get('as2_async_mdn_ca', '')
        self._as2_require_http_authentication = external.get(
            'as2_require_http_authentication', True)
        raw_required_credentials = external.get(
            'required_credentials', (CONFIGURATION_INHERIT[0],))
        self._required_credentials = tuple((
            six.text_type(v).lower().strip()
            for v in raw_required_credentials if v
            ))

    @property
    def kind_name(self):
        """
        Human readable name for the type of this account configuration.
        """
        raise NotImplementedError('kind_name not implemented')

    @property
    def type(self):
        """
        By default external accounts are always application type.
        """
        return TYPE_NAME.APPLICATION

    @property
    def enabled(self):
        """
        External accounts are always enabled as if they were not enabled,
        they would have not been authenticated in the first place and it would
        not reach this place.
        """
        return True

    @property
    def uuid(self):
        """
        See: EntityConfiguration.
        """
        return self._uuid

    @property
    def name(self):
        """
        See: EntityConfiguration.
        """
        return self._name

    @property
    def description(self):
        """
        See: EntityConfiguration.
        """
        return self._name

    # FIXME:5585:
    # This should be renamed to `groups`.
    @property
    def group(self):
        """
        See: `IAccountConfiguration`.
        """
        return self._group

    @property
    def home_folder_path(self):
        return self._home_folder_path

    @property
    def lock_in_home_folder(self):
        """
        External accounts are locked by default in home folder.
        """
        return True

    @property
    def create_home_folder(self):
        return self._create_home_folder

    @property
    def create_home_folder_owner(self):
        return self._create_home_folder_owner

    @property
    def create_home_folder_group(self):
        return self._create_home_folder_group

    @property
    def home_folder_structure(self):
        return self._home_folder_structure

    @property
    def virtual_folders(self):
        return self._virtual_folders

    @property
    def permissions(self):
        return self._permissions

    @property
    def file_write_size(self):
        return self._file_write_size

    @property
    def amend_write_name(self):
        return self._amend_write_name

    @property
    def ssh_authorized_keys_path(self):
        return CONFIGURATION_INHERIT[0]

    @property
    def ssh_authorized_keys_content(self):
        return ''

    @property
    def source_ip_filter(self):
        """
        External accounts have no source filter.
        """
        return ()

    @property
    def allow_certificate_authentication(self):
        return CONFIGURATION_INHERIT[0]

    @property
    def allow_own_password_change(self):
        """
        External accounts can't have their password changed.
        """
        return False

    @property
    def password(self):
        """
        See: `IApplicationAccountConfiguration`
        """
        return None

    @property
    def password_update(self):
        """
        See: `IApplicationAccountConfiguration`
        """
        return None

    @property
    def last_password_update(self):
        """
        We don't know when the password was changed for external accounts.
        """
        return 0

    @property
    def password_lifetime(self):
        """
        External passwords never expire.
        """
        return 0

    @property
    def disable_on_inactivity(self):
        """
        External accounts are never disabled.
        """
        return 0

    @property
    def last_passwords(self):
        """
        We don't know what passwords were used for external accounts.
        """
        return ()

    @property
    def multi_factor_authentication(self):
        """
        External accounts don't have TOTP.
        """
        return None

    @property
    def expire_datetime(self):
        return self._expire_datetime

    @property
    def created(self):
        return self._created

    @property
    def email(self):
        return self._email

    @property
    def as2_certificates(self):
        return self._as2_certificates

    @property
    def as2_async_mdn_ca(self):
        return self._as2_async_mdn_ca

    @property
    def as2_require_http_authentication(self):
        return self._as2_require_http_authentication

    @property
    def required_credentials(self):
        return self._required_credentials


@implementer(IAccountsConfiguration)
class AccountsConfiguration(SetWithTypeConfiguration):
    """
    All accounts configured for the server.
    """

    SECTION_PREFIX = CONFIGURATION_SECTION_ACCOUNTS
    __type_interface__ = IAccountConfiguration
    __DEFAULT_TYPE__ = TYPE_NAME.APPLICATION
    _PASSWORD_CRYPT_ROUNDS = PASSWORD_CRYPT_ROUNDS

    def getApplicationAccountConfiguration(self, account_name):
        """
        Return configuration for application account.
        """
        account = self.getSectionByName(account_name)
        if not account:
            return None

        if IApplicationAccountConfiguration.providedBy(account):
            return account
        else:
            return None

    def onChildCreate(self, section_name, value, uuid):
        """
        Called just after the section was created, but before the values
        are applied to the section.
        """
        now = datetime.utcnow().replace(microsecond=0).isoformat()
        self._proxy.setString(section_name, 'created', now + 'Z')

        return super(AccountsConfiguration, self).onChildCreate(
            section_name, value, uuid)

    def beforeChildDelete(self, uuid):
        """
        See: `ISetConfiguration`.
        """
        attached_components = []
        authentications = self.root.configuration.authentications
        for authentication_uuid in authentications.getPublicSectionNames():
            authentication = authentications.getSection(authentication_uuid)
            if authentication.type != TYPE_NAME.ANONYMOUS:
                continue

            if authentication.anonymous_account_uuid != uuid:
                continue

            attached_components.append(authentication.name)

        if attached_components:
            raise ConfigurationDeleteInUseException(uuid, attached_components)


class GroupsConfiguration(SetConfiguration):
    """
    All groups configured for the server.
    """

    SECTION_PREFIX = CONFIGURATION_SECTION_GROUPS
    __type_interface__ = IGroupConfiguration


class AdministratorsConfiguration(SetConfiguration):
    """
    All administrators configured for the server.
    """
    SECTION_PREFIX = CONFIGURATION_SECTION_ADMINISTRATORS
    __type_interface__ = IAdministratorConfiguration
    _PASSWORD_CRYPT_ROUNDS = PASSWORD_CRYPT_ROUNDS


@implementer(IExternalAdministratorConfiguration)
class ExternalAdministratorConfigurationAbstract(EntityConfiguration):
    """
    Base code for administrators authenticated from external sources.
    """
    SECTION_PREFIX = CONFIGURATION_SECTION_ADMINISTRATORS

    def __init__(self, parent, uuid, external=None):
        super(ExternalAdministratorConfigurationAbstract, self).__init__(
            parent=parent, uuid=uuid)
        if external is None:
            external = {}
        self._roles = external.get('roles', (CONFIGURATION_DEFAULT_ROLE,))
        self._name = external.get('name', self._uuid)

    @property
    def kind_name(self):
        """
        Human readable name of this administrator type.
        """
        raise NotImplementedError('kind_name not implemented')

    @property
    def enabled(self):
        """
        See `IExternalAdministratorConfiguration`.
        """
        return True

    @property
    def name(self):
        """
        See `IExternalAdministratorConfiguration`.
        """
        return self._name

    @property
    def description(self):
        """
        See `IExternalAdministratorConfiguration`.
        """
        return 'External application administrator.'

    @property
    def roles(self):
        """
        See `IExternalAdministratorConfiguration`.
        """
        return self._roles

    @property
    def password(self):
        """
        See: `IApplicationAccountConfiguration`
        """
        return None

    @property
    def password_update(self):
        """
        See: `IApplicationAccountConfiguration`
        """
        return None

    @property
    def last_password_update(self):
        """
        We don't know when the password was changed for external accounts.
        """
        return 0

    @property
    def last_passwords(self):
        """
        We don't know what password were used.
        """
        return ()

    @property
    def multi_factor_authentication(self):
        """
        See: `IApplicationAccountConfiguration`

        External accounts don't have MFA.
        """
        return None

    @property
    def allow_certificate_authentication(self):
        """
        External administrators just inherit certificate authentication
        configuration.
        """
        return _INHERIT_MARKER

    @property
    def source_ip_filter(self):
        """
        External administrators just inherit IP filtering.
        """
        return ()

    @property
    def node_variables(self):
        """
        No external node configuration for external admins.
        """
        return ''


@implementer(IOSAdministratorConfiguration)
class OSAdministratorConfiguration(EntityConfiguration):
    """
    All configurations for an administrator authenticated by the OS.

    All OS administrators are external to the configuration file.
    """

    SECTION_PREFIX = CONFIGURATION_SECTION_ADMINISTRATORS

    @property
    def kind_name(self):
        """
        Human readable name of this administrator type.
        """
        return 'os-administrator'

    @property
    def enabled(self):
        """
        See `IOSAdministratorConfiguration`.
        """
        return True

    @property
    def name(self):
        """
        See `IOSAdministratorConfiguration`.
        """
        return self._uuid

    @property
    def description(self):
        """
        See `IOSAdministratorConfiguration`.
        """
        return 'External OS administrator.'

    @property
    def roles(self):
        """
        For now, OS administrator are only member of the default role.

        They are outside of our configuration so we can't define a list of
        roles.
        In the future we might have a configuration for the default list of
        roles.
        """
        return (CONFIGURATION_DEFAULT_ROLE,)

    @property
    def multi_factor_authentication(self):
        """
        See: `IOSAdministratorConfiguration`
        """
        return None

    @property
    def allow_certificate_authentication(self):
        """
        For now, administrators don't have certificate authentication.
        """
        return _INHERIT_MARKER

    @property
    def source_ip_filter(self):
        """
        For now, administrators don't have IP filtering.
        """
        return ()

    @property
    def password(self):
        """
        See: `IOSAdministratorConfiguration`
        """
        return None

    @property
    def password_update(self):
        """
        See: `IOSAdministratorConfiguration`
        """
        return None

    @property
    def last_password_update(self):
        """
        We don't know when external password was updated.
        """
        return 0

    @property
    def last_passwords(self):
        """
        We don't know what password were used.
        """
        return ()

    @property
    def node_variables(self):
        """
        OS admins should not be used for node authentication.
        """
        return ''


class RolesConfiguration(SetConfiguration):
    """
    All groups of administrators configured for the server.
    """

    SECTION_PREFIX = CONFIGURATION_SECTION_ROLES
    __type_interface__ = IRoleConfiguration
