# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
JSON coding and decoding.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

from datetime import datetime
from json import JSONEncoder
from twisted.internet.address import IPv4Address, IPv6Address

from chevah.server import native_string
from chevah.server.commons.exception import ServerError
from chevah.server.commons.interface import (
    IPublicObject,
    PublicAttribute,
    )
from chevah.server.configuration.model import get_attribute_names
import six


class ChevahJSONEncoder(JSONEncoder):
    """
    JSON Encoder for Chevah data types.
    """

    def default(self, obj):
        """
        Convert/Break complex types into JSON primitive types.
        """
        if isinstance(obj, (set, tuple)):
            return list(obj)

        if isinstance(obj, ServerError):
            return self._convertServerError(obj)

        if isinstance(obj, IPv4Address):
            return self._convertIPv4Address(obj)

        if isinstance(obj, IPv6Address):
            return self._convertIPv6Address(obj)

        if IPublicObject.providedBy(obj):
            return self._convertPublicObject(obj)

        if isinstance(obj, datetime):
            return self._convertDateTime(obj)

        # Fall back with code from super().
        return JSONEncoder.default(self, obj)

    def _convertServerError(self, error):
        """
        Return a copy of error's data attribute.
        """
        values = {
            'id': error.id,
            'data': error.data.copy(),
            }

        return values

    def _convertIPv4Address(self, peer):
        """
        Convert an `IPv4Address` to a dictionary.
        """
        values = {
            'address': six.text_type(peer.host),
            'port': peer.port,
            'family': 'IPv4',
            'protocol': six.text_type(peer.type),
            }

        return values

    def _convertIPv6Address(self, peer):
        """
        Convert an `IPv6Address` to a dictionary.
        """
        values = {
            'address': six.text_type(peer.host),
            'port': peer.port,
            'family': 'IPv6',
            'protocol': six.text_type(peer.type),
            }

        return values

    def _convertPublicObject(self, obj):
        """
        Return a dictionary with `PublicAttributes` and
        `PublicWritableAttributes` as keys with their respective values from
        the `obj` instance.
        """
        result = {}
        for name in get_attribute_names(obj, PublicAttribute):
            result[name] = getattr(obj, name)

        return result

    def _convertDateTime(self, obj):
        """
        Return the ISO 8601 representation of datetime.
        """
        return obj.isoformat(native_string('T'))
