# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
Module for working with JSON files.
"""
from __future__ import absolute_import
from __future__ import unicode_literals
import json

from chevah.server import force_unicode
from chevah.server.commons.constant import CONFIGURATION_DISABLED_VALUES
from chevah.server.commons.exception import ServerException


class JSONFile(object):
    """
    A file containing JSON serialized data.

    This is a slightly modified version of JSON, which allow single line
    comments. A commented line should start with "# ", note the space after
    the "#".
    Inline comments are not supported.
    """

    def __init__(self, path=None, file=None):
        """
        It can be initialized with a path of a file like object.

        File like object initialization is mainly for testing.
        """
        from chevah_compat import local_filesystem
        self._path = path
        if self._path:
            self._segments = local_filesystem.getSegmentsFromRealPath(
                self._path)
        else:
            self._segments = None

        self._file = file
        if not (self._path or self._file) or (self._path and self._file):
            raise RuntimeError('You must specify a path or a file.')

        self._data = {}

    @property
    def data(self):
        """
        Data stored by the JSONFile.
        """
        return self._data

    def load(self):
        """
        Load the JSON from input file. Deserialize data.
        """
        from chevah_compat import local_filesystem
        if self._segments:
            try:
                self._file = (
                    local_filesystem.openFileForReading(
                        self._segments))
            except EnvironmentError as error:
                data = {
                    'path': self._path,
                    'details': force_unicode(error.strerror),
                    }
                message = (
                    'Failed to load JSON file "%(path)s". %(details)s' % data)
                raise ServerException(message)

        try:
            result = json.load(self._file)
        except ValueError as error:
            # We use tell() to find the length of the file, assuming
            # that the previous json.load has already done a seek to EOF.
            if not self._file.tell():
                # We have an empty file, so just ignore the error and
                # initialize an empty JSON structure.
                result = {}
            else:
                data = {
                    'path': self._path,
                    'details': force_unicode(error),
                    }
                message = (
                    'Bad format for JSON file "%(path)s". %(details)s' % data)
                raise ServerException(message)
        finally:
            self._file.close()

        self._data = result

    def getValueOrNone(self, dictionary, key):
        """
        Return the value stored in `dictionary` at `key`.

        If no key is empty or key is disabled, it will return None.
        """
        if key not in dictionary:
            raise RuntimeError('Dictionary has no key "%s".' % (key))

        value = dictionary[key]
        if value.lower() in CONFIGURATION_DISABLED_VALUES:
            return None
        elif len(value) == 0:
            return None
        else:
            return value
