# Copyright (c) 2011 Adi Roiban.
# See LICENSE for details.
"""
JSON-RPC 2.0 Twisted Web Resource.

An implemented JSON-RPC method will return a result that could be serialized
as JSON-RPC or raise an
"""
import inspect

from twisted.internet import defer
from twisted.python.failure import Failure
from twisted.web import server

from chevah.server import json, force_unicode
from chevah.server.commons.exception import ServerException
from chevah.server.http.resource import ChevahResource


def _parseError():
    """
    Parse error response.
    """
    message = 'Parse error.'
    return {'code': -32700, 'message': message, 'data': None}


def _invalidRequest(message):
    """
    Invalid JSON-RPC request.
    """
    return {'code': -32600, 'message': message, 'data': None}


def _methodNotFound():
    """
    Method not found.
    """
    message = 'Method not found.'
    return {'code': -32601, 'message': message, 'data': None}


def _invalidArguments(message):
    """
    Invalid parameters.
    """
    return {'code': -32602, 'message': message, 'data': None}


def _internalError(code, message=None):
    """
    Internal error.
    """
    if not message:
        message = 'Internal error.'
    return {'code': code, 'message': message, 'data': None}


def _noSessionError():
    '''Missing session error.'''
    # FIXME:3119:
    # Move this out of commons as this is an server id.
    return {
        'code': 50000,
        'message': 'No session exists.',
        'data': None,
        }


class JSONRPCError(Exception):
    """
    An error in processing an JSON-RPC.
    """

    def __init__(self, value):
        self.value = value


class JSONRPCResource(ChevahResource):
    """
    JSON-RPC 2.0 resource.
    """

    isLeaf = True

    def __init__(self):
        super(JSONRPCResource, self).__init__()
        self.public_methods = []
        self._deferred = None

    def render_GET(self, request):
        """
        Convert a GET request into an JSON-RPC request.
        """
        # Remove trailing slash.
        # FIXME:6201:
        # request.postpath will be unicode.
        if len(request.postpath) > 0 and request.postpath[-1] == b'':
            request.postpath.pop()

        # Process index request
        if len(request.postpath) < 1:
            method_name = 'get_index'
        else:
            # FIXME:6201:
            # request.postpath will be unicode.
            method_name = 'get_' + request.postpath[0].decode('utf-8')

        # Create the JSON-RPC request data.
        json_content = {
            'method': method_name,
            'params': {},
            'id': 1,
            'jsonrpc': 2.0,
            }
        return self._renderJSONRPCOverHTTP(request, json_content)

    def render_POST(self, request):
        """
        Execute the requested JSON-RPC method.
        """
        try:
            content = request.content.getvalue()
            json_content = json.loads(content)
        except ValueError:
            response = {'id': None, 'result': None, 'error': _parseError()}
            return json.dumps(response).encode('utf-8')

        return self._renderJSONRPCOverHTTP(request, json_content)

    def _renderJSONRPCOverHTTP(self, request, json_content):
        """
        Execute the method and return a response for the HTTP request.

        Returns the JSON-RPC response or a deferred with NOT_DONE_YET.
        """
        # Check if we have a request id.
        # Notifications have no ID.
        # For notification the ID is set to None.
        try:
            request.id = json_content['id']
        except KeyError:
            json_content['id'] = None
            request.id = None

        # We don't want to cache the JSON-RPC conversation.
        request.setDisableCacheHeaders()

        try:
            json_content['jsonrpc']
        except KeyError:
            error_response = {
                'result': None,
                'error': _invalidRequest('Missing "jsonrpc".'),
                }
            return _getHTTPResponse(error_response, request)

        def _triggerRequest(request, json_content):
            """
            Get and call the method inside a deferred.
            """
            request_method = self._getMethod(json_content)
            result = self._callMethod(request_method, request, json_content)
            return result

        def _cbPackResult(result, request):
            return {'result': result, 'error': None}

        def _cbRender(result, request):
            """
            Send the response.
            """
            if request.id is None:
                # Do nothing for notifications.
                return

            if not request.channel:
                # Do nothing as the connection is disconnected.
                return

            response = _getHTTPResponse(result, request)
            request.write(response)
            request.finish()

        def _ebRenderServerException(failure, request):
            """
            Send the error response or do nothing if the request is a
            notification.
            """
            failure.trap(ServerException)

            code = self.logInternalFailure(failure, peer=request.client)

            if request.id is None:
                # Do nothing for notifications.
                return

            result = {
                'result': None,
                'error': {
                    'code': code,
                    'message': force_unicode(failure.value),
                    'data': None,
                    },
                }
            _cbRender(result, request)
            return None

        def _ebRenderJSONRPCError(failure, request):
            """
            Send the error response or do nothing if the request is a
            notification.
            """
            failure.trap(JSONRPCError)

            if request.id is None:
                # Do nothing for notifications.
                return

            result = {
                'result': None,
                'error': failure.value.value,
                }
            _cbRender(result, request)
            return None

        def _ebRenderInternalError(failure, request):
            """
            Send the error response or do nothing if the request is a
            notification.
            """
            code = self.logInternalFailure(failure, peer=request.client)

            if request.id is None:  # noqa:cover
                # Do nothing for notifications.
                # We should not end up here as any error should be
                # previously converted into JSON-RPC errors.
                return

            error = _internalError(
                message='Internal server error. %s' % (failure.value),
                code=code,
                )

            result = {
                'result': None,
                'error': error,
                }
            _cbRender(result, request)

        # The deferred called is stored in the resource so that we can
        # use it in tests.
        self._deferred = defer.maybeDeferred(
            _triggerRequest, request, json_content)
        self._deferred.addCallback(_cbPackResult, request)
        self._deferred.addCallback(_cbRender, request)
        self._deferred.addErrback(_ebRenderServerException, request)
        self._deferred.addErrback(_ebRenderJSONRPCError, request)
        self._deferred.addErrback(_ebRenderInternalError, request)

        # For notification we return an empty response right away.
        # Request ID is None for notifications.
        if request.id is None:
            return b''
        else:
            return server.NOT_DONE_YET

    def _getMethod(self, json_content):
        '''Get the method for the JSON-RPC request.'''
        try:
            method_name = json_content['method']
        except KeyError:
            raise JSONRPCError(_invalidRequest('Missing "method".'))

        request_method = getattr(self, 'jsonrpc_' + method_name, None)
        if request_method is None:
            raise JSONRPCError(_methodNotFound())

        return request_method

    def _callMethod(self, request_method, request, json_content):
        """
        Execute the JSON-RPC method.
        """
        # Check valid session.
        # We don't want to create a new session, just get the existing one.
        # Twisted handles the session ID as bytes.
        session_id = json_content.get('session_id', '').encode('ascii')
        if session_id:
            session = request.getSessionByID(session_id)
        else:
            session = request.getExistingSession()

        request._setSession(session)
        request.touchSession()

        if (
            request.session is None and
            not json_content['method'] in self.public_methods
                ):
            raise JSONRPCError(_noSessionError())

        try:
            arguments = json_content['params']
        except KeyError:
            raise JSONRPCError(_invalidRequest('Missing "params".'))

        try:
            if type(arguments) is list:
                return _call_with_list(request_method, request, arguments)
            elif type(arguments) is dict:
                return _call_with_dict(request_method, request, arguments)
            else:
                raise JSONRPCError(
                    _invalidArguments('"params" must be list or dict.'))
        except JSONRPCError:
            # Re-raise specific JSONRPC error as otherwise it will be
            # converted into internal server error.
            raise
        except ServerException as error:
            raise error
        except Exception as error:
            import traceback

            failure = Failure(error)
            code = self.logInternalFailure(failure, peer=request.client)
            error_details = traceback.format_exc()
            error_response = _internalError(code=code)
            error_response.update({'data': error_details})
            raise JSONRPCError(error_response)

    def logInternalFailure(self, failure, peer=None):
        """
        Log an internal error.

        This method should be overwritten to provide custom handling.

        It can return a error code to be used instead of the standard
        JSON-RPC internal error code.
        """
        raise NotImplementedError('Please define logInternalFailure.')


def _getHTTPResponse(result, request):
    """
    Return the JSON-RPC result for an HTTP request.

    If 'jsonrpc_result' is None it returns an empty result.
    """
    result.update({
        'jsonrpc': 2.0,
        'id': request.id,
        })
    response_content = json.dumps(result).encode('utf-8')

    request.setHeader(
        b"content-length", str(len(response_content)).encode('ascii'))
    request.setHeader(b"content-type", b"application/json-rpc")
    return response_content


def _check_arguments(method, arguments):
    """
    Check that arguments are valid to be called with method.
    """
    method_arguments_spec = inspect.getfullargspec(method)
    argument_names = method_arguments_spec[0]
    default_values = method_arguments_spec[3]

    if default_values is None:
        default_values = ()

    number_or_mandatory_arguments = (
        len(argument_names) - 2 - len(default_values))

    if len(arguments) + 2 > len(argument_names):
        raise JSONRPCError(
            _invalidArguments('Too many values in "params".'))

    if len(arguments) < number_or_mandatory_arguments:
        raise JSONRPCError(
            _invalidArguments('Too few values in "params".'))

    # Check that all non-default arguments are present.
    if type(arguments) is dict:
        non_default_names = argument_names[2:(-1 * len(default_values))]
        for mandatory_argument in non_default_names:
            if mandatory_argument not in arguments:
                raise JSONRPCError(
                    _invalidArguments(
                        'Missing "%s" argument in "params".' % (
                            mandatory_argument)))
        # Arguments accepted in RPC (without self and response).
        rpc_arguments = argument_names[2:]
        for requested_argument in arguments.keys():
            if requested_argument not in rpc_arguments:
                raise JSONRPCError(
                    _invalidArguments(
                        'Argument "%s" is not accepted in "params".' % (
                            requested_argument)))


def _get_result(method, request, *args, **kvargs):
    """
    Simple wrapper for formatting a result.
    """
    result = method(request, *args, **kvargs)
    return result


def _call_with_list(method, request, arguments):
    """
    Call the method with list arguments.

    Raise JSONRPC error if arguments are not valid.
    """
    _check_arguments(method, arguments)
    return _get_result(method, request, *arguments)


def _call_with_dict(method, request, arguments):
    """
    Call the method with keyword arguments.

    Raise JSONRPC error if arguments are not valid.
    """
    # Only str keys are accepted. Convert unicode to str.
    for key, value in list(arguments.items()):
        del arguments[key]
        arguments[str(key)] = value

    _check_arguments(method, arguments)
    return _get_result(method, request, **arguments)
