# Copyright (c) 2018 Adi Roiban.
# See LICENSE for details.
"""
Sharaed code for Azure Storage based locations.

This is just the location adapter.
The actual client-side protocol is found in the chevah.server.http package.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
from base64 import b64decode

from chevah.server.commons.constant import (
    CONFIGURATION_SECTION_LOCATIONS,
    )
from chevah.server.commons.exception import ServerException
from chevah.server.commons.ssl import (
    SSLClientMixin,
    )
from chevah.server.configuration.model import (
    DefaultOnly,
    )
from chevah.server.configuration.option import (
    UsernamePasswordPublicOptionsMixin,
    )
from chevah.server.location.commons import SSLLocationBase
from chevah.server.location.configuration import _LocationConfigurationBase


class _AzureStorageAccountLocationConfiguration(
    _LocationConfigurationBase,
    UsernamePasswordPublicOptionsMixin,
        ):
    """
    Shared configuration for a services of an Azure Storage account.

    It has the SSL configuration options, but they are read-only as we enforce
    the validation against our embedded MS root CAs.
    We still need to allow writing them, to comply with the ISSLOptions
    interface.

    Is an abstract class.
    Classes inheriting from this need to implement `address` property
    and implement the dedicated configuration interface.
    """
    SECTION_PREFIX = CONFIGURATION_SECTION_LOCATIONS

    @property
    def address(self):
        raise NotImplementedError('Azure Storage address not implemented.')

    @property
    def port(self):
        # Always over default HTTPS.
        return 443

    # No domains for now.
    # Allow writing to comply with the interface, but ignore the operation.
    ssl_domains = DefaultOnly(())
    # For this we might want to allow customization.
    # For now is fixed to keep it simple.
    ssl_cipher_list = DefaultOnly('secure')
    # For this we might want to allow customization.
    # For now is fixed to keep it simple.
    ssl_allowed_methods = DefaultOnly(('secure',))
    # Allow writing to comply with the interface, but ignore the operation.
    ssl_certificate = DefaultOnly('')
    ssl_key = DefaultOnly('')
    ssl_key_password_update = DefaultOnly('')
    ssl_certificate_authority = DefaultOnly('${MICROSOFT_IT_CA}')
    ssl_certificate_revocation_list = DefaultOnly(())
    ssl_certificate_revocation_list_refresh = DefaultOnly(0)

    @property
    def ssl_certificate_value(self):
        return ''

    @property
    def ssl_key_value(self):
        return ''

    @property
    def ssl_key_password(self):
        return ''


class _AzureStorageLocation(SSLLocationBase, SSLClientMixin):
    """
    Common code for locations based on Azure Storage account services.

    It has SSL configuration option, but for now are private.
    """

    PROPERTIES_REQUIRING_RESTART = (
        'password_update',
        'username',
        'idle_connection_timeout',
        'idle_connection_keepalive_interval',
        'connection_retry_count',
        'connection_retry_interval',
        )

    _HTTPClient = None

    def _getTrustedURLs(self):
        """
        Return the list of base URL accepted for HTTPS connections.
        """
        # This is called when is start and before (re)connect,
        # so we use it to validate the configuration.
        if not self.configuration.username:
            raise ServerException('Azure Storage account name is not defined.')

        try:
            self.configuration.username.encode('ascii')
        except Exception:
            raise ServerException(
                'Azure Storage account name has invalid format.')

        if not self.configuration.password:
            raise ServerException(
                'Azure Storage account shared key is not defined.')

        try:
            b64decode(self.configuration.password)
        except Exception:
            raise ServerException(
                'Azure Storage account access key has invalid format.')

        url = 'https://{}.{}:443'.format(
            self.configuration.username,
            self._HTTPClient.SERVICE_DNS_BASE,
            )
        return [url]
