# Copyright (c) 2015 Adi Roiban.
# See LICENSE for details.
"""
Shared code for all locations.
"""
from __future__ import unicode_literals

from __future__ import absolute_import
from datetime import datetime
import os
import ntpath

from lazr.delegates import delegate_to
from twisted.internet import defer
from twisted.internet.error import ConnectError
from twisted.internet.task import deferLater
from zope.interface import implementer
import posixpath

from chevah.server import (
    force_unicode,
    get_path_handling_module,
    emit as _emit,
    )
from chevah.server.commons.exception import (
    ConnectException,
    ServerError,
    ServerException,
    )
from chevah.server.commons.interface import (
    ILocalLocation,
    ILocation,
    ITransferStream,
    )
from chevah.server.event.handler import get_exception_details
from chevah.server.identity.avatar import ClientAvatar
from chevah.server.location.watch import SnapshotDriver, SnapshotLoop
from chevah.server.resource.runnable import ResourceClientRunnable


def convertToServerException(func):
    """
    Generic decorator to attach an errback that converts any non-ServerError
    failure into a ServerException.
    """

    def wrapper(*args, **kwargs):
        deferred = func(*args, **kwargs)
        deferred.addErrback(eb_convert_to_ServerException)
        return deferred

    def eb_convert_to_ServerException(failure):
        if failure.check(ServerError, ServerException, defer.CancelledError):
            return failure

        failure.value = ServerException(force_unicode(failure.value))
        return failure

    return wrapper


class OnConnectionLostMixin(object):
    """
    Shared code for ILocationShell
    """

    # List of callbacks for the connection lost.
    _on_connection_lost_calls = None

    def callOnConnectionLost(self, callback):
        """
        See: `ILocationShell`.
        """
        if self._on_connection_lost_calls is None:
            self._on_connection_lost_calls = []

        self._on_connection_lost_calls.append(callback)

    def _onConnectionLost(self, reason):
        """
        Called when connection to the remote subsystem was lost.
        """
        if not self._on_connection_lost_calls:
            return

        for call in self._on_connection_lost_calls:
            call(reason)


@implementer(ILocation)
class LocationBase(ResourceClientRunnable):
    """
    Common code for all locations.

    A location is just a wrapper providing a common API around the low level
    file transfer client.

    It provides handling for connection/disconnection of the remote location.

    When a location is started and connected it is in STARTED state.
    When started and not connected it is in STALLED state.

    Classes inheriting from this calls should define the `_path_module`.

    Instances can define a `_client` member that implements ILocationShell.
    """

    # os.path like module used for path operations.
    _path_module = None

    def __init__(self, parent, configuration):
        super(LocationBase, self).__init__(parent, configuration)
        # List of watched paths.
        self._watchers = set()

    def _onStop(self, avatar=None):
        """
        See: `IRunnable`.
        """
        for watched_handler in self._watchers.copy():
            self.unwatch(watched_handler)
        return super(LocationBase, self)._onStop(avatar=avatar)

    def splitPath(self, path):
        """
        See: `ILocation`.
        """
        self._hintPathHandlingModule(path)
        return self._path_module.split(path)

    def joinPath(self, *parts):
        """
        See: `ILocation`.
        """
        for part in parts:
            self._hintPathHandlingModule(part)
        # Try to guess how to convert separators.
        # For now only Posix and Windows path styles are supported.
        if self._path_module.sep == '/':
            our_separator = '/'
            other_separator = '\\'
        else:
            our_separator = '\\'
            other_separator = '/'

        converted_parts = []
        for part in parts:
            converted_parts.append(
                part.replace(other_separator, our_separator))

        full_path = self._path_module.join(*converted_parts)

        # See if we have a URL so that we don't normalize the
        # URL protocol part.
        parts = full_path.split(':' + our_separator * 2, 1)
        if len(parts) == 2:
            # We have an URL which should be handled as Unix path and
            # have its protocol part preserved.
            protocol = parts[0] + '://'
            path = parts[1].replace('\\', '/')
            return protocol + posixpath.normpath(path)

        return self._path_module.normpath(full_path)

    def relativePath(self, base, path):
        """
        See: `ILocation`.
        """
        self._hintPathHandlingModule(base)
        self._hintPathHandlingModule(path)

        canonical_base = base.rstrip(self._path_module.sep)
        if not canonical_base:
            # We have a root base.
            canonical_base = base

        if (
            len(canonical_base) == 2
            and canonical_base[1] == ':'
            and self._path_module.sep == '\\'
                ):
            # We have a windows drive path, from which we have
            # previously stripped the root separator.
            # So make the path valid again.
            canonical_base = canonical_base + '\\'

        if not path.lower().startswith(canonical_base.lower()):
            raise ServerException(
                'There is no relations between paths. %s and %s' % (
                    canonical_base, path))

        if os.name == 'nt':
            result = self._path_module.relpath(path, start=canonical_base)
        else:
            result = self._path_module.relpath(
                path.encode('utf-8'),
                start=canonical_base.encode('utf-8')
                ).decode('utf-8')

        if result == '.':
            return ''

        return result

    def watch(self, configuration, handler, on_error, on_snapshot=None):
        """
        See: `ILocation`.
        """
        watched_handler = self._getSourceWatch(
            configuration=configuration,
            handler=handler,
            on_error=on_error,
            on_snapshot=on_snapshot,
            )
        self._watchers.add(watched_handler)
        self.callWhenConnected(watched_handler.start)
        return watched_handler

    def _getSourceWatch(self, configuration, handler, on_error, on_snapshot):
        """
        Get the watcher for the source.
        """
        watch_driver = SnapshotDriver(self, configuration)
        return SnapshotLoop(
            driver=watch_driver,
            handler=handler,
            scheduler=self._scheduler,
            on_error=on_error,
            on_snapshot=on_snapshot,
            )

    def unwatch(self, watched_handler):
        """
        See: `ILocation`.
        """
        # FIXME:2293:
        # We should wait for watcher to stop.
        path = watched_handler.stop()

        data = {
            'location': self.name,
            'path': path,
            }

        try:
            self._watchers.remove(watched_handler)
        except KeyError:
            # Most probably it was already removed.
            # All ok.
            pass
        else:
            # Only emit event if we really did the unwatch
            # and not when unwatch is called again later.
            # This is here mainly because FilteredWatchAdapter doesn't
            # know when the location is stopped and its low-level watch is
            # stopped.
            self.emitEvent('60040', data)

    def _hintPathHandlingModule(self, path):
        """
        Try to set the path handling module if not already set.

        Raise an exception if path can not be detected.
        """
        if self._path_module:
            return

        if '/' in path or '\\' in path:
            self._setPathHandlingModule(path)
            return

        raise ServerException(
            'Can not detect path handling from "%s".' % (path,))

    def _setPathHandlingModule(self, path):
        """
        Based on the provided path, set the path handling module.
        """
        self._path_module = get_path_handling_module(path)

    def exists(self, path):
        """
        See: ILocation.
        """
        return self._client.exists(path)

    def rename(self, source_path, destination_path):
        """
        See: ILocation.
        """
        return self._client.rename(source_path, destination_path)

    def deleteFile(self, path):
        """
        See: ILocation.
        """
        return self._client.deleteFile(path)

    def deleteFolder(self, path, recursive=False):
        """
        See: ILocation.
        """
        return self._client.deleteFolder(path)

    def createFolder(self, path):
        """
        See: ILocation.
        """
        return self._client.createFolder(path)

    def getAttributes(self, path):
        """
        See: ILocation.
        """
        return self._client.getAttributes(path)

    def getFolderMemberAttributes(self, path):
        """
        See: ILocation.
        """
        return self._client.getFolderMemberAttributes(path)

    def getStreamForReading(self, path):
        """
        See: ILocation.
        """
        return self._client.getStreamForReading(path=path)

    def getStreamForWriting(self, path, length=None):
        """
        See: ILocation.
        """
        return self._client.getStreamForWriting(path=path, length=length)

    def getStreamForAppending(self, path):
        """
        See: ILocation.
        """
        return self._client.getStreamForAppending(path=path)


class SSLLocationBase(LocationBase):
    """
    Common code for all HTTP locations with a ssl context.
    """

    # Class of specialized HTTP client implementing operations for this
    # location.
    _HTTPClient = None

    def emitEvent(self, event_id, data=None, avatar=None):
        """
        Helper to emit an event for this location.
        """
        if avatar is not None:
            # Some locations (ex HTTP) might connect to different peers due
            # to DNS load balancing.
            # Here we preserve the provided avatar, but update the
            # component for which it was created.
            # A manager can also trigger an action for this location, like
            # start or stop.
            if isinstance(self._avatar, ClientAvatar):
                avatar = ClientAvatar(self._avatar._credentials)
                avatar._peer = avatar.peer
        else:
            avatar = self._avatar

        _emit(event_id, avatar=avatar, data=data)

    def _getTrustedURLs(self):
        """
        List of URLs trusted by this location for HTTPS connections.
        """
        # Default with just the main url.
        return [self.configuration.url]

    @defer.inlineCallbacks
    def _onConnect(self):
        """
        See: LocationBase.
        """
        try:
            result = yield self._initStartedContextFactory(
                urls=self._getTrustedURLs())

            try:
                # Set the self._client right aways to that on errors it will be
                # closed.
                self._client = self._HTTPClient(
                    location=self,
                    context_factory=result,
                    scheduler=self._scheduler,
                    )
                yield self._client.authenticate()
            except ConnectError as error:
                # Called when the connection was closed.
                message = 'Connected to %s:%s. %s' % (
                    self.configuration.address,
                    self.configuration.port,
                    force_unicode(error),
                    )
                raise ConnectException(message)
        except Exception as error:
            # We delay the failure to allow for the HTTP agent connection
            # to be processed.
            yield deferLater(self._scheduler, 0, lambda: None)
            raise error
        defer.returnValue(self._client)

    @defer.inlineCallbacks
    def _onDisconnect(self):
        """
        See: `LocationBase`.
        """
        if self._context_factory:
            context, self._context_factory = self._context_factory, None
            context.stop()

        if not self._client:
            return
        yield self._client.close()


@implementer(ILocation)
@delegate_to(ILocation, context='_context')
class LocationAuditor(object):
    """
    A wrapper for ILocation which emits events and handles errors.
    """
    def __init__(self, context):
        self._context = context

    def watch(self, configuration, handler, on_error, on_snapshot=None):
        """
        See: `ILocation`.
        """
        data = {
            'path': configuration.source_path,
            'recursive': configuration.recursive,
            }
        result = self._context.watch(
            configuration=configuration,
            handler=handler,
            on_error=on_error,
            on_snapshot=on_snapshot,
            )
        # This is emitted after watch is called so that it is only called if
        # watch was successful.
        # FIXME:5430:
        # This event should be emitted with transfer or the service monitor
        # avatar.
        self._cbEmitSuccess(None, '60039', data)
        return result

    def unwatch(self, watched_handler):
        """
        See: `ILocation`.
        """
        # No success event is emitted here.
        # It is emitted in the location shared code as unwatch is called
        # internally (outside of the auditor).
        result = self._context.unwatch(watched_handler=watched_handler)
        return result

    @convertToServerException
    def exists(self, path, with_event=True):
        """
        See: `ILocation`.
        """
        deferred = self.heartbeat()
        data = {'path': path}
        deferred.addCallback(lambda _: self._context.exists(path=path))
        if with_event:
            deferred.addCallback(self._cbEmitSuccess, '60041', data)
        deferred.addErrback(self._ebHandleFailure, '60042', data)
        return deferred

    @convertToServerException
    def rename(self, source_path, destination_path):
        """
        See: `ILocation`.
        """
        deferred = self.heartbeat()
        data = {
            'path': destination_path,
            'to': destination_path,
            'from': source_path,
            }
        deferred.addCallback(lambda _: self._context.rename(
            source_path=source_path, destination_path=destination_path))
        deferred.addCallback(self._cbEmitSuccess, '60043', data)
        deferred.addErrback(self._ebHandleFailure, '60044', data)
        return deferred

    @convertToServerException
    def deleteFile(self, path):
        """
        See: `ILocation`.
        """
        deferred = self.heartbeat()
        data = {'path': path}
        deferred.addCallback(lambda _: self._context.deleteFile(path=path))
        deferred.addCallback(self._cbEmitSuccess, '60045', data)
        deferred.addErrback(self._ebHandleFailure, '60046', data)
        return deferred

    @convertToServerException
    def deleteFolder(self, path, recursive=False):
        """
        See: `ILocation`.
        """
        deferred = self.heartbeat()
        data = {
            'path': path,
            'recursive': recursive,
            }
        deferred.addCallback(lambda _: self._context.deleteFolder(
            path=path, recursive=recursive))
        deferred.addCallback(self._cbEmitSuccess, '60047', data)
        deferred.addErrback(self._ebHandleFailure, '60048', data)
        return deferred

    @convertToServerException
    def createFolder(self, path):
        """
        See: `ILocation`.
        """
        deferred = self.heartbeat()
        data = {'path': path}
        deferred.addCallback(lambda _: self._context.createFolder(path=path))
        deferred.addCallback(self._cbEmitSuccess, '60049', data)
        deferred.addErrback(self._ebHandleFailure, '60050', data)
        return deferred

    @convertToServerException
    def getFolderMemberAttributes(self, path):
        """
        See: `ILocation`.
        """
        deferred = self.heartbeat()
        data = {'path': path}
        deferred.addCallback(lambda _: self._context.getFolderMemberAttributes(
            path=path))
        deferred.addCallback(self._cbEmitSuccess, '60051', data)
        deferred.addErrback(self._ebHandleFailure, '60052', data)
        return deferred

    @convertToServerException
    def getAttributes(self, path):
        """
        See: `ILocation`.
        """
        deferred = self.heartbeat()
        data = {'path': path}
        deferred.addCallback(lambda _: self._context.getAttributes(path=path))
        deferred.addCallback(self._cbEmitSuccess, '60071', data)
        deferred.addErrback(self._ebHandleFailure, '60072', data)
        return deferred

    @convertToServerException
    def getStreamForReading(self, path):
        """
        See: `ILocation`.
        """
        deferred = self.heartbeat()
        data = {'path': path}
        deferred.addCallback(lambda _: self._context.getStreamForReading(
            path=path))
        deferred.addCallback(self._cbEmitSuccess, '60053', data)
        deferred.addCallback(self._cbWrapStream)
        deferred.addErrback(self._ebHandleFailure, '60054', data)
        return deferred

    @convertToServerException
    def getStreamForWriting(self, path, length=None):
        """
        See: `ILocation`.
        """
        deferred = self.heartbeat()
        data = {'path': path}
        deferred.addCallback(lambda _: self._context.getStreamForWriting(
            path=path, length=length))
        deferred.addCallback(self._cbEmitSuccess, '60055', data)
        deferred.addCallback(self._cbWrapStream)
        deferred.addErrback(self._ebHandleFailure, '60056', data)
        return deferred

    @convertToServerException
    def getStreamForAppending(self, path):
        """
        See: `ILocation`.
        """
        deferred = self.heartbeat()
        data = {'path': path}
        deferred.addCallback(lambda _: self._context.getStreamForAppending(
            path=path))
        deferred.addCallback(self._cbEmitSuccess, '60057', data)
        deferred.addCallback(self._cbWrapStream)
        deferred.addErrback(self._ebHandleFailure, '60058', data)
        return deferred

    def _cbWrapStream(self, stream):
        """
        Wraps the stream into an TransferStreamAuditor.
        """
        return TransferStreamAuditor(context=stream)

    def _cbEmitSuccess(self, result, event_id, data):
        """
        Callback for emitting a successful event.
        """
        data['location'] = self.name
        # FIXME:6292:
        # Add here the avatar of the original caller.
        self._context.emitEvent(event_id, data=data)
        return result

    def _ebHandleFailure(self, failure, event_id, data):
        """
        Callback when operation failed.
        """
        if failure.check(defer.CancelledError):
            return failure

        if failure.check(ServerError):
            # If this is a ServerError, just forward it and it will be
            # emitted later in the general error handling.
            failure.value.avatar = self._context.avatar
            return failure

        data['location'] = self.name

        if failure.check(ServerException):
            data['details'] = failure.value.message
            # ServerException should be a known error and we should not
            # need a traceback to identify it's origin.
            data['tb'] = ''
        else:
            data['details'] = force_unicode(failure.value)
            data['tb'] = force_unicode(failure.getBriefTraceback())
        # FIXME:6292:
        # Add here the avatar of the original caller.
        raise ServerError(event_id, data=data, avatar=self._context.avatar)


@implementer(ILocalLocation)
@delegate_to(ILocalLocation, context='_context')
class LocalLocationAuditor(LocationAuditor):
    """
    A wrapper for ILocalLocation which emits events and handles errors.

    Local location is always connected so we don't wait for the heartbeat to
    make the connection.
    """

    @convertToServerException
    def touch(self, path):
        """
        See: `ILocalLocation`.
        """
        self.heartbeat()
        data = {'path': path}
        deferred = self._context.touch(path=path)
        deferred.addCallback(self._cbEmitSuccess, '60059', data)
        deferred.addErrback(self._ebHandleFailure, '60060', data)
        return deferred

    @convertToServerException
    def copyFile(self, source_path, destination_path, overwrite=False):
        """
        See: `ILocalLocation`.
        """
        self.heartbeat()
        data = {
            'path': source_path,
            'from': source_path,
            'to': destination_path,
            'overwrite': overwrite,
            'real_path': destination_path,
            }
        deferred = self._context.copyFile(
            source_path=source_path,
            destination_path=destination_path,
            overwrite=overwrite,
            )
        deferred.addCallback(self._cbEmitSuccess, '60061', data)
        deferred.addErrback(self._ebHandleFailure, '60062', data)
        return deferred


@implementer(ITransferStream)
class TransferFileBase(object):
    """
    Shared code for a file on an ILocation.
    """

    def __init__(self, stream, location, path, mode):
        self._stream = stream
        self._path = path
        self._location = location
        self._mode = mode

    @property
    def path(self):
        """
        See: `ITransferStream`.
        """
        return self._path

    @property
    def location(self):
        """
        See: `ITransferStream`.
        """
        return self._location

    @property
    def mode(self):
        """
        See: `ITransferStream`.
        """
        return self._mode

    def close(self):
        """
        See: ITransferStream.
        """
        return self._stream.close()


@implementer(ITransferStream)
@delegate_to(ITransferStream, context='_context')
class TransferStreamAuditor(object):
    """
    Wrapper for ITransferStream to audit its actions and handle errors.

    The TransferStream does not try to re-connect the location as once the
    connection is lost the file handlers are also lost.
    """

    def __init__(self, context):
        self._context = context
        self._written = 0
        self._read = 0

    @property
    def written(self):
        """
        Size of the data written to this stream.
        """
        return self._written

    @convertToServerException
    def close(self):
        """"
        See: `ITransferStream`.
        """
        self.location.heartbeat()
        deferred = self._context.close()
        deferred.addCallback(self._cbEmitSuccess, '60035')
        deferred.addErrback(self._ebEmitFailure, '60036')
        return deferred

    @convertToServerException
    def read(self, length):
        """
        See: `ITransferStream`.
        """
        def cb_read(result):
            """
            Called after data was read with success.
            """
            self._read += len(result)
            return result

        self.location.heartbeat()
        deferred = self._context.read(length=length)
        deferred.addCallback(cb_read)
        deferred.addErrback(self._ebEmitFailure, '60037')
        return deferred

    @convertToServerException
    def write(self, data):
        """
        See: `ITransferStream`.
        """
        def cb_wrote(result):
            """
            Called after data was wrote with success.
            """
            self._written += len(data)
            return result

        self.location.heartbeat()
        deferred = self._context.write(data=data)
        deferred.addCallback(cb_wrote)
        deferred.addErrback(self._ebEmitFailure, '60038')
        return deferred

    def _cbEmitSuccess(self, result, event_id):
        """
        Callback for emitting a successful event.
        """
        data = {
            'path': self.path,
            'location': self.location.name,
            'mode': self.mode,
            'read_size': self._read,
            'write_size': self._written,
            }
        self._context.location.emitEvent(event_id, data=data)
        return result

    def _ebEmitFailure(self, failure, event_id):
        """
        Callback for emitting a failed event.
        """
        details = get_exception_details(failure.value)
        data = {
            'path': self.path,
            'mode': self.mode,
            'location': self.location.name,
            'details': details,
            'tb': force_unicode(failure.getBriefTraceback()),
            }
        self._context.location.emitEvent(event_id, data=data)
        # FIXME:3517:
        # Here we should just convert it into a ServerError, but with current
        # code this has the unwanted effect of emitting the error only after
        # the transfer has failed.
        return failure


def get_file_name(path):
    """
    Extract the name of a path.
    """
    path = path.strip()

    # Try with the Unix separator as it can't be found on Windows path.
    path_separator = '/'
    if path.count('/') == 0:
        path_separator = '\\'

    name = path.rstrip(path_separator).split(path_separator)[-1]

    if len(name) > 1 and name[1] == ':':
        name = ''

    if name in ['.', '..']:
        name = ''

    if not name:
        raise ServerException('Failed to extract name from path.')

    return name


def get_file_name_with_timestamp(name, seconds, random_mark):
    """
    Return the name of a file with appended timestamp, keeping the
    file extension.

    Timestamp is in milliseconds, but in case we get 2 events in the
    same millisecond, a random mark would make the distinction.
    """
    date = datetime.fromtimestamp(seconds)
    timestamp = date.strftime('%Y-%m-%d-%H-%M-%S-%f')
    base, extension = posixpath.splitext(name)
    return '%s.%s-%03d%s' % (base, timestamp, random_mark, extension)


def get_file_name_with_unixtime(name, seconds):
    """
    Return the name of a file with prepended Unix timestamp.
    """
    return '%017.6f-%s' % (seconds, name)


def get_target_path(member_path, base):
    """
    Return the path for `member_path` as if it would be
    extracted in path `base`.

    The returned `path` is the normalized OS path.

    It raises an exception if archive file has malicious paths.
    """
    base = os.path.normpath(base)

    # Force path to be relative.
    arcname = member_path.lstrip('/')

    if not arcname:
        # This member is just the root.
        return None

    for part in arcname.strip('/').split('/'):
        part = part.strip()
        if not part:
            raise ServerException('Archive with empty member name.')

    # build the destination pathname, replacing
    # forward slashes to platform specific separators.
    arcname = arcname.strip().replace('/', os.path.sep)

    # We might have malicious zips like:
    # /c:/abs/win
    #
    # interpret absolute pathname as relative, remove drive letter or
    # UNC path, redundant separators, "." and ".." components.
    drive, arcname = ntpath.splitdrive(arcname)

    if drive:
        raise ServerException('Archive has absolute paths.')

    targetpath = os.path.normpath(os.path.join(base, arcname))

    if not targetpath.startswith(base):
        raise ServerException(
            'Archive has members with path outside of the base path.')

    return targetpath


@defer.inlineCallbacks
def create_destination_parents(location, base_path, target_path):
    """
    Try to create destination parents for `target_path` as relative to
    `base_path`.

    We start with the first parent all the the last parent, and then create
    the whole parents again.

    For example for /one/two/three/four we try to create them in the following
    order:
    * /one/two/three/four -> stop if successful
    * /one/two/three -> stop if successful
    * /one/two -> stop if successful
    * /one -> fail if this operations fails.
    * /one/two -> fail if this operation fails
    * /one/two/three -> fail if this operation fails
    * /one/two/three/four -> fail if this operation fails

    The exceptions raised by this function are usually ignored, as it is
    designed to be called when writing the `target_path` fails, and we think
    that it failed because the parents are missing.

    Raise ServerException if no parents need to created.

    Raise the error if it fails to create the parents.
    """
    error = None
    parent_path = target_path
    parents = []
    while True:
        new_path, _ = location.splitPath(parent_path)
        if not new_path:
            # End of relative base
            break
        if new_path == parent_path:
            raise RuntimeError('Absolute target path.')
        parent_path = new_path
        parents.append(parent_path)

    if not parents:
        raise ServerException('No parents to create.')

    remaining_parents = []
    for member in parents:
        # Reset the error on each try
        # as maybe we can create the current child.
        error = None
        try:
            yield location.createFolder(location.joinPath(
                base_path, member))
            break

        except Exception as create_error:
            remaining_parents.insert(0, member)
            # We ignore the error as maybe the parent already exists.
            # In this case it will success for next child and the
            # error is reset.
            error = create_error

    if error:
        # This means that we failed to create the base parent.
        # so there is no point to try to create its children.
        raise error

    for member in remaining_parents:
        yield location.createFolder(location.joinPath(
            base_path, member))
