# Copyright (c) 2014 Adi Roiban.
# See LICENSE for details.
"""
Configuration for locations used by client side transfers.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
from twisted.internet import reactor
from zope.interface import implementer

from chevah.server.commons.constant import (
    CONFIGURATION_SECTION_LOCATIONS,
    DEFAULT_LOCAL_FILESYSTEM_UUID,
    LOCATION_CONNECTION_DEFAULTS,
    TYPE_NAME,
    )
from chevah.server.commons.interface import (
    ILocationsConfiguration,
    ILocationConfiguration,
    IFilesystemLocationConfiguration,
    ISFTPLocationConfiguration,
    )
from chevah.server.configuration.option import (
    AddressPortOptionsMixin,
    EntityConfiguration,
    TypeOptionsMixin,
    UsernamePasswordPublicOptionsMixin,
    )
from chevah.server.configuration.set import SetWithTypeConfiguration
from chevah.server.configuration.model import (
    DefaultOnly,
    WritableInteger,
    WritableString,
    WritableStringOrEmpty,
    WritableCSVUniqueInsensitive,
    )


class _LocationConfigurationBase(
        EntityConfiguration, TypeOptionsMixin):
    """
    Shared attributes by all location configurations.
    """
    SECTION_PREFIX = CONFIGURATION_SECTION_LOCATIONS

    # Locations are always not-enabled at startup as they are started
    # on demand.
    enabled = DefaultOnly(False)

    # FIXME:6411:
    # See how to implement a minimum value.
    # Maybe something like 5 seconds.
    idle_connection_timeout = WritableInteger(
        'idle_connection_timeout',
        default=LOCATION_CONNECTION_DEFAULTS)
    idle_connection_keepalive_interval = WritableInteger(
        'idle_connection_keepalive_interval',
        default=LOCATION_CONNECTION_DEFAULTS)
    connection_retry_count = WritableInteger(
        'connection_retry_count',
        default=LOCATION_CONNECTION_DEFAULTS)
    connection_retry_interval = WritableInteger(
        'connection_retry_interval',
        default=LOCATION_CONNECTION_DEFAULTS)


@implementer(IFilesystemLocationConfiguration)
class FilesystemLocationConfiguration(_LocationConfigurationBase):
    """
    Configuration for a local filesystem location.
    """
    # Local filesystem can't time out.
    idle_connection_timeout = 0
    address = '0.0.0.0'
    port = 0
    # For now, just a placeholder to know that this is the main process
    # file access.
    username = 'Process'


class DefaultFilesystemLocationConfiguration(FilesystemLocationConfiguration):
    """
    Configuration representing the local filesystem as available to
    main process.
    """
    type = TYPE_NAME.FILESYSTEM
    uuid = DEFAULT_LOCAL_FILESYSTEM_UUID
    name = 'Local filesystem'
    description = 'Filesystem available to main process.'
    idle_connection_timeout = 0
    idle_connection_keepalive_interval = 0
    connection_retry_count = 0
    connection_retry_interval = 0
    reactor = reactor

    def __init__(self, parent):
        # It should not touch the low-level file configuration as it has no
        # attached file.
        self._parent = parent
        self._proxy = None


@implementer(ISFTPLocationConfiguration)
class SFTPLocationConfiguration(
        _LocationConfigurationBase,
        AddressPortOptionsMixin,
        UsernamePasswordPublicOptionsMixin,
        ):
    """
    Configuration for an SFTP location.
    """

    SECTION_PREFIX = CONFIGURATION_SECTION_LOCATIONS
    PORT_DEFAULT = 22

    ssh_private_key = WritableStringOrEmpty('ssh_private_key')

    @property
    def ssh_server_identity(self):
        """
        One pubic key or fingerprint per line.

        Support backward compatiblity for `ssh_server_fingerprint`
        """
        value = self._proxy.getString(
            self._section_name, 'ssh_server_identity')

        if value:
            return [line.strip() for line in value.splitlines()]

        # Fallback to old configuration option.
        return [self._proxy.getString(
            self._section_name, 'ssh_server_fingerprint')]

    @ssh_server_identity.setter
    def ssh_server_identity(self, value):
        """
        Set as a single text, with multiple lines.
        """
        self._proxy.removeOption(
            self._section_name, 'ssh_server_fingerprint')
        return self._proxy.setString(
            self._section_name, 'ssh_server_identity', '\n'.join(value))

    proxy = WritableString('proxy')
    ssh_cipher_list = WritableCSVUniqueInsensitive(
        'ssh_cipher_list'
        )


@implementer(ILocationsConfiguration)
class LocationsConfiguration(SetWithTypeConfiguration):
    """
    See: `ILocationsConfiguration`.
    """
    SECTION_PREFIX = CONFIGURATION_SECTION_LOCATIONS

    __type_interface__ = ILocationConfiguration

    def __init__(self, parent):
        super(LocationsConfiguration, self).__init__(parent=parent)
        self._default_location = DefaultFilesystemLocationConfiguration(
            parent=self)

    def getSection(self, name):
        """
        See: SetWithTypeConfiguration.
        """
        if name == DEFAULT_LOCAL_FILESYSTEM_UUID:
            return self._default_location
        else:
            return super(LocationsConfiguration, self).getSection(name)

    def getPublicSectionNames(self):
        """
        See: SetWithTypeConfiguration.

        Also add default local filesystem.
        """
        result = super(LocationsConfiguration, self).getPublicSectionNames()
        result.append(DEFAULT_LOCAL_FILESYSTEM_UUID)
        return result
