# Copyright (c) 2014 Adi Roiban.
# See LICENSE for details.
"""
Location for a local filesystem.
"""
import os

from twisted.internet import defer, reactor, task, threads
from zope.interface import implementer

from chevah_compat import local_filesystem
from chevah.server.commons.interface import ILocalLocation
from chevah.server.identity.avatar import FilesystemAvatar
from chevah.server.location.commons import (
    LocationBase,
    OnConnectionLostMixin,
    TransferFileBase,
    )


class FilesystemLocationClient(OnConnectionLostMixin):
    """
    The client which handles local filesystem.

    For now, it does nothing as filesystem is always connected.

    In the future we might allow a specific folder which when missing
    the location is not connected and once it appears the location is
    connected.
    """

    def __init__(self, location):
        self.avatar = FilesystemAvatar(location)

    def loseConnection(self):
        """
        Call all callbacks subscribed for connection lost.
        """
        self._onConnectionLost(Exception('done'))


@implementer(ILocalLocation)
class FilesystemLocation(LocationBase):
    """
    A location on the local filesystem.

    It notifies a transfer when files are changed on this location.
    Notifications are raised with absolute path.

    It also provides primitive file transfer operation on location. All
    operations are done using absolute path.

    There are lot of threads generated here:
    * Single main tread to process the event queue.
    * One thread for each watched path.

    `_timeout` is only used to speed up testing.
    """

    _path_module = os.path
    PROPERTIES_REQUIRING_RESTART = (
        'idle_connection_keepalive_interval',
        )

    def __init__(self, parent, configuration):
        super(FilesystemLocation, self).__init__(parent, configuration)
        self._filesystem = local_filesystem

    def _onConnect(self):
        """
        See: `Runnable`.
        """
        # Filesystem location is now always connected as it represent a
        # local filesystem.
        return FilesystemLocationClient(self)

    def _onDisconnect(self, avatar=None):
        """
        See: `Runnable`.
        """
        if not self._client:
            # The connection was already lost.
            return

        self._client.loseConnection()

    def deleteFile(self, path):
        """
        See: ILocation.
        """
        segments = self._filesystem.getSegmentsFromRealPath(path)
        return defer.maybeDeferred(
            self._filesystem.deleteFile, segments)

    def deleteFolder(self, path, recursive=False):
        """
        See: ILocation.
        """
        segments = self._filesystem.getSegmentsFromRealPath(path)
        return defer.maybeDeferred(
            self._filesystem.deleteFolder, segments, recursive=recursive)

    def createFolder(self, path):
        """
        See: ILocation.
        """
        segments = self._filesystem.getSegmentsFromRealPath(path)
        return defer.maybeDeferred(
            self._filesystem.createFolder, segments)

    @defer.inlineCallbacks
    def getFolderMemberAttributes(self, path):
        """
        See: ILocation.
        """
        segments = self._filesystem.getSegmentsFromRealPath(path)
        iterator = self._filesystem.iterateFolderContent(segments)

        result = {}
        count = 0
        for attributes in iterator:
            result[attributes.name] = attributes

            count += 1
            if count > 1000:
                count = 0
                yield task.deferLater(self._scheduler, 0, lambda: None)

        defer.returnValue(result)

    def getAttributes(self, path):
        """
        See: ILocation.
        """
        segments = self._filesystem.getSegmentsFromRealPath(path)
        return defer.maybeDeferred(
            self._filesystem.getAttributes, segments)

    def getStreamForReading(self, path):
        """
        See: ILocation.
        """
        segments = self._filesystem.getSegmentsFromRealPath(path)
        deferred = defer.maybeDeferred(
            self._filesystem.openFileForReading, segments)
        deferred.addCallback(self._cbOpenFile, path, 'reading')
        return deferred

    def getStreamForWriting(self, path, length=None):
        """
        See: ILocation.

        `length` is ignored as for local filesystem we don't need to
        announce in advance how much we will write.
        """
        segments = self._filesystem.getSegmentsFromRealPath(path)
        deferred = defer.maybeDeferred(
            self._filesystem.openFileForWriting, segments, mode=0o666)
        deferred.addCallback(self._cbOpenFile, path, 'writing')
        return deferred

    def getStreamForAppending(self, path):
        """
        See: ILocation.
        """
        segments = self._filesystem.getSegmentsFromRealPath(path)
        deferred = defer.maybeDeferred(
            self._filesystem.openFileForAppending, segments)
        deferred.addCallback(self._cbOpenFile, path, 'appending')
        return deferred

    def exists(self, path):
        """
        See: ILocation.
        """
        segments = self._filesystem.getSegmentsFromRealPath(path)
        return defer.maybeDeferred(self._filesystem.exists, segments)

    def rename(self, source_path, destination_path):
        """
        See: ILocation.
        """
        from_segments = self._filesystem.getSegmentsFromRealPath(source_path)
        to_segments = self._filesystem.getSegmentsFromRealPath(
            destination_path)
        return defer.maybeDeferred(
            self._filesystem.rename, from_segments, to_segments)

    def _cbOpenFile(self, stream, path, mode):
        """
        Called when stream is opened.
        """
        return LocalTransferFile(
            stream=stream,
            location=self,
            path=path,
            mode=mode,
            )

    def touch(self, path):
        """
        See: ILocalLocation.
        """
        def do_touch():
            """
            Do the actual operation.
            """
            segments = self._filesystem.getSegmentsFromRealPath(path)
            self._filesystem.touch(segments)
        return defer.maybeDeferred(do_touch)

    def copyFile(self, source_path, destination_path, overwrite=False):
        """
        See: ILocalLocation.
        """
        def do_copy():
            """
            Do the actual copy, in a thread.
            """
            source_segments = self._filesystem.getSegmentsFromRealPath(
                source_path)
            destination_segments = self._filesystem.getSegmentsFromRealPath(
                destination_path)
            self._filesystem.copyFile(
                source_segments, destination_segments, overwrite=overwrite)

        return threads.deferToThread(do_copy)


class LocalTransferFile(TransferFileBase):
    """
    A file on local filesystem.

    It wraps calls into deferreds.
    """

    def close(self):
        """
        See: ITransferStream.
        """
        return defer.maybeDeferred(self._stream.close)

    def read(self, length):
        """
        See: ITransferStream.
        """
        # FIXME:2510:
        # Look at a better way to create an async read.
        # Maybe using twisted.internet.fdesc
        return self._deferLater(self._stream.read, length)

    def write(self, data):
        """
        See: ITransferStream.
        """
        return self._deferLater(self._stream.write, data)

    def _deferLater(self, function, *args, **kwargs):
        """
        Call deferred with callLater to break the deferred chain and
        allow other deferreds to execute.
        """
        return task.deferLater(reactor, 0, function, *args, **kwargs)
