# Copyright (c) 2014 Adi Roiban.
# See LICENSE for details.
"""
Location for a remote FTP server.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from twisted.internet import defer, endpoints, reactor
from twisted.internet.error import ConnectError
from twisted.internet.protocol import Protocol
from twisted.protocols import ftp
from zope.interface import implementer

from chevah_compat import FileAttributes
from chevah.server import force_unicode
from chevah.server.commons.constant import (
    CONFIGURATION_SECTION_LOCATIONS,
    STATE,
    TYPE_NAME,
    )
from chevah.server.commons.exception import (
    ConnectException,
    ServerException,
    )
from chevah.server.commons.interface import (
    IFTPLocationConfiguration,
    IFTPSELocationConfiguration,
    IFTPSILocationConfiguration,
    )
from chevah.server.commons.factory import FactoryForProtocol
from chevah.server.commons.ssl import (
    SSLClientMixin,
    SSLConfigurationSectionMixin,
    )
from chevah.server.ftp.client import (
    FTPClientProtocol,
    FTPSEClientProtocol,
    FTPSIClientProtocol,
    )
from chevah.server.configuration.model import (
    WritableBoolean,
    WritableString,
    )
from chevah.server.configuration.option import (
    AddressPortOptionsMixin,
    UsernamePasswordPublicOptionsMixin,
    )
from chevah.server.location.commons import LocationBase
from chevah.server.location.configuration import _LocationConfigurationBase


@implementer(IFTPLocationConfiguration)
class FTPLocationConfiguration(
        _LocationConfigurationBase,
        AddressPortOptionsMixin,
        UsernamePasswordPublicOptionsMixin,
        ):
    """
    Configuration for an FTP location.
    """

    SECTION_PREFIX = CONFIGURATION_SECTION_LOCATIONS
    PORT_DEFAULT = 21

    debug = WritableBoolean('debug', default=False)
    acct = WritableString('acct', default='')
    ignore_passive_address = WritableBoolean(
        'ignore_passive_address', default=True)


@implementer(IFTPSELocationConfiguration)
class FTPSELocationConfiguration(
        FTPLocationConfiguration, SSLConfigurationSectionMixin):
    """
    Configuration for an Implicit FTPS location.
    """

    SECTION_PREFIX = CONFIGURATION_SECTION_LOCATIONS
    PORT_DEFAULT = 21

    ftps_reuse_session = WritableBoolean('ftps_reuse_session', default=True)


@implementer(IFTPSILocationConfiguration)
class FTPSILocationConfiguration(
        FTPLocationConfiguration, SSLConfigurationSectionMixin):
    """
    Configuration for an Implicit FTPS location.
    """

    SECTION_PREFIX = CONFIGURATION_SECTION_LOCATIONS
    PORT_DEFAULT = 990
    ftps_reuse_session = WritableBoolean('ftps_reuse_session', default=True)


class AbortingEndpoint(endpoints.TCP4ClientEndpoint):
    """
    An endpoint which disconnects after connection.
    """

    def connect(self, factory):
        """
        Called when we should connect to the remote server.
        """
        class DisconnectingProtocol(Protocol):
            """
            A simple protocol which closes the connection.
            """
            def connectionMade(self):
                # Instead of a nice close, brutally abort the connection here
                # to let the client know that things are bad.
                # Otherwise the client might think that it got an empty
                # transfer.
                self.transport.abortConnection()

        def cb_connected(ignored):
            """
            Called when we are connected.
            """
            raise ftp.ConnectionLost('Client in shutdown.')

        deferred = super(AbortingEndpoint, self).connect(
            FactoryForProtocol(DisconnectingProtocol()))
        deferred.addCallback(cb_connected)
        return deferred


class FTPLocationAbstract(LocationBase):
    """
    Shared code by FTP/FTPS based locations.

    This has 2 attributes for keeping track of the connected client:
    * _protocol = raw client, which might not be connected or authorized
    * _client = defined only when the client is connected and authorized.
    """

    # Used by FTPS for each connection.
    _context_factory = None

    def _getProtocol(self):
        """
        Return the client protocol used to connect to the FTP server.

        This is called before `_getClientEndpoint` so it can used for preparing
        things which are shared by the protocol and endpoint.
        """
        raise NotImplementedError('_getProtocol not implemented.')

    def _getClientEndpoint(self):
        """
        Return the client endpoint used to connect to the FTP server for the
        command channel.
        """
        raise NotImplementedError('_getClientEndpoint not implemented.')

    def _getClientPassiveEndpoint(self, host, port):
        """
        Return the client endpoint used to connect to the FTP server for the
        data channel via passive node.

        The client should call _getPassiveEndpoint.
        """
        raise NotImplementedError('_getClientPassiveEndpoint not implemented.')

    def _beforeConnection(self):
        """
        Called before the FTP command channel connection is initiated.

        Should return a deferred.
        """
        raise NotImplementedError('_beforeConnection not implemented.')

    # Reactor used for TCP connections.
    _reactor = reactor

    PROPERTIES_REQUIRING_RESTART = (
        'address',
        'password_update',
        'port',
        'username',
        'idle_connection_timeout',
        'idle_connection_keepalive_interval',
        'connection_retry_count',
        'connection_retry_interval',
        )

    def __init__(self, parent, configuration):
        super(FTPLocationAbstract, self).__init__(parent, configuration)
        # Low level FTP transport protocol.
        self._protocol = None

    def _debug(self, details):
        """
        By default, don't do nothing.
        """

    def _debugDisabled(self, details):
        """
        Method used when debug is disabled.
        """

    def _debugEnabled(self, details):
        """
        Method used when debug is enabled.
        """
        self.emitEvent(
            '20000',
            data={'message': details},
            )

    def _getPassiveEndpoint(self, client, address, port):
        """
        Called when client needs to open a passive connection.

        This is a wrapper for _getClientPassiveEndpoint
        """
        if self.state != STATE.STARTED:
            # Called when the location is stopping, but the server is still
            # waiting for us to connect.
            # Fox example when do a passive request but then client is closed
            # before finalizing the LIST or RETR.
            # We are nice and connect to the server so that it will not wait
            # for data connection, but after connection we abort.
            #
            # If you are not doing that, and close the client after server has
            # started the data port, you will have a bad time with the
            # tests as the server will hang waiting for data connection.
            return AbortingEndpoint(
                reactor=self._reactor,
                host=address,
                port=port,
                timeout=1,
                )

        return self._getClientPassiveEndpoint(address, port)

    def _onConnect(self):
        """
        See: LocationBase.
        """
        if self.configuration.debug:
            self._debug = self._debugEnabled
        else:
            self._debug = self._debugDisabled

        # Convert a FTP ConnectionLost into ServerException, so that we don't
        # trigger an ISE error.
        def eb_connection_lost(failure):
            """
            Called when the connection was closed.
            """
            # Beside the Twisted ConnectError we might also get a FTP error
            # due to TLS negotiation.
            failure.trap(ftp.ConnectionLost, ConnectError)
            message = 'Connected to %s:%s. %s' % (
                self.configuration.address,
                self.configuration.port,
                force_unicode(failure.value),
                )

            raise ConnectException(message)

        deferred = self._doConnect()
        deferred.addErrback(eb_connection_lost)
        return deferred

    @defer.inlineCallbacks
    def _doConnect(self):
        """
        Do the actual connection.
        """
        yield self._beforeConnection()
        # Make sure getProtocol is called before _getClientEndpoint.
        self._protocol = self._getProtocol()
        endpoint = self._getClientEndpoint(
            address=self.configuration.address,
            port=self.configuration.port,
            )
        yield endpoints.connectProtocol(endpoint, self._protocol)
        # Once protocol is connected we can call the specialized method to get
        # the client.
        client = yield self._protocol.authenticate()

        # Connection was authenticated so we can set the connection for binary
        # transfers, in "file" structure and "stream" mode.
        try:
            yield client.queueStringCommand('STRU F')
        except Exception:
            """
            Any error is ignored.
            """

        yield client.queueStringCommand('TYPE I')

        try:
            yield client.queueStringCommand('MODE S')
        except Exception:
            """
            Any error is ignored.
            """

        features = yield client.getFeatures()

        if 'CLNT' in features:
            # Server provides support for CLNT command.
            # Better use it as without it some servers will fail at the
            # OPTS UTF8 stage.
            yield client.queueStringCommand('CLNT SFTPPlus')

        if 'UTF8' in features:
            # Server support UTF-8.
            # For RFC 2640 this is enough to know that server will use UTF-8
            # but so servers implement the draft-ietf-ftpext-utf-8-option-00
            # so we also try to send the OPTS UTF8 command, but ignoring
            # any error.
            try:
                yield client.queueStringCommand('OPTS UTF8 ON')
            except ftp.CommandFailed:
                # Server might not support UTF-8 option and just implement
                # RFC 2640 without the draft stuff.
                # At this time we ignore the error and pretend like UTF-8 is
                # available.
                pass

        # Detect server path format.
        try:
            lines = yield client.queueStringCommand('PWD')
            self._setPathHandlingModule(lines[0])
        except ftp.CommandFailed:
            # If PWD is not supported,
            # just fallback to Unix paths.
            self._setPathHandlingModule('/')

        defer.returnValue(client)

    @defer.inlineCallbacks
    def _onDisconnect(self):
        """
        See: `LocationBase`.
        """
        if self._context_factory:
            context_factory = self._context_factory
            self._context_factory = None
            context_factory.stop()

        if not (self._protocol and self._protocol.connected):
            # Not yet connected. Nothing to disconnect.
            return

        # Send the QUIT command to inform the server that we want to
        # close the connection in a clean way.
        # This might delay the close, but we have idle_connection_timeout
        # which can cut the connection when it takes to long.
        try:
            if self._client:
                # Only send the quit command if we are connected.
                # Client failing, will try to send the QUIT command.
                yield self._client.fail(ServerException('Disconnected.'))
        except ftp.ConnectionLost:
            """
            Even if we send the quit command when we are connected, it
            might still fail... but we don't care.
            """
        finally:
            # In any case make sure that connection is closed.
            if self._protocol and self._protocol.transport:
                # We have a client which was connected at some time.
                # Force disconnection in case we are not yet disconnected.
                self._protocol.transport.loseConnection()

            # But wait for the close confirmation.
            yield self._protocol.disconnect_deferred

    def rename(self, source_path, destination_path):
        """
        See: ILocation.
        """
        # FIXME:4455:
        # The client should take Unicode input.
        # Here it takes UTF-8.
        return self._client.rename(
            source_path, destination_path)

    def deleteFile(self, path):
        """
        See: ILocation.
        """
        return self._client.removeFile(path)

    def deleteFolder(self, path, recursive=False):
        """
        See: ILocation.
        """
        return self._client.removeDirectory(path)

    def createFolder(self, path):
        """
        See: ILocation.
        """
        return self._client.makeDirectory(path)

    @defer.inlineCallbacks
    def getFolderMemberAttributes(self, path):
        """
        See: ILocation.
        """
        result = {}

        members = yield self._client.list(path)
        for member in members:
            result[member.name] = member

        defer.returnValue(result)

    @defer.inlineCallbacks
    def getAttributes(self, path):
        """
        See: ILocation.
        """
        parts = self.splitPath(path)
        name = parts[-1]
        if len(parts) > 1:
            parent = self.joinPath(*parts[:-1])
        else:
            # No parent exists.
            parent = ''

        # FIXME:3887:
        # Add support for relative and for root.
        if not name and parent == '/':
            defer.returnValue(FileAttributes(
                name='',
                path='/',
                is_folder=True,
                ))
        # FTP does not have a dedicated method to get the attributes.
        # We get the attributes from human readable folder listing by
        # listing the parent.
        # FIXME:3905:
        # If machine listing is available, use that.
        result = yield self.getFolderMemberAttributes(parent)
        if name not in result:
            raise ftp.CommandFailed(['500 Path not found to get attributes'])

        defer.returnValue(result[name])


class FTPLocation(FTPLocationAbstract):
    """
    Location to a FTP server.
    """

    def _getProtocol(self):
        """
        See: `FTPLocationAbstract`.
        """
        return FTPClientProtocol(
            location=self,
            dtp_endpoint_factory=self._getPassiveEndpoint,
            debug=self._debug,
            )

    def _getClientEndpoint(self, address, port):
        """
        See: `FTPLocationAbstract`.
        """
        result = endpoints.TCP4ClientEndpoint(
            reactor=self._reactor,
            host=address,
            port=port,
            # FIXME:2672:
            # Add configurable timeout.
            timeout=10,
            )
        result.use_ssl = False
        return result

    def _getClientPassiveEndpoint(self, address, port):
        """
        See: `FTPLocationAbstract`.
        """
        return self._getClientEndpoint(address, port)

    def _beforeConnection(self):
        """
        See: `FTPLocationAbstract`.
        """
        # Nothing special for plain FTP.
        return defer.succeed(None)


class _FTPSLocation(FTPLocation, SSLClientMixin):
    """
    Based code shared by implicit and explicit ftps locations.
    """

    def _getClientPassiveEndpoint(self, address, port):
        """
        See: `FTPLocationAbstract`.

        FTPS Explicit and Implicit always use SSL protected data communication.
        """
        return self._getSSLClientEndpoint(address, port)

    def _getSSLClientEndpoint(self, address, port):
        """
        Return a SSL/TLS protected endpoint.
        """
        result = endpoints.SSL4ClientEndpoint(
            reactor=self._reactor,
            host=address,
            port=port,
            sslContextFactory=self._context_factory,
            # FIXME:2672:
            # Add configurable timeout.
            timeout=10,
            )
        result.use_ssl = True
        return result

    def _beforeConnection(self):
        """
        See: `FTPLocationAbstract`.
        """
        # The context factory is created each time we connect-reconnect.
        # It is shared between the command (set in _getClientEndpoint) and data
        # connection (set in the protocol here).
        return self._initStartedContextFactory(remote_servers=[(
            self.configuration.address, self.configuration.port)])


class FTPSELocation(_FTPSLocation):
    """
    Location to an explicit FTP server.

    It connects to the command channel just like FTP, but it has an updated
    protocol which knows some FTPS commands.
    """

    PROPERTIES_REQUIRING_RESTART = (
        FTPLocationAbstract.PROPERTIES_REQUIRING_RESTART +
        SSLConfigurationSectionMixin.PROPERTIES_REQUIRING_RESTART
        )

    def _beforeConnection(self):
        """
        See: `FTPLocationAbstract`.
        """
        # FIXME:5459:
        # This can be removed once we can get CA chain for Explicit FTPS.
        if not self.configuration.ssl_certificate_authority:
            # Since we don't support FTPSE for CA chain,
            # we default to public CA roots.
            self.configuration.ssl_certificate_authority = (
                TYPE_NAME.MOZILLA_CA_ROOTS)

        # The context factory is created each time we connect-reconnect.
        # It is shared between the command (set in _getClientEndpoint) and data
        # connection (set in the protocol here).
        return self._initStartedContextFactory(remote_servers=[(
            self.configuration.address, self.configuration.port)])

    def _getProtocol(self):
        """
        See: `FTPLocationAbstract`.
        """
        return FTPSEClientProtocol(
            location=self,
            dtp_endpoint_factory=self._getPassiveEndpoint,
            ssl_context_factory=self._context_factory,
            debug=self._debug,
            )


class FTPSILocation(_FTPSLocation):
    """
    Location to an Implicit FTP server.

    It always connect directly with TLS/SSL and after connection the protocol
    will set up some FTPS specific commands.
    """

    PROPERTIES_REQUIRING_RESTART = (
        FTPLocationAbstract.PROPERTIES_REQUIRING_RESTART +
        SSLConfigurationSectionMixin.PROPERTIES_REQUIRING_RESTART)

    def _getProtocol(self):
        """
        See: `FTPLocationAbstract`.
        """
        return FTPSIClientProtocol(
            location=self,
            dtp_endpoint_factory=self._getPassiveEndpoint,
            ssl_context_factory=self._context_factory,
            debug=self._debug,
            )

    def _getClientEndpoint(self, address, port):
        """
        See: `FTPLocationAbstract`.
        """
        return self._getSSLClientEndpoint(address, port)
