# Copyright (c) 2022 Adi Roiban.
# See LICENSE for details.
"""
Locations for a remote HTTP server with a simple API consisting of a single
request.

It is designed to be used only as source.

This is just the location adapter without any HTTP protocol code.
The actual client-side protocol is found in the chevah.server.http package.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

from twisted.internet import defer
from zope.interface import implementer

from chevah_compat.posix_filesystem import FileAttributes

from chevah.server.commons.constant import (
    DEFAULT_LOCAL_FILESYSTEM_UUID,
    EVENT_TYPE,
    TYPE_NAME,
    )
from chevah.server.commons.exception import (
    HTTPClientException,
    ServerException,
    )
from chevah.server.commons.interface import (
    IHTTPPullLocationConfiguration,
    IWatchHandler,
    )
from chevah.server.commons.match import get_matched_groups
from chevah.server.commons.ssl import (
    SSLClientMixin,
    SSLConfigurationSectionMixin,
    )
from chevah.server.configuration.model import (
    WritableLines,
    WritableString,
    WritableStringCaseInsensitive,
    )
from chevah.server.configuration.option import (
    URLOptions,
    UsernamePasswordPublicOptionsMixin,
    )
from chevah.server.location.commons import (
    SSLLocationBase,
    )
from chevah.server.location.configuration import _LocationConfigurationBase
from chevah.server.location.watch import (
    ChevahLoopingCall,
    FilesystemEvent,
    )
from chevah.server.http.client import HTTPDownloadedFile
from chevah.server.http.client_shell import _HTTPClientShell


@implementer(IHTTPPullLocationConfiguration)
class HTTPPullLocationConfiguration(
        _LocationConfigurationBase,
        UsernamePasswordPublicOptionsMixin,
        SSLConfigurationSectionMixin,
        URLOptions,
        ):
    """
    Configuration for the HTTP/HTTPS location from where files are pulled.
    """
    request_method = WritableStringCaseInsensitive(
        'request_method', default='get')
    request_body = WritableStringCaseInsensitive('request_body', default='')
    request_headers = WritableLines('request_headers')
    # We use `2**` which has the same effect as `2*` to make it easier to read.
    response_code = WritableString('response_code', default='2**')
    response_file_name = WritableString(
        'response_file_name', default='http-pull.file')


@implementer(IWatchHandler)
class HTTPPullWatchHandler(object):
    """
    A placeholder watcher that doesn't really watch the remote server,
    but instead fakes a fixed snapshop at each changes_poll_interval.

    This is called by the transfer to start monitoring the source.
    `configuration` is the transfer configuration.
    """

    def __init__(
        self, configuration, handler, on_error, on_snapshot, scheduler,
        location_configuration,
            ):
        self._configuration = configuration
        self._scheduler = scheduler
        self._handler = handler
        self._onError = on_error
        self._onSnasphot = on_snapshot

        if self._configuration.recursive:
            raise ServerException(
                'http-pull location has no support for recursive transfers.')

        self._loop = ChevahLoopingCall(self._lcPull)
        self._loop.clock = scheduler

        self._location_configuration = location_configuration

    def start(self):
        """
        Called when we are connected to the remote server.
        """
        self._loop.start(self._configuration.changes_poll_interval)

    def stop(self):
        """
        Called when we should no longer trigger pull file transfers.

        This is also called before start.
        """
        if self._loop.running:
            self._loop.stop()
        return self._location_configuration.url

    @property
    def running(self):
        """
        True if watch snapshots are still scheduled.
        """
        return self._loop.running

    def _lcPull(self):
        """
        Called each change interval.

        This should never fail, and instead call the `on_error` callback
        provided at init.
        """
        # We can't detect the file on the remote HTTP server.
        # We just push a fixed event.
        file_path = '/' + self._location_configuration.response_file_name
        event = FilesystemEvent(
            type=EVENT_TYPE.CREATED,
            source_path=file_path,
            is_folder=False,
            node_id=None,
            size=0,
            modified=0,
            )
        # FIXME:5837:
        # `handler` is the StableFilesystemEventHandler
        # This means that our requests are not pulled right away, but there
        # is a wait for the stable file.
        # One fix is tot add a marker for configuration.stable_interval
        # so that StableFilesystemEventHandler will consider the snaphost stable
        # right away.
        # Another hack is that instad of callling handler.handle(event)
        # we call handler._onChange(event) to bypass the stability checks.
        self._handler._onChange(event)

        new_snapshot = {'/': []}
        visit_order = ('/',)
        self._onSnasphot((new_snapshot, visit_order))


class HTTPPullShell(_HTTPClientShell):
    """
    A shell for a location that only supports downloading of remote files.
    This can only be used for a location that is the source for a transfer.

    The request download can be made using GET/POST/PUT and with an empty
    or a fixed request body.
    """

    def _isAuthenticationSuccessful(self, code):
        """
        Return True if http `code` is a valid response code for the HTTP server.
        """
        actual_code = str(code)
        expected_code = self._location.configuration.response_code
        return bool(
            get_matched_groups(actual_code, expected_code, 'response_code'))

    @defer.inlineCallbacks
    def exists(self, path):
        """
        See: ILocation.
        """
        # We first try to see if we can get the attributes,
        # as a basic check that file exists.
        yield self.getAttributes(path)
        defer.returnValue(True)

    def getAttributes(self, path):
        """
        See: ILocation.
        """
        # FIXME:5839:
        # This is used in other places, so before getting the attributes,
        # we should make a HEAD request to see if the URL is still valid.

        # Just fake the name, based on the path.
        name = path.rsplit('/', 1)[-1]
        return FileAttributes(
            name=name,
            path=path,
            size=0,
            is_file=True,
            is_folder=False,
            is_link=False,
            modified=0,
            mode=0,
            hardlinks=1,
            uid=0,
            gid=0,
            node_id=None,
            )

    def getFolderMemberAttributes(self, path):
        """
        See: ILocation.
        """
        # This is called before the transfer start to make sure
        # the URL exits.
        return {}

    @defer.inlineCallbacks
    def getStreamForReading(self, path):
        """
        Return a deferred with an `ITransferStream` opened for
        reading at the start of the stream.
        """
        body = yield self._getRequestBody(path)
        response = yield self._getResponse(body)

        # We use the same code as for auth to identity a successfull pull.
        if not self._isAuthenticationSuccessful(response.code):
            yield self._agent.readBodyWithIgnore(response)
            raise HTTPClientException(
                message='Failed to download. Unexpected response code.',
                path=path,
                code=response.code,
                phrase=response.phrase,
                )

        path = '/' + self._location.configuration.response_file_name
        defer.returnValue(HTTPDownloadedFile(
            path=path,
            location=self._location,
            response=response,
            ))

    def _getRequestBody(self, path):
        """
        Return the body used to request the file download.
        """
        # We have simple, fixed a request body.
        # The `path` is ignored, as we alwasy have the same body.
        body = self._location.configuration.request_body
        return defer.succeed(body.encode('utf-8'))

    def _getResponse(self, body):
        """
        Make the HTTP request and return the HTTP response.
        """
        headers = get_request_headers(
            self._location.configuration.request_headers)
        method_name = self._location.configuration.request_method
        if method_name in ['post', 'put']:
            request_method = getattr(self._agent, method_name)
            return request_method(
                self._location.configuration.url,
                headers=headers,
                body=body,
                )

        # Default to GET
        return self._agent.get(
            self._location.configuration.url, headers=headers)


def get_request_headers(raw_headers):
    """
    Retrun a dict with the HTTP headers to be used in the request, based on
    the values extracted from `raw_headers`.
    """
    result = {}
    for line in raw_headers:
        try:
            key, value = line.split(':', 1)
        except ValueError:
            raise ServerException(
                'Invalid value "%s" configured for "request_headers".' % (
                    line,))
        result[key.strip()] = value.strip()

    return result


class HTTPSeedPullShell(HTTPPullShell):
    """
    Download a remote file with a POST/PUT request for which the body
    is a local file.

    This is a mix for a local filesystem location and an http-pull location.
    """

    # Local uses to handle local files used as the seed for the download
    # requests.
    # This is injected after init by the location.
    _seed_location = None

    def exists(self, path):
        """
        See: ILocation.
        """
        return self._seed_location.exists(path)

    def getFolderMemberAttributes(self, path):
        """
        See: ILocation.

        This is called before the transfer start to make sure the source
        exists.
        """
        return self._seed_location.getFolderMemberAttributes(path)

    def getAttributes(self, path):
        """
        See: ILocation.

        This is called before a file transfer starts to check that the file
        still exists in the source.
        """
        return self._seed_location.getAttributes(path)

    @defer.inlineCallbacks
    def _getRequestBody(self, path):
        """
        Return the body used to request the file download.
        """
        # We extract the body from the local file.
        # This is a big ugly, as we go back to the source location
        # to obtain the file, but I don't have a better design.
        source = None
        result = []
        try:
            source = yield self._seed_location.getStreamForReading(path)
            while True:
                chunk = yield source.read(1024 * 10)
                if not chunk:
                    break
                result.append(chunk)
        finally:
            if source:
                yield source.close()

        defer.returnValue(b''.join(result))

    def deleteFile(self, path):
        """
        Delete the original file used to trigger this transfer.
        """
        # We can't delete the file on the remote URL.
        # only the local seed file.
        return self._seed_location.deleteFile(path)


class HTTPPullLocation(SSLLocationBase, SSLClientMixin):
    """
    Location for a remote HTTP server for pulling files from.

    This is Frankenstein as it combines a local filesystem location with an
    HTTP location.

    It uses the general location code so that the URL and authentication
    is valided and the source files are generated only when the URL
    is available.

    The HTTP server has no way to watch for files.
    This location can have a dummy watch that will always find a dummy
    file used to trigger the transfer.

    The location can also watch local files, and send them to the remote HTTP
    server in order to obtain the actual file.
    The local file serve as the "seed" for the resulting pulled file.
    The local "seed" file is pushed to the remote server with the goal of
    obtaining the targeted destination file payload.
    """

    PROPERTIES_REQUIRING_RESTART = (
        SSLConfigurationSectionMixin.PROPERTIES_REQUIRING_RESTART +
        (
            'url',
            'password_update',
            'username',
            'request_body',
            ))

    _SUPPORTED_METHODS = ['get', 'put', 'post']

    def _HTTPClient(self, location, context_factory, scheduler):
        """
        Return an instance that should do the high-level HTTP client operation.
        """
        klass = HTTPPullShell
        if self._configuration.request_body == TYPE_NAME.SEED_FILE:
            klass = HTTPSeedPullShell

        client = klass(
            location=location,
            context_factory=context_factory,
            scheduler=scheduler,
            )
        # Inject the seed location, in any shell.
        client._seed_location = self.root.locations.getSection(
            DEFAULT_LOCAL_FILESYSTEM_UUID)._context

        return client

    def _onStart(self, avatar=None):
        """
        Called when the location is started for the first time.
        """
        request_method = self._configuration.request_method
        if request_method not in self._SUPPORTED_METHODS:
            raise ServerException(
                'Invalid value "%s" for "request_method".' % (request_method,))

        if request_method == 'get' and self._configuration.request_body:
            raise ServerException(
                'When the "request_method = get" only empty "request_body" '
                'configuration is supported.'
                )

        # Validate the matching expression as start.
        get_matched_groups(
            '', self._configuration.response_code, 'response_code')

        # Check that we can extract the headers.
        get_request_headers(self._configuration.request_headers)

        return super(HTTPPullLocation, self)._onStart(avatar=avatar)

    def _getSourceWatch(self, configuration, handler, on_error, on_snapshot):
        """
        Get the watcher for the source.

        This is called when the transfer starts, in order to detect stable
        files on the source.
        """
        if self._configuration.request_body != TYPE_NAME.SEED_FILE:
            # This just users a rigged watch handler that will return
            # a pre-defined set of files.
            configuration.source_path = '/'
            configuration.source_filter = ''
            return HTTPPullWatchHandler(
                configuration,
                handler,
                on_error,
                on_snapshot,
                self._scheduler,
                location_configuration=self._configuration,
                )

        # Allow the normal source watch with filtering as stable period.
        return super(HTTPPullLocation, self)._getSourceWatch(
            configuration, handler, on_error, on_snapshot)
