# Copyright (c) 2014 Adi Roiban.
# See LICENSE for details.
"""
Location for a remote SFTP server.
"""
import stat
from six.moves.urllib.parse import urlparse

from twisted.conch.ssh import filetransfer
from twisted.internet import defer, endpoints, reactor
from twisted.internet.error import ConnectError
from txtorcon.socks import SocksError, TorSocksEndpoint

from chevah_compat import FileAttributes
from chevah.server.commons.constant import ADDRESS_UNKNOWN
from chevah.server.commons.credential import (
    PasswordCredentials,
    SSHPrivateKeyCredentials,
    )
from chevah.server.commons.exception import (
    ConnectException,
    ServerException,
    )
from chevah.server.location.commons import (
    convertToServerException,
    get_file_name,
    LocationBase,
    TransferFileBase,
    )
from chevah.server.ssh.configuration import load_ssh_key
from chevah.server.ssh.session import (
    SFTPClientSSHSession,
    SingleSSHConnectionService,
    )
from chevah.server.ssh.protocol import (
    ConnectionDoneBeforeProtocol,
    SSHClientProtocol,
    )


class SFTPLocation(LocationBase):
    """
    Location to a SFTP server.
    """

    PROPERTIES_REQUIRING_RESTART = (
        'address',
        'password_update',
        'port',
        'username',
        'ssh_cipher_list',
        'ssh_private_key',
        'ssh_server_identity',
        'idle_connection_timeout',
        'idle_connection_keepalive_interval',
        'connection_retry_count',
        'connection_retry_interval',
        'proxy',
        )

    def __init__(self, parent, configuration):
        super(SFTPLocation, self).__init__(parent, configuration)
        # Low level SSH transport protocol used by the SFTP client.
        self._protocol = None
        # Encryption validation is done when object starts.
        self._encryption = None

    def _onConnect(self):
        """
        See: `LocationBase`.

        Connect to the remote SSH server using a single SFTP session.

        To initiate a SFTP subsystem there a multiple steps to be done:
        * create TCP connection
        * create SSH transport/protocol
        * request user-auth service (handled by SSHClientProtocol)
        * create SFTP subsystem
        * request SFTP subsystem (done by SSHClientProtocol)

        Return a deferred which fires when SFTP connection is done and ready
        to process SFTP commands.
        """
        self._encryption = SSHClientProtocol.getEncryptionConfiguration(
            self.configuration.ssh_cipher_list)

        credentials = self._getCredentials()
        endpoint = self._getEndpoint()

        service = SingleSSHConnectionService(channel=SFTPClientSSHSession())
        # Since we use ssh-connection service with SFTPClientSSHSession we need
        # to set the deferred which will be called when SFTP starts.
        # This deferred is also used by SSHClientProtocol to signal errors.
        service.deferred = defer.Deferred()
        self._protocol = SSHClientProtocol(
            configuration=self.configuration,
            credentials=credentials,
            service=service,
            )

        self._encryption.setFor(self._protocol)

        deferred = endpoints.connectProtocol(endpoint, self._protocol)
        deferred.addErrback(self._ebConnectError, endpoint)
        deferred.addErrback(self._ebProxyError, endpoint)
        deferred.addCallback(self._cbConnected, service.deferred)
        # Service.deferred is fired with the high level SFTP client.
        deferred.addCallback(self._cbClientStarted)
        deferred.addErrback(self._ebConnected)
        return deferred

    def _cbConnected(self, ignored, service_deferred):
        """
        Called when we are connected to the server.
        """
        # Wait for SSH protocol to set up the transport, to authenticate and
        # set up SFTP session.
        return service_deferred

    def _ebConnected(self, failure):
        """
        Called when there is a failure after connection.

        For example no cipher could be negotiated.
        """
        failure.trap(ConnectionDoneBeforeProtocol)
        raise ServerException(str(failure.value))

    @defer.inlineCallbacks
    def _cbClientStarted(self, client):
        """
        Called when we have successfully initiated a SFTP session.
        """
        # See what kind of paths are in used by the server by getting the
        # absolute path for the current working directory.
        path = yield client.realPath('.')
        self._setPathHandlingModule(path)
        defer.returnValue(client)

    def _ebConnectError(self, failure, endpoint):
        """
        Called when we fail to connect to the server.
        """
        failure.trap(ConnectError)

        if isinstance(endpoint, endpoints.TCP4ClientEndpoint):
            # We are connected directly to the server.
            message = 'Connected to %s:%s.' % (endpoint._host, endpoint._port)
        else:
            # We are connected using the proxy.
            message = 'Connected to %s:%s via proxy %s:%s.' % (
                endpoint._host,
                endpoint._port,
                endpoint._proxy_ep._host,
                endpoint._proxy_ep._port,
                )

        raise ConnectException(message)

    def _ebProxyError(self, failure, endpoint):
        """
        Called when we fail to connect to the server.
        """
        failure.trap(SocksError)
        message = (
            'Proxy failed to process the connection. '
            'Connected to %s:%s via proxy %s:%s. %s.'
            ) % (
                endpoint._host,
                endpoint._port,
                endpoint._proxy_ep._host,
                endpoint._proxy_ep._port,
                failure.value.message,
                )

        raise ServerException(message)

    def _onDisconnect(self):
        """
        See: `LocationBase`.
        """
        # FIXME:2501:
        # Add real code here/wait for protocol to stop.
        # Maybe update client Class to have its own quit/close method.
        if not self._protocol:
            # SSH layer not initiated.
            return

        protocol = self._protocol
        self._protocol = None

        if not protocol.transport:
            # SSH layer not yet connected.
            return

        if not protocol.connected:
            # We are no longer connected.
            return

        deferred = defer.Deferred()
        protocol._service.deferred = deferred
        protocol.loseConnection()

        # Wait for connection to be closed, ignoring errors.
        deferred.addErrback(lambda _: None)
        return deferred

    def _getCredentials(self):
        """
        Called by the SSH client when it needs to send credentials to the
        server.
        """
        private_key = self.configuration.ssh_private_key
        if private_key:
            key = load_ssh_key(
                configuration=private_key,
                usage='ssh_private_key',
                password=self.configuration.password,
                )
            return SSHPrivateKeyCredentials(
                username=self.configuration.username,
                peer=ADDRESS_UNKNOWN,
                creator=self,
                key=key,
                )

        # Default to password authentication
        return PasswordCredentials(
            username=self.configuration.username,
            password=self.configuration.password,
            creator=self,
            peer=ADDRESS_UNKNOWN,
            )

    def _getEndpoint(self):
        """
        Return the endpoint used to connect to the server.
        """
        proxy = self.configuration.proxy
        if not proxy:
            # Do a direct connection.
            return endpoints.TCP4ClientEndpoint(
                reactor,
                self.configuration.address,
                self.configuration.port,
                # FIXME:2672:
                # Add configurable timeout.
                timeout=10,
                )

        url = urlparse(proxy)
        if url.scheme != 'socks5':
            raise ServerException('Unsupported proxy type: %s' % (url.scheme,))

        proxy_host = url.hostname
        if not proxy_host:
            raise ServerException('Proxy URL requires a host.')

        proxy_port = url.port
        if not proxy_port:
            # Go with default port.
            proxy_port = 1080

        # We make a connection the the proxy server and pass that connection
        # to the socks handlers.
        # The location will use the socks handler as its main endpoint.
        proxy_endpoint = endpoints.TCP4ClientEndpoint(
            reactor,
            proxy_host,
            proxy_port,
            # FIXME:2672:
            # Add configurable timeout.
            timeout=10,
            )
        payload_endpoint = TorSocksEndpoint(
            host=self.configuration.address.encode('ascii'),
            port=self.configuration.port,
            socks_endpoint=proxy_endpoint,
            )
        return payload_endpoint

    @defer.inlineCallbacks
    def exists(self, path):
        """
        See: ILocation.

        SFTP protocol does not have a dedicated request so we use the
        `Retrieving File Attributes` request.
        """
        try:
            yield self._client.getAttrs(path.encode('utf-8'))
            defer.returnValue(True)
        except defer.CancelledError:
            raise
        except Exception:
            defer.returnValue(False)

    # FIXME:4455:
    # The client should take Unicode input.
    # Here it takes UTF-8.
    def rename(self, source_path, destination_path):
        """
        See: ILocation.
        """
        return self._client.renameFile(
            source_path.encode('utf-8'), destination_path.encode('utf-8'))

    def createFolder(self, path):
        """
        See: ILocation.

        Folder is created with default attributes.
        """
        return self._client.makeDirectory(path.encode('utf-8'), {})

    def deleteFolder(self, path, recursive=False):
        """
        See: ILocation.
        """
        return self._client.removeDirectory(path.encode('utf-8'))

    def deleteFile(self, path):
        """
        See: ILocation.
        """
        return self._client.removeFile(path.encode('utf-8'))

    @defer.inlineCallbacks
    def getFolderMemberAttributes(self, path):
        """
        See: ILocation.
        """
        remote_path_encoded = path.encode('utf-8')
        result = {}

        folder_iterator = yield self._client.openDirectory(remote_path_encoded)

        for deferred_member in folder_iterator:
            try:
                filename, longname, attributes = yield deferred_member
            except (EOFError, StopIteration):
                # FIXME:5169:
                # See why stop iteration is needed now.
                # No more files.
                break

            if filename in [b'.', b'..']:
                # Ignore current and parent and other `meta` members.
                continue

            is_folder = False
            is_link = False
            is_file = False
            if longname.startswith(b'd'):
                is_folder = True
            elif longname.startswith(b'l'):
                is_link = True
            else:
                is_file = True

            name = filename.decode('utf-8')

            member_path = self.joinPath(path, name)

            result[name] = FileAttributes(
                name=name,
                path=member_path,
                size=attributes['size'],
                is_file=is_file,
                is_folder=is_folder,
                is_link=is_link,
                modified=attributes['mtime'],
                mode=attributes['permissions'],
                hardlinks=1,
                uid=attributes['uid'],
                gid=attributes['gid'],
                )

        folder_iterator.close()
        defer.returnValue(result)

    def _openTransferFile(self, path, flags, mode):
        """
        Called when a remote SFTP file needs to be opened.
        """
        remote_path_encoded = path.encode('utf-8')
        deferred = self._client.openFile(remote_path_encoded, flags, {})
        deferred.addCallback(lambda remote_file: SFTPTransferFile(
            stream=remote_file,
            location=self,
            path=path,
            mode=mode,
            ))
        return deferred

    def getAttributes(self, path):
        """
        See: ILocation.
        """
        def cb_got_raw_attributes(result):
            """
            Called when we got the attributes.
            """
            name = get_file_name(path)
            mode = result['permissions']
            return FileAttributes(
                name=name,
                path=path,
                size=result['size'],
                modified=result['mtime'],
                uid=result['uid'],
                gid=result['gid'],
                mode=mode,
                is_file=stat.S_ISREG(mode),
                is_folder=stat.S_ISDIR(mode),
                )

        deferred = self._client.getAttrs(path.encode('utf-8'))
        deferred.addCallback(cb_got_raw_attributes)
        return deferred

    def getStreamForReading(self, path):
        """
        See: ILocation.
        """
        flags = filetransfer.FXF_READ
        return self._openTransferFile(path, flags, 'reading')

    def getStreamForWriting(self, path, length=None):
        """
        See: ILocation.

        `length` is ignored as for SFTP we don't need to announce how much
        we will write.
        """
        flags = (
            filetransfer.FXF_WRITE |
            filetransfer.FXF_CREAT |
            filetransfer.FXF_TRUNC
            )
        return self._openTransferFile(path, flags, 'writing')

    def getStreamForAppending(self, path):
        """
        See: ILocation.
        """
        flags = (
            filetransfer.FXF_WRITE |
            filetransfer.FXF_APPEND |
            filetransfer.FXF_CREAT
            )
        return self._openTransferFile(path, flags, 'appending')


class SFTPTransferFile(TransferFileBase):
    """
    A file on remote SFTP location.

    `_stream` is a twisted.conch.interfaces.ISFTPFile and we need to keep the
    remote file cursor in the `_offset`.
    """

    def __init__(self, stream, location, path, mode):
        super(SFTPTransferFile, self).__init__(stream, location, path, mode)
        self._offset = 0

    @convertToServerException
    @defer.inlineCallbacks
    def read(self, length):
        """
        See: ITransferStream.
        """
        try:
            data = yield self._stream.readChunk(self._offset, length)
        except EOFError:
            # Return an empty data as a signal that there is no more data to
            # read.
            defer.returnValue(b'')

        self._offset += len(data)
        defer.returnValue(data)

    @convertToServerException
    @defer.inlineCallbacks
    def write(self, data):
        """
        See: ITransferStream.
        """
        initial_offset = self._offset
        # See https://tools.ietf.org/html/rfc4253#section-6.1
        CHUNK_SIZE = 32 * 1024
        relative_cursor = 0
        deferreds = []
        while True:
            chunk = data[relative_cursor:relative_cursor + CHUNK_SIZE]
            if not chunk:
                break
            deferreds.append(self._stream.writeChunk(self._offset, chunk))
            relative_cursor += len(chunk)
            self._offset += len(chunk)

        try:
            # Some SFTP servers don't accept big chunks so we split into
            # smaller chunks and send them in parallel.
            yield defer.DeferredList(
                deferreds, fireOnOneErrback=True, consumeErrors=True)
        except Exception as error:
            if isinstance(error, defer.FirstError):
                error = error.args[0]
            self._offset = initial_offset
            raise error
