# Copyright (c) 2015 Adi Roiban.
# See LICENSE for details.
"""
Command line shell for client side of a file transfer location.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import os
import sys
from cmd import Cmd
from getpass import getpass
from io import StringIO

import six
from six.moves import input
from bunch import Bunch
from twisted.internet import defer, reactor
from twisted.internet.task import LoopingCall
from unidecode import unidecode

from chevah_compat import local_filesystem, process_capabilities
from chevah.server import (
    emit as _emit,
    emit_error_details,
    emit_from_error,
    event_dispatcher,
    force_unicode,
    )
from chevah.server.commons.configuration import (
    ApplicationConfiguration,
    parse_url,
    )
from chevah.server.commons.constant import (
    CONFIGURATION_DEFAULTS,
    LOCATION_NAMES,
    PRODUCT_NAME,
    PRODUCT_VERSION,
    SSL_DEFAULTS,
    )
from chevah.server.commons.csv import from_csv_row
from chevah.server.commons.exception import ServerError, ServerException
from chevah.server.event.model import ChevahTimeStamp
from chevah.server.location.manager import LocationsManager


class StringIOWithLimit(StringIO, object):
    """
    A StringIO which prevent using all memory.

    It handles everything as unicode text.
    """

    def __init__(self, initial_value='', maximum_size=100 * 100):
        super(StringIOWithLimit, self).__init__(initial_value=initial_value)
        self._maximum_size = maximum_size

    def write(self, s):
        if isinstance(s, bytes):
            # On Py2 we receive data as bytes.
            s = s.decode('utf-8')

        if len(s) + len(self.getvalue()) > self._maximum_size:
            # Just ignore more data.
            return
        super(StringIOWithLimit, self).write(s)


class StringIOWithoutClose(StringIO, object):
    """
    A StringIO that will ignore any close operation and will emit an info
    message instead.
    """

    # The file save operation might be done over multiple write calls.
    # We only want to emit an event on the first write operation.
    # This is a quick hack to get something working.
    _already_write = False

    def write(self, data):
        """
        Write data, but also make sure we don't have data left
        after previous closes.
        """
        result = super(StringIOWithoutClose, self).write(data)
        self.truncate(self.tell())
        if not self._already_write:
            self._already_write = True
            _emit('20000', data={
                'message': 'Configuration save operation ignored.'})
        return result

    def close(self):
        """
        Store the content and reset the cursor without closing the file.
        """
        self._already_write = False
        self.seek(0)


class ClientShell(object):
    """
    Command line shell for an ILocation.

    It can be started with command line options, or without any argument.

    The shell exposes a set of `options` which are configuration options
    either for the ILocation or the shell itself.
    Any option is available as a command line argument.
    """

    # Low level commands kept to help with testing.
    _cmd = None

    def __init__(self):
        # Here to avoid circular import due to EmbeddedClientShellCommands
        # usage in the process via the transfers.
        from chevah.server.commons.command import UnicodeArgumentParser
        self._cmd = ClientShellCommands()
        script_name = 'client-shell'
        self._parser = UnicodeArgumentParser(prog=script_name)
        self._populateParser(self._parser)

    def _populateParser(self, parser):
        """
        Called to add options to the parser.

        All command lines defined as text as they are use internally in a
        similar way to the `set` command.
        """
        parser.add_argument(
            'url', nargs='?',
            action='store',
            metavar="proto://(user@)address(:port)(/path)", default=None,
            help=(
                'URL used by the connection step. Initial working '
                'path, user and port are optional. '
                'Available protocols: %s.'
                ) % (', '.join(list(LOCATION_NAMES.keys()),)),
            )

        # Generic options for a location.
        parser.add_argument(
            '-u', '--username',
            action='store', dest='username',
            metavar="NAME", default='',
            help='Username used by the authentication step.',
            )
        parser.add_argument(
            '-p', '--password',
            action='store', dest='password_update',
            metavar="PASS", nargs='?', default='',
            help=(
                'Password used by the authentication step. '
                'If no value is provided on the command line, '
                'it will be asked at runtime in a secure way.'
                ),
            )
        parser.add_argument(
            '--idle-connection-timeout',
            action='store', dest='idle_connection_timeout',
            metavar="SECONDS", default='30',
            help=(
                'Number of seconds after which inactive connections are '
                'closed.'
                ),
            )
        parser.add_argument(
            '--idle-connection-keepalive-interval',
            action='store', dest='idle_connection_keepalive_interval',
            metavar="SECONDS", default='0',
            help=(
                'Number of seconds after which a keep alive command is sent '
                'to prevent the server from disconnecting the session.'
                ),
            )
        parser.add_argument(
            '--connection-retry-count',
            action='store', dest='connection_retry_count',
            metavar="NUMBER", default='0',
            help='Number of time to retry a connection.',
            )
        parser.add_argument(
            '--connection-retry-interval',
            action='store', dest='connection_retry_interval',
            metavar="SECONDS", default='0',
            help='Number of seconds to wait between failed connection.',
            )
        parser.add_argument(
            '--proxy',
            action='store', dest='proxy',
            metavar="URI", default='',
            help=(
                'Proxy expression to configure the proxy to be used.'
                ),
            )

        # FTP(S) options.
        parser.add_argument(
            '--acct',
            action='store', dest='acct',
            metavar="FTP-ACCT-VALUE", default='',
            help=(
                'Only for FTP or FTPS. '
                'ACCT value if requested by the server. '
                'Default: empty.'
                ),
            )
        parser.add_argument(
            '--ignore-pasv-ip',
            action='store_true', dest='ignore_passive_address',
            default=False,
            help=(
                'Only for FTP or FTPS. '
                'Ignore the IP address returned by the PASV command. '
                'Default: yes.'
                ),
            )
        parser.add_argument(
            '--no-reuse-session',
            action='store_false', dest='ftps_reuse_session',
            default=True,
            help=(
                'Only for FTPS. '
                'Don\'t enforce TLS session resumption. '
                'Default: yes.'
                ),
            )

        # SSH options.
        parser.add_argument(
            '--ssh-server-identity',
            action='store', dest='ssh_server_identity',
            metavar="FINGERPRINT", default='',
            help='Fingerprint of the remote SSH server (MD5, SHA1 or SHA256).',
            )
        parser.add_argument(
            '--ssh-private-key',
            action='store', dest='ssh_private_key',
            metavar="KEY_PATH", default='',
            help='Path to private key used by the authentication step.',
            )

        # SSL options.
        ssl_methods_default = ','.join(SSL_DEFAULTS['ssl_allowed_methods'])
        parser.add_argument(
            '--ssl-allowed-methods',
            action='store', dest='ssl_allowed_methods',
            metavar="COMMA_DELIMITED_METHODS", default=ssl_methods_default,
            help=(
                'Comma separated list with TLS/SSL accepted methods. '
                'Default: %s' % (ssl_methods_default,)),
            )

        default_cipher = SSL_DEFAULTS['ssl_cipher_list']
        parser.add_argument(
            '--ssl-cipher-list',
            action='store', dest='ssl_cipher_list',
            metavar="OPENSSL:CIPHERS", default=default_cipher,
            help=(
                'List of accepted ciphers in OpenSSL format. '
                'Default: %s' % (default_cipher,)
                ),
            )
        parser.add_argument(
            '--ssl-certificate',
            action='store', dest='ssl_certificate',
            metavar="CERT_PATH", default='',
            help=(
                'Path to X.509 certificate to be used by the client. '
                'Default: do not send a certificate.'
                ),
            )
        parser.add_argument(
            '--ssl-key',
            action='store', dest='ssl_key',
            metavar="KEY_PATH", default='',
            help=(
                'Path to private to be used by the client for SSL/TLS. '
                'Default: do not use a key.'
                ),
            )
        parser.add_argument(
            '--ssl-key-password',
            action='store', dest='ssl_key_password_update',
            metavar="PASSWORD", default='',
            help='Password used to decrypt the private key.',
            )
        parser.add_argument(
            '--ssl-certificate-authority',
            action='store', dest='ssl_certificate_authority',
            metavar="CRL_PATH", default='',
            help='Path to X.509 CA certificate accepted by the client.',
            )
        parser.add_argument(
            '--ssl-certificate-revocation-list',
            action='store', dest='ssl_certificate_revocation_list',
            metavar="PATH_OR_URL", default='',
            help=(
                'Path or URL to a CRL used for validation. '
                'Can be called multiple times for using multiple CRLs.'
                ),
            )
        parser.add_argument(
            '--ssl-certificate-revocation-list-refresh',
            action='store', dest='ssl_certificate_revocation_list_refresh',
            metavar="SECONDS", default=0,
            help=(
                'Time after which the CRL is reloaded. '
                'Default: 0'
                ),
            )

        parser.add_argument(
            '-d', '--debug',
            action='store_true', dest='debug', default=False,
            help='Enable low-level protocol debug messages.',
            )

        # Non location related.
        # These are exceptions and are not text.
        parser.add_argument(
            '-b', '--batch',
            action='store', dest='batch',
            metavar="PATH", default=None,
            help='Execute commands from a batch file.',
            )
        parser.add_argument(
            '-q', '--quiet',
            action='store_true', dest='quiet', default=False,
            help='Run without sending any message to standard output.',
            )

    def start(self, args=None):
        """
        Called when shell should be started.

        `args` is a list of arguments used to start the shell.
        """
        arguments = self._parseArguments(args)

        # The client shell is not refreshing the CRL.
        arguments.ssl_certificate_revocation_list_refresh = '0'

        if arguments.password_update is None:
            if arguments.batch:
                message = (
                    'Interactive password input is not compatible'
                    ' with the batch mode.'
                    )
                self._exit(1, message)
                return

            arguments.password_update = self._cmd._getPasswordInput(
                'Password: ')

        # Load batch command if available.
        if arguments.batch:
            lines = self._readLinesFromPath(arguments.batch)
            # Make sure command will exit at the end.
            lines.append('exit')
            self._cmd.setBatchCommands(lines)
        else:
            # FIXME:2912:
            # We should add the option for predefined commands like
            # upload/download/list.
            pass

        # Update command variables based on command line arguments.
        # For now all available options should have an associated command line
        # argument.
        for name in self._cmd.available_options.keys():
            value = getattr(arguments, name, None)
            if value is None:
                # Option not provided in CLI arguments.
                continue
            self._cmd.updateFromCLI(name, value)

        self._cmd.start()
        self._exit(0)

    def _parseArguments(self, args=None):
        """
        Parse command line arguments.

        Return a namespace with parsed options.
        """
        if args is None:
            args = sys.argv[1:]

        # Convert all arguments to Unicode as this is an entry point.
        unicode_args = []
        for arg in args:
            if not isinstance(arg, six.text_type):
                arg = arg.decode('utf-8')
            unicode_args.append(arg)

        arguments = self._parser.parse_args(unicode_args)
        # Protocol, host, port and path are extracted from the URL.
        self._parseURL(arguments)
        return arguments

    def _parseURL(self, arguments):
        """
        Parse the URL and update the arguments.
        """
        raw_url = arguments.url

        # Initialize the arguments as they are going to be translated into
        # variables even if we don't have an URL.
        arguments.protocol = ''
        arguments.address = ''
        arguments.port = '1'  # Needs to be a text value that can be an int.
        arguments.pwd = ''

        if not raw_url:
            return

        try:
            url = parse_url(raw_url)
        except ServerException as error:
            self._exit(1, error.message)
            return  # noqa:cover

        # Update arguments so they can be translated into variables.
        arguments.protocol = url['protocol']
        arguments.address = url['address']
        arguments.port = six.text_type(url['port'])
        arguments.pwd = url['path']

        # Only update username if defined inside the URL.
        if url['username']:
            if arguments.username and arguments.username == url['username']:
                self._exit(1, (
                    'Either define the username as argument or as part of '
                    'the URL.'),
                    )
                return  # noqa:cover
            arguments.username = url['username']

    def _readLinesFromPath(self, path):
        """
        Return a list of all lines from the file at `path`.
        """
        file = None
        lines = []
        try:
            segments = local_filesystem.getSegmentsFromRealPath(path)
            file = local_filesystem.openFileForReading(segments)
            for line in file:
                lines.append(line.decode('utf-8').strip())
        except Exception:
            self._exit(1, 'Fail to read: %s' % (path,))
        finally:
            if file:
                file.close()

        return lines

    def _exit(self, code, message=''):
        """
        Called when shell should end.
        """
        if message:
            if six.PY2 and process_capabilities.os_family != 'posix':
                # On Windows on Py2 we get IOError if we try to send Unicode
                # characters even if they are UTF-8 encoded so we transliterate
                # to them plain ASCII.
                message = unidecode(message)

            message = message + '\n'

            if six.PY2 and isinstance(message, six.text_type):
                # On py2 make sure value are encoded before putting
                # them on the wire.
                # Py3 should just work with Unicode.
                message = message.encode('utf-8')

            sys.stderr.write(message)

        sys.exit(code)


class ClientShellProcess(object):
    """
    Minimal process to hold components required by the client shell.
    """
    def __init__(self):
        # Shell process as a volatile configuration.
        configuration = ApplicationConfiguration(
            configuration_file=StringIOWithoutClose(''), parent=self)
        configuration.loadExternalConfiguration()

        self.configuration = configuration
        self._proxy = configuration._proxy

        self.locations = LocationsManager(parent=self)

        self.event_dispatcher = event_dispatcher

        if not self.event_dispatcher.configured:
            # Dispatcher is a singleton.
            # We only configure the dispatcher if not already configured to
            # prevent configuring it multiple time in a test.
            self.event_dispatcher.configure(
                definitions=self.configuration.events,
                configuration=self.configuration,
                )

        from chevah.server.commons.process import ProcessPublicIdentity
        self.public_identity = ProcessPublicIdentity(self)
        configuration.server.name = 'client-shell'
        configuration.server._setUUID('client-shell')
        self.public_identity.setConfiguration(configuration)

    @property
    def root(self):
        return self

    @property
    def parent(self):
        return None


class BatchStopException(Exception):
    """
    Used to signal that shell should stop processing further commands in
    batch mode.
    """


def connection_required(func):
    """
    Decorator to set a method as requiring an active connection.
    """
    func.connection_required = True
    return func


class ClientShellCommands(Cmd, object):
    """
    Command handling for the client shell.

    It is based on `cmd` module but is modified without support for readline
    and with support for commands which might return a deferred.

    Commands are implemented using do_COMMAND_NAME methods.
    On errors command can raise ServerException.
    Command can return a deferred and the deferred result is sent to the
    shell output.
    """

    # All option available to this shell with their default values.
    # All values are received as text in CLI arguments.
    # Here we keep the actual type to make the conversion.
    available_options = {
        'acct': '',
        'address': '',
        'debug': False,
        'idle_connection_keepalive_interval': 0,
        'idle_connection_timeout': 30,
        'connection_retry_count': 0,
        'connection_retry_interval': 60,
        'ignore_passive_address': True,
        'ftps_reuse_session': True,
        'password_update': '',
        'port': 1,
        'protocol': '',
        'proxy': '',
        'pwd': '/',
        'quiet': False,
        'ssh_server_identity': '',
        'ssh_private_key': '',
        'ssh_cipher_list': ('secure',),
        'username': '',
        'ssl_certificate': '',
        'ssl_key': '',
        'ssl_key_password_update': '',
        'ssl_certificate_authority': '',
        'ssl_certificate_revocation_list': (),
        'ssl_certificate_revocation_list_refresh': 0,
        'ssl_cipher_list': 'secure',
        'ssl_allowed_methods': 'secure',
        'authentication_method': '',
        'allow_insecure_http': False,
        'require_encryption': True,
        }

    # Options which are not set to the location.
    _ignored_options = [
        'description',
        'enabled',
        'name',
        'type',
        'ssl_domains',  # server-side only config.
        ]
    _ignore_azure = [
        'address',
        'port',
        'ssl_allowed_methods',
        'ssl_cipher_list',
        'ssl_certificate',
        'ssl_key',
        'ssl_key_password_update',
        'ssl_certificate_authority',
        'ssl_certificate_revocation_list',
        'ssl_certificate_revocation_list_refresh',
        ]
    _ignored_protocol_options = {
        'azure-file': _ignore_azure,
        'azure-blob': _ignore_azure,
        }

    use_rawinput = True
    prompt = '> '
    intro = '%s (%s) file transfer client shell' % (
        PRODUCT_NAME, PRODUCT_VERSION)

    doc_leader = (
        'For details about a command, type `help <command>`\n'
        'Command names are case insensitive and can be abbreviated \n'
        'as long as there are no conflicts.\n'
        )
    doc_header = 'Available commands:'

    _CHUNK_SIZE = 16 * 1024

    _reactor = reactor

    def __init__(self):
        # Initiate a new copy for the working options.
        self._options = self.available_options.copy()
        # Process used by the shell.
        self._process = None

        self._stderr = sys.stderr
        self._stdout = sys.stdout

        # List of batch commands.
        self._batch_queue = []
        # Deferred looping call for the prompt.
        self._loop_call = None
        # Current location used by the shell.
        self._location = None
        # Delayed called use to start the loop in the reactor.
        # This is here to help with testing.
        self._start_delayed_call = None
        self._is_batch_mode = False
        self._current_batch_line = 0

    def _onEmitSendFailure(self, signal):
        """
        Called when a new event is triggered as a signal.
        """
        event = signal.event
        dispatcher = self._process.event_dispatcher
        configuration = Bunch(
            shell_output_format=CONFIGURATION_DEFAULTS['entry_content'] + '\n')
        shell_output_format = configuration.shell_output_format
        output = event.renderFromTemplate(
            shell_output_format, 'shell_output_format')

        if dispatcher.definitions.eventHasGroup(event.id, 'failure'):
            if six.PY2:
                output = output.encode('utf-8')
            self._stderr.write(output)
            return

        self._sendFeedback(output)

    @property
    def stdout(self):
        """
        Expose this as read-only since the python cmd module is using it.
        """
        return self._stdout

    def setBatchCommands(self, commands):
        """
        Set the batch commands which should be executed when shell starts.
        """
        self._batch_queue = commands[:]

    def emitEvent(self, *args, **kwargs):
        """
        Emit an event as shell client.
        """
        _emit(*args, **kwargs)

    def start(self):
        """
        Called when command shell should start processing either interactive
        or batch mode.
        """
        # Only create a client process if we are running in interactive mode.
        self._process = ClientShellProcess()
        self._sendFeedback(self.intro)

        dispatcher = self._process.event_dispatcher
        dispatcher.subscribe('emit', self._onEmitSendFailure)

        if self._batch_queue:
            mode_callback = self._processBatch
        else:
            mode_callback = self._processInput
        self._loop_call = LoopingCall(
            self._loopCallback, mode_callback=mode_callback)
        self._loop_call.clock = self._reactor

        self._start_delayed_call = self._reactor.callLater(
            0, self._loop_call.start, interval=0, now=True)
        self._reactorStart()

    def _stop(self):
        """
        Called when we should stop waiting for command and exit.
        """
        dispatcher = self._process.event_dispatcher
        dispatcher.unsubscribe('emit', self._onEmitSendFailure)
        if self._loop_call and self._loop_call.running:
            self._loop_call.stop()
        self._reactorStop()

    def _reactorStart(self):  # noqa:cover
        """
        Helper for testing.
        """
        self._reactor.run()

    def _reactorStop(self):  # noqa:cover
        """
        Helper for testing.
        """
        self._reactor.stop()

    def _getRawInput(self, prompt):
        """
        Test helper.
        """
        if self._is_batch_mode:
            raise RuntimeError(
                'Interactive input not supported in batch mode at line %s.' % (
                    self._current_batch_line,))

        if six.PY2 and isinstance(prompt, six.text_type):
            prompt = prompt.encode('utf-8')

        return input(prompt)

    def _getPasswordInput(self, prompt):
        """
        Test helper.
        """
        if self._is_batch_mode:
            raise RuntimeError(
                'Interactive input not supported in batch mode at line %s.' % (
                    self._current_batch_line,))

        if six.PY2 and isinstance(prompt, six.text_type):
            prompt = prompt.encode('utf-8')

        return getpass(prompt)

    def get_names(self):
        """
        Patch python version to return all names from the instance,
        and not from the class.
        """
        return dir(self)

    def _loopCallback(self, mode_callback):
        """
        Called when a new command should be read and processed.

        `mode_callback` is the method used to process a new command.
        """
        # This just wraps the low level lopped call to make sure we catch
        # all errors.
        deferred = defer.maybeDeferred(mode_callback)
        deferred.addCallback(self._cbLogResult)
        deferred.addErrback(self._ebLogFailure)
        return deferred

    def _processInput(self):
        """
        Low level process of input for a single command in interactive mode.
        """
        if self._batch_queue:
            line = self._batch_queue.pop(0)
        else:
            try:
                line = self._getRawInput(self.prompt)
            except EOFError:
                self._stop()
                return
            except KeyboardInterrupt:
                self._stop()
                return

        return self._processLine(line)

    def _processBatch(self):
        """
        Low level processing of a single command from the batch/non-interactive
        mode.
        """
        self._is_batch_mode = True
        self._current_batch_line += 1

        if not self._batch_queue:
            raise BatchStopException()

        line = self._batch_queue.pop(0)
        return self._processLine(line)

    def _processLine(self, line):
        """
        Called to process and execute command from `line`.
        """
        line = line.strip()

        if not line:
            # No command to process. Wait for next command.
            return

        if line.startswith('#'):
            # Ignore comments.
            return

        parts = line.split(' ', 1)
        try:
            command, arguments = parts
        except ValueError:
            command = line
            arguments = ''

        # Set it as lower case.
        command = command.lower()

        # Resolve incomplete commands.
        candidate_commands = self.completenames(command)

        if len(candidate_commands) == 0:
            raise ServerException(
                'Unknown command. Type `help` to see all commands.')
        if len(candidate_commands) != 1:
            if command not in candidate_commands:
                raise ServerException(
                    'Abbreviation has multiple commands: %s' % (
                        ' '.join(candidate_commands),))
            candidate_commands[0] = command

        # We always get a command_method as command name was already
        # validated.
        # Command is requested with normalized name.
        command_method = getattr(self, 'do_' + candidate_commands[0])

        # Validate command prerequisites.
        self._validateCommand(command_method)

        return command_method(arguments)

    def _validateCommand(self, command):
        """
        Called before processing a command.

        Can be used to reject the command.
        """
        connection_required = getattr(command, 'connection_required', False)
        if (
            connection_required and
            (not self._location or not self._location.started)
                ):
            raise ServerException(
                'You need to open a connection before issuing this command.')

    def _sendOutput(self, message):
        """
        Send output to the user.
        """
        if not self._stdout:
            return
        if message is None:
            return
        if not message.endswith('\n'):
            message += '\n'

        if six.PY2:
            message = message.encode('utf-8')

        self._stdout.write(message)

    def _cbLogResult(self, result):
        """
        Called when command returns a result.
        """
        self._sendOutput(result)

    def _sendFeedback(self, message):
        """
        Send output to the user. Feedback is optional and muted when quiet.
        """
        if self._options['quiet']:
            return
        self._sendOutput(message)

    def _ebLogFailure(self, failure):
        """
        Called when a deferred has failed and need to emit an event.
        """
        if failure.check(BatchStopException):
            return failure

        if failure.check(ServerError):
            emit_from_error(failure.value, emitter=self.emitEvent)
            return

        # A command failed, we emit an event
        emit_error_details(
            error=failure.value,
            event_id='60002',
            data={},
            emitter=self.emitEvent,
            )

        if not failure.check(ServerException):
            self._stop()

    def _getConfiguration(self, protocol):
        """
        Return location configuration for `protocol`.
        """
        locations = self._process.configuration.locations
        configuration = locations.getSectionByName(protocol)

        if not configuration:
            # If we don't have yet a configuration for this protocol,
            # create one.
            settings = {
                'type': protocol,
                'name': protocol,
                }
            configuration = locations.createProperty('', settings)
            # We don't need delayed updates for this configuration.
            configuration._UPDATE_DELAY = 0

        return configuration

    def _getLocation(self):
        """
        Return the location to be used by the shell.
        """
        sane_options = self._options.copy()

        # Add custom handling for ignoring SSH server idenity.
        if sane_options['ssh_server_identity'] == 'set-on-first-connection':
            raise ServerException(
                'The value "set-on-first-connection" is not supported '
                'for the CLI tool. You can use IGNORE as a replacement.'
                )
        if sane_options['ssh_server_identity'] == 'IGNORE':
            sane_options['ssh_server_identity'] = 'set-on-first-connection'

        # Manually convert the options that should be a sequence.
        sane_options['ssh_server_identity'] = [
            sane_options['ssh_server_identity']]
        sane_options['ssl_allowed_methods'] = (
            sane_options['ssl_allowed_methods'].split(','))

        protocol = sane_options['protocol']

        available_protocols = list(LOCATION_NAMES.keys())
        if protocol not in list(LOCATION_NAMES.keys()):
            raise ServerException(
                'Protocol "%s" is not supported. Available protocols: %s' % (
                    protocol, ', '.join(available_protocols)))

        configuration = self._getConfiguration(protocol)

        for name in configuration.getPublicWritableAttributeNames():
            if name in self._ignored_options:
                continue

            protocol_ignored = self._ignored_protocol_options.get(protocol, [])
            if name in protocol_ignored:
                continue

            # The interface might have a configuration option which is not
            # exposed by the shell options, but we should fail in this case
            # as we want all configuration options to also be available
            # via the shell.
            value = sane_options[name]
            configuration.setValue(name, value)

        return self._process.locations.getSection(configuration.uuid)

    def do_exit(self, arguments):
        """
        Exit the shell.
        """
        self._stop()

    # Exit aliases
    do_bye = do_quit = do_exit

    def do_show(self, arguments):
        """
        Show variables or a subset of the variables.

        Password are masked when showing multiple variables.

        To unmaks the value, request to show a specific variable.
        """
        if arguments:
            # Keep the keys in the requested order.
            keys = arguments.lower().split(' ')
        else:
            # Use a well defined order to help with testing.
            keys = sorted(self._options.keys())

        result = []
        for key in keys:
            key = key.strip()
            if not key:
                continue

            if key.endswith('password'):
                key += '_update'

            try:
                value = self._options[key]
                if isinstance(value, (tuple, list, set)):
                    value = ', '.join(value)

                if 'password' in key and not arguments and value:
                    value = '*MASKED*'
                result.append('%s: %s' % (key, value))
            except KeyError:
                raise ServerException(
                    'Variable "%s" doesn\'t exists.' % (key,))
            except Exception as error:
                raise ServerException('Failed to show %s: %r' % (key, error))

        return '\n'.join(result)

    def do_set(self, arguments):
        """
        Set a variable.
        """
        arguments = arguments.strip()
        if not arguments:
            raise ServerException('Missing variable name.')

        parts = arguments.split(' ', 1)
        if len(parts) == 1:
            # We only have the name of the variable to set.
            # We ask for the value in an interactive way.
            key = parts[0]
            if 'password' in key:
                value = self._getPasswordInput(
                    'New %s (hidden input): ' % (key,))
            else:
                value = self._getRawInput('New value for %s: ' % (key,))

        else:
            key, value = parts

        if key.endswith('password'):
            key += '_update'

        if key not in self._options:
            raise ServerException('Unknown option "%s".' % (key,))

        self.updateFromCLI(key, value)

    def updateFromCLI(self, name, value):
        """
        Update configuration option `name` with text value provided by the
        shell or CLI arguments.

        Most of the time, value is text.
        It can also be boolean, when flags are received from the CLI
        """
        if isinstance(value, bool):
            self._options[name] = value
            return

        # See if we need to covert from text to actual type.
        # Cast text value to same type as current value.
        value = value.strip()
        cast = type(self._options[name])
        if cast is bool:
            value = value.lower()
            if value in ['yes', 'true', '1']:
                value = True
            else:
                value = False
        elif cast in (tuple, set, list):
            value = cast(from_csv_row(value))
        else:
            value = cast(value)

        self._options[name] = value

    def do_unset(self, arguments):
        """
        Set a variable to its default value.
        """
        try:
            self._options[arguments] = self.available_options[arguments]
        except Exception:
            raise ServerException('Failed to unset: %s' % (arguments,))

    @defer.inlineCallbacks
    def do_open(self, arguments):
        """
        Open a remote connection.
        """
        if self._location and self._location.started:
            # Close current location before opening a new connection.
            yield self.do_close('')

        self._location = self._getLocation()
        yield self._location.start()

    # Open aliases.
    do_connect = do_open

    @defer.inlineCallbacks
    def do_close(self, arguments):
        """
        Close the current connection without exiting the shell.
        """
        if self._location and self._location.started:
            yield self._location.stop()
        else:
            defer.returnValue('Nothing to close.')

    # Close aliases.
    do_disconnect = do_close

    @connection_required
    @defer.inlineCallbacks
    def do_ls(self, arguments):
        """
        Get the content of a remote path.
        """
        members = yield self._location.getFolderMemberAttributes(arguments)

        result = []
        for name, attributes in members.items():
            if attributes.is_file:
                kind = 'file'
            elif attributes.is_link:
                kind = 'link'
            else:
                kind = 'directory'

            result.append('%s %s %s %s' % (
                kind, attributes.size, attributes.modified, name))

        defer.returnValue('\n'.join(result))

    # List aliases.
    do_dir = do_ls

    @connection_required
    @defer.inlineCallbacks
    def do_mkdir(self, arguments):
        """
        Create a new directory on the remote location.
        """
        yield self._location.createFolder(arguments)

    # mkdir aliases.
    do_md = do_mkdir

    @connection_required
    @defer.inlineCallbacks
    def do_rmdir(self, arguments):
        """
        Remove a directory on the remote location.
        """
        yield self._location.deleteFolder(arguments)

    # rmdir aliases.
    do_rd = do_rmdir

    @connection_required
    @defer.inlineCallbacks
    def do_rm(self, arguments):
        """
        Remove a file on the remote location.
        """
        yield self._location.deleteFile(arguments)

    # rm aliases.
    do_delete = do_rm

    @connection_required
    @defer.inlineCallbacks
    def do_mv(self, arguments):
        """
        Rename a file or directory on the remote location.
        """
        source, destination = self._parsePaths(arguments)
        if not (source and destination):
            raise ServerException('Fail to get source and destination.')

        yield self._location.rename(source, destination)

    # mv aliases.
    do_move = do_rename = do_mv

    def _parsePaths(self, arguments):
        """
        Parse arguments for paths and return a tuple of (source, destination).
        """
        source = None
        destination = None

        if "'" in arguments:
            parts = arguments.split("'")
            parts = [part.strip() for part in parts if part.strip()]
            arguments_count = len(parts)
            if arguments_count == 2:
                source = parts[0]
                destination = parts[1]
            elif arguments_count == 1:
                source = parts[0]
                destination = None
            else:
                raise ServerException(
                    'Only a maximum of 2 paths are supported.')
        else:
            try:
                source, destination = arguments.split(' ', 1)
            except ValueError:
                source = arguments
                destination = None

        return (source, destination)

    @connection_required
    @defer.inlineCallbacks
    def do_put(self, arguments):
        """
        Upload a local file to the remote server.
        """
        source, destination = self._parsePaths(arguments)
        if not destination:
            # Since we don't have destination we do a simple guess about
            # the path destination.
            # This will fail when path system is different from source and
            # destination.
            destination = os.path.basename(source)

        source_segments = local_filesystem.getSegmentsFromRealPath(source)

        attributes = local_filesystem.getAttributes(source_segments)
        source_stream = local_filesystem.openFileForReading(
            source_segments)
        try:
            destination_stream = yield self._location.getStreamForWriting(
                destination, length=attributes.size)
            try:
                while True:
                    data = source_stream.read(self._CHUNK_SIZE)
                    if not data:
                        break
                    yield destination_stream.write(data)
            finally:
                yield destination_stream.close()
        finally:
            source_stream.close()

    # put aliases.
    do_upload = do_put

    @connection_required
    def do_get(self, arguments):
        """
        Download a remote file on local system.
        """
        source, destination = self._parsePaths(arguments)
        if not destination:
            # Since we don't have destination we do a simple guess about
            # the path destination.
            # This will fail when path system is different from source and
            # destination.
            destination = os.path.basename(source)

        deferred = self._location.getStreamForReading(source)
        deferred.addCallback(self._cbGotSourceStream, destination)
        return deferred

    def _cbGotSourceStream(self, source_stream, destination_path):
        """
        Called when we got the source stream.
        """
        try:
            destination_segments = local_filesystem.getSegmentsFromRealPath(
                destination_path)
            destination_stream = local_filesystem.openFileForWriting(
                destination_segments)
        except Exception:
            # We failed to get the destination stream so close the source
            # stream and just re-raise the error.
            self._bbCloseStream(None, source_stream)
            raise

        deferred = source_stream.read(self._CHUNK_SIZE)
        deferred.addCallback(
            self._cbReceiveMore, source_stream, destination_stream)

        deferred.addBoth(self._bbCloseStream, destination_stream)
        deferred.addBoth(self._bbCloseStream, source_stream)
        return deferred

    def _cbReceiveMore(self, data, source_stream, destination_stream):
        """
        Called when we received read data.
        """
        if not data:
            # We are done with the stream.
            return

        destination_stream.write(data)

        # Trigger another read request.
        deferred = source_stream.read(self._CHUNK_SIZE)
        deferred.addCallback(
            self._cbReceiveMore, source_stream, destination_stream)
        return deferred

    def _bbCloseStream(self, result_or_failure, stream):
        """
        Close a stream.
        """
        deferred = defer.maybeDeferred(stream.close)
        deferred.addBoth(lambda _: result_or_failure)
        return deferred

    # put aliases.
    do_download = do_get

    @connection_required
    @defer.inlineCallbacks
    def do_get_attributes(self, arguments):
        """
        Get the attributes of a path.
        """
        result = yield self._location.getAttributes(arguments)
        timestamp = ChevahTimeStamp(result.modified)
        template = (
            'name: %(name)s\n'
            'path: %(path)s\n'
            'size: %(size)s\n'
            'modified: %(modified)s\n'
            'is_file: %(is_file)s\n'
            'is_folder: %(is_folder)s\n'
            )

        output = template % {
            'name': result.name,
            'path': result.path,
            'size': result.size,
            'modified': timestamp.cwa_14051,
            'is_file': result.is_file,
            'is_folder': result.is_folder,
            }
        defer.returnValue(output)

    # aliases.
    do_gattrs = do_get_attributes

    @connection_required
    @defer.inlineCallbacks
    def do_exists(self, arguments):
        """
        Check if path exists
        """
        result = yield self._location.exists(arguments)
        if result:
            output = 'Yes\n'
        else:
            output = 'No\n'
        defer.returnValue(output)


class EmbeddedClientShellCommands(ClientShellCommands):
    """
    Specialized ClientShellCommands for processing commands in
    non-interactive mode.
    """
    def _ebLogFailure(self, failure):
        """
        Called when a deferred has failed and need to be logged.
        """
        # FIXME:3349:
        # We should emit an event on erros like the super class.
        if failure.check(ServerException):
            if not self._stderr:
                return
            message = force_unicode(failure.value) + '\n'
            if six.PY2:
                message = message.encode('utf-8')
            self._stderr.write(message)
            return
        return super(EmbeddedClientShellCommands, self)._ebLogFailure(failure)

    @defer.inlineCallbacks
    def start(self, location):
        """
        Called when command shell should start processing non-interactive
        embedded batch mode.

        `location` is used to connect the shell to the location.
        It should be already started.

        Return a deferred which is fired when command execution is done.
        """
        # Use the same process as the one used by the location.
        self._process = location.root
        self._location = location
        self._stderr = StringIOWithLimit()
        self._stdout = None

        try:
            while True:
                yield self._loopCallback(mode_callback=self._processBatch)
        except BatchStopException:
            # Normal stop was requested.
            pass

    def getErrors(self):
        """
        Returns any error messages.
        """
        return force_unicode(self._stderr.getvalue().strip())

    def _stop(self):
        """
        Called when we should stop processing command.
        """
        raise BatchStopException()
