# Copyright (c) 2021 Adi Roiban.
# See LICENSE for details.
"""
High level code for SMB location and configuration.
"""
from twisted.internet import defer
from twisted.internet.address import IPv4Address
from zope.interface import implementer

from chevah.server.commons.constant import (
    CONFIGURATION_SECTION_LOCATIONS,
    )
from chevah.server.commons.credential import PasswordCredentials
from chevah.server.commons.interface import (
    ISMBLocationConfiguration,
    )
from chevah.server.configuration.option import (
    AddressPortOptionsMixin,
    UsernamePasswordPublicOptionsMixin,
    )
from chevah.server.configuration.model import WritableBoolean

from chevah.server.location.commons import LocationBase
from chevah.server.location.configuration import _LocationConfigurationBase
from chevah.server.smb.client import SMBClient


@implementer(ISMBLocationConfiguration)
class SMBLocationConfiguration(
        _LocationConfigurationBase,
        AddressPortOptionsMixin,
        UsernamePasswordPublicOptionsMixin,
        ):
    """
    Configuration for a SMB location.
    """
    SECTION_PREFIX = CONFIGURATION_SECTION_LOCATIONS
    PORT_DEFAULT = 445

    require_encryption = WritableBoolean('require_encryption', default=True)


class SMBLocation(LocationBase):
    """
    Location for a SMB server.
    """

    PROPERTIES_REQUIRING_RESTART = (
        'address',
        'port',
        'password_update',
        'username',
        'idle_connection_timeout',
        'idle_connection_keepalive_interval',
        'connection_retry_count',
        'connection_retry_interval',
        'require_encryption',

        )

    _client = None

    def _onConnect(self):
        """
        Called when client should connect.

        It should return a deferred which is fired with the low level client.
        """
        credentials = PasswordCredentials(
            username=self._configuration.username,
            password=self._configuration.password,
            peer=IPv4Address(
                type='TCP',
                host=self._configuration.address,
                port=self._configuration.port,
                ),
            creator=self,
            )
        client = SMBClient(
            credentials,
            require_encryption=self._configuration.require_encryption,
            )
        return client.authenticate()

    def _onDisconnect(self):
        """
        Called when connection should disconnect.
        """
        if self._client:
            deferred = self._client.close()
        else:
            deferred = defer.succeed(None)
        self._client = None
        return deferred
