# Copyright (c) 2015 Adi Roiban.
# See LICENSE for details.
"""
Code to support watching any ILocations and notifying the changes.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from collections import deque
import sys

from twisted.internet import defer, reactor
from twisted.internet.base import DelayedCall
from twisted.internet.task import deferLater, LoopingCall

from chevah.server import force_unicode
from chevah.server.commons.constant import EVENT_TYPE
from chevah.server.commons.exception import ServerException
from chevah.server.commons.interface import IWatchHandler
from chevah.server.commons.match import get_glob_rule, get_matched_groups
from zope.interface.declarations import implementer


class FilesystemEvent(object):
    """
    An event emitted by the location watcher.
    """
    def __init__(
        self, type, source_path, is_folder=False, node_id=None,
        size=0, modified=0,
            ):
        self.type = type
        self.source_path = source_path
        self.destination_path = None
        self.is_folder = is_folder
        self.node_id = node_id
        self.current_path = source_path
        self.size = size
        self.modified = modified

        # Use as fix to still handling MODIFY + DELETE + CREATE for the
        # same node_id with using only 2 events.
        self._previously_modified = False

        # Use to merge CREATE + MOVE
        self._previously_created = False

        if self.type == EVENT_TYPE.MOVED:
            raise RuntimeError(
                'Direct creation of move events is not yet supported.')

    def __str__(self):
        text = '<FilesystemEvent %d:%s> %s (folder %s) %s %s %s' % (
            id(self),
            self.node_id,
            self.type,
            self.is_folder,
            self.source_path,
            self.destination_path,
            self.current_path,
            )
        return text

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        if (
            self.type != other.type or
            self.is_folder != other.is_folder or
            self.source_path != other.source_path or
            self.destination_path != other.destination_path or
            self.current_path != other.current_path or
            self.node_id != other.node_id
                ):
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)


class StableFilesystemEventHandler(object):
    """
    An event handler which only react on the last event emitted for a path.

    For example if a file is created and then is continue to change for 10
    seconds and closed, only a single event is emitted after 10 + TIMEOUT.

    It will call `on_change` for each stable event. `on_change` callable
    should never fail as nobody is handling its result.

    `timeout` is the grace period in which changes for same path are merged.
    """
    #: Scheduler used to manage delays.
    _scheduler = reactor

    def __init__(self, on_change, timeout=1):
        # Events waiting to be stabilized by path
        self._queue = {}
        self._timeout = timeout
        self._onChange = on_change

        # Initialized a delayed call which does nothing and which is not
        # registered to any reactor.
        # In this way we always have a last call and reduce the number of
        # if calls in the _callLaterMonotonic.
        # We also reduce the usage of None :)
        self._last_monotonic_call = DelayedCall(
            time=0,
            func=lambda: None,
            args=(),
            kw={},
            cancel=lambda _: None,
            reset=lambda _: None,
            seconds=0,
            )

    def handle(self, event):
        """
        Dispatch event to specialized methods.

        This is called from main thread.
        """
        self._processMergeQueue(event)

    def clear(self):
        """
        Clear all pending calls.
        """
        while True:
            try:
                (_, (call, _)) = self._queue.popitem()
            except KeyError:
                return
            call.cancel()

    def _processMergeQueue(self, event):
        """
        Merge duplicate / redundant events based on path or node_id.

        Modify events for folders are ignored.

        Merge results for same path:

        * existing + modify = existing
        * existing + delete = ignore
        * existing + create = ??? (this should not be possible)

        * modify + modify = modify
        * modify + delete = delete (without delay)
        * modify + create = ??? - create (not sure how is this possible)
        * modify + move = move + modify (modify is later)

        * create + create = create (if this ever happens)
        * create + modify = create
        * create + delete = discard
        * create + move = create at the moved path

        * delete + create = modify !!! (a file was completely replaced)
        * DELETE.folder + CREATED.file  = 2 separate event.

        * delete + delete = ??? - delete (not sure how is this possible)
        * delete + modify = ??? - delete (not sure how is this possible)
        * delete + move = ??? - not sure how is this possible

        * moved + created = moved + created (as 2 events)
        * moved + modified = moved + modified (Windows ReadDirectoryChanges)
        * moved + deleted = deleted for the source path

        Merge events for the same node_id to create move events.

        * delete + create = moved to created
        * created + delete = moved to created

        There is a great chance for a delete event to be last, but when a
        file is completely replaced, we will have a delete, shortly followed
        by a create. For this case, we might use a separate, shorter timeout.
        """
        if event.type == EVENT_TYPE.MODIFIED and event.is_folder:
            # We ignore modified event for folders as they always come in pair
            # with create / delete.
            return

        queued_call, queued_event = self._queue.get(
            event.source_path, (None, None))
        if queued_call and queued_call.active():
            # We already have an event for that path.
            # And the event was not yet trigerred.
            self._pathBasedUpdate(queued_call, queued_event, event)
            return

        # We don't care the path, as we will do node-id matching.
        for _, value in self._queue.items():
            moved_call, moved_event = value
            if not moved_event.node_id:
                continue
            if moved_event.node_id != event.node_id:
                continue
            if not moved_call.active():
                continue

            # We already have an event for that node_id and the event was
            # not previously matched based on path.

            # This can happen especially for deep recursive watchers in which
            # the move is done while the watcher is slowly taking snapshots
            # of each directory... and the file is moved from a directory
            # which was already scanned, and moved into a directory which
            # was not yet scanned.
            # For this, the watcher will just produce a created event... and
            # on the second run will produce the deleted event.

            self._nodeBasedUpdate(moved_call, moved_event, event)
            return

        self._queueEvent(event, cancel=None)

    def _pathBasedUpdate(self, queued_call, queued_event, event):
        """
        Consider handling a new event based on the the existing queue.
        """
        if queued_event.type == event.type:
            # For similar events, just reset the timeout.
            queued_call.reset(self._timeout)
            return

        if queued_event.type == EVENT_TYPE.EXISTING:
            if event.type == EVENT_TYPE.DELETED:
                # EXISTING + DELETE = discard
                # If a create event if followed, then the path will be
                # considered as newly created.
                self._cancelQueue(event.source_path)
                return

            # For any other combination, discard the new event and
            # reset the initial one.
            queued_call.reset(self._timeout)
            return

        if queued_event.type == EVENT_TYPE.MODIFIED:
            if event.type == EVENT_TYPE.DELETED:
                # MODIFIED + DELETE = DELETE (discard modify).
                # MODIFIED + MOVE (CREATE-new-path + DELETE)
                # MODIFIED + MOVE (DELETE + CREATE)

                # We don't emit the event right away as it might be
                # followed or have an associated CREATE
                # and will be turned into MODIFY.
                self._cancelQueue(event.source_path)

                # Mark the delete event as modified, so that if it turns
                # out that this is a move, we will emit the modify event.
                event._previously_modified = True
                # We call handle again to merge the queue in the case
                # we previously had move + modified.
                # As for now only the previous modified event was merged.
                self.handle(event)
                return

            # MODIFY + CREATE = should not happen
            # For any other combination, discard the new event and
            # reset the initial one.
            queued_call.reset(self._timeout)
            return

        if queued_event.type == EVENT_TYPE.CREATED:
            if event.type == EVENT_TYPE.DELETED:
                # CREATE + DELETE = discard
                self._cancelQueue(event.source_path)
                return

            # CREATE + MODIFY = CREATE + reset
            # CREATE + CREATE
            # CREATE + MOVE = still create firsts
            # For any other combination, discard the new event and
            # reset the initial one.
            queued_call.reset(self._timeout)
            return

        if queued_event.type == EVENT_TYPE.DELETED:
            if event.type == EVENT_TYPE.CREATED:
                # DELETED + CREATED = MODIFIED (as new event, discard old)
                event.type = EVENT_TYPE.MODIFIED
                self._queueEvent(event, cancel=event.source_path)
                return

            # Ignore the event without updating the timeout.
            # We should not reach here after delete we can only see create.
            return

        if queued_event.type == EVENT_TYPE.MOVED:
            if event.type == EVENT_TYPE.CREATED:
                # Path was renamed and then a new path is created at the
                # original name.
                # Both event should be kept and not merged as we now have
                # 2 separate files.
                self._queueEvent(event, cancel=None)
                return

            if event.type == EVENT_TYPE.DELETED:
                event.source_path = queued_event.source_path
                event.current_path = queued_event.source_path
                self._queueEvent(event, cancel=queued_event.current_path)
                return

            if event.type == EVENT_TYPE.MODIFIED:
                queued_call._previously_modified = True
                queued_call.reset(self._timeout)
                return

            # I can't think of any case in which we have a moved event
            # for a source path, and then another event for the same path.
            return

        # We should never reach this.
        queued_event.cancel()
        raise RuntimeError(
            'Unhandled event %s based on %s in %s' % (
                event, queued_event, self._queue))

    def _nodeBasedUpdate(self, queued_call, queued_event, event):
        """
        Convert the events based on the node id.

        `event` is the latest event which was created for the same node id.
        `queued_event` in an event in the queue with the same node ID, but
        with a different path.

        Node IDs can be reused. We can end up with any condition here.

        Before calling this make sure events have the same node_id.
        """
        if (
            queued_event.type == EVENT_TYPE.DELETED and
            event.type == EVENT_TYPE.CREATED
                ):
            # Delete was observed first.

            # Might have MOVED while MODIFIED.
            # We have MODIFIED + DELETE + CREATE -> MOVE + MODIFIED.
            event._previously_modified = queued_event._previously_modified

            # Just a simple DELETE + CREATE -> MOVE
            event.type = EVENT_TYPE.MOVED
            event.source_path = queued_event.source_path
            event.destination_path = event.current_path
            self._queueEvent(event, cancel=queued_event.source_path)
            return

        if queued_event.type == EVENT_TYPE.CREATED:
            if event.type == EVENT_TYPE.CREATED:
                # The initial path was created, but we no longer have it in
                # latest snapshots.
                # Might be a CREATE + CREATE as first part of move
                # Ignore it and record the latest event.
                # CREATE + CREATE with different path but same ID.
                event._previously_modified = queued_event._previously_modified
                event._previously_created = True
                self._queueEvent(event, cancel=queued_event.source_path)
                return

            if event.type == EVENT_TYPE.DELETED:
                if queued_event._previously_created:
                    # CREATE + MOVE = CREATE as updated path.
                    queued_call.reset(self._timeout)
                    return

                if not queued_event._previously_modified:
                    # Just a simple CREATE + DELETE -> MOVE
                    queued_event.type = EVENT_TYPE.MOVED
                    queued_event.source_path = event.source_path
                    queued_event.destination_path = queued_event.current_path
                    queued_call.reset(self._timeout)
                    return

                # We have MODIFIED + CREATE + DELETE -> MOVE + MODIFIED.
                # Last DELETE -> MOVED.
                event.type = EVENT_TYPE.MOVED
                event.destination_path = queued_event.current_path
                event.current_path = queued_event.current_path
                event._previously_modified = True
                self._queueEvent(event, cancel=queued_event.source_path)
                return

            # Should not reach this.
            # CREATE + MODIFY -> Created
            # CREATE + MOVED -> Created
            queued_call.reset(self._timeout)
            return

        if queued_event.type == EVENT_TYPE.MOVED:
            if event.type == EVENT_TYPE.MODIFIED:
                # MOVE + MODIFIED = MOVE-modify-marker
                queued_event._previously_modified = True
                queued_call.reset(self._timeout)
                return

            if event.type == EVENT_TYPE.CREATED:
                # MOVE + CREATE = MOVE with updated destination
                queued_event.current_path = event.current_path
                queued_event.destination_path = event.current_path
                queued_event._previously_created = True
                queued_call.reset(self._timeout)
                return

            if event.type == EVENT_TYPE.DELETED:
                if queued_event._previously_created:
                    # This is the second part of a move after another move.
                    queued_event._previously_created = False
                    queued_call.reset(self._timeout)
                    return

                # Rename then delete.
                event.current_path = queued_event.source_path
                event.source_path = queued_event.source_path
                event._previously_modified = queued_event._previously_modified
                self._queueEvent(event, cancel=queued_event.source_path)
                return

            # Should not reach this.
            # MOVE + CREATE = MOVE
            # MOVE + MOVE = MOVE
            queued_call.reset(self._timeout)
            return

        if queued_event.type == EVENT_TYPE.MODIFIED:
            if event.type == EVENT_TYPE.CREATED:
                # MODIFY + CREATE
                # This create might the half of a moved.
                event._previously_modified = True
                self._queueEvent(event, cancel=queued_event.source_path)
                return

            if event.type == EVENT_TYPE.DELETED:
                # We have a modify + MOVE twice.
                # Cancel the modify and set the delete for the modify path.
                event._previously_modified = True
                event.source_path = queued_event.source_path
                event.current_path = queued_event.source_path
                self._queueEvent(event, cancel=queued_event.source_path)
                return

            # Should not reach this.
            # MODIFY + MOVE
            # MODIFY + MODIFY
            queued_call.reset(self._timeout)
            return

        # We should not reach this place.
        # The initial node has an operation but we no longer observe the
        # original path..
        # Mark it as delete.
        queued_event.current_path = queued_event.source_path
        queued_event.destination_path = None
        queued_event.type = EVENT_TYPE.DELETED
        self._queueEvent(event, cancel=None)

    def _queueEvent(self, event, cancel):
        """
        Schedule a fresh event.
        """
        def execute_event(path, callback, event):
            """
            Execute and remove call from queue.
            """
            self._queue.pop(path, None)
            callback(event)
            if event.type == EVENT_TYPE.MOVED and event._previously_modified:
                modify_event = FilesystemEvent(
                    EVENT_TYPE.MODIFIED,
                    source_path=event.destination_path,
                    is_folder=event.is_folder,
                    size=event.size,
                    modified=event.modified,
                    node_id=event.node_id,
                    )
                callback(modify_event)

        call = self._callLaterMonotonic(
            self._timeout,
            execute_event,
            event.source_path,
            self._onChange,
            event,
            )

        if cancel:
            self._cancelQueue(cancel)

        self._queue[event.source_path] = (call, event)

    def _cancelQueue(self, path):
        """
        Cancel queued event for path, together with the move tracking.
        """
        (call, event) = self._queue.pop(path)
        call.cancel()

    def _callLaterMonotonic(self, *args, **kwargs):
        """
        Make sure that we don't have delayed call which is scheduled at
        a time earlier or equal to the last delayed call.
        """
        call = self._scheduler.callLater(*args, **kwargs)
        last_call = self._last_monotonic_call

        if call.getTime() <= last_call.getTime():
            # Current time is not scheduled after the last call.
            # Reset it with base time and delay of the last call.
            call.time = last_call.time
            # We should increase with a value smaller than the resolution of
            # the best clock.
            # I have no idea why 0.0000001 will not work on Windows.
            call.delayed_time = last_call.delayed_time + 0.000001

        self._last_monotonic_call = call
        return call


class FilteredWatchAdapter(object):
    """
    A watch on a location with filtered events.

    Folder changes are unfiltered.
    """

    def __init__(
        self, source, configuration, on_change, on_error, event_emitter,
        on_snapshot=None, scheduler=None,
            ):
        """
        `source` location.
        `configuration` is an ILocationWatchConfiguration.
        `on_change` is called when an event is raised. This callback should
            never raise an error.
        `on_error` is called with an exception when there is an error
            processing the events.
        `on_snapshot` is called when a new snapshot is made with a dict indexed
            by path and containing member attributes.
        """
        stable_interval = configuration.stable_interval
        if stable_interval <= configuration.changes_poll_interval:
            stable_interval = configuration.changes_poll_interval + 1
            event_emitter('60083', data={'stable_interval': stable_interval})

        self._event_handler = StableFilesystemEventHandler(
            timeout=stable_interval,
            on_change=self._onUnfilteredChange,
            )
        if scheduler:
            self._event_handler._scheduler = scheduler

        self._onFilteredChange = on_change
        self._onError = on_error
        self._onSnapshot = on_snapshot
        self._configuration = configuration
        self._source = source
        self._watch_handler = None

    def start(self):
        """
        Start the watcher.
        """
        # Validate that filter expression is valid before launching the
        # whole watch process.
        get_matched_groups(
            value='',
            rule=self._configuration.source_filter,
            usage='source_filter',
            )

        on_snapshot = self._onSnapshot
        if on_snapshot:
            on_snapshot = self._onUnfilteredSnapshot

        self._watch_handler = self._source.watch(
            configuration=self._configuration,
            handler=self._event_handler,
            on_error=self._onError,
            on_snapshot=on_snapshot,
            )

    def stop(self):
        """
        Stop the watcher and don't emit any other events.
        """
        if not self._watch_handler:
            return
        self._event_handler.clear()
        self._source.unwatch(self._watch_handler)

    @property
    def running(self):
        """
        True if the low-level watcher is still active.
        """
        if not self._watch_handler:
            return False
        return self._watch_handler.running

    def _onUnfilteredSnapshot(self, result):
        """
        Called when a new snapshot was made.
        """
        new_snapshot, visit_order = result
        filtered_changes = {}

        # Filter and convert the changes to a format more useful for the
        # snapshot usage.
        # The initial data format is optimize for detecting changes between
        # snapshots.
        for path in visit_order:
            filtered_members = {}
            for member in new_snapshot[path]:
                for name, attributes in member.items():
                    if self._isPathIgnored(attributes.path):
                        continue
                    filtered_members[name] = attributes
            filtered_changes[path] = filtered_members

        # Notify the changes.
        return self._onSnapshot(filtered_changes)

    def _onUnfilteredChange(self, event):
        """
        Called when a source event is emitted.

        This should never raise an error as nobody is waiting or handling
        the result of this call.
        """
        try:
            if event.is_folder:
                # For now, only file changes are filtered.
                # Result is returned to help with testing.
                return self._onFilteredChange(event)

            if self._isPathIgnored(event.current_path):
                return

            # Result is returned to help with testing.
            return self._onFilteredChange(event)

        except Exception as error:
            # This call should never fail, so we forward all calls to the
            # error handler.
            self._onError(error)

    def _isPathIgnored(self, path):
        """
        Return `True` if `path` is not matched by the filtering rule.
        """
        rule = self._configuration.source_filter

        # FIXME:6447:
        # Be default only match based on file name.
        # With version 5 release we should find a way to enable
        # full path matching for regex.
        # We are not doing a full path matching now, as the heuristic used
        # to keep backward compatibility is more complicated.
        # For v5 we should match on full path be default.
        # And if we see that the expression is only for a single path,
        # automatically prepend the match anything
        # rule like "*" (glob) or ".*" (regex).
        _, value_to_check = self._source.splitPath(path)

        if (get_glob_rule(rule) is not None) and ('/' in rule or '\\' in rule):
            # We have a globbing rule which checks for path separators
            # so we guess that the user wants to do a full path match.
            # FIXME:6447:
            value_to_check = path

        return not get_matched_groups(
            value=value_to_check,
            rule=rule,
            usage='source_filter',
            )


class ChevahLoopingCall(LoopingCall):
    """
    Patched LoopingCall to keep a record of the deferred on which the loop
    is waiting.

    After loop is started you should attach an errback to the returned
    deferred so that we don't have unhandled errors in the deferred.
    """
    # Deferred on which the loop is waiting to reschedule.
    _ongoingDeferred = None

    def __call__(self):
        """
        Called when the wrapped method should be called.
        """
        def cb(result):
            if self.running:
                self._scheduleFrom(self.clock.seconds())
            else:
                d, self._deferred = self._deferred, None
                d.callback(self)

        def eb(failure):
            self.running = False
            d, self._deferred = self._deferred, None
            d.errback(failure)

        def bb_reset_deferred(result_or_failure):
            """
            Called when the ongoing deferred was fired.
            """
            self._ongoingDeferred = None
            return result_or_failure

        self.call = None
        deferred = defer.maybeDeferred(self.f, *self.a, **self.kw)
        deferred.addCallback(cb)
        deferred.addErrback(eb)
        deferred.addBoth(bb_reset_deferred)
        self._ongoingDeferred = deferred


class _LoopStop(Exception):
    """
    The snapshot loop was stopped while the looped code was running.
    """


class SnapshotDriver:
    """
    Wraps a location that is used to take snapshots for directories.
    """

    def __init__(self, location, configuration):
        self._location = location
        self.emitEvent = location.emitEvent
        self.path = configuration.source_path
        self.recursive = configuration.recursive
        self.interval = configuration.changes_poll_interval
        self.retry_count = configuration.retry_count

    @defer.inlineCallbacks
    def getSnapshot(self, path):
        """
        Method used to get the members of a directory.
        """
        yield self._location.heartbeat()
        result = yield self._location.getFolderMemberAttributes(path)
        defer.returnValue(result)


@implementer(IWatchHandler)
class SnapshotLoop(object):
    """
    A generic loop which polls for the state of a path using the `driver` and
    trigger the raw observed changes using the event `handler`.

    It will not detect moved events event if the location supports node_id.
    It will just emit primitive create/delete/modified events.
    When supported events will include the node_id.

    `driver` is an SnapshotDriver
    `handler` is an EventHandler
    """
    def __init__(
        self, driver, handler, on_error, scheduler=reactor, on_snapshot=None,
            ):
        if not driver.path:
            raise ServerException('No path configured to monitor.')
        self._snapshot = None
        self._driver = driver
        self._handler = handler
        self._error_count = 0
        self._loop = ChevahLoopingCall(self._onSnapshot)
        self._loop.clock = scheduler
        self._onError = on_error
        self._onSnapshotExternal = on_snapshot

    def start(self):
        """
        Start watching the path.
        """
        deferred = self._loop.start(self._driver.interval)
        deferred.addErrback(self._ebLoopingCall)

    def stop(self):
        """
        Stop watching the path.
        """
        if self._loop.running:
            self._loop.stop()
        return self._driver.path

    @property
    def running(self):
        """
        True if watch snapshots are still scheduled.
        """
        return self._loop.running

    def _ebLoopingCall(self, failure):
        """
        Called when the looping call failed.
        """
        if failure.check(StopIteration):
            # Loop is done.
            return failure

        # Looping might call due to hitting the recursion limit so
        # we temporary increase the recursion limit to allow catching
        # the error.
        initial = sys.getrecursionlimit()
        try:
            sys.setrecursionlimit(initial + 100)
            self._onError(failure.value)
        finally:
            sys.setrecursionlimit(initial)

    def _cbNotifyNewSnapshot(self, result):
        """
        Called when a new snapshot is done to inform the external hook.
        """
        if not self._onSnapshotExternal:
            return result

        # External snapshot might return a deferred and we ignore its result.
        deferred = defer.maybeDeferred(self._onSnapshotExternal, result)
        deferred.addCallback(lambda _: result)
        return deferred

    @defer.inlineCallbacks
    def _getNewSnapshot(self):
        """
        Traverse the path and return the new snapshot together with the
        order in which the path was traversed.

        Return tuple (new_snapshot, visit_order),
        Raise _LoopStop if loop was stopped while getting the snapshot.
        """
        path = self._driver.path
        # Snapshot with current state.
        new_snapshot = {}
        # Order in which recursive directories were visited.
        visit_order = [path]
        # Get root / initial changes.
        members = yield self._getMembers(path)
        # Record the new path content.
        new_snapshot[path] = members

        if self._driver.recursive:
            # Only visit directories.
            visit_queue = deque(list(members[1].values()))

            while visit_queue:
                directory_attributes = visit_queue.popleft()
                path = directory_attributes.path

                members = yield self._getMembers(path)
                # Record the new path content.
                new_snapshot[path] = members

                visit_queue.extendleft(list(members[1].values()))
                visit_order.append(path)

        defer.returnValue((new_snapshot, visit_order))

    def _emitIgnoredException(self, event_emitter, details, path):
        """
        Helper to raise the same event ID from all watch errors which are
        ignored and for which the location is continued to be watched.
        """
        event_emitter(
            event_id='60005',
            data={
                'details': details,
                'path': path,
                },
            )

    def _onSnapshot(self):
        """
        Called when a new snapshot should be made.
        """
        def cb_get_all_changes(result):
            """
            Called when we got the results from a snapshot.

            Return a sequence with all changes in `path`.
            """
            new_snapshot, visit_order = result
            changes = []

            if self._snapshot is None:
                # This is the first snapshot. Create events for existing
                # members in the root folder.
                # We compare files with files and directories with directories.
                # Files first.
                path = visit_order[0]
                changes.extend(self._existingSameType(
                    path, new_snapshot, is_folder=False))
                changes.extend(self._existingSameType(
                    path, new_snapshot, is_folder=True))

                # Initiate an empty snapshot.
                self._snapshot = {}

            for path in visit_order:
                changes.extend(self._getChanges(path, new_snapshot))

            self._snapshot = new_snapshot
            return changes

        def cb_got_changes(result):
            # FIXME:2973:
            # Look at all changes and match move events.
            for event in result:
                self._handler.handle(event)

        def eb_loop_stop(failure):
            """
            Do nothing if loop was stopped while we were running.
            """
            failure.trap(_LoopStop)

        def eb_loop_fail(failure):
            """
            Forward all errors to the error handler.
            """
            self._onError(failure.value)

        deferred = self._getNewSnapshot()
        deferred.addCallback(self._cbNotifyNewSnapshot)
        deferred.addCallback(cb_get_all_changes)
        deferred.addCallback(cb_got_changes)
        deferred.addErrback(eb_loop_stop)
        deferred.addErrback(eb_loop_fail)
        return deferred

    @defer.inlineCallbacks
    def _getMembers(self, path):
        """
        Get members for path.

        Return a deferred which fires with a tuple of
        (file_members_dict, folder_members_dict)
        """
        # To make sure we don't call this in a continuous loop,
        # we put it at the back of the queue.
        yield deferLater(self._loop.clock, 0)

        if not self._loop.running:
            # Loop was stopped while we were rescheduling the watch.
            raise _LoopStop()

        try:
            result = yield self._driver.getSnapshot(path)
        except Exception as error:
            if self._error_count >= self._driver.retry_count:
                # Too many consecutive errors.
                # No longer ignore the error.
                raise error

            self._error_count += 1

            # We ignore any error and consider that the directory is
            # empty.
            self._emitIgnoredException(
                event_emitter=self._driver.emitEvent,
                details=force_unicode(error),
                path=path,
                )
            defer.returnValue(({}, {}))

        self._error_count = 0
        if not self._loop.running:
            # Loop was already stopped while we were getting
            # the directory's content.
            # Abort getting the snapshot.
            raise _LoopStop()

        files = {}
        directories = {}
        for name, attributes in result.items():
            if attributes.is_folder:
                directories[name] = attributes
            else:
                files[name] = attributes

        defer.returnValue((files, directories))

    def _getChanges(self, path, new_snapshot):
        """
        Return a list with changes for path, based on new_snapshot.
        """
        snapshot = self._snapshot.get(path, None)

        if snapshot is None:
            # This is the first time we see this path so there are no
            # changes to observe.
            return []

        # We compare files with files and directories with directories.
        # If a file was converted as a folder 2 events are generated.
        changes = []
        changes.extend(
            self._diffSameType(path, new_snapshot, is_folder=False))
        changes.extend(
            self._diffSameType(path, new_snapshot, is_folder=True))

        return changes

    def _diffSameType(self, path, new_snapshot, is_folder):
        """
        Called to return the list of changes between two snapshots.

        Return a list of `FilesystemEvent`
        """
        results = []

        previous_snapshot = self._snapshot.get(path)
        current_snapshot = new_snapshot.get(path)
        new_member_event_type = EVENT_TYPE.CREATED

        if is_folder:
            previous = previous_snapshot[1]
            current = current_snapshot[1]
        else:
            previous = previous_snapshot[0]
            current = current_snapshot[0]

        # Check for removed or modified
        for name, previous_attributes in sorted(previous.items()):
            if name not in current:
                # If name is not part of current snapshot, then it was
                # removed.
                if is_folder:
                    results.extend(self.__deleteDirectory(previous_attributes))
                else:
                    results.append(self.__deleteFile(previous_attributes))
                continue

            # Name still exists
            current_attributes = current[name]

            if previous_attributes.node_id != current_attributes.node_id:
                # It was replaced. So we created the 2 events required to
                # track the exact changes.
                results.append(FilesystemEvent(
                    type=EVENT_TYPE.DELETED,
                    source_path=previous_attributes.path,
                    is_folder=is_folder,
                    node_id=previous_attributes.node_id,
                    size=previous_attributes.size,
                    modified=previous_attributes.modified,
                    ))
                results.append(FilesystemEvent(
                    type=EVENT_TYPE.CREATED,
                    source_path=current_attributes.path,
                    is_folder=is_folder,
                    node_id=current_attributes.node_id,
                    size=current_attributes.size,
                    modified=current_attributes.modified,
                    ))
                continue

            # Check to see if it was not modified.
            if (
                previous_attributes.size != current_attributes.size or
                previous_attributes.modified != current_attributes.modified
                    ):
                results.append(FilesystemEvent(
                    type=EVENT_TYPE.MODIFIED,
                    source_path=previous_attributes.path,
                    is_folder=is_folder,
                    node_id=previous_attributes.node_id,
                    size=previous_attributes.size,
                    modified=previous_attributes.modified,
                    ))
                continue

        # Check for created members.
        for name, attributes in current.items():
            if name in previous:
                # This is not a new member.
                continue

            if is_folder:
                results.extend(self.__createDirectory(
                    attributes, new_snapshot, new_member_event_type))
            else:
                results.append(self.__createFile(
                    attributes, new_member_event_type))

        return results

    def _existingSameType(self, path, new_snapshot, is_folder):
        """
        Called to return the list of existing events for new_snapshot.

        Return a list of `FilesystemEvent`
        """
        results = []
        current_snapshot = new_snapshot.get(path)
        new_member_event_type = EVENT_TYPE.EXISTING
        if is_folder:
            current = current_snapshot[1]
        else:
            current = current_snapshot[0]

        # Check for members using a well defined order to help with the
        # testing.
        for _, attributes in sorted(current.items()):
            if is_folder:
                results.extend(self.__createDirectory(
                    attributes, new_snapshot, new_member_event_type))
            else:
                results.append(self.__createFile(
                    attributes, new_member_event_type))

        return results

    def __deleteFile(self, attributes):
        """
        Return the event for the file which was removed.
        """
        return FilesystemEvent(
            type=EVENT_TYPE.DELETED,
            source_path=attributes.path,
            is_folder=False,
            node_id=attributes.node_id,
            size=attributes.size,
            modified=attributes.modified,
            )

    def __deleteDirectory(self, attributes):
        """
        Generates a list of events for a directory which was removed.
        """
        if self._driver.recursive:
            # Events for directory's content are first.
            snapshot = self._snapshot[attributes.path]
            for directory_name, directories_attributes in snapshot[1].items():
                # FIXME:3002:
                # Rewrite this using non-recursive call so that we can
                # support deeply nested directories.
                for event in self.__deleteDirectory(
                        directories_attributes,
                        ):
                    yield event
            for directory_name, directories_attributes in snapshot[0].items():
                yield self.__deleteFile(
                    directories_attributes,
                    )

        # Followed by the event for the directory itself.
        yield FilesystemEvent(
            type=EVENT_TYPE.DELETED,
            source_path=attributes.path,
            is_folder=True,
            node_id=attributes.node_id,
            size=attributes.size,
            modified=attributes.modified,
            )

    def __createFile(self, attributes, event_type):
        """
        Return the event for a file which was created.
        """
        return FilesystemEvent(
            type=event_type,
            source_path=attributes.path,
            is_folder=False,
            node_id=attributes.node_id,
            size=attributes.size,
            modified=attributes.modified,
            )

    def __createDirectory(self, attributes, new_snapshot, event_type):
        """
        Generates a list of events for the directory which was created.
        """
        # Directory creating event is first.
        yield FilesystemEvent(
            type=event_type,
            source_path=attributes.path,
            is_folder=True,
            node_id=attributes.node_id,
            size=attributes.size,
            modified=attributes.modified,
            )

        if not self._driver.recursive:
            return

        # Followed by its content, but only when we are in recursive.
        snapshot = new_snapshot[attributes.path]
        for directory_name, directories_attributes in snapshot[1].items():
            # FIXME:3002:
            # Rewrite this using non-recursive call so that we can
            # support deeply nested directories.
            for event in self.__createDirectory(
                    directories_attributes,
                    new_snapshot,
                    event_type,
                    ):
                yield event
        for directory_name, directories_attributes in snapshot[0].items():
            yield self.__createFile(
                directories_attributes,
                event_type,
                )
