# Copyright (c) 2017 Adi Roiban.
# See LICENSE for details.
"""
Locations for a remote WebDAV server.

This is just the location adapter. The actual client-side protocol is found
in the chevah.server.http package.
"""
from zope.interface import implementer

from chevah.server.commons.constant import (
    CONFIGURATION_SECTION_LOCATIONS,
    TYPE_NAME,
    )
from chevah.server.commons.interface import (
    IWebDAVSLocationConfiguration,
    )

from chevah.server.commons.ssl import (
    SSLClientMixin,
    SSLConfigurationSectionMixin,
    )
from chevah.server.configuration.model import (
    WritableBoolean,
    WritableString,
    )
from chevah.server.configuration.option import (
    UsernamePasswordPublicOptionsMixin,
    URLOptions,
    )

from chevah.server.http.webdav import WebDAVClient
from chevah.server.location.commons import SSLLocationBase
from chevah.server.location.configuration import _LocationConfigurationBase


@implementer(IWebDAVSLocationConfiguration)
class WebDAVSLocationConfiguration(
        _LocationConfigurationBase,
        UsernamePasswordPublicOptionsMixin,
        SSLConfigurationSectionMixin,
        URLOptions,
        ):
    """
    Configuration for a WebDAVS location.
    """
    SECTION_PREFIX = CONFIGURATION_SECTION_LOCATIONS

    proxy = WritableString('proxy')
    authentication_method = WritableString(
        'authentication_method', default=TYPE_NAME.SHAREPOINT_ONLINE)
    allow_insecure_http = WritableBoolean(
        'allow_insecure_http', default=False)

    # FIXME:6521:
    # To be removed in v6.
    @property
    def url(self):
        """
        The base URL used for HTTP webdav requests.
        """
        value = self._proxy.getString(self._section_name, 'url')
        if value:
            return value

        if not self._proxy.definesOption(self._section_name, 'address'):
            # We don't have legacy address/port configuration.
            return value

        # Fallback to legacy address/port configuration.
        # Address will be 127.0.0.1 even when not configured.
        address = self._proxy.getString(self._section_name, 'address')
        port = self._proxy.getString(self._section_name, 'port')

        if not address:
            return ''

        scheme = 'https'
        if self.allow_insecure_http:
            scheme = 'http'

        if not port or port == '1':
            # 1 is the legacy marker for a default port.
            return '%s://%s' % (scheme, address)

        return '%s://%s:%s' % (scheme, address, port)

    @url.setter
    def url(self, value):
        # Remove the v4 address/port configuration cleanup.
        self._proxy.removeOption(self._section_name, 'address')
        self._proxy.removeOption(self._section_name, 'port')
        return self._proxy.setString(self._section_name, 'url', value)


class WebDAVSLocation(SSLLocationBase, SSLClientMixin):
    """
    Location to a WebDAV server.

    The path sent to this location might be URL encoded.
    """

    PROPERTIES_REQUIRING_RESTART = (
        SSLConfigurationSectionMixin.PROPERTIES_REQUIRING_RESTART +
        (
            'url',
            'password_update',
            'proxy',
            'username',
            'idle_connection_timeout',
            'idle_connection_keepalive_interval',
            'connection_retry_count',
            'connection_retry_interval',
            'authentication_method',
            'allow_insecure_http',
            ))

    _HTTPClient = WebDAVClient

    def _getTrustedURLs(self):
        """
        List of urls trusted by this location for HTTPS.
        """
        trusted = [self.configuration.url]

        if (self.configuration.authentication_method ==
                TYPE_NAME.SHAREPOINT_ONLINE):
            # We also trust login.microsoftonline.com,
            # which has `graph.windows.net` as the non-SNI name,
            # but we have SNI so we use that name.
            trusted.append(('https://login.microsoftonline.com:443'))

        return trusted
