# Copyright (c) 2011-2015 Adi Roiban.
# See LICENSE for details.
"""
Web Manager resource for past accessing the activity log over a  REST
interface.
"""
from __future__ import absolute_import, unicode_literals
from base64 import b64decode

from twisted.internet import defer
from twisted.web import http, server

from chevah.server import (
    emit_error_details,
    force_unicode,
    json,
    )
from chevah.server.commons.constant import (
    LOG_TABLE_NAME,
    )
from chevah.server.commons.exception import ServerException
from chevah.server.commons.interface import ILocalManagerAvatar
from chevah.server.commons.csv import UnicodeCSVProducer
from chevah.server.event.database import LogEntryDBTable
from chevah.server.http.protocol import _CONTINUE
from chevah.server.http.resource import ChevahResource
from chevah.server.manager.db import DatabaseFetcher


class ActivityLogResource(ChevahResource):
    """
    Activity log download resource.
    """

    isLeaf = True

    def __init__(self, factory):
        super(ActivityLogResource, self).__init__()
        self._factory = factory
        # FIXME:5949:
        # We need to have it called `_table` to allow access from JSON-RPC.
        self._table = LogEntryDBTable(name=LOG_TABLE_NAME)
        self._fetcher = DatabaseFetcher(
            process=factory.service_configuration.root,
            )

        # Authenticated account requesting this resource.
        self._avatar = None

        self._cursor = 0
        self._criteria = None
        self._sorting = None

        # Here to help with testing as render_GET and render_POST return a
        # string.
        self._last_deferred = None

    def emitEvent(self, event_id, avatar=None, data=None):
        """
        Emit an event for this resource.
        """
        if avatar is None:
            avatar = self._avatar

        self._factory.emitEvent(event_id, avatar=avatar, data=data)

    def headersReceived(self, request):
        """
        See: `IChevahResource`.

        Called when the header part of the request was received.
        """
        # We don't want to create new session for download request(s).
        # All request should have been authenticated previously.
        session = request.getExistingSession()
        if not session:
            return http.BAD_REQUEST, 'Request is not authorized.'

        avatar = getattr(session, 'avatar', None)
        if not avatar:
            return http.BAD_REQUEST, 'Invalid session.'

        if not ILocalManagerAvatar.providedBy(avatar):
            return http.BAD_REQUEST, 'Bad authorization.'

        self._avatar = avatar
        return _CONTINUE

    def render_GET(self, request):
        """
        See: `Resource`

        Handles a GET request.
        """
        # GET requests are cached by default so we inform the client to
        # not cache this resource.
        request.setDisableCacheHeaders()

        options = {'db_uuid': 'undefined'}

        def decode_options():
            try:
                # FIXME:6201:
                # request.postpath will be unicode.
                options = json.loads(b64decode(request.postpath[0]))
            except IndexError:
                raise ServerException(
                    'Invalid request for logs download.')
            except (TypeError, ValueError) as error:
                raise ServerException(
                    'Unable to decode the download options. %s' % (
                        force_unicode(error),))

            if not isinstance(options, dict):
                raise ServerException('Invalid download options.')

            db_uuid = options.get('db_uuid', None)
            if not db_uuid:
                raise ServerException('No database selected for download.')

            return options

        try:
            options = decode_options()
            self._last_deferred = self._handle(request, options)

        except ServerException as error:
            self._last_deferred = defer.fail(error)

        self._last_deferred.addErrback(
            self._cbRaiseError, request, options['db_uuid'])
        self._last_deferred.addBoth(lambda _: request.finish())
        return server.NOT_DONE_YET

    def _handle(self, request, options):
        """
        Process request and generate CSV data.
        """
        # The input validation is done by the lower fetch method.
        # And options structure was validated before.
        self._criteria = options.get('criteria', ())
        self._sorting = options.get('sorting', {'id': 'asc'})
        db_uuid = options['db_uuid']

        # RFC 7111 for CSV content type.
        request.setHeader('content-type', 'text/csv')
        request.setHeader(
            'content-disposition',
            'inline; filename="activity-log-export.csv"',
            )

        def cb_write_content(request, db_uuid):
            """
            Called when we should generate the content for the body.
            """
            csv_producer = UnicodeCSVProducer(consumer=request.write)

            def write_rows(batch):
                """
                Called when we got a batch from the DB.
                """
                for entry in batch:
                    row = []
                    for column in self._table.columns:
                        # Table ID or datetime are not text.
                        # We force everything as text.
                        value = force_unicode(entry[column.name])
                        row.append(value)
                    csv_producer.makeRow(row)

            deferred = self._fetcher.fetch(
                db_uuid=db_uuid,
                processor=write_rows,
                criteria=self._criteria,
                sorting=self._sorting,
                avatar=self._avatar
                )
            # Pass the request to the next callback.
            deferred.addCallback(lambda _: request)
            return deferred

        def cb_handle(result):
            """
            Called when generation of CSV file is successful.
            """
            self.emitEvent(
                event_id='50026', data={'database_uuid': db_uuid})

        deferred = defer.succeed(request)
        deferred.addCallback(cb_write_content, db_uuid)
        deferred.addCallback(cb_handle)
        return deferred

    def _cbRaiseError(self, failure, request, db_uuid):
        """
        Emit error and raise a dedicated ServerError with details.
        """
        error = failure.value
        data = {
            'database_uuid': db_uuid,
            'tb': force_unicode(failure.getBriefTraceback()),
            }
        emit_error_details(
            error=error,
            event_id='50027',
            data=data,
            avatar=self._avatar,
            emitter=self.emitEvent,
            )
        request.setResponseCode(http.BAD_REQUEST, b'Request failed.')
