# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
This is the representation of an administrator after the authentication
process.

No authentication related code should be placed here.
It's OK to have authorization or accounting related code.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
from zope.interface import implementer

from chevah.server.commons.constant import (
    ROLE_PERMISSION,
    ROLE_PERMISSIONS,
    )
from chevah.server.commons.exception import AuthorizationException
from chevah.server.commons.interface import (
    IApplicationLocalManagerAvatar,
    ILocalManagerAvatar,
    IOSLocalManagerAvatar,
    )
from chevah.server.commons.match import get_matched_groups


@implementer(ILocalManagerAvatar)
class LocalManagerAvatarBase(object):
    """
    Avatar used by local manager service.
    """

    def __init__(self, name, roles, component, peer=None, uuid='Unknown'):
        self._name = name
        self._peer = peer
        self._roles = roles
        self._uuid = uuid
        self._component = component

    @property
    def name(self):
        '''Username for this avatar.'''
        return self._name

    @property
    def uuid(self):
        """
        See: ILocalManagerAvatar.
        """
        return self._uuid

    @property
    def email(self):
        """
        No email address implemented for this avatar.
        """
        return ''

    @property
    def emails(self):
        """
        No email address implemented for this avatar.
        """
        return []

    @property
    def roles_uuid(self):
        """
        See: ILocalManagerAvatar.

        List of associated roles.
        The admin might not be stored in the config so this is the only
        way to know what groups he has.
        """
        return [r.uuid for r in self._roles]

    @property
    def peer(self):
        '''Peer for this avatar.'''
        return self._peer

    @property
    def component(self):
        """
        See: `ILocalManagerAvatar`.
        """
        return self._component

    def checkPermission(self, target, action=ROLE_PERMISSION.ALL):
        """
        Raise an exception if administrator doesn't have the `action`
        permissions for `target`.

        It will normalize both target and permissions path to leading
        slashes.

        At configuration level we have:

        * /configuration/ - access to the configuration
        * /operation/ - access to the runnables
        """
        if not target.startswith('/'):
            target = '/' + target

        if not target.endswith('/'):
            # All targets should end in `/` so they will match a rule like
            # `/some/target/*`.
            target = target + '/'

        # The name as it should be set in the configuration and as it is
        # displayed in the error messages.
        canonical_target = target
        if target.startswith('/operation/'):
            # The end-user configuration uses the "operation/" prefix for
            # runnable permissions,
            # while in the low-level process the runnables are set in the
            # root of the process.
            target = target[len('/operation'):]

        permissions = []
        for role in self._roles:
            permissions.extend(role.permissions)

        for definition in permissions:
            target_rule = definition[0]
            allowed_actions = definition[1:]

            if not target_rule.startswith('/'):
                target_rule = '/' + target_rule

            if target_rule.startswith('/operation/'):
                # The end-user configuration uses the "operation/" prefix for
                # runnable permissions,
                # while in the low-level process the runnables are set in the
                # root of the process.
                target_rule = target_rule[len('/operation'):]

            if not allowed_actions:
                # Allow any action for a target by default.
                allowed_actions = (ROLE_PERMISSION.ALL,)

            if not (target_rule.endswith('/') or '*' in target_rule):
                # Automatically accept descendants for simple rules.
                target_rule = target_rule + '/*'

            if target_rule.startswith('/own/'):
                target_rule = '/configuration/administrators/{}/{}'.format(
                    self.uuid, target_rule[5:])

            if not get_matched_groups(target, target_rule, 'role_permissions'):
                # Not a definition for this target.
                continue

            # Do a quick validation to fail on invalid configurations.
            for action_validation in allowed_actions:
                if action_validation not in ROLE_PERMISSIONS:
                    raise AuthorizationException(
                        'Unknown action "%s".' % (action_validation,))

            if ROLE_PERMISSION.DENY in allowed_actions:
                # Deny any action for this target.
                raise AuthorizationException(
                    'Action "%s" denied for "%s".' % (action, canonical_target))

            if ROLE_PERMISSION.ALL in allowed_actions:
                # Anything is allowed, no need to check for a specific action.
                return

            if action in allowed_actions:
                # We got a match. Don't look any further.
                return

            # Current action not allowed.
            # Continue, maybe it is allowed in a following definition.

        # After checking all the rules, none of them matched the action.
        raise AuthorizationException(
            'Action "%s" not allowed for "%s".' % (action, canonical_target))


@implementer(IApplicationLocalManagerAvatar)
class ApplicationLocalManagerAvatar(LocalManagerAvatarBase):
    """
    Local manager that was authenticated based on application administrator.
    """


@implementer(IOSLocalManagerAvatar)
class OSLocalManagerAvatar(LocalManagerAvatarBase):
    """
    Local manager that was authenticated based on OS administrator.
    """
