# Copyright (c) 2011-2015 Adi Roiban.
# See LICENSE for details.
"""
Web Manager paginated database fetching controllers.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
from twisted.internet import defer

from chevah.server.commons.exception import ServerException


class DatabaseFetcher(object):
    """
    Database entries fetcher with support for pagination and batch processing.

    It received requests from JSON-RPC and convert them to low level API
    and then back to JSON-RPC data.

    It will fail if the DB is not already started.
    """
    #: Default number of entries to be processed in a SQL request (LIMIT).
    _batch_size = 1024

    def __init__(self, process, batch_size=None):
        self._process = process

        if batch_size:
            self._batch_size = batch_size

    def _getSource(self, db_uuid):
        """
        Return interaction for `db_uuid`.

        Here to help with testing.
        """
        if db_uuid is None:
            return self._process.resources.getAnalytics()

        for handler in self._process.event_handlers.getAll():
            if handler.uuid == db_uuid:
                return handler

        return None

    @defer.inlineCallbacks
    def fetch(self, db_uuid, processor, criteria=None, sorting=None,
              limit=None, start=0, avatar=None):
        """
        Returns a deferred which will fire with the database entries matching
        the filter and sort criteria as a result.

        db_uuid, criteria and sorting are untrusted values.

        Raises an error if no matching fetcher is found.

        If `limit` is none, it will generate files from `start` page up to the
        end.
        """
        component = self._getSource(db_uuid)

        if db_uuid is None:
            db_uuid = 'Default DB'

        if not component:
            raise ServerException(
                'Failed to get database for "%s".' % (db_uuid,))

        if not component.operating:
            raise ServerException(
                'Database for "%s" is not started.' % (component.name,))

        # FIXME:5949:
        # Find an API for this private variable access.
        table = component._table
        interaction = component._interaction

        sort_order = self._convertSortOrder(sorting)

        table = component._table
        interaction = component._interaction

        self._cursor = start

        result_size = 0
        # We might not have a limit, so default to batch size.
        remaining_limit = limit or self._batch_size
        while True:
            # We do multiple select in small batches, to make it easier
            # to support streaming.
            # For example for the CSV download.
            # If we have a limit, we make sure the last match is the size
            # up to the limit.
            # This means that for short results, we end up running the query
            # twice. Once for the first resul, and second time that will
            # return an empty result and will break the loop.
            batch_size = min(remaining_limit, self._batch_size)
            result = yield interaction.selectRows(
                table=table,
                criteria=criteria,
                limit=batch_size,
                offset=self._cursor,
                orderBy=sort_order,
                )

            if not result:
                # We reached the end of the table.
                break

            processor(result)

            select_size = len(result)
            result_size += select_size

            if limit:
                remaining_limit -= select_size
                if result_size >= limit:
                    # We have already returned enough results.
                    break

            # Continue with the next page.
            self._cursor += select_size

    @staticmethod
    def _convertSortOrder(sort_order=None):
        """
        Return a tuple in the form (column_name, sort_direction) from the
        LM UI sort order configuration.

        If no direction is provided, ASCENDING will be used by default.
        """
        if not sort_order:
            return None

        column_name = list(sort_order.keys())[0]
        direction = sort_order[column_name].lower()
        return [(column_name, direction)]
