# Copyright (c) 2011 Adi Roiban.
# See LICENSE for details.
"""
Chevah Web Manager Web Site.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
import os

from twisted.cred.portal import Portal
from twisted.web import http

from chevah_compat import local_filesystem

from chevah.server import MODULE_PATH
from chevah.server.authentication.credential_checker import (
    ServiceCredentialChecker,
    )
from chevah.server.commons.constant import (
    PRODUCT_VERSION,
    STATIC_FOLDER_NAME,
    )
from chevah.server.commons.jinja import (
    JinjaEnvironment,
    ViewModel,
    )
from chevah.server.commons.ssl import SSLServiceFactoryMixin
from chevah.server.http.factory import HTTPFactoryBase
from chevah.server.http.flowjs import InMemorySessionFlowJS
from chevah.server.http.protocol import HTTPSProtocol
from chevah.server.http.request import ChevahRequest
from chevah.server.http.resource import (
    JinjaResource,
    RedirectResource,
    RootResource,
    StaticFilesystemResource,
    StaticTextResource,
    )
from chevah.server.manager.activity_log import ActivityLogResource
from chevah.server.manager.oidc import AdministrationOIDCResource
from chevah.server.manager.protocol import AdministrationJSONRPC
from chevah.server.manager.realm import LocalManagerRealm

ADMIN_STATIC_FOLDER_NAME = 'manager_service'
IN_MEMORY_UPLOAD_CHILD_NAME = 'in-memory-upload'
JSON_CHILD_NAME = 'json'
OIDC_CHILD_NAME = 'oidc'
MANAGER_CHILD_NAME = 'manager'
# If we are going to add more REST resources, we should put them as
# /rest/v1/RESOURCE1
# /rest/v1/RESOURCE2
# /rest/v1/PARENT/RESOURCE1
ACTIVITY_LOG_CHILD_NAME = 'activity-log-export'
DOCUMENTATION_CHILD_NAME = 'documentation'

# The CSP policy is similar to the default one, but we inject 2 inline
# style generated by angular.
_LM_CSP_POLICY = ChevahRequest._CSP_POLICY[:]


class LocalManagerHTTPRequest(ChevahRequest):
    """
    A request designed for unsecured Web Manager connections.
    """
    _CSP_POLICY = _LM_CSP_POLICY
    session_cookie_force_secure = False
    session_cookie_basename = 'CHEVAH_MANAGER_INSECURED_SESSION'


class LocalManagerHTTPSRequest(ChevahRequest):
    """
    A request designed for secured Web Manager connections.

    It will always force secured session cookies.
    """
    _CSP_POLICY = _LM_CSP_POLICY
    session_cookie_basename = 'CHEVAH_MANAGER_SESSION'


class LocalManagerViewModel(ViewModel):
    """
    View manager for local manager main admin page.
    """
    #: This is the default value as resource will update it based on
    #:
    CACHE_BUSTER = PRODUCT_VERSION

    def __init__(self, base_url_path):
        self._base_url_path = base_url_path

    @property
    def BASE_URL_PATH(self):
        """
        Name of the OpenID connect method enabled that is enabled.

        Empty is not enabled.
        """
        return self._base_url_path


class LoginViewModel(LocalManagerViewModel):
    """
    View manager for local manager login page.
    """

    def __init__(self, base_url_path, oidc_resource):
        self._oidc_resource = oidc_resource
        super(LoginViewModel, self).__init__(base_url_path=base_url_path)

    @property
    def oidc_name(self):
        """
        Name of the OpenID connect method enabled that is enabled.

        Empty is not enabled.
        """
        oidc = self._oidc_resource.getOIDC()
        if not oidc:
            return ''
        return oidc.name

    @property
    def oidc_url(self):
        """
        URL to initiate the OpenID connect auth.

        Empty if no OpenID connect method is enabled.
        """
        oidc = self._oidc_resource.getOIDC()
        if not oidc:
            return ''
        return '{}/{}?{}'.format(
            self.BASE_URL_PATH,
            OIDC_CHILD_NAME,
            oidc.request_argument,
            )


class LocalManagerJinjaResource(JinjaResource):
    """
    Jinja template for local manager.

    This is used by both the login page and the manager page.
    """
    #: This is a class member to share a single environment for all resources.
    _jinja_environment = JinjaEnvironment(ADMIN_STATIC_FOLDER_NAME, 'jinja')

    def __init__(self, template, model):
        super(LocalManagerJinjaResource, self).__init__(
            environment=self._jinja_environment,
            template=template,
            model=model,
            )

    def render_POST(self, request):
        """
        The root resource should usually only receive GET request to load the
        JS single page app.

        This is here to support OpenID Connect response handling
        and simplify the redirect URL structure.
        """
        auth_url = '{}/{}{}'.format(
            request.site.base_url_path,
            OIDC_CHILD_NAME,
            request.uri.decode('ascii'),
            )
        request.setHeader(b"location", auth_url.encode('ascii'))
        request.setResponseCode(http.TEMPORARY_REDIRECT)
        return b'Redirecting to the our OpenID Connect handler.'


class LocalManagerFactoryBase(HTTPFactoryBase):
    """
    Local manager's base class for both HTTP and HTTPS.
    """

    # Reference to the OIDC resource for the local manager.
    _oidc_resource = None

    def _getRootResource(self):
        """
        Return root resource for HTTP/HTTPS service.
        """
        root_path = os.path.join(
            MODULE_PATH,
            STATIC_FOLDER_NAME,
            ADMIN_STATIC_FOLDER_NAME,
            )

        other_resource = StaticFilesystemResource(
            path=root_path, ignoredExts=['.html'])
        other_resource.childNotFound = StaticTextResource(
            code=404,
            content_type='text/html',
            content=(
                '<html>'
                '<head><title>Page not found.</title></head><body></body>'
                '</html>'
                ),
            )

        # We initiate the resource here as it is needed by the view model.
        # but only add it to the tree later as part of _addRootChildren.
        # This is due to the way our HTTP server shared code is designed.
        self._oidc_resource = AdministrationOIDCResource(factory=self)
        # Root with login index and static files.
        root = RootResource(
            root_resource=LocalManagerJinjaResource(
                template='index.j2',
                model=LoginViewModel(
                    oidc_resource=self._oidc_resource,
                    base_url_path=self.base_url_path,
                    ),
                ),
            other_resource=other_resource,
            )

        self._addRootChildren(root)
        return root

    def _addRootChildren(self, root):
        """
        Add default children and local manager specific children.
        """
        credential_checkers = [
            ServiceCredentialChecker(self.service_configuration),
            ]
        realm = LocalManagerRealm(self.service_configuration)
        self.portal = Portal(realm, credential_checkers)

        # Add documentation.
        doc_segments = local_filesystem.installation_segments
        # FIXME:4291:
        # 'doc' should be renamed to 'documentation'
        doc_segments.extend(['doc', 'html'])
        root.putChild(
            DOCUMENTATION_CHILD_NAME.encode('ascii'),
            StaticFilesystemResource(
                path=local_filesystem.getRealPathFromSegments(doc_segments),
                ignoredExts=['.html'],
                ),
            )

        # Add JSON-RPC.
        root.putChild(
            JSON_CHILD_NAME.encode('ascii'),
            AdministrationJSONRPC(factory=self))

        root.putChild(OIDC_CHILD_NAME.encode('ascii'), self._oidc_resource)

        # Add handler for uploaded files which are stored in memory.
        root.putChild(
            IN_MEMORY_UPLOAD_CHILD_NAME.encode('ascii'),
            InMemorySessionFlowJS())

        # Add main manager page.
        manager_model = LocalManagerViewModel(base_url_path=self.base_url_path)
        root.putChild(
            MANAGER_CHILD_NAME.encode('ascii'),
            LocalManagerJinjaResource(
                template='manager.j2',
                model=manager_model,
                ),
            )

        # Add activity log streaming page.
        root.putChild(
            ACTIVITY_LOG_CHILD_NAME.encode('ascii'),
            ActivityLogResource(factory=self))

        self._populateStaticFolder(root)

    def renderLoginError(self, request, error_code):
        """
        Render a redirect to the login page in case of errors.

        `error_code` is an LM_ERROR value shared between backend and frontend.

        The URL from here needs to be synchronized with the LM GUI.
        """
        # The login page will show the error via the GET JS error processing.
        # This is why we redirect with SEE_OTHER so that a POST is converted
        # to GET.
        # We pass the URL as bytes since we don't want any URL escaping.
        resource = RedirectResource(
            '/#!?backend_error={}'.format(error_code).encode('ascii'),
            http.SEE_OTHER)
        return resource.render(request)

    def renderLoginSuccess(self, request, avatar):
        """
        Render a redirect page for a successful authentication.
        """
        session = request.getSession()
        session.avatar = avatar
        request._setSession(session)

        # We need to redirect via HTML as otherwise the session cookie is
        # not set.
        # We redirect with SEE_OTHER so that a POST is converted to GET.
        redirect = RedirectResource(
            self.base_url_path + '/' + MANAGER_CHILD_NAME,
            http.SEE_OTHER)
        return redirect.render(request)


class LocalManagerHTTPFactory(LocalManagerFactoryBase):
    """
    Local manager site over HTTP.

    For now, HTTP is not officially supported and used only for testing
    and development.
    """
    requestFactory = LocalManagerHTTPRequest


class LocalManagerHTTPSFactory(
        LocalManagerFactoryBase, SSLServiceFactoryMixin):
    """
    Local manager site over HTTPS.
    """
    protocol = HTTPSProtocol
    requestFactory = LocalManagerHTTPSRequest
