# Copyright (c) 2022 Adi Roiban.
# See LICENSE for details.
"""
Handle OpenID Connect authentication for the local manager services.

This links our internal credentials checker system with the external
OpenID Connect services.

Mainly it helps generate the link where users are redirected,
and parse/consumes the OAuth2 response fro, the remote OIDC servers.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
from twisted.internet import defer
from twisted.web import http
from twisted.web.server import NOT_DONE_YET

from chevah.server import emit_error_details
from chevah.server.commons.constant import LM_ERROR
from chevah.server.commons.exception import ServerException
from chevah.server.commons.interface import ILocalManagerCredentials
from chevah.server.http.resource import (
    ChevahResource,
    )
from chevah.server.identity.avatar import PeerAvatar


class AdministrationOIDCResource(ChevahResource):
    """
    Handles the UX for session and OpenID connect based authentication.

    There is no UI, and on any result it redirects to the login
    or manager page.

    The actual login process is delegated to the `portal.login()`.
    This is just the adapter for OpenID connect to our standard login process.

    Note that the manager also performs a login operation over JSON-RPC.
    """
    isLeaf = True

    def __init__(self, factory):
        self._factory = factory
        super(AdministrationOIDCResource, self).__init__()

    def render_GET(self, request):
        """
        Called when user loads the page to initiate the first step of OIDC.
        """
        authentication = self.getOIDC()
        if not authentication:
            return self._factory.renderLoginError(
                request, error_code=LM_ERROR.OIDC_NOT_ENABLED)

        auth_url = authentication.getAuthorizationURL(request)
        if not auth_url:
            return self._factory.renderLoginError(
                request, error_code=LM_ERROR.OIDC_NO_URL)

        request.setResponseCode(http.TEMPORARY_REDIRECT)
        request.setHeader(b"location", auth_url)
        return b'Redirecting to the OpenID Connect provider.'

    def render_POST(self, request):
        """
        Process the authentication request from the browser or API.
        """
        # This is to prevent someone having the ID of on old session, which
        # now is re-used with another credentials.
        request.removeSession()
        oidc = self.getOIDC()
        if not oidc:
            return self._factory.renderLoginError(
                request, error_code=LM_ERROR.OIDC_NOT_ENABLED)

        # Try to get the credentials from the OIDC response.
        try:
            credentials = oidc.getAdministratorCredentials(request)
            if not credentials:
                raise ServerException('OIDC response is for another method.')
        except ServerException as error:
            self._factory.emitEvent(
                '50001',
                peer=request.client,
                data={'details': error.message},
                )
            return self._factory.renderLoginError(
                request, error_code=LM_ERROR.OIDC_GET_CREDENTIALS)

        self._doLogin(credentials, request)
        return NOT_DONE_YET

    def getOIDC(self):
        """
        Return the first authentication method that will handle OpenID connect
        for this local manager site.

        This is public API as it is also used by the Jinja renderer.

        Return None if no OIDC is enabled.
        """
        root = self._factory.root
        return root.authentications.getOIDC(
            self._factory.service_configuration.parent.authentications,
            # For manager, we don't have a default list.
            # Manager services should always have an explicit list.
            # Disable OIDC by default, just to reuse the code.
            default=[],
            )

    @defer.inlineCallbacks
    def _doLogin(self, credentials, request):
        """
        Called after we have the credentials.

        This is called outside of the simple request rendering flow.
        So it should explicitly write the content to the request and finalize
        the request.

        This should never fail.
        Instead it should display the error as a page and finalize the request.
        """
        try:
            result = yield self._factory.portal.login(
                credentials, None, ILocalManagerCredentials)
            (interface, avatar, logout) = result
            result = self._factory.renderLoginSuccess(
                request=request, avatar=avatar)
            request.write(result)

        except Exception as error:
            emit_error_details(
                event_id='50001',
                avatar=PeerAvatar(request.client),
                error=error,
                data={},
                emitter=self._factory.emitEvent,
                )

            # Redirect to the login page.
            result = self._factory.renderLoginError(
                request, error_code=LM_ERROR.CREDENTIALS_DENIED)
            request.write(result)

        finally:
            request.finish()
