# Copyright (c) 2011-2013 Adi Roiban.
# See LICENSE for details.
"""
Chevah Web Manager protocol.

The JSON-RPC layer should  be as light as possible.
It should just check permissions, forward the request to dedicate internal API
and components and present the result of the errors.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

import base64
import io
import struct
import time
import traceback
from contextlib import contextmanager


from bunch import Bunch
from twisted.internet import defer, threads
from twisted.python.failure import Failure

from chevah_keycert.exceptions import BadKeyError, KeyCertException
from chevah_keycert.ssh import Key, FingerprintFormats
from chevah_keycert.ssl import (
    generate_csr,
    generate_ssl_self_signed_certificate,
    )

from chevah.server import emitFromError, force_unicode
from chevah.server.commons.constant import (
    DEFAULT_KEY_TYPE,
    DEFAULT_KEY_SIZE,
    ROLE_PERMISSION,
    )
from chevah.server.commons.exception import (
    AuthorizationException,
    ServerError,
    ServerException,
    NoSuchSectionError,
    )
from chevah.server.commons.credential import LocalManagerCredentials
from chevah.server.commons.interface import ILocalManagerCredentials, IRunnable
from chevah.server.commons.ssl import (
    dump_certificate,
    dump_key,
    load_PKCS12,
    )
from chevah.server.json.rpc import JSONRPCError, JSONRPCResource
from chevah.server.manager.db import DatabaseFetcher
from chevah.server.openpgp.util import generate_pgp_key
import six


FAILED_AUTHENTICATION = {
    'code': 50001,
    'message': 'Authentication failed.',
    'data': None,
    }

# Marker when API reas access is not permitted.
_NO_READ_PERMISSION = object()


class ServerJSONRPCError(JSONRPCError):
    """
    JSRONRPCError raised by the server.

    It has an 'id' instead or JSON-RPC 'code' attribute.
    """

    def __init__(self, id, data):
        self.id = id
        self.data = data
        self.value = {
            'id': id,
            'data': self.data,
            'message': 'See code and data for details.',
            }

    def _repr(self):  # noqa:cover
        """
        Here to support generic testing assertions.
        """
        return repr(self)


class AdministrationJSONRPC(JSONRPCResource):
    """
    JSON-RPC resource for the local manager backend API.

    This is used by the web GUI and also by other management functions like
    synchronizing the configuration to an external server.
    """

    def __init__(self, factory):
        super(AdministrationJSONRPC, self).__init__()
        self.factory = factory
        # These are the methods that don't need authentication.
        # All other methods are automatically protected and the session is
        # automatically refreshed.
        self.public_methods = ['login', 'is_valid_session']

        # This is used to filter the properties based on read permissions,
        # but still leave stubs if no read permissions is available for
        # a part of the process.
        self._BRANCHES = [
            '/' + n
            for n in self.main_configuration.root.getPublicSectionNames()
            ]
        self._BRANCHES.extend([
            '/configuration/' + n
            for n in self.main_configuration.getPublicSectionNames()
            ])

    @property
    def service_configuration(self):
        """
        Configuration of this service.
        """
        return self.factory.service_configuration

    @property
    def main_configuration(self):
        """
        Application configuration for the server under which this service
        is running.
        """
        return self.service_configuration.root.configuration

    @property
    def _process(self):
        """
        Process under which the JSON-RPC services is running.
        """
        return self.main_configuration.parent

    def logInternalFailure(self, failure, peer=None, avatar=None):
        '''
        Log an internal JSON-RCP error.
        '''
        if failure.check(ServerException):
            details = failure.value.message
            code = '50014'
        else:
            details = force_unicode(failure.getBriefTraceback())
            code = '50004'

        data = {'details': details}
        self.factory.emitEvent(code, peer=peer, avatar=avatar, data=data)
        return int(code)

    def _getAvatar(self, request):
        """
        Return avatar from session associated with request.
        """
        if request.session:
            return request.session.avatar

        return None

    def jsonrpc_ping(self, request):
        """
        Notification from local web admin page.
        """
        # Access the session to update the session timeout.
        request.getSession()
        return 'pong'

    def jsonrpc_get_index(self, request):
        """
        Return full configuration.
        """
        return self.jsonrpc_get_property(request, path='/')

    def jsonrpc_get_property(self, request, path):
        """
        Return full configuration.
        """
        with self._handlePropertyError(
            request, 'Failed to retrieve the process properties.'
                ) as avatar:
            result = self._getPermittedRead(avatar, path=path)
            self.factory.emitEvent(
                '50002',
                data={'path': path},
                avatar=avatar,
                )
            return result

    def _getPermittedRead(self, avatar, path='/', target=None):
        """
        Return a filtered list of properties from `target` for which
        `avatar` has read access, starting at `path`.

        If `target` is None, it will start from the whole process properties.
        """
        result = {}
        if target is None:
            target = self._process.getProperty(path.lstrip('/'))

        for key, value in target.items():
            candidate = value
            try:
                avatar.checkPermission(path + key, ROLE_PERMISSION.READ)
            except AuthorizationException:
                # No access to this key.
                # While we don't have access to this key,
                # we will continue as we might have permission to a sub-key.
                candidate = _NO_READ_PERMISSION

            if isinstance(value, dict):
                candidate = self._getPermittedRead(
                    avatar, path=path + key + '/', target=value)
                if candidate:
                    result[key] = candidate
                elif path + key in self._BRANCHES:
                    # We don't have permissions for the branch itself.
                    # For now, we start with an empty value as later we
                    # might have read permissions to sub-nodes of this.
                    result[key] = {}
            else:
                if candidate is not _NO_READ_PERMISSION:
                    result[key] = candidate

        return result

    def jsonrpc_sync_pull(self, request):
        """
        Return the full configuration to be used by a node for synchronization.
        """
        # FIXME:4786:
        # For now, we just return the full configuration each time.
        # In the future, we can have a long lasting request that will send
        # the time since last synchronization and if no changes were made
        # delay the response until changes are made... or until a timeout.
        with self._handlePropertyError(
                request, 'Failed to sync configuration.') as avatar:
            avatar.checkPermission('sync')

            # FIXME:3540:
            # Besides sending the configuration we should also send the
            # runnable state to see if you manually start/stop a component.
            result = self._process.configuration.getPropertyForSync()

            # Update the remote node base server config based on the
            # local node configuraiton.
            # For security reasons, for the future sync might not be allowed
            # to export the list of administrators.
            # We will only export a single administrator, which is the
            # node itself.
            node_uuid = request.session.avatar.uuid
            node_config = self.main_configuration.administrators.getSection(
                node_uuid)
            result['server']['name'] = node_config.name
            result['server']['uuid'] = node_config.uuid
            result['server']['description'] = node_config.description

            # As part of the sync, we don't want to export the controller
            # instance node variables.
            result['node_variables'] = {
                'error': 'Node variables sync via "raw_node_variables"'}
            # Make sure we export the node variables for the remote node
            # and not for this node
            result['raw_node_variables'] = node_config.node_variables

            self.factory.emitEvent('50017', avatar=avatar)
            return result

    @contextmanager
    def _handlePropertyError(self, request, message):
        """
        Handle the errors raised while getting the properties of the system.
        """
        try:
            avatar = self._getAvatar(request)
            yield avatar
        except Exception as error:
            # This should not normally raise any errors, but
            # in case it does, we need to inform the remote client and
            # log an internal error.
            code = self.logInternalFailure(Failure(error), avatar=avatar)
            response = {
                'code': code,
                'message': message,
                'data': None,
                }
            raise JSONRPCError(response)

    def jsonrpc_db_fetch(
        self, request, page=0, limit=100, criteria=None,
        sorting=None, db_uuid=None,
            ):
        """
        Returns the database entries matching the filter query sorted
        according to the specified sorting clause.

        Raises an error if no matching fetcher is found.

        Valid sources:
            - logs: log entries
            - last_successful_login: last successful login report entries
        """
        return self._db_fetch(
            request, page, limit, criteria, sorting, db_uuid)

    @defer.inlineCallbacks
    def _db_fetch(self, request, page, limit, criteria, sorting, db_uuid):
        """
        Our JSON-RPC framework doesn't have support for inline callbacks.
        So this is the wrapper.
        """
        start_time = time.time()
        avatar = self._getAvatar(request)
        fetcher = DatabaseFetcher(self._process, batch_size=limit)
        result = []

        def process(items):
            """
            Called to process entries.
            """
            if not items:
                return

            result.extend(items)

        start = max(0, page * limit)
        try:
            yield fetcher.fetch(
                db_uuid=db_uuid,
                processor=process,
                criteria=criteria,
                sorting=sorting,
                limit=limit,
                start=start,
                avatar=avatar,
                )
        except ServerException as error:
            raise error
        except Exception as error:
            details = force_unicode(error)
            raise ServerException('Failed to get data. %s' % (details,))

        event_data = {
            'db_uuid': db_uuid,
            'duration': time.time() - start_time,
            'size': len(result),
            'cursor': start,
            'sorting': sorting,
            }
        self.factory.emitEvent('50028', data=event_data, avatar=avatar)
        defer.returnValue({'data': result})

    def jsonrpc_apply(self, request, changes):
        """
        Apply server configuration changes.

        `changes` is a dict keyed using 'change_id' and containing dict
        with the following format:
        * operation: update | create | delete
        * key: path to object that needs update, relative to the
          configuration tree
        * value: value that should be updated. Can be None (ex for delete)
          but should be defined as such.

        `key` is in the format 'section/subsection/option_name'.

        Returns a dict with:
        * configuration : full configuration as changed after update
        * results: a dict keyed using 'change_id' and containing dict with
          the following format:
          * error - error of operation on error, None on success.
        """
        avatar = self._getAvatar(request)

        try:
            results = self._applyChanges(avatar, changes)
        finally:
            try:
                self.main_configuration.save()
                self.factory.emitEvent('20046', avatar=avatar)
            except Exception as error:
                server_error = ServerError(
                    '20049',
                    data={
                        'details': force_unicode(error),
                        'avatar': avatar,
                        },
                    )
                results = self._saveFailed(server_error, changes)

        if self._hasErrors(results):
            serialized_results = {}
            for key, value in results.items():
                error = value['error']
                if not error:
                    serialized_results[key] = value
                else:
                    emitFromError(error, emitter=self.factory.emitEvent)
                    serialized_results[key] = {
                        'error': {'id': error.id, 'data': error.data},
                        'result': None,
                        }

            data = {
                'changes': self._maskPassword(changes),
                'results': serialized_results,
                }
            self.factory.emitEvent(
                '50006', avatar=avatar, data=data)
        else:
            data = {
                'changes': self._maskPassword(changes),
                'results': results,
                }
            self.factory.emitEvent(
                '50005', avatar=avatar, data=data)

        response = self._getPermittedRead(avatar)
        response['results'] = results
        return response

    def _saveFailed(self, error, changes):
        """
        Return results when failing to save the changes.

        It only show the error once, for the first change and
        make it look like all other changes were successfully applied.

        All changes are applied, but they are not saved.
        """
        results = {}
        first = True
        for change_id, value in changes.items():
            if first:
                results[change_id] = {'error': error, 'result': None}
                first = False
            else:
                results[change_id] = {'error': None, 'result': None}
        return results

    def _hasErrors(self, results):
        """
        Return `True` if the result contains one or more errors.
        """
        for change_id, change_result in results.items():
            if change_result['error']:
                return True

        return False

    def _applyChanges(self, avatar, changes):
        """
        Apply changes and return a list of errors.

        Empty list is returned for no errors.
        """
        if not isinstance(changes, dict):
            data = {
                'changes': changes,
                'details': str(type(changes)),
                }
            error = ServerError('50011', data=data)
            return {'0': {'error': error}}

        results = {}
        for change_id, change_value in changes.items():
            change_result = {
                'result': None,
                'error': None,
                }
            try:
                change_result['result'] = self._applyChange(
                    avatar, change_value)
            except AuthorizationException as error:
                data = {'details': force_unicode(error)}
                change_result['error'] = ServerError(
                    '50009', avatar=avatar, data=data)
            except ServerError as error:
                change_result['error'] = error
            except ServerException as error:
                try:
                    operation = change_value['operation']
                except Exception:
                    operation = 'unknown'

                try:
                    path = change_value['path']
                except Exception:
                    path = 'unknown'

                data = {
                    'operation': operation,
                    'target': path,
                    'details': error.message,
                    }
                change_result['error'] = ServerError(
                    '50018', avatar=avatar, data=data)
            except Exception as error:
                data = {
                    'details': force_unicode(error),
                    'tb': traceback.format_exc(),
                    }
                change_result['error'] = ServerError(
                    '50003', avatar=avatar, data=data)

            results[change_id] = change_result

        return results

    def _applyChange(self, avatar, change):
        """
        Apply a specific change.

        Return the result from configuration object for change operation.

        Raise ServerException if operation is unknown.
        """
        if not isinstance(change, dict):
            raise ServerException('Bad format for change: %s' % change)

        if (
            'operation' not in change or
            'path' not in change or
            'value' not in change
                ):
            raise ServerException('Missing fields for change: %s' % change)

        if change['operation'] == 'update':
            avatar.checkPermission(
                'configuration/' + change['path'], ROLE_PERMISSION.UPDATE)
            result = self.main_configuration.setProperty(
                property_path=change['path'], value=change['value'])
        elif change['operation'] == 'create':
            avatar.checkPermission(
                'configuration/' + change['path'], ROLE_PERMISSION.CREATE)
            result = self.main_configuration.createProperty(
                property_path=change['path'], value=change['value'])
            result = result.uuid

            # This is async code, but we don't need to worry about it
            # as it uses Runnable.start() which will handle any errors.
            self._autoStartIfRunnable(result)

        elif change['operation'] == 'delete':
            avatar.checkPermission(
                'configuration/' + change['path'], ROLE_PERMISSION.DELETE)
            result = self.main_configuration.deleteProperty(
                property_path=change['path'])
        else:
            raise ServerException(
                'Unknown change request "%(operation)s" for "%(path)s".' % (
                    change))

        return result

    def _autoStartIfRunnable(self, uuid):
        """
        Start the runnable with given UUID, if found and enabled.
        """
        for runner_manager in self._process.getAll():
            try:
                component = runner_manager.getSection(uuid)
            except NoSuchSectionError:
                continue

            if not component:
                # Not the component for this type.
                continue

            if not IRunnable.providedBy(component):  # noqa:cover
                # Not a runnable.
                raise RuntimeError(
                    'Non-Runnable created in ChevahServerProcess from UI.')

            # We have a Runnable, not just any component.
            runnable = component

            if not runnable.enabled:
                # No need to start.
                return

            runnable.start()
            # No need to search further.
            return

    def _maskPassword(self, changes):
        """
        Mask password from `changes`.
        """
        if not isinstance(changes, dict):
            return changes

        # We make a copy of the dict since we are going to change its content.
        changes = changes.copy()

        # Update for password field.
        path = changes.get('path', '')
        if path.endswith('password'):
            changes['value'] = '***'

        if path.endswith('secret'):
            changes['value'] = '***'

        if 'password_update' in path:
            changes['value'] = '***'

        if 'multi_factor_authentication' in path:
            changes['value'] = '***'

        if 'private_key' in path:
            changes['value'] = '***'

        if 'ssl_key' in path:
            changes['value'] = '***'

        if 'ssl_certificate' in path and 'ssl_certificate_' not in path:
            # We only care about `ssl_certificate` but not more.
            changes['value'] = '***'

        # Recursive filter the content.
        for key, value in changes.items():
            if isinstance(value, dict):
                changes[key] = self._maskPassword(value)
            elif isinstance(value, six.string_types) and 'password' in key:
                changes[key] = '***'
            elif isinstance(value, six.string_types) and 'secret' in key:
                changes[key] = '***'
            elif (
                isinstance(value, six.string_types)
                and 'multi_factor_authentication' in key
                    ):
                changes[key] = '***'
            elif isinstance(value, six.string_types) and key == 'ssl_key':
                changes[key] = '***'
            elif (
                isinstance(value, six.string_types)
                and key == 'ssl_certificate'
                    ):
                changes[key] = '***'
            elif isinstance(value, six.string_types) and 'private_key' in key:
                changes[key] = '***'
            else:
                pass

        return changes

    def jsonrpc_login(self, request, username, password):
        '''Create and return a new session token based on provided
        username and password credentials.'''
        username = six.text_type(username)
        password = six.text_type(password)

        def cb_authenticationDone(result):
            (interface, avatar, logout) = result
            session = request.getSession()
            session.avatar = avatar
            request._setSession(session)
            return {'session_id': session.uid.decode('ascii')}

        def cb_authenticationError(failure):
            failure.trap(ServerError)
            emitFromError(
                failure.value,
                avatar=self._getAvatar(request),
                emitter=self.factory.emitEvent,
                )
            raise JSONRPCError(FAILED_AUTHENTICATION)

        credentials = LocalManagerCredentials(
            username=username,
            password=password,
            request=request,
            creator=self.factory,
            )

        deferred_login = self.factory.portal.login(
            credentials, None, ILocalManagerCredentials)
        deferred_login.addCallbacks(
            cb_authenticationDone, cb_authenticationError)
        return deferred_login

    def jsonrpc_logout(self, request):
        """
        Delete the current session.
        """
        try:
            avatar = request.session.avatar
            request.session.expire()
        except (AttributeError, KeyError):
            # Session was already expired.
            # Ignore logout requests for un-authenticated sessions.
            return True

        self.factory.emitEvent('50010', avatar=avatar)
        return True

    def jsonrpc_is_valid_session(self, request):
        """
        Return True if session attached to the request is valid.
        """
        session = request.getExistingSession()
        if session and hasattr(session, 'avatar'):
            return {
                'is_valid': True,
                'username': session.avatar.name,
                'uuid': session.avatar.uuid,
                'roles_uuid': session.avatar.roles_uuid,
                }
        else:
            return {
                'is_valid': False
                }

    def _actOnRunnableManager(self, uuid, family, callback, request):
        """
        Run `callback` for runnable with `uuid`.
        """
        avatar = self._getAvatar(request)

        try:
            avatar.checkPermission(
                '/operation/' + family + '/' + uuid, ROLE_PERMISSION.UPDATE)
        except AuthorizationException as error:
            return defer.fail(error)

        def cb_get_runnable(parent):
            """
            Return runnable with uuid.
            """
            if not parent:
                raise ServerException('Runnable without a family.')

            runnable = parent.getSection(uuid)
            if not runnable:
                raise ServerException('No such runnable.')

            deferred = callback(runnable, avatar)
            # The action result will return the new runnable state.
            deferred.addCallback(lambda _: runnable.getProperty())
            return deferred

        def eb_wrap_in_jsonrpc_error(failure):
            """
            Forward the error to JSON-RPC client..
            """
            details = (
                'Failed to process action for runnable "{}".'.format(uuid))
            self._fail(request, details, failure.value)

        parent = getattr(self._process, family, None)
        deferred = defer.succeed(parent)
        deferred.addCallback(cb_get_runnable)
        deferred.addErrback(eb_wrap_in_jsonrpc_error)
        return deferred

    def _cbStartRunnable(self, runnable, avatar):
        """
        Start the runnable.
        """
        return runnable.start(avatar=avatar)

    def _cbStopRunnable(self, runnable, avatar):
        """
        Stop the runnable.
        """
        return runnable.stop(avatar=avatar)

    def _cbRestartRunnable(self, runnable, avatar):
        """
        Restart the runnable.
        """
        return runnable.restart(avatar=avatar)

    def jsonrpc_runnable_start(self, request, family, uuid):
        """
        Start runnable with `uuid` from family.
        """
        return self._actOnRunnableManager(
            uuid=uuid,
            family=family,
            callback=self._cbStartRunnable,
            request=request,
            )

    def jsonrpc_runnable_stop(self, request, family, uuid):
        """
        Stop runnable with `uuid` from family.
        """
        return self._actOnRunnableManager(
            uuid=uuid,
            family=family,
            callback=self._cbStopRunnable,
            request=request,
            )

    def jsonrpc_runnable_restart(self, request, family, uuid):
        """
        Restart runnable with `uuid` from family.
        """
        return self._actOnRunnableManager(
            uuid=uuid,
            family=family,
            callback=self._cbRestartRunnable,
            request=request,
            )

    def jsonrpc_runnable_check(self, request, family, configuration):
        """
        Check that the `configuration` dict is a valid for runnable.
        """
        return self._runnable_check(request, family, configuration)

    @defer.inlineCallbacks
    def _runnable_check(self, request, family, configuration):
        """
        Used as a wrapper since our JSON-RPC base code doesn't support
        inline callbacks.
        """
        try:
            uuid = configuration.get('uuid', 'missing-uuid')
            parent = getattr(self.factory.root, family)
            resource = parent.getSection(uuid)
            if not resource:
                raise ServerException(
                    'Runnable with uuid "{}" was not found.'.format(uuid))

            result = yield resource.check(configuration)
            defer.returnValue(result)
        except ServerError as error:
            self._failWithError(request, error)
        except Exception as error:
            self._fail(request, 'Failed to check configuration.', error)

    def _getUploadedFileData(self, request, uuid):
        """
        Return the content of the file just uploaded via flow js.
        """
        session = request.getSession()
        if not hasattr(session, 'flowjs'):
            self._fail(request, 'No file was uploaded.', error=None)

        file_content = session.flowjs.get(uuid, None)
        if not file_content:
            self._fail(request, 'File is empty or does not exist.', error=None)

        return file_content

    def jsonrpc_ssh_key_export(
            self, request, uuid, import_password=None, export_password=None):
        """
        Export SSH key stored as file with `uuid` in all supported formats.

        If key is an encrypted private key, `import_password` is used to
        decrypt the file.

        If `export_password` is provided, it will be used to encrypt the
        exported private keys.
        """
        file_content = self._getUploadedFileData(request, uuid)

        if import_password:
            import_password = import_password.encode('utf-8')

        try:
            import_format = Key.getKeyFormat(file_content)
            key = Key.fromString(file_content, passphrase=import_password)
            result = self._exportKey(key, password=export_password)
            result['import_format'] = import_format
            return result
        except (KeyCertException, struct.error) as error:
            self._fail(request, 'Fail to export SSH key.', error)

    def _exportKey(self, key, password=None):
        """
        Export the key in all supported formats.
        """
        if password:
            password = password.encode('utf-8')

        is_public = key.isPublic()
        public_key = key.public()

        openssh_private = None
        sshcom_private = None
        putty_private = None
        putty_v3_private = None
        openssh_v1_private = None

        if not is_public:

            openssh_v1_private = key.toString(
                type='openssh_v1', passphrase=password).decode('ascii')
            putty_private = key.toString(
                type='putty', passphrase=password).decode('ascii')
            putty_v3_private = key.toString(
                type='putty_v3', passphrase=password).decode('ascii')

            try:
                openssh_private = key.toString(
                    type='openssh', passphrase=password).decode('ascii')
            except BadKeyError:
                openssh_private = (
                    'Key type not supported by OpenSSH PEM format.')

            try:
                sshcom_private = key.toString(
                    type='sshcom', passphrase=password).decode('ascii')
            except BadKeyError:
                sshcom_private = 'Key type not supported by SSH.com format.'

        openssh_public = public_key.toString(type='openssh').decode('ascii')

        try:
            sshcom_public = public_key.toString(type='sshcom').decode('ascii')
        except BadKeyError:
            sshcom_public = 'Key type not supported by SSH.com format.'

        return {
            'fingerprint_md5': key.fingerprint(FingerprintFormats.MD5_HEX),
            'fingerprint_sha1': key.fingerprint(
                FingerprintFormats.SHA1_BASE64),
            'fingerprint_sha256': key.fingerprint(
                FingerprintFormats.SHA256_BASE64),
            'is_public': is_public,
            'openssh_private': openssh_private,
            'openssh_v1_private': openssh_v1_private,
            'openssh_public': openssh_public,
            'sshcom_private': sshcom_private,
            'sshcom_public': sshcom_public,
            'putty_private': putty_private,
            'putty_v3_private': putty_v3_private,
            }

    def jsonrpc_ssh_key_generate(
        self, request, type=DEFAULT_KEY_TYPE, size=DEFAULT_KEY_SIZE,
        password=None,
            ):
        """
        Generate a new SSH key of `type` and `size`, then export it in all
        supported formats.

        If `password` is provided, it will be used to encrypt the
        exported private keys.
        """
        def generate_key_in_thread():
            """
            Called from thread since key generation is CPU intensive.
            """
            try:
                key = Key.generate(key_type=type, key_size=size)
            except (KeyCertException, struct.error) as error:
                self._fail(request, 'Fail to generate SSH key.', error)

            return self._exportKey(key, password=password)

        return threads.deferToThread(generate_key_in_thread)

    def jsonrpc_pgp_key_generate(
        self, request, name, email, comment,
        type=DEFAULT_KEY_TYPE, size=DEFAULT_KEY_SIZE,
        password=None,
            ):
        """
        Generate a new PGP key of `type` and `size`, then export it in
        armored format.

        If `password` is provided, it will be used to encrypt the
        exported private keys.
        """
        def generate_pgp_key_in_thread():
            """
            Called from thread since key generation is CPU intensive.
            """
            options = Bunch(
                key_type=type,
                key_size=size,
                name=name,
                email=email,
                comment=comment,
                password=password,
                )
            try:
                return generate_pgp_key(options)
            except ServerException as error:  # noqa:cover
                self._fail(request, 'Fail to generate GPG key.', error)

        return threads.deferToThread(generate_pgp_key_in_thread)

    def jsonrpc_generate_csr(
        self, request, common_name, key='',
        size=DEFAULT_KEY_SIZE, password=None,
        alternative_name=None, constraints='', key_usage='',  email=None,
        organization=None, organization_unit=None,
        locality=None, state=None, country=None, sign_algorithm='sha256',
            ):
        """
        Generate a new SSL key and CSR.

        If `password` is provided, it will be used to encrypt the
        exported private keys.
        """
        def generate_csr_in_thread():
            """
            Called from thread since key generation is CPU intensive.
            """
            try:
                options = Bunch(
                    key_size=size,
                    key=key.encode('ascii'),
                    key_password=password,
                    common_name=common_name,
                    email=email,
                    alternative_name=alternative_name,
                    constraints=constraints,
                    key_usage=key_usage,
                    organization=organization,
                    organization_unit=organization_unit,
                    locality=locality,
                    state=state,
                    country=country,
                    sign_algorithm=sign_algorithm,
                    )
                result = generate_csr(options)

                return {
                    'csr': result['csr_pem'].decode('ascii'),
                    'key': result['key_pem'].decode('ascii'),
                    }
            except KeyCertException as error:
                self._fail(request, 'Fail to generate CSR.', error)

        return threads.deferToThread(generate_csr_in_thread)

    def jsonrpc_generate_self_signed_certificate(
        self, request, common_name,
        size=DEFAULT_KEY_SIZE,
        alternative_name=None, constraints='', key_usage='', email=None,
        organization=None, organization_unit=None,
        locality=None, state=None, country=None, sign_algorithm='sha256',
            ):
        """
        Generate a new self signed certificate key.
        """
        def generate_csr_in_thread():
            """
            Called from thread since key generation is CPU intensive.
            """
            try:
                options = Bunch(
                    key_size=size,
                    common_name=common_name,
                    email=email,
                    alternative_name=alternative_name,
                    constraints=constraints,
                    key_usage=key_usage,
                    organization=organization,
                    organization_unit=organization_unit,
                    locality=locality,
                    state=state,
                    country=country,
                    sign_algorithm=sign_algorithm,
                    )
                result = generate_ssl_self_signed_certificate(options)
                return {
                    'cert': result[0],
                    'key': result[1],
                    }
            except KeyCertException as error:
                self._fail(
                    request,
                    'Fail to generate self-signed certificate.',
                    error,
                    )

        return threads.deferToThread(generate_csr_in_thread)

    def jsonrpc_ssl_convert_pkcs12(self, request, uuid, import_password=None):
        """
        Convert the upload P12 stored as file with `uuid` and export its
        content as PEM.

        `import_password` is used to decrypt the file.
        """
        file_content = self._getUploadedFileData(request, uuid)

        try:
            cert, key, chain = load_PKCS12(
                data=file_content,
                password=import_password,
                source='imported data'
                )
            result = {
                'certificate': dump_certificate(cert),
                'key': dump_key(key),
                'chain': '\n'.join(dump_certificate(c) for c in chain),
                }
            return result
        except ServerException as error:
            self._fail(request, 'Fail to convert PKCS#12.', error)

    def jsonrpc_send_email(self, request, to, subject, body, attachments=None):
        """
        Send an email using the default email client.

        Attachment is a dict mapping names to BASE64 encoded data.
        """

        def parse_data_uri(name, data):
            """
            Parse the data in data_uri format and return the file name with
            extension extracted from the data-uri.

            data-uri format: data:image/gif;base64,BASE64_DATA
            """
            extension = ''
            parts = data.split(',')
            encoded = parts[-1]
            mime_marker = parts[0][:5]
            if mime_marker == b'data:':
                # We have a data-uri.
                if 'image/gif' in parts[0]:
                    extension = '.gif'
                elif 'image/png' in parts[0]:
                    extension = '.png'
                else:
                    extension = '.bin'

            value = io.BytesIO(base64.b64decode(encoded))
            return name + extension, value

        byte_attachments = {}
        if attachments is None:
            attachments = {}

        for name, value in attachments.items():
            name, value = parse_data_uri(name, value)
            byte_attachments[name] = value

        client = self.factory.root.resources.getEmailClient()
        return client.send(
            to=to,
            subject=subject,
            body=body,
            attachments=byte_attachments,
            )

    def _failWithError(self, request, error):
        """
        Log error and inform client.
        """
        avatar = self._getAvatar(request)
        emitFromError(error, avatar=avatar, emitter=self.factory.emitEvent)
        raise ServerJSONRPCError(error.id, error.data)

    def _fail(self, request, details, error):
        """
        Generate a generic error and inform client.
        """
        tb = ''
        details = details

        if error:
            details = details + ' ' + force_unicode(error)

        if error and error.__traceback__:
            tb = ''.join(traceback.format_tb(error.__traceback__))

        server_error = ServerError('50023', data={
            'details': details,
            'tb': tb,
            })
        self._failWithError(request, server_error)
