# Copyright (c) 2020 Adi Roiban.
# See LICENSE for details.
"""
Command line shell for process administration.

This is the CLI version of the local manager.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
import sys
from cmd import Cmd
from datetime import datetime
from getpass import getpass

import six
from six.moves import input
from unidecode import unidecode
import requests

from chevah_compat import local_filesystem, process_capabilities
from chevah.server import force_unicode, json
from chevah.server.commons.command import UnicodeArgumentParser
from chevah.server.commons.constant import (
    AUTHENTICATION_NAMES,
    EVENT_HANDLER_NAMES,
    LOCATION_NAMES,
    PRODUCT_NAME,
    PRODUCT_VERSION,
    SERVICE_NAMES,
    )
from chevah.server.commons.exception import ServerException


def _is_ignored(filter, component):
    """
    Return `True` if component should be ignored.
    """
    if not filter:
        return False

    for key, value in filter.items():
        if component[key] == value:
            return False

    return True


class AdminShell(Cmd, object):
    """
    Command handling for the admin shell.

    Commands are implemented using do_COMMAND_NAME methods.
    On errors command can raise ServerException.
    Command can return a deferred and the deferred result is sent to the
    shell output.
    """
    use_rawinput = True
    prompt = '> '
    intro = '%s (%s) command line administration shell' % (
        PRODUCT_NAME, PRODUCT_VERSION)

    doc_leader = (
        'For details about a command, type `help <command>`\n'
        'Command names are case insensitive and can be abbreviated \n'
        'as long as there are no conflicts.\n'
        )
    doc_header = 'Available commands:'

    def __init__(self):
        super(AdminShell, self).__init__()
        script_name = 'admin-shell'
        self._stdin = sys.stdin
        self._stderr = sys.stderr
        self._rpc_id = 0
        self._session_id = None
        self._url = None
        self._http = requests.Session()
        self._parser = UnicodeArgumentParser(prog=script_name)
        self._populateParser(self._parser)
        self._verbose = True

    def emptyline(self):
        """
        Just show the help message on empty lines.
        """
        return self.do_help('')

    def _populateParser(self, parser):
        """
        Called to add options to the parser.
        """
        parser.add_argument(
            'command', nargs='*',
            action='store',
            metavar="COMMAND",
            default='',
            help=(
                'Command to execute in non interactive mode. '
                'Ex: show failures'
                ),
            )
        parser.add_argument(
            '-a', '--url',
            action='store', dest='url',
            metavar='https://address(:port)',
            default='https://localhost:10020',
            help='URL to the local manager service.',
            )
        parser.add_argument(
            '-u', '--name',
            action='store', dest='name',
            metavar="NAME", default='',
            help='Administrator name used by the authentication step.',
            )
        parser.add_argument(
            '-p', '--password',
            action='store', dest='password',
            metavar="PASS|path|-", default='',
            help=(
                'Password used by the authentication step. '
                'If no value is provided on the command line, '
                'it will be asked at runtime in a secure way. '
                'You can specify the actual password or the path to a file '
                'containing the password. '
                'Use - value to read the password from standard input.'
                ),
            )
        parser.add_argument(
            '-c', '--cacert',
            action='store', dest='ca_cert',
            metavar='PATH',
            default='configuration/self_signed_certificate.pem',
            help='The CA certificates file used to verify the remote server.',
            )
        parser.add_argument(
            '-k', '--insecure',
            action='store_true', dest='insecure',
            default=False,
            help=(
                'This options allows the admin-shell command to proceed '
                'without validating the identity of the remote server.'
                ),
            )
        parser.add_argument(
            '-q', '--quiet',
            action='store_true', dest='quiet',
            default=False,
            help=(
                'Suppress any additional command information.'
                ),
            )
        parser.add_argument(
            '-t', '--totp',
            action='store', dest='mfa_password',
            metavar="TOTP", default='',
            help='Two-factor authentication code.',
            )

    def _parseArguments(self, args=None):
        """
        Parse command line arguments.

        Return a namespace with parsed options.
        """
        if args is None:
            args = sys.argv[1:]

        # Convert all arguments to Unicode as this is an entry point.
        unicode_args = []
        for arg in args:
            if not isinstance(arg, six.text_type):
                arg = arg.decode('utf-8')
            unicode_args.append(arg)

        arguments = self._parser.parse_args(unicode_args)
        return arguments

    def _sendOutput(self, message):
        """
        Send output to the user.
        """
        if not message.endswith('\n'):
            message += '\n'

        if six.PY2:
            message = message.encode('utf-8')

        self.stdout.write(message)

    def _sendInfo(self, message):
        """
        Send a message which might be not display in non verbose mode.
        """
        if not self._verbose:
            return
        return self._sendOutput(message)

    def _sendError(self, message):
        """
        Send message as error.
        """
        if not message.endswith('\n'):
            message += '\n'
        if six.PY2:
            message = message.encode('utf-8')
        self._stderr.write(message)

    def _rpc(self, command, **kwargs):
        """
        Make a JSON RPC request and return the result.

        It will exit on failure.
        """
        self._rpc_id += 1
        data = {
            'jsonrpc': '2.0',
            'id': self._rpc_id,
            'method': command,
            'params': kwargs,
            }

        if self._session_id:
            data['session_id'] = self._session_id

        headers = {
            'content-type': 'application/json;charset=utf-8',
            'user-agent': '%s/admin-shell/%s' % (
                PRODUCT_NAME, PRODUCT_VERSION),
            }

        verify = self._arguments.ca_cert
        if self._arguments.insecure:
            verify = False

        try:
            response = self._http.post(
                self._url,
                json=data,
                headers=headers,
                verify=verify,
                )
        except requests.exceptions.RequestException as error:
            message = str(error)
            raise ServerException(message=message)

        if response.status_code != 200:
            message = 'Got an unknown response from the server: %s %s' % (
                response.status_code, response.reason)
            raise ServerException(message=message)

        try:
            result = response.json()
        except ValueError:
            raise ServerException('Invalid JSON response.')

        error = result.get('error', {})
        if error:
            error_id = error.get('code', None)
            if not error_id:
                error_id = error.get('id', 'UnknownID')
            message = error.get('message', 'No error details.')

            raise ServerException(
                '%s: %s' % (error_id, message))

        try:
            return result['result']
        except KeyError:
            raise ServerException(
                'Invalid JSON-RPC response. No "result" member.')

    def start(self, args=None):
        """
        Called when shell should be started.

        `args` is a list of arguments used to start the shell.
        """
        try:
            self._start(args)
        except Exception as error:
            self._sendError('ERROR: %s' % (force_unicode(error),))
            self._exit(1)

    def _start(self, args):
        """
        Start the shell and loop for commands.
        """
        arguments = self._parseArguments(args)
        self._arguments = arguments
        self._verbose = not arguments.quiet

        if not self._verbose:
            self.intro = ''

        self._url = '%s/json' % (arguments.url,)

        self._sendInfo(
            'Connecting to %s ...' % (self._url,))
        try:
            self._rpc('is_valid_session')
        except ServerException as error:
            self._sendError('Connection failed.')
            self._sendError(error.message)
            self._exit(1)
            return  # noqa:cover

        if not arguments.name:
            arguments.name = self._getRawInput('Administrator name: ')

        if arguments.password == '-':
            # Password read with echo from stdin.
            arguments.password = self._stdin.readline().strip()
        elif not arguments.password:
            # Password read without echo. Safe.
            arguments.password = self._getPasswordInput('Password: ')
        else:
            # See if we have the password from command line, or is a file.
            try:
                encoded_path = local_filesystem.getEncodedPath(
                    arguments.password)
                with open(encoded_path, 'rb') as stream:
                    arguments.password = stream.read().strip().decode('utf-8')
            except EnvironmentError:
                """
                We might not have a path.
                """

        password = arguments.password + arguments.mfa_password

        result = self._rpc('login', username=arguments.name, password=password)
        self._session_id = result['session_id']
        self._sendInfo('Connected.')

        if self._arguments.command:
            command = ' '.join(self._arguments.command)
            self.cmdqueue = [command, 'exit']

        self.cmdloop()
        self._exit(0)

    def stop(self):
        """
        Called to stop the shell.
        """
        self._exit(0)

    def _exit(self, code=0, message=''):
        """
        Called when shell should end.
        """
        if message:
            if six.PY3:
                message = message + '\n'
            elif process_capabilities.os_family != 'posix':
                # On Windows we get IOError if we try to send Unicode
                # characters even if they are UTF-8 encoded so we transliterate
                # to them plain ASCII.
                message = unidecode(message) + '\n'
            elif isinstance(message, six.text_type):
                # Make sure value are encoded before putting them on the wire.
                message = message.encode('utf-8') + b'\n'

            sys.stderr.write(message)

        sys.exit(code)

    def _getRawInput(self, prompt):
        """
        Test helper.
        """
        if six.PY2 and isinstance(prompt, six.text_type):
            prompt = prompt.encode('utf-8')

        return input(prompt)

    def _getPasswordInput(self, prompt):
        """
        Test helper.
        """
        if six.PY2 and isinstance(prompt, six.text_type):
            prompt = prompt.encode('utf-8')

        return getpass(prompt)

    def _getTime(self, timestamp):
        """
        Return human readable timestamp.
        """
        if not timestamp:
            return None
        return datetime.utcfromtimestamp(timestamp).strftime(
            '%Y-%m-%d %H:%M:%S')

    def onecmd(self, line):
        """
        Patch stdlib implementation to print the error and continue the loop.
        """
        try:
            super(AdminShell, self).onecmd(line)
        except ServerException as error:
            self._sendError(error.message)

    def get_names(self):
        """
        Patch python version to return all names from the instance,
        and not from the class.

        It also skip some aliases.
        """
        result = dir(self)
        result.remove('do_EOF')
        result.remove('do_quit')
        return result

    def do_exit(self, arguments):
        """
        Exit the shell.
        """
        try:
            self._rpc('logout')
        finally:
            self._exit(0)

    # Exit aliases
    do_quit = do_EOF = do_exit

    def do_show(self, arguments):
        """
        Show the status of components.

        show server -> show the general status for the process.
        show service [NAME|UUID]-> show the status of the services
            or a specific service.
        show transfer [NAME|UUID] -> show the status of the transfers
            or a specific transfer.
        show location [NAME|UUID] -> show the status of the locations
            or a specific location.
        show handler [NAME|UUID] -> show the status of the event handlers
            or a specific handler.
        show resource [NAME|UUID] -> show the status of the resources
            or a specific resource.
        show authentication [NAME|UUID] -> show the status of the authentication
            methods or a specific authentication method.
        show failures -> show all components in a failed state.
        """
        if not arguments:
            return self.do_help('show')

        arguments = [a.strip() for a in arguments.split(' ', 1) if a.strip()]

        target = arguments[0]
        try:
            filter = {
                'uuid': arguments[1],
                'name': arguments[1]
                }
        except IndexError:
            filter = None

        target = '_show_' + arguments[0].strip().lower()
        method = getattr(self, target, None)
        if not method:
            self._sendError('Unknown target.')
            return self.do_help('show')

        # FIXME:6318:
        # Instead of getting the whole tree.
        # It would be best to get only the details for the requested path.
        tree = self._rpc('get_index')
        return method(tree, filter)

    def _show_server(self, tree, filter=None):
        """
        Show the server status.
        """
        self._sendOutput('server_name: %s' % (tree['status']['name'],))
        self._sendOutput('server_uuid: %s' % (tree['status']['uuid'],))

        for key, value in tree['status'].items():
            if key == 'startup_configuration':
                continue

            if key == 'start_time':
                value = self._getTime(value)

            self._sendOutput('%s = %s' % (key, value))

    def _show_service(self, tree, filter):
        """
        Show the services status.
        """
        return self._showProperties('services', tree, filter)

    def _show_authentication(self, tree, filter):
        """
        Show the authentication methods status.
        """
        return self._showProperties('authentications', tree, filter)

    def _show_transfer(self, tree, filter):
        """
        Show the transfers status.
        """
        return self._showProperties('transfers', tree, filter)

    def _show_handler(self, tree, filter):
        """
        Show the event handlers status.
        """
        return self._showProperties('event_handlers', tree, filter)

    def _show_location(self, tree, filter):
        """
        Show the location status.
        """
        return self._showProperties('locations', tree, filter)

    def _show_resource(self, tree, filter):
        """
        Show the resources status.
        """
        return self._showProperties('resources', tree, filter)

    # Backward compatible names
    _show_services = _show_service
    _show_authentications = _show_authentication
    _show_transfers = _show_transfer
    _show_handlers = _show_handler
    _show_locations = _show_location
    _show_resources = _show_resource

    def _show_failures(self, tree, filter):
        """
        Show the failed components.
        """
        filter = {'failed': True}
        for family in self._TYPE_TO_FAMILY.values():
            self._showProperties(family, tree, filter)

    def _showProperties(self, family, result, filter=None):
        """
        Show the status for all the components from a family.
        """
        for uuid, value in result[family].items():
            if _is_ignored(filter, value):
                continue
            self._showProperty(value)

    def _showProperty(self, component):
        """
        Show the details for a single component/configuration.
        """
        self._sendOutput('\n' + '-' * 50 + '\n')
        for key in sorted(component.keys()):
            value = component[key]
            if key in ['startup_configuration']:
                continue

            if key in ['start_time', 'stop_time']:
                value = self._getTime(value)

            self._sendOutput('%s = %s' % (key, json.dumps(value)))

    _TYPE_TO_FAMILY = {
        'service': 'services',
        'authentication': 'authentications',
        'transfer': 'transfers',
        'location': 'locations',
        'resource': 'resources',
        'handler': 'event_handlers',
        }

    def do_start(self, arguments):
        """
        Start a component.

        start service UUID -> start service with UUID.
        start transfer UUID -> start transfer with UUID.
        start location UUID -> start location with UUID.
        start handler UUID -> start event handler with UUID.
        start resource UUID -> start resource with UUID.
        start authentication UUID -> start authentication with UUID.
        """
        self._actOnRunnable('start', arguments)

    def do_stop(self, arguments):
        """
        Stop a component.

        stop service UUID -> stop service with UUID.
        stop transfer UUID -> stop transfer with UUID.
        stop location UUID -> stop location with UUID.
        stop handler UUID -> stop event handler with UUID.
        stop resource UUID -> stop resource with UUID.
        stop authentication UUID -> stop authentication with UUID.
        """
        self._actOnRunnable('stop', arguments)

    def do_restart(self, arguments):
        """
        Restart a component.

        restart service UUID -> restart service with UUID.
        restart transfer UUID -> restart transfer with UUID.
        restart location UUID -> restart location with UUID.
        restart handler UUID -> restart event handler with UUID.
        restart resource UUID -> restart resource with UUID.
        restart authentication UUID -> restart authentication with UUID.
        """
        self._actOnRunnable('restart', arguments)

    def _actOnRunnable(self, action, arguments):
        """
        Make an operation on a component.
        """
        if not arguments:
            return self.do_help(action)

        arguments = [a.strip() for a in arguments.strip().split(' ')]

        if len(arguments) != 2:
            self._sendError(
                'Too few arguments for {} command.'.format(action))
            return self.do_help(action)

        family = self._TYPE_TO_FAMILY.get(arguments[0].lower(), None)
        uuid = arguments[1]
        if not family:
            self._sendError(
                'Unknown component family "{}".'.format(arguments[0]))
            return self.do_help(action)

        result = self._rpc('runnable_' + action, family=family, uuid=uuid)

        self._sendInfo('Operation was successfully.')
        self._sendInfo('Below is the new status.')
        self._showProperty(result)

    def do_add(self, arguments):
        """
        Add a new component with minimal configuration.

        add account NAME
        add group NAME
        add administrator NAME
        add role NAME
        add service TYPE NAME
        add transfer NAME
        add location TYPE NAME
        add handler TYPE NAME
        add authentication TYPE NAME
        """
        return self._doSubCommand('add', arguments)

    def _doSubCommand(self, name, line):
        """
        Handle a subcommand with name.
        """
        if not line:
            return self.do_help(name)

        command, rest = _parse_command(line)

        target = '_{}_{}'.format(name, command)
        method = getattr(self, target, None)
        if not method:
            self._sendError('Unknown target for {} command.'.format(name))
            return self.do_help(name)

        return method(rest)

    def _add_transfer(self, line):
        """
        Add a new transfer.
        """
        self._addProperty(family='transfers', name=line)

    def _add_service(self, line):
        """
        Add a new service.
        """
        return self._addPropertyWithType(
            command_name='service',
            line=line,
            family='services',
            names=SERVICE_NAMES,
            )

    def _add_authentication(self, line):
        """
        Add a new authentication.
        """
        return self._addPropertyWithType(
            command_name='authentication',
            line=line,
            family='authentications',
            names=AUTHENTICATION_NAMES,
            )

    def _add_handler(self, line):
        """
        Add a new event handler.
        """
        return self._addPropertyWithType(
            command_name='handler',
            line=line,
            family='event_handlers',
            names=EVENT_HANDLER_NAMES,
            )

    def _add_location(self, line):
        """
        Add a new location.
        """
        return self._addPropertyWithType(
            command_name='location',
            line=line,
            family='locations',
            names=LOCATION_NAMES,
            )

    def _add_account(self, line):
        """
        Add a new account.
        """
        self._addProperty(family='accounts', name=line)

    def _add_group(self, line):
        """
        Add a new event group.
        """
        self._addProperty(family='groups', name=line)

    def _add_administrator(self, line):
        """
        Add a new administrator.
        """
        self._addProperty(family='administrators', name=line)

    def _add_role(self, line):
        """
        Add a new event role.
        """
        self._addProperty(family='roles', name=line)

    def _addPropertyWithType(self, command_name, line, family, names):
        """
        Add a new component with types
        """
        if not line:
            self._sendError('Usage: {} TYPE NAME'.format(command_name))
            return self._showPropertyTypes(family, names)

        service_type, name = _parse_command(line)
        self._addProperty(
            family=family,
            name=name,
            component_type=service_type,
            names=names,
            )

    def _addProperty(self, family, name, component_type=None, names=None):
        """
        Add a generic component and show the error or the UUID of the
        created component for `family`.
        """
        if not name:
            self._sendError('A name is required for new {}.'.format(family))
            return

        value = {
            'name': name,
            'enabled': False,
            }

        if component_type:
            if component_type not in names:
                return self._errorUnknowType(component_type, family, names)

            value['type'] = component_type

        change = {
            'operation': 'create',
            'path': family + '/',
            'value': value,
            }
        response = self._rpc('apply', changes={'0': change})
        result = response['results']['0']
        if result['error']:
            # FIXME:6297:
            # Display error based on events DB.
            self._sendError(
                'Failed to add {}: {}'.format(family, result['error']))
            return
        uuid = result['result']
        self._sendOutput(
            'New {} created with UUID: {}'.format(family, uuid))

    def _errorUnknowType(self, component_type, family, names):
        """
        Handle the error when an unknow type is requested for `family`.

        `names` is a dict with available types and their human representation.
        """
        self._sendError(
            'Unknown {} type: {}'.format(family, component_type))
        return self._showPropertyTypes(family, names)

    def _showPropertyTypes(self, family, names):
        """
        Return the list of available service types.
        """
        self._sendError(
            'Available {} types:'.format(family))
        for key in sorted(names.keys()):
            self._sendError(
                '  {} - {}'.format(key, names[key]))

    def do_delete(self, arguments):
        """
        Delete an existing component by UUID.

        delete account UUID
        delete group UUID
        delete administrator UUID
        delete role UUID
        delete service UUID
        delete transfer UUID
        delete location UUID
        delete handler UUID
        delete authentication UUID
        """
        return self._doSubCommand('delete', arguments)

    def _delete_transfer(self, line):
        """
        Delete an existing transfer
        """
        return self._deleteComponent(
            kind='transfer',
            family='transfers',
            target=line,
            )

    def _delete_location(self, line):
        """
        Delete an existing location.
        """
        return self._deleteComponent(
            kind='location',
            family='locations',
            target=line,
            )

    def _delete_service(self, line):
        """
        Delete an existing service.
        """
        return self._deleteComponent(
            kind='service',
            family='services',
            target=line,
            )

    def _delete_authentication(self, line):
        """
        Delete an existing authentication.
        """
        return self._deleteComponent(
            kind='authentication',
            family='authentications',
            target=line,
            )

    def _delete_handler(self, line):
        """
        Delete an existing event handler.
        """
        return self._deleteComponent(
            kind='handler',
            family='event_handlers',
            target=line,
            )

    def _delete_account(self, line):
        """
        Delete an existing account.
        """
        return self._deleteComponent(
            kind='account',
            family='accounts',
            target=line,
            )

    def _delete_group(self, line):
        """
        Delete an existing group.
        """
        return self._deleteComponent(
            kind='group',
            family='groups',
            target=line,
            )

    def _delete_administrator(self, line):
        """
        Delete an existing admin.
        """
        return self._deleteComponent(
            kind='administrator',
            family='administrators',
            target=line,
            )

    def _delete_role(self, line):
        """
        Delete an existing role.
        """
        return self._deleteComponent(
            kind='role',
            family='roles',
            target=line,
            )

    def _deleteComponent(self, kind, family, target):
        """
        Delete a component with `target` UUID  for `family`.
        """
        # FIXME:6318:
        # It would be nice to also delete by name.
        if not target:
            self._sendError(
                'An UUID for {} is required to delete.'.format(kind))
            return

        path = '{}/{}'.format(family, target)

        change = {
            'operation': 'delete',
            'path': path,
            'value': '',
            }
        response = self._rpc('apply', changes={'0': change})
        result = response['results']['0']
        if result['error']:
            # FIXME:6297:
            # Display error based on events DB.
            self._sendError(
                'Failed to delete {}: {}'.format(kind, result['error']))
            return
        uuid = result['result']
        self._sendOutput(
            '{} with UUID {} was removed.'.format(kind, uuid))

    def do_configure(self, arguments):
        """
        Show or update the configuration of an existing component.

        When OPTION_NAME is not provided, it will show the complete current
        configuration for component.

        When VALUE is not provided, it will show the current configuration
        for OPTION_NAME.

        To update the configuration, the VALUE must be defined in JSON
        format.

        configure server [[OPTION_NAME [VALUE]]]
        configure account UUID [[OPTION_NAME [VALUE]]]
        configure group UUID [[OPTION_NAME [VALUE]]]
        configure administrator UUID [[OPTION_NAME [VALUE]]]
        configure role UUID [[OPTION_NAME [VALUE]]]
        configure service UUID [[OPTION_NAME [VALUE]]]
        configure transfer UUID [[OPTION_NAME [VALUE]]]
        configure location UUID [[OPTION_NAME [VALUE]]]
        configure handler UUID [[OPTION_NAME [VALUE]]]
        configure authentication UUID [[OPTION_NAME [VALUE]]]
        configure resource UUID [[OPTION_NAME [VALUE]]]
        """
        return self._doSubCommand('configure', arguments)

    def _configure_server(self, line):
        """
        Show or update configuration for server.
        """
        if not line:
            return self._showConfiguration('server', uuid=None, key=None)

        key, value = _parse_command(line)
        if not value:
            self._showConfiguration('server', uuid=None, key=key)
            return

        key, value = _parse_command(line)
        self._updateConfiguration(
            'server', uuid=None, key=key, value=value)

    def _configure_account(self, line):
        """
        Show or update configuration for an account.
        """
        return self._configureProperty(
            family='accounts',
            line=line,
            )

    def _configure_group(self, line):
        """
        Show or update configuration for a group.
        """
        return self._configureProperty(
            family='groups',
            line=line,
            )

    def _configure_administrator(self, line):
        """
        Show or update configuration for an administrator.
        """
        return self._configureProperty(
            family='administrators',
            line=line,
            )

    def _configure_role(self, line):
        """
        Show or update configuration for a role.
        """
        return self._configureProperty(
            family='roles',
            line=line,
            )

    def _configure_service(self, line):
        """
        Show or update configuration for a service.
        """
        return self._configureProperty(
            family='services',
            line=line,
            )

    def _configure_authentication(self, line):
        """
        Show or update configuration for a authentication.
        """
        return self._configureProperty(
            family='authentications',
            line=line,
            )

    def _configure_handler(self, line):
        """
        Show or update configuration for a handler.
        """
        return self._configureProperty(
            family='event_handlers',
            line=line,
            )

    def _configure_transfer(self, line):
        """
        Show or update configuration for a service.
        """
        return self._configureProperty(
            family='transfers',
            line=line,
            )

    def _configure_location(self, line):
        """
        Show or update configuration for a location.
        """
        return self._configureProperty(
            family='locations',
            line=line,
            )

    def _configure_resource(self, line):
        """
        Show or update configuration for a resource.
        """
        return self._configureProperty(
            family='resources',
            line=line,
            )

    def _configureProperty(self, family, line):
        """
        Show configuration or update configuration for a property.
        """
        if not line:
            self._sendError(
                'UUID is required to show/update configuration for {}.'.format(
                    family))
            return

        uuid, rest = _parse_command(line, lowercase=False)
        if not rest:
            self._showConfiguration(family, uuid=uuid, key=None)
            return

        key, value = _parse_command(rest)
        if not value:
            self._showConfiguration(family, uuid=uuid, key=key)
            return

        self._updateConfiguration(
            family, uuid=uuid, key=key, value=value)

    def _showConfiguration(self, family, uuid, key):
        """
        Show configuration for component from family with uuid.
        """
        # FIXME:6318:
        # Instead of getting the whole configuration tree
        # we can make a request just for family + uuid.
        tree = self._rpc('get_index')
        component_configuration = tree['configuration'][family]
        if uuid is not None:
            try:
                component_configuration = component_configuration[uuid]
            except KeyError:
                self._sendError(
                    '{} UUID not found for {}'.format(uuid, family))
                return

        if key:
            try:
                value = json.dumps(component_configuration[key])
            except KeyError:
                self._sendError(
                    'Unknown configuration "{}" for {} with UUID "{}".'.format(
                        key, family, uuid))
                return
            self._sendOutput('{} = {}'.format(key, value))
            return

        return self._showProperty(component_configuration)

    def _updateConfiguration(self, family, uuid, key, value):
        """
        Update a single configuration option for a component.
        """
        parent = family
        if uuid is not None:
            parent = '{}/{}'.format(family, uuid)
        path = '{}/{}'.format(parent, key)

        try:
            value = json.loads(value)
        except Exception:
            self._sendError(
                'Invalid JSON format requested to update "{}" for "{}".'.format(
                    key, path))
            return

        change = {
            'operation': 'update',
            'path': path,
            'value': value,
            }
        response = self._rpc('apply', changes={'0': change})
        result = response['results']['0']
        if result['error']:
            # FIXME:6297:
            # Display error based on events DB.
            self._sendError(
                'Failed to update {} with UUID {}: {}'.format(
                    family, uuid, result['error']))
            return
        self._sendOutput('Configuration change was saved.')


def _parse_command(line, lowercase=True):
    """
    Return a tuple of command_name, argument_line

    If no arguments are provided, arguments_line is empty.
    """
    line = line.strip()
    arguments = [a.strip() for a in line.split(' ', 1) if a.strip()]
    command_name = arguments[0]
    if lowercase:
        command_name = command_name.lower()

    if len(arguments) < 2:
        return command_name, ''

    return command_name, arguments[1].strip()
