# Copyright (c) 2020 Adi Roiban.
# See LICENSE for details.
"""
Helpers for GPG operations.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

import uuid
import six

from chevah.server import force_unicode
from chevah.server.commons.exception import ServerException
from chevah.server.openpgp.wrapper import ChevahGPG


_DSA_KEY_SIZES = [1024, 2048, 3072]
_RSA_KEY_SIZES = [1024, 2048, 3072, 4096, 6144, 7168, 8192]


def generate_pgp_key(options):
    """
    Generate a new PGP key and return a dict with the armored private and
    public key.

    Raise ServerException on errors.
    """
    key_type = options.key_type.upper()

    if key_type not in ['RSA', 'DSA']:
        raise ServerException('Unknown key type: "%s".' % (key_type,))

    if key_type == 'DSA' and options.key_size not in _DSA_KEY_SIZES:
        raise ServerException('Unsupported DSA key size. Use any of %s.' % (
            ', '.join([str(k) for k in _DSA_KEY_SIZES])))

    if key_type == 'RSA' and options.key_size not in _RSA_KEY_SIZES:
        raise ServerException('Unsupported RSA key size. Use any of %s.' % (
            ', '.join([str(k) for k in _RSA_KEY_SIZES])))

    if not (options.name or options.email):
        raise ServerException(
            'You need either a name or an email for the new key.')

    gpg_commands = """
Key-Type: %s
Key-Length: %s
Expire-Date: 0
""" % (key_type, options.key_size)

    if options.name:
        gpg_commands += 'Name-Real: %s\n' % (options.name)

    if options.email:
        gpg_commands += 'Name-Email: %s\n' % (options.email)

    if options.comment:
        gpg_commands += 'Name-Comment: %s\n' % (options.comment)

    if options.password:
        gpg_commands += 'Passphrase: %s\n' % (options.password)

    if key_type == 'DSA':
        gpg_commands += (
            'Subkey-Type: ELG-E\n'
            'Subkey-Length: %s\n'
            ) % (options.key_size,)

    gpg_commands += '%commit\n'

    if six.PY2:
        gpg_commands = gpg_commands.encode('utf-8')

    gpg = None
    try:
        gpg = ChevahGPG(uuid=six.text_type(uuid.uuid4()))
        key = gpg.gen_key(gpg_commands)
        if not key:
            details = force_unicode(key.stderr.splitlines()[-1])
            raise ServerException(
                'Failed to generate PGP key. %s' % (details,))

        private = gpg.export_keys(key.fingerprint, True)
        public = gpg.export_keys(key.fingerprint)
    finally:
        if gpg:
            gpg.stop()

    return {
        'private': private,
        'public': public,
        }
