# Copyright (c) 2019 Adi Roiban.
# See LICENSE for details.
"""
Wrapper around the upstream GNU GPG binary.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
from __future__ import print_function

import os
import platform
import subprocess  # noqa:bandit
import threading
import six

import gnupg

from chevah import chevah_gpg
from chevah_compat import local_filesystem, process_capabilities
from chevah.server.commons.exception import ServerException


# Patch upstream so that we have data.
gnupg.SearchKeys.data = b''


def _write_passphrase(stream, passphrase, encoding=None):
    """
    Patch upstream to not try re-encoding.
    """
    stream.write(passphrase + b'\n')


gnupg._write_passphrase = _write_passphrase


def _get_gnupg_binary():
    """
    Return the OS path to the GPG binary.
    """
    os_name = process_capabilities.os_name
    gpg_executable = ''
    if os_name == 'windows':
        gpg_executable = 'gpg1.exe'
    elif os_name == 'linux':
        arch = platform.processor().lower()
        if not arch:
            arch = platform.machine().lower()
        # FIXME:4226:
        # See how to differentiate between GLIBC and MUSL Linux.
        if arch not in ['x86_64', 'aarch64']:
            raise ServerException(
                'Only Linux on 64bit is supported by the OpenPGP handler.'
                '"%s" detected.' % (arch,))
        if arch == 'aarch64':
            gpg_executable = 'gpg1-linux-arm64'
        else:
            gpg_executable = 'gpg1-linux-x64'
    elif os_name == 'openbsd':
        gpg_executable = 'gpg1-openbsd'
    elif os_name == 'osx':
        gpg_executable = 'gpg1-macos-arm64'
    else:
        raise ServerException('OS not supported by OpenPGP handler.')

    gpg_module = chevah_gpg.MODULE_PATH
    if six.PY2 and os_name != 'windows':
        gpg_module = gpg_module.decode('utf-8')

    return local_filesystem.getEncodedPath(os.path.join(
        gpg_module, gpg_executable))


class ChevahGPG(gnupg.GPG):
    """
    Patching for upstream GPG.
    """

    _CHUNK_SIZE = 16000
    encoding = 'ascii'

    def __init__(self, uuid, working_segments=None):

        if not working_segments:
            working_segments = local_filesystem.installation_segments + [
                'configuration', 'openpgp']

        self._gnupg_home_segments = working_segments + [uuid]
        gnupg_home = local_filesystem.getEncodedPath(
            local_filesystem.getRealPathFromSegments(
                self._gnupg_home_segments))

        self._cleanGPGHome()
        super(ChevahGPG, self).__init__(
            gnupghome=gnupg_home,
            gpgbinary=_get_gnupg_binary(),
            )

    def stop(self):
        """
        Stop handling any other commands.
        """
        self._cleanGPGHome()

    def _cleanGPGHome(self):
        """
        Remove the GnuPG home folder for this event handler.
        """
        try:
            local_filesystem.deleteFolder(
                self._gnupg_home_segments, recursive=True)
        except EnvironmentError:
            pass

    def is_valid_passphrase(self, passphrase):
        """
        Confirm that the passphrase doesn't contain newline-type characters -
        it is passed in a pipe to gpg, and so not checking could lead to
        spoofing attacks by passing arbitrary text after passphrase and
        newline.
        """
        if not passphrase:
            return True

        return (b'\n' not in passphrase and b'\r' not in passphrase and
                b'\x00' not in passphrase)

    def encrypt_file(
        self, file, recipients, sign=None,
        always_trust=False, passphrase=None,
        armor=True, output=None, symmetric=False, extra_args=None
            ):
        """
        Modified from upstream to output to a file like object.
        """
        if passphrase and not self.is_valid_passphrase(passphrase):
            raise ValueError('Invalid passphrase')
        args = [b'--encrypt']
        if symmetric:
            # can't be False or None - could be True or a cipher algo value
            # such as AES256
            args = [b'--symmetric']
            if symmetric is not True:
                args.extend([b'--cipher-algo', gnupg.no_quote(symmetric)])
            # else use the default, currently CAST5
        else:
            if not recipients:
                raise ValueError(b'No recipients specified with asymmetric '
                                 b'encryption')
            if not gnupg._is_sequence(recipients):
                recipients = (recipients,)
            for recipient in recipients:
                args.extend([b'--recipient', gnupg.no_quote(recipient)])
        if armor:   # create ascii-armored output - False for binary output
            args.append(b'--armor')

        if sign is True:  # pragma: no cover
            args.append(b'--sign')
        elif sign:  # pragma: no cover
            args.extend([b'--sign', b'--default-key', gnupg.no_quote(sign)])
        if always_trust:  # pragma: no cover
            args.append(b'--always-trust')
        if extra_args:
            args.extend(extra_args)
        result = self.result_map['crypt'](self)

        if output:
            result.data = output
        else:
            result.data = b''

        self._handle_io(args, file, result, passphrase=passphrase, binary=True)
        return result

    def decrypt_file(
        self, file, always_trust=False, passphrase=None,
        output=None, extra_args=None,
            ):
        """
        Modified from upstream to output to a file like object.
        """
        if passphrase and not self.is_valid_passphrase(passphrase):
            raise ValueError('Invalid passphrase')
        args = [b"--decrypt"]

        if always_trust:  # pragma: no cover
            args.append(b"--always-trust")
        if extra_args:
            args.extend(extra_args)
        result = self.result_map['crypt'](self)

        if output:
            result.data = output
        else:
            result.data = b''

        self._handle_io(args, file, result, passphrase, binary=True)
        return result

    def _open_subprocess(self, args, passphrase=False):
        """
        open a pipe to a GPG subprocess and return
        the file objects for communicating with it.
        """
        cmd = self.make_args(args, passphrase)
        if self.verbose:
            print(cmd)

        def fully_drop_root():
            """
            Helper to fully drop privileges of the forked sub-process.
            """
            if os.getuid() != 0:
                # We are already not root.
                return

            current_uid = os.geteuid()
            current_gid = os.getegid()

            # Set the group while root.
            os.seteuid(0)
            os.setegid(0)
            os.setgroups([])
            os.setgid(current_gid)

            os.setreuid(current_uid, current_uid)

        if not gnupg.STARTUPINFO:
            # We are on Unix.
            si = None
            preexec_fn = fully_drop_root
        else:
            # We are on Windows.
            preexec_fn = None
            si = gnupg.STARTUPINFO()
            si.dwFlags = gnupg.STARTF_USESHOWWINDOW
            si.wShowWindow = gnupg.SW_HIDE

        result = subprocess.Popen(
            cmd,
            shell=False,  # noqa:bandit
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            preexec_fn=preexec_fn,
            startupinfo=si
            )
        gnupg.logger.debug("%s: %s", result.pid, cmd)
        return result

    def _handle_io(self, args, fileobj, result, passphrase=None, binary=False):
        """
        Handle a basic data call - pass data to GPG, handle the output
        including status information. Garbage In, Garbage Out :)
        """
        p = self._open_subprocess(args, passphrase is not None)
        stdin = p.stdin
        if passphrase:
            _write_passphrase(stdin, passphrase, self.encoding)
        writer = gnupg._threaded_copy_data(fileobj, stdin)
        self._collect_output(p, result, writer, stdin)
        return result

    def _read_response(self, stream, result):
        # Internal method: reads all the stderr output from GPG, taking notice
        # only of lines that begin with the magic [GNUPG:] prefix.
        #
        # Calls methods on the response object for each valid token found,
        # with the arg being the remainder of the status line.
        lines = []
        while True:
            # This is the only patch,
            # in which we decode the output.
            line = stream.readline().decode('utf-8')
            if len(line) == 0:
                break
            lines.append(line)
            line = line.rstrip()
            if self.verbose:  # pragma: no cover
                print(line)
            if line[0:9] == '[GNUPG:] ':
                # Chop off the prefix
                line = line[9:]
                L = line.split(None, 1)
                keyword = L[0]
                if len(L) > 1:
                    value = L[1]
                else:
                    value = ""
                result.handle_status(keyword, value)
        result.stderr = ''.join(lines)

    def _collect_output(self, process, result, writer=None, stdin=None):
        """
        Drain the subprocesses output streams, writing the collected output
        to the result. If a writer thread (writing to the subprocess) is given,
        make sure it's joined before returning. If a stdin stream is given,
        close it before returning.
        """
        stderr = process.stderr
        rr = threading.Thread(
            target=self._read_response, args=(stderr, result))
        rr.setDaemon(True)
        rr.start()

        stdout = process.stdout
        dr = threading.Thread(
            target=self._read_data, args=(stdout, result, self.on_data))
        dr.setDaemon(True)
        dr.start()

        dr.join()
        rr.join()
        if writer is not None:
            writer.join()
        process.wait()
        if stdin is not None:
            try:
                stdin.close()
            except IOError:  # pragma: no cover
                pass
        stderr.close()
        stdout.close()

    def _read_data(self, stream, result, on_data=None):
        """
        Read the contents of the file from GPG's stdout.
        """
        while True:
            data = stream.read(self._CHUNK_SIZE)
            if data == b'':
                break

            try:
                result.data.write(data)
            except AttributeError:
                # Not a stream.
                try:
                    result.data += data
                except AttributeError:
                    # Not yet initialized.
                    result.data = data
