# Copyright (c) 2014 Adi Roiban.
# See LICENSE for details.
"""
Execute external processes.

Most of the code from here  is executed from the child process.

It should import as few things as possible to keep the memory usage low.
"""
from __future__ import absolute_import
from __future__ import unicode_literals
import os
import sys
import subprocess


if os.name == 'nt' or os.getuid() != 0:
    preexec_fn = None
else:
    import pwd

    def preexec_fn():
        """
        Switch context of the process to the effective user, so that the
        following executed process will start under the same user.
        """
        user_id = os.geteuid()
        group_id = os.getegid()
        pwnam = pwd.getpwuid(user_id)
        username_encoded = pwnam.pw_name
        os.seteuid(0)
        os.setegid(0)
        os.initgroups(username_encoded, group_id)
        os.setgid(group_id)
        os.setuid(user_id)

# Timeout in seconds to wait for external command to end.
SPAWN_PROCESS_TIMEOUT = 15
# Total timeout in seconds to wait for a command to stay in queue and be
# processed.
QUEUE_PROCESS_TIMEOUT = 86400


def unicode_error(error):
    """
    Return the unicode representation of the error.

    This is called in the sub-process and should not depend on the main
    server code.
    """
    try:
        value = str(error)
        if not value:
            value = repr(error)
    except (UnicodeDecodeError, UnicodeEncodeError):
        value = error.args[0]

    if isinstance(value, str):
        return value

    # Since we replace on errors, this will never fail.
    return str(value, encoding='utf-8', errors='replace')


def kill_tree(process):
    """
    Kill the whole tree for process, including process.
    """
    # We delay the import to reduce memory size.
    # As long as we don't have a timeout, there is no need to import this.
    try:
        import psutil
    except (ImportError, NotImplementedError, RuntimeError):
        # Placeholders for when we don't have the psutil.
        psutil = None

    # We refresh the process state, as the process might be already dead.
    process.poll()

    if not psutil:
        process.kill()
        return

    if sys.platform.startswith('aix'):
        # FIXME:5238:
        # See how to run psutils as non-root on AIX.
        process.kill()
        return

    children = []
    try:
        parent = psutil.Process(process.pid)
        children = parent.children(recursive=True)
    except psutil.NoSuchProcess:
        pass
    finally:
        # Start with the parent, so that if we have a script,
        # it will not continue.
        process.kill()

    if not children:
        return

    for child in children:
        try:
            child.kill()
        except Exception:
            """
            Child might already end, of we don't have permissions.
            """


def _spawn(command, timeout=SPAWN_PROCESS_TIMEOUT, environment=None):
    """
    Execute `command` killing the process if it does not complete after
    `timeout`.

    Return a tuple of (exit_code, standard_output, error_output)

    This is blocking and executed inside the external spawn process.
    """
    command_name = command[0]
    arguments = command[:]

    # Environment variables exposed for all operating system.
    run_env = {
        'PATH': os.environ.get('PATH', ''),
        'PYTHONUTF8': '1',
        }

    if sys.platform.startswith('win'):
        # Special env variables exposed only on Windows.
        run_env['SystemDrive'] = os.environ.get('SystemDrive', '')
        run_env['SystemRoot'] = os.environ.get('SystemRoot', '')
    else:
        # Special env variables exposed on Unix.
        run_env['PWD'] = os.environ.get('PWD', '')

    if environment is None:
        environment = {}

    run_env.update(environment)

    try:
        process = subprocess.Popen(
            arguments,
            shell=False,
            preexec_fn=preexec_fn,
            env=run_env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            )
    except EnvironmentError as error:
        raise RuntimeError('Failed to start %s. %s' % (
            command_name, unicode_error(error),))

    try:
        output, error = process.communicate(timeout=timeout)
    except subprocess.TimeoutExpired:
        kill_error = ''
        try:
            kill_tree(process)
        except EnvironmentError as error:
            # We might not have the permission to kill the process.
            kill_error = ' Failed to kill the sub-process. %s' % (
                unicode_error(error),)

        raise RuntimeError('Process execution timed out after %ss.%s' % (
            timeout, kill_error))

    exit_code = process.returncode
    return (exit_code, output, error)


def defer_spawn(
    command,
    timeout=SPAWN_PROCESS_TIMEOUT,
    environment=None,
    queue_timeout=QUEUE_PROCESS_TIMEOUT,
        ):
    """
    Execute `command` as list of executable and arguments
    using an external/forked process.

    This is executed in the main process.

    Note that we have 2 timeouts here.

    The execute.timeout for which the deferred waits in the queue.

    The defer_spawn.timeout for which the process can execute and after
    which the spawned process is killed.
    """
    from chevah.slim_server.pool import spawn_pool
    from chevah.server import force_unicode
    from chevah.server.commons.exception import ServerException

    def eb_spawn(failure):
        failure.trap(RuntimeError)
        raise ServerException(failure.value.args[0])

    def cb_spawn(result):
        exit_code, output, error = result
        output = force_unicode(output)
        error = force_unicode(error)
        return (exit_code, output, error)

    # FIXME:2477:
    # Create a queue in the main process used for the tasks waiting to be
    # executed.
    # The queue should have a maximum size.
    # Also, with a queue in the main process, on timeout, we can remove
    # the task from the queue.

    kwds = {
        'command': command,
        'timeout': timeout,
        'environment': environment
        }

    # To troubleshoot the spawn code, you can use the following lines
    # to launch the code in the main process.
    # from twisted.internet import defer
    # deferred = defer.maybeDeferred(_spawn, **kwds)
    deferred = spawn_pool.execute(
        _spawn,
        kwds=kwds,
        # For now there is a hard timeout of one day for which a command
        # can stay in the queue.
        timeout=queue_timeout,
        )

    deferred.addCallback(cb_spawn)
    deferred.addErrback(eb_spawn)
    return deferred
