# Copyright (c) 2019 Adi Roiban.
# See LICENSE for details.
"""
Code executed in a worker from a pool.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from bunch import Bunch
from twisted.internet import defer

from chevah.slim_server.pool import ChevahPool
from chevah.server import json
from chevah.server.extension import load_extension_object


class ChevahWorkers(object):
    """
    A worker pool which can execute tasks in a process pool.

    It is designed to returned deferred for most of its methods.

    This is the code executed in the main process.
    """

    def __init__(self, name):
        self._pool = None
        self._name = name

    @property
    def started(self):
        return self._pool is not None

    def start(self, timeout=30):
        """
        Start the workers.

        Return a deferred which fires with `True` if the pool was started and
        `False` if it was already started.
        """
        if self._pool:
            # Already started.
            return defer.succeed(False)

        # The process pool is only used to handle extension as for now
        # we only support 2 concurrent handling.
        self._pool = ChevahPool(
            processes=2,
            maxtasksperchild=None,
            name=self._name,
            )
        return self._pool.execute(_noop, timeout=timeout)

    def stop(self):
        """
        Return a deferred which fires when the pool is stopped.
        """
        if not self._pool:
            return defer.succeed(None)

        def execute_stop(pool):
            pool.terminate()
            pool.join()

        pool, self._pool = self._pool, None
        return execute_stop(pool)

    def handleEvent(self, entry_point, event_json, configuration):
        """
        Handle the event using an extension a worker.

        `debug` should only be used for development and troubleshooting.
        """
        kwds = {
            'entry_point': entry_point,
            'event_json': event_json,
            'configuration': configuration,
            }

        if False:  # noqa:cover
            # Run the code in the same process to make it easy to use PDB.
            # To be manually updated during development.
            return defer.succeed(_handle_event(**kwds))

        return self._pool.execute(
            _handle_event,
            kwds=kwds,
            timeout=600,
            )


#
# Code executed in the worker sub-process.
#

# Keeps the loaded extensions
_worker_extensions = {}


def _noop():  # noqa:cover
    """
    Does nothing, other than proves that at least one worker was available.
    """
    return True


def _handle_event(entry_point, event_json, configuration):
    """
    Handle the event using an extension a worker.
    """
    extension = _worker_extensions.get(entry_point, None)
    if not extension:
        extension = load_extension_object(entry_point)()
        _worker_extensions[entry_point] = extension

    public_event = Bunch.fromDict(json.loads(event_json))
    return extension.handle(public_event, configuration)
