# Copyright (c) 2014 Adi Roiban.
# See LICENSE for details.
"""
HTTP proxy.
"""
from __future__ import absolute_import, unicode_literals

import re
from six.moves.urllib.parse import urlparse

from twisted.internet import defer
from twisted.internet.endpoints import TCP4ClientEndpoint

from chevah.server.commons.exception import (
    HTTPClientException,
    ServerException,
    )


def proxify_http_connect_agent(agent, proxy, context_factory):
    """
    It transforms an HTTP agent to make requests via an HTTP proxy using the
    CONNECT method as a tunnel.

    Is initialized with the `proxy` configuration passed as URI.
    """

    def get_endpoint_with_http_connect(uri):
        """
        See: twisted.web.client.Agent.
        """
        # Import from parent scope to local as the parent
        # of this function is changed during the patching process.
        ssl_context_factory = context_factory
        return _HTTPConnectClientEndpoint(
            agent._reactor,
            uri=uri,
            proxyHost=proxy_host,
            proxyPort=proxy_port,
            contextFactory=ssl_context_factory,
            )

    url = urlparse(proxy)
    if url.scheme != 'connect':
        raise ServerException('Unsupported proxy type: %s' % (url.scheme,))

    proxy_host = url.hostname

    if not proxy_host:
        raise ServerException('Proxy URL requires a host.')

    proxy_port = url.port
    if not proxy_port:
        proxy_port = 8080

    agent._getEndpoint = get_endpoint_with_http_connect


class _HTTPConnectClientEndpoint(TCP4ClientEndpoint):
    """
    An endpoint that tunnels through HTTP/1.1 proxy to allow HTTPS downloads.

    To accomplish that, this endpoint sends an HTTP CONNECT to the proxy.

    CONNECT is sent for each new connection.
    If a connection is reused in a pool, CONNECT is not re-sent as no new
    connection is made.

    When CONNECT fails the transport is closed to prevent keeping it in the
    pool and being reused without a previous successful CONNECT.

    See: https://tools.ietf.org/html/rfc2817#section-5.2
    """
    _responseMatcher = re.compile(br'HTTP/1\.. 200')

    def __init__(self, reactor, uri, proxyHost, proxyPort, contextFactory):
        super(_HTTPConnectClientEndpoint, self).__init__(
            reactor, proxyHost, proxyPort)

        # Called when the request was made or failed via the HTTP proxy.
        self._tunnelReadyDeferred = defer.Deferred()
        # URI of the original request..
        self._uri = uri
        # Factory used for the tunneled connection.
        self._protocolFactory = None
        # The protocol which should be tunneled.
        self._protocol = None
        self._contextFactory = contextFactory

    def _cbProxyConnect(self, protocol):
        """
        Called when we get a new connection to the proxy.
        """
        # Send the HTTP CONNECT to initiate the tunnel.
        tunnelCommand = (
            'CONNECT %(host)s:%(port)s HTTP/1.1\r\n'
            'User-Agent: MFT Proxy\r\n'
            'Proxy-Connection: keep-alive\r\n'
            'Connection: keep-alive\r\n'
            'Host: %(host)s:%(port)s\r\n'
            '\r\n'
            ) % {
                'host': self._uri.host.decode('ascii'),
                'port': self._uri.port,
                }
        protocol.transport.write(tunnelCommand.encode('ascii'))
        # Intercept the response,
        self._protocolDataReceived = protocol.dataReceived
        protocol.dataReceived = self._gotHTTPConnectResponse
        self._protocol = protocol

        return protocol

    def _gotHTTPConnectResponse(self, data):
        """
        Called when we receive the response for the HTTP CONNECT command.
        """
        # Switch back the original protocol data received handling as we no
        # longer need to intercept the data.
        self._protocol.dataReceived = self._protocolDataReceived
        if self._responseMatcher.match(data):
            # Advertise that proxy is connected, before we start the SSL, in
            # case SSL handshake fails, we can see the remote IP.
            self._protocol.proxyConnectionMade(
                address=self._uri.host.decode('ascii'),
                port=self._uri.port,
                )

            if self._uri.scheme == b'https':
                self._protocol.transport.startTLS(
                    self._contextFactory, self._protocolFactory)

            self._tunnelReadyDeferred.callback(self._protocol)
        else:
            # We close the connection to make sure it is not kept in the
            # pool.
            self._protocol.transport.loseConnection()
            # Inform the client that the proxy was not able to open the
            # connection.
            # We don't parse the whole response.
            response = data[:50]
            path = 'no-path'
            code = 502
            phrase = b'Bad Gateway'
            if response[:6].lower() == b'http/1':
                first_line = response.split(b'\r\n', 1)[0]
                parts = first_line.split(b' ', 2)
                try:
                    code = int(parts[1])
                except (ValueError, IndexError):
                    pass

                try:
                    phrase = parts[2].strip()
                except (ValueError, IndexError):
                    pass

                if not phrase:
                    phrase = b'No reason was provided by the proxy.'

            self._tunnelReadyDeferred.errback(HTTPClientException(
                'Could not open the HTTP CONNECT tunnel.',
                path=path,
                code=code,
                phrase=phrase,
                ))

    def connect(self, protocolFactory):
        """
        Connect to the remote host.
        """
        self._protocolFactory = protocolFactory

        deferred = super(_HTTPConnectClientEndpoint, self).connect(
            protocolFactory)

        deferred.addCallback(self._cbProxyConnect)
        deferred.addErrback(
            lambda failure: self._tunnelReadyDeferred.errback(failure))
        return self._tunnelReadyDeferred
