# Copyright (c) 2019 Adi Roiban.
# See LICENSE for details.
"""
Acts as an analytics engine for the process.

It started as a process-monitor to check CPU, memory and file usage.

We want to use it to record logins, file transfer size, count, frequency,
and other information which might help to get an idea on how the product is
used.

Later it might serve as the source for feeding an AI engine to detect abnormal
behaviour and raise security issues.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
from collections import deque
from datetime import datetime, timedelta
from time import mktime

from twisted.internet import defer, reactor
from twisted.internet.error import AlreadyCalled, AlreadyCancelled
from twisted.internet.task import LoopingCall
from zope.interface import implementer

try:
    import psutil
    from psutil import Process
except (ImportError, NotImplementedError, RuntimeError):
    # Placeholders for when we don't have the psutil.
    # NotImplementedError on HPUX
    # RuntimeError on XP.
    Process = None

from chevah_compat import process_capabilities
from chevah.server import emit_error_details, force_unicode
from chevah.server.commons.constant import (
    LAST_LOGIN_TABLE_NAME,
    )
from chevah.server.commons.exception import ServerException
from chevah.server.commons.interface import (
    IAnalyticsConfiguration,
    IAnalyticsResource,
    IProcessMonitorConfiguration,
    IProcessMonitorResource,
    )
from chevah.server.commons.runnable import DeprecatedRunnable, Runnable
from chevah.server.configuration.model import (
    WritableBoolean,
    WritableInteger,
    get_relaxed_datetime,
    is_inherited,
    )
from chevah.server.db.entity import (
    DBFunctionColumn,
    DBTable,
    DBTableColumn,
    DBTableIndex,
    )
from chevah.server.resource.configuration import _ResourceConfigurationBase

# Number of seconds in a day.
_ONE_DAY = 86400


@implementer(IProcessMonitorConfiguration)
class DeprecatedProcessMonitorConfiguration(_ResourceConfigurationBase):
    """
    Deprecated process-monitor configuration to help with the migration
    """
    enabled = False
    # This is deprecated and is the same as in the migrated analytics
    # configuration.
    # jscpd:ignore-start
    monitor_interval = WritableInteger('monitor_interval', default=60)
    memory_resident_trigger = WritableInteger(
        'memory_resident_trigger', default=0)
    memory_virtual_trigger = WritableInteger(
        'memory_virtual_trigger', default=0)
    connection_count_trigger = WritableInteger(
        'connection_count_trigger', default=0)
    file_count_trigger = WritableInteger('file_count_trigger', default=0)
    thread_count_trigger = WritableInteger('thread_count_trigger', default=0)
    # jscpd:ignore-end


@implementer(IProcessMonitorResource)
class DeprecatedProcessMonitorResource(DeprecatedRunnable):
    """
    Just a runnable which is used as a placeholder during the migration
    process.
    """


@implementer(IAnalyticsConfiguration)
class AnalyticsConfiguration(_ResourceConfigurationBase):
    """
    Configurations for the process monitor resource.
    """
    enabled = WritableBoolean('enabled')
    monitor_interval = WritableInteger('monitor_interval', default=60)
    memory_resident_trigger = WritableInteger(
        'memory_resident_trigger', default=0)
    memory_virtual_trigger = WritableInteger(
        'memory_virtual_trigger', default=0)
    connection_count_trigger = WritableInteger(
        'connection_count_trigger', default=0)
    file_count_trigger = WritableInteger('file_count_trigger', default=0)
    thread_count_trigger = WritableInteger('thread_count_trigger', default=0)


class LastLoginDBTable(DBTable):
    """
    Entry persisted in the db for a successful login of an account/admin.
    """

    # Clock used measure time since Unix epoch.
    # Here to help with testing.
    _clock = reactor

    def __init__(self, name):
        super(LastLoginDBTable, self).__init__(name)

        # Columns definitions.
        self.columns.add(DBTableColumn(
            'id', 'INTEGER', primary_key=True, auto_increment=True))

        self.columns.add(DBTableColumn(
            'account_name',
            'VARCHAR',
            200,
            description='Name of the user.'
            ))
        self.columns.add(DBTableColumn(
            'account_uuid',
            'VARCHAR',
            50,
            description='UUID of user account.'
            ))

        # Last successful attempt columns.
        self.columns.add(DBTableColumn(
            'last_login_successful_timestamp',
            'DOUBLE',
            description='Timestamp of successful login attempt.'))
        self.columns.add(DBTableColumn(
            'last_login_successful_peer',
            'VARCHAR',
            30,
            description='IP/port of successful login attempt.'))
        self.columns.add(DBTableColumn(
            'last_login_successful_service',
            'VARCHAR',
            180,
            description='Name of the service of successful login.'))
        self.columns.add(DBTableColumn(
            'last_login_successful_server',
            'VARCHAR',
            180,
            description='Server\'s UUID of successful login.'))

        # Indexes definitions.
        idx_account_name = DBTableIndex('idx_account_name')
        idx_account_name.addColumn(self.columns.column('account_name'))
        self.indexes.addIndex(idx_account_name)

        idx_account_uuid = DBTableIndex('idx_account_uuid')
        idx_account_uuid.addColumn(self.columns.column('account_uuid'))
        self.indexes.addIndex(idx_account_uuid)

        idx_timestamp = DBTableIndex('idx_timestamp')
        idx_timestamp.addColumn(self.columns.column(
            'last_login_successful_timestamp'))
        self.indexes.addIndex(idx_timestamp)

        idx_peer = DBTableIndex('idx_peer')
        idx_peer.addColumn(self.columns.column('last_login_successful_peer'))
        self.indexes.addIndex(idx_peer)

    def getAccountValues(self, avatar):
        """
        Initialize instance members with data from the service avatar instance.
        """
        return self._getSharedValues(avatar)

    def getAdministratorValues(self, avatar):
        """
        Initialize instance members with data from the admin avatar instance.
        """
        return self._getSharedValues(avatar)

    def _getSharedValues(self, avatar):
        """
        For now, code shared to store both account and admins data to the
        last login table.
        """
        success_timestamp = self._clock.seconds()
        success_service = avatar.component.name
        return {
            'account_name': avatar.name,
            'account_uuid': avatar.uuid,
            'last_login_successful_peer': avatar.peer.host,
            'last_login_successful_timestamp': success_timestamp,
            'last_login_successful_service': success_service,
            'last_login_successful_server': 'unknown',
            }


class TransfersDBTable(DBTable):
    """
    Entry persisted in the db for every file that was transferred,
    including failed transfers.

    This is just a very specific, pre-filtered audit/log table.
    Optimized for transfer reports.

    We try to store raw data.
    There will be separate table with agregations.
    """

    # Clock used measure time since Unix epoch.
    # Here to help with testing.
    _clock = reactor

    # ID of events used for generate analytics.
    TRANSFER_EVENTS = [
        '60012',  # Tranfser succeed for a single file.
        '60011',  # Transfer failed for a single file - will retry.
        '60016',  # Job failed after all retries.
        '60017',  # Job success with all files.
        ]

    def __init__(self, name):
        super(TransfersDBTable, self).__init__(name)
        # This is a table without primary key.

        self.columns.add(DBTableColumn(
            'timestamp',
            'DOUBLE',
            description='Timestamp of this event.'))

        # From event ID, we know if it's a failure or a success.
        self.columns.add(DBTableColumn(
            'event_id',
            'VARCHAR',
            size=6,  # Right ow we only need 5. Leave 6 for future growth.
            description='The event ID associated with the file transfer.'))

        self.columns.add(DBTableColumn(
            'transfer_uuid',
            'VARCHAR',
            size=80,
            description='UUID of transfer'
            ))
        idx_transfer_uuid = DBTableIndex('idx_transfer_name')
        idx_transfer_uuid.addColumn(self.columns.column('transfer_uuid'))
        self.indexes.addIndex(idx_transfer_uuid)

        # This is here to allow quick search for a certain filename.
        self.columns.add(DBTableColumn(
            'filename',
            'VARCHAR',
            size=256,
            description='Just the filename + extension that was transferred.'))

        self.columns.add(DBTableColumn(
            'source_path',
            'VARCHAR',
            size=256,
            description='The original source path.'))

        self.columns.add(DBTableColumn(
            'destination_path',
            'VARCHAR',
            size=256,
            description='The final destination path.'))

        self.columns.add(DBTableColumn(
            'transfer_size',
            'INTEGER',
            size=8,
            description='The size transfer.'))

        self.columns.add(DBTableColumn(
            'transfer_duration',
            'REAL',
            description='The duration transfer.'))

    def getRows(self, events):
        """
        Return the raw values that can be used to insert the events in the DB.
        """
        result = []
        for event in events:
            filename = ''
            source_path = ''
            destination_path = ''
            transfer_size = 0
            transfer_duration = 0

            if event.id == '60012':
                # Single file with success.
                filename = event.data['real_file_name']
                transfer_size = event.data['size']
                transfer_duration = event.data['duration']
                source_path = event.data['source_path']
                # FIXME:5934:
                # Add final destination, and not the temporary name to this
                # step.
                destination_path = event.data['destination_path']

            if event.id == '60011':
                # Single step failed.
                filename = event.data['file_name']
                source_path = event.data['path']
                # FIXME:5934:
                # Add final destination to this event.

            result.append({
                'timestamp': event.timestamp.timestamp,
                'event_id': event.id,
                'transfer_uuid': event.creator.uuid,
                'filename': filename,
                'source_path': source_path,
                'destination_path': destination_path,
                'transfer_size': transfer_size,
                'transfer_duration': transfer_duration,
                })
        return result


@implementer(IAnalyticsResource)
class AnalyticsResource(Runnable):
    """
    Takes care of collecting information about how the product is used
    and how it behaves.
    """

    PROPERTIES_REQUIRING_RESTART = (
        'monitor_interval',
        )

    # These are here to allow table name injection during tests.
    _LAST_LOGIN_TABLE_NAME = LAST_LOGIN_TABLE_NAME
    # Later we might need to make this name public to allow the JSON-RPC
    # to access it.
    _TRANSFER_FILES_TABLE_NAME = 'chevah_transfer_files'
    # Number of seconds at which we check the events queue.
    _EVENT_INTERVAL = 5
    # Added here to support testing or using without a started analytics.
    _interaction = None

    def _resetState(self):
        """
        Reset the internal state.
        """
        self._monitor_loop = None
        self._event_loop = None
        self._event_queue = deque()
        self._midnight_call = None
        self._process = None
        self._getOpenFiles = lambda: 0
        self._getOpenConnections = lambda: 0
        self._root_avatar = None
        # The reference to the DB resource.
        self._interaction = None
        # The table used by the analytics.
        self._last_logins = LastLoginDBTable(self._LAST_LOGIN_TABLE_NAME)
        self._last_logins._clock = self._scheduler
        # FIXME:5949:
        # We need to have it called `_table` to allow access from JSON-RPC.
        self._table = self._last_logins
        self._transfer_files = TransfersDBTable(self._TRANSFER_FILES_TABLE_NAME)
        # Just a helper to a low a test to wait the internal loop.
        self._insert_transfers = None
        # Helpers to wait for midnight job during testing.
        self._midnight_deferred = None
        # Helper to wait for check inactive accounts during tesing.
        self._ongoing_check_accounts = None

    @defer.inlineCallbacks
    def _onStart(self, avatar=None):
        """
        See: Runnable.
        """
        self._resetState()

        interval = self.configuration.monitor_interval
        if interval < 1:
            raise ServerException(
                '"monitor_interval" needs to be greater than 0.')

        # First create our tables.
        self._interaction = self.root.resources.getDatabase()
        yield self._interaction.migrateTable(self._last_logins)
        yield self._interaction.migrateTable(self._transfer_files)

        if Process and process_capabilities.os_name != 'aix':
            # Enable process monitoring when supported.
            self._initCompatibleProcess()
            self._monitor_loop = LoopingCall(self._clMonitor)
            self._monitor_loop.clock = self._scheduler
            monitor_deferred = self._monitor_loop.start(interval, now=False)
            monitor_deferred.addErrback(self._ebMonitor)

        self._event_loop = LoopingCall(self._clEvent)
        self._event_loop.clock = self._scheduler
        self._event_loop.start(self._EVENT_INTERVAL, now=False)

        self.root.event_dispatcher.subscribe('emit', self._onEvent)
        self._scheduleMidnight()

    @defer.inlineCallbacks
    def _onStop(self, avatar=None):
        """
        See: Runnable.
        """
        # Stop monitoring for new events and flush the latest events.
        self.root.event_dispatcher.unsubscribe('emit', self._onEvent)
        yield self._clEvent()

        if self._event_loop and self._event_loop.running:
            self._event_loop.stop()

        if self._monitor_loop and self._monitor_loop.running:
            self._monitor_loop.stop()

        if self._midnight_call:
            try:
                self._midnight_call.cancel()
            except (AlreadyCalled, AlreadyCancelled):
                "Already stopped. All good."

        self._resetState()

    def _initCompatibleProcess(self):
        """
        Create a process instance which has the same function for all OSes.
        """
        self._process = Process()
        # Make an initial call so that future calls have valid values.
        self._process.cpu_percent()

        if process_capabilities.os_name == 'windows':
            self._getOpenFiles = self._process.num_handles
        else:
            self._getOpenFiles = self._process.num_fds

    @defer.inlineCallbacks
    def _clMonitor(self):
        """
        Called regularly to do the resource monitoring.
        """
        self._ongoing_check_accounts = self._checkInactiveAccounts()
        yield self._ongoing_check_accounts
        self._ongoing_check_accounts = None
        self._checkProcess()

    def _ebMonitor(self, failure):
        """
        Called when the monitor loop generated a failure.
        """
        return self.fail(failure.value)

    def _checkProcess(self):
        """
        Called the check the process status.
        """
        with self._process.oneshot():
            # Get raw data.
            file_count = self._getOpenFiles()
            connections = self._process.connections()
            memory_info = self._process.memory_info()
            thread_count = self._process.num_threads()
            cpu_percent = self._process.cpu_percent()

        global_memory = psutil.virtual_memory()
        global_cpus = '|'.join([
            str(c) for c in
            psutil.cpu_percent(interval=None, percpu=True)
            ])
        usage = {
            'cpu_percent': cpu_percent,
            'memory_resident': memory_info.rss,
            'memory_virtual': memory_info.vms,
            'connection_count': len(connections),
            'file_count': file_count,
            'thread_count': thread_count,
            'global_cpus': global_cpus,
            'global_memory_available': global_memory.available,
            }
        self.emitEvent('20079', data=usage)

        targets = [
            'memory_resident',
            'memory_virtual',
            'connection_count',
            'file_count',
            'thread_count',
            ]
        triggers = []
        for target in targets:
            trigger = getattr(self.configuration, '%s_trigger' % (target,))
            if trigger < 1:
                # Trigger not enabled.
                continue

            current_value = usage[target]
            if current_value < trigger:
                # We are below the limit.
                continue

            triggers.append((target, current_value))

        if triggers:
            details = ', '.join('%s=%s' % a for a in triggers)
            self.emitEvent(
                '20080',
                data={
                    'details': details,
                    'triggers': triggers,
                    },
                )

    @defer.inlineCallbacks
    def _checkInactiveAccounts(self):
        """
        Called to check and disable any inactive accounts.
        First, handle the accounts with recorded logins.
        Then, handle the rest of the accounts, looking at creation date.
        """
        targeted_accounts = self._getTargetedInactivityAccount()

        if not targeted_accounts:
            # We don't have any account that should be auto-disabled
            # due to inactivity.
            return

        interaction = self._interaction
        result = yield interaction.selectRows(
            table=self._last_logins)
        have_update = False
        now = self._scheduler.seconds()
        for row in result:
            account_uuid = row['account_uuid']
            target = targeted_accounts.pop(account_uuid, None)
            if target is None:
                # Account not targeted for auto-disabling.
                # Maybe the account was removed.
                # We still keep the last-login in the DB.
                continue

            account, disable_on_inactivity = target
            timestamp = row['last_login_successful_timestamp']
            disable_timestamp = timestamp + disable_on_inactivity * 60 * 60 * 24

            if disable_timestamp > now:
                # Auto-disabling scheduled for sometime in the future.
                continue

            # Auto-disable the account.
            account.enabled = False
            have_update = True
            self.emitEvent(
                '20195',
                data={
                    'account_name': account.name,
                    'account_uuid': account.uuid,
                    'day_count': disable_on_inactivity,
                    'last_login': datetime.fromtimestamp(timestamp).isoformat(),
                    },
                )

        for value in targeted_accounts.values():
            # Look at the accounts for which we don't have any last
            # login information.
            account, disable_on_inactivity = value
            try:
                disable_timestamp = self._getAccountDisableTimestamp(
                    account, disable_on_inactivity)
            except Exception as error:
                details = force_unicode(error)
                if isinstance(error, ServerException):
                    # Account has no created date or has invalid created date.
                    # We set the date to now, to prevent raising this type of
                    # error at the next monitoring cycle.
                    created_now = datetime.utcnow().replace(
                        microsecond=0).isoformat() + 'Z'
                    account.created = created_now
                    have_update = True
                    details = details + ' Created date set as current date.'

                self.emitEvent(
                    '20196',
                    data={
                        'account_name': account.name,
                        'account_uuid': account.uuid,
                        'details': details,
                        },
                    )
                continue

            if disable_timestamp > now:
                # Auto-disabling scheduled for sometime in the future.
                continue

            # Auto-disable the account.
            account.enabled = False
            have_update = True
            self.emitEvent(
                '20195',
                data={
                    'account_name': account.name,
                    'account_uuid': account.uuid,
                    'day_count': disable_on_inactivity,
                    'last_login': 'unknown',
                    },
                )

        if have_update:
            self.root.configuration.save()

    def _getAccountDisableTimestamp(self, account, disable_on_inactivity):
        """
        Return the earliest timestamp when the account may be auto-disabled
        because of inactivity, assuming there is no known login event.
        """
        if not account.created:
            raise ServerException(
                'Auto-disable configured for {} days. '
                'No last login or created date found.'.format(
                    disable_on_inactivity))

        created = get_relaxed_datetime(
            account.created, 'Account created')
        disable_date = created + timedelta(days=disable_on_inactivity)
        disable_timestamp = mktime(disable_date.timetuple())
        return disable_timestamp

    def _getTargetedInactivityAccount(self):
        """
        Return a dict with the accounts that have enabled the option to
        disable on inactivity.
        """
        targeted_accounts = {}
        groups = self.root.configuration.groups
        for account in self.root.configuration.accounts.getAll():
            if not account.enabled:
                continue

            if not account.disable_on_inactivity:
                # Configured not to be disabled.
                continue

            if not is_inherited(account.disable_on_inactivity):
                disable_on_inactivity = account.disable_on_inactivity
                targeted_accounts[account.uuid] = account, disable_on_inactivity
                continue

            # Account has an inherited value.
            group = groups.getSection(account.group[0])

            if group.disable_on_inactivity < 1:
                # Primary group configured not to auto-disable.
                continue

            disable_on_inactivity = group.disable_on_inactivity
            targeted_accounts[account.uuid] = account, disable_on_inactivity

        return targeted_accounts

    def accountAuthenticated(self, avatar):
        """
        Called when the a file transfer user is authenticated.
        """
        if not self._interaction:
            # Resource not yet started or DB not yet avaialbe.
            # Nothing to do.
            return defer.succeed(None)

        values = self._last_logins.getAccountValues(avatar)
        return self._recordAuthentication(avatar.uuid, values)

    def administratorAuthenticated(self, avatar):
        """
        Called when the an administrator is authenticated.
        """
        if not self._interaction:
            # Resource not yet started or DB not yet avaialbe.
            # Nothing to do.
            return defer.succeed(None)

        values = self._last_logins.getAdministratorValues(avatar)
        return self._recordAuthentication(avatar.uuid, values)

    @defer.inlineCallbacks
    def _recordAuthentication(self, uuid, values):
        """
        Store both account and admins data after login in the same table.
        """
        # This is called after checking  that we have a DB.
        interaction = self._interaction
        criteria = [[('account_uuid', '=', uuid)]]
        try:
            result = yield interaction.selectRows(
                table=self._last_logins, criteria=criteria, limit=1)

            if not result:
                yield interaction.insertRow(
                    table=self._last_logins, values=values)
                return

            yield interaction.updateRows(
                table=self._last_logins, criteria=criteria, values=values)
            return
        except Exception as error:
            # Updating of an event failed.
            # For now we just log the event, without stopping or prevent future
            # similar errors to happen.
            emit_error_details(error, '20084', {}, emitter=self.emitEvent)

    def _onEvent(self, signal):
        """
        Called when a new event was emitted.

        This should be fast, not to block the event handling.
        """
        event = signal.event
        if event.id not in self._transfer_files.TRANSFER_EVENTS:
            return

        self._event_queue.append(event)

    @defer.inlineCallbacks
    def _clEvent(self):
        """
        Called regularly to handle the latest recorded events.
        """
        if not self._event_queue:
            return

        interaction = self._interaction
        # We save a copy of the current queue and will process only this batch.
        # if events are added while we process the queue, they will be
        # handled at the next loop.
        # The next loop is triggered only after we finish process this batch.
        # No need to worry about overlaping calls.
        current_queue = [e for e in self._event_queue]
        self._event_queue.clear()

        try:
            rows = self._transfer_files.getRows(current_queue)
            self._insert_transfers = interaction.insertRows(
                table=self._transfer_files, rows=rows)
            yield self._insert_transfers

        except Exception as error:
            # We ignore the errors for now, since we want to process
            # the next events.
            self._emitDBError('record transfer', error)
            # Make sure we don't do more proccessing.
            # For now, the failed events are lost.
            # Later we might want to recover then,
            # Since we use SQLite, it is less likely to have a failure and then
            # have an auto-healing.
            # This is why we don't try to recover.
            return

    def _emitDBError(self, operation, error):
        """
        Emit an error on DB operations.

        This consumes the error and converts it into an event.
        """
        self.emitEvent(
            '20191',
            data={
                'operation': operation,
                'details': force_unicode(error),
                },
            )

    def _scheduleMidnight(self):
        """
        Called the schedule the next midnight call.

        FIXME:6371:
        For now his is hardcoded to midnight; in the future we might add
        a configuration and a separate pool.
        """
        timestamp = self._scheduler.seconds()
        now = datetime.fromtimestamp(timestamp)
        tomorrow = (now + timedelta(days=1)).replace(
            hour=0, minute=0, second=0, microsecond=0)
        until_midnight = (tomorrow - now).total_seconds()
        self._midnight_call = self._scheduler.callLater(
            until_midnight, self._onMidnight)

    @defer.inlineCallbacks
    def _onMidnight(self):
        """
        Called at midnight.
        """
        self._midnight_deferred = self._onMidnightCall()
        yield self._midnight_deferred

    @defer.inlineCallbacks
    def _onMidnightCall(self):
        """
        Wrapper to allow havig a single deferred for the midnight call.

        This should never fail.
        """
        self._scheduleMidnight()
        interaction = self._interaction
        now = self._scheduler.seconds()

        # SELECT
        #     COUNT(CASE WHEN event_id = "60012" THEN 1 END),
        #     COUNT(CASE WHEN event_id = "60011" THEN 1 END),
        #     COUNT(CASE WHEN event_id = "60017" THEN 1 END),
        #     COUNT(CASE WHEN event_id = "60016" THEN 1 END),
        #     sum(transfer_size),
        #     sum(transfer_duration)
        # FROM chevah_transfer_files
        # Where timestamp > "1664452230"
        columns = [
            DBFunctionColumn(
                function='COUNT(CASE WHEN event_id = "60012" THEN 1 END)',
                name='success_files',
                kind='INT',
                ),
            DBFunctionColumn(
                function='COUNT(CASE WHEN event_id = "60011" THEN 1 END)',
                name='retried_files',
                kind='INT',
                ),
            DBFunctionColumn(
                function='COUNT(CASE WHEN event_id = "60017" THEN 1 END)',
                name='success_jobs',
                kind='INT',
                ),
            DBFunctionColumn(
                function='COUNT(CASE WHEN event_id = "60016" THEN 1 END)',
                name='failure_jobs',
                kind='INT',
                ),
            DBFunctionColumn(
                function='SUM(transfer_size)',
                name='sum_size',
                kind='INT',
                ),
            DBFunctionColumn(
                function='SUM(transfer_duration)',
                name='sum_duration',
                kind='INT',
                ),
            ]

        try:
            # Have it 1 minute after midnight.
            yesterday_timestamp = now - _ONE_DAY

            criteria = [[('timestamp', '>', yesterday_timestamp)]]
            result = yield interaction.selectRows(
                table=self._transfer_files,
                columns=columns,
                criteria=criteria,
                limit=1,
                )
            total_size = result[0]['sum_size']

            # SQlite returs None if no data is found for the SUM.
            if not total_size:
                total_size = 0
            total_duration = result[0]['sum_duration']
            if not total_duration:
                total_duration = 0

            data = {
                'success_files': result[0]['success_files'],
                'retried_files': result[0]['retried_files'],
                'success_jobs': result[0]['success_jobs'],
                'failure_jobs': result[0]['failure_jobs'],
                'total_size': total_size,
                'total_duration': total_duration,
                }
            self.emitEvent('20192', data=data)
        except Exception as error:
            # We emit the error and continue other operations.
            self._emitDBError('daily summary', error)

        try:
            # Delete older records.
            # FIXME:5633:
            # We should make sure the data is agregated in a separate table
            # before we remove it.
            purge_threashold = now - _ONE_DAY * 31
            criteria = [[('timestamp', '<', purge_threashold)]]
            yield interaction.deleteRows(
                table=self._transfer_files,
                criteria=criteria,
                )
        except Exception as error:
            self._emitDBError('delete old transfers', error)
