# Copyright (c) 2015 Adi Roiban.
# See LICENSE for details.
"""
Email resources.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from email.utils import parseaddr

from twisted.internet import defer
from zope.interface import implementer

from chevah.server import force_unicode
from chevah.server.commons.constant import TYPE_NAME
from chevah.server.commons.exception import ServerException
from chevah.server.commons.interface import (
    IEmailClientConfiguration,
    IEmailClientResource,
    )
from chevah.server.commons.runnable import Runnable
from chevah.server.commons.ssl import (
    SSLClientMixin,
    SSLConfigurationSectionMixin,
    )
from chevah.server.configuration.option import (
    AddressPortOptionsMixin,
    UsernamePasswordPublicOptionsMixin,
    )
from chevah.server.configuration.model import (
    WritableString,
    WritableCSVUniqueInsensitive,
    )
from chevah.server.resource.configuration import _ResourceConfigurationBase
from chevah.server.resource.runnable import CheckMixin

from chevah.server.smtp.client import ESMTPClient


@implementer(IEmailClientConfiguration)
class EmailClientConfiguration(
    _ResourceConfigurationBase,
    AddressPortOptionsMixin,
    SSLConfigurationSectionMixin,
    UsernamePasswordPublicOptionsMixin,
        ):
    """
    Configuration for the email client.
    """
    PORT_DEFAULT = 25

    email_from_address = WritableString('email_from_address')
    email_to_recipients = WritableCSVUniqueInsensitive(
        'email_to_recipients')


@implementer(IEmailClientResource)
class EmailClientResource(Runnable, SSLClientMixin, CheckMixin):
    """
    An SMTP/ESMTP email client as a resource.
    """
    PROPERTIES_REQUIRING_RESTART = (
        'address',
        'port',
        'username',
        'password_update',
        'email_from_address',
        ) + SSLConfigurationSectionMixin.PROPERTIES_REQUIRING_RESTART

    # Used in testing to inject a SMTP client.
    _EmailClient = ESMTPClient

    def __init__(self, parent, configuration):
        super(EmailClientResource, self).__init__(parent, configuration)
        self._client = None

    @defer.inlineCallbacks
    def _onStart(self, avatar=None):
        """
        See: Runnable.
        """
        client, name, address = yield self._getClient(self.configuration)
        self._client = client
        self._sender = name, address

    @defer.inlineCallbacks
    def _getClient(self, configuration):
        """
        Return an email client that can be used to send email.
        """
        if not self.configuration.ssl_certificate_authority:
            # FIXME:6566:
            # This can be removed once we add support for
            # set-on-first-connection.
            # For now we default to Mozilla CA as this is a better default.
            # And most SMTP servers should have public certificates.
            self.configuration.ssl_certificate_authority = (
                TYPE_NAME.MOZILLA_CA_ROOTS)

        context_factory = yield self._initStartedContextFactory(
            remote_servers=[(configuration.address, self.configuration.port)])

        client = self._EmailClient(
            host=configuration.address,
            port=configuration.port,
            username=configuration.username,
            password=configuration.password,
            context_factory=context_factory,
            )
        email_from_address = configuration.email_from_address
        if not email_from_address:
            raise ServerException(
                'Missing "email_from_address" configuration.')

        name, address = parseaddr(email_from_address)
        # Once we no longer need py2, we should remove unicode force.
        name = force_unicode(name)
        address = force_unicode(address)

        if not address:
            raise ServerException(
                'Invalid configuration for "email_from_address".')

        defer.returnValue((client, name, address))

    def _onStop(self, avatar=None):
        """
        See: Runnable.
        """
        self._client = None
        self._sender = None

    def send(self, subject, body, to=(), attachments=None, cc=(), bcc=()):
        """
        See: IEmailClientResource.

        `to` is a list of recipients.
        `attachments` is a dict mapping the attachment to a data stream.
        """
        if not to:
            to = self._configuration.email_to_recipients

        if not to:
            return defer.fail(ServerException(
                'At least one recipient is required for sending emails. '
                'Configure it via "email_to_recipients".'))

        # FIXME:3178:
        # Use a generic helper as similar code will be used in all public
        # methods.
        if not self._client:
            # For now, start will never fail.
            # Auto-start it.
            deferred = self.start()
        else:
            deferred = defer.succeed(None)

        def cb_send_email(result):
            """
            Called when we have a connected client.
            """
            return self._client.send(
                sender=self._sender,
                to=to,
                cc=cc,
                bcc=bcc,
                subject=subject,
                body=body,
                attachments=attachments,
                )

        def cb_email_sent(result):
            """
            Called when email was sent.
            """
            if result:
                return
            raise ServerException('Failed to send the email.')

        deferred.addCallback(cb_send_email)
        deferred.addCallback(cb_email_sent)
        return deferred

    @defer.inlineCallbacks
    def _onCheck(self, checked_configuration):
        """
        See: `IEmailClientResource`.

        `checked_configuration` is a new configuration that we need to check
        that is valid.
        """
        subject = "SFTPPlus Test Email"
        body = """
This is a test message to check that SFTPPlus email configuration is functional.

--
SFTPPlus application
"""
        client, name, address = yield self._getClient(checked_configuration)
        sender = name, address
        to = checked_configuration.email_to_recipients

        if not to:
            raise ServerException(
                'You need to define at least one default recipient '
                'via the "email_to_recipients" option.'
                )

        yield client.send(
            sender=sender,
            to=to,
            subject=subject,
            body=body,
            )

        recipients = ', '.join(to)
        defer.returnValue(
            'Test message successfully sent to {}'.format(recipients))
