# Copyright (c) 2018 Adi Roiban.
# See LICENSE for details.
"""
Act as ACME client to generate and renew SSL certificates from Let's Encrypt.

The certificates are then set to each SSL service and the service
is restarted if required.

The task of updating the service configuration is done by the
certificate storage.
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from zope.interface import implementer, Interface

import pem
from OpenSSL import SSL
from twisted.internet import defer
from twisted.internet.endpoints import TCP4ServerEndpoint
from twisted.python.failure import Failure
from twisted.python.url import URL
from twisted.web import http
from twisted.web.iweb import IResponse

# Let's Encrypt is not available everywhere
try:
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.primitives.asymmetric import rsa
    from josepy.jwa import RS256
    from josepy.jwk import JWKRSA
    from txacme.client import Client, JWSClient
    from acme.errors import ClientError
    from txacme.interfaces import ICertificateStore, IResponder
    from txacme.service import AcmeIssuingService

except ImportError:
    # Placeholders for when we don't have the lets encrypt libraries.
    JWSClient = None
    ICertificateStore = Interface
    IResponder = Interface
    AcmeIssuingService = object

from chevah.server import force_unicode
from chevah.server.commons import DEFAULT_BACKEND
from chevah.server.commons.constant import _PEMS_PLACEHOLDER
from chevah.server.commons.exception import ServerException
from chevah.server.commons.interface import (
    ILetsEncryptClientConfiguration,
    ILetsEncryptClientResource,
    ISSLOptions,
    )
from chevah.server.commons.runnable import Runnable
from chevah.server.commons.service import listen
from chevah.server.configuration.model import (
    WritableBoolean,
    WritableString,
    )
from chevah.server.configuration.option import (
    AddressPortOptionsMixin,
    )
from chevah.server.http.client import PersistentAgent
from chevah.server.http.resource import (
    RedirectResource,
    StaticTextResource,
    )
from chevah.server.http.factory import HTTPFactoryBase
from chevah.server.http.configuration import HTTPServiceConfigurationSection
from chevah.server.resource.configuration import _ResourceConfigurationBase


@implementer(ILetsEncryptClientConfiguration)
class LetsEncryptClientConfiguration(
        _ResourceConfigurationBase,
        AddressPortOptionsMixin,
        HTTPServiceConfigurationSection,
        ):
    """
    Configurations for the Let's Encrypt resource.

    It is a resource configuration and has all the basic HTTP server
    configurations.
    """
    PORT_DEFAULT = 80

    # HTTPServiceConfigurationSection is messing with proxy.
    _proxy = None

    # The resources are started on demand, but for Let's Encrypt we want
    # to be a service and a resource.
    enabled = WritableBoolean('enabled')
    debug = WritableBoolean('debug')
    acme_url = WritableString('acme_url')
    contact_email = WritableString('contact_email')
    redirect_url = WritableString('redirect_url')
    _account_key = WritableString('_account_key')
    _storage = WritableString('_storage')

    @property
    def use_ssl(self):
        """
        See `IServiceConfiguration`.
        """
        return False


# FIXME:2599:
# This should be removed once service configuration is flat.
class _ServiceConfgurationProxy(object):
    """
    This tried to simulate the service configuration in which the
    configuration if found at 2 levels: base and base.configuration.

    For Let's Encrypt we only have `base`.
    """

    _OWN_ATTRIBUTES = ['parent']

    def __init__(self, context):
        self._me = context

    def __getattr__(self, attr):
        if attr in self._OWN_ATTRIBUTES:
            target = self
        else:
            target = self._me
        return getattr(target, attr)

    @property
    def parent(self):
        return self._me


@implementer(ILetsEncryptClientResource)
class LetsEncryptClientResource(Runnable):
    """
    Manage certificate issued by Let's Encrypt.

    It uses the configured account key to register/update an account with that
    key.

    It is an HTTP server for HTTP-01 challenge and ACME client.

    The certificate storage is available even when the resource is stopped.
    """

    PROPERTIES_REQUIRING_RESTART = (
        HTTPServiceConfigurationSection.PROPERTIES_REQUIRING_RESTART + (
            'acme_url',
            'address',
            'port',
            'redirect_url',
            )
        )

    # Size of the account key when auto-generated.
    _KEY_SIZE = 3072

    _JWSClient = JWSClient

    def __init__(self, parent, configuration):
        super(LetsEncryptClientResource, self).__init__(parent, configuration)
        self._resetState()

    def _resetState(self):
        """
        Reset the internal state.
        """
        self._http_server = None
        self._agent = None
        # The high level ACME service taking care of updating the certs.
        self._acme_service = None
        # The acme client used to make high level requests.
        self._client = None
        # The HTTP Server.
        self._factory = LetsEncryptHTTPFactory(
            _ServiceConfgurationProxy(self.configuration),
            on_error=self.fail,
            )
        # The storage for the domain certificates.
        self._store = CertificateStore(
            self.configuration,
            on_update=self._onUpdate,
            domains_getter=self._getConfiguredDomains,
            )

    def _onStart(self, avatar=None):
        """
        See: Runnable.
        """
        self._resetState()

        if self._JWSClient is None:
            raise ServerException(
                'Let\'s Encrypt is not yet supported on this OS.')

        if not self.configuration.acme_url:
            raise ServerException(
                'Missing URL to the Let\'s Encrypt Server Directory.')

        endpoint = TCP4ServerEndpoint(
            reactor=self._scheduler,
            interface=self.configuration.address,
            port=self.configuration.port,
            )

        deferred = listen(endpoint, self._factory, self.configuration.port)
        deferred.addErrback(self._ebHTTPServerStarted)
        deferred.addCallback(self._cbHTTPServerStarted)
        return deferred

    def _ebHTTPServerStarted(self, failure):
        """
        Called when failing to start the HTTP server.
        """
        raise ServerException(
            'Failed to start the local HTTP-01 challenge server. %s' % (
                force_unicode(failure.value),))

    def _cbHTTPServerStarted(self, port):
        """
        Called when HTTP server was started.
        """
        self._http_server = port
        return self._startACMEClient()

    def _startACMEClient(self):
        """
        Start the ACME client service.
        """
        acme_key = self._getAccountKey()

        self._agent = PersistentAgent(
            event_emitter=self.emitEvent,
            ssl_context=SSL.Context(SSL.SSLv23_METHOD),
            )

        jws_client = JWSClient(
            agent=self._agent, key=acme_key, alg=RS256)

        def get_message(failure):
            """
            Return the best human readable message we can get out of failure.
            """
            error = failure.value
            if failure.check(ClientError):
                try:
                    error = error.args[0]
                except IndexError:
                    pass

                if IResponse.providedBy(failure.value.args[0]):
                    response = failure.value.args[0]
                    error = 'Server responded %s:%s' % (
                        response.code, response.phrase)

            return force_unicode(error)

        def cb_client_creation(client):
            """
            Called when the client has validated the ACME server.
            """
            self._client = client
            email = self.configuration.contact_email
            if not email:
                # For no email the configuration is using empty string while
                # txacme needs None.
                email = None

            self._acme_service = AcmeIssuingService(
                cert_store=self._store,
                client=self._client,
                email=email,
                clock=self._scheduler,
                responders=[self._factory.responder],
                panic=self._onPanic,
                )

            deferred = self._acme_service.start()
            deferred.addErrback(eb_service_start)
            return deferred

        def eb_client_creation(failure):
            """
            Called when failing to create the client.
            """
            raise ServerException('Failed to discover the ACME server. %s' % (
                get_message(failure),))

        def eb_service_start(failure):
            """
            Called when the service fails to start.
            """
            raise ServerException('Failed to register the ACME server. %s' % (
                get_message(failure),))

        # To simplify the upgrade, we try to automatically migrate from
        # public V1 to V2.
        acme_url = self.configuration.acme_url
        if acme_url.lower() == 'https://acme-v1.api.letsencrypt.org/directory':
            acme_url = 'https://acme-v02.api.letsencrypt.org/directory'
            self.configuration.acme_url = acme_url

        deferred = Client.from_url(
            self._scheduler,
            URL.fromText(acme_url),
            key=acme_key,
            alg=RS256,
            jws_client=jws_client,
            )
        deferred.addErrback(eb_client_creation)
        deferred.addCallback(cb_client_creation)
        return deferred

    @defer.inlineCallbacks
    def _onStop(self, avatar=None):
        """
        See: Runnable.
        """
        if self._agent:
            yield self._agent.closePersistentConnections()
            self._agent = None

        if self._acme_service:
            yield self._acme_service.stopService()
            self._acme_service = None

        if self._http_server:
            yield self._http_server.stopListening()

    def getCertificatePEMs(self, name, sans=None):
        """
        Return a PEM serialized of all the certificate objects for `name`.

        This include cert itself, key and cert chain.

        This will always return a certificate.
        If no certificate is yet found, it will return a self-signed one.
        """
        if not sans:
            sans = []
        # Is important to join using a comma and space, as this is the
        # normalized name of the certificate.
        certificate_name = ', '.join([name] + sans)

        def cb_store_get(pem_objects):
            """
            Called when we got valid object in the cert storage.
            """
            if not pem_objects:
                # No certificate yet...but should get one soon.
                return _PEMS_PLACEHOLDER

            return '\n'.join(o.as_text() for o in pem_objects)

        def eb_store_get(failure):
            """
            Called when `name` is not in the certificate storage.
            """
            failure.trap(ServerException)

            if self._acme_service:
                # Service is started.
                # Trigger a new cert create. Don't wait for it.
                # Make sure errors are handled.
                deferred = self._acme_service.issue_cert(certificate_name)
                deferred.addErrback(
                    lambda failure: self._onPanic(failure, certificate_name))
                # Here to help with testing.
                self._ongoing_new_cert = deferred

            return _PEMS_PLACEHOLDER

        deferred = self._store.get(certificate_name)
        deferred.addCallback(cb_store_get)
        deferred.addErrback(eb_store_get)

        result = deferred.result
        if isinstance(result, Failure):
            # This is a hack as `getCertificatePEMs` is a sync call.
            # But the store is designed to be async.
            # We are using a custom store that has async return value, but
            # in fact it's a sync code.
            # But by the time we are here, we should already have a result: the
            # real one or the placeholder.
            # Also, the store should never trigger a failure.
            # If a get fails because the cert is not found, it will lauch an
            # auto-cert creation for that certificate.
            # This code is here to catch internal server errors.

            # Tell Twisted failure is handled.
            deferred.addErrback(lambda _: None)

            # Bubble the error in a sync way.
            raise result.value

        return result

    def _getDomainsConfiguration(self):
        """
        Return a dict with mapping of the configured domains to the targeted
        services.
        """
        result = {}

        for service in self.root.services.getAll():
            configuration = service.configuration.configuration
            if not ISSLOptions.providedBy(configuration):
                continue

            if not configuration.ssl_domains:
                continue

            # Is important to join using a comma and space, as this is the
            # normalized name of the certificate.
            domains = ', '.join(configuration.ssl_domains)
            services = result.get(domains, [])
            services.append(service)
            result[domains] = services

        return result

    def _getConfiguredDomains(self):
        """
        Return the list of the domains configured for the services and
        which need to be managed by the resource.
        """
        return list(self._getDomainsConfiguration().keys())

    def _onUpdate(self, certificate_name, pem_objects):
        """
        Called when we got updated certificates/keys for a certificate.
        """
        for name, services in self._getDomainsConfiguration().items():
            if name != certificate_name:
                continue

            for service in services:
                update = ''.join(o.as_text() for o in pem_objects)
                # The service runnable has a service configuration which
                # has the configuration section for the protocol.
                # We don't save the configuration as it is saved
                # in another call.
                service.configuration.configuration.ssl_certificate = update
                service.configuration.configuration.ssl_key = ''

                self.emitEvent(
                    '20016',
                    data={'domains': certificate_name, 'service': service.name}
                    )

                if not service.restart_required:
                    # We either got the same cert,
                    # but most probably the service is stopped.
                    continue

                service.restart()

    def _getAccountKey(self):
        """
        Return the JWK key used by the acme client account.

        This can take some time to run and will block the server when
        generating a new key, but we expect that the key is not generated
        often so for now, don't bother with threading.
        """
        content = self.configuration._account_key
        if content:
            try:
                key = serialization.load_pem_private_key(
                    content.encode('ascii'),
                    password=None,
                    backend=DEFAULT_BACKEND,
                    )
            except Exception as error:
                raise ServerException(
                    'Failed to load the account key. {}'.format(
                        force_unicode(error)))
        else:
            try:
                key = rsa.generate_private_key(
                    public_exponent=65537,
                    key_size=self._KEY_SIZE,
                    backend=DEFAULT_BACKEND,
                    )
                account_key = key.private_bytes(
                    encoding=serialization.Encoding.PEM,
                    format=serialization.PrivateFormat.TraditionalOpenSSL,
                    encryption_algorithm=serialization.NoEncryption(),
                    )
                self._configuration._account_key = account_key.decode('ascii')
                self._configuration._proxy.save()

            except Exception as error:
                raise ServerException(
                    'Failed to generate and save the account key. {}'.format(
                        force_unicode(error)))

        return JWKRSA(key=key)

    def _onPanic(self, failure, certificate_name):
        """
        Called when failing to get a certificate for `domain`.
        """
        if failure.check(defer.FirstError):
            failure = failure.value.subFailure

        self.emitEvent(
            '20017',
            data={
                'domains': certificate_name,
                'details': force_unicode(failure.value),
                'tb': failure.getBriefTraceback(),
                },
            )


class LetsEncryptHTTPFactory(HTTPFactoryBase):
    """
    A simple HTTP site/factory which does Let's Encrypt over HTTP-01.
    """
    def __init__(self, service_configuration, on_error):
        self._responder = HTTP01ChevahResponder()
        super(LetsEncryptHTTPFactory, self).__init__(
            service_configuration, on_error)

    @property
    def responder(self):
        return self._responder

    def _getRootResource(self):
        """
        Root is empty and only has .well-known/acme-challenge/ .
        """
        redirect_url = self.service_configuration.redirect_url
        if redirect_url:
            root = RedirectResource(
                url=redirect_url, code=http.MOVED_PERMANENTLY)
        else:
            root = StaticTextResource('')

        well_known = StaticTextResource('')
        root.putChild(b'.well-known', well_known)
        well_known.putChild(b'acme-challenge', self._responder)

        return root


@implementer(IResponder)
class HTTP01ChevahResponder(StaticTextResource):
    """
    Web resource for ``http-01`` challenge responder.

    Beside the challenge pages, it displays empty pages.
    """
    challenge_type = 'http-01'

    def __init__(self):
        super(HTTP01ChevahResponder, self).__init__('')
        # Add a static response to help with connection troubleshooting.
        self.putChild(b'test.txt', StaticTextResource('Let\'s Encrypt Ready'))

    def start_responding(self, server_name, challenge, response):
        """
        Prepare for the ACME server to validate the challenge.
        """
        self.putChild(
            challenge.encode('token').encode('ascii'),
            StaticTextResource(response.key_authorization),
            )

    def stop_responding(self, server_name, challenge, response):
        """
        Remove the child resource once the process is done.
        """
        encoded_token = challenge.encode('token').encode('ascii')
        if self.getStaticEntity(encoded_token) is not None:
            self.delEntity(encoded_token)


@implementer(ICertificateStore)
class CertificateStore(object):
    """
    A certificate store that keeps certificates in the configuration as a
    single value.

    For certificates with multiple domains, the key is a comma separated list
    of domains (with a space after the comma)

    Return a `ServerException` failure when a requested domain has no
    certificate.
    """

    def __init__(self, configuration, on_update, domains_getter):
        self._storage = {}
        self._configuration = configuration
        self._onUpdate = on_update
        self._getConfiguredDomains = domains_getter
        self._load()

    def _getNormalizedName(self, name):
        """
        Return the normalized comma separated certificate name.
        """
        domains = []
        for part in name.split(','):
            part = part.strip()
            if not part:
                continue
            domains.append(part)

        return ', '.join(domains)

    def as_dict(self):
        """
        Return all the managed certificates as a dict,
        indexed by the certificate name.

        It will not return certificates which were managed in the past
        for which no longer have a service using them.

        This is done so that when a new check is triggered for the certs,
        old certs are not scheduled for renew.
        """
        self._refresh()
        return defer.succeed(self._storage.copy())

    def get(self, certificate_name):
        """
        Get the X.509 objects for `certificate_name`.
        """
        certificate_name = self._getNormalizedName(certificate_name)
        try:
            return defer.succeed(self._storage[certificate_name])
        except KeyError:
            return defer.fail(ServerException(
                'No certificate for "%s".' % (certificate_name,)))

    def store(self, certificate_name, pem_objects):
        """
        Keep and persist the X.509 objects for `certificate_name`.
        """
        certificate_name = self._getNormalizedName(certificate_name)
        self._storage[certificate_name] = pem_objects
        self._save()
        self._onUpdate(certificate_name, pem_objects)
        return defer.succeed(None)

    def _refresh(self):
        """
        Update the domains which needs to be managed.
        """
        names = self._getConfiguredDomains()
        # Delete domains which are no longer handled.
        existing_domains = list(self._storage.keys())
        for domain in existing_domains:
            if domain not in names:
                del self._storage[domain]

        for domain in names:
            if domain not in existing_domains:
                self._storage[domain] = []

    def _load(self):
        """
        Populate the storage with the persisted PEM objects.
        """
        lines = []
        cert_names = None
        lines = self._configuration._storage.splitlines()
        if not lines:
            # Cache is completely empty.
            return

        storage_acme_url = lines[0].strip()
        if storage_acme_url != self._configuration.acme_url:
            # We have a new acme_url so all certificates should be reloaded.
            # So we don't load anything as storage values are no longer valid.
            return

        for line in lines[1:]:
            if '|||' in line:
                # We got a certificate delimiter.
                if cert_names:
                    # Load the current certificate.
                    content = '\n'.join(lines)
                    self._storage[cert_names] = pem.parse(
                        content.encode('utf-8'))

                # Start a new certificate.
                # The certificate name is guarded by '|||'.
                lines = []
                cert_names = self._getNormalizedName(
                    line.strip()[3:-3].strip())
                continue
            lines.append(line)

        if cert_names:
            # Load the last certificate.
            content = '\n'.join(lines)
            self._storage[cert_names] = pem.parse(content.encode('utf-8'))

    def _save(self):
        """
        Persist the current domains to the configuration.
        """
        serialization = []
        for key, value in self._storage.items():
            cert_names = '|||%s|||\n' % (key,)
            serialization.append(
                cert_names
                + ''.join(o.as_text() for o in value)
                + '\n\n')

        acme_url = self._configuration.acme_url + '\n'
        self._configuration._storage = acme_url + '\n'.join(serialization)
        self._configuration._proxy.save()
