# Copyright (c) 2015 Adi Roiban.
# See LICENSE for details.
"""
Monitor is a service which monitor files activity outside of
server process.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
from zope.interface import implementer

from twisted.internet import defer

from chevah_compat import local_filesystem
from chevah.server import emit_error_details
from chevah.server.commons.constant import (
    DEFAULT_LOCAL_FILESYSTEM_UUID,
    EVENT_TYPE,
    )
from chevah.server.commons.exception import ServerException
from chevah.server.commons.interface import (
    IMonitorResource,
    IMonitorConfiguration,
    )
from chevah.server.commons.runnable import Runnable
from chevah.server.configuration.model import (
    PropertyMixin,
    )
from chevah.server.configuration.model import (
    WritableBoolean,
    WritableInteger,
    WritableCSVUniqueInsensitive,
    )
from chevah.server.configuration.option import (
    LocationWatchConfigurationMixin,
    )
from chevah.server.identity.avatar import FilesystemAvatar
from chevah.server.location.watch import FilteredWatchAdapter
from chevah.server.resource.configuration import _ResourceConfigurationBase


@implementer(IMonitorConfiguration)
class MonitorConfiguration(
        _ResourceConfigurationBase,
        LocationWatchConfigurationMixin,
        PropertyMixin):
    """
    Configurations for the service section within monitor service.
    """

    # The resources are started on demand, but for monitor we want
    # to be a service and a resource.
    enabled = WritableBoolean('enabled')

    @property
    def file_age_notification(self):
        """
        Read `file_age_notification` with fallback to v3 name.
        """
        value = self._proxy.getInteger(
            self._section_name, 'file_age_notification')

        if value < 0:
            # Fallback to old configuration option.
            return self._proxy.getInteger(
                self._section_name, 'warn_non_modified_files_interval')

        return value

    @file_age_notification.setter
    def file_age_notification(self, value):
        return self._proxy.setInteger(
            self._section_name, 'file_age_notification', value)

    file_age_auto_delete = WritableInteger('file_age_auto_delete', default=0)
    monitored_operations = WritableCSVUniqueInsensitive(
        'monitored_operations')
    retry_count = WritableInteger('retry_count', default=5)


@implementer(IMonitorResource)
class MonitorResource(Runnable):
    """
    A simple service which monitors a local filesystem and emits
    events on changes.
    """

    PROPERTIES_REQUIRING_RESTART = (
        'recursive',
        'source_path',
        'source_uuid',
        'changes_poll_interval',
        'stable_interval',
        )

    def __init__(self, parent, configuration):
        super(MonitorResource, self).__init__(parent, configuration)
        self._monitor_handler = None
        # Dispatcher based on event types.
        self._onChangeMap = {
            EVENT_TYPE.EXISTING: self._onExisting,
            EVENT_TYPE.MODIFIED: self._onModified,
            EVENT_TYPE.MOVED: self._onMoved,
            EVENT_TYPE.CREATED: self._onCreated,
            EVENT_TYPE.DELETED: self._onDeleted,
            }

    def _emitEventWithPath(self, event_id, data):
        """
        Helper for emitting an event for this monitor with a path
        attribute.
        """
        data['real_path'] = data['path']
        self.emitEvent(event_id, data=data)

    def emitEvent(self, event_id, data=None, avatar=None):
        """
        We accept the avatar to match the function signature,
        but it is always replaced with an avatar for this service.
        """
        if avatar is None:
            avatar = FilesystemAvatar(component=self)

        return super(MonitorResource, self).emitEvent(
            event_id, data=data, avatar=avatar)

    @defer.inlineCallbacks
    def _onStart(self, avatar=None):
        """
        See: Runnable.
        """

        source_uuid = self._configuration.source_uuid
        if not source_uuid:
            source_uuid = DEFAULT_LOCAL_FILESYSTEM_UUID
        source = self.root.locations.getSection(source_uuid)

        source_path = self._configuration.source_path

        source_exists = yield source.exists(source_path)

        if not source_exists:
            raise ServerException('Source path does not exist.')

        self._monitor_handler = FilteredWatchAdapter(
            source=source,
            configuration=self._configuration,
            on_change=self._onSourceChange,
            on_error=self._onSourceError,
            on_snapshot=self._onSnapshot,
            event_emitter=self.emitEvent,
            )
        self._monitor_handler.start()

    def _onStop(self, avatar=None):
        """
        See: Runnable.
        """
        if self._monitor_handler:
            self._monitor_handler.stop()
        self._monitor_handler = None

    def _onSourceChange(self, event):
        """
        Emit event when monitored folder contains changes.
        """
        self._onChangeMap[event.type](event)

    def _onSnapshot(self, snapshot):
        """
        Called each time a snapshot is made.
        """
        total_directories = len(snapshot) - 1
        total_files = 0
        warn_age = self._configuration.file_age_notification
        auto_delete = self._configuration.file_age_auto_delete

        if warn_age < 1 and auto_delete < 1:
            # This is an an optimization.
            return

        now = self._scheduler.seconds()
        for path, members in snapshot.items():

            for name, attributes in members.items():
                if attributes.is_folder:
                    # Don't handle folders.
                    continue

                total_files += 1

                file_age = now - attributes.modified

                if warn_age > 0 and file_age > warn_age:
                    data = {
                        'path': attributes.path,
                        'age': warn_age,
                        }
                    self._emitEventWithPath('20115', data=data)

                if auto_delete > 0 and file_age > auto_delete:
                    data = {
                        'path': attributes.path,
                        'age': auto_delete,
                        }
                    try:
                        segments = local_filesystem.getSegmentsFromRealPath(
                            attributes.path)
                        local_filesystem.deleteFile(segments)
                        self._emitEventWithPath('20082', data=data)
                    except Exception as error:
                        emit_error_details(
                            event_id='20083',
                            error=error,
                            data=data,
                            emitter=self._emitEventWithPath,
                            )

        self.emitEvent(
            '20065',
            data={
                'path': self._configuration.source_path,
                'total_directories': total_directories,
                'total_files': total_files,
                },
            )

    def _onModified(self, event):
        """
        Called when a file/folder was modified.
        """
        # For now, it ignores modified events for folders as they are somehow
        # a duplicate of create / delete events.
        if event.is_folder:
            return

        if self._isIgnored('modify'):
            return

        data = {
            'name': self.name,
            'path': event.source_path,
            }
        self._emitEventWithPath('20166', data=data)

    def _onMoved(self, event):
        """
        Called when a file / folder was moved.
        """
        if self._isIgnored('rename'):
            return

        data = {
            'name': self.name,
            'path': event.destination_path,
            'from_path': event.source_path,
            'to_path': event.destination_path,
            }
        if event.is_folder:
            self._emitEventWithPath('20168', data=data)
        else:
            self._emitEventWithPath('20167', data=data)

    def _onCreated(self, event):
        """
        Called when a file / folder was created.
        """
        if self._isIgnored('create'):
            return

        data = {
            'name': self.name,
            'path': event.source_path,
            }
        if event.is_folder:
            self._emitEventWithPath('20170', data=data)
        else:
            self._emitEventWithPath('20169', data=data)

    def _onDeleted(self, event):
        """
        Called when a delete action occurred on the filesystem.
        """
        if self._isIgnored('delete'):
            return

        data = {
            'name': self.name,
            'path': event.source_path,
            }
        if event.is_folder:
            self._emitEventWithPath('20172', data=data)
        else:
            self._emitEventWithPath('20171', data=data)

    def _onExisting(self, event):
        """
        Called when a file / folder already exist at the monitored location.
        """
        if self._isIgnored('exist'):
            return

        data = {
            'name': self.name,
            'path': event.source_path,
            }
        if event.is_folder:
            self._emitEventWithPath('20180', data=data)
        else:
            self._emitEventWithPath('20179', data=data)

    def _onSourceError(self, error):
        """
        Called when source watcher had an error.
        """
        self.fail(error)

    def _isIgnored(self, event_name):
        """
        Return `True` if `event_name` should be ignored.
        """
        monitored_operations = self._configuration.monitored_operations
        if 'all' in monitored_operations:
            return False

        if 'disable' in monitored_operations:
            return True

        if 'disabled' in monitored_operations:
            # This is a variation.
            return True

        return event_name not in monitored_operations
