# Copyright (c) 2016 Adi Roiban.
# See LICENSE for details.

from __future__ import absolute_import
from __future__ import unicode_literals

from io import StringIO

from twisted.internet import defer, task
from twisted.internet.address import IPv4Address
from zope.interface import implementer

from chevah.server import emit_error_details, force_unicode
from chevah.server.commons.constant import STATE
from chevah.server.commons.configuration import (
    ApplicationConfiguration,
    )
from chevah.server.commons.credential import PasswordCredentials
from chevah.server.commons.exception import ServerException
from chevah.server.commons.interface import IResourceRunnable
from chevah.server.commons.observer import Signal
from chevah.server.commons.runnable import Runnable
from chevah.server.identity.avatar import ClientAvatar


class CheckMixin(object):
    """
    Shared code for testing the configuration of a runnable without
    saving the configuration.

    This is now used only for email, but it is designed to be reused
    for other resource or even other components.
    """

    # FIXME:5893:
    # See what needs to be changed to use the same code for locations.
    # Maybe also standardize the `getClient(checked_configuration)` part.
    def check(self, configuration):
        """
        Check that `configuration` passed as a dict has valid content
        that can be used to successfully start this resource.

        The configuration is not saved, and the existing configuration is not
        changed.
        """
        checked_configuration = self._getCheckConfiguration(
            input_config=configuration)
        return self._onCheck(checked_configuration)

    def _getCheckConfiguration(self, input_config):
        """
        Return the resource configuration object based on the
        `input_config` dictionary.
        """
        # We start by creating a separate configuration tree that is not
        # associated to the running process.
        # This should help process and check the input_config across
        # the full configuration procedure.
        storage = StringIO()
        app_config = ApplicationConfiguration(
            # Not having a parent might get us into trouble, since this is
            # not how the configuration looks in main process.
            # In theory the configuration should not depend on the process.
            configuration_file=storage, parent=None)
        storage.close()

        target_configuration = {}

        for key in self._configuration.getPublicWritableAttributeNames():
            if key not in input_config:
                # Configuration change not requested by the check operation.
                # go with default value.
                continue
            target_configuration[key] = input_config[key]

        if (
            self._configuration.password
            and not input_config['password_update']
                ):
            # We already have a saved configuration and the test is not
            # asking to set or disaable the password.
            # Use this existing password.
            target_configuration['password_update'] = (
                self._configuration.password)

        # Make sure we don't accidentally change the type.
        # Type is also not a public (re)writable so we
        # manually have to add it outside of the writable interface.
        target_configuration['type'] = self._configuration.type

        return app_config.resources.createProperty('', target_configuration)


@implementer(IResourceRunnable)
class ResourceClientRunnable(Runnable):
    """
    A runnable that wraps a resource which connects to something (usually
    a server) to provide its functionality.

    The connection is usually done as a client side connection.

    The wrapped resource is stored as `_client` instance member.
    A hook from `_client.callOnConnectionLost` is auto setup to the runnable.

    Resource runnable are auto-started and auto-connect when their
    functionality is required.

    They can also keep the connection alive while the resource is idle.

    It has the following extra states:
    * CONNECTING - connecting to the client resource
    * DISCONNECTING - in process of disconnecting the client resource while the
        runnable should still be operational.
    * DISCONNECTED - the runnable itself is running but the client resource
        is not connected and not held. Will auto-connect when an action is
        requested.

    Signals:
    * connect - called when resource is connected, at start or at runtime as
      a reconnection.
    """
    # `CONNECTING` will overwrite the `STARTING` state during the start
    # process and even after start it has similar semantics.
    _STARTED_STATES = Runnable._STARTED_STATES + [
        STATE.CONNECTING,
        STATE.DISCONNECTED,
        STATE.DISCONNECTING,
        ]

    _OPERATING_STATES = Runnable._OPERATING_STATES + [STATE.DISCONNECTED]

    # Who is using this resource.
    _avatar = None

    def __init__(self, parent, configuration):
        super(ResourceClientRunnable, self).__init__(parent, configuration)
        # High level client.
        # When is `None` it means that resource is not yet
        # connected and authenticated.
        self._client = None
        self._updateAvatar(None)
        # List of calls to do after client is connected + authenticated.
        self._afterConnectCalls = []
        # Call to the keep alive
        self._keep_alive_call = None
        # Call to disconnect on idle connection.
        self._idle_connection_call = None
        # Deferred of the auto connect process. Used for testing.
        self._auto_connect_deferred = None
        # Deferred use to cancel a connection attempt.
        # We can't use the `self._auto_connect_deferred` since it also holds
        # the callbacks for the start or stop operations.
        self._ongoing_connection = None
        self._ongoing_disconnection = None
        self._connection_retries_left = 0

    def _onStart(self, avatar=None):
        """
        See: `IRunnable`.

        The client resource should not be started directly, but only as
        part of the auto-connection process.
        """
        # FIXME:2501:
        # If client is already connected we should first disconnect
        # but maybe this is already solved by Runnable framework.
        # _onStop should continue only after the connection was closed.
        self._auto_connect_deferred = self._reconnect(_at_start=True)
        return self._auto_connect_deferred

    def _onStop(self, avatar=None):
        """
        See: `IRunnable`.
        """
        return self._disconnect()

    def _onConnect(self):
        """
        Called when client should connect.

        It should return a deferred which is fired with the low level client.
        """
        raise NotImplementedError('_onConnect not implemented')

    def _onDisconnect(self):
        """
        Called when connection should disconnect.
        """
        raise NotImplementedError('_onDisconnect not implemented')

    def emitEvent(self, event_id, data=None, avatar=None):
        """
        Helper to emit an event for this client.
        """
        super(ResourceClientRunnable, self).emitEvent(
            event_id, avatar=self._avatar, data=data)

    @property
    def avatar(self):
        """
        The identity of the credentials for this resource.
        """
        return self._avatar

    def _updateAvatar(self, client):
        """
        Called to update the avatar associated with this resource.

        We can start with a generic ClientAvatar, but after authentication
        and connection we can have a more specific avatar.
        """
        if client:
            # FIXME:4950:
            # See how to have a single avatar,
            # and update the peer of that avatar,
            # instead of replacing it.
            self._avatar = self._client.avatar
            return

        # Go with a default/placeholder avatar for this location.
        # This is a placeholder, until the connection is made
        # and we know more about the remote server.
        credentials = PasswordCredentials(
            username=self._configuration.username,
            creator=self,
            peer=IPv4Address(
                'TCP',
                self._configuration.address,
                self._configuration.port,
                ),
            )

        self._avatar = ClientAvatar(credentials)

    def callWhenConnected(self, call, *args, **kwargs):
        """
        See: `IResourceRunnable`.

        If client is not connected, return a deferred which will be chained
        to the call which is made when client is connected.

        If the resource is not on the process to connect, it will trigger
        a connection.
        """
        if self.state == STATE.STARTED:
            return defer.maybeDeferred(call, *args, **kwargs)

        # We don't have a client so this resource is not connected.
        deferred = defer.Deferred()
        self._afterConnectCalls.append((deferred, call, args, kwargs))

        if self.state is STATE.CONNECTING:
            # We are already connecting.
            # Prevent circular loop from _reconnect()
            return deferred

        self._auto_connect_deferred = self._reconnect()
        # FIXME:3352:
        # Free free to refactor this.
        # If reconnection has failed, it will already trigger the errbacks
        # on the waiting deferred, so we can just mute this deferred as nobody
        # is waiting for it.
        self._auto_connect_deferred.addErrback(lambda _: False)
        return deferred

    def heartbeat(self):
        """
        See: `IResourceRunnable(`.
        """
        if self._keep_alive_call and self._keep_alive_call.active():
            # Delay the keep alive as the heartbeat is always called before
            # issuing a command.
            keep_alive = self.configuration.idle_connection_keepalive_interval
            self._keep_alive_call.reset(keep_alive)

        if self._idle_connection_call and self._idle_connection_call.active():
            # Delay disconnecting due to idle connection as heartbeat is
            # always called before executing a command.
            timeout = self.configuration.idle_connection_timeout
            self._idle_connection_call.reset(timeout)

        if self.state == STATE.STARTED:
            # We are already connected and have previously updated
            # the timeouts.
            # No need for anything else.
            return defer.succeed(None)

        # Most probably we should try to connect/reconnect.
        self._auto_connect_deferred = self._reconnect(_at_heartbeat=True)
        return self._auto_connect_deferred

    @defer.inlineCallbacks
    def _reconnect(self, _at_start=False, _at_heartbeat=False):
        """
        Connect and retry failed connection to the remote server.

        This handles start / heartbeat / callWhenConnected internal connection.

        Return a deferred which is fired when the resource is connected or
        errbacked when connection fails.
        """
        if self._ongoing_disconnection:
            # We are already disconnecting.
            # Wait for the disconnection to be done,
            # before taking any other reconnection action.
            try:
                yield self._ongoing_disconnection
            except Exception:
                """
                Ignore any disconnection error.
                """

        if self.state is STATE.STARTED:
            # By the time disconnection was done,
            # we are already connected.
            defer.returnValue(None)

        if self.state is STATE.STOPPING:
            # Wait to stop and restart.
            # FIXME:3313:
            # Let's hope the resource use will retry and by that time the
            # resource will be stopped.
            error = ServerException('Resource not stopped yet.')
            self._callOnFailedConnection(error)
            raise error

        if self.failed:
            # Fail.
            error = ServerException('Resource has failed.')
            self._callOnFailedConnection(error)
            raise error

        if self.state is STATE.STOPPED:
            # Start this resource.
            # Start will re-call reconnect with the right flags.
            result = yield self.start()
            if not result:
                # Start has failed so reconnect should also fail.
                raise self.last_error
            else:
                defer.returnValue(result)

        # When started, the runnable is placed in CONNECTING state so here
        # we also handle the case in which the runnable is in the process
        # of being started.
        if self.state is STATE.CONNECTING:
            # We are already connecting, so just wait for the main connection
            # Use a do nothing callWhenConnected()
            # callWhenConnected will not recurse here as we have a special
            # handling in callWhenConnected to avoid the recursion.
            yield self.callWhenConnected(lambda: None)
            defer.returnValue(None)

        self._connection_retries_left = (
            self.configuration.connection_retry_count + 1)
        # First attempt is called right away.
        connect_delay = 0
        error = None
        result = None
        while self._connection_retries_left > 0:
            try:
                if connect_delay == 0:
                    # No need to trigger the whole deferLater if we don't
                    # need to.
                    self._ongoing_connection = self.__connect()
                else:
                    self._ongoing_connection = task.deferLater(
                        self._scheduler,
                        connect_delay,
                        self.__connect,
                        )

                yield self._ongoing_connection
                # All good so we can clear any previous errors and continue.
                self._ongoing_connection = None
                error = None
                break

            except defer.CancelledError:
                # Connection was canceled.
                # No need to re-try to reconnect.
                break
            except Exception as last_error:
                # In Py3 we need a separate variable name.
                error = last_error
                connect_delay = self.configuration.connection_retry_interval
                self._connection_retries_left -= 1
                if self._connection_retries_left > 0:
                    emit_error_details(
                        event_id='20035',
                        error=error,
                        data={
                            'retries_left': self._connection_retries_left,
                            'delay': connect_delay
                            },
                        emitter=self.emitEvent,
                        )

        if not error:
            # All good. Nothing to do as _reconnect takes cares of post
            # connection actions.
            return

        # Reconnect failed after all attempts.

        # Fail all callbacks waiting for the connection to start.
        self._callOnFailedConnection(error)

        # FIXME:3352:
        # The usage of _at_start and _at_heartbeat is ugly.
        # Free free to refactor.

        if _at_start:
            # If we are just starting, we don't need to fail as someone
            # else will update the state, but the start is waiting for the
            # reconnect status.
            raise error

        if self.state != STATE.STOPPING:
            # If we have not got an explicit request to stop,
            # fail and wait for runnable to fail, before continuing.
            result = yield self.fail(error=error)

        if _at_heartbeat:
            # The heartbeat is waiting for the reconnect status.
            raise error

        defer.returnValue(result)

    @defer.inlineCallbacks
    def __connect(self):
        """
        Connect in a single attempt to the remote server.

        Return a deferred which is fired when the resource is connected or
        errbacked when connection fails.
        """

        # FIXME:4950:
        # See if _updateAvatar is needed here,
        # as _makeClientConnection will call anyway.
        self._updateAvatar(self._client)
        self._state = STATE.CONNECTING
        self.emitEvent('20140', data={'name': self.name})
        client = yield self._onConnect()
        self.emitEvent('20141', data={'name': self.name})
        self._makeClientConnection(client)

    def _disconnect(self):
        """
        Disconnect from the remote server.
        """
        if self._ongoing_connection:
            self._ongoing_connection.cancel()

        if self._ongoing_disconnection:
            return self._ongoing_disconnection

        def bb_disconnect(result):
            self._ongoing_disconnection = None
            return result

        def cb_disconnect_done(result):
            """
            Called when the resource is disconnected.
            """
            if self.started:
                # By the time the disconnection is done, the resource is
                # back online due to auto-connection.
                return result

            self._client = None
            return result

        if self.started:
            # We are disconnecting only when we are already started.
            # When we stop/fail just keep the state.
            self._state = STATE.DISCONNECTING

        self._ongoing_disconnection = defer.Deferred()
        self._ongoing_disconnection.addBoth(bb_disconnect)
        self._ongoing_disconnection.addCallback(cb_disconnect_done)

        deferred = defer.maybeDeferred(self._onDisconnect)
        deferred.chainDeferred(self._ongoing_disconnection)

        return self._ongoing_disconnection

    def _callOnSuccessfulConnection(self):
        """
        Called when client is connected to trigger all calls which are waiting
        for client connection.
        """
        while self._afterConnectCalls:
            # Call them in the same order as the were added to the queue.
            deferred, call, args, kwargs = self._afterConnectCalls.pop(0)
            deferred_call = defer.maybeDeferred(call, *args, **kwargs)
            deferred_call.chainDeferred(deferred)

    def _callOnFailedConnection(self, error):
        """
        Called when client fails to connect.
        """
        while self._afterConnectCalls:
            deferred, call, args, kwargs = self._afterConnectCalls.pop()
            deferred.errback(error)

    def _makeClientConnection(self, client):
        """
        Called when high level client connection was done.
        """
        self._client = client
        self._updateAvatar(client)

        self._state = STATE.STARTED
        self._client.callOnConnectionLost(self._clientConnectionLost)
        self._callOnSuccessfulConnection()

        # Make sure we notify, after we mark the resource as started.
        self.notify('connect', Signal(self, client=client))

        # Once connected, schedule the first keep alive call.
        self._tickKeepAlive()

        if self.configuration.idle_connection_timeout > 0:
            self._idle_connection_call = self._scheduler.callLater(
                self.configuration.idle_connection_timeout,
                self._disconnectAsIdle,
                )

    def _clientConnectionLost(self, reason):
        """
        Called when the high level client connection was lost.
        """
        if self.state not in [STATE.STOPPED, STATE.STOPPING, STATE.FAILING]:
            # Set disconnected state only if we don't disconnect when we want
            # to stop or fail.
            self._state = STATE.DISCONNECTED

        # We no longer have a valid client.
        self._client = None
        # Once client connection was lost we can stop the keep alive calls and
        # idle disconnection.
        # They should be enabled again when client is re-connected.
        self._stopKeepAlive()
        self._stopIdleTimeout()

        if not self.started:
            # Don't auto-reconnect as runnable is already stopped.
            return

        if self.configuration.idle_connection_timeout > 0:
            # Don't auto-reconnect based on configuration.
            return

        # Try to keep the connection active.
        self.emitEvent('60068', data={'name': self.name})
        # Nobody is waiting for this deferred and this should never return
        # a failure as fail() should always succeed.
        # Here for helping with testing.
        self._auto_connect_deferred = self._reconnect()

    def _disconnectAsIdle(self):
        """
        Disconnect the resource as it became idle.
        """
        timeout = self.configuration.idle_connection_timeout
        self.emitEvent('60067', data={
            'name': self.name,
            'seconds': timeout,
            })
        return self._disconnect()

    def _tickKeepAlive(self):
        """
        Called when the keep alive should be triggered.
        """
        keep_alive = self.configuration.idle_connection_keepalive_interval
        if keep_alive < 1:
            self._keep_alive_call = None
            return

        def eb_keep_alive_call(failure):
            """
            Called when keep alive call failed.
            """
            self.emitEvent('60024', avatar=self._client.avatar, data={
                'name': self.name,
                'details': force_unicode(failure.value),
                })

        def bb_schedule_next(result):
            """
            Called to reschedule the tick.
            """
            self._keep_alive_call = self._scheduler.callLater(
                keep_alive, self._tickKeepAlive)
            return result

        self.emitEvent(
            '60063',
            avatar=self._client.avatar,
            data={'name': self.name},
            )
        deferred = self.exists('.')

        # On errors emit an event.
        deferred.addErrback(eb_keep_alive_call)
        # Only schedule next call after we get the response for the NOOP
        # We schedule the next keep-alive even when the current one failed.
        deferred.addBoth(bb_schedule_next)

    def _stopKeepAlive(self):
        """
        Called when keep alive loop should be stopped.
        """
        if not self._keep_alive_call or not self._keep_alive_call.active():
            return
        self._keep_alive_call.cancel()
        self._keep_alive_call = None

    def _stopIdleTimeout(self):
        """
        Called when idle timeout should be stopped.
        """
        if (
            not self._idle_connection_call or
            not self._idle_connection_call.active()
                ):
            return
        self._idle_connection_call.cancel()
        self._idle_connection_call = None
