# Copyright (c) 2020 Adi Roiban.
# See LICENSE for details.
"""
Code imported for side effects and patching and preparing the process.
"""
from __future__ import print_function, unicode_literals
from __future__ import absolute_import

import base64
import collections
import errno
import locale
import logging
import inspect
import os
import select
import socket
import sys
import timeit
import warnings

import bunch
from ldaptor.protocols.pureber import UnknownBERTag

# We import compat as it has the workaround for using cryptography with
# older OpenSSL version.
from chevah_compat import local_filesystem, process_capabilities
import six

#  This is here to get the same behaviour on Windows for string format.
if os.name == 'nt':
    if sys.getwindowsversion().build <= 14393:
        # We are on Windows Server 2016
        locale.setlocale(locale.LC_ALL, ('English_United States', '1252'))
    else:
        # We are on newer Window.
        locale.setlocale(locale.LC_ALL, ('English_United States', 'utf8'))

elif process_capabilities.os_name == 'osx':
    locale.setlocale(locale.LC_ALL, 'en_US.UTF-8')
else:
    locale.setlocale(locale.LC_ALL, 'C.UTF-8')

# To double-check tht the locale is valid,
# we retrieve it here.
# Python will allow to set any locale, but then fails later when getting it.
locale.getlocale()

# Enable UTF-8 by default for sub-processes.
# See https://peps.python.org/pep-0540/
# Make sure to also propagate this variable to spawned processes.
os.environ['PYTHONUTF8'] = '1'

warnings.filterwarnings(
    'ignore',
    category=UserWarning,
    module='OpenSSL.crypto',
    )

try:
    # hashlib must be imported first to work around some strange problem on
    # RedHat 4.
    # hashlib and SSL is loaded here only for testing the openssl library.
    import hashlib
    from OpenSSL import crypto, SSL
    # pyflakes should no longer complain if we use it somehow.
    hashlib
    crypto
    SSL
except ImportError:  # noqa:cover
    import traceback
    print(traceback.format_exc())
    print(
        'Failed to load the OpenSSL module. Make sure openssl is installed '
        'on your system.')
    sys.exit(-1)

from twisted.python.log import (  # noqa:pycodestyle
    startLoggingWithObserver,
    theLogPublisher,
    )

from twisted.internet import tcp  # noqa:pycodestyle

# Disable the loggers for various libraries.
logging.getLogger('dicttoxml').setLevel(60)
logging.getLogger('docker.utils.config').setLevel(60)
logging.getLogger('passlib.registry').setLevel(600)
logging.getLogger('passlib.utils.compat').setLevel(600)
logging.getLogger('pika').setLevel(600)
logging.getLogger('radius').setLevel(600)
logging.getLogger('smbprotocol').setLevel(600)
logging.getLogger('spnego').setLevel(600)

if os.name == 'nt':
    # On Windows, load pywintypes first, as win32api can not load it
    # from the right folder.
    import pywintypes
    pywintypes

# FIXME:5544:
# Warnings in chevah_compat are affecting functional tests which run in
# a separate process so we can not ignore it only in chevah.server.tests
# package.
warnings.filterwarnings(
    'ignore',
    category=UnicodeWarning,
    module='chevah_compat.*',
    )


#
# Convert Twisted error messages into events.
#
def _handle_twisted_log(eventDict):
    """
    Called when an message is logged by Twisted.
    """
    from chevah.server import _emit
    from chevah.server.helpers import force_unicode
    is_error = eventDict.get('isError', False)
    if not is_error:
        # Ignore log events which are not errors.
        return

    try:
        details = eventDict['format'] % eventDict
    except (AttributeError, KeyError):
        details = six.text_type(eventDict)

    if details == 'Unhandled error in Deferred:':
        # This is part-1 so we ignore it.
        # Twisted will emit another log next with the actual failure.
        return

    failure = eventDict.get('failure', None)
    if failure:
        details = force_unicode(failure.getBriefTraceback())

    reason = eventDict.get('why', 'No reason.')
    if not reason:
        reason = 'No reason.'

    if 'debugInfo' in eventDict:
        # This is part-2 for Unhandled error in deferred.
        failure = eventDict['failure']
        _emit('20024', data={
            'details': force_unicode(failure.value),
            'tb': force_unicode(failure.getBriefTraceback()),
            })
        return

    # We have a generic Twisted error log.
    _emit('20033', data={'reason': reason, 'details': details})


# Twisted logger should only be added once, but somehow it is called
# twice, at least in `./brink.sh debug`.
if _handle_twisted_log not in theLogPublisher.observers:
    startLoggingWithObserver(_handle_twisted_log, setStdout=False)


def patch_resolveIPv6(ip, port):
    """
    Patch upstream code as it does not work on Solaris 10 and Solaris 11.

    See: https://twistedmatrix.com/trac/ticket/9404
    See: https://twistedmatrix.com/trac/ticket/7798
    """
    usedPort = port
    if isinstance(port, int):
        # On macOS systems, `getaddrinfo` will reject an invalid port number,
        # while other systems accept it as long as it is an int.
        # We pass a valid port which is later overwritten.
        usedPort = 1
    else:  # noqa:cover
        # This raises a similar error to the error raised when using
        # the socket.AI_NUMERICSERV flag.
        # The socket.AI_NUMERICSERV flag is not used as it fails on Solaris
        # even when a numeric port is provided.
        raise socket.gaierror(socket.EAI_NONAME, 'Name or service not known')

    result = socket.getaddrinfo(
        ip,
        usedPort,
        socket.AF_UNSPEC,  # Any family.
        0,  # Any socket type.
        0,  # Any protocol.
        socket.AI_NUMERICHOST,
        )[0][4]

    # On non macOS systems, `getaddrinfo` will also `resolve`
    # an invalid port number into a valid one.
    # Example 123456 is resolved as 57920, but we want to preserve the initial
    # port number.
    result = (result[0], port, result[2], result[3])
    return result


tcp._resolveIPv6 = patch_resolveIPv6


def patch_Bunch_repr(self):
    """
    Upstream bunch is using List.sort() which is no longer available in Py3.
    """
    keys = sorted(self.keys())
    args = ', '.join(['%s=%r' % (key, self[key]) for key in keys])
    return '%s(%s)' % (self.__class__.__name__, args)


def bunchify(x):
    """
    Recursively transforms a dictionary into a Bunch via copy.
    """
    if isinstance(x, dict):
        return bunch.Bunch((k, bunchify(v)) for k, v in x.items())
    elif isinstance(x, (list, tuple)):
        return type(x)(bunchify(v) for v in x)
    else:
        return x


def unbunchify(x):
    """
    Recursively converts a Bunch into a dictionary.
    """
    if isinstance(x, dict):
        return dict((k, unbunchify(v)) for k, v in x.items())
    elif isinstance(x, (list, tuple)):
        return type(x)(unbunchify(v) for v in x)
    else:
        return x


@staticmethod
def patch_fromDict(d):
    return bunchify(d)


def patch_toDict(self):
    return unbunchify(self)


# We should really just remove Bunch and create own implementation,
# instead of all this patching.
bunch.Bunch.__repr__ = patch_Bunch_repr
bunch.Bunch.fromDict = patch_fromDict
bunch.Bunch.toDict = patch_toDict


# Insert the extensions path, before doing any chevah.server import.
from chevah_compat import local_filesystem  # noqa
_extension_path = local_filesystem.getRealPathFromSegments(
    local_filesystem.installation_segments + ['extension'])

if six.PY2:

    def fix_force_unicode(data, encoding):
        from chevah.server.helpers import force_unicode
        return force_unicode(data)

    # On py2 this is bytes, while on Py3 we have text.
    _extension_path = _extension_path.encode('utf-8')

    if not hasattr(base64, 'decodebytes'):
        base64.decodebytes = base64.decodestring

sys.path.append(_extension_path)

# FIXME:6229:
# This can be removed once we no longer need to support py2.
# And also update all our dependencies.
if six.PY3:
    # On Py2 we don't have ABC.
    for member in ['Callable', 'Iterable', 'Mapping', 'Sequence']:
        if not hasattr(collections, member):
            setattr(collections, member, getattr(collections.abc, member))
    import cryptography.utils
    if not hasattr(cryptography.utils, 'int_from_bytes'):
        cryptography.utils.int_from_bytes = int.from_bytes

    if not hasattr(base64, 'encodestring'):
        base64.encodestring = base64.encodebytes

    if not hasattr(base64, 'decodestring'):
        base64.decodestring = base64.decodebytes

    if not hasattr(inspect, "getargspec"):
        inspect.getargspec = lambda func: inspect.getfullargspec(func)[:4]


# FIXME:6193:
# See: https://github.com/jborean93/smbprotocol/issues/263
def _patch_recv(self, length, timeout):
    """
    Patch SMB protocol not to block when recv overlaps with close.
    """
    buffer = bytearray(length)
    offset = 0
    while offset < length:
        read_len = length - offset

        start_time = timeit.default_timer()

        if not self.connected:
            return None, timeout  # The socket was closed

        try:
            read = select.select([self._sock], [], [], max(timeout, 1))[0]
        except OSError as error:
            if error.errno == errno.EBADF:
                # Connection was already closed.
                return None, timeout
            # So far, we only see this error on macOS with EBADF
            # In the case in which we got other errors, we raise them to
            # be aware of what's going on there.
            raise  # noqa:cover
        timeout = timeout - (timeit.default_timer() - start_time)

        if not read:
            raise TimeoutError()

        if not self.connected:
            return None, timeout  # The socket was closed

        with self._close_lock:
            if not self.connected:
                return None, timeout  # The socket was closed

            try:
                b_data = self._sock.recv(read_len)
            except socket.error as e:
                # Windows will raise this error if the socket has been shutdown,
                # Linux return returns an empty byte
                # string so we just replicate that.
                if e.errno not in [errno.ESHUTDOWN, errno.ECONNRESET]:
                    raise
                b_data = b''

        read_len = len(b_data)

        if read_len == 0:
            self.close()
            return None, timeout  # The socket has been shutdown

        buffer[offset:offset + read_len] = b_data
        offset += read_len

    return bytes(buffer), timeout


from smbprotocol.transport import Tcp  # noqa:pycodestyle
Tcp._recv = _patch_recv


# FIXME:6193:
# See https://github.com/jborean93/smbprotocol/issues/254
def _patch_process_message_thread(self):
    import threading
    self._t_worker = threading.current_thread()
    return self._unpatched_process_message_thread()


from smbprotocol.connection import Connection  # noqa:pycodestyle
Connection._t_worker = None
Connection._unpatched_process_message_thread = (
    Connection._process_message_thread)
Connection._process_message_thread = _patch_process_message_thread

# FIXME:6568:
# This can be removed once ldaptor is updated.
# See: https://github.com/twisted/ldaptor/issues/243
UnknownBERTag.__str__ = lambda self: ''
