# Copyright (c) 2021 Adi Roiban.
# See LICENSE for details.
"""
Low level code for SMB client implementation.

For now it adapts the thread-based smbprotocol code to deferred.
It also converts the smbprotocol protocol errors into ServerExceptions,
"""
from __future__ import unicode_literals
from __future__ import absolute_import
import datetime
import uuid
import stat
import struct
import types

from smbprotocol.connection import Connection, Ciphers
from smbprotocol.exceptions import (
    AccessDenied,
    BadNetworkName,
    EndOfFile,
    NoMoreFiles,
    ObjectPathNotFound,
    ObjectNameNotFound,
    SMBException,
    SMBConnectionClosed,
    )
from smbprotocol import file_info, Dialects
from smbprotocol.open import (
    CreateDisposition,
    CreateOptions,
    DirectoryAccessMask,
    FileAttributes as SMBFileAttributes,
    FilePipePrinterAccessMask,
    ImpersonationLevel,
    Open,
    QueryDirectoryFlags,
    ShareAccess,
    SMB2SetInfoResponse,
    SMB2SetInfoRequest,
    SMB2QueryInfoRequest,
    SMB2QueryInfoResponse,
    )
from smbprotocol.session import Session
from smbprotocol.structure import DateTimeField
from smbprotocol.tree import TreeConnect
from twisted.internet import defer, reactor, threads
from twisted.internet.error import ConnectionDone, ConnectionClosed
from zope.interface import implementer

from chevah_compat import FileAttributes
from chevah.server import force_unicode
from chevah.server.commons import UTC
from chevah.server.commons.exception import ServerException
from chevah.server.commons.interface import ILocationShell, ITransferStream
from chevah.server.commons.queue import SingleProcessQueue
from chevah.server.commons.session import RemoteConnectionMixin
from chevah.server.helpers import get_timestamp
from chevah.server.identity.avatar import ClientAvatar
from chevah.server.location.commons import OnConnectionLostMixin
import six


_EPOCH_FILETIME = 116444736000000000  # epoch as a MS FILETIME int.


class SMBNotFound(ServerException):
    """
    A marker for a path that was not found.
    """


class SMBCancelled(ServerException):
    """
    A marker for when the operation was cancelled.
    """


class SMBConnectionLost(ServerException):
    """
    A marker that the connection was lost unexpected.
    """


_DIALECT_NAMES = {
    Dialects.SMB_2_0_2: 'SMB 2.0.2',
    Dialects.SMB_2_1_0: 'SMB 2.1.0',
    Dialects.SMB_3_0_0: 'SMB 3.0.0',
    Dialects.SMB_3_0_2: 'SMB 3.0.2',
    Dialects.SMB_3_1_1: 'SMB 3.1.1',
    }


def _patched_pack_value(self, value):  # noqa:cover
    utc_tz = datetime.timezone.utc
    utc_dt = value.replace(tzinfo=value.tzinfo if value.tzinfo else utc_tz)
    td = (
        utc_dt.astimezone(utc_tz) -
        datetime.datetime(1970, 1, 1, 0, 0, 0, tzinfo=utc_tz)
        )
    epoch_time_ms = (
        td.microseconds + (td.seconds + td.days * 24 * 3600) * 10 ** 6)
    ns100 = self.EPOCH_FILETIME + (epoch_time_ms * 10)

    format = self._get_struct_format(8)
    struct_string = "%s%s"\
                    % ("<" if self.little_endian else ">", format)
    bytes_value = struct.pack(struct_string, ns100)

    return bytes_value


def _patched_parse_value(self, value):
    if value is None:
        datetime_value = datetime.datetime.today()
    elif isinstance(value, types.LambdaType):  # noqa:cover
        datetime_value = value
    elif isinstance(value, bytes):
        format = self._get_struct_format(8)
        struct_string = "%s%s"\
                        % ("<" if self.little_endian else ">", format)
        int_value = struct.unpack(struct_string, value)[0]
        return self._parse_value(int_value)  # just parse the value again
    elif isinstance(value, six.integer_types):

        time_microseconds = (value - self.EPOCH_FILETIME) // 10
        try:
            datetime_value = (
                datetime.datetime(1970, 1, 1) +
                datetime.timedelta(microseconds=time_microseconds)
                )
        except OverflowError:  # noqa:cover
            # Some servers might return crazy datetime values.
            # This is unfortunately but 9999 is the max value a datetime can
            # be so we just default to that.
            datetime_value = datetime.datetime.max

    elif isinstance(value, datetime.datetime):
        datetime_value = value
    else:  # noqa:cover
        raise TypeError("Cannot parse value for field %s of type %s to a "
                        "datetime" % (self.name, type(value).__name__))
    return datetime_value.replace(tzinfo=UTC)


# See https://github.com/jborean93/smbprotocol/pull/125/
# This can be removed once we migrate to Python 3 and can update the
# smbprotocol library.
DateTimeField._pack_value = _patched_pack_value
DateTimeField._parse_value = _patched_parse_value


@implementer(ILocationShell)
class SMBClient(OnConnectionLostMixin, RemoteConnectionMixin):
    """
    Adapter for smbprotocol low-level API to deferred based API.

    We have 3 stacked layers in smbprotocol API:
    * Connection - The raw TCP connection.
    * Session - A session authenticated to the server.
    * Tree - Authorized access to a share.

    The low-level paths are using Windows '\' separators.
    Our high-level API should always accept only '/' separators and produce
    paths with unix ('/') separators.
    """

    def __init__(self, credentials, require_encryption):
        self._credentials = credentials
        self._location = credentials.creator
        self._avatar = None
        self._connection = None
        self._session = None
        # Cache for authenticated shares.
        self._trees = {}
        self._server_name = None
        self._operations = SingleProcessQueue()
        self._connected = False
        self._require_encryption = require_encryption

    @property
    def avatar(self):
        """
        See: ILocationShell.
        """
        return self._avatar

    def emitEvent(self, event_id, data=None):
        """
        Helper to emit an event for this SMB client.
        """
        self._location.emitEvent(event_id, data=data)

    def connectionMade(self):
        """
        Called when we are connected to the server.

        This is called before the authentication.
        """
        self._connected = True

        # In the future maybe also add the capabilities
        # from smbprotocol.connection.Capabilities and
        # _connection.server_capabilities
        dialect = self._connection.dialect
        dialect = _DIALECT_NAMES.get(dialect, str(dialect))
        self.emitEvent('70001', data={
            'dialect': dialect,
            })

    def connectionLost(self, reason):
        """
        Called when connection to server was lost.

        This tries to emulates the same API as Twisted.
        For now, `reason` is an exception and not a failure.
        """
        # FIXME:5006:
        # For now, this is not called right away when the connection is lost.
        # It is only called when we need to do an operation over a closed
        # connection.
        # This should be fixed once we have async network.

        self._operations.stop()

        if not self._connected:
            return defer.succeed(None)

        self._connected = False
        self.emitEvent('70002', data={'details': str(reason)})
        # Notify anyone interested in this event.
        # This will allow for an external connection retry.
        self._onConnectionLost(reason)
        return threads.deferToThread(self._tcDisconnect)

    def loseConnection(self, reason):
        """
        Called when connection to server should be closed.
        """
        # FIXME:5006:
        # For now this is just an alias for the connection lost.
        return self.connectionLost(reason)

    @defer.inlineCallbacks
    def authenticate(self):
        """
        Connect and authenticate to the server attached to credentials.
        """
        self._avatar = ClientAvatar(self._credentials)

        try:
            self._session = yield threads.deferToThread(
                self._tcConnectAndAuthenticate)
        except Exception as error:
            self.loseConnection(error)
            raise error

        cipher, signature = self._getEncryptionDescription()

        self.emitEvent('70003', data={
            'cipher': cipher,
            'signature': signature,
            })
        self._operations.start()
        defer.returnValue(self)

    def _tcConnectAndAuthenticate(self):
        """
        Called from the thread to do the blocking connection and authentication
        process.
        """
        self._connection = Connection(
            uuid.uuid4(),
            self._credentials.peer.host,
            self._credentials.peer.port,
            require_signing=True,
            )

        try:
            self._connection.connect()
        except Exception as error:
            error = ServerException('Failed to connect to SMB server. %s' % (
                force_unicode(error),))
            raise error

        reactor.callFromThread(self.connectionMade)

        session = Session(
            self._connection,
            self._credentials.username,
            self._credentials.password,
            require_encryption=self._require_encryption,
            )

        try:
            session.connect()
        except SMBException as error:
            message = force_unicode(error)
            if message.startswith(
                'SMB encryption or signing was required'
                    ):
                # It looks like when the user is not found some servers
                # like Samba will fallback is to anonymous session.
                # Present an unified error.
                message = (
                    'Received unexpected status from the server: '
                    'The attempted logon is invalid.'
                    )

            error = ServerException(
                'Failed to authenticate. %s' % (message,))
            raise error

        return session

    def _getEncryptionDescription(self):
        """
        Return a tuple with (cipher_name, signature_name) used for
        the connection.
        """
        if self._connection.dialect < Dialects.SMB_3_0_0:
            signature = 'SHA256'
        else:
            signature = 'CMAC-AES128'

        if not self._session.signing_key:
            # We don't yet have a negotiated key.
            signature = 'pending-' + signature

        if not self._session.signing_required:
            # The connection will not need a signature.
            signature = 'optional-' + signature

        if self._connection.dialect < Dialects.SMB_3_1_1:
            # Older version have fixed cipher.
            cipher = 'AES-128-CCM'
        else:
            cipher_id = self._connection.cipher_id
            if not cipher_id:
                cipher = 'no-cipher'
            else:
                cipher_map = {
                    Ciphers.AES_128_CCM: 'AES-128-CCM',
                    Ciphers.AES_128_GCM: 'AES-128-GCM',
                    Ciphers.AES_256_CCM: 'AES-256-CCM',
                    Ciphers.AES_256_GCM: 'AES-256-GCM',
                    }
                cipher = cipher_map.get(cipher_id, six.text_type(cipher_id))

        return cipher, signature

    def _tcDisconnect(self):
        """
        Disconnects, ignoring any errors.

        To be called from a thread.
        """
        if self._connection:  # pragma: no branch
            try:
                # We don't wait for the existing sessions to be logged-out
                # and force closing the connection.
                self._connection.disconnect(close=False)
                # FIXME:6193:
                # Might be removed with updated smbprotocol.
                # See https://github.com/jborean93/smbprotocol/issues/254
                self._connection._t_worker.join(timeout=5)
            except SMBException:  # noqa:cover
                """
                We ignore disconnection errors as the server might have
                already disconnected us.

                This does not always happen in our CI tests.
                """

        self._session = None
        self._connection = None

    def close(self):
        """
        Called when the SMB should no longer be used.
        """
        return self.loseConnection(ConnectionDone('SMB client closed'))

    @defer.inlineCallbacks
    def getAttributes(self, path):
        """
        See: ILocation.
        """

        def get_attributes(handler):
            """
            Called from within the thread.
            """
            handler.create(
                impersonation_level=ImpersonationLevel.Impersonation,
                desired_access=(
                    FilePipePrinterAccessMask.FILE_READ_ATTRIBUTES),
                file_attributes=SMBFileAttributes.FILE_ATTRIBUTE_NORMAL,
                share_access=ShareAccess.FILE_SHARE_READ,
                create_disposition=CreateDisposition.FILE_OPEN,
                create_options=0,
                )

            basic_response = _tc_query_info(
                handler, file_info.FileBasicInformation)
            internal_response = _tc_query_info(
                handler, file_info.FileInternalInformation)
            standard_response = _tc_query_info(
                handler, file_info.FileStandardInformation)
            return basic_response, internal_response, standard_response

        path, result = yield self._threadedOperation(path, get_attributes)
        basic, internal, standard = result

        name = path.rsplit('/', 1)[-1]

        is_file, is_folder, mode = _get_attributes(basic['file_attributes'])

        defer.returnValue(FileAttributes(
            name=name,
            path=path,
            mode=mode,
            is_folder=is_folder,
            is_file=is_file,
            is_link=False,
            # SMB has last access, last write, and last attribute_write times.
            modified=_from_smb_time(basic['last_write_time']),
            size=standard['end_of_file'].get_value(),
            hardlinks=standard['number_of_links'].get_value(),
            node_id=internal['index_number'].get_value(),
            ))

    @defer.inlineCallbacks
    def exists(self, path):
        """
        See: ILocation.

        Implement via get attributes.
        """
        try:
            yield self.getAttributes(path)
            defer.returnValue(True)
        except SMBNotFound:
            # It's OK if not found.
            defer.returnValue(False)

    def createFolder(self, path):
        """
        See: ILocation.
        """
        def create_folder(handler):
            """
            Called from within the thread.
            """
            handler.create(
                impersonation_level=ImpersonationLevel.Impersonation,
                desired_access=DirectoryAccessMask.GENERIC_WRITE,
                file_attributes=SMBFileAttributes.FILE_ATTRIBUTE_DIRECTORY,
                share_access=0,
                create_disposition=CreateDisposition.FILE_CREATE,
                create_options=CreateOptions.FILE_DIRECTORY_FILE,
                )

        return self._threadedOperation(path, create_folder)

    def deleteFolder(self, path, recursive=False):
        """
        See: ILocation.

        Recursive delete is not supported.
        """

        def delete_folder(handler):
            """
            Called from within the thread.
            """
            _tc_delete_path(handler, is_file=False)

        return self._threadedOperation(path, delete_folder)

    def deleteFile(self, path):
        """
        See: ILocation.
        """

        def delete_file(handler):
            """
            Called from within the thread.
            """
            _tc_delete_path(handler, is_file=True)

        return self._threadedOperation(path, delete_file)

    def rename(self, source_path, destination_path):
        """
        See: ILocation.
        """
        source_share, _ = _split_share(source_path)
        destination_share, rename_path = _split_share(destination_path)

        if source_share != destination_share:
            return defer.fail(ServerException(
                'Rename is only supported inside the same share. '
                '%s -> %s' % (source_share, destination_share)))

        def tc_rename(handler):
            """
            Called from within the thread.
            """
            handler.create(
                impersonation_level=ImpersonationLevel.Impersonation,
                desired_access=DirectoryAccessMask.DELETE,
                file_attributes=0,
                share_access=(
                    ShareAccess.FILE_SHARE_DELETE
                    | ShareAccess.FILE_SHARE_READ
                    | ShareAccess.FILE_SHARE_WRITE
                    ),
                # I don't know why this is needed.
                create_disposition=CreateDisposition.FILE_OPEN,
                create_options=CreateOptions.FILE_OPEN_REPARSE_POINT,
                )

            file_rename = file_info.FileRenameInformation()
            file_rename['replace_if_exists'] = False
            # This needs to be a relative path to the share's root.
            file_rename['file_name'] = rename_path
            _tc_set_info(handler, file_rename)

        return self._threadedOperation(source_path, tc_rename)

    @defer.inlineCallbacks
    def getStreamForWriting(self, path, length=None):
        """
        See: ILocation.

        `length` is ignored.
        """

        def open_for_writing(handler):
            """
            Called from within the thread.
            """
            handler.create(
                impersonation_level=ImpersonationLevel.Impersonation,
                desired_access=(
                    FilePipePrinterAccessMask.FILE_WRITE_DATA
                    | FilePipePrinterAccessMask.FILE_WRITE_ATTRIBUTES
                    | FilePipePrinterAccessMask.FILE_WRITE_EA
                    ),
                file_attributes=SMBFileAttributes.FILE_ATTRIBUTE_NORMAL,
                share_access=ShareAccess.FILE_SHARE_WRITE,
                # Overwrite any existing file.
                create_disposition=CreateDisposition.FILE_OVERWRITE_IF,
                create_options=(
                    CreateOptions.FILE_NON_DIRECTORY_FILE
                    | CreateOptions.FILE_OPEN_REPARSE_POINT
                    ),
                )
            return handler

        path, handler = yield self._threadedOperation(
            path, open_for_writing, auto_close=False)
        defer.returnValue(SMBWriteStream(
            handler,
            self._location,
            path,
            max_chunk=self._connection.max_write_size,
            on_disconnect=self.connectionLost,
            ))

    @defer.inlineCallbacks
    def getStreamForReading(self, path):
        """
        See: ILocation.
        """

        def open_for_reading(handler):
            """
            Called from within the thread.
            """
            handler.create(
                impersonation_level=ImpersonationLevel.Impersonation,
                desired_access=(
                    FilePipePrinterAccessMask.FILE_READ_DATA
                    | FilePipePrinterAccessMask.FILE_READ_ATTRIBUTES
                    | FilePipePrinterAccessMask.FILE_READ_EA
                    ),
                file_attributes=SMBFileAttributes.FILE_ATTRIBUTE_NORMAL,
                share_access=ShareAccess.FILE_SHARE_READ,
                # Fail if file doesn't exist.
                create_disposition=CreateDisposition.FILE_OPEN,
                create_options=(
                    CreateOptions.FILE_NON_DIRECTORY_FILE
                    | CreateOptions.FILE_OPEN_REPARSE_POINT
                    ),
                )
            return handler

        path, handler = yield self._threadedOperation(
            path, open_for_reading, auto_close=False)
        defer.returnValue(SMBReadStream(
            handler,
            self._location,
            path,
            max_chunk=self._connection.max_read_size,
            on_disconnect=self.connectionLost,
            ))

    def getStreamForAppending(self, path):
        """
        See: ILocation.
        """
        return defer.fail(
            ServerException('Appending is not supported for SMB.'))

    @defer.inlineCallbacks
    def getFolderMemberAttributes(self, path):
        """
        See: ILocation.
        """
        def list_folder(handler):
            """
            Called from within the thread.
            """
            base_path = path.rstrip('/')
            handler.create(
                impersonation_level=ImpersonationLevel.Impersonation,
                desired_access=(
                    DirectoryAccessMask.FILE_LIST_DIRECTORY
                    | DirectoryAccessMask.FILE_READ_ATTRIBUTES
                    | DirectoryAccessMask.FILE_READ_EA
                    ),
                file_attributes=SMBFileAttributes.FILE_ATTRIBUTE_DIRECTORY,
                share_access=ShareAccess.FILE_SHARE_READ,
                create_disposition=CreateDisposition.FILE_OPEN,
                create_options=(
                    CreateOptions.FILE_DIRECTORY_FILE
                    | CreateOptions.FILE_OPEN_REPARSE_POINT
                    ),
                )

            file_info_class = file_info.FileInformationClass
            info_flags = file_info_class.FILE_ID_FULL_DIRECTORY_INFORMATION
            query_flags = QueryDirectoryFlags.SMB2_RESTART_SCANS

            result = {}

            while True:
                try:
                    entries = handler.query_directory(
                        '*',
                        info_flags,
                        flags=query_flags,
                        )
                except NoMoreFiles:
                    break

                # Only the first request should have set SMB2_RESTART_SCANS
                query_flags = 0

                for entry in entries:
                    name = entry['file_name'].get_value().decode('utf-16-le')
                    if name in ['.', '..']:
                        # We don't care about current path and its parent.
                        continue

                    # Server returns the value in UTC.
                    modified = get_timestamp(
                        entry['last_write_time'].get_value())
                    is_file, is_folder, mode = _get_attributes(
                        entry['file_attributes'])
                    member_path = base_path + '/' + name
                    result[name] = FileAttributes(
                        name=name,
                        path=member_path,
                        size=entry['end_of_file'].get_value(),
                        is_file=is_file,
                        is_folder=is_folder,
                        modified=modified,
                        node_id=entry['file_id'].get_value(),
                        mode=mode,
                        )
            return result

        path, result = yield self._threadedOperation(path, list_folder)
        defer.returnValue(result)

    def _threadedOperation(self, path, callback, auto_close=True):
        """
        Execute a single operation at a time in a thread.
        """
        deferred = self._operations.queue(
            self._executeOperation,
            path=path, callback=callback, auto_close=auto_close,
            )

        def eb_cancel(failure):
            failure.trap(defer.CancelledError)
            raise SMBCancelled('SMB operation cancelled.')

        deferred.addErrback(eb_cancel)
        return deferred

    @defer.inlineCallbacks
    def _executeOperation(self, path, callback, auto_close=True):
        """
        Return a deferred with the result of `callback` executed inside
        a thread.

        `path` is the absolute path, including the share name.

        The callback is called with a SMB handler as the first argument.
        The handler is not automatically created.
        The handler is automatically closed.

        It does common error handling for SMB errors.

        it returns a tuple of (path_with_share, operation_result)
        """
        path = path.rstrip('/')
        tree, share_path = yield self._getTreeAndPath(path)

        # Construct as much as possible outside of the thread.
        handler = Open(tree, share_path)

        def tc_operate():
            try:
                return callback(handler)
            except (
                ObjectNameNotFound,
                ObjectPathNotFound,
                    ) as error:
                message = force_unicode(error).replace(
                    'Received unexpected status from the server: ', '')
                raise SMBNotFound(message)

            except AccessDenied as error:
                # On Windows Share and maybe also on Azure with Kerberos auth
                # you can get access denied after the share is opened.
                message = force_unicode(error).replace(
                    'Received unexpected status from the server: ', '')
                raise ServerException('Access denied to "%s". %s' % (
                    path, message))

            except SMBConnectionClosed:
                # Connection closed during operation.
                # Raise it to not be converted into a general error.
                # It will be handled by the helper code.
                raise

            except SMBException as error:
                message = force_unicode(error).replace(
                    'Received unexpected status from the server: ', '')
                raise ServerException(message)

            finally:
                if auto_close:
                    handler.close()

        result = yield _defer_to_thread(tc_operate, self.connectionLost)
        defer.returnValue((path, result))

    def _getTreeAndPath(self, path):
        """
        Return an already-authenticated share tree and path inside the share,
        for the path containing both the share name and path.
        """
        share, path = _split_share(path)

        tree = self._trees.get(share, None)
        if tree:
            return defer.succeed((tree, path))

        def tc_connect():
            share_address = '\\\\%s\\%s' % (
                self._connection.server_name, share)
            tree = TreeConnect(self._session, share_address)
            try:
                tree.connect(require_secure_negotiate=True)

            except SMBConnectionClosed as error:
                # Connection closed during operation.
                reactor.callFromThread(
                    self.connectionLost, ConnectionClosed('SMB client lost.'))
                raise SMBConnectionLost(force_unicode(error))

            except BadNetworkName as error:
                raise ServerException('Remote share not found: "%s". %s' % (
                    share_address, force_unicode(error)))

            except AccessDenied as error:
                raise ServerException('Access denied to "%s". %s' % (
                    share_address, force_unicode(error)))

            self._trees[share] = tree
            return tree, path

        return _defer_to_thread(tc_connect, self.connectionLost)


@implementer(ITransferStream)
class _SMBStreamBase(object):
    """
    Share code for a stream which is used to read/write data.

    The `path` is only used for event handling as the file is already
    opened and the path is not used for any file operation.
    Constructing the path from `handler` is possible, but complicated.
    """

    def __init__(self, handler, location, path, max_chunk, on_disconnect):
        self._handler = handler
        self._location = location
        self._path = path
        self._offset = 0
        self._max_chunk = max_chunk
        self._connectionLost = on_disconnect

    @property
    def path(self):
        return self._path

    @property
    def location(self):
        return self._location

    def close(self):
        """"
        Close the stream.
        """
        if not self._handler:
            return defer.fail(ServerException('SMB file is already closed.'))

        def tc_close():
            self._handler.close()
            self._handler = None

        return _defer_to_thread(tc_close, self._connectionLost)


@implementer(ITransferStream)
class SMBWriteStream(_SMBStreamBase):
    """
    A file/stream which is used to write data.
    """

    @property
    def mode(self):
        """
        Always in writing.
        """
        return 'writing'

    def read(self, length):
        """
        No read operation.
        """
        return defer.fail(ServerException('SMB stream opened for write-only.'))

    def write(self, data):
        """
        Write data.
        """
        if not self._handler:
            return defer.fail(ServerException('SMB write after close.'))

        def tc_write():
            """
            Called from the thread to send the data to the server.
            """
            # FIXME:4986:
            # Add a timeout if it takes longer to send the data.
            # FIXME:4983:
            # See if we can optimize the SMB credit usage.
            # Avoid writing too much.
            left_data = data
            while left_data:
                try:
                    # Only write the up the negotiated size.
                    to_write = left_data[:self._max_chunk]
                    written = self._handler.write(
                        to_write, offset=self._offset)
                    self._offset += written
                    left_data = left_data[written:]
                except SMBException as error:
                    if 'credits are available' not in error.args[0]:
                        # Not a credit issue.
                        raise

                    if self._max_chunk < 2:
                        # We don't know how to read less.
                        raise error

                    # Try to write half of the data.
                    self._max_chunk = self._max_chunk // 2

        return _defer_to_thread(tc_write, self._connectionLost)


@implementer(ITransferStream)
class SMBReadStream(_SMBStreamBase):
    """
    A file/stream which is used to read data.
    """

    @property
    def mode(self):
        """
        Always in reading.
        """
        return 'reading'

    def read(self, length):
        """
        Return data of maximum `length`.
        """
        if not self._handler:
            return defer.fail(ServerException('SMB read after close.'))

        # FIXME:4986:
        # Add a timeout if it takes longer to receive the data.

        # FIXME:4983:
        # See if we can optimize the SMB credit usage.
        # Avoid reading too much.
        length = min(length, self._max_chunk)

        def tc_read():
            try:
                result = self._handler.read(self._offset, length)
            except EndOfFile:
                # We reached the end of file.
                return b''

            self._offset += len(result)
            return result

        return _defer_to_thread(tc_read, self._connectionLost)

    def write(self, data):
        """
        No write operation.
        """
        return defer.fail(ServerException('SMB stream opened for read-only.'))


def _defer_to_thread(callback, on_disconnect):
    """
    Helper to call `callback` in a thread and call `on_disconnect` if the
    connection was lost.
    """
    def eb_connection_lost(failure):
        """
        Handle the case when the connection was lost.
        """
        failure.trap(SMBConnectionClosed)
        on_disconnect(ConnectionClosed('SMB client lost.'))
        # Convert it into our own error.
        message = force_unicode(failure.value).replace(
            'Received unexpected status from the server: ', '')

        raise SMBConnectionLost(force_unicode(message))

    deferred = threads.deferToThread(callback)
    deferred.addErrback(eb_connection_lost)
    return deferred


def _from_smb_time(smb_time):
    """
    Convert from SMB time to local unix unit time.

    The time fields are 100s of nanoseconds since 1601-01-01 UTC
    and we need to convert to nanoseconds since EPOCH.
    """
    return int((smb_time.get_value() - _EPOCH_FILETIME) * 100 // 1000000000)


def _split_share(path):
    """
    Return a tuple of (share-name, path-inside-share).

    `share-name` is returned in lower case.
    `path-inside-share` is returned using Windows separators and without the
    leading separator.
    """
    # The provided path is always absolute.
    # This is why we split in 3 to extract the share name and path name.
    parts = [p for p in path.split('/', 2) if p]
    if not parts:
        raise ServerException(
            'Path provided without a share name.')

    try:
        share, path = parts
    except ValueError:
        # We have a root path.
        share = parts[0]
        path = ''

    # No SMB path should start or end with the delimiter.
    # Samba and Azure are happy with '/' separators.
    # Windows Share wants '\' as separator.
    path = path.strip('/').replace('/', '\\')

    # Share names are case insensitive.
    share = share.lower()

    return share, path


def _tc_query_info(handler, info_class):
    """
    Sends a QUERY INFO request and return the parsed response.

    This is designed to be called inside a thread.
    """
    info_obj = info_class()
    query_req = SMB2QueryInfoRequest()
    query_req['info_type'] = info_obj.INFO_TYPE
    query_req['file_info_class'] = info_obj.INFO_CLASS
    query_req['file_id'] = handler.file_id
    query_req['output_buffer_length'] = len(info_obj)
    query_req['flags'] = 0

    request = handler.connection.send(
        query_req,
        handler.tree_connect.session.session_id,
        handler.tree_connect.tree_connect_id)
    response = handler.connection.receive(request)

    query_resp = SMB2QueryInfoResponse()
    query_resp.unpack(response['data'].get_value())
    return query_resp.parse_buffer(info_class).fields


def _tc_set_info(handler, info_buffer):
    """
    Sends a SET INFO request to the server with the information class input
    and return the response.
    """
    set_req = SMB2SetInfoRequest()
    set_req['info_type'] = info_buffer.INFO_TYPE
    set_req['file_info_class'] = info_buffer.INFO_CLASS
    set_req['file_id'] = handler.file_id
    set_req['buffer'] = info_buffer

    request = handler.connection.send(
        set_req,
        handler.tree_connect.session.session_id,
        handler.tree_connect.tree_connect_id)
    response = handler.connection.receive(request)

    set_resp = SMB2SetInfoResponse()
    set_resp.unpack(response['data'].get_value())


def _tc_delete_path(handler, is_file):
    """
    Delete a file or a directory.

    `delete_type` is a CreateOptions to be used for either a file or a
    directory.

    Called from the thread.
    """
    if is_file:
        create_type = CreateOptions.FILE_NON_DIRECTORY_FILE
        file_attributes = SMBFileAttributes.FILE_ATTRIBUTE_NORMAL
    else:
        create_type = CreateOptions.FILE_DIRECTORY_FILE
        file_attributes = SMBFileAttributes.FILE_ATTRIBUTE_DIRECTORY

    handler.create(
        impersonation_level=ImpersonationLevel.Impersonation,
        desired_access=(
            # This works for both files and directory as
            # DirectoryAccessMask has the same values for
            # DELETE and FILE_WRITE_ATTRIBUTES.
            FilePipePrinterAccessMask.DELETE
            # Also open for write attributes to reset the
            # read-only attribute.
            | FilePipePrinterAccessMask.FILE_WRITE_ATTRIBUTES
            ),
        file_attributes=0,
        share_access=(
            ShareAccess.FILE_SHARE_DELETE
            | ShareAccess.FILE_SHARE_READ
            | ShareAccess.FILE_SHARE_WRITE
            ),
        # I don't know why this is needed, but maybe is needed for
        # create_options.
        create_disposition=CreateDisposition.FILE_OPEN,
        # FILE_DELETE_ON_CLOSE is requires to handle shared
        # folder deletion.
        # It should also not fail when a directory is already opened.
        create_options=(
            CreateOptions.FILE_OPEN_REPARSE_POINT
            | create_type
            | CreateOptions.FILE_DELETE_ON_CLOSE
            ),
        )

    # Reset the attributes before delete so that on Windows Share
    # we can also delete read-only directories.
    basic_info = file_info.FileBasicInformation()
    basic_info['creation_time'] = 0
    basic_info['last_access_time'] = 0
    basic_info['last_write_time'] = 0
    basic_info['change_time'] = 0
    basic_info['file_attributes'] = file_attributes
    _tc_set_info(handler, basic_info)

    # This is required in order to receive an error if the
    # directory is not empty.
    disposition_info = file_info.FileDispositionInformation()
    disposition_info['delete_pending'] = True
    _tc_set_info(handler, disposition_info)


def _get_attributes(file_attributes):
    """
    Helper to extract the attributes from FileBasicInformation.
    """
    if file_attributes.has_flag(
            SMBFileAttributes.FILE_ATTRIBUTE_DIRECTORY):
        is_folder = True
        is_file = False
        mode = stat.S_IFDIR | 0o111
    else:
        is_folder = False
        is_file = True
        mode = stat.S_IFREG

    if file_attributes.has_flag(SMBFileAttributes.FILE_ATTRIBUTE_READONLY):
        mode |= 0o444
    else:
        mode |= 0o666

    return is_file, is_folder, mode
