# Copyright (c) 2015 Adi Roiban.
# See LICENSE for details.
"""
(E)SMTP high level client.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from email.header import Header
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formataddr, formatdate, make_msgid
from io import BytesIO
from time import time

from twisted.internet import defer, reactor
from twisted.internet.error import ConnectError, ConnectionDone, TimeoutError
from twisted.internet.protocol import connectionDone
from twisted.mail import smtp

from chevah.server import force_unicode
from chevah.server.commons.exception import ServerException
from twisted.python.failure import Failure


class ESMTPClient(object):
    """
    High level API for sending messages over the ESMTP protocol.

    Supports only ascii username/password combinations.

    FIXME:3067:
    Twisted does not allow unicode usernames/passwords.
    """
    _reactor = reactor

    # For testing purposes.
    _boundary = None

    def __init__(
            self,
            host,
            port=25,
            username='',
            password='',
            retries=3,
            timeout=15,
            context_factory=None,
            ):
        self._host = host
        self._port = port
        self._username = username
        self._password = password
        self._retries = retries
        self._timeout = timeout
        self._context_factory = context_factory

    def _formatAddressesAsHeader(self, addresses):
        """
        Returns an email Header instance containing a comma separated list
        of mailboxes`.

        `Addresses` is a list of tuples in the form (name, email). ASCII
        only email addresses are supported.

        https://tools.ietf.org/html/rfc788#page-25
        http://www.rfc-base.org/txt/rfc-2821.txt
        """
        # FIXME:3067:
        # Implement support for non ASCII email name/address.
        charset = 'us-ascii'

        header = Header(charset=charset)
        for name, email in addresses:
            try:
                email.encode('ascii')
            except UnicodeEncodeError:
                raise ServerException(
                    'Non-ASCII email addresses are not supported.')
            address = (name, email)
            item = '%s,' % formataddr(address)
            header.append(item, charset=charset, errors='strict')
        return header

    def _buildContent(
            self,
            sender,
            to,
            subject,
            body,
            id=None,
            attachments=None,
            cc=(),
            bcc=(),
            ):
        """
        Returns plain text email data header and body as a utf-8 encoded
        byte string.

        Unicode subject and body values are supported and will be encoded
        as utf-8.
        """
        if not id:
            id = make_msgid()

        if not attachments:
            attachments = {}

        destinations = []
        for email in to:
            # FIXME:3068:
            # Named email addresses not supported.
            destinations.append(('', email))

        destinations_cc = []
        for email in cc:
            # FIXME:3068:
            # Named email addresses not supported.
            destinations_cc.append(('', email))

        destinations_bcc = []
        for email in bcc:
            # FIXME:3068:
            # Named email addresses not supported.
            destinations_bcc.append(('', email))

        charset = 'utf-8'
        message = MIMEMultipart(boundary=self._boundary)
        message['From'] = self._formatAddressesAsHeader([sender])
        message['To'] = self._formatAddressesAsHeader(destinations)

        # Use UTF-8 as default.
        subject_charset = charset
        try:
            # Try to see if we can simplify the subject as ASCII.
            subject.encode('ascii')
            subject_charset = 'ascii'
        except UnicodeEncodeError:
            """
            Nope. We don't have ascii.
            So keep the previous encoding.
            """

        message['Subject'] = Header(subject, charset=subject_charset)
        message['Date'] = formatdate(time(), localtime=True)
        message['Message-Id'] = id

        # Only TO and CC headers are set in the message.
        if destinations_cc:
            message['CC'] = self._formatAddressesAsHeader(destinations_cc)

        message_body = MIMEText(body.encode(charset), 'plain', charset)

        message.attach(message_body)

        for name, attachment in attachments.items():
            attachment.seek(0)
            part = MIMEApplication(attachment.read(), Name=name)
            part['Content-Disposition'] = (
                'attachment; filename="%s"' % (name,))
            message.attach(part)

        return message.as_string()

    def send(
        self,
        sender,
        to,
        subject,
        body,
        id=None,
        attachments=None,
        cc=(),
        bcc=(),
            ):
        """
        Sends email with `subject`, `body` and (optionally) `attachments`
        to `to`, from specified `sender`.

        `sender` is a tuple of strings in the format
        ('SOME NAME', 'address@example.com').

        `to` a list of string email addresses in the format
        `address@example.com`.

        `attachments` is a dictionary each key of which contains a
        file name and each value contains file-like object
        (e.g. `io.BytesIO`).

        Final recipients are all to, cc and bcc addresses.

        Returns a deferred that will fire once the email is send and the
        server connection terminated. If email was delivered to all specified
        recipients the result will be `True`, otherwise `False`.

        For each call a new connection to the (E)SMTP server is made and
        subsequently dropped.
        """
        connector = None
        recipients = tuple(to) + tuple(cc) + tuple(bcc)

        def bb_disconnect(result_or_failure):
            """
            If not connected, return `result_or_failure` immediately;
            otherwise wait for disconnection from server to occur and return
            `result_or_failure`.
            """
            if not connector:
                return result_or_failure

            if not connector.transport:
                # Not yet connected.
                return result_or_failure

            if not connector.factory.currentProtocol:
                # Connected, but SMTP was not started or was already stopped.
                return result_or_failure

            # Wait for current connection to close.
            deferred = connector.factory.currentProtocol.disconnected_deferred
            deferred.addCallback(lambda _: result_or_failure)
            return deferred

        send_deferred = defer.Deferred()
        send_deferred.addCallback(self._cbSend, recipients)
        send_deferred.addErrback(self._ebSend, to, subject)
        send_deferred.addBoth(bb_disconnect)

        try:
            content = self._buildContent(
                sender,
                to,
                subject,
                body,
                id=id,
                attachments=attachments,
                cc=cc,
                bcc=bcc,
                )
        except ServerException as error:
            send_deferred.errback(error)
            return send_deferred

        factory = SenderFactory(
            username=self._username,
            password=self._password,
            # fromEmail is just the address.
            # the content will contain the full FROM header.
            fromEmail=sender[1],
            toEmail=recipients,
            file=BytesIO(content.encode('utf-8')),
            deferred=send_deferred,
            retries=self._retries,
            timeout=self._timeout,
            context_factory=self._context_factory
            )
        connector = self._reactor.connectTCP(
            self._host, self._port, factory, timeout=self._timeout)

        return send_deferred

    def _cbSend(self, result, recipients):
        """
        Called when email was delivered successfully.

        Returns `True` if mail was delivered to all recipients, `False`
        otherwise.

        FIXME:3071:
        Return detailed success/failure results.
        """
        successful, details = result

        if len(recipients) == successful:
            # Email delivered successfully to all recipients.
            return True

        return False

    def _ebSend(self, failure, to, subject):
        """
        Called when email delivery failed or there are connection errors.
        """
        failure.trap(ConnectError, TimeoutError, smtp.SMTPClientError)
        error = failure.value
        if isinstance(error, smtp.SMTPClientError):
            message = '%d %s' % (
                failure.value.code, force_unicode(failure.value.resp))
        else:
            message = force_unicode(failure.value)

        message = 'Failed to send email to "%s" with subject "%s": %s' % (
            ', '.join(to), subject, message)
        raise ServerException(message=message)


class ESMTPClientProtocol(smtp.ESMTPSender):
    """
    A ESMTP client protocol that fires an internal deferred once connection
    with server is terminated.
    """

    #: A deferred which is called when protocol is not connected.
    disconnected_deferred = defer.succeed(None)

    def connectionMade(self):
        """
        See: `SMTPSender`.
        """
        smtp.ESMTPSender.connectionMade(self)
        self.disconnected_deferred = defer.Deferred()

    def connectionLost(self, reason=connectionDone):
        """
        See: `SMTPClient`.
        """
        smtp.SMTPClient.connectionLost(self, reason)
        self.disconnected_deferred.callback(None)


class SenderFactory(smtp.ESMTPSenderFactory):
    """
    A factory for simple sending of emails.
    """
    protocol = ESMTPClientProtocol

    def __init__(
        self, username, password, fromEmail, toEmail, file, deferred,
        retries, timeout, context_factory
            ):
        encoded_username = username.encode('utf-8')
        encoded_password = b''
        if encoded_username:
            # We ignore the password if no username is defined.
            encoded_password = password.encode('utf-8')

        smtp.ESMTPSenderFactory.__init__(
            self,
            username=encoded_username,
            password=encoded_password,
            fromEmail=fromEmail,
            toEmail=toEmail,
            file=file,
            deferred=deferred,
            retries=retries,
            timeout=timeout,
            requireAuthentication=False,
            requireTransportSecurity=False,
            heloFallback=1,
            contextFactory=context_factory,
            )

    def _processConnectionError(self, connector, err):
        """
        This is upstream Twisted code with a fix for handling connection
        done error.

        FIXME:6502:
        https://twistedmatrix.com/trac/ticket/9340
        Should be removed once we use newer version of Twisted which has
        this fix.
        """
        self.currentProtocol = None
        if (self.retries < 0) and (not self.sendFinished):
            # Rewind the file in case part of it was read while attempting to
            # send the message.
            self.file.seek(0, 0)
            connector.connect()
            self.retries += 1
        elif not self.sendFinished:
            # If we were unable to communicate with the SMTP server a
            # ConnectionDone will be returned.
            # We want a more clear error message for debugging
            if err.check(ConnectionDone):
                value = smtp.SMTPConnectError(
                    -1, "Unable to connect to server.")
            else:
                value = err.value
            failure = Failure(value, exc_tb=err.tb)
            # Mark the current failure as handled.
            # Make sure to called this after the traceback was
            # moved to the new failure.
            err.cleanFailure()
            self.result.errback(failure)
