# Copyright (c) 2010 Adi Roiban.
# See LICENSE for details.
"""
Avatar used by SSH service.
"""

from __future__ import absolute_import
from __future__ import unicode_literals
from lazr.delegates import delegate_to
from twisted.conch.avatar import ConchUser
from zope.interface import Attribute, Interface, implementer

from chevah.server import force_unicode
from chevah.server.commons.interface import (
    IServiceAvatar,
    )
from chevah.server.identity.avatar import ClientAvatar
from chevah.server.ssh.session import FileTransferServerSSHSession
from chevah.server.ssh.sftp import SFTPSubsystem
from chevah.server.ssh.scp import SCPServerProtocol
from chevah.server.ssh.passwd import PasswdProtocol


class _ISSHAvatar(Interface):
    """
    An avatar used by the SSH protocol/connection for both client and server.
    """

    def gotGlobalRequest(requestType, data):
        """
        A global request was sent from the other side.

        Returns either:
            - 1: request accepted
            - 1, <data>: request accepted with request specific data
            - 0: request denied
        """


class ISSHServiceAvatar(IServiceAvatar, _ISSHAvatar):
    """
    Adapter of ServiceAvatar to be used by the SSH service.
    """

    ignore_create_permissions = Attribute(
        'True if file create requests will ignore requested permissions and '
        'attributes.'
        )

    ssh_delete_delay = Attribute(
        """
        Number of seconds after which the delete is actually performed.

        When 0, the delete is performed right away.
        When greater than 0, the delete response is return right away with a
        success, but the actual delete is delayed.
        """)

    def lookupSubsystem(name, data):
        """
        Return the handler for `name` subsystem.

        Return None if subsystem is not supported or not allowed.
        """

    def lookupExec(name, data):
        """
        Return the handler for `name` command.

        Return None if the command is not supported or not allowed.
        """


@implementer(_ISSHAvatar)
class GlobalRequestsMixin(object):
    """
    Mixin for handling SSH global requests. This needs to be mixed with an
    avatar.

    https://tools.ietf.org/html/rfc4254#page-4
    """
    _GLOBAL_REQUEST_DENIED = 0

    def gotGlobalRequest(self, requestType, data):
        """
        See: `ISSHAvatar`.

        It always responds with _GLOBAL_REQUEST_DENIED as we do not support
        global requests.
        """
        if (
            requestType and
            (b'keepalive' in requestType or b'keep-alive' in requestType)
                ):
            # We still reject the keepalive messages,
            # but we don't log an event.
            return self._GLOBAL_REQUEST_DENIED

        self.component.emitEvent(
            '30004', data={'request_type': force_unicode(requestType)})

        return self._GLOBAL_REQUEST_DENIED


@implementer(ISSHServiceAvatar)
@delegate_to(IServiceAvatar)
class SSHServerAvatar(ConchUser, GlobalRequestsMixin):
    """
    Adapter for `ServiceAvatar` into an SSH server avatar.
    """

    def __init__(self, context, service_configuration):
        ConchUser.__init__(self)
        self.context = context
        self._service_configuration = service_configuration

        self.listeners = {}  # dict mapping (interface, port) -> listener
        self.channelLookup.update({b'session': FileTransferServerSSHSession})

        self._subsystemLookup = {
            b'sftp': SFTPSubsystem,
            }
        self._execLookup = {
            b'scp': SCPServerProtocol,
            b'passwd': PasswdProtocol,
            }

    def lookupSubsystem(self, name, data):
        """
        See: ISSHAvatar.

        This is a strict lookup with support only for SFTP, and only when
        SFTP is enabled.
        """
        protocol_class = None

        if name == b'sftp' and self.service_configuration.sftp:
            protocol_class = self._subsystemLookup.get(b'sftp', None)

        if not protocol_class:
            return None

        return protocol_class(data, avatar=self)

    def lookupExec(self, name, data):
        """
        See: ISSHAvatar.

        Since this is the avatar, it does not start the service, if account
        has not access to it.
        """
        if name == b'scp' and not self.service_configuration.scp:
            return None

        if name == b'passwd' and not self.can_change_password:
            return None

        protocol_class = self._execLookup.get(name, None)

        if not protocol_class:
            return None

        return protocol_class(data, avatar=self)

    @property
    def service_configuration(self):
        """
        See: `IServiceAvatar`.
        """
        return self._service_configuration

    @property
    def ignore_create_permissions(self):
        """
        Return true if permissions should be ignored when creating
        files and folders.
        """
        return self.service_configuration.ignore_create_permissions

    @property
    def ssh_delete_delay(self):
        """
        See: ISSHServiceAvatar.

        This is extracted only from the group, and only for the primary group.
        """
        return self.context._groups[0].ssh_delete_delay

    def getUserGroupId(self):
        raise NotImplementedError('shell not available')

    def getOtherGroups(self):
        raise NotImplementedError('shell not available')

    def getShell(self):
        raise NotImplementedError('shell not available')

    def logout(self):
        """
        Action performed when users logs out.
        """

    def gotGlobalRequest(self, requestType, data):
        """
        See: `ISSHAvatar`.
        """
        # We don't want to use the Twisted behaviour.
        return GlobalRequestsMixin.gotGlobalRequest(self, requestType, data)


class SSHClientAvatar(ClientAvatar, GlobalRequestsMixin):
    """
    An avatar for a client connected to a remote server via SSH protocol.
    """
