# Copyright (c) 2012 Adi Roiban.
# See LICENSE for details.
"""
SSH service configuration.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
import struct

from chevah_compat import local_filesystem
from chevah_keycert.exceptions import KeyCertException
from chevah_keycert.ssh import Key
from zope.interface import implementer

from chevah.server import force_unicode
from chevah.server.commons.exception import ServerException
from chevah.server.commons.interface import ISSHServiceOptions
from chevah.server.commons.service_configuration import (
    ConfigurationWithConnections,
    )
from chevah.server.configuration.model import (
    WritableString,
    WritableCSVUniqueInsensitive,
    )


@implementer(ISSHServiceOptions)
class SSHServiceConfigurationSection(ConfigurationWithConnections):
    """
    Configurations for the service section within SSH service.
    """

    PROPERTIES_REQUIRING_RESTART = (
        # We don't require a restart on ssh_host_private_keys
        # since the service might use the global keys.
        # So we are observing the `ssh_host_keys` computed value.
        'ssh_host_keys',
        'dh_prime_size',
        'ssh_cipher_list',
        )

    banner = WritableString('banner')

    # FIXME:6521:
    # Should be removed in v6 and replaced with writtable string.
    @property
    def ssh_host_private_keys(self):
        """
        Return the direct `ssh_host_private_keys` or fallback to deprecated
        RSA and DSA keys.
        """
        value = self._proxy.getString(
            self._section_name, 'ssh_host_private_keys')
        if value:
            # We have a direct configuration.
            return value

        rsa_key = self._proxy.getString(self._section_name, 'rsa_private_key')
        dsa_key = self._proxy.getString(self._section_name, 'dsa_private_key')

        keys = []
        if rsa_key:
            keys.append(rsa_key)
        if dsa_key:
            keys.append(dsa_key)

        return '\n\n'.join(keys)

    @ssh_host_private_keys.setter
    def ssh_host_private_keys(self, value):
        """
        Set value and remove deprecated RSA and DSA
        """
        self._proxy.setStringOrNone(
            self._section_name, 'ssh_host_private_keys', value)

        # FIXME:6521:
        # Should be removed in v6.
        self._proxy.removeOption(self._section_name, 'rsa_private_key')
        self._proxy.removeOption(
            self._section_name, 'rsa_private_key_password')
        self._proxy.removeOption(self._section_name, 'dsa_private_key')
        self._proxy.removeOption(
            self._section_name, 'dsa_private_key_password')

    @property
    def ssh_host_keys(self):
        """
        Get the host keys to be used by this service.

        The local one, or the global one.
        """
        result = self.ssh_host_private_keys
        if result:
            return result
        return self.root.configuration.server.ssh_host_private_keys

    dh_prime_size = WritableCSVUniqueInsensitive(
        'dh_prime_size')

    def _get_ignore_create_permissions(self):
        '''Return the value for `ignore_create_permissions`.'''
        return self._proxy.getBoolean(
            self._section_name,
            'ignore_create_permissions',
            )

    def _set_ignore_create_permissions(self, value):
        '''Set the value for `ignore_create_permissions`.'''
        self._proxy.setBoolean(
            self._section_name,
            'ignore_create_permissions',
            value,
            )

    ignore_create_permissions = property(
        _get_ignore_create_permissions, _set_ignore_create_permissions)

    @property
    def sftp(self):
        """
        See: ISSHServiceOptions.
        """
        return self._proxy.getBoolean(self._section_name, 'sftp')

    @sftp.setter
    def sftp(self, value):
        self._proxy.setBoolean(self._section_name, 'sftp', value)

    @property
    def scp(self):
        """
        See: ISSHServiceOptions.
        """
        return self._proxy.getBoolean(self._section_name, 'scp')

    @scp.setter
    def scp(self, value):
        self._proxy.setBoolean(self._section_name, 'scp', value)

    ssh_cipher_list = WritableCSVUniqueInsensitive(
        'ssh_cipher_list'
        )


def load_ssh_key(configuration, usage, password=None):
    """
    Load a DSA/RSA key from configuration value.
    """
    if password:
        password = password.encode('utf-8')

    # The UI might have extra spaces.
    configuration = configuration.strip()

    # Try to load the configuration as text.
    key_type = Key._guessStringType(configuration.encode('utf-8'))
    if key_type:
        # Only try as string if we can detect the type.
        try:
            key = Key.fromString(
                data=configuration,
                passphrase=password,
                )
        except (KeyCertException, struct.error) as error:
            details = force_unicode(error)
            raise ServerException(
                'Failed to load content of %s. Invalid format. %s' % (
                    usage, details))

    else:
        # Fallback to path.
        path = configuration
        try:
            encoded_path = local_filesystem.getEncodedPath(path)
            key = Key.fromFile(filename=encoded_path, passphrase=password)
        except (OSError, IOError) as error:
            details = force_unicode(error)
            raise ServerException(
                'Failed to load %s from "%s". File not found. %s' % (
                    usage, path, details))
        except (KeyCertException, struct.error) as error:
            details = force_unicode(error)
            raise ServerException(
                'Failed to load %s from "%s". Invalid format. %s' % (
                    usage, path, details))

    if key.isPublic():
        raise ServerException(
            'Failed to load %s. '
            'Need a private key. '
            'Public key found instead.' % (usage,)
            )

    return key
