# Copyright (c) 2010-2015 Adi Roiban.
# See LICENSE for details.
"""
The modules hosts SSH service top level components.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

from collections import OrderedDict
import re

import pem
from twisted.cred.portal import Portal
from twisted.conch.ssh.connection import SSHConnection
from twisted.conch.ssh.factory import SSHFactory
from twisted.python import components

from chevah_keycert.ssh import Key, KeyCertException
from chevah_compat import local_filesystem
from chevah.server import force_unicode
from chevah.server.authentication.credential_checker import (
    ServiceCredentialChecker,
    )
from chevah.server.commons.exception import ServerException
from chevah.server.commons.factory import ServiceFactoryBase
from chevah.server.commons.interface import ISFTPServer
from chevah.server.commons.realm import ServiceRealm
from chevah.server.ssh.avatar import ISSHServiceAvatar
from chevah.server.ssh.protocol import (
    SSHServerProtocol,
    SSHOverflowTransport,
    SSHTransportBasePatcher,
    )
from chevah.server.ssh.sftp import ChevahSFTPShell
from chevah.server.ssh.userauth import SSHUserAuthServerService
from chevah.server.static.ssh_service import moduli

# Register the SFTPShell.
components.registerAdapter(ChevahSFTPShell, ISSHServiceAvatar, ISFTPServer)


class PublicBlobCache(object):
    """
    A temporary fix to cache the public while still exposing the .blob() API.

    FIXME:6502:
    See if this was included upstream.
    https://github.com/twisted/twisted/issues/10205
    """
    def __init__(self, blob):
        self._blob = blob

    def blob(self):
        return self._blob

    def __eq__(self, other):
        """
        It can be compared with another instance or with bytes.
        """
        if isinstance(other, PublicBlobCache):
            return self._blob == other.blob()
        return self._blob == other

    def __ne__(self, other):
        if isinstance(other, PublicBlobCache):
            return self._blob != other.blob()
        return self._blob != other


class SSHServerFactory(ServiceFactoryBase, SSHFactory):
    """
    SSH factory for file transfer server side protocols.
    """

    # List of supported prime number sizes.
    _PRIME_SIZES = [1024, 1536, 2048, 3072, 4096, 6144, 8192]

    protocol = SSHServerProtocol
    overflowProtocol = SSHOverflowTransport

    privateKeys = None
    # Cache of public keys, already serialized and ready to be used in KEX.
    publicKeys = None

    def __init__(self, service_configuration, on_error):
        super(SSHServerFactory, self).__init__(service_configuration, on_error)

        self.services = {
            b'ssh-userauth': SSHUserAuthServerService,
            b'ssh-connection': SSHConnection,
            }

        # These should be private.
        # We use OrderedDict to keep the same order as the one defined
        # in the configuration file.
        self.privateKeys = OrderedDict()
        self.publicKeys = {}

        ssh_host_keys = self._service_configuration.ssh_host_keys

        keys = _load_ssh_private_keys(
            configuration=ssh_host_keys,
            )

        for key in keys:
            # Set all private host keys.
            for alg_name in key.supportedSignatureAlgorithms():
                # Set all variants of a host key
                self.privateKeys[alg_name] = key
                # FIXME:6502:
                # See if it was included upstream.
                # https://github.com/twisted/twisted/issues/10205
                # Here we should cache NS(key.public().blob()),
                self.publicKeys[alg_name] = PublicBlobCache(key.public().blob())

        if not self.privateKeys:
            raise ServerException(
                'No private key configured.')

        self.primes = moduli.STATIC_MODULI
        dh_prime_size = list(self._service_configuration.dh_prime_size)

        if len(dh_prime_size) != 2:
            raise ServerException(
                'Only 2 values should be defined for "dh_prime_size".')

        try:
            self._dh_prime_minimum = int(dh_prime_size[0])
        except ValueError:
            raise ServerException(
                'The first value for "dh_prime_size" must be a number.')

        if self._dh_prime_minimum not in self._PRIME_SIZES:
            raise ServerException(
                'The minimum value for "dh_prime_size" is not supported.')

        self._dh_prime_preference = dh_prime_size[1]
        if self._dh_prime_preference not in ['minimum', 'ideal', 'maximum']:
            raise ServerException(
                'The second value of "dh_prime_size" is not supported.')

        # Build the ciphers configuration at start to trigger its validation.
        # ciphers.
        # Here `protocol` is a class.
        # We always allow any public host key, as for server-side connections
        # the public key is defined via `ssh_host_private_keys`.
        allowed_ciphers = (
            list(self.service_configuration.ssh_cipher_list) +
            [c.decode('ascii') for c in
                SSHTransportBasePatcher.supportedPublicKeys]
            )
        self._encryption = self.protocol.getEncryptionConfiguration(
            allowed_ciphers)

        credential_checker = [
            ServiceCredentialChecker(self.service_configuration)]
        realm = ServiceRealm(self.service_configuration)
        self.portal = Portal(realm, credential_checker)

    def _configureProtocol(self, protocol):
        """
        See: `ServiceFactoryBase`.
        """
        # FIXME:2580:
        # We need a formalized way of providing logout support instead of
        # dynamically patching the instance.
        protocol.avatar = None
        protocol.logoutFunction = lambda: None

        self._encryption.setFor(protocol)

        # We overwrite the public key algorithms configured with the
        # ssh_cipher_list as for server side we defined them via the
        # configured host keys.
        # The public keys have no preference order and we sort them to
        # help with testing.
        protocol.supportedPublicKeys = list(self.privateKeys.keys())

    def getDHPrime(self, ideal, minimum=None, maximum=None):
        """
        Return the prime number to be used in DH group key exchange.
        """
        if minimum is None:
            minimum = ideal
        if maximum is None:
            maximum = ideal

        if not minimum <= ideal <= maximum:
            raise ServerException(
                'Invalid prime number parameters sent by the peer.'
                )

        if (maximum < self._dh_prime_minimum):
            raise ServerException(
                'The remote peer doesn\'t support at least the minimum '
                'prime number size.'
                )

        if self._dh_prime_preference == 'minimum':
            target = max(self._dh_prime_minimum, minimum)
        elif self._dh_prime_preference == 'ideal':
            target = ideal
        else:
            target = maximum

        for candidate in self._PRIME_SIZES:
            if candidate > target:
                break
            selected = candidate

        return super(SSHServerFactory, self).getDHPrime(selected)


def _load_ssh_private_keys(configuration):
    """
    Load all the SSH private keys in configuration.

    Return a list of SSH private keys instances.
    """
    configuration = configuration.strip()
    result = []

    # See if configuration is text.
    pem_objects = pem.parse(configuration.encode('utf-8'))

    for line in _strip_pem(configuration):
        line = line.strip()
        if not line:
            # Ignore empty lines.
            continue
        segments = local_filesystem.getSegmentsFromRealPath(line)
        ssh_file = None
        try:
            # See if we have a file.
            ssh_file = local_filesystem.openFileForReading(segments)
            content = ssh_file.read().strip()
        except EnvironmentError as error:
            raise ServerException(
                'Failed to read SSH key file. %s' % (error,))
        finally:
            if ssh_file:
                ssh_file.close()

        result.append(_load_ssh_key(content))

    for member in pem_objects:
        result.append(_load_ssh_key(member.as_bytes()))

    return result


def _load_ssh_key(content):
    """
    Try to load the SSH key.
    """
    try:
        return Key.fromString(data=content)
    except KeyCertException as error:
        raise ServerException(
            'Failed to load SSH key. %s' % (force_unicode(error),))


_PEM_RE = re.compile(
    b"""----[- ]BEGIN (.+)[- ]----\r?
.+?\r?
----[- ]END \\1[- ]----\r?\n?""",
    re.DOTALL,
    )


def _strip_pem(value):
    """
    Return a list of lines from multi-line value that are not in PEM format.
    """
    value = value.encode('utf-8')
    return [d.decode('utf-8') for d in re.sub(_PEM_RE, b'', value).splitlines()]
