# Copyright (c) 2019 Adi Roiban.
# See LICENSE for details.
"""
passwd command execution handler.
"""
from __future__ import absolute_import
from __future__ import unicode_literals
import struct

from twisted.internet.protocol import Protocol

from chevah.server import force_unicode
from chevah.server.commons.exception import ServerException
from chevah.server.commons.session import RemoteConnectionMixin


class TextCommandProtocol(Protocol, RemoteConnectionMixin):
    """
    Shared code for all text commands available over SSH.

    This is a line receiver which works with multiple delimiters.
    It does not support mixed delimiters.

    Once we have more commands, we can move it from here.
    """
    DELIMITERS = [b'\n', b'\r']
    _buffer = b''
    MAX_LENGTH = 16384

    #: True when protocol is successfully connected.
    #: This is defined in BaseProtocol, but as integer.
    connected = False

    _NAME = 'not-yet-defined'
    _raw_arguments = b''

    def __init__(self, arguments, avatar):
        """
        `arguments` with which command is called.
        """
        self._avatar = avatar
        self._raw_arguments = arguments
        self.connected = False

    @property
    def avatar(self):
        """
        Proxy to transport/session avatar.
        """
        return self._avatar

    def connectionMade(self):
        """
        Called when process is started.
        """
        self.connected = True
        self.emitEvent(
            '30009',
            data={
                'command': self._NAME,
                'arguments': force_unicode(self._raw_arguments),
                },
            )

    def eofReceived(self):
        """
        Called when the other side has sent an EOF message.
        """
        self._endProcess(exit_code=1)

    def connectionLost(self, reason=None):
        """
        Called when process was stopped either by client or server side.
        """
        self.connected = False
        self.emitEvent('30010', data={'command': self._NAME})

    def dataReceived(self, data):
        """
        Translates bytes into lines, and calls lineReceived.
        """
        for delimiter in self.DELIMITERS:
            if delimiter not in data:
                continue
            else:
                break

        lines = (self._buffer + data).split(delimiter)
        self._buffer = lines.pop(-1)

        for line in lines:
            if self.transport.disconnecting:
                # this is necessary because the transport may be told to lose
                # the connection by a line within a larger packet, and it is
                # important to disregard all the lines in that packet following
                # the one that told it to close.
                return
            if len(line) > self.MAX_LENGTH:
                return self.lineLengthExceeded(line)
            else:
                self.lineReceived(line)

        if len(self._buffer) > self.MAX_LENGTH:
            return self.lineLengthExceeded(self._buffer)

    def lineReceived(self, line):
        """
        Called when a full line was received.
        """

    def lineLengthExceeded(self, line):
        """
        Called when the maximum line length has been reached.
        Override if it needs to be dealt with in some special way.
        """
        self._sendError('Line length exceeded.')

    def _sendError(self, message):
        """
        Send the error message and close.
        """
        self.emitEvent(
            '30066',
            data={'command': self._NAME, 'details': message})
        self._send(message)
        self._endProcess(exit_code=1)

    def _send(self, message):
        """
        Send a message to the remote peer.
        """
        # We force a new line, for plink or other clients which try to
        # force TTY over a non-TTY session.
        self.transport.write(b'\r\n' + message.encode('utf-8'))

    def _endProcess(self, exit_code):
        """
        End the current command execution.
        """
        self.connected = False
        self.transport.conn.sendRequest(
            self.transport,
            b'exit-status',
            struct.pack('>L', exit_code),
            )
        self.transport.loseConnection()


class PasswdProtocol(TextCommandProtocol):
    """
    In memory handling of the `passwd` command use to change the password
    for a user.

    The transport used by the protocol is a SSHSession which initializes the
    protocol.
    """
    # Public name of this command.
    _NAME = 'passwd'

    def __init__(self, arguments, avatar):
        """
        `arguments` with which command is called.
        """
        super(PasswdProtocol, self).__init__(arguments, avatar)
        self._current_password = None
        self._new_password = None

    def connectionMade(self):
        """
        Called when process is started.
        """
        super(PasswdProtocol, self).connectionMade()

        if self._raw_arguments:
            self._sendError('Command arguments are not supported.')
            return

        # Start the dialog.
        self._send('Current password: ')

    def lineReceived(self, line):
        """
        See: LineReceiver.
        """
        try:
            self._handleInput(line.strip())
        except ServerException as error:
            self._sendError(error.message)
            return

    def _handleInput(self, line):
        """
        Handle input from password update.
        """
        if not line:
            raise ServerException('Empty passwords are not supported.')

        try:
            line = line.decode('utf-8')
        except Exception:
            raise ServerException('Failed to process input.')

        if self._current_password is None:
            self._current_password = line
            self._send('New password: ')
            return

        if self._new_password is None:
            self._new_password = line
            self._send('Confirm new password: ')
            return

        confirmed_password = line

        if self._new_password != confirmed_password:
            raise ServerException('New passwords don\'t match.')

        if self.avatar.updatePassword(
            self._current_password,
            confirmed_password,
            event_emitter=self.emitEvent
                ):
            self._send('Password successfully updated.')
            self._endProcess(exit_code=0)
            return

        raise ServerException('Failed to update password.')
