# Copyright (c) 2010 Adi Roiban.
# See LICENSE for details.
"""
SSH top transport protocol implementation for both client and server.
"""
from __future__ import absolute_import
from __future__ import unicode_literals
import struct
from itertools import chain
from hashlib import sha1, sha256, sha512

from twisted.conch.error import ConchError
from twisted.conch.ssh import (
    connection as ssh_connection,
    userauth,
    transport as ssh_transport,
    _kex,
    )

from twisted.conch.ssh.common import NS, MP, getMP, ffs, getNS
from twisted.conch.ssh.transport import (
    DISCONNECT_CONNECTION_LOST,
    DISCONNECT_TOO_MANY_CONNECTIONS,
    messages,
    MSG_KEX_DH_GEX_GROUP,
    MSG_KEXDH_REPLY,
    MSG_UNIMPLEMENTED,
    SSHClientTransport,
    SSHCiphers,
    SSHTransportBase,
    SSHServerTransport,
    )
from twisted.internet import defer
from twisted.internet.error import ConnectionDone
from twisted.protocols import policies
from zope.interface import implementer

from chevah_keycert.exceptions import KeyCertException
from chevah_keycert.ssh import FingerprintFormats, Key

from chevah.server import force_unicode
from chevah.server.commons.constant import PRODUCT_NAME, TYPE_NAME, SSH_CIPHERS
from chevah.server.commons.exception import ServerError, ServerException
from chevah.server.commons.interface import IIdleTimeoutProtocol
from chevah.server.commons.session import RemoteConnectionMixin
from chevah.server.identity.avatar import PeerAvatar
from chevah.server.ssh.avatar import SSHClientAvatar
from chevah.server.ssh.userauth import SSHUserAuthClientService


MSG_ID_TO_TEXT = {}
for member in dir(ssh_transport):
    if not member.startswith('MSG_'):
        continue
    value = getattr(ssh_transport, member)
    MSG_ID_TO_TEXT[value] = member

for member in dir(ssh_connection):
    if not member.startswith('MSG_'):
        continue
    value = getattr(ssh_connection, member)
    MSG_ID_TO_TEXT[value] = member

for member in dir(userauth):
    if not member.startswith('MSG_'):
        continue
    value = getattr(userauth, member)
    MSG_ID_TO_TEXT[value] = member


# FIXME:6502:
# Send this patch upstream.
# https://github.com/twisted/twisted/issues/10202
@implementer(_kex._IFixedGroupKexAlgorithm)
class _DHGroup14SHA256(object):
    """
    Like diffie-hellman-group14-sha1 but with SHA256 hash.

    See RFC 8268 for SSH.
    See RFC 3526 Section 3 for Group 14 value.
    """
    preference = 12
    hashProcessor = sha256
    prime = int("""
        FFFFFFFF FFFFFFFF C90FDAA2 2168C234 C4C6628B 80DC1CD1
        29024E08 8A67CC74 020BBEA6 3B139B22 514A0879 8E3404DD
        EF9519B3 CD3A431B 302B0A6D F25F1437 4FE1356D 6D51C245
        E485B576 625E7EC6 F44C42E9 A637ED6B 0BFF5CB6 F406B7ED
        EE386BFB 5A899FA5 AE9F2411 7C4B1FE6 49286651 ECE45B3D
        C2007CB8 A163BF05 98DA4836 1C55D39A 69163FA8 FD24CF5F
        83655D23 DCA3AD96 1C62F356 208552BB 9ED52907 7096966D
        670C354E 4ABC9804 F1746C08 CA18217C 32905E46 2E36CE3B
        E39E772C 180E8603 9B2783A2 EC07A28F B5C55DF0 6F4C52C9
        DE2BCBF6 95581718 3995497C EA956AE5 15D22618 98FA0510
        15728E5A 8AACAA68 FFFFFFFF FFFFFFFF
        """.replace(" ", "").replace("\n", ""), 16)
    generator = 2


# FIXME:6502:
# Send this patch upstream.
# https://github.com/twisted/twisted/issues/10202
@implementer(_kex._IFixedGroupKexAlgorithm)
class _DHGroup15SHA512(object):
    """
    See RFC 8268 for SSH.
    See RFC 3526 Section 4 for Group 14 value.
    """

    preference = 11
    hashProcessor = sha512
    prime = int("""
        FFFFFFFF FFFFFFFF C90FDAA2 2168C234 C4C6628B 80DC1CD1
        29024E08 8A67CC74 020BBEA6 3B139B22 514A0879 8E3404DD
        EF9519B3 CD3A431B 302B0A6D F25F1437 4FE1356D 6D51C245
        E485B576 625E7EC6 F44C42E9 A637ED6B 0BFF5CB6 F406B7ED
        EE386BFB 5A899FA5 AE9F2411 7C4B1FE6 49286651 ECE45B3D
        C2007CB8 A163BF05 98DA4836 1C55D39A 69163FA8 FD24CF5F
        83655D23 DCA3AD96 1C62F356 208552BB 9ED52907 7096966D
        670C354E 4ABC9804 F1746C08 CA18217C 32905E46 2E36CE3B
        E39E772C 180E8603 9B2783A2 EC07A28F B5C55DF0 6F4C52C9
        DE2BCBF6 95581718 3995497C EA956AE5 15D22618 98FA0510
        15728E5A 8AAAC42D AD33170D 04507A33 A85521AB DF1CBA64
        ECFB8504 58DBEF0A 8AEA7157 5D060C7D B3970F85 A6E1E4C7
        ABF5AE8C DB0933D7 1E8C94E0 4A25619D CEE3D226 1AD2EE6B
        F12FFA06 D98A0864 D8760273 3EC86A64 521F2B18 177B200C
        BBE11757 7A615D6C 770988C0 BAD946E2 08E24FA0 74E5AB31
        43DB5BFC E0FD108E 4B82D120 A93AD2CA FFFFFFFF FFFFFFFF
        """.replace(" ", "").replace("\n", ""), 16)
    generator = 2


# FIXME:6502:
# Send this patch upstream.
# https://github.com/twisted/twisted/issues/10202
@implementer(_kex._IFixedGroupKexAlgorithm)
class _DHGroup16SHA512(object):
    """
    See RFC 8268 for SSH.
    See RFC 3526 Section 5 for Group 16 value.
    """

    preference = 10
    hashProcessor = sha512
    prime = int("""
        FFFFFFFF FFFFFFFF C90FDAA2 2168C234 C4C6628B 80DC1CD1
        29024E08 8A67CC74 020BBEA6 3B139B22 514A0879 8E3404DD
        EF9519B3 CD3A431B 302B0A6D F25F1437 4FE1356D 6D51C245
        E485B576 625E7EC6 F44C42E9 A637ED6B 0BFF5CB6 F406B7ED
        EE386BFB 5A899FA5 AE9F2411 7C4B1FE6 49286651 ECE45B3D
        C2007CB8 A163BF05 98DA4836 1C55D39A 69163FA8 FD24CF5F
        83655D23 DCA3AD96 1C62F356 208552BB 9ED52907 7096966D
        670C354E 4ABC9804 F1746C08 CA18217C 32905E46 2E36CE3B
        E39E772C 180E8603 9B2783A2 EC07A28F B5C55DF0 6F4C52C9
        DE2BCBF6 95581718 3995497C EA956AE5 15D22618 98FA0510
        15728E5A 8AAAC42D AD33170D 04507A33 A85521AB DF1CBA64
        ECFB8504 58DBEF0A 8AEA7157 5D060C7D B3970F85 A6E1E4C7
        ABF5AE8C DB0933D7 1E8C94E0 4A25619D CEE3D226 1AD2EE6B
        F12FFA06 D98A0864 D8760273 3EC86A64 521F2B18 177B200C
        BBE11757 7A615D6C 770988C0 BAD946E2 08E24FA0 74E5AB31
        43DB5BFC E0FD108E 4B82D120 A9210801 1A723C12 A787E6D7
        88719A10 BDBA5B26 99C32718 6AF4E23C 1A946834 B6150BDA
        2583E9CA 2AD44CE8 DBBBC2DB 04DE8EF9 2E8EFC14 1FBECAA6
        287C5947 4E6BC05D 99B2964F A090C3A2 233BA186 515BE7ED
        1F612970 CEE2D7AF B81BDD76 2170481C D0069127 D5B05AA9
        93B4EA98 8D8FDDC1 86FFB7DC 90A6C08F 4DF435C9 34063199
        FFFFFFFF FFFFFFFF
        """.replace(" ", "").replace("\n", ""), 16)
    generator = 2


# FIXME:6502:
# Send this patch upstream.
@implementer(_kex._IFixedGroupKexAlgorithm)
class _DHGroup17SHA512(object):
    """
    See RFC 8268 for SSH.
    See RFC 3526 Section 6 for Group 17 value.
    """

    preference = 9
    hashProcessor = sha512
    prime = int("""
        FFFFFFFF FFFFFFFF C90FDAA2 2168C234 C4C6628B 80DC1CD1 29024E08
        8A67CC74 020BBEA6 3B139B22 514A0879 8E3404DD EF9519B3 CD3A431B
        302B0A6D F25F1437 4FE1356D 6D51C245 E485B576 625E7EC6 F44C42E9
        A637ED6B 0BFF5CB6 F406B7ED EE386BFB 5A899FA5 AE9F2411 7C4B1FE6
        49286651 ECE45B3D C2007CB8 A163BF05 98DA4836 1C55D39A 69163FA8
        FD24CF5F 83655D23 DCA3AD96 1C62F356 208552BB 9ED52907 7096966D
        670C354E 4ABC9804 F1746C08 CA18217C 32905E46 2E36CE3B E39E772C
        180E8603 9B2783A2 EC07A28F B5C55DF0 6F4C52C9 DE2BCBF6 95581718
        3995497C EA956AE5 15D22618 98FA0510 15728E5A 8AAAC42D AD33170D
        04507A33 A85521AB DF1CBA64 ECFB8504 58DBEF0A 8AEA7157 5D060C7D
        B3970F85 A6E1E4C7 ABF5AE8C DB0933D7 1E8C94E0 4A25619D CEE3D226
        1AD2EE6B F12FFA06 D98A0864 D8760273 3EC86A64 521F2B18 177B200C
        BBE11757 7A615D6C 770988C0 BAD946E2 08E24FA0 74E5AB31 43DB5BFC
        E0FD108E 4B82D120 A9210801 1A723C12 A787E6D7 88719A10 BDBA5B26
        99C32718 6AF4E23C 1A946834 B6150BDA 2583E9CA 2AD44CE8 DBBBC2DB
        04DE8EF9 2E8EFC14 1FBECAA6 287C5947 4E6BC05D 99B2964F A090C3A2
        233BA186 515BE7ED 1F612970 CEE2D7AF B81BDD76 2170481C D0069127
        D5B05AA9 93B4EA98 8D8FDDC1 86FFB7DC 90A6C08F 4DF435C9 34028492
        36C3FAB4 D27C7026 C1D4DCB2 602646DE C9751E76 3DBA37BD F8FF9406
        AD9E530E E5DB382F 413001AE B06A53ED 9027D831 179727B0 865A8918
        DA3EDBEB CF9B14ED 44CE6CBA CED4BB1B DB7F1447 E6CC254B 33205151
        2BD7AF42 6FB8F401 378CD2BF 5983CA01 C64B92EC F032EA15 D1721D03
        F482D7CE 6E74FEF6 D55E702F 46980C82 B5A84031 900B1C9E 59E7C97F
        BEC7E8F3 23A97A7E 36CC88BE 0F1D45B7 FF585AC5 4BD407B2 2B4154AA
        CC8F6D7E BF48E1D8 14CC5ED2 0F8037E0 A79715EE F29BE328 06A1D58B
        B7C5DA76 F550AA3D 8A1FBFF0 EB19CCB1 A313D55C DA56C9EC 2EF29632
        387FE8D7 6E3C0468 043E8F66 3F4860EE 12BF2D5B 0B7474D6 E694F91E
        6DCC4024 FFFFFFFF FFFFFFFF
        """.replace(" ", "").replace("\n", ""), 16)
    generator = 2


# FIXME:6502:
# Send this patch upstream.
# https://github.com/twisted/twisted/issues/10202
@implementer(_kex._IFixedGroupKexAlgorithm)
class _DHGroup18SHA512(object):
    """
    See RFC 8268 for SSH.
    See RFC 3526 Section 7 for Group 18 value.
    """

    preference = 8
    hashProcessor = sha512
    prime = int("""
        FFFFFFFF FFFFFFFF C90FDAA2 2168C234 C4C6628B 80DC1CD1
        29024E08 8A67CC74 020BBEA6 3B139B22 514A0879 8E3404DD
        EF9519B3 CD3A431B 302B0A6D F25F1437 4FE1356D 6D51C245
        E485B576 625E7EC6 F44C42E9 A637ED6B 0BFF5CB6 F406B7ED
        EE386BFB 5A899FA5 AE9F2411 7C4B1FE6 49286651 ECE45B3D
        C2007CB8 A163BF05 98DA4836 1C55D39A 69163FA8 FD24CF5F
        83655D23 DCA3AD96 1C62F356 208552BB 9ED52907 7096966D
        670C354E 4ABC9804 F1746C08 CA18217C 32905E46 2E36CE3B
        E39E772C 180E8603 9B2783A2 EC07A28F B5C55DF0 6F4C52C9
        DE2BCBF6 95581718 3995497C EA956AE5 15D22618 98FA0510
        15728E5A 8AAAC42D AD33170D 04507A33 A85521AB DF1CBA64
        ECFB8504 58DBEF0A 8AEA7157 5D060C7D B3970F85 A6E1E4C7
        ABF5AE8C DB0933D7 1E8C94E0 4A25619D CEE3D226 1AD2EE6B
        F12FFA06 D98A0864 D8760273 3EC86A64 521F2B18 177B200C
        BBE11757 7A615D6C 770988C0 BAD946E2 08E24FA0 74E5AB31
        43DB5BFC E0FD108E 4B82D120 A9210801 1A723C12 A787E6D7
        88719A10 BDBA5B26 99C32718 6AF4E23C 1A946834 B6150BDA
        2583E9CA 2AD44CE8 DBBBC2DB 04DE8EF9 2E8EFC14 1FBECAA6
        287C5947 4E6BC05D 99B2964F A090C3A2 233BA186 515BE7ED
        1F612970 CEE2D7AF B81BDD76 2170481C D0069127 D5B05AA9
        93B4EA98 8D8FDDC1 86FFB7DC 90A6C08F 4DF435C9 34028492
        36C3FAB4 D27C7026 C1D4DCB2 602646DE C9751E76 3DBA37BD
        F8FF9406 AD9E530E E5DB382F 413001AE B06A53ED 9027D831
        179727B0 865A8918 DA3EDBEB CF9B14ED 44CE6CBA CED4BB1B
        DB7F1447 E6CC254B 33205151 2BD7AF42 6FB8F401 378CD2BF
        5983CA01 C64B92EC F032EA15 D1721D03 F482D7CE 6E74FEF6
        D55E702F 46980C82 B5A84031 900B1C9E 59E7C97F BEC7E8F3
        23A97A7E 36CC88BE 0F1D45B7 FF585AC5 4BD407B2 2B4154AA
        CC8F6D7E BF48E1D8 14CC5ED2 0F8037E0 A79715EE F29BE328
        06A1D58B B7C5DA76 F550AA3D 8A1FBFF0 EB19CCB1 A313D55C
        DA56C9EC 2EF29632 387FE8D7 6E3C0468 043E8F66 3F4860EE
        12BF2D5B 0B7474D6 E694F91E 6DBE1159 74A3926F 12FEE5E4
        38777CB6 A932DF8C D8BEC4D0 73B931BA 3BC832B6 8D9DD300
        741FA7BF 8AFC47ED 2576F693 6BA42466 3AAB639C 5AE4F568
        3423B474 2BF1C978 238F16CB E39D652D E3FDB8BE FC848AD9
        22222E04 A4037C07 13EB57A8 1A23F0C7 3473FC64 6CEA306B
        4BCBC886 2F8385DD FA9D4B7F A2C087E8 79683303 ED5BDD3A
        062B3CF5 B3A278A6 6D2A13F8 3F44F82D DF310EE0 74AB6A36
        4597E899 A0255DC1 64F31CC5 0846851D F9AB4819 5DED7EA1
        B1D510BD 7EE74D73 FAF36BC3 1ECFA268 359046F4 EB879F92
        4009438B 481C6CD7 889A002E D5EE382B C9190DA6 FC026E47
        9558E447 5677E9AA 9E3050E2 765694DF C81F56E8 80B96E71
        60C980DD 98EDD3DF FFFFFFFF FFFFFFFF
        """.replace(" ", "").replace("\n", ""), 16)
    generator = 2


_kex._DHGroup14SHA1.preference = 13


# FIXME:6502:
# See if this was added back in Twisted. Very unlikely.
# https://github.com/twisted/twisted/issues/9019
@implementer(_kex._IFixedGroupKexAlgorithm)
class _DHGroup1SHA1(object):
    """
    Diffie-Hellman key exchange with SHA-1 as HASH, and Oakley Group 2
    (1024-bit MODP Group). Defined in RFC 4253, 8.1.
    """

    preference = 14
    hashProcessor = sha1
    # Diffie-Hellman primes from Oakley Group 2 (RFC 2409, 6.2).
    prime = int(
        '17976931348623159077083915679378745319786029604875601170644'
        '44236841971802161585193689478337958649255415021805654859805036464405'
        '48199239100050792877003355816639229553136239076508735759914822574862'
        '57500742530207744771258955095793777842444242661733472762929938766870'
        '9205606050270810842907692932019128194467627007')
    generator = 2


# Add our extra key exchanges not supported upstream by twisted.
_kex._kexAlgorithms.update({
    b"diffie-hellman-group1-sha1": _DHGroup1SHA1(),
    b"diffie-hellman-group14-sha256": _DHGroup14SHA256(),
    b"diffie-hellman-group15-sha512": _DHGroup15SHA512(),
    b"diffie-hellman-group16-sha512": _DHGroup16SHA512(),
    b"diffie-hellman-group17-sha512": _DHGroup17SHA512(),
    b"diffie-hellman-group18-sha512": _DHGroup18SHA512(),
    })


class SSHTransportBasePatcher(object):
    """
    Code overwritten from SSHTransportBase.

    It is used by both client and server protocols.
    """

    # Initialize to have an attribute even before the KEX_INIT.
    # Set to `None` if there is no match.
    otherVersionString = None
    _kexAlg = 'not-negotiated-yet'
    keyAlg = 'not-negotiated-yet'
    incomingCompressionType = 'not-negotiated-yet'
    outgoingCompressionType = 'not-negotiated-yet'
    nextEncryptions = SSHCiphers(
        'not-negotiated-yet',
        'not-negotiated-yet',
        'not-negotiated-yet',
        'not-negotiated-yet',
        )

    supportedMACs = [
        b'hmac-sha2-512',
        # SHA2 384 is not supported by OpenSSH,
        # so we just skip it for now.
        # b'hmac-sha2-384',
        b'hmac-sha2-256',
        b'hmac-sha1',
        b'hmac-md5',
        # Remove the comment below to support clear SSH.
        # b'none',
        ]

    # See if the fix was accepted upstream.
    # We load our own KEX as they include legacy algorithms.
    supportedKeyExchanges = _kex.getSupportedKeyExchanges()

    # Base list of supported algorithms.
    # This is updated at runtime for client and server based on the
    # configuration.
    supportedPublicKeys = [
        b"ssh-ed25519",
        b'ecdsa-sha2-nistp256',
        b'ecdsa-sha2-nistp384',
        b'ecdsa-sha2-nistp521',
        b"rsa-sha2-256",
        b"rsa-sha2-512",
        b"ssh-rsa",
        b"ssh-dss",
        ]

    _avatar = None

    @classmethod
    def getEncryptionConfiguration(cls, configuration):
        """
        Return a new encryption configuration.
        """
        return _SSHEncryptionConfiguration(cls, configuration)

    @property
    def avatar(self):
        """
        See: `RemoteConnectionMixin.avatar`.
        """
        return self._avatar

    @avatar.setter
    def avatar(self, value):
        """
        Required to update the avatar after authentication.
        """
        self._avatar = value

    def debug(self, details):
        """
        Nothing is debugged by default.

        This is overwritten by the factory.
        """

    def dataReceived(self, data):
        """
        First, check for the version string (SSH-2.0-*).  After that has been
        received, this method adds data to the buffer, and pulls out any
        packets.

        @type data: L{bytes}
        @param data: The data that was received.

        FIXME:6502:
        Keep this code in sync with upstream.
        See if there is a way to send an upstream patch for debugging,
        maybe using Twisted logging.
        This is here to add support for debugging log.
        """
        self.buf = self.buf + data
        if not self.gotVersion:

            if len(self.buf) > 10 * 1024:
                self.sendDisconnect(
                    DISCONNECT_CONNECTION_LOST,
                    b'Peer version string longer than 10KB. '
                    b'Preventing a denial of service attack.'
                    )
                return

            if self.buf.find(b'\n', self.buf.find(b'SSH-')) == -1:
                return

            # RFC 4253 section 4.2 ask for strict `\r\n` line ending.
            # Here we are a bit more relaxed and accept implementations ending
            # only in '\n'.
            # https://tools.ietf.org/html/rfc4253#section-4.2
            self.debug('< ' + force_unicode(repr(self.buf)))
            lines = self.buf.split(b'\n')
            for p in lines:
                if p.startswith(b'SSH-'):
                    self.gotVersion = True
                    # Since the line was split on '\n' and most of the time
                    # it uses '\r\n' we may get an extra '\r'.
                    self.otherVersionString = p.rstrip(b'\r')
                    remoteVersion = p.split(b'-')[1]
                    if remoteVersion not in self.supportedVersions:
                        self._unsupportedVersionReceived(remoteVersion)
                        return
                    i = lines.index(p)
                    self.buf = b'\n'.join(lines[i + 1:])
        packet = self.getPacket()
        while packet:
            messageNum = ord(packet[0:1])
            message_text = MSG_ID_TO_TEXT.get(messageNum, messageNum)
            self.debug('< %s: %r' % (message_text, packet[1:100]))
            self.dispatchMessage(messageNum, packet[1:])
            packet = self.getPacket()

    def sendPacket(self, messageType, payload):
        """
        Patched to add support for debugging log.
        """
        message_text = MSG_ID_TO_TEXT.get(messageType, messageType)
        self.debug('> %s: %r' % (message_text, payload[:100]))
        return super(SSHTransportBasePatcher, self).sendPacket(
            messageType, payload)

    def loseConnection(self):
        """
        Lose the connection to the other side, sending a
        DISCONNECT_CONNECTION_LOST message.

        Patched to have a better error message.
        """
        self.sendDisconnect(
            DISCONNECT_CONNECTION_LOST,
            b"We have triggered the closed operation."
            )

    def dispatchMessage(self, messageNum, payload):
        """
        Send the received message to the appropriate method.

        It is called by `dataReceived` once a complete command was
        received.

        It is the root dispatcher for all the SSH commands.

        FIXME:6502:
        See if this needs to be updated.
        It is a reimplementation for Twisted upstream code.
        It adds support for emitting events and disconnecting on
        critical errors.

        @type messageNum: C{int}
        @type payload: c{str}
        """
        deferred = defer.maybeDeferred(
            self._dispatchMessage, messageNum, payload)

        def eb_dispatch_failure(failure):
            """
            Called when we failed to handle the SSH command.
            """
            # At this level we should not have any ServerError... and if
            # we do have one, it is still an ISE.
            if failure.check(ServerException):
                details = failure.value.message
                debug = ''
            else:
                details = force_unicode(failure.getBriefTraceback())
                debug = repr(payload)

            self.emitEvent(
                '30006',
                data={
                    'message_id': messageNum,
                    'payload': debug,
                    'details': details,
                    },
                )
            # This is a critical error and we can not continue.
            self.sendDisconnect(
                DISCONNECT_CONNECTION_LOST,
                'Internal error while processing %s.' % (messageNum,),
                )

        deferred.addErrback(eb_dispatch_failure)
        # Deferred is returned only to help with testing as there should
        # be no other code wanting to use this result.
        return deferred

    def _dispatchMessage(self, message_id, payload):
        """
        Low level handling of a SSH command.

        It always return a deferred, which under normal operation should not
        fail.

        Message IDs allocation convention:
        Transport layer protocol:
            1 to 19    Transport layer generic (e.g., disconnect, ignore,
                       debug, etc.)
            20 to 29   Algorithm negotiation
            30 to 49   Key exchange method specific (numbers can be reused
                       for different authentication methods)
        User authentication protocol:
            50 to 59   User authentication generic
            60 to 79   User authentication method specific (numbers can be
                       reused for different authentication methods)
        Connection protocol:
            80 to 89   Connection protocol generic
            90 to 127  Channel related messages
        Reserved for client protocols:
            128 to 191 Reserved
        Local extensions:
            192 to 255 Local extensions
        """
        if message_id < 50 and message_id in messages:
            # We have a supported transport layer command.

            # Messages values are in the format: MSG_COMMAND_NAME.
            # Convert from message ID to message name.
            message_name = messages[message_id][4:]
            transport_handler = getattr(self, 'ssh_%s' % message_name, None)
            if not transport_handler:
                # We don't have an implementation for that request.
                return self._sendUnimplementedResponse(message_id, payload)

            return transport_handler(payload)

        if not self.service:
            # It is not a SSH transport message and no service was
            # previously requested and initialized to handle the command.
            return self._sendUnimplementedResponse(message_id, payload)

        # A service was already initialized using the transport commands.
        # Forward the request to be handled by the service.
        return self.service.packetReceived(message_id, payload)

    def _sendUnimplementedResponse(self, message_id, payload):
        """
        Send a message to the other side that the last packet was not
        understood.

        It is based on `sendUnimplementedCommand` from Twisted, but it also
        emit an event and will always return a deferred.
        """
        self.emitEvent(
            '30005',
            data={
                'message_id': message_id,
                'payload': repr(payload),
                },
            )
        seqnum = self.incomingPacketSequence
        self.sendPacket(MSG_UNIMPLEMENTED, struct.pack('!L', seqnum))
        return defer.succeed(None)

    def connectionMade(self):
        """
        Called when a new connection was made.

        This is called before any handshake is done.
        """
        super(SSHTransportBasePatcher, self).connectionMade()
        # This is a hack to detect when handshake failed and to know
        # where to look for the negotiated encryption.
        self._initEncryption = self.currentEncryptions

    def connectionLost(self, reason):
        """
        Called when connection was lost.
        """
        if self.service:
            self.service.serviceStopped()

        # avatar member is used by Conch to signal that the connection
        # was authenticated ... otherwise we just have a connection
        # that was closed before finalizing the authentication.
        if self.avatar:
            self.logoutFunction()

    def sendDisconnect(self, reason, desc):
        """
        Patch upstream code to add our event.

        `desc` can be text and will be auto-encoded to UTF-8.
        """
        self.emitEvent('30050', data={'details': force_unicode(desc)})
        super(SSHTransportBasePatcher, self).sendDisconnect(reason, desc)

    def _getEncryptionEventData(self):
        """
        Return the event data members describing the SSH encryption in used
        by the connection.
        """
        # We should normally look at currentEncryption, but we use next
        # to also see what was negotiated and not agreed.
        # The way Twisted works now is that when encryption was agreed,
        # current is the same as next, and only during rekey it is different.
        ciphers = self.nextEncryptions

        def get_description(value):
            if value is None:
                # twisted.conch uses `None` when there is no match.
                return 'no-match'
            else:
                return force_unicode(value)

        return {
            'host_key': get_description(self.keyAlg),
            'key_exchange': get_description(self.kexAlg),
            'in_hmac': get_description(ciphers.inMACType),
            'in_cipher': get_description(ciphers.inCipType),
            'in_compression': get_description(self.incomingCompressionType),
            'out_hmac': get_description(ciphers.outMACType),
            'out_cipher': get_description(ciphers.outCipType),
            'out_compression': get_description(self.outgoingCompressionType),
            }

    # FIXME:6502:
    # This changes the Twisted implementation so that we have a better
    #
    # Client-initiated rekeying looks like this:
    #
    #  C> MSG_KEXINIT
    #  S> MSG_KEXINIT
    #  C> MSG_KEX_DH_GEX_REQUEST  or   MSG_KEXDH_INIT
    #  S> MSG_KEX_DH_GEX_GROUP    or   MSG_KEXDH_REPLY
    #  C> MSG_KEX_DH_GEX_INIT     or   --
    #  S> MSG_KEX_DH_GEX_REPLY    or   --
    #  C> MSG_NEWKEYS
    #  S> MSG_NEWKEYS
    #
    # Server-initiated rekeying is the same, only the first two messages are
    # switched.
    def ssh_KEXINIT(self, packet):
        """
        Called when we receive a MSG_KEXINIT message.  Payload::
            bytes[16] cookie
            string keyExchangeAlgorithms
            string keyAlgorithms
            string incomingEncryptions
            string outgoingEncryptions
            string incomingAuthentications
            string outgoingAuthentications
            string incomingCompressions
            string outgoingCompressions
            string incomingLanguages
            string outgoingLanguages
            bool firstPacketFollows
            unit32 0 (reserved)

        Starts setting up the key exchange, keys, encryptions, and
        authentications.  Extended by ssh_KEXINIT in SSHServerTransport and
        SSHClientTransport.

        @type packet: L{bytes}
        @param packet: The message data.

        @return: A L{tuple} of negotiated key exchange algorithms, key
        algorithms, and unhandled data, or L{None} if something went wrong.
        """
        self.otherKexInitPayload = bytes((ssh_transport.MSG_KEXINIT,)) + packet
        # This is useless to us:
        # cookie = packet[: 16]
        k = getNS(packet[16:], 10)
        strings, rest = k[:-1], k[-1]
        (
            kexAlgs,
            keyAlgs,
            encCS,
            encSC,
            macCS,
            macSC,
            compCS,
            compSC,
            langCS,
            langSC,
            ) = (s.split(b",") for s in strings)
        # These are the server directions
        outs = [encSC, macSC, compSC]
        ins = [encCS, macCS, compCS]
        if self.isClient:
            outs, ins = ins, outs  # Switch directions
        server = (
            self.supportedKeyExchanges,
            self.supportedPublicKeys,
            self.supportedCiphers,
            self.supportedCiphers,
            self.supportedMACs,
            self.supportedMACs,
            self.supportedCompressions,
            self.supportedCompressions,
            )
        client = (
            kexAlgs, keyAlgs, outs[0], ins[0], outs[1], ins[1], outs[2], ins[2])
        if self.isClient:
            server, client = client, server
        self.kexAlg = ffs(client[0], server[0])
        self.keyAlg = ffs(client[1], server[1])
        self.nextEncryptions = SSHCiphers(
            ffs(client[2], server[2]),
            ffs(client[3], server[3]),
            ffs(client[4], server[4]),
            ffs(client[5], server[5]),
            )
        self.outgoingCompressionType = ffs(client[6], server[6])
        self.incomingCompressionType = ffs(client[7], server[7])

        client_algorithms = b', '.join(list(
            chain.from_iterable(client))).decode('ascii')
        server_algorithms = b', '.join(list(
            chain.from_iterable(server))).decode('ascii')

        if self.keyAlg is None:
            self.sendDisconnect(
                ssh_transport.DISCONNECT_KEY_EXCHANGE_FAILED,
                b"Couldn't agree common public key algorithm."
                )
            self.emitEvent(
                '30013',
                data={
                    'type': 'public key',
                    'client_algorithms': client_algorithms,
                    'server_algorithms': server_algorithms,
                    })
            return

        if self.kexAlg is None:
            self.sendDisconnect(
                ssh_transport.DISCONNECT_KEY_EXCHANGE_FAILED,
                b"Couldn't agree common kex exchange algorithm."
                )
            self.emitEvent(
                '30013',
                data={
                    'type': 'key exchange',
                    'client_algorithms': client_algorithms,
                    'server_algorithms': server_algorithms,
                    })
            return

        if None in (self.outgoingCompressionType, self.incomingCompressionType):
            self.sendDisconnect(
                ssh_transport.DISCONNECT_KEY_EXCHANGE_FAILED,
                b"Couldn't agree common compression."
                )
            self.emitEvent(
                '30013',
                data={
                    'type': 'compression',
                    'client_algorithms': client_algorithms,
                    'server_algorithms': server_algorithms,
                    })
            return

        if None in self.nextEncryptions.__dict__.values():
            self.sendDisconnect(
                ssh_transport.DISCONNECT_KEY_EXCHANGE_FAILED,
                b"Couldn't agree common encryption."
                )
            self.emitEvent(
                '30013',
                data={
                    'type': 'encryption',
                    'client_algorithms': client_algorithms,
                    'server_algorithms': server_algorithms,
                    })
            return

        # We MUST disconnect if an extension negotiation indication ends
        # up being negotiated as a key exchange method (RFC 8308,
        # section 2.2).
        if self.kexAlg in (self._EXT_INFO_C, self._EXT_INFO_S):
            self.sendDisconnect(
                ssh_transport.DISCONNECT_KEY_EXCHANGE_FAILED,
                b'Extension negotiation found in common '
                b'key exchange algorithm.'
                )
            self.emitEvent(
                '30013',
                data={
                    'type': 'extension',
                    'client_algorithms': client_algorithms,
                    'server_algorithms': server_algorithms,
                    })
            return

        self._peerSupportsExtensions = (
            self._EXT_INFO_S if self.isClient else self._EXT_INFO_C
            ) in kexAlgs

        if self._keyExchangeState == self._KEY_EXCHANGE_REQUESTED:
            self._keyExchangeState = self._KEY_EXCHANGE_PROGRESSING
        else:
            self.sendKexInit()

        return kexAlgs, keyAlgs, rest  # For SSHServerTransport to use


@implementer(IIdleTimeoutProtocol)
class SSHServerProtocol(
        SSHTransportBasePatcher,
        SSHServerTransport,
        policies.TimeoutMixin,
        RemoteConnectionMixin,
        ):
    """
    SSH server protocol adapted for Chevah SSH service.
    """
    protocolVersion = '2.0-%s' % (PRODUCT_NAME,)
    ourVersionString = ('SSH-' + protocolVersion).strip().encode('utf-8')

    supportedSigAlgs = [
        b"ssh-ed25519",
        b'ecdsa-sha2-nistp256',
        b'ecdsa-sha2-nistp384',
        b'ecdsa-sha2-nistp521',
        b"rsa-sha2-256",
        b"rsa-sha2-512",
        b"ssh-rsa",
        b"ssh-dss",
        ]

    def __init__(self):
        super(SSHServerProtocol, self).__init__()
        self._avatar = None

    def dataReceived(self, data):
        """
        Called when new data is received on the connection.
        """
        self.resetTimeout()
        return super(SSHServerProtocol, self).dataReceived(data)

    def connectionMade(self):
        """
        Called when a new connection was made.

        This is called before any handshake is done.
        """
        self._avatar = PeerAvatar(
            self.transport.getPeer(),
            component=self.factory,
            )

        super(SSHServerProtocol, self).connectionMade()
        self.setTimeout(self.idle_connection_timeout)
        self.emitEvent('30014')

    def receiveError(self, reasonCode, description):
        """
        Called when we receive a disconnect error message from the other
        side.

        @param reasonCode: the reason for the disconnect, one of the
                           DISCONNECT_ values.
        @type reasonCode: L{int}
        @param description: a human-readable description of the
                            disconnection.
        @type description: L{str}
        """
        data = {
            'code': force_unicode(reasonCode),
            'details': force_unicode(description),
            }
        self.emitEvent('30082', data)

    def connectionLost(self, reason):
        """
        Called when the current command connection was lost.
        """
        self.setTimeout(None)
        try:
            return super(SSHServerProtocol, self).connectionLost(reason)
        finally:
            # We emit the event last, as while closing the connection,
            # we might get events for the connected sessions.
            data = self._getEncryptionEventData()
            data['client_version'] = force_unicode(self.otherVersionString)
            self.emitEvent('30015', data=data)

    def timeoutConnection(self):
        """
        Called when the connection times out.
        """
        self.sendDisconnect(
            DISCONNECT_CONNECTION_LOST,
            'Client connection idle time out.',
            )

    # FIXME:6502:
    # This is upstream code, with the only exception that we call
    # SSHTransportBasePatcher.
    def ssh_KEXINIT(self, packet):
        """
        Called when we receive a MSG_KEXINIT message.  For a description
        of the packet, see SSHTransportBase.ssh_KEXINIT().  Additionally,
        this method checks if a guessed key exchange packet was sent.  If
        it was sent, and it guessed incorrectly, the next key exchange
        packet MUST be ignored.
        """
        retval = SSHTransportBasePatcher.ssh_KEXINIT(self, packet)
        if not retval:  # Disconnected
            return

        kexAlgs, keyAlgs, rest = retval
        if ord(rest[0:1]):  # noqa:cover
            # Maybe this is the first_kex_packet_follows flag
            if (
                kexAlgs[0] != self.supportedKeyExchanges[0]
                or keyAlgs[0] != self.supportedPublicKeys[0]
                    ):
                self.ignoreNextPacket = True  # Guess was wrong

    def _ssh_KEXDH_INIT(self, packet):
        """
        Called to handle the beginning of a non-group key exchange.

        Unlike other message types, this is not dispatched automatically.  It
        is called from C{ssh_KEX_DH_GEX_REQUEST_OLD} because an extra check is
        required to determine if this is really a KEXDH_INIT message or if it
        is a KEX_DH_GEX_REQUEST_OLD message.

        The KEXDH_INIT payload::

                integer e (the client's Diffie-Hellman public key)

        We send the KEXDH_REPLY with our host key and signature.

        @type packet: L{bytes}
        @param packet: The message data.
        """
        clientDHpublicKey, foo = getMP(packet)
        self.g, self.p = _kex.getDHGeneratorAndPrime(self.kexAlg)
        self._startEphemeralDH()
        sharedSecret = self._finishEphemeralDH(clientDHpublicKey)
        # FIXME:6502:
        # This was already merged upstream
        # at https://github.com/twisted/twisted/issues/10203
        # but we are blocked on the other part.
        # https://github.com/twisted/twisted/issues/10205
        h = _kex.getHashProcessor(self.kexAlg)()
        h.update(NS(self.otherVersionString))
        h.update(NS(self.ourVersionString))
        h.update(NS(self.otherKexInitPayload))
        h.update(NS(self.ourKexInitPayload))

        # FIXME:6502:
        # See if this was included upstream.
        # https://twistedmatrix.com/trac/ticket/10205
        h.update(NS(self.factory.publicKeys[self.keyAlg].blob()))
        h.update(MP(clientDHpublicKey))
        h.update(self.dhSecretKeyPublicMP)
        h.update(sharedSecret)
        exchangeHash = h.digest()
        self.sendPacket(
            MSG_KEXDH_REPLY,
            NS(self.factory.publicKeys[self.keyAlg].blob()) +
            self.dhSecretKeyPublicMP +
            NS(self.factory.privateKeys[self.keyAlg].sign(exchangeHash)))
        self._keySetup(sharedSecret, exchangeHash)

    def _keySetup(self, sharedSecret, exchangeHash):
        """
        See SSHTransportBase._keySetup().
        """
        firstKey = self.sessionID is None
        SSHTransportBase._keySetup(self, sharedSecret, exchangeHash)
        # RFC 8308 section 2.4 says that the server MAY send EXT_INFO at
        # zero, one, or both of the following opportunities: the next packet
        # following the server's first MSG_NEWKEYS, or immediately preceding
        # the server's MSG_USERAUTH_SUCCESS.  We have no need for the
        # latter, so make sure we only send it in the former case.
        if firstKey:
            self.sendExtInfo(
                [(b"server-sig-algs", b",".join(self.supportedSigAlgs))]
                )

    def ssh_KEX_DH_GEX_REQUEST(self, packet):
        """
        Called when we receive a MSG_KEX_DH_GEX_REQUEST message.  Payload::
            integer minimum
            integer ideal
            integer maximum

        The client is asking for a Diffie-Hellman group between minimum and
        maximum size, and close to ideal if possible.  We reply with a
        MSG_KEX_DH_GEX_GROUP message.

        If we were told to ignore the next key exchange packet by ssh_KEXINIT,
        drop it on the floor and return.

        @type packet: L{bytes}
        @param packet: The message data.
        """
        if self.ignoreNextPacket:
            self.ignoreNextPacket = 0
            return
        self.dhGexRequest = packet
        min, ideal, max = struct.unpack('>3L', packet)
        # FIXME:6502:
        # See if this was accepted upstream.
        # https://github.com/twisted/twisted/issues/10204
        self.g, self.p = self.factory.getDHPrime(ideal, min, max)
        self._startEphemeralDH()
        self.sendPacket(MSG_KEX_DH_GEX_GROUP, MP(self.p) + MP(self.g))


class SSHOverflowTransport(SSHServerProtocol):
    """
    Simple SSH transport protocol that will disconnect users informing
    that server is full.
    """

    def connectionMade(self):
        """
        Called when a new command connection was made.
        """
        self._avatar = PeerAvatar(
            self.transport.getPeer(),
            component=self.factory,
            )

        self.emitEvent('30051')
        self.transport.write(b'%s\r\n' % (self.ourVersionString,))
        self.currentEncryptions = SSHCiphers(b'none', b'none', b'none', b'none')
        self.currentEncryptions.setKeys(b'', b'', b'', b'', b'', b'')
        self.sendDisconnect(
            DISCONNECT_TOO_MANY_CONNECTIONS,
            'To many server connections. Try again later.',
            )


class ConnectionDoneBeforeProtocol(ConnectionDone):
    """SSH handshake failed. Maybe no common algorithms or not a SSH server."""
    # The docstring is used as error message so keep it on one line
    # and without the final full stop.


class SSHClientProtocol(
    SSHTransportBasePatcher,
    SSHClientTransport,
    RemoteConnectionMixin,
        ):
    """
    SSH protocol / transport that will verify remote host for `fingerprint`,
    start ssh-auth service using `credentials` and then start the final
    `service`
    """

    def __init__(self, configuration, credentials, service):
        super(SSHClientProtocol, self).__init__()
        self._configuration = configuration
        self._credentials = credentials
        self._service = service
        self._location = credentials.creator
        self._avatar = SSHClientAvatar(credentials)

    def connectionMade(self):
        """
        Called when a new connection was made to the server.
        """
        super(SSHClientProtocol, self).connectionMade()
        self.emitEvent('30072', data={'name': self._location.name})

    def receiveError(self, reasonCode, description):
        """
        Called when we receive a disconnect error message from the other
        side.

        @param reasonCode: the reason for the disconnect, one of the
                           DISCONNECT_ values.
        @type reasonCode: L{int}
        @param description: a human-readable description of the
                            disconnection.
        @type description: L{str}
        """
        data = {
            'code': force_unicode(reasonCode),
            'details': force_unicode(description),
            }
        self.emitEvent('30083', data)

    def connectionLost(self, reason):
        """
        Called when the current connection was lost.
        """
        # We need to signal that connection is lost as soon as possible since
        # some services might check this during their stop process.
        self.connected = False
        data = {
            'name': self._location.name,
            'server_version': force_unicode(self.otherVersionString),
            }
        data.update(self._getEncryptionEventData())
        self.emitEvent('30073', data=data)
        if not self._service.deferred.called:
            # Connection was lost before initializing the high level protocol.
            reason = ConnectionDoneBeforeProtocol()

        self._service.serviceStopped(reason)

    def verifyHostKey(self, hostKey, fingerprint):
        """
        Called when server host key was received.

        See: SSHClientTransport
        """
        # FIXME:4694:
        # Here we receive any public key that we advertise as supported
        # in supportedPublicKeys.
        # Change supportedPublicKeys for set-on-first-connection.
        try:
            remote_key = Key.fromString(hostKey)
        except KeyCertException:
            # FIXME:4372:
            # Use shared `fail()` code to emit an event on corrupted keys.
            self._service.serviceStopped(ServerException(
                'Corrupted host key sent by the remote server.'))
            return defer.fail(ConchError(b'corrupt host key'))

        remote_sha256 = remote_key.fingerprint(
            FingerprintFormats.SHA256_BASE64)

        def fail(actual, expected):
            """
            Shared code to return an error in validating the server key.
            """
            # FIXME:4372:
            # Add information about negotiated ciphers.
            error = ServerError('30071', data={
                'name': self._location.name,
                'expected_fingerprint': expected,
                'actual_fingerprint': actual,
                # We add the details so that the event data is standard.
                'details': fingerprint.decode('ascii'),
                },
                avatar=self.avatar
                )
            self._service.serviceStopped(error)
            return defer.fail(ConchError('bad host key'))

        last_actual = ''
        last_expected = ''
        for server_identity in self._configuration.ssh_server_identity:

            if not server_identity:
                # We have an empty line.
                continue

            if server_identity == TYPE_NAME.SET_ON_FIRST_CONNECTION:
                # Accept any remote key.
                # FIXME:4694:
                # Save key type together with public key.
                self._configuration.ssh_server_identity = [remote_sha256]
                self._configuration._proxy.save()
                return defer.succeed(None)

            if len(server_identity) == 44:
                # we have SHA256 BASE64 configuration
                if remote_sha256 == server_identity:
                    return defer.succeed(None)
                last_actual = remote_sha256
                last_expected = server_identity

            if len(server_identity) == 47:
                # we have MD5 configuration
                remote_md5 = remote_key.fingerprint(
                    FingerprintFormats.MD5_HEX)

                if remote_md5 == server_identity:
                    return defer.succeed(None)
                last_actual = remote_md5
                last_expected = server_identity

            if len(server_identity) == 28:
                # we have SHA1 BASE64 configuration
                remote_sha1 = remote_key.fingerprint(
                    FingerprintFormats.SHA1_BASE64)
                if remote_sha1 == server_identity:
                    return defer.succeed(None)
                last_actual = remote_sha1
                last_expected = server_identity

        if last_actual:
            return fail(actual=last_actual, expected=last_expected)

        # As a fallback, try to load it as a multi-line key.
        try:
            configured_key = Key.fromString(
                '\n'.join(self._configuration.ssh_server_identity).strip())
        except KeyCertException:
            # Not a valid public key.
            return fail(actual=remote_sha256, expected=server_identity)

        if configured_key.public() == remote_key:
            return defer.succeed(None)

        expected_sha256 = configured_key.fingerprint(
            FingerprintFormats.SHA256_BASE64)
        return fail(actual=remote_sha256, expected=expected_sha256)

    def connectionSecure(self):
        """
        Called when the encryption is set up on initial connection or at rekey.

        This is the place after which the protocol is initialized and
        we can start using it.
        """
        if self.service is not None:
            # self.service is defined but SSHTransportBase and if it is not
            # None it means there is already a service running and this was
            # called after re-keying, so we don't do anything.

            # FIXME:4687:
            # This should be an event for both server and client side to
            # inform about (re)negotiated encryption.
            # Maybe try to call it at a lower level to share the code.
            self.emitEvent('30080')
            return

        self._auth_service = SSHUserAuthClientService(
            credentials=self._credentials,
            instance=self._service,
            )
        self.requestService(self._auth_service)

    # FIXME:6502:
    # This is upstream code, with the only exception that we call
    # SSHTransportBasePatcher.
    def ssh_KEXINIT(self, packet):
        """
        Called when we receive a MSG_KEXINIT message.  For a description
        of the packet, see SSHTransportBase.ssh_KEXINIT().  Additionally,
        this method sends the first key exchange packet.

        If the agreed-upon exchange is ECDH, generate a key pair for the
        corresponding curve and send the public key.

        If the agreed-upon exchange has a fixed prime/generator group,
        generate a public key and send it in a MSG_KEXDH_INIT message.
        Otherwise, ask for a 2048 bit group with a MSG_KEX_DH_GEX_REQUEST
        message.
        """
        if SSHTransportBasePatcher.ssh_KEXINIT(self, packet) is None:
            # Connection was disconnected while doing base processing.
            # Maybe no common protocols were agreed.
            return

        # Are we using ECDH?
        if _kex.isEllipticCurve(self.kexAlg):
            # Generate the keys
            self.ecPriv = self._generateECPrivateKey()
            self.ecPub = self.ecPriv.public_key()

            # DH_GEX_REQUEST_OLD is the same number we need.
            self.sendPacket(
                ssh_transport.MSG_KEX_DH_GEX_REQUEST_OLD,
                NS(self._encodeECPublicKey(self.ecPub))
                )
        elif _kex.isFixedGroup(self.kexAlg):
            # We agreed on a fixed group key exchange algorithm.
            self.g, self.p = _kex.getDHGeneratorAndPrime(self.kexAlg)
            self._startEphemeralDH()
            self.sendPacket(
                ssh_transport.MSG_KEXDH_INIT, self.dhSecretKeyPublicMP)
        else:
            # We agreed on a dynamic group. Tell the server what range of
            # group sizes we accept, and what size we prefer; the server
            # will then select a group.
            self.sendPacket(
                ssh_transport.MSG_KEX_DH_GEX_REQUEST,
                struct.pack(
                    "!LLL",
                    self._dhMinimalGroupSize,
                    self._dhPreferredGroupSize,
                    self._dhMaximalGroupSize,
                    ),
                )


class _SSHEncryptionConfiguration(object):
    """
    Handles the configuration of SSH protocol encryption.

    Keep the implementation in sync with reference capabilities.

    The primary usage is to set the encryption for a protocol instance.
    """

    # We want to have special names at the start.
    # We use `set` to reduce duplicate names.
    # This uses a list of `str`, but internally we work with bytes.
    SUPPORTED_CIPHERS = (
        ['all', 'fips', 'secure'] +
        [a.decode('utf-8') for a in
            SSHServerProtocol.supportedKeyExchanges] +
        [a.decode('utf-8') for a in SSHTransportBasePatcher.supportedMACs] +
        [a.decode('utf-8') for a in SSHServerProtocol.supportedCiphers] +
        [a.decode('utf-8') for a in SSHTransportBasePatcher.supportedPublicKeys]
        )

    def __init__(self, reference, ssh_cipher_list):
        """
        Configure allowed encryption ciphers, key exchange and hmac algorithms.

        Raise ServerException if ssh_cipher_list is not configured properly.
        """
        ssh_cipher_list = [c.encode('utf-8') for c in ssh_cipher_list]
        # We keep a reference to the lists to help with tests.
        # Keep it up to date with the documentation.
        self._secure_cipher_list = (
            # Host keys
            b'ssh-ed25519',
            b'ecdsa-sha2-nistp256',
            b'ecdsa-sha2-nistp384',
            b'ecdsa-sha2-nistp521',
            b'rsa-sha2-512',
            b'rsa-sha2-256',
            b'ssh-rsa',
            # Ciphers
            b'aes256-ctr', b'aes192-ctr', b'aes128-ctr',
            # MACs
            b'hmac-sha2-512', b'hmac-sha2-256',
            # KeyExchanges
            # See RFC for current recommendation (check updates).
            # This is based on:
            # https://tools.ietf.org/id/draft-ietf-curdle-ssh-kex-sha2-09.html
            b'diffie-hellman-group-exchange-sha256',
            b'diffie-hellman-group18-sha512',
            b'diffie-hellman-group17-sha512',
            b'diffie-hellman-group16-sha512',
            b'diffie-hellman-group15-sha512',
            b'diffie-hellman-group14-sha256',
            b'curve25519-sha256',
            b'curve25519-sha256@libssh.org',
            b'ecdh-sha2-nistp256',
            b'ecdh-sha2-nistp384',
            b'ecdh-sha2-nistp521',
            )
        # We keep two lists containing all known ciphers, MACs and
        # key exchange algorithms, dividing them into FIPS 140-2 approved
        # and rejected. This should contain all supported ciphers.
        # It is here to help us when adding new ciphers not forget
        # to classify it as FIPS approved or rejected.
        # For more information:
        # http://archimedesnewton.com/2010/12/20/fips-140-2-for-mere-mortals/
        self._fips_cipher_list = (
            b'ssh-ed25519',
            b'ecdsa-sha2-nistp256',
            b'ecdsa-sha2-nistp384',
            b'ecdsa-sha2-nistp521',
            b'rsa-sha2-512',
            b'rsa-sha2-256',
            b'ssh-rsa',
            # Ciphers
            b'aes256-ctr', b'aes256-cbc',
            b'aes192-ctr', b'aes192-cbc',
            b'aes128-ctr', b'aes128-cbc',
            b'3des-cbc',
            # MACs
            b'hmac-sha2-512', b'hmac-sha2-256', b'hmac-sha1',
            # KeyExchanges
            b'diffie-hellman-group-exchange-sha1',
            b'diffie-hellman-group-exchange-sha256',
            b'diffie-hellman-group1-sha1',
            b'diffie-hellman-group14-sha1',
            b'diffie-hellman-group14-sha256',
            b'diffie-hellman-group15-sha512',
            b'diffie-hellman-group16-sha512',
            b'diffie-hellman-group17-sha512',
            b'diffie-hellman-group18-sha512',
            b'curve25519-sha256',
            b'curve25519-sha256@libssh.org',
            b'ecdh-sha2-nistp256',
            b'ecdh-sha2-nistp384',
            b'ecdh-sha2-nistp521',
            )

        # This is the list with any other algorithm that we support,
        # and that is not part of secure or FIPS predefined lists.
        # The following ciphers are not supported by cryptography.
        # cast128-ctr blowfish-ctr 3des-ctr
        self._non_fips_cipher_list = (
            # Host keys
            b'ssh-dss',
            # Ciphers
            b'cast128-cbc', b'blowfish-cbc',
            # MACs
            b'hmac-md5',
            )

        all_supported = [c.encode('utf-8') for c in self.SUPPORTED_CIPHERS]
        unsupported = set(ssh_cipher_list) - set(all_supported)
        if unsupported:
            raise ServerException(
                'Unsupported ciphers in ssh_cipher_list: %s.'
                % ', '.join(sorted([c.decode('utf-8') for c in unsupported])))

        def filter_cipher_list(from_list, list_name):
            """
            Return a filtered list from `from_list` based on the
            `ssh_cipher_list` configuration option.
            This received a list of bytes and returns a list of bytes.
            """
            if b'all' in ssh_cipher_list:
                return from_list
            elif b'secure' in ssh_cipher_list:
                cipher_list = self._secure_cipher_list
            elif b'fips' in ssh_cipher_list:
                cipher_list = self._fips_cipher_list
            else:
                cipher_list = ssh_cipher_list

            ciphers = []
            for cipher in from_list:
                if cipher in cipher_list:
                    ciphers.append(cipher)

            if not ciphers:
                raise ServerException(
                    'Could not find a suitable %s from ssh_cipher_list.' % (
                        list_name)
                    )

            return ciphers

        self._supportedCiphers = (
            filter_cipher_list(
                reference.supportedCiphers,
                'cipher'
                )
            )
        self._supportedMACs = (
            filter_cipher_list(
                reference.supportedMACs,
                'hmac algorithm'
                )
            )
        self._supportedKeyExchanges = (
            filter_cipher_list(
                reference.supportedKeyExchanges,
                'kex algorithm'
                )
            )
        self._supportedPublicKeys = (
            filter_cipher_list(
                SSHTransportBase.supportedPublicKeys,
                'host keys'
                )
            )

    def setFor(self, instance):
        """
        Defines the supported ciphers to the protocol instance to be used
        when connecting,
        """
        instance.supportedCiphers = self._supportedCiphers
        instance.supportedMACs = self._supportedMACs
        instance.supportedKeyExchanges = self._supportedKeyExchanges
        instance.supportedPublicKeys = self._supportedPublicKeys


# Update the constants value, but keep the reference.
del SSH_CIPHERS[:]
SSH_CIPHERS.extend(_SSHEncryptionConfiguration.SUPPORTED_CIPHERS)
