# Copyright (c) 2014 Adi Roiban.
# See LICENSE for details.
"""
SCP request handler.

Reference: https://blogs.oracle.com/janp/entry/how_the_scp_protocol_works

Download a single file (remote) path -> (local) local-path
scp user@host:path local-path

Download a single file (remote) "path with space" -> (local)
scp user@host:"'path with space'" local-path

Download multiple files into a single local file... overwriting local files:
    * path1 -> local-path
    * path2 -> local-path
scp user@host:"path1 path2" local-file

Download multiple files into multiple local files:
    * path1 -> local-dir/path1
    * path2 -> local-dir/path2
scp user@host:"path1 path2" local-dir

Download multiple files with globbing:
    * path1 -> local-dir/path1
    * path2 -> local-dir/path2
scp user@host:path* local-dir

When multiple files are requested for download and some files are not found,
it just send a warning and continue processing remaining files.
"""
from __future__ import absolute_import
from __future__ import unicode_literals
import struct

from twisted.conch.ssh import filetransfer
from twisted.internet import defer, reactor
from twisted.internet.error import (
    ConnectionAborted,
    ProcessDone,
    ProcessTerminated,
    )
from twisted.internet.interfaces import IConsumer, IPushProducer
from twisted.internet.protocol import BaseProtocol
from twisted.protocols.basic import LineReceiver
from twisted.python.failure import Failure
from zope.interface import implementer

from chevah.server import force_unicode
from chevah.server.commons.exception import (
    AuthorizationException,
    ServerError,
    ServerException,
    )
from chevah.server.commons.interface import ISFTPServer
from chevah.server.commons.session import RemoteConnectionMixin
from chevah.server.event.handler import emit_error_details


@implementer(IPushProducer)
class SCPServerProtocol(LineReceiver, RemoteConnectionMixin):
    """
    The server side of the protocol which act like BSD SCP command line binary.

    It handles both source (-f) and sink (-t) mode.
    It ignores verbose or quiet arguments.

    Text lines are only received for sink mode. For source mode we only
    receive binary and send both text and binary.

    The transport used by the protocol is a SSHSession which initializes the
    protocol.

    It is a producer of uploaded data.

    It is not a consumer of downloaded data, since it delegates this to the
    transport (SSHSession).
    """
    ACK_OK = b'\x00'
    ACK_WARNING = b'\x01'
    ACK_ERROR = b'\x02'

    MODE_SOURCE = 'source'
    MODE_SINK = 'sink'

    delimiter = b'\n'

    _DEFAULT_OPTIONS = {
        'mode': None,
        'path': None,
        'recursive': False,
        }

    #: SFTPShell used for filesystem access and emitting fs related events.
    _shell = None
    #: Push producer which generates data for download requests.
    producer = None
    #: Consumer which received data for upload requests.
    _consumer = None
    #: List of files pending download request.
    _send_queue = None

    #: True when protocol is successfully connected.
    #: This is defined in BaseProtocol, but as integer.
    connected = False

    def __init__(self, arguments, avatar):
        """
        `arguments` with which command is called.
        """
        # Class use for initializing SCP file upload consumer.
        # Here to help with testing.
        self.SCPFileConsumer = SCPFileConsumer
        self._avatar = avatar
        self._raw_arguments = arguments
        self._options = self._parseArguments(arguments)

    @property
    def avatar(self):
        """
        Proxy to transport/session avatar.
        """
        return self._avatar

    def _internalError(self, error):
        """
        Called when an internal server error is produced.

        Make sure no code is called after this as it will end the process.
        """
        self.connected = False
        # Internal server error.
        emit_error_details(
            error, '30065', data={}, emitter=self.emitEvent)

        if isinstance(error, ServerException):
            message = error.message
        else:
            message = 'Internal server error.'

        self._sendStatus(self.ACK_ERROR, message)

    def _parseArguments(self, arguments):
        """
        Parse SCP arguments and return a dictionary with options.

        Fill values with None when they can not be parsed or are missing.
        """
        result = self._DEFAULT_OPTIONS.copy()

        arguments = arguments.strip()
        # Try with 'COMMANDS -- FILES
        parts = arguments.split(b' -- ', 1)

        if arguments.endswith(b'--'):
            # We have no files, only options.
            options = arguments[:-2]
            files = ''
        elif len(parts) != 2:
            # Try with COMMANDS FILES or COMMANDS 'FILES'
            parts = arguments.split(b' ', 1)
            options = []
            while parts[0].startswith(b'-'):
                options.append(parts[0])
                if len(parts) < 2:
                    # No more arguments.
                    parts = ''
                    break

                parts = parts[1].split(b' ', 1)

            files = b' '.join(parts).strip(b' \'"')
            options = b' '.join(options)
        else:
            options = parts[0]
            files = parts[1]

        if b'-t' in options:
            mode = self.MODE_SINK
        elif b'-f' in options:
            mode = self.MODE_SOURCE
        else:
            mode = None
        result['mode'] = mode

        if b'-r' in options:
            result['recursive'] = True

        # When path is missing or empty, assume request for current folder.
        path = files
        if not path:
            path = b'.'
        result['path'] = force_unicode(path)

        return result

    def _sendStatus(self, code=ACK_OK, message=''):
        """
        Send confirmation for previous command.

        3 response values are supported with binary value.

        * 0 (OK) continue
        * 1 (warning) continue with a message
        * 2 (error) end the connection).

        Messages 1 and 2 can be followed by a text message to be printed
        on the other side, followed by a new line character.
        The new line character is mandatory whether the text is empty or not.
        """
        response = self.ACK_OK
        message = message.encode('utf-8')

        if code != self.ACK_OK:
            response = code + message + self.delimiter

        self.transport.write(response)

        if code == self.ACK_ERROR:
            self._processEnded(Failure(ProcessTerminated(exitCode=1)))

    def connectionMade(self):
        """
        Called when process is started.
        """
        try:
            self._connectionMade()
            self.connected = True
        except Exception as error:
            self._internalError(error)

    def _connectionMade(self):
        """
        See: connectionMade.

        Wrapped code to handle internal server errors.
        """
        self._shell = ISFTPServer(self.avatar)
        encryption_data = (
            self.transport.conn.transport._getEncryptionEventData())
        self._shell.start('SCP', encryption_data)
        if self._options['mode'] == self.MODE_SOURCE:
            # Pipe all data data to the file transfer producer via the
            # rawDataReceived.
            self.setRawMode()
            self.producer = SCPFileProducer(self.transport, self._shell)
            self._startSource()
            return

        if self._options['mode'] == self.MODE_SINK:
            self.setLineMode()
            # Let source know that we are ready.
            self._sendStatus()
            return

        raise ServerException(
            'Bad arguments: %s' % force_unicode(self._raw_arguments))

    def eofReceived(self):
        """
        Called when the other side has sent an EOF message.
        """
        self._processEnded(reason=Failure(ProcessDone(0)))

    def connectionLost(self, reason=None):
        """
        Called when process was stopped either by client or server side.
        """
        # By the time the connection is closed, the stream should have been
        # already closed.
        self._stopConsumer(is_failure=True)
        self._stopProducer(reason)
        self.emitEvent('30064')

    def _stopProducer(self, reason):
        """
        Stop sending data to the connection as the connection was closed.
        """
        if not self.producer:
            return

        self.producer.connectionLost(reason)
        self.producer = None

    def _stopConsumer(self, is_failure=False):
        """
        Unregister from consumer and clear consumer.
        """
        if not self._consumer:
            return

        if is_failure:
            self._consumer.fail()

        self._consumer.unregisterProducer()
        self._consumer = None

    def lineReceived(self, line):
        """
        See: LineReceiver.

        Line mode is enabled only for sink mode and only for the initial
        file header part.
        """
        try:
            self._sink_header(line)
        except Exception as error:
            # Unhanded error while processing headers.
            self._consumer = None
            self._internalError(error)

    def rawDataReceived(self, data):
        """
        See: LineReceiver.
        """
        if self.transport.closing:
            # Ignore data received while closing the transport.
            # This can happen since SCP has no method to signal an error
            # in the middle of a transfer, so we just close the connection
            # while remote peer continues to send us data.
            return

        if self._options['mode'] == self.MODE_SOURCE:
            # FIXME:2135:
            # Here we get confirmation from the remote peer.
            # Confirmation that the file was opened and that it was received.
            if data == b'\x00':
                # FIXME:2135:
                # We should forward this to the file producer so that it will
                # wait for confirmation, before closing the process or
                # proceeding to the next file.
                pass
            elif data == b'\x00\x00':
                # FIXME:2135:
                # We should forward this to the file producer so that it will
                # wait for confirmation, before closing the process or
                # proceeding to the next file.
                pass
            else:
                # FIXME:2135:
                # Client should send 0 to confirm that it has received the
                # file and then we should continue with next file.
                # Have no idea what to do when client send another message.
                pass
            return

        # If not in source mode, we are an sink and should write to consumer.
        self._consumer.write(data)

    def _startSource(self):
        """
        Start SCP files transfer in source mode.
        """
        # FIXME:2127:
        # For now it supports transferring a single file.
        # Files to send is a list to support future recursive, globbing or
        # multiple files.
        self._send_queue = [self._options['path']]
        return self._sendNextPath()

    def _sendNextPath(self):
        """
        Transfer next path in queue.

        It is only called when transfer starts and after the client confirms
        that it has received the previous path (file or folder).

        Before sending the content, the protocol require to first send
        a message describing what is sent:

        * For file: Cmmmm <length> <filename>
        * For folder: Dmmmm 0 <dirname>
        * For end of folder: E
        """
        try:
            next_file = self._send_queue.pop()
        except IndexError:
            # All files were successfully transferred.
            self._processEnded(reason=Failure(ProcessDone(0)))
            return defer.succeed(None)

        # FIXME:2127:
        # We should first check that path is really a file and not a folder
        # and for folders expand the list for recursive transfer.
        return self._sendFile(next_file)

    def _sendFile(self, path):
        """
        Send file at `path`.

        Deferred is only returned to help with testing.
        """
        deferred = self.producer.beginFileTransfer(path)
        deferred.addCallback(self._cbSendFile)
        deferred.addErrback(self._ebSendFileWarning)
        deferred.addErrback(self._ebSendFileError)
        return deferred

    def _cbSendFile(self, result):
        """
        Called when file was successfully sent.
        """
        # Confirm that file was sent.
        self._sendStatus()
        self._sendNextPath()

    def _ebSendFileWarning(self, failure):
        """
        Called when failed to send file, but the error is not critical so
        transfer can continue with next file.
        """
        failure.trap(SCPFileProducerWarning)
        self._sendStatus(self.ACK_WARNING, failure.value.message)
        self._sendNextPath()

    def _ebSendFileError(self, failure):
        """
        Called when failed to send file and the failure is not
        a warning.
        """
        self._sendStatus(self.ACK_ERROR, force_unicode(failure.value.args[0]))

    def _processEnded(self, reason):
        """
        Handle SCP process termination by setting an exit code.
        """
        self.transport.conn.sendRequest(
            self.transport,
            b'exit-status',
            struct.pack('>L', reason.value.exitCode),
            )
        self.transport.loseConnection()

    def _sink_header(self, line):
        """
        Process headers received in sink mode.
        """
        # For now we ignore timestamp requests.
        if line.startswith(b'T'):
            self._sendStatus()
            return

        # Process final message before raw data.
        if line.startswith(b'C'):
            try:
                mode, size, local_path = self._parseFileCommandLine(
                    line[1:])
            except ServerException as error:
                self._sendStatus(self.ACK_WARNING, error.message)
                return

            try:
                self._consumer = self.SCPFileConsumer(
                    shell=self._shell,
                    mode=mode,
                    size=size,
                    path=local_path,
                    )
                # All good, ready to receive raw data.
                self._consumer.registerProducer(self, streaming=True)
                self.setRawMode()
                self._sendStatus()
            except (
                AuthorizationException,
                ServerError,
                EnvironmentError,
                    ):
                self._consumer = None
                # These errors should already be logged by the shell.
                self._sendStatus(
                    self.ACK_WARNING, 'Failed to open file for writing.')

            return

        # Inform remote peer that we don't know about the message.
        self._sendStatus(
            self.ACK_WARNING, force_unicode(b'Unknown message: ' + line))

    def _parseFileCommandLine(self, line):
        """
        Return a tuple of (mode, size, path) from file command line.

        Line is without trailing 'C'.
        """
        # We use a more secure default mode.
        mode = 0o700
        size = 0
        local_path = None

        parts = line.split(b' ', 2)
        if len(parts) != 3:
            raise ServerException(
                'Bad format: {}.'.format(force_unicode(line)))

        try:
            mode = int(parts[0], 8)
        except ValueError:
            raise ServerException(
                'Bad mode: {}.'.format(force_unicode(parts[0])))

        try:
            size = int(parts[1], 10)
        except ValueError:
            raise ServerException('Bad size: %s.' % parts[1])

        remote_path = force_unicode(parts[2])
        # FIXME:2152:
        # When self._options['path'] is a folder the local_path should
        # be joined with requested remote path.
        # For now, we do a naive guess to see if path is a folder, without
        # asking the filesystem.
        if self._options['path'] == '.':
            local_path = remote_path
        elif self._options['path'].endswith('/'):
            local_path = self._options['path'] + remote_path
        else:
            local_path = self._options['path']

        return mode, size, local_path

    def stopProducing(self):
        """
        Called by consumer when all data was received.
        """
        self._stopConsumer()
        # Enter in line mode to receive header for next file.
        self.setLineMode()
        # Confirm remote source that all data was received.
        self._sendStatus()

    def failedConsuming(self, message):
        """
        Called by consumer when something went wrong.
        """
        # No event is emitted as there should already be a event from
        # the lower level shell.
        self._stopConsumer(is_failure=True)
        self._sendStatus(self.ACK_ERROR, message)


class SCPFileProducerWarning(ServerException):
    """
    Warning raised during SCP protocol.

    A warning does not closes the connection.
    """


class SCPFileProducerError(ServerException):
    """
    Error raised by SCPFileProducer.
    """


@implementer(IPushProducer)
class SCPFileProducer(BaseProtocol):
    """
    Send content of local file in chunks to avoid blocking the process.
    """
    CHUNK_SIZE = 2 ** 14

    #: SFTPFile which is transferred.
    _file = None
    #: Cursor up until which the file was read.
    _cursor = 0

    #: A deferred which fires when file is transferred. with either error
    #: or failu0-re.
    _result = None

    _scheduler = reactor

    delimiter = b'\n'

    def __init__(self, consumer, shell):
        self._consumer = consumer
        self._shell = shell
        self._path = None
        self._file = None
        self._result = defer.Deferred()
        self._result.addBoth(self._bbClose)
        self._send_delayed_callback = None

    def beginFileTransfer(self, path):
        """
        Start sending data from file at `path`.

        Return a deferred which is called once file is completely transferred.
        """
        if self._path:
            raise RuntimeError('Only one file can be sent at a time.')

        self._path = path
        try:
            # FIXME:2135:
            # We should open the local file, only after the remote peer
            # has confirmed that it can receive the file.
            self._file = self._shell.openFile(
                self._path, filetransfer.FXF_READ, {})
        except Exception:
            self._result.errback(SCPFileProducerWarning(
                'Failed to open file for reading: %s' % self._path))
            return self._result

        try:
            attributes = self._file.getAttrs()
        except Exception:
            self._result.errback(SCPFileProducerWarning(
                'Failed to get file attributes: %s' % self._path))
            return self._result

        self._consumer.registerProducer(self, streaming=True)

        # Send file properties.
        # Only last 4 digits are required.
        # And we need to have exactly 4.
        mode = "{:o}".format(attributes['permissions'])[-4:].zfill(4)

        file_description_message = 'C%s %d %s' % (
            mode, attributes['size'], self._path)
        self._sendLine(file_description_message)

        # FIXME:2135:
        # We should start sending the data only after description line was
        # confirmed.
        self._scheduleNextChunk()

        return self._result

    def connectionLost(self, reason=None):
        """
        Called when the remote peer was disconnected.
        """
        if self._file:
            # Close the opened file with a failure.
            self._file.fail()
            self._file = None

        self.pauseProducing()

        # The file might have been already fully read.
        if not self._result.called:
            self._result.errback(ConnectionAborted(
                'Peer closes the connection before transfer end.'))

    def _clSendChunk(self):
        """
        Send a new chunk of file.
        """
        try:
            data = self._file.readChunk(self._cursor, self.CHUNK_SIZE)
        except Exception:
            self._result.errback(SCPFileProducerError(
                'Failed to read file data: %s' % self._path))
            return

        if not data:
            # Trigger callback as we are ready.
            # This should also trigger the file closing.
            self._result.callback(None)
            return
        else:
            self._cursor += len(data)
            self._consumer.write(data)
            self._scheduleNextChunk()

    def _bbClose(self, result_or_failure):
        """
        Close the file.
        """
        self._consumer.unregisterProducer()
        path = self._path
        self._path = None

        try:
            if self._file:
                self._file.close()
                self._file = None
        except Exception:
            raise SCPFileProducerError('Failed to close file: %s' % path)

        return result_or_failure

    def _sendLine(self, line):
        """
        Send an Unicode line of text as UTF-8 appending the new line delimiter.
        """
        # Each write is sent as a separate SSH data package.
        self._consumer.write(line.encode('utf-8') + self.delimiter)

    def _scheduleNextChunk(self):
        """
        Schedule transfer for the next data chunk.
        """
        # This might be called multiple times, while transfer is still
        # active.
        if (
            self._send_delayed_callback and
            self._send_delayed_callback.active()
                ):
            return

        self._send_delayed_callback = self._scheduler.callLater(
            0, self._clSendChunk)

    def pauseProducing(self):
        """
        Cancel in progress send, if still active.

        This is called by consumer.
        """
        if not self._send_delayed_callback:
            return
        if not self._send_delayed_callback.active():
            return
        self._send_delayed_callback.cancel()

    def resumeProducing(self):
        """
        This is called by consumer to tell us that we should continue.
        """
        self._scheduleNextChunk()

    def stopProducing(self):
        """
        This is required by IProducer but we don't have special logic.
        """
        self.pauseProducing()

        # The file might have been already fully read.
        if not self._result.called:
            self._result.callback(None)


@implementer(IConsumer)
class SCPFileConsumer(object):
    """
    Consumer for SCP uploaded data.
    """
    _scheduler = reactor
    _producer = None
    #: Buffer for data received but not consumed.
    _buffer = None
    #: Buffer limit at which producer is paused.
    BUFFER_PAUSE_LIMIT = 2 ** 20  # 1 MB.

    def __init__(self, shell, path, size, mode=0o777):
        flags = (
            filetransfer.FXF_WRITE |
            filetransfer.FXF_CREAT |
            filetransfer.FXF_TRUNC
            )
        self._file = shell.openFile(path, flags, {'permissions': mode})
        self._size = size
        self._cursor = 0
        self._buffer = b''
        self._all_data_received = False
        self._null_byte_received = False
        # Delayed called for buffer flushing. To help with testing.
        self._flush_call = None

    def registerProducer(self, producer, streaming=True):
        """
        Called by producer when it want to send us data.
        """
        if not streaming:
            raise RuntimeError('Only push producers are supported.')
        self._producer = producer

    def fail(self):
        """
        Called when the transfer failed.

        This is also called when connection was lost without sending the
        last end of file marker.
        """
        if self._file:
            if self._all_data_received:
                # Most probably the connection was lost without sending
                # the EOF marker.
                self._file.close()
            else:
                self._file.fail()
            self._file = None

    def unregisterProducer(self):
        """
        Called by producer when it stop sending data.
        """
        if self._file:
            self._file.close()
            self._file = None
        self._producer = None

    def write(self, data):
        """
        Called by producer when data is received.
        """
        to_write = self._buffer + data

        requested_size = self._cursor + len(to_write)

        if requested_size >= self._size:
            self._all_data_received = True

            if requested_size > self._size:
                if requested_size - self._size > 1 or to_write[-1:] != b'\x00':
                    # Got unexpected bytes, inform the producer and
                    # stop writing.
                    self._producer.failedConsuming(
                        'Got more than advertised.')
                    return
                self._null_byte_received = True
                # Don't write the NULL byte.
                to_write = to_write[:-1]

        try:
            self._write(to_write)
        except Exception:
            # The error is already emitted by the SFTP shell.
            self._all_data_received = False
            self._producer.failedConsuming('Failed to write file data.')
            return

        if self._cursor >= self._size and self._null_byte_received:
            # Finished writing and got the NULL byte, inform producer.
            self._producer.stopProducing()

        # All data was received so write will no longer be triggered by
        # remote SCP. This means that we need to manually trigger buffer
        # flush.
        if self._all_data_received and self._buffer:
            # To flush the buffer, just call write without any new data.
            # Not sure how to delay this to avoid a loop bomb.
            self._flush_call = self._scheduler.callLater(0.2, self.write, b'')

        # When buffer is too big, we need to pause the producer,
        # flush the buffer and resume once buffer is done.
        if len(self._buffer) > self.BUFFER_PAUSE_LIMIT:
            # FIXME:2151:
            # For now, don't do anything and hope the system has
            # enough memory.
            pass

    def _write(self, data):
        """
        Low level write code, which handles cursor and buffer.
        """
        written = self._file.writeChunk(self._cursor, data)
        self._cursor += written
        self._buffer = data[written:]
