# Copyright (c) 2011 Adi Roiban.
# See LICENSE for details.
"""
SSH Session dispatcher.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from twisted.conch.ssh.channel import SSHChannel
from twisted.conch.ssh.common import getNS, NS
from twisted.conch.ssh.connection import SSHConnection
from twisted.conch.ssh.session import (
    SSHSession,
    SSHSessionProcessProtocol,
    wrapProcessProtocol,
    wrapProtocol,
    )
from twisted.internet import reactor
from twisted.internet.error import ConnectionDone
from twisted.internet.interfaces import IConsumer
from zope.interface import implementer

from chevah.server import force_unicode
from chevah.server.commons.session import RemoteConnectionMixin
from chevah.server.ssh.sftp import SFTPClientSubsystem


@implementer(IConsumer)
class FileTransferServerSSHSession(SSHSession, RemoteConnectionMixin, object):
    """
    This is the SSH Server Session used to force file transfer only services.

    For now it only accepts SFTP sub-system requests and SCP command.

    `self.client` is the high-level handler for requested SSH session.
    `self.session` is not used.
    """
    disconnecting = False
    producer = None
    _scheduler = reactor

    def _emitRequestReject(self, details):
        """
        Emit an event for the SFTP request rejection.
        """
        self.emitEvent('30008', data={'details': details})

    def request_subsystem(self, data):
        """
        Handler requested subsystem.

        Code copied from twisted.conch to use our own
        SFTPSessionProcessProtocol.
        """
        subsystem_name, ignored = getNS(data)
        subsystem = self.avatar.lookupSubsystem(subsystem_name, data)
        if not subsystem:
            message = (
                'Subsystem "%s" not enabled for user or not supported.' % (
                    force_unicode(subsystem_name),))
            self._emitRequestReject(message)
            return 0

        session = SFTPSessionProcessProtocol(self)
        subsystem.makeConnection(wrapProcessProtocol(session))
        session.makeConnection(wrapProtocol(subsystem))
        session.subsystem = subsystem

        self.client = session
        return 1

    def request_shell(self, data):
        """
        Disable shell requests.
        """
        self._emitRequestReject('User is not allowed to obtain a shell.')
        return 0

    def request_exec(self, data):
        """
        Only allow SCP requests.
        """
        command, _ = getNS(data)

        try:
            name, arguments = command.split(b' ', 1)
        except ValueError:
            name = command
            arguments = b''

        client = self.avatar.lookupExec(name, arguments)
        if not client:
            self._emitRequestReject(
                'Exec "%s" not enabled for user or not supported.' % (
                    force_unicode(name),))
            return 0

        # All good and we can register the client.
        self.client = client

        # The SSH client might be waiting for the response from request_exec,
        # so inform that all is OK and start the client activity a bit later.
        self._scheduler.callLater(0, client.makeConnection, self)
        return 1

    def request_pty_req(self, data):
        """
        Disable PTY requests.
        """
        self._emitRequestReject('User is not allowed to obtain a PTY.')
        return 0

    def dataReceived(self, data):
        """
        Ignore data until a client is established.
        """
        # twisted.conch is accumulating data until the client
        # is present and then client should get/flush this data when client
        # connection is made.
        if not self.client:
            self.buf += data
            return

        self.client.dataReceived(data)

    def eofReceived(self):
        """
        Called when the other side has sent an EOF message.
        """
        if self.client:
            self.client.eofReceived()

    def loseConnection(self):
        """
        Skip SSHSession method as we close the client only once in self.closed.
        """
        self.disconnecting = True
        SSHChannel.loseConnection(self)

    def closed(self):
        """
        Called when session is closed.
        """
        self.conn.sendClose(self)

        if not self.client:
            return

        self.client.connectionLost()
        self.client = None

    def registerProducer(self, producer, streaming=True):
        """
        Called by producer when it want to send us data.
        """
        if not streaming:
            raise RuntimeError('Only push producers are supported.')
        self.producer = producer

    def unregisterProducer(self):
        """
        Called by producer when it stop sending data.
        """
        self.producer = None

    def stopWriting(self):
        """
        Called when the remote buffer is full, as a hint to stop writing.
        This can be ignored, but it can be helpful.
        """
        if not self.producer:
            return
        self.producer.pauseProducing()

    def startWriting(self):
        """
        Called when the remote buffer has more room, as a hint to continue
        writing.
        """
        if not self.producer:
            return
        self.producer.resumeProducing()


class SFTPSessionProcessProtocol(SSHSessionProcessProtocol):
    """
    Patch SSHSessionProcessProtocol.

    SSHSessionProcessProtocol does not implement dataReceived or eofReceived.
    """

    subsystem = None

    def loseConnection(self):
        """
        Closes the session.
        """
        self.connectionLost(reason=ConnectionDone())

    def dataReceived(self, data):
        """
        Proxy to transport.
        """
        self.transport.write(data)

    def eofReceived(self):
        """
        Called when an EOF message is sent.

        This is triggered by the client's QUIT command.
        """
        self.session.conn.sendClose(self.session)

    def connectionLost(self, reason=None):
        """
        Called when session is closed.
        """
        self.connected = False
        if self.subsystem:
            self.subsystem.connectionLost(reason)
        self.session.loseConnection()
        self.transport.loseConnection()


class SingleSSHConnectionService(SSHConnection, object):
    """
    A SSH connection is the implementation of 'ssh-connection' service that
    is multiplexing multiple ssh channels over a single ssh transport.

    Designed to be used on the client side.

    This only support a single channel/session.

    The user of this service will set the `deferred` member in order to wait
    for signals from the remote connection.
    """

    def __init__(self, channel):
        super(SingleSSHConnectionService, self).__init__()
        self._channel = channel

    def serviceStarted(self):
        """
        Called by protocol when the service was started.
        """
        self.openChannel(self._channel)

    def serviceStopped(self, reason=None):
        """
        Called by protocol when the service should be stopped.
        """
        if reason and not self.deferred.called:
            # Service was stopped due to an error.
            self.deferred.errback(reason)
        super(SingleSSHConnectionService, self).serviceStopped()


class SFTPClientSSHSession(SSHChannel, object):
    """
    A SSH client channel that request a remote sftp subsystem and link
    the SFTP client to the transport.

    Session initialization is linked to the `deferred` attribute of parent
    'ssh-connection' service so that users can be notified when the SFTP
    session starts.
    """
    name = 'session'
    # High level SFTP subsystem associated with this session.
    _client = None

    def channelOpen(self, data):
        """
        Called when the channel is opened. `data` is any data that the
        other side sent us when opening the channel and for SFTP it is ignored.
        """
        deferred = self.conn.sendRequest(
            self, 'subsystem', NS('sftp'), wantReply=1)
        deferred.addCallback(self._cbSubsystem)
        deferred.addErrback(self._ebSubsystem)

    def closed(self):
        """
        Called when the channel is closed.  This means that both our side and
        the remote side have closed the channel.
        """
        if self._client:
            self._client.connectionLost()

    def _cbSubsystem(self, result):
        """
        Called when we got the SFTP subsystem.
        """
        client = SFTPClientSubsystem()
        client.makeConnection(self)
        self._client = client

        # Link sftp subsystem data to channel data.
        self.dataReceived = client.dataReceived

        # Make sure connection's callback is called at the end.
        self.conn.deferred.callback(client)

    def _ebSubsystem(self, failure):
        """
        Called when we fail to get the SFTP subsystem.
        """
        self.conn.deferred.errback(failure)
