# Copyright (c) 2010 Adi Roiban.
# See LICENSE for details.
"""
SFTP server subsystem.
"""
from __future__ import unicode_literals
from __future__ import absolute_import

from contextlib import contextmanager
import errno
import os
import stat
import struct

from zope.interface import implementer
from twisted.internet import defer, reactor
from twisted.internet.error import ConnectionDone, ConnectionLost
from twisted.conch.interfaces import ISFTPFile
from twisted.conch.ls import lsLine
from twisted.conch.ssh.filetransfer import (
    FileTransferClient,
    FileTransferServer,
    FX_OP_UNSUPPORTED,
    FX_FILE_IS_A_DIRECTORY,
    FX_FAILURE,
    FX_NO_SUCH_FILE,
    FX_NOT_A_DIRECTORY,
    FX_PERMISSION_DENIED,
    FX_EOF,
    FX_FILE_ALREADY_EXISTS,
    FXF_READ,
    FXF_WRITE,
    FXF_APPEND,
    FXF_CREAT,
    FXF_TRUNC,
    FXF_EXCL,
    SFTPError,
    )

from chevah_compat import LocalFilesystem
from chevah_compat.exceptions import CompatError
from chevah.server import (
    add_timeout,
    emit_from_error,
    emit as _emit,
    force_unicode,
    )
from chevah.server.commons.authorization import AuthorizationAvatarAdapter
from chevah.server.commons.constant import DEFAULT_FILE_MODE
from chevah.server.commons.exception import AuthorizationException, ServerError
from chevah.server.commons.interface import (
    ISFTPServer,
    ILocationShell,
    )
from chevah.server.commons.session import (
    emit_account_login_paths,
    RemoteConnectionMixin,
    )
from chevah.server.event.handler import emit_error_details
from chevah.server.location.commons import OnConnectionLostMixin

# Not all SFTP codes are found in Twisted upstream
# https://datatracker.ietf.org/doc/html/draft-ietf-secsh-filexfer-13#section-9.1
# The is in version 13 of the draft SSH_FX_QUOTA_EXCEEDED = 15
# but we only support v2
FX_WRITE_PROTECT = 12


class SFTPShellBase(RemoteConnectionMixin):
    """
    Shared code for all SFTP shell objects.
    """

    _scheduler = reactor

    def emitEvent(self, event_id, data=None, avatar=None):
        """
        Emit an event for this shell.
        """
        if data is None:
            data = {}

        path = data.get('path', None)
        if path:
            segments = self.filesystem.getSegments(path)
            data['real_path'] = self.filesystem.getRealPathFromSegments(
                segments)
        super(SFTPShellBase, self).emitEvent(
            event_id,
            avatar=avatar,
            data=data,
            )

    def _getCompatAttributes(self, attributes):
        """
        Convert attributes from SFTP shell to compat.filesystem.
        """
        result = attributes.copy()
        try:
            # SFTP shell uses permissions while compat uses `mode`.
            permissions = result.pop('permissions')
            result['mode'] = permissions
        except KeyError:
            """
            No need to convert.
            """

        return result


@implementer(ISFTPFile)
class ChevahSFTPFile(SFTPShellBase):
    """
    File handling shell for SFTP subsystem.
    """

    def __init__(
        self, server, authorization, segments, flags, attrs, scheduler=reactor,
            ):
        self._scheduler = scheduler
        self._avatar = server.avatar
        self.filesystem = server.filesystem
        self._authorization = authorization
        # The open_mode text is exposed to the events and is part of our
        # API. Take extra care when changing its value.
        self._open_mode = 'unknown'

        self._read_total = 0
        self._write_total = 0
        self._start_time = self._scheduler.seconds()

        # Record if the last read/write request was successful.
        # Use to have a better close message.
        self._write_success = True
        self._read_success = True

        if os.name == 'nt':
            openFlags = os.O_BINARY
        else:
            openFlags = 0
        if flags & FXF_READ == FXF_READ and flags & FXF_WRITE == 0:
            openFlags = self.filesystem.OPEN_READ_ONLY
            self._open_mode = 'reading'

        if flags & FXF_WRITE == FXF_WRITE and flags & FXF_READ == 0:
            openFlags = self.filesystem.OPEN_WRITE_ONLY
            self._open_mode = 'writing'

        if flags & FXF_WRITE == FXF_WRITE and flags & FXF_READ == FXF_READ:
            openFlags = self.filesystem.OPEN_READ_WRITE
            self._open_mode = 'reading and writing'

        if flags & FXF_APPEND == FXF_APPEND:
            openFlags |= self.filesystem.OPEN_APPEND
            self._open_mode += ' with append'

        if flags & FXF_CREAT == FXF_CREAT:
            openFlags |= self.filesystem.OPEN_CREATE
        if flags & FXF_TRUNC == FXF_TRUNC:
            openFlags |= self.filesystem.OPEN_TRUNCATE
        if flags & FXF_EXCL == FXF_EXCL:
            openFlags |= self.filesystem.OPEN_EXCLUSIVE

        # This is the path as requested by the remote client.
        # It might be changed as part of the server MFT process.
        self._requested_path = self.filesystem.getPath(segments)

        if 'writing' in self._open_mode:
            # We have a write request, so see if the path needs mangling.
            amended_segments, was_path_amended = (
                self.avatar.getAmendedUploadSegments(segments))

            if was_path_amended:
                # We always want to create a new file and fail if file
                # exists.
                openFlags |= self.filesystem.OPEN_CREATE
                openFlags |= self.filesystem.OPEN_EXCLUSIVE

            segments = amended_segments

        self.path = self.filesystem.getPath(segments)

        init_event_data = {
            'requested_path': self._requested_path,
            'path': self.path,
            'mode': self._open_mode,
            }

        init_event_data['mode'] = self._open_mode

        # twisted.conch use `permissions` instead of `mode`.
        mode = attrs.pop('permissions', DEFAULT_FILE_MODE)
        if self.avatar.ignore_create_permissions:
            mode = DEFAULT_FILE_MODE
            attrs = {}

        self._write_limit = self.avatar.file_write_size
        self._check_quota = lambda _, ignored: None
        if self._write_limit > 0:
            self._check_quota = self._checkQuota

        with ChevahSFTPShell.emitDetailsOnError(
                '30044', init_event_data, self.emitEvent):

            # Check high level permission.
            if 'read' in self._open_mode:
                self._authorization.canRead(self.path)
            if 'append' in self._open_mode or 'writ' in self._open_mode:
                self._authorization.canWrite(self.path)

            self.fd = self.filesystem.openFile(segments, openFlags, mode)
            try:
                if attrs:
                    compat_attributes = self._getCompatAttributes(attrs)
                    self.filesystem.setAttributes(segments, compat_attributes)
                self.emitEvent('30043', data=init_event_data)
            except Exception:
                # Make sure we close the file on errors.
                os.close(self.fd)
                raise

    @property
    def avatar(self):
        """
        See: `RemoteConnectionMixin.avatar`.
        """
        return self._avatar

    def close(self):
        """
        Called when client request file closing and at the end of a transfer.
        """
        if not self._write_success:
            return self.fail()

        if not self._read_success:
            return self.fail()

        return self._close(
            read_event='30067',
            write_event='30068',
            other_event='30017',
            )

    def fail(self):
        """
        Called when the file should be closed due to a failure.
        """
        return self._close(
            read_event='30041',
            write_event='30042',
            other_event='30040',
            )

    def _close(self, read_event, write_event, other_event):
        """
        Close the low level file and emit the events.
        """
        duration = self._scheduler.seconds() - self._start_time
        read_speed = 0
        write_speed = 0
        if duration > 0:
            read_speed = self._read_total / duration
            write_speed = self._write_total / duration

        data = {
            'path': self.path,
            'mode': self._open_mode,
            'total_read': self._read_total,
            'total_write': self._write_total,
            'read_speed': read_speed,
            'write_speed': write_speed,
            'duration': duration,
            'list': ['a', 'b'],
            }
        with ChevahSFTPShell.emitDetailsOnError(
                '30047', data, self.emitEvent):

            result = os.close(self.fd)
            if self._open_mode == 'reading':
                self.emitEvent(read_event, data=data)
            elif self._open_mode == 'writing':
                self.emitEvent(write_event, data=data)
            else:
                self.emitEvent(other_event, data=data)
            return result

    def readChunk(self, offset, length):
        """
        Called during download transfers.
        """
        data = {'path': self.path}
        self._read_success = False
        with ChevahSFTPShell.emitDetailsOnError(
                '30045', data, self.emitEvent):

            os.lseek(self.fd, offset, 0)
            data = os.read(self.fd, length)
            self._read_total += len(data)
            self._read_success = True
            return data

    def writeChunk(self, offset, data):
        """
        Called during upload transfers.
        """
        event_data = {'path': self.path}
        self._write_success = False
        with ChevahSFTPShell.emitDetailsOnError(
                '30046', event_data, self.emitEvent):

            self._check_quota(self._write_total, len(data))
            os.lseek(self.fd, offset, 0)
            written_size = os.write(self.fd, data)
            self._write_success = True
            self._write_total += written_size
            return written_size

    def getAttrs(self):
        """
        Return all attributes for this file.
        """
        data = {'path': self.path}
        with ChevahSFTPShell.emitDetailsOnError(
                '30061', data, self.emitEvent):

            status = os.fstat(self.fd)
            result = ChevahSFTPShell.getAttributesFromStatus(status)
            self.emitEvent('30081', data=data)
            return result

    def setAttrs(self, attrs):
        """
        Set attributes for this file.
        """
        data = {'path': self.path}
        if self.avatar.ignore_create_permissions:
            self.emitEvent('30074', data=data)
            return

        self.emitEvent('30062', data=data)
        raise NotImplementedError(
            'Setting attributes for opened files is not supported.')

    def _checkQuota(self, written_size, new_chunk_size):
        """
        Check that we are allowed to write another `size` bytes.
        """
        # FIXME:6587:
        # For SFTP you can overwrite a file and the total size on disk
        # is the same.
        if written_size + new_chunk_size > self._write_limit:
            raise SFTPError(
                errorCode=FX_WRITE_PROTECT,
                errorMessage='Single file write limit is "{}" bytes.'.format(
                    self._write_limit),
                )


class ChevahSFTPFolder(object):
    """
    An iterator over a folder.
    """

    def __init__(self, filesystem, segments, only_folders=False):
        self._files = filesystem.iterateFolderContent(segments)
        self._segments = segments
        self._show_only_folders = only_folders

    def __iter__(self):
        return self

    class _UnknownStat(object):
        """
        A placeholder for stats for a member which we know that it exists
        but we can not get its details.
        """
        def __init__(self):
            self.st_size = 0
            self.st_mode = 0
            self.st_atime = 0
            self.st_mtime = 0
            self.st_ctime = 0
            self.st_uid = 0
            self.st_gid = 0
            self.st_nlink = 0
            self.set_dev = 0

    class _StatFromAttributes(object):
        """
        A placeholder for stats obtained via chevah_compat.
        """
        def __init__(self, attributes):
            self.st_size = attributes.size
            self.st_mode = attributes.mode
            self.st_atime = attributes.modified
            self.st_mtime = attributes.modified
            self.st_ctime = attributes.modified
            self.st_uid = attributes.uid
            self.st_gid = attributes.gid
            self.st_nlink = attributes.hardlinks
            self.set_dev = 0

    def __next__(self):
        try:
            child_attributes = next(self._files)
            name = child_attributes.name
            child_stats = self._StatFromAttributes(child_attributes)
        except OSError:
            name = '???'
            child_stats = self._UnknownStat()

        child_name_encoded = name.encode('utf-8')
        longname = lsLine(child_name_encoded, child_stats)

        attrs = ChevahSFTPShell.getAttributesFromStatus(child_stats)

        if self._show_only_folders and not stat.S_ISDIR(child_stats.st_mode):
            # We only have traversal, so hide non-folders.
            return next(self)

        return (child_name_encoded, longname, attrs)

    next = __next__

    def close(self):
        """
        Close the folder listing.

        Required by Twisted ChevahSFTPShell.
        """
        self._files = []


@implementer(ISFTPServer)
class ChevahSFTPShell(SFTPShellBase):
    """
    Shell for Chevah SFTP/SCP service.

    Most filesystem access is implemented here, but it also delegates to
    SFTPFile and SFTPFolder.
    """
    SFTPFile = ChevahSFTPFile

    def __init__(self, avatar):
        self._avatar = avatar
        self._authorization = AuthorizationAvatarAdapter(avatar)
        self.filesystem = LocalFilesystem(avatar=self._avatar)
        self._remote_version = 0

    @property
    def avatar(self):
        """
        See: `RemoteConnectionMixin.avatar`.
        """
        return self._avatar

    @property
    def remote_version(self):
        """
        The version used when communicating with the remote client.
        """
        return self._remote_version

    def start(self, service_name, encryption_data={}):
        """
        See: `ISFTPServer`.
        """
        start_path = self.filesystem.getPath(self.filesystem.home_segments)
        data = {'service_name': service_name}
        data.update(encryption_data)
        # FIXME:4687:
        # We pass here the encryption data, but the encryption data should be
        # recorded after a successful negotiation and renegotiations.
        emit_account_login_paths(
            id='30011',
            avatar=self._avatar,
            emitter=self.emitEvent,
            path=start_path,
            data=data,
            )

    @staticmethod
    def getAttributesFromStatus(status):
        """
        Convert os.stats into a dict used by the shell.

        The shell uses `permissions` while in chevah_compat we use mode.
        """
        return {
            "size": status.st_size,
            "uid": status.st_uid,
            "gid": status.st_gid,
            "permissions": status.st_mode,
            "atime": int(status.st_atime),
            "mtime": int(status.st_mtime),
            }

    @staticmethod
    @contextmanager
    def emitDetailsOnError(event_id, data, emit_method):
        """
        When execution fails, emit 'event_id' with `data` augmented with
        error's details.
        """
        try:
            yield
        except Exception as error:
            emit_error_details(
                error=error,
                event_id=event_id,
                data=data,
                emitter=emit_method,
                )
            # Since we have already logged the details we don't want to
            # trigger an internal server error in _ebStatus so we force / fake
            # unknown errors as errno.EACCES to be accepted by _ebStatus.
            if (
                isinstance(error, (OSError, IOError)) and
                error.errno not in SFTPSubsystem.HANDLED_ERROR_NUMBERS
                    ):
                error.errno = errno.EACCES

            # Same for CompatErrors.
            if isinstance(error, CompatError):
                error = OSError(errno.EACCES, error.message)

            raise error

    def _emitDetailsOnError(self, event_id, data):
        """
        When execution fails, emit 'event_id' with `data` augmented with
        error's details.
        """
        return self.emitDetailsOnError(event_id, data, self.emitEvent)

    def gotVersion(self, otherVersion, extData):
        """
        See: `ISFTPServer`.

        No extensions are supported for now.
        """
        self._remote_version = otherVersion
        return {}

    def openDirectory(self, path):
        """
        See: `ISFTPServer`.
        """
        segments = self.filesystem.getSegments(path)
        data = {'path': self.filesystem.getPath(segments)}
        self.emitEvent('30019', data=data)

        with self._emitDetailsOnError('30021', data):
            self._authorization.canTraverse(data['path'])
            if not self.filesystem.exists(segments):
                raise SFTPError(
                    errorCode=FX_NO_SUCH_FILE,
                    errorMessage='No such file or directory.',
                    )
            if not self.filesystem.isFolder(segments):
                raise SFTPError(
                    errorCode=FX_OP_UNSUPPORTED,
                    errorMessage=(
                        'Listing files is not supported. '
                        'Get files attributes instead.'))

            try:
                self._authorization.canList(data['path'])
                only_folders = False
            except AuthorizationException:
                # We don't have full listing permissions, so we only
                # show folders for traversal.
                only_folders = True

            result = ChevahSFTPFolder(
                self.filesystem, segments, only_folders)
            self.emitEvent('30020', data=data)
            return result

    def openFile(self, path, flags, attrs):
        """
        See: `ISFTPServer`.
        """
        segments = self.filesystem.getSegments(path)
        # Permission is handled in the SFTPFile init.
        return self.SFTPFile(
            self, self._authorization, segments, flags, attrs,
            scheduler=self._scheduler,
            )

    def removeFile(self, path):
        """
        See: `ISFTPServer`.
        """
        segments = self.filesystem.getSegments(path)
        data = {'path': self.filesystem.getPath(segments)}
        self.emitEvent('30022', data=data)

        def remove_file(segments, data):
            """
            Do the delete via the filesystem.
            """
            with self._emitDetailsOnError('30024', data):
                self._authorization.canDeleteFile(data['path'])
                result = self.filesystem.deleteFile(segments)
                self.emitEvent('30023', data=data)
            return result

        def remove_from_delay(segment, data):
            """
            This is called by the delete delayed callback.
            """
            try:
                remove_file(segment, data)
            except Exception:
                """
                There is nothing we can do on error, as we already told the
                client that all was ok, when we faked the retunr in the first
                place.
                """

        ssh_delete_delay = self.avatar.ssh_delete_delay
        if ssh_delete_delay > 0:
            # We delay the delete operation,
            # but we tell the client that all was ok.
            # This is to work around a defect in Azure Data Factory, when
            # the file is first delete and then closed.
            self._scheduler.callLater(
                ssh_delete_delay,
                remove_from_delay, segments, data)
            return

        # No need to delay or fake.
        # Just execute the remove operation and return the result
        return remove_file(segments, data)

    def renameFile(self, old_path, new_path):
        """
        See: `ISFTPServer`.
        """
        old_segments = self.filesystem.getSegments(old_path)
        new_segments = self.filesystem.getSegments(new_path)
        old_path_name = self.filesystem.getPath(old_segments)
        new_path_name = self.filesystem.getPath(new_segments)
        data = {
            'path': new_path_name,
            'from': old_path_name,
            'to': self.filesystem.getPath(new_segments),
            }
        self.emitEvent('30025', data=data)

        with self._emitDetailsOnError('30027', data):
            self._authorization.canRename(data['from'], data['to'])
            result = self.filesystem.rename(old_segments, new_segments)
            self.emitEvent('30026', data=data)
            return result

    def makeDirectory(self, path, attributes):
        """
        See: `ISFTPServer`.
        """
        segments = self.filesystem.getSegments(path)
        data = {'path': self.filesystem.getPath(segments)}
        self.emitEvent('30028', data=data)

        compat_attributes = self._getCompatAttributes(attributes)

        with self._emitDetailsOnError('30030', data):
            self._authorization.canCreateFolder(data['path'])
            self.filesystem.createFolder(segments)
            if not self.avatar.ignore_create_permissions:
                self.filesystem.setAttributes(segments, compat_attributes)
            self.emitEvent('30029', data=data)

    def removeDirectory(self, path):
        """
        See: `ISFTPServer`.
        """
        segments = self.filesystem.getSegments(path)
        data = {'path': self.filesystem.getPath(segments)}
        self.emitEvent('30031', data=data)

        with self._emitDetailsOnError('30033', data):
            self._authorization.canDeleteFolder(data['path'])
            result = self.filesystem.deleteFolder(segments, recursive=False)
            self.emitEvent('30032', data=data)
            return result

    def getAttrs(self, path, followLinks):
        """
        See: `ISFTPServer`.
        """
        segments = self.filesystem.getSegments(path)
        data = {'path': self.filesystem.getPath(segments)}
        self.emitEvent('30034', data=data)

        with self._emitDetailsOnError('30036', data):
            self._authorization.canGetAttributes(data['path'])
            status = self.filesystem.getStatus(segments)
            result = self.getAttributesFromStatus(status)
            self.emitEvent('30035', data=data)
            return result

    def setAttrs(self, path, attributes):
        """
        See: `ISFTPServer`.
        """
        segments = self.filesystem.getSegments(path)
        data = {'path': self.filesystem.getPath(segments)}
        self.emitEvent('30037', data=data)

        compat_attributes = self._getCompatAttributes(attributes)

        with self._emitDetailsOnError('30039', data):
            self._authorization.canSetAttributes(data['path'])
            self.filesystem.setAttributes(segments, compat_attributes)
            self.emitEvent('30038', data=data)

    def readLink(self, path):
        """
        See: `ISFTPServer`.
        """
        segments = self.filesystem.getSegments(path)
        data = {'path': self.filesystem.getPath(segments)}
        self.emitEvent('30053', data=data)

        with self._emitDetailsOnError('30055', data):
            self._authorization.canList(data['path'])
            target_segments = self.filesystem.readLink(segments)
            result = self.filesystem.getPath(target_segments)
            self.emitEvent('30054', data=data)
            return result

    def makeLink(self, link_path, target_path):
        """
        See: `ISFTPServer`.
        """
        link_segments = self.filesystem.getSegments(link_path)
        target_segments = self.filesystem.getSegments(target_path)

        target_name = self.filesystem.getRealPathFromSegments(target_segments)
        link_name = self.filesystem.getPath(link_segments)
        data = {
            'path': link_name,
            'link': link_name,
            'destination_path': target_name,
            }
        self.emitEvent('30056', data=data)

        with self._emitDetailsOnError('30058', data):
            self._authorization.canFull(data['path'])
            self.filesystem.makeLink(target_segments, link_segments)
            self.emitEvent('30057', data=data)

    def realPath(self, path):
        """
        See: `ISFTPServer`.

        Return the absolute ChevahPath starting from a relative path.

        The path is returned in encoded UTF-8.
        """
        name = path.decode('utf-8')
        data = {'path': name}
        self.emitEvent('30060', data=data)
        segments = self.filesystem.getSegments(path)
        result = self.filesystem.getPath(segments).encode('utf-8')
        return result

    def extendedRequest(self, extName, extData):
        """
        See: `ISFTPServer`.
        """
        self.emitEvent('30059')
        raise NotImplementedError('Extended request are not supported.')


class SFTPSubsystem(FileTransferServer, object):
    """
    Implementation of server side SFTP SSH subsystem.

    It handled low level SSH packages using `packet_TYPE` methods and pass
    them to the SFTPShell.

    It is initialized by the SSH lookup code with (data, avatar).
    data is the raw data received as subsystem request and can be ignored.
    avatar is a generic SSHAvatar.
    """

    def emitEvent(self, event_id, data=None, avatar=None):
        """
        Emit the event via the SSH session.
        """
        if avatar is None:
            avatar = self.client.avatar
        _emit(event_id, avatar=avatar, data=data)

    def makeConnection(self, protocol):
        """
        Called just after initialization.
        """
        super(SFTPSubsystem, self).makeConnection(protocol)
        # This is one huge chain as Twisted SSH high level object don't get
        # access to the transport.
        encryption_data = (
            protocol.session.conn.transport._getEncryptionEventData())
        self.client.start('SFTP', encryption_data)

    def connectionLost(self, reason):
        """
        Called when the SSH session (which is SFTP subsystem) is closed.
        """
        for fileObj in self.openFiles.values():
            fileObj.fail()
        self.openFiles = {}

        for (dirObj, dirIter) in self.openDirs.values():
            # FIXME:5102:
            # Here we should call fail.
            dirObj.close()
        self.openDirs = {}

        if self.connected:
            # SSHConnection.ssh_CHANNEL_CLOSE is closing the system/channel
            # twice but we only want to emit events once.
            remote_version = self.client.remote_version
            self.emitEvent('30012', data={'client_version': remote_version})
        self.connected = False

    def _logSFTPError(self, failure):
        # FIXME:3231:
        # Add peer/avatar to data when emitting the error and not when
        # raising it.
        if failure.check(ServerError):
            emit_from_error(failure.value, emitter=self.emitEvent)
        else:
            data = {'details': force_unicode(failure.getBriefTraceback())}
            self.emitEvent('30016', data=data)

    #: List with error number which are handled by SFTP subsystem and for
    #: which a dedicated response error is created.
    HANDLED_ERROR_NUMBERS = [
        errno.ENOENT,
        errno.EACCES,
        errno.EEXIST,
        errno.EISDIR,
        errno.ENOTDIR,
        errno.EFBIG,
        ]

    def _ebStatus(self, reason, requestId, msg="request failed"):
        code = FX_FAILURE
        message = msg
        if isinstance(reason.value, (IOError, OSError)):
            # OS might raise non-English messages.
            # We we always want to send back English error messages.
            error_code = reason.value.errno
            if error_code == errno.ENOENT:
                code = FX_NO_SUCH_FILE
                message = b'No such file or directory'
            elif error_code == errno.EACCES:
                code = FX_PERMISSION_DENIED
                message = b'Permission denied'
            elif error_code == errno.EEXIST:
                code = FX_FILE_ALREADY_EXISTS
                message = b'File exists'
            elif error_code == errno.EISDIR:
                code = FX_FILE_IS_A_DIRECTORY
                message = b'Is a directory'
            elif error_code == errno.ENOTDIR:
                code = FX_NOT_A_DIRECTORY
                message = b'Not a directory'
            else:
                self._logSFTPError(reason)
        elif reason.type == EOFError:
            code = FX_EOF
            message = b'End of file'
            if reason.value.args:
                message = reason.value.args[0]
        elif isinstance(reason.value, AuthorizationException):
            # Denied by our authorization layer.
            code = FX_PERMISSION_DENIED
            message = b'Permission denied.'
        elif reason.type == NotImplementedError:
            code = FX_OP_UNSUPPORTED
            if reason.value.args:
                message = reason.value.args[0]
        elif reason.type == SFTPError:
            code = reason.value.code
            message = reason.value.message
        else:
            self._logSFTPError(reason)

        message = force_unicode(message).encode('utf-8')
        self._sendStatus(requestId, code, message)


@implementer(ILocationShell)
class SFTPClientSubsystem(
        FileTransferClient, OnConnectionLostMixin, RemoteConnectionMixin):
    """
    Implementation of client side SFTP subsystem and a SFTP client side shell.

    Only a single request can be active in the same time.
    All other requests are queues and acted once the previous request was done.
    """

    _scheduler = reactor

    def __init__(self, extData=None):
        # Fix initialization of upstream code which is initialized with a
        # mutable dict instead of None.
        if extData is None:
            extData = {}
        super(SFTPClientSubsystem, self).__init__(extData=extData)

    @property
    def avatar(self):
        """
        Avatar associated with this protocol.
        """
        return self._ssh_protocol.avatar

    @property
    def _ssh_protocol(self):
        """
        Lowest level SSH protocol/transport used by this subsystem.
        """
        session = self.transport
        connection_service = session.conn
        return connection_service.transport

    @property
    def _location(self):
        """
        Location associated with this client.
        """
        return self.avatar._credentials.creator

    def authenticate(self):
        """
        See: ILocationShell

        The SFTP subsystem is already authenticated as it is created after
        the SSH connection is authenticated.

        Here to comply with the interface.
        """
        return defer.succeed(self)

    def connectionMade(self):
        """
        Called when client sftp subsystem is connected and authenticated
        to the remote subsystem.
        """
        super(SFTPClientSubsystem, self).connectionMade()
        self.connected = True
        # FIXME:4372:
        # Add information about negotiated ciphers.
        # Also add SSH key fingerprint.
        credentials_type = self._ssh_protocol._auth_service.authentication_type
        self.emitEvent(
            '30076',
            data={
                'name': self._location.name,
                'credentials_type': credentials_type,
                },
            )

    def connectionLost(self, reason=ConnectionDone):
        """
        Called when connection to the remote subsystem was lost.

        The way Twisted SSH is implemented,  this is never call with anything
        else that a ConnectionDone, even when the connection is lost.
        """
        self.emitEvent('30077', data={'name': self._location.name})
        self.connected = False
        # FIXME:6502:
        # https://github.com/twisted/twisted/pull/1110
        # This is already fixed upstream.
        # But we have this difference to have a custom message on these errors.
        # See if we can send this change upstream.
        for _, deferred in self.openRequests.items():
            # There are still requests waiting to be executed while the
            # client is disconnected.
            # We assume that the connection was lost, as otherwise, we should
            # have not seen open requests here.
            deferred.errback(ConnectionLost('SFTP client request canceled'))
        self.openRequests = {}

        self._onConnectionLost(reason)
        super(SFTPClientSubsystem, self).connectionLost(reason)

    def _sendRequest(self, msg, data):
        """
        Send a SSH request and return a deferred which waits for the result.
        """
        if not self.connected:
            return defer.fail(ConnectionLost('Not connected to the server'))

        data = struct.pack('!L', self.counter) + data
        deferred = defer.Deferred()
        # Make sure that if the server hangs and is not responding to our
        # request we don't wait forever.
        add_timeout(
            self._scheduler,
            deferred,
            delay=60,
            message='No server response.',
            )

        self.openRequests[self.counter] = deferred
        self.counter += 1
        self.sendPacket(msg, data)
        return deferred

    def _parseAttributes(self, data):
        """
        Parses the file attributes, as received from the remote server,
        and returns a tuple containing a dictionary with the attributes and
        the remaining data.

        We overrride the upstream Twisted code so that we can provide an API
        that will always contain all attributes available to open (with the
        exception of the extension attributes):

            size: the size of the file in bytes.
            uid: the user ID of the file as an integer.
            gid: the group ID of the file as an integer.
            permissions: the permissions of the file as an integer. the bit
            representation of this field is defined by POSIX.
            atime: the access time of the file as seconds since epoch.
            mtime: the modification time of the file as seconds since epoch.
        """
        attributes = {
            'size': 0,
            'uid': None,
            'gid': None,
            'permissions': 0,
            'atime': 0,
            'mtime': 0,
            }
        attrs, data = super(SFTPClientSubsystem, self)._parseAttributes(data)
        attributes.update(attrs)

        return attributes, data
