# Copyright (c) 2010 Adi Roiban.
# See LICENSE for details.
"""
SSH user authentication service.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import six
from twisted.cred.error import UnauthorizedLogin
from twisted.conch.error import ConchError, IgnoreAuthentication
from twisted.conch.ssh import transport
from twisted.conch.ssh.common import NS, getNS
from twisted.conch.ssh.transport import (
    DISCONNECT_NO_MORE_AUTH_METHODS_AVAILABLE,
    )
from twisted.conch.error import ValidPublicKey
from twisted.conch.ssh.userauth import (
    MSG_USERAUTH_FAILURE,
    MSG_USERAUTH_SUCCESS,
    MSG_USERAUTH_REQUEST,
    SSHUserAuthClient,
    SSHUserAuthServer,
    )
from twisted.internet import defer

from chevah_keycert.ssh import Key
from chevah.server import emit_from_error, force_unicode
from chevah.server.commons.constant import (
    TYPE_NAME,
    )
from chevah.server.commons.credential import (
    SSHPasswordCredentials,
    SSHKeyCredentials,
    )
from chevah.server.commons.exception import (
    InsufficientAuthentication,
    ServerError,
    ServerException,
    )
from chevah.server.commons.interface import (
    IPasswordCredentials,
    ISSHKeyCredentials,
    ISSHPrivateKeyCredentials,
    )
from chevah.server.ssh.avatar import SSHServerAvatar

# Mapping from Chevah credential names to SSH RFC names.
_CHEVAH_MAPPING = {
    'password': b'password',
    'ssh-key': b'publickey'
    }


def _get_ssh_names(chevah_credentials):
    """
    Convert the chevah credential names into SSH credentials names.

    Raises an exception when receiving SSH credentials that are not supported.
    """
    result = []
    for name in chevah_credentials:
        try:
            ssh_name = _CHEVAH_MAPPING[name]
        except KeyError:
            raise ServerException(
                'Credentials "%s" can\'t be handled by SSH.' % (name,))

        result.append(ssh_name)

    return result


class SSHUserAuthServerService(SSHUserAuthServer, object):
    """
    Server side implementation of SSH user-auth service.
    """

    # Name of the user for which the authentication was currently requested.
    _username = None
    # The public key for which we are waiting the signature validation.
    _public_key_auth_result = None

    interfaceToMethod = SSHUserAuthServer.interfaceToMethod
    interfaceToMethod.update({
        IPasswordCredentials: b'password',
        ISSHKeyCredentials: b'publickey',
        })

    def ssh_USERAUTH_REQUEST(self, packet):
        """
        The client has requested to be authenticated.

        Payload:
            string user
            string next service
            string method
            <authentication specific data>
        """
        user, next_service, method, rest = getNS(packet, 3)

        try:
            username = user.decode('utf-8')
        except UnicodeDecodeError:
            deferred = defer.fail(ServerException(
                'Authentication requested for username with invalid '
                'encoding: "%s".' % (force_unicode(user,))))
        else:
            if user != self.user or next_service != self.nextService:
                # This is a request for a different user or a service, so we
                # reset the authentication state.
                self.authenticatedWith = []

            self.user = user
            self._username = username
            self.nextService = next_service
            self.method = method
            deferred = self.tryAuth(method, user, rest)

        deferred.addCallback(self._cbFinishedAuth)
        deferred.addErrback(self._ebMaybeBadAuth)
        deferred.addErrback(self._ebBadAuth)
        return deferred

    def tryAuth(self, kind, user, data):
        """
        Try to authenticate the user with the given method.

        This is a private method called by ssh_USERAUTH_REQUEST and
        dispatches to a auth_* method.

        @param kind: the authentication method to try.
        @type kind: C{str}
        @param user: the username the client is authenticating with.
        @type user: C{str}
        @param data: authentication specific data sent by the client.
        @type data: C{str}
        @return: A Deferred called back if the method succeeded, or erred back
            if it failed.
        @rtype: C{defer.Deferred}
        """
        if kind == b'none':
            # We don't support the `none` auth method.
            # Client will usually first try to see if the server will accept
            # them without an authentication. When rejected the response
            # will include the list of supported method.
            # This is a way to discover the method supported by the server.
            # `none` auth method is used by client to request the list of
            # supported method.
            self.transport.sendPacket(
                MSG_USERAUTH_FAILURE,
                NS(b','.join(self.supportedAuthentications)) + b'\x00',
                )
            return defer.fail(IgnoreAuthentication())

        if kind not in self.supportedAuthentications:
            return defer.fail(ServerException(
                'Authentication method "%s" is not supported.' % (
                    force_unicode(kind),)))

        kind = kind.replace(b'-', b'_')
        auth_method = getattr(self, 'auth_{}'.format(kind.decode('ascii')))
        return auth_method(data)

    def auth_password(self, packet):
        """
        Password authentication.  Payload::
            string password

        Make a UsernamePassword credential and verify it with our portal.
        """
        password = getNS(packet[1:])[0]
        try:
            password = password.decode('utf-8')
        except Exception:
            return defer.fail(ServerException(
                'Authentication requested with a password with invalid '
                'encoding.'
                ))

        credentials = SSHPasswordCredentials(
            username=self._username,
            password=password,
            peer=self.transport.avatar.peer,
            creator=self.transport.factory,
            )
        deferred = self.portal.login(credentials, None)
        deferred.addErrback(self._ebPassword)
        return deferred

    def auth_publickey(self, packet):
        """
        Public key authentication.  Payload::
            byte has signature
            string algorithm name
            string key data
            [string signature] (if has signature is True)

        Create a SSHPublicKey credential and verify it using our portal.

        Most of the time this is called twice.
        First the client will send just the public-key without a signature,
        to see if the key is accepted in the first place.
        The second time, the client will send the previous key and signed data
        to prove they have the private key.
        """
        peer = self.transport.avatar.peer
        has_signature = six.byte2int(packet)
        key_algorithm, key_data, rest = getNS(packet[1:], 2)

        try:
            pubKey = Key.fromString(key_data)
        except Exception as error:
            data = {'details': force_unicode(error)}
            self.transport.emitEvent('30049', data=data)
            return defer.fail(UnauthorizedLogin())

        signature = has_signature and getNS(rest)[0] or None
        if has_signature:
            key_signed_data = (
                NS(self.transport.sessionID) +
                six.int2byte(MSG_USERAUTH_REQUEST) +
                NS(self.user) +
                NS(self.nextService) +
                NS(b'publickey') +
                six.int2byte(has_signature) +
                NS(key_algorithm) +
                NS(key_data)
                )
            credentials = SSHKeyCredentials(
                username=self._username,
                peer=peer,
                creator=self.transport.factory,
                key_algorithm=key_algorithm,
                key_data=key_data,
                key_signed_data=key_signed_data,
                key_signature=signature,
                )

            if (
                self._public_key_auth_result
                and self._public_key_auth_result[0] == pubKey
                    ):
                deferred = defer.succeed(self._public_key_auth_result[1])
            else:
                deferred = self.portal.login(credentials, None)
        else:
            credentials = SSHKeyCredentials(
                username=self._username,
                peer=peer,
                creator=self.transport.factory,
                key_algorithm=key_algorithm,
                key_data=key_data,
                )
            deferred = self.portal.login(credentials, None)

        deferred.addCallback(self._cbCheckSignature, pubKey, credentials)
        deferred.addErrback(self._ebCheckKey, packet[1:])
        return deferred

    def _cbCheckSignature(self, result, public_key, credentials):
        """
        Called after the public key was accepted by the authentication.
        """
        if not credentials.key_signature:
            # Public key was accepted by the authentication method,
            # but we don't know yet if the remote peer also has the private
            # part.
            self._public_key_auth_result = (public_key, result)
            raise ValidPublicKey()

        if public_key.verify(
                credentials.key_signature, credentials.key_signed_data):
            # All good.
            self._public_key_auth_result = None
            return result

        raise ServerException('Invalid user public key signature.')

    def _cbFinishedAuth(self, result):
        """
        The callback when user has successfully been authenticated.  For a
        description of the arguments, see L{twisted.cred.portal.Portal.login}.
        We start the service requested by the user.

        This is a hook into the SSHUserAuthServer callback to convert the
        avatar from ServiceAvatar to an SSHConchUser.
        """
        interface, avatar, logout = result
        service_configuration = self.transport.factory.service_configuration
        self.transport.logoutFunction = logout
        service = self.transport.factory.getService(
            self.transport, self.nextService)
        if not service:
            raise ConchError(
                'could not get next service: %s' % self.nextService)
        self.transport.avatar = SSHServerAvatar(
            context=avatar,
            service_configuration=service_configuration,
            )
        self.transport.sendPacket(MSG_USERAUTH_SUCCESS, b'')
        self.transport.setService(service())

    def _ebMaybeBadAuth(self, reason):
        """
        Called when authentication fails.

        This handles accounts that require multi-credential authentication.
        """
        reason.trap(InsufficientAuthentication)

        # See https://tools.ietf.org/html/rfc4252#section-5.1
        self.supportedAuthentications = _get_ssh_names(
            reason.value.remaining_credentials)

        partial_success = b'\xff'  # True.
        self.transport.sendPacket(
            MSG_USERAUTH_FAILURE,
            NS(b','.join(self.supportedAuthentications)) + partial_success,
            )

    def _ebBadAuth(self, reason):
        """
        The final errback in the authentication chain.  If the reason is
        error.IgnoreAuthentication, we simply return; the authentication
        method has sent its own response.  Otherwise, send a failure message
        and (if the method is not 'none') increment the number of login
        attempts.

        For failures other than `UnauthorizedLogin` an event is also
        emitted.

        @type reason: L{twisted.python.failure.Failure}
        """
        if reason.check(IgnoreAuthentication):
            return

        if reason.check(UnauthorizedLogin):
            # For now we don't log these errors.
            # There should be an event raised in the Authenticator.
            # log.msg('unauthorized login: %s' % reason.getErrorMessage())
            pass
        else:
            self._emitEventFromFailure(reason)

        self.loginAttempts += 1
        if self.loginAttempts > self.attemptsBeforeDisconnect:
            self.transport.sendDisconnect(
                transport.DISCONNECT_NO_MORE_AUTH_METHODS_AVAILABLE,
                'too many bad auths'
                )
            return

        # We return all the supported methods to not leak on errors
        # any extra security info about the user.
        partial_success = b'\x00'  # False.
        self.transport.sendPacket(
            MSG_USERAUTH_FAILURE,
            NS(b','.join(self.supportedAuthentications)) + partial_success
            )

    def _emitEventFromFailure(self, reason):
        """
        Handle event emission from a generic Twisted failure.
        """
        if reason.check(ServerError):
            emit_from_error(reason.value, emitter=self.transport.emitEvent)
        elif reason.check(ServerException):
            self.transport.emitEvent(
                '30007', data={'details': reason.value.message})
        else:
            data = {
                'username': self._username,
                'details': force_unicode(reason.getBriefTraceback())
                }
            self.transport.emitEvent('30018', data=data)


class SSHUserAuthClientService(SSHUserAuthClient, object):
    """
    Implementation of ssh userauth service.

    Usually this service should be called before any other service.

    This is initialized with a credential and then various authentication
    methods will be chose based on credentials.

    `instance` is the service to be started on successful authentication.

    On failure `instance.deferred` is called to inform that authentication
    failed.
    """
    def __init__(self, credentials, instance):
        self._credentials = credentials
        self._authentication_type = 'not-yet-authenticated'
        self._sign_algorithm = b''
        super(SSHUserAuthClientService, self).__init__(
            credentials.username.encode('utf-8'), instance)

        if IPasswordCredentials.providedBy(self._credentials):
            self.preferredOrder = [b'password']
        elif ISSHPrivateKeyCredentials.providedBy(self._credentials):
            self.preferredOrder = [b'publickey']
        else:
            self.preferredOrder = []

    @property
    def authentication_type(self):
        """
        Return the authentication that was succesful.
        """
        if self._sign_algorithm:
            return (
                self._authentication_type + '-' +
                self._sign_algorithm.decode('ascii')
                )
        return self._authentication_type

    def ssh_USERAUTH_BANNER(self, packet):
        """
        We receive a MSG_USERAUTH_BANNER with a message and language.

        Payload::
            string message
            string language

        See: https://tools.ietf.org/html/rfc4252#section-5.4
        """
        message, _ = getNS(packet)

        banner = force_unicode(message)

        data = {'message': banner}
        self.transport.emitEvent('30079', data=data)

    def ssh_USERAUTH_SUCCESS(self, packet):
        """
        We received a MSG_USERAUTH_SUCCESS.  The server has accepted our
        authentication, so start the next service.
        """
        self.transport.setService(self.instance)
        self._authentication_type = self._credentials.kind_name

    def ssh_USERAUTH_FAILURE(self, packet):
        """
        We received a MSG_USERAUTH_FAILURE.

        Payload::
            string methods
            byte partial success

        This is a patch of upstream code. See upstream docstring.

        `partial` is ignored.
        """
        canContinue, partial = getNS(packet)
        # We never retry a failed method, as this is automated transfer
        # and we go for a single method.
        self.authenticatedWith.append(self.lastAuth)

        def order_by_preference(value):
            """
            Invoked once per authentication method in order to extract a
            comparison key which is then used for sorting.
            """
            if value in self.preferredOrder:
                return self.preferredOrder.index(value)
            else:
                # Put the element at the end of the list.
                # This is from upstream, but this should not happen in
                # our non-interactive client.
                return len(self.preferredOrder)

        # Upstream code will try to continue even if method is not in
        # our preferred list... but there is no point in trying as we don't
        # support it.
        remaining_methods = iter(sorted([
            method for method in canContinue.split(b',')
            if method in self.preferredOrder and
            method not in self.authenticatedWith
            ],
            key=order_by_preference,
            ))

        deferred = self._cbUserauthFailure(None, remaining_methods)
        deferred.addErrback(self._ebAuth)
        return deferred

    def _ebAuth(self, failure, *args):
        """
        Generic callback for a failed authentication attempt
        """
        # We don't retry or try any other method,
        self.instance.deferred.errback(ServerError('30078', data={
            'name': self._credentials.creator.name,
            'methods': force_unicode(b', '.join(self.authenticatedWith)),
            'details': force_unicode(failure.value),
            },
            peer=self._credentials.peer,
            ))
        # Inform the server that we stop trying to authenticate.
        self.transport.sendDisconnect(
            DISCONNECT_NO_MORE_AUTH_METHODS_AVAILABLE,
            'no more authentication methods available',
            )

    def _cbUserauthFailure(self, result, iterator):
        """
        Called from ssh_USERAUTH_FAILURE when the sent credentials were not
        accepted.

        This is a patch of upstream code to allow calling a deferred.
        I have also tried to add some comment while struggling to understand
        the code.
        """
        # This is from our recursive call. The idea is that if tryAuth
        # return None, we will try the next method.
        if result:
            return

        try:
            method = next(iterator)
        except StopIteration:
            return defer.fail(
                ServerException('No more authentication methods available.'))
        else:
            # Try to authenticate using the next method.
            deferred = defer.maybeDeferred(self.tryAuth, method)
            deferred.addCallback(self._cbUserauthFailure, iterator)
            return deferred

    def getPassword(self):
        """
        Return password to be sent to server or None to not use password
        authentication.
        """
        if (
            IPasswordCredentials.providedBy(self._credentials) and
            self._credentials.password
                ):
            return defer.succeed(self._credentials.password.encode('utf-8'))
        else:
            return None

    def getPublicKey(self):
        """
        See: `SSHUserAuthClient`.
        """
        if not ISSHPrivateKeyCredentials.providedBy(self._credentials):
            return None
        return defer.succeed(self._credentials.getPublicKey())

    def getPrivateKey(self):
        """
        See: `SSHUserAuthClient`.
        """
        if not ISSHPrivateKeyCredentials.providedBy(self._credentials):
            return None
        return defer.succeed(self._credentials.getPrivateKey())

    def _cbGetPublicKey(self, publicKey):
        """
        This is a patched upstream version so that it works with
        chevah-keycert.

        It only removes the `isinstance` check for conch own keys.Key
        """
        def get_key_algorighm(publicKey):
            """
            Return the key algorithm to use for `publicKey`.
            """
            variants = publicKey.supportedSignatureAlgorithms()
            if len(variants) == 1:
                # Only a single method is supported.
                # Nothing to select.
                # Just return it.
                return variants[0]

            # We have multiple options, use the extension to select one.
            server_sig_algs = self.transport.peerExtensions.get(
                b'server-sig-algs', b'')
            server_algorithms = [
                a.strip() for a in server_sig_algs.split(b',') if a.strip()]

            if not server_algorithms:
                # Remote server has no supported methods.
                if TYPE_NAME.SSH_RSA.encode('ascii') in variants:
                    # Fallback to standard/simple ssh-rsa method for
                    # RSA.
                    return TYPE_NAME.SSH_RSA.encode('ascii')

                # This should never happen, as we should only support RSA
                # keys.
                variants_text = force_unicode(b','.join(variants))
                raise ServerException(  # noqa:cover
                    'Remote server has no "server-sig-algs" '
                    'for our "{}" key.'.format(variants_text)
                    )

            for variant in variants:
                if variant in server_algorithms:
                    return variant

            variants_text = force_unicode(b','.join(variants))
            raise ServerException(
                'No match for "server-sig-algs" extension. '
                'Server support: "{}". '
                'Our support: "{}".'.format(
                    force_unicode(server_sig_algs), variants_text))

        # FIXME:6502:
        # Check if upstream code is different
        # This also adds support for RSA SHA2 based signature, if the
        # server supports it.
        # See https://github.com/twisted/twisted/issues/11808
        if publicKey is not None:
            self._sign_algorithm = get_key_algorighm(publicKey)
            self.lastPublicKey = publicKey
            self.triedPublicKeys.append(publicKey)
            self.askForAuth(
                b'publickey',
                b'\x00' + NS(self._sign_algorithm) + NS(publicKey.blob()))
            return True
        else:
            return False
