/*
Tests for table UI components.
*/

describe('TableFooterCtrl', function() {
    var table_ui = require('./table_ui.js')

    /*
    When going to negative page numbers, it will force the page to value
    0 as the first page..
    */
    test('onGoToPage_keep_start_page', function() {
        var sut = mk.getComponent('tableFooter', table_ui.TableFooterBindings)
        sut.ctrl.tableOptions.page = 3

        sut.ctrl.onGoToPage(-1)

        expect(sut.ctrl.tableOptions.page).toEqual(0)
    })

})


describe('TableSortCtrl', function() {
    var table_ui = require('./table_ui.js')

    /*
    Return an already initialized component.
    */
    var getTableSort = function(sortOptions) {
        var sut = mk.getComponent('tableSort', table_ui.TableSortBindings)
        sut.ctrl.sortOptions = sortOptions
        sut.ctrl.$onInit()
        return sut
    }

    /*
    Will generate UI options based on the input configuration and will
    populate the input with reset and sort up and down callbacks.
    */
    test('$onInit', function() {
        var sut = mk.getComponent('tableSort', table_ui.TableSortBindings)
        sut.ctrl.sortOptions = {
            // This are populated by the controller.
            'reset': null,
            'sort_up': null,
            'sort_down': null,
            // DB description.
            'default_column': 'USER_ID',
            'default_order': 'asc',
            'columns': [
                {'value': 'USER_NAME', 'text': 'Username'},
                {'value': 'USER_ID', 'text': 'User ID'},
                {'value': 'USER_TYPE', 'text': 'User type'},
            ],
        }

        sut.ctrl.$onInit()

        expect(sut.ctrl.options[4]).toBe(sut.ctrl.sortOptions.sort_up)
        expect(sut.ctrl.options[5]).toBe(sut.ctrl.sortOptions.sort_down)
        expect(sut.ctrl.options).toEqual([
            {
                "active": false,
                "text": "Username",
                "type": "col",
                "value": "USER_NAME",
            },
            {
                "active": true,
                "text": "User ID",
                "type": "col",
                "value": "USER_ID",
            },
            {
                "active": false,
                "text": "User type",
                "type": "col",
                "value": "USER_TYPE",
            },
            {"active": false, "text": "", "type": "sep", "value": "sep"},
            {
                "active": true,
                "text": "Ascending",
                "type": "order",
                "value": "asc",
            },
            {
                "active": false,
                "text": "Descending",
                "type": "order",
                "value": "desc",
            }])
    })

    /*
    Will update both the UI and JSON-RPC model and will close the dropdown.
    */
    test('onSort', function() {
        var sortOptions = {
            // This are populated by the controller.
            'reset': null,
            'sort_up': null,
            'sort_down': null,
            // DB description.
            'default_column': 'USER_ID',
            'default_order': 'asc',
            'columns': [
                {'value': 'USER_NAME', 'text': 'Username'},
                {'value': 'USER_ID', 'text': 'User ID'},
                {'value': 'USER_TYPE', 'text': 'User type'},
            ],
        }
        var sut = getTableSort(sortOptions)
        sut.ctrl.sort_dropdown = mk.flowbiteDropdown(true)

        // Sort based on column.
        sut.ctrl.onSort({'type': 'col', 'value': 'USER_TYPE'})

        expect(sut.ctrl.sort_dropdown.isVisible()).toBeFalse()
        expect(sut.ctrl.tableOptions.sorting).toEqual({
            'USER_TYPE': 'asc',
        })
        expect(sut.ctrl.options).toEqual([
            {
                "active": false,
                "text": "Username",
                "type": "col",
                "value": "USER_NAME",
            },
            {
                "active": false,
                "text": "User ID",
                "type": "col",
                "value": "USER_ID",
            },
            {
                "active": true,  // The sorted column is not active.
                "text": "User type",
                "type": "col",
                "value": "USER_TYPE",
            },
            {"active": false, "text": "", "type": "sep", "value": "sep"},
            {
                "active": true,
                "text": "Ascending",
                "type": "order",
                "value": "asc",
            },
            {
                "active": false,
                "text": "Descending",
                "type": "order",
                "value": "desc",
            }])

        // Change sort order.
        sut.ctrl.onSort({'type': 'order', 'value': 'asc'})

        expect(sut.ctrl.sort_dropdown.isVisible()).toBeFalse()
        expect(sut.ctrl.tableOptions.sorting).toEqual({
            'USER_TYPE': 'asc',
        })
        expect(sut.ctrl.options).toEqual([
            {
                "active": false,
                "text": "Username",
                "type": "col",
                "value": "USER_NAME",
            },
            {
                "active": false,
                "text": "User ID",
                "type": "col",
                "value": "USER_ID",
            },
            {
                "active": true,  // The sorted column is not active.
                "text": "User type",
                "type": "col",
                "value": "USER_TYPE",
            },
            {"active": false, "text": "", "type": "sep", "value": "sep"},
            {
                "active": true,
                "text": "Ascending",
                "type": "order",
                "value": "asc",
            },
            {
                "active": false,
                "text": "Descending",
                "type": "order",
                "value": "desc",
            }])

    })

    /*
    Will inject a method to the sortOptions that can be used to reset
    the sorting.
    */
    test('onSort_reset', function() {
        var sortOptions = {
            // This are populated by the controller.
            'reset': null,
            'sort_up': null,
            'sort_down': null,
            // DB description.
            'default_column': 'USER_TYPE',
            'default_order': 'asc',
            'columns': [
                {'value': 'USER_NAME', 'text': 'Username'},
                {'value': 'USER_ID', 'text': 'User ID'},
                {'value': 'USER_TYPE', 'text': 'User type'},
            ],
        }
        var sut = getTableSort(sortOptions)
        sut.ctrl.sort_dropdown = mk.flowbiteDropdown(true)

        // Update sorting.
        sut.ctrl.onSort({'type': 'col', 'value': 'USER_NAME'})
        sut.ctrl.onSort({'type': 'order', 'value': 'desc'})

        sut.ctrl.sortOptions.reset()

        expect(sut.ctrl.options).toEqual([
            {
                "active": false,
                "text": "Username",
                "type": "col",
                "value": "USER_NAME",
            },
            {
                "active": false,
                "text": "User ID",
                "type": "col",
                "value": "USER_ID",
            },
            {
                "active": true,
                "text": "User type",
                "type": "col",
                "value": "USER_TYPE",
            },
            {"active": false, "text": "", "type": "sep", "value": "sep"},
            {
                "active": true,
                "text": "Ascending",
                "type": "order",
                "value": "asc",
            },
            {
                "active": false,
                "text": "Descending",
                "type": "order",
                "value": "desc",
            }])
    })


})
