/*
Tests for trial warning component.
*/

describe('TrialWarningCtrl', function() {

    var dayjs = require('dayjs')
    var storage_key = 'SFTPPlus_trial_confirmation'
    /*
    Will return an empty text if version is not trial,
    which is what we have for the default run.
    */
    test('init_no_expiration', function() {
        var sut = mk.getComponent('trialWarning', {})
        expect(sut.ctrl.expiration_description).toEqual('')
    })


    /*
    Will clear the expiration time and will set a local storage with the
    current timestamp.
    */
    test('onConfirmExpiration', function() {
        var sut = mk.getComponent('trialWarning', {})
        sut.ctrl.expiration_description = 'some-value'
        sut.ctrl.onConfirmExpiration()
        expect(sut.ctrl.expiration_description).toEqual('')
        var confirmation = parseInt(sut.$window.localStorage.getItem(
            storage_key))

        // Warning text is reset.
        expect(sut.ctrl.expiration_description).toEqual('')
        var now = dayjs().unix()
        // The confirmation is set for almost now.
        expect(confirmation).toBeWithin(now -1, now + 1)

    })

    /*
    If there is previous expiration confirmation and the current version
    is no longer a trial, clear the confirmation.
    */
    test('no_longer_trial', function() {
        var sut = mk.getComponent('trialWarning', {})
        sut.$window.localStorage.setItem(storage_key, 'any')

        var result = sut.ctrl._getExpirationDescription()
        expect(result).toEqual('')
        var confirmation = sut.$window.localStorage.getItem(
            storage_key)
        expect(confirmation).toBeNull()
    })

    /*
    If there is previous expiration confirmation and currend date is not
    one day past that confirmation, do nothing.
    */
    test('not_show_again_yet', function() {
        var sut = mk.getComponent('trialWarning', {})
        var now = dayjs().unix()
        // This is last day... but just one second less.
        var past_confirmation = now - 24 * 60 * 60 + 1
        sut.$window.localStorage.setItem(storage_key, past_confirmation)
        // Trial only expires in a long time.
        sut.utils.constants.EXPIRATION_TIMESTAMP = now + 10999

        var result = sut.ctrl._getExpirationDescription()
        expect(result).toEqual('')
        var confirmation = sut.$window.localStorage.getItem(storage_key)
        expect(confirmation).toEqual(past_confirmation.toString())
    })

    /*
    If there is previous expiration confirmation and currend date is
    more than 1 day after that confirmation, show the info about the
    expiration time.
    */
    test('show_again_yet', function() {
        var sut = mk.getComponent('trialWarning', {})
        var now = dayjs().unix()
        var past_confirmation = now - 24 * 60 * 60 - 1
        sut.$window.localStorage.setItem(
            storage_key, past_confirmation)
        // Trial only expires in a long time.
        sut.utils.constants.EXPIRATION_TIMESTAMP = now + 10999

        var result = sut.ctrl._getExpirationDescription()
        expect(result).toStartWith('in 3 hours on')
        // The last confirmation is left as it is.
        var confirmation = sut.$window.localStorage.getItem(storage_key)
        expect(confirmation).toEqual(past_confirmation.toString())
    })

    /*
    If there is no previous confirmation return the description of the
    time when the trial expires.
    */
    test('no_previous_confirmation', function() {
        var sut = mk.getComponent('trialWarning', {})
        var now = dayjs().unix()
        sut.utils.constants.EXPIRATION_TIMESTAMP = now + 12

        var result = sut.ctrl._getExpirationDescription()
        expect(result).toStartWith('in a few seconds on')
        // No confirmation is set.
        var confirmation = sut.$window.localStorage.getItem(storage_key)
        expect(confirmation).toBeNull()
    })

})
