/*
Directives used in the code

When scope is `false` (default), directive has the exact/same scope as parent.
When scope is `true` - start with a copy of the parent, then separate
When scope is {} - has a private empty scope, unless bindings are defined
based on:
‘@’ – Text binding / one-way binding
‘=’ – Direct model binding / two-way binding
‘&’ – Behavior binding / Method binding

*/

(function(exports){

var init = require('../../js_commons/__init__.js', 'init')
/*
Formatter and parser for a model holding the configuration value for the
rotate_on options.
*/
exports.logRotateOn = function() {
    return {
        require: 'ngModel',
        link: function(scope, element, attrs, ngModel) {

            var _badLogRotateOnValue = function(calendar_type) {
                ngModel.$setValidity('log-rotate-on', false)
                return [0, calendar_type]
            }

            /*
            Convert data from view to model.
            */
            ngModel.$parsers.push(function(data) {
                ngModel.$setValidity('log-rotate-on', true)

                if (!ngModel.$modelValue) {
                    var calendar_type = 'day-of-month'
                } else {
                    var calendar_type = ngModel.$modelValue[1]
                }

                if (calendar_type == 'time-of-day') {
                    // Validates the string format (should be H:M) and
                    // convert to seconds.
                    var timeRegExp = /^\d{1,2}:\d{1,2}$/

                    if (!timeRegExp.test(data)) {
                        return _badLogRotateOnValue(calendar_type)
                    }

                    var time = data.split(':')
                    var hour = time[0]
                    var minute = time[1]

                    if (hour > 23 || minute > 59) {
                        return _badLogRotateOnValue(calendar_type)
                    }

                    data = (parseInt(time[0]) * 60 + parseInt(time[1])) * 60
                } else {
                    // Validates it is an integer between 0 and 31.
                    var intRegExp = /^\d+$/
                    if (!intRegExp.test(data) || data < 0 || data > 31) {
                        return _badLogRotateOnValue(calendar_type)
                    }
                    data = parseInt(data)
                }
                return [data, calendar_type]
            })

            /*
            Convert data from model to view.
            */
            ngModel.$formatters.push(function(data) {
                if (!data) {
                    return 0
                }
                if (data[1] == 'day-of-month') {
                    return data[0]
                } else if (data[1] == 'time-of-day') {
                    var minutes = Math.floor(data[0] / 60);
                    var hours = Math.floor(minutes / 60);
                    minutes = minutes % 60;
                    if (hours < 10) {
                        hours = '0' + hours
                    }
                    if (minutes < 10) {
                        minutes = '0' + minutes
                    }
                    return hours + ':' + minutes
                }
            })
        }
    }
}

/*
Directive for configuration log rotation based on calendar.

It allows to configure calendar type that defines the period of time
when log rotation should occur.
*/
exports.logRotateOnCalendarType = function() {
    return {
        require: 'ngModel',
        link: function(scope, element, attrs, ngModel) {

            /*
            Convert data from view to model.
            */
            ngModel.$parsers.push(function(data) {
                if (!data) {
                    data = 'day-of-month'
                }
                return [0, data]
            })

            /*
            Convert data from model to view.
            */
            ngModel.$formatters.push(function(data) {
                if (!data) {
                    return 'day-of-month'
                }
                return data[1]
            })
        }
    }
}

/*
Only allow input data in integer format.
*/
exports.integer = function() {
    var INTEGER_REGEXP = /^\-?\d*$/;
    return {
        require: 'ngModel',
        link: function(scope, elm, attrs, ngModel) {
            ngModel.$render = function() {
                if (ngModel.$viewValue == 'inherit') {
                    elm.val('')
                    return
                }
                elm.val(ngModel.$viewValue)
            }
            ngModel.$parsers.unshift(function(viewValue) {
                if (!viewValue || ! INTEGER_REGEXP.test(viewValue)) {
                    ngModel.$setValidity('integer', false)
                    return undefined
                } else {
                    ngModel.$setValidity('integer', true)
                    return viewValue
                }
            })
        }
    }
}


/*
Only allow input data in integer or float.
*/
exports.floatOrInteger = function() {
    var FLOAT_REGEXP = /^\-?\d+((\.|\,)\d+)?$/;
    return {
        require: 'ngModel',
        link: function(scope, elm, attrs, ngModel) {
            var parser = function(viewValue) {
                if (!viewValue || !FLOAT_REGEXP.test(viewValue)) {
                    ngModel.$setValidity('floatOrInteger', false)
                    return undefined
                } else {
                    ngModel.$setValidity('floatOrInteger', true)
                    var value = parseFloat(viewValue.replace(',', '.'))
                    var minimum = parseFloat(attrs.min)
                    var maximum = parseFloat(attrs.max)
                    if (!isNaN(minimum) && value < minimum) {
                        value = minimum
                        ngModel.$setValidity('floatOrInteger', false)
                    } else if (!isNaN(maximum) && value > maximum) {
                        value = maximum
                        ngModel.$setValidity('floatOrInteger', false)
                    }
                    return value
                }
            }

            ngModel.$parsers.unshift(parser)

        }
    }
}


/*
Force default non-empty value for input field.
*/
exports.defaultEmpty = function(utils) {
    return {
        require: 'ngModel',
        restrict: 'A',
        link: function(scope, element, attrs, ngModel) {
            var placeholder = attrs.defaultEmpty
            var unique_marker = attrs.defaultEmptyUnique
            element.bind('blur', function(event) {
                // Only set default value if value is empty so that we can
                // allow '0' as a valid value.
                // Only set default if model is invalid, ie set to undefined.
                if (ngModel.$viewValue !== '' &&
                    init.typeOf(ngModel.$modelValue) !== 'undefined'
                        ) {
                    return
                }
                if (unique_marker) {
                    // The field is still invalid... we have just put the
                    // placeholder to avoid 'null' value and keep it unique.
                    placeholder = placeholder + utils.getRandomMark()
                }

                // Force updating the UI with the placeholder.
                // This is the Angular way of doing
                // event.currentTarget.value = placeholder
                // but via the pipeline
                ngModel.$setViewValue(placeholder)
                ngModel.$render()

                // Since the form was set with a default value, we mark it
                // as invalid to draw attention to it.
                ngModel.$setValidity('empty', false)

            })

            var validator = function(value) {
                if (value || value === '0' || value === 0) {
                    ngModel.$setValidity('empty', true)
                    return value
                }
                ngModel.$setValidity('empty', false)
                return ''
            }

            ngModel.$formatters.push(validator)
            ngModel.$parsers.unshift(validator)
        }
    }
}


exports.loadingOverlay = function() {
    return {
        restrict: 'A',
        scope: false,
        link: function(scope, element, attrs) {
            var loadingLayer = (
                $('<div class="loading"></div>').appendTo(element))
            $(element).addClass('loading-container')
            scope.$watch(attrs.loadingOverlay, function(value) {
                loadingLayer.toggle(value)
            })
        }
    }
}


/*
Formatter and parser for a model holding an array of strings, each string
on a separate line.
*/
exports.multiLine = function(utils) {
    return {
        require: 'ngModel',
        link: function(scope, element, attrs, ngModel) {

            /*
            Convert data from view to model.
            */
            ngModel.$parsers.push(function(data) {
                if (!data) {
                    return []
                }
                return utils.textToMultiLines(data)
            })

            /*
            Convert data from model to view.
            */
            ngModel.$formatters.push(function(data) {
                if (!data) {
                    return ''
                }
                return utils.multiLinesToText(data)
            })
        }
    }
}

/*
Formatter and parser for a model holding an array of arrays.

The view value is a multi-line text with space separated values on each line.
*/
exports.multiLineSSV = function(utils) {
    return {
        require: 'ngModel',
        link: function(scope, element, attrs, ngModel) {
            /*
            Convert data from model to view.
            */
            ngModel.$formatters.push(function(data) {
                if (!data) {
                    return ''
                }
                var lines = []
                angular.forEach(
                    data,
                    function(line) {
                        lines.push(line.join(' '))
                    }
                )
                return utils.multiLinesToText(lines)
            })

            /*
            Convert data from view to model.
            */
            ngModel.$parsers.push(function(data) {
                if (!data) {
                    return []
                }
                var result = []
                angular.forEach(
                    utils.textToMultiLines(data),
                    function(line) {
                        result.push(line.split(/ +/g))
                    }
                )
                return result
            })
        }
    }
}


/*
Formatter and parser for a model holding an array from comma separated fields.
*/
exports.commaSeparatedFields = function(utils) {
    return {
        require: 'ngModel',
        link: function(scope, element, attrs, ngModel) {

            /*
            Convert data from view to model.
            */
            ngModel.$parsers.push(function(data) {
                if (!data) {
                    return []
                }
                var result = utils.textToMultiSequence(data)
                return result[0]
            })

            /*
            Convert data from model to view.
            */
            ngModel.$formatters.push(function(data) {
                if (!data) {
                    return ''
                }
                return utils.multiSequenceToText([data])
            })
        }
    }
}


/*
Formatter and parser for a model holding an array of arrays
from comma separated values.
*/
exports.commaSeparatedTextarea = function(utils) {
    return {
        require: 'ngModel',
        link: function(scope, element, attrs, ngModel) {

            /*
            Convert data from view to model.
            */
            ngModel.$parsers.push(function(data) {
                return utils.textToMultiSequence(data)
            })

            /*
            Convert data from model to view.
            */
            ngModel.$formatters.push(function(data) {
                if (!data) {
                    return ''
                }
                return utils.multiSequenceToText(data)
            })
        }
    }
}


/*
Simple wrapper to handle the HTML element attribute evaluation with
support for default values.
*/
var evaluate_attribute = function(scope, attribute, default_value) {
    if (!attribute) {
        return default_value
    }
    return scope.$eval(attribute)
}

/*
Simple wrapper to simulate an HTML element update via UI so that we can
trigger the ng-model internal update via common methods.

This is synchronous on purpose, so that after calling this you can do
other manipulations.
*/
var trigger_ui_update = function(scope, ngModel, value) {
    scope.$apply(function() {
        ngModel.$setViewValue(value)
    })
}


/*
This takes an ng-model enabled HTML element and handles the model
updates via tagify.

The directive has no `$render` function as tagify takes care of its own
rendering.

The whitelist can be a list of raw values or a list of UUID values.
The UUID values are displayed using the name of the component with that UUID,
*/

var TAGIFY_DROPDOWN_OPTIONS = {
    enabled: 0,
    position: 'all',
    maxItems: 100,
    // For now we allow a single selection.
    // In the future, this might be configured via the
    // element attributes.
    closeOnSelect: true,
    searchKeys: ['value'],
    caseSensitive: false,
    // Not only start of the string.
    fuzzySearch: true
}
exports.tagify = function(utils) {

    /*
    Convert our component into something that can be handled
    internally by tagify.

    It can also handle simple non-component string values.
    */
    var get_item = function(component) {

        if (typeof(component) != 'object') {
            // We have a simple array.
            return {
                'value': component,
                'uuid': component
            }
        }

        // We have a component.
        return {
            'value': component.name,
            'uuid': component.uuid
        }
    }

    /*
    This is just helper code extracted from the link function.
    */
    var _get_whitelist = function(raw_whitelist) {
        var whitelist = []
        var uuid_whitelist = {}
        angular.forEach(
            raw_whitelist,
            function(component) {

                if (component.uuid) {
                    uuid_whitelist[component.uuid] = component
                } else {
                    /*
                    We don't have an UUID base whitelist.
                    the whitelist is based on exact tag value.
                    */
                    uuid_whitelist = null
                }
                whitelist.push(get_item(component))

            })
        return {
            'whitelist': whitelist,
            'uuid_whitelist': uuid_whitelist
        }
    }

    return {
        require: 'ngModel',
        link: function(scope, element, attrs, ngModel) {

            if (!scope.configuration) {
                // Configuration not yet loaded, no need to try.
                return
            }

            var raw_whitelist = evaluate_attribute(
                scope, attrs.tagifyWhitelist, [])

            var result = _get_whitelist(raw_whitelist)
            var uuid_whitelist = result.uuid_whitelist

            var user_input = false
            if (attrs.tagifyUserInput === "") {
                user_input = true
            }
            var placeholder = 'Click to select from list'
            if  (attrs.tagifyPlaceholder) {
                placeholder = attrs.tagifyPlaceholder
            }

            // Create the tagify handler.
            // See https://github.com/yairEO/tagify/tree/v3.22.3#settings
            var tagify_element = new Tagify(element[0], {
                whitelist: result.whitelist,
                enforceWhitelist: !user_input,
                userInput: user_input,
                editTags: false,
                addTagOnBlur: user_input,
                placeholder: placeholder,
                dropdown: TAGIFY_DROPDOWN_OPTIONS
            })
            /*
            Let angular know of any UI change.
            */
            tagify_element.on('change', function(event) {
                if (tagify_element.inProgress) {
                    // We don't want to update the model when the UI is
                    // updated based on the model itself.
                    // So here we avoid a recursion.
                    return
                }
                trigger_ui_update(
                    scope, ngModel, JSON.stringify(event.detail.tagify.value))
            })

            /*
            Convert data from view to model.

            This is triggerred when we programmatically update the view via
            the tagify 'change' event.
            */
            ngModel.$parsers.push(function(data) {
                var model = []
                if (!data) {
                    return model
                }
                var data = JSON.parse(data)
                angular.forEach(
                    data,
                    function(tag) {
                        if (!tag.uuid) {
                            // We got some unknown value from tagify.
                            if (!user_input) {
                                // We don't allow random input.
                                // don't forward it to our model.
                                return
                            }
                            tag.uuid = tag.value
                        }
                        model.push(tag.uuid)
                    })
                return model
            })

            /*
            Convert data from model to view.

            This is called when the page is loaded and we get the starting
            value for the UI.
            */
            ngModel.$formatters.push(function(data) {
                if (!data) {
                    // Nothing to update.
                    return ''
                }
                var tags = []
                angular.forEach(
                    data,
                    function(uuid) {
                        var component = uuid
                        if (uuid_whitelist) {
                            component = uuid_whitelist[uuid]
                        }
                        if (user_input && !component) {
                            component = uuid
                        }
                        tags.push(get_item(component))
                    })

                if (tags.length == 0) {
                    // No tags yet in the model.
                    // We don't push empty tags as tagify fails.
                    return ''
                }

                tagify_element.removeAllTags()
                // We don't want to trigger model updates while updating
                // the UI based on the model itself.
                tagify_element.inProgress = true
                tagify_element.addTags(tags)
                tagify_element.inProgress = false
                // We don't store anything inside the actual element's value.
                return ''
            })
        }
    }
}


/*
This is the UI widget that uses tagify to configure SSL cipher list.
*/
exports.sslCiphers = function(utils) {

    /*
    Convert our component into something that can be handled
    internally by tagify.

    It can also handle simple non-component string values.
    */
    var _getAllSSLCiphers = function() {
        var result = []
        result.push.apply(result, utils.constants.SSL_CIPHERS)
        return result
    }

    return {
        require: 'ngModel',
        link: function(scope, element, attrs, ngModel) {

            // Create the tagify handler.
            // See https://github.com/yairEO/tagify/tree/v3.22.3#settings
            var tagify_element = new Tagify(element[0], {
                whitelist: _getAllSSLCiphers(),
                enforceWhitelist: false,
                skipInvalid: false,
                userInput: true,
                editTags: true,
                addTagOnBlur: true,
                placeholder: 'Click here to select ciphers.',
                dropdown: TAGIFY_DROPDOWN_OPTIONS
            })

            /*
            Let angular know of any UI change.
            */
            tagify_element.on('change', function(event) {
                if (tagify_element.inProgress) {
                    // We don't want to update the model when the UI is
                    // updated based on the model itself.
                    return
                }
                trigger_ui_update(
                    scope, ngModel, JSON.stringify(event.detail.tagify.value))
            })

            /*
            Convert data from view to model.
            */
            ngModel.$parsers.push(function(data) {
                if (!data) {
                    return ''
                }
                var data = JSON.parse(data)
                var model = []
                angular.forEach(
                    data,
                    function(tag) {
                        model.push(tag.value)
                    })

                return model.join(':')
            })

            /*
            Convert data from model to view.

            This is called when the page is loaded and we get the starting
            value for the UI.
            */
            ngModel.$formatters.push(function(data) {
                if (!data) {
                    // Nothing to update.
                    return ''
                }
                var tags = []

                angular.forEach(
                    data.split(':'),
                    function(cipher) {
                        cipher = cipher.trim()
                        if (!cipher) {
                            return
                        }

                        tags.push(cipher)
                    })

                if (tags.length == 0) {
                    // No tags yet in the model.
                    // We don't push empty tags as tagify fails.
                    return ''
                }

                tagify_element.removeAllTags()
                // We don't want to trigger model updates while updating
                // the UI based on the model itself.
                tagify_element.inProgress = true
                tagify_element.addTags(tags)
                tagify_element.inProgress = false
                // We don't store anything inside the actual element's value.
                return ''
            })
        }
    }
}


/*
This is designed to be used for <input> element.
It will trigger the delayed callback when the current input value is stable.
It takes a required value which is the name of function that will be called.

For now, the ng-model is unidirectional.
It only supports reading the current filter value from the controller.
To be extended to bidirectional later, if needed.

Usage:
    <input
      input-trigger="filterVariables"
      ng-model="some_scope_view_model"
      >
*/
exports.inputTrigger = function(utils, $timeout) {

    // Number of miliseconds to wait for triggering,
    // after entering the input value.
    var default_wait_time = 400
    // The minimum number of characters to trigger.
    var minimum_length = 3

    return {
        require: 'ngModel',
        link: function(scope, element, attrs, ngModel) {

            // The pending delayed filtering while typing the filter.
            var pending_trigger
            var filterCallback = scope[attrs.inputTrigger]

            // Called when value changed.
            // It handles scheduling of the trigger.
            element.on('input', function(event) {
                var input_value = event.currentTarget.value

                if (!input_value) {
                    cancelPreviousTrigger()
                    scheduleTrigger()
                    return
                }
                if (input_value.length < minimum_length) {
                    cancelPreviousTrigger()
                    return
                }

                scheduleTrigger(event.currentTarget.value)
            })

            var scheduleTrigger = function(input_value) {
                cancelPreviousTrigger()
                pending_trigger = $timeout(function() {
                    filterCallback(input_value)
                    },
                    default_wait_time
                )
            }
            var cancelPreviousTrigger = function() {
                if (!pending_trigger) {
                    return
                }
                $timeout.cancel(pending_trigger)
            }
        }
    }
}

/*
This is designed to update the UI of any configuration option
that is overridden by node variables.
*/
exports.disableOverriddenOptions = function() {

    return {
        require: [],
        link: function(scope, element, attrs) {

            var node_variables = []
            if (scope.configuration && scope.configuration.node_variables) {
                angular.forEach(
                    scope.configuration.node_variables,
                    function(value, key) {
                        node_variables.push(value)
                    })
            }

            angular.forEach(
                element[0].querySelectorAll('input, textarea, button'),
                function(child) {
                    var is_disabled = false
                    angular.forEach(node_variables, function(variable) {
                        if (variable.section != scope.section_uuid) {
                            return
                        }
                        if (variable.key != child.id) {
                            return
                        }
                        is_disabled = true
                    })

                    child.disabled = is_disabled
                })

        }
    }
}


/*
Create a flowbite based dropdown menu.

It can inject the dropdown reference to the local scope,
using the "data-scope-name" attribute.

When `own-reference` is use, the scope must have that member defined,
even if initialized just with `null`, but it need to be present.
It will not auto-add it to the scope.

Usage:

<dropdown
  button-name="'Text of dropdown' | scope.reference"
  own-reference="location_dropdown"
  data-placement="bottom"
  >
  <dropdown-button>
    dropdown button
  </dropdown-button>
  <any-content-wrapper>
    <div>Some Header</div>
    <ul>
        <li ng-repeat="db_type in getLocationsTypes()">
            <a
            ng-click="onAddLocation(db_type)"
            ng-bind="db_type | componentType:'location'"
            >
            </a>
        </li>
    </ul>
  </any-content-wrapper>
</dropdown>
*/
exports.dropdown = function(utils) { return {
    require: [],
    restrict: 'AE',
    scope: {
        'buttonName': '=',
        'ownReference': '=',
    },
    link: function(scope, element, attrs) {

        var placement = 'left-start'
        if (attrs.placement && !utils.ie_version.IsIE) {
            // On IE, only left-start placement works.
            placement = attrs.placement
        }
        var button = document.createElement('button')
        button.setAttribute('type', 'button')

        var button_text = document.createElement('span')
        button_text.textContent = scope.buttonName || ''
        button.appendChild(button_text)

        scope.$parent.$watch(attrs.buttonName, function(current, old) {
            button_text.textContent = scope.buttonName || ''
        })

        var caret_icon = document.createElement('i')
        caret_icon.setAttribute('class', 'fas fa-caret-down ml-2')

        button.appendChild(caret_icon)

        // See if we have a dedicated <dropdown-button> tag.
        angular.forEach(element[0].childNodes, function(child) {
            if (child.nodeName == 'DROPDOWN-BUTTON') {
                button = element[0].children[0]
            }
        })

        button.setAttribute('data-dropdown-toggle', 'dropdown')

        var target = document.createElement('div')

        // Move the test attribute to the trigger button.
        button.setAttribute('data-test', attrs.test)
        element[0].removeAttribute('data-test')

        target.setAttribute('class', 'dropdown-content hidden')

        // Move all content to the target drowpdown element.
        while (element[0].childNodes.length > 0) {
            target.appendChild(element[0].childNodes[0]);
        }

        // Button at the start.
        element.prepend(button)

        // Target dropdown after the button.
        element[0].appendChild(target)

        // Trigger flowbite login.
        var dropdown_instance = new Dropdown(
            target, button, {'placement': placement})

        if (!angular.isUndefined(scope.ownReference)) {
            scope.ownReference = dropdown_instance
        }
    }
}}


var SECONDS_IN_MINUTE = 60
var SECONDS_IN_HOUR = SECONDS_IN_MINUTE * 60
var SECONDS_IN_DAY = SECONDS_IN_HOUR * 24

var DISABLED = 'disabled'
var MINUTES = 'minutes'
var SECONDS = 'seconds'
var HOURS = 'hours'
var DAYS = 'days'
/*
Configuration UI for seconds.

If the value can be 0 (disabled), don't define a `data-min`.
If the value doesn't need a restart, don't define the `data-restart-required`.

When model contains invalid data, it will default to the minimum value.

Usage:
  <config-seconds
    ng-model="some_parent.scope_view_model"
    data-label="TEST FOR THE LABEL"
    data-min="42"
    data-on-update="someFunction()"
    data-restart-required
    >
    <help-block>
      RAW HTML USED FOR HELP.
    </help-block>
  </config-seconds>

*/
exports.configSeconds = function(utils, CACHE_BUSTER) {

    var linkFn = function(scope, element, attrs) {
        scope.label_text = attrs.label
        scope.input_id = attrs.ngModel
        scope.is_restart_required = attrs.restartRequired != undefined
        scope.has_disabled = true

        var default_value = 0
        var value_input = element.find('input')
        value_input.attr('min', '0')
        if (attrs.min) {
            value_input.attr('min', attrs.min)
            scope.has_disabled = false
            // Min will be the default value.
            default_value = parseInt(attrs.min)
        }

        var select = element.find('select')

        /*
        Called when the model was changed and the view should be updated.

        Updating the view should not trigger any other update event.
        */
        var onModelChange = function() {
            var value = scope.ngModel
            select = element.find('select')
            if (!value || value <= 0) {
                // The model should not normally be negative.
                if (scope.has_disabled) {
                    // Set it to disable.
                    select.val(DISABLED)
                    value_input.val('0')
                    scope.ngModel = 0
                    return
                }
                // Or to the minimum or default value.
                value_input.val(default_value.toString())
                select.val(SECONDS)
                scope.ngModel = default_value
                return
            }

            if (value % SECONDS_IN_DAY == 0) {
                // We have a fixed day value.
                select.val(DAYS)
                value_input.val(value / SECONDS_IN_DAY + '')
                return
            }

            if (value % SECONDS_IN_HOUR == 0) {
                // We have a fixed number of hours
                select.val(HOURS)
                value_input.val(value / SECONDS_IN_HOUR + '')
                return
            }

            if (value % SECONDS_IN_MINUTE == 0) {
                // We have a fixed number of minutes
                select.val(MINUTES)
                value_input.val(value / SECONDS_IN_MINUTE + '')
                return
            }

            // We have raw seconds.
            select.val(SECONDS)
            value_input.val(value.toString())
        }


        /*
        Called when the view was changed and the model needs to be updated.
        */
        var onViewChange = function(args) {
            var value = parseInt(value_input.val())
            if (value <= 0) {

                if (scope.has_disabled) {
                    // Any negative number is set to disabled.
                    select.val(DISABLED)
                    write_model(0)
                    return
                }

                // Or to the minimim or default value.
                select.val(SECONDS)
                value_input.val(default_value.toString())
                write_model(default_value)
                return
            }
            if (select.val() == DISABLED) {
                // It used to be disabled,
                // now switch to seconds.
                select.val(SECONDS)
                return write_model(value)
            }

            if (select.val() == MINUTES) {
                // We have minutes
                return write_model(value * SECONDS_IN_MINUTE)
            }
            if (select.val() == HOURS) {
                // We have hours
                return write_model(value * SECONDS_IN_HOUR)
            }
            if (select.val() == DAYS) {
                // We have days
                return write_model(value * SECONDS_IN_DAY)
            }
            // We have seconds
            return write_model(value)
        }

        /*
        Called when the model is updated.
        */
        scope.$parent.$watch(attrs.ngModel, function(current, old) {
            if (current === undefined) {
                return
            }

            onModelChange()
        })

        /*
        Update the model and notify others that the value was
        changed.
        */
        var write_model = function(value) {
            scope.$apply(function() {
                scope.ngModel = value
            })
        }

        /*
        Called when value is changed.
        */
        value_input.on('input', function(event) {
            onViewChange({'from_input': true})
        })

        /*
        Called when measurement unit is changed.
        */
        select.on('change', function(event) {
            // Update based on curently selected unit measurements.
            // The UI is already updated, so just trigger a model update.
            var result = value_input.val()

            if (select.val() == DISABLED) {
                // Changed to disabled state.
                value_input.val('0')
            } else if (result == '0') {
                // Previous value was disabled and a new unit was
                // selected
                // Go with 10 as an arbitrary default value.
                value_input.val('10')
            }
            onViewChange({'from_input': false})
        })

        // Set initial value to get the view in sync with model.
        onModelChange()
    }

    return {
        transclude: true,
        scope: {'ngModel': '='},
        templateUrl: (
            utils.constants.BASE_URL_PATH +
            '/directive/config-seconds.html?v=' + CACHE_BUSTER),
        link: linkFn,
    }
}


/* c8 ignore next */
}(typeof exports === 'undefined' ? this.directives = {} : exports))
