/*
Tests for directives.

For these test we load a real testing app for which we enable the directives
under test.
*/


describe('Directives', function() {
    // We using our real application to make sure all directives
    // are registered in our production app.
    require('../js/app.js')

    // Helper variables used for all the tests.
    var scope
    var dom
    var input_field

    // Helper methods to handle the Angular internal processing/
    var viewUpdate
    // Raw compile.
    var compile
    // Compile while wrapping in an ng-form
    var compileForm
    var utils
    var $templateCache

    beforeEach(function() {

        angular.mock.module('app')
        var fs = require('fs')
        angular.module('app').run(function($templateCache) {
            $templateCache.put(
                '/rp/path/directive/config-seconds.html?v=TEST_BUSTER',
                fs.readFileSync(
                    ('chevah/server/static/manager_service' +
                    '/directive/config-seconds.html'),
                    'utf8')
                    )
        })

        inject(function($injector, _$compile_, $rootScope) {
            utils = $injector.get('utils')
            $templateCache = $injector.get('$templateCache')
            scope = $rootScope.$new()
            // Setup the scope for all tests.
            // In an ideal case each directive will have its own scope
            // but for now we got with this legacy code.
            scope.integer = 0
            scope.float_or_integer = 0
            scope.calendar_value = undefined
            scope.calendar_type = ''

            compile = function (template) {

                var element = _$compile_(template)(scope)
                scope.$digest()
                return element
            }
            compileForm = function (template) {
                var element = _$compile_(
                    '<div ng-form="testForm">' + template+ '</div')(scope)
                scope.$digest()
                return element
            }
            viewUpdate = mk.getViewUpdate(scope)
        })
    })


    describe('integer', function() {

        beforeEach(function() {

            dom = compile(
                '<div ng-form="testForm">' +
                '<input ' +
                'type="number" ' +
                'integer ' +
                'ng-model="integer" ' +
                'name="integer" ' +
                'required ' +
                '>' +
                '</div>'
                )
            input_field = dom.find('input')[0]
        })

        test(
        'No errors are set for valid scope integer data and view is updated.',
        function() {
            scope.$apply(function() {
                scope.integer = 123
            })

            expect(input_field.value).toBe('123')
            expect(scope.testForm.$valid).toBeTrue()
        })

        test(
        'Updating the view, will update the scope.',
        function() {
            viewUpdate(input_field, '123')

            expect(scope.integer).toBe(123)
            expect(scope.testForm.$valid).toBeTrue()
        })

        test(
        'Updating the view with a text value, will delete the value from ' +
        'scope and update the form validation data.',
        function() {
            viewUpdate(input_field, 'no-integer')

            expect(scope.integer).toBeUndefined()
            expect(scope.testForm.$valid).toBeFalse()
            expect(scope.testForm.integer.$valid).toBeFalse()
            expect(scope.testForm.integer.$error.integer).toBeTrue()
        })

        test(
        'Updating the view with a float value, will delete the value from ' +
        'scope and update the form validation data.',
        function() {
            viewUpdate(input_field, '10.4')

            expect(scope.integer).toBeUndefined()
            expect(scope.testForm.$valid).toBeFalse()
            expect(scope.testForm.integer.$valid).toBeFalse()
            expect(scope.testForm.integer.$error.integer).toBeTrue()
        })

    })


    describe('defaultEmpty', function() {

        /*
        Return an input field with default empty directive.
        */
        getInputField = function(placeholder, unique) {
            if (!placeholder) {
                placeholder = 'Empty Place-holder'
            }
            if (unique) {
                unique = 'default-empty-unique="string"'
            } else {
                unique = ''
            }
            var html = '<div ng-form="testForm">' +
                '<input ' +
                'name="test_field" ' +
                'ng-model="test_variable" ' +
                'default-empty="%(placeholder)s" ' +
                '%(unique)s ' +
                '>' +
                '</div>'
            dom = compile(html.supplant({
                'placeholder': placeholder,
                'unique': unique
                }))
            return dom.find('input')
        }

        test(
        'No errors are set for non-empty model and view is updated.',
        function() {
            input_field = getInputField()
            scope.$apply(function() {
                  scope.test_variable = 'something';
            })

            expect(input_field.val()).toBe('something')
            expect(scope.testForm.$valid).toBeTrue()
        })

        test(
        'No errors are set for non-empty view and model is updated.',
        function() {
            input_field = getInputField()
            viewUpdate(input_field, 'new-value')

            expect(scope.test_variable).toBe('new-value')
            expect(scope.testForm.$valid).toBeTrue()
        })

        test(
        '`0` string is accepted as non-empty and model is updated.',
        function() {
            input_field = getInputField()
            viewUpdate(input_field, '0')

            expect(scope.test_variable).toBe('0')
            expect(scope.testForm.$valid).toBeTrue()
        })

        test(
        'For empty model values the invalid form is set and model value is ' +
        'kept at it is only updated on blur event.',
        function() {
            input_field = getInputField()

            scope.$apply(function() {
                  scope.test_variable = '';
            })

            expect(scope.test_variable).toBe('')
            expect(input_field.val()).toBe('')
            expect(scope.testForm.$valid).toBeFalse()
            expect(scope.testForm.test_field.$error.empty).toBeTrue()
        })

        /*
        For empty view values the model value is not updated until blur
        event when is set to placeholder.
        */
        test('empty_no_update', function() {
            var placeholder = mk.string()
            input_field = getInputField(placeholder)
            viewUpdate(input_field, '')

            expect(scope.test_variable).toBeUndefined()
            expect(scope.testForm.$valid).toBeFalse()
            expect(scope.testForm.test_field.$error.empty).toBeTrue()

            viewUpdate(input_field, '', 'blur')

            expect(input_field.val()).toBe(placeholder)
            expect(scope.test_variable).toBe(placeholder)
            // Form is not valid as it was set with a default value.
            expect(scope.testForm.$valid).toBeFalse()
            expect(scope.testForm.test_field.$error.empty).toBeTrue()
        })

        /*
        For empty view values the invalid form is set and model value is
        updated as empty. When blur is triggered it is set to placeholder
        while keeping the error state.

        This is also a test to check that with an unique plceholder a random
        value is added.
        */
        test('empty_input_with_unique_default', () => {
            scope.test_variable = 'some-initial-value';
            var placeholder = 'Some placeholder value'
            input_field = getInputField(placeholder, 'unique')

            // Delete the initial value and see that model is updated.
            viewUpdate(input_field, '')

            expect(scope.test_variable).toEqual('')
            expect(scope.testForm.$valid).toBeFalse()
            expect(scope.testForm.test_field.$error.empty).toBeTrue()

            // Then the field input UI is left via blur.
            viewUpdate(input_field, '', 'blur')

            // It is not the exact placeholder, but start with it.
            expect(input_field.val()).not.toBe(placeholder)
            expect(scope.test_variable).not.toBe(placeholder)
            expect(input_field.val()).toStartWith(placeholder)
            expect(scope.test_variable).toStartWith(placeholder)
            // Form is now invalid.
            expect(scope.testForm.$valid).toBeFalse()
            expect(scope.testForm.test_field.$error.empty).toBeTrue()
        })

        test(
        'When model is undefined, consider this an empty value and for the ' +
        'default non-empty value.',
        function() {
            var placeholder = '-10.4'
            input_field = getInputField(placeholder)
            scope.test_variable = undefined
            viewUpdate(input_field, '23a', 'blur')

            expect(scope.test_variable).toEqual(placeholder)
            expect(input_field.val()).toEqual(placeholder)
            expect(scope.testForm.$valid).toBeFalse()
        })

    })


    describe('commaSeparatedFields', function() {

        beforeEach(function() {
            dom = compile(
                '<div ng-form="testForm">' +
                '<input ' +
                'type="text" ' +
                'comma-separated-fields ' +
                'ng-model="field" ' +
                'name="field" ' +
                '>' +
                '</div>'
                )
            input_field = dom.find('input')[0]
        })

        test(
        'An array is converted to a string with items separated by comma.',
        function() {
            scope.$apply(function() {
                  scope.field = ['field1', 'fie"ld2', 'field3, comma']
            })

            expect(input_field.value).toEqual(
                'field1, "fie""ld2", "field3, comma"')
            expect(scope.testForm.$valid).toBeTrue()
        })

        test(
        'Updating the view, will update the scope.',
        function() {
            viewUpdate(
                input_field,
                'field1,field2, field3, field"4, "field, 5", "field6"')

            expect(scope.field).toEqual([
                'field1', 'field2', 'field3', 'field"4', 'field, 5', 'field6'])
            expect(scope.testForm.$valid).toBeTrue()
        })

        test(
        'An empty array is converted an empty string.',
        function() {
            scope.$apply(function() {
                  scope.field = []
            })

            expect(input_field.value).toEqual('')
            expect(scope.testForm.$valid).toBeTrue()
        })

        test(
        'An empty input is update to empty array',
        function() {
            // We need to put some value first, as is empty by default
            // and will not trigger the update.
            viewUpdate(input_field, 'field1,field2')
            expect(scope.field).toEqual(['field1', 'field2'])
            // Then we delete the input.
            viewUpdate(input_field, '')

            expect(scope.field).toEqual([])
        })

    })


    describe('multiLineSSV', function() {

        beforeEach(function() {
            dom = compile(
                '<div ng-form="testForm">' +
                '<textarea ' +
                'multi-line-ssv ' +
                'ng-model="field" ' +
                '></textarea>' +
                '</div>'
                )
            input_field = dom.find('textarea')[0]
        })

        test(
        'An array or arrays is converted to a string with items separated ' +
        'spaces and newlinse.', () => {
            scope.$apply(() => {
                  scope.field = [
                      ['field1', 'fie"ld2'],
                      ['field3 comma', 'more']
                    ]
            })

            expect(input_field.value).toEqual(
                'field1 fie"ld2\nfield3 comma more')
            expect(scope.testForm.$valid).toBeTrue()
        })

        test(
            'Lines are converted to arraysa and ' +
            'empty input is update to empty array',
            function() {
                // We need to put some value first, as is empty by default
                // and will not trigger the update.
                viewUpdate(input_field, 'field1 field2\nother-line')
                expect(scope.field).toEqual(
                    [['field1', 'field2'], ['other-line']])
                // Then we delete the input.
                viewUpdate(input_field, '')

                expect(scope.field).toEqual([])
            })

    })

    describe('multiLine', function() {

        beforeEach(function() {
            dom = compile(
                '<div ng-form="testForm">' +
                '<textarea ' +
                'multi-line ' +
                'ng-model="field" ' +
                '></textarea>' +
                '</div>'
                )
            input_field = dom.find('textarea')[0]
        })

        // An array of strings from the movde is converted to a string
        // to the view with items separated by newlines.
        test('model_to_view', () => {
            scope.$apply(() => {
                  scope.field = [
                      ['field1 fie"ld2'],
                      ['field3-more']
                    ]
            })

            expect(input_field.value).toEqual(
                'field1 fie"ld2\nfield3-more')
            expect(scope.testForm.$valid).toBeTrue()
        })

        // Lines from the view are converted to an array of strings.
        test('view_to_model', () => {
                // We need to put some value first, as is empty by default
                // and will not trigger the update.
                viewUpdate(input_field, 'field1 field2\nother-line')
                expect(scope.field).toEqual(
                    ['field1 field2', 'other-line'])
                // Then we delete the input.
                viewUpdate(input_field, '')

                expect(scope.field).toEqual([])
            })

    })


    describe('floatOrInteger', function() {

        var getInputField = function(options) {
            if (options === undefined) {
                options = {
                    'min': -10.5,
                    'max': 10.5
                }
            }
            var markup = '<div ng-form="testForm">' +
                '<input ' +
                'type="text" ' +
                'float-or-integer ' +
                'ng-model="float_or_integer" ' +
                'name="float_or_integer" ' +
                'min="%(min)s" ' +
                'max="%(max)s" ' +
                '>' +
                '</div>'
            dom = compile(markup.supplant(options))
            return dom.find('input')[0]
        }

        test(
        'Return undefined when not a float or integer',
        function() {
            input_field = getInputField()
            viewUpdate(input_field, 'abc')

            expect(scope.float_or_integer).toBeUndefined()
            expect(scope.testForm.$valid).toBeFalse()
            expect(scope.testForm.float_or_integer.$valid).toBeFalse()
        })

        test(
        'Return floating value when input is valid.',
        function() {
            input_field = getInputField()
            viewUpdate(input_field, '1.5')

            expect(scope.float_or_integer).toEqual(1.5)
            expect(scope.testForm.$valid).toBeTrue()
            expect(scope.testForm.float_or_integer.$valid).toBeTrue()
        })

        test(
        'Parse float number with comma and negative values.',
        function() {
            input_field = getInputField()
            viewUpdate(input_field, '-1,6')

            expect(scope.float_or_integer).toEqual(-1.6)
            expect(scope.testForm.$valid).toBeTrue()
        })

        test(
        'Return minimum value when input is less that minimum and set ' +
        'input as invalid.',
        function() {
            input_field = getInputField()
            viewUpdate(input_field, '-10,6')

            expect(scope.float_or_integer).toEqual(-10.5)
            expect(scope.testForm.$valid).toBeFalse()
            expect(scope.testForm.float_or_integer.$valid).toBeFalse()
        })

        test(
        'Return maximum value when input is more that maximum and set ' +
        'input as invalid.',
        function() {
            input_field = getInputField()
            viewUpdate(input_field, '10,6')

            expect(scope.float_or_integer).toEqual(10.5)
            expect(scope.testForm.$valid).toBeFalse()
            expect(scope.testForm.float_or_integer.$valid).toBeFalse()
        })

        test(
        'Minimum value can be 0 and is accepted as valid.',
        function() {
            input_field = getInputField({'min': 0, 'max': 10})
            viewUpdate(input_field, '0')

            expect(scope.float_or_integer).toEqual(0)
            expect(scope.testForm.$valid).toBeTrue()
            expect(scope.testForm.float_or_integer.$valid).toBeTrue()
        })

    })

    describe('logRotateOn', function() {

        /*
        Return an input field with `logRotateOn` directive.
        */
        var getInputField = function() {
            var html = '<div ng-form="testForm">' +
                '<input ' +
                'log-rotate-on ' +
                'type="text" ' +
                'name="calendar_value" ' +
                'ng-model="calendar_value" ' +
                '>' +
                '</div>'
            dom = compile(html)

            return dom.find('input')[0]
        }

        var setRotateOnType = function(type) {
            // The state of the logRotateOn directive depends on the state
            // of two scope variables, calendar_value and calendar_type.
            // Here we inject the calendar_type so we can test its behavior
            // with different types.
            inject(function($rootScope) {
                // Changes the value of calendar_value
                scope = $rootScope.$new();
                scope.calendar_value = [undefined, type]
            })
        }

        test(
        'Return 17 day of the month when input is 17 ' +
        'and type is day-of-month',
        function() {
            setRotateOnType('day-of-month')
            input_field = getInputField()
            viewUpdate(input_field, '17')
            scope.$digest()

            expect(scope.calendar_value).toEqual([17, 'day-of-month'])
        })

        test(
        'Accepts time expressed as single digit when type is time-of-day',
        function() {
            setRotateOnType('time-of-day')
            input_field = getInputField()
            viewUpdate(input_field, '2:3')

            expect(scope.calendar_value).toEqual([7380, 'time-of-day'])
        })

        test(
        'Accepts time expressed as double digits when type is time-of-day',
        function() {
            setRotateOnType('time-of-day')
            input_field = getInputField()
            viewUpdate(input_field, '02:30')

            expect(scope.calendar_value).toEqual([9000, 'time-of-day'])
        })


        test(
        'Validation fails when value is less than zero and ' +
        'type is day-of-month',
        function() {
            setRotateOnType('day-of-month')
            input_field = getInputField()
            viewUpdate(input_field, '-1')

            expect(scope.testForm.$valid).toBeFalse()
            expect(scope.testForm.calendar_value.$valid).toBeFalse()
        })

        test(
        'Validation fails when input is empty and type is day-of-month',
        function() {
            setRotateOnType('day-of-month')
            input_field = getInputField()
            viewUpdate(input_field, '')

            expect(scope.testForm.$valid).toBeFalse()
            expect(scope.testForm.calendar_value.$valid).toBeFalse()
        })

        test(
        'Validation fails when input is greater than 31 and ' +
        'type is day-of-month',
        function() {
            setRotateOnType('day-of-month')
            input_field = getInputField()
            viewUpdate(input_field, '33')

            expect(scope.testForm.$valid).toBeFalse()
            expect(scope.testForm.calendar_value.$valid).toBeFalse()
            expect(scope.calendar_value).toEqual(
                [0, 'day-of-month'])
        })

        test(
        'Validation fails when setting a non integer for calendar type ' +
        'day-of-month',
        function() {
            setRotateOnType('day-of-month')
            input_field = getInputField()
            viewUpdate(input_field, '02:30')

            expect(scope.testForm.$valid).toBeFalse()
            expect(scope.testForm.calendar_value.$valid).toBeFalse()
            expect(scope.calendar_value).toEqual(
                [0, 'day-of-month'])
        })

        test(
        'Validation fails when setting a value not fitting the ' +
        'time format (H:M) for calendar type time-of-day',
        function() {
            setRotateOnType('time-of-day')
            input_field = getInputField()
            viewUpdate(input_field, '15')

            expect(scope.testForm.$valid).toBeFalse()
            expect(scope.testForm.calendar_value.$valid).toBeFalse()
            expect(scope.calendar_value).toEqual(
                [0, 'time-of-day'])
        })

        test(
        'Validation fails when setting a value bigger than 23 for hour ' +
        'when calendar type is time-of-day',
        function() {
            setRotateOnType('time-of-day')
            input_field = getInputField()
            viewUpdate(input_field, '25:00')

            expect(scope.testForm.$valid).toBeFalse()
            expect(scope.testForm.calendar_value.$valid).toBeFalse()
            expect(scope.calendar_value).toEqual(
                [0, 'time-of-day'])
        })

        test(
        'Validation fails when setting a value bigger than 59 for minute ' +
        'when calendar type is time-of-day',
        function() {
            setRotateOnType('time-of-day')
            input_field = getInputField()
            viewUpdate(input_field, '13:60')

            expect(scope.testForm.$valid).toBeFalse()
            expect(scope.testForm.calendar_value.$valid).toBeFalse()
            expect(scope.calendar_value).toEqual(
                [0, 'time-of-day'])
        })

        test(
        'Validation fails when setting a negative value for hour ' +
        'when calendar type is time-of-day',
        function() {
            setRotateOnType('time-of-day')
            input_field = getInputField()
            viewUpdate(input_field, '-1:60')

            expect(scope.testForm.$valid).toBeFalse()
            expect(scope.testForm.calendar_value.$valid).toBeFalse()
            expect(scope.calendar_value).toEqual(
                [0, 'time-of-day'])
        })
    })

    describe('logRotateOnCalendarType', function() {

        /*
        Return an select field with `logRotateOnCalendarType` directive.
        */
        var getInputField = function() {
            var html = '<div ng-form="testForm">' +
                '<select ' +
                'log-rotate-on-calendar-type ' +
                'name="calendar_type" ' +
                'ng-model="calendar_type" ' +
                '>' +
                '<option value="day-of-month">day of the month</option>' +
                '<option value="time-of-day">time of the month</option>' +
                '</select>' +
                '</div>'
            dom = compile(html)

            // Initialize the scope.
            scope.calendar_type = []
            scope.$apply()

            return dom.find('select')[0]
        }

        test(
        'Return zero day of the month when the first item is selected',
        function() {
            input_field = getInputField()
            viewUpdate(input_field, 'day-of-month', 'change')

            expect(input_field.value).toEqual('day-of-month')
            expect(scope.calendar_type).toEqual([0, 'day-of-month'])
        })

        test(
        'Return zero second of the day when the second item is selected',
        function() {
            input_field = getInputField()
            viewUpdate(input_field, 'time-of-day', 'change')

            expect(input_field.value).toEqual('time-of-day')
            expect(scope.calendar_type).toEqual([0, 'time-of-day'])
        })

        test(
        'Return 0 day of the month when no item is selected',
        function() {
            input_field = getInputField()
            viewUpdate(input_field, 'day-of-month', 'change')

            expect(scope.calendar_type).toEqual([0, 'day-of-month'])
        })

    })

    describe('dropdown', function() {

        require('flowbite/dist/flowbite.js')
        /*
        Return an dropdown directive with a very simple structur.
        */
        var getDropdown = function(attributes) {
            var html = (
                '<div><dropdown ' + attributes + '>' +
                '<ul> ' +
                '<li>Option 1</li>' +
                '<li>Option 2</li>' +
                '</ul>' +
                '</dropdown></div>'
            )
            return compile(html)
        }

        /*
        Initially hides the content and makes it visible on button click.

        The class is maintained in the top-level element.
        */
        test('show_on_click', function() {
            var dropdown = getDropdown(
                'button-name="\'Some Name\'" data-test="1 2 a" ' +
                'class="button button--primary"'
                )
            var top_wrapper = dropdown.find('dropdown')
            var button = dropdown.find('button')
            var content = dropdown.find('div')

            // CSS is kept to high level.
            expect(top_wrapper).toHaveClass('button button--primary')

            // Data test is moved to button.
            expect(dropdown).not.toHaveAttribute('data-test')
            expect(button).toHaveAttribute('data-test', '1 2 a')

            // A button is created and content is hidden,
            expect(content).toHaveClass('dropdown-content')
            expect(content).not.toBeVisible()
            expect(button.text()).toEqual('Some Name')

            viewUpdate(button, '', 'click')
            expect(content).toBeVisible()
        })

        /*
        It injects the dropdown object into the scope based on the
        data-scope-name attribute value.

        Can also be created with text content for the button and with a
        custom placement.
        */
        test('scope_name_custom_placement', function() {
            scope.drop_down = null
            scope.drop_name = 'Some text'
            var dropdown = getDropdown(
                'button-name="drop_name" ' +
                'data-placement="bottom" ' +
                'own-reference="drop_down"'
                )
            var button = dropdown.find('button')

            expect(scope.drop_down).toBeDefined()
            expect(scope.drop_down._options.placement).toEqual('bottom')
            expect(button.text()).toEqual('Some text')

            // Once the model is updated the view is also updated.
            scope.drop_name = 'New text'
            scope.$digest()
            expect(button.text()).toEqual('New text')
        })

        /*
        On IE11 the placement is ignored.
        */
        test('ie_ignore_placement', function() {
            scope.drop_down = null
            utils.ie_version.IsIE = true
            var dropdown = getDropdown(
                'data-placement="bottom" own-reference="drop_down"')
            scope.$digest()

            expect(scope.drop_down).toBeDefined()
            expect(scope.drop_down._options.placement).toEqual('left-start')
        })


        /*
        It can be created with a custom button and any tag cn serve as
        the content wrapper
        */
        test('dropdown_custom_button', function() {
            var html = (
                '<dropdown>' +
                '<dropdown-button><i>Can be anything</i></dropdown-button>' +
                '<ul data-test="this can be any other tag"> ' +
                '<li>Option 1</li>' +
                '<li>Option 2</li>' +
                '</ul>' +
                '</dropdown>'
            )
            var dropdown = compile(html)

            var button = dropdown.find('dropdown-button')
            var content = dropdown.find('div')

            expect(content).not.toBeVisible()
            expect(button.text()).toEqual('Can be anything')

            viewUpdate(button, '', 'click')
            expect(content).toBeVisible()
            expect(content[0].childNodes[0].nodeName).toEqual('UL')
        })

    })


    /*
    Tagify tests should focus on the scope/model updated.

    We have some checks for the tagify DOM, but they should only be init checks.
    */
    describe('tagify', function() {

        global.Tagify = require('@yaireo/tagify/dist/tagify.min')
        /*
        Return a tagify directive.
        */
        var getTagify = function(options) {

            var model = options.model
            if (!model) {
                model = []
            }

            var whitelist = options.whitelist
            if (!whitelist) {
                whitelist = []
            }

            var attributes = options.attributes
            if (!attributes) {
                attributes = ''
            }

            var html = (
                '<div>' +
                '<input ' +
                'tagify ' +
                'tagify-placehoder="Add tag" ' +
                attributes + ' ' +
                'ng-model="tagify_model" ' +
                '/></div>'
            )

            // Initialize the scope.
            scope.tagify_model = model
            scope.tagify_whitelist = whitelist
            scope.configuration = {}

            var dom = compile(html)

            // This has 2 childers
            // * tags - tagify behind the scene
            // * input - the original input for which the value is updated.
            return {
                'tags': dom.find('tags')[0],
                'input': dom.find('input')[0],
            }
        }

        /*
        It can be initilized with an empty list.
        */
        test('tagify_init_empty', function() {
            var tagify = getTagify({})

            expect(tagify.input.value).toEqual('')
        })

        /*
        It can be initilized with a custom placeholder.
        */
        test('tagify_init_placeholder', function() {
            var tagify = getTagify({
                model:[],
                attributes: 'tagify-placeholder="Some placeholder text"'},
                )

            // There are no tag childer, so the last one is the placeholder.
            expect(
                tagify.tags.children[0].getAttribute('data-placeholder')
                ).toEqual('Some placeholder text')
        })

        /*
        When user input is is enabled, any value from the list is allowed,
        event if we don't have an allowed-list.
        */
        test('tagify_init_user_input', function() {
            var tagify = getTagify({
                model: ['initial', 'options'],
                attributes: 'tagify-user-input'},
                )

            // Tags are populated with initial tags with same value and uui.
            expect(
                tagify.tags.children[0].getAttribute('value')
                ).toEqual('initial')
            expect(
                tagify.tags.children[0].getAttribute('uuid')
                ).toEqual('initial')
            expect(
                tagify.tags.children[1].getAttribute('value')
                ).toEqual('options')
            expect(
                tagify.tags.children[1].getAttribute('uuid')
                ).toEqual('options')
            // The last one is the data placeholder with a default text.
            expect(
                tagify.tags.children[2].getAttribute('data-placeholder')
                ).toEqual('Click to select from list')
        })

        /*
        It can be initialized with arbitrary uuid+value when user-input
        is enabled and when no whiltelist is defined.
        */
        test('tagify_init_uuid', function() {
            var tagify = getTagify({
                // Model only has list of uuids.
                model: ['uuid1', 'uuid3'],
                // Whitelist has uuid and name, and name is visible in the UI.
                whitelist: [
                    {'uuid': 'uuid1', 'name': 'initial'},
                    {'uuid': 'uuid2', 'name': 'other'},
                    {'uuid': 'uuid3', 'name': 'options'},
                ],
                attributes: 'tagify-whitelist=tagify_whitelist'
            })

            // Tags are populated with initial tags with same value and uui.
            expect(
                tagify.tags.children[0].getAttribute('value')
                ).toEqual('initial')
            expect(
                tagify.tags.children[0].getAttribute('uuid')
                ).toEqual('uuid1')
            expect(
                tagify.tags.children[1].getAttribute('value')
                ).toEqual('options')
            expect(
                tagify.tags.children[1].getAttribute('uuid')
                ).toEqual('uuid3')
        })

        /*
        It will ignore uuids found in the model but not in the whitelist,
        when user-input is not enabled.
        */
        test('tagify_init_unknown_uuid', function() {
            var tagify = getTagify({
                // Model only has list of uuids.
                model: ['uuid1', 'other-uuid', 'uuid2'],
                // Whitelist has uuid and name, and name is visible in the UI.
                whitelist: [
                    {'uuid': 'uuid1', 'name': 'initial'},
                    {'uuid': 'uuid2', 'name': 'other'},
                ],
                attributes: 'tagify-whitelist=tagify_whitelist'
            })

            // Tags are populated with initial tags with same value and uuid.
            expect(
                tagify.tags.children[0].getAttribute('value')
                ).toEqual('initial')
            expect(
                tagify.tags.children[0].getAttribute('uuid')
                ).toEqual('uuid1')
            expect(
                tagify.tags.children[1].getAttribute('value')
                ).toEqual('other')
            expect(
                tagify.tags.children[1].getAttribute('uuid')
                ).toEqual('uuid2')
        })

        /*
        It will ignore updated values without uuid when user-input
        is not enabled.
        */
        test('tagify_update_unknown_uuid', function() {
            var tagify = getTagify({
                // Model only has list of uuids.
                model: ['uuid1'],
                // Whitelist has uuid and name, and name is visible in the UI.
                whitelist: [
                    {'uuid': 'uuid1', 'name': 'initial'},
                    {'uuid': 'uuid2', 'name': 'new option'},
                ],
                attributes: 'tagify-whitelist=tagify_whitelist'
            })

            // Just check state before the change.
            expect(scope.tagify_model).toEqual(['uuid1'])

            // Tagify will send updates from the view as full JSON list.
            // In this example the previous tag is removed and we have
            // one uknown tag and one new one.
            viewUpdate(
                tagify.input,
                '[{"value": "unknown"}, {"uuid": "uuid2"}]',
                'change',
                )

            // Previous value is removed and new one is added.
            // the uknown is ignored.
            expect(scope.tagify_model).toEqual(['uuid2'])
        })

        /*
        It can handle a whitelist as simple array.
        */
        test('tagify_simple_whitelist', function() {
            var tagify = getTagify({
                // Model only has list of uuids.
                model: ['value 2'],
                // Whitelist is just a list of raw values.
                whitelist: ['value-1', 'value 2', 'value_3'],
                attributes: 'tagify-whitelist=tagify_whitelist'
            })

            // Just check state before the change.
            expect(scope.tagify_model).toEqual(['value 2'])

            // Tagify will send updates from the view as full JSON list.
            // In this example the previous tag is removed and we have
            // one uknown tag and one new one.
            viewUpdate(
                tagify.input,
                '[{"uuid": "value-1"}]',
                'change',
                )

            // Previous value is removed and new one is added.
            // the uknown is ignored.
            expect(scope.tagify_model).toEqual(['value-1'])
        })


        /*
        When user input is enabled, any value is allowed to be updated.
        */
        test('tagify_update_user_input', function() {
            var tagify = getTagify({
                model: ['initial'],
                attributes: 'tagify-user-input'},
                )

            // Just check state before the change.
            expect(scope.tagify_model).toEqual(['initial'])

            viewUpdate(
                tagify.input,
                '[{"value": "new value"}, {"value": "initial"}]',
                'change',
                )

            // Previous value is removed and new one is added.
            // the uknown is ignored.
            expect(scope.tagify_model).toEqual(['new value', 'initial'])
        })


    })

    /*
    Compile the config seconds directive and return the main components.
    */
    var compileConfigSeconds = (extra_attributes) => {
        if (!extra_attributes) {
            extra_attributes = ''
        }
        var html = (
            '<config-seconds data-test="some-test" ' + extra_attributes +
            ' ng-model="total_seconds" '+
            'data-label="Value for this Label"> '+
            '<help-block>Some help text</help-block> '+
            '</config-seconds>'
        )
        var dom_directive = compileForm(html)
        return {
            'input': dom_directive.find('input'),
            'select': dom_directive.find('select'),
        }
    }

    /*
    When model is zero it is initialized as disabled.

    It used the attributes from the top tag to configure the config and the
    inner tab as help.

    When there is no min attribute selected, it will default to 0.

    Any extra attributes are left to the top level

    The help block is inserted as HTML.
    */
    test('config_seconds_init_disabled', function() {

        var html = (
            '<config-seconds data-test="some-test"' +
            'ng-model="config.total_seconds" '+
            'data-label="Value for this Label"> '+
            '<help-block>Some <em>text</em></help-block> '+
            '</config-seconds>'
        )
        scope.config = { 'total_seconds': 0}
        var dom_directive = compileForm(html)

        var wrapper = dom_directive.find('config-seconds')
        var config_label = dom_directive.find('label')
        var input_seconds = dom_directive.find('input')
        var select_seconds = dom_directive.find('select')
        var restart_required = dom_directive.find('restart-required')

        var help_block_wrapper = dom_directive.find('help-block-wrapper')
        expect(help_block_wrapper.html()).toEqual(
            ' <help-block class="ng-scope">Some <em>text</em></help-block> ')

        // The input ID is selected based on model and link to the label.
        expect(config_label.text()).toEqual('Value for this Label')
        expect(config_label.attr('for')).toEqual('config.total_seconds')
        expect(input_seconds.attr('id')).toEqual('config.total_seconds')

        // The top level attributes are kept.
        expect(wrapper.attr('data-test')).toEqual('some-test')
        // It is augmented to a numeric input with default min value.
        expect(input_seconds.attr('type')).toEqual('number')
        expect(input_seconds.attr('min')).toEqual('0')

        // Restart required is hiddne.
        expect(restart_required.hasClass('ng-hide')).toBeTrue()

        expect(input_seconds.val()).toEqual("0")
        expect(scope.config.total_seconds).toEqual(0)
        expect(select_seconds.val()).toEqual('disabled')
        // The disable selection is visible.
        expect(
            angular.element(select_seconds.children()[0]).hasClass('ng-hide')
        ).toBeFalse()
    })

    /*
    When min attribute is defined, it will be set to the input and disable
    option is hidden.

    The `restart-required` will make the restart required icon visible.

    When model already has a value, and the value is not a multiple
    of minutes / hours / days, it is initialized as seconds.

    The model can be initialized from a sub-object of the main scope,
    it doesn't have to be on the main scope.
    */
    test('config_seconds_init_seconds', function() {

        var html = (
            '<config-seconds data-min="42" data-restart-required ' +
            'ng-model="config.total_seconds" '+
            'data-label="TEST FOR THE LABEL"> '+
            '<help-block>Some <span> text.</help-block> '+
            '</config-seconds>'
        )
        scope.config = { 'total_seconds': 79}
        var dom_directive = compileForm(html)

        var restart_required = dom_directive.find('restart-required')
        var config_label = dom_directive.find('label')
        var input_seconds = dom_directive.find('input')
        var select_seconds = dom_directive.find('select')

        // The input ID is selected based on model and link to the label.
        expect(config_label.attr('for')).toEqual('config.total_seconds')
        expect(input_seconds.attr('id')).toEqual('config.total_seconds')

        // It is augmented to a numeric input with custom min value.
        expect(input_seconds.attr('type')).toEqual('number')
        expect(input_seconds.attr('min')).toEqual('42')

        // Restart required is visible.
        expect(restart_required.hasClass('ng-hide')).toBeFalse()

        expect(input_seconds.val()).toEqual("79")
        expect(scope.config.total_seconds).toEqual(79)
        expect(select_seconds.val()).toEqual('seconds')
        // The disable selection is hidden.
        expect(
            angular.element(select_seconds.children()[0]).hasClass('ng-hide')
        ).toBeTrue()

        // Updating  the view from seconds to minutes will update the model.
        viewUpdate(select_seconds, 'minutes', 'change')
        expect(scope.config.total_seconds).toEqual(79 * 60)
    })

    /*
    When model value is a negative number, it is started as disabled.

    Later if updated to a negative value, it is set again to disabled.
    */
    test('config_seconds_init_negative_no_min', function() {
        scope.total_seconds = -2
        var sut = compileConfigSeconds()

        expect(sut.input.val()).toEqual("0")
        expect(scope.total_seconds).toEqual(0)
        expect(sut.select.val()).toEqual('disabled')

        viewUpdate(sut.input, '2')
        expect(sut.input.val()).toEqual("2")
        expect(scope.total_seconds).toEqual(2)
        expect(sut.select.val()).toEqual('seconds')

        viewUpdate(sut.input, '-4')
        expect(sut.input.val()).toEqual("0")
        expect(scope.total_seconds).toEqual(0)
        expect(sut.select.val()).toEqual('disabled')
    })

    /*
    When model value is a negative number and a custom minimum is defined,
    it is set to that minimum value.
    */
    test('config_seconds_init_negative_with_min', function() {
        scope.total_seconds = -2
        var sut = compileConfigSeconds('data-min="120"')

        expect(scope.total_seconds).toEqual(120)
        expect(sut.select.val()).toEqual('minutes')
        expect(sut.input.val()).toEqual("2")
    })


    /*
    When view is set to a negative value, it will fallback to minium value
    and the view is updated.
    */
    test('config_seconds_view_negative_with_min', function() {
        scope.total_seconds = 180
        var sut = compileConfigSeconds('data-min="42"')

        expect(scope.total_seconds).toEqual(180)
        expect(sut.select.val()).toEqual('minutes')
        expect(sut.input.val()).toEqual("3")

        viewUpdate(sut.input, '-4')

        expect(scope.total_seconds).toEqual(42)
        expect(sut.select.val()).toEqual('seconds')
        expect(sut.input.val()).toEqual("42")
    })

    /*
    When model value is a multiple of minutes it is initialized as minutes.
    */
    test('config_seconds_init_minutes', function() {
        scope.total_seconds = 180
        var sut = compileConfigSeconds()

        // This is the seconds selector.
        expect(sut.select.val()).toEqual('minutes')
        expect(sut.input.val()).toEqual("3")
        expect(scope.total_seconds).toEqual(180)
    })

    /*
    When model value is a multiple of hours it is initialized as hours.
    */
    test('config_seconds_init_hours', function() {
        scope.total_seconds = 3600 * 4
        var sut = compileConfigSeconds()

        expect(sut.select.val()).toEqual('hours')
        expect(sut.input.val()).toEqual("4")
        expect(scope.total_seconds).toEqual(3600 * 4)
    })

    /*
    When model value is a multiple of days it is initialized as days.
    */
    test('config_seconds_init_days', function() {
        scope.total_seconds = 3600 * 24
        var sut = compileConfigSeconds()

        expect(sut.select.val()).toEqual('days')
        expect(sut.input.val()).toEqual("1")
        expect(scope.total_seconds).toEqual(3600 * 24)
    })

    /*
    When model is zero it is started in disabled mode, but once updated,
    it will move to seconds or the closest measurement unit.
    */
    test('config_seconds_enable_from_model', function() {
        scope.total_seconds = 0
        var sut = compileConfigSeconds()

        // Started as disabled
        expect(sut.select.val()).toEqual('disabled')
        expect(sut.input.val()).toEqual("0")
        expect(scope.total_seconds).toEqual(0)

        // Update to seconds
        scope.total_seconds = 4
        scope.$digest()

        expect(sut.select.val()).toEqual('seconds')
        expect(sut.input.val()).toEqual("4")
        expect(scope.total_seconds).toEqual(4)
    })

    /*
    When model is undefined, or later updated to undefined,
    the view is left as it is.
    */
    test('config_seconds_enable_from_model', function() {
        scope.total_seconds = undefined
        var sut = compileConfigSeconds()

        // Started as disabled
        expect(sut.select.val()).toEqual('disabled')
        expect(sut.input.val()).toEqual("0")
        expect(scope.total_seconds).toEqual(0)

        // Update to seconds
        scope.total_seconds = 3600
        scope.$digest()

        expect(sut.select.val()).toEqual('hours')
        expect(sut.input.val()).toEqual("1")
        expect(scope.total_seconds).toEqual(3600)

        // Update to seconds
        scope.total_seconds = undefined
        scope.$digest()

        expect(sut.select.val()).toEqual('hours')
        expect(sut.input.val()).toEqual("1")
        expect(scope.total_seconds).toEqual(undefined)
    })

    /*
    When model is zero it is started in disabled mode, but once updated,
    it will move to seconds or the closest measurement unit.
    */
    test('config_seconds_enable_from_view', function() {
        scope.total_seconds = 0
        var sut = compileConfigSeconds()

        // Started as disabled
        expect(sut.select.val()).toEqual('disabled')
        expect(sut.input.val()).toEqual("0")
        expect(scope.total_seconds).toEqual(0)

        viewUpdate(sut.input, '123')

        // Updated to seconds.
        expect(sut.select.val()).toEqual('seconds')
        expect(sut.input.val()).toEqual("123")
        expect(scope.total_seconds).toEqual(123)
    })

    /*
    When model is zero it is started in disabled mode, but once a unit is
    selected it is updated to 10 times that unit.
    */
    test('config_seconds_enable_from_unit', function() {
        scope.total_seconds = 0
        var sut = compileConfigSeconds()

        // Started as disabled
        expect(sut.select.val()).toEqual('disabled')
        expect(sut.input.val()).toEqual("0")
        expect(scope.total_seconds).toEqual(0)

        viewUpdate(sut.select, 'hours', 'change')

        // Updated to seconds.
        expect(sut.select.val()).toEqual('hours')
        expect(sut.input.val()).toEqual("10")
        expect(scope.total_seconds).toEqual(36000)
    })

    /*
    When unit is changed to disabled, value is set to 0.
    */
    test('config_seconds_disable_from_unit', function() {
        scope.total_seconds = 123
        var sut = compileConfigSeconds()

        expect(sut.select.val()).toEqual('seconds')
        expect(sut.input.val()).toEqual("123")
        expect(scope.total_seconds).toEqual(123)

        viewUpdate(sut.select, 'disabled', 'change')

        // Updated to seconds.
        expect(sut.select.val()).toEqual('disabled')
        expect(sut.input.val()).toEqual("0")
        expect(scope.total_seconds).toEqual(0)
    })

    /*
    When unit is changed to another type, the model value is updated,
    but the view is kept with the old value.
    */
    test('config_seconds_update_from_unit', function() {
        scope.total_seconds = 4
        var sut = compileConfigSeconds()

        expect(sut.select.val()).toEqual('seconds')
        expect(sut.input.val()).toEqual("4")
        expect(scope.total_seconds).toEqual(4)

        viewUpdate(sut.select, 'minutes', 'change')

        // Updated to seconds.
        expect(sut.select.val()).toEqual('minutes')
        expect(sut.input.val()).toEqual("4")
        expect(scope.total_seconds).toEqual(240)
    })

    /*
    When input is changed, the unit is kept and model is updated acording
    to the unit.

    Setting the view to a float will conver the value to an integer.

    */
    test('config_seconds_update_from_input', function() {
        scope.total_seconds = 3600 * 24 * 3
        var sut = compileConfigSeconds()

        expect(sut.select.val()).toEqual('days')
        expect(sut.input.val()).toEqual("3")
        expect(scope.total_seconds).toEqual(3600 * 24 * 3)

        viewUpdate(sut.input, '1.9')

        // Updated to seconds.
        expect(sut.select.val()).toEqual('days')
        expect(sut.input.val()).toEqual("1")
        expect(scope.total_seconds).toEqual(3600 * 24)
    })

    /*
    When input is changed, it will auto-convert to closes unit.

    When model is a float, it is initialized as float.
    */
    test('config_seconds_update_from_input', function() {
        scope.total_seconds = 123.4
        var sut = compileConfigSeconds()

        expect(sut.select.val()).toEqual('seconds')
        expect(sut.input.val()).toEqual("123.4")
        expect(scope.total_seconds).toEqual(123.4)

        viewUpdate(sut.input, '120')

        expect(sut.select.val()).toEqual('minutes')
        expect(sut.input.val()).toEqual("2")
        expect(scope.total_seconds).toEqual(120)
    })

})
