/*
Tests for account edit page controller.
*/


describe('accountCtrl', function() {


    /*
    Set the configuration to already have an account.

    Returns the account.
    */
    var setExistingAccount = function(sut) {
        var account = {
            'uuid': 'some-uuid',
            'name': 'existing-user',
        }
        sut.scope.$parent.configuration.accounts = {'some-uuid': account}
        sut.$location.path('/accounts/some-uuid')
        sut.route.uuid = 'some-uuid'
        return account
    }

    /*
    __init__ redirect to accounts index page if no users exist.
    */
    test('init_no_accounts', function() {
        var sut = mk.getController('accountCtrl')

        sut.scope.$parent.configuration.accounts = {}
        sut.scope.__init__()

        expect(sut.$location.path()).toBe('/accounts')
    })

    /*
    __init__ redirect to users page if some users are defined but requested
    user does not exist.
    */
    test('init_account_does_not_exist', function() {
        var sut = mk.getController('accountCtrl')

        sut.scope.$parent.configuration.accounts = {
            'some-old-uuid': {'name': 'account in config'}}
        sut.route.uuid = 'not-exist-uuid'

        sut.scope.__init__()

        expect(sut.$location.path()).toBe('/accounts')
    })

    /*
    It load the page of an account for which the UUID is extracted from the
    URL.
    */
    test('init_account_existing', function() {
        var sut = mk.getController('accountCtrl')
        var account = setExistingAccount(sut)
        var groups = {
            'group1-uuid': {
                'uuid': 'group1-uuid',
                'home_folder_structure': ['one', 'two'],
            },
            'group2-uuid': {
                'uuid': 'group2-uuid',
                'home_folder_structure': ['three'],
            },
        }
        sut.scope.$parent.configuration.groups = groups
        account.group = ['group1-uuid', 'group2-uuid', 'uuid-ignored']
        // Account not yet set before init.
        expect(sut.scope.account).toEqual({})

        sut.scope.__init__()

        expect(sut.$location.path()).toBe('/accounts/some-uuid')
        expect(sut.scope.section_uuid).toBe('accounts/some-uuid')
        // View model updated after init.
        expect(sut.scope.account).toEqual(account)
        expect(sut.scope.groups_location).toEqual('groups')
        expect(sut.scope.last_password_update_description).toEqual(
            'Date and time when password was last updated is unknown.')
        expect(sut.scope.show_disable_on_inactivity).toBeTrue()
        expect(sut.scope.home_folder_structure).toEqual([
            'one', 'two', 'three'])
    })

    /*
    It load the page of an account from an external authentication file.
    */
    test('init_account_existing_local_file', function() {
        var sut = mk.getController('accountCtrl')

        var account = {
            'uuid': 'some-uuid',
            'name': 'existing-user',
            // Set to a very old/initial unix time.
            'last_password_update': 12234,
            'group': ['group1-uuid']
        }
        var groups = {
            'group1-uuid': {'uuid': 'group1-uuid'},
        }
        // No accounts are defined in the main config
        sut.scope.$parent.configuration.accounts = {}
        sut.scope.$parent.configuration.groups = {}
        sut.scope.$parent.configuration.authentications = {
            'file-uuid': {
                'uuid': 'file-uuid',
                'type': 'local-file',
                'accounts': {'some-uuid': account},
                'groups': groups
            }
        }
        sut.$location.path('/accounts/some-uuid')
        sut.route.uuid = 'some-uuid'

        // Account not yet set before init.
        expect(sut.scope.account).toEqual({})

        sut.scope.__init__()

        expect(sut.$location.path()).toBe('/accounts/some-uuid')
        expect(sut.scope.section_uuid).toBe('accounts/some-uuid')
        // View model updated after init.
        expect(sut.scope.account).toEqual(account)
        expect(sut.scope.groups_location).toEqual(
            'authentications/file-uuid/groups')
        // Just check that we got the year right.
        expect(sut.scope.last_password_update_description).toStartWith(
            'Password was last updated on 1970-01')
        expect(sut.scope.show_disable_on_inactivity).toBeFalse()
    })

    /*
    It updated the UI model based on the account configuration.
    */
    test('init_account_expirate_datetime_ui', function() {
        var sut = mk.getController('accountCtrl')
        var account = setExistingAccount(sut)
        account['expire_datetime'] = '2022-11-16T03:07:29.000Z'

        sut.scope.__init__()

        expect(sut.scope.expire_datetime_ui.date_value).toEqual(
            new Date('2022-11-16T03:07:29.000Z'))
        expect(sut.scope.expire_datetime_ui.time_value).toEqual(
            new Date('2022-11-16T03:07:29.000Z'))
    })

    /*
    It set the expire_datetime and removes the expire_datetime configuration
    based on UI changes.
    */
    test('expire_datetime_date_value_watch', function() {
        var sut = mk.getController('accountCtrl')
        setExistingAccount(sut)
        sut.scope.__init__()
        sut.scope.$digest()

        // Update the configuration based on UI model.
        sut.scope.expire_datetime_ui.date_value = new Date(
            '2022-11-16T03:07:29.000Z')
        sut.scope.$digest()
        expect(sut.scope.account.expire_datetime).toEqual(
            '2022-11-16T03:07:29.000Z')

        // When UI model is reset, the configuration is also reset.
        sut.scope.expire_datetime_ui.date_value = null
        sut.scope.$digest()
        expect(sut.scope.account.expire_datetime).toEqual('')
    })

    /*
    The configuration is not updated when time value is updated without
    having a date value set first.
    */
    test('expire_datetime_time_value_no_date', function() {
        var sut = mk.getController('accountCtrl')
        setExistingAccount(sut)
        sut.scope.__init__()
        sut.scope.$digest()

        // Update the configuration based on UI model.
        sut.scope.expire_datetime_ui.time_value = new Date(
            '2022-11-16T03:07:29.000Z')
        sut.scope.$digest()
        // Not changed.
        expect(sut.scope.account.expire_datetime).toEqual('')

        // Update again only the time.
        sut.scope.expire_datetime_ui.time_value = new Date(
            '2022-01-16T03:07:29.000Z')
        sut.scope.$digest()
        // Not changed.
        expect(sut.scope.account.expire_datetime).toEqual('')

        // When UI model is reset, the configuration is also reset.
        sut.scope.expire_datetime_ui.time_value = null
        sut.scope.$digest()
        expect(sut.scope.account.expire_datetime).toEqual('')
    })

    /*
    The configuration is updated when time value is updated and we have
    an already defined date value.
    */
    test('expire_datetime_time_value_watch', function() {
        var sut = mk.getController('accountCtrl')
        setExistingAccount(sut)
        sut.scope.__init__()
        sut.scope.$digest()

        // The time from the date is ignored.
        sut.scope.expire_datetime_ui.date_value = new Date(
            '2022-11-16T13:17:19.000Z')
        // The date from time is ignored.
        sut.scope.expire_datetime_ui.time_value = new Date(
            '2002-01-16T03:07:29.000Z')
        sut.scope.$digest()
        // Date from date, time from time :)
        expect(sut.scope.account.expire_datetime).toEqual(
            '2022-11-16T03:07:29.000Z')

        // Update again only the time.
        sut.scope.expire_datetime_ui.time_value = new Date(
            '2002-01-16T13:17:19.000Z')
        sut.scope.$digest()
        // Time is updated and date is not updated from the time value.
        expect(sut.scope.account.expire_datetime).toEqual(
            '2022-11-16T13:17:19.000Z')

        // When UI model is reset, the date is set at the start of the day.
        sut.scope.expire_datetime_ui.time_value = null
        sut.scope.$digest()
        expect(sut.scope.account.expire_datetime).toEqual(
            '2022-11-16T00:00:00.000Z')
    })

    /*
    When on the page for a new account that is not yet saved in the config,
    it will remove it from the pending changes.
    */
    test('onDeleteAccount_new', function() {
        var sut = mk.getController('accountCtrl')

        sut.scope.$parent.configuration.accounts = {
            'some-old-uuid': {'name': 'account in config'}}
        sut.scope.$parent.new_entities.accounts = [{
            'name': 'some-new-account'
        }]
        sut.route.uuid = 'some-new-account'
        sut.scope.__init__()

        sut.scope.onDeleteAccount()

        // The new account is removed.
        expect(sut.scope.$parent.new_entities.accounts).toHaveLength(0)
        // Existing account is left.
        expect(sut.scope.$parent.configuration.accounts).toContainKey(
            'some-old-uuid')
    })

    /*
    For an account that is found in the config, it will remove it from the
    configuration object.
    */
    test('onDeleteAccount_existing', function() {
        var sut = mk.getController('accountCtrl')

        sut.scope.$parent.configuration.accounts = {
            'some-old-uuid': {'name': 'old account', 'uuid': 'some-old-uuid'}}
        sut.scope.$parent.new_entities.accounts = [{
            'name': 'some-new-account'
        }]
        sut.route.uuid = 'some-old-uuid'
        sut.scope.__init__()

        sut.scope.onDeleteAccount()

        // The new account is not removed.
        expect(sut.scope.$parent.new_entities.accounts).toHaveLength(1)
        // Existing account is removed.
        expect(sut.scope.$parent.configuration.accounts).not.toContainKey(
            'some-old-uuid')
    })

    /*
    For an external account that is found in the config, it will remove it from
    the configuration object.
    */
    test('onDeleteAccount_existing_and_external', function() {
        var sut = mk.getController('accountCtrl')
        // Set the default account.
        sut.scope.$parent.configuration.accounts = {
            'default-account': {
                'name': 'default account',
                'uuid': 'default-account'
            }
        }

        sut.scope.$parent.configuration.authentications = {
            'local-file-auth-uuid': {
                'name': 'local-file-auth',
                'type': 'local-file',
                'uuid' :'local-file-auth-uuid',
                'enabled': true,
                'accounts': {
                    'ext-account-uuid': {
                        'name': 'old account',
                        'uuid': 'ext-account-uuid'
                    }
                }

            }
        }

        sut.scope.$parent.new_entities.accounts = [{'name': 'some-new-account'}]
        sut.route.uuid = 'ext-account-uuid'
        sut.scope.__init__()

        sut.scope.onDeleteAccount()

        // External account is removed.
        expect(
            sut.scope.$parent.configuration
                .authentications['local-file-auth-uuid'].accounts
            )
            .not.toContainKey('ext-account-uuid')
        // The new account is left.
        expect(sut.scope.$parent.new_entities.accounts).toHaveLength(1)
    })

    /*
    On changes into `os` type, it  will delete the
    password_update property.
    */
    test('account_switch_to_os', function() {
        var sut = mk.getController('accountCtrl')
        sut.scope.account['type'] = 'application'
        sut.scope.account['description'] = ''
        sut.scope.account['name'] = ''
        sut.scope.account['password_update'] = ''
        sut.scope.$digest()
        expect(sut.scope.account).toContainKey('password_update')
        expect(sut.scope.application_account).toBeTrue()

        sut.scope.account['type'] = 'os'
        sut.scope.$digest()

        expect(sut.scope.account).not.toContainKey('password_update')
        expect(sut.scope.application_account).toBeFalse()
    })


    /*
    On changes into `application` type, it will keep the
    password_update property.
    */
    test('account_switch_to_app', function() {
        var sut = mk.getController('accountCtrl')
        sut.scope.account['type'] = 'os'
        sut.scope.account['description'] = ''
        sut.scope.account['name'] = ''
        sut.scope.$digest()
        expect(sut.scope.application_account).toBeFalse()

        sut.scope.account['type'] = 'application'
        sut.scope.$digest()

        expect(sut.scope.application_account).toBeTrue()
    })

    /*
    Return a list with available groups..
    */
    test('getAllGroups_no_groups', function() {
        var group_1 = {
            'name': 'Group 1',
            'uuid': 'uuid1',
        }
        var sut = mk.getController('accountCtrl')
        setExistingAccount(sut)
        sut.scope.$parent.configuration.groups = {'uuid1': group_1}
        sut.scope.__init__()

        var result = sut.scope.getAllGroups()

        expect(result).toEqual([group_1])
    })

    /*
    Will automatically update the primary group when the group is changed.
    */
    test('watch_accounts.group', function() {
        var group_1 = {
            'name': 'Group 1',
            'uuid': 'uuid1',
            'disable_on_inactivity': 24,
            'file_write_size': 42,
            'password_lifetime': 0,
        }
        var group_2 = {
            'name': 'Group 2',
            'uuid': 'uuid2',
            'disable_on_inactivity': 0,
            'file_write_size': 0,
            'password_lifetime': 34,
        }
        var sut = mk.getController('accountCtrl')
        var account = setExistingAccount(sut)
        sut.scope.$parent.configuration.groups = {
            'uuid1': group_1,
            'uuid2': group_2,
        }
        account.disable_on_inactivity = 'inherit'
        account.password_lifetime = 'inherit'
        account.file_write_size = 'inherit'
        account.group = ['uuid2', 'uuid1']

        // Just some pre-checks.
        expect(sut.scope.group).toBe('')
        expect(sut.scope.disable_on_inactivity_group).toBe('')
        expect(sut.scope.password_lifetime_group).toBe('')

        sut.scope.__init__()
        sut.scope.$digest()

        expect(sut.scope.group).toBe(group_2)
        expect(sut.scope.disable_on_inactivity_group).toBe('unlimited')
        expect(sut.scope.file_write_size_group).toBe('unlimited')
        expect(sut.scope.password_lifetime_group).toBe(34)

        // When the group is updated,
        // the primary group and inherited values are also updated.
        account.group = ['uuid1', 'uuid2']
        sut.scope.$digest()

        expect(sut.scope.group).toBe(group_1)
        expect(sut.scope.disable_on_inactivity_group).toBe(24)
        expect(sut.scope.file_write_size_group).toBe(42)
        expect(sut.scope.password_lifetime_group).toBe('unlimited')

    })

    /*
    When account has no inherited value, its group-inherited computed
    values are empty.
    */
    test('watch_accounts.group_no_inherited', function() {
        var group_1 = {
            'name': 'Group 1',
            'uuid': 'uuid1',
            'disable_on_inactivity': 24,
            'file_write_size': 25,
            'password_lifetime': 23,
        }
        var sut = mk.getController('accountCtrl')
        var account = setExistingAccount(sut)
        sut.scope.$parent.configuration.groups = {
            'uuid1': group_1,
        }
        account.disable_on_inactivity = 42
        account.password_lifetime = 0
        account.group = ['uuid1']

        sut.scope.__init__()
        sut.scope.$digest()

        expect(sut.scope.group).toBe(group_1)
        expect(sut.scope.disable_on_inactivity_group).toBe('')
        expect(sut.scope.file_write_size_group).toBe('')
        expect(sut.scope.password_lifetime_group).toBe('')
    })


})
