/*
Tests for accounts controller.
*/


describe('AccountsCtrl', function() {


    /*
    Will redirect to the start page when there is no configuration.
    The start page can wait longer for the configuration to be available.
    */
    test('__init__no_config', function() {
        var sut = mk.getController('accountsCtrl')
        sut.$location.path('/accounts')
        expect(sut.$location.path()).toEqual('/accounts')

        sut.scope.__init__()

        expect(sut.$location.path()).toEqual('/')
    })

    /*
    Will remain on the same page if there is a configuration.
    */
    test('__init__with_config', function() {
        var sut = mk.getController('accountsCtrl')
        sut.$location.path('/accounts')
        sut.scope.configuration.server = {'uuid': 'some-id'}
        expect(sut.$location.path()).toEqual('/accounts')

        sut.scope.__init__()

        expect(sut.$location.path()).toEqual('/accounts')
    })

    /*
    Will add a new account in new accounts cache and then
    go to the page of the new account.
    */
    test('onAddAccount', function() {
        var sut = mk.getController('accountsCtrl')
        sut.utils.constants['CONFIGURATION_ACCOUNT_DEFAULTS'] = {
            'home_folder_path': 'inherit',
            'create_home_folder': 'No',
            'lock_in_home_folder': 'Yes'
            }
        sut.scope.$parent.new_entities['accounts'] = []
        var new_accounts = sut.scope.$parent.new_entities['accounts']
        var parent = sut.scope.$parent.component_categories.accounts
        var tab = {
            'container': parent,
            'default_group_uuids': ['some-uuid', 'other-uuid']
        }
        sut.scope.onAddAccount(tab)

        expect(new_accounts.length).toBe(1)
        expect(new_accounts[0]['enabled']).toBeTrue()
        expect(new_accounts[0]['uuid']).toBeUndefined()
        expect(new_accounts[0]['type']).toBe('application')
        expect(new_accounts[0]['home_folder_path']).toBe('inherit')
        expect(new_accounts[0]['create_home_folder']).toBe('inherit')
        expect(new_accounts[0]['lock_in_home_folder']).toBe('inherit')
        expect(new_accounts[0]['disable_on_inactivity']).toBe('inherit')
        expect(new_accounts[0]['password_lifetime']).toBe('inherit')
        expect(new_accounts[0]['file_write_size']).toBe('inherit')
        expect(new_accounts[0]['name']).toStartWith('New Account ')
        expect(new_accounts[0]['group']).toEqual(['some-uuid', 'other-uuid'])
        expect(sut.$location.path()).toStartWith('/accounts/New Account ')
    })

    /*
    When no filter is provided it will have available all the accounts,
    including newly created accounts.

    This is also a test to check the details getter for accounts.
    */
    test('onFilterAccounts_all', function() {
        var sut = mk.getController('accountsCtrl')

        sut.scope.$parent.configuration.groups = {
            'group-uuid-1':{
                'uuid': 'group-uuid-1',
                'name': 'Group-1',
                'description': 'Some group 1 description',
                'source_ip_filter': [
                    ['allow', '6.5.6.4'],
                    ['deny', '1.2.3.4/16'],
                ],
                'home_folder_path': '/some/base-Path',
                'allow_certificate_authentication': true,
            },
            'group-uuid-2':{
                'uuid': 'group-uuid-2',
                'name': 'Group-2',
                'description': '',
                'source_ip_filter': [['allow', '1.2.3.4/24']],
                'home_folder_path': 'Not userd',
                'allow_certificate_authentication': false,
            },
        }

        sut.scope.$parent.configuration.accounts = {
            'uuid-1': {
                'uuid': 'uuid-1',
                'name': 'Last User-1',
                'group': ['group-uuid-1', 'group-uuid-2'],
                'kind_name': 'application',
                'description': 'Some description',
                'source_ip_filter': [
                    ['allow', '1.2.3.4'],
                    ['deny', '4.5.6.7/16'],
                ],
                'home_folder_path': '/some/Unix/path with spaces',
                'multi_factor_authentication': 'otpauth://some-secret',
                'expire_datetime': 'ISO-DATETIME',
                'ssh_authorized_keys_content': 'rsa some-base64 comment',
                'allow_certificate_authentication': false,
            },
            'uuid-2': {
                'uuid': 'uuid-2',
                'name': 'First User 2',
                'kind_name': 'os',
                'group': ['group-uuid-1'],
                'description': '',
                'source_ip_filter': [],
                'home_folder_path': 'inherit',
                'multi_factor_authentication': '',
                'expire_datetime': '',
                'ssh_authorized_keys_content': '',
                'allow_certificate_authentication': 'inherit',
            },

        }
        sut.scope.$parent.new_entities.accounts = [{
            'name': 'Intermediary New 4',
            'kind_name': 'application',
            'group': ['group-uuid-2'],
            'description': 'Some description',
            'source_ip_filter': [],
            'home_folder_path': 'C:\\Win \\Path',
            'multi_factor_authentication': '',
            'expire_datetime': '',
            'ssh_authorized_keys_content': '',
            'allow_certificate_authentication': true,
        }]

        sut.scope.onFilterAccounts()

        expect(sut.scope.filtered_accounts).toHaveLength(3)

        expect(sut.scope.filtered_accounts[0].config.name).toEqual(
            'First User 2')
        expect(sut.scope.filtered_accounts[0].details).toEqual([
            {"key": "Type", "value": "OS"},
            {"key": "", "value": ""},
            {"key": "Source IPs", "value": (
                "(inherited)allow-6.5.6.4 (inherited)deny-1.2.3.4/16")},
            {"key": "Home folder", "value": 'inherit'},
            {"key": "TOTP enabled", "value": "No"},
            {"key": "Password expiration", "value": ""},
            {"key": "SSH keys", "value": "No"},
            {"key": "SSL cert", "value": "(inherited) Yes"},
        ])

        expect(sut.scope.filtered_accounts[1].config.name).toEqual(
            'Intermediary New 4')
        expect(sut.scope.filtered_accounts[1].details).toEqual([
            // For application, the type is empty and hidden from UI.
            {"key": "Type", "value": ""},
            {"key": "", "value": "Some description"},
            // Source IPs are from direct configuration + all associated groups.
            {"key": "Source IPs", "value": "(inherited)allow-1.2.3.4/24"},
            {"key": "Home folder", "value": 'C:\\Win \\Path'},
            {"key": "TOTP enabled", "value": "No"},
            {"key": "Password expiration", "value": ""},
            {"key": "SSH keys", "value": "No"},
            {"key": "SSL cert", "value": "Yes"},
        ])

        expect(sut.scope.filtered_accounts[2].config.name).toEqual(
            'Last User-1')
        expect(sut.scope.filtered_accounts[2].details).toEqual([
            // For application, the type is empty and hidden from UI.
            {"key": "Type", "value": ""},
            {"key": "", "value": "Some description"},
            {"key": "Source IPs", "value": (
                "allow-1.2.3.4 deny-4.5.6.7/16 " +
                "(inherited)allow-6.5.6.4 (inherited)deny-1.2.3.4/16 " +
                "(inherited)allow-1.2.3.4/24"
                )},
            {"key": "Home folder", "value": '/some/Unix/path with spaces'},
            {"key": "TOTP enabled", "value": "Yes"},
            {"key": "Password expiration", "value": "ISO-DATETIME"},
            {"key": "SSH keys", "value": "Yes"},
            {"key": "SSL cert", "value": "No"},
        ])

    })


})
