/*
Tests for authentication method controller.
*/


describe('AuthenticationCtrl', function() {

    var compiler
    var scope
    var utils
    var ctrl
    var result
    var routeParams
    var $location

    require('../app.js')
    beforeEach(angular.mock.module('app'))
    beforeEach(inject(function($injector, _$compile_, $rootScope, $controller) {
        /*
        Prepare the scope and controller before each test.
        */
        utils = mk.servicesUtils($injector)

        compiler = _$compile_
        var parent_scope = $rootScope.$new();
        ctrl = $controller(
            'mainCtrl', {
            $scope: parent_scope,
            utils: utils
            })

        routeParams = {}
        scope = parent_scope.$new();
        ctrl = $controller(
            'authenticationCtrl', {
            $routeParams: routeParams,
            $scope: scope,
            utils: utils
        })

        $location = $injector.get('$location')
        // Run an initial digest to 'settle' all watchers.
        scope.$digest()
    }))

    /*
    Update the config so that it has an authentication.

    It returns the configuration for the new auth method.
    */
    createAuthentication = function(os_name, type) {
        if (!os_name) {
            os_name = 'linux'
        }
        if (!type) {
            type = 'some-type'
        }
        var auth_method = {
            'uuid': 'some-uuid',
            'name': 'some-name',
            'type': type,
            'target': ['123', '345']
            }
        scope.$parent.configuration = {'authentications': {
            'some-uuid': auth_method
        }}
        scope.$parent.process.runnables = {
            'authentications': {'some-uuid': {'uuid': 'some-uuid'}},
            'status': {'os_name': os_name}
        }
        routeParams['uuid'] = 'some-uuid'
        return auth_method
    }

    /*
    Check that PAM configuration is not visible for `os_name`
    */
    assertNoPAMConfiguration = function(os_name) {
        createAuthentication(os_name)

        scope.__init__()

        expect(scope.have_configurable_pam).toBeFalse()
    }

    /*
    Check that PAM configuration is visible for `os_name`.
    */
    assertPAMConfiguration = function(os_name) {
        createAuthentication(os_name)

        scope.__init__()

        expect(scope.have_configurable_pam).toBeTrue()
    }

    /*
    Goes to authentications page if could not find the method.
    */
    test('init_no_uuid', function() {
        scope.$parent.configuration = {'authentications': []}
        routeParams['uuid'] = 'some-name'

        scope.__init__()

        expect($location.path()).toBe('/authentications')
    })

    /*
    It can handle authentication method details from newly created objects,
    that are not applied to the server.
    They are only in config pre-stage.

    New components have no runnable.

    This is also a tests for the HTTP method for which the SSL UI is
    visible.
    */
    test('init_new_http_method', function() {
        var new_method = {
            'name': 'some-name',
            'type': 'http',
            'target': ['123', '345']
            }
        scope.$parent.configuration = {'authentications': []}
        scope.$parent.new_entities['authentications'] = [new_method]
        routeParams['uuid'] = 'some-name'

        scope.__init__()

        expect(scope.authentication).toBe(new_method)
        // It has the SSL configuration alias.
        expect(scope.ssl_configuration).toBe(new_method)
        expect(scope.template_path).toEqual(
            'partials/authentication_http.html?v=%(cache)s')
        expect(scope.runnable).toBeFalse()
    })

    /*
    It shows authentication method edit from existing configuration.

    They have a runnable.

    This is also a test for the Entra ID for which the SSL ui is not
    visible.
    */
    test('init_existing_entra_id', function() {
        var method = {
            'uuid': 'some-uuid',
            'name': 'some-name',
            'type': 'entra-id',
            'target': ['123', '345']
            }
        scope.$parent.configuration = {
            'authentications': {'some-uuid': method}}
        scope.$parent.new_entities['authentications'] = []
        scope.$parent.process.runnables = {
            'authentications': {'some-uuid': {'uuid': 'some-uuid'}}
        }
        routeParams['uuid'] = 'some-uuid'

        scope.__init__()

        expect(scope.authentication).toBe(method)
        expect(scope.runnable).toBeDefined()
    })

    /*
    Updates the allowed_groups for OS auth.
    */
    test('init_os_allowed_groups', function() {
        var method = {
            'uuid': 'some-uuid',
            'name': 'some-name',
            'type': 'os',
            'allowed_groups': ['aDm', 'root']
            }
        scope.$parent.configuration = {
            'authentications': {'some-uuid': method}}
        scope.$parent.new_entities['authentications'] = []
        scope.$parent.process.runnables = {
            'authentications': {'some-uuid': {'uuid': 'some-uuid'}}
        }
        routeParams['uuid'] = 'some-uuid'

        scope.__init__()

        expect(scope.authentication).toBe(method)
    })

    /*
    PAM options are not displayed on systems on which it does not make sense.
    */
    test('init_pam_hide', function() {
        assertNoPAMConfiguration('windows')
        assertNoPAMConfiguration('osx')
    })

    /*
    PAM options are displayed on systems supporting it.
    */
    test('init_pam_show', function() {
        assertPAMConfiguration('linux')
    })

    /*
    It will remove the current authentication from the config and
    moved to authentication index page.
    */
    test('onDeleteAuthentication_existing', function() {
        var auth_method = createAuthentication()
        var authentications = scope.$parent.configuration.authentications
        scope.__init__()
        expect(authentications).toContainKey(auth_method.uuid)

        scope.onDeleteAuthentication()

        expect(authentications).not.toContainKey(auth_method.uuid)
        expect($location.path()).toBe('/authentications')
    })

})
