/*
Tests for authentications controller.
*/


describe('AuthenticationsCtrl', function() {

    var result

    /*
    Go to start page if no configuration is defined.
    */
    test('__init__no_config', function() {
        var sut = mk.getController('authenticationsCtrl')
        sut.$location.path('/authentications')
        expect(sut.$location.path()).toEqual('/authentications')

        sut.scope.__init__()

        expect(sut.$location.path()).toEqual('/')
    })

    /*
    OS authentication method is not available when server runs on macOS.
    */
    test('__init__macos_no_os_auth', function() {
        var sut = mk.getController('authenticationsCtrl')
        sut.scope.configuration.server = {'name': 'something'}
        sut.scope.process.runnables.status = {'os_name': (
            sut.utils.constants.TYPE_NAME.OSX
        )}
        sut.scope.__init__()

        expect(sut.scope.authentication_types).toEqual([
            "anonymous",
            "application",
            "deny-username",
            "http",
            "ldap",
            "local-file",
            "entra-id",
            "radius",
            "ip-time-ban",
        ])
    })

    /*
    OS authentication method is available when running on Windows.
    */
    test('__init__windows_auth', function() {
        var sut = mk.getController('authenticationsCtrl')
        sut.scope.configuration.server = {'name': 'something'}
        sut.scope.process.runnables.status = {'os_name': (
            sut.utils.constants.TYPE_NAME.WINDOWS
        )}
        sut.scope.__init__()

        expect(sut.scope.authentication_types).toEqual([
            "anonymous",
            "application",
            "deny-username",
            "http",
            "ldap",
            "local-file",
            "entra-id",
            "os",
            "radius",
            "ip-time-ban",
        ])
    })

    /*
    It will navigate to the default manager page.
    */
    test('onGoToLocalManager', function() {
        var sut = mk.getController('authenticationsCtrl')
        sut.$location.path('/authentications')
        sut.scope.configuration.services = {'DEFAULT-MANAGER': {
            'uuid': 'manager-uuid',
        }}
        expect(sut.$location.path()).toEqual('/authentications')

        sut.scope.onGoToLocalManager()

        expect(sut.$location.path()).toEqual('/services/manager-uuid')
    })


    /*
    Keeps on the same page and sets the server configuration on the local
    scope with all the auth methos available in the filtered list.
    */
    test('__init__', function() {
        var server = {
            'uuid': 'server-uuid',
            'name': 'server-name',
            'authentications': [],
        }
        var sut = mk.getController('authenticationsCtrl')
        sut.$location.path('/authentications')
        expect(sut.$location.path()).toEqual('/authentications')
        sut.scope.$parent.configuration.server = server
        sut.scope.$parent.configuration.authentications = {
            'uuid-1': {'uuid': 'uuid1'},
            'uuid-2': {'uuid': 'uuid2'},
            }
        sut.scope.$parent.process.runnables.authentications = {
            'uuid-1': {'uuid': 'uuid1', 'name': 'Runnable 1'},
            'uuid-2': {'uuid': 'uuid2', 'name': 'Runnable 2'},
        }
        sut.scope.$parent.new_entities.authentications = [
            {'name': 'new auth 3'}
        ]

        sut.scope.__init__()

        expect(sut.$location.path()).toEqual('/authentications')
        expect(sut.scope.server).toBe(server)
        expect(sut.scope.filtered_authentications).toHaveLength(3)
    })

    /*
    Return the list with only the auth methods that .
    */
    test('getExistingAuthentications', function() {
        var server = {
            'uuid': 'server-uuid',
            'name': 'server-name',
        }
        var sut = mk.getController('authenticationsCtrl')
        sut.$location.path('/authentications')
        expect(sut.$location.path()).toEqual('/authentications')
        sut.scope.$parent.configuration.server = {'uuid': 'server-uuid'}
        sut.scope.$parent.configuration.authentications = {
            'uuid-1': {'uuid': 'uuid1'},
            'uuid-2': {'uuid': 'uuid2'},
            }
        sut.scope.$parent.new_entities.authentications = [
            {'name': 'new auth 3'}
        ]

        result = sut.scope.getExistingAuthentications()

        expect(result).toEqual([{'uuid': 'uuid1'}, {'uuid': 'uuid2'}])
    })

    /*
    When no filter is provided it will have availabe all the auth methods,
    including newly created methods.
    */
    test('onFilterAuthentications_all', function() {
        var sut = mk.getController('authenticationsCtrl')
        sut.scope.$parent.configuration.server = {
            'uuid': 'server-uuid',
            'name': 'server-name',
            'authentications': ['uuid1', 'uuid2'],
        }
        sut.scope.$parent.configuration.authentications = {
            'uuid-1': {'uuid': 'uuid1', 'type': 'http'},
            'uuid-2': {'uuid': 'uuid2'},
            }
        sut.scope.$parent.process.runnables.authentications = {
            'uuid-1': {'uuid': 'uuid1', 'name': 'Runnable 1'},
            'uuid-2': {'uuid': 'uuid2', 'name': 'Runnable 2'},
        }
        sut.scope.$parent.configuration.services = {
            'uuid-3': {
                'uuid': 'uuid3',
                'name': 'https Server',
                'authentications': ['uuid2'],
            },
            'uuid-4': {'uuid': 'uuid4', 'authentications': []},
            }
        sut.scope.$parent.new_entities.authentications = [
            {'name': 'new auth 3'}
        ]

        sut.scope.onFilterAuthentications()

        expect(sut.scope.filtered_authentications).toHaveLength(3)
        expect(
            sut.scope.filtered_authentications[0].is_delete_blocked).toEqual('')
        expect(
            sut.scope.filtered_authentications[1].is_delete_blocked).toEqual(
                'Used by: Default authentication.')
        expect(
            sut.scope.filtered_authentications[2].is_delete_blocked).toEqual(
                'Used by: Default authentication, https Server.')
    })


    /*
    onAddAuthentication will add a new HTTP authentication in the cache and
    then go to the page of the new method.
    */
    test('onAddAuthentication_http', function() {
        var sut = mk.getController('authenticationsCtrl')
        var new_authentications = sut.scope.$parent.new_entities[
            'authentications']

        sut.scope.onAddAuthentication('http')

        expect(new_authentications.length).toBe(1)
        expect(new_authentications[0]['enabled']).toBeTrue()
        expect(new_authentications[0]['uuid']).toBeUndefined()
        expect(new_authentications[0]['url']).toEqual(
            ['http://ws.example.org/auth-handler'])
        expect(new_authentications[0]['name']).toStartWith('New http method ')
        expect(sut.$location.path()).toStartWith(
            '/authentications/New http method ')
    })

    /*
    onAddAuthentication can add a radius auth method.
    */
    test('onAddAuthentication_radius', function() {
        var sut = mk.getController('authenticationsCtrl')
        var new_authentications = sut.scope.$parent.new_entities[
            'authentications']

        sut.scope.onAddAuthentication('radius')

        expect(new_authentications.length).toBe(1)
        expect(new_authentications[0]['enabled']).toBeTrue()
        expect(new_authentications[0]['uuid']).toBeUndefined()
        expect(new_authentications[0]['address']).toEqual('radius.example.com')
        expect(new_authentications[0]['name']).toStartWith('New radius method ')
        expect(sut.$location.path()).toStartWith(
            '/authentications/New radius method ')
    })

    /*
    An Entra ID can be added.
    */
    test('onAddAuthentication_entra_id', function() {
        var sut = mk.getController('authenticationsCtrl')
        var new_authentications = sut.scope.$parent.new_entities[
            'authentications']

        sut.scope.onAddAuthentication('entra-id')

        expect(new_authentications.length).toBe(1)
        expect(new_authentications[0]['enabled']).toBeTrue()
        expect(new_authentications[0]['uuid']).toBeUndefined()
        expect(new_authentications[0]['name']).toStartWith(
            'New entra-id method ')
        expect(sut.$location.path()).toStartWith(
            '/authentications/New entra-id method ')

        // Some default values are checked here.
        expect(new_authentications[0]['directory_id']).toEqual(
            'YOUR-ENTRA-ID-DIRECTORY-ID')
    })

    describe('_getDefault', function() {

        test(
        'Get default for unknown.',
        function() {
            var sut = mk.getController('authenticationsCtrl')
            result = sut.scope._getDefault('unknown-type')

            expect(result).toEqual({
                'type': 'unknown-type',
                'enabled': true
            })
        })

        test(
        'Get default for http.',
        function() {
            var sut = mk.getController('authenticationsCtrl')
            result = sut.scope._getDefault('http')

            expect(result).toEqual({
                'enabled': true,
                'type': 'http',
                'url': ['http://ws.example.org/auth-handler'],
                'timeout': 120,
                'ssl_allowed_methods': ['secure'],
                'ssl_cipher_list': 'secure',
                'ssl_certificate': '',
                'ssl_key': '',
                'ssl_certificate_authority': '',
                'ssl_certificate_revocation_list': '',
                'ssl_certificate_revocation_list_refresh': 0
            })
        })

        test(
        'Get default for LDAP.',
        function() {
            var sut = mk.getController('authenticationsCtrl')
            result = sut.scope._getDefault('ldap')

            expect(result).toEqual({
                'enabled': true,
                'type': 'ldap',
                'address': '127.0.0.1',
                'port': 389,
                'bind_dn_type': 'parent',
                'bind_dn': ['CN=Users,dc=domain,dc=example,dc=com'],
                'username_attribute': 'cn',
                'email_attribute': 'mail',
                'home_folder_attribute': ['homeDirectory']
            })
        })

        test(
        'Get default for OS.',
        function() {
            var sut = mk.getController('authenticationsCtrl')
            result = sut.scope._getDefault('os')

            expect(result).toEqual({
                'enabled': true,
                'type': 'os',
                'allowed_groups': [
                    'example-os-group-name',
                    'other-example-group'
                ],
                'manager_allowed_groups': [
                    'admin-example-group3',
                    'admin-example-group4'
                ],
                'group_association': 'base-groups',
                'base_groups': ['DEFAULT_GROUP'],
                'pam_usage': 'fallback',
                'pam_service': 'login'
            })
        })

    })

})
