/*
Tests for event handlers controller.
*/


describe('EventHandlersCtrl', function() {

    const allHandlers = [
        'account-interaction',
        'create-archive',
        'database',
        'email-sender',
        'extension',
        'external-executable',
        'extract-archive',
        'file-digest',
        'file-dispatcher',
        'http',
        'local-file',
        'openpgp',
        'rabbitmq',
        'standard-stream',
        'syslog',
        'windows-eventlog'
    ];


    /*
    When there is no configuration it list all available types.
    */
    test('__init__no_types', function() {
        var sut = mk.getController('eventHandlersCtrl')

        sut.scope.__init__()

        expect(sut.scope.event_handler_types).toIncludeSameMembers(allHandlers)
    })

    test(
    'When configuration is empty it list all available types.',
    function() {
        var sut = mk.getController('eventHandlersCtrl')
        sut.scope.$parent.upstream_configuration = {'event_handlers': {}}

        sut.scope.__init__()

        expect(sut.scope.event_handler_types).toIncludeSameMembers(allHandlers)
    })


    test(
    'When there is no configured account activity event handler ' +
    'it list all available types.',
    function() {
        var sut = mk.getController('eventHandlersCtrl')

        sut.scope.$parent.upstream_configuration = {'event_handlers': {
            'some-uuid': {'type': 'other-type'}
            }}

        sut.scope.$parent.new_entities = {'event_handlers': {
            'some-name': {'type': 'some-other-type'}
            }}

        sut.scope.__init__()

        expect(sut.scope.event_handler_types).toIncludeSameMembers(allHandlers)
    })

    test(
    'When server is running on Linux/Unix the Windows Event handler, ' +
    'it is not available.',
    function() {
        var sut = mk.getController('eventHandlersCtrl')
        sut.scope.$parent.process.runnables = {
            'status': {'os_family': 'Unix'}
        }
        sut.scope.__init__()

        expect(sut.scope.event_handler_types).toIncludeSameMembers([
            'account-interaction',
            'create-archive',
            'database',
            'email-sender',
            'extension',
            'external-executable',
            'extract-archive',
            'file-dispatcher',
            'file-digest',
            'http',
            'local-file',
            'openpgp',
            'rabbitmq',
            'standard-stream',
            'syslog'
            ])
    })

    test(
    'When server is running on Windows the Windows Event handler, ' +
    'is available.',
    function() {
        var sut = mk.getController('eventHandlersCtrl')
        sut.scope.$parent.process.runnables = {
            'status': {'os_family': 'Windows'}
        }
        sut.scope.__init__()

        expect(sut.scope.event_handler_types).toIncludeSameMembers(allHandlers)
    })

    /*
    onAddEventHandler will add a handler in the cache and then go
    to the page of the new handler.
    */
    test('onAddEventHandler_http', function() {
        var sut = mk.getController('eventHandlersCtrl')
        var new_handlers = sut.scope.$parent.new_entities['event_handlers']

        sut.scope.onAddEventHandler('http')

        expect(new_handlers.length).toBe(1)
        expect(new_handlers[0]['enabled']).toBeTrue()
        expect(new_handlers[0]['uuid']).toBeUndefined()
        expect(new_handlers[0]['url']).toEqual(
            ['http://ws.example.org/handler'])
        expect(new_handlers[0]['name']).toStartWith('New http handler ')
        expect(sut.$location.path()).toStartWith(
            '/event-handlers/New http handler ')
    })

    /*
    onAddEventHandler can add a new account-interaction event handler.
    */
    test('onAddEventHandler_account_interaction', function() {
        var sut = mk.getController('eventHandlersCtrl')
        var new_handlers = sut.scope.$parent.new_entities['event_handlers']

        sut.scope.onAddEventHandler('account-interaction')

        // General options.
        expect(new_handlers.length).toBe(1)
        expect(new_handlers[0]['enabled']).toBeTrue()
        expect(new_handlers[0]['uuid']).toBeUndefined()
        expect(new_handlers[0]['name']).toStartWith(
            'New account-interaction handler ')
        expect(sut.$location.path()).toStartWith(
            '/event-handlers/New account-interaction handler ')
        // Specific options.
        expect(new_handlers[0]['monitored_operations']).toEqual(['created'])
    })

    /*
    It can add an event of type file digest.
    */
    test('onAddEventHandler_file_digest', function() {
        var sut = mk.getController('eventHandlersCtrl')
        var new_handlers = sut.scope.$parent.new_entities['event_handlers']
        expect(new_handlers.length).toBe(0)

        sut.scope.onAddEventHandler('file-digest')

        // General options.
        expect(new_handlers.length).toBe(1)
        expect(new_handlers[0]['enabled']).toBeTrue()
        expect(new_handlers[0]['uuid']).toBeUndefined()
        expect(new_handlers[0]['name']).toStartWith(
            'New file-digest handler ')
        expect(sut.$location.path()).toStartWith(
            '/event-handlers/New file-digest handler ')
        // Specific options.
        expect(new_handlers[0]['digest_algorithm']).toEqual('sha512')
        expect(new_handlers[0]['output_format']).toEqual('gnu')
        expect(new_handlers[0]['source_attribute']).toEqual('real_path')
    })

    /*
    onFilterHandlers can return info for account-interaction event handler
    both configuration and runnable.
    */
    test('onFilterHandlers_account_interaction', function() {
        var sut = mk.getController('eventHandlersCtrl')
        var config = {
            'uuid': 'uuid1',
            'type': sut.utils.constants.TYPE_NAME.ACCOUNT_INTERACTION,
            'name': 'Some account interaction',
            'monitored_operations': [
                sut.utils.constants.EVENT_TYPE.DELETED,
                sut.utils.constants.EVENT_TYPE.MOVED,
            ]
        }
        var runnable = {
            'uuid': 'uuid1',
            'status': 'started',
        }
        sut.scope.$parent.configuration.event_handlers = {'uuid1': config}
        sut.scope.$parent.process.runnables.event_handlers = {
            'uuid1': runnable}

        expect(sut.scope.filtered_handlers).toEqual([])

        // Is enough to trigger the filter with 3 letters.
        sut.scope.onFilterHandlers('acc')

        // A single handler is filtered.
        expect(sut.scope.filtered_handlers).toHaveLength(1)
        var result = sut.scope.filtered_handlers[0]
        expect(result.uuid).toEqual('uuid1')
        expect(result.just_created).toBeFalse()
        expect(result.runnable).toBe(runnable)
        expect(result.config).toBe(config)
        expect(result.details).toIncludeAllMembers([
            {
                'key': 'Type',
                'value': 'Monitor inter-account file operations',
            },
            {
                'key': 'Operations',
                'value': 'deleted, moved',
            },
        ])
    })

    /*
    onFilterHandlers can return info for file-digest event handler
    both configuration and runnable.
    */
    test('onFilterHandlers_file_digest', function() {
        var sut = mk.getController('eventHandlersCtrl')
        var config = {
            'uuid': 'uuid1',
            'type': sut.utils.constants.TYPE_NAME.FILE_DIGEST,
            'name': 'Some file digest',
            'fail_after_errors': 24,
            'target': ['123', '!345'],
            'digest_algorithm': 'md5',
            'source_attribute': 'destination_path',
        }
        var runnable = {
            'uuid': 'uuid1',
            'status': 'started',
        }
        sut.scope.$parent.configuration.event_handlers = {'uuid1': config}
        sut.scope.$parent.process.runnables.event_handlers = {
            'uuid1': runnable}

        expect(sut.scope.filtered_handlers).toEqual([])

        // Is enough to trigger the filter with 3 letters.
        sut.scope.onFilterHandlers('dige')

        // A single handler is filtered.
        expect(sut.scope.filtered_handlers).toHaveLength(1)
        var result = sut.scope.filtered_handlers[0]
        expect(result.uuid).toEqual('uuid1')
        expect(result.just_created).toBeFalse()
        expect(result.runnable).toBe(runnable)
        expect(result.config).toBe(config)
        expect(result.details).toIncludeAllMembers([
            {
                'key': 'Type',
                'value': 'Generate file digest message',
            },
            {
                'key': 'Algorithm',
                'value': 'md5',
            },
            {
                'key': 'Source attribute',
                'value': 'destination_path',
            },
        ])
    })

    /*
    onFilterHandlers can return info for HTTP event handler and the
    enabled event filters.
    */
    test('onFilterHandlers_http', function() {
        var sut = mk.getController('eventHandlersCtrl')
        var config = {
            'uuid': 'uuid-http-1',
            'type': sut.utils.constants.TYPE_NAME.HTTP,
            'name': 'Web Hook',
            'http_content_type': 'xml',
            'url': ['http://first.url' , 'https://second.com/query'],
            'target': ['123', '!234'],
            'data_filter': [['path', '*.pdf'], ['username', 'some-user']],
            'components': ['uuid-1', 'uuid-http-1'],
            'usernames': ['some-user-1'],
            'groups': ['failure', 'failure-critical'],
            'source_addresses': ['allow 12.12/12', 'deny 1.2.3.4'],
            'fail_after_errors': 123,
        }

        sut.scope.$parent.configuration.locations = {'uuid-1': {
            'uuid': 'uuid-1',
            'type': 'sftp',
            'name': 'SFTP Server',
        }}
        var runnable = {
            'uuid': 'uuid-http-1',
            'status': 'started',
        }
        sut.scope.$parent.configuration.event_handlers = {'uuid-http-1': config}
        sut.scope.$parent.process.runnables.event_handlers = {
            'uuid-http-1': runnable}

        expect(sut.scope.filtered_handlers).toEqual([])

        // It can filter based on type name description.
        sut.scope.onFilterHandlers('http req')

        // A single handler is filtered.
        expect(sut.scope.filtered_handlers).toHaveLength(1)
        var result = sut.scope.filtered_handlers[0]
        expect(result.uuid).toEqual('uuid-http-1')
        expect(result.just_created).toBeFalse()
        expect(result.runnable).toBe(runnable)
        expect(result.config).toBe(config)
        expect(result.details).toIncludeSameMembers([
            {
                'key': 'Type',
                'value': 'Send as HTTP request',
            },
            {
                'key': 'URL',
                'value': 'http://first.url https://second.com/query',
            },
            {
                'key': 'Content',
                'value': 'xml',
            },
        ])
        expect(result.filters).toIncludeSameMembers([
            {
                'key': 'Events',
                'value': '123, !234',
            },
            {
                'key': 'Event groups',
                'value': 'failure, failure-critical',
            },
            {
                'key': 'Accounts',
                'value': 'some-user-1',
            },
            {
                'key': 'Components',
                'value': 'SFTP Server, Web Hook',
            },
            {
                'key': 'Source IPs',
                'value': 'allow 12.12/12, deny 1.2.3.4',
            },
            {
                'key': 'Data',
                'value': 'path = *.pdf AND username = some-user',
            },
            {
                'key': 'Stop after errors',
                'value': 123,
            },
        ])
    })

    /*
    onFilterHandlers can return info for email sender event handler
    */
    test('onFilterHandlers_email_sender', function() {
        var sut = mk.getController('eventHandlersCtrl')
        var config = {
            'uuid': 'email-1',
            'type': sut.utils.constants.TYPE_NAME.EMAIL_SENDER,
            'name': 'Email Notify',
            'email_subject': '{id} some subject',
            'email_to_recipients': ['first@ex.com' , 'other@domain.org'],
        }
        var runnable = {
            'uuid': 'email-1',
            'status': 'stopped',
        }
        sut.scope.$parent.configuration.event_handlers = {'email-1': config}
        sut.scope.$parent.process.runnables.event_handlers = {
            'email-1': runnable}

        expect(sut.scope.filtered_handlers).toEqual([])

        // Filter is alwasy triggered on empty input.
        sut.scope.onFilterHandlers('')

        // A single handler is filtered.
        expect(sut.scope.filtered_handlers).toHaveLength(1)
        var result = sut.scope.filtered_handlers[0]
        expect(result.uuid).toEqual('email-1')
        expect(result.just_created).toBeFalse()
        expect(result.runnable).toBe(runnable)
        expect(result.config).toBe(config)
        expect(result.details).toIncludeSameMembers([
            {
                'key': 'Type',
                'value': 'Send as email',
            },
            {
                'key': 'Subject',
                'value': '{id} some subject',
            },
            {
                'key': 'Recipients',
                'value': 'first@ex.com, other@domain.org',
            },
        ])

    })


    /*
    onFilterHandlers can return info for file dispatcher event handler
    */
    test('onFilterHandlers_file_dispatcher', function() {
        var sut = mk.getController('eventHandlersCtrl')
        var config = {
            'uuid': 'dispatch-1',
            'type': sut.utils.constants.TYPE_NAME.FILE_DISPATCHER,
            'name': 'File Dispatch',
            'overwrite_rule': 'skip',
            'retry_count': 24,
            'retry_wait': 42,
            'dispatch_delay': 10,
            'destination_temporary_suffix': '.filepart',
        }
        var runnable = {
            'uuid': 'dispatch-1',
            'status': 'stopped',
        }
        sut.scope.$parent.configuration.event_handlers = {'dispatch-1': config}
        sut.scope.$parent.process.runnables.event_handlers = {
            'dispatch-1': runnable}

        expect(sut.scope.filtered_handlers).toEqual([])

        // Can match on minium 3 characters.
        sut.scope.onFilterHandlers('Dis')

        // A single handler is filtered.
        expect(sut.scope.filtered_handlers).toHaveLength(1)
        var result = sut.scope.filtered_handlers[0]
        expect(result.uuid).toEqual('dispatch-1')
        expect(result.just_created).toBeFalse()
        expect(result.runnable).toBe(runnable)
        expect(result.config).toBe(config)
        expect(result.details).toIncludeSameMembers([
            {
                'key': 'Type',
                'value': 'Dispatch file',
            },
            {
                "key": "Overwrite rule",
                "value": "skip",
            },
            {
                "key": "Retry count",
                "value": 24,
            },
            {
                "key": "Retry wait",
                "value": 42,
            },
            {
                "key": "Dispatch delay",
                "value": 10,
            },
            {
                "key": "Temporary suffix",
                "value": ".filepart",
            }
        ])

    })

    test(
    'onDeleteEventHandler will remove the newly created handler.',
    function() {
        var sut = mk.getController('eventHandlersCtrl')
        var handler = {
            'uuid': undefined,
            'name': mk.string(),
            'type': 'http'
            }
        sut.scope.$parent.new_entities['event_handlers'].push(handler)

        sut.scope.onDeleteEventHandler(handler)

        expect(sut.scope.$parent.new_entities['event_handlers'].length).toBe(0)
    })

    test(
    'onDeleteEventHandler will remove an existing handler.',
    function() {
        var sut = mk.getController('eventHandlersCtrl')
        var uuid = mk.string()
        var handler = {
            'uuid': uuid,
            'name': 'existing name',
            'type': 'http'
            }
        sut.scope.$parent.configuration = {'event_handlers': {}}
        sut.scope.$parent.configuration['event_handlers'][uuid] = handler

        sut.scope.onDeleteEventHandler(handler)

        expect(sut.scope.$parent.configuration['event_handlers']).toBeEmpty()
    })

    describe('_getDefault', function() {

        test(
        'Get default for unknown.',
        function() {
            var sut = mk.getController('eventHandlersCtrl')
            var result = sut.scope._getDefault('unknown-type')

            expect(result).toEqual({
                'type': 'unknown-type',
                'enabled': true,
                'target': ['123456'],
                'usernames': [],
                'components': [],
                'fail_after_errors': 10
            })
        })

        allHandlers.forEach(function(handler) {
            test(
            'Get default for ' + handler + ' includes common defaults.',
            function() {
                var sut = mk.getController('eventHandlersCtrl')
                var result = sut.scope._getDefault(handler)
                expect(result).toContainEntries([
                    ['type', handler],
                    ['enabled', true],
                    ['target', ['123456']],
                    ['usernames', []],
                    ['components', []],
                    ['fail_after_errors', 10]
                ])
            })
        })

        test(
        'Get default for http.',
        function() {
            var sut = mk.getController('eventHandlersCtrl')
            var result = sut.scope._getDefault('http')

            expect(result).toEqual({
                'enabled': true,
                'type': 'http',
                'url': ['http://ws.example.org/handler'],
                'http_content_type': 'json',
                'extra_data': '{"text": "{message}"}',
                'retry_count': 2,
                'retry_increase': 10,
                'usernames': [],
                'components': [],
                'target': ['123456'],
                'fail_after_errors': 10,
                'timeout': 120,
            })
        })

        test(
        'Get default for rabbitmq.',
        function() {
            var sut = mk.getController('eventHandlersCtrl')
            var result = sut.scope._getDefault('rabbitmq')

            expect(result).toEqual({
                'enabled': true,
                'type': 'rabbitmq',
                'url': 'amqps://YOUR.RABBITMQ.HOST/VIRTUAL_HOST_NAME',
                'routing_key': 'test-queue-name',
                'payload_template': 'New message {id} {data_json}',
                'retry_count': 2,
                'retry_increase': 10,
                'usernames': [],
                'components': [],
                'target': ['123456'],
                'fail_after_errors': 10
            })
        })

        test(
        'Get default for syslog.',
        function() {
            var sut = mk.getController('eventHandlersCtrl')
            var result = sut.scope._getDefault('syslog')

            expect(result).toEqual({
                'enabled': true,
                'type': 'syslog',
                'url': 'udp://127.0.0.1:514',
                'usernames': [],
                'components': [],
                'target': ['123456'],
                'fail_after_errors': 10
            })
        })

        test(
        'Get default for file-dispatcher.',
        function() {
            var sut = mk.getController('eventHandlersCtrl')
            var result = sut.scope._getDefault('file-dispatcher')

            expect(result).toEqual({
                'enabled': true,
                'type': 'file-dispatcher',
                'usernames': [],
                'components': [],
                'target': ['123456'],
                'fail_after_errors': 10,
                'dispatch_attribute': 'real_path',
                'dispatch_rules': [[
                    'move', '*.txt', 'c:\\some-path', '/other/path']],
                'retry_count': 2,
                'dispatch_delay': 5,
                'fallback_rule': '',
                'create_destination_folder': '',
                'executable_timeout': 60,
            })
        })

    })

})
