/*
Tests for group edit page.
*/
describe('groupCtrl', function() {


    /*
    It goes to the start page when there are no groups, as this is a sign
    that the configuration was not yet loaded, and there is no
    point in searching for the group with the UUID from the URL.
    */
    test('init_no_groups', function() {
        var sut = mk.getController('groupCtrl')

        sut.scope.$parent.configuration.groups = {}
        sut.scope.$parent.new_entities.groups = []
        sut.route.uuid = 'some-new-group'

        sut.scope.__init__()

        expect(sut.$location.path()).toBe('/groups')
    })

    /*
    It goes to the groups list page when we have groups, but the one
    from the URL is not found.
    */
    test('init_group_does_not_exist', function() {
        var sut = mk.getController('groupCtrl')

        sut.scope.$parent.configuration.groups = {}
        sut.scope.$parent.new_entities.groups = [{
            'name': 'new-group'
        }]
        sut.route.uuid = 'unknown-uuid'

        sut.scope.__init__()

        expect(sut.$location.path()).toBe('/groups')
    })

    /*
    It loads the group page from a new group based on the name extracted
    from the URL.
    */
    test('init_group_new', function() {
        var sut = mk.getController('groupCtrl')
        var new_group = {
            'name': 'new-group'
        }
        sut.scope.$parent.configuration.groups = {}
        sut.scope.$parent.new_entities.groups = [new_group]
        sut.route.uuid = 'new-group'
        expect(sut.scope.group).toBeEmpty()

        sut.scope.__init__()

        expect(sut.scope.group).toBe(new_group)
        expect(sut.scope.section_uuid).toBe('')
        expect(sut.scope.show_disable_on_inactivity).toBeTrue()

    })

    /*
    It loads the group page from an existing group based on the UUID extracted
    from the URL.
    */
    test('init_group_existing', function() {
        var sut = mk.getController('groupCtrl')
        var group = {
            'uuid': 'some-uuid',
            'name': 'existing group',
        }
        sut.scope.$parent.configuration.groups = {'some-uuid': group}
        sut.scope.$parent.new_entities.groups = [{
            'name': 'new-group'
        }]
        sut.route.uuid = 'some-uuid'

        sut.scope.__init__()

        expect(sut.scope.group).toBe(group)
        expect(sut.scope.section_uuid).toBe('groups/some-uuid')
        expect(sut.scope.show_disable_on_inactivity).toBeTrue()
    })

    /*
    It loads the group page from the configuration of an external local
    file authentication method.
    */
    test('init_group_existing_external_file', function() {
        var sut = mk.getController('groupCtrl')
        var group = {
            'uuid': 'some-uuid',
            'name': 'existing group',
        }
        sut.scope.$parent.configuration.groups = {}
        sut.scope.$parent.configuration.authentications = {
            'file-uuid': {
                'uuid': 'file-uuid',
                'type': 'local-file',
                'accounts': {},
                'groups': {'some-uuid': group}
            }
        }

        sut.route.uuid = 'some-uuid'
        sut.scope.__init__()

        expect(sut.scope.group).toBe(group)
        expect(sut.scope.section_uuid).toBe('groups/some-uuid')
        expect(sut.scope.show_disable_on_inactivity).toBeFalse()
    })


    /*
    When on the page for a new group that is not yet saved in the config,
    it will remove it from the pending changes.
    */
    test('onDeleteGroup_new', function() {
        var sut = mk.getController('groupCtrl')

        sut.scope.$parent.configuration.groups = {
            'some-old-uuid': {'name': 'group in config'}}
        sut.scope.$parent.new_entities.groups = [{
            'name': 'some-new-group'
        }]
        sut.route.uuid = 'some-new-group'
        sut.scope.__init__()

        sut.scope.onDeleteGroup()

        // The new group is removed.
        expect(sut.scope.$parent.new_entities.groups).toHaveLength(0)
        // Existing group is left.
        expect(sut.scope.$parent.configuration.groups).toContainKey(
            'some-old-uuid')
    })

    /*
    For a group that is found in the config, it will remove it from the
    configuration object.
    */
    test('onDeleteGroup_existing', function() {
        var sut = mk.getController('groupCtrl')

        sut.scope.$parent.configuration.groups = {
            'some-old-uuid': {'name': 'old group', 'uuid': 'some-old-uuid'}}
        sut.scope.$parent.new_entities.groups = [{
            'name': 'some-new-group'
        }]
        sut.route.uuid = 'some-old-uuid'
        sut.scope.__init__()

        sut.scope.onDeleteGroup()

        // Old group is removed.
        expect(sut.scope.$parent.configuration.groups).not.toContainKey(
            'some-old-uuid')
        // The new group is kept.
        expect(sut.scope.$parent.new_entities.groups).toHaveLength(1)
    })

    /*
    For an external group that is found in the config, it will remove it from
    the configuration object.
    */
    test('onDeleteGroup_existing_and_external', function() {
        var sut = mk.getController('groupCtrl')
        // Set the default group.
        sut.scope.$parent.configuration.groups = {
            'default-group': {'name': 'default group', 'uuid': 'default-group'}
        }

        sut.scope.$parent.configuration.authentications = {
            'local-file-auth-uuid': {
                'name': 'local-file-auth',
                'type': 'local-file',
                'uuid' :'local-file-auth-uuid',
                'enabled': true,
                'groups': {
                    'ext-group-uuid': {
                        'name': 'old group',
                        'uuid': 'ext-group-uuid'
                    }
                }

            }
        }

        sut.scope.$parent.new_entities.groups = [{'name': 'some-new-group'}]
        sut.route.uuid = 'ext-group-uuid'
        sut.scope.__init__()

        sut.scope.onDeleteGroup()

        // External group is removed.
        expect(
            sut.scope.$parent.configuration
                .authentications['local-file-auth-uuid'].groups
            )
            .not.toContainKey('ext-group-uuid')
        // The new group is left.
        expect(sut.scope.$parent.new_entities.groups).toHaveLength(1)
    })

})
