/*
Tests for ssh key manager controller.
*/


describe('KeyCertCtrl', function() {

    var Flow = require('@flowjs/ng-flow/dist/ng-flow-standalone')

    /*
    Check default form values.
    */
    function assertDefaultForm(form) {
        expect(form).toEqual({
            'upload_action': 'ssh_upload',
            'ssl_new_key': 'csr',
            'sign_algorithm': 'sha256',
            'constraints': 'CA:FALSE',
            'key_usage': (
                'critical,digital-signature,' +
                'key-agreement,server-authentication'
                ),
            'generate_type': 'rsa',
            'generate_size': 3072,
            'pgp_generate_type': 'rsa',
            'pgp_generate_size': 3072,
            'pgp_generate_name': 'Autogenerated Key',
            'pgp_generate_email': 'email@example.com',
            'pgp_generate_comment': 'Created by SFTPPlus'
            })
    }

    /*
    Return a new FlowFile.
    */
    function makeFlowFile(scope, extra_options) {
        // We can not instantiate a File object, on the browser can,
        // so we fake th    e object.
        var file = {
            'name': 'some-file.txt',
            'size': 134,
            'type': 'text/txt'
        }
        if (extra_options) {
            angular.extend(file, extra_options)
        }
        return new Flow.FlowFile(scope.$flow, file)
    }

    /*
    Helper to generate a new controller for the keycert.
    */
    var getController = function() {
        var sut = mk.getController('keyCertCtrl')
        sut.scope.$flow = new Flow({})
        return sut
    }

    /*
    __init__ will setup the file uploader paramenter and reset the form.
    */
    test('__init__file_upload_config', function() {
        var sut = getController()

        sut.scope.__init__()

        expect(sut.scope.$flow.opts['target']).toEqual(
            '/rp/path/in-memory-upload')
        expect(sut.scope.$flow.opts['singleFile']).toBeTrue()
        assertDefaultForm(sut.scope.form)
        expect(sut.scope.sent_form).toBeEmpty()
        expect(sut.scope.key).toBeEmpty()
    })


    /*
    It clears any existing errors.
    */
    test('onFileAdded_clear_errors', function() {
        var sut = getController()

        sut.scope.addErrorAlert('Critical failure!')

        sut.scope.onFileAdded(makeFlowFile(sut.scope))

        expect(sut.scope.alerts).toBeEmpty()
    })

    /*It accepts, small non empty files.
    */
    test('onFileAdded_small_file', function() {
        var sut = getController()

        var result = sut.scope.onFileAdded(makeFlowFile(sut.scope, {
            'size': sut.utils.constants['MAXIMUM_IN_MEMORY_RECEIVED_DATA']}))

        expect(result).toBeTrue()
    })

    /*
    It rejects, empty and big files.
    */
    test('onFileAdded_reject', function() {
        var sut = getController()

        var result = sut.scope.onFileAdded(makeFlowFile(sut.scope, {'size': 0}))

        expect(result).toBeFalse()

        result = sut.scope.onFileAdded(makeFlowFile(sut.scope, {
            'size': sut.utils.constants['MAXIMUM_IN_MEMORY_RECEIVED_DATA'] + 1
        }))

        expect(result).toBeFalse()
    })

    /*
    Show error as an alert and clear the uploader, the current loaded
    key and previous alerts.
    */
    test('onUploadError_message', function() {
        var sut = getController()

        sut.scope.addErrorAlert('previous-error')
        sut.scope.$flow.files.push(makeFlowFile(sut.scope))
        sut.scope.key = {'fingerprint_md5': 'value'}
        sut.scope.form = {'import_password': 'password'}

        sut.scope.onUploadError('file-ignored', 'error-message')

        expect(sut.scope.$flow.files).toBeEmpty()
        expect(sut.scope.key).toBeEmpty()
        assertDefaultForm(sut.scope.form)
        expect(sut.scope.alerts).toEqual([
            {'type': 'error', 'msg': 'error-message'}
            ])
    })

    /*
    Show a default message when error had no content.
    */
    test('onUploadError_default', function() {
        var sut = getController()

        sut.scope.onUploadError('file-ignored', '')

        expect(sut.scope.$flow.files).toBeEmpty()
        expect(sut.scope.alerts).toEqual([{
            'type': 'error',
            'msg': 'Server has not accepted the file. Maybe it is too big.'
            }])
    })

    /*
    When response is an error, show that error.
    */
    test('onFileSuccess_error_response', function() {
        var sut = getController()

        sut.scope.addErrorAlert('previous-error')

        sut.scope.onFileSuccess('file-ignored', 'ERROR: custom-error')

        expect(sut.scope.$flow.files).toBeEmpty()
        expect(sut.scope.alerts).toEqual([
            {'type': 'error', 'msg': 'previous-error'},
            {'type': 'error', 'msg': 'ERROR: custom-error'}
            ])
    })

    /*
    If response has no content show default error.
    */
    test('onFileSuccess_empty_response', function() {
        var sut = getController()

        sut.scope.onFileSuccess('file-ignored', '')

        expect(sut.scope.$flow.files).toBeEmpty()
        expect(sut.scope.alerts).toEqual([{
            'type': 'error',
            'msg': 'ERROR: Server has not accepted the file. ' +
                'Maybe it is too big.'
            }])
    })

    /*
    When response is not an error, trigger conversion of the newly
    uploaded file and on success, update the page content and clear
    previous alerts while not reseting the form data
    */
    test('onFileSuccess_ok', function() {
        var sut = getController()

        var new_file_uuid = 'file-uuid-ds43af'
        var converted_key = {
            'fingerprint_md5': '12:23:db'
            }
        var json_result = {
            'result': converted_key
            }
        var initial_form = {
            'import_password': 'import pass',
            'export_password': 'export pass',
            'upload_action': 'ssh_upload'
            }
        sut.scope.__init__()
        sut.scope.form = initial_form
        sut.scope.sent_form = {}
        sut.$httpBackend.expectPOST(
            '/rp/path/json',
            {
                'jsonrpc': '2.0',
                'method': 'ssh_key_export',
                'params': {
                    'uuid': new_file_uuid,
                    'import_password': 'import pass',
                    'export_password': 'export pass'
                    },
                'id': 1
            }
            ).respond(200, json_result)
        // I have no idea why the default router is triggered here.
        sut.expectDefaultRouter()

        sut.scope.$parent.alerts = [{'type': 'info', 'msg': 'Sending key...'}]

        sut.scope.onFileSuccess('file-ignored', new_file_uuid)

        // Alerts are not cleared unless we receive the JSON RCP response
        // and page content is not updated.
        expect(sut.scope.sent_form).toBeEmpty()
        expect(sut.scope.alerts).not.toBeEmpty()
        sut.$httpBackend.flush()

        expect(sut.scope.form).toEqual(initial_form)
        expect(sut.scope.sent_form).toEqual(initial_form)
        expect(sut.scope.key).toEqual(converted_key)
        expect(sut.scope.alerts).toBeEmpty()
    })

    /*
    When server fails to export the key that was just uploaded, the
    server errors is displayed as an alert and page is reset.
    */
    test('onFileSuccess_error', function() {
        var sut = getController()
        var new_file_uuid = 'file-uuid-ds43af'
        var json_result = {
            'error': {
                'code': '50023',
                'message': 'error-message'
                }
        }
        sut.scope.__init__()
        sut.scope.key = {'fingerprint_md5': 'previous value'}
        sut.$httpBackend.expectPOST(
            '/rp/path/json',
            {
                'jsonrpc': '2.0',
                'method': 'ssh_key_export',
                'params': {'uuid': new_file_uuid},
                'id': 1
            }
            ).respond(200, json_result)
        // I have no idea why the default router is triggered here.
        sut.expectDefaultRouter()
        sut.scope.$parent.alerts = [{'type': 'info', 'msg': 'Sending key...'}]

        sut.scope.onFileSuccess('file-ignored', new_file_uuid)

        sut.$httpBackend.flush()
        assertDefaultForm(sut.scope.form)
        expect(sut.scope.key).toBeEmpty()
        expect(sut.scope.alerts).toEqual([{
            'type': 'error',
            'msg': '50023: error-message'
            }])
    })


    /*
    When server fails to convert the P12 file that was just uploaded,
    the server errors is displayed as an alert and page is reset.
    */
    test('onFileSuccess_ssl_convert_pkcs12_error', function() {
        var sut = getController()
        var new_file_uuid = 'file-uuid-ds43af'
        var json_result = {
            'error': {
                'code': '50023',
                'message': 'error-message'
                }
            }
        sut.scope.__init__()
        sut.scope.form.upload_action = 'ssl_upload'
        sut.scope.form.ssl_convert_password = 'some-pass'
        sut.scope.key = {'fingerprint_md5': 'previous value'}
        sut.$httpBackend.expectPOST(
            '/rp/path/json',
            {
                'jsonrpc': '2.0',
                'method': 'ssl_convert_pkcs12',
                'params': {
                    'uuid': new_file_uuid,
                    'import_password': 'some-pass'
                },
                'id': 1
            }
            ).respond(200, json_result)
        // I have no idea why the default router is triggered here.
        sut.expectDefaultRouter()
        sut.scope.$parent.alerts = [{'type': 'info', 'msg': 'Sending key...'}]

        sut.scope.onFileSuccess('file-ignored', new_file_uuid)

        sut.$httpBackend.flush()
        assertDefaultForm(sut.scope.form)
        expect(sut.scope.key).toBeEmpty()
        expect(sut.scope.alerts).toEqual([{
            'type': 'error',
            'msg': '50023: error-message'
            }])
    })

    /*
    Clear all alerts but keeps the form data,
    while push an in process alert and calls JSON-RPC
    to generate the key.
    */
    test('onSSHKeyGenerate_ok', function() {
        var sut = getController()
        var generated_key = {
            'fingerprint_md5': '12:23:db'
            }
        var json_result = {
            'result': generated_key
            }
        sut.scope.addErrorAlert('previous-error')
        sut.scope.key = {'fingerprint_md5': '12:34'}
        var submit_form = {
            'generate_type': 'dsa',
            'generate_size': 42,
            'generate_password': 'blue',
            'upload_action': 'ssh_upload'
            }
        sut.scope.form = submit_form
        sut.$httpBackend.expectPOST(
            '/rp/path/json',
            {
                'jsonrpc': '2.0',
                'method': 'ssh_key_generate',
                'params': {
                    'type': 'dsa',
                    'size': 42,
                    'password': 'blue'
                    },
                'id': 1
            }
            ).respond(200, json_result)
        // I have no idea why the default router is triggered here.
        sut.expectDefaultRouter()

        sut.scope.onSSHKeyGenerate()

        sut.scope.$parent.alerts = [{
            'type': 'info',
            'msg': 'Generating key (can take up to 5 minutes)...'
            }]
        // Key is empty, until the new key is returned by the server.
        expect(sut.scope.key).toBeEmpty()

        sut.$httpBackend.flush()
        expect(sut.scope.alerts).toBeEmpty()
        expect(sut.scope.key).toEqual(generated_key)
        expect(sut.scope.form).toEqual(submit_form)
    })

    /*
    On JSON-RPC errors generating the key clears alerts and push a new
    error alert.
    */
    test('onSSHKeyGenerate_alert_renew', function() {
        var sut = getController()
        sut.scope.form = {
            'generate_type': 'dsa',
            'generate_size': 42,
            'generate_password': ''
            }
        var json_result = {
            'error': {
                'code': '100',
                'message': 'error-message'
                }
            }
        sut.$httpBackend.expectPOST(
            '/rp/path/json',
            {
                'jsonrpc': '2.0',
                'method': 'ssh_key_generate',
                'params': {
                    'type': 'dsa',
                    'size': 42,
                    'password': ''
                    },
                'id': 1
            }
            ).respond(200, json_result)
        // I have no idea why the default router is triggered here.
        sut.expectDefaultRouter()

        sut.scope.onSSHKeyGenerate()
        sut.$httpBackend.flush()

        expect(sut.scope.alerts).toEqual([{
            'type': 'error',
            'msg': '100: error-message'
            }])
        expect(sut.scope.key).toBeEmpty()
        assertDefaultForm(sut.scope.form)
    })


    /*
    Does nothing when form in invalid.
    */
    test('onSSLGenerate invalid_form', function() {
        var sut = getController()
        sut.scope.onSSLGenerate({'$invalid': true})
    })

    /*
    When failing to generate the CSR will show the error received from
    the server while keeping input form fields
    */
    test('onSSLGenerate_fail', function() {
        var sut = getController()
        sut.scope.form = {
            'common_name': 'Common Name',
            'generate_size': 1024
        }
        sut.scope.addErrorAlert('previous-error')
        var json_result = {
            'error': {
                'code': '100',
                'message': 'error-message'
                }
            }
        sut.$httpBackend.expectPOST(
            '/rp/path/json',
            {
                'jsonrpc': '2.0',
                'method': 'generate_csr',
                'params': {
                    'size': 1024,
                    'common_name': 'Common Name'
                    },
                'id': 1
            }
            ).respond(200, json_result)
        // I have no idea why the default router is triggered here.
        sut.expectDefaultRouter()

        sut.scope.onSSLGenerate({'$invalid': false})

        // Info message is shown while waiting for server answer.
        expect(sut.scope.alerts).toEqual([{
            'type': 'info',
            'msg': (
                'Generating SSL key and CSR (can take up to 5 minutes)...')
            }])

        sut.$httpBackend.flush()
        expect(sut.scope.alerts).toEqual([{
            'type': 'error',
            'msg': '100: error-message'
            }])
        expect(sut.scope.form['common_name']).toEqual('Common Name')
        expect(sut.scope.form['generate_size']).toEqual(1024)
    })

    /*
    When CSR is successfully generated it is show on the page and
    previous errors are cleared together with form content
    */
    test('onSSLGenerate csr_ok', function() {
        var sut = getController()
        sut.scope.form = {
            'generate_size': 1024,
            'generate_password': 'pass',
            'common_name': 'Common Name',
            'alternative_name': 'Alt Name',
            'email': 'user@domain.tld',
            'organization': 'Some ORG',
            'organization_unit': 'Some Unit',
            'locality': 'London',
            'state': 'England',
            'country': 'GB'
        }
        sut.scope.addErrorAlert('previous-error')
        var result = {
            'key': 'KEY_PEM',
            'csr': 'CSR_PEM'
        }
        sut.$httpBackend.expectPOST(
            '/rp/path/json',
            {
                'jsonrpc': '2.0',
                'method': 'generate_csr',
                'params': {
                    'size': 1024,
                    'password': 'pass',
                    'common_name': 'Common Name',
                    'alternative_name': 'Alt Name',
                    'email': 'user@domain.tld',
                    'organization': 'Some ORG',
                    'organization_unit': 'Some Unit',
                    'locality': 'London',
                    'state': 'England',
                    'country': 'GB'
                    },
                'id': 1
            }
            ).respond(200, {'result': result})
        // I have no idea why the default router is triggered here.
        sut.expectDefaultRouter()

        sut.scope.onSSLGenerate({'$invalid': false})
        sut.$httpBackend.flush()

        expect(sut.scope.alerts).toEqual([])
        assertDefaultForm(sut.scope.form)
        expect(sut.scope.ssl).toEqual(result)
    })


    /*
    Check the UI model update when a RSA SSH key is selected to be generated.
    */
    test('onSSHGenerateTypeChange_rsa',function() {
        var sut = getController()
        // Set some initial values to check they are updated.
        sut.scope.form = {
            'generate_type': 'rsa',  // used as input for UI update.
            'generate_size': '42'
            }

        sut.scope.onSSHGenerateTypeChange()

        expect(sut.scope.form.generate_size).toEqual(3072)
        expect(sut.scope.ssh_key_sizes).toEqual([
            {'size': 1024, 'name': '1024'},
            {'size': 2048, 'name': '2048'},
            {'size': 3072, 'name': '3072'},
            {'size': 4096, 'name': '4096'},
            {'size': 8192, 'name': '8192'}
        ])
    })

    /*
    Check the UI model update when a DSA SSH key is selected to be generated.
    */
    test('onSSHGenerateTypeChange_dsa',function() {
        var sut = getController()
        // Set some initial values to check they are updated.
        sut.scope.form = {
            'generate_type': 'dsa',  // used as input for UI update.
            'generate_size': '42'
        }

        sut.scope.onSSHGenerateTypeChange()

        expect(sut.scope.form.generate_size).toEqual(3072)
        expect(sut.scope.ssh_key_sizes).toEqual([
            {'size': 1024, 'name': '1024'},
            {'size': 2048, 'name': '2048'},
            {'size': 3072, 'name': '3072'}
        ])

    })

    /*
    Check the UI model update when a ecdsa SSH key is selected to be generated.
    */
    test('onSSHGenerateTypeChange_ecdsa',function() {
        var sut = getController()
        // Set some initial values to check they are updated.
        sut.scope.form = {
            'generate_type': 'ecdsa',  // used as input for UI update.
            'generate_size': '42'
        }

        sut.scope.onSSHGenerateTypeChange()

        expect(sut.scope.form.generate_size).toEqual(256)
        expect(sut.scope.ssh_key_sizes).toEqual([
            {'size': 256, 'name': '256'},
            {'size': 384, 'name': '384'},
            {'size': 521, 'name': '521'}
        ])

    })

    /*
    Check the UI model update when a ed25519 SSH key is selected to be
    generated.
    */
    test('onSSHGenerateTypeChange_ed25519',function() {
        var sut = getController()
        // Set some initial values to check they are updated.
        sut.scope.form = {
            'generate_type': 'ed25519',  // used as input for UI update.
            'generate_size': '42'
        }

        sut.scope.onSSHGenerateTypeChange()

        expect(sut.scope.form.generate_size).toEqual(256)
        expect(sut.scope.ssh_key_sizes).toEqual([
            {'size': 256, 'name': '256'}
        ])

    })


    /*
    Check the UI model update when a DSA PGP key is selected to be generated.'
    */
    test('onPGPGenerateTypeChange_dsa',function() {
        var sut = getController()
        // Set some initial values to check they are updated.
        sut.scope.form = {
            'pgp_generate_type': 'dsa',  // used as input for UI update.
            'pgp_generate_size': '42'
        }

        sut.scope.onPGPGenerateTypeChange()

        expect(sut.scope.form.pgp_generate_size).toEqual(3072)
        expect(sut.scope.pgp_key_sizes).toEqual([
            {'size': 1024, 'name': '1024'},
            {'size': 2048, 'name': '2048'},
            {'size': 3072, 'name': '3072'}
        ])

    })

    /*
    Check the UI model update when a RSA PGP key is selected to be generated.'
    */
    test('onPGPGenerateTypeChange_rsa', function() {
        var sut = getController()
        // Set some initial values to check they are updated.
        sut.scope.form = {
            'pgp_generate_type': 'rsa',  // used as input for UI update.
            'pgp_generate_size': '42'
        }

        sut.scope.onPGPGenerateTypeChange()

        expect(sut.scope.form.pgp_generate_size).toEqual(3072)
        expect(sut.scope.pgp_key_sizes).toEqual([
            {'size': 1024, 'name': '1024'},
            {'size': 2048, 'name': '2048'},
            {'size': 3072, 'name': '3072'},
            {'size': 4096, 'name': '4096'},
            {'size': 8192, 'name': '8192'}
        ])
    })

    /*
    Does nothing if the clipboad opeation fails.
    */
    test('onCopyToClipboard_fail', function() {
        var sut = getController()
        sut.scope.__init__()
        sut.expectDefaultRouter()
        // Just a double check before the test action.
        expect(sut.scope.clipboard_text.openssh_public).toEqual(
            'Copy to clipboard')
        //
        sut.utils.writeClipboard = function(text) { return false}

        sut.scope.onCopyToClipboard('some-content', 'openssh_public')
        expect(sut.scope.clipboard_text.openssh_public).toEqual('Copy failed')
        // After 3 seconds the button text is still not changed
        sut.$flushPendingTasks(3000)
        expect(sut.scope.clipboard_text.openssh_public).toEqual('Copy failed')
        // After some time the error text is still present.
        sut.$flushPendingTasks(2000)
        expect(sut.scope.clipboard_text.openssh_public).toEqual('Copy failed')

    })

    /*
    Will update the clipboard and temporary update the text of the
    targeted button as a way to confirm the sucess operation.
    */
    test('onCopyToClipboard_ok', function() {
        var sut = getController()
        sut.scope.__init__()
        sut.expectDefaultRouter()
        // Just a double check before the test action.
        expect(sut.scope.clipboard_text.openssh_public).toEqual(
            'Copy to clipboard')

        sut.scope.onCopyToClipboard('some-content', 'openssh_public')
        expect(sut.scope.clipboard_text.openssh_public).toEqual('Copied')
        // After 3 seconds the button text is still not changed
        sut.$flushPendingTasks(3000)
        expect(sut.scope.clipboard_text.openssh_public).toEqual('Copied')
        // After some time the text is reset.
        sut.$flushPendingTasks(2000)
        expect(sut.scope.clipboard_text.openssh_public).toEqual(
            'Copy to clipboard')

    })



})
