/*
Tests for locations controller.
*/


describe('LocationCtrl', function() {


    /*
    __init__ goes to locations if could not find the requested location.
    */
    test('__init__uuid_not_found', function() {
        var sut = mk.getController('locationCtrl')
        sut.scope.$parent.new_entities['locations'] = [{'name': 'Another'}]
        sut.scope.$parent.configuration = {'locations': []}
        sut.route.uuid = 'some-name'

        sut.scope.__init__()

        expect(sut.$location.path()).toBe('/locations')
    })

    /*
    __init__ loads newly created locations based their name from the URL.
    */
    test('__init__new_location_pending_save', function() {
        var sut = mk.getController('locationCtrl')
        var new_location = {
            'name': 'some-name',
            'type': 'filesystem',
            }
        sut.scope.$parent.configuration = {'locations': []}
        sut.scope.$parent.new_entities['locations'] = [new_location]
        sut.route.uuid = 'some-name'

        sut.scope.__init__()

        expect(sut.scope.location).toBe(new_location)
        expect(sut.scope.ssl_configuration).toEqual(new_location)
        // Newly created have no runnable.
        expect(sut.scope.runnable).toBeFalse()
        expect(sut.scope.template_path).toEqual(
            'partials/location_filesystem.html?v=%(cache)s')
    })

    /*
    __init__ AS2 location is loaded with SSL config.
    */
    test('__init__as2_existing', function() {
        var sut = mk.getController('locationCtrl')
        var config = {
            'uuid': 'some-uuid',
            'name': 'some-name',
            'type': 'as2',
            }
        var runnable = {
            'uuid': 'some-uuid',
        }
        sut.scope.$parent.configuration = {
            'locations': {'some-uuid': config}}
        sut.scope.$parent.process.runnables.locations = {'some-uuid': runnable}
        sut.route.uuid = 'some-uuid'

        sut.scope.__init__()

        expect(sut.scope.location).toBe(config)
        expect(sut.scope.ssl_configuration).toEqual(config)
        expect(sut.scope.runnable).toBe(runnable)
        expect(sut.scope.template_path).toEqual(
            'partials/location_as2.html?v=%(cache)s')
        expect(sut.scope.ssl_no_set_on_first_connection).toBeFalse()
    })

    /*
    FIXME:5459:
    set-on-first-connection is not supported for FTPSE Explicit location.
    */
    test('__init__ftpse', function() {
        var sut = mk.getController('locationCtrl')
        var config = {
            'uuid': 'some-uuid',
            'name': 'some-name',
            'type': 'ftpse',
            }
        var runnable = {
            'uuid': 'some-uuid',
        }
        sut.scope.$parent.configuration = {
            'locations': {'some-uuid': config}}
        sut.scope.$parent.process.runnables.locations = {'some-uuid': runnable}
        sut.route.uuid = 'some-uuid'

        sut.scope.__init__()

        expect(sut.scope.location).toBe(config)
        expect(sut.scope.ssl_configuration).toEqual(config)
        expect(sut.scope.runnable).toBe(runnable)
        expect(sut.scope.template_path).toEqual(
            'partials/location_ftpse.html?v=%(cache)s')
        expect(sut.scope.ssl_no_set_on_first_connection).toBeTrue()
    })

    /*
    It can delete a location that was just created and not yet saved.
    */
    test('onDeleteLocation_new', function() {
        var sut = mk.getController('locationCtrl')

        sut.scope.$parent.configuration.locations = {
            'some-old-uuid': {'name': 'location in config'}}
        sut.scope.$parent.new_entities.locations = [{
            'name': 'some-new-location'
        }]
        sut.route.uuid = 'some-new-location'

        sut.scope.__init__()

        sut.scope.onDeleteLocation()

        // The new location is removed.
        expect(sut.scope.$parent.new_entities.locations).toHaveLength(0)
        // Existing location is left.
        expect(sut.scope.$parent.configuration.locations).toContainKey(
            'some-old-uuid')
    })

    /*
    For a location that is found in the config, it will remove it from the
    configuration object.
    */
    test('onDeleteLocation_existing', function() {
        var sut = mk.getController('locationCtrl')

        sut.scope.$parent.configuration.locations = {
            'some-old-uuid': {'name': 'old location', 'uuid': 'some-old-uuid'}}
        sut.scope.$parent.process.runnables.locations = {
            'some-old-uuid': {
                'name': 'old location',
                'uuid': 'some-old-uuid',
                'state': 'started',
            }}
        sut.scope.$parent.new_entities.locations = [{
            'name': 'some-new-location'
        }]
        // Location UUID is extracted from the URL route info.
        sut.route.uuid = 'some-old-uuid'

        sut.scope.__init__()

        sut.scope.onDeleteLocation()

        // The new location is kept.
        expect(sut.scope.$parent.new_entities.locations).toHaveLength(1)
        // Existing location is removed.
        expect(sut.scope.$parent.configuration.locations).not.toContainKey(
            'some-old-uuid')
    })


})
