(function(exports){

exports.locationsCtrl = function($scope, utils, componentTypeFilter, $timeout) {

    var init = require('../../../js_commons/__init__.js', 'init')
    // The current list of filtered and sorted locations
    // Will be initialized later.
    $scope.filtered_locations = []
    // Used by the search input tag.
    $scope.filter_input = ''

    // All available service types that can be added.
    $scope.location_types = utils.getFamilyTypes(
        utils.constants.TYPE_NAME.LOCATION)

    $scope.$parent.component = {
        'name': 'Locations',
        'summary': 'Represent connection details to remote servers.',
    }

    var details_getter = {}
    var defaults_getter = {}

    var category = $scope.$parent.component_categories.locations

    /*
    This is called each time the page is accessed,
    including when using the browser back/forward.
    */
    $scope.__init__ = function() {
        // Trigger an initial filter that will show all.
        $scope.onFilterLocations()

        // The configuration might not be loaded so we force a dumb reload
        // We should find a better way to auto-reload the page but for
        // now it should be ok.
        // In an ideal case, we should use a watch on the configuration,
        // but that might have some side-effects and performance issues.
        // Most of the times, users should not land directly on the locations
        // page, so this hack should be ok.
        if (init.isEmpty($scope.filtered_locations)) {
            $timeout(function() {$scope.onFilterLocations()}, 3000)
        }
    }

    /*
    Called the update the list of locations to be displayed on the page.
    */
    $scope.onFilterLocations = function(filter_value) {
        var view_model
        $scope.filtered_locations = []

        if (!$scope.configuration.locations) {
            // Configuration not yet loaded.
            return
        }

        angular.forEach(
            getAllLocations(),
            function(location) {
                if (utils.isComponentExcluded(filter_value, location)) {
                    return
                }
                view_model = getLocationView(
                    location, $scope.process.runnables.locations)
                $scope.filtered_locations.push(view_model)
            })
    }

    /*
    Return a list with configuration for all the locations.
    */
    var getAllLocations = function() {
        return $scope.getEntities('locations')
    }

    /*
    Update the `locations` based on the current filter.
    */
    var getLocationView = function(configuration, runnables) {
        var view_model = utils.getBaseViewModel(configuration, runnables)
        return {
            'uuid': view_model.uuid,
            'type_description': componentTypeFilter(
                configuration.type, 'location'),
            'details': getLocationDetails(configuration),
            'connection_details': getLocationConnectionDetails(configuration),
            'is_delete_blocked': utils.isDeleteBlocked(
                configuration, 'location'),
            'config': configuration,
            'runnable': view_model.runnable,
            'just_created': view_model.just_created
        }
    }

    /*
    Return the details to be visible in the UI for each location type.

    Return empty list no extra details are required for the UI.
    */
    var getLocationDetails = function(configuration) {
        var get_details = details_getter[configuration.type]
        if (!get_details) {
            // Fallback to nothing.
            return []
        }
        return get_details(configuration)
    }

    details_getter.as2 = function(configuration) {
        return [
            {'key': 'URL',
            'value': configuration.url
            },
            {'key': 'partner id',
            'value': configuration.as2_partner_identifier
            },
            {'key': 'Security',
            'value': configuration.as2_send_security
            },
            {'key': 'MDN',
            'value': configuration. as2_mdn_receipt
            }
        ]
    }

    details_getter['azure-blob'] = function(configuration) {
        return [
            {'key': 'Storage account',
            'value': configuration.username
            }
        ]
    }

    details_getter['azure-file'] = function(configuration) {
        return [
            {'key': 'Storage account',
            'value': configuration.username
            }
        ]
    }

    details_getter.filesystem = function() {
        return [
            {'key': '',
            'value': (
                'Access to filesystems provided by the operating system.'
                )
            }
        ]
    }

    details_getter.ftp = function(configuration) {
        return getAddressAndUsernameDetails(configuration)
    }

    details_getter.ftpse = function(configuration) {
        return getAddressAndUsernameDetails(configuration)
    }

    details_getter.ftpsi = function(configuration) {
        return getAddressAndUsernameDetails(configuration)
    }

    details_getter.sftp = function(configuration) {
        var result = getAddressAndUsernameDetails(configuration)
        result.push({
            'key': 'Server identities',
            'value': configuration.ssh_server_identity.join(', ')
        })
        result.push({
            'key': 'Proxy',
            'value': configuration.proxy
        })
        return result
    }

    details_getter.smb = function(configuration) {
        var result = getAddressAndUsernameDetails(configuration)
        result.push({
            'key': 'Require encryption',
            'value': configuration.require_encryption ? 'Yes' : 'No'
        })
        return result
    }

    details_getter.webdavs = function(configuration) {
        return [
            {'key': 'url',
            'value': configuration.url
            },
            {'key': 'Username',
            'value': configuration.username
            },
        ]
    }

    var getAddressAndUsernameDetails = function(configuration) {
        return [
            {'key': 'address',
            'value': configuration.address +':'+ configuration.port
            },
            {'key': 'username',
            'value': configuration.username
            },
        ]
    }

    /*
    Return the connection details to be visible in the UI for each
    location type.
    */
    var getLocationConnectionDetails = function(configuration) {
        if (configuration.type == 'filesystem') {
            return []
        }

        return [
            {'key': 'idle',
            'value': configuration.idle_connection_timeout + 's'
            },
            {'key': 'keep-alive',
            'value': configuration.idle_connection_keepalive_interval + 's'
            },
            {'key': 'retry',
            'value': configuration.connection_retry_count + ' times'
            },
            {'key': 'wait',
            'value': configuration.connection_retry_interval + 's'
            },
            ]
    }


    /*
    Return default configuration for a location.

    This is on the scope to simplify testing.
    */
    $scope._getDefault = function(type) {
        var get_defaults = defaults_getter[type]
        var result = {
            'type': type
        }
        if (get_defaults) {
            angular.extend(result, get_defaults())
        }
        angular.extend(result, utils.constants.LOCATION_CONNECTION_DEFAULTS)
        return result
    }

    defaults_getter.sftp = function() {
        var result = {
            'address': 'sftp.example.com',
            'port': 22,
            'proxy': '',
            'ssh_server_identity': ['REMOTE-SERVER-IDENTITY-NOT-SET'],
            'ssh_cipher_list': ['secure']
            }
        return result
    }

    defaults_getter.ftp = function() {
        var result = {
            'address': 'ftp.example.com',
            'port': 21,
            'ignore_passive_address': false,
            'acct': '',
            'debug': false
            }
        return result
    }

    defaults_getter.ftpse = function() {
        var result = {
            'address': 'ftps.example.com',
            'port': 21,
            'ignore_passive_address': false,
            'acct': '',
            'ftps_reuse_session': true,
            'debug': false
            }
        angular.extend(result, utils.constants.SSL_DEFAULTS)
        delete result['ssl_key_password']
        return result
    }

    defaults_getter.ftpsi = function() {
        var result = {
            'address': 'ftps.example.com',
            'port': 990,
            'ignore_passive_address': false,
            'acct': '',
            'ftps_reuse_session': true,
            'debug': false
            }
        angular.extend(result, utils.constants.SSL_DEFAULTS)
        delete result['ssl_key_password']
        return result
    }

    defaults_getter.webdavs = function() {
        var result = {
            'url': 'https://SITE-NAME.sharepoint.com',
            'authentication_method': 'sharepoint-online'
            }
        angular.extend(result, utils.constants.SSL_DEFAULTS)
        delete result['ssl_key_password']
        return result
    }

    defaults_getter.smb = function() {
        var result = {
            'address': 'server.address.example.com',
            'port': 445,
            'require_encryption': true
            }
        return result
    }

    defaults_getter.as2 = function() {
        var result = {
            'as2_own_identifier': 'Your AS2 identifier',
            'as2_partner_identifier': 'Remote AS2 partner identifier',
            'url': 'https://as2.server.org/as2-receive'
            }
        angular.extend(result, utils.constants.LOCATION_AS2_DEFAULTS)
        angular.extend(result, utils.constants.SSL_DEFAULTS)
        return result
    }

    details_getter['http-pull'] = function(configuration) {
        return [
            {'key': 'url',
            'value': configuration.url
            },
            {'key': 'Username',
            'value': configuration.username
            },
        ]
    }

    defaults_getter['http-pull'] = function() {
        var result = {
            'url': 'https://transfer.example.com/resource/path',
            'request_method': 'get',
            'response_code': '2**',
            'response_file_name': 'http-pull.file'
            }
        angular.extend(result, utils.constants.SSL_DEFAULTS)
        delete result['ssl_key_password']
        return result
    }

    defaults_getter['azure-files'] = function() {
        var result = {
            'username': 'update your Azure storage account'
            }
        return result
    }

    /*
    Called when add location button is triggered.
    */
    $scope.onAddLocation = function(location_type) {
        var location = $scope._getDefault(location_type)
        location['name'] = 'New %(type)s location %(random)s'.supplant({
            'type': location_type,
            'random': utils.getRandomMark()
            })

        category.add(location)
        category.goToConfiguration(location)
    }

    /*
    Go to configuration page.
    */
    $scope.onGoToConfiguration = function(location) {
        category.goToConfiguration(location)
    }

    /*
    Delete a location.
    */
    $scope.onDeleteLocation = function(location) {
        category.delete(location)
        // Update the table after one entry was removed.
        $scope.onFilterLocations($scope.filter_input)
    }
}

/* c8 ignore next */
}(typeof exports === 'undefined' ? this.locationsCtrl = {} : exports))
