/*
Tests for locations controller.
*/


describe('LocationsCtrl', function() {


    /*
    It can start without any locations and will schedule to reload the
    location
    */
    test('__init__', function() {
        var sut = mk.getController('locationsCtrl')
        sut.utils.constants.LOCATION_NAMES = {
            'some-kind': 'Human readable.',
            'other-kind': 'ignored'
            }
        sut.scope.__init__()

        expect(sut.scope.location_types).toEqual([
            'as2',
            'azure-file',
            'ftp',
            'ftpse',
            'ftpsi',
            'http-pull',
            'filesystem',
            'webdavs',
            'sftp',
            'smb',
        ])
        // No locations listed.
        expect(sut.scope.filtered_locations).toEqual([])

        // We add the default location.
        mk.minimalProcess(sut)

        // Still not updated after 1 second.
        sut.$flushPendingTasks(1000)
        expect(sut.scope.filtered_locations).toEqual([])

        // After a total of 3 seconds, it is loaded.
        sut.expectDefaultRouter()
        sut.$httpBackend.expectGET(/partials\/location.html\?v=.+/).respond('')

        sut.$flushPendingTasks(2000)
        expect(sut.scope.filtered_locations.length).toEqual(1)
        var location_view = sut.scope.filtered_locations[0]
        expect(location_view.details).toEqual([
            {
                'key': '',
                'value': (
                    'Access to filesystems provided by the operating system.')
            }
        ])
        expect(location_view.type_description).toEqual('Local filesystem')

        })

    /*
    onAddLocation will add a location in the cache and then go ' +
    to the page of the location.'
    */
    test('onAddLocation_sftp', function() {
        var sut = mk.getController('locationsCtrl')
        var new_locations = sut.scope.$parent.new_entities['locations']

        sut.scope.onAddLocation('sftp')

        expect(new_locations.length).toBe(1)
        expect(new_locations[0]['uuid']).toBeUndefined()
        expect(new_locations[0]['name']).toStartWith('New sftp location ')
        expect(sut.$location.path()).toStartWith(
            '/locations/New sftp location ')

        expect(new_locations[0]).toContainEntries([
            ['type', 'sftp'],
            ['address', 'sftp.example.com'],
            ['ssh_server_identity', ['REMOTE-SERVER-IDENTITY-NOT-SET']],
            ['ssh_cipher_list', ['secure']]
        ])
    })

    // Will still return a new location, even for an unknow type.
    test('onAddLocation_unknow', function() {
        var sut = mk.getController('locationsCtrl')
        var new_locations = sut.scope.$parent.new_entities['locations']

        sut.scope.onAddLocation('unknown-type')

        expect(new_locations.length).toBe(1)
        expect(new_locations[0]['uuid']).toBeUndefined()
        expect(new_locations[0]['name']).toStartWith(
            'New unknown-type location ')
        expect(sut.$location.path()).toStartWith(
            '/locations/New unknown-type location ')

        expect(new_locations[0]).toContainEntries([
            ['type', 'unknown-type'],
            ['idle_connection_keepalive_interval', 0],
            ['connection_retry_interval', 300],
            ['connection_retry_count', 12],
            ['idle_connection_timeout', 300]
        ])
    })


    /*
    It can generate a new location for HTTP pull.'
    */
    test('onAddLocation_http_pull', function() {
        var sut = mk.getController('locationsCtrl')
        var new_locations = sut.scope.$parent.new_entities['locations']

        sut.scope.onAddLocation('http-pull')

        expect(new_locations.length).toBe(1)
        expect(new_locations[0]['uuid']).toBeUndefined()
        expect(new_locations[0]['name']).toStartWith('New http-pull location ')
        expect(sut.$location.path()).toStartWith(
            '/locations/New http-pull location ')

        expect(new_locations[0]).toContainEntries([
            ['url', 'https://transfer.example.com/resource/path'],
            ['request_method', 'get'],
            ['response_code', '2**'],
            ['response_file_name', 'http-pull.file']
        ])
    })

    test(
    'Get default for FTP.',
    function() {
        var sut = mk.getController('locationsCtrl')

        var result = sut.scope._getDefault('ftp')

        expect(result.address).toEqual('ftp.example.com')
        expect(result.port).toEqual(21)

    })

    test(
    'Get default for FTPSE.',
    function() {
        var sut = mk.getController('locationsCtrl')

        var result = sut.scope._getDefault('ftpse')

        expect(result.address).toEqual('ftps.example.com')
        expect(result.port).toEqual(21)
        expect(result.ssl_cipher_list).toEqual('secure')

    })

    test(
    'Get default for FTPSI.',
    function() {
        var sut = mk.getController('locationsCtrl')

        var result = sut.scope._getDefault('ftpsi')

        expect(result.address).toEqual('ftps.example.com')
        expect(result.port).toEqual(990)
        expect(result.ssl_cipher_list).toEqual('secure')

    })

    test(
    'Get default for WebDAVs.',
    function() {
        var sut = mk.getController('locationsCtrl')

        var result = sut.scope._getDefault('webdavs')

        expect(result.url).toEqual('https://SITE-NAME.sharepoint.com')
        expect(result.ssl_cipher_list).toEqual('secure')
    })



    // onDeleteLocation will remove the newly created location.
    test('onDeleteLocation_new', function() {
        var location = {
            'uuid': undefined,
            'name': mk.string(),
            'type': 'sftp'
        }
        var sut = mk.getController('locationsCtrl')

        sut.scope.$parent.new_entities['locations'].push(location)

        sut.scope.onDeleteLocation(location)

        expect(sut.scope.$parent.new_entities['locations'].length).toBe(0)
    })

    // 'onDeleteLocation will remove an existing location from configuration.
    test('onDeleteLocation_existing', function() {
        var sut = mk.getController('locationsCtrl')
        var uuid = mk.string()
        var location = {
            'uuid': uuid,
            'name': 'existing name',
            'type': 'sftp'
            }
        sut.scope.$parent.configuration = {'locations': {}}
        sut.scope.$parent.configuration['locations'][uuid] = location

        sut.scope.onDeleteLocation(location)

        expect(sut.scope.$parent.configuration['locations']).toBeEmpty()
    })

})
