/*
* Tests for login controller.
*/

require('../login_app.js')

describe('LoginCtrl', function() {

    var scope
    var utils
    var $location
    var $httpBackend
    var $window
    var $location

    beforeEach(angular.mock.module('login'))

    beforeEach(inject(function($injector, $rootScope, $controller) {
        /*
        Prepare the scope and controller before each test.
        */
        $httpBackend = $injector.get('$httpBackend')
        $location = $injector.get('$location')
        $window = $injector.get('$window')
        $location = $injector.get('$location')
        $window.location.href = ''
        utils = mk.servicesUtils($injector)
        scope = $rootScope.$new()
        $controller('loginCtrl', {
            $scope: scope,
            $location: $location,
            utils: utils
            })
    }))

    /*
    Setup the mocked HTTP for an invalid session HTTP response.
    */
    var setupInvalidSession = function() {
      $httpBackend.expectPOST(
        '/rp/path/json', {
        'jsonrpc': '2.0',
        'method': 'is_valid_session',
        'params': {},
        'id': 1
        }
      ).respond(200, {'result': {'is_valid': false}})
    }

    // Initializes with no errors and blank values
    test('init', function() {
        expect(scope.username).toBe('')
        expect(scope.password).toBe('')
        expect(scope.show_form).toBeFalse()
        expect(scope.alerts).toEqual([])
    })


    /*
    When cookies are not enabled the error is displayed and login form
    is not show.
    */
    test('init_no_cookies', function() {
        utils.areCookiesEnabled = function() { return false }

        scope.__init__()
        scope.$digest()

        expect(scope.show_form).toBeFalse()

        expect(scope.alerts[0].msg).toBe(
          'To use this service, ' +
          'the browser must accept cookies from this site.')

    })

  /*
  At load it checks that the current session is valid.

  If not valid, it will show the login page.
  */
  test('init_invalid_session', function() {

      setupInvalidSession()

      scope.__init__()
      expect(scope.show_form).toBeFalse()

      // Trigger the HTTP response.
      $httpBackend.flush()

      // The form is visible to allow a new login.
      expect(scope.show_form).toBeTrue()
      // Remain on the same page.
      expect($window.location.href).toBe('')
      // The password is hidden on page load.
      expect(scope.show_password).toBeFalse()
      // No alerts are visible.
      expect(scope.alerts).toEqual([])
  })

  /*
  It will check the external error codes from the URL and show them as an
  alert.
  */
  test('init_external_errors', function() {
    $location.$$search = {
      'backend_error': utils.constants.LM_ERROR.CREDENTIALS_DENIED
    }
    // The existing session is not valid. That is ok for a new login.
    setupInvalidSession()

    scope.__init__()
    // Trigger the HTTP response.
    $httpBackend.flush()

    // Remain on the same page.
    expect($window.location.href).toBe('')

    expect(scope.alerts).toHaveLength(1)
    expect(scope.alerts[0]).toEqual({
      'type': 'error',
      'msg': 'Credentials denied',
    })
})

  /*
  A generic alert is displayed if the external error code is not known.
  */
  test('init_external_unknown_errors', function() {
    $location.$$search = {
      'backend_error': 'no-such-code'
    }
    // The existing session is not valid. That is ok for a new login.
    setupInvalidSession()

    scope.__init__()
    // Trigger the HTTP response.
    $httpBackend.flush()

    // Remain on the same page.
    expect($window.location.href).toBe('')

    expect(scope.alerts).toHaveLength(1)
    expect(scope.alerts[0]).toEqual({
      'type': 'error',
      'msg': 'Unknown error: no-such-code',
    })
})


  /*
  At load it checks that the current session is valid.

  If JSON-RPC returns an error, it will not show any error and remain on
  the login page.
  */
  test('init_error_session', function() {

    $httpBackend.expectPOST(
      '/rp/path/json', {
      'jsonrpc': '2.0',
      'method': 'is_valid_session',
      'params': {},
      'id': 1
      }
    ).respond(200, {'error': {'id': 123, 'message': 'details'}, result: null})

    scope.__init__()
    expect(scope.show_form).toBeFalse()

    // Trigger the HTTP response.
    $httpBackend.flush()

    // The form is visible to allow a new login.
    expect(scope.show_form).toBeTrue()
    // Remain on the same page.
    expect($window.location.href).toBe('')
})

  /*
  At load it checks the URL for error codes and shows the text message
  as an error alert.
  */
  test('init_url_error_code', function() {

    $httpBackend.expectPOST(
      '/rp/path/json', {
      'jsonrpc': '2.0',
      'method': 'is_valid_session',
      'params': {},
      'id': 1
      }
    ).respond(200, {'result': {'is_valid': false}, error: null})
    $location.search({'backend_error': 'session-not-valid'})

    scope.__init__()
    expect(scope.show_form).toBeFalse()

    // Trigger the HTTP response.
    $httpBackend.flush()

    // The form is visible to allow a new login.
    expect(scope.show_form).toBeTrue()
    expect(scope.alerts).toEqual([{
        'type': 'error',
        'msg': 'Session is no longer valid. It has probably expired.',
      }])
    // Remain on the same page.
    expect($window.location.href).toBe('')
})


  /*
  At load it checks that the current session is valid.

  If valid, it will keep the login form hidden and redire to manager page.
  */
  test('init_valid_session', function() {

    $httpBackend.expectPOST(
      '/rp/path/json', {
      'jsonrpc': '2.0',
      'method': 'is_valid_session',
      'params': {},
      'id': 1
      }
    ).respond(200, {'result': {'is_valid': true}})

    scope.__init__()
    expect(scope.show_form).toBeFalse()

    // Trigger the HTTP response.
    $httpBackend.flush()

    expect(scope.show_form).toBeFalse()
    expect($window.location.href).toBe('/rp/path/manager')
})


  // _setCriticalError hides the form and sets the error message.
  test('form_error', function() {
      var message = mk.string()

      scope._setCriticalError(message)

      expect(scope.show_form).toBeFalse()
      expect(scope.alerts[0].msg).toBe(message)
  })

  // onAlertClose removes an alert.
  test('alert_close', function() {
      scope.alerts = ['alert-1', 'alert-2', 'alert-3']

      scope.onAlertClose(1)

      expect(scope.alerts).toEqual(['alert-1', 'alert-3'])
  })

  /*
  Returns the full list of CSS classes as text.
  */
  test('getAlertClasses',
    function() {
      var result = scope.getAlertClasses('info')

      expect(result).toEqual('alert alert-info')
  })

  /*
  Flips the current state of the show_password flag.
  */
  test('onTogglePasswordVisibility', function() {
      scope.show_password = true

      scope.onTogglePasswordVisibility()

      expect(scope.show_password).toBeFalse()

      scope.onTogglePasswordVisibility()

      expect(scope.show_password).toBeTrue()

      scope.onTogglePasswordVisibility()

      expect(scope.show_password).toBeFalse()

    })

    // 'Clears all alerts, even if the form is not valid.
    test('login_submit_clear_alerts', function() {
      scope.alerts = [{'type': 'error', 'msg': mk.string()}]

      scope.onLogin(false)
      scope.$digest()

      expect(scope.alerts).toEqual([])
    })

    // Login request is not performed if form is not valid.
    test('login_submit_not_valid', function() {
      utils.jsonrpc = jest.fn()

      // The form is called with `false` argument based on general form
      // validation.
      // Markup is: ng-click="login(loginForm.$valid)"
      scope.onLogin(false)
      scope.$digest()

      expect(utils.jsonrpc.mock.calls.length).toBe(0)
    })

    // An error is displayed if the JSON RPC request could not be processed.
    test('login_fail_http_response', function() {
        var message = mk.string()
        var status = 500
        $httpBackend.expectPOST(
          '/rp/path/json', {
          'jsonrpc': '2.0',
          'method': 'login',
          'params': {'username': '', 'password': ''},
          'id': 1
          }
        ).respond(status, message)

        scope.onLogin(true)
        // Trigger the HTTP response.
        $httpBackend.flush()

        expect(scope.alerts[0]).toEqual({
          'type': 'error',
          'msg': [status, ': ', message].join('')
        })
        expect($window.location.href).toBe('')
    })

    // An error is displayed if JSON RPC response is an error.
    test('login_fail_at_jsonrpc', function() {
        var response = {
          'id': 1,
          'error': {'code': mk.number(), 'message': mk.string()}
          }
        $httpBackend.expectPOST(
          '/rp/path/json', {
          'jsonrpc': '2.0',
          'method': 'login',
          'params': {'username': '', 'password': ''},
          'id': 1
          }
        ).respond(200, response)

        scope.onLogin(true)
        $httpBackend.flush()

        expect(scope.alerts[0].msg).toBe('%(code)s: %(message)s'.supplant({
          'code': response['error']['code'],
          'message': response['error']['message']
          }))
        expect($window.location.href).toBe('')
    })

    // Session cookies are set and page redirected on successful login.
    test('login_ok', function() {
      var session_id = mk.number()
      var response = {
        id: 1,
        error: null,
        result: {session_id: session_id}
        }
      scope.username = mk.string()
      $httpBackend.expectPOST(
        '/rp/path/json', {
        'jsonrpc': '2.0',
        'method': 'login',
        'params': {'username': scope.username, 'password': ''},
        'id': 1
        }
      ).respond(response)
      scope.alerts = [{'type': 'error', 'msg': 'Some error.'}]

      scope.onLogin(true)
      $httpBackend.flush()

      expect(scope.alerts).toEqual([])
      expect($window.location.href).toBe('/rp/path/manager')
    })

})
