/*
Tests for main controller.
*/

describe('mainCtrl', function() {

    require('../app.js')
    var scope
    var utils
    var ctrl
    var $httpBackend
    var result
    var $cookies
    var $location
    var $window
    var $flushPendingTasks
    var $anchorScroll

    beforeEach(angular.mock.module('app'))
    beforeEach(inject(function($injector, $rootScope, $controller, $route) {

        // FIXME:3932:
        // See how to test without removing the default route.
        delete $route.routes[null]

        $httpBackend = $injector.get('$httpBackend')
        $cookies = $injector.get('$cookies')
        utils = mk.servicesUtils($injector)
        $window = $injector.get('$window')
        $location = $injector.get('$location')
        $flushPendingTasks = $injector.get('$flushPendingTasks')
        $anchorScroll = $injector.get('$anchorScroll')

        // Prepare the scope and controller before each test.
        scope = $rootScope.$new();
        ctrl = $controller(
          'mainCtrl', {
            $scope: scope,
            utils: utils
          })
        // Run an initial digest to 'settle' all watchers.
        scope.$digest()
        // Make sure each test starts with a clear storage.
        $window.localStorage.clear()
    }))

    /*
    Make sure we are not left with pending HTTP requests or expectation for
    which we don't have matching actual HTTP requests.
    */
    afterEach(() => {
        // All of the requests defined via the expect were made.
        $httpBackend.verifyNoOutstandingExpectation()
        // No outstanding requests that need to be flushed.
        $httpBackend.verifyNoOutstandingRequest()
    })

    /*
    Clear the expected request after an init for a valid session.
    */
    assertInitRequests = () => {
        $httpBackend.expectPOST(
          '/rp/path/json', {
            'jsonrpc': '2.0',
            'method': 'is_valid_session',
            'params': {},
            'id': 1
            }
          ).respond({result: {'is_valid': true, 'username': 'johnD-admin'}})

        $httpBackend.expectPOST(
          '/rp/path/json', {
            'jsonrpc': '2.0',
            'method': 'get_index',
            'params': {},
            'id': 1
            }
          ).respond({result: true})

        $httpBackend.flush()
    }

    /*
    Expect that a ping HTTP JSON-RPC is called by the code.
    */
    var expectPing = function() {
        $httpBackend.expectPOST('/rp/path/json', {
          'jsonrpc': '2.0',
          'method': 'get_property',
          'params': {'path': '/runnables/'},
          'id': 1
        }).respond({})
    }

    // Return a human readable error if timestamp is not available.
    test('unixTimeToISO8601_no_value',
    function() {
        expect(scope._unixTimeToISO8601()).toBe('no date yet')
    })

    // Return ISO format for timestamp.
    test('unixTimeToISO8601_timestamp', function() {
        expect(scope._unixTimeToISO8601(1)).toStartWith('1970-01-01T')
    })

    /*
    When service is started, return service start details.
    */
    test('getStateDetails_started', function() {
        var service_status   = {
            'state': utils.constants.STATE.STARTED,
            'start_time': 1
        }

        result = scope.getStateDetails(service_status)

        expect(result).toContain('years ago at 1970-01-01T')
    })

    /*
    When component is in a failed state return the stop time and the error
    details
    */
    test('getStateDetails_fail', function() {
        var service_status = {
            'state': utils.constants.STATE.FAILED_AT_RUN,
            'stop_time': 1,
            'last_error': {'id': 20000, 'data': {'message': 'test error'}}
        }

        result = scope.getStateDetails(service_status)

        expect(result).toContain('years ago at 1970-01-01T')
        expect(result).toContain(' 20000: DEBUG: test error')
    })

  test(
  'Initializes with no errors and default values',
  function() {
    expect(scope.avatar).toEqual({
      'name': 'NOT-YET-VALIDATED',
      'uuid': '',
      'roles_uuid': [],
    })
    expect(scope.alerts).toBeEmpty()
    expect(scope.changes.length).toEqual(0)
    expect(scope.new_entities['authentications'].length).toBe(0)
    expect(scope.new_entities['accounts'].length).toBe(0)
    expect(scope.new_entities['groups'].length).toBe(0)
    expect(scope.new_entities['event_handlers'].length).toBe(0)
    expect(scope.new_entities['services'].length).toBe(0)
    expect(scope.new_entities['locations'].length).toBe(0)
    expect(scope.new_entities['resources'].length).toBe(0)
    expect(scope.new_entities['transfers'].length).toBe(0)
    expect(scope.process).toEqual({'runnables': {}})
    expect(scope.length).toBeDefined()
  })

  /*
  The human readable uptime value is updated based on the process status.
  */
  test('product_uptime', function() {
      scope.process.status = {'start_time': 0}
      scope.__init__()
      assertInitRequests()
      // Check value at init.
      expect(scope.product.uptime).toEqual('no uptime yet')

      // Value is updated at runtime.
      // We just put a very old value at the beginning of Unix time.
      scope.process.status = {'start_time': 12133}
      scope.$digest()
      expect(scope.product.uptime).toContain(' years ago')
  })

  test('validateUmask', function() {
    expect(scope.validateUmask('012')).toBeTrue()
    expect(scope.validateUmask('777')).toBeTrue()
    expect(scope.validateUmask('008')).toBeFalse()
    expect(scope.validateUmask('080')).toBeFalse()
    expect(scope.validateUmask('800')).toBeFalse()
    expect(scope.validateUmask('0001')).toBeFalse()
    expect(scope.validateUmask('abc')).toBeFalse()
  })

  test(
  'validateHTTPURL',
  function() {
    expect(scope.validateHTTPURL('http://ceva')).toBeTrue()
    expect(scope.validateHTTPURL('http://')).toBeTrue()
    expect(scope.validateHTTPURL('')).toBeTrue()
    expect(scope.validateHTTPURL('https://ceva')).toBeTrue()
    expect(scope.validateHTTPURL('ceva')).toBeFalse()
    expect(scope.validateHTTPURL('http:/')).toBeFalse()
  })

  describe('getMenuItemClass', function() {

    test(
    'No class is returned for menu items that are not related to current ' +
    'location.path.',
    function() {
      $location.path('/other/path')

      result = scope.getMenuItemClass('start')

      expect(result).toBe('')
    })

    test(
    'Active class is returned for menu items that are related to current ' +
    'location.path.',
    function() {
      $location.path('/start/path');

      result = scope.getMenuItemClass('start')

      expect(result).toBe('active')
    })

  })


  describe('__init__', function() {

    /*
    Redirect to login page if session cookie is present but not valid.
    */
    test('session_not_valid', function() {
      $httpBackend.expectPOST(
        '/rp/path/json', {
          'jsonrpc': '2.0',
          'method': 'is_valid_session',
           'params': {},
          'id': 1
          }
        ).respond({result: {'is_valid': false}})

      scope.__init__()
      $httpBackend.flush()

      expect($window.location.href).toBe('/rp/path/')
    })

    /*
    Remain on same page if session cookie is valid and call initialization.

    it will not re-initialized if the page was already initialized.
    */
    test('session_valid_ok', function() {
      var main_url = '/index'
      var user = mk.string()
      $location.path(main_url)

      $httpBackend.expectPOST(
        '/rp/path/json', {
          'jsonrpc': '2.0',
          'method': 'is_valid_session',
          'params': {},
          'id': 1
          }
        ).respond({result: {
          'is_valid': true,
          'username': user,
          'uuid': 'some-uuid',
          'roles_uuid': ['role-1-uuid', 'role-2-uuid'],
        }})

      $httpBackend.expectPOST(
        '/rp/path/json', {
          'jsonrpc': '2.0',
          'method': 'get_index',
          'params': {},
          'id': 1
          }
        ).respond({result: true})

      scope.__init__()
      $httpBackend.flush()

      expect($location.path()).toBe(main_url)
      // We only check that further initialization is done by
      // checking that username was updated.
      expect(scope.avatar).toEqual({
        'name': user,
        'uuid': 'some-uuid',
        'roles_uuid': ['role-1-uuid', 'role-2-uuid'],
      })

      // Re-access the page.
      // It will only check the session validation.
      $httpBackend.expectPOST(
        '/rp/path/json', {
          'jsonrpc': '2.0',
          'method': 'is_valid_session',
          'params': {},
          'id': 1
          }
        ).respond({result: {
          'is_valid': true,
          'username': 'other-name',
          'uuid': 'other-uuid',
          'roles_uuid': ['role-2-uuid', 'role-3-uuid'],
        }})
      scope.__init__()
      $httpBackend.flush()

      // The future updates from the session are ignored,
      // as it will record the info only from the start of the session.
      expect(scope.avatar).toEqual({
        'name': user,
        'uuid': 'some-uuid',
        'roles_uuid': ['role-1-uuid', 'role-2-uuid'],
      })

    })

  })


  describe('resolveError', function() {

    beforeEach(function() {
      utils.constants.EVENTS_DB.events = {
        '123': {'message': 'Start %(replace)s end.'}
      }
    })

    test(
    'Expand the error based in id and data.',
    function() {
      var error = {
        'id': '123',
        'data': {'replace': 'replacement'}
        }

      result = scope.resolveError(error)

      expect(result).toBe('123: Start replacement end.')
    })

    test(
    'Expand the error based in code and data.',
    function() {
      var error = {
        'code': '123',
        'message': 'Message text.'
        }

      result = scope.resolveError(error)

      expect(result).toBe('123: Message text.')
    })

    // Sidebar is not condensed for a new browser that has no settings
    // in local storage.
    test('sidebar_no_local_storage', function() {
        // Do some checks to make sure browser is fresh.
        expect(
          $window.localStorage.getItem('sidebar_condensed')).toBeNull()
        expect(scope.sidebar_css).toBe('')

        scope.__init__()
        assertInitRequests()

        expect(scope.sidebar_css).toBe('')
      })

    // Sidebar will be condensed when toggled on a new browser that has no
    // settings in local storage.
    test('sidebar_toggle_on_fresh', function() {
        // Do some checks to make sure browser is fresh.
        expect(
          $window.localStorage.getItem('sidebar_condensed')).toBeNull()
        expect(scope.sidebar_css).toBe('')
        scope.__init__()
        assertInitRequests()

        scope.toggleCondensedSidebar()

        expect($window.localStorage.getItem('sidebar_condensed')).toBe('yes')
        expect(scope.sidebar_css).toBe('sidebar_condensed')
      })

    // Whe the local storage contains a stored configuratin, the UI is
    // updated on init.
    test('sidebar_no_local_storage', function() {
      // Setup the browser to have an already set configuration.
      $window.localStorage.setItem('sidebar_condensed', 'yes')
      expect(scope.sidebar_css).toBe('')

      scope.__init__()
      assertInitRequests()

      expect(scope.sidebar_css).toBe('sidebar_condensed')
    })

    // When the sidebar is in condensed state, toggling its state will result
    // in the sidebar no longer being condensed.
    test('sidebar_toggle_while_condensed', function() {
        // Do some checks to make sure browser is fresh.
        expect(
          $window.localStorage.getItem('sidebar_condensed')).toBeNull()
        expect(scope.sidebar_css).toBe('')

        scope.__init__()
        assertInitRequests()

        // Trigger a first toggle to condense it.
        scope.toggleCondensedSidebar()
        expect($window.localStorage.getItem('sidebar_condensed')).toBe('yes')
        expect(scope.sidebar_css).toBe('sidebar_condensed')

        // Toggle it again.
        scope.toggleCondensedSidebar()

        expect($window.localStorage.getItem('sidebar_condensed')).toBe('no')
        expect(scope.sidebar_css).toBe('')
      })

  })


    /*
    Return empty class if no input is provided.
    */
    test('getEntityStateBadgeClass_not_defined', function() {
      expect(scope.getEntityStateBadgeClass(null)).toBe('')
    })

    /*
    Return a placeholder if no input is provided.
    */
    test('getEntityStartupText_not_defined', function() {
      expect(scope.getEntityStartupText(null)).toBe('UNKNOWN')
    })

    /*
    Add simple text for entities that already have an UUID, ie the exists
    on the server.
    */
    test('getEntityStartupText', function() {
      expect(
        scope.getEntityStartupText({'uuid': 'sdf', 'enabled': true})
        ).toBe('Start at startup')

      expect(
        scope.getEntityStartupText({'uuid': 'sdf'})
        ).toBe('Leave stopped')
    })

    /*
    Append a prefix for entities that are just created.
    */
    test('getEntityStartupText_just_created', function() {
      expect(
        scope.getEntityStartupText({'enabled': true})
        ).toBe('New - Start at startup')

      expect(
        scope.getEntityStartupText({})
        ).toBe('New - Leave stopped')
    })

    /*
    Return empty string if no input is provided.
    */
    test('getStateDetails_not_defined', function() {
      expect(scope.getStateDetails(null)).toBe('')
    })

  /*
  Will not include the local file auth method, when managed externally.
  */
  test('createIdentityTabs_auth_local_file_external', function() {

      scope.process = {'runnables': {}}
      scope.process.runnables.authentications = {
        'some-uuid': {
          'type': utils.constants.TYPE_NAME.LOCAL_FILE,
          'started': true,
          'startup_configuration': {'external_management': true},
        },
      }
      var result = scope.createIdentityTabs('accounts')

      expect(result).toHaveLength(1)
      expect(result[0]['uuid']).toEqual('main-configuration')
  })

    describe('getAllMainApplicationAccounts', function() {

        test(
        'Return empty list when configuration is not yet received.',
        function() {
            scope.configuration = null

            result = scope.getAllMainApplicationAccounts()

            expect(result).toEqual([])
        })

        test(
        'Return empty list when the configuration was received but the ' +
        'identity configuration is not yet updated.',
        function() {
            scope.configuration = {
                'services': {},
                'status': {}
                }

            result = scope.getAllMainApplicationAccounts()

            expect(result).toEqual([])
        })

        test(
        'Return a list with all application accounts, ordered by name. ' +
        'Other account types are ignored',
        function() {
            var alpha = {
                'uuid': 'uuid3',
                'kind_name': 'application',
                'name': 'Alpha'
                }
            var bravo = {
                'uuid': 'uuid2',
                'kind_name': 'application',
                'name': 'bravo'
                }
            var charlie = {
                'uuid': 'uuid1',
                'kind_name': 'application',
                'name': 'Charlie'
                }
            scope.configuration = {
                'services': {},
                'accounts': {
                    'uuid1': charlie,
                    'uuid2': bravo,
                    'uuid3': alpha,
                    'uuid4': {
                        'uuid': 'uuid4',
                        'name': 'john',
                        'kind_name': 'os'
                    }
                }
            }

            result = scope.getAllMainApplicationAccounts()

            expect(result).toEqual([alpha, bravo, charlie])
        })

    })

    /*
    Return all groups in alphabetical order,
    including the groups that are pending to be added.
    */
    test('getAllMainGroups', function() {

      utils.constants['CONFIGURATION_DEFAULT_GROUP'] = 'DEFAULT_GROUP'
      scope.configuration = {
        'groups': {
          'uuid-1': {'uuid': 'uuid-1', 'name': 'bravo'},
          'uuid-2': {'uuid': 'uuid-2', 'name': 'echo'},
          'DEFAULT_GROUP': {'uuid': 'DEFAULT_GROUP', 'name': 'DEFAULT_GROUP'}
          }
        }

      scope.new_entities = {
        'groups': [
          {'name': 'Charlie New'},
          {'name': 'alpha New'}
          ]
        }

      result = scope.getAllMainGroups()

      expect(result).toEqual([
        {'name': 'alpha New'},
        {'uuid': 'uuid-1', 'name': 'bravo'},
        {'name': 'Charlie New'},
        {'uuid': 'DEFAULT_GROUP', 'name': 'DEFAULT_GROUP'},
        {'uuid': 'uuid-2', 'name': 'echo'}
        ])
    })


    test('getEventComponents', function() {
        scope.configuration = {
            'services': {},
            'locations': {
              'locations1-uuid': {
                  'uuid': 'locations1-uuid',
                  'name': 'test-sftp',
              },
            },
            'transfers': {
                'transfer1-uuid': {
                    'uuid': 'transfer1-uuid',
                    'name': 'test-push',
                },
            },
            'event_handlers': {
              'event1-uuid': {
                  'uuid': 'event1-uuid',
                  'name': 'handler-http',
              },
            },
            'services': {
              'service1-uuid': {
                  'uuid': 'service1-uuid',
                  'name': 'service-https',
              },
            },
            'resources': {
              'email1-uuid': {
                  'uuid': 'email1-uuid',
                  'name': 'email-client',
              },
            },
        }

        result = scope.getEventComponents()

        expect(result).toIncludeSameMembers([
          {
            'uuid': 'locations1-uuid',
            'name': 'test-sftp',
            'family': 'location',
          },
          {
            'uuid': 'transfer1-uuid',
            'name': 'test-push',
            'family': 'transfer',
          },
          {
            'uuid': 'event1-uuid',
            'name': 'handler-http',
            'family': 'handler',
          },
          {
            'uuid': 'service1-uuid',
            'name': 'service-https',
            'family': 'service',
          },
          {
            'uuid': 'email1-uuid',
            'name': 'email-client',
            'family': 'resource',
          },
        ])
    })

  describe('_findEntity', function() {

    test(
    'Return null if search list is empty.',
    function() {
      var all = []

      result = scope._findEntity(all, 'anything')

      expect(result).toEqual(null)
    })

    test(
    'Return null if search term is empty.',
    function() {
      var all = [{'name': '', 'uuid': ''}]

      result = scope._findEntity(all, '')

      expect(result).toEqual(null)
    })

    test(
    'Return the element for which the uuid match the search term.',
    function() {
      var search = mk.string()
      var target = {'name': 'something', 'uuid': search}
      var all = [{'name': search, 'uuid': 'other-uuid'}, target]

      result = scope._findEntity(all, search)

      expect(result).toEqual(target)
    })

    test(
    'When no UUI is defined, return the element for which the name match ' +
    'the search term.',
    function() {
      var search = mk.string()
      var target = {'name': search}
      var all = [{'name': 'other', 'uuid': 'other-uuid'}, target]

      result = scope._findEntity(all, search)

      expect(result).toEqual(target)
    })

  })


  describe('isServiceSpecialManager', function() {

    test(
    'Return `true` for local manager services.',
    function() {
      var value = mk.string()
      utils.constants['DEFAULT_LOCAL_MANAGER_UUID'] = value
      var service = {
        'uuid': value
      }

      result = scope.isServiceSpecialManager(service)

      expect(result).toBeTrue()
    })

    test(
    'Return `false` for all other services.',
    function() {
      var service = {
        'uuid': 'other_uuid'
      }

      result = scope.isServiceSpecialManager(service)

      expect(result).toBeFalse()
    })

    test(
    'Return `false` if service is not defined.',
    function() {
      result = scope.isServiceSpecialManager(undefined)

      expect(result).toBeFalse()
    })

  })


  describe('_getChangesForCreate', function() {

    test(
    'Return empty list when no new entities are present.',
    function() {
      result = scope._getChangesForCreate()

      expect(result).toEqual([])
    })

    test(
    'Return changes for new entities from all categories.',
    function() {
      scope.new_entities = {
        'category-1': [{'name': 'name-1'}, {'name': 'name-2'}],
        'sub/category-2': [{'name': 'name-3'}]
      }
      result = scope._getChangesForCreate()

      expect(result.length).toBe(3)
      expect(result[0]).toEqual({
        'operation' : 'create',
        'path' : 'category-1/',
        'old' : '',
        'value' : { 'name' : 'name-1' }
      })
      expect(result[1]).toEqual({
        'operation' : 'create',
        'path' : 'category-1/',
        'old' : '',
        'value' : { 'name' : 'name-2' }
      })
      expect(result[2]).toEqual({
        'operation' : 'create',
        'path' : 'sub/category-2/',
        'old' : '',
        'value' : { 'name' : 'name-3' }
      })

    })

  })

  test(
  'getEntities return all entities in alphabetical order, including ' +
  'those that are pending to be added',
  function() {
    scope.configuration.family = {
      'uuid-1': {'uuid': 'uuid-1', 'name': 'Bravo'},
      'uuid-2': {'uuid': 'uuid-2', 'name': 'echo'}
      }

    scope.new_entities = {
      'family': [
        {'name': 'charlie'},
        {'name': 'alpha'}
        ]
      }

    result = scope.getEntities('family')

    expect(result).toEqual([
      {'name': 'alpha'},
      {'uuid': 'uuid-1', 'name': 'Bravo'},
      {'name': 'charlie'},
      {'uuid': 'uuid-2', 'name': 'echo'}
      ])
  })


  test(
  'getTemplatePath return dedicated path for a specific component.',
  function() {
      result = scope.getTemplatePath('comp', 'some-kind')

      expect(result).toBe('partials/comp_some-kind.html?v=%(cache)s')
  })

  test(
  'hasInterface return true or false depending on service type.',
  function() {
      expect(scope.hasInterface()).toBeFalse()
      expect(scope.hasInterface({'type': 'monitor'})).toBeFalse()
      expect(scope.hasInterface({'name': 'something'})).toBeTrue()
      expect(scope.hasInterface({'type': 'sftp'})).toBeTrue()
  })

  describe('getStateClass', function() {

    test(
    'Return the css class name associated with a given runnable state.',
    function() {
      utils.constants = CONSTANTS

      var runnable = {
        'state': utils.constants.STATE.FAILED_AT_START
      }

      result = scope.getStateClass(runnable)

      expect(result).toEqual('label-important')
    })

    test(
    'Return empty string for unknow states.',
    function() {

      var runnable = {
        'state': 'Some random state'
      }

      result = scope.getStateClass(runnable)

      expect(result).toEqual('')
    })

  })


    test(
      'Component is added to the list with new entities',
      function() {
        scope.new_entities = {
          'resources': [
            {'name': 'test-name'}
          ]
        }

        var category = scope.component_categories['resources']
        category.add({'name': 'test-name-2'})

        expect(scope.new_entities).toEqual(
          {
            'resources': [
              {'name': 'test-name'},
              {'name': 'test-name-2'}
            ]
          }
        )
      })

    test(
      'Component without UUID is deleted from the list',
      function() {
        scope.new_entities = {
          'resources': [
            {'name': 'test-name'},
            {'name': 'test-name-2'}
          ]
        }

        var category = scope.component_categories['resources']
        category.delete({'name': 'test-name-2'})

        expect(scope.new_entities).toEqual(
          {
            'resources': [
              {'name': 'test-name'}
            ]
          }
        )
      })

    test(
      'Component with UUID is deleted from the configuration list',
      function() {
        scope.configuration = {
          'resources': {
            'test-uuid': {},
            'test-uuid-2': {}
          }
        }

        var category = scope.component_categories['resources']
        category.delete({'uuid': 'test-uuid-2'})

        expect(scope.configuration).toEqual({
          'resources': {'test-uuid': {}}
        })
      })

    test(
      'Identity component with UUID is deleted from the configuration list',
      function() {
        scope.configuration = {
            'groups': {
            'test-uuid': {},
            'test-uuid-2': {}}
          }

        var category = scope.component_categories['groups']
        category.delete({'uuid': 'test-uuid-2'})

        expect(scope.configuration).toEqual({
            'groups': {'test-uuid': {}}
        })
      })

    test(
      'Will redirect to the resource configuration page',
      function() {
        scope.new_entities = {
          'resources': [
            {'name': 'test-name'}
          ]
        }
        expectPing()
        $httpBackend.expectGET(/\/resource.html\?v=.+/).respond('')

        var category = scope.component_categories['resources']
        category.goToConfiguration({'name': 'test-name'})

        $httpBackend.flush()
        expect($location.path()).toBe('/resources/test-name')
      })

    test(
      'Will redirect to the group configuration page',
      function() {
        scope.configuration = {
            'groups': {
              'test-uuid': {}
            }
          }
        expectPing()
        $httpBackend.expectGET(/\/group.html\?v=.+/).respond('')

        var category = scope.component_categories['groups']
        category.goToConfiguration({'uuid': 'test-uuid'})

        $httpBackend.flush()
        expect($location.path()).toBe('/groups/test-uuid')
      })


    /*
    Will redirect to service configuration page while pinging to keep
    the session alive.
    */
    test('onGoToRunnable', function() {
        scope.configuration = {
          'services': [
            {'uuid': 'test-uuid'}
          ]
        }
        expectPing()
        $httpBackend.expectGET(/\/service.html\?v=.+/).respond('')

        var runnable = {'uuid': 'test-uuid', 'family': 'service'}
        scope.onGoToRunnable(runnable)

        $httpBackend.flush()
        expect($location.path()).toBe('/services/test-uuid')
      })

    /*
    Will redirect to admin configuration page while pinging to keep
    the session alive.
    */
    test('onGoToAdministrator', function() {
      scope.configuration = {
        'administrators': [
          {'uuid': 'test-uuid'}
        ]
      }
      expectPing()
      $httpBackend.expectGET(/\/administrator.html\?v=.+/).respond('')

      scope.onGoToAdministrator('test-uuid')

      $httpBackend.flush()
      expect($location.path()).toBe('/administrators/test-uuid')
    })


    /*
    Navigate to the location page when a location object is used.
    */
    test('_goToComponent_with_uuid', function() {
        var uuid = 'some-uuid'
        expectPing()
        $httpBackend.expectGET(/.+\/location.html\?v=.+/).respond('')

        scope._goToComponent({'uuid': uuid}, '/locations/')

        $httpBackend.flush()
        expect($location.path()).toBe('/locations/some-uuid')
    })

    /*
    Will use the object name to navigate to the location
    when this is a newly created object which does not yet have an ID.
    */
    test('goToComponent_no_id' , () => {
        var name = 'some-name'
        expectPing()
        $httpBackend.expectGET(/.+\/location.html\?v=.+/).respond('')

        scope._goToComponent({'name': name}, '/locations/')

        $httpBackend.flush()
        expect($location.path()).toBe('/locations/some-name')
    })


  describe('getRunnablesRequiringRestart', function() {

    test(
      'Will return all runnables which require restart.',
      function() {
        /*
         Make a runnable with the given
         `changed_configurations`. `changed_configurations` is a set
         of changed configurations which requiring the runnable to be
         restarted.
         */
        var mk_runnable = function(changed_configurations) {
          return {
            'name': mk.string(),
            'restart_required': changed_configurations
          }
        }
        var event_handler1 = mk_runnable([
          'changed-configuration-1'
        ])
        var event_handler2 = mk_runnable([
            'changed-configuration-2',
            'changed-configuration-3'
          ]
        )
        var event_handler3 = mk_runnable([])
        var event_handlers = {}
        event_handlers[event_handler1.name] = event_handler1
        event_handlers[event_handler2.name] = event_handler2
        event_handlers[event_handler3.name] = event_handler3
        scope.process.runnables.event_handlers = event_handlers
        var service1 = mk_runnable([
          'changed-configuration-1'
        ])
        var service2 = mk_runnable([
          'changed-configuration-2',
          'changed-configuration-3'
        ]);
        var service3 = mk_runnable([])
        var services = {}
        services[service1.name] = service1
        services[service2.name] = service2
        services[service3.name] = service3
        scope.process.runnables.services = services
        scope.process.without_restart_required = {
          'property': 'value'
        }

        var runnables = scope.getRunnablesRequiringRestart()

        var expected_runnables = []
        expected_runnables.push(event_handler1)
        expected_runnables.push(event_handler2)
        expected_runnables.push(service1)
        expected_runnables.push(service2)
        expect(runnables).toEqual(expected_runnables)
      })
  })

  describe('_actOnRunnableManager', function() {

    test(
      'Request method is sent and the service status is updated with ' +
        'the response.',
      function() {
        var resource_uuid = mk.string()
        scope.configuration['resources'] = {}
        scope.configuration['resources'][resource_uuid] = {
          'uuid': resource_uuid,
          'name': 'some-name'
        }
        scope.process.runnables = {'resources': {}}
        scope.process.runnables.resources[resource_uuid] = {
          'uuid': resource_uuid,
          'state': utils.constants.STATE.STARTED,
          'start_time': 1,
          'family': 'resource'
        }
        var resource = scope.process.runnables.resources[resource_uuid]
        var new_state = utils.constants.STATE.STOPPING
        var jsonrpc_method = 'fake_method'
        $httpBackend.expectPOST(
          '/rp/path/json', {
            'jsonrpc': '2.0',
            'method': jsonrpc_method,
            'params': {'uuid': resource_uuid, 'family': 'resources'},
            'id': 1
          }
        ).respond({
          'result': {'state': utils.constants.STATE.STOPPED},
          'error': null
        })

        // Will trigger an update of all runnables.
        $httpBackend.expectPOST(
          '/rp/path/json', {
            'jsonrpc': '2.0',
            'method': 'get_property',
            'params': {'path': '/runnables/'},
            'id': 1
          }
        ).respond({
          'result': {},
          'error': null
        })

        scope._actOnRunnableManager(
          resource,
          new_state,
          jsonrpc_method
        )

        // Before getting the response, the intermediary state is set.
        expect(scope.process.runnables.resources[resource_uuid]['state']
              ).toBe(utils.constants.STATE.STOPPING)

        $httpBackend.flush()
        expect(scope.process.runnables.resources[resource_uuid]['state']
              ).toBe(utils.constants.STATE.STOPPED)
      })

    test(
      'On error in RPC method, the previous state is set and an alert set.',
      function() {
        var service_uuid = mk.string()
        scope.configuration['services'] = {}
        scope.configuration['services'][service_uuid] = {
          'uuid': service_uuid,
          'name': 'some-name'
        }
        scope.services = {}
        scope.services[service_uuid] = {
          'uuid': service_uuid,
          'state': 1,
          'start_time': 1,
          'family': 'services'
        }
        var service = scope.services[service_uuid]
        var new_state = 101
        var jsonrpc_method = 'fake_method'
        $httpBackend.expectPOST(
          '/rp/path/json', {
            'jsonrpc': '2.0',
            'method': jsonrpc_method,
            'params': {'uuid': service_uuid},
            'id': 1
          }
        ).respond({
          'result': null,
          'error': {'id': '123'}
        })
        // Will trigger an update of all runnables.
        $httpBackend.expectPOST(
          '/rp/path/json', {
            'jsonrpc': '2.0',
            'method': 'get_property',
            'params': {'path': '/runnables/'},
            'id': 1
          }
        ).respond({
          'result': {},
          'error': null
        })

        utils.constants.EVENTS_DB.events = {
          '123': {'message': 'Error message.'}
        }

        scope._actOnRunnableManager(
          service, new_state, jsonrpc_method)
        $httpBackend.flush()

        expect(scope.services[service_uuid]['state']).toBe(1)
        expect(scope.alerts[0]['msg']).toBe('123: Error message.')
      })

  })

  /*
  If there is no active auto-update,
  it will trigger another round of update,
  but will auto-stop after a number of consecutive updates.
  */
  test('onStartAutoUpdate_stop_loop',
    function() {
      scope.auto_updates_left = -1

      $httpBackend.expectPOST(
        '/rp/path/json', {
          'jsonrpc': '2.0',
          'method': 'get_property',
          'params': {'path': '/runnables/'},
          'id': 1
        }
      ).respond({
        'result': {},
        'error': null
      })

      scope.onStartAutoUpdate()

      expect(scope.auto_updates_left).toBe(30)

      // As soon as a response is received, it decreases the number of
      // auto updates left.
      $httpBackend.flush()
      expect(scope.auto_updates_left).toBe(29)

      // Rig the internal counter to speed up the test.
      scope.auto_updates_left = 1

      // A new update is scheduled after 2 seconds
      $httpBackend.expectPOST(
        '/rp/path/json', {
          'jsonrpc': '2.0',
          'method': 'get_property',
          'params': {'path': '/runnables/'},
          'id': 1
        }
      ).respond({
        'result': {},
        'error': null
      })
      $flushPendingTasks(scope._AUTO_UPDATE_INTERVAL)
      $httpBackend.flush()
      // And continues to decrease the counter.
      expect(scope.auto_updates_left).toBe(0)


      $httpBackend.expectPOST(
        '/rp/path/json', {
          'jsonrpc': '2.0',
          'method': 'get_property',
          'params': {'path': '/runnables/'},
          'id': 1
        }
      ).respond({
        'result': {},
        'error': null
      })
      $flushPendingTasks(scope._AUTO_UPDATE_INTERVAL)
      $httpBackend.flush()

      // Will not decrease and there no more requests.
      expect(scope.auto_updates_left).toBe(0)
      $flushPendingTasks(scope._AUTO_UPDATE_INTERVAL)
      $httpBackend.verifyNoOutstandingRequest()
    })

  /*
  If there is an active auto-update,
  it will not trigger a new update,
  but will reset the number of auto-updates left.
  */
  test('onStartAutoUpdate_overlap', function() {
      scope.auto_updates_left = -1

      $httpBackend.expectPOST(
        '/rp/path/json', {
          'jsonrpc': '2.0',
          'method': 'get_property',
          'params': {'path': '/runnables/'},
          'id': 1
        }
      ).respond({
        'result': {},
        'error': null
      })

      scope.onStartAutoUpdate()

      expect(scope.auto_updates_left).toBe(30)

      // As soon as a response is received, it decreases the number of
      // auto updates left.
      $httpBackend.flush()
      expect(scope.auto_updates_left).toBe(29)

      // Rig the internal counter to speed up the test.
      scope.auto_updates_left = 23

      // Call the auto update again.
      scope.onStartAutoUpdate()

      // A previous update is scheduled after 2 seconds
      $httpBackend.expectPOST(
        '/rp/path/json', {
          'jsonrpc': '2.0',
          'method': 'get_property',
          'params': {'path': '/runnables/'},
          'id': 1
        }
      ).respond({
        'result': {},
        'error': null
      })
      $flushPendingTasks(scope._AUTO_UPDATE_INTERVAL)
      $httpBackend.flush()
      // And continues to decrease the counter.
      expect(scope.auto_updates_left).toBe(29)
    })

  /*
  If by the time the JSON-RPC response is receive we no longer have a
  process object, it does not decrease the counter and the auto-update
  continues.
  */
  test('onStartAutoUpdate_no_more_process', function() {
    scope.auto_updates_left = -1

    $httpBackend.expectPOST(
      '/rp/path/json', {
        'jsonrpc': '2.0',
        'method': 'get_property',
        'params': {'path': '/runnables/'},
        'id': 1
      }
    ).respond({
      'result': {},
      'error': null
    })

    scope.onStartAutoUpdate()

    expect(scope.auto_updates_left).toBe(30)

    // Clear the process.
    var orig_runnables = scope.process.runnables
    scope.process.runnables = null

    // As soon as a response is received the counter is kept.
    $httpBackend.flush()
    expect(scope.auto_updates_left).toBe(30)

    // It continues to schedule an auto-update.
    $httpBackend.expectPOST(
      '/rp/path/json', {
        'jsonrpc': '2.0',
        'method': 'get_property',
        'params': {'path': '/runnables/'},
        'id': 1
      }
    ).respond({
      'result': {},
      'error': null
    })
    // And if the process is available it will decrease the counter.
    scope.process.runnables = orig_runnables
    $flushPendingTasks(scope._AUTO_UPDATE_INTERVAL)
    $httpBackend.flush()
    // And continues to decrease the counter.
    expect(scope.auto_updates_left).toBe(29)
  })

    describe('validateSSLMethod', function() {

        test(
        'Return `true` if value is not defined.',
        function() {
            var result = scope.validateSSLMethod()

            expect(result).toBeTrue()
        })

        test(
        'Return `true` if value is in accepted SSL methods.',
        function() {
            utils.constants['SSL_SUPPORTED_METHODS'] = ['some', 'valid-ssl']
            var result = scope.validateSSLMethod('valid-ssl')

            expect(result).toBeTrue()
        })

        test(
        'Return `true` if values are in accepted SSL methods.',
        function() {
            utils.constants['SSL_SUPPORTED_METHODS'] = ['some', 'valid-ssl']
            var result = scope.validateSSLMethod('valid-ssl some')

            expect(result).toBeTrue()
        })

        test(
        'Return `false` if values contains methods that are not accepted.',
        function() {
            utils.constants['SSL_SUPPORTED_METHODS'] = ['some', 'valid-ssl']
            var result = scope.validateSSLMethod('valid-ssl some other')

            expect(result).toBeFalse()
        })

    })



    describe('_getRequestedChanges', function() {

        test(
        'Keeps only changes attributes that need to be sent to the server ' +
        'and get an id for each change',
        function() {
            var key_value = mk.string()
            var source = {'key-name': key_value}
            var changes = [
                {
                    'operation': 'delete',
                    'path': 'some-path',
                    'other-key': 'other-value'
                },
                {
                    'operation': 'update',
                    'path': 'some-path',
                    'value': 'some-value',
                    'other-key': 'other-value'
                }
                ]

            result = scope._getRequestedChanges(changes)

            expect(result).toEqual({
                '0': {
                    'operation': 'delete',
                    'path': 'some-path',
                    'value': undefined
                },
                '1':
                {
                    'operation': 'update',
                    'path': 'some-path',
                    'value': 'some-value'
                }
                })
        })

    })


    describe('_cbApplyUpstreamChanges', function() {

        test(
        'On critical error, it just shows an error message and keeps the ' +
        'previous changes.',
        function() {
            scope.new_entities['services'] = [
                {
                    'operation': 'create',
                    'path': 'services/',
                    'value': {'name': 'service1'}
                }
                ]
            // Create new changes.
            scope._updateChanges()
            expect(scope.changes.length).toEqual(1)
            var change = scope.changes[0]
            var requested_changes = scope._getRequestedChanges([change])
            var data = {
                'error': {'code': 123, 'message': 'some-error'},
                'result': null
            }
            scope.alerts.push({type: 'info', msg: 'previous' })

            scope._cbApplyUpstreamChanges(data, null, requested_changes)
            scope.$digest()

            // Changes are kept unchanged.
            expect(scope.changes).toEqual([change])
            // An error message is listed.
            expect(scope.alerts).toEqual(
                [{type: 'error', msg: '123 : some-error' }])

        })

        test(
        'Apply a single change from 2 new entities from same type will ' +
        'keep the change for the other new entity.',
        function() {
            scope.new_entities['services'] = [
                {
                    'operation': 'create',
                    'path': 'services/',
                    'value': {'name': 'service1'}
                },
                {
                    'operation': 'create',
                    'path': 'services/',
                    'value': {'name': 'service2'}
                }
                ]
            // Create new changes.
            scope._updateChanges()
            expect(scope.changes.length).toEqual(2)
            var changes = scope.changes
            var change_1 = changes[0]
            var change_2 = changes[1]
            var requested_changes = scope._getRequestedChanges([change_1])
            var data = {
                'error': null,
                'result': {
                    'configuration': {},
                    'services': {},
                    'status': {},
                    'results': {'0': {'error': null}}
                }
            }

            scope._cbApplyUpstreamChanges(data, null, requested_changes)
            scope.$digest()

            // Only second change remains.
            expect(scope.changes).toEqual([change_2])
        })


        test(
        'Apply a single change from 2 removed entities from same type will ' +
        'keep the change for the other removed entity.',
        function() {
            scope.upstream_configuration['services'] = {
                'uuid1': {'name': 'service1'},
                'uuid2': {'name': 'service2'}
                }
            scope.configuration['services'] = {}
            // Create new changes.
            scope._updateChanges()
            expect(scope.changes.length).toEqual(2)
            var changes = scope.changes
            var change_1 = changes[0]
            var change_2 = changes[1]
            var requested_changes = scope._getRequestedChanges([change_1])
            var data = {
                'error': null,
                'result': {
                    'configuration': {
                        'services': {
                            'uuid2': {
                                'name': 'service2'
                            }
                        }
                    },
                    'services': {},
                    'status': {},
                    'results': {'0': {'error': null}}
                }
            }

            scope._cbApplyUpstreamChanges(data, null, requested_changes)
            scope.$digest()

            // Only second change remains.
            expect(scope.changes).toEqual([change_2])
        })

    })


    describe('_restoreChangeRequest', function() {

        test(
        'Revert update changes.',
        function() {
            var new_value = mk.string()
            var old_value = mk.string()
            var change = {
                'operation': 'update',
                'path': 'section/subsection/key',
                'value': new_value
                }
            scope.configuration = {
                'section' : {
                    'subsection': {'key': old_value}
                    }
                }

            result = scope._restoreChangeRequest(change)

            expect(
                scope.configuration['section']['subsection']['key']
                ).toBe(new_value)
        })

        test(
        'Revert delete changes.',
        function() {
            var change = {
                'operation': 'delete',
                'path': 'section/key-name',
                'value': {'attribute': 'some-value'}
                }
            scope.configuration = {
                'section' : {'key-name': {'attribute': 'some-value'}}
                }

            result = scope._restoreChangeRequest(change)

            expect(
                scope.configuration['section']
                ).not.toContainKey('key-name')
        })

        test(
        'Revert create changes.',
        function() {
            var new_value = mk.string()
            var change = {
                'operation': 'create',
                'path': 'section/key',
                'value': new_value
                }
            scope.new_entities = {'section' : []}

            result = scope._restoreChangeRequest(change)

            expect(
                scope.new_entities['section'][0]).toBe(new_value)
        })

    })


    describe('_getNewEntitiesSource', function() {

        test(
        'Get the array for new services.',
        function() {
            var new_services = [{'name': 'new-service'}]
            scope.new_entities = {'services': new_services}

            result = scope._getNewEntitiesSource('services/new-service')

            expect(result).toBe(new_services)
        })

        test(
        'Get the array for new groups or accounts.',
        function() {
            var path
            var new_accounts = [{'name': 'new-account'}]
            var new_administrator = [{'name': 'new-administrator'}]
            var new_groups = [{'name': 'new-group'}]
            scope.new_entities = {
                'accounts': new_accounts,
                'administrators': new_administrator,
                'groups': new_groups
                }

            result = scope._getNewEntitiesSource(
                'accounts/new-account')

            expect(result).toBe(new_accounts)

            result = scope._getNewEntitiesSource(
                'administrators/new-administrator')

            expect(result).toBe(new_administrator)

            result = scope._getNewEntitiesSource(
                'groups/new-group')

            expect(result).toBe(new_groups)
        })

    })


    /*
    Return a `tuple` with the parent and key for configuration at path.
    */
    test('_getConfigurationSource', function() {
        var key_value = mk.string()
        var source = {'key-name': key_value}
        scope.configuration = {'account': source}

        result = scope._getConfigurationSource('account/key-name')

        expect(result[0]).toBe(source)
        expect(result[1]).toBe('key-name')
    })


    /*
    Does nothing when the new value is empty.
    This is just some error handling when angular passes
    */
    test('onConfigurationChange_empty_change', function() {
        scope.product.instance = 'some existing name'

        result = scope._onConfigurationChange({}, 'ignored')

        expect(scope.product.instance).toBe('some existing name')
    })

    /*
    Will update the instance name based on the latest configuration.
    */
    test('_onConfigurationChange_instance_name', function() {
        var new_value =  {
            'server': {'name': 'new-name'}
        }

        scope._onConfigurationChange(new_value, 'old-value-not-used')

        expect(scope.product.instance).toBe('new-name')
        expect(scope.documentation_target).toBe('_blank')
    })

    /*
    Will update the instance name based on the latest configuration.
    */
    test('_onConfigurationChange_selenium', function() {
      var new_value =  {
          'server': {'name': 'chevah-selenium-testing'}
      }
      expect(scope.documentation_target).toBe('_blank')

      scope._onConfigurationChange(new_value, 'old-value-not-used')

      expect(scope.documentation_target).toBe('')
    })


    /*
    Will change the page hash sub-page navigation to the requested anchor
    and will update the current page DOM
    */
    test('goToSection_new_section', function() {
      // The code works wit any TAG type, since it looks for data attributes.
      var nav = document.createElement('div')
      nav.setAttribute('data-role', 'page-nav')

      var item1 = document.createElement('div')
      item1.setAttribute('data-tab', 'item_1')
      item1.classList.add('active')
      item1.classList.add('other')

      var item2 = document.createElement('div')
      // There is a bug in $anchorScroll where it will create a global
      // variable for the target element.
      // We use the `leak_ignore` to let our testing framework know
      // that is ok to ignore it.
      item2.setAttribute('data-tab', 'leak_ignore')
      item2.classList.add('another')

      nav.appendChild(item1)
      nav.appendChild(item2)

      var target = document.createElement('target')
      target.setAttribute('id', 'leak_ignore')

      document.body.appendChild(nav)
      document.body.appendChild(target)
      $anchorScroll.yOffset = 0

      // The first element is active, and we navigate to the second.
      scope.goToSection('leak_ignore')

      // The CSS is updated.
      expect(item1.classList + '').toEqual('other')
      expect(item2.classList + '').toEqual('another active')

      // The scroll is updated.
      expect($anchorScroll.yOffset).toEqual(150)
      // And document is scrolled to the target
      expect(global._scroll_into_view).toBe(target)

    })

})
