/*
Tests for ActivityLogCtrl controller.
*/


describe('ActivityLogCtrl', function() {

    var parent_scope
    var scope
    var utils
    var result
    var $httpBackend

    var dayjs = require('dayjs')


    /*
    Return a result similar to the one returned by the JSON-RPC API.
    */
    getJSONResult = function(total) {
        var data = []
        for (var i=0; i < total; i++) {
            data.push({'id': i, 'data': '{}'})
        }
        var result = {
            'result': {'data': data}
        }
        return result
    }

    /*
    Make sure page is loaded with current changes.
    */
    var settlePage = function() {
        $httpBackend.expectPOST('/rp/path/json').respond(200, getJSONResult(34))
        scope.$digest()
        $httpBackend.flush()
    }

    require('../app.js')
    beforeEach(angular.mock.module('app'))
    beforeEach(inject(function($injector, $rootScope, $controller) {
        /*
        Prepare the scope and controller before each test.
        */
        utils = mk.servicesUtils($injector)
        $httpBackend = $injector.get('$httpBackend')
        parent_scope = $rootScope.$new();
        $controller(
            'mainCtrl', {
            '$scope': parent_scope,
            'utils': utils,

        })

        scope = parent_scope.$new();
        $controller(
            'activityLogCtrl', {
            $scope: scope,
            utils: utils,
            'dayjs': mk.getDayJS(),
            })
        scope.$digest()
    }))

    /*
    Will select the first DB handler.
    */
    test('init_multiple_db_handlers', function() {

        scope.table_options.db_uuid = 'something'

        scope.configuration.event_handlers = {
            'first-uuid': {'uuid': 'first-uuid', 'type': 'local-file'},
            'first-db-uuid': {'uuid': 'first-db-uuid', 'type': 'database'},
            'other-uuid': {'uuid': 'other-uuid', 'type': 'database'},
        }
        scope.__init__()

        expect(scope.table_options.db_uuid).toEqual('first-db-uuid')
    })


    /*
    When date start is updated and there is no time set,
    the time is set to the start of the day.
    */
    test('date_start_ui_watch_no_time', function() {
        const new_date = new Date('2002-01-16T13:17:19.000Z')
        expect(scope.date_start_ui.time_value).toBeNull()

        scope.date_start_ui.date_value = new_date
        scope.$digest()

        expect(scope.date_start_ui.time_value).toEqual(new Date(
            '2002-01-16T00:00:00.000Z'))
    })

    /*
    When date start is updated and there is already a time set,
    the time is left as it is.
    */
    test('date_start_ui_watch_existing_time', function() {
        const new_date = new Date('2002-01-16T13:17:19.000Z')
        const existing_time = new Date('2002-01-17T23:07:29.000Z')
        scope.date_start_ui.time_value = existing_time

        scope.date_start_ui.date_value = new_date
        scope.$digest()

        expect(scope.date_start_ui.time_value).toEqual(existing_time)
    })

    /*
    When date end is updated and there is no time set for end range,
    the time is set to the end of the day.
    */
    test('date_end_ui_watch_no_time', function() {
        const new_date = new Date('2002-01-16T13:17:19.000Z')
        expect(scope.date_end_ui.time_value).toBeNull()

        scope.date_end_ui.date_value = new_date
        scope.$digest()

        expect(scope.date_end_ui.time_value).toEqual(new Date(
            '2002-01-16T23:59:59.000Z'))
    })

    /*
    When date end is updated and there is already a time set for end range,
    the time is left as it is.
    */
    test('date_end_ui_watch_existing_time', function() {
        const new_date = new Date('2002-01-16T13:17:19.000Z')
        const existing_time = new Date('2002-01-15T12:19:42.000Z')
        scope.date_end_ui.time_value = existing_time

        scope.date_end_ui.date_value = new_date
        scope.$digest()

        expect(scope.date_end_ui.time_value).toEqual(existing_time)
    })


    /*
    Return an URL with the current selected filters and options, in base64
    encoding.
    */
    test('getDownloadHREF', function() {
        scope.date_start_ui.date_value = new Date('2023-12-24')
        scope.filter_ui.criteria = {'event': [['event', '~', 'some-value']]}
        scope.table_options = {
            'any': ['extra', 'is', 'included'],
            'cursor': 2,
            'sorting': {
                'name': 'asc'
                },
        }

        var result = scope.getDownloadHREF()

        expect(result).toStartWith('/rp/path/activity-log-export/')
        // The filter is last part of the URL, in base64.
        // We decode it for testing.
        var url_parts = result.split('/')
        var options = JSON.parse(atob(url_parts[url_parts.length - 1]))
        expect(options).toEqual({
            'any': ['extra', 'is', 'included'],
            'cursor': 2,
            'sorting': {
                'name': 'asc'
                },
            'criteria': [
                [['event_id', 'contains', 'some-value']],
                // The start timestamp is extracted from the UI.
                [["timestamp", '>=', 1703376000]],
            ],
        })
    })

    /*
    Will reset the total result size when the filter is started and
    will make a new backend request.
    */
    test('onFilter', function() {
        mk.expectDefaultRouter($httpBackend)
        $httpBackend.flush()

        $httpBackend.expectPOST(
            '/rp/path/json', {
            'jsonrpc': '2.0',
            'method': 'db_fetch',
            'params': {
                'db_uuid': 'some-uuid',
                'page': 0,
                'limit': 50,
                'criteria': [],
                'sorting': {'timestamp': 'desc'}
            },
            'id': 1
        }).respond(200, getJSONResult(50))


        scope.table_options.db_uuid = 'some-uuid'
        scope.table_ui.total = 123

        scope.onFilter()
        scope.$digest()
        $httpBackend.flush()

        expect(scope.table_ui.total).toBe('many')
    })


    /*
    It will reset all filtering value and date time ranges and then will
    reload the table with the first page from JSON-RPC.
    */
    test('onReset_ok', function() {
        mk.expectDefaultRouter($httpBackend)
        $httpBackend.flush()

        // We need to have an uuid, as otherwise the whole table reload
        // is ignored.
        scope.table_options.db_uuid = 'some-uuid'

        // We can have any filters or sorting.
        scope.filter_ui.criteria = {'event': [['event', '=', 'some-value']]}
        scope.table_options.sorting = {'event_id': 'asc'}

        const existing_time = new Date('2002-01-15T12:19:42.000Z')
        scope.date_start_ui.date_value = existing_time
        scope.date_start_ui.time_value = existing_time
        scope.date_end_ui.date_value = existing_time
        scope.date_end_ui.time_value = existing_time

        scope.table_options.page = 13

        settlePage(34)

        $httpBackend.expectPOST(
            '/rp/path/json', {
            'jsonrpc': '2.0',
            'method': 'db_fetch',
            'params': {
                'db_uuid': 'some-uuid',
                'page': 0,
                'limit': 50,
                'criteria': [],
                'sorting': {'timestamp': 'desc'}
            },
            'id': 1
        }).respond(200, getJSONResult(50))

        scope.onReset()
        scope.$digest()
        $httpBackend.flush()

        expect(scope.entries.length).toEqual(50)
        expect(scope.table_options.page).toEqual(1)
        expect(scope.table_ui.total).toEqual('many')

        expect(scope.date_start_ui.date_value).toBeNull()
        expect(scope.date_start_ui.time_value).toBeNull()
        expect(scope.date_end_ui.date_value).toBeNull()
        expect(scope.date_end_ui.time_value).toBeNull()

        expect(scope.filter_ui.criteria).toEqual({})
        expect(scope.table_options.sorting).toEqual({'timestamp': 'desc'})
    })

    /*
    It will reset all date filter, but will keep the other filters.
    The page is reset, since we will have a different set of data.
    */
    test('onDateReset_ok', function() {
        mk.expectDefaultRouter($httpBackend)
        $httpBackend.flush()

        // We need to have an uuid, as otherwise the whole table reload
        // is ignored.
        scope.table_options.db_uuid = 'some-uuid'

        // We can have any filters.
        scope.filter_ui.criteria = {'event': [['event', '=', 'value']]}
        scope.table_options.sorting = {'event_id': 'asc'}

        const existing_time = new Date('2002-01-15T12:19:42.000Z')
        scope.date_start_ui.date_value = existing_time
        scope.date_start_ui.time_value = existing_time
        scope.date_end_ui.date_value = existing_time
        scope.date_end_ui.time_value = existing_time

        scope.table_options.page = 13
        settlePage(34)

        $httpBackend.expectPOST(
            '/rp/path/json', {
            'jsonrpc': '2.0',
            'method': 'db_fetch',
            'params': {
                'db_uuid': 'some-uuid',
                'page': 0, // We go to first page.
                'limit': 50,
                'criteria': [[["event_id", "=", "value"]]], // Keep the filters.
                'sorting': {'event_id': 'asc'} // Keep the sorting.
            },
            'id': 1
        }).respond(200, getJSONResult(50))

        scope.onDateReset()
        scope.$digest()
        $httpBackend.flush()

        expect(scope.entries.length).toEqual(50)
        expect(scope.table_options.page).toEqual(1)
        expect(scope.table_ui.total).toEqual('many')

        expect(scope.date_start_ui.date_value).toBeNull()
        expect(scope.date_start_ui.time_value).toBeNull()
        expect(scope.date_end_ui.date_value).toBeNull()
        expect(scope.date_end_ui.time_value).toBeNull()

        // The non-date filters and sorting are kept.
        expect(scope.filter_ui.criteria).toEqual(
            {'event': [['event', '=', 'value']]})
        expect(scope.table_options.sorting).toEqual({'event_id': 'asc'})
    })

    /*
    It will reset all date filter and will trigger a DB update,
    even when we are already on the first page.
    */
    test('onDateReset_first_page', function() {
        mk.expectDefaultRouter($httpBackend)
        $httpBackend.flush()

        // We need to have an uuid, as otherwise the whole table reload
        // is ignored.
        scope.table_options.db_uuid = 'some-uuid'

        // We can have any filters.
        scope.filter_ui.criteria = {'event': [['event', '=', 'value']]}
        scope.table_options.sorting = {'event_id': 'asc'}

        const existing_time = new Date('2002-01-15T12:19:42.000Z')
        scope.date_start_ui.date_value = existing_time
        scope.date_start_ui.time_value = existing_time
        scope.date_end_ui.date_value = existing_time
        scope.date_end_ui.time_value = existing_time

        scope.table_options.page = 1 // We are already on first page.

        settlePage(34)

        // Do the reset.
        $httpBackend.expectPOST(
            '/rp/path/json', {
            'jsonrpc': '2.0',
            'method': 'db_fetch',
            'params': {
                'db_uuid': 'some-uuid',
                'page': 0, // We go to first page.
                'limit': 50,
                'criteria': [[["event_id", "=", "value"]]], // Keep the filters.
                'sorting': {'event_id': 'asc'} // Keep the sorting.
            },
            'id': 1
        }).respond(200, getJSONResult(50))

        scope.onDateReset()
        scope.$digest()
        $httpBackend.flush()

        expect(scope.entries.length).toEqual(50)
        expect(scope.table_options.page).toEqual(1)
        expect(scope.table_ui.total).toEqual('many')

        expect(scope.date_start_ui.date_value).toBeNull()
        expect(scope.date_start_ui.time_value).toBeNull()
        expect(scope.date_end_ui.date_value).toBeNull()
        expect(scope.date_end_ui.time_value).toBeNull()

        // Non date filters are kept.
        expect(scope.filter_ui.criteria).toEqual({
            'event': [['event', '=', 'value']]})
        expect(scope.table_options.sorting).toEqual({'event_id': 'asc'})
    })

    /*
    It will trigger a JSON-RPC fetch based on the selected fixed range.

    It will keep the existing sorting and filtering.
    */
    test('onDateFilter_range', function() {
        mk.expectDefaultRouter($httpBackend)
        $httpBackend.flush()

        // We need to have an uuid, as otherwise the whole table reload
        // is ignored.
        scope.table_options.db_uuid = 'some-uuid'

        // We can have any filters.
        scope.filter_ui.criteria = {'event': [['event', '=', 'value']]}
        scope.table_options.sorting = {'event_id': 'asc'}

        scope.table_options.page = 1 // We are already on first page.
        settlePage(34)

        $httpBackend.expectPOST(
            '/rp/path/json', {
            'jsonrpc': '2.0',
            'method': 'db_fetch',
            'params': {
                'db_uuid': 'some-uuid',
                'page': 0, // We go to first page.
                'limit': 50,
                'criteria': [
                    [["event_id", "=", "value"]],  // Keep the filters.
                    [["timestamp", ">=" , 1234564290]], // Add time filter.
                ],
                'sorting': {'event_id': 'asc'} // Keep the sorting.
            },
            'id': 1
        }).respond(200, getJSONResult(43))

        scope.onDateFilter(1, 'hour')
        scope.$digest()
        $httpBackend.flush()

        expect(scope.entries.length).toEqual(43)
        expect(scope.table_options.page).toEqual(1)
        expect(scope.table_ui.total).toEqual(43)

        var start_date = new Date('2009-02-13T22:31:30.000Z')
        expect(scope.date_start_ui.date_value).toEqual(start_date)
        expect(scope.date_start_ui.time_value).toEqual(start_date)
        expect(scope.date_end_ui.date_value).toBeNull()
        expect(scope.date_end_ui.time_value).toBeNull()

        // Non-date filters are kept.
        expect(scope.filter_ui.criteria).toEqual({
            'event': [['event', '=', 'value']]})
        expect(scope.table_options.sorting).toEqual({'event_id': 'asc'})
    })

    /*
    It will trigger a JSON-RPC fetch based on the selected start date.
    */
    test('onDateFilter_only_start', function() {
        mk.expectDefaultRouter($httpBackend)
        $httpBackend.flush()

        // We need to have an uuid, as otherwise the whole table reload
        // is ignored.
        scope.table_options.db_uuid = 'some-uuid'

        // We can have any filters.
        scope.filter_ui.criteria = {'event': [['event', '=', 'value']]}
        scope.table_options.sorting = {'event_id': 'asc'}

        scope.table_options.page = 1 // We are already on first page.
        settlePage(34)

        var start_date = new Date('2009-02-13T22:31:30.000Z')
        scope.date_start_ui.time_value = start_date
        scope.date_start_ui.date_value = start_date

        $httpBackend.expectPOST(
            '/rp/path/json', {
            'jsonrpc': '2.0',
            'method': 'db_fetch',
            'params': {
                'db_uuid': 'some-uuid',
                'page': 0, // We go to first page.
                'limit': 50,
                'criteria': [
                    [["event_id", "=", "value"]],  // Keep the filters.
                    [["timestamp", ">=" , 1234564290]], // Add time filter.
                ],
                'sorting': {'event_id': 'asc'} // Keep the sorting.
            },
            'id': 1
        }).respond(200, getJSONResult(43))

        scope.onDateFilter()
        scope.$digest()
        $httpBackend.flush()

        expect(scope.entries.length).toEqual(43)
        expect(scope.table_options.page).toEqual(1)
        expect(scope.table_ui.total).toEqual(43)

        expect(scope.date_start_ui.date_value).toEqual(start_date)
        expect(scope.date_start_ui.time_value).toEqual(start_date)
        expect(scope.date_end_ui.date_value).toBeNull()
        expect(scope.date_end_ui.time_value).toBeNull()

        // Non-date filters are kept.
        expect(scope.filter_ui.criteria).toEqual({
            'event': [['event', '=', 'value']]})
        expect(scope.table_options.sorting).toEqual({'event_id': 'asc'})
    })

    /*
    It will trigger a JSON-RPC fetch based on the selected end date.
    */
    test('onDateFilter_only_end', function() {
        mk.expectDefaultRouter($httpBackend)
        $httpBackend.flush()

        // We need to have an uuid, as otherwise the whole table reload
        // is ignored.
        scope.table_options.db_uuid = 'some-uuid'

        // We can have any filters.
        scope.filter_ui.criteria = {'event': [['event', '=', 'value']]}
        scope.table_options.sorting = {'event_id': 'asc'}

        scope.table_options.page = 1 // We are already on first page.
        settlePage(34)

        var end_date = new Date('2009-02-13T22:31:30.000Z')
        scope.date_end_ui.time_value = end_date
        scope.date_end_ui.date_value = end_date

        $httpBackend.expectPOST(
            '/rp/path/json', {
            'jsonrpc': '2.0',
            'method': 'db_fetch',
            'params': {
                'db_uuid': 'some-uuid',
                'page': 0, // We go to first page.
                'limit': 50,
                'criteria': [
                    [["event_id", "=", "value"]],  // Keep the filters.
                    [["timestamp", "<=" , 1234564290]], // Add time filter.
                ],
                'sorting': {'event_id': 'asc'} // Keep the sorting.
            },
            'id': 1
        }).respond(200, getJSONResult(43))

        scope.onDateFilter()
        scope.$digest()
        $httpBackend.flush()

        expect(scope.entries.length).toEqual(43)
        expect(scope.table_options.page).toEqual(1)
        expect(scope.table_ui.total).toEqual(43)

        expect(scope.date_start_ui.date_value).toBeNull()
        expect(scope.date_start_ui.time_value).toBeNull()
        expect(scope.date_end_ui.date_value).toEqual(end_date)
        expect(scope.date_end_ui.time_value).toEqual(end_date)

        // Non-date filters are kept.
        expect(scope.filter_ui.criteria).toEqual({
            'event': [['event', '=', 'value']]})
        expect(scope.table_options.sorting).toEqual({'event_id': 'asc'})
    })

    /*
    It will trigger a JSON-RPC fetch based on the selected start and end date.
    */
    test('onDateFilter_start_and_end', function() {
        mk.expectDefaultRouter($httpBackend)
        $httpBackend.flush()

        // We need to have an uuid, as otherwise the whole table reload
        // is ignored.
        scope.table_options.db_uuid = 'some-uuid'

        // We can have any filters.
        scope.filter_ui.criteria = {'event': [['event', '=', 'value']]}
        scope.table_options.sorting = {'event_id': 'asc'}

        scope.table_options.page = 1 // We are already on first page.
        settlePage(34)

        var start_date = new Date('2009-02-13T21:31:30.000Z')
        scope.date_start_ui.time_value = start_date
        scope.date_start_ui.date_value = start_date
        var end_date = new Date('2009-02-13T22:31:30.000Z')
        scope.date_end_ui.time_value = end_date
        scope.date_end_ui.date_value = end_date

        $httpBackend.expectPOST(
            '/rp/path/json', {
            'jsonrpc': '2.0',
            'method': 'db_fetch',
            'params': {
                'db_uuid': 'some-uuid',
                'page': 0, // We go to first page.
                'limit': 50,
                'criteria': [
                    [["event_id", "=", "value"]],  // Keep the filters.
                    [["timestamp", ">=" , 1234560690]],
                    [["timestamp", "<=" , 1234564290]],
                ],
                'sorting': {'event_id': 'asc'} // Keep the sorting.
            },
            'id': 1
        }).respond(200, getJSONResult(43))

        scope.onDateFilter()
        scope.$digest()
        $httpBackend.flush()

        expect(scope.entries.length).toEqual(43)
        expect(scope.table_options.page).toEqual(1)
        expect(scope.table_ui.total).toEqual(43)

        expect(scope.date_start_ui.date_value).toEqual(start_date)
        expect(scope.date_start_ui.time_value).toEqual(start_date)
        expect(scope.date_end_ui.date_value).toEqual(end_date)
        expect(scope.date_end_ui.time_value).toEqual(end_date)

        // Non-date filters are kept.
        expect(scope.filter_ui.criteria).toEqual({
            'event': [['event', '=', 'value']]})
        expect(scope.table_options.sorting).toEqual({'event_id': 'asc'})
    })

    /*
    It will not trigger a JSON-RPC fetch if no range or start and end date
    were selected..
    */
    test('onDateFilter_noop', function() {
        mk.expectDefaultRouter($httpBackend)
        $httpBackend.flush()

        // We need to have an uuid, as otherwise the whole table reload
        // is ignored.
        scope.table_options.db_uuid = 'some-uuid'

        // We can have any filters.
        scope.table_options.filter = {'some': 'value'}
        scope.table_options.sorting = {}
        scope.table_options.sorting['event_id'] = 'asc'

        scope.table_options.page = 1 // We are already on first page.
        settlePage(34)

        scope.date_start_ui.time_value = null
        scope.date_start_ui.date_value = null

        scope.onDateFilter()
        scope.$digest()
        $httpBackend.verifyNoOutstandingRequest()

    })

    /*
    Total number of entries is zero if we are on the first page and
    no results were returned.
    */
    test('_cbTableUpdate_total_no_results', function() {
        var data = getJSONResult(0)
        scope.table_options.page = 1
        scope.table_options.limit = 5

        scope._cbTableUpdate(data)
        scope.$digest()

        expect(scope.table_ui.total).toBe(0)
        expect(scope.table_ui.range_start).toBe(1)
        expect(scope.table_ui.range_end).toBe(0)
        expect(scope.table_ui.have_previous).toBeFalse()
        expect(scope.table_ui.have_next).toBeFalse()
    })

    /*
    Total number of entries is 10 if we are on the 3rds page and no result
    were return for a batch of 5 entries.

    When we are on the last page, and no result are visible, it will
    show the result from the previous page.
    */
    test('_cbTableUpdate_total_empty_last_page', function() {
        var data = getJSONResult(0)
        scope.table_options.page = 3
        scope.table_options.limit = 5
        scope.table_ui.have_next = true
        scope._previous_entries = [{}, {}, {}, {}, {}]

        scope._cbTableUpdate(data)
        scope.$digest()

        expect(scope.table_ui.total).toBe(10)
        expect(scope.table_ui.range_start).toBe(6)
        expect(scope.table_ui.range_end).toBe(10)
        expect(scope.table_ui.have_next).toBeFalse()
        expect(scope.table_ui.have_previous).toBeTrue()

    })

    /*
    Will compute the total number or rows when the returned result is less
    than the limit.

    And we don't have a next page.
    */
    test('_cbTableUpdateTotal_last_less_limit', function() {
        var data = getJSONResult(4)
        scope.table_options.page = 3
        scope.table_options.limit = 5

        scope._cbTableUpdate(data)

        expect(scope.table_ui.total).toBe(14)
        expect(scope.table_ui.range_start).toBe(11)
        expect(scope.table_ui.range_end).toBe(14)
        expect(scope.table_ui.have_previous).toBeTrue()
        expect(scope.table_ui.have_next).toBeFalse()
    })

    /*
    Compute datetime member and set it
    */
    test('_cbTableUpdate_transformation', function() {
        var now = dayjs()
        var timestamp = now.valueOf() / 1000
        var data = {
            'result': {
                'data': [{'timestamp': timestamp, 'data': '{}'}],
                'total': 0
                }
            }

        scope._cbTableUpdate(data)

        result = scope.entries[0]
        expect(result['datetime'].format()).toEqual(now.format())
        expect(result['timestamp']).toEqual(timestamp)
    })

    /*
    Does nothing when the page number is invalid.

    The UI will reset
    */
    test('_updateTable_invalid_page', function() {
        mk.expectDefaultRouter($httpBackend)
        $httpBackend.flush()

        scope.table_options.page = null
        scope.table_options.db_uuid = 'something'

        scope._updateTable()
        scope.$digest()

        $httpBackend.verifyNoOutstandingRequest()

    })

    /*
    Will ignore any empty rules
    */
    test('_sanitizeCriteria_date_end_start', function() {
        scope.filter_ui.criteria = {
            'event': [],
            'user': null,
            'text': [['text', '=', 'some-value']],
        }

        result = scope._sanitizeCriteria()

        expect(result).toEqual([
            [['message', '=', 'some-value']],
        ])
    })


    /*
    Will convert current filtered column to DB criteria.

    Empty values are ignored.
    */
    test('_sanitizeCriteria_date_end_start', function() {
        scope.filter_ui.criteria = {
            'event': [['event', '=', '1234']],
            'user': [['user', '~', 'john D']],
            'text': [['text', '=', 'some-value']],
        }

        result = scope._sanitizeCriteria()

        expect(result).toEqual([
            [['event_id', '=', '1234']],
            [['user_name', 'contains', 'john D']],
            [['message', '=', 'some-value']],
        ])
    })

    /*
    Will convert date and time values to timestamp.
    */
    test('_sanitizeCriteria_date_end_start', function() {
        var d = new Date()
        var timestamp = d.getTime() / 1000
        scope.date_start_ui.date_value = d
        scope.date_start_ui.time_value = d
        scope.date_end_ui.date_value = d
        scope.date_end_ui.time_value = d

        result = scope._sanitizeCriteria()

        expect(result).toEqual([
            [['timestamp', '>=', timestamp]],
            [['timestamp', '<=', timestamp]],
        ])
    })


    /*
    Deserialize the data part of the json, removing the account member.

    Events with unknown ID are considered as `failure-critical`.

    Date and time are present as separate values.
    */
    test('_updateEntryData_data_unknown_event_id', function() {
        var db_event = {
            'event_id': 123,  // There should be no event with this ID
            'timestamp': 1,
            'data': (
                '{"source": "logs", "account": {"peer": ' +
                '{"protocol": "TCP", "port": 57860, "address": ' +
                '"127.0.0.1"}, "uuid": "DEFAULT-ADMINISTRATOR-UUID", ' +
                '"name": "admin"}, "database_uuid": "db-uuid"}'
                )
        }

        scope._updateEntryData(db_event)

        var expected = {
            'event_id': 123,
            'timestamp': 1,
            'datetime': dayjs.unix(1),
            'date': '1970-01-01',
            'time': '00:00:01',
            'main_group': 'failure-critical',
            'data': {
                'source': 'logs',
                'database_uuid': 'db-uuid'
            }
        }
        expect(db_event).toEqual(expected)
    })

    /*
    It can detect event fom the "success" group.

    If there is not data, an empty dict is generated.
    */
    test('_updateEntryData_success_no_data', function() {
        var db_event = {
            // This ID might be removed in a future release,
            // but for now it's helps with this test and we can update
            // this tests when 20039 is removed.
            'event_id': 20039,
            'timestamp': 1,
            'data': '{}',
        }

        scope._updateEntryData(db_event)

        var expected = {
            'event_id': 20039,
            'timestamp': 1,
            'datetime': dayjs.unix(1),
            'date': '1970-01-01',
            'time': '00:00:01',
            'main_group': 'success',
            'data': {},
        }
        expect(db_event).toEqual(expected)
    })

    /*
    It can detect event fom the "failure" group.
    */
    test('_updateEntryData_failure', function() {
        var db_event = {
            'event_id': 20021,
            'timestamp': 1,
            'data': '{}',
        }

        scope._updateEntryData(db_event)

        var expected = {
            'event_id': 20021,
            'timestamp': 1,
            'datetime': dayjs.unix(1),
            'date': '1970-01-01',
            'time': '00:00:01',
            'main_group': 'failure',
            'data': {},
        }
        expect(db_event).toEqual(expected)
    })

    /*
    When an event is both "failure" and "failure-high",
    the main group is considered "failure-high",
    */
    test('_updateEntryData_failure_high', function() {
        var db_event = {
            'event_id': 20031,
            'timestamp': 1,
            'data': '{}',
        }

        scope._updateEntryData(db_event)

        var expected = {
            'event_id': 20031,
            'timestamp': 1,
            'datetime': dayjs.unix(1),
            'date': '1970-01-01',
            'time': '00:00:01',
            'main_group': 'failure-high',
            'data': {},
        }
        expect(db_event).toEqual(expected)
    })


    /*
    When an event is both "failure" and "failure-critical",
    the main group is considered "failure-critical",
    */
    test('_updateEntryData_failure', function() {
        var db_event = {
            'event_id': 20033,
            'timestamp': 1,
            'data': '{}',
        }

        scope._updateEntryData(db_event)

        var expected = {
            'event_id': 20033,
            'timestamp': 1,
            'datetime': dayjs.unix(1),
            'date': '1970-01-01',
            'time': '00:00:01',
            'main_group': 'failure-critical',
            'data': {},
        }
        expect(db_event).toEqual(expected)
    })

    /*
    Will go to the first page, if page number is negative.
    */
    test('onGoToPage_negative', function() {

        scope.table_options.page = 8

        scope.onGoToPage(-1)

        expect(scope.table_options.page).toBe(1)
    })

    /*
    Will reload the table when the page size is changed,
    going to the first page.
    */
    test('table_options_limit_update', function() {

        scope.table_options.page = 6
        scope.table_options.limit = 123

        scope.$digest()

        expect(scope.table_options.page).toBe(1)
    })

    /*
    Will reload the table when the sorting order is changed,
    going to the first page.
    */
    test('table_options_sorting_update', function() {

        scope.table_options.page = 4
        scope.table_options.sorting = {'event_id': 'asc'}

        scope.$digest()

        expect(scope.table_options.page).toBe(1)
    })

    /*
    Will reset the comparison for empty column names.
    */
    test('filter_ui.column_watch_no_Value', function() {
        scope.$digest()
        scope.filter_ui.column = ''
        scope.filter_ui.comparison = 'anything'

        scope.$digest()

        expect(scope.filter_ui.comparison).toEqual('=')
    })

    /*
    Will keep the comparison for a column that has the comparison.
    */
    test('filter_ui.column_watch_keep', function() {
        scope.$digest()
        scope.filter_ui.column = 'text'
        scope.filter_ui.comparison = '~'

        scope.$digest()

        expect(scope.filter_ui.comparison).toEqual('~')
    })

    /*
    Will reset the comparison if the current column does not support
    the previous comparison.
    */
    test('filter_ui.column_watch_reset', function() {
        // Make sure we trigger a watch event.
        // as the event value is the default one, so we need to
        // change it first, and then set it back again.
        scope.filter_ui.column = 'text'
        scope.$digest()
        scope.filter_ui.column = 'event'
        scope.filter_ui.comparison = '~'

        scope.$digest()

        expect(scope.filter_ui.comparison).toEqual('=')
    })

    /*
    Does nothing if random key events are generated.
    */
    test('onFilterValueKeyDown_other_key', function() {
        scope.filter_ui.column = 'anything'
        scope.filter_ui.comparison = '~'
        scope.filter_ui.value = 'same-old'

        var event = {'keyCode': 123}
        scope.onFilterValueKeyDown(event)

        expect(scope.filter_ui.value).toEqual('same-old')
        expect(scope.filter_ui.criteria).toEqual({})
    })

    /*
    Add the current value if Enter key is pressed.
    */
    test('onFilterValueKeyDown_enter_key', function() {
        scope.filter_ui.column = 'user'
        scope.filter_ui.comparison = '~'
        scope.filter_ui.value = 'same-old'

        var event = {'keyCode': 13}
        scope.onFilterValueKeyDown(event)

        // The filter column and comparison are kept.
        expect(scope.filter_ui.column).toEqual('user')
        expect(scope.filter_ui.comparison).toEqual('~')
        // Value is reset.
        expect(scope.filter_ui.value).toEqual('')
        expect(scope.filter_ui.criteria).toEqual({'user': [
            ['user', '~', 'same-old']
        ]})
    })

    /*
    Does nothing if no value is selected for the new filter rule.
    */
    test('onAddFilter_empty_value', function() {
        scope.filter_ui.column = 'anything'
        scope.filter_ui.comparison = '~'
        scope.filter_ui.value = ''

        scope.onAddFilter()

        expect(scope.filter_ui.column).toEqual('anything')
        expect(scope.filter_ui.comparison).toEqual('~')
        expect(scope.filter_ui.value).toEqual('')
        expect(scope.filter_ui.criteria).toEqual({})
    })

    /*
    Add a new filter rule based on current input.
    */
    test('onAddFilter_value', function() {
        scope.filter_ui.column = 'event'
        scope.filter_ui.comparison = '='
        scope.filter_ui.value = '1223'

        scope.onAddFilter()

        // The column UI is kept.
        expect(scope.filter_ui.column).toEqual('event')
        expect(scope.filter_ui.comparison).toEqual('=')
        // But the value is reset.
        expect(scope.filter_ui.value).toEqual('')
        // And a new filter rule is added.
        expect(scope.filter_ui.criteria).toEqual({
            'event': [['event', '=', '1223']],
        })

        // Updating the value, will add a new rule
        scope.filter_ui.value = '456'

        scope.onAddFilter()

        // The column UI is kept.
        expect(scope.filter_ui.column).toEqual('event')
        expect(scope.filter_ui.comparison).toEqual('=')
        // But the value is reset.
        expect(scope.filter_ui.value).toEqual('')
        // And a new filter rule is added.
        expect(scope.filter_ui.criteria).toEqual({
            'event': [['event', '=', '1223'], ['event', '=', '456']],
        })

        // Updating the value and rule type
        scope.filter_ui.column = 'user'
        scope.filter_ui.comparison = '~'
        scope.filter_ui.value = 'some-text'

        scope.onAddFilter()

        // The column UI is kept.
        expect(scope.filter_ui.column).toEqual('user')
        expect(scope.filter_ui.comparison).toEqual('~')
        // But the value is reset.
        expect(scope.filter_ui.value).toEqual('')
        // And a new filter rule is added.
        expect(scope.filter_ui.criteria).toEqual({
            'event': [['event', '=', '1223'], ['event', '=', '456']],
            'user': [['user', '~', 'some-text']],
        })

    })

    /*
    Will remove one rule from the existing ones.
    */
    test('onDeleteFilter_more', function() {
        // Rule to be removed must be the exact same instance.
        var rule = ['event', '=', '456']
        scope.filter_ui.criteria = {}
        scope.filter_ui.criteria['event'] = [
            ['event', '=', '1223'],
            rule,
            ['event', '=', '789'],
        ]
        scope.filter_ui.criteria['user'] = [['user', '~', 'some-text']]

        scope.onDeleteFilter(rule)

        expect(scope.filter_ui.criteria).toEqual({
            'event': [['event', '=', '1223'], ['event', '=', '789']],
            'user': [['user', '~', 'some-text']],
        })
    })


    /*
    Will remove the whole category if the removed rule is the last from that
    type.
    */
    test('onDeleteFilter_more', function() {
        var rule = ['user', '~', 'some-text']
        scope.filter_ui.criteria = {}
        scope.filter_ui.criteria['event'] = [
            ['event', '=', '1223'],
            ['event', '=', '789'],
        ]
        scope.filter_ui.criteria['user'] = [rule]

        scope.onDeleteFilter(rule)

        expect(scope.filter_ui.criteria).toEqual({
            'event': [['event', '=', '1223'], ['event', '=', '789']],
        })
    })


    /*
    Goes to the previous event from the current.
    */
    test('onGoToPrevious', function() {

        scope.event_ui.cursor = 3

        scope.onGoToPrevious()

        expect(scope.event_ui.cursor).toBe(2)
    })

    /*
    Goes to the next event from the current.
    */
    test('onGoToNext', function() {

        scope.event_ui.cursor = 3

        scope.onGoToNext()

        expect(scope.event_ui.cursor).toBe(4)
    })

    /*
    Nulifies the event for which the details should be displayed.
    */
    test('onShowEventsOverview', function() {

        scope.event = {'something': ''}

        scope.onShowEventsOverview()

        expect(scope.event).toBeNull()
    })

    /*
    Will set the event for which the details should be displayed as an index
    based on the current page.
    */
    test('onShowEventDetails_middle_event', function() {

        scope.table_options.limit = 5
        scope.event = null
        scope.event_details_cursor = null
        var event_details = {'event_id': '123', 'data': '{}'}
        scope.entries = [{}, {}, event_details, {}]
        scope.event_ui.have_previous = false
        scope.event_ui.have_next = false

        scope.onShowEventDetails(2)

        expect(scope.event).toBe(event_details)
        expect(scope.event_ui.cursor).toBe(2)
        expect(scope.event_ui.have_previous).toBeTrue()
        expect(scope.event_ui.have_next).toBeTrue()
    })

    /*
    When we have the first event from the first page, there is no
    previous button.
    */
    test('onShowEventDetails_first_event', function() {

        scope.table_options.limit = 5
        scope.event = null
        scope.event_details_cursor = null
        var event_details = {'event_id': '123', 'data': '{}'}
        scope.entries = [event_details, {}, {}, {}]
        scope.event_ui.have_previous = true
        scope.event_ui.have_next = false
        scope.table_options.page = 1

        scope.onShowEventDetails(0)
        scope.$digest()

        expect(scope.event).toBe(event_details)
        expect(scope.event_ui.cursor).toBe(0)
        expect(scope.event_ui.have_previous).toBeFalse()
        expect(scope.event_ui.have_next).toBeTrue()
    })

    /*
    When the event is last from the current page, and current page has
    fewer entries than the page limit, it means that we are on the last event.
    */
    test('onShowEventDetails_last_event', function() {

        scope.table_options.limit = 5
        scope.event = null
        scope.event_details_cursor = null
        var event_details = {'event_id': '123', 'data': '{}'}
        scope.entries = [{}, {}, {}, event_details]
        scope.event_ui.have_previous = false
        scope.event_ui.have_next = true
        scope.table_options.page = 2

        scope.onShowEventDetails(3)

        expect(scope.event).toBe(event_details)
        expect(scope.event_ui.cursor).toBe(3)
        expect(scope.event_ui.have_previous).toBeTrue()
        expect(scope.event_ui.have_next).toBeFalse()
    })

    /*
    When we are past the last event from the page, it will automatically
    go to the first element from the next page.
    */
    test('onShowEventDetails_auto_next_page', function() {
        mk.expectDefaultRouter($httpBackend)
        $httpBackend.flush()

        // Will reload the table when we set a new limit.
        $httpBackend.expectPOST(
            '/rp/path/json', {
            'jsonrpc': '2.0',
            'method': 'db_fetch',
            'params': {
                'db_uuid': 'db-uuid',
                'page': 0,
                'limit': 5,
                'criteria': [],
                'sorting': {'timestamp': 'desc'}
            },
            'id': 1
            }
          ).respond(200, '')

        scope.table_options.db_uuid = 'db-uuid'
        scope.table_options.limit = 5
        scope.event = null
        scope.event_details_cursor = null
        scope.event_ui.have_previous = false
        scope.event_ui.have_next = false
        scope.$digest()
        $httpBackend.flush()

        // Pretend that we are already on second page.
        scope.table_options.page = 2
        var response = {'result': {'data': [
            {"event_id": "123", "data": "{}"},
            {"event_id": "332", "data": "{}"},
        ]}}
        // We expect to load the 3rd page.
        $httpBackend.expectPOST(
            '/rp/path/json', {
            'jsonrpc': '2.0',
            'method': 'db_fetch',
            'params': {
                'db_uuid': 'db-uuid',
                'page': 2, // This is UI page 3.
                'limit': 5,
                'criteria': [],
                'sorting': {'timestamp': 'desc'}
            },
            'id': 1
        }).respond(200, JSON.stringify(response))
        // This is loaded twice.
        $httpBackend.expectPOST(
            '/rp/path/json', {
            'jsonrpc': '2.0',
            'method': 'db_fetch',
            'params': {
                'db_uuid': 'db-uuid',
                'page': 2, // This is UI page 3.
                'limit': 5,
                'criteria': [],
                'sorting': {'timestamp': 'desc'}
            },
            'id': 1
        }).respond(200, JSON.stringify(response))

        scope.onShowEventDetails(5)

        scope.$digest()
        $httpBackend.flush()

        expect(scope.event.event_id).toEqual('123')
        expect(scope.event_ui.cursor).toBe(0)
        expect(scope.table_options.page).toBe(3)
        expect(scope.event_ui.have_previous).toBeTrue()
        // The next page has more than 1 element.
        expect(scope.event_ui.have_next).toBeTrue()
    })

    /*
    When we are before the first event from the page, it will automatically
    go to the last element from the previous page.
    */
    test('onShowEventDetails_auto_previous_page', function() {
        mk.expectDefaultRouter($httpBackend)
        $httpBackend.flush()

        var response = {'result': {'data': [
            {"event_id": "123", "data": "{}"},
        ]}}
        // Will reload the table when we set a new limit.
        $httpBackend.expectPOST(
            '/rp/path/json', {
            'jsonrpc': '2.0',
            'method': 'db_fetch',
            'params': {
                'db_uuid': 'db-uuid',
                'page': 0,
                'limit': 5,
                'criteria': [],
                'sorting': {'timestamp': 'desc'}
            },
            'id': 1,
        }).respond(200, response)

        scope.table_options.db_uuid = 'db-uuid'
        scope.table_options.limit = 5
        scope.event = null
        scope.event_details_cursor = null
        scope.event_ui.have_previous = false
        scope.event_ui.have_next = false
        scope.$digest()
        $httpBackend.flush()

        // Pretend that we are already on second page.
        scope.table_options.page = 2
        var response = {'result': {'data': [
            {"event_id": "123", "data": "{}"},
            {"event_id": "232", "data": "{}"},
            {"event_id": "332", "data": "{}"},
            {"event_id": "432", "data": "{}"},
            {"event_id": "532", "data": "{}"},
        ]}}
        // We expext to load the 3rd page.
        $httpBackend.expectPOST(
            '/rp/path/json', {
            'jsonrpc': '2.0',
            'method': 'db_fetch',
            'params': {
                'db_uuid': 'db-uuid',
                'page': 0,
                'limit': 5,
                'criteria': [],
                'sorting': {'timestamp': 'desc'}
            },
            'id': 1
            }
          ).respond(200, JSON.stringify(response))

        scope.onShowEventDetails(-1)

        $httpBackend.flush()

        expect(scope.event.event_id).toEqual('532')
        expect(scope.event_ui.cursor).toBe(4)
        expect(scope.table_options.page).toBe(1)
        expect(scope.event_ui.have_previous).toBeTrue()
        expect(scope.event_ui.have_next).toBeTrue()
    })


})
