/*
Tests for reports controller.
*/


describe('ReportsCtrl', function() {

    /*
    It will load the default db.
    */
    test('__init__', function() {
        var sut = mk.getController('reportsCtrl')

        sut.$httpBackend.expectPOST(
            '/rp/path/json', {
            'jsonrpc': '2.0',
            'method': 'db_fetch',
            'params': {
                'db_uuid': null,
                'page': 0,
                'limit': 50,
                'criteria': [],
                'sorting': {'last_login_successful_timestamp': 'desc'}
            },
            'id': 1
            }
          ).respond(200, '')

        // I don't know why the status page is loaded here.
        mk.expectDefaultRouter(sut.$httpBackend)

        // Any DB uuid is ignored for reports,
        // since the reports are using the default DB.
        sut.scope.table_options.db_uuid = 'some-uuid'

        sut.scope.__init__()
        sut.$httpBackend.flush()
    })

    /*
    Augments the data received from JSON-RPC backend.
    */
    test('__init__update_fetched_data', function() {
        var sut = mk.getController('reportsCtrl')
        var json_rpc_row = {'last_login_successful_timestamp': 12345}

        sut.scope._updateEntryData(json_rpc_row)

        expect(json_rpc_row).toEqual({
            'last_login_successful_timestamp': 12345,
            'datetime': sut.dayjs.unix(12345),
        })

    })

    /*
    It will search based on account name or IP using the current text filter.
    */
    test('onTextFilter', function() {
        var sut = mk.getController('reportsCtrl')
        sut.$httpBackend.expectPOST('/rp/path/json', {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'db_fetch',
            'params': {
                "db_uuid": null,
                "page":0,
                "limit":50,
                "criteria": [[
                    ['account_name', 'contains', 'some-value'],
                    ['last_login_successful_peer', 'contains', 'some-value'],
                    ['last_login_successful_service', 'contains', 'some-value'],
                ]],
                "sorting": {"last_login_successful_timestamp":"desc"},
            }
        }).respond(200, {'result': {'data': []}})
        sut.expectDefaultRouter()
        sut.scope.current_filter.text = 'some-value'

        sut.scope.onTextFilter()

        sut.$httpBackend.flush()
    })

    /*
    It will search starting with last hour.
    */
    test('onDateFilter_hour', function() {
        var sut = mk.getController('reportsCtrl')
        sut.$httpBackend.expectPOST('/rp/path/json', {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'db_fetch',
            'params': {
                "db_uuid": null,
                "page":0,
                "limit":50,
                "criteria": [[[
                    'last_login_successful_timestamp',
                    '>=',
                    sut.dayjs().subtract(1, 'hour').unix()
                ]]],
                "sorting": {"last_login_successful_timestamp":"desc"},
            }
        }).respond(200, {'result': {'data': []}})
        sut.expectDefaultRouter()
        sut.scope.date_button_name = 'some value'
        var option = {'value': 'hour', 'text': 'Last hour'}

        sut.scope.onDateFilter(option)

        expect(sut.scope.date_button_name).toEqual('Showing last hour')
        sut.$httpBackend.flush()
    })

    /*
    It will search starting with last day.
    */
    test('onDateFilter_day', function() {
        var sut = mk.getController('reportsCtrl')
        sut.$httpBackend.expectPOST('/rp/path/json', {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'db_fetch',
            'params': {
                "db_uuid": null,
                "page":0,
                "limit":50,
                "criteria": [[[
                    'last_login_successful_timestamp',
                    '>=',
                    sut.dayjs().subtract(24, 'hour').unix()
                ]]],
                "sorting": {"last_login_successful_timestamp":"desc"},
            }
        }).respond(200, {'result': {'data': []}})
        sut.expectDefaultRouter()
        sut.scope.date_button_name = 'some value'
        var option = {'value': 'day', 'text': 'Last day'}

        sut.scope.onDateFilter(option)

        expect(sut.scope.date_button_name).toEqual('Showing last day')
        sut.$httpBackend.flush()
    })

    /*
    It will remove the time limitation if time value is not known.
    */
    test('onDateFilter_something_else', function() {
        var sut = mk.getController('reportsCtrl')
        sut.$httpBackend.expectPOST('/rp/path/json', {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'db_fetch',
            'params': {
                "db_uuid": null,
                "page":0,
                "limit":50,
                "criteria": [],
                "sorting": {"last_login_successful_timestamp":"desc"},
            }
        }).respond(200, {'result': {'data': []}})
        sut.expectDefaultRouter()
        sut.scope.date_button_name = 'some value'
        var option = {'value': 'anything', 'text': 'Last time'}

        sut.scope.onDateFilter(option)

        expect(sut.scope.date_button_name).toEqual('Date filter options')
        sut.$httpBackend.flush()
    })

    /*
    It will search starting with last week.
    */
    test('onDateFilter_week', function() {
        var sut = mk.getController('reportsCtrl')
        sut.$httpBackend.expectPOST('/rp/path/json', {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'db_fetch',
            'params': {
                "db_uuid": null,
                "page":0,
                "limit":50,
                "criteria": [[[
                    'last_login_successful_timestamp',
                    '>=',
                    sut.dayjs().subtract(7, 'day').unix()
                ]]],
                "sorting": {"last_login_successful_timestamp":"desc"},
            }
        }).respond(200, {'result': {'data': []}})
        sut.expectDefaultRouter()
        sut.scope.date_button_name = 'some value'
        var option = {'value': 'week', 'text': 'Last week'}

        sut.scope.onDateFilter(option)

        expect(sut.scope.date_button_name).toEqual('Showing last week')
        sut.$httpBackend.flush()
    })

    /*
    It will search starting with last month.
    */
    test('onDateFilter_month', function() {
        var sut = mk.getController('reportsCtrl')
        sut.$httpBackend.expectPOST('/rp/path/json', {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'db_fetch',
            'params': {
                "db_uuid": null,
                "page":0,
                "limit":50,
                "criteria": [[[
                    'last_login_successful_timestamp',
                    '>=',
                    sut.dayjs().subtract(31, 'day').unix()
                ]]],
                "sorting": {"last_login_successful_timestamp":"desc"},
            }
        }).respond(200, {'result': {'data': []}})
        sut.expectDefaultRouter()
        sut.scope.date_button_name = 'some value'
        var option = {'value': 'month', 'text': 'Last month'}

        sut.scope.onDateFilter(option)

        expect(sut.scope.date_button_name).toEqual('Showing last month')
        sut.$httpBackend.flush()
    })

    /*
    It will search starting with last year.
    */
    test('onDateFilter_year', function() {
        var sut = mk.getController('reportsCtrl')
        sut.$httpBackend.expectPOST('/rp/path/json', {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'db_fetch',
            'params': {
                "db_uuid": null,
                "page":0,
                "limit":50,
                "criteria": [[[
                    'last_login_successful_timestamp',
                    '>=',
                    sut.dayjs().subtract(356, 'day').unix()
                ]]],
                "sorting": {"last_login_successful_timestamp":"desc"},
            }
        }).respond(200, {'result': {'data': []}})
        sut.expectDefaultRouter()
        sut.scope.date_button_name = 'some value'
        var option = {'value': 'year', 'text': 'Last year'}

        sut.scope.onDateFilter(option)

        expect(sut.scope.date_button_name).toEqual('Showing last year')
        sut.$httpBackend.flush()
    })

    /*
    It can trigger a search ordered by peer and will use the current,
    order direction.
    */
    test('table_options.sorting', function() {
        var sut = mk.getController('reportsCtrl')
        sut.$httpBackend.expectPOST('/rp/path/json', {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'db_fetch',
            'params': {
                "db_uuid": null,
                "page":0,
                "limit":50,
                "criteria": [],
                "sorting": {"last_login_successful_peer": "desc"},
            }
        }).respond(200, {'result': {'data': []}})
        sut.expectDefaultRouter()

        // This is updated by the sort UI component.
        sut.scope.table_options.sorting = {
            'last_login_successful_peer': 'desc',
        }

        sut.scope.$digest()
        sut.$httpBackend.flush()
        // Changing the order, will keep the column.
        sut.$httpBackend.expectPOST('/rp/path/json', {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'db_fetch',
            'params': {
                "db_uuid": null,
                "page":0,
                "limit":50,
                "criteria": [],
                "sorting": {"last_login_successful_peer": "asc"},
            }
        }).respond(200, {'result': {'data': []}})
    })

    /*
    It can trigger a DB fetch with an updated page.

    Page value is in UI mode (starting with 1),
    while JSON-RPC request is done starting with 0.
    */
    test('table_options.page', function() {
        var sut = mk.getController('reportsCtrl')
        sut.expectDefaultRouter()
        sut.$httpBackend.flush()

        // This is updated by the sort UI component.
        sut.scope.table_options.page = 3
        sut.$httpBackend.expectPOST('/rp/path/json', {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'db_fetch',
            'params': {
                "db_uuid": null,
                "page": 2,
                "limit": 50,
                "criteria": [],
                "sorting": {"last_login_successful_timestamp": "desc"},
            }
        // Make sure the response contains something,
        // as otherwise it will navigate to page 0.
        }).respond(200, {'result': {'data': [{'some': 'content'}]}})
        sut.scope.$digest()
        sut.$httpBackend.flush()

    })



})
