/*
Tests for resource controller.
*/


describe('ResourceCtrl', function() {

    /*
    Go to resources page if could not find requested resource.
    */
    test('__init__not_found', function() {
        var sut = mk.getController('resourceCtrl')
        sut.scope.$parent.configuration = {'resources': []}
        sut.scope.$parent.new_entities['resources'] = [
            {'name': 'some-name'}]
        sut.$location.path('/resources/some-other-name')

        sut.scope.__init__()

        expect(sut.$location.path()).toBe('/resources')
    })

    /*
    Loads the resource based on URL name of a new resource,
    and loads the template associated with the resource type.

    New resources don't have UUID yet, but they still can be loaded, based on
    the name.
    */
    test('__init__new', function() {
        var sut = mk.getController('resourceCtrl')
        var new_resource = {
            'name': 'some-name',
            'type': 'email-client'
        }
        sut.scope.$parent.process.resources = {}
        sut.scope.$parent.configuration = {'resources': []}
        sut.scope.$parent.new_entities['resources'] = [new_resource]
        sut.route.uuid = 'some-name'

        sut.scope.__init__()

        expect(sut.scope.resource).toBe(new_resource)
        // New resources don't have a runnable yet.
        expect(sut.scope.runnable).toBeFalse()
        expect(sut.scope.has_check).toBeTrue()
        expect(sut.scope.section_uuid).toBe('')
        expect(sut.scope.template_path).toEqual(
            'partials/resource_email-client.html?v=%(cache)s')
    })

    var setExistingResource = function(scope) {
        var existing_resource = {
            'name': 'some-name',
            'type': 'some-type',
            'uuid': 'some-uuid',
        }
        var existing_runnable = {
            'name': 'some-name',
            'type': 'some-type',
            'uuid': 'some-uuid',
            'state': 'stopped',
        }
        scope.$parent.process.runnables.resources = {
            'some-uuid': existing_runnable}
        scope.$parent.configuration.resources = {'some-uuid': existing_resource}
        scope.new_entities['resources'] = []
        return {
            'configuration': existing_resource,
            'runnable': existing_runnable}
    }

    /*
    Loads the resource based on URL uuid for an existing resource
    and loads the template associated with the resource type.
    */
    test('__init__existing', function() {
        var sut = mk.getController('resourceCtrl')
        var resource = setExistingResource(sut.scope)
        sut.route.uuid = resource.configuration.uuid

        sut.scope.__init__()

        expect(sut.scope.resource).toBe(resource.configuration)
        expect(sut.scope.runnable).toBe(resource.runnable)
        expect(sut.scope.has_check).toBeFalse()
        expect(sut.scope.section_uuid).toBe('resources/some-uuid')
        expect(sut.scope.template_path).toEqual(
            'partials/resource_some-type.html?v=%(cache)s')
    })

    /*
    It can load the monitor service as a resource..
    */
    test('__init__monitor', function() {
        var sut = mk.getController('resourceCtrl')
        mk.minimalProcess(sut)
        var config = {
            'uuid': 'some-uuid',
            'type': 'monitor',
            'name': 'Local Monitor',
            'configuration': {
                'source_path': 'some/path',
            }
        }
        var runnable = {
            'uuid': 'some-uuid',
            'type': 'monitor',
            'name': 'Local Monitor',
            'state': 'stopped',
        }
        sut.scope.configuration.services = {'some-uuid': config}
        sut.scope.$parent.process.runnables.services = {'some-uuid': runnable}
        sut.route.uuid = 'some-uuid'

        sut.scope.__init__()

        expect(sut.scope.resource).toBe(config)
        expect(sut.scope.configuration).toBe(config)
        expect(sut.scope.ssl_configuration).toBe(config)
        expect(sut.scope.$parent.component).toBe(config)
        expect(sut.scope.$parent.runnable).toBe(runnable)
        expect(sut.scope.has_check).toBeFalse()
        // Delete is not blocked for monitors.
        expect(sut.scope.is_delete_blocked).toEqual('')
        expect(sut.scope.section_uuid).toBe('resources/some-uuid')
        expect(sut.scope.template_path).toEqual(
            'partials/resource_monitor.html?v=%(cache)s')
    })


    /*
    Will call the JSON-RPC once with the current resource configuration
    and set the response message and icon.
    */
    test('onCheckResource_success', function() {
        var sut = mk.getController('resourceCtrl')
        var resource = setExistingResource(sut.scope)
        sut.route.uuid = resource.configuration.uuid
        sut.scope.__init__()

        sut.$httpBackend.expectPOST(
            '/rp/path/json', {
            'jsonrpc': '2.0',
            'method': 'runnable_check',
            'params': {
                'family': 'resources',
                'configuration': resource.configuration,
            },
            'id': 1
            }
          ).respond(200, {'result': 'Message from server.'})
        mk.expectDefaultRouter(sut.$httpBackend)

        sut.scope.onCheckResource()
        // Calling it again, while the response was not received yet will
        // not trigger another requst.
        sut.scope.onCheckResource()

        sut.$httpBackend.flush()

        expect(sut.scope.check_result).toEqual({
            'message': 'Message from server.',
            'class_icon': "fas fa-check-circle text--success",
        })
    })

    /*
    Will set the error details using the events DB,
    when the JSON-RPC returns an error.
    */
    test('onCheckResource_error', function() {
        var sut = mk.getController('resourceCtrl')
        var resource = setExistingResource(sut.scope)
        sut.route.uuid = resource.configuration.uuid
        sut.scope.__init__()

        sut.$httpBackend.expectPOST(
            '/rp/path/json', {
            'jsonrpc': '2.0',
            'method': 'runnable_check',
            'params': {
                'family': 'resources',
                'configuration': resource.configuration,
            },
            'id': 1
            }
          ).respond(
            200,
            {'error': {
                'id': '20000',
                'message': 'JSON-RPC message not used',
                'data': {'message': 'Boom from server!'},
            }},
            )
        mk.expectDefaultRouter(sut.$httpBackend)

        sut.scope.onCheckResource()
        // Calling it again, while the response was not received yet will
        // not trigger another requst.
        sut.scope.onCheckResource()

        sut.$httpBackend.flush()

        expect(sut.scope.check_result).toEqual({
            'message': 'Test failed. 20000: DEBUG: Boom from server!',
            'class_icon': "fas fa-times-circle text--important",
        })
    })

    /*
    It can delete the normal a resource.
    */
    test('onDeleteResource_email', function() {
        var sut = mk.getController('resourceCtrl')
        mk.minimalProcess(sut)
        var config = {
            'uuid': 'some-uuid',
            'type': 'email-client',
            'name': 'Some email',
        }
        var runnable = {
            'uuid': 'some-uuid',
            'type': 'email-client',
            'name': 'Some email',
            'state': 'stopped',
        }
        sut.scope.$parent.configuration.resources = {'some-uuid': config}
        sut.scope.$parent.process.resources = {'some-uuid': runnable}
        sut.route.uuid = 'some-uuid'
        sut.scope.__init__()

        sut.scope.onDeleteResource()

        expect(sut.scope.$parent.configuration.resources).toEqual({})
        expect(sut.$location.path()).toBe('/resources')
    })

    /*
    It can delete the monitor resource.
    */
    test('onDeleteResource_monitor', function() {
        var sut = mk.getController('resourceCtrl')
        mk.minimalProcess(sut)
        var config = {
            'uuid': 'some-uuid',
            'type': 'monitor',
            'name': 'Local Monitor',
            'source_path': 'some/path',
        }
        var runnable = {
            'uuid': 'some-uuid',
            'type': 'monitor',
            'name': 'Local Monitor',
            'state': 'stopped',
        }
        sut.scope.$parent.configuration.resources = {'some-uuid': config}
        sut.scope.$parent.process.resources = {'some-uuid': runnable}
        sut.route.uuid = 'some-uuid'
        sut.scope.__init__()

        sut.scope.onDeleteResource()

        expect(sut.scope.$parent.configuration.resources).toEqual({})
        expect(sut.$location.path()).toBe('/resources')
    })

})
