(function(exports){

exports.resourcesCtrl = function($scope, utils, componentTypeFilter, $timeout) {

    var init = require('../../../js_commons/__init__.js', 'init')
    // The current list of filtered and sorted resources
    // Will be initialized later.
    $scope.filtered_resources = []
    // Used by the search input tag.
    $scope.filter_input = ''

    $scope.$parent.component = {
        'name': 'Resources',
        'summary': (
            'Specialized embedded components supporting various file ' +
            'transfer operations.'
        ),
    }

    var category = $scope.$parent.component_categories.resources
    var _auth_uuid = utils.constants.CONFIGURATION_DEFAULT_AUTHENTICATION
    var details_getter = {}

    // List of service types that are considered resources.
    var _resource_services = [
        utils.constants.TYPE_NAME.LOCAL_MANAGER,
        utils.constants.TYPE_NAME.LOCAL_MANAGER_UNSECURED
        ]

    /*
    This is called each time the page is accessed,
    including when using the browser back/forward.
    */
    $scope.__init__ = function() {
        // Trigger an initial filter that will show all.
        $scope.onFilterResources()

        // The configuration might not be loaded so we force a dumb reload
        // We should find a better way to auto-reload the page but for
        // now it should be ok.
        // In an ideal case, we should use a watch on the configuration,
        // but that might have some side-effects and performance issues.
        // Most of the times, users should not land directly on the resources
        // page, so this hack should be ok.
        if (init.isEmpty($scope.filtered_resources)) {
            $timeout(function() {$scope.onFilterResources()}, 3000)
        }
    }

    /*
    Go to configuration page for `target` configuration.

    Handles all the configuration types available on the resources page.
     */
    $scope.onGoToConfiguration = function(target) {
        var categories = $scope.$parent.component_categories

        if (target.type == utils.constants.TYPE_NAME.APPLICATION) {
            // Go to the authentication section.
            categories.authentications.goToConfiguration(target)
            return
        }

        if (
            (target.type == utils.constants.TYPE_NAME.LOCAL_MANAGER) ||
            (target.type == utils.constants.TYPE_NAME.LOCAL_MANAGER_UNSECURED)
            ) {
            // Go to  services section for manager services,
            // as they are not yet moved to resources.
            categories.services.goToConfiguration(target)
            return
        }

        category.goToConfiguration(target)
    }

    /*
    Return a list with a {config, parent_runnables} object
    with all the available resources that can be visible on the
    resources index page.

    This includes the monitor and default authentication.
    */
    var getAllResources = function() {
        var result = []

        // Then normal processes
        angular.forEach(
            $scope.getEntities('services'),
            function(configuration) {
                if (!_resource_services.contains(configuration.type)) {
                    return
                }
                result.push({
                    'configuration': configuration,
                    'parent_runnables': $scope.process.runnables.services
                })
            })

        // Then normal processes
        angular.forEach(
            $scope.getEntities('resources'),
            function(configuration) {
                result.push({
                    'configuration': configuration,
                    'parent_runnables': $scope.process.runnables.resources
                })
        })

        // Add the authentication last as it has less importance for the
        // configuration.
        var authentication = $scope.configuration.authentications[_auth_uuid]
        result.push({
            'configuration': authentication,
            'parent_runnables': $scope.process.runnables.authentications
        })

        return result
    }

    /*
    Called the update the list of resources to be displayed on the page.
    */
    $scope.onFilterResources = function(filter_value) {
        $scope.filtered_resources = []

        if (!$scope.configuration.resources) {
            // Configuration not yet loaded.
            return
        }

        angular.forEach(
            getAllResources(),
            function(result) {
                var view_model = getFilteredResourceView(
                    filter_value, result.configuration, result.parent_runnables)
                if (!view_model) {
                    // Most probably filtered out.
                    return
                }
                $scope.filtered_resources.push(view_model)
            })
    }

    /*
    Update the `resources` based on the current filter.
    */
    var getFilteredResourceView = function(
            filter_value, configuration, runnables) {

        if (utils.isComponentExcluded(filter_value, configuration)) {
            return
        }

        var view_model = utils.getBaseViewModel(configuration, runnables)
        return {
            'uuid': view_model.uuid,
            'details': getResourceDetails(configuration),
            'enabled_text': $scope.getEntityStartupText(configuration),
            'is_delete_blocked': utils.isDeleteBlocked(
                configuration, 'resource'),
            'config': configuration,
            'runnable': view_model.runnable
        }
    }

    /*
    Return the details to be visible in the UI for each resource type.
    */
    var getResourceDetails = function(configuration) {

        var type_description = componentTypeFilter(
            configuration.type, utils.constants.TYPE_NAME.RESOURCE)
        if (_resource_services.contains(configuration.type)) {
            type_description = componentTypeFilter(
                configuration.type, utils.constants.TYPE_NAME.SERVICE)
        }
        var result = [{
            'key': 'Type',
            'value': type_description
        }]

        var get_details = details_getter[configuration.type]
        if (get_details) {
            Array.prototype.push.apply(
                result, get_details(configuration))
        }

        return result
    }

    details_getter.manager = function(configuration) {
        var result = [
            {'key': '',
            'value': 'The configuration for this web console network access.' },
            {'key': 'Listening address',
            'value': utils.getIPAddressDescription(configuration.address)},
            {'key': 'Port',
            'value': configuration.port},
        ]
        if (configuration.configuration.ssl_domains) {
            result.push({
                'key': 'Let\'s Encrypt',
                'value': configuration.configuration.ssl_domains.join(', ')
            })
        }
        return result

    }
    details_getter._manager_unsecured = details_getter.manager

    details_getter.application = function(configuration) {
        return [
            {'key': '',
            'value': 'Embedded authentication for users and administrators.' }
            ]
    }

    details_getter.sqlite = function(configuration) {
        return [
            {'key': '',
            'value': 'Embedded database used by file transfer processes.' },
            {'key': 'DB path',
            'value': configuration.path }
            ]
    }

    details_getter.analytics = function(configuration) {
        return [
            {'key': '',
            'value': 'Monitoring and analytics for transfer processes.' },
            {'key': 'Monitor internal',
            'value': configuration.monitor_interval + ' seconds' }
            ]
    }

    details_getter['email-client'] = function(configuration) {
        return [
            {'key': '',
            'value': 'Integration with an email server for sending emails.' },
            {'key': 'Server address',
            'value': configuration.address + ':' + configuration.port
                },
            {'key': 'Source address',
            'value': configuration.email_from_address
                },
            {'key': 'Default recipients',
            'value': configuration.email_to_recipients.join(', ')
                }
            ]
    }

    details_getter['lets-encrypt'] = function(configuration) {
        return [
            {'key': '',
            'value': 'Automatically obtain and renew TLS certificates.' },
            {'key': 'Validation own address',
            'value':configuration.address + ':' + configuration.port
            },
            {'key': 'Lets Encrypt URL',
            'value': configuration.acme_url
                }
            ]
    }

    details_getter.monitor = function(configuration) {
        var result = [
            {
                'key': 'Path',
                'value': configuration.source_path
            },
            {
                'key': 'Recursive',
                'value': configuration.recursive ? 'Yes' : ''
            },
            {
                'key': 'Filter',
                'value': configuration.source_filter
            }
        ]
        return result
    }

    /*
    Delete a resource.
    */
    $scope.onDeleteResource = function(configuration) {
        category.delete(configuration)
        $scope.onFilterResources($scope.filter_input)
    }

    /*
    Called when add monitor button is triggered.
    */
    $scope.onAddMonitor = function() {
        var resources = $scope.$parent.component_categories.resources
        var resource = {
            'enabled': true,
            'type': utils.constants.TYPE_NAME.MONITOR,
            'source_path': 'ADD YOUR MONITORED PATH HERE',
            'changes_poll_interval': 120,
            'stable_interval': 241,
            'retry_count': 5,
            'monitored_operations': ['disabled'],
            'file_age_notification': 0,
            'file_age_auto_delete': 0,
            'recursive': false
        }
        resource['name'] = 'New filesystem monitor %(random)s'.supplant({
            'random': utils.getRandomMark()
            })
        resources.add(resource)
        category.goToConfiguration(resource)
    }

}
/* c8 ignore next */
}(typeof exports === 'undefined' ? this.resourcesCtrl = {} : exports))
