/*
Tests for resources controller.
*/
describe('ResourcesCtrl', function() {

    /*
    Nothing happens when loading the page, even when no resources
    are yet available.
    */
    test('__init__', function() {
        var sut = mk.getController('resourcesCtrl')
        sut.scope.__init__()
    })

    /*
    It can remove an existing resource.
    */
    test('onDeleteResource_resource', function() {
        var sut = mk.getController('resourcesCtrl')
        mk.minimalProcess(sut)
        var email_client = {
            'uuid': 'some-uuid',
            'type': 'email-client',
            'name': 'An email client',
            'email_to_recipients': [],
        }
        sut.scope.$parent.configuration.resources['some-uuid'] = email_client
        sut.scope.onDeleteResource(email_client)

        expect(sut.scope.configuration.resources['some-uuid']).toBeUndefined()
    })

    /*
    It can remove an existing monitor service.
    */
    test('onDeleteResource_monitor', function() {
        var sut = mk.getController('resourcesCtrl')
        mk.minimalProcess(sut)
        var monitor = {
            'uuid': 'some-uuid',
            'type': 'monitor',
            'name': 'A monitor',
        }
        sut.scope.$parent.configuration.resources['some-uuid'] = monitor
        sut.scope.onDeleteResource(monitor)

        expect(sut.scope.configuration.resources['some-uuid']).toBeUndefined()
    })

    /*
    It can remove an existing monitor service.
    */
    test('onDeleteResource_manager_unsercured', function() {
        var sut = mk.getController('resourcesCtrl')
        mk.minimalProcess(sut)
        var manager = {
            'uuid': 'some-uuid',
            'type': '_manager_unsecured',
            'name': 'A manager',
        }
        sut.scope.$parent.configuration.resources['some-uuid'] = manager
        sut.scope.onDeleteResource(manager)

        expect(sut.scope.configuration.resources['some-uuid']).toBeUndefined()
    })

    /*
    It can add a monitor service.
    */
    test('onAddMonitor', function() {
        var sut = mk.getController('resourcesCtrl')
        mk.minimalProcess(sut)
        var new_resources = sut.scope.$parent.new_entities.resources

        expect(new_resources).toEqual([])
        sut.scope.onAddMonitor()
        expect(new_resources).toHaveLength(1)
        expect(new_resources[0].name).toStartWith('New filesystem monitor ')
        expect(new_resources[0]).toContainEntries([
            ['type', 'monitor'],
            ['enabled', true],
            ['source_path', 'ADD YOUR MONITORED PATH HERE'],
            ['changes_poll_interval', 120],
            ['stable_interval', 241],
            ['monitored_operations', ['disabled']],
            ['file_age_notification', 0],
            ['file_age_auto_delete', 0],
            ['recursive', false],
        ])
    })

    /*
    When no filter is provided,
    Will include the local manager and monitor services, and as lost item,
    the default authentication.
    */
    test('onFilterResources_no_filter', function() {
        var sut = mk.getController('resourcesCtrl')
        mk.minimalProcess(sut)
        sut.scope.$parent.configuration.resources['monitor-uuid'] = {
            'uuid': 'monitor-uuid',
            'type': 'monitor',
            'name': 'Some monitor service',
            'source_path': 'some/path',
            'recursive': true,
            'source_filter': '*.txt',
        }
        sut.scope.$parent.process.runnables.resources['monitor-uuid'] = {
            'uuid': 'monitor-uuid',
            'type': 'monitor',
            'name': 'Some monitor service',
        }

        sut.scope.onFilterResources()

        expect(sut.scope.filtered_resources).toHaveLength(7)
        expect(sut.scope.filtered_resources[0].config.name).toEqual(
            'Zoro Default Web Manager')
        expect(sut.scope.filtered_resources[1].config.name).toEqual(
            'Analitics resources')
        expect(sut.scope.filtered_resources[2].config.name).toEqual(
            'Email client')
        expect(sut.scope.filtered_resources[3].config.name).toEqual(
            'Lets Encrypt')
        expect(sut.scope.filtered_resources[4].config.name).toEqual(
            'Some monitor service')
        expect(sut.scope.filtered_resources[5].config.name).toEqual(
            'SQlite DB')
        expect(sut.scope.filtered_resources[6].config.name).toEqual(
            'A Default Auth Method')
    })

    /*
    Will filter case insentive starting with 3 letter.
    */
    test('onFilterResources_filter', function() {
        var sut = mk.getController('resourcesCtrl')
        mk.minimalProcess(sut)

        sut.scope.onFilterResources('def')

        expect(sut.scope.filtered_resources).toHaveLength(2)
        expect(sut.scope.filtered_resources[0].config.name).toEqual(
            'Zoro Default Web Manager')
        expect(sut.scope.filtered_resources[1].config.name).toEqual(
            'A Default Auth Method')
    })

    /*
    Will got to the configuration page in the resource URL
    */
    test('onGoToConfiguration_resource', function() {
        var sut = mk.getController('resourcesCtrl')
        mk.minimalProcess(sut)
        var config = {
            'uuid': 'some-email-uuid',
            'type': 'email-client',
            'name': 'An email client',
            'email_to_recipients': [],
        }

        sut.scope.configuration.resources['some-email-uuid'] = config
        sut.scope.onGoToConfiguration(config)

        expect(sut.$location.path()).toEqual('/resources/some-email-uuid')
    })

    /*
    Monitor services are configured using the resource URL space.
    */
    test('onGoToConfiguration_monitor', function() {
        var sut = mk.getController('resourcesCtrl')
        mk.minimalProcess(sut)
        var config = {
            'uuid': 'some-monitor-uuid',
            'type': 'monitor',
            'name': 'A monitor',
        }

        sut.scope.configuration.resources['some-monitor-uuid'] = config
        sut.scope.onGoToConfiguration(config)

        expect(sut.$location.path()).toEqual('/resources/some-monitor-uuid')
    })

    /*
    Will redirect to the authentication edit on the auth URL.
    */
    test('onGoToConfiguration_authentication', function() {
        var sut = mk.getController('resourcesCtrl')
        mk.minimalProcess(sut)
        var config = {
            'uuid': 'some-auth-uuid',
            'type': 'application',
            'name': 'The app auth',
        }

        sut.scope.configuration.authentications['some-auth-uuid'] = config
        sut.scope.onGoToConfiguration(config)

        expect(sut.$location.path()).toEqual('/authentications/some-auth-uuid')
    })

    /*
    Will got to the configuration page in the services URL
    */
    test('onGoToConfiguration_manager', function() {
        var sut = mk.getController('resourcesCtrl')
        mk.minimalProcess(sut)
        var config = {
            'uuid': 'some-lm-uuid',
            'type': 'manager',
            'name': 'The LM',
        }

        sut.scope.configuration.services['some-lm-uuid'] = config
        sut.scope.onGoToConfiguration(config)

        expect(sut.$location.path()).toEqual('/services/some-lm-uuid')
    })

    /*
    Will got to the configuration page in the services URL
    */
    test('onGoToConfiguration_manager_unsecured', function() {
        var sut = mk.getController('resourcesCtrl')
        mk.minimalProcess(sut)
        var config = {
            'uuid': 'some-lmu-uuid',
            'type': '_manager_unsecured',
            'name': 'The LM',
        }

        sut.scope.configuration.services['some-lmu-uuid'] = config
        sut.scope.onGoToConfiguration(config)

        expect(sut.$location.path()).toEqual('/services/some-lmu-uuid')
    })
})
