/*
Tests for review controller.
*/


describe('reviewCtrl', function() {

    /*
    For random operations it returns the full path.
    */
    test('getChangeName_new', function() {
        var change = {
            'operation': 'not-create',
            'path': 'some/path'
            }
        var sut = mk.getController('reviewCtrl')

        var result = sut.scope.getChangeName(change)

        expect(result).toBe('some > path')
    })

    /*
    Can detect a change for a service configuration.
    */
    test('getChangeType_services', function() {
        var change = {
            'operation': 'not-create',
            'path': 'services/path'
            }
        var sut = mk.getController('reviewCtrl')

        var result = sut.scope.getChangeType(change)

        expect(result).toBe('Protocol')
    })


    /*
    For create changes is return the path with change name appended.
    */
    test('getChangeName_create', function() {
        var change = {
            'operation': 'create',
            'path': 'accounts/',
            'value': {'name': 'some-name'}
            }
        var sut = mk.getController('reviewCtrl')

        var result = sut.scope.getChangeName(change)

        expect(result).toBe('some-name')
    })

    /*
    For update of node_variables to the base server config, it returns
    the change as it was done directly to the server.
    */
    test('getChangeName_update_node_variables_server', function() {
        var change = {
            'operation': 'update',
            'path': 'node_variables/server|name/value',
            'value': {'old': 'some-name', 'new': 'some-other'}
            }
        var sut = mk.getController('reviewCtrl')

        var result = sut.scope.getChangeName(change)

        expect(result).toBe('Server > name')
    })

    /*
    For update of node_variables to a service, it returns the name of the
    service and the updated variable name.
    */
    test('getChangeName_update_node_variables_service', function() {
        var sut = mk.getController('reviewCtrl')
        var change = {
            'operation': 'update',
            'path': 'node_variables/services|ftp-uuid|name/value',
            'value': {'old': 'some-name', 'new': 'some-other'}
            }
        sut.scope.$parent.configuration.services = {'ftp-uuid': {
            'name': 'FTP Srv'}
        }

        var result = sut.scope.getChangeName(change)

        expect(result).toBe('FTP Srv > name')
    })

    /*
    For delete of node_variables to a service, it returns the name of the
    service.
    */
    test('getChangeName_delete_node_variables_service', function() {
        var sut = mk.getController('reviewCtrl')
        var change = {
            'operation': 'delete',
            'path': 'node_variables/services|ftp-uuid|name',
            'value': {'old': 'some-name', 'new': 'some-other'}
            }
        sut.scope.$parent.configuration.services = {'ftp-uuid': {
            'name': 'FTP Srv'}
        }

        var result = sut.scope.getChangeName(change)

        expect(result).toBe('FTP Srv > name')
    })


    describe('getChangeOldValue', function() {

        test(
        'For delete operations just returns as removed.',
        function() {
            var change = {
                'operation': 'delete',
                'path': 'some/path'
                }
            var sut = mk.getController('reviewCtrl')

            var result = sut.scope.getChangeOldValue(change)

            expect(result).toBe('Just removed.')
        })

        test(
        'When changing password returns masked value.',
        function() {
            var change = {
                'path': 'accounts/password'
                }
            var sut = mk.getController('reviewCtrl')

            var result = sut.scope.getChangeOldValue(change)

            expect(result).toBe('Masked value')
        })

        test(
        'When changing from an empty value, returns EMPTY.',
        function() {
            var change = {
                'old': '',
                'path': 'some/path'
                }
            var sut = mk.getController('reviewCtrl')

            var result = sut.scope.getChangeOldValue(change)

            expect(result).toBe('EMPTY')
        })

        test(
        'When changing from an undefined value, returns NOT DEFINED.',
        function() {
            var change = {
                'old': null,
                'path': 'some/path'
                }
            var sut = mk.getController('reviewCtrl')

            var result = sut.scope.getChangeOldValue(change)

            expect(result).toBe('NOT DEFINED')
        })


    })


    describe('getChangeNewValue', function() {

        test(
        'For create operations just returns as created.',
        function() {
            var change = {
                'operation': 'create',
                'path': 'some/path'
                }
            var sut = mk.getController('reviewCtrl')

            var result = sut.scope.getChangeNewValue(change)

            expect(result).toBe('Just created.')
        })

        test(
        'When changing password returns masked value.',
        function() {
            var change = {
                'path': 'accounts/password'
                }
            var sut = mk.getController('reviewCtrl')

            var result = sut.scope.getChangeNewValue(change)

            expect(result).toBe('Masked value')
        })

        test(
        'When changing to an empty value, returns EMPTY.',
        function() {
            var change = {
                'value': '',
                'path': 'some/path'
                }
            var sut = mk.getController('reviewCtrl')

            var result = sut.scope.getChangeNewValue(change)

            expect(result).toBe('EMPTY')
        })

        test(
        'When changing to an undefined value, returns NOT DEFINED.',
        function() {
            var change = {
                'value': null,
                'path': 'some/path'
                }
            var sut = mk.getController('reviewCtrl')

            var result = sut.scope.getChangeNewValue(change)

            expect(result).toBe('NOT DEFINED')
        })

    })



    /*
    It goes to the authentications index page for general server auth config.
    */
    test('onGoToChange_general_server_auth', function() {

        var sut = mk.getController('reviewCtrl')
        var change = {
            'path': 'server/authentications',
            'value': 'ignored'
        }
        expect(sut.$location.path()).toBe('')

        sut.scope.onGoToChange(change)

        expect(sut.$location.path()).toBe('/authentications')
    })

    /*
    It goes to the general server page for general server configurations.
    */
    test('onGoToChange_general_server_option', function() {

        var sut = mk.getController('reviewCtrl')
        var change = {
            'path': 'server/name',
            'value': 'new-name'
        }
        expect(sut.$location.path()).toBe('')

        sut.scope.onGoToChange(change)

        expect(sut.$location.path()).toBe('/server')
    })

    /*
    It goes to the specific component page for an updated value.

    Underscores from config values are replaced with dashed for URL compoment
    familty.
    The UUID is not changed.
    */
    test('onGoToChange_component_update', function() {

        var sut = mk.getController('reviewCtrl')
        var change = {
            'path': 'event_handlers/some-uuid_here/name',
            'operation': 'update',
            'value': 'new-name'
        }
        expect(sut.$location.path()).toBe('')

        sut.scope.onGoToChange(change)

        expect(sut.$location.path()).toBe('/event-handlers/some-uuid_here')
    })

    /*
    It goes to the specific component page based on name for newly
    created components.

    Underscores from config values are replaced with dashed for URL.
    */
    test('onGoToChange_component_create', function() {

        var sut = mk.getController('reviewCtrl')
        var change = {
            'path': 'authentications',
            'operation': 'create',
            'value': {'name': 'New Auth-name_1'}
        }
        expect(sut.$location.path()).toBe('')

        sut.scope.onGoToChange(change)

        expect(sut.$location.path()).toBe('/authentications/New Auth-name_1')
    })


    describe('_getChangeLink', function() {

        test(
        'Link for server and reports.',
        function() {
            var change = {
                'path': 'server/name',
                'value': 'some-other-name',
                'operation': 'update'
                }
            var sut = mk.getController('reviewCtrl')

            var result = sut.scope._getChangeLink(change)

            expect(result).toEqual('server')

            change = {
                'path': 'reports/database',
                'value': 'some-other-name',
                'operation': 'update'
                }

            var result = sut.scope._getChangeLink(change)

            expect(result).toEqual('server')
        })

        test(
        'Link for identity members.',
        function() {
            var change = {
                'path': 'roles/role-uuid/name',
                'value': 'some-name',
                'operation': 'update'
                }
            var sut = mk.getController('reviewCtrl')

            var result = sut.scope._getChangeLink(change)

            expect(result).toEqual('roles/role-uuid')
        })

    })


    describe('_getChangeEntity', function() {

        test(
        'Return null for an unknown path.',
        function() {
            var change = {
                'path': 'unknow-root/some-path'
                }
            var sut = mk.getController('reviewCtrl')

            var result = sut.scope._getChangeEntity(change)

            expect(result).toBe(null)
        })

        test(
        'Return null for server configuration.',
        function() {
            var change = {
                'path': 'server/name'
                }
            var sut = mk.getController('reviewCtrl')

            var result = sut.scope._getChangeEntity(change)

            expect(result).toBe(null)
        })

        test(
        'Return configured entity if found.',
        function() {
            var sut = mk.getController('reviewCtrl')
            var entity = {
                'name': 'other-name',
                'uuid': 'some-uuid'
                }
            sut.scope.$parent.configuration = {
                'entities': {
                    'some-uuid': entity
                    }
                }

            var change = {
                'path': 'entities/some-uuid/name'
                }

            var result = sut.scope._getChangeEntity(change)

            expect(result).toBe(entity)
        })

        test(
        'Return entry from identity configuration.',
        function() {
            var sut = mk.getController('reviewCtrl')
            var role = {
                'name': 'other-name',
                'uuid': 'role-uuid'
                }
            sut.scope.$parent.configuration = {'roles': {
                        'role-uuid': role
                        }
                    }
            var change = {
                'path': 'roles/role-uuid'
                }

            var result = sut.scope._getChangeEntity(change)

            expect(result).toEqual(role)
        })

        test(
        'Return null for an unknown identity type.',
        function() {
            var sut = mk.getController('reviewCtrl')
            sut.scope.$parent.configuration = {
                    'roles': {}
                    }
            var change = {
                'path': 'unknown-type/uuid'
                }

            var result = sut.scope._getChangeEntity(change)

            expect(result).toBe(null)
        })

        test(
        'Return undefined when identity entry is not configured.',
        function() {
            var sut = mk.getController('reviewCtrl')
            sut.scope.$parent.configuration = {
                    'roles': {}
                    }
            var change = {
                'path': 'roles/uuid'
                }

            var result = sut.scope._getChangeEntity(change)

            expect(result).toBeUndefined()
        })


    })

    /*
    Returns a human readable text describing the node variables change
    category.
    */
    test('getChangeType_node_variables', function() {
        var sut = mk.getController('reviewCtrl')
        var change = {
            'operation': 'delete',
            'path': 'node_variables/services/ftp-uuid/name',
            'value': {}
            }

        var result = sut.scope.getChangeType(change)

        expect(result).toBe('Node variables')
    })


    /*
    Returns a human readable name for  the event handlers changes.
    */
    test('getChangeType_event_handler', function() {
        var sut = mk.getController('reviewCtrl')
        var change = {
            'operation': 'delete',
            'path': 'event_handlers/uuid',
            'value': {}
            }

        var result = sut.scope.getChangeType(change)

        expect(result).toBe('Event Handler')
    })

    /*
    Returns a generic text when the change type is not expected.

    This is done not to generate a failure.
    */
    test('getChangeType_generic', function() {
        var sut = mk.getController('reviewCtrl')
        var change = {
            'operation': 'delete',
            'path': 'random-root/uuid',
            'value': {}
            }

        var result = sut.scope.getChangeType(change)

        expect(result).toBe('General')
    })


})
