/*
Tests for service controller.
*/


describe('ServiceCtrl', function() {

    /*
    Add a new service to the existing upstream configuration and return
    the new service.
    */
    var createService = function(sut, kind) {
        var uuid = 'some-uuid'
        var service = {
            'uuid': uuid,
            'name': 'one-service',
            'type': kind,
            'configuration': {'option': 'value'}
            }
        sut.scope.$parent.configuration = {
            'services': {'some-uuid': service }
            }
        // Also set the runnable.
        sut.scope.$parent.process.runnables = {
            'services': {'some-uuid': {'uuid': uuid} }
            }
        sut.route.uuid = uuid

        sut.scope.__init__()
        return service

    }

    /*
    Redirects to services when the service was not found.
    */
    test('init_not_found', function() {
        var sut = mk.getController('serviceCtrl')
        sut.route.uuid = 'no-such-service'
        sut.scope.$parent.configuration['services'] = {'some-uuid': {}}

        sut.scope.__init__()

        expect(sut.$location.path()).toBe('/services')
    })

    /*
    Load template for implicit ftps as normal FTP.

    It has connection UI.
    */
    test('ftps_implicit_as_ftp', function() {
        var sut = mk.getController('serviceCtrl')
        var service = createService(sut, 'ftpsi')

        expect(sut.scope.template_path).toBe(
            'partials/service_ftp.html?v=%(cache)s')
        // It has the SSL configuration alias.
        expect(sut.scope.ssl_configuration).toBe(service.configuration)
    })

    /*
    Load http service and will set headers as text value for UI.'
    */
    test('init_http', function() {
        var sut = mk.getController('serviceCtrl')
        var uuid = mk.string()
        var service = {
            'uuid': uuid,
            'name': 'one-service',
            'type': 'http',
            'authentications': [],
            'configuration': {
                'headers': [['one: header'], ['other: header']]}
            }
        sut.scope.$parent.configuration = {
            'services': {uuid: service }
            }
        // Also set the runnable.
        sut.scope.$parent.process.runnables = {
            'services': {uuid: {'uuid': uuid} }
        }
        sut.route.uuid = uuid

        sut.scope.__init__()
        expect(sut.scope.service).toBe(service)
        expect(sut.scope.service_configuration).toBe(service.configuration)
        expect(sut.scope.template_path).toBe(
            'partials/service_http.html?v=%(cache)s')
    })

    /*
    Retun True of False depending on the current value of the service.address.
    */
    test('hasCustomAddress', function() {
        var sut = mk.getController('serviceCtrl')
        var service = createService(sut, 'sftp')

        service.address = '0.0.0.0'
        expect(sut.scope.hasCustomAddress()).toBeFalse()

        service.address = '::'
        expect(sut.scope.hasCustomAddress()).toBeFalse()

        service.address = '10.1.2.3'
        expect(sut.scope.hasCustomAddress()).toBeTrue()
    })

    /*
    Will set the service address to a non default value.
    */
    test('onCustomAddress', function() {
        var sut = mk.getController('serviceCtrl')
        var service = createService(sut, 'sftp')

        service.address = '0.0.0.0'
        sut.scope.onCustomAddress()

        expect(sut.scope.service.address).toEqual('192.168.0.123')
    })

    /*
    It will remove the current service from the config and
    moved to services index page.
    */
    test('onDeleteService_existing', function() {
        var sut = mk.getController('serviceCtrl')
        var service = createService(sut, 'ssh')
        var services = sut.scope.$parent.configuration.services
        expect(services).toContainKey(service.uuid)

        sut.scope.onDeleteService()

        expect(services).not.toContainKey(service.uuid)
        expect(sut.$location.path()).toBe('/services')

    })

})
