/*
Tests for services controller.
*/

describe('ServicesCtrl', function() {

    describe('onDeleteService', function() {

        test(
        'For services that already created and have an UUID, the service ' +
        'is removed from the configuration object.',
        function() {
            var sut = mk.getController('servicesCtrl')
            var uuid = mk.string()
            var service = {'uuid': uuid}
            sut.scope.$parent.configuration['services'] = {}
            sut.scope.$parent.configuration['services'][uuid] = service

            sut.scope.onDeleteService(service)

            expect(
                sut.scope.$parent.configuration['services'][uuid]
                ).toBeUndefined()
        })

        test(
        'When service does not exists, it does nothing.',
        function() {
            var sut = mk.getController('servicesCtrl')
            var uuid = mk.string()
            var service = {'uuid': uuid}
            var other_service = {'uuid': 'other'}
            sut.scope.$parent.configuration['services'] = {}
            sut.scope.$parent.configuration['services'][uuid] = service
            sut.scope.$parent.process.runnables = {'services': {}}

            sut.scope.onDeleteService(other_service)

            expect(
                sut.scope.$parent.configuration['services'][uuid]
                ).toBe(service)
        })

        test(
        'For new services, it removes the service from the list of new ' +
        'services.',
        function() {
            var sut = mk.getController('servicesCtrl')
            var service = {'name': 'new_service_without_uuid'}
            var first_service = {'name': 'new_first_sevice'}
            var last_service = {'name': 'new_last_sevice'}
            sut.scope.$parent.new_entities['services'] = [
                first_service, service, last_service]

            sut.scope.onDeleteService(service)

            expect(
                sut.scope.$parent.new_entities['services']
                ).toEqual([first_service, last_service])
        })

    })


    describe('onServiceAdd', function() {

        test(
        'Creates a new service and populate it with default basic values. ' +
        'No configuration is created if no default values are available ' +
        'for protocol.',
        function() {
            var sut = mk.getController('servicesCtrl')
            var new_services = []
            sut.scope.$parent.new_entities['services'] = new_services

            sut.scope.onServiceAdd('other-protocol')

            expect(new_services.length).toBe(1)
            var new_service = new_services[0]
            expect(new_service['name']).toStartWith(
                'New other-protocol Service ')
            expect(new_service['enabled']).toBeTrue()
            expect(new_service['type']).toBe('other-protocol')
            expect(new_service['address']).toBe('0.0.0.0')
            expect(new_service['port']).toBe(12345)
            expect(new_service['configuration']).toEqual({
                'idle_connection_timeout': 300,
            })
        })

        /*
        For FTP and FTPSI service it initiates a valid passive port range
        and SSL configuration.
        */
        test('onServiceAdd_ftp', function() {
            var sut = mk.getController('servicesCtrl')
            var new_services = []
            sut.scope.$parent.new_entities['services'] = new_services

            sut.scope.onServiceAdd('ftp')

            var new_service = new_services[0]
            expect(new_service['type']).toStartWith('ftp')
            expect(new_service['configuration']).toEqual({
                'passive_port_range': [0, 0],
                'ssl_cipher_list': 'secure',
                'ssl_certificate': '',
                'idle_connection_timeout': 300,
                'idle_data_connection_timeout': 30,
                'ftps_require_session_reuse': true,
                })
        })

        test(
        'For SSH it initiates the default value for the ciphers list.',
        function() {
            var sut = mk.getController('servicesCtrl')
            var new_services = []
            sut.scope.$parent.new_entities['services'] = new_services

            sut.scope.onServiceAdd('ssh')

            var new_service = new_services[0]
            expect(new_service['type']).toStartWith('ssh')
            expect(new_service['configuration']).toEqual({
                'ssh_cipher_list': ['secure'],
                'dh_prime_size': ['1024', 'ideal'],
                'idle_connection_timeout': 300,
                })
        })

        /*
        For HTTPS service it initiates HTTP configuration
        and SSL configuration.
        */
        test('onServiceAdd_https', function() {
                var sut = mk.getController('servicesCtrl')
                var new_services = []
                sut.scope.$parent.new_entities['services'] = new_services

                sut.scope.onServiceAdd('https')

                var new_service = new_services[0]
                expect(new_service['type']).toStartWith('http')
                expect(new_service['configuration']).toEqual({
                    'as2_receive_name': 'as2receive',
                    'languages': ['en'],
                    'public_account_uuid': '',
                    'public_name': 'public',
                    'ui_version': 'ui-gen-2',
                    'login_footnotes': '',
                    'ssl_cipher_list': 'secure',
                    'ssl_certificate': '',
                    'idle_connection_timeout': 300,
                    })
            })

        test(
        'For Lets Encrypt service it initiates with public URL and an ' +
        'example for domains config.',
        function() {
            var sut = mk.getController('servicesCtrl')
            sut.utils.constants['CONFIGURATION_DEFAULTS']['acme_url'] = (
                'http://api.acme.com/dir')
            var new_services = []
            sut.scope.$parent.new_entities['services'] = new_services

            sut.scope.onServiceAdd('lets-encrypt')

            var new_service = new_services[0]
            expect(new_service['type']).toEqual('lets-encrypt')
            expect(new_service['configuration']).toEqual({
                'acme_url': 'http://api.acme.com/dir',
                'contact_email': '',
                'domains': [[
                    'files.example.com', '1c17-44-878c', '17c9-7a-2f35']],
                'idle_connection_timeout': 300,
                })
        })

    })

    /*
    When no filter is provided it will have available all the services,
    including newly created methods.

    This is also a test to check the details getter for services.
    */
    test('onFilterServices_all', function() {
        var sut = mk.getController('servicesCtrl')

        sut.scope.$parent.configuration.services = {
            'uuid-1': {
                'uuid': 'uuid1',
                'type': 'https',
                'address': '0.0.0.0',
                'port': 10080,
                'idle_connection_timeout': 2345,
                'configuration': {
                    'as2_receive_name': '/path/as2receive.pub',
                    'ssl_domains': ['web.test.com', 'files.test.org'],
                },
            },
            'uuid-2': {
                'uuid': 'uuid-2',
                'type': 'manager',
                'address': '127.0.0.1',
                'port': 10020,
                'idle_connection_timeout': 25,
                'configuration': {
                    'ssl_domains': ['admin.test.com', 'adm.test.org'],
                },
            },
            'uuid-ignored': {
                'uuid': 'uuid-ignored',
                'type': 'monitor',
                'address': '0.0.0.0',
                'port': 0,
                'idle_connection_timeout': 134,
                'configuration': {},
            },
            'uuid-3': {
                'uuid': 'uuid3',
                'type': 'ssh',
                'address': '1.2.3.4',
                'port': 1234,
                'idle_connection_timeout': 42,
                'configuration': {
                    'sftp': true,
                    'scp': false,
                },
            },
        }
        sut.scope.$parent.process.runnables = {}
        sut.scope.$parent.process.runnables.services = {
            'uuid-1': {'uuid': 'uuid1', 'name': 'Runnable 1'},
            'uuid-2': {'uuid': 'uuid2', 'name': 'Runnable 2'},
        }
        sut.scope.$parent.new_entities.services = [{
            'name': 'new auth 3',
            'type': 'ftp',
            'address': '::',
            'port': 10021,
            'idle_connection_timeout': 24,
            'configuration': {
                'ftps_explicit_enabled': true,
                'passive_address': '1.2.3.4',
                'passive_port_range': [1230, 1330],
                'idle_data_connection_timeout': 982,
                'ssl_domains': ['ftp.example.com'],
            },
        }]

        sut.scope.onFilterServices()

        // The monitor service is not listed here.
        expect(sut.scope.filtered_services).toHaveLength(4)
        expect(sut.scope.filtered_services[0].details).toEqual([
            {"key": "Type", "value": "FTP and Explicit FTPS"},
            {"key": "Port", "value": 10021},
            {"key": "Listening", "value": "All IPv6"},
            {"key": "Timeout", "value": 24},
            {"key": "FTPS Secure", "value": "Enabled"},
            {"key": "Passive address", "value": "1.2.3.4"},
            {"key": "Passive range", "value": "1230 - 1330"},
            {"key": "Data timeout", "value": "982s"},
            {"key": "Let's Encrypt", "value": "ftp.example.com"},
        ])
        expect(sut.scope.filtered_services[1].details).toEqual([
            {"key": "Type", "value": "HTTPS"},
            {"key": "Port", "value": 10080},
            {"key": "Listening", "value": "All IPv4"},
            {"key": "Timeout", "value": 2345},
            {"key": "Let's Encrypt", "value": "web.test.com, files.test.org"},
            {"key": "AS2 Enabled", "value": "/path/as2receive.pub"},
        ])
        expect(sut.scope.filtered_services[2].details).toEqual([
            {"key": "Type", "value": "Local Management"},
            {"key": "Port", "value": 10020},
            {"key": "Listening", "value": "127.0.0.1"},
            {"key": "Timeout", "value": 25},
            {"key": "Let's Encrypt", "value": "admin.test.com, adm.test.org"},
        ])
        expect(sut.scope.filtered_services[3].details).toEqual([
            {"key": "Type", "value": "SSH (SFTP/SCP)"},
            {"key": "Port", "value": 1234},
            {"key": "Listening", "value": "1.2.3.4"},
            {"key": "Timeout", "value": 42},
            {"key": "SFTP", "value": "Enabled"},
            {"key": "SCP", "value": "Disabled"},
        ])

    })


})
