/*
Tests for status controller.
*/

describe('StatusCtrl', function() {

    /*
    No alerts are loaded when there is no process.
    */
    test('__init__', function() {
        var sut = mk.getController('statusCtrl')

        sut.scope.__init__()

        expect(sut.scope.$parent.alerts).toEqual([])
    })

    /*
    No alerts are loaded when there is process,
    but the proccess is not running the legacy win service.
    */
    test('__init__', function() {
        var sut = mk.getController('statusCtrl')
        mk.minimalProcess(sut)

        sut.scope.__init__()

        expect(sut.scope.$parent.alerts).toEqual([])
    })

    /*
    Show alert on windows legacy, only once.
    */
    test('__init___win_legacy_service', function() {
        var sut = mk.getController('statusCtrl')
        mk.minimalProcess(sut)
        sut.scope.$parent.process.runnables.status.windows_legacy_service = true

        sut.scope.__init__()

        expect(sut.scope.$parent.alerts).toHaveLength(1)
        expect(sut.scope.$parent.alerts[0].type).toEqual('warning')
        expect(sut.scope.$parent.alerts[0].msg).toStartWith(
            'The SFTPPlus Windows service was started using the legacy')

        // Loading the page again, will not show another alert.
        sut.scope.__init__()
        expect(sut.scope.$parent.alerts).toHaveLength(1)
        expect(sut.scope.$parent.alerts[0].type).toEqual('warning')
    })

    /*
    Show alert on windows legacy, already shown.
    */
    test('__init___win_legacy_service_session', function() {
        var sut = mk.getController('statusCtrl')
        mk.minimalProcess(sut)
        sut.scope.$parent.process.runnables.status.windows_legacy_service = true
        expect(sut.$window.sessionStorage.getItem(
            'windows_legacy_service_warning')).toBeNull()
        sut.$window.sessionStorage.setItem(
            'windows_legacy_service_warning', 'done')
        sut.scope.__init__()

        expect(sut.scope.$parent.alerts).toEqual([])
    })
    /*
    Will auto-update the list of failed components when the process object
    is change with a deep-watch.
    */
    test('failed_components', function() {
        var sut = mk.getController('statusCtrl')

        var initial_failed_components = sut.scope.failed_components
        expect(sut.scope.failed_components).toEqual([])

        sut.scope.configuration.services = {'uuid1': {
            'uuid': 'uuid1',
            'name': 'Failing service',
        }}
        sut.scope.process.runnables = {}
        sut.scope.process.runnables.services = {'uuid1': {
            'uuid': 'uuid1',
            'last_error': {'id': '1200', data: {'details': 'Error info'}}
        }}
        // If the runnable has no error, it is not included.
        sut.scope.process.runnables.services['uuid2'] = {
            'uuid': 'uuid2',
            'name': 'Ok service',
            'last_error': null,
        }
        // The transfer is a runnable, but no longer a config,
        // this means that it is pending removal.
        sut.scope.configuration.transfers = null
        sut.scope.process.runnables.transfers = {'uuid3': {
            'uuid': 'uuid3',
            'name': 'Failed transfer that was removed',
            'last_error': {'id': '1200', data: {'details': 'Error info'}},
        }}

        // All other configuration are empty for this test.
        sut.scope.process.runnables.authentications = {}
        sut.scope.process.runnables.event_handlers = {}
        sut.scope.process.runnables.locations = {}
        sut.scope.process.runnables.resources = {}

        sut.scope.$digest()

        // The list is auto updated in place.
        expect(sut.scope.failed_components).toBe(initial_failed_components)
        expect(sut.scope.failed_components).toEqual([{
            'config': sut.scope.configuration.services.uuid1,
            'runnable': sut.scope.process.runnables.services.uuid1,
            'just_created': false,
            'name': 'Failing service',
            'uuid': 'uuid1',
        }])

    })

    /*
    Return interface and port as string.
    */
    test('getServiceInterface_service', function() {
        var sut = mk.getController('statusCtrl')
        var service_uuid = mk.string()
        sut.scope.$parent.process.runnables = {}
        sut.scope.$parent.process.runnables.services = {}
        sut.scope.$parent.upstream_configuration = {'services': {}}
        sut.scope.$parent.upstream_configuration.services[service_uuid] = {
            'type': 'sftp',
            'address': '123.0.0.2',
            'port': 80
            }

        var result = sut.scope.getServiceInterface(service_uuid)

        expect(result).toEqual('123.0.0.2:80')
    })

    /*
    Return empty string for monitor services.
    */
    test('getServiceInterface_monitor', function() {
        var sut = mk.getController('statusCtrl')
        var service_uuid = mk.string()
        sut.scope.$parent.process.runnables = {'services': {}}
        sut.scope.$parent.upstream_configuration = {'services': {}}
        sut.scope.$parent.upstream_configuration['services'][service_uuid] = {
            'type': 'monitor',
            'address': '123.0.0.2',
            'port': 80
            }

        var result = sut.scope.getServiceInterface(service_uuid)

        expect(result).toEqual('')
    })

    /*
    Return empty object if configuration was not yet loaded.
    */
    test('getServiceConfiguration_no_config_yet', function() {
        var sut = mk.getController('statusCtrl')

        var result = sut.scope.getServiceConfiguration('any-uuid')

        expect(result).toEqual({})
    })

    /*
    Return an empty object if the upstream configuration is not yet loaded.
    */
    test('getServiceStartupConfiguration_not_init', function() {
        var sut = mk.getController('statusCtrl')
        sut.scope.process.runnables.services = {}
        sut.scope.process.upstream_configuration = {}

        var result = sut.scope.getServiceStartupConfiguration('some-uuid')

        expect(result).toEqual({})
      })


    /*
    Return the configuration advertised by the service as being used
    at startup.
    */
    test('getServiceStartupConfiguration_found', function() {
        var sut = mk.getController('statusCtrl')
        var startup_config = {'name': 'something'}
        sut.scope.process.runnables.services = {'some-uuid': {
          'startup_configuration': startup_config,
        }}

        var result = sut.scope.getServiceStartupConfiguration('some-uuid')

        expect(result).toBe(startup_config)
      })

    /*
    Return empty object if configuration was not yet loaded.
    */
    test('getServiceUpstreamConfiguration_no_config_yet', function() {
        var sut = mk.getController('statusCtrl')

        var result = sut.scope.getServiceUpstreamConfiguration('any-uuid')

        expect(result).toEqual({})
    })
})
