(function(exports){

var init = require('../../../js_commons/__init__.js', 'init')

/*
Purge `text` for known DOM XSS attacks.

Angular `ng-bind-html` already does a good job at removing any unsafe HTML
markup, but it still tries to best leave in place the safe HTML markup.

This is a more aggressive fix.

Normally, doing conversion for this should be enough.
If this doesn't work, look at using a library like dompurify.
This was used in a previous version of this code, and you can check
the PR history.
*/
var _dom_purge = function(text) {

    var dirty_text = text
    if (init.typeOf(dirty_text) == 'array') {
      dirty_text = dirty_text.join(', ')
    }

    return dirty_text.toString()
      .replace(/&/g, '&amp;')
      .replace(/</g, '&lt;')
      .replace(/>/g, '&gt;')
      .replace(/\"/g, '&quot;')
      .replace(/\'/g, '&#39;')
      .replace(/\//g, '&#x2F;')
}

/*
Return human readable enabled/disabled state.
*/
exports.enableState = function() {
  return function(input) {
    if (input) {
      return 'enabled'
    } else {
      return 'disabled'
    }
  }
}


/*
Return human readable boolean as yes/no
*/
exports.confirmation = function() {
  return function(input) {
    if (input) {
      return 'yes'
    } else {
      return 'no'
    }
  }
}


/*
Return 'not defined' if value is not defined.
*/
exports.checkDefined = function() {
  return function(input) {
    if (typeof(input) === 'undefined') {
      return 'not defined'
    }

    if (input === null) {
      return 'not defined'
    }

    return input
  }
}

/*
Convert of dict members into an array
*/
exports.dictToArray = function() {
    return function(input, sort_key) {
        var result = []

        for (var key in input) {
            result.push(input[key])
        }

        if (sort_key) {
            result.sort(init.sort_by(sort_key, false, init.toUpperCaseFn))
        }
        return result
    }
}


exports.length = function() {
  return function(input) {
    return init.objectLength(input)
  }
}


/*
Generate HTML diff markup based on words.

`type` values:
* `added`: only show unchanged and added values
* `removed`: only show unchanged and removed values
* UNDEFINED: show inline changes (both removed and added)
*/
exports.textDiff = function() {
  return function(old_text, new_text, type) {

    var Diff = require('diff', 'Diff')

    if (old_text === null) {
        old_text = ''
    }

    if (new_text === null) {
        new_text = ''
    }
    var old_text = _dom_purge(old_text)
    var new_text = _dom_purge(new_text)

    function getDiffMarkup(changes) {

      var markup = []
      angular.forEach(changes, function(change) {
        var change_markup
        if (change.added && (!type || type === 'added')) {
          change_markup = '<ins>' + change.value + '</ins>'
        } else if (change.removed && (!type || type === 'removed')) {
          change_markup = '<del>' + change.value + '</del>'
        } else if (!change.added && !change.removed) {
          change_markup = change.value
        }
        markup.push(change_markup)
      })
      return markup.join('')
    }

    var changes = Diff.diffWords(old_text, new_text)
    return getDiffMarkup(changes, type)
  }
}

exports.cssForState = function(utils) {
  return function(state) {
    if (!state) {
      return ''
    }
    var map = {}
    map[utils.constants.STATE.STARTED] = 'label-success'
    map[utils.constants.STATE.FAILED_AT_START] = 'label-important'
    map[utils.constants.STATE.FAILED_AT_STOP] = 'label-important'
    map[utils.constants.STATE.FAILED_AT_RUN] = 'label-important'
    map[utils.constants.STATE.CONNECTING] = 'label-warning'
    map[utils.constants.STATE.DISCONNECTED] = 'label-warning'
    map[utils.constants.STATE.DISCONNECTING] = 'label-warning'
    map[utils.constants.STATE.SOURCE_FAILED] = 'label-warning'
    map[utils.constants.STATE.SOURCE_STOPPED] = 'label-warning'
    map[utils.constants.STATE.DESTINATION_FAILED] = 'label-warning'
    map[utils.constants.STATE.SUSPENDED] = 'label-warning'
    var result = map[state]
    if (!result) {
      return ''
    }
    return result
  }
}

/*
This is the highlight filter generator.
*/
exports.highlight = function() {


  return function (text, search, caseSensitive) {

    if (!search) {
        // Nothing to highlight.
        // Just do the escape.
        return _dom_purge(text)
      }

      text = text.toString()
      search = search.toString()

      if (search.length < 3) {
          // Don't trigger the highlight for small text.
          // This is basically designed to avoid overloading the UI.
          return _dom_purge(text)
      }

      var split_rule = new RegExp(search, 'gi')
      if (caseSensitive) {
         // Use exact value for case sensitive.
         split_rule = search
      }

      // We reinvent the `TEXT.join()` function here, but we escape each part.
      var result = []
      angular.forEach(text.split(split_rule), function(value) {
        result.push(_dom_purge(value))
        // FIXME:5873:
        // For case insensitive search this is an problem as the original
        // case is not preserved, and instead replaced with the search term.
        result.push(
          '<code class="highlight">' + _dom_purge(search) + '</code>')
      })
      // Remove the last insert to be like `TEXT.join()`.
      result.pop()

      return result.join('')
      }
}

/*
Retun a description of the timestamp in comparison with current date and time.
*/
exports.fromNow = function(dayjs) {
  return function(input) {
    var target_date = dayjs.unix(input)
    return target_date.fromNow()
  }
}


/*
Retun an ISO8601 representaiont of timestamp .
*/
exports.iso8601 = function(dayjs) {
  return function(input) {
    var target_date = dayjs.unix(input)
    return target_date.format()
  }
}

exports.componentType = function(utils) {
  return function(type, family) {
    return utils.getReadableType(family, type)
  }
}
/* c8 ignore next */
}(typeof exports === 'undefined' ? this.filters = {} : exports))
