/*
Tests for filters.
*/

describe('services.filters', function() {

    // We using our real application to make sure all filters
    // are registered in our production app.
    require('../app.js')

    /*
    Return filter with `name`.
    This can only be called once per test due to angularjs injector
    limitation.
    */
    var getFilter = function(name) {
        var target
        angular.mock.module('app')
        angular.mock.inject(function($filter) {
            target = $filter(name)
        })
        return target
    }

    /*
    Return length of objects.
    */
    test('length', function() {
        var lengthFilter = getFilter('length')
        expect(lengthFilter()).toBe(0)
        expect(lengthFilter('abc')).toBe(3)
    })

    /*
    fromNow converts an unix timestamp into an easy to understand text.
    */
    test('fromNow', function() {
        var fromNowFitler = getFilter('fromNow')
        expect(fromNowFitler(0)).toContain('years ago')
        expect(fromNowFitler(Date.now()/1000)).toEqual('a few seconds ago')
    })

    /*
    iso8601 converts an unix timestamp into an ISO format text.
    */
    test('iso8601', function() {
        var iso8601Filter = getFilter('iso8601')
        expect(iso8601Filter(0)).toEqual('1970-01-01T00:00:00+00:00')
        // The actual result depensds on the local timezone
        // we just check the start part.
        expect(iso8601Filter(1680714606)).toStartWith('2023-04-05T1')
    })


    /*
    return a list from all values from objects ignoring the object keys
    */
    test('dictToArray', function() {
        var dictToArrayFilter = getFilter('dictToArray')
        expect(dictToArrayFilter({})).toEqual([])

        var result = dictToArrayFilter({'b': 2, 'a': 1})
        expect(result).toEqual([2, 1])

        // Can also sort array based on dict key.
        result = dictToArrayFilter({
            'b': {'sort_key': 3, 'k': 'v'},
            'a': {'sort_key': 1, 'c': 'd'}
            }, 'sort_key')
        expect(result).toEqual([
            {'sort_key': 1, 'c': 'd'},
            {'sort_key': 3, 'k': 'v'}
            ])
    })

    /*
    Return a list sorted case insensitive.
    */
    test('dictToArray_sort', function() {
        var dictToArrayFilter = getFilter('dictToArray')
        var result = dictToArrayFilter({
            'ignored_1': {'sort_key': 'Z', 'k': 'v'},
            'ignored_2': {'sort_key': 'b', 'a': 'A'},
            'ignored_3': {'sort_key': 'a', 'c': 'd'}
            }, 'sort_key')
        expect(result).toEqual([
            {'sort_key': 'a', 'c': 'd'},
            {'sort_key': 'b', 'a': 'A'},
            {'sort_key': 'Z', 'k': 'v'}
            ])
    })

    /*
    The `highlight` filter will return the source with escaped HTML
    and no highlight if nothing is to be highlighted.
    */
    test('highlight_nothing', function() {
        var highlightFilter = getFilter('highlight')
        var result = highlightFilter('source <a b=\'c-&-\'/>"text')
        expect(result).toEqual(
            'source &lt;a b=&#39;c-&amp;-&#39;&#x2F;&gt;&quot;text')
    })

    /*
    The `highlight` filter will enclose the selected text in markup,
    with default search as case insensitive.
    The result has HTML escapes.
    */
    test('highlight_case_insensitive', function() {
        var highlightFilter = getFilter('highlight')
        var result = highlightFilter('source-quo-"here" Quo <a more', 'quo')
        // FIXME:5873:
        // For now the result is also case insensitive.
        expect(result).toEqual(
            'source-<code class=\"highlight\">quo</code>-&quot;here&quot; ' +
            '<code class=\"highlight\">quo</code> &lt;a more'
            )
    })

   /*
    The `highlight` filter will enclose the selected text in markup,
    with search option as case sensitive.
    The result has HTML escapes.
    */
    test('highlight_case_sensitive', function() {
        var highlightFilter = getFilter('highlight')
        var result = highlightFilter(
            'source-TexT-here text <a more', 'text', true)
        expect(result).toEqual(
            'source-TexT-here <code class=\"highlight\">text</code> &lt;a more')
    })

    test(
    'In added mode, it can handle the case when old text equals to null,' +
    ' handling it as an empty text.',
    function() {
        var textDiffFilter = getFilter('textDiff')
        var old_text = null
        var new_text = 'value'

        var result = textDiffFilter(old_text, new_text, 'added')

        expect(result).toEqual('<ins>value</ins>')
    })

    test(
    'In added mode, it can handle the case when old text is empty.',
    function() {
        var textDiffFilter = getFilter('textDiff')
        var old_text = ''
        var new_text = 'value'

        var result = textDiffFilter(old_text, new_text, 'added')

        expect(result).toEqual('<ins>value</ins>')
    })

    test(
    'In added mode, when text was only removed it only show unchanged ' +
    'text.',
    function() {
        var textDiffFilter = getFilter('textDiff')
        var old_text = 'some value'
        var new_text = 'value'

        var result = textDiffFilter(old_text, new_text, 'added')

        expect(result).toEqual('value')
    })

    test(
    'In added mode, the previous value which was unchanged is kept.',
    function() {
        var textDiffFilter = getFilter('textDiff')
        var old_text = 'old'
        var new_text = 'old value'

        var result = textDiffFilter(old_text, new_text, 'added')

        expect(result).toEqual('old<ins> value</ins>')
    })

    test(
    'In added mode, when words were removed from the old text, ' +
    'they are not listed.',
    function() {
        var textDiffFilter = getFilter('textDiff')
        var old_text = 'other old'
        var new_text = 'some old value'

        var result = textDiffFilter(old_text, new_text, 'added')

        expect(result).toEqual('<ins>some</ins> old<ins> value</ins>')
    })

    // In removed mode, when all text is removed it is marked as such.
    test('remove_mode_all', function() {
        var textDiffFilter = getFilter('textDiff')
        var old_text = 'value'
        var new_text = ''

        var result = textDiffFilter(old_text, new_text, 'removed')

        expect(result).toEqual('<del>value</del>')
    })

    // In removed mode, when no text is removed it is not marked.
    test('remove_mode_no_remove', function() {
        var textDiffFilter = getFilter('textDiff')
        var old_text = 'value'
        var new_text = 'value new-value'

        var result = textDiffFilter(old_text, new_text, 'removed')

        expect(result).toEqual('value')
    })

    // In removed mode, when words are added and other are removed,
    // it only shows removed words.
    test('remove_mode_only_remove', function() {
        var textDiffFilter = getFilter('textDiff')
        var old_text = 'old value'
        var new_text = 'value new-value'

        var result = textDiffFilter(old_text, new_text, 'removed')

        expect(result).toEqual('<del>old</del> value')
    })

    // In inline mode, removed words are listed before added words.
    test('inline_remove_before_added', function() {
        var textDiffFilter = getFilter('textDiff')
        var old_text = 'other old'
        var new_text = 'some old value'

        var result = textDiffFilter(old_text, new_text, undefined)

        expect(result).toEqual(
            '<del>other</del><ins>some</ins> old<ins> value</ins>')
    })

    // Lists are shown as comma separated values.
    test('list_diff', function() {
        var textDiffFilter = getFilter('textDiff')
        var old_text = ['other', 'old']
        var new_text = ['some', 'old', 'value']

        var result = textDiffFilter(old_text, new_text, undefined)

        expect(result).toEqual(
            '<del>other</del><ins>some</ins>, old<ins>, value</ins>')
    })

    // Integers are shown as strings.
    test('integer_diff', function() {
        var textDiffFilter = getFilter('textDiff')
        var old_text = 123
        var new_text = 1245

        var result = textDiffFilter(old_text, new_text, undefined)

        expect(result).toEqual('<del>123</del><ins>1245</ins>')
    })

    // Both old and new values are sanitized for HTML markup.
    test('html_sanitization', function() {
        var textDiffFilter = getFilter('textDiff')
        var old_text = 'other <a href="some.url">link text</a> old'
        var new_text = 'some <a href="some.url">link text</a>old value'

        var result = textDiffFilter(old_text, new_text, undefined)

        expect(result).toEqual(
            '<del>other</del><ins>some</ins> ' +
            '&lt;a href=&quot;some.url&quot;&gt;link text&lt;&#x2F;a&gt;' +
            '<ins>old</ins> <del>old</del><ins>value</ins>')
    })


})
