/*
* Tests for 'utils' service.
*
* For now the utils service contains a lot of code interacting with the
* browser. We need a set of system test for this.
*/


describe('services.Utils', function() {

    // Create a simple app to support the utils tests.
    // This is here to load the injector with the dependencies of utils.
    angular.module('utils_app', [
        'ngCookies'
    ])

    var utils
    var $window

    beforeEach(angular.mock.module('utils_app'))
    beforeEach(inject(function($injector) {
        utils = mk.servicesUtils($injector)
        $window = $injector.get('$window')
    }))

    describe('random', function() {

        test(
        'getGUIDUnsafe return a valid UID.',
        function() {
            expect(utils.getGUIDUnsafe()).toBeTruthy()
        })

        test(
        'getRandomUUIDUnit return a random value.',
        function() {
            var result

            result = utils.getGUIDUnsafe()

            expect(result).toMatch(/^\w{8}-\w{4}-\w{4}-\w{4}-\w{12}$/)
        })

        test(
        'getRandomMark return a 4 characters random value.',
        function() {
            var result

            result = utils.getRandomMark()

            expect(result).toBeTruthy()
            expect(result.length).toBe(4)
        })


        test(
        'getRandomMark return a 4 characters random value with prefix.',
        function() {
            var result

            result = utils.getRandomMark('bla')

            expect(result).toBeTruthy()
            expect(result).toStartWith('bla')
            expect(result.length).toBe(4 + 3)
        })

    })

    describe('maybeParseBoolean', function() {

        test(
        'Return exact value when not a string.',
        function() {
            expect(utils.maybeParseBoolean(3)).toBe(3)
        })

        test(
        'Return exact value when not a boolean string.',
        function() {
            expect(utils.maybeParseBoolean('some')).toBe('some')
        })

        test(
        'Return `true` value when a boolean string for `true`.',
        function() {
            expect(utils.maybeParseBoolean('yes')).toBeTrue()
            expect(utils.maybeParseBoolean('Yes')).toBeTrue()
            expect(utils.maybeParseBoolean('true')).toBeTrue()
        })

        test(
        'Return `false` value when a boolean string for `false`.',
        function() {
            expect(utils.maybeParseBoolean('no')).toBeFalse()
            expect(utils.maybeParseBoolean('nO')).toBeFalse()
            expect(utils.maybeParseBoolean('fAlSe')).toBeFalse()
        })

    })

    describe('multiSequenceToText', function() {

        test(
        'Return empty string for an empty sequence.',
        function() {
            expect(utils.multiSequenceToText([])).toEqual('')
        })

        test(
        'Return a text with each element on a new line.',
        function() {
            expect(utils.multiSequenceToText([
                ['first line'], ['second line']])).toEqual(
            'first line\nsecond line')
        })


        test(
        'Return a text with each element on a new line.',
        function() {
            expect(utils.multiSequenceToText([
                ['first', 'line'], ['second', 'line']])).toEqual(
            'first, line\nsecond, line')
        })

        test(
        'When values have commas or newlines, they are quoted.',
        function() {
            expect(utils.multiSequenceToText([
                ['fir , st', 'li\r\nne'], ['second', 'line']])).toEqual(
            '"fir , st", "li\r\nne"\nsecond, line')
        })

    })

    describe('textToMultiSequence', function() {

        test(
        'Return empty list for an empty text.',
        function() {
            expect(utils.textToMultiSequence('')).toEqual([])
        })

        test(
        'Return a text with each element on a new line.',
        function() {
            expect(utils.textToMultiSequence(
                'first line\nsecond line')).toEqual(
                [['first line'], ['second line']]
                )
        })

        test(
        'A single line will still return a list of lists.',
        function() {
            expect(utils.textToMultiSequence(
                'first line, "other element"  , end\n')).toEqual(
                [['first line', 'other element', 'end']]
                )
        })

        test(
        'Return a text with each element on a new line.',
        function() {
            expect(utils.textToMultiSequence(
                'first, li"ne\nsecond , line')).toEqual(
                [['first', 'li"ne'], ['second', 'line']]
                )
        })

        test(
        'New lines in values are replaced by spaces.',
        function() {
            expect(utils.textToMultiSequence(
                'first, "line\ncharacter"\nsecond , line')).toEqual(
                [['first', 'line character'], ['second', 'line']]
                )
        })

        test(
        'It support commas in values as long as they are escaped.',
        function() {
            expect(utils.textToMultiSequence(
                '"first", "line , comma" \nsec"ond , "*/ch/*.csv"')).toEqual(
                [['first', 'line , comma'], ['sec"ond', '*/ch/*.csv']]
                )
        })

    })

    describe('isDeleteBlocked', function() {

        test(
        'Return empty string for generic entity.',
        function() {
            var config = {'uuid': 'some random uuid'}
            expect(utils.isDeleteBlocked(config, 'entity')).toEqual('')
        })

        test(
        'Return a message when we have a specia entity.',
        function() {
            var config = {'uuid': 'DEFAULT-ADMINISTRATOR-UUID'}
            expect(utils.isDeleteBlocked(config, 'entity')).toEqual(
                'Default entity can\'t be removed.')
        })

    })

    /*
    Return a false object when we don't have IE.
    */
    test('getIEVersion_non_IE', function() {
        var result = utils.getIEVersion()
        expect(result).toEqual({
            "ActingVersion": 0,
            "CompatibilityMode": false,
            "IsIE": false,
            "TrueVersion": 0,
            })
    })

    /*
    It can detect the version from IE in non compatible mode.
    */
    test('getIEVersion_IE11', function() {
        utils._user_agent = (
            'Mozilla/5.0 (Windows NT 10.0; WOW64; ' +
            'Trident/7.0; .NET4.0C; .NET4.0E; rv:11.0) like Gecko'
        )
        var result = utils.getIEVersion()
        expect(result).toEqual({
            "ActingVersion": 11,
            "CompatibilityMode": false,
            "IsIE": true,
            "TrueVersion": 11,
            })
    })

    /*
    It can detect the version from IE in compatible mode.
    */
    test('getIEVersion_IE11_compabitle', function() {
        utils._user_agent = (
            'Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 10.0; ' +
            'WOW64; Trident/7.0; .NET4.0C; .NET4.0E)'
        )
        var result = utils.getIEVersion()
        expect(result).toEqual({
            "ActingVersion": 7,
            "CompatibilityMode": true,
            "IsIE": true,
            "TrueVersion": 11,
            })
    })

    /*
    Writes the data to the navigator's clipboard.
    */
    test('writeClipboard_chrome', function() {
        var result = utils.writeClipboard('some-text')
        expect(result).toBeTrue()
        expect($window.navigator.clipboard.readText()).toEqual('some-text')
    })

    /*
    Return False without a crash when when browser has no support for
    clibpborad.
    */
    test('writeClipboard_unknow_browser', function() {
        var original_clipboard = $window.navigator.clipboard
        $window.navigator.clipboard = null

        try {
            var result = utils.writeClipboard('some-text')
        } finally {
            $window.navigator.clipboard = original_clipboard
        }

        expect(result).toBeFalse()
    })

    /*
    When clipboard API is not found it will fallback to IE API.
    */
    test('writeClipboard_ie', function() {
        var original_clipboard = $window.navigator.clipboard
        $window.navigator.clipboard = null
        $window.clipboardData = {
            'setData': function(target, text) {
                if (target != 'Text') {
                    throw "Only text is supported for clipboard"
                }
                global._clipboard_content = text
            }
        }

        try {

            var result = utils.writeClipboard('some-text')
        } finally {
            $window.navigator.clipboard = original_clipboard
            delete $window.clipboardData
        }

        expect(result).toBeTrue()
        expect($window.navigator.clipboard.readText()).toEqual('some-text')
    })

    /*
    Return an empty list when object is empty.
    */
    test('keyValueArray_empty', function() {
        expect(utils.keyValueArray({})).toEqual([])
    })

    /*
    Return an array of objects with key and value fields.
    */
    test('keyValueArray_content', function() {
        var source = {
            'key1': 'value1',
            'key2': 'a value 2',
            'key3': 'More values'
        }
        expect(utils.keyValueArray(source)).toEqual([
            {'key': 'key2', 'value': 'a value 2'},
            {'key': 'key3', 'value': 'More values'},
            {'key': 'key1', 'value': 'value1'},
        ])
    })


    /*
    Returns a placeholder when we don't have a type
    */
    test('getReadableType_undefined', function() {
        var result = utils.getReadableType('family-ignored', undefined)

        expect(result).toEqual('NO-TYPE-YET')
    })

    /*
    Returns the uppercase text when family is not known.
    */
    test('getReadableType_unknown_family', function() {
        var result = utils.getReadableType('unknown-family', 'some-type')

        expect(result).toEqual('SOME-TYPE')
    })


    /*
    Returns the uppercase text when family known, but type is not found
    for family.
    */
    test('getReadableType_unknown_family', function() {
        var result = utils.getReadableType('service', 'unknown-type')

        expect(result).toEqual('UNKNOWN-TYPE')
    })

    /*
    Returns the uppercase text when family known, but type is not found
    for family.
    */
    test('getReadableType_unknown_family', function() {
        var result = utils.getReadableType('service', 'http')

        expect(result).toEqual('HTTP (unsecured)')
    })

    /*
    Returns True if value is one of the inherited markers.
    */
    test('isInherited', function() {
        expect(utils.isInherited('some-value')).toBeFalse()
        expect(utils.isInherited(42)).toBeFalse()
        expect(utils.isInherited('')).toBeFalse()

        expect(utils.isInherited('inherit')).toBeTrue()
        expect(utils.isInherited('inherited')).toBeTrue()
    })
})
