# Copyright (c) 2014 Adi Roiban.
# See LICENSE for details.
"""
Jobs management.

A job is a chain of steps executed in order to do a managed transfer for
a file.

If a step fails, the transfer will also fail.
See `step` package for more details about a step.

There is a manager that coordinates which job is executed at a certain time.
Only one job is active at a certain time.

Jobs can be suspended if they fail and need a retry later.
The jobs manager will suspend the current job and select another job from the
queue (either from the new jobs or already suspended jobs.)
"""
from __future__ import absolute_import, unicode_literals
from collections import OrderedDict
import uuid

from bunch import Bunch
from ordered_set import OrderedSet
from twisted.internet import defer, reactor
from twisted.internet.error import AlreadyCalled, AlreadyCancelled
from zope.interface import implementer

from chevah.server import (
    emit_error_details,
    emit_from_error,
    force_unicode,
    get_path_handling_module,
    )
from chevah.server.commons.constant import (
    CONDITION,
    STATE,
    TYPE_NAME,
    )
from chevah.server.commons.exception import (
    EmittedError,
    ServerError,
    ServerException,
    )
from chevah.server.commons.interface import ILocalLocation, ITransferJob
from chevah.server.commons.match import get_matched_groups
from chevah.server.commons.public_format import NoStrBunch, public_format
from chevah.server.commons.runnable import Runnable, RunnablesManager
from chevah.server.event.model import ChevahTimeStamp
from chevah.server.process.spawn import defer_spawn
from chevah.server.identity.avatar import JobAvatar
from chevah.server.location.commons import (
    get_file_name,
    get_file_name_with_timestamp,
    )
from chevah.server.transfer.step.local import CopyTransferredFile, ExecuteStep
from chevah.server.transfer.step.location import (
    DeleteSourceStep,
    EnforceOverwritingRuleStep,
    ExecuteDestinationStep,
    RenameTemporaryDestination,
    StartDestinationStep,
    TransferFileStep,
    )
from chevah.server.transfer.step.statement import IterateBatchStep, RetryStep
from chevah.server.transfer.transformer import get_transformers


class StepFirstException(Exception):
    """
    Internal exception used to pass the first error caused by a step execution.
    """
    def __init__(self, name, exception):
        self.name = name
        self.exception = exception


@implementer(ITransferJob)
class TransferJob(Runnable):
    """
    Transfer jobs are created by a transfer job manager.
    Jobs represent a transfer of a specific file.

    Each job has a unique uuid.

    Jobs share configuration with the transfer which initiated them.

    It is a finite state machine.
    It is started/ended automatically by the `TransferJobsManager`.
    It can also be stopped/canceled manually.

    A job transfer is done in multiple steps.

    A step is the most low level system operation done by this job.

    A phase is a set of atomic steps done by the job.
    When a step inside a phase fails, the whole phase is retried.
    When all steps from a phase succeed, the next phase is executed and
    current phase is no longer retried.

    execute-before
    foreach-file
        retry
            file-transfer
        archive-success
        archive-failure
        [delete-source-after-transfer]
    execute-after-success
    execute-after-failure
    """
    PROPERTIES_REQUIRING_RESTART = ()

    _RESERVED_CONDITIONS = (
        CONDITION.ALWAYS,
        CONDITION.FAILURE,
        CONDITION.SUCCESS,
        )

    def __init__(self, parent, paths, source, destination, configuration):
        self._uuid = str(uuid.uuid4())
        self._parent = parent
        self._source = source
        self._destination = destination
        self._configuration = configuration

        # Since the configuration can change at runtime, used the value
        # validated at transfer start.
        self._destination_path_actions = (
            self._parent.parent._destination_path_actions)

        # All paths which needs to be transferred in this job.
        # There is also the option of using an ordered dict,
        # since the keys can act like a set.
        # But for now we use this extenal dependency.
        self._paths = OrderedSet(paths)

        # Paths which were successfully transferred to destination,
        # not necessarily with a successful archive or post execute.
        # Mapping from source to destination of files that were successfully
        # transferred, but perhaps not successfully renamed (yet).
        self._success_paths = OrderedDict()

        self._state = STATE.STOPPED

        # Auto-kept in sync with `_path`. This is the path of the file as
        # transferred to destination.
        self._destination_path = None
        # Temporary name on destination.
        self._temporary_path = None
        # Path to file which is currently transferred.
        self._path = None
        # Content transformer for the current path.
        self._transformer = None
        # We start with first path so that it is available in before-execute
        # or any other step call before
        self.setNextTransferPath(next(iter(self._paths)))

        # This is the main deferred, which signals when the job is done.
        self._deferred = None
        # This is the deferred of the current step.
        self._current_deferred = None
        # This is the current step which is executed.
        self._current_step = None
        # We start with a successful transfer and also
        # add `always` condition.
        # Conditions needs to be private and should only
        # be change using dedicated methods.
        self._conditions = {'always', 'success'}
        self._avatar = JobAvatar(
            uuid=self._uuid, parent=self.parent.parent)

    def emitEvent(self, *args, **kwargs):
        """
        See: IRunnable
        """
        kwargs['avatar'] = self._avatar
        self.parent.parent.emitEvent(*args, **kwargs)

    @property
    def uuid(self):
        """
        Job has an unique UUID and does not reflect the one from transfer
        configuration.
        """
        return self._uuid

    @property
    def avatar(self):
        """
        The job session.
        """
        return self._avatar

    @property
    def source_file_path(self):
        """
        Path to source file of current file from this transfer.
        """
        return self._path

    @property
    def destination_path(self):
        """
        Return path on destination location of current file for this transfer.
        """
        return self._destination_path

    @property
    def temporary_path(self):
        """
        Return the temporary path on destination.
        """
        return self._temporary_path

    def addTimestampToDestinationPath(self):
        """
        Amend the destination path to include a timestamp.
        """
        self._destination_path = self.getPathWithTimestamp(
            self._destination, self._destination_path)

    def _setDestinationPath(self, root, path, prefix='', suffix=''):
        """
        Set the destination path and the temporary destination path.
        """
        self._temporary_path = ''
        self._destination_path = self._destination.joinPath(root, path)

        temp_root, name = self._destination.splitPath(self._destination_path)

        temporary_name = name
        if prefix:
            temporary_name = prefix + temporary_name
        if suffix:
            temporary_name = temporary_name + suffix

        if temporary_name != name:
            # We have a temporary name
            self._temporary_path = self._destination.joinPath(
                temp_root, temporary_name)

    @property
    def source(self):
        """
        Source location.
        """
        return self._source

    @property
    def destination(self):
        """
        Destination location.
        """
        return self._destination

    @property
    def local_location(self):
        """
        Return the local location between source and destination.

        Return None when no location is local.
        """
        if ILocalLocation.providedBy(self._source):
            return self._source

        if ILocalLocation.providedBy(self._destination):
            return self._destination

        return None

    @property
    def transferred_file_local_path(self):
        """
        Return the absolute local path of the file which was transferred and
        the relative path to its source.
        """
        if ILocalLocation.providedBy(self._source):
            result = self.source_file_path
            relative_path = self._source.relativePath(
                self._configuration.source_path, result)
            return result, relative_path

        if ILocalLocation.providedBy(self._destination):
            result = self.destination_path
            relative_path = self._destination.relativePath(
                self._configuration.destination_path, result)

            return result, relative_path

        raise NotImplementedError(
            'Remote to remote transfers are not supported.')

    @property
    def transformer(self):
        """
        Return the transformer for this job.
        """
        return self._transformer

    def hasLocalSource(self):
        """
        Return True if source is local location.
        """
        return ILocalLocation.providedBy(self._source)

    def addCondition(self, name):
        """
        Add a condition to this job.
        """
        if name in self._RESERVED_CONDITIONS:
            raise RuntimeError('Trying to add a reserved condition.')
        self._conditions.add(name)

    def removeCondition(self, name):
        """
        Remove a condition from this job.
        """
        if name in self._RESERVED_CONDITIONS:
            raise RuntimeError('Trying to remove a reserved condition.')
        self._conditions.discard(name)

    def hasCondition(self, name):
        """
        Return `True` if job has condition with `name`.
        """
        return name in self._conditions

    def _getConditions(self):
        """
        Return the current conditions.
        """
        return self._conditions.copy()

    def _setConditions(self, conditions):
        """
        Set the new conditions.
        """
        self._conditions = conditions.copy()

    def addTemporaryTransfer(self, attributes):
        """
        Called when the current file was copied to the destination
        with the `attributes` of the source file.

        This is called before a future possible rename operation.

        Returns the path that was transferred, which might be the temporary
        path.
        """
        destination_path = self._temporary_path or self._destination_path

        self._success_paths[self._path] = destination_path, attributes

        return destination_path

    def onFinalTransfer(self):
        """
        Called by a step after the current file was transferred to destination
        with the final name.
        """
        _, attributes = self._success_paths[self._path]
        # We set attributes to None to signal that they were forwarded to
        # the transfer.
        # It also simplifies assertions inside the test code.
        self._success_paths[self._path] = self._destination_path, None

        # Notify the parent transfer that the file was transferred to
        # destination.
        #
        # For the partner-to-partner transfer we consider that the file
        # is done.
        # Local post transfer steps might fail, but the file was already
        # fully pushed to the partner and we should not re-transfer it.
        #
        # Our parent is the job manager, and his parent is the transfer.
        self.parent.parent.onFileSuccess(attributes)

    def getAllPaths(self):
        """
        Return all paths for this job.
        """
        # We keep the OrderedSet for internal usage and export it as a
        # simple list to keep the order.
        return list(self._paths)

    def setNextTransferPath(self, path):
        """
        Set the source and destination paths for the file which should be
        transferred next and the content transformer for the
        new path.
        """
        self._applyDestinationPath(path)

        relative_source_path = self._source.relativePath(
            self._configuration.source_path, path)
        self._transformer = get_transformers(
            relative_source_path,
            self._configuration.destination_content_actions,
            )

    def setFallbackPath(self):
        """
        Set the source to be transferred to fallback destination path.

        This needs `setNextTransferPath` to be called first.
        Is designed to be called by the transfer after delivery to
        `setNextTransferPath` fails.
        """
        self._applyDestinationPath(self._path, fallback=True)

    def _applyDestinationPath(self, path, fallback=False):
        """
        Set the destination path based on the configuration.
        """
        if path not in self._paths:
            raise RuntimeError(
                'Only set paths which are inside the source path.')

        self._path = path

        # Also set the destination path.
        relative_source_path = self._source.relativePath(
            self._configuration.source_path, self._path)

        # Start with the simple configuration.
        # Later it might be overwritten by destination actions.

        temporary_prefix = ''
        temporary_suffix = self._configuration.destination_temporary_suffix

        if fallback:
            root = self._configuration.destination_fallback_path
            self._setDestinationPath(
                root, relative_source_path, temporary_prefix, temporary_suffix)
            return

        root = self._configuration.destination_path

        if self._destination_path_actions:
            # Apply dynamic destination path rules.
            self._applyDestinationPathAction(
                relative_source_path, destination_base=root)
            return

        # No need for any action.
        # joinPath will normalize the relative path and will also convert
        # to destination path separators.
        self._setDestinationPath(
            root, relative_source_path, temporary_prefix, temporary_suffix)

    def _applyDestinationPathAction(
            self, relative_source_path, destination_base):
        """
        Set the destination path to be transferred.

        Based on the advanced configuration.
        """
        root = destination_base

        # Start with basic suffix and prefix.
        temporary_prefix = ''
        temporary_suffix = self._configuration.destination_temporary_suffix

        # Start with the source name.
        result = relative_source_path
        action_taken = False

        # The rule matching is done on the original path, but the result
        # from a rule is passed to the next.

        # The configuration is validate at the time when the transfer
        # starts.
        for rule in self._destination_path_actions:
            groups = get_matched_groups(
                value=relative_source_path,
                rule=rule[0],
                usage='destination_path_actions',
                )

            if not groups:
                # Not a rule for this path.
                continue

            action = rule[1].lower()
            action_taken = True

            if action == TYPE_NAME.NO_ACTION:
                # Do nothing.
                continue

            if action == TYPE_NAME.TRANSFORM:
                option = rule[2]
                result = _transform_path(
                    source_path=relative_source_path,
                    groups=groups,
                    template=option,
                    now=self._scheduler.seconds(),
                    )
                continue

            if action == TYPE_NAME.ONLY_FILENAME:
                result = get_file_name(result)
                continue

            if action == TYPE_NAME.LOWERCASE:
                result = result.lower()
                continue

            if action == TYPE_NAME.REPLACE_SEPARATOR:
                option = rule[2]
                if '/' in relative_source_path:
                    separator = '/'
                else:
                    separator = '\\'
                result = result.replace(separator, option)
                continue

            if action == TYPE_NAME.TEMPORARY_PREFIX:
                temporary_prefix = rule[2]
                result = result
                continue

            if action == TYPE_NAME.TEMPORARY_SUFFIX:
                temporary_suffix = rule[2]
                result = result
                continue

            if action == TYPE_NAME.FIXED_PATH:
                root = rule[2]
                result = ''
                continue

        if not action_taken:
            raise ServerException(
                'No destination_path_actions defined for "%s".' % (
                    relative_source_path,))

        self._setDestinationPath(
            root, result, temporary_prefix, temporary_suffix)

    def start(self, avatar=None):
        """
        Simple code to start a transfer.

        Each start method should handle the success path. For errors, it
        should log the error and then return the failure.

        This method always succeeds.
        `avatar` is not used as the jobs are automatically started by the
        transfers.

        Returns True or False to signal that the job was successful.

        It does not uses the infrastructure from Runnable.
        """
        self._state = STATE.STARTING
        if not self._deferred:
            self._deferred = defer.succeed(None)

        deferred = self._deferred
        deferred.addCallback(self._cbStart)
        deferred.addCallback(self._cbJobSucceeded)
        deferred.addErrback(self._ebJobFailed)
        return deferred

    def _cbStart(self, result):
        """
        Called to trigger the steps execution.
        """
        self._state = STATE.STARTED
        data = {
            'name': self.parent.parent.name,
            'count': len(self._paths),
            'files': '"' + '", "'.join(self._paths) + '"',
            }
        self.emitEvent('60073', data=data)

        steps = self._getSteps()
        return self.runSteps(steps=steps)

    def stop(self, avatar=None):
        """
        Stop/cancel the job.

        This should be called only when the job needs to be stopped while
        still running.

        After a job is started, it will just end it execution without a stop.

        It does not uses the convention from Runnable.
        `avatar` is not used as the jobs are automatically stopped by the
        transfers.
        """
        # Under normal circumstances, we should not reach here, so if we
        # are requested to stop, it means that we should cancel the current
        # job.
        if not self._deferred:
            return defer.fail(RuntimeError('Stop called before start.'))

        if self.state == STATE.DONE:
            return self._deferred

        self._state = STATE.FAILED_AT_RUN
        self._deferred.cancel()
        return self._deferred

    @defer.inlineCallbacks
    def runSteps(self, steps):
        """
        Run each step from `steps`.

        If one step fail continue to execute the remaining steps but in the
        end raise the first error.

        When first step fails, the CONDITION.FAILURE is set to the job while
        CONDITION.SUCCESS is removed.
        """
        self._state = STATE.STARTED

        first_error = None
        for step in steps:
            if not (self._conditions & step.conditions):
                # We ignore steps as it does not match current condition.
                continue

            self._current_step = step
            try:
                self._current_deferred = step.run()
                yield self._current_deferred
            except defer.CancelledError as error:
                # Stop executing other step when the job is canceled.
                raise error

            except Exception as error:
                # All step errors, beside first error
                # are ignored.
                # We might be interested in latest error,
                # but once a step fails, there is a great
                # change of other steps to fail so there is
                # no much use in seeing multiple errors, all
                # related to the initial error.
                if first_error:
                    continue

                if isinstance(error, StepFirstException):
                    # A child runSteps has failed. Just record the error at
                    # this upper level.
                    first_error = error
                    continue

                if isinstance(error, ServerError):
                    # Emit the error as soon as possible
                    error = emit_from_error(error, emitter=self.emitEvent)
                else:
                    """
                    Pass the error along to be converted into an ISE.
                    """

                first_error = StepFirstException(
                    name=step.name, exception=error)
                self._conditions.add(CONDITION.FAILURE)
                self._conditions.discard(CONDITION.SUCCESS)

        if not first_error:
            return

        raise first_error

    def getPathWithTimestamp(self, location, path):
        """
        Return the path with a file name with amended timestamp.
        """
        dir_path, file_name = location.splitPath(path)
        file_name = self.getFileNameWithTimestamp(file_name)
        return location.joinPath(dir_path, file_name)

    def getFileNameWithTimestamp(self, file_name):
        """
        Return the name of a file with amended timestamp, keeping the
        file extension.
        """
        return get_file_name_with_timestamp(
            name=file_name,
            seconds=self._scheduler.seconds(),
            random_mark=self._random.randint(0, 100),
            )

    # A description of the available placeholder.
    # It is used for example to validate the configuration, before calling
    # the step.
    REMOTE_COMMAND_PLACEHOLDERS = NoStrBunch(
        source=NoStrBunch(path='', name=''),
        destination=NoStrBunch(path='', name=''),
        )

    def resolveRemoteCommands(self, raw_commands):
        """
        Resolve the commands with current data from the job.
        """
        source_path, _ = self.transferred_file_local_path
        source_parts = self.source.splitPath(source_path)

        destination_path = self.destination_path
        destination_parts = self.destination.splitPath(destination_path)

        # WARNING: Keep this in sync with REMOTE_COMMAND_PLACEHOLDERS.
        data = {
            'source': {
                'path': source_path,
                'name': source_parts[1],
                },
            'destination': {
                'path': destination_path,
                'name': destination_parts[1],
                },
            }
        resolved_commands = []
        for command in raw_commands:
            resolved_commands.append(
                public_format(command, **Bunch.fromDict(data)))

        return resolved_commands

    def _cbJobSucceeded(self, result):
        """
        Called when transfer execution was done.
        """
        self._state = STATE.DONE

        if self.hasCondition(CONDITION.SKIP):
            return True

        source_paths = []
        destination_paths = []
        for source_path, destination in self._success_paths.items():
            destination_path, _ = destination
            source_paths.append(source_path)
            destination_paths.append(destination_path)
        self.emitEvent(
            '60017',
            data={
                'paths': ', '.join(source_paths),
                'destination_paths': destination_paths,
                'success_list': source_paths,
                },
            )
        return True

    def _ebJobFailed(self, failure):
        """
        Called when transfer has failed after all retries.

        It should always succeed.
        """
        self._state = STATE.FAILED_AT_RUN

        data = {'path': self.source_file_path}

        if failure.check(defer.CancelledError):
            # We need to take the current step before canceling the current
            # step, as this will trigger a failure which will call the
            # execution of the other steps.
            if not self._current_step:
                step_name = 'initialization'
            else:
                step_name = self._current_step.name
            data['step_name'] = step_name

            # The job was canceled/stopped and we need to forward the
            # request to the current step.
            if self._current_deferred:
                self._current_deferred.cancel()

            error = ServerError('60078', data=data)
            emit_from_error(error, emitter=self.emitEvent)
            error = EmittedError(error)
        else:
            # We have a StepFirstException.
            step_name = failure.value.name
            data['step_name'] = step_name
            error = failure.value.exception

        if isinstance(error, ServerException):
            data['details'] = error.message
            error = ServerError('60081', data=data)
            emit_from_error(error, emitter=self.emitEvent)
            error = EmittedError(error)

        if not isinstance(error, EmittedError):
            # ServerErrors should be emitted earlier as otherwise if we try
            # to emit them here, we will see the errors to late when some
            # other steps (set for run on failure) were already executed.
            # All errors should have already been emitted so any error which
            # is not emitted is an ISE.
            emit_error_details(
                error=error,
                event_id='60018',
                data=data,
                emitter=self.emitEvent,
                )
            # We convert the error into a ServerError so that it can be set
            # as last error.
            data['details'] = force_unicode(error)
            error = EmittedError(ServerError(
                '60018', data=data, avatar=self.avatar))

        self._setLastError(error)

        # FIXME:5219:
        # Add separate event for partial transfer - transfer with success
        # on destination, but for which rename or execute after failed.
        source_paths = list(self._success_paths.keys())

        failed_paths = ', '.join(self._paths - set(source_paths))
        success_paths = ', '.join(source_paths)
        if not success_paths:
            success_paths = "all files failed"

        data = {
            'failed_paths': failed_paths,
            'success_paths': success_paths,
            'failed_list': list(self._paths - set(source_paths)),
            'success_list': source_paths,
            'step_name': step_name,
            }
        self.emitEvent('60016', data=data)
        return False

    def _getSteps(self):
        """
        Return the steps to execute for a transfer.
        """
        delete_source = self._configuration.delete_source_on_success
        execute_timeout = self._configuration.execute_timeout

        # The steps repeated for each file.
        transfer_steps = [
            RetryStep(
                EnforceOverwritingRuleStep(
                    conditions={CONDITION.SUCCESS},
                    job=self,
                    name='enforce overwriting rule',
                    ),
                retry_count=self._configuration.retry_count,
                retry_wait=self._configuration.retry_wait,
                ),
            RetryStep(
                step=TransferFileStep(
                    conditions={CONDITION.SUCCESS},
                    job=self,
                    name='file-transfer',
                    ),
                retry_count=self._configuration.retry_count,
                retry_wait=self._configuration.retry_wait,
                ),
            RetryStep(
                step=RenameTemporaryDestination(
                    conditions={CONDITION.SUCCESS},
                    job=self,
                    name='rename-temporary-destination',
                    ),
                retry_count=self._configuration.retry_count,
                retry_wait=self._configuration.retry_wait,
                ),
            # `archive-on-failure` is placed first so that when
            # `archive-on-success` fails the `archive-on-failure` was already
            # executed and the file is not archived as a failure.
            CopyTransferredFile(
                job=self,
                conditions={CONDITION.FAILURE},
                name='archive-on-failure',
                path=self._configuration.archive_failure_path,
                ),
            CopyTransferredFile(
                job=self,
                conditions={CONDITION.SUCCESS},
                name='archive-on-success',
                path=self._configuration.archive_success_path,
                ),
            ]
        if delete_source:
            transfer_steps.append(
                RetryStep(
                    DeleteSourceStep(
                        conditions={CONDITION.SUCCESS},
                        job=self,
                        name='delete-source-after-transfer',
                        ),
                    retry_count=self._configuration.retry_count,
                    retry_wait=self._configuration.retry_wait,
                    ))

        # Final assembly of steps.
        steps = [
            StartDestinationStep(
                conditions={CONDITION.ALWAYS},
                name='start-destination',
                job=self,
                ),
            ExecuteStep(
                conditions={CONDITION.SUCCESS},
                name='before',
                command=self._configuration.execute_before,
                timeout=execute_timeout,
                job=self,
                ),
            RetryStep(
                ExecuteDestinationStep(
                    conditions={CONDITION.SUCCESS},
                    name='on-destination-before',
                    commands=split_remote_commands(
                        self._configuration.execute_on_destination_before),
                    job=self,
                    ),
                retry_count=self._configuration.retry_count,
                retry_wait=self._configuration.retry_wait,
                ),
            IterateBatchStep(
                conditions={CONDITION.SUCCESS},
                job=self,
                steps=transfer_steps,
                name='iterate-batch',
                ),
            ExecuteStep(
                conditions={CONDITION.SUCCESS},
                name='after-success',
                command=self._configuration.execute_after_success,
                timeout=execute_timeout,
                job=self,
                ),
            ExecuteDestinationStep(
                conditions={CONDITION.SUCCESS},
                name='on-destination-after-success',
                commands=split_remote_commands(
                    self._configuration.execute_on_destination_after_success),
                job=self,
                ),
            ExecuteStep(
                conditions={CONDITION.FAILURE},
                name='after-failure',
                command=self._configuration.execute_after_failure,
                timeout=execute_timeout,
                job=self,
                ),
            ExecuteDestinationStep(
                conditions={CONDITION.FAILURE},
                name='on-destination-after-failure',
                commands=split_remote_commands(
                    self._configuration.execute_on_destination_after_failure),
                job=self,
                ),
            ]
        return steps

    def _callProcess(self, *args, **kwargs):
        """
        This is here to help with testing.
        """
        return defer_spawn(*args, **kwargs)


class TransferJobsManager(RunnablesManager):
    """
    Collection manager of all active transfer jobs.

    It receives new job request from a transfer and acts like a job
    factory.

    For now, jobs can not be explicitly started.
    They are started when created.
    Also they can not be stopped/restarted.
    """
    #: Maximum number of batches in the queue of files to be transferred.
    #: The queue only keeps the source and destination and some minor info
    #: needed to fire a job.
    #: The actual job is not added to the queue.
    #: If we have more files pending to be transferred, the transfer will be
    #: stopped.
    _MAX_CAPACITY = 100000
    #: Maximum number of jobs executed simultaneously.
    _MAX_PARALLEL_JOBS = 1

    _TransferJob = TransferJob

    def __init__(self, parent, scheduler=reactor):
        super(TransferJobsManager, self).__init__(parent)
        self._scheduler = scheduler
        self._state = STATE.STOPPED
        self._resetState()

    def _resetState(self):
        """
        Reset the internal state.
        """
        # List of jobs to be executed.
        self._queue = []
        # List of jobs that are suspended.
        self._suspended = []
        # Ongoing job is active.
        self._processing = False
        # Delayed call to resume.
        self._resume_call = None

    def _initializeRunnables(self):
        """
        See: RunnablesManager.

        No runnables are created at init time.
        """
        pass

    def isQueued(self, path):
        """
        Return True if the job manager already contains a queued job for
        `path`.
        """
        for proto_job in self._queue:
            batch_files = proto_job[0]
            if path in batch_files:
                return True

        return False

    def isParentUsed(self, path):
        """
        Return True if the job manager already contains any active or
        queued job with a target path inside the parent `path`.
        """
        def is_parent(source, paths):
            """
            Return `True` if the base parent path
            is a parent of any file from `paths` list.
            """
            for batched_file in paths:
                try:
                    source.relativePath(path, batched_file)
                    return True
                except ServerException:
                    # Paths that aren't relative raise an exception and we
                    # just continue searching for relative paths.
                    continue
            return False

        # Check pending jobs.
        for job in self.getAll():
            if is_parent(job.source, job.getAllPaths()):
                return True

        # Check queued jobs.
        for proto_job in self._queue:
            batch_files = proto_job[0]
            source = proto_job[1]
            if is_parent(source, batch_files):
                return True

        return False

    def isActive(self, path):
        """
        Return True if the job manager already contains job for `path`
        that is currently operating.
        """
        # Check pending jobs.
        for job in self.getAll():
            if path in job.getAllPaths():
                return True

        return False

    def dequeue(self, path):
        """
        Remove path from the queue.

        Return True if file was actually dequeued, and False if file was not
        in the queue.
        """
        was_removed = False
        for proto_job in self._queue:
            if path in proto_job[0]:
                was_removed = True
                break

        if was_removed:
            # We modify the buffer outside of the iteration.
            if len(proto_job[0]) < 2:
                self._queue.remove(proto_job)
            else:
                proto_job[0].remove(path)

        return was_removed

    def queue(self, paths, source, destination, configuration):
        """
        Add the request for job creation to the processing queue.
        """

        # We were requested to transfer a new set of path.
        # We need to filter the set of duplicated and if a new request is
        # made for the current active job, the current job should be canceled
        # and started again.

        values = (
            paths,
            source,
            destination,
            configuration,
            )
        self._queue.append(values)
        self._process()

    def start(self, avatar=None):
        """
        See: IRunnablesManager.
        """
        self._resetState()
        self._state = STATE.STARTED
        return defer.succeed(None)

    @defer.inlineCallbacks
    def stop(self, avatar=None):
        """
        See: IRunnablesManager.
        """
        self._resetState()
        self._state = STATE.STOPPING
        # The jobs will never fail themselves.
        # On failure they trigger the transfer failure.
        # Trigger the finalization of the ongoing jobs.
        for job in self.getAll():
            yield job.stop()

        self._state = STATE.STOPPED

    def restart(self, avatar=None):
        """
        See: IRunnablesManager.

        Does nothing as jobs cannot be stopped.
        """
        return defer.succeed(None)

    def suspend(self, seconds):
        """
        Suspend the current job execution for `seconds` and continue
        processing next job.

        Return a deferred that is called when the job is resumed.
        """
        # The caller can wait for this to know when resumed.
        suspended_deferred = defer.Deferred()
        # When to execute the suspended job.
        scheduled = self._scheduler.seconds() + seconds

        # Add the current job in the suspended queue, based on the
        # scheduled time.
        start_schedule = []
        end_schedule = []
        for job in self._suspended:
            if job[0] < scheduled:
                start_schedule.append(job)
            else:
                end_schedule.append(job)
        # Start with sooner jobs.
        self._suspended = start_schedule
        # Then add our job.
        self._suspended.append((
            scheduled,
            suspended_deferred,
            ))
        # Then add jobs scheduled for later.
        self._suspended.extend(end_schedule)

        # Continue processing the queue.
        self._processing = False
        self._process()
        return suspended_deferred

    @defer.inlineCallbacks
    def _process(self):
        """
        Create and start `TransferJob` for each batch item. Return a deferred
        that will fire when all jobs finish.

        Fail the transfer that originated the job on errors.
        """
        if self._processing:
            # Ignore if already processing the queue.
            return

        if not self._queue and not self._suspended:
            # Nothing queued or suspended.
            return

        if (
            self._suspended
            and self._suspended[0][0] <= self._scheduler.seconds()
                ):
            # We have one or more suspended jobs
            # that are scheduled to be executed now.
            # Or has the time in the past, which means that it was scheduled
            # but we were busy to trigger it at the exact scheduled time.
            # Run it now.
            job = self._suspended.pop(0)
            resume_deferred = job[1]
            self._processing = True
            try:
                resume_deferred.callback(None)
                yield resume_deferred
            finally:
                self._processing = False

            # Try to process again.
            # This will prioritize the queue.
            # As long as we have jobs in the suspended queue we don't break
            # the recursive chain.
            # It should be fine as we should not have too many files in the
            # queue... and if we do have, then the transfer should be
            # stopped anyway.
            self._process()
            return

        if not self._queue:
            # Nothing in the transfer queue.
            # Stop further processing but schedule a new processing based on
            # the soonest to be executed job from the suspended queue.
            if self._resume_call:
                # Cancel the existing resume schedule as we will
                # refresh it.
                try:
                    self._resume_call.cancel()
                except (AlreadyCancelled, AlreadyCalled):
                    """
                    Ignore if the call was already cancelled or doesn't
                    need cancel.
                    """

            # Nothing queue and no suspended job to resume.
            # Schedule a wakeup for the next suspended job.
            next_time = self._suspended[0][0]
            next_resume = next_time - self._scheduler.seconds() + 0.01
            self._resume_call = self._scheduler.callLater(
                next_resume, self._process)
            return

        self._processing = True

        try:
            paths, source, destination, configuration = self._queue.pop(0)
            job = self._TransferJob(
                parent=self,
                paths=paths,
                source=source,
                destination=destination,
                configuration=configuration,
                )

            # Keep track of running jobs.
            self._runnables[job.uuid] = job
            try:
                yield job.start()
            finally:
                # Current job is done, mark it as soon as possible.
                self._processing = False

                # FIXME:2474:
                # Right now we discard the job from manager.
                # In the future we will want to show them in LM for a while and
                # allow retrying. At least the failed ones.
                del self._runnables[job.uuid]

        except Exception as error:
            # The job should never fail, but if it does, this is critical.
            if self._state != STATE.STARTED:
                # Already stopping or failing.
                # Don't overwrite the existing state.
                return
            self.parent.fail(error)

        finally:
            # Make the current job as done.
            self._processing = False

        # Continue processing the remaining files.
        # But call it a bit later to make sure we break the recursive chain.
        self._scheduler.callLater(0.001, self._process)


def split_remote_commands(commands):
    """
    Split the `commands` configuration option and return them as list.

    Command can be defined on separate lines or multiple commands on
    single line delimited with semicolon (;).
    """
    result = []
    if not commands:
        return result

    for line in commands.splitlines():
        for command in line.split(';'):
            command = command.strip()
            if not command:
                continue
            result.append(command)
    return result


def validate_remote_commands(value):
    """
    Check that `value` is a valid configuration for remote commands.
    """
    commands = split_remote_commands(value)
    for command in commands:
        public_format(command, **TransferJob.REMOTE_COMMAND_PLACEHOLDERS)


def _transform_path(source_path, groups, template, now):
    """
    Return the transformed `source_path` based on `template` and using regex
    `groups` value to expand the template.

    Besides the regex groups which are available as {group.1}, {group.2},
    the following substitutions are available:

    {group.0} - the whole source path. Placeholder to allow 1 based numbering.
                Its usage should be avoided.
    {now.TIMESTAMP_TYPE} - timestamp based on current date and time.
    {path.source} - source path relative to base source folder for the transfer
    {path.directory_path} - source directory name, relative to base source
      folder.
    {path.file_name} - Source file name and extension.
    {path.file_base} - Source file name without the extension.
    {path.file_extension} - Source file extension, including the dot.

    """
    module = get_path_handling_module(source_path)
    directory_path, file_name = module.split(source_path)
    file_base, file_extension = module.splitext(file_name)
    if not file_extension:
        # On Python 2 the empty extension is bytes.
        file_extension = ''

    path = {
        'source': source_path,
        'directory_path': directory_path,
        'file_name': file_name,
        'file_base': file_base,
        'file_extension': file_extension,
        }

    group = {}
    for index, value in enumerate(groups):
        group[str(index)] = value

    data = {
        'path': path,
        'group': group,
        'now': ChevahTimeStamp(now),
        }

    try:
        result = public_format(template, **Bunch.fromDict(data))
    except ServerException as error:
        raise ServerException(
            'Failed to transform the destination path. %s' % error.message)

    if not result:
        raise ServerException(
            'Transformation resulted in an empty path.')

    is_absolute_path = _is_absolute_path(result)
    if not is_absolute_path:
        # If destination is not absolute, return it as it is.
        return result

    is_absolute_path = _is_absolute_path(template)
    if is_absolute_path:
        # If template is an absolute path return the result as it is.
        return result

    # Force the result to be a relative path.
    # For example if the template uses `{path.directory_path}/TARGET` for a file
    # in source root we end up with `/TARGET` which is an absolute path.
    return _make_relative_path(result)


def _is_absolute_path(value):
    """
    Return a True if `value` is an absolute path, Unix or Windows.

    It also return true for "almost" Windows absolute
    paths line \\some\\path which is an absolute Windows Share server path.
    """
    if value[0] == '/':
        # Have a Unix absolute.
        return True

    if value[0] == '\\':
        # Have a Windows almost absolute,
        # from our templating point if view
        return True

    if len(value) < 3:
        # A 2 character path that is not a Unix absolute,
        # can only be a relative path.
        return False

    if value[1] == ':' and (value[2] == '\\' or value[2] == '/'):
        # Have a Windows absolute path.
        return True

    # Anything else is relative.
    return False


def _make_relative_path(value):
    """
    Make `value` absolute path into a relative path.
    """
    if value[0] == '/':
        # Have a Unix absolute.
        return value.lstrip('/')

    if value[0] == '\\':
        # Have a Windows almost absolute.
        return value.lstrip('\\')

    # Have Windows local drive absolute path.
    return value[3:]
