# -*- test-case-name: normal.transfer.test_transfer -*-
# Copyright (c) 2014 Adi Roiban.
# See LICENSE for details.
"""
Transfer metrics.

Record and query transfer metrics and logs.

For now it uses sqlite.
In the future we might look at using influxdb or victoria metrics or another
dedicated db.
"""
from bunch import Bunch
from twisted.internet import defer
from twisted.internet.task import LoopingCall

from chevah.server.commons.constant import (
    TABLE_PREFIX,
    )
from chevah.server.db.entity import (
    DBTable,
    DBTableColumn,
    DBTableIndex,
    )

PATH_STATE = Bunch(
    SUCCESS=10,
    START_MARKER=220,
    )


class _SkipEvent(Exception):
    """
    Raised when an event should be skipped and not handled.
    """

    def __init__(self, reason):
        self.reason = reason


class _OverlapEvent(Exception):
    """
    Raised when an event was generated for an active transfer.
    """


class TransferMemory(DBTable):
    """
    Table for each transfer with files which were transferred.
    """

    def __init__(self, name):
        super(TransferMemory, self).__init__(name)

        self.columns.add(DBTableColumn(
            'timestamp',
            'DOUBLE',
            description='Timestamp of the transfer.'))

        self.columns.add(DBTableColumn(
            'path',
            'VARCHAR',
            2000,
            description='Path of the file that was transferred.'
            ))

        self.columns.add(DBTableColumn(
            'modified',
            'DOUBLE',
            description='Timestamp when transfer file was modified.'))

        self.columns.add(DBTableColumn(
            'size',
            'DOUBLE',
            description='Size of the transfer file.'))

        self.columns.add(DBTableColumn(
            'state',
            'INTEGER',
            description='State of the transfer success | failure | retry.'))

        timestamp_index = DBTableIndex(
            'idx_transfer_memory_timestamp')
        timestamp_index.addColumn(self.columns.column('timestamp'))
        self.indexes.addIndex(timestamp_index)

        path_index = DBTableIndex('idx_transfer_memory_path')
        path_index.addColumn(self.columns.column('path'))
        self.indexes.addIndex(path_index)


class TransferMetrics:
    """
    Keep a record of transferred files and their metadata.
    """

    # Run cleanup operations twice per day.
    _CLEANUP_INTERVAL = 12 * 60 * 60
    # Check for minimum transfer count every 10 minutes.
    _CHECK_INTERVAL_FILE_COUNT = 10 * 60
    # Number of seconds to keep a record of transferred files.
    _TRANSFER_ANALYTICS_TIMESPAN = 370 * 24 * 60 * 60

    def __init__(self, transfer):
        self._transfer = transfer
        self._scheduler = transfer._scheduler
        self._configuration = transfer.configuration

        # Table used by the DB.
        self._table = TransferMemory(
            TABLE_PREFIX.TRANSFER_MEMORY + self._transfer.uuid)

        self._resetInternalState()

    def _resetInternalState(self):
        """
        Set the internal state the initial values.
        """
        # DB interaction used by this transfers.
        self._db = None

        # Path used inside the DB to record the start of the transfer.
        self._start_marker_path = 'START-MARKER-' + self._transfer.uuid

        # Local cache of the DB table with the last successfully
        # transferred file.
        # We only record the last transfer for each filename, to reduce
        # memory usage.
        self._last_path_memory = {}
        # Looping call to delete older files.
        self._memory_cleanup = None
        self._has_transfer_memory = 0
        # Set to true if in the last time interval,
        # the minium file count was missed.
        self._minimum_count_missed = False
        # Looping call to check minimum transferred files.
        self._minimum_count_check = None

    @property
    def _last_path_age(self):
        """
        Return the timestamp in Unix time with the older path
        that should be kept in memory.
        """
        memory_duration_seconds = (
            self._configuration.transfer_memory_duration * 24 * 60 * 60)
        return int(self._scheduler.seconds() - memory_duration_seconds)

    @defer.inlineCallbacks
    def start(self):
        """
        Called when we should initialize the DB.
        """
        # This is called at init and after stop,
        # but just to be safe.
        # We might want to raise an error if started without a clean state.
        self._resetInternalState()

        self._has_transfer_memory = (
            (self._configuration.transfer_memory_duration > 0) or
            (self._configuration.minimum_transfer_count > 0)
            )

        if self._has_transfer_memory <= 0:
            # No need to start the DB as the memory is not used.
            return

        self._db = self._transfer.root.resources.getDatabase()
        yield self._db.start()
        yield self._db.migrateTable(self._table)
        yield self._startMinimumTransferCount()
        yield self._startTransferMemory()

    def stop(self):
        """
        Called when transfer is stopped.
        """
        if self._memory_cleanup and self._memory_cleanup.running:
            self._memory_cleanup.stop()

        if self._minimum_count_check and self._minimum_count_check.running:
            self._minimum_count_check.stop()

        self._resetInternalState()

    def checkSkipEvent(self, event, operation):
        """
        See if file associated the event should be ignored.

        Raise _SkipEvent if file should be skipped.
        Raise _OverlapEvent if the file was changes while pending transfer.
        """
        target_path = event.current_path

        if self._transfer._jobs.isQueued(target_path):
            raise _SkipEvent(operation + ' path already queued.')

        if self._transfer._jobs.isActive(target_path):
            # FIXME:5437:
            # Here we should raise an error, or emit a different event
            # if path exists for a transfer which is a the 'transfer-step'
            # or beyond.
            raise _OverlapEvent()

        was_already_transferred = (
            self._configuration.transfer_memory_duration
            and self._last_path_memory.get(target_path, None)
            )
        if (
            was_already_transferred
            and self._configuration.ignore_duplicate_paths
                ):
            # We already have a file transferred with the same path.
            raise _SkipEvent(operation + ' path was already transferred.')

        source_filter_age = self._configuration.source_filter_age
        if source_filter_age > 0:
            age_limit = self._scheduler.seconds() - source_filter_age
            if event.modified < age_limit:
                raise _SkipEvent(operation + ' with older modified time.')

        # All good.
        return was_already_transferred

    def updateSuccess(self, attributes):
        """
        Record that the file was successfully transferred.

        This never blocks and will not wait for the DB operation to finalize.
        This should never raise an exception.
        """
        if self._has_transfer_memory <= 0:
            # Transfer memory is not used.
            return

        self._minimum_count_missed = False

        row = {
            'timestamp': int(self._scheduler.seconds()),
            'path': attributes.path,
            'modified': int(attributes.modified),
            'size': attributes.size,
            'state': PATH_STATE.SUCCESS,
            }

        # Update the last transferred path.
        self._last_path_memory[attributes.path] = row

        def eb_update_memory(failure):
            """
            Called when updating the memory fails.
            """
            self._transfer.fail(failure.value)

        criteria = [[
            (self._table.columns.column('path'), '=',  attributes.path),
            ]]
        deferred = self._db.deleteRows(self._table, criteria)
        deferred.addCallback(lambda _: self._db.insertRow(self._table, row))
        deferred.addErrback(eb_update_memory)
        self._update_transfer_deferred = deferred

    @defer.inlineCallbacks
    def _startTransferMemory(self):
        """
        Called at the start of the transfer manager, after DB is started,
        to load the transfer history into memory and schedule cleanups.
        """
        # Only keep in memory a record last path, based on the configured
        # memory duration.
        criteria = [[(
            self._table.columns.column('timestamp'), '>=', self._last_path_age
            )]]
        rows = yield self._db.selectRows(self._table, criteria)
        for row in rows:
            self._last_path_memory[row['path']] = row

        # Always run the DB cleanup.
        self._memory_cleanup = LoopingCall(self._clMemoryCleanup)
        self._memory_cleanup.clock = self._scheduler
        self._memory_cleanup.start(
            self._CLEANUP_INTERVAL, now=True)
        yield self._transfer_memory_cleanup_deferred

    @defer.inlineCallbacks
    def _startMinimumTransferCount(self):
        """
        Called af the start of the transfer manager, after DB is started,
        to setup the process for checking minimum transferred files.
        """
        if self._configuration.minimum_transfer_count < 1:
            # Not configured for minimum transfer.
            return

        start_marker = {
            'timestamp': int(self._scheduler.seconds()),
            'path': self._start_marker_path,
            'modified': 0,
            'size': 0,
            'state': PATH_STATE.START_MARKER,
            }
        yield self._db.insertRow(self._table, start_marker)

        # Only check if we have a configuration for minimum files.
        self._minimum_count_check = LoopingCall(self._clMinimumCountCheck)
        self._minimum_count_check.clock = self._scheduler
        self._minimum_count_check.start(
            self._CHECK_INTERVAL_FILE_COUNT, now=False)

    @defer.inlineCallbacks
    def _clMemoryCleanup(self):
        """
        Called regularly to do the auto delete.
        """
        self._cleanupLastPathMemory()

        max_db_age = int(
            self._scheduler.seconds() - self._TRANSFER_ANALYTICS_TIMESPAN)

        # Delete from DB without waiting for delete operation to finalize.
        criteria = [[
            (self._table.columns.column('timestamp'), '<', max_db_age),
            ]]
        try:
            deferred = self._db.deleteRows(self._table, criteria)
            # To help with testing to wait for the delete action.
            self._transfer_memory_cleanup_deferred = deferred
            yield deferred
        except Exception as error:
            self._transfer.fail(error)
            return

    def _cleanupLastPathMemory(self):
        """
        Called to cleanup memory for last file path transferred.
        """
        # Do a faster cache purge.
        purge_paths = []
        for path, attributes in self._last_path_memory.items():
            if attributes['timestamp'] < self._last_path_age:
                purge_paths.append(path)

        for path in purge_paths:
            del self._last_path_memory[path]

    @defer.inlineCallbacks
    def _clMinimumCountCheck(self):
        """
        Called regularly to check if the minimum files were transferred.
        """
        if self._minimum_count_missed:
            # We have already recorded a miss event.
            return

        minimum_count = self._configuration.minimum_transfer_count
        success_interval = self._configuration.minimum_transfer_interval
        file_count_age = int(self._scheduler.seconds() - success_interval)

        # Check to see if it's too soon to check.
        # If there is no start transfer event older than file_count_age,
        # it means that the transfer was just started and we need to wait
        # more.
        columns = self._table.columns
        start_marker_criteria = [
            [(columns.column('timestamp'), '<=', file_count_age)],
            [(columns.column('state'), '=', PATH_STATE.START_MARKER)],
            ]

        criteria = [[(columns.column('timestamp'), '>=', file_count_age)]]
        try:
            deferred = self._db.selectRows(
                self._table, start_marker_criteria, limit=1)
            self._minimum_count_deferred = deferred
            rows = yield deferred

            # Let the test know that we are done and no other DB operation
            # was triggered yet.
            self._minimum_count_deferred = None

            if len(rows) < 1:
                # Transfer started sooner than the count age.
                # Nothing to check yet.
                return

            deferred = self._db.selectRows(self._table, criteria)
            # To help with testing to wait for the DB action.
            self._minimum_count_deferred = deferred
            rows = yield deferred

            # Let the test know that we are done and no other DB operation
            # was triggered yet.
            self._minimum_count_deferred = None

        except Exception as error:
            self._transfer.fail(error)
            return

        if len(rows) >= minimum_count:
            # All ok.
            return

        self._minimum_count_missed = True
        self._transfer.emitEvent('60019', data={
            'minimum_count': minimum_count,
            'success_interval': success_interval,
            'transfer_count': len(rows),
            })
