# -*- test-case-name: normal.transfer.test_transfer -*-
# Copyright (c) 2021 Adi Roiban.
# See LICENSE for details.
"""
Handled the removal of parent source folders for a transfer

This is not supposed to be used outside of a transfer.
"""

from __future__ import absolute_import
from __future__ import unicode_literals

from twisted.internet import defer

from chevah.server import emit_error_details
from chevah.server.commons.exception import ServerException


class _ParentRemoval(object):
    """
    It handles the removal of source directories.

    Does nothing if the transfer is not recursive or not configured to
    delete source directory.

    When a path is added for removal, it does nothing if there is already
    a path for the same source directory in the queue of the transfer.
    Those queued files will trigger a removal once they are done.
    """

    # The state of the parent source directory scheduled for deletion.
    # This is a dict, but is define here as None so that each instance to
    # create its own queue.
    _delete_parent = None

    def __init__(
            self, scheduler, configuration, event_emitter, parent_checker):
        self._scheduler = scheduler
        self._configuration = configuration
        self._isParentUsed = parent_checker
        self.emitEvent = event_emitter
        self.reset(None)

    def reset(self, source):
        """
        Called to (re)start or stop the parent removal.
        """
        self._source = source
        self._clearDeleteParent()

    def _clearDeleteParent(self):
        """
        Cancel any pending delete operation and start with an empty state.
        """
        if self._delete_parent is None:
            # Initialization for this instance.
            self._delete_parent = {}
            return

        if not self._delete_parent:
            #  Nothing to clear.
            return

        for delayed_call in self._delete_parent.values():
            delayed_call.cancel()

        self._delete_parent = {}

    def add(self, path):
        """
        Called to schedule or update the deletion of a source directory based
        on file at `path`.
        """
        if not self._configuration.recursive:
            # Nothing to delete for non-recursive transfers.
            return

        if not self._configuration.delete_source_on_success:
            # If source file is not removed, parent will not be empty.
            # Maybe this disabled at run-time so we disable any previously
            # scheduled delete operations.
            self._clearDeleteParent()
            return

        delete_delay = self._configuration.delete_source_parent_delay
        if delete_delay <= 0:
            # Deletion of parent is disabled.
            # Maybe this disabled at run-time so we disable any previously
            # scheduled delete operations.
            self._clearDeleteParent()
            return

        parent_path, _ = self._source.splitPath(path)
        relative_parent_path = self._source.relativePath(
            self._configuration.source_path, parent_path)

        if not relative_parent_path:
            # We are in the base source directory.
            return

        scheduled_delete = self._delete_parent.get(parent_path, None)

        if scheduled_delete:
            scheduled_delete.reset(delete_delay)
            return

        self._delete_parent[parent_path] = self._scheduler.callLater(
            delete_delay,
            self._deleteSourceFileParent, self._source, parent_path)

    @defer.inlineCallbacks
    def _deleteSourceFileParent(self, source, path):
        """
        Called when it's time to delete `path` which is the parent path of
        a source file that was transferred and already removed.
        """
        # This should never fail as it's called by the scheduler out of order.
        # Instead emit an error so wrap all the code.
        try:
            # Start by removing the callback from the queue.
            del self._delete_parent[path]

            # Don't delete if there are still files in the queue
            # or active jobs.
            # Just ignore, as the parent will be rescheduled when the files
            # from the queue are transferred.
            if self._isParentUsed(path):
                return

            # Don't delete directories that still have files.
            members = yield source.getFolderMemberAttributes(path)
            if members:
                raise ServerException(
                    'Source directory not removed as '
                    'it still contains %s members.' % (len(members),))

            yield source.deleteFolder(path)

        except Exception as error:
            emit_error_details(
                event_id='60084',
                data={'path': path},
                error=error,
                emitter=self.emitEvent,
                )
