# -*- test-case-name: normal.transfer.test_schedule -*-
# Copyright (c) 2022 Adi Roiban.
# See LICENSE for details.
"""
Transfer high level scheduler.

Handles resume/suspend a transfer based on the schedule.
"""
from __future__ import absolute_import
from __future__ import unicode_literals
import datetime
import functools
import itertools

from ordered_set import OrderedSet
from twisted.internet import defer, reactor

from chevah.server.commons.constant import STATE
from chevah.server.commons.exception import ServerException

# Special value to denote a daily action.
_DAILY = 'daily'
# Name of week days and they ISO value.
_WEEKDAYS = {
    'monday': 1,
    'tuesday': 2,
    'wednesday': 3,
    'thursday': 4,
    'friday': 5,
    'saturday': 6,
    'sunday': 7,
    }
# Valid values for week days configuration.
_VALID_WEEKDAYS = (_DAILY,) + tuple(_WEEKDAYS.keys())
# All day values sorted.
_ALL_DAYS = [1, 2, 3, 4, 5, 6, 7]


@functools.total_ordering
class TransferAction(object):
    """
    Base class for all actions.

    week_day is a number between 0 and 6 to help with sorting.
    """

    name = 'not-defined'

    def __init__(self, week_day, time, callback, scheduler=reactor):
        self._week_day = week_day
        self._time = time
        self._callback = callback
        self._scheduler = scheduler

    def __eq__(self, other):
        if self.name == 'not-defined':
            raise RuntimeError('No equality for generic transfer actions.')

        return (
            self.name == other.name and
            self.week_day == other.week_day and
            self.time == other.time
            )

    def __lt__(self, other):
        if self.week_day == other.week_day:
            # For same day, arranged based on time.
            return self.time < other.time
        else:
            # For different days, arranged based on day.
            return self.week_day < other.week_day

    def __gt__(self, other):
        if self.week_day == other.week_day:
            # For same day, arranged based on time.
            return self.time > other.time
        else:
            # For different days, arranged based on day.
            return self.week_day > other.week_day

    def __repr__(self):
        return '<%s at %s> %s at %s' % (
            self.__class__, id(self), self.week_day, self.time)

    def _now(self):
        """
        Return the current local datetime.
        """
        return datetime.datetime.fromtimestamp(self._scheduler.seconds())

    @property
    def week_day(self):
        """
        Day of the week in which this action is active as 0 to 6 number.
        """
        return self._week_day

    @property
    def time(self):
        """
        Time at which action should be executed.

        It is an datetime.time.
        """
        return self._time

    @property
    def delay(self):
        """
        Number of seconds (as float) from now up to the time when action
        should be executed in the future.
        """
        base_offset = self.offset
        if base_offset < 0:
            return 7 * 86400 + base_offset
        else:
            # This is in the near future, no need to wrap the week.
            return base_offset

    @property
    def offset(self):
        """
        Number of seconds from now to the time when action should be executed
        as positive or negative offset within the current week.

        It is like the distance since the execution time, in the frame of a
        week, with negative values in the past.

        This is used to decide if the action should be executed in this week,
        or we should look for another action for this week.
        """
        now_local = self._now()
        now_week_day = now_local.isoweekday()
        scheduled_datetime = datetime.datetime.combine(now_local, self.time)
        delta = scheduled_datetime - now_local
        seconds_difference = delta.total_seconds()

        day_diference = self.week_day - now_week_day
        base_difference = day_diference * 86400

        return base_difference + seconds_difference

    def onExecute(self):
        """
        Called when action needs to execute.
        """
        return self._callback()


class StartAction(TransferAction):
    """
    Marker class for resuming a transfer.
    """
    name = 'resume'


class StopAction(TransferAction):
    """
    Marker class for suspending a transfer.
    """
    name = 'suspend'


class TransferActionScheduler(object):
    """
    Scheduler of TransferActions.

    This is the scheduler part decomposed from the Transfer class.
    """

    def __init__(
        self, configuration, on_resume, on_suspend, transfer,
        scheduler=reactor,
            ):
        self._scheduler = scheduler
        self._transfer = transfer
        self._delayed_call = None
        self._actions = self._parseScheduledActions(
            configuration, on_resume=on_resume, on_suspend=on_suspend)
        self._iterator = itertools.cycle(self._actions)

    def _now(self):
        """
        Return the current local datetime.
        """
        return datetime.datetime.fromtimestamp(self._scheduler.seconds())

    def start(self, on_already_started):
        """
        Called when scheduler should start processing the configured actions.

        `on_already_started` is called if transfer should be started before
        the first scheduled action is executed.
        """
        if self._delayed_call:
            raise RuntimeError('Should not be started multiple times')

        if not self._actions:
            # No scheduled actions, so keep the transfer always active
            # Without emitting any special event.
            return on_already_started()

        return self._handleFirstAction(on_already_started)

    def stop(self):
        """
        Called when scheduler should stop executing actions and should stop
        scheduling actions.
        """
        if self._delayed_call and self._delayed_call.active():
            self._delayed_call.cancel()

    def _handleFirstAction(self, on_already_started):
        """
        Called when scheduler starts.

        It schedules the first action and makes sure the transfer is in the
        right state.
        """
        first_action = next(self._iterator)

        last_action = self._actions[-1]
        if isinstance(last_action, StartAction):
            # Last action for the day is a start action, so the day begins
            # with a started transfer.
            already_started = True
        else:
            already_started = False

        current_action = first_action
        while True:
            delay = current_action.offset
            if delay >= 0:
                # We hit an action which is going to be scheduled today.
                # Go ahead and schedule it.
                break

            # Current action should have been already scheduled for today.
            if isinstance(current_action, StartAction):
                # Action was already executed for today and it started the
                # transfer.
                already_started = True
            elif isinstance(current_action, StopAction):
                # Action was already executed for today and it stopped the
                # transfer
                already_started = False
            else:
                # Ignore other actions.
                pass

            # Advance the loop.
            current_action = next(self._iterator)

            if current_action is first_action:
                # We made a loop. This mean that no action is going to be
                # executed today so get the final delay...which is for next
                # day.
                delay = current_action.delay
                break

        if delay < 30:
            # Next scheduled action is going to be run
            # after 30 seconds so no longer start the transfer as it will
            # be handled by the scheduled action.
            already_started = False

        start_state = STATE.SUSPENDED
        if already_started:
            start_state = STATE.STARTED

        # Inform the state in which the transfer was started.
        self._transfer.emitEvent('60021', data={
            'name': self._transfer.name,
            'state': start_state,
            })

        result = defer.succeed(None)
        if already_started:
            # Transfer should be started before next action is executed.
            result = on_already_started()

        self._scheduleAction(delay, current_action)
        return result

    def _scheduleAction(self, delay, action):
        """
        Called when an action needs to be scheduled.
        """
        scheduled_date = self._now() + datetime.timedelta(seconds=delay)
        self._transfer.emitEvent('60020', data={
            'name': self._transfer.name,
            'action': action.name,
            'date': scheduled_date.isoformat(),
            'seconds': delay,
            })
        self._delayed_call = self._scheduler.callLater(
            delay, self._onExecute, action)

    def _onExecute(self, action):
        """
        Called when scheduled action needs to be executed.
        """
        self._transfer.emitEvent('60022', data={
            'name': self._transfer.name,
            'action': action.name,
            })

        # Execute scheduled action and fail the transfer on execution failures.
        deferred = action.onExecute()

        def eb_action_execute(failure):
            """
            Called when scheduled action fails to execute.
            """
            return self._transfer.fail(failure.value)

        deferred.addErrback(eb_action_execute)

        if not self._transfer.started:
            # If transfer was already stopped, no longer schedule
            # the next action.
            return

        # Once action is executed, schedule next action without waiting for
        # it to complete.
        action = next(self._iterator)
        delay = action.offset
        if delay < 0:
            # Action is in the past.
            # We need to check if we need to catch up or schedule it for
            # tomorrow
            if action is self._actions[0]:
                # This is for tomorrow
                delay = action.delay
            else:
                # Catch up.
                expected_date = self._now() + datetime.timedelta(seconds=delay)
                self._transfer.emitEvent('60023', data={
                    'name': self._transfer.name,
                    'action': action.name,
                    'expected_date': expected_date.isoformat(),
                    })
                delay = 0

        self._scheduleAction(delay, action)

    def _parseScheduledActions(self, configuration, on_resume, on_suspend):
        """
        Return a list of TransferAction ordered by time.

        Format for an actions is WEEKDAY-24HOUR:MINUTE-ACTION_NAME.
        When WEEKDAY is missing it is assumed that that action is valid
        for each day of the week.
        """
        raw_schedule = []

        # Merge schedule configuration.
        # It is configured as multi-line to make it easier to introduce data
        # but is used as a single list.
        for line in configuration:
            for value in line:
                value = value.strip().lower()
                raw_schedule.append(value)

        if not raw_schedule:
            # Transfer is always active.
            return ()

        raw_schedule = OrderedSet(raw_schedule)

        action_mapper = {
            'start': (StartAction, on_resume),
            'stop': (StopAction, on_suspend),
            }

        actions = []

        for raw_event in raw_schedule:
            try:
                week_day, event_time, action_name = self._parseEvent(
                    raw_event, list(action_mapper.keys()))
            except RuntimeError as error:
                raise ServerException(
                    'Invalid "schedule" configuration "%s". %s' % (
                        ', '.join(raw_schedule), error.args[0]))

            klass, callback = action_mapper[action_name]

            if week_day == _DAILY:
                # Create event for each day of the week.
                for day in _ALL_DAYS:
                    actions.append(klass(
                        week_day=day,
                        time=event_time,
                        callback=callback,
                        scheduler=self._scheduler,
                        ))
            else:
                actions.append(klass(
                    week_day=_WEEKDAYS[week_day],
                    time=event_time,
                    callback=callback,
                    scheduler=self._scheduler,
                    ))

        return sorted(actions)

    def _parseEvent(self, raw_configuration, valid_actions):
        """
        It parse and validates the configuration.

        Return a tuple of weekday, event_time, action_name.

        Raise RuntimeError on invalid values.
        """
        parts = raw_configuration.split('-')

        try:
            # See if we have a weekday.
            week_day, raw_time, action_name = parts
        except ValueError:
            # Fallback to just time and consider the weekday daily.
            try:
                raw_time, action_name = parts
                week_day = _DAILY
            except ValueError:
                raise RuntimeError('Invalid format.')

        if week_day not in _VALID_WEEKDAYS:
            raise RuntimeError('Invalid week day "{}".'.format(week_day))

        action_name = action_name.strip()

        if action_name not in valid_actions:
            raise RuntimeError('Invalid action.')

        try:
            event_datetime = datetime.datetime.strptime(raw_time, "%H:%M")
            event_time = event_datetime.time()
        except ValueError:
            raise RuntimeError('Invalid time.')

        return (
            week_day.strip(),
            event_time,
            action_name,
            )
