# Copyright (c) 2015 Adi Roiban.
# See LICENSE for details.
"""
Steps used for a file transfer.

Step is a stateless action performed as part of a transfer.
Steps are executed by the job only if conditions match.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from twisted.internet import defer

from chevah.server import emit_error_details
from chevah.server.commons.exception import (
    EmittedError,
    ServerError,
    )


class StepBase(object):
    """
    Base class for all steps.
    """
    _name = 'no step name defined'

    def __init__(self, job, conditions, name):
        self._job = job
        self._conditions = conditions
        self._name = name

    @property
    def job(self):
        """
        Name of this step.
        """
        return self._job

    @property
    def conditions(self):
        """
        Conditions in which this step should be executed by the external job.

        This value will not affect the test execution, but rather it
        affects when the external job should execute it.

        Under a clean design, the conditions should not be part of the Step
        class, but to simplify the code, we attach the conditions to a step.
        """
        return self._conditions

    @property
    def name(self):
        """
        Name of this step.
        """
        return self._name

    def run(self):
        """
        Called when step should execute.
        """
        raise NotImplementedError('run not implemented')

    def _ebEmitAndConvertToEmittedError(self, failure, event_id, data):
        """
        Helper to emit a dedicated error together with generic or
        ServerError details and return the EmittedError.

        This can be used to log the error of this step and forward it to the
        job.
        """
        # Always emit the event.
        data['tb'] = failure.getBriefTraceback()
        emit_error_details(
            error=failure.value,
            event_id=event_id,
            data=data,
            emitter=self._job.emitEvent,
            )

        if failure.check(defer.CancelledError):
            return failure

        # Propagate error so that retry manager will know that job was
        # failed.
        if failure.check(ServerError):
            error = failure.value
        else:
            error = ServerError(event_id, data=data)

        # Keep the ServerError avatar in sync with the emitted one
        # from the job.
        error.avatar = self._job.avatar
        raise EmittedError(error)
