# Copyright (c) 2015 Adi Roiban.
# See LICENSE for details.
"""
Steps executed on local location.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
from twisted.internet import defer

from chevah.server.commons.constant import (
    CONDITION,
    TYPE_NAME,
    )
from chevah.server.commons.exception import (
    EmittedError,
    ServerError,
    ServerException,
    )
from chevah.server.transfer.step import StepBase
from chevah.server.location.commons import create_destination_parents


class ExecuteStep(StepBase):
    """
    Execute a local command.
    """

    def __init__(self, job, conditions, name, command, timeout):
        super(ExecuteStep, self).__init__(job, conditions, name)
        self._command = command
        self._spawn_timeout = timeout

    def run(self):
        """
        Try to execute command and emit events to inform execution's progress.

        Return a deferred which fires when command exit, on timeout or on
        errors.
        """
        if not self._command:
            return defer.succeed(None)

        source_path = self._job.source_file_path
        destination_path = self._job.destination_path
        transferred_local_path, _ = self._job.transferred_file_local_path
        data = {
            'path': transferred_local_path,
            'source_path': source_path,
            'destination_path': destination_path,
            'name': self._name,
            'command': self._command,
            'output': 'No output yet.',
            'error': 'No error yet.',
            'exit_code': -1,
            }

        self._job.emitEvent('60007', data=data)

        deferred = self._job._callProcess(
            command=[self._command, transferred_local_path],
            timeout=self._spawn_timeout,
            environment={
                'SOURCE_PATH': source_path,
                'DESTINATION_PATH': destination_path,
                },
            )

        def cb_done(result):
            """
            Called when command was successful.
            """
            exit_code, output, error = result
            data['output'] = output
            data['error'] = error
            data['exit_code'] = exit_code

            if exit_code != 0:
                raise ServerException('Non zero exit code.')

            self._job.emitEvent('60008', data=data)

        def emit_error(details):
            """
            Emit an error with `details`.
            """
            data['details'] = details
            self._job.emitEvent('60009', data=data)
            return EmittedError(ServerError('60009', data=data))

        def eb_done(failure):
            """
            Called when command failed.

            Log error and return failure.
            """
            if failure.check(defer.CancelledError):
                # Just forward the cancellation.
                return failure

            if failure.check(ServerException):
                raise emit_error(failure.value.message)

            raise emit_error('Internal error. %r' % (failure.value,))

        deferred.addCallback(cb_done)
        deferred.addErrback(eb_done)
        return deferred


class CopyTransferredFile(StepBase):
    """
    Make a copy of the transferred file into the specified folder.
    """

    def __init__(self, job, conditions, name, path):
        super(CopyTransferredFile, self).__init__(job, conditions, name)
        self._destination_parent_path = path

    def run(self):
        """
        See: StepBase.
        """
        if not self._destination_parent_path:
            return defer.succeed(None)

        has_failed = self._job.hasCondition(CONDITION.FAILURE)

        deferred, target_path = self._onArchiveFile(has_failed=has_failed)

        data = {
            'archive_path': target_path,
            'path': self._job.source_file_path,
            }

        def cb_success(destination_path):
            data['archive_path'] = destination_path
            self._job.emitEvent('60026', data=data)

        deferred.addCallback(cb_success)
        deferred.addErrback(
            self._ebEmitAndConvertToEmittedError,
            event_id='60025',
            data=data,
            )

        return deferred

    def _onArchiveFile(self, has_failed=False):
        """
        Do the actual archiving operation.

        Returns a tuple of (deferred, archive_file_path)
        Deferred which fires when archive is done.
        """

        archive_format = self._job.configuration.archive_format
        local_location = self._job.local_location
        if not local_location:
            deferred = defer.fail(ServerException(
                'Archiving is not supported for remote to remote transfers.'))
            return deferred, 'Skipped archive.'

        local_path, relative_source = self._job.transferred_file_local_path

        source_parent_path, source_name = local_location.splitPath(
            relative_source)

        archive_path = local_location.joinPath(
            self._destination_parent_path, source_parent_path)

        if archive_format != TYPE_NAME.EXACT:
            destination_name = self._job.getFileNameWithTimestamp(source_name)
        else:
            destination_name = source_name

        destination_path = local_location.joinPath(
            archive_path, destination_name)

        def cb_source_exists(result):
            """
            Called when we know if file to be archived exists or not.
            """
            if result is True:
                # All good, we have the transferred file.
                return on_source_exists(local_path, destination_path)

            elif self._job.hasLocalSource():
                # For successful transfers from local sources we should always
                # have the file present, so the fact that it was not present
                # at archive time is an critical error.
                return defer.fail(ServerException(
                    'Could not find the source local file.'))

            elif has_failed:
                # When transfer has failed, is OK if we don't have a file
                # when source is remote.
                deferred = local_location.touch(destination_path)
                deferred.addCallback(lambda _: destination_path)
                return deferred

            else:
                # Has remote source but transfer was not failed.
                return defer.fail(ServerException(
                    'Could not find the local file which was just '
                    'successfully transferred.'))

        def cb_destination_exists(result):
            """
            Called when we know if destination file exists to return the
            updated name of the actual destination path.
            """
            if result is False:
                return destination_path

            destination_name = self._job.getFileNameWithTimestamp(source_name)
            return local_location.joinPath(
                archive_path, destination_name)

        def on_source_exists(local_path, destination_path):
            """
            Called when we know that source exists.
            """
            deferred = defer.succeed(destination_path)

            if archive_format == TYPE_NAME.EXACT:
                deferred = local_location.exists(
                    destination_path, with_event=False)
                deferred.addCallback(cb_destination_exists)

            deferred.addCallback(cb_copy_archive)
            return deferred

        def cb_copy_archive(destination_path):
            """
            Called to do the actual archive copy operation.
            """
            deferred = local_location.copyFile(local_path, destination_path)
            deferred.addErrback(eb_copy_archive, local_path, destination_path)
            deferred.addCallback(lambda _: destination_path)
            return deferred

        def eb_copy_archive(failure, local_path, destination_path):
            """
            Called when copy operation fails at first try.
            """
            # Try to create the parent folders and then try again.
            deferred = create_destination_parents(
                location=local_location,
                base_path=self._destination_parent_path,
                target_path=relative_source,
                )
            deferred.addCallback(lambda _: local_location.copyFile(
                local_path, destination_path))
            deferred.addErrback(lambda _: failure)
            return deferred

        deferred = local_location.exists(local_path, with_event=False)
        deferred.addCallback(cb_source_exists)
        return (deferred, destination_path)
