# Copyright (c) 2015 Adi Roiban.
# See LICENSE for details.
"""
Location independent steps.
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from twisted.internet import defer
from twisted.python.failure import Failure

from chevah.server import emit_error_details, emit_from_error
from chevah.server.commons.constant import CONDITION, OVERWRITE_RULE, STATE
from chevah.server.commons.interface import ILocalLocation
from chevah.server.commons.exception import (
    AbortFileException,
    EmittedError,
    ServerError,
    ServerException,
    )
from chevah.server.event.handler import get_exception_details
from chevah.server.transfer.step import StepBase
from chevah.server.location.commons import create_destination_parents
from chevah.server.location.shell import EmbeddedClientShellCommands


class TransferFileStep(StepBase):
    """
    Copy file from source to destination.
    """
    _CHUNK_SIZE = 320 * 1024

    @property
    def _path(self):
        """
        Proxy for job source file.
        """
        return self._job.source_file_path

    @defer.inlineCallbacks
    def run(self):
        """
        Do the actual content transfer.
        """
        yield wait_location_start(
            self._job, self._name, location=self._job.source)
        yield wait_location_start(
            self._job, self._name, location=self._job.destination)
        attributes, source_stream = yield self._getSourceStream()
        destination_stream = yield self._openDestination(
            attributes, source_stream)
        yield self._startTransfer(source_stream, destination_stream, attributes)

    @defer.inlineCallbacks
    def _getSourceStream(self):
        """
        Return the source stream and its attributes.
        """
        # We first get the attributes of the source, as otherwise at least for
        # FTP, we can't run any other commands after we finalize
        # getting the source.
        #
        # With this, we also check the file still exists by the time we try
        # to transfer it.
        try:
            attributes = yield self._job.source.getAttributes(self._path)
        except defer.CancelledError:
            raise
        except Exception as error:
            # Called when we fail to get the attributes for source.
            #
            # If source doesn't exist, `getAttributes`` returns a failure.
            # But `getAttributes`` also returns a failure when connection fails,
            # while the file might still exist on the source.
            #
            # So we double check with an exists operation.
            # Any failure for the `exists` operation is raised right away
            # and this will trigger a retry.
            still_exists = yield self._job.source.exists(self._path)
            if not still_exists:
                # We don't have a connection failure, but file no longer exists.
                # This is confirmed by both `getAttributes` and `exists`
                # operations.
                # Raise a special exception to ignore this file and stop
                # retrying it.
                raise AbortFileException(
                    'File no longer available in source. %s' % (
                        get_exception_details(error)))

            # Raise the original error.
            # We might not have permissions to read the attributes of the file.
            raise error

        try:
            source_stream = yield self._job.source.getStreamForReading(
                self._path)
        except Exception as error:
            self._ebEmitAndConvertToEmittedError(
                Failure(error),
                event_id='60014',
                data={'path': self._path},
                )

        defer.returnValue((attributes, source_stream))

    def _startTransfer(self, source_stream, destination_stream, attributes):
        """
        Called after we have source and destination streams and data transfer
        can start.

        In the end, it will close both streams.
        """
        size = attributes.size
        start_time = self._job._scheduler.seconds()
        if (
            ILocalLocation.providedBy(self._job.source) and
            ILocalLocation.providedBy(self._job.destination)
                ):
            transfer_type = 'local-to-local'
        elif ILocalLocation.providedBy(self._job.destination):
            transfer_type = 'pull'
        elif ILocalLocation.providedBy(self._job.source):
            transfer_type = 'push'
        else:
            transfer_type = 'remote-to-remote'

        data = {
            'source_path': self._job.source_file_path,
            'destination_path': self._job.destination_path,
            'path': self._path,
            'size': size,
            'transfer_type': transfer_type,
            'content_action': self._job.transformer.name,
            }

        self.job.emitEvent('60010', data=data)

        def cb_transfer_done(result):
            """
            Called when the transfer was successful.
            """

            def eb_unwrap_FirstError(failure):
                failure.trap(defer.FirstError)
                return failure.value.subFailure

            deferred = close_streams()
            deferred.addErrback(eb_unwrap_FirstError)
            return deferred

        def eb_transfer_done(failure):
            """
            Called when the transfer has failed.
            """
            # We try to close the streams, and ignore any errors and forward
            # the original transfer failure.
            deferred = close_streams()
            deferred.addBoth(lambda _: failure)
            return deferred

        @defer.inlineCallbacks
        def close_streams():
            """
            Called when transfer is done (with success or failure) to close
            the source and destination streams.

            If source fails, it raises the error for the source,
            while still closing the destination.
            If destination also fails, the destination error is ignored.

            If only destination fails, it raises the error for the destination.
            """
            duration = self._job._scheduler.seconds() - start_time + 0.0001

            written = destination_stream.written
            speed = int(written / duration / 1024)

            data.update({
                'duration': duration,
                'speed': speed,
                'transferred_size': written,
                })

            source_error = None
            destination_error = None
            try:
                yield source_stream.close()
            except Exception as error:
                source_error = error

            try:
                yield destination_stream.close()
            except Exception as error:
                destination_error = error

            if source_error:
                raise source_error

            if destination_error:
                raise destination_error

        def cb_success(result):
            """
            Called as last step when transfer is done with success.
            """
            step_destination = self._job.addTemporaryTransfer(attributes)
            data.update({
                # FIXME:3230:
                # This should be removed once file dispatcher can select
                # the location and the source/destination.
                # At the same time, this will be kept for a while for
                # backward compatibility.
                'real_path': step_destination,
                # Add user friendly names.
                # Note that destination might be the temporary path.
                'source_path': self._path,
                'destination_path': step_destination,
                })

            self.job.emitEvent('60012', data=data)

        deferred = self._doTransfer(source_stream, destination_stream)
        deferred.addCallbacks(cb_transfer_done, eb_transfer_done)
        deferred.addCallback(cb_success)
        deferred.addErrback(
            self._ebEmitAndConvertToEmittedError, event_id='60013', data=data)
        return deferred

    def _openDestination(self, attributes, source_stream):
        """
        The the best to get a destination for the transfer.
        """
        deferred = self._getDestinationStream(source_attributes=attributes)
        deferred.addErrback(
            self._ebGetDestinationFirst, source_stream, attributes)
        return deferred

    def _getDestinationStream(self, source_attributes):
        """
        Get destination stream for this transfer.
        """
        transfer_path = self._job.destination_path
        temporary_path = self._job.temporary_path

        if temporary_path:
            transfer_path = temporary_path

        deferred = self._job.destination.getStreamForWriting(
            path=transfer_path,
            length=source_attributes.size,
            )
        deferred.addErrback(
            self._ebEmitAndConvertToEmittedError,
            event_id='60015',
            data={'path': transfer_path},
            )
        return deferred

    @defer.inlineCallbacks
    def _doTransfer(self, source, destination):
        """
        Copy data from source to destination stream.
        """
        self._job.transformer.reset()
        while True:
            data = yield source.read(self._CHUNK_SIZE)
            if not data:
                break

            data = self._job.transformer.consume(data)
            yield destination.write(data)

        data = self._job.transformer.finalize()
        if data:
            yield destination.write(data)

    @defer.inlineCallbacks
    def _ebGetDestinationFirst(self, failure, source_stream, attributes):
        """
        Called when we fail to get the destination for the first time.

        Transfer will not start so source will not be closed in the
        normal callback.
        """
        try:
            destination_stream = yield self._getRecoveredDestination(
                initial_error=failure.value,
                attributes=attributes,
                )
        except Exception as error:
            self._cleanupSourceOnError(source_stream)
            raise error

        defer.returnValue(destination_stream)

    @defer.inlineCallbacks
    def _getRecoveredDestination(self, initial_error, attributes):
        """
        Try to get the destination stream after an initial failure.

        This is here to make it easy to wrap any error and close the source.
        """
        destination = self._job.destination

        if self._job.configuration.destination_fallback_path:
            # Try to deliver to fallback.
            self._job.setFallbackPath()
            # We try to open the fallback and on error we report the fallback
            # error as the main error.
            result = yield self._getDestinationStream(
                source_attributes=attributes)
            defer.returnValue(result)

        if not self._job.configuration.recursive:
            # We don't try to create destination parent for non
            # recursive transfers.
            raise initial_error

        # Maybe we failed to open the destination as parent folder
        # don't exists.
        # With all the protocols and error messages,
        # it is hard to tell why we failed to open the destination,
        # so we try to create the parent folder and to open destination again.
        try:
            base_path = self._job.configuration.destination_path
            relative_path = destination.relativePath(
                base_path, self._job.destination_path)

            yield create_destination_parents(
                location=destination,
                base_path=base_path,
                target_path=relative_path)
        except Exception:
            # We hide the error for creating parent and just report the
            # initial error.
            raise initial_error

        result = yield self._getDestinationStream(source_attributes=attributes)
        defer.returnValue(result)

    def _cleanupSourceOnError(self, source_stream):
        """
        Called when we failed to open the destination.

        Designed to be called as cleanup.
        It doesn't return or block.
        But it handles/logs source close errors.

        This is also called when we fail to create the parent folder,
        after an initial attempt to open destination.
        """
        # We don't wait for close operation, but we do need to
        # handle its errors.
        def eb_close_source(failure):
            data = {
                'path': self._path,
                'destination_path': self._job.destination_path,
                }
            emit_error_details(
                error=failure.value,
                event_id='60069',
                data=data,
                emitter=self.job.emitEvent,
                )

        deferred = source_stream.close()
        # Just handle the error without returning anything.
        deferred.addErrback(eb_close_source)


class DeleteSourceStep(StepBase):
    """
    Delete source file.
    """
    def run(self):
        """
        See: StepBase.
        """
        deferred = wait_location_start(
            self._job, self._name, location=self._job.source)
        deferred.addCallback(
            lambda _: self._job.source.deleteFile(self._job.source_file_path))
        return deferred


class RenameTemporaryDestination(StepBase):
    """
    Rename the file to final destination, if a temporary file was used.
    """

    def run(self):
        """
        See: StepBase.
        """
        temporary_path = self._job.temporary_path
        final_path = self._job.destination_path

        if not temporary_path:
            deferred = defer.succeed(None)
        else:
            deferred = wait_location_start(self._job, self._name)
            deferred.addCallback(lambda _: self._job.destination.rename(
                source_path=temporary_path,
                destination_path=final_path,
                ))

        deferred.addCallback(self._cbRenameOK)
        return deferred

    def _cbRenameOK(self, result):
        """
        Called when the rename operation was done with success.
        Also called when rename is not needed and this succeeded due to a noop.
        """
        self._job.onFinalTransfer()
        return result


class EnforceOverwritingRuleStep(StepBase):
    """
    It check if remote file should be overwritten and handler the overwriting
    rules.

    Executed before a file it transferred to destination.
    """
    @defer.inlineCallbacks
    def run(self):
        """
        Called after before step and does the actual content transfer.
        """
        configuration = self._job.configuration.overwrite_rule.lower()

        if configuration == OVERWRITE_RULE.DISABLE:
            # Nothing to do.
            return

        def emitted_error(error):
            """
            Emit and return the emitted error.
            """
            emit_from_error(error, emitter=self.job.emitEvent)
            # Sync the avatar for the error, as the job.emitEvent is injecting
            # the avatar.
            error.avatar = self._job._avatar
            return EmittedError(error)

        if configuration not in list(OVERWRITE_RULE.values()):
            raise emitted_error(ServerError('60027', data={
                'name': self._job.name,
                'details': configuration,
                }))

        if configuration == OVERWRITE_RULE.TIMESTAMP_ALWAYS:
            # Just amend the destination path
            self._job.addTimestampToDestinationPath()
            return

        yield wait_location_start(self._job, self._name)

        exists = yield self._job.destination.exists(self._job.destination_path)

        if not exists:
            # All good, destination file does not already exist.
            return

        if configuration == OVERWRITE_RULE.SKIP:
            self.job._setConditions({CONDITION.SKIP})
            self.job.emitEvent('60076', data={
                'name': self._job.name,
                'path': self._job.source_file_path,
                })
            return

        if configuration == OVERWRITE_RULE.FAIL:
            # Destination exist and we are configured not to allow
            # overwriting.
            raise emitted_error(ServerError('60028', data={
                'name': self._job.name,
                'path': self._job.destination_path,
                }))

        if configuration == OVERWRITE_RULE.OVERWRITE:
            # Inform that remote file exist and is going to be overwritten.
            self.job.emitEvent('60029', data={
                'name': self._job.name,
                'path': self._job.source_file_path,
                })
            return

        if configuration == OVERWRITE_RULE.TIMESTAMP_TO_NEW_FILE:
            # New file is copied with a timestamp.
            self._job.addTimestampToDestinationPath()
            self.job.emitEvent('60031', data={
                'name': self._job.name,
                'path': self._job.source_file_path,
                'destination_path': self._job.destination_path,
                })
            return

        if configuration == OVERWRITE_RULE.TIMESTAMP_TO_EXISTING_FILE:
            existing_path = self._job.destination_path
            renamed_path = self._job.getPathWithTimestamp(
                self._job.destination, existing_path)
            yield self._job.destination.rename(existing_path, renamed_path)
            # Inform that existing file was renamed on destination.
            self.job.emitEvent('60030', data={
                'name': self._job.name,
                'existing_path': existing_path,
                'path': renamed_path,
                })
            return

        raise RuntimeError('Execution should not reach this place.')


def wait_location_start(job, step_name, location=None):
    """
    Return a deferred which is fired when `job` location is
    started.

    It will try to auto-start the location if now started.
    """
    if not location:
        location = job.destination

    if location.state is STATE.STARTED:
        # Already started and operational.
        return defer.succeed(True)

    destination_start = defer.Deferred()

    def on_connect(signal):
        """
        Called when location was connected.
        """
        job.emitEvent('60082', data={
            'path': job.source_file_path,
            'step_name': step_name,
            'location': location.name,
            })
        destination_start.callback(None)

    def bb_unsubscribe(result_or_failure):
        location.unsubscribe('connect', on_connect)
        return result_or_failure

    destination_start.addBoth(bb_unsubscribe)
    location.subscribe('connect', on_connect)

    job.emitEvent('60070', data={
        'path': job.source_file_path,
        'step_name': step_name,
        'location': location.name,
        })

    if location.failed:
        # We don't try to manually restart a failed location as most
        # probably it will fail again.
        return destination_start

    def eb_heartbeat(failure):
        """
        Called when heartbeat failed.

        Ignore any errors as we only care when the location is successfully
        connected.
        """
        return

    deferred = location.heartbeat()
    deferred.addErrback(eb_heartbeat)

    return destination_start


class StartDestinationStep(StepBase):
    """
    Make sure destination is started and ready for the transfer.

    If destination is not available, the whole queue should be blocked
    as we don't want to waste file transfer retired due to failing destination.
    """

    def run(self):
        """
        See: StepBase.
        """
        return wait_location_start(self._job, self._name)


class ExecuteDestinationStep(StepBase):
    """
    Execute a list of commands on destination.
    """

    def __init__(self, job, conditions, name, commands):
        super(ExecuteDestinationStep, self).__init__(job, conditions, name)
        self._configured_commands = commands
        self._shell = EmbeddedClientShellCommands()

    def run(self):
        """
        See: StepBase.
        """
        if not self._configured_commands:
            return defer.succeed(None)

        deferred = wait_location_start(self._job, self._name)
        deferred.addCallback(self._cbDestinationStarted)
        return deferred

    def _cbDestinationStarted(self, ignored):
        # We need to resolve them as run time, as at init time we don't
        # have the path set, and the path can change during job execution.
        resolved_commands = self._job.resolveRemoteCommands(
            self._configured_commands)
        self._shell.setBatchCommands(resolved_commands)

        data = {
            'condition': self.name,
            'name': self._job.name,
            }
        self.job.emitEvent('60064', data=data)
        deferred = self._shell.start(location=self._job.destination)
        deferred.addCallback(self._cbCommandsDone, data)
        deferred.addErrback(
            self._ebEmitAndConvertToEmittedError, '60066', data)
        return deferred

    def _cbCommandsDone(self, result, data):
        """
        Called when commands processing was done.
        """
        # We get message without any leading of trailing spaces of new lines.
        error_message = self._shell.getErrors()
        if error_message:
            # There were errors while executing the commands.
            raise ServerException(error_message)
        self.job.emitEvent('60065', data=data)
