# Copyright (c) 2015 Adi Roiban.
# See LICENSE for details.
"""
Steps which are part of the transfer language.
"""
from __future__ import unicode_literals
from __future__ import absolute_import
from twisted.internet import defer

from chevah.server import emit_error_details
from chevah.server.transfer.step import StepBase
from chevah.server.commons.exception import (
    AbortFileException,
    ServerException,
    )


class RetryStep(object):
    """
    Retry a step for a number of times and waiting for a number of seconds.

    If step still fails, it will forward the last error.
    """

    def __init__(self, step, retry_count, retry_wait):
        self._step = step
        self._retry_count = retry_count
        self._retry_wait = retry_wait
        self._current_retries = 0

    @property
    def _job(self):
        """
        See: StepBase.

        Just mirrors the wrapped step.
        """
        return self._step.job

    @property
    def conditions(self):
        """
        See: StepBase.

        Just mirrors the wrapped step.
        """
        return self._step.conditions

    @property
    def name(self):
        """
        See: StepBase.

        Just mirrors the wrapped step.
        """
        return self._step.name

    def run(self):
        """
        See: StepBase.
        """
        deferred = self._step.run()
        deferred.addErrback(self._ebRetry)
        return deferred

    def _ebRetry(self, failure):
        """
        Called when wrapped step fails.
        """
        if failure.check(AbortFileException, defer.CancelledError):
            # Forward specific failures to be handled by the general job
            # code.
            return failure

        if self._current_retries >= self._retry_count:
            # We have retried enough.
            # Just return the failure without emitting it as it will be
            # emitted as the final failure.
            return failure

        # Emit failure to know that transfer was retried.
        data = {
            'path': self._job.source_file_path,
            'count': self._retry_count - self._current_retries,
            'wait': self._retry_wait,
            }
        emit_error_details(
            failure.value,
            '60011',
            data=data,
            emitter=self._job.emitEvent,
            )

        # Try at least one more time
        self._current_retries += 1

        # FIXME:5441:
        # Should not suspend if transfer order is enforces.
        # Suspend the job until the next retry.
        # This needs to return a deferred so that we pause the execution
        # of this job.
        deferred = self._job.parent.suspend(self._retry_wait)
        deferred.addCallback(lambda _: self.run())
        return deferred


class IterateBatchStep(StepBase):
    """
    Execute steps for each file from the batch.
    """

    def __init__(self, job, conditions, name, steps):
        self._steps = steps
        super(IterateBatchStep, self).__init__(
            job=job, conditions=conditions, name=name)

    @defer.inlineCallbacks
    def run(self):
        """
        See: `StepBase`.
        """
        first_error = None
        # We get a copy of the conditions to reset them before
        # executing each iteration.
        conditions = self._job._getConditions()

        for path in self._job.getAllPaths():
            try:
                self._job.setNextTransferPath(path)
            except ServerException as error:
                raise error

            try:
                self._job._setConditions(conditions)
                yield self._job.runSteps(self._steps[:])
            except Exception as error:
                if first_error:
                    continue
                first_error = (error, self._job._getConditions())

        if first_error:
            # When we have an error, we raise the error and
            # set conditions to the conditions found after first failed path.
            error, conditions = first_error
            self._job._setConditions(conditions)
            raise error
