# -*- test-case-name: normal.transfer.test_transfer -*-
# Copyright (c) 2014 Adi Roiban.
# See LICENSE for details.
"""
Transfer management.

A transfer should rely only on locations to perform operations.
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from chevah.server.event.handler import emit_error_details

from twisted.internet import error as twisted_error, defer, reactor
from twisted.internet.task import LoopingCall
from zope.interface import implementer

from chevah_compat import local_filesystem

from chevah.server import emit_from_error, force_unicode
from chevah.server.commons.constant import (
    ADDRESS_UNKNOWN,
    DEFAULT_LOCAL_FILESYSTEM_UUID,
    EVENT_TYPE,
    STATE,
    TRANSFER_ARCHIVE_FORMATS,
    TYPE_NAME,
    )
from chevah.server.commons.exception import (
    ConnectException,
    EmittedError,
    ServerError,
    ServerException,
    )

from chevah.server.commons.interface import (
    ITransfer,
    ITransfersManager,
    )

from chevah.server.commons.match import get_matched_groups
from chevah.server.commons.runnable import (
    Runnable,
    RunnablesManagerWithConfigurationSet,
    )
from chevah.server.identity.avatar import TransferAvatar
from chevah.server.location.watch import FilteredWatchAdapter
from chevah.server.transfer.job import (
    TransferJobsManager,
    validate_remote_commands,
    )
from chevah.server.transfer.parent_removal import _ParentRemoval
from chevah.server.transfer.scheduler import TransferActionScheduler
from chevah.server.transfer.transformer import validate_transformers
from chevah.server.transfer.metrics import (
    TransferMetrics,
    _OverlapEvent,
    _SkipEvent,
    )


@implementer(ITransfersManager)
class TransfersManager(RunnablesManagerWithConfigurationSet):
    """
    Management of available transfers.
    """

    def _getRunnablesConfiguration(self):
        """
        See: `RunnablesManagerWithConfigurationSet`.
        """
        return self.parent.configuration.transfers

    def _createRunnable(self, configuration):
        """
        Create a new transfer instance and return it.
        """
        return Transfer(parent=self, configuration=configuration)


@implementer(ITransfer)
class Transfer(Runnable):
    """
    A manager of client side file transfers.

    A transfer has the following extra states:
    * suspended - started but on pause/not acting on source changes

    * SOURCE_FAILED - waiting for source to become available, for example due
        to network failures. Once the source is available the transfer will
        auto resume
    * DESTINATION_FAILED - waiting for destination to become available.

    It handles availability of source and destination of the transfer.

    It is a long lived object that for the files created/changed in the source
    it creates TransferJob for each file or batch of files that needs to be
    transferred.

    It only creates the job, but does not manages the jobs.

    It manages the time at which the transfer is activated / suspended /
    resumed / run.
    """

    PROPERTIES_REQUIRING_RESTART = (
        'recursive',
        'source_path',
        'source_uuid',
        'destination_uuid',
        'destination_path_actions',
        'changes_poll_interval',
        'stable_interval',
        'schedule',
        'transfer_memory_duration',
        'minimum_transfer_count',
        'archive_format',
        )

    _STARTED_STATES = Runnable._STARTED_STATES + [
        STATE.SOURCE_FAILED,
        STATE.DESTINATION_FAILED,
        ]

    # This is here only to help validate the configuration.
    # The actual action implementation is in job.py.
    _SUPPORTED_ACTIONS = [
        TYPE_NAME.NO_ACTION,
        TYPE_NAME.FIXED_PATH,
        TYPE_NAME.LOWERCASE,
        TYPE_NAME.REPLACE_SEPARATOR,
        TYPE_NAME.ONLY_FILENAME,
        TYPE_NAME.TEMPORARY_PREFIX,
        TYPE_NAME.TEMPORARY_SUFFIX,
        TYPE_NAME.TRANSFORM,
        ]
    _SINGLE_ARGUMENT_ACTIONS = [
        TYPE_NAME.FIXED_PATH,
        TYPE_NAME.REPLACE_SEPARATOR,
        TYPE_NAME.TEMPORARY_PREFIX,
        TYPE_NAME.TEMPORARY_SUFFIX,
        TYPE_NAME.TRANSFORM,
        ]

    # Run cleanup operations twice per day.
    _CLEANUP_INTERVAL = 12 * 60 * 60

    # Low level schedule use to schedule future action.
    _scheduler = reactor
    # The avatar of this transfer.
    _avatar = None

    def __init__(self, parent, configuration):
        super(Transfer, self).__init__(parent, configuration)

        self._resetInternalState()

        # Dispatcher based on event types.
        self._onChangeMap = {
            EVENT_TYPE.CREATED: self._onCreated,
            EVENT_TYPE.EXISTING: self._onExisting,
            EVENT_TYPE.MODIFIED: self._onModified,
            EVENT_TYPE.MOVED: self._onMoved,
            }

        self._jobs = TransferJobsManager(
            parent=self, scheduler=self._scheduler)

        self._parent_removal = _ParentRemoval(
            self._scheduler,
            configuration,
            event_emitter=self.emitEvent,
            parent_checker=self._jobs.isParentUsed,
            )

    @property
    def type(self):
        """
        See: `IRunnable`.
        """
        return ''

    @property
    def avatar(self):
        """
        The client identity for this transfer.
        """
        return self._avatar

    def _resetInternalState(self):
        """
        Set the internal state the initial values.
        """
        # A transfer can be reconfigured, so source and destination are only
        # set on start.
        self._source = None
        # Cached path which is monitored.
        self._source_path = None
        self._destination = None
        self._monitor_handler = None
        self._action_scheduler = None
        self._destination_path_actions = None

        # Mapping of batched transfers to the delayed.
        # For time bases, there is a single call.
        # For file marker, the batches are split based on the marker.
        self._batchCalls = {}
        # List of the paths queued for a batch transfer.
        self._batchPaths = []

        self._archive_cleanup = None
        self._metrics = None

    def emitEvent(self, event_id, data=None, avatar=None, peer=None):
        """
        See: IRunnable.

        `avatar` can be the job representation.
        When we don't have an avatar, we use the process representation.
        """
        if not avatar:
            avatar = self._avatar

        self.root.emitEvent(
            event_id,
            data=data,
            avatar=avatar,
            creator=self,
            )

    @property
    def state(self):
        """
        See: `IRunnable`.
        """
        if self._action_scheduler and not self._monitor_handler:
            # Scheduler exists but source is not monitored.
            return STATE.SUSPENDED
        elif self._source and self._source.failed:
            # Source is not available, so the transfer is stalled for now.
            return STATE.SOURCE_FAILED
        elif self._destination and self._destination.failed:
            # Destination is not available, so all transferred files will
            # fail.
            return STATE.DESTINATION_FAILED
        elif (
            self._state == STATE.STARTED
            and self._source
            and not self._source.operating
                ):
            return STATE.SOURCE_STOPPED
        else:
            return self._state

    @defer.inlineCallbacks
    def _onStart(self, avatar=None):
        """
        See: `Runnable`.

        This is called when the general transfer starts. It will then
        schedule the resume operation.
        """
        self._source = self._getLocation(self._configuration.source_uuid)
        if not self._source:
            raise ServerException('Unknown source UUID "%s".' % (
                self._configuration.source_uuid))

        self._destination = self._getLocation(
            self._configuration.destination_uuid)
        if not self._destination:
            raise ServerException('Unknown destination UUID "%s".' % (
                self._configuration.destination_uuid))

        if not self._configuration.source_path:
            raise ServerException('No source path defined.')

        if not self._configuration.destination_path:
            raise ServerException('No destination path defined.')

        if (
            self._configuration.archive_format
            not in list(TRANSFER_ARCHIVE_FORMATS.keys())
                ):
            raise ServerException('Unknown archive format configured.')

        # Check that the configured transformers are ok at start time.
        validate_transformers(self._configuration.destination_content_actions)

        # We use a cached value which is already validated.
        self._destination_path_actions = (
            self._configuration.destination_path_actions)
        for rule in self._destination_path_actions:
            if len(rule) < 2:
                raise ServerException(
                    'The destination_path_actions needs to have at least a '
                    'path matching expression and the name of the action.')

            action_name = rule[1].lower()
            if action_name not in self._SUPPORTED_ACTIONS:
                raise ServerException(
                    'Unknown action name "%s".' % (action_name,))

            if (
                action_name in self._SINGLE_ARGUMENT_ACTIONS
                    ):
                if len(rule) != 3:
                    raise ServerException(
                        'Action "%s" has only one parameter.' % (action_name,))
                else:
                    """
                    It has the right number of options.
                    """
            elif len(rule) != 2:
                raise ServerException(
                    'Action "%s" need no options.' % (action_name,))
            else:
                """
                Rule has the right number of options.
                """

            # We do a generic validation of the matching rule.
            get_matched_groups(
                value='', rule=rule[0], usage='destination_path_actions')

        # Validate the configuration for batch marker.
        for rule in self.configuration.batch_marker_path:
            get_matched_groups(
                value='',
                rule=rule,
                usage='batch_marker_path',
                )

        # Validate the configuration for remote commands.
        remote_commands = [
            'execute_on_destination_before',
            'execute_on_destination_after_success',
            'execute_on_destination_after_failure',
            ]
        for command in remote_commands:
            value = getattr(self._configuration, command)
            try:
                validate_remote_commands(value)
            except Exception as error:
                raise ServerException('Invalid "%s": %s' % (command, error))

        self._avatar = TransferAvatar(
            uuid=self._configuration.uuid,
            name=self._configuration.name,
            component=self,
            # By default we use the source location.
            # There are also some events emitted by the transfer for the
            # destination, but most for source.
            peer=self._source.avatar.peer,
            )
        # This is the avatar for the events that are not associated with
        # the source or destination locations.
        self._avatar_peerless = TransferAvatar(
            uuid=self._configuration.uuid,
            name=self._configuration.name,
            component=self,
            peer=ADDRESS_UNKNOWN,
            )

        self._parent_removal.reset(self._source)

        # Start with clean batch state.
        self._batchCalls = {}
        self._batchPaths = []

        self._action_scheduler = TransferActionScheduler(
            configuration=self._configuration.schedule,
            on_resume=self._onResume,
            on_suspend=self._onSuspend,
            transfer=self,
            scheduler=self._scheduler,
            )

        self._metrics = TransferMetrics(self)
        yield self._metrics.start()
        yield self._jobs.start(avatar=avatar)

    def _clArhiveCleanup(self):
        """
        Called in a loop to clean the archive folders.
        """
        retention_period = self.configuration.archive_retention_period
        if not retention_period > 0:
            # Keep all files.
            return

        self._cleanArchiveFolder(
            self.configuration.archive_success_path, retention_period)

        self._cleanArchiveFolder(
            self.configuration.archive_failure_path, retention_period)

    def _cleanArchiveFolder(self, path, retention_period):
        """
        Clean folder at `path` of all files older than `retention_period` days.
        """
        if not path:
            return

        age_limit = self._scheduler.seconds() - retention_period * 24 * 60 * 60

        segments = local_filesystem.getSegmentsFromRealPath(path)

        try:
            self._deleteOldFiles(segments, age_limit, retention_period)
        except Exception as error:
            self.fail(error)

    def _deleteOldFiles(self, segments, age_limit, retention_period):
        """
        Recursive delete old files from `segments`.
        """
        for member in local_filesystem.iterateFolderContent(segments):
            if member.is_folder:
                # Do recursive search.
                self._deleteOldFiles(
                    segments + [member.name],
                    age_limit,
                    retention_period,
                    )
                # We don't even try to delete the directory itself.
                continue

            # Here we do file handling.
            if member.modified > age_limit:
                # Not old enough.
                continue

            local_filesystem.deleteFile(segments + [member.name])
            self.emitEvent(
                '60077',
                # The archive deletion event is not associated with the
                # source location.
                avatar=self._avatar_peerless,
                data={'path': member.path, 'days': retention_period},
                )

    def _afterStart(self, avatar=None):
        """
        See: `Runnable`.
        """
        self._archive_cleanup = LoopingCall(self._clArhiveCleanup)
        self._archive_cleanup.clock = self._scheduler
        self._archive_cleanup.start(self._CLEANUP_INTERVAL, now=True)

        # Scheduler is started after runnable is started so that the
        # event informing that runnable was started is emitted before any
        # file is transferred.
        if not self._action_scheduler:
            # Start might already fail.
            return
        return self._action_scheduler.start(on_already_started=self._onResume)

    def _getLocation(self, uuid):
        """
        Return location with `uuid` or default location when `uuid` is None.
        """
        if not uuid:
            uuid = DEFAULT_LOCAL_FILESYSTEM_UUID
        return self._configuration.root.locations.getSection(uuid)

    @defer.inlineCallbacks
    def _onStop(self, avatar=None):
        """
        See: `Runnable`.

        This is also called on failure.
        """
        if self._metrics:
            self._metrics.stop()

        if self._archive_cleanup and self._archive_cleanup.running:
            self._archive_cleanup.stop()
        self._archive_cleanup = None

        if self._action_scheduler:
            self._action_scheduler.stop()
            self._action_scheduler = None
        self._onSuspend()

        for batched_call in self._batchCalls.values():
            if batched_call.active():
                batched_call.cancel()

        if self._batchPaths:
            self.emitEvent('60034', data={
                'name': self.name,
                'count': len(self._batchPaths),
                })

        self._parent_removal.reset(source=None)
        yield self._jobs.stop(avatar=avatar)
        self._resetInternalState()

    def _onResume(self):
        """
        Called when transfer needs to be resumed.

        It can be called when transfer is already active.
        """
        if self._monitor_handler:
            # Already running.
            return defer.succeed(None)

        self._source_path = self.configuration.source_path
        # It is created but only started after the initial walk.
        self._monitor_handler = FilteredWatchAdapter(
            source=self._source,
            configuration=self._configuration,
            on_change=self._onSourceChange,
            on_error=self._onSourceError,
            on_snapshot=self._onSnapshot,
            scheduler=self._scheduler,
            event_emitter=self.emitEvent,
            )
        return self._startWatching()

    def onFileSuccess(self, attributes):
        """
        Called when file with `attributes` was successfully transferred.
        """
        self._parent_removal.add(attributes.path)
        self._metrics.updateSuccess(attributes)

    def _onSourceConnect(self, signal):
        """
        Called when source is connected (most common after a failure).
        """
        # There are 3 states from which the source can connect:
        # * after a failure - the path is unwatched we need to re-watch.
        # * after a manual stop - the path is unwatched and we need
        #   to re-watch it.
        # * after an idle disconnection - the path is still watched and we
        #   don't need to re-watch it.

        if self._monitor_handler.running:
            # We are still monitoring the source, so we need to setup the
            # monitor again.
            # This is for the case when the location was disconnected due to
            # idle/timeout.
            return

        deferred = self._startWatching()
        deferred.addErrback(lambda failure: self.fail(failure.value))

    @defer.inlineCallbacks
    def _startWatching(self):
        """
        Called when we should start watching the source location.
        """
        # Make sure we don't recurse here due to signals.
        self._source.unsubscribe('connect', self._onSourceConnect)
        # We might re-start watching after a source failure, so make sure
        # this is stopped before starting again.
        self._monitor_handler.stop()

        try:
            # Trigger a connection to the source location and
            # check that source path exists as the watcher will ignore
            # missing paths and we want to fail on missing source path
            # to help detect invalid configuration.
            # The operation on the source will automatically trigger the
            # connection.
            yield self._source.callWhenConnected(
                self._source.getFolderMemberAttributes,
                self._configuration.source_path,
                )
        except Exception as error:
            if self._source.operating:
                # The location was already started and working,
                # most probably source path doesn't exist.
                # We raise an error to stop the transfer.
                if isinstance(error, ServerError):
                    # We emit the error here, as emitting it later on failure
                    # it will have a different avatar.
                    error.avatar = self._source.avatar
                    emit_from_error(error, emitter=self.emitEvent)
                    # Source is operational, yet it has failed to perform the
                    # requested operation or it has a non ServerError.
                    raise EmittedError(error)

                # If a non ServerError is raised,
                # it means that this is a critical (unexpected) error.
                raise error

            # We have a generic error, like "location not yet connected".
            # just emit the error and let the source know that we are
            # interested when it is ready again.
            self._source.subscribe('connect', self._onSourceConnect)

            data = {
                'path': self._configuration.source_path,
                'location': self._source.name,
                }
            emit_error_details(
                error, event_id='60085', data=data, emitter=self.emitEvent)
            # We "consume" the error here, not to trigger a transfer failure,
            # so that the transfer is autoamtically resumed when location is
            # ready
            return

        # If the path exists, we can watch it.
        try:
            self._monitor_handler.start()
            # When a source stops, it will stop all monitors.
            # Here we hook into any future source starts,
            # to re-enable our source monitor.
            self._source.subscribe('connect', self._onSourceConnect)
        except Exception as error:
            # If somehow we can't watch the path,
            # just fail the transfer without any retry.
            self.fail(error)

    def _onSuspend(self):
        """
        Called when transfer needs to be suspended or stopped.

        Can be called when transfer is already suspended/stopped.

        If the transfer is suspended, the files which are currently batched
        are left to be transferred.
        """
        self._source_path = None
        if self._source:
            # Suspended before getting to have a source. Maybe errors during
            # source initialization.
            self._source.unsubscribe('connect', self._onSourceConnect)

        if not self._monitor_handler:
            # Nothing to stop.
            return defer.succeed(None)

        if (
            not self._batchCalls and
            self._batchPaths and
            self.state == STATE.STARTED
                ):
            # We don't have any pending files.
            self._batchPaths = []

        self._monitor_handler.stop()
        self._monitor_handler = None
        return defer.succeed(None)

    def _initiateJobTransfer(self, paths, kind):
        """
        Called when files should be transferred.

        Each file from `paths` is transferred in order.

        It can be called as a delayed call all exceptions should be handled
        here.
        """
        # Inform that transfer is queued.
        self.emitEvent('60033', data={
            'kind': kind,
            'name': self.name,
            'count': len(paths),
            'queue': paths,
            'files': '"' + '", "'.join(paths) + '"',
            })

        self._jobs.queue(
            paths=paths,
            source=self._source,
            destination=self._destination,
            configuration=self._configuration,
            )

    def _initiateMarkerBatchJobTransfer(self, base_path):
        """
        Called when a marker based batch should be transferred.

        `base_path` is the path where the marker file
        which has triggered this file is located.

        This will trigger the transfer for any files in the current folder or
        a subfolder of the file marker.
        """
        del self._batchCalls[base_path]

        paths = []
        for path in self._batchPaths[:]:
            if not path.startswith(base_path):
                # Path is for another marker, so we don't transfer in this
                # batch.
                continue

            # Move from the queue to the list of files to be transferred.
            self._batchPaths.remove(path)
            paths.append(path)

        return self._initiateJobTransfer(paths=paths, kind='batch marker')

    def _initiateTimeBatchJobTransfer(self):
        """
        Called when a batch should be transferred.

        It transfer all the batched files.
        """
        paths = sorted(self._batchPaths)
        self._batchPaths = []
        self._batchCalls = {}
        return self._initiateJobTransfer(paths=paths, kind='time batch')

    def _onSourceError(self, error):
        """
        Called when source watcher encounters an error.
        """
        if isinstance(error, (
            twisted_error.ConnectionClosed,
            twisted_error.ConnectError,
            ConnectException,
                )):
            # Connection errors are ignored as they will automatically
            # stop the watcher.
            # The state of the source should be updated and this should be
            # propagated to the state of the transfer.
            # Transfer is not stopped, so that when the source location is
            # reconnected, the transfer will be automatically resumed.
            return

        # All other errors on source are critical and will fail the transfer.
        # Most of the errors should be handled by the watch itself so if we
        # are here, the we have an unexpected error which shouldn't be
        # ignored... and we should investigate if this is a legit error
        # or the error reached this point due to a bug.
        self.fail(error)

    def _onSnapshot(self, result):
        """
        Called when the watch got a full snapshot for source.
        """
        total_directories = 0
        total_files = 0

        for _, members in result.items():
            for _, attributes in members.items():
                if attributes.is_folder:
                    # Don't count folders.
                    total_directories += 1
                    continue
                total_files += 1

        self.emitEvent(
            '60074',
            data={
                'path': self._source_path,
                'total_directories': total_directories,
                'total_files': total_files,
                },
            )

    def _onSourceChange(self, event):
        """
        Called when source folder contains changes.

        This should never raise an exception.
        """
        try:
            return self._handleSourceChange(event)
        except ServerException as error:
            self.fail(error)
        except Exception as error:
            message = 'Fail to handle source changes. %s' % (
                force_unicode(error),)
            self.fail(ServerException(message))

    def _handleSourceChange(self, event):
        """
        Low level code to handle source change events.
        """
        # Ignore events for folder as we don't transfer folders.
        if event.is_folder:
            return

        # We ignore the delete event as we can not transfer deleted files
        # but check to see if we should not cancel a queued job.
        if event.type == EVENT_TYPE.DELETED:
            self._onDelete(event)
            return

        try:
            # Emit informational event for the new source change.
            self._onChangeMap[event.type](event)
        except _OverlapEvent:
            self.emitEvent(
                '60079',
                data={
                    'path': event.source_path,
                    'details': (
                        'File might not be transferred with '
                        'the latest content.'
                        ),
                    },
                )
            return
        except _SkipEvent as error:
            self.emitEvent(
                '60075',
                data={
                    'path': event.source_path,
                    'reason': error.reason,
                    },
                )
            return

        file_markers = self.configuration.batch_marker_path
        if file_markers:
            return self._handleMarkerBatch(event, file_markers)

        timeout = self.configuration.batch_interval
        if timeout > 0:
            return self._handleTimeBatch(event, timeout)

        # Execute the job right away.
        # Newly created job is returned to help with testing.
        return self._initiateJobTransfer(
            paths=[event.current_path], kind='single file')

    def _get_marker_path(self, marker, base_path, paths):
        """
        Return the path matching `marker` from `paths`.

        None is returned if no match is found.
        """
        matches = []
        for target_path in paths:
            target_base = self._source.splitPath(target_path)[0]
            if (
                not target_base.startswith(base_path)
                and not base_path.startswith(target_base)
                    ):
                # Files from different subdirectories.
                # We ignore them as it will be done in another batch.
                continue

            if get_matched_groups(
                value=target_path,
                rule=marker,
                usage='batch_marker_path',
                    ):
                matches.append(target_path)

        if not matches:
            # No match
            return None

        if len(matches) != 1:
            raise ServerException(
                'Multiple markers detected for the same batch.')

        # Have a single match
        return matches[0]

    def _handleMarkerBatch(self, event, file_markers):
        """
        Called when we have a file name based batching.
        """
        current_path = event.current_path
        if not self._batchPaths:
            self._batchPaths = [current_path]
        else:
            self._batchPaths.append(current_path)

        # We are only looking for marker path in subdir of current file.
        base_path = self._source.splitPath(current_path)[0]

        markers = ', '.join(file_markers)
        for marker in file_markers:
            target_path = self._get_marker_path(
                marker, base_path, self._batchPaths)
            if not target_path:
                # File not yet found for this marker.
                self._informTransferDelay(
                    current_path, 'Waiting for "{}" markers.'.format(markers))
                return

            # We have a marker for the current file.
            # Make sure marker is last in the list.
            self._batchPaths.remove(target_path)
            self._batchPaths.append(target_path)

        delayed_call = self._batchCalls.get(base_path, None)
        if delayed_call:
            # We already have the marker for this path scheduled.
            # Reschedule the job.
            #
            # This is here mostly to handle the case in which when the transfer
            # is started, all the files are already in the source.
            #
            # The +3 is here to help with testing, and it should be fine
            # for our customer if they get a bit of extra delay.
            delayed_call.reset(self._configuration.stable_interval + 3)
            self._informTransferDelay(
                current_path,
                'Waiting for possible extra "{}" markers.'.format(markers))
            return

        self._batchCalls[base_path] = self._scheduler.callLater(
            self._configuration.stable_interval + 3,
            self._initiateMarkerBatchJobTransfer, base_path)
        return self._batchCalls[base_path]

    def _informTransferDelay(self, path, details):
        self.emitEvent('60006', data={
            'path': path,
            'details': details,
            })

    def _handleTimeBatch(self, event, timeout):
        """
        Called when we have a time based batching.
        """
        if self._batchCalls:
            # We already have a batch in progress.
            # For time based, we have a single batch.
            self._batchPaths.append(event.current_path)
            self._batchCalls['time-based'].reset(timeout)
            self._informTransferDelay(
                event.current_path,
                'Waiting {} seconds for possible other files.'.format(timeout))
            return

        # If we are here then we have the start of a batch process.
        self._batchPaths = [event.current_path]
        self._batchCalls['time-based'] = self._scheduler.callLater(
            timeout, self._initiateTimeBatchJobTransfer)

        # Inform that a batch was started.
        self.emitEvent('60032', data={
            'name': self.name,
            'seconds': timeout,
            'path': event.current_path,
            })
        return self._batchCalls['time-based']

    def _onDelete(self, event):
        """
        Called when a file is removed.
        """
        target = event.source_path
        if not self._jobs.dequeue(target):
            # Do nothing. Path was not queued.
            return

        self.emitEvent(
            '60080',
            data={'path': event.source_path},
            )

    def _onModified(self, event):
        """
        Called when a file was modified.

        For now, it ignores modified events for folders as they are somehow
        a duplicate of create / delete events.
        """
        self._metrics.checkSkipEvent(event, 'Modified file')
        self.emitEvent('60000', data={'path': event.source_path})

    def _onMoved(self, event):
        """
        Called when a file was moved.
        """
        self._metrics.checkSkipEvent(event, 'Moved file')

        data = {
            'path': event.source_path,
            'from_path': event.source_path,
            'to_path': event.destination_path,
            }
        self.emitEvent('60001', data=data)

    def _onCreated(self, event):
        """
        Called when a file is created.
        """
        self._metrics.checkSkipEvent(event, 'Created file')
        self.emitEvent('60003', data={'path': event.source_path})

    def _onExisting(self, event):
        """
        Called when a file is transferred as it was present when transfer
        was started.
        """
        already_transferred = self._metrics.checkSkipEvent(
            event, 'Existing file')

        if (
            already_transferred
            and already_transferred['size'] == event.size
            and already_transferred['modified'] == int(event.modified)
                ):
            # The message is similar to the one from metrics.checkSkipEvent but
            # without the `path` info.
            raise _SkipEvent('Existing file was already transferred.')

        self.emitEvent('60004', data={'path': event.source_path})
