# Copyright (c) 2023 Adi Roiban.
# See LICENSE for details.
"""
Various data transformation for transfers.
"""
from __future__ import absolute_import, unicode_literals
import codecs

from zope.interface import implementer

from chevah.server.commons.constant import TYPE_NAME
from chevah.server.commons.exception import (
    ServerException,
    )
from chevah.server.commons.interface import ITransferTransformer
from chevah.server.commons.match import get_matched_groups


_SUPPORTED_ACTIONS = [
    TYPE_NAME.NO_ACTION,
    TYPE_NAME.ENCODING,
    ]


def validate_transformers(rules):
    """
    Raise an error if the configured transformers are not valid.
    """
    for rule in rules:
        get_matched_groups(
            value='',
            rule=rule[0],
            usage='destination_content_actions',
            )
        if len(rule) < 2:
            raise ServerException(
                'Missing path and action definition for '
                'destination_content_actions: {}'.format(', '.join(rule)))

        action = rule[1].lower()
        if action not in _SUPPORTED_ACTIONS:
            raise ServerException(
                'Unknown action "{}" for '
                'destination_content_actions.'.format(action))

        if action == TYPE_NAME.NO_ACTION:
            if len(rule) != 2:
                raise ServerException(
                    'The "no-action" transformation for '
                    '"destination_content_actions" needs to be configured '
                    'without any parameter.'
                    )
            # PY3 will optimize this line.
            continue  # noqa:cover

        if action == TYPE_NAME.ENCODING:
            if len(rule) != 3:
                raise ServerException(
                    'The "encoding" transformation for '
                    '"destination_content_actions" needs to be configured '
                    'with one parameter.'
                    )

            # We try to parse, to validate the configuration.
            ConvertEncoding.parseParameter(rule[2])

            # PY3 will optimize this line.
            continue  # noqa:cover


def get_transformers(path, rules):
    """
    Return an ITransferTransformer instance that will handle the content
    transformation for `path` based on the list of `rules`.
    """
    validate_transformers(rules)

    if not rules:
        return NoopTransformer()

    for rule in rules:
        groups = get_matched_groups(
            value=path,
            rule=rule[0],
            usage='destination_content_actions',
            )

        if not groups:
            # Not a rule for this path.
            continue

        action = rule[1].lower()

        if action == TYPE_NAME.NO_ACTION:
            # Do nothing.
            return NoopTransformer()

        if action == TYPE_NAME.ENCODING:
            return ConvertEncoding(rule[2])

    raise ServerException(
        'No destination_content_actions defined for "%s".' % (
            path,))


@implementer(ITransferTransformer)
class TransformMixin(object):
    """
    Shared code for transformers.
    """

    def _postFinalize(self):
        """
        Called after finalized.
        """
        self.consume = self._postFinalizeConsume
        self.finalize = self._postFinalizeFinalize

    def _postFinalizeConsume(self, data):
        """
        Helper to raise and error if operations are done after finalize.
        """
        raise RuntimeError('Transform consume operation called after finalize.')

    def _postFinalizeFinalize(self):
        """
        Helper to raise and error if operations are done after finalize.
        """
        raise RuntimeError(
            'Transform finalize operation called after finalize.')


class NoopTransformer(TransformMixin):
    """
    Does no transformation.
    """

    name = 'no action'

    def __init__(self):
        self._preFinalizeConsume = self.consume
        self._preFinalizeFinalize = self.finalize

    def reset(self):
        """
        It has no internal state,
        so we just update the handlers.
        """
        self.consume = self._preFinalizeConsume
        self.finalize = self._preFinalizeFinalize

    def consume(self, data):
        return data

    def finalize(self):
        self._postFinalize()
        return b''


class ConvertEncoding(TransformMixin):
    """
    Converts from UTF-16 to ASCII.

    If the BOM is missing, RFC 2781 recommends that big-endian (BE)
    encoding be assumed.
    In practice, due to Windows using little-endian (LE) order by default,
    many apps just write the content without the BOM and the result is LE.
    """
    def __init__(self, parameter, default_bom=codecs.BOM_UTF16_LE):
        self._default_bom = default_bom
        self._preFinalizeFinalize = self.finalize
        self._source_encoding, self._destination_encoding = self.parseParameter(
            parameter)

        self._name = "{}-to-{}".format(
            self._source_encoding, self._destination_encoding)
        self.reset()

    @property
    def name(self):
        """
        See: ITransferTransformer
        """
        return self._name

    _SUPPORTED_CONVERSIONS = [
        'utf-16-to-ascii',
        'utf-16-to-utf-8',
        ]

    @classmethod
    def parseParameter(cls, parameter):
        """
        Return the source and destination encoding.
        """
        parameter = parameter.lower()
        if parameter not in cls._SUPPORTED_CONVERSIONS:
            # This should not be called in production,
            # as the transfer will validate the input before calling this.
            raise ServerException('Unknown encoding conversion: "{}".'.format(
                parameter))

        parts = parameter.split('-to-')
        return parts

    def reset(self):
        """
        Get ready for a new run.
        """
        self._decoder = codecs.getincrementaldecoder(
            self._source_encoding)(errors='strict')
        self._buffer = b''
        # To optimize the code, we have a function that
        # handles BOM detection and after that it switches to
        # the main consumeContent function.
        self.consume = self.consumeBOM
        self.finalize = self._preFinalizeFinalize

    def consumeBOM(self, data):
        """
        Handle the initial data and BOM detection.
        """
        self._buffer = self._buffer + data
        if len(self._buffer) < 2:
            return b''

        current_start = self._buffer[0:2]
        if current_start not in [
                codecs.BOM_UTF16_BE, codecs.BOM_UTF16_LE]:
            self._buffer = self._default_bom + self._buffer
        data = self._buffer
        self._buffer = b''
        self.consume = self.consumeContent
        return self.consume(data)

    def consumeContent(self, data):
        """
        This is the main consume call.

        See ITransferTransformer.
        """
        try:
            result = self._decoder.decode(data, final=False)
        except UnicodeDecodeError:
            raise ServerException(
                'The source contains invalid {} characters. '
                'Maybe this is a file with a different encoding '
                'or it has a different byte order.'.format(
                    self._source_encoding)
                )
        return self._toDestination(result)

    def finalize(self):
        """
        See ITransferTransformer.
        """
        self._postFinalize()
        try:
            result = self._decoder.decode(self._buffer, final=True)
        except UnicodeDecodeError:
            raise ServerException(
                'The source contains partial {} characters. '
                'Maybe it is truncated.'.format(self._source_encoding)
                )
        return self._toDestination(result)

    def _toDestination(self, text):
        """
        Convert Unicode text to destination encoding.
        """
        try:
            return text.encode(self._destination_encoding)
        except UnicodeEncodeError:
            raise ServerException(
                'The source contains non {} characters.'.format(
                    self._destination_encoding))
