"""
Fast launcher for commands not needing to import the rest of SFTPPlus.
"""
from __future__ import unicode_literals

from __future__ import absolute_import
import sys
import webbrowser
import six

from chevah_compat import local_filesystem
from chevah.slim_server.raw_config_parser import (
    ChevahRawConfigParser,
    configparser,
    )


def _launch_browser(url):
    """
    Open the URL in a new browser tab.
    """
    webbrowser.open_new_tab(url)


def get_option(config, section, option, fallback):
    """
    Get an option from the config, with a given fallback.
    """
    try:
        return config.get(section, option)
    except configparser.NoSectionError:
        return fallback
    except configparser.NoOptionError:
        return fallback


def get_manager_url(args):
    """
    Read the configuration file, and return the URL of the Web Manager.
    """
    config_path = _get_config_path(args)
    config = ChevahRawConfigParser()
    with open(config_path) as config_stream:
        config.read_file(config_stream)

        address = get_option(
            config,
            section='services/DEFAULT-MANAGER',
            option='address',
            fallback='localhost',
            ).strip()
        if address in ['0.0.0.0', '::']:
            address = 'localhost'

        port = get_option(
            config, 'services/DEFAULT-MANAGER', 'port', '10020').strip()

        return 'https://%s:%s/' % (address, port)


def _get_config_path(args):
    """
    Return the path of the configuration file.

    If an argument is passed with `-c` or `--config`, the argument is the path,
    otherwise a default path is used.
    """
    result = None

    try:
        result = args[args.index('-c') + 1]
    except ValueError:
        try:
            result = args[args.index('--config') + 1]
        except Exception:
            result = 'configuration/server.ini'

    if isinstance(result, six.text_type):
        result = local_filesystem.getEncodedPath(result)

    return result


def subcommand_manager(args):
    """
    The `manager` subcommand.
    Launch the local web browser with the Web Manager page started.
    """
    url = get_manager_url(args)
    _launch_browser(url)


def subcommand_documentation(args):
    """
    The `documentation` subcommand.
    Launch the local web browser with the documentation page started.
    """
    url = get_manager_url(args)
    _launch_browser(url + 'documentation/')


def commands(args):
    """
    Dispatch any command line operation.
    If a fast handler does not exist, fall back to the default handler.
    """
    args = [_force_unicode(a) for a in args]
    try:
        action, rest = args[0], args[1:]

        handler_name = 'subcommand_%s' % (action,)
        handler = globals().get(handler_name)

        if handler:
            handler(rest)
            return
    except IndexError:
        """
        There have been no arguments passed.
        Keep going with `admin_commands`.
        """

    # No fast command available.
    # Forward the command line options to the standard admin_commands.

    from chevah.server.commons import component
    from chevah.server.commons.command import admin_commands
    component.register()
    admin_commands()


def _force_unicode(value):
    """
    Always return Unicode for `value`.
    """
    if isinstance(value, six.text_type):
        return value
    return value.decode('utf-8')


if __name__ == '__main__':
    commands(sys.argv[1:])
