# Copyright (c) 2020 Adi Roiban.
# See LICENSE for details.
"""
Execute external code on processes from a pool.

This is outside of the chevah.server package to reduce memory usage.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import os
import signal
import sys
import time
from multiprocessing.pool import ApplyResult, MaybeEncodingError, Pool
from multiprocessing.queues import SimpleQueue
from six.moves import range
import six

try:
    from multiprocessing import get_context
except ImportError:
    get_context = None


def worker(
    own_tasks, inqueue, outqueue, initializer=None, initargs=(), maxtasks=None,
        ):
    """
    Code executed in each sub-process.

    It will try to execute with priority its own tasks, before executing the
    tasks shared with the other workers from the pool.
    """
    put = outqueue.put
    get = inqueue.get
    own = own_tasks.get

    if hasattr(inqueue, '_writer'):
        inqueue._writer.close()
        outqueue._reader.close()
        own_tasks._writer.close()

    if initializer is not None:
        initializer(*initargs)

    signal.signal(signal.SIGINT, signal.SIG_DFL)
    signal.signal(signal.SIGTERM, signal.SIG_DFL)

    def execute(task):
        """
        Execute `tasks` and send the result back to main process.
        """
        if task == 'skip':
            # Let other processes to execute, so that we are not the only
            # processing consuming all the skip commands.
            time.sleep(0.001)
            return

        # FIXME:5235:
        # Before starting to execute the task, announce which worker is
        # handling the task.
        # In this way the pool can kill the worker on timeout, or notify
        # if the worker dies while executing the task.
        job, i, func, args, kwds = task
        try:
            result = (True, func(*args, **kwds))
        except Exception as e:
            result = (False, e)

        try:
            put((job, i, result))
        except Exception as e:
            wrapped = MaybeEncodingError(e, result[1])
            put((job, i, (False, wrapped)))

    def do_own_tasks():
        """
        Execute the tasks assigned to this specific worker.

        Return False, when the worker should no longer process any tasks.
        """
        while not own_tasks.empty():
            try:
                task = own()
            except (EOFError, IOError):
                # Communication to master was closed.
                return False

            if task is None:
                return False

            execute(task)

        return True

    completed = 0
    while maxtasks is None or (maxtasks and completed < maxtasks):

        if not do_own_tasks():
            break

        try:
            task = get()
        except (EOFError, IOError):
            break

        if task is None:
            break

        execute(task)

        if not do_own_tasks():
            break

        task = None
        completed += 1


class WorkerResult(ApplyResult):
    """
    The result returned by executing a function on a worker.

    This is modified version to play nice with Twisted deferred and Py2.
    """

    def __init__(self, cache, callback=None, errback=None):
        super(WorkerResult, self).__init__(cache, callback)
        self._errback = errback

    def _set(self, i, obj):
        """
        Called when we got a result from the worker.
        """
        self._success, self._value = obj

        if self._callback and self._success:
            self._callback(self._value)

        if self._errback and not self._success:
            self._errback(self._value)

        self._cond.acquire()
        try:
            self._ready = True
            self._cond.notify()
        finally:
            self._cond.release()
        del self._cache[self._job]


class _DefaultInitializer():
    """
    Generator for sub-process initialization on Unix.
    """
    def __init__(self, name):
        self.name = name

    def initializer(self):  # noqa:cover
        """
        Called when each worker starts.

        This is called in the sub-process and we don't have code coverage.
        """
        try:
            from setproctitle import setproctitle
            setproctitle
        except ImportError:
            def setproctitle(name):
                """
                Does nothing.
                """

        # Reset any special signal handler that might be set by the
        # main process.
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        signal.signal(signal.SIGTERM, signal.SIG_DFL)

        setproctitle(self.name)


class ChevahPool(Pool):
    """
    A worker pool which can execute tasks in a process pool.

    On top of the standard Pool is does:
    * methods which return a deferred, and have a timeout by default.
    * method to execute the same function on each worker.

    The timeout will only fail the deferred, but it can't cancel the current
    work from the workers.

    FIXME:5235:
    Maybe kill the worker on timeout, so that the pool will not end up with
    workers which are in an infinite wait/loop.
    """

    def __init__(self, name='Pool', *args, **kwargs):
        self._name = name
        initializer = kwargs.get('initializer', None)

        if os.name != 'nt' and not initializer:
            kwargs['initializer'] = _DefaultInitializer(name).initializer

        try:
            # On Windows, if the arguments are long the pool is initialized
            # and jobs are dispatched, but no result is returned.
            sys_argv = sys.argv
            sys.argv = ['']
            super(ChevahPool, self).__init__(*args, **kwargs)
        finally:
            sys.argv = sys_argv
        self._worker_handler.name = self._name + '-WorkerHandler'
        self._task_handler.name = self._name + '-TaskHandler'
        self._result_handler.name = self._name + '-ResultHandler'

    def get_worker_pids(self):
        """
        Return a list with PIDs for all the current workers.
        """
        result = []
        for p in self._pool:
            result.append(p.pid)
        return result

    def execute(self, func, args=(), kwds={}, timeout=30):
        """
        Call `func` and a random worker.

        Return deferred which fires when we got the result from the worker.

        FIXME:5235:
        `timeout` is the total time in which the job should be executed.
        Counting from the moment it is placed in the queue.
        We might want to have a separate execution timeout not counting
        the time it waits in the queue.
        """
        # Imported to reduce the size of the process pool workers.
        from twisted.internet import defer, reactor
        from chevah.server import add_timeout

        deferred = defer.Deferred()
        if timeout:
            add_timeout(
                reactor, deferred,
                delay=timeout,
                message='Timeout while waiting in the queue.',
                )

        def callback(result):
            reactor.callFromThread(deferred.callback, result)

        def errback(error):
            reactor.callFromThread(deferred.errback, error)

        if six.PY2:
            result = WorkerResult(self._cache, callback, errback)
            self._taskqueue.put(([(result._job, None, func, args, kwds)], None))
        else:
            self.apply_async(
                func, args=args, kwds=kwds,
                callback=callback,
                error_callback=errback,
                )
        return deferred

    def _repopulate_pool(self):
        """
        Bring the number of pool processes up to the specified number,
        for use after reaping workers which have exited.

        This is also used to start the initial pool.
        """
        for i in range(self._processes - len(self._pool)):
            ctx = None
            kwargs = {}
            if get_context:
                # Only Python3 has the context.
                ctx = get_context()
                own_tasks = SimpleQueue(ctx=ctx)
                kwargs = {'ctx': ctx}
            else:
                # In Py2 don't usue the context.
                own_tasks = SimpleQueue()

            w = self.Process(
                target=worker,
                args=(
                    own_tasks,
                    self._inqueue,
                    self._outqueue,
                    self._initializer,
                    self._initargs,
                    self._maxtasksperchild,
                    ),
                **kwargs
                )
            self._pool.append(w)
            # Queue with task to be executed on this worker.
            w.own_tasks = own_tasks
            w.name = w.name.replace('Process', self._name + '-Worker')
            w.daemon = True
            w.start()

    @staticmethod
    def _help_stuff_finish(inqueue, task_handler, size):
        # task_handler may be blocked trying to put items on inqueue
        while task_handler.is_alive() and inqueue._reader.poll():
            try:
                inqueue._reader.recv()
                time.sleep(0.0001)
            except EOFError:
                """
                We ignore any error. Most probably the worker is closed.
                """


class PoolWrapper(object):
    """
    A wrapper for the pool used to spawn sub-process.

    It has a separate start method.
    It is needed to delay initialization on Windows and reduce memory usage.
    We also delay the initialization to allow setting up the unpriviledged
    user context under which the process will execute.
    """

    def __init__(self, name):
        self._name = name
        self._context = None

    def start(self):
        """
        Start the child processes if required.
        """
        if self._context:
            return

        self._context = ChevahPool(
            processes=3,
            maxtasksperchild=None,
            name=self._name,
            )

    def stop(self):
        """
        Stop and close the processes from the pool.
        """
        if not self._context:
            return

        self._context.terminate()
        self._context.join()
        if six.PY3:
            # Release any extra resources.
            self._context.close()
        self._context = None

    def execute(self, *args, **kwargs):
        """
        Proxy to the pool.
        """
        return self._context.execute(*args, **kwargs)


spawn_pool = PoolWrapper('sftpplus-spawn-worker')
