"""
Customization for our .INI configuration file.

This is outside of the chevah.server package to reduce loading times
for launcher commands.
"""

from __future__ import unicode_literals
from __future__ import absolute_import

import six

if six.PY2:
    from backports import configparser
else:
    import configparser


class NewLineWrapper(object):
    """
    Patch the stdlib parser to accept other type of new-line characters are
    valid, but only consider `\n` as the .INI line separator.
    """
    def __init__(self, stream):
        self._stream = stream

    def _getNextLine(self):
        """
        Return the next line from a file descriptor,
        but only consider '\n' as a line terminator.
        """
        line = next(self._stream)

        if line and not line.endswith('\n'):
            try:
                return line + self._getNextLine()
            except StopIteration:
                # This can be raise for a file that has no new line at the
                # end.
                # Just return what we have so far.
                # The StopIterator will be re-raise later on the followin
                # invocation.
                return line

        return line

    def __iter__(self):
        return self

    def __next__(self):
        return self._getNextLine()

    next = __next__


class ChevahRawConfigParser(configparser.RawConfigParser):
    """
    Upstream fixes used for the Chevah project.

    * Allow empty lines in multi line configuration.
    """
    def __init__(self, *args, **kwargs):
        super(ChevahRawConfigParser, self).__init__(*args, **kwargs)
        # Later we might want to enforce string,
        # but for now we have this to move on with the Python3 migration.
        self._strict = False

    def _read(self, fp, fpname):
        """
        Just inject our stream wrapper.
        """
        wrapper = NewLineWrapper(fp)
        super(ChevahRawConfigParser, self)._read(fp=wrapper, fpname=fpname)

    def has_option(self, section, option):
        """
        See `IConfigurationProxy'.

        This overwrites the upstream not to consider a default value
        as always existing.
        """
        if not self.has_section(section):
            return False
        return option.lower() in self._sections[section]
